(**
*Certified Round Complexity of Self-Stabilizing Algorithms*
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Bool.
From Coq Require Import ZArith.
From Coq Require Import Lia.

(** **** Local Imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import OptionUtil.
From Padec Require Import NatUtils.
From Padec Require Import ListUtils.
From Padec Require Import Algorithm.

From Padec Require Import Graph_Diameter. (* min_list *)

(* end show *)
Open Scope nat_scope.
Open Scope signature_scope.
Set Implicit Arguments.

(** * BFS Spanning Tree: the Algoritm 

  nb: Two algorithms are given, one for the root, the other for
  other processes. We merge them, using conditional structure.

  Algorithm for a process p:
    Input: p.N contains channels 
       assumption: p.N is the list of channels of every neighbor towards 
                   p in the network
    Input: p.root Boolean
       assumption: True for a unique process, 
                   represents the root of the rooted network
    Variables: p.d natural number, p.par channel
       nb: p.par has no meaning if root
 *)

Section BFS_algo.

  Context {Chans: Channels}.

  (** State of node for the algorithm *)
  Record BFSROState :=
    mkROState {
        (** is the process considered as the root? *)
        root: bool;
        (** neighbors of the process, implemented as a list. 
         The order in the list represents the local on those
         channels (required in the algorithm) *)
        neighbors: list Channel
      }.

  Notation eqROS :=
    (fun (x y: BFSROState) => root x == root y /\
                              neighbors x == neighbors y).

  Global Instance eqROS_equiv: Equivalence eqROS.
  Proof.
    split; hnf; intuition;etransitivity;eauto.
  Qed.

  Global Instance BFSROState_setoid: Setoid BFSROState :=
    {| equiv := eqROS |}.

  Record BFSState :=
    mkState {
        (** distance to the root *)
        d: nat;
        (** parent pointer; meaningless if root *)
        par: Channel;
        (** read-only part of the state *)
        BFSROpart:> BFSROState
      }.

  Notation eqS :=
    (fun (x y: BFSState) =>
       d x == d y /\ par x == par y /\ BFSROpart x == BFSROpart y).

  Global Instance eqS_equiv: Equivalence eqS.
  Proof.
    split; hnf; intuition;
      etransitivity; eauto.
  Qed.

  Global Instance BFSState_setoid: Setoid BFSState :=
    {| equiv := eqS |}.

  Lemma BFSROState_dec: Decider (equiv (A := BFSROState)).
  Proof.
    intros [r1 p1] [r2 p2]; simpl equiv.
    destruct (bool_dec r1 r2);[subst|right;tauto].
    destruct (eqlistA_dec Channel_dec p1 p2);
      [left; split; auto|right; tauto].
  Qed.

  Lemma BFSState_dec: Decider (equiv (A := BFSState)).
  Proof.
    intros [d1 p1 ro1] [d2 p2 ro2]; simpl.
    destruct (Nat.eq_dec d1 d2); [|right; tauto].
    destruct (Channel_dec p1 p2); [|right; tauto].
    destruct (BFSROState_dec ro1 ro2);
      [left;auto|right;simpl in n; intuition].
  Qed.

  (** **** Compatibility on states *)
  Global Instance d_compat: Proper fequiv d.
  Proof. intros x y [h _];auto. Qed.

  Global Instance par_compat: Proper fequiv par.
  Proof. intros x y [_ [p pp]];auto. Qed.

  Global Instance root_compat: Proper fequiv root.
  Proof. intros x y [h _];auto. Qed.

  Global Instance neighbors_compat: Proper fequiv neighbors.
  Proof. intros x y [_ h];auto. Qed.

  Global Instance BFSROpart_compat: Proper fequiv BFSROpart.
  Proof. intros x y exy; apply exy. Qed.

  (** ** Function run *)
  Definition Dist_p (dists: list nat): nat :=
    match min_list dists with None => 0 | Some m => m end.

  (** Par_dist = q s.t. q in p.Neigh and q.d + 1 = p.d *)
  Definition Par_dist (pd: nat) (dum: Channel)
             (neigh_dist: list (Channel * nat)): Channel :=
    fst (hd (dum, 0)
            (filter (fun cn => if Nat.eq_dec (S (snd cn)) pd
                               then true else false) neigh_dist)).

  Global Instance Par_dist_compat: Proper pequiv Par_dist.
  Proof.
    intros pd1 pd2 epd d1 d2 ed n1 n2 en; unfold Par_dist.
    simpl in ed, epd; rewrite ed, epd; clear epd pd1 ed d1.
    eapply (fst_compat (eqA:=equiv) (eqB:=equiv)).
    assert ((d2, 0) == (d2, 0)) by auto.
    apply (hd_compat _ _ H).
    apply filter_compat; auto.
    intros [a x] [b y] ab.
    assert (snd (a, x) = snd (b, y)).
    change (snd (a, x) == snd (b, y)).
    eapply snd_compat, ab.
    rewrite H0; auto.
  Qed.
  
  (** Root Algorithm *)
  Definition simple_run_Root (s: BFSState): BFSState :=
    {| d := 0; par := par s; BFSROpart := BFSROpart s |}.

  (** Non-Root Algorithm:
      guard for CD: p.d <> Distp
      guard for CP: p.d = Distp /\ p.par.d + 1 <> p.d  
   *)
  Definition simple_run (d: nat) (par: Channel) (par_d: nat)
             (ro: BFSROState)
             (neigh: list (Channel * nat)): BFSState :=
    if Nat.eq_dec d (Dist_p (map S (map snd neigh))) then
      (* p.d = Distp *)
      if Nat.eq_dec d (S par_d) then
        (* p.par.d + 1 = p.d *)
        (** Do Nothing *) {| d := d; par := par; BFSROpart := ro |}
      else
        (* p.par.d + 1 <> p.d *)
        (** CP Action *)
        {| d := d; par := Par_dist d par neigh; BFSROpart := ro |}
    else
      (* p.d <> Distp *)
      {| (** CD Action *)
        d := Dist_p (map S (map snd neigh));
        par := par;
        BFSROpart := ro |}.
  
  Definition neighbor_list
             (state: BFSState)
             (peer_state: Channel -> option BFSState):
    list (Channel * nat) :=
    map_filter (fun c =>
                  match peer_state c with
                  | None => None (* should never occur *)
                  | Some s => Some (c, d s) end)
               (neighbors state).

  Definition run (state: BFSState)
             (peer_state: Channel -> option BFSState): BFSState :=
    if root state then simple_run_Root state
    else simple_run (d state)
                    (par state)
                    (match peer_state (par state) with
                     | None => d state (* should never occur *)
                     | Some s => d s end)
                    (BFSROpart state)
                    (neighbor_list state peer_state).

  (** ** Compatibility of definitions used in run_algorithm *)

  Global Instance Dist_p_compat: Proper pequiv Dist_p.
  Proof.
    intros l1 l2 el; unfold Dist_p.
    assert (min_list l1 == min_list l2) by (apply min_list_compat; auto).
    destruct (min_list l1); destruct (min_list l2); auto;
      inversion H.
  Qed.

  Global Instance simple_run_Root_compat: Proper pequiv simple_run_Root.
  Proof.
    intros s1 s2 es; unfold simple_run_Root.
    split; [|split].
    simpl; auto.
    simpl; apply par_compat; auto.
    simpl; split; [apply root_compat|apply neighbors_compat]; apply es.
  Qed.
  
  Global Instance simple_run_compat: Proper pequiv simple_run.
  Proof.
    intros d1 d2 ed c1 c2 ec dp1 dp2 edp ro1 ro2 ero l1 l2 el;
      unfold simple_run.
    rewrite ed; clear ed d1.
    assert (D2: Dist_p (map S (map snd l1)) =
                Dist_p (map S (map snd l2))).
    { apply Dist_p_compat; auto.
      eapply map_compat.
      { intros x y xy; rewrite xy; auto. }
      eapply map_compat.
      apply snd_compat.
      apply el. }
    rewrite D2. 
    destruct (Nat.eq_dec d2 (Dist_p (map S (map snd l2)))).
    assert (D3: S dp1 = S dp2) by (simpl in edp; lia).
    rewrite D3.
    destruct (Nat.eq_dec d2 (S dp2)).
    split; [|split; [|split]]; simpl; auto.
    apply root_compat; auto.
    apply neighbors_compat; auto.
    split; [|split; [|split]]; simpl; auto.
    apply Par_dist_compat; auto.
    apply root_compat; auto.
    apply neighbors_compat; auto.
    split; [|split; [|split]]; simpl; auto.
    apply root_compat; auto.
    apply neighbors_compat; auto.
  Qed.      
  
  Global Instance neighbor_list_compat: Proper pequiv neighbor_list.
  Proof.
    intros s1 s2 es ps1 ps2 ep; unfold neighbor_list.
    eapply map_filter_compat_eqlistA.
    2: { rewrite es; reflexivity. }
    intros c1 c2 ec.
    case_eq (ps1 c1); case_eq (ps2 c2).
    { intros a1 h1 a2 h2.
      specialize (ep _ _ ec).
      rewrite h1, h2 in ep; inversion ep.
      simpl; split; auto. }
    { intros h2 a1 h1;
        specialize (ep _ _ ec);
        rewrite h1, h2 in ep; inversion ep. }
    { intros a2 h2 h1;
        specialize (ep _ _ ec);
        rewrite h1, h2 in ep; inversion ep. }
    { intros _ _; reflexivity. }
  Qed.
  
  Global Instance run_compat: Proper pequiv run.
  Proof.
    intros s1 s2 es ps1 ps2 ep; unfold run.
    assert (root s1 == root s2).
    apply root_compat, es.
    rewrite H.
    destruct (root s2).
    apply simple_run_Root_compat; auto.
    apply simple_run_compat; auto.
    apply d_compat, es. 
    apply par_compat, es.
    assert (ps1 (par s1) == ps2 (par s2)).
    apply ep, par_compat, es.
    destruct (ps1 (par s1)); destruct (ps2 (par s2));
      simpl in H0; try tauto.
    apply d_compat, es.
    apply es.
    apply neighbor_list_compat; auto.
  Qed.
  
  Lemma ROpart_stable: forall s sn, BFSROpart (run s sn) =~= BFSROpart s.
  Proof.
    intros s sn; unfold run, simple_run.
    destruct (root s); [reflexivity|].
    destruct (Nat.eq_dec
                (d s) (Dist_p (map S (map snd (neighbor_list s sn))))); [|reflexivity].
    destruct (Nat.eq_dec (d s)
                         (S match sn (par s) with
                            | Some s0 => d s0 | None => d s
                            end)); reflexivity.
  Qed.

  (** ** The Algorihtm *)
  Global Instance BFS_algo: Algorithm.
  Proof.
    esplit.
    apply BFSState_dec.
    apply run_compat.
  Defined.

  Global Instance BFS_Stable:
    Stable_Variable BFS_algo BFSROpart.
  Proof.
    esplit.
    apply BFSROpart_compat.
    apply ROpart_stable.
  Defined.

End BFS_algo.

Close Scope signature_scope.
Close Scope nat_scope.
Unset Implicit Arguments.
