(**
*Certified Round Complexity of Self-Stabilizing Algorithms*
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Bool.
From Coq Require Import ZArith.
From Coq Require Import Lia.

(** **** Local Imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import NatUtils.
From Padec Require Import ListUtils.
From Padec Require Import OptionUtil.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Count.
From Padec Require Import Graph_Diameter.
From Padec Require Import Tree_topology_ALT.
From Padec Require Import BFS_algo_bis.

(* end show *)
Open Scope nat_scope.
Open Scope signature_scope.
Set Implicit Arguments.

(** * BFS Spanning Tree Under a Distributed Weakly Daemon: 
      Common definitions and results
 *)

(** ** Assumption about the topology:
  The network should be: bidirected, connected and rooted
 *)
Section BFS_Network.

  Context {Chans: Channels}
          {Net: Network}.
  
  Class BFS_Assume :=
    mkBFS {
        G: relation Node;
        G_compat: Proper pequiv G;
        G_dec: Decider G;
        
        Bidir: forall x y, G x y -> G y x;
        Connected: is_connected G;
        
        Root: Node;

        Diam := compute_diameter'
                  all_nodes_prop G_compat G_dec Root (DA := Node_dec);

      }.
  
  Context {Graph_assumption: BFS_Assume}.
  
  Definition Graph_Network_matching: Prop :=
    forall x y, G x y <-> exists c, is_channel x c y.

  Definition Input_matching
             (_root: Node -> bool)
             (_neighbors: Node -> list Channel): Prop :=
    Network_assumption Net _neighbors /\
    (forall x, _root x = true <-> x == Root).
  
  Global Instance Input_matching_compat: Proper pequiv Input_matching.
  Proof.
    intros r1 r2 er n1 n2 en.
    apply and_compat.
    { apply Network_assumption_compat. 
      intros x x' ex; apply eqlistA_equivlistA.
      apply setoid_equiv.
      apply en, ex. }
    { apply all_compat; intros x.
      apply iff_compat; auto.
      apply eq_compat; [apply er|]; auto. }
  Qed.

#[refine] Global Instance BFS_RO_assumption: Stable_Assumption
                                       (Algo := BFS_algo)
                                       (Net := Net)
    := {|  Assume_RO :=
            (fun ro: Node -> BFSROState => 
               Input_matching
                 (fun n: Node => root (ro n))
                 (fun n: Node => neighbors (ro n))
            )|}.
  Proof.
    apply BFSROpart.
    apply BFS_Stable.
    intros ro1 ro2 ero.
    apply Input_matching_compat.
    intros n1 n2 en; apply root_compat, ero; auto.
    intros n1 n2 en; apply neighbors_compat, ero; auto.
  Defined.

End BFS_Network.

(** ** Common results *)
Section BFS_common.
  
  Context {Chans: Channels}
          {Net: Network}
          {Graph_assumption: BFS_Assume}
          {Graph_network_assumption: Graph_Network_matching}.
  Existing Instance G_compat.
  
  Notation Algo := (BFS_algo).
  Notation Env := (Env (Net := Net) (Algo := Algo)).

  Definition Input_Assume (g: Env): Prop :=
    Input_matching (fun n => (root (g n))) (fun n => neighbors (g n)).

  Global Instance Input_Assume_compat: Proper pequiv Input_Assume.
  Proof.
    intros g g' eg; unfold Input_Assume;
      apply Input_matching_compat; auto.
    intros n n' en; apply root_compat, eg, en.
    intros n n' en; apply neighbors_compat, eg, en.
  Qed.

  Lemma Input_Assume_Stable:
    forall g' g, Step g' g -> Input_Assume g -> Input_Assume g'.
  Proof.
    intros g' g st [ass1 ass2].
    assert (XX := Stable_Variable_step (BFS_Stable) st).
    split; intros p;
      specialize (XX _ _ (Equivalence_Reflexive p));
      simpl in XX; destruct XX as [XX XXX].
    - specialize (ass1 p).
      rewrite <- ass1; apply eqlistA_equivlistA;
        [apply setoid_equiv|apply XXX].
    - rewrite <- ass2, XX; tauto.
  Qed.
  
  (** Access to parent node *)
  Fixpoint Par_ (c: Channel) (p: Node) (l: list Channel): Node :=
    match l with
    | nil => p
    | cc :: ll => if Channel_dec c cc
                  then match peer p c with None => p | Some q => q end
                  else Par_ c p ll
    end.
  
  Definition Par' (s: BFSState) (p: Node): Node :=
    Par_ (par s) p (neighbors s).

  Definition Par (g: Env) (p: Node): Node := Par' (g p) p.

  Instance Par__compat: Proper pequiv Par_.
  Proof.
    intros g g' eg p p' ep l.
    revert eg ep; revert g g' p p'; induction l as [| c l ih];
      intros g g' p p' eg ep [| c' l'] el;
      inversion_clear el; simpl.
    { auto. }
    { case (Channel_dec g c);
        case (Channel_dec g' c'); intros hc hc'.
      { assert (h: peer p g == peer p' g') by (apply peer_compat; auto).
        revert h; case_eq (peer p g); case_eq (peer p' g').
        intros q _ q' _ hq; simpl in hq; auto.
        intros _ q' _ hq; inversion hq. 
        intros q _ _ hq; inversion hq. 
        auto. }
      contradict hc.
      rewrite <- H, <- eg; auto.
      contradict hc'.
      rewrite H, eg; auto.
      apply ih; auto.
    }
  Qed.

  Global Instance Par'_compat: Proper pequiv Par'.
  Proof.
    intros g g' eg p p' ep; apply Par__compat; auto.
    apply par_compat; auto.
    apply neighbors_compat, eg.
  Qed.
  
  Global Instance Par_compat: Proper pequiv Par.
  Proof. intros g g' eg p p' ep; apply Par'_compat; auto. Qed.

  Lemma Par__spec_chan:
    forall c p l, InA equiv c l -> InA equiv c (peers p) ->
                  is_channel p c (Par_ c p l).
  Proof.
    intros c p l; revert c p; induction l as [|cc l ih]; simpl.
    intros c p h.
    rewrite InA_nil in h; tauto.
    intros c p h inpeers.
    dup inpeers hh.
    rewrite peers_spec_peer in hh.
    destruct hh as [q hq]; unfold is_channel in hq.
    unfold is_channel.
    case (Channel_dec c cc); intros hc.
    { revert hq; case_eq (peer p c); auto. }
    { rewrite InA_cons in h; destruct h as [h|h].
      contradict hc; auto.
      apply ih; auto. }
  Qed.

  Lemma Par__chan_conv:
    forall c l p q,
      InA equiv c l -> peer p c == Some q -> Par_ c p l == q.
  Proof.
    intros c l p q; induction l as [|a l ih].
    intros h; rewrite InA_nil in h; tauto.
    intros h; rewrite InA_cons in h; destruct h as [h|h];
      intros hp; simpl.
    case (Channel_dec c a); [intros _|intros; tauto].
    destruct (peer p c); [|simpl in hp; tauto].
    apply hp.
    case (Channel_dec c a); intros; auto.
    destruct (peer p c); [apply hp|simpl in hp; tauto].
  Qed.    

  Lemma Par__None:
    forall c p l, peer p c == None -> Par_ c p l == p.
  Proof.
    intros c p l; induction l as [|a l ih]; simpl; auto.
    destruct (peer p c); intros h.
    inversion h.
    destruct (Channel_dec c a); auto.
  Qed.

  Lemma Par__spec:
    forall c p l, InA equiv c l -> InA equiv c (peers p) ->
                  G p (Par_ c p l).
  Proof.
    intros c p l ina inp.
    apply Graph_network_assumption.
    exists c; apply Par__spec_chan; auto.
  Qed.
  
  Lemma Par'_spec_chan:
    forall (s: BFSState) p (ass: equivlistA equiv (neighbors s) (peers p)),
      InA equiv (par s) (peers p) ->
      is_channel p (par s) (Par' s p).
  Proof.
    intros; unfold Par'; apply Par__spec_chan; auto.
    rewrite ass; auto.
  Qed.
  
  Lemma Par'_spec:
    forall (s: BFSState) p (ass: equivlistA equiv (neighbors s) (peers p)),
      InA equiv (par s) (peers p) -> G p (Par' s p).
  Proof.
    intros s p h h'.
    apply Par__spec; auto.
    rewrite h; auto.
  Qed.
  
  Lemma Par_spec_chan:
    forall g (ass: Input_Assume g) p,
      InA equiv (par (g p)) (peers p) -> is_channel p (par (g p)) (Par g p).
  Proof.
    intros; unfold Par; apply Par'_spec_chan; auto.
    apply ass.
  Qed.
  
  Lemma Par_spec:
    forall g (ass: Input_Assume g) p,
      InA equiv (par (g p)) (peers p) -> G p (Par g p).
  Proof. intros g [ass _] p h; unfold Par; apply Par'_spec; auto. Qed.

  Lemma Dist_p_spec:
    forall l, l <> nil ->
              InA equiv (Dist_p l) l /\
              forall a, InA equiv a l -> Dist_p l <= a.
  Proof.
    intros [| a l] hn; unfold Dist_p; [congruence|].
    case_eq (min_list (a :: l)).
    2: intros h; rewrite min_list_None in h; congruence.
    intros m hm; rewrite min_list_Some in hm; auto.
  Qed.

  Lemma Dist_p_spec':
    forall l dd, l <> nil ->
                 dd = Dist_p l <->
                 InA equiv dd l /\
                 forall a, InA equiv a l -> dd <= a.
  Proof.
    intros l dd hl; apply Dist_p_spec in hl.
    destruct hl as [hl hl'].
    split; [intros h|intros [h hh]].
    split; subst; auto.
    specialize (hl' _ h).
    specialize (hh _ hl).
    lia.
  Qed.    

  Lemma Dist_p_ge:
    forall l a, l <> nil ->
                (forall x, InA equiv x l -> x >= a)
                -> Dist_p l >= a.
  Proof.
    intros l a hn h.
    unfold Dist_p, min_list.
    destruct l as [|b l]; [congruence|clear hn].
    induction l as [| bb l ih]; simpl; auto.
    assert (bb >= a); auto.
    assert (fold_right Init.Nat.min b l >= a).
    apply ih.
    intros x hx; apply h.
    rewrite InA_cons.
    rewrite InA_cons in hx; destruct hx as [hx|hx]; auto.
    destruct (Nat.min_spec bb (fold_right Init.Nat.min b l)); lia.
  Qed.
  
  Notation neigh g p := (neighbor_list (g p) (local_env g p)).

  Lemma RUN_unfold:
    forall (g: Env) (p: Node),
      RUN g p =
      (if root (g p)
       then simple_run_Root (g p)
       else
         if Nat.eq_dec (d (g p)) (Dist_p (map S (map snd (neigh g p)))) then
           if Nat.eq_dec (d (g p)) (S match local_env g p (par (g p)) with
                                      | Some s => d s | None => d (g p) end) then
             {| d := d (g p); par := par (g p); BFSROpart := g p |}
           else
             {| d := d (g p); par := Par_dist (d (g p)) (par (g p)) (neigh g p);
                BFSROpart := g p |}
         else
           {| d := Dist_p (map S (map snd (neigh g p))); par := par (g p);
              BFSROpart := g p |}).
  Proof.
    intros g p; unfold RUN, Algo, Algorithm.run, run, simple_run.
    auto.
  Qed.
  
  (** RUN: root *)
  Lemma RUN_root:
    forall g: Env, Input_Assume g -> d (RUN g Root) = 0.
  Proof.
    intros g [_ ass]; rewrite RUN_unfold.
    case_eq (root (g Root)); simpl; [lia|].
    assert (Root == Root) by auto.
    apply ass in H; congruence.
  Qed. 

  (** neighbor list *)
  Lemma neighbor_list_prop:
    forall (g: Env) (ass: Input_Assume g) (pg: Proper pequiv g)
           p c dd,
      InA equiv (c, dd) (neigh g p) <->
      exists q, is_channel p c q /\ dd = d (g q).
  Proof.
    intros g [ass _] pg p c dd.
    unfold neighbor_list.
    rewrite InA_map_filter_iff.
    2: { intros x y xy.
         assert (h: local_env g p x == local_env g p y) by
             (rewrite xy; auto).
         revert h; destruct (local_env g p x); destruct (local_env g p y);
           intros h; inversion h.
         apply Some_compat; simpl in H; rewrite xy, H; auto.
         auto. }
    split.
    { intros [cc [hc hcc]].
      rewrite (ass p) in hc.
      unfold local_env in hcc.    
      revert hcc; case_eq (peer p cc); simpl; [|tauto].
      intros q hq [h1 h2];
        unfold RelationPairs.RelCompFun in h1, h2; simpl in h1, h2.
      exists q; split.
      rewrite <- h1; unfold is_channel; rewrite hq; auto.
      lia.
    }
    { intros [q [hq hd]].
      exists c; split.
      apply ass.
      apply peers_spec_peer.
      exists q; auto.
      revert hq; unfold local_env, is_channel; destruct (peer p c); simpl.
      intros hn.
      assert (hd' : dd == d (g q)) by (rewrite hd; auto).
      rewrite <- hn in hd'.
      simpl in hd'; rewrite hd'; auto.
      tauto.
    }
  Qed.

  Lemma neighbor_list_proj:
    forall s l (l_compat: Proper pequiv l)
           c n, InA equiv (c, n) (neighbor_list s l) ->
                InA equiv c (neighbors s).
  Proof.
    intros s l lc c d ina; unfold neighbor_list in ina.
    rewrite InA_map_filter_iff in ina.
    2: { intros x y xy.
         assert (l x == l y) by (rewrite xy; auto).
         revert H; destruct (l x); destruct (l y); intros H;
           simpl in H; try tauto.
         simpl; destruct H as [H _]; rewrite H, xy; auto.
    }         
    destruct ina as [cc [h hh]].
    assert (cc == c).
    destruct (l cc); simpl in hh; try tauto.
    apply hh.
    rewrite <- H; auto.
  Qed.

  Lemma get_neigh: forall p r, ~ p == r -> exists q, G p q.
  Proof.
    intros p r hr;
      destruct (is_connected_has_neighbor
                  Connected _ _ hr (DA := Node_dec) p) as [q hq].
    exists q; auto.
  Qed.

  Lemma has_neigh':
    forall p r, ~ p == r -> exists c q, is_channel p c q.
  Proof.
    intros p r hr; destruct (get_neigh _ _ hr) as [q hq].
    apply Graph_network_assumption in hq.
    destruct hq as [c hc].
    exists c; exists q; auto.
  Qed.    

  Lemma technical_neigh:
    forall p q (g: Env), Proper pequiv g -> 
                         Input_Assume g -> ~ p == q ->
                         map S (map snd (neigh g p)) <> nil.
  Proof.
    intros p r g pg ass hr F; apply map_eq_nil, map_eq_nil in F.
    destruct (has_neigh' _ _ hr) as [c [q hc]].
    assert (FF: InA equiv (c, d (g q)) (neigh g p)).
    apply InA_map_filter_iff.
    { intros x y xy; unfold local_env.
      assert (peer p x == peer p y) by (rewrite xy; auto).
      destruct (peer p x); destruct (peer p y); simpl in H;
        try tauto; simpl.
      rewrite xy.
      assert (dum := d_compat (pg _ _ H)); simpl in dum; rewrite dum; auto. }
    exists c; split.
    apply ass, peers_spec_peer.
    exists q; auto.
    unfold local_env; unfold is_channel in hc.
    destruct (peer p c); [| inversion hc]; simpl.
    simpl in hc. assert (dum := d_compat (pg _ _ hc)).
    simpl in dum; rewrite dum; auto.
    rewrite F, InA_nil in FF; auto.
  Qed.
  
  Definition Neigh p: list Node :=
    filter (fun q => if G_dec p q then true else false) all_nodes.

  Global Instance Neigh_compat: Proper pequiv Neigh.
  Proof.
    intros p p' ep; unfold Neigh.
    apply filter_compat; [|reflexivity].
    intros q q' eq; case (G_dec p q); case (G_dec p' q');
      intros h h'; auto; rewrite eq, ep in h'; tauto.
  Qed.    

  Lemma Neigh_spec (p: Node):
    forall q, InA equiv q (Neigh p) <-> G p q.
  Proof.
    intros q; unfold Neigh; rewrite filter_InA.
    case (G_dec p q); intros h.
    split; [intros; auto| split; [apply all_nodes_prop|auto]].
    split; [intros [_ hh]; congruence|intros hh; tauto].
    { intros x y xy;
        destruct (G_dec p x); destruct (G_dec p y); auto.
      contradict n; rewrite <- xy; auto.
      contradict n; rewrite xy; auto. }
  Qed.

  Lemma Neigh_not_nil: forall p, ~ p == Root -> Neigh p <> nil.
  Proof.
    intros p hp; destruct (get_neigh _ _ hp) as [q hq].
    cut (InA equiv q (Neigh p)).
    { intros h F; rewrite F, InA_nil in h; tauto. }
    apply filter_InA.
    { intros pp pp' ep; case (G_dec p pp); case (G_dec p pp');
        intros h h'; auto; rewrite ep in h'; tauto. }
    split; [apply all_nodes_prop|].
    case (G_dec p q); [auto| intros h; tauto].
  Qed.

  Definition dist_Neigh g p := (map (fun q => d (g q)) (Neigh p)).

  Lemma dist_Neigh_spec:
    forall g (pg: Proper pequiv g) p x,
      InA equiv x (dist_Neigh g p) <->
      exists q, G p q /\ d (g q) = x.
  Proof.
    intros g pg p x; unfold dist_Neigh.
    rewrite InA_map_iff.
    split; intros [q [hq hqq]]; simpl in hq.
    exists q; split; [eapply Neigh_spec|]; eauto.
    exists q; simpl; split; [|eapply Neigh_spec]; eauto.
    {intros xx y xy; apply d_compat, pg, xy. }
  Qed.

  Lemma Sdist_Neigh_spec:
    forall g (pg: Proper pequiv g) p x,
      InA equiv x (map S (dist_Neigh g p)) <->
      exists q, G p q /\ S (d (g q)) = x.
  Proof.
    intros g pg p x.
    rewrite InA_map_iff; simpl.
    2: intros xx y xy; rewrite xy; auto.
    split; intros [a [ha hn]].
    apply dist_Neigh_spec in hn; auto.
    destruct hn as [q [hq hd]].
    exists q; split; [auto|lia].
    exists (d (g a)); split; [|eapply dist_Neigh_spec]; eauto.
  Qed.      
  
  Global Instance min_list_compat_equivlist:
    Proper (equivlistA equiv ==> eq) min_list.
  Proof.
    intros [| a l] [|b ll] ab; unfold min_list; auto.
    { edestruct ab with b.
      assert (F: InA equiv b nil) by auto.
      rewrite InA_nil in F; tauto. }
    { edestruct ab with a.
      assert (F: InA equiv a nil) by auto.
      rewrite InA_nil in F; tauto. }
    cut (fold_right Init.Nat.min a l =
         fold_right Init.Nat.min b ll).
    intros; rewrite H; auto.
    assert (ltd: forall a b, { a < b } + { ~ a < b }).
    { intros aa bb; destruct (le_gt_dec bb aa).
      right; lia.
      left; lia.
    }
    cut (fold_right (FoldRight.min _ ltd) a l ==
         fold_right (FoldRight.min _ ltd) b ll).
    { assert (forall y x, fold_right (FoldRight.min lt ltd) x y
                          = fold_right Init.Nat.min x y).
      { induction y as [| z y ih]; simpl; intros x; auto.
        unfold FoldRight.min at 1.
        destruct (ltd z (fold_right (FoldRight.min lt ltd) x y)).
        rewrite min_l; auto.
        rewrite <- ih; lia.
        rewrite min_r; auto.
        rewrite <- ih; lia.
      }        
      rewrite H, H.
      intros ok; simpl in ok; auto.
    }
    eapply FoldRight.fold_right_min_cons.
    { intros aa bb aab xx yy xxy; rewrite aab, xxy; auto. }
    { intros x y z h hh; lia. }
    { intros x y h hh; simpl in hh; lia. }
    { intros x y; case (Nat.eq_dec x y); intros h.
      left; auto.
      right; destruct (le_gt_dec x y); [left|right]; lia. }
    auto.
    Unshelve. { intros aa bb. case (Nat.eq_dec aa bb); intros h.
                left; auto. right; intros f; simpl in f; tauto. }
  Qed.
  
  Global Instance Dist_p_compat_equivlist:
    Proper (equivlistA equiv ==> eq) Dist_p.
  Proof.
    intros a b ab; unfold Dist_p.
    assert (TT := min_list_compat_equivlist ab).
    destruct (min_list a); destruct (min_list b); inversion TT;
      auto.
  Qed.

  Lemma Dist_p2:
    forall g (ass: Input_Assume g) (pg: Proper pequiv g) p,
      Dist_p (map S (map snd (neigh g p))) =
      Dist_p (map S (dist_Neigh g p)).
  Proof.
    intros g ass pg p; apply Dist_p_compat_equivlist.
    apply map_compat_equiv.
    { intros x y xy; rewrite xy; auto. }
    split; intros h.
    { apply dist_Neigh_spec; auto.
      rewrite InA_map_iff in h.
      2: apply snd_compat.
      destruct h as [[c dd] [hdd hc]].
      rewrite neighbor_list_prop in hc; simpl in hdd; auto.
      destruct hc as [q [hq hd]].
      exists q; split; simpl; [|lia].
      apply Graph_network_assumption.
      exists c; auto. }
    { rewrite InA_map_iff; rewrite dist_Neigh_spec in h; auto.
      2: apply snd_compat.
      destruct h as [q [hd hq]]; simpl in hd.
      apply Graph_network_assumption in hd.
      destruct hd as [c hc].
      exists (c, x); split; [simpl; auto|].
      rewrite neighbor_list_prop; auto.
      exists q; split; auto.
    }
  Qed.

  Lemma technical_not_nil0:
    forall g p, ~ p == Root -> dist_Neigh g p <> nil.
  Proof.
    intros g p hp F.
    unfold dist_Neigh in F; apply map_eq_nil in F.
    contradict F; apply Neigh_not_nil; auto.
  Qed.

  Lemma technical_not_nil:
    forall g p, ~ p == Root -> map S (dist_Neigh g p) <> nil.
  Proof.
    intros g p hp F; apply map_eq_nil in F.
    contradict F; apply technical_not_nil0; auto.
  Qed.

  Lemma DistOK:
    forall (g: Env) (ass: Input_Assume g) (pg: Proper pequiv g)
           p dd,
      ~ p == Root ->
      Dist_p (map S (dist_Neigh g p)) = dd -> 
      (exists q, G p q /\ S (d (g q)) = dd) /\
      forall q, G p q -> S (d (g q)) >= dd.
  Proof.
    intros g ass pg p dd hp hdd.
    assert (H := technical_not_nil g p hp).
    apply Dist_p_spec in H.
    rewrite hdd in H; destruct H as [H1 H2].
    split.
    { apply Sdist_Neigh_spec in H1; auto. }
    { intros q hq; apply H2.
      apply Sdist_Neigh_spec; auto.
      exists q; split; auto. }
  Qed.

  (** RUN: non root nodes -- CD Action *)
  Lemma RUN_CD0:
    forall (g: Env) (ass: Input_Assume g) (pg: Proper pequiv g)
           p (hr: ~ p == Root),
      d (RUN g p) = Dist_p (map S (dist_Neigh g p)).
  Proof.
    intros g ass pg p hr.
    rewrite <- Dist_p2; auto.
    rewrite RUN_unfold.
    case_eq (root (g p)); intros hhr.
    { contradict hr; destruct ass as [_ ass]; rewrite <- ass;
        auto. }
    clear hhr.
    destruct (Nat.eq_dec (d (g p))
                         (Dist_p (map S (map snd (neigh g p))))).
    destruct (Nat.eq_dec (d (g p))
                         (S
                            match local_env g p (par (g p)) with
                            | Some s => d s | None => d (g p)
                            end)).
    simpl; auto.
    simpl; auto.
    simpl; auto.
  Qed.

  Lemma RUN_CD2:
    forall (g: Env) (ass: Input_Assume g) (pg: Proper pequiv g)
           p (hr: ~ p == Root),
      (exists q, G p q /\ S (d (g q)) = d (RUN g p)) /\
      forall q, G p q -> S (d (g q)) >= d (RUN g p).
  Proof.
    intros g ass pg p hr.
    apply DistOK; auto.
    symmetry; apply RUN_CD0; auto.
  Qed.

  Lemma filter_cons:
    forall A (f: A -> bool) a l,
      filter f (a :: l) = if f a then a :: filter f l
                          else filter f l.
  Proof. intros; simpl; auto. Qed.
  
  Lemma Par_dist_prop:
    forall l pd dum,
      (exists c n, InA equiv (c, n) l /\ S n = pd) ->
      exists nn, InA equiv (Par_dist pd dum l, nn) l /\ S nn = pd.
  Proof.
    intros l pd dum [c [n [h hh]]].
    unfold Par_dist.
    case_eq (filter (fun cn => if Nat.eq_dec (S (snd cn)) pd
                               then true else false) l).
    intros F.
    assert (InA equiv (c, n)
                (filter (fun cn => if Nat.eq_dec (S (snd cn)) pd
                                   then true else false) l)).
    apply filter_InA.
    { intros x y xy.
      assert (S (snd x) == S (snd y)).      
      apply f_equal; change (snd x == snd y).
      eapply snd_compat, xy. 
      rewrite H; auto. }
    split; auto; case (Nat.eq_dec (S (snd (c, n))) pd);
      [auto|simpl; tauto].
    rewrite F, InA_nil in H; tauto.
    intros p ll hp.
    exists (snd p).
    simpl hd.
    assert (H: InA equiv p
                   (filter (fun cn => if Nat.eq_dec (S (snd cn)) pd
                                      then true else false) l)).
    rewrite hp; auto.
    rewrite filter_InA in H.
    2: { intros x y xy.
         assert (S (snd x) == S (snd y)).      
         apply f_equal; change (snd x == snd y).
         eapply snd_compat, xy. 
         rewrite H0; auto. }
    destruct H as [H HH].
    assert ((fst p, snd p) = p) by (destruct p; simpl; auto).
    split.
    rewrite H0; apply H.
    destruct (Nat.eq_dec (S (snd p)) pd); [auto|congruence].
  Qed.
  
  (** RUN: non root nodes -- CP Action *)
  Lemma RUN_CP:
    forall (g: Env) (ass: Input_Assume g) (pg: Proper pequiv g)
           p, ~ p == Root ->
              d (g p) = Dist_p (map S (dist_Neigh g p)) ->
              G p (Par' (RUN g p) p) /\
              d (g p) = S (d (g (Par' (RUN g p) p))).
  Proof.
    intros g ass pg p hr hd.
    rewrite <- Dist_p2 in hd; auto.
    rewrite RUN_unfold.
    case_eq (root (g p)); intros hhr.
    { contradict hr; destruct ass as [_ ass]; rewrite <- ass;
        auto. }
    clear hhr.
    destruct (Nat.eq_dec (d (g p))
                         (Dist_p (map S (map snd (neigh g p))))).
    2: lia.
    destruct (Nat.eq_dec (d (g p))
                         (S match local_env g p (par (g p)) with
                            | Some s => d s | None => d (g p)
                            end)).
    { unfold local_env in e0; case_eq (peer p (par (g p))).
      2: { intros h; rewrite h in e0; simpl in e0; lia. }
      intros n hn; rewrite hn in e0; simpl in e0.
      split.
      apply Par'_spec; [apply ass|]; simpl.
      apply peers_spec_peer.
      exists n; unfold is_channel; rewrite hn; auto.
      rewrite e0 at 1; apply f_equal, d_compat, pg.
      unfold Par'; simpl par; symmetry; apply Par__chan_conv.
      2: { rewrite hn; auto. }
      apply ass, peers_spec_peer.
      exists n; unfold is_channel; rewrite hn; auto. }
    { destruct (@Par_dist_prop (neigh g p) (d (g p)) (par (g p))).
      { assert (h := technical_neigh p Root pg ass hr).
        apply Dist_p_spec in h.
        destruct h as [ina _].
        rewrite <- hd in ina.
        rewrite InA_map_iff in ina.
        2: intros x y xy; simpl in xy; simpl; lia.
        destruct ina as [a [ha ina]].
        rewrite InA_map_iff in ina.
        2: apply snd_compat.
        destruct ina as [[c nn] [hc ina]].
        exists c; exists nn; split; auto.
        simpl in hc, ha; lia. }
      { destruct H as [H HH].
        assert (TT: InA equiv
                        (Par_dist (d (g p))
                                  (par (g p)) (neigh g p))
                        (neighbors (g p))).
        { apply neighbor_list_proj in H; auto.
          intros xx y xy; rewrite xy; auto. }
        unfold Par'; simpl par; simpl neighbors.
        split.
        - apply Par__spec; auto.
          revert TT; apply InA_compat; auto.
          apply setoid_equiv.
          symmetry; apply ass.
        - rewrite <- HH at 1.
          apply f_equal.
          rewrite neighbor_list_prop in H; auto.
          destruct H as [q [hq hqq]].
          rewrite hqq; apply d_compat, pg.
          symmetry; apply Par__chan_conv; auto.
      }
    }
  Qed.
  
  (** enabled / disabled nodes *)
  Lemma disabled_root:
    forall g: Env, Input_Assume g ->
                   enabled_b g Root = false <-> d (g Root) = 0.
  Proof.
    intros g ass; unfold enabled_b.
    case (State_dec (RUN g Root) (g Root)); intros R; split;
      try congruence.
    intros _; rewrite <- R; apply RUN_root; auto.
    intros h; contradict R; rewrite RUN_unfold.
    case_eq (root (g Root)); simpl.
    repeat split; auto; reflexivity.
    assert (Root == Root) by auto.
    apply ass in H; congruence.
  Qed.

  Lemma enabled_root:
    forall g: Env,
      Input_Assume g ->
      enabled_b g Root = true <-> d (g Root) <> 0.
  Proof.
    intros g ass.
    apply disabled_root in ass.
    destruct (enabled_b g Root); intuition.
  Qed.
  
  Lemma disabled_non_root':
    forall p, ~ p == Root -> forall g: Env,
        Input_Assume g -> Proper pequiv g ->
        enabled_b g p = false ->
        d (g p) = Dist_p (map S (dist_Neigh g p)) /\
        d (g p) = S (d (g (Par g p))).
  Proof.
    intros p hp g ass pg; unfold enabled_b.
    case (State_dec (RUN g p) (g p)).
    2: { intros _ f; congruence. }
    intros R _.
    assert (R0 := RUN_CD0 ass pg _ hp).
    rewrite R in R0.
    split; auto.
    assert (R2 := RUN_CP ass pg _ hp R0).
    destruct R2 as [_ R2].
    unfold Par; rewrite <- R at 2; auto.
  Qed.

  Lemma disabled_non_root_CP:
    forall p, ~ p == Root -> forall g: Env,
        Input_Assume g -> Proper pequiv g ->
        enabled_b g p = false ->
        G p (Par g p) /\
        forall q, G p q -> d (g p) <= S (d (g q)).
  Proof.
    intros p hp g ass pg en.
    dup en R.
    revert R; unfold enabled_b; case (State_dec (RUN g p) (g p)).
    2: { intros _ f; congruence. }
    intros R _.
    apply disabled_non_root' in en; auto.
    destruct en as [D1 D2].
    assert (D4 := RUN_CP ass pg _ hp D1).
    destruct D4 as [D4 _].
    split.
    unfold Par; rewrite <- R; auto.
    apply DistOK; auto.
  Qed.

  Lemma disabled_non_root:
    forall p, ~ p == Root -> forall g: Env,
        Input_Assume g -> Proper pequiv g ->
        enabled_b g p = false ->
        G p (Par g p) /\ d (g p) = S (d (g (Par g p))) /\
        forall q, G p q -> d (g p) <= S (d (g q)).
  Proof.
    intros p hp g ass pg en.
    dup en en'.
    apply disabled_non_root' in en; auto.
    apply disabled_non_root_CP in en'; auto.
    intuition.
  Qed.
  
  Lemma enabled_non_root':
    forall p, ~ p == Root -> forall g: Env,
        Input_Assume g -> Proper pequiv g ->
        enabled_b g p = true ->
        d (g p) <> Dist_p (map S (dist_Neigh g p)) \/
        d (g p) <> S (d (g (Par g p))).
  Proof.
    intros p hp g ass pg; unfold enabled_b.
    case (State_dec (RUN g p) (g p)).
    { intros _ f; congruence. }
    rewrite RUN_unfold.
    case_eq (root (g p)); intros hhr.
    { contradict hp; destruct ass as [_ ass]; rewrite <- ass;
        auto. }
    clear hhr.
    destruct (Nat.eq_dec (d (g p))
                         (Dist_p (map S (map snd (neigh g p))))).    
    2: { intros _ _; left; rewrite <- Dist_p2; auto. }
    destruct (Nat.eq_dec (d (g p))
                         (S match local_env g p (par (g p)) with
                            | Some s => d s | None => d (g p)
                            end)).
    { intros F; contradict F; repeat split; reflexivity. }
    { intros _ _; right; rewrite Dist_p2 in e; auto.
      contradict n; rewrite n at 1.
      unfold local_env.
      case_eq (peer p (par (g p))); simpl.
      { intros q hq.
      apply f_equal, d_compat, pg.
      unfold Par, Par'; apply Par__chan_conv.
      2: { rewrite hq; auto. }
      apply ass, peers_spec_peer.
      exists q; unfold is_channel; rewrite hq; auto. }
      { intros F; contradict n.
        cut (d (g p) = d (g (Par g p))); [lia|].
        apply d_compat, pg; unfold Par, Par'.
        symmetry; apply Par__None; rewrite F; auto.
      }
    }
  Qed.

  Lemma enabled_non_root_CD1:
    forall p, ~ p == Root -> forall g: Env,
        Input_Assume g -> Proper pequiv g ->
        ~ InA equiv (d (g p)) (map S (dist_Neigh g p))
        -> enabled_b g p = true.
  Proof.
    intros p hp g ass pg H.
    case_eq (enabled_b g p); intros en; [auto|].
    eapply (disabled_non_root _ hp) in en; auto.
    contradict H.
    apply Sdist_Neigh_spec; auto.
    exists (Par g p); split; [|symmetry]; apply en.
  Qed.
  
  Lemma enabled_non_root_CD2:
    forall p, ~ p == Root -> forall g: Env,
        Input_Assume g -> Proper pequiv g ->
        (exists q, G p q /\ d (g p) > S (d (g q)))
        -> enabled_b g p = true.
  Proof.
    intros p hp g ass pg.
    case_eq (enabled_b g p); intros en; [auto|].
    eapply (disabled_non_root _ hp) in en; auto.
    intros [q [hq H1]]. 
    destruct en as [_ [_ h]].
    specialize (h _ hq); lia.
  Qed.

  Lemma enabled_non_root_CP:
    forall p, ~ p == Root -> forall g: Env,
        Input_Assume g -> Proper pequiv g ->
        d (g p) = Dist_p (map S (dist_Neigh g p)) (* see DistOK *)
        -> d (g p) <> S (d (g (Par g p)))
        -> enabled_b g p = true.
  Proof.
    intros p hp g ass pg H.
    case_eq (enabled_b g p); intros en; [auto|].
    eapply (disabled_non_root _ hp) in en; auto.
    intros F; destruct en as [h1 [h2 h3]]; auto.
  Qed.

  (** Step *)
  Lemma step_Root:
    forall (g g': Env),
      Input_Assume g ->
      Step g' g -> has_moved_b g' g Root = true ->
      d (g' Root) = 0.
  Proof.
    intros g g' ass [pg pg' _ hR] hm.
    specialize (hR Root); destruct hR as [hR|hR].
    rewrite hR.
    apply RUN_root; auto.
    { revert hm; unfold has_moved_b;
        case (State_dec (g' Root) (g Root)).
      congruence.
      tauto.
    }
  Qed.

  Lemma step_CD0:
    forall (g g': Env) (ass: Input_Assume g) (st: Step g' g)
           p, ~ p == Root -> has_moved_b g' g p = true ->
              d (g' p) = Dist_p (map S (dist_Neigh g p)).
  Proof.
    intros g g' ass [pg pg' _ hR] p hp hm.
    specialize (hR p); destruct hR as [hR|hR].
    rewrite hR.
    apply RUN_CD0; auto.
    { revert hm; unfold has_moved_b;
        case (State_dec (g' p) (g p)).
      congruence.
      tauto.
    }
  Qed.

  Lemma step_CD2:
    forall (g g': Env) (ass: Input_Assume g) (st: Step g' g)
           p, ~ p == Root -> has_moved_b g' g p = true ->
              (exists q, G p q /\ S (d (g q)) = d (g' p)) /\
              (forall q, G p q -> S (d (g q)) >= d (g' p)).
  Proof.
    intros g g' ass [pg pg' _ hR] p hp hm.
    specialize (hR p); destruct hR as [hR|hR].
    2: { revert hm; unfold has_moved_b;
         case (State_dec (g' p) (g p)); 
         [congruence|tauto]. }
    apply RUN_CD2 with (g := g) in hp; auto.
    destruct hp as [[q [hq h2]] h3].
    split; [exists q; split; auto| intros qq hqq]; rewrite hR; auto.
  Qed.

  Lemma step_CP:
    forall (g g': Env) (ass: Input_Assume g) (st: Step g' g)
           p, ~ p == Root -> has_moved_b g' g p = true ->
              d (g p) = Dist_p (map S (dist_Neigh g p)) ->
              G p (Par g' p) /\ d (g p) = S (d (g (Par g' p))).
  Proof.
    intros g g' ass [pg pg' _ hR] p hp hm h.
    specialize (hR p); destruct hR as [hR|hR].
    2: { revert hm; unfold has_moved_b;
         case (State_dec (g' p) (g p)); 
         [congruence|tauto]. }
    apply RUN_CP with (g := g) in hp; auto.
    destruct hp as [h1 h2].
    unfold Par; rewrite hR; split; auto.
  Qed.

  Definition Par_Rel (gg: Env): relation Node :=
    (fun n n' => if Node_dec n Root then False else
                   Par gg n == n').
  
  Global Instance Par_Rel_compat: Proper pequiv Par_Rel.
  Proof.
    intros g1 g2 eg n1 n2 en n'1 n'2 en'; unfold Par_Rel.
    case (Node_dec n1 Root); case (Node_dec n2 Root);
      intros h1 h2; auto; try (rewrite <- en in h1; tauto).
    apply equiv_compat; auto; apply Par_compat; auto.
  Qed.

  (** Par_Rel: no edge from root *)
  Lemma Par_Rel_is_root: forall g, is_root (Par_Rel g) Root.
  Proof.
    intros g p; unfold Par_Rel;
      case (Node_dec Root Root); intros hp.
    tauto.
    contradict hp; auto.
  Qed.

End BFS_common.

Close Scope signature_scope.
Close Scope nat_scope.
Unset Implicit Arguments.
