(**
*Certified Round Complexity of Self-Stabilizing Algorithms*
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Bool.
From Coq Require Import ZArith.
From Coq Require Import Lia.

(** **** Local Imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import NatUtils.
From Padec Require Import ListUtils.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import BFS_algo.
From Padec Require Import BFS_common.
From Padec Require Import BFS_specification.
From Padec Require Import Graph_Diameter.
From Padec Require Import Tree_topology_ALT.

(* end show *)
Open Scope nat_scope.
Open Scope signature_scope.
Set Implicit Arguments.

(** * BFS Spanning Tree Under a Distributed Unfair Daemon: Correctness
 
 This proof shows that every terminal configuration satisfies
 the predicate: [BFS_Spannig_Tree_spec]
 meaning that the par variables design a BFS spanning tree.
 *)

Section BFS_correctness.
  
  Context {Chans: Channels}
          {Net: Network}
          {Graph_assumption: BFS_Assume}
          {Graph_network_assumption: Graph_Network_matching}
          {D_Assumption: BFS_D}.
  Existing Instance G_compat.

  Notation Algo := (BFS_algo D).
  Notation Env := (Env (Net := Net) (Algo := Algo)).
  Notation terminal := (terminal (Net := Net) (Algo := Algo)).

  Notation dist := (dist G_compat G_dec Connected all_nodes_prop (DA := Node_dec)).
  
  (** We consider a terminal configuration, g *)
  Variable (g: Env) (term: terminal g) (pg: Proper pequiv g)
           (ass: Input_Assume g).
  
  (** *** Variables d are not underestimated in g *)
  Lemma d_not_underestimated: forall p, d (g p) >= dist p Root.
  Proof.
    intros p.
    case (le_gt_dec (dist p Root) (d (g p))); intros h; [lia|].
    pose (FF := filter
                  (fun n => if le_gt_dec (dist n Root) (d (g n))
                            then false else true) all_nodes).
    assert (pp: Proper pequiv
                       (fun n =>
                          if le_gt_dec (dist n Root) (d (g n))
                          then false else true)).
    { intros x y xy.
      destruct (le_gt_dec (dist x Root) (d (g x)));
        destruct (le_gt_dec (dist y Root) (d (g y))); auto;
          [rewrite xy in l|rewrite <- xy in l]; lia. }
    case_eq (argmin (fun n => d (g n)) FF).
    2: { intros hh; eapply argmin_None in hh.
         assert (InA equiv p FF).
         apply filter_InA; auto.
         split; [apply all_nodes_prop|].
         case (le_gt_dec (dist p Root) (d (g p))); intros hhh;
           [lia|auto].
         rewrite hh, InA_nil in H; tauto. }
    intros pmin hpmi.
    assert (hpmin: argmin (fun n => d (g n)) FF == Some pmin).
    rewrite hpmi; auto.
    clear hpmi.
    assert (h1: d (g pmin) < dist pmin Root).
    { apply argmin_min1 in hpmin.
      unfold FF in hpmin; apply filter_InA in hpmin; auto.
      destruct hpmin as [_ hh].
      destruct (le_gt_dec (dist pmin Root) (d (g pmin)));
        [congruence|auto]. }
    assert (F: forall q, G pmin q -> d (g pmin) < S (d (g q))).
    { intros q hq.
      case (le_gt_dec (dist q Root) (d (g q))); intros hd.
      - assert (h2: dist pmin Root <= S (dist q Root)).
        { apply Bidir in hq.
          destruct (dist_neighbors G_compat G_dec Connected all_nodes_prop Bidir Root _ _ hq (DA := Node_dec)) as [hh|[hh|hh]];
            lia. }
        lia.
      - cut (d (g pmin) <= d (g q)); [lia|].
        change ((fun n => d (g n)) pmin <= (fun n => d (g n)) q).
        eapply argmin_min2.
        { intros n1 n2 en; apply d_compat, pg, en. }
        apply hpmin.
        apply filter_InA; auto.
        split; [apply all_nodes_prop|].
        case (le_gt_dec (dist q Root) (d (g q))); [intros f; lia|auto]. }
    assert (F': d (g pmin) < D).
    { assert (X := D_dist pmin Root); lia. }
    assert (TT := terminal_all_disabled term).
    specialize (TT pmin); unfold enabled in TT.
    contradict TT.
    eapply enabled_non_root_CD1; auto.
    { intros hr; rewrite <- hr, dist_nil in h1; lia. }
    intros hi; rewrite InA_cons in hi; destruct hi as [hi|hi].
    - simpl in hi; lia.
    - rewrite Sdist_Neigh_spec in hi; auto.
      destruct hi as [q [hc hq]].
      apply F in hc.
      lia.
      Unshelve. apply Graph_network_assumption.
  Qed.

  Lemma d_is_dist: forall p, d (g p) = dist p Root.
  Proof.
    cut (forall dd, forall p, dist p Root = dd -> d (g p) = dd).
    { intros ok p; auto. }
    induction dd as [| k ih].
    { intros p hd.
      assert (TTT: enabled_b g Root = false).
      cut (~ enabled g Root).
      { intros h; unfold enabled in h.
        apply not_true_is_false in h; auto. }
      apply terminal_all_disabled; auto. 
      apply disabled_root in TTT; auto.
      rewrite <- TTT.
      apply d_compat, pg; rewrite <- is_dist_nul.
      rewrite <- hd; apply is_dist_dist. }
    { intros q hq.
      assert (hr: ~ q == Root).
      { intros hr; rewrite hr in hq; rewrite dist_nil in hq; lia. }
      destruct (dist_neighbor_ex G_compat G_dec Connected all_nodes_prop 
                                 q Root hq) as [q' [hq' hq'']].
      apply ih in hq''.
      cut (Dist_p (D :: map S (dist_Neigh g q)) = S k).
      { intros ok.
        assert (TT := terminal_all_disabled term).
        unfold enabled in TT. specialize (TT q).
        assert (TTT: enabled_b g q = false).
        { apply not_true_is_false in TT; auto. }
        eapply disabled_non_root' in TTT; auto.
        lia. }
      symmetry; rewrite Dist_p_spec'; auto.
      2: congruence.
      split.
      - rewrite InA_cons; right; rewrite Sdist_Neigh_spec; auto.
        exists q'; split; [auto|lia].
      - intros a ina; rewrite InA_cons in ina; destruct ina as [ina|ina].
        + simpl in ina.
          rewrite <- hq, ina; transitivity Diam;
            [apply diam_dist|apply HD].
        + rewrite Sdist_Neigh_spec in ina; auto.
          destruct ina as [q'' [hqq hqqq]].
          cut (d (g q'') >= k); [lia|].
          assert (AA := d_not_underestimated q'').
          destruct (dist_neighbors G_compat G_dec Connected all_nodes_prop Bidir Root (DA := Node_dec)
                                   q q'' hqq); lia.
          Unshelve. apply Graph_network_assumption.   
    }
  Qed.

  Lemma DistOK_true:
    forall p, ~ p == Root ->
              d (g p) = Dist_p (map S (dist_Neigh g p)).
  Proof.
    intros p hp.
    eapply Dist_p_spec'; auto. 
    eapply technical_not_nil; auto.
    split.
    { eapply Sdist_Neigh_spec; auto.
      assert (EE: exists k, dist p Root = S k).
      { case_eq (dist p Root).
        intros h; contradict hp.
        eapply is_dist_nul.
        rewrite <- h; apply is_dist_dist.
        intros n hn; exists n; auto. }
      destruct EE as [k hk].
      destruct (dist_neighbor_ex G_compat G_dec Connected all_nodes_prop (DA := Node_dec)
                                 _ _ hk) as [q [hq hd]].
      exists q; split; [auto|].
      rewrite d_is_dist, d_is_dist; auto.
      lia. }
    { intros k hk; eapply Sdist_Neigh_spec in hk; auto.
      destruct hk as [q [hq hd]].
      rewrite d_is_dist in hd; rewrite d_is_dist; auto.
      destruct (dist_neighbors G_compat G_dec Connected all_nodes_prop Bidir Root (DA := Node_dec)
                               _ _ hq); lia.
    }
  Qed.

  Lemma parOK:
    forall p, ~ p == Root -> d (g p) = S (d (g (Par g p))) /\
                             G p (Par g p).
  Proof.
    intros p hp.
    assert (dum := terminal_all_disabled term).
    specialize (dum p); unfold enabled in dum.
    apply not_true_is_false in dum; auto.
    eapply disabled_non_root_CP in dum; auto.
    split; apply dum.
    apply DistOK_true; auto.
    Unshelve. apply Graph_network_assumption.    
  Qed.

  (** About BFS TREE *)
  
  Lemma path_Par_ex:
    forall p, ~ p == Root ->
              exists pt, is_path (Par_Rel g) p pt Root /\
                         S (length pt) = dist p Root.
  Proof.
    cut (forall dd p,
            ~ p == Root -> dist p Root = dd ->
            exists pt, is_path (Par_Rel g) p pt Root /\
                       S (length pt) = dd).
    { intros ok p; auto. }
    induction dd as [| dd ih].
    { intros p hp hd.
      contradict hp.
      eapply is_dist_nul.
      rewrite <- hd.
      apply is_dist_dist. }
    intros p hp hd.
    destruct (parOK p hp) as [DSP GP].
    assert (DP: dist (Par g p) Root = dd).
    { rewrite d_is_dist, d_is_dist in DSP; lia. }
    clear DSP.
    case (Node_dec (Par g p) Root); intros hr.
    { exists nil; simpl.
      repeat split.
      unfold Par_Rel.
      case (Node_dec p Root); [tauto|auto].
      apply f_equal; rewrite <- DP, hr; symmetry; apply dist_nil.
    }
    { specialize (ih _ hr DP); destruct ih as [pt [h1 h2]].
      exists (Par g p :: pt); repeat split; auto.
      unfold Par_Rel; case (Node_dec p Root); [tauto|auto].
    }
  Qed.

  Lemma path_Par:
    forall p pt, is_path (Par_Rel g) p pt Root ->
                 S (length pt) = dist p Root.
  Proof.
    cut (forall dd p,
            dist p Root = dd ->
            forall pt, is_path (Par_Rel g) p pt Root ->
                       S (length pt) = dd).
    { intros ok p pt; apply ok; auto. }
    induction dd as [| dd ih].
    { intros p hd pt hpt.
      assert (p == Root).
      eapply is_dist_nul; rewrite <- hd; apply is_dist_dist.
      destruct pt as [| a pt]; simpl in hpt.
      revert hpt; unfold Par_Rel; case (Node_dec p Root); tauto.
      destruct hpt as [hpt _].
      revert hpt; unfold Par_Rel; case (Node_dec p Root); tauto.
    }
    intros p hd.
    case (Node_dec p Root); intros hp.
    { assert (DD := dist_nil G_compat G_dec Connected all_nodes_prop Root (DA := Node_dec)).
      rewrite <- hp in DD at 1.
      contradict DD; lia. }
    destruct (parOK p hp) as [DSP GP].
    assert (DP: dist (Par g p) Root = dd).
    { rewrite d_is_dist, d_is_dist in DSP; lia. }
    clear DSP.
    intros [| a pt] hpt; simpl in hpt.
    { simpl; apply f_equal.
      unfold Par_Rel in hpt.
      revert hpt; case (Node_dec p Root); [tauto|]; intros hpp hpt.
      rewrite hpt in DP; rewrite <- DP; symmetry; apply dist_nil. }
    destruct hpt as [hh hpt].
    unfold Par_Rel in hh.
    revert hh; case (Node_dec p Root); [tauto|]; intros hpp hh.
    rewrite hh in DP.      
    specialize (ih _ DP _ hpt).
    simpl; lia.
  Qed.

  (* Par_Rel: dist to Root decreases along any Par_Rel path *)
  Lemma Par_Rel_dist_decreases:
    forall p q pt, is_path (Par_Rel g) p pt q ->
                   dist p Root > dist q Root.
  Proof.
    intros p q pt; revert p q; induction pt as [| a pt ih];
      intros p q; simpl.
    - unfold Par_Rel; case (Node_dec p Root); [tauto|]; intros hr.
      intros h.
      assert (PP := parOK p hr).
      destruct PP as [PP _]; rewrite d_is_dist, d_is_dist in PP.
      rewrite h in PP; lia.
    - unfold Par_Rel at 1; case (Node_dec p Root); [tauto|]; intros hr.
      intros [h hh].
      assert (PP := parOK p hr).
      destruct PP as [PP _]; rewrite d_is_dist, d_is_dist in PP.
      rewrite h in PP; transitivity (dist a Root).
      lia.
      apply ih; auto.
  Qed.
  
  (* Par_Rel is_dag: *)
  Lemma Par_Rel_is_dag: is_dag (Par_Rel g).
  Proof.
    intros p h.
    rewrite RelationA.transitive_closure_iff_exists_non_empty_chain in h.
    destruct h as [pt hp].
    assert (PP: is_path (Par_Rel g) p pt p).
    apply hp.
    apply Par_Rel_dist_decreases in PP; lia.
  Qed.
  
  (* Par_Rel: a single parent for every non-root node *)
  Lemma Par_Rel_single_parent: single_parent (Par_Rel g).
  Proof.
    intros p q q'; unfold Par_Rel.
    case (Node_dec p Root); intros hp.
    tauto.
    intros h1 h2; rewrite <- h1, <- h2; auto.
  Qed.
  
  Lemma Par_Rel_subgraph_G:
    is_subgraph (Par_Rel g) G.
  Proof.
    intros p q; unfold Par_Rel; case (Node_dec p Root); intros hp.
    tauto.
    intros h.
    assert (P := parOK _ hp).
    rewrite <- h; apply P.
  Qed.

  Lemma G_is_R_Net:
    G =~= fun n n' => exists c, is_channel n c n'.
  Proof.
    intros a b ab x y xy; rewrite xy, ab; split; intros h;
      apply Graph_network_assumption; auto.
  Qed.

  (* Par_Rel_path_to_root: every node has a Par_Rel path to the root *)
  Lemma Par_Rel_path_to_root: forall x : Node,
      ~ x == Root ->
      exists p : list Node, is_path (Par_Rel g) x p Root.
  Proof.
    intros x hx.
    destruct (path_Par_ex _ hx) as [pt hpt].
    exists pt; apply hpt.
  Qed.

  (* Par_Rel is BFS : is_BFS (Par_Rel g) G Root *)
  Lemma Par_Rel_BFS: is_BFS (Par_Rel g) G Root.
  Proof.
    intros x d hd.
    case (Node_dec x Root); intros hr.
    { rewrite hr; rewrite hr in hd.
      apply is_dist_nul' in hd; rewrite hd.
      apply is_dist_nul; auto.
    }
    destruct (path_Par_ex x hr) as [pt [hpt hpd]].
    destruct hd as [[hd _]|[_ [[ptt [hd1 hd3]] hd2]]]; [tauto|].
    assert (EP:  S (length pt) = d). 
    { rewrite <- hd3; apply f_equal, length_compat; auto.
      eapply is_path_unique_in_tree.
      apply Par_Rel_compat, pg.
      apply Par_Rel_is_dag.
      apply Par_Rel_single_parent.
      apply hpt.
      apply hd1. }
    rewrite <- EP, hpd.
    apply is_dist_dist.
  Qed.
  
  (** ** Correctness *)
  Theorem BFS_correctness:
    (BFS_Spanning_Tree_spec (Net := Net)) Root Par_Rel g.
  Proof.
    repeat split.
    - apply Par_Rel_is_root.
    - transitivity G.
      apply Par_Rel_subgraph_G.
      intros a b ab; eapply G_is_R_Net.
      3: apply ab.
      auto. auto.
    - apply Par_Rel_is_dag.
    - apply Par_Rel_single_parent.
    - intros x; apply Par_Rel_path_to_root.
    - assert (PP := Par_Rel_BFS).
      revert PP; apply is_BFS_compat; auto.
      apply Par_Rel_compat, pg.
      symmetry; apply G_is_R_Net.
  Qed.
  
End BFS_correctness.

Close Scope signature_scope.
Close Scope nat_scope.
Unset Implicit Arguments.
