
(**
*Certified Round Complexity of Self-Stabilizing Algorithms*
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Bool.
From Coq Require Import ZArith.
From Coq Require Import Lia.

(** **** Local Imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import NatUtils.
From Padec Require Import ListUtils.
From Padec Require Import BoolSet.
From Padec Require Import All_In.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Stream.
From Padec Require Import Exec.
From Padec Require Import Fairness.
From Padec Require Import Rounds.
From Padec Require Import BFS_algo_bis.
From Padec Require Import BFS_common_bis.
From Padec Require Import Graph_Diameter.
From Padec Require Import BFS_specification.
From Padec Require Import BFS_correctness_bis.
From Padec Require Import Tree_topology_ALT.

(* end show *)
Open Scope nat_scope.
Open Scope signature_scope.
Set Implicit Arguments.

(** * BFS Spanning Tree: Round Complexity *)

Section BFS_rounds.
  
  Context {Chans: Channels}
          {Net: Network}
          {Graph_assumption: BFS_Assume}
          {Graph_network_assumption: Graph_Network_matching}.
  Existing Instance G_compat.
  Existing Instance BFS_RO_assumption.
  
  Notation Algo := (BFS_algo).
  Notation Env := (Env (Net := Net) (Algo := Algo)).
  Notation terminal := (terminal (Net := Net) (Algo := Algo)).
  Notation Exec := (Exec (Net := Net) (Algo := Algo)).

  (** Tools for dist *)
  Notation dist := (dist G_compat G_dec Connected all_nodes_prop (DA := Node_dec)).
  Notation dist_neighbors :=
    (dist_neighbors G_compat G_dec Connected all_nodes_prop
                    Bidir Root (DA := Node_dec)).
  Notation dist_neighbor_ex :=
    (dist_neighbor_ex G_compat G_dec Connected all_nodes_prop
                      (DA := Node_dec)).
  Lemma diam_dist: forall p q, dist p q <= Diam.
  Proof. apply diam_dist. Qed.
  Lemma dist_nil': forall p q, dist p q = 0 <-> p == q.
  Proof. apply dist_nil'. Qed.
  (** *)

  Notation NodeSet := (BoolSet (A:=Node)).
  Notation eqNodeSet := (eqBoolSet (eqA := (equiv (A:=Node)))).
  Notation leNodeSet := (leBoolSet (eqA := (equiv (A:=Node)))).

  Definition CD1 := (fun (k: nat) (p: Node) (g: Env) =>
                       (dist p Root < k -> d (g p) = dist p Root)).
  Definition CD2 := (fun (k: nat) (p: Node) (g: Env) =>
                       (dist p Root >= k -> d (g p) >= k)).

  Definition CD :=
    (fun (k: nat) (p: Node) (g: Env) => CD1 k p g /\ CD2 k p g).
  
  Definition check_dist (k: nat) (g: Env): Prop :=
    forall p, CD k p g.

  Instance CD1_compat: Proper pequiv CD1.
  Proof.
    intros k1 k ek; simpl in ek; subst.
    intros p1 p2 ep g1 g2 eg.
    apply impl_compat; [| rewrite (eg _ _ ep)]; rewrite ep; tauto.
  Qed.

  Instance CD2_compat: Proper pequiv CD2.
  Proof.
    intros k1 k ek; simpl in ek; subst.
    intros p1 p2 ep g1 g2 eg.
    apply impl_compat; [rewrite ep| rewrite (eg _ _ ep)]; tauto.
  Qed.
  
  Instance CD_compat: Proper pequiv CD.
  Proof.
    intros k1 k ek; simpl in ek; subst.
    intros p1 p2 ep g1 g2 eg.
    apply and_compat; [apply CD1_compat|apply CD2_compat]; auto.
  Qed.

  Instance check_dist_compat: Proper pequiv check_dist.
  Proof.
    intros k1 k ek g1 g2 eg.
    apply all_compat; intros p;     apply and_compat; 
      [apply CD1_compat|apply CD2_compat]; auto.
  Qed.
  
  Lemma CD1_dec: forall k p g, {CD1 k p g} + {~ CD1 k p g}.
  Proof.
    intros k p g; unfold CD1.
    case (le_gt_dec k (dist p Root)); intros hk.
    left; intros h; lia.
    case (Nat.eq_dec (d (g p)) (dist p Root)); intros hd.
    left; lia.
    right; intros F; lia.    
  Qed.   

  Lemma CD2_dec: forall k p g, {CD2 k p g} + {~ CD2 k p g}.
  Proof.
    intros k p g; unfold CD2.
    case (le_gt_dec k (dist p Root)); intros hk.
    case (le_gt_dec k (d (g p))); intros hkk.
    left; lia.
    right; intros  F; lia.    
    left; intros h; lia.
  Qed.   

  Lemma CD_dec: forall k p g, {CD k p g} + {~ CD k p g}.
  Proof.
    intros k p g; unfold CD.
    destruct (CD1_dec k p g).
    destruct (CD2_dec k p g).
    left; split; auto.
    right; intuition.
    right; intuition.
  Qed.   

  Lemma check_dist_dec:
    forall k g,
      Proper pequiv g ->
      { check_dist k g } + { ~check_dist k g }.
  Proof.
    intros k g pg.
    destruct 
      (All_In_dec_forall_ex
         all_nodes_prop (P := fun p => ~ CD k p g))
      as [[p h]|h].
    { intros p p' ep.
      apply not_compat.
      apply CD_compat; auto. }
    { intros p; destruct (CD_dec k p g); [right|left]; auto. }
    - right.
      contradict h.
      specialize (h p); destruct h as [h1 h2].
      split; [apply h1|apply h2].
    - left; intros p; specialize (h p).
      destruct (CD_dec k p g); auto; intuition.
  Qed.

  Lemma check_dist_clas:
    forall k g,
      Proper pequiv g ->
      check_dist k g \/ ~check_dist k g .
  Proof.
    intros k g pg; destruct (check_dist_dec k pg);
      [left|right]; auto.
  Qed.

  Lemma check_dist_incr:
    forall k g, check_dist k g ->
                forall k', k' <= k -> check_dist k' g.
  Proof.
    unfold check_dist, CD, CD1, CD2.
    intros k g ac k' hk p.
    destruct (ac p) as [ac1 ac2].
    split; intros h; lia.
  Qed.

  Lemma check_dist_0: forall g, check_dist 0 g.
  Proof.
    intros g p; unfold CD, CD1, CD2; lia.
  Qed.

  Lemma check_dist_Diam:
    forall g, check_dist (S Diam) g <->
              forall p, d (g p) = dist p Root.
  Proof.
    unfold check_dist, CD, CD1, CD2.
    intros g; split; intros h p.

    specialize (h p); destruct h as [h1 _].
    apply h1.
    assert (dd := diam_dist p Root).
    lia.

    split; intros hh; auto.
    assert (dd := diam_dist p Root).
    lia.
  Qed.

  Lemma check_dist_ge_Diam:
    forall g, check_dist (S Diam) g ->
              forall k, S Diam <= k -> check_dist k g.
  Proof.
    unfold check_dist, CD, CD1, CD2.
    intros g ac k hk p.
    specialize (ac p); destruct ac as [ac1 ac2].
    split; assert (dd := diam_dist p Root); intros h.
    apply ac1; lia.
    lia.    
  Qed.

  Lemma CD1_step:
    forall g' g, Assume g -> Step g' g ->
                 forall k p,
                   (forall q, G p q -> CD k q g) ->
                   CD1 (S k) p g -> CD1 (S k) p g'.
  Proof.
    unfold CD1.
    intros g' g ass st k p hneigh hp.
    assert (pg' := Step_proper_l st).
    assert (pg := Step_proper_r st).

    case (Node_dec p Root); intros hr.
    
    { (*Root*)
      rewrite hr; rewrite hr in hp.
      rewrite dist_nil; rewrite dist_nil in hp.
      intros h.
      destruct (Step_correct st Root); rewrite H.
      apply RUN_root; auto.
      auto.
    }
    
    (*non-Root*)
    destruct (Step_correct st p); rewrite H; clear H.
    2: { apply hp; auto. }
    erewrite RUN_CD0; auto.
    Unshelve. 2: { auto. }
    remember (dist p Root) as n; symmetry in Heqn.
    destruct n as [| n].
    { rewrite dist_nil' in Heqn; tauto. }
    destruct (dist_neighbor_ex  _ _ Heqn) as [q [hg hd]].
    intros h.
    symmetry; rewrite Dist_p_spec'.
    2: apply technical_not_nil; auto.
    split.
    - rewrite Sdist_Neigh_spec; auto.
      exists q; split; auto.
      apply f_equal.
      specialize (hneigh q hg).
      rewrite <- hd; apply hneigh; lia.
    - intros a hha.
      apply Sdist_Neigh_spec in hha; auto.
      destruct hha as [q' [hq' hd']].
      specialize (hneigh q' hq'); destruct hneigh
        as [hneigh1 hneigh2].
      case (le_gt_dec k (dist q' Root)); intros lg.
      * specialize (hneigh2 lg); lia.
      * specialize (hneigh1 lg).
        apply dist_neighbors in hq'; eauto.
        destruct hq' as [hq'|[hq'|hq']]; lia.
  Qed.
  
  Lemma CD2_step:
    forall g' g, Assume g -> Step g' g ->
                 forall k p,
                   (forall q, G p q -> CD k q g) ->
                   CD2 (S k) p g -> CD2 (S k) p g'.
  Proof.
    unfold CD2.
    intros g' g ass st k p hneigh hp.
    assert (pg' := Step_proper_l st).
    assert (pg := Step_proper_r st).

    case (Node_dec p Root); intros hr.
    
    { (*Root*)
      rewrite hr, dist_nil; lia. 
    }
    
    (*non-Root*)
    destruct (Step_correct st p); rewrite H; clear H.
    2: { apply hp; auto. }
    erewrite RUN_CD0; auto.
    Unshelve. 2: { auto. }
    remember (dist p Root) as n; symmetry in Heqn.
    destruct n as [| n].
    { rewrite dist_nil' in Heqn; tauto. }
    destruct (dist_neighbor_ex _ _ Heqn) as [q [hg hd]].
    intros h.
    { apply Dist_p_ge.
      apply technical_not_nil; auto.
      intros x hx.
      apply Sdist_Neigh_spec in hx; auto.
      destruct hx as [qq [hqq hdd]].
      specialize (hneigh qq hqq); destruct hneigh
        as [hneigh1 hneigh2].
      case (le_gt_dec k (dist qq Root)); intros lg.
      * specialize (hneigh2 lg); lia.
      * specialize (hneigh1 lg).
        apply dist_neighbors in hqq.
        destruct hqq as [hq'|[hq'|hq']]; lia.
    }
  Qed.

  Lemma CD_step:
    forall g' g, Assume g -> Step g' g ->
                 forall k p,
                   (forall q, G p q -> CD k q g) ->
                   CD (S k) p g -> CD (S k) p g'.
  Proof.
    intros g' g ass ste k p hn [h1 h2]; split.
    apply CD1_step with g; auto.
    apply CD2_step with g; auto.
  Qed.    
  
  Lemma check_dist_step:
    forall g' g, Assume g -> Step g' g ->
                 forall k, check_dist k g -> check_dist k g'.
  Proof.
    intros g' g ass st [|k] h p.
    split; intros hh; lia.
    eapply CD_step; eauto.
    intros q hh.
    apply check_dist_incr with (S k); [auto|lia].
  Qed.
  
  Lemma CD1_exec_Always:
    forall e, Assume (s_hd e) -> is_exec e ->
              forall k p,
                check_dist k (s_hd e) ->
                CD1 (S k) p (s_hd e) ->
                Always (fun c => CD1 (S k) p (s_hd c)) e.
  Proof.
    cofix h; intros [g | g e] ass ise k p hcd h1.
    { left; auto. }
    { right; auto.
      apply h.
      { eapply step_Assume.
        apply ass.
        apply is_exec_inv_1 in ise; auto. }
      { apply is_exec_inv_2 in ise; auto. }
      { apply check_dist_step with g; auto.
        apply is_exec_inv_1 in ise; auto. }
      apply CD1_step with g; auto. 
      { apply is_exec_inv_1 in ise; auto. }
    }
  Qed.

  Lemma CD2_exec_Always:
    forall e, Assume (s_hd e) -> is_exec e ->
              forall k p,
                check_dist k (s_hd e) ->
                CD2 (S k) p (s_hd e) ->
                Always (fun c => CD2 (S k) p (s_hd c)) e.
  Proof.
    cofix h; intros [g | g e] ass ise k p hcd h2.
    { left; auto. }
    { right; auto.
      apply h.
      { eapply step_Assume.
        apply ass.
        apply is_exec_inv_1 in ise; auto. }
      { apply is_exec_inv_2 in ise; auto. }
      { apply check_dist_step with g; auto.
        apply is_exec_inv_1 in ise; auto. }
      apply CD2_step with g; auto. 
      { apply is_exec_inv_1 in ise; auto. }
    }
  Qed.

  Lemma check_dist_exec_Always:
    forall e, Assume (s_hd e) -> is_exec e ->
              forall k, check_dist k (s_hd e) ->
                        Always (fun c => check_dist k (s_hd c)) e.
  Proof.
    cofix h; intros [g | g e] ass ise k hcd.
    { left; auto. }
    { right; auto.
      apply h.
      { eapply step_Assume.
        apply ass.
        apply is_exec_inv_1 in ise; auto. }
      { apply is_exec_inv_2 in ise; auto. }
      apply check_dist_step with g; auto.
      { apply is_exec_inv_1 in ise; auto. }
    }
  Qed.

  Lemma check_dist_terminal:
    forall g, Proper pequiv g -> terminal g ->
              Assume g ->
              forall k, check_dist k g.
  Proof.
    intros g pg ht ass k.
    cut (check_dist (S Diam) g).
    { intros ok.
      destruct (le_gt_dec k (S Diam)).
      eapply check_dist_incr; eauto.
      apply check_dist_ge_Diam; auto; lia.
    }
    { intros p; split;
        assert (dd := diam_dist p Root); intros h; [|lia].
      eapply d_is_dist; auto.
      Unshelve. auto.
    }
  Qed.
  
  (***** Induction : base case (1) 
   - for Root (1a)
   - for other nodes (1b)
   *)

  Lemma Root_remains_0:
    forall e, is_exec e -> Assume (s_hd e) ->
              d (s_hd e Root) = 0 ->
              Always (fun c => d (s_hd c Root) = 0) e.
  Proof.
    intros e ise ass hd.
    assert (X: CD1 1 Root (s_hd e)).
    { intros _; rewrite dist_nil; auto. }
    eapply CD1_exec_Always with (k := 0) (p := Root) in X; auto.
    2: apply check_dist_0.
    revert X; apply Always_compat; auto.
    2: apply is_exec_proper in ise; auto.
    { intros x y xy; unfold CD1.
      rewrite dist_nil, xy.
      split; intros h; auto.
    }
  Qed.
  
  Lemma Root_act_neut:
    forall g (e: Exec),
      Assume g -> Step (s_hd e) g ->
      act_neut_b (s_cons g e) Root = true->
      d (s_hd e Root) = 0.
  Proof.
    intros g e ass st h.
    unfold act_neut_b, AN_b in h.
    apply orb_prop in h; destruct h as [act|neut].
    - unfold ACT_b in act; simpl in act.
      apply step_Root in act; auto.
    - simpl in neut; unfold NEUT_b.
      apply andb_prop in neut; destruct neut as [_ neut].
      apply andb_prop in neut; destruct neut as [_ neut].
      rewrite negb_true_iff in neut.
      apply disabled_root; auto.
      { change (Assume (s_hd e)).
        apply step_Assume with g; auto. }
  Qed.

  Lemma BFS_round_base_case_1a:
    forall e (ise: is_exec e) (ass: Assume (s_hd e)),
      at_most_rounds (Always
                         (fun c: Exec => d (s_hd c Root) = 0)) 1 e.
  Proof.
    intros e ise ass.

    apply (at_most_rounds_scheme_per_node
             ise ass
             (fun p e => Always (fun e => d (s_hd e p) = 0) e)).
    { (* cnd_here *)
      case (Nat.eq_dec (d (s_hd e Root)) 0); intros hn.
      { intros hen; apply Root_remains_0; auto. }
      { intros hen; apply  disabled_root in hen; tauto. }
    }
    { (* cnd_act *)
      revert ise ass; revert e; cofix h; intros [g | g e] ise ass.
      left; tauto.
      right.
      { intros hact.
        apply Root_remains_0; auto.
        inversion ise; auto.
        { eapply step_Assume.
          apply ass.
          apply is_exec_inv_1 in ise; auto. }
        apply Root_act_neut with g; auto.
        { apply is_exec_inv_1 in ise; auto. }
      }
      { apply h.
        inversion ise; auto.
        { eapply step_Assume.
          apply ass.
          apply is_exec_inv_1 in ise; auto. }
      }
    }
  Qed.
  
  Lemma non_Root_positive:
    forall e,
      is_exec e -> Assume (s_hd e) -> 
      forall p, ~ p == Root -> d (s_hd e p) > 0 ->
                Always (fun c => d (s_hd c p) >= 1) e.
  Proof.
    intros e ise ass p hnr hpos.
    assert (cd2: CD2 1 p (s_hd e)) by (intros _; lia).
    eapply CD2_exec_Always with (k := 0) (p := p) in cd2; auto.
    unfold CD2 in cd2; revert cd2; apply Always_compat.
    { intros x y xy; rewrite xy; split; intros h; auto.
      cut (d (s_hd y p) >= 1); [intros hh; auto|].
      apply h.
      rewrite <- dist_nil' in hnr; auto; lia. }
    { apply is_exec_proper in ise; auto. }
    { intros q; split; intros h; lia. }
  Qed.

  Lemma non_Root_d0_enabled:
    forall (g: Env) p, Proper pequiv g -> Assume g ->
                       ~ p == Root -> d (g p) = 0 ->
                       enabled_b g p = true.
  Proof.
    intros g p pg ass hnr hd.
    eapply enabled_non_root_CD1; auto.
    rewrite hd.
    intros F. 
    remember (dist_Neigh g p) as l.
    clear - F.
    induction l as [| a l]; simpl in F.
    rewrite InA_nil in F; tauto.
    rewrite InA_cons in F; destruct F as [hx|hx].
    lia.
    auto.
    Unshelve. apply Graph_network_assumption.
  Qed.

  Lemma non_Root_act:
    forall g (e: Exec) p,
      Assume g -> Step (s_hd e) g ->
      ~ p == Root -> ACT_b p (s_cons g e) = true->
      d (s_hd e p) = Dist_p (map S (dist_Neigh g p)).
  Proof.
    intros g e p ass st hnr act.
    unfold ACT_b in act; simpl in act.
    apply has_moved_true in act; auto.
    rewrite act.
    eapply RUN_CD0; eauto.
    apply Step_proper_r in st; auto.
    Unshelve. apply Graph_network_assumption.
  Qed.

  Lemma non_Root_neut:
    forall g (e: Exec) p,
      Assume g -> Step (s_hd e) g ->
      ~ p == Root -> NEUT_b p (s_cons g e) = true->
      d (s_hd e p) = Dist_p (map S (dist_Neigh (s_hd e) p)).
  Proof.
    intros g e p ass st hnr neut.
    simpl in neut; unfold NEUT_b.
    apply andb_prop in neut; destruct neut as [_ neut].
    apply andb_prop in neut; destruct neut as [_ neut].
    rewrite negb_true_iff in neut.
    eapply disabled_non_root' in neut; eauto.
    destruct neut as [neut _]; auto.
    { change (Assume (s_hd e)).
      apply step_Assume with g; auto. }
    { apply Step_Proper_l in st; auto. } 
    Unshelve. apply Graph_network_assumption.
  Qed.

  Lemma non_Root_act_neut_ge1:
    forall g (e: Exec) p,
      Assume g -> Step (s_hd e) g ->
      ~ p == Root -> act_neut_b (s_cons g e) p = true->
      d (s_hd e p) >= 1.
  Proof.
    intros g e p ass st hnr an.
    assert (hip: exists l, l <> nil /\ d (s_hd e p) = Dist_p (map S l)).
    { unfold act_neut_b, AN_b in an.
      apply orb_prop in an; destruct an as [act|neut].
      - apply non_Root_act in act; auto.
        exists (dist_Neigh g p); auto.
        split; [apply technical_not_nil0|]; auto.
      - apply non_Root_neut in neut; auto.
        exists (dist_Neigh (s_hd e) p).
        split; [apply technical_not_nil0|]; auto.
    }
    destruct hip as [l [hn hl]]; rewrite hl.
    apply Dist_p_ge.
    - contradict hn; eapply map_eq_nil; eauto.
    - intros x hx.
      rewrite InA_map_iff in hx.
      destruct hx as [a [ha _]]; simpl in ha; lia.
      apply NPeano.Nat.succ_wd.
  Qed.

  Lemma BFS_round_base_case_1b:
    forall e (ise: is_exec e) (ass: Assume (s_hd e)),
    forall p,
      ~ p == Root ->
      at_most_rounds (Always (fun c: Exec => d (s_hd c p) >= 1)) 1 e.
  Proof.
    intros e ise ass p hnp.

    apply (at_most_rounds_scheme_per_node
             ise ass
             (fun p e => Always (fun e => d (s_hd e p) >= 1) e)).
    { (* cnd_here *)
      intros hen.
      case (Nat.eq_dec (d (s_hd e p)) 0); intros h0.
      - apply non_Root_d0_enabled with (g := s_hd e) in hnp; auto.
        rewrite hnp in hen; congruence.
        apply Stream.hd_compat; apply is_exec_proper in ise; auto.
      - apply non_Root_positive; auto.
        lia.
    }
    { (* cnd_act *)
      revert ise ass; revert e; cofix h; intros [g | g e] ise ass.
      left; tauto.
      right.
      { intros hact.
        apply non_Root_positive; auto.
        inversion ise; auto.
        { eapply step_Assume.
          apply ass.
          apply is_exec_inv_1 in ise; auto. }
        apply non_Root_act_neut_ge1 with g; auto.
        { apply is_exec_inv_1 in ise; auto. }
      }
      { apply h.
        inversion ise; auto.
        { eapply step_Assume.
          apply ass.
          apply is_exec_inv_1 in ise; auto. }
      }
    }
  Qed.
    
  Lemma BFS_round_base_case_1:
    forall e (ise: is_exec e) (ass: Assume (s_hd e)),
      at_most_rounds (Always (fun c => check_dist 1 (s_hd c))) 1 e.
  Proof.
    intros e ise ass.
    apply at_most_rounds_forall'.
    { intros p q pq x y xy; apply CD_compat; auto.
      apply Stream.hd_compat; auto. }
    { apply is_exec_proper in ise; auto. }
    intros p.
    case (Node_dec p Root); intros hr.
    {
      assert (X := BFS_round_base_case_1a ise ass).
      revert X; apply at_most_rounds_compat; auto.
      2: apply is_exec_proper in ise; auto.
      apply Always_compat.
      intros x y xy.
      rewrite xy.
      assert (py: Proper pequiv y) by proper_.
      unfold CD, CD1, CD2; rewrite hr, dist_nil; split; intros h.
      apply h; lia.
      split; intros hh; auto.
      lia.
    }
    {
      assert (X := BFS_round_base_case_1b ise ass _ hr).      
      revert X; apply at_most_rounds_compat; auto.
      2: apply is_exec_proper in ise; auto.
      apply Always_compat.
      intros x y xy.
      rewrite xy.
      assert (py: Proper pequiv y) by proper_.
      unfold CD, CD1, CD2; split; intros h.
      apply h.
      case (Nat.eq_dec (dist p Root) 0); intros hn.
      contradict hr; apply dist_nil'; auto.
      lia.      
      split; intros hh; auto.
      assert (dd: dist p Root = 0) by lia.
      contradict hr; apply dist_nil'; auto.
    }
  Qed.      
  
  (***** Induction : induction case (k -> k+1) 
   - for nodes at length <= k from the root (ka)
   - for other nodes (kb)
   *)
  
  Lemma node_remains_k:
    forall e,
      is_exec e -> Assume (s_hd e) ->
      forall k p,
        k >= 1 -> check_dist k (s_hd e) ->
        dist p Root = k -> d (s_hd e p) = k ->
        Always (fun c => d (s_hd c p) = k) e.
  Proof.
    intros e ise ass k p hk hcd hd hdd.
    assert (XX: CD1 (S k) p (s_hd e)).
    { intros h; lia. }
    assert (X := CD1_exec_Always ass ise hcd XX).
    revert X; apply Always_compat.
    intros x y xy; rewrite xy.
    split; intros h.
    intros hh; subst; auto.
    unfold CD1 in h; subst; auto.
    apply is_exec_proper in ise; auto.
  Qed.

  Lemma not_k_enabled:
    forall e,
      is_exec e -> Assume (s_hd e) ->
      forall k p, k >= 1 -> check_dist k (s_hd e) ->
                  dist p Root = k -> d (s_hd e p) <> k ->
                  enabled_b (s_hd e) p = true.
  Proof.
    intros e ise ass k p hk hc hdst hd.
    case_eq (enabled_b (s_hd e) p); intros hen; auto.
    contradict hd.

    destruct k as [| k]; [lia|].
    destruct (dist_neighbor_ex _ _ hdst) as [q [pq hq]].
    assert (hqq: d (s_hd e q) = k).
    { specialize (hc q); destruct hc as [hc _].
      rewrite hc; [|lia]; auto. }
    
    eapply disabled_non_root in hen; auto.
    2: { intros F; rewrite F, dist_nil in hdst; lia. }
    2: { apply is_exec_proper, Stream.hd_compat in ise; auto. }
    destruct hen as [h1 [h2 h3]].
    pose (PAR := Par (s_hd e) p).
    fold PAR in h2, h3.
    specialize (h3 _ pq).
    rewrite hqq in h3; clear pq hq hqq q.
    destruct (dist_neighbors p PAR) as [hpar|[hpar|hpar]]; auto;
      rewrite hdst in hpar.
    - specialize (hc PAR); destruct hc as [_ hc2].
      assert (dum: dist PAR Root >= S k) by lia.
      specialize (hc2 dum); lia.
    - specialize (hc PAR); destruct hc as [hc1 _].
      assert (dum: dist PAR Root < S k) by lia.
      specialize (hc1 dum); lia.
    - specialize (hc PAR); destruct hc as [_ hc2].
      assert (dum: dist PAR Root >= S k) by lia.
      specialize (hc2 dum); lia.
      Unshelve. apply Graph_network_assumption.
  Qed.

  Lemma act_neut_k_aux:
    forall g p, Proper pequiv g ->
                ~ p == Root ->
                check_dist (dist p Root) g ->
                dist p Root = Dist_p (map S (dist_Neigh g p)).
  Proof.
    intros g p pg hnr hc.
    apply Dist_p_spec'; [apply technical_not_nil|split]; auto.
    - rewrite Sdist_Neigh_spec; auto.
      remember (dist p Root) as k; symmetry in Heqk.
      destruct k as [| k].
      { rewrite dist_nil' in Heqk; tauto. }
      destruct (dist_neighbor_ex _ _ Heqk) as [q [pq hq]].
      exists q; split; auto.
      apply f_equal.
      rewrite <- hq.
      specialize (hc q); destruct hc as [hc _]; apply hc; lia.
    - intros x h.
      rewrite Sdist_Neigh_spec in h; auto.
      destruct h as [q [pq hq]]; rewrite <- hq; clear hq x.
      destruct (dist_neighbors p q) as [hq|[hq|hq]]; auto.
      + specialize (hc q); destruct hc  as [_ hc].
        unfold CD2 in hc; lia.
      + specialize (hc q); destruct hc  as [hc _].
        unfold CD1 in hc; lia.
      + specialize (hc q); destruct hc  as [_ hc].
        unfold CD2 in hc; lia.
  Qed.
  
  Lemma act_neut_k:
    forall g e p, Assume g -> Step (s_hd e) g ->
                  ~ p == Root ->
                  check_dist (dist p Root) g ->
                  act_neut_b (s_cons g e) p = true ->
                  d (s_hd e p) = dist p Root.
  Proof.
    intros g e p ass st hnr hc han.
    unfold act_neut_b, AN_b in han.
    apply orb_prop in han; destruct han as [act|neut].
    { apply non_Root_act in act; auto.
      rewrite act; symmetry; apply act_neut_k_aux; auto.
      apply Step_proper_r in st; auto. }
    { apply non_Root_neut in neut; auto.
      rewrite neut; symmetry; apply act_neut_k_aux; auto.
      apply Step_proper_l in st; auto.
      apply check_dist_step with g; auto.
    }
  Qed.
  
  Lemma BFS_round_induction_case_ka:
    forall e (ise: is_exec e) (ass: Assume (s_hd e)) k,
      k >= 1 ->
      check_dist k (s_hd e) ->
      forall p, dist p Root < S k ->
                at_most_rounds
                  (Always (fun c: Exec => d (s_hd c p) = dist p Root)) 1 e.
  Proof.
    intros e ise ass k hk hcd p hdp.

    case (le_gt_dec k (dist p Root)); intros hd.
    2: { apply check_dist_exec_Always in hcd; auto.
         apply rnd_here.
         revert hcd; apply Always_impl_P.
         intros x hx.
         apply hx; auto. }

    assert (hdd: dist p Root = k) by lia.
    clear hdp hd.
    
    apply (at_most_rounds_scheme_per_node
             ise ass
             (fun p e => Always (fun e => d (s_hd e p) = dist p Root) e)).

    { (* cnd_here *)
      intros hen.
      case (Nat.eq_dec (d (s_hd e p)) k); intros hdk.
      { rewrite hdd.
        apply node_remains_k; auto.
      }
      { apply not_k_enabled in hdk; auto.
        rewrite hen in hdk; congruence.
      }
    }

    { (* cnd_act *)
      revert ise ass hcd; revert e; cofix h.
      intros [g | g e] ise ass hcd.
      left; tauto.
      right.
      { intros hact.
        apply node_remains_k; auto.
        inversion ise; auto.
        { eapply step_Assume.
          apply ass.
          apply is_exec_inv_1 in ise; auto. }
        lia.
        rewrite hdd; apply check_dist_step with g; auto.
        apply is_exec_inv_1 in ise; auto.
        apply act_neut_k with g; auto.
        apply is_exec_inv_1 in ise; auto.
        { intros F; rewrite F, dist_nil in hdd; lia. }
        rewrite hdd; auto.
      }
      { apply h.
        inversion ise; auto.
        { eapply step_Assume.
          apply ass.
          apply is_exec_inv_1 in ise; auto. }
        apply check_dist_step with g; auto.
        apply is_exec_inv_1 in ise; auto.
      }
    }
  Qed.

  Lemma node_remains_ge_Sk:
    forall e,
      is_exec e -> Assume (s_hd e) ->
      forall k p, k >= 1 -> check_dist k (s_hd e) ->
                  dist p Root >= S k -> d (s_hd e p) >= S k ->
                  Always (fun c => d (s_hd c p) >= S k) e.
  Proof.
    intros e ise ass k p hk hc hnr hd.
    destruct k as [| k]; [lia|].
    assert (X: CD2 (S (S k)) p (s_hd e)) by (unfold CD2; auto).
    apply CD2_exec_Always in X; auto.
    revert X; apply Always_impl_P.
    intros x; unfold CD2; auto.
  Qed.

  Lemma lt_Sk_enabled:
    forall e,
      is_exec e -> Assume (s_hd e) ->
      forall k p, k >= 1 -> check_dist k (s_hd e) ->
                  dist p Root >= S k -> d (s_hd e p) < S k ->
                  enabled_b (s_hd e) p = true.
  Proof.
    intros e ise ass k p hk hc hdst hd.
    eapply enabled_non_root_CD1; auto.
    { intros F; rewrite F, dist_nil in hdst; lia. }
    { apply is_exec_proper, Stream.hd_compat in ise; auto. }
    intros ina.
    apply Sdist_Neigh_spec in ina.
    2: { apply is_exec_proper, Stream.hd_compat in ise; auto. }
    destruct ina as [q [pq hq]].
    cut (d (s_hd e q) >= k); [lia|].
    specialize (hc q); destruct hc as [_ hc]; apply hc.
    destruct (dist_neighbors p q) as [hqq|[hqq|hqq]]; auto;
      lia.
    Unshelve. apply Graph_network_assumption.
  Qed.

  Lemma act_neut_ge_Sk_aux:
    forall g k p, Proper pequiv g ->
                  dist p Root >= S k ->
                  check_dist k g ->
                  Dist_p (map S (dist_Neigh g p)) >= S k.
  Proof.
    intros g k p pg hdst hc.
    apply Dist_p_ge; [apply technical_not_nil; auto|].
    { intros F; apply dist_nil' in F; lia. }
    { intros x h.
      apply Sdist_Neigh_spec in h; auto.
      destruct h as [q [pq hq]].
      rewrite <- hq.
      cut (d (g q) >= k); [lia|].
      specialize (hc q); destruct hc as [_ hc]; apply hc.
      destruct (dist_neighbors p q) as [hqq|[hqq|hqq]]; auto;
        lia.
    }
  Qed.

  Lemma act_neut_ge_Sk:
    forall g e k p, Assume g -> Step (s_hd e) g ->
                    dist p Root >= S k ->
                    check_dist k g ->
                    act_neut_b (s_cons g e) p = true ->
                    d (s_hd e p) >= S k.
  Proof.
    intros g e k p ass st hdst hc han.
    unfold act_neut_b, AN_b in han.
    apply orb_prop in han; destruct han as [act|neut].
    { apply non_Root_act in act; auto.
      2: { intros F; rewrite F, dist_nil in hdst; lia. }
      rewrite act.
      apply act_neut_ge_Sk_aux; auto.
      { apply Step_proper_r in st; auto. }
    }
    { apply non_Root_neut in neut; auto.
      2: { intros F; rewrite F, dist_nil in hdst; lia. }
      rewrite neut.
      apply act_neut_ge_Sk_aux; auto.
      { apply Step_proper_l in st; auto. }
      { eapply check_dist_step; eauto. }
    }
  Qed.

  Lemma BFS_round_induction_case_kb:
    forall e (ise: is_exec e) (ass: Assume (s_hd e)) k,
      k >= 1 ->
      check_dist k (s_hd e) ->
      forall p, dist p Root >= S k ->
                at_most_rounds (Always (fun c => CD2 (S k) p (s_hd c))) 1 e.
  Proof.
    intros e ise ass k hk hcd p hp.

    apply (at_most_rounds_scheme_per_node
             ise ass
             (fun p e => Always (fun e => CD2 (S k) p (s_hd e)) e)).

    { (* cnd_here *)
      intros hen.
      case (le_gt_dec (S k) (d (s_hd e p))); intros hdk.
      {
        apply node_remains_ge_Sk with (p := p) in hcd; auto.
        revert hcd; apply Always_impl_P.
        intros x; unfold CD2; auto.
      }
      { apply lt_Sk_enabled with (p := p) in hcd; auto.
        congruence.
      }
    }
    { (* cnd_act *)
      revert hcd ise ass; revert e; cofix h;
        intros [g | g e] hcd ise ass.
      left; tauto.
      right.
      { intros hen.
        cut (Always (fun c => d (s_hd c p) >= S k) e).
        { apply Always_impl_P; intros x; unfold CD2; auto. }
        apply node_remains_ge_Sk; auto.
        inversion ise; auto.
        { eapply step_Assume.
          apply ass.
          apply is_exec_inv_1 in ise; auto. }
        apply check_dist_step with g; auto.
        apply is_exec_inv_1 in ise; auto.
        apply act_neut_ge_Sk with g; auto.
        apply is_exec_inv_1 in ise; auto.
      }
      { apply h.
        apply check_dist_step with g; auto.
        apply is_exec_inv_1 in ise; auto.
        inversion ise; auto.
        { eapply step_Assume.
          apply ass.
          apply is_exec_inv_1 in ise; auto. }
      }        
    }
  Qed.
  
  Lemma BFS_round_induction_case_k:
    forall e (ise: is_exec e) (ass: Assume (s_hd e)) k,
      k >= 1 ->
      check_dist k (s_hd e) ->
      at_most_rounds (Always (fun c => check_dist (S k) (s_hd c))) 1 e.
  Proof.
    intros e ise ass k hk hcd.
    unfold check_dist.
    apply at_most_rounds_forall'; auto.
    { intros p q pq x y xy; apply CD_compat; auto.
      apply Stream.hd_compat; auto. }
    apply is_exec_proper in ise; auto.
    intros p.
    case (le_gt_dec (S k) (dist p Root)); intros hp.
    { assert (hp': dist p Root >= S k) by lia.
      assert (X := BFS_round_induction_case_kb ise ass hk hcd _ hp').
      revert X; apply at_most_rounds_compat; auto.
      2: apply is_exec_proper in ise; auto.
      apply Always_compat.
      intros x y xy; unfold CD, CD1, CD2.
      rewrite xy.
      split.
      intros [h hh]; auto.
      intros h; split; intros hh; lia.
    }
    { assert (hp': dist p Root < S k) by lia.
      assert (X := BFS_round_induction_case_ka ise ass hk hcd _ hp').
      revert X; apply at_most_rounds_compat; auto.
      2: apply is_exec_proper in ise; auto.
      apply Always_compat.
      intros x y xy; unfold CD, CD1, CD2.
      rewrite xy.
      split.
      intros [h hh]; auto.      
      intros h; lia. 
    }
  Qed.
  
  Lemma not_check_dist_en:
    forall k g, Assume g -> Proper pequiv g ->
                ~ check_dist k g -> ~ terminal g.
  Proof.
    intros k g ass pg h; contradict h.
    apply check_dist_terminal; auto; tauto.
  Qed.

  Lemma BFS_round_ind_ccl:
    forall e,
      is_exec e -> Assume (s_hd e) ->
      at_most_rounds
        (fun c: Exec => forall p, d (s_hd c p) = dist p Root) (Diam + 1) e.
  Proof.
    intros e ise ass.
    cut (at_most_rounds (Always (fun c => check_dist (S Diam) (s_hd c))) (Diam + 1) e).
    { apply at_most_rounds_equiv_context
        with (fun e => is_exec e /\ Assume (s_hd e)).
      apply is_exec_proper in ise; auto.
      { apply Always_conj; split.
        apply is_exec_Always; auto.
        apply is_exec_Assume; auto. }
      revert ise ass; revert e; cofix h; intros [g | g e] ise ass.
      - left.
        intros ha.
        apply check_dist_Diam.
        apply Always_P_P in ha; auto.
      - right.
        + intros ha.
          apply check_dist_Diam.
          apply Always_P_P in ha; auto.
        + apply h; auto.
          apply is_exec_inv_2 in ise; auto.
          { eapply step_Assume in ass.
            apply ass.
            apply is_exec_inv_1 in ise; auto. }
    }          
    assert (dum: Diam + 1 = S Diam) by lia.
    rewrite dum; clear dum.
    apply schema_round_induction with 
        (P := fun n e => Always (fun c =>
                                   check_dist n (s_hd c)) e); auto.
    { intros i j ij x y xy; apply Always_compat; auto.
      intros z t zt; apply check_dist_compat; auto.
      apply Stream.hd_compat; auto.
    }
    { apply is_exec_proper in ise; auto. }
    { apply check_dist_exec_Always; auto.
      apply check_dist_0. }
    { revert ise ass; revert e; cofix h; intros [g | g e] ise ass.
      { left; intros k hk ha.
        apply rnd_here.
        apply check_dist_exec_Always; auto.
        apply check_dist_terminal; auto.
        apply is_exec_proper in ise; apply Stream.hd_compat; auto.
        apply is_exec_inv_0 in ise; auto.
      }
      { right.
        - intros k hk ha.
          destruct k as [|k].
          + apply BFS_round_base_case_1; auto.
          + apply BFS_round_induction_case_k; auto.
            lia.
            apply Always_P_P in ha; auto.
        - apply h; auto.
          apply is_exec_inv_2 in ise; auto.
          { eapply step_Assume in ass.
            apply ass.
            apply is_exec_inv_1 in ise; auto. }
      }          
    }
  Qed.
  
  (** Round Complexity: Action CD sets every distance to the root *)
  Theorem BFS_rounds_CD:
    Round_complexity (fun e => is_exec e /\ Assume (s_hd e))
                     (fun e =>
                        forall p, d (s_hd e p) = dist p Root
                     )
                     (Diam + 1).
  Proof.
    intros e [ise ass].
    apply BFS_round_ind_ccl; auto.
  Qed.

  Lemma check_dist_Diam_disabled_CD:
    forall g, Proper pequiv g -> Assume g ->
              check_dist (S Diam) g ->
              forall p,
                ~ p == Root ->
                d (g p) = Dist_p (map S (dist_Neigh g p)).
  Proof.
    intros g pg ass hc p hnr.
    apply Dist_p_spec'.
    apply technical_not_nil; auto.
    split.
    apply Sdist_Neigh_spec; auto.
    { remember (dist p Root) as n; symmetry in Heqn.
      destruct n as [| n].
      { rewrite dist_nil' in Heqn; tauto. }
      destruct (dist_neighbor_ex _ _ Heqn) as [q [hg hd]].
      exists q; split; auto.
      dup hc hcq.
      specialize (hcq q); destruct hcq as [hcq _].
      specialize (hc p); destruct hc as [hc _].
      rewrite hc, hcq; subst; auto.
      assert (tt := diam_dist q Root); lia. 
      assert (tt := diam_dist p Root); lia. 
    }
    { intros x hx.
      apply Sdist_Neigh_spec in hx; auto.
      destruct hx as [q [pq hq]].
      rewrite <- hq; clear hq x.
      dup hc hcq.
      specialize (hcq q); destruct hcq as [hcq _].
      specialize (hc p); destruct hc as [hc _].
      rewrite hc, hcq.
      2: assert (tt := diam_dist q Root); lia. 
      2: assert (tt := diam_dist p Root); lia.
      destruct (dist_neighbors _ _ pq); lia.
    }
  Qed.

  Lemma check_dist_Diam_disabled_CD':
    forall g, Proper pequiv g -> Assume g ->
              check_dist (S Diam) g ->
              forall p,
                ~ p == Root ->
                d (g p) = Dist_p (map S (dist_Neigh g p)).
  Proof.
    intros g pg ass hc p hnr.
    dup hc hp; specialize (hp p); destruct hp as [hp _].
    apply check_dist_Diam_disabled_CD with (p := p) in hc; auto.
  Qed.

  Lemma check_dist_Diam_d_unchanged:
    forall g, Proper pequiv g -> Assume g ->
              check_dist (S Diam) g ->
              forall p, d (g p) = d (RUN g p).
  Proof.
    intros g pg ass hc p.

    case (Node_dec p Root); intros hnr.

    { rewrite hnr, RUN_root; auto.
      specialize (hc Root); destruct hc as [hc _];
        rewrite hc; auto; rewrite dist_nil; lia.
    }
    { erewrite RUN_CD0; auto.
      apply check_dist_Diam_disabled_CD'; auto.
      Unshelve. apply Graph_network_assumption.
    }
  Qed.

  Lemma check_dist_Diam_d_unchanged':
    forall g' g, Step g' g -> Assume g ->
                 check_dist (S Diam) g ->
                 forall p, d (g p) = d (g' p).
  Proof.
    intros g' g st ass hc p.
    destruct (Step_correct st p); rewrite H; auto.
    apply check_dist_Diam_d_unchanged; auto.
    apply Step_proper_r in st; auto.
  Qed.

    Lemma disabled_non_Root:
    forall g,
      Proper pequiv g -> Assume g ->
      forall p,
        ~ p == Root ->
        G p (Par g p) ->
        d (g p) = S (d (g (Par g p))) ->
        (forall q : Node, G p q -> d (g p) <= S (d (g q)))
        -> enabled_b g p = false.
  Proof.
    intros g pg ass p hnr hG hds hf.
    case_eq (enabled_b g p); intros h; [|auto].
    eapply enabled_non_root' in h; auto.
    destruct h as [h|h].
    { contradict h; apply Dist_p_spec'; [apply technical_not_nil|]; auto.
      split.
      - apply Sdist_Neigh_spec; auto.
        exists (Par g p); split; auto.
      - intros x hx.
        apply Sdist_Neigh_spec in hx; auto.
        destruct hx as [q [pq hq]]; rewrite <- hq; auto.
    }
    { tauto. }
    Unshelve. apply Graph_network_assumption.    
  Qed.
  
  Lemma check_dist_Diam_Par_OK:
    forall g p, Assume g -> Proper pequiv g -> ~ p == Root ->
                check_dist (S Diam) g ->
                d (g p) = S (d (g (Par' (RUN g p) p))) /\
                G p (Par' (RUN g p) p).
  Proof.
    intros g p ass pg hnr hc.
    unfold Par, Par'.
    repeat rewrite RUN_unfold.
    repeat erewrite Dist_p1; auto.
    repeat erewrite Dist_p2; auto.
    repeat rewrite <- check_dist_Diam_disabled_CD'; auto.
    repeat rewrite <- check_dist_Diam_disabled_CD; auto.
    case_eq (root (g p)); intros hhr.
    { contradict hnr; destruct ass as [_ ass]; rewrite <- ass;
        auto. }
    clear hhr.
    destruct (Nat.eq_dec (d (g p)) (d (g p))); [|lia]; clear e.
    destruct (Nat.eq_dec (d (g p))
                         (S
                            match local_env g p (par (g p)) with
                            | Some s => d s
                            | None => d (g p)
                            end)); simpl.
    {  
      unfold local_env in e.   
      case_eq (peer p (par (g p))).
      2: { intros f; rewrite f in e; simpl in e; lia. }
      intros q hq; rewrite hq in e; simpl in e.
      assert (Q: Par_ (par (g p)) p (neighbors (g p)) == q).
      { apply Par__chan_conv; auto.
        destruct ass as [ass _].
        unfold Network_assumption in ass.
        specialize (ass p (par (g p))); apply ass.
        apply peers_spec_peer; exists q.
        unfold is_channel; rewrite hq; auto.
        rewrite hq; auto. }
      rewrite Q.
      split; auto.
      apply Graph_network_assumption.
      exists (par (g p)); unfold is_channel; rewrite <- hq; auto.
    }
    { clear n.
      assert (T: exists nn,
                 InA equiv (Par_dist (d (g p)) (par (g p))
                                     (neighbor_list (g p) (local_env g p)), nn) (neighbor_list (g p) (local_env g p)) /\ S nn = (d (g p))).
      { apply Par_dist_prop.
        case_eq (dist p Root).
        { intros h; contradict hnr; apply dist_nil'; auto. }
        intros n hn; dup hn hq; apply dist_neighbor_ex in hq.
        destruct hq as [q [hq hdq]].
        specialize (Graph_network_assumption p q).
        rewrite Graph_network_assumption in hq.
        destruct hq as [c hq].
        exists c; exists n; split; [|].
        2: { rewrite check_dist_Diam in hc; rewrite hc; lia. }
        apply neighbor_list_prop; auto.
        exists q; split; auto.
        rewrite check_dist_Diam in hc; rewrite hc; lia.
      }
      destruct T as [n T].
      rewrite neighbor_list_prop in T; auto.
      destruct T as [[q [hq hn]] hnn].
      unfold is_channel in hq.
      assert (Q: Par_ (Par_dist (d (g p)) (par (g p))
                                (neighbor_list (g p) (local_env g p))) p (neighbors (g p)) == q).
      { apply Par__chan_conv; auto.
        destruct ass as [ass _].
        unfold Network_assumption in ass.
        specialize (ass p (Par_dist (d (g p)) (par (g p))
                                    (neighbor_list (g p) (local_env g p)))); apply ass.
        apply peers_spec_peer; exists q.
        unfold is_channel; rewrite hq; auto. }
      rewrite Q.
      split; [lia|].
      apply Graph_network_assumption.
      exists (Par_dist (d (g p)) (par (g p))
                       (neighbor_list (g p) (local_env g p))); auto.
    }
    Unshelve. apply Graph_network_assumption.
  Qed.
  
  Lemma check_dist_Diam_step_disabled:
    forall g e p, Assume g -> Step (s_hd e) g ->
                  check_dist (S Diam) g ->
                  (enabled_b g p = false \/
                   act_neut_b (s_cons g e) p = true) ->
                  enabled_b (s_hd e) p = false.
  Proof.
    intros g e p ass st hc hen.
    assert (pg: Proper pequiv g) by (apply Step_proper_r in st; auto).
    assert (pg': Proper pequiv (s_hd e)) by (apply Step_proper_l in st; auto).
    assert (ass': Assume (s_hd e)) by (apply step_Assume with g; auto).
    assert (hc' := check_dist_step ass st hc).

    case (Node_dec p Root); intros hnr.
    { rewrite hnr, disabled_root.
      rewrite check_dist_Diam in hc'; rewrite hc', dist_nil; auto.
      apply ass'.
    }

    assert (okpar: G p (Par (s_hd e) p) /\
                   d (s_hd e p) = S (d (s_hd e (Par (s_hd e) p)))).
    { destruct (Step_correct st p) as [S|S]; unfold Par.
      - rewrite S.
        dup hc hpar.
        apply check_dist_Diam_Par_OK with (p := p) in hpar; auto.
        split; [apply hpar|].
        rewrite <- check_dist_Diam_d_unchanged; auto.
        symmetry; rewrite <- check_dist_Diam_d_unchanged' with (g := g); auto;
          symmetry; apply hpar.
      - destruct hen as [hen|han].
        + apply enabled_b_false in hen.
          rewrite S, <- hen.
          dup hc hpar.
          apply check_dist_Diam_Par_OK with (p := p) in hpar; auto.
          split; [apply hpar|].
          rewrite <- check_dist_Diam_d_unchanged; auto.
          symmetry; rewrite <- check_dist_Diam_d_unchanged' with (g := g);
            auto; symmetry; apply hpar.
        + unfold act_neut_b, AN_b in han; apply orb_prop in han.
          destruct han as [han|han].
          * apply act_b_true in han; auto.
            rewrite <- han.
            dup hc hpar.
            apply check_dist_Diam_Par_OK with (p := p) in hpar; auto.
            split; [apply hpar|].
            rewrite <- check_dist_Diam_d_unchanged; auto.
            symmetry; rewrite <- check_dist_Diam_d_unchanged' with (g := g);
              auto; symmetry; apply hpar.
          * apply neut_b_true in han; auto.
            rewrite <- han.
            dup hc' hpar.
            apply check_dist_Diam_Par_OK with (p := p) in hpar; auto.
            split; [apply hpar|].
            rewrite <- check_dist_Diam_d_unchanged; auto.
            apply hpar.
    }
    
    apply disabled_non_Root; auto.

    { apply okpar. }
    { apply okpar. }
    { apply DistOK; auto.
      symmetry; apply check_dist_Diam_disabled_CD; auto. }
  Qed.

  Lemma node_remains_disabled:
    forall e,
      is_exec e -> Assume (s_hd e) -> 
      forall p, check_dist (S Diam) (s_hd e) ->
                enabled_b (s_hd e) p = false ->
                Always (fun c => enabled_b (s_hd c) p = false) e.
  Proof.
    cofix h; intros [g | g e] ise ass p hc hen.
    - left; auto.
    - assert (S: Step (s_hd e) g).
      { apply is_exec_inv_1 in ise; auto. }
      right; auto.
      apply h; auto.
      inversion ise; auto.
      apply step_Assume with g; auto. 
      apply check_dist_step with g; auto. 
      apply check_dist_Diam_step_disabled with g; auto.
  Qed.

  Lemma last_round_action_CP:
    forall e (ise: is_exec e) (ass: Assume (s_hd e)),
      (forall p, d ((s_hd e) p) = dist p Root) ->
      at_most_rounds (fun c => terminal (s_hd c)) 1 e.
  Proof.
    intros e ise ass hdst.
    apply check_dist_Diam in hdst.

    (*    *)
    cut (at_most_rounds
           (Always (fun c => forall p, enabled_b (s_hd c) p = false)) 1 e).
    { apply at_most_rounds_cov; auto.
      { apply is_exec_proper in ise; auto. }
      clear ass hdst.
      revert ise; revert e; cofix h; intros [g | g e] ise.
      left; apply is_exec_inv_0 in ise; auto.
      right.
      { clear h.
        intros h.
        apply Always_P_P in h.
        apply terminal_all_disabled_reverse.
        apply is_exec_proper in ise; inversion ise; auto.
        intros p F; specialize (h p); unfold enabled in F.
        congruence.
      }
      { apply h.
        inversion ise; auto. }
    }

    (*    *)
    apply at_most_rounds_forall'.
    { intros p q pq x y xy; apply eq_compat; auto.
      apply enabled_b_compat; auto; apply Stream.hd_compat; auto.
    }
    { apply is_exec_proper in ise; auto. }

    (*    *)
    intros p.
    apply (at_most_rounds_scheme_per_node
             ise ass
             (fun p e => Always (fun e => enabled_b (s_hd e) p = false) e)).
    { (* cnd_here *)
      intros hen.
      apply node_remains_disabled; auto. }
    { (* cnd_act *)
      revert ise ass hdst; revert e; cofix h;
        intros [g | g e] ise ass hdst.
      left; tauto.
      right.
      { intros hact.
        apply node_remains_disabled; auto.
        inversion ise; auto.
        { eapply step_Assume.
          apply ass.
          apply is_exec_inv_1 in ise; auto. }
        { apply check_dist_step with g; auto.
          apply is_exec_inv_1 in ise; auto. }
        { apply check_dist_Diam_step_disabled with g; auto.
          apply is_exec_inv_1 in ise; auto. }
      }        
      { apply h.
        inversion ise; auto.
        { eapply step_Assume.
          apply ass.
          apply is_exec_inv_1 in ise; auto. }
        { apply check_dist_step with g; auto.
          apply is_exec_inv_1 in ise; auto. }
      }        
    }
  Qed.
  
  Lemma Par_Rel_dec:
    forall g, forall x y, {Par_Rel g x y} + {~ Par_Rel g x y}.
  Proof.
    intros g x y; unfold Par_Rel.
    case (Node_dec x Root); intros hn; [right; auto|].
    case (Node_dec (Par g x) y); intros hp; [left|right]; auto.
  Qed.

  Lemma G_channel_dec:
    forall p q,
      {exists c, is_channel p c q} +
      {~ (exists c, is_channel p c q)}.
  Proof.
    intros p q; destruct (G_dec p q); [left|right].
    {  apply Graph_network_assumption in g; auto. }
    { contradict n.
      apply Graph_network_assumption; auto. }
  Qed.  
  
  (** Round Complexity: Final Theorem *)
  Lemma BFS_rounds_terminal:
    Round_complexity (fun e => is_exec e /\ Assume (s_hd e))
                     (fun e => terminal (s_hd e))
                     (Diam + 2).
  Proof.
    intros e [ise ass].
    assert (dum: Diam + 2 = S (Diam + 1)) by lia.
    rewrite dum; clear dum.
    eapply schema_round_step.
    5: { apply BFS_rounds_CD; auto. }
    { intros x y xy.
      apply all_compat; intros p.
      rewrite xy; tauto. }
    { intros x y xy; apply terminal_compat, Stream.hd_compat; auto. }
    { apply is_exec_proper in ise; auto. }

    (****************                                    *****)
    apply Always_forall' with (fun e => is_exec e /\ Assume (s_hd e)).
    intros ee [isee asss] h; apply last_round_action_CP; auto.
    apply Always_conj; split.
    apply is_exec_Always; auto.
    apply is_exec_Assume; auto.
  Qed.
  
  Theorem BFS_rounds:
    Round_complexity (fun e => is_exec e /\ Assume (s_hd e))
                     (fun e =>
                        (BFS_Spanning_Tree_spec (Net := Net))
                          Root Par_Rel (s_hd e))
                     (Diam + 2).
  Proof.
    eapply Round_complexity_compat_cov.
    9: apply BFS_rounds_terminal.
    auto.    
    { intros e f ef; apply terminal_compat, Stream.hd_compat, ef. }
    { intros e f ef; apply BFS_Spanning_Tree_spec_compat, Stream.hd_compat; auto.
      apply Par_Rel_compat.
    }
    { intros e f ef; rewrite <- ef; auto.  }
    { intros e f ef; rewrite <- ef; auto.  }
    { intros e f ef [ise ass]; rewrite <- ef; auto. }
    { cofix h.
      intros [g | g e] [ise ass].
      - left; intros ht.
        eapply BFS_correctness; auto.
        apply Stream.hd_compat; apply is_exec_proper in ise; auto.
        Unshelve. apply Graph_network_assumption.
      - right.
        intros ht.
        eapply BFS_correctness; auto.
        apply Stream.hd_compat; apply is_exec_proper in ise; auto.
        apply h.
        split.
        apply is_exec_inv_2 in ise; auto.
        apply step_Assume with g; auto.
        apply is_exec_inv_1 in ise; auto.
        Unshelve. apply Graph_network_assumption.
    }        
    { intros e [ise _]; apply is_exec_proper in ise; auto. }
  Qed.

End BFS_rounds.

Close Scope signature_scope.
Close Scope nat_scope.
Unset Implicit Arguments.
