(**
*Certified Round Complexity of Self-Stabilizing Algorithms*
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Lia.

(** **** Local Imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import Stream.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Exec.
From Padec Require Import Fairness.
From Padec Require Import Self_Stabilization.
From Padec Require Import BFS_algo_bis.
From Padec Require Import BFS_common_bis.
From Padec Require Import BFS_specification.
From Padec Require Import BFS_correctness_bis.
From Padec Require Import BFS_rounds_bis.
From Padec Require Import Rounds.

(* end show *)

Open Scope nat_scope.
Set Implicit Arguments.

(**
 * BFS: Proof of self-stabilization and specification
 **)

(**
* Specifications Proved

   /contains:/
   Proof that the algorithm is self-stabilizing wrt BFS 
   construction (the algorithm is silent).

   /assumes:/     
   G = (V, E) connected, rooted (r), semi-anonymous,
   Weakly fair daemon

   - Closure: silent algorithm

   - Convergence: The result is direct from complexity.

   - Specification: BFS construction
 **)

Section BFS_spec_proved.

  Context {Chans: Channels}
          {Net: Network}
          {Graph_assumption: BFS_Assume}
          {Graph_network_assumption: Graph_Network_matching}.

  Existing Instance G_compat.
  
  Notation Algo := (BFS_algo).
  Notation Env := (Env (Net := Net) (Algo := Algo)).

  Existing Instance BFS_RO_assumption.
  
  (** ** Convergence: eventually a legitimate configuration is 
    reached
    Obtained from round complexity *)
  Lemma BFS_convergence:
    convergence BFS_RO_assumption weakly_fair terminal.
  Proof.
    intros e ass ise wf.
    eapply at_most_rounds_weakly_fair; auto.
    apply is_exec_proper; auto.
    eapply BFS_rounds_terminal.
    split; auto.
    Unshelve. apply Graph_network_assumption.
  Qed.
  
  (** ** Specification *)
  Theorem BFS_self_stab_prop:
    self_stabilization
      BFS_RO_assumption weakly_fair
      (only_one (BFS_Spanning_Tree_spec Root Par_Rel)).
  Proof.
    apply silent_self_stabilization.
    apply convergence_silence.
    apply BFS_convergence.
    intros g ass ht.
    eapply BFS_correctness; auto.
    apply ht.
    Unshelve. apply Graph_network_assumption.
  Qed.

End BFS_spec_proved.

Unset Implicit Arguments.
Close Scope nat_scope.
