(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- March 2017
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import ZArith.
From Coq Require Import Lia.
From Coq Require Import Bool.

(** **** Local Imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import NatUtils.
From Padec Require Import ListUtils.
From Padec Require Import Count.
From Padec Require Import All_In.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Exec.
From Padec Require Import P_Q_Termination.
From Padec Require Import Fairness.
From Padec Require Import Self_Stabilization.
From Padec Require Import Ring_topology.
From Padec Require Import Dijkstra_algo.
From Padec Require Import Dijkstra_common.

(* end show *)

Open Scope nat_scope.
Set Implicit Arguments.

(**
 * Diskstra's first token ring algorithm
  - Proof of self-stabilization and specification
  - Proof of a complexity bound in steps
  - Proof that this complexity bound is exact (by building a worst case
    prefix of execution)
 **)

(**
* Potential Z

   /contains:/
   Definitions and results about Z and convexity

   /assumes:/ Parameter k>1; the network is a unidirectional rooted
   ring; assumption: N <= k

   - convex: A configuration is said convex w.r.t. to a given value vr if
     vr is the value of Root and all the nodes holding this value vr
     shape a convex sequence of vr-valued nodes made of consecutive
     successors of the root.  Convex for a value [vr] is equivalent to
     - [vr] is the value of the root and it is actually convex
     - otherwise, no node holds the value [vr]
     There always exists a value that satisfies convex (This is due to
     the fact that N <= k). 

   - Z: The potential Z of a configuration g is 
        - 0 if g is convex for the value of the Root
        - the distance to the closest absent value from the Root (modulo k),
          otherwise
     i.e. Z g = min { z < k | convex g ((v g Root + z) mod k) }

   - When Z is null: 'Z g = 0' is equivalent to 'g is convex for the
     root value'; and a configuration where Z is 0 and where the root
     holds a token is legitimate. 

   - Z evolution during a step from an illegitimate configuration:
     - Z decreases when the root executes 
     - Z does not increase.

   - Bound on Z: Z is upper bounded by N-1. If the configuration is
     convex, the result is trivial.  We consider a non-convex
     configuration. We use counting arguments which stands on Lemma
     all_here (that asserts that every value between 0 and Z-1 is
     represented in the ring).
   
   - When Z g is N - 1: g is of a particular shape. There exists a
     non-root node, which is not the successor of the root, that holds
     the root value; moreover, every value from (v g Root + 1) mod k
     to (v g Root + N-2) mod k is held by a single node.  From this
     observation, easily follows that Z g' <= N - 2.

 **)

Section Dijkstra_Z.

  Context {Chans: Channels}
          {Net: Network}
          {RURN: Rooted_Unidirectional_Ring_Network Net}
          {DP: Dijkstra_parameters}.

  Notation Algo := Dijkstra_algo.
  Notation Root := (RURN_root (Net := Net)).
  Notation Ring := (URN_topo (Net := Net)).
  Notation Env := (Env (Net := Net) (Algo := Dijkstra_algo)).
  Let H_k := H_k.

  Existing Instance Dijkstra_RO_assumption.

  (** This part of the proof requires the additional property that
   the number of nodes in the ring is smaller than or equal to k.
   *)
  Notation N := (Ring_size Root).
  Hypothesis HNk: N <= k.
  Hypothesis HNk1: 1 < N.

  (** ** Convexity 
     A configuration is said convex wrt to a given value [vr] if 
     - [vr] is the value of Root and 
     - all the nodes holding this value [vr] shape a convex sequence of [vr] 
       valued nodes made of consecutive successors of the root.

     For example, 
           - Root->5 5 5 6 3 4 0 is convex for vr = 5
           - Root->5 5 5 6 3 4 0 is not convex for any vr <> 5
           - Root->5 5 5 6 3 4 5 is not convex for any vr
           - Root->5 5 6 2 5 3 0 is not convex for any vr
   *)
  Definition convex (g: Env) (vr: nat): Prop :=
    forall n n', Between Root n' n -> v (g n) = vr -> v (g n') = vr.

  (** The convex predicate is weakly-decidable *)
  Lemma convex_dec: forall (g: Env) (vr: nat),
      {Proper fequiv g -> convex g vr} + {~convex g vr}.
  Proof.
    intros g vr.
    pose (P n n':= Between Root n' n ->  v (g n) = vr -> v (g n') = vr).
    change ({Proper fequiv g -> forall n n', P n n'} +
            {~ forall n n', P n n'}).
    assert (Pdec: forall n n', {P n n'} + {~P n n'}).
    {
      intros n n'.
      destruct (PeanoNat.Nat.eq_dec (Dist Root n' + Dist n' n) (Dist Root n))
        as [B|NB].
      2: unfold P, Between; tauto.
      destruct (PeanoNat.Nat.eq_dec (v (g n)) vr).
      2: unfold P; tauto.
      destruct (PeanoNat.Nat.eq_dec (v (g n')) vr).
      2: unfold P; tauto.
      unfold P; tauto.
    }
    assert (h := All_In_dec_forall_notforall_2
                   all_nodes_prop all_nodes_prop P Pdec).
    assert (hh: Proper fequiv g -> Proper fequiv P).
    {
      intros pg n1 n2 en n1' n2' en'.
      repeat apply impl_compat.
      apply Between_compat; reflexivity || auto.
      apply eq_compat;reflexivity ||
                                  apply v_compat, pg, en.
      apply eq_compat; reflexivity ||
                                   apply v_compat, pg, en'.
    }
    tauto.
  Qed.

  (** Boolean getters for convex *)
  Definition convex_b (g: Env) (vr: nat): bool :=
    match convex_dec g vr with left _ => true | right _ => false end.
  
  Lemma convex_b_true:
    forall g (pg: Proper fequiv g) V, convex g V <-> convex_b g V = true.
  Proof.
    intros g pg V; unfold convex_b;
      destruct (convex_dec g V) as [C|C]; split;
        try tauto; congruence.
  Qed.
  
  Lemma convex_b_false:
    forall g (pg: Proper fequiv g) V, ~convex g V <-> convex_b g V = false.
  Proof.
    intros g pg V; unfold convex_b;
      destruct (convex_dec g V) as [C|C]; split;
        try tauto; congruence.
  Qed.

  (** Convex for a value [vr] is equivalent to 
      - [vr] is the value of the root and it is actually convex
      - otherwise, no node holds the value [vr] 
   *)
  Lemma convex_equiv:
    forall (g: Env) (pg: Proper fequiv g) (V: nat),
      convex g V <->
      V = v (g Root) /\ convex g (v (g Root)) \/
                        forall n, V <> v (g n).
  Proof.
    intros g pg V; split; [intros C | intros [[VR C] | VN]].
    { destruct (PeanoNat.Nat.eq_dec V (v (g Root))) as [VR | VNR].
      - left; rewrite <- VR; split; auto.
      - right; intros n F; contradict VNR.
        specialize (C n Root).
        symmetry; apply C.
        apply Between_self_l.
        symmetry; auto. }
    { rewrite VR; auto. }
    { intros n n' B Vn; specialize (VN n); contradict VN; auto. }
  Qed.

  (** There always exists a value that satisfies convex.

    This is due to the fact that N <= k.

    Here follows the outline of the proof.

    If the value of the root is convex, then we are done.
    Otherwise, the value we are looking for is not in the ring 
    (see Lemma [convex_equiv]).

    We now focus on the case when the root value is not convex.
    We first prove that the number of values in the ring is smaller
    or equal to N (the ring size), using the Library [Count].
        [present x := exists n, v (g n) = x.
         Num_Card Smaller (@eqSig nat eq present) N]
   In particular, we exhibit the following injection relation:
   -  a present value [x] corresponds to a natural value (less than N)
      [j] when [x] is the value of the node located at [j] successor
      edges from the root.

   Let nV be the number values present in the ring: 
       [Num_Card Same (eqSig present) nV]
   In particular, we have [nV <= N <= k] (using transitivity of Card).

   Let nP be the number of values that satisfy convex (i.e. absent from 
   the ring).
       [P := fun V : nat => V < k /\ convex g V
        Num_Card Same (eqSig P) nP]
   We show that [nV + nP = k] using (disjoint union of Card).

   Now comes the proof: it is split in two cases, depended if nV is
   less than k or equal to k.

   - Case nV < k. We show that [nP >0] and then pick an element [V] 
     from P (i.e., such that V < k /\ convex g V].

   - Case nV = k. Then the number of present is exactly N, the number
     of nodes in the ring: [Card Same (@eqSig _ eq present) eqN]
     From this hypothesis, we can show that [convex g (v (g Root))], 
     a contradiction.

     Indeed, if the root is the unique node to hold this value, we 
     are done. Otherwise, we count the number of values present in 
     the ring, except the one of the root, using: 
        [present_outside_Root := 
           fun x => exists n : Node, ~eqN n Root /\ v (g n) = x]
     and obtain that: it is less than N and also equal to the number
     of values present in the ring, that is N, a contradiction.
   **)
  Lemma convex_exists:
    forall g, Proper fequiv g -> Assume g ->
              exists V, V < k /\ convex g V.
  Proof.
    intros g pg assume.
    destruct (convex_dec g (v (g Root))) as [YES|NO].
    {
      specialize (YES pg).
      split with (v (g Root)).
      split;auto using v_ltk.
    }
    assert (hN: Num_Card Same Node N)
      by (apply Ring_elements_card; auto using eqN_equiv).
    pose (present x := exists n, v (g n) = x).
    assert (h: Num_Card Smaller (sig present) N).
    { split with (fun x j =>
                    proj1_sig x =
                    v (g (power Succ (proj1_sig j) Root))); hnf.
      * intros   [x1 hx1] [x2 hx2] ex [j1 hj1] [j2 hj2] ej.
        apply eq_compat.
        apply ex.
        apply v_compat, pg, power_compat.
        apply Succ_compat.
        apply ej.
        reflexivity.
      * intros [x1 hx1] [j1 hj1] h1 [x2 hx2] [j2 hj2] h2 ej.
        hnf; rewrite h1, h2.
        apply v_compat, pg, power_compat.
        apply Succ_compat.
        apply ej.
        reflexivity.
      * intros   [x_ hx_] [x hx] ex.
        cbv in ex;simpl in*;subst.
        clear hx_.
        destruct hx as [n hn].
        pose (j := Dist Root n).
        assert (hj: j < N) by apply Dist_lt_Ring_size.
        do 2 exists (exist (x:=j) hj);simpl.
                    assert (h: x = v (g (power Succ j Root))).
                    {
                      rewrite <- hn.
                      apply v_compat, pg.
                      unfold j.
                      symmetry.
                      apply Dist_power_Succ; auto.
                    }
                    intuition; reflexivity.
    }
    assert (clas_eq_nat:forall n1 n2:nat, n1 == n2 \/ ~ n1 == n2)
      by (intros;simpl;lia).
    assert (clas_present: forall x, present x \/ ~ present x).
    { intro x;
        case_eq (existsb (fun n => Nat.eqb x (v (g n))) all_nodes);
        intro hh; [left | right];
          try (rewrite <- Bool.not_true_iff_false in hh; unfold not in hh);
          rewrite existsb_exists_InA in hh.
      destruct hh as [n [_ hh]].
      rewrite PeanoNat.Nat.eqb_eq in hh.
      exists n; congruence.
      intros n1 n2 en; apply PeanoNat.Nat.eqb_compat; auto; apply v_compat, pg, en.
      intros [n hn].
      contradict hh; split with n; split.
      apply all_nodes_prop.
      rewrite PeanoNat.Nat.eqb_eq; congruence.
      intros n1 n2 en; apply PeanoNat.Nat.eqb_compat; auto; apply v_compat, pg, en.
    }
    pose (absent V := V < k /\ convex g V).
    assert (clas_abs: forall x, absent x \/ ~ absent x).
    { intro x; unfold absent.
      assert (_h: x < k \/ ~(x < k)) by lia.
      assert (_hh: convex g x \/ ~convex g x)
        by (destruct (convex_dec g x); auto).
      destruct _h; [destruct _hh|]; intuition.
    }
    destruct (Classical_subset_Num_Card
                clas_eq_nat present _
                clas_present (map (fun n => v (g n)) all_nodes))
      as [nV hnV].
    { intros x [n hnx].
      replace eq with (equiv (A:=nat)) by reflexivity.
      rewrite InA_map_iff.
      exists n; split; auto using all_nodes_prop.
      intros n1 n2 en; apply v_compat, pg, en.
    }
    destruct (Classical_subset_Num_Card clas_eq_nat absent _
                                        clas_abs (range k)) as [nA hnA].
    {
      intros a [ha _].
      apply In_InA;try apply eq_equivalence.
      rewrite In_range;auto.
    }
    assert (HnVk: nV <= k).
    {
      apply PeanoNat.Nat.le_trans with N;auto.
      apply Card_smaller_le_iff.
      unfold Num_Card in *.
      apply Card_trans with (sig present) _;auto.
      apply Card_weak.
      apply Card_sym.
      assumption.
    }
    assert (HnVPk: nV + nA = k).
    { rewrite <- Card_same_eq_iff.
      eapply Card_trans with (sig (fun x => present x \/ absent x)) _.
      eapply Card_sym.
      simpl sym_Card_Prop.
      eapply (Disjoint_union_Num_Card present _ absent _ nV hnV nA hnA).
      {
        intros x [n hnx] [hkx conv].
        apply NO.
        assert (hhh := conv n Root (Between_self_l _ _) hnx).
        congruence.
      }
      apply Inj_iff.
      intros x1 x2 ex;simpl in *;lia.
      { intros x1 x2 ex;hnf in ex;hnf; unfold present, absent; intuition.
        subst.
        destruct H0 as [n hn].
        rewrite <- hn.
        apply v_ltk; auto.
        subst.
        case_eq (existsb (fun n => Nat.eqb x2 (v (g n))) all_nodes);
          intro hh; [left | right];
            try (rewrite <- Bool.not_true_iff_false in hh; unfold not in hh);
            (rewrite existsb_exists_InA in hh;
             [|intros n1 n2 en; apply PeanoNat.Nat.eqb_compat; auto;
               apply v_compat, pg, en]).
        { destruct hh as [n [_ hh]].
          rewrite PeanoNat.Nat.eqb_eq in hh.
          exists n; congruence.
        }
        { split;auto.
          intros n1 n2 Bn.
          intro hhh; contradict hh.
          exists n1; split.
          apply all_nodes_prop.
          rewrite PeanoNat.Nat.eqb_eq; congruence.
        }
      }
    }
    destruct (Compare_dec.le_lt_eq_dec HnVk) as [LT|EQ].
    {  assert (ex_not_present: Num_Card Larger (sig absent) 1).
       { assert (hP: nA > 0) by lia.
         unfold Num_Card.
         apply Card_sym.
         eapply Card_trans with (ltN nA) _.
         rewrite Card_smaller_le_iff.
         apply hP.
         apply Card_sym, Card_weak, hnA.
       }
       hnf in ex_not_present.
       destruct ex_not_present.
       hnf in inj_left_total.
       pose (_0ltN1 := (exist (PeanoNat.Nat.le_refl 1) : ltN 1)).
       specialize (inj_left_total _0ltN1 _0ltN1 (refl_equal 0)).
       destruct inj_left_total as [[b1 hb1]];apply (ex_intro _ b1 hb1) .
    }
    { rewrite EQ in *.
      clear nV HnVk EQ.
      assert (bijNpresent: Card Same (sig present) Node).
      {
        eapply Card_trans with (ltN k) _.
        apply hnV.
        replace  k with N.
        apply Card_sym.
        simpl sym_Card_Prop.
        apply Ring_elements_card.
        apply PeanoNat.Nat.le_antisymm; auto.
        apply Card_smaller_le_iff.
        apply Card_trans with (sig present) _; auto.
        apply Card_weak, Card_sym, hnV.
      }
      contradict NO.
      intros n n' B hv.
      destruct (Node_dec Root n) as [hn|hn]. 
      { apply v_compat, pg.
        generalize (Between_eq B); intuition.
      }
      { elimtype False.
        pose (present_outside_Root x :=
                exists n, ~ n == Root /\ v (g n) = x).
        assert (H1: Num_Card Same (sig present_outside_Root) k).
        { eapply Card_trans with (sig present) _.
          apply Inj_iff.
          intros x1 x2 ex; hnf in ex; apply ex_compat; intro nn; intuition.
          intros x_ x ex; hnf in ex;subst;
            unfold present, present_outside_Root.
          split.
          intros [nn [_ hnn]]; exists nn; congruence.
          intros [nn hnn]; destruct (Node_dec nn Root) as [_hn|_hn];
            [exists n|exists nn]; intuition.
          rewrite hv, <- hnn.
          symmetry; apply v_compat, pg, _hn.
          apply hnV.
        }
        assert (H2: Num_Card Smaller (sig present_outside_Root) (pred N)).
        {
          split with (fun x j => proj1_sig x =
                                 v (g (power Succ (proj1_sig j)
                                             (Succ Root)))); hnf;
            unfold Ring_size; simpl.
          * intros   [x1 hx1] [x2 hx2] ex [j1 hj1] [j2 hj2] ej.
            apply eq_compat.
            apply ex.
            apply v_compat, pg, power_compat.
            apply Succ_compat.
            apply ej.
            reflexivity.
          * intros [x1 hx1] [j1 hj1] h1 [x2 hx2] [j2 hj2] h2 ej.
            hnf; rewrite h1, h2.
            apply v_compat, pg, power_compat.
            apply Succ_compat.
            apply ej.
            reflexivity.
          * intros [x_ hx_] [x hx] ex.
            cbv in ex; simpl in *; subst.
            clear hx_.
            destruct hx as [nx [hnx hnx_]].
            pose (j := Dist (Succ Root) nx).
            assert (hj_: j < Dist (Succ Root) Root).
            { assert (_h:j <= Dist (Succ Root) Root).
              { apply Lt.lt_n_Sm_le.
                apply Dist_lt_Ring_size.
              }
              destruct (Compare_dec.le_lt_eq_dec _h); auto.
              contradict hnx.
              rewrite <-  (Dist_unique _ _ _ _ e).
              reflexivity.
            }
            do 2 exists (exist (x:=j) hj_); simpl.
                        assert (h_: x = v (g (power Succ j (Succ Root)))).
                        { rewrite <- hnx_.
                          apply v_compat, pg.
                          unfold j.
                          symmetry.
                          apply Dist_power_Succ.
                          auto.
                        }
                        intuition; reflexivity.
        }
        cut (k <= pred N); [lia|].
        apply Card_smaller_le_iff.
        eapply Card_trans with (sig present_outside_Root) _.
        apply Card_weak, Card_sym; auto.
        apply H2.
      }
    }
  Qed.
     
  (** ** Non-convex properties *)
  
  (** When the configuration is not convex, there exists a non-root
      node that holds the root value 
      (this directly comes from the definition)
   *)
  Lemma nonconvex_Root_direct:
    forall g, Proper pequiv g ->
              ~ convex g (v (g Root)) ->
              exists n n', Between Root n' n /\
                           v (g n) = v (g Root) /\
                           v (g n') <> v (g Root).
  Proof.
    intros g pg nc.
    unfold convex in nc.
    pose (ff := fun n n' =>
                  if Between_dec Root n' n then
                    if PeanoNat.Nat.eq_dec (v (g n)) (v (g Root))
                    then if PeanoNat.Nat.eq_dec (v (g n')) (v (g Root))
                         then true
                         else false
                    else true
                  else true).
    assert (ff_compat: Proper pequiv ff).
    { intros n1 n2 en nn1 nn2 enn; unfold ff.
      destruct (Between_dec Root nn1 n1);
        destruct (Between_dec Root nn2 n2); auto.
      2: { contradict n.
           eapply Between_rewrite_m; [symmetry; apply enn|].
           eapply Between_rewrite_r; [symmetry; apply en|].
           auto. }
      2: { contradict n.
           eapply Between_rewrite_m; [apply enn|].
           eapply Between_rewrite_r; [apply en|].
           auto. }
      destruct (PeanoNat.Nat.eq_dec (v (g n1)) (v (g Root)));
        destruct (PeanoNat.Nat.eq_dec (v (g n2)) (v (g Root))); auto.
      2: { contradict n; rewrite <- e.
           symmetry; apply v_compat, pg, en. }
      2: { contradict n; rewrite <- e.
           apply v_compat, pg, en. }
      destruct (PeanoNat.Nat.eq_dec (v (g nn1)) (v (g Root)));
        destruct (PeanoNat.Nat.eq_dec (v (g nn2)) (v (g Root))); auto.
      2: { contradict n; rewrite <- e1.
           apply v_compat, pg, enn. }
      { contradict n; rewrite <- e1.
        symmetry; apply v_compat, pg, enn. }
    }
    assert (F: forallb (fun n => forallb (ff n) all_nodes)
                       all_nodes = false).
    { rewrite <- not_true_iff_false.
      contradict nc.
      rewrite ListUtils.forallb_forall_InA in nc.
      2: { intros n1 n2 en.
           eapply ListUtils.forallb_compat. 
           intros nn1 nn2 enn; apply ff_compat; auto.
           apply enn.
           reflexivity.
      }
      intros n.
      specialize (nc n (all_nodes_prop _)).
      rewrite ListUtils.forallb_forall_InA in nc.
      2: { intros n1 n2 en; apply ff_compat; auto. }
      intros n'.
      specialize (nc n' (all_nodes_prop _)).
      unfold ff in nc.
      destruct (Between_dec Root n' n); [|tauto].
      destruct (PeanoNat.Nat.eq_dec (v (g n)) (v (g Root))); [|tauto].
      destruct (PeanoNat.Nat.eq_dec (v (g n')) (v (g Root))); [|congruence].
      tauto.
    }
    erewrite ListUtils.forallb_existsb
      with (notf := fun n => negb (forallb (ff n) all_nodes)) in F.
    2:{ intros n.
        rewrite negb_involutive; auto. }
    rewrite negb_false_iff in F.
    rewrite ListUtils.existsb_exists_InA in F.
    2: { intros n1 n2 en.
         simpl.
         rewrite BoolUtil.negb_lr, negb_involutive.
         eapply ListUtils.forallb_compat.
         intros nn1 nn2 enn; apply ff_compat; auto.
         apply enn.
         reflexivity.
    }
    destruct F as [n [_ F]].
    rewrite negb_true_iff in F.
    erewrite ListUtils.forallb_existsb
      with (notf := fun n' => negb (ff n n')) in F.
    2:{ intros nn.
        rewrite negb_involutive; auto. }
    rewrite negb_false_iff in F.
    rewrite ListUtils.existsb_exists_InA in F.
    2:{ intros n1 n2 en.
        simpl.
        rewrite BoolUtil.negb_lr, negb_involutive.
        apply ff_compat; auto.
    }
    destruct F as [n' [_ F]].
    rewrite negb_true_iff in F.
    unfold ff in F. 
    destruct (Between_dec Root n' n); [| congruence].
    destruct (PeanoNat.Nat.eq_dec (v (g n)) (v (g Root))); [|congruence].
    destruct (PeanoNat.Nat.eq_dec (v (g n')) (v (g Root))); [congruence|].
    clear F.
    exists n.
    exists n'.
    split; [|split]; auto.
  Qed.

  Lemma nonconvex_Root: 
    forall g, Proper pequiv g ->
              ~ convex g (v (g Root)) ->
              exists n, ~ n == Root /\
                        v (g n) = v (g Root).
  Proof.
    intros g pg nc.
    assert (NC := nonconvex_Root_direct pg nc).
    destruct NC as [n [n' [B [vn vn']]]].
    exists n.
    split; auto.
    contradict vn'.
    eapply Between_rewrite_r in B; [|symmetry; apply vn'].
    apply Between_self in B.
    apply v_compat, pg; symmetry; auto.
  Qed.

  (** ** Z
     The potential [Z] of a configuration [g] is 
        - 0 if [g] is convex for the value of the Root
        - the distance to the closest absent value from the Root (modulo k),
          otherwise
     i.e. [Z g = min { z < k | convex g ((v g Root + z) mod k) }]

     For example, with k=7,
           - Root->5 5 5 6 3 4 0 has potential 0
           - Root->5 5 5 6 3 4 5 has potential 2 (since the closest absent 
              value is 0, at distance 2 from 5) 
           - Root->5 5 6 2 5 3 0 has potential 3 (since the closest absent 
              value is 1, at distance 3 from 5) 
   *)

  (** Conv: Boolean function, (Conv g j) returns true when g is convex 
      for the value (vR + j) mod k, where vR is the value of the root. 
   *) 
  Notation Conv g := (fun j => convex_b g ((v (g Root) + j) mod k)).

  Definition Z (g: Env): nat  := min_below (Conv g) k.
  
  (** n.b.: for a function f and a value k, min_below f n is specified by:
             [min_below f n = n /\ (forall x, x < n -> f x = false) \/
              min_below f n < n /\ is_minimum f (min_below f n)]

        where, as expected, (is_minimum f m) means: 
             [f m = true /\ (forall m', m' < m -> f m' = false)]

        (is_minimum f m) is hence the minimum natural value on which f 
        returns true.
        (min_below f n) returns 
           - either n if no value above n allows f to return true
           - or the minimum value above n on which f is true.
   *)
  
  (** The potential [Z] is always smaller than k since there 
     always exists a value < k, such that convex holds 
     (see Lemma convex_exists above). **)
  Lemma Z_ltk:
    forall (g: Env) (pg: Proper fequiv g) (assume: Assume g),
      Z g < k.
  Proof.
    intros g pg assume; unfold Z.
    destruct (min_below_spec (Conv g) k) as [[M M']|]; [|lia].
    destruct (convex_exists pg assume) as [V [hV CV]].
    destruct (mod_plus_inv (v_ltk (g Root)) hV)
      as [[x [hx hhx]] _].
    specialize (M' _ hx); rewrite hhx in M'.
    rewrite <- convex_b_false in M'; tauto.
  Qed.

  (** The potential [Z] is actually the minimum among the 
     convex values *)
  Lemma Z_min:
    forall (g: Env) (pg: Proper fequiv g) (assume: Assume g) (p: nat),
      Z g = p <-> is_minimum (Conv g) p.
  Proof.
    intros g pg assume p;
      destruct (min_below_spec (Conv g) k) as [[M _]|[M1 M2]].
    - assert (d := Z_ltk pg assume);
        unfold Z in d; lia.
    - unfold Z; split; [intros P | intros M].
      + rewrite <- P; auto.
      + eapply is_minimum_unique; [apply M2 | apply M].
  Qed.

  (** ** When [Z] is zero... *)
  
  (** 'Potential [Z] is 0' is equivalent to 
        'the configuration is convex for the root value'. 
   *)
  Lemma pot_0_equiv:
    forall (g: Env) (pg: Proper fequiv g) (assume: Assume g),
      Z g = 0 <-> convex g (v (g Root)).
  Proof.
    intros g pg assume.
    rewrite Z_min; auto.
    split; [intros p0 | intros C].
    { rewrite convex_equiv; auto.
      left; split; auto.
      destruct p0 as [min _].
      unfold convex_b in min;
        destruct (convex_dec g ((v (g Root) + 0) mod k)) as [C|C];
        [|congruence].
      rewrite <- (v_plus_0 (g Root)); apply C; auto.
    }
    { split.
      unfold convex_b; rewrite (v_plus_0 (g Root)).
      destruct (convex_dec g (v (g Root))); auto.
      intros m' hm'; lia.
    }
  Qed.
  
  (** In a legitimate configuration, Z is 0, since 
      the configuration is convex. To prove this, we use the 
      characterization of a legitimate configuration [ls_charac]. *)
  Lemma Leg_state_pot_0:
    forall g, Proper pequiv g -> Assume g -> Leg_state g ->
              Z g = 0.
  Proof.
    intros g pg ass ls.
    apply pot_0_equiv; auto.
    dup ls ls'.
    apply ls_charac in ls'; auto.
    destruct ls' as [ntok [tok [val1 val2]]].
    intros n n' hb hv.
    case (Node_dec Root ntok); intros hr.
    { assert (V: v (g (Pred Root)) = v (g Root)).
      { apply (Token_compat pg _ _ hr) in tok.
        unfold Token in tok; intuition. }
      rewrite <- V; symmetry; apply val2.
      eapply Between_rewrite_l; [symmetry; apply hr|].
      apply all_Between_Pred.
    }
    { assert (nt: ~ Token g Root).
      { contradict hr; apply ls; auto. }
      case (Between_dec ntok n' (Pred Root)); intros B.
      { assert (BB: Between ntok n (Pred Root)).
        { apply Between_trans1 with n'; auto.
          revert hb; apply Between_compat; auto; apply Succ_Pred.
        }
        specialize (val2 _ BB).
        rewrite <- val2 in hv.
        unfold Token in nt; intuition.        
      }
      { destruct (Between_alt_m ntok n' Root) as [BB|[BB|BB]].
        symmetry in BB; tauto.
        apply Between_Pred_r in BB; destruct BB as [BB|BB].
        { apply v_compat, pg; auto. }
        tauto.
        apply Between_Pred_r in BB; destruct BB as [BB|BB].
        { contradict B; eapply Between_rewrite_l.
          symmetry; apply BB.
          apply Between_self_l.
        }
        { symmetry; apply val1; auto. }
      }
    }
  Qed.

  (** A configuration where Z is 0 and where the root holds a 
      token is legitimate. 
      Indeed, in such configuration, every node holds one and the
      same value (the root one).
   *)
  Lemma Z_0_Token_Root_Leg_state:
    forall g,  Assume g -> Proper fequiv g ->
               Z g = 0 -> Token g Root -> Leg_state g. 
  Proof.
    intros g ass pg hZ tR.
    destruct tR as [[_ tR]|[tR _]]; [|contradict tR; auto].   
    apply pot_0_equiv in hZ; auto.
    unfold convex in hZ.
    specialize (hZ (Pred Root)).
    assert (T: forall n', v (g n') = v (g Root)).
    { intros n'; apply hZ.
      apply all_Between_Pred.
      symmetry; auto.
    }
    clear hZ.
    intros n1 n2 tok1 tok2.
    assert (t1 := T n1).
    assert (tp1 := T (Pred n1)).
    rewrite <- tp1 in t1; clear tp1.
    assert (t2 := T n2).
    assert (tp2 := T (Pred n2)).
    rewrite <- tp2 in t2; clear tp2.
    unfold Token in tok1, tok2; intuition.
    rewrite <- H0, <- H2; auto.
  Qed.

  (** ** [Z]-evolution during a step from an illegitimate configuration *)
  
  (** The potential [Z] decreases when the root executes from a 
     illegitimate configuration.

     We first show that the potential at g is not null, using
     Z_0_Token_Root_Leg_state.  (Indeed, every value in the ring would
     be equal to the one of the root, hence there would be a unique
     token, hence this would be a legitimate configuration.)

     Now, we have: 
     - the Root holds a Token, i.e., its value is equal to the value of its 
       predecessor 
       (a) [Token g Root] and (b) [v (g Root) = v (g (Pred Root))] 
     - the Root increments its value modulo k 
       (c) [v (g' Root) = (v (g Root) + 1) mod k]

     To show that the potential is g' has decreased, 
     we rewrite potential as minimum (using
     [Z_min]) and applies Lemma [is_minimum_lt] that says
     that if M1 (resp. M2) is the minimum of some set A1 (resp. A2)
     and if for every value in A1, there exists a smaller value in A2,
     then M2 is smaller than M1.  Hence, for any value [x] such that
     (x + v (g Root)) mod k is absent from the ring in configuration
     g, we have to build a smaller value [x'] such that (x + v (g
     Root)) mod k is absent from the ring in configuration g.  

     If x >= k, we just pick the value given by convex_exists in g'; as this
     value is smaller than k, we are done.

     If x < k, x cannot be 0 (resp. 1 otherwise the potential in g
     (resp. in g') is null.  Then the value we are looking for is
     (x-1), that is to say: we prove that the configuration g' is
     convex for value (v (g' Root) + (x - 1)) mod k = (v (g Root) + x)
     mod k.  Namely, we prove that 
       [ forall n, (v (g Root) + x) mod k <> v (g' n) ] 
     since no value appears in the ring but ((v (g Root) + 1) mod k) and 
     1 < x < k.  
   **)
  Lemma Z_decreases:
    forall g, Assume g -> ~Leg_state g ->
              forall g', Step g' g ->
                         has_moved_b g' g Root = true ->
                         Z g' < Z g.
  Proof.
    intros g ass nonleg g' step hmr.
    assert (pg := Step_Proper_r step).
    assert (pg' := Step_Proper_l step).
    assert (ass' := step_Assume ass step).
    assert (Z_pos: Z g > 0).
    { case (Nat.eq_dec (Z g) 0); intros hZ; [|lia].
      apply Z_0_Token_Root_Leg_state in hZ; auto; [tauto|].
      apply has_moved_b_enabled_b in hmr; auto.
      apply enabled_Token; auto.
    }
    assert (tokr: Token g Root)
      by (rewrite <- enabled_Token; auto;
          apply has_moved_b_enabled_b with g'; auto).
    destruct tokr as [[_ tokr]|[f _]]; [|contradict f; reflexivity].
    eapply has_moved_v_Root in hmr; auto.
    destruct (Nat.eq_dec (Z g') 0) as [P'|P']; [lia|].
    remember (Z g) as p; symmetry in Heqp;
      remember (Z g') as p'; symmetry in Heqp'.
    rewrite Z_min in Heqp;
      rewrite Z_min in Heqp'; auto.
    eapply is_minimum_lt.
    2: apply Heqp.
    2: apply Heqp'.
    intros x hx.
    rewrite <- convex_b_true, convex_equiv in hx; auto.
    destruct hx as [[rx hx]|hx].
    {
      destruct Heqp as [h hh].
      rewrite convex_b_true, <- v_plus_0 in hx; auto.
      assert (d: 0 < p) by lia; specialize (hh _ d); congruence.
    }
    { destruct (le_gt_dec k x) as [hxk|hxk].
      { destruct (convex_exists pg') as [C [hC hhC]]; auto.
        destruct (mod_plus_inv (v_ltk (g' Root)) hC)
          as [[V [hV hhV]] _].
        exists V; rewrite hhV; split.
        rewrite <- convex_b_true; auto.
        lia.
      }

      (**** *)
      assert (hhx: x > 0).
      { destruct (le_gt_dec x 0); auto.
        - assert (d: x = 0) by lia; rewrite d in hx; clear d.
          specialize (hx Root); contradict hx; rewrite v_plus_0; lia.
      }
      exists (x - 1); split; [|lia].
      rewrite hmr, <- tokr.
      assert (T: ((v (g Root) + 1) mod k + (x - 1)) mod k =
                 (v (g Root) + x) mod k).
      { assert (vr := v_ltk (g Root)).
        destruct (le_gt_dec k (v (g Root) + 1)) as [vk|vk].
        - assert (e: v (g Root) + 1 = k) by lia; rewrite e; clear e.
          assert (e: v (g Root) = k - 1) by lia; rewrite e; clear e.
          assert (e: (k - 1 + x) = k + (x - 1)) by lia; rewrite e;
            clear e.
          rewrite Nat.add_mod_idemp_l; [reflexivity|lia].
        - rewrite (Nat.mod_small (v (g Root) + 1)); [|lia].
          assert (e: v (g Root) + 1 + (x - 1) = v (g Root) + x) by lia;
            rewrite e; clear e.
          reflexivity.
      }
      rewrite T; clear T.
      rewrite <- convex_b_true, convex_equiv; auto.
      destruct (le_gt_dec x 1). 
      { assert (d: x = 1) by lia; subst; clear l hhx.
        left; split; [rewrite tokr, hmr; auto|].
        intros n n' B.
        rewrite hmr, <- tokr; intros V.
        destruct (Node_dec Root n) as [nr|nr];
          [| destruct (Node_dec Root n') as [nr'|nr']].
        { rewrite <- (Between_compat
                        _ _ (Equivalence_Reflexive Root)
                        _ _ (Equivalence_Reflexive n') _ _ nr) in B.
          rewrite Between_self in B;
            rewrite <- (v_compat (pg' _ _ B)); auto.
          rewrite hmr, <- tokr; auto.
        }
        { rewrite <- (v_compat (pg' _ _ nr')); auto.
          rewrite hmr, <- tokr; auto.
        }
        { case_eq (has_moved_b g' g n); intros hmn.
          + eapply has_moved_v_non_Root in hmn; auto.
            specialize (hx (Pred n)); contradict hx; lia.
          + apply has_moved_false in hmn; auto; destruct hmn as [hmn _].
            specialize (hx n); contradict hx.
            hnf in hmn;unfold v in *.
            lia.
        }
      }
      { right.
        intros n; destruct (Node_dec Root n) as [nr|nr].
        - rewrite <- (v_compat (pg' _ _ nr)).
          rewrite hmr, <- tokr.
          intros F.
          assert (d: k <> 0) by lia.
          destruct
            (mod_plus_inv (v_ltk (g (Root)))
                          (Nat.mod_upper_bound (v (g Root) + 1) _ d))
            as [_ V]; clear d.
          assert (d: x < k) by lia; assert (dd: 1 < k) by lia.
          specialize (V  _ _ d dd F eq_refl); lia.
        - case_eq (has_moved_b g' g n); intros hmn.
          + eapply has_moved_v_non_Root in hmn; auto.
            rewrite hmn; apply hx.
          + apply has_moved_false in hmn; auto; destruct hmn as [hmn _].
            unfold v in *;rewrite hmn; apply hx.
      }
    }
  Qed.
  
  (** The potential [Z] does not increase from a illegitimate
  configuration.

   We rewrite potentials as minimum (using [Z_min]) and applies Lemma
   [is_minimum_le] that says that if M1 (resp. M2) is minimum of some
   set A1 (resp. A2) and if for every value in A1, there exists a
   smaller or equal value in A2, then M2 is smaller than or equal to
   M1.

   We first restrict to values smaller than k in M1 (i.e. values x
   such that (convex g ((v (g Root) + x) mod k)) holds).  - If x is 0,
   we prove that (convex g' (v (g Root))): since the value is copied
   from the predecessor when a node moves, this cannot break the
   convex property for (v (g Root)).

   - If x <> 0 then, no node in the ring holds value ((v (g Root) + x)
     mod k) = ((v (g' Root) + x) mod k).  As no new value appears in
     the step, (convex g ((v (g Root) + x) mod k)) holds.  
   *)
  Lemma Z_does_not_increase:
    forall g, Assume g -> ~Leg_state g ->
              forall g', Step g' g ->
                         Z g' <= Z g.
  Proof.
    assert (H_k := H_k).
    intros g ass nls g' step.
    assert (pg: Proper fequiv g) by apply step.
    assert (pg': Proper fequiv g') by (apply Step_Proper_l with g, step).
    case_eq (has_moved_b g' g Root); intros hr.
    { apply Z_decreases in hr; auto; lia. }
    apply has_moved_false, v_compat in hr; simpl in hr.
    eapply is_minimum_le.
    2: apply Z_min; [apply pg| apply ass|reflexivity].
    2: apply Z_min; [apply pg'| |reflexivity].
    2: apply step_Assume with g; auto. 
    rewrite hr.
    cut (forall v1, v1 < k ->
                    convex_b g ((v (g Root) + v1) mod k) = true ->
                    exists v2,
                      convex_b g' ((v (g Root) + v2) mod k) = true /\
                      v2 <= v1).
    { intros ok x Cx.
      assert (d: k <> 0) by lia;
        specialize (ok (x mod k) (Nat.mod_upper_bound x _ d)); clear d.
      rewrite Nat.add_mod_idemp_r in ok; [|lia].
      specialize (ok Cx).
      destruct ok as [y [Cy hy]].
      exists y; split; auto.
      transitivity (x mod k); auto; apply Nat.mod_le; lia.
    }
    intros x hhx Cx.
    rewrite <- convex_b_true in Cx; auto.
    destruct (Nat.eq_dec x 0) as [hx|hx].
    { rewrite hx; rewrite hx in Cx; clear x hhx hx.
      exists 0; split; [|lia].
      rewrite <- convex_b_true; auto.
      rewrite v_plus_0; rewrite v_plus_0 in Cx; auto.
      intros n n' B V.
      case_eq (has_moved_b g' g n); intros hmn.
      - assert (nr: ~ Root == n).
        { rewrite has_moved_v_change in hmn; auto.
          contradict hmn.
          rewrite <- (v_compat (pg' _ _ hmn));
            rewrite <- (v_compat (pg _ _ hmn)); lia. }
        eapply has_moved_v_non_Root in hmn; auto.
        rewrite hmn in V.
        case_eq (has_moved_b g' g n'); intros hmn'.
        + assert (nr': ~ Root == n').
          { rewrite has_moved_v_change in hmn'; auto.
            contradict hmn'.
            rewrite <- (v_compat (pg' _ _ hmn'));
              rewrite <- (v_compat (pg _ _ hmn')); lia. }
          eapply has_moved_v_non_Root in hmn'; auto.
          rewrite hmn'.
          destruct (Between_Pred_r B) as [B'|B'].
          rewrite (v_compat (pg _ _ (Pred_compat _ _ B'))); lia.
          destruct (Between_Pred_m B') as [B''|B''].
          tauto.
          specialize (Cx _ _ B''); apply Cx, V.
        + apply has_moved_false in hmn';auto;destruct hmn' as [hmn' _]; auto.
          destruct (Between_Pred_r B) as [B'|B'].
          rewrite <- (v_compat (pg' _ _ B')) in hmn; lia.
          unfold v in *;rewrite hmn'.
          specialize (Cx _ _ B'); apply Cx, V.
      - apply has_moved_false in hmn;auto;destruct hmn as [hmn _]; auto.
        unfold v in V;rewrite hmn in V.
        case_eq (has_moved_b g' g n'); intros hmn'.
        + assert (nr': ~ Root == n').
          { rewrite has_moved_v_change in hmn'; auto.
            contradict hmn'.
            rewrite <- (v_compat (pg' _ _ hmn'));
              rewrite <- (v_compat (pg _ _ hmn')); lia. }
          eapply has_moved_v_non_Root in hmn'; auto.
          rewrite hmn'.
          destruct (Between_Pred_m B) as [B'|B'].
          tauto.
          specialize (Cx _ _ B'); apply Cx, V.
        + apply has_moved_false in hmn';auto;destruct hmn' as [hmn' _]; auto.
          unfold v;rewrite hmn'.
          specialize (Cx _ _ B); apply Cx, V.
    }
    { rewrite convex_equiv in Cx; auto.
      destruct Cx as [[Cx _]|Cx].
      { assert (d: v (g Root) = (v (g Root) + 0) mod k) by
            (rewrite v_plus_0; auto); rewrite d in Cx at 2; clear d.
        assert (d: k <> 0) by lia.
        destruct (mod_plus_inv
                    (v_ltk (g Root))
                    (Nat.mod_upper_bound (v (g Root) + 0) _ d)) as [_ M];
          clear d.
        assert (d: 0 < k) by lia.
        specialize (M _ _ hhx d Cx eq_refl); clear d; lia.
      }
      exists x.
      split; [|lia].
      rewrite <- convex_b_true, convex_equiv; auto.
      right.
      intros n.
      case_eq (has_moved_b g' g n); intros hmn.
      - assert (nr: ~ Root == n).
        { rewrite has_moved_v_change in hmn; auto.
          contradict hmn.
          rewrite <- (v_compat (pg' _ _ hmn));
            rewrite <- (v_compat (pg _ _ hmn)); lia. }
        eapply has_moved_v_non_Root in hmn; auto.
        rewrite hmn; apply Cx.
      - apply has_moved_false in hmn;auto;destruct hmn as [hmn _]; auto.
        unfold v;rewrite hmn; apply Cx.
    }
  Qed.

  Lemma Z_evolves:
    forall g, Assume g -> ~Leg_state g ->
              forall g', Step g' g ->
                         if has_moved_b g' g Root then Z g' < Z g
                         else Z g' <= Z g.
  Proof.
    intros g ass nls g' step.
    assert (Z1 := Z_decreases ass nls step).
    assert (Z2 := Z_does_not_increase ass nls step).
    destruct (has_moved_b g' g Root); auto.
  Qed.
       
  (** ** Bound on Z *)

  (** Every value between 0 and pot_with_Root-1 is present at 
      some node in the ring.
      Indeed, Z is minimum among the convex values (v+j)mod k.
   *)
  Lemma all_here:
    forall g, Assume g -> Proper fequiv g ->
              forall x, x < Z g ->
                        exists n, v (g n) = (v (g Root) + x) mod k.
  Proof.
    intros g h_assume pg x x_ltZ.
    destruct (All_In.All_In_dec_forall_exists
                all_nodes_prop
                (P := fun n => v (g n) = (v (g Root) + x) mod k)) as
        [hh|hh]; auto.
    {
      intros n1 n2 en; apply eq_compat.
      apply v_compat, pg, en.
      apply f_equal2; auto.
    }
    {
      intros a; apply Nat.eq_dec.
    }
    contradict hh; intro hh.
    assert (h := proj1 (Z_min pg h_assume (Z g)) (refl_equal _)).
    unfold is_minimum in h.
    assert (conv_x: convex g ((v (g Root) + x) mod k)).
    { rewrite convex_equiv; auto. }
    apply proj2 in h.
    specialize (h x x_ltZ).
    rewrite <- convex_b_false in h; auto.
  Qed.

  (** Z is upper bounded by N-1
     If the configuration is convex, the result is trivial.
     We consider a non convex configuration. We use counting 
     arguments which stands on [all_here] (that
     asserts that every value between 0 and Z-1 is represented in the
     ring).

     - presentNR are the nodes which hold a value different from the
       root. Cardinal = NR
     - presentR are the nodes which hold the root value. Cardinal = R

     We show that R >=2 due to [nonconvex_Root]. And that N = NR + R.
     Hence NR <= N - 2.

     Now, Z g - 1 <= NR due to pot_with_root definition.
     Hence the result.
   *)
  Lemma Z_lt_N:
    forall g, Assume g -> Proper fequiv g -> Z g <= N - 1.
  Proof.
    intros g ass pg.
    destruct (le_gt_dec (Z g) 0) as [|hZ]; [lia|].

    destruct (convex_dec g (v (g Root))) as [c |NC].
    { specialize (c pg).
      rewrite <- pot_0_equiv in c; auto.
      lia. }
    assert (AH := all_here ass pg).

    pose (presentNR n := v (g n) <> v (g Root)).
    destruct (All_In.All_In_exists_Card
                all_nodes_prop (P := presentNR) (A_dec := Node_dec))
      as [NR HNR].
    { intros n1 n2 en; unfold presentNR;
        rewrite <- (v_compat (pg _ _ en)); auto. }
    { intros n.
      case (eq_nat_dec (v (g n)) (v ( g Root))); intros h;
        [right|left]; unfold presentNR; lia. }
    
    assert (h: NR + 2 <= N).
    {
      assert (HN := Ring_elements_card Root).

      pose (presentR n := v (g n) = v (g Root)).
      destruct (All_In.All_In_exists_Card
                  all_nodes_prop (P := presentR) (A_dec := Node_dec))
        as [R HR].
      { intros n1 n2 en; unfold presentR;
          rewrite <- (v_compat (pg _ _ en)); auto. }
      { intros n.
        case (eq_nat_dec (v (g n)) (v ( g Root))); intros h;
          [left|right]; unfold presentR; lia. }
      
      assert (hR: 2 <= R).
      { apply Card_smaller_le_iff.
        eapply Card_trans with (B :=  {x : Node | presentR x}).
        2: { apply HR. }
        destruct (nonconvex_Root pg NC) as [xR hxR].
        split with (fun (x: ltN 2) (y: {n: Node | presentR n}) =>
                      proj1_sig x = 0 /\ proj1_sig y == Root \/
                                         proj1_sig x = 1 /\ proj1_sig y == xR).
        * intros x1 x2 hx n1 n2 hn.
          simpl in hn, hx; unfold eqSig in hn, hx;
            destruct x1 as [x1 hx1]; destruct x2 as [x2 hx2];
              destruct n1 as [n1 hn1]; destruct n2 as [n2 hn2];
                simpl in *.
          rewrite hx, hn; tauto.
        * intros x1 n1 [[h1 hh1]|[h1 hh1]]
                 x2 n2 [[h2 hh2]|[h2 hh2]] ej; hnf; try lia;
            simpl in ej; unfold eqSig in ej;
              destruct x1 as [x1 hx1]; destruct x2 as [x2 hx2];
                destruct n1 as [n1 hn1]; destruct n2 as [n2 hn2];
                  simpl in *;
                  destruct hxR as [hxR _];
                  contradict hxR;
                  rewrite <- hh2, <- hh1.
          symmetry; auto.
          auto.
        * intros x1 x2 hx.
          simpl in hx; unfold eqSig in hx;
            destruct x1 as [x1 hx1]; destruct x2 as [x2 hx2];
              simpl in *.
          destruct (le_gt_dec x1 0).
          { assert (dum: presentR Root) by reflexivity.
            exists (exist dum).
            exists (exist dum).
            split; [|split].
            left; split; [lia|reflexivity].
            left; split; [lia|reflexivity].
            reflexivity.
          }              
          { assert (dum: presentR xR) by apply hxR.
            exists (exist dum).
            exists (exist dum).
            split; [|split].
            right; split; [lia|reflexivity].
            right; split; [lia|reflexivity].
            reflexivity.
          }              
      }
      assert (N = NR + R).
      { apply Subset_Num_Card_iff_eq with
            (PA1 := fun _ => True)
            (PA2 := fun n => presentNR n \/ presentR n).        
        unfold Num_Card.
        eapply Card_trans with (B :=  Node).
        apply Trivial_subset_Num_Card.
        apply HN.          
        apply Disjoint_union_Num_Card; auto.
        { intros n1 n2 en; unfold presentNR;
            rewrite <- (v_compat (pg _ _ en)); auto. }
        { intros n1 n2 en; unfold presentR;
            rewrite <- (v_compat (pg _ _ en)); auto. }
        { intros _ n _.
          split; [intros _|tauto].
          case (eq_nat_dec (v (g n)) (v (g Root)));
            intros hh; [right|left]; auto. }
      }            
      lia.
    }
    assert (hh: Z g - 1 <= NR).
    { apply Card_smaller_le_iff.
      eapply Card_trans with (B :=  {x : Node | presentNR x}).
      2: apply HNR.
      split with
          (fun (x: ltN (Z g -1)) (y: {n: Node | presentNR n}) =>
             v (g (proj1_sig y)) = (v (g Root) + S (proj1_sig x)) mod k).
      { intros x1 x2 hx n1 n2 hn.
        simpl in hn, hx; unfold eqSig in hn, hx;
          destruct x1 as [x1 hx1]; destruct x2 as [x2 hx2];
            destruct n1 as [n1 hn1]; destruct n2 as [n2 hn2];
              simpl in *.
        rewrite hx.
        rewrite (v_compat (pg _ _ hn)).
        tauto.
      }
      { intros x1 n1 h1 x2 n2 h2 hh.
        simpl in hh; unfold eqSig in hh;
          destruct x1 as [x1 hx1]; destruct x2 as [x2 hx2];
            destruct n1 as [n1 hn1]; destruct n2 as [n2 hn2];
              hnf; simpl in *.
        rewrite (v_compat (pg _ _ hh)) in h1.
        assert (Zltk := Z_ltk pg ass).
        cut (S x1 = S x2); [lia|].
        eapply mod_plus_inv.
        6: symmetry in h2; apply h2.
        5: symmetry in h1; apply h1.
        apply v_ltk.
        apply v_ltk.
        lia. lia.
      }
      { intros x1 x2 hx.
        simpl in hx; unfold eqSig in hx;
          destruct x1 as [x1 hx1]; destruct x2 as [x2 hx2];
            hnf; simpl in *.
        subst.
        destruct (AH (S x2)) as [n hn]; try lia.
        assert (hnn: presentNR n).
        { intros F.
          rewrite F in hn.
          assert (v (g Root) = (v (g Root) + 0) mod k).
          { assert (v (g Root) = (v (g Root) + 0)) by lia.
            rewrite <- H; clear H.
            symmetry; apply NPeano.Nat.mod_small.
            apply v_ltk. }
          assert (FF: 0 = S x2).
          { eapply mod_plus_inv.
            6: symmetry; apply hn.
            5: symmetry; apply H.
            apply v_ltk.
            apply v_ltk.
            lia.
            assert (Zltk := Z_ltk pg ass);
              lia.
          }
          lia.
        }
        exists (exist hnn).
        exists (exist hnn).
        split; [|split]; auto.
        reflexivity.
      }
    }
    lia.
  Qed.
  
  (** ** When Z is N - 1: in a step from a non-legitimate
    configuration, it decreases whatever be the node which moves.
   *)

  (** SKETCH 

       Situation : (vR = value of the Root) 

       There exists a unique non-root node (which neither the successor of 
       the root) which holds vR.
       For every value between vR+1 et vR+N-2, there exists a unique node which
       holds this value.

       Counting arguments.

       In a step of the algorithm,
       - If the root moves then
           the value of the root becomes vR+1 
           as vR+1+N-2 is absent (and no value can appear)
           we obtain Z g' <= N-2

       - Otherwise, there exists a node n which moves but whose successor 
         does not.
         (induction which exhibits such a node, the closest from the root)
       
       If the value of n is vR, then this vR disappears:
       - the root still holds vR
       - its successort may also holds vR if it has moved
       In both cases, we get a convex configuration hence Z g' = 0.

       If the value vn of n is not vR, then we have
       - vR < vn < vR+N-1
       n is the only node to hold vn and vn is absent from g'. 
       Hence Z g' < N-1
   *)

  Section Z_Fully_Decreases.

    Variable HN4: N >= 4.
    Variable (g: Env) (ass: Assume g).
    Variable (g': Env) (step: Step g' g).

    Variable (HZ: Z g = N - 1).

    Notation pg' := (Step_Proper_l step).
    Notation pg := (Step_Proper_r step).

    (** We note presentX x n the predicate which means that node
        n holds present value x *)
    Definition presentX :=
      (fun x n => (v (g Root) + x) mod k = v (g n)).

    Instance presentX_compat: Proper pequiv presentX.
    Proof.
      intros x1 x2 ex nn1 nn2 enn; unfold presentX.
      simpl in ex; subst.
      rewrite (v_compat (pg _ _ enn)); auto.
    Qed.

    Lemma v_inv: forall n, exists x,
          x < k /\ (v (g Root) + x) mod k = v (g n) /\
          forall x', x' < k -> x' <> x -> ~ presentX x' n.
    Proof.
      intros n.
      destruct (@mod_plus_inv' (v (g Root)) (v (g n)) k).
      apply v_ltk.
      apply v_ltk.
      exists x; apply H.
    Qed.
    
    Lemma v_inv_unique: forall n,
        forall x x', presentX x n -> presentX x' n ->
                     x < k -> x' < k -> x = x'.
    Proof.
      intros n.
      destruct (@mod_plus_inv (v (g Root)) (v (g n)) k).
      apply v_ltk.
      apply v_ltk.
      intros x x' hp hp' hx hx'; apply H0; auto.
    Qed.

    (** The number of non-root nodes is N-1. *)
    Lemma HNR: Num_Card Same {x : Node | ~ x == Root} (N - 1).
    Proof.
      destruct (All_In.All_In_exists_Card
                  all_nodes_prop (A_dec := Node_dec)                  
                  (P := fun n => ~ n == Root))
        as [NR HNR].
      { intros n1 n2 en; rewrite en; auto. }
      { intros n; destruct (Node_dec n Root); [right|left]; auto. }
      assert (hhnr: NR = N - 1).
      { assert (Num_Card Same {x : Node | x == Root} 1).
        eapply Singleton_Num_Card.
        reflexivity.
        intros a2 h; symmetry; auto.
        cut (N = 1 + NR); [lia|].
        assert (HN := Ring_elements_card Root).
        apply Subset_Num_Card_iff_eq with
            (PA1 := fun _ => True)
            (PA2 := fun n =>  n == Root \/ ~ n == Root).        
        unfold Num_Card.
        eapply Card_trans with (B :=  Node).
        apply Trivial_subset_Num_Card.
        apply HN.
        apply Disjoint_union_Num_Card; auto.
        { intros n1 n2 en; rewrite en; auto. }
        { intros n1 n2 en; rewrite en; auto. }
        { intros _ n _.
          split; [intros _|tauto].
          destruct (Node_dec n Root); [left|right]; auto. }
      }
      subst; auto.
    Qed.
    
    (**
        Let Ua = { n, ~n==Root /\ 
                      exists x < a, x<N-1 /\ presentX x n }
        Ua_card defines the number of non-root nodes in Ua.
     *)
    Lemma Ua_card:
      forall a, exists Na,
          Num_Card Same {n | ~ n == Root /\
                             (exists x, x < a /\ presentX x n)} Na.
    Proof.
      intros a.
      destruct (All_In.All_In_exists_Card
                  all_nodes_prop (A_dec := Node_dec)
                  (P := fun nn => ~ nn == Root /\
                                  exists x, x < a /\ presentX x nn)) as [Na HNa].
      { intros nn1 nn2 enn.
        apply and_compat; [rewrite enn; tauto|].
        apply ex_compat; intros x.
        rewrite enn; tauto.
      }
      { intros x.
        case (Node_dec x Root); intros hr.
        - right; intros F; tauto.
        - destruct (v_inv x) as [vx [hvx [hhvx _]]].
          case (le_gt_dec a vx); intros ha.
          + right; intros [_ [vvx [hhvv hvv]]].
            assert (vx = vvx).
            apply (v_inv_unique (n := x)); auto; lia.
            lia.
          + left; split; auto.
            exists vx; split; auto.
      }
      exists Na; auto.
    Qed.

    
    (** From Lemmas all_here and nonconvex_Root, we know that for
        each present value in {0, ..., N-2}, there exists a non-root
        node which holds it. *)
    Lemma AB:
      forall x, x < N-1 ->
                exists n, ~ n == Root /\ presentX x n.
    Proof.
      intros x hx; unfold presentX.
      case (eq_nat_dec x 0); intros hhx.
      - subst; clear hx; rewrite v_plus_0.
        destruct (nonconvex_Root pg).
        assert (~ Z g = 0) by lia.
        contradict H; apply pot_0_equiv; auto.
        apply pg.
        exists x; split; [|symmetry]; apply H.
      - edestruct (all_here ass pg).
        rewrite HZ; apply hx.
        exists x0; split; auto.
        intros F; rewrite (v_compat (pg _ _ F)) in H.
        clear F x0.
        contradict hhx.
        eapply mod_plus_inv.
        6: rewrite v_plus_0; reflexivity.
        5: symmetry; apply H.
        4: lia.
        3: lia.
        apply v_ltk.
        apply v_ltk.
    Qed.

    (** Let Up_a = { n | ~ n == Root /\ presentX a n } for any 
        value a. Up_a has a cardinality (see Upa_card) which is
        positive (see Upa). The results are direct from the 
        definition of the set and from Lemma AB. *)
    Lemma Upa_card:
      forall a, a < k -> exists Na,
          Num_Card Same {n | ~ n == Root /\ presentX a n} Na.
    Proof.
      intros a ha.
      destruct (All_In.All_In_exists_Card
                  all_nodes_prop (A_dec := Node_dec)
                  (P := fun nn => ~ nn == Root /\ presentX a nn))
        as [Na HNa].
      { intros nn1 nn2 enn; apply and_compat; rewrite enn; tauto. }
      { intros x.
        case (Node_dec x Root); intros hhr.
        - right; intros F; tauto.
        - destruct (v_inv x) as [vx [hvx [hhvx _]]].
          case (eq_nat_dec vx a); intros hhha.
          + left; split; subst; auto.
          + right; intros [_ hhvv].
            contradict hhha.
            apply (v_inv_unique (n := x)); auto; lia.
      }
      exists Na; auto.
    Qed.
    
    Lemma Upa:
      forall a, a < N-1 ->
                Num_Card Larger {n | ~ n == Root /\ presentX a n} 1.
    Proof.
      intros a ha.
      destruct (AB ha).
      apply (Nonempty_Num_Card
               (@exist Node (fun x => ~ x == Root /\ presentX a x)
                       _ H)).
      unfold Proper; reflexivity.
    Qed.

    (** We show that forall a in {1, ..., N-1}, |Ua| >= a
        Induction on a. 
        The base case, for a=1, is provided by nonconvex_Root: 
        this exhibits a non-root node withe the same value as the 
        root.
        For induction step, we write 
          Ua = U(a-1) Union {n, ~n==Root /\ presentX a n}
        (This is a disjoint union.)
        By induction hypothesis, |U(a-1)| >= a - 1 and we can prove 
        that |{n, ~n==Root /\ presentX a n}| >= 1 since Lemma 
        all_here provides a node with present value a.
     **)
    Lemma Ua:
      forall a, a <= N-1 ->
                Num_Card Larger
                         {n | ~ n == Root /\
                              exists x, x < a /\ presentX x n} a.
    Proof.
      intros b hr.
      pose (a := b); fold a.
      assert (ha: a <= b) by lia.
      clearbody a.
      induction a as [| a ih].
      { destruct (Ua_card 0) as [Na HNa].
        eapply Card_trans with (B := ltN Na).
        apply Card_weak, HNa.
        apply Card_larger_ge_iff; lia.
      }
      { (* induction step *)
        destruct (Ua_card a) as [Na HNa].
        destruct (Ua_card (S a)) as [Nap HNap].
        assert (d: a < k) by lia.
        destruct (Upa_card d) as [Naa HNaa]; clear d.
        assert (hlea: Na >= a).
        { apply Card_larger_ge_iff.
          eapply Card_trans with
              (B := {n | ~ n == Root /\ (exists x, x < a /\ presentX x n)}).
          apply Card_weak, Card_sym, HNa.
          apply ih; lia.
        }
        assert (hsum: Nap = Na + Naa).
        { apply Subset_Num_Card_iff_eq with
              (PA1 := fun nn => ~ nn == Root /\
                                exists x, x < S a /\ presentX x nn)
              (PA2 := fun nn => (~ nn == Root /\
                                 exists x, x < a /\ presentX x nn)
                                \/
                                (~ nn == Root /\ presentX a nn)); auto.
          apply Disjoint_union_Num_Card; auto.
          { intros nn1 nn2 enn; apply and_compat;
              [rewrite enn; tauto|].
            apply ex_compat; intros x.
            rewrite enn; tauto.
          }
          { intros nn1 nn2 enn; apply and_compat;
              rewrite enn; tauto. }
          { intros nn [hhr [x [hexx hexxx]]] [_ hp].
            assert (F := v_inv_unique hp hexxx).
            lia.
          }
          { intros nn1 nn2 enn; split; intros h.
            - destruct h as [hrr [x [hx hpx]]].
              case (eq_nat_dec x a); intros hhha.
              + subst; right; split; rewrite <- enn; tauto.
              + left; split; [rewrite <- enn; tauto|].
                exists x; split; [lia|rewrite <- enn; auto].
            - destruct h as [[hrr hex] | [hrr hp]].
              + split; [rewrite enn; tauto|].
                destruct hex as [x [hxa hpx]].
                exists x; split; [lia|rewrite enn; auto]. 
              + split; [rewrite enn; tauto|].
                exists a; split; [lia|rewrite enn; auto]. 
          }
        }
        assert (hone: Naa >= 1).
        { apply Card_larger_ge_iff.
          eapply Card_trans with (B := {x | ~ x == Root /\ presentX a x}).
          apply Card_weak, Card_sym, HNaa.
          apply Upa; lia.
        }
        eapply Card_trans with (B := ltN Nap).
        apply Card_weak, HNap.
        apply Card_larger_ge_iff; lia.
      }
    Qed.
    
    (** The number of non-root nodes with a present value smaller
        than N-1 is exactly N-1.

        Indeed, it is smaller than or equal to the number of non-root
        nodes, ie N-1 (see HNR)
        -- we use the implication of predicates and Inj_impl.
        It is also greater or equal due to Lemma Ua.
     **)
    Lemma cc:
      Num_Card Same {n | ~ n == Root /\
                         exists x, x < N-1 /\ presentX x n} (N-1).
    Proof.
      apply Card_antisym.
      eapply Card_trans with (B :=  {x | ~ x == Root }).
      apply Inj_impl.
      { intros nn1 nn2 enn h; rewrite <- enn; apply h. }
      apply Card_weak, HNR.
      apply Ua; lia. 
    Qed.

    Lemma interval_clas:
      forall (P: nat -> Prop) (n: nat) 
             (P_clas: forall x, x < n -> P x \/ ~P x),
        (exists x, x < n /\ P x) \/ (forall x, x < n -> ~P x).
    Proof.
      intros P n P_clas.
      pose (nn := n); fold nn.
      assert (hn: nn <= n) by lia.
      clearbody nn.
      induction nn as [| nn ih].
      - right; intros x hx; lia.
      - destruct (P_clas nn); [lia| |].
        + left; exists nn; split; [lia|auto].
        + assert (d: nn <= n) by lia.
          specialize (ih d); destruct ih as [ih|ih].
          * destruct ih as [x [hx hhx]].
            left; exists x; split; [lia|auto].
          * right; intros x hx.
            case (le_gt_dec nn x); intros hle.
            assert (x = nn) by lia; subst; auto.
            apply ih; lia.
    Qed.

    Lemma interval_clas':
      forall (P: nat -> Prop) (n: nat) 
             (P_clas: forall x, x < n -> ~P x \/ P x),
        (exists x, x < n /\ ~P x) \/ (forall x, x < n -> P x).
    Proof.
      intros P n P_clas.
      pose (nn := n); fold nn.
      assert (hn: nn <= n) by lia.
      clearbody nn.
      induction nn as [| nn ih].
      - right; intros x hx; lia.
      - destruct (P_clas nn); [lia| |].
        + left; exists nn; split; [lia|auto].
        + assert (d: nn <= n) by lia.
          specialize (ih d); destruct ih as [ih|ih].
          * destruct ih as [x [hx hhx]].
            left; exists x; split; [lia|auto].
          * right; intros x hx.
            case (le_gt_dec nn x); intros hle.
            assert (x = nn) by lia; subst; auto.
            apply ih; lia.
    Qed.

    (** Using cc above, we know that the number of non-root nodes
        with a present value x<N-1 is (N-1) If there exists a non-root
        node with a non-present value, then by disjoint union of the
        sets the number of non-root nodes with a present value x<N-1
        would be smaller than (N-1).

        This proof uses the fact that we can enumerate every natural
        number between 0 and N-2 in order to check if the value is
        present of not.  **)
    Lemma T1:
      forall n, ~ n == Root -> exists x, x < N-1 /\ presentX x n.
    Proof.
      intros n nr.
      destruct (All_In.All_In_exists_Card
                  all_nodes_prop (A_dec := Node_dec)
                  (P := fun nn => ~ nn == Root /\
                                  forall x, x < N-1 -> ~presentX x nn))
        as [NF HNF].
      { intros nn1 nn2 enn.
        apply and_compat; [rewrite enn; tauto|].
        apply all_compat; intros x.
        rewrite enn; tauto.
      }
      { intros nn.
        case (Node_dec nn Root); intros hnn.
        right; intuition.
        destruct (v_inv nn).
        case (le_gt_dec (N-1) x); intros hx.
        - left; split; auto.
          intros xx hxx.
          apply H; lia.
        - right; intros [_ F].
          specialize (F x hx).
          contradict F; apply H.
      }
      assert (NF = 0).
      { cut (N-1 = NF + (N-1)); [lia|].
        apply Subset_Num_Card_iff_eq with
            (PA1 := fun nn => ~ nn == Root)
            (PA2 := fun nn => (~ nn == Root /\
                               forall x, x < N - 1 -> ~presentX x nn)
                              \/
                              (~ nn == Root /\
                               exists x, x < N-1 /\ presentX x nn)).
        apply HNR.
        apply Disjoint_union_Num_Card; auto.
        { intros nn1 nn2 enn.
          apply and_compat; [rewrite enn; tauto|].
          apply all_compat; intros x.
          rewrite enn; tauto.
        }
        { intros nn1 nn2 enn.
          apply and_compat; [rewrite enn; tauto|].
          apply ex_compat; intros x.
          rewrite enn; tauto.
        }
        apply cc.
        { intros nn [nnr hf] [x [he [hhe hhhe]]].
          specialize (hf _ hhe); tauto.
        }
        { intros a b ab; split; intros h.
          {  destruct interval_clas with
                 (fun x => presentX x b) (N-1) as [hi|hi].
             { intros x hx; case (eq_nat_dec ((v (g Root) + x) mod k) (v (g b)));
                 intros he; [left|right]; apply he. }
             - right; destruct hi as [x [hx hhx]].
               split; [rewrite <- ab; auto|exists x; split; auto].
             - left; split; [rewrite <- ab; auto|intros x hx; auto].
          }
          { destruct h as [h|h]; rewrite ab; apply h. }
        }
      }

      destruct interval_clas with (fun x => presentX x n) (N-1) as [D|D]; auto.
      { intros x hx; case (eq_nat_dec ((v (g Root) + x) mod k) (v (g n)));
          intros he; [left|right]; apply he. }
      assert (NF >= 1).
      { apply Card_larger_ge_iff.
        eapply Card_trans
          with (B :=  {x | ~ x == Root /\
                           (forall x0, x0 < N - 1 -> ~ presentX x0 x)}).
        apply Card_weak, Card_sym, HNF.
        apply (Nonempty_Num_Card
                 (@exist Node
                         (fun x => ~ x == Root /\
                                   (forall x0, x0 < N - 1 -> ~ presentX x0 x))
                         _ (conj nr D))).
        unfold Proper; reflexivity. 
      }
      lia.
    Qed.

    Lemma T2:
      forall x, x < N-1 -> Num_Card Same {n | ~ n == Root /\
                                              presentX x n} 1.
    Proof.
      destruct interval_clas' with
          (fun x => Num_Card Same {n | ~ n == Root /\
                                       presentX x n} 1) (N-1)
        as [D|D]; auto.
      { intros x hx.
        assert (d: x < k) by lia.
        destruct (Upa_card d) as [Nx HNx]; clear d.
        case (eq_nat_dec Nx 1); intros hnx.
        right; subst; auto.
        left; contradict hnx.
        apply Subset_Num_Card_iff_eq with
            (PA1 := fun n => ~ n == Root /\ presentX x n)
            (PA2 := fun n => ~ n == Root /\ presentX x n); auto.
        { intros nn1 nn2 enn; rewrite enn; auto. }
      }
      { destruct D as [bi [hbi hhbi]].
        assert (d: bi < k) by lia.
        destruct (Upa_card d) as [Nbi HNbi]; clear d.
        assert (HHNb: Nbi > 1).
        { assert (Nbi >= 1).
          apply Card_larger_ge_iff.
          eapply Card_trans with
              (B := { n | ~ n == Root /\ presentX bi n}).
          apply Card_weak, Card_sym, HNbi.
          apply Upa; lia.
          cut (~Nbi <= 1); [lia|].
          contradict hhbi.
          eapply Card_trans with (B := ltN Nbi). 
          apply HNbi.
          apply Card_same_eq_iff; lia.
        }
        cut (forall a, S bi <= a -> a <= N-1 ->
                       Num_Card Larger
                                {n | ~ n == Root /\
                                     exists x, x < a /\ presentX x n} (S a)).
        { intros ok.
          assert (T := cc).
          assert (F: N-1 >= S (N-1)).
          apply Card_larger_ge_iff.
          eapply Card_trans with
              (B := { n | ~ n == Root /\ exists x, x < N-1 /\ presentX x n}).
          apply Card_weak, Card_sym, T.
          apply ok ; lia.
          contradict F; lia.
        }

        intros b hl hr.
        pose (a := b); fold a.
        assert (ha: a <= b) by lia.
        assert (ha': S bi <= a) by lia.
        clearbody a.
        induction a as [| a ih].
        {  lia. }
        case (eq_nat_dec a bi); intros hsa.
        { (* base case *)
          subst.
          destruct (Ua_card (S bi)) as [ni hni].
          destruct (Ua_card bi) as [nl hnl].
          eapply Card_trans with (B := ltN ni).
          apply Card_weak, hni.
          apply Card_larger_ge_iff.
          
          assert (ni = nl + Nbi).
          apply Subset_Num_Card_iff_eq with
              (PA1 := fun nn => ~ nn == Root /\
                                exists x, x < S bi /\ presentX x nn)
              (PA2 := fun nn => (~ nn == Root /\
                                 exists x, x < bi /\ presentX x nn)
                                \/
                                (~ nn == Root /\ presentX bi nn)); auto.
          apply Disjoint_union_Num_Card; auto.
          { intros nn1 nn2 enn; apply and_compat;
              [|apply ex_compat; intros x]; rewrite enn; tauto. }
          { intros nn1 nn2 enn; rewrite enn; auto. }
          { intros nn [nr [x [hx hhx]]]  [_ hp].
            assert (bi = x).
            apply (v_inv_unique (n := nn)); auto; lia.
            lia.
          }
          { intros nn1 nn2 enn; split; intros H.
            - destruct H as [nr [x [hx hhx]]].
              case (eq_nat_dec x bi); intros hxi.
              + subst; right; rewrite <- enn; auto.
              + left; split; [|exists x]; rewrite <- enn; auto.
                split; [lia|auto].
            - destruct H as [[nr [x [hx hhx]]]|[nr hp]].
              + split; [|exists x]; rewrite enn; auto.
              + split; [|exists bi]; rewrite enn; auto.
          }
          subst.
          assert (hnll: nl >= bi).
          { apply Card_larger_ge_iff.
            eapply Card_trans with
                (B := {n | ~ n == Root /\
                           exists x, x < bi /\ presentX x n}).
            apply Card_weak, Card_sym, hnl.
            apply Ua; lia.
          }
          lia.            
        }
        { (* induction step *)
          destruct (Ua_card (S a)) as [ns hs].
          eapply Card_trans with (B := ltN ns).
          apply Card_weak, hs.
          apply Card_larger_ge_iff.

          destruct (Ua_card a) as [na hna].
          assert (d: a < k) by lia.
          destruct (Upa_card d) as [ia hia]; clear d.
          assert (ns = na + ia).
          apply Subset_Num_Card_iff_eq with
              (PA1 := fun nn => ~ nn == Root /\
                                exists x, x < S a /\ presentX x nn)
              (PA2 := fun nn => (~ nn == Root /\
                                 exists x, x < a /\ presentX x nn)
                                \/
                                (~ nn == Root /\ presentX a nn)); auto.
          apply Disjoint_union_Num_Card; auto.
          { intros nn1 nn2 enn; apply and_compat;
              [|apply ex_compat; intros x]; rewrite enn; tauto. }
          { intros nn1 nn2 enn; rewrite enn; auto. }
          { intros nn [nr [x [hx hhx]]]  [_ hp].
            assert (a = x).
            apply (v_inv_unique (n := nn)); auto; lia.
            lia.
          }
          { intros nn1 nn2 enn; split; intros H.
            - destruct H as [nr [x [hx hhx]]].
              case (eq_nat_dec x a); intros hxi.
              + subst; right; rewrite <- enn; auto.
              + left; split; [|exists x]; rewrite <- enn; auto.
                split; [lia|auto].
            - destruct H as [[nr [x [hx hhx]]]|[nr hp]].
              + split; [|exists x]; rewrite enn; auto.
              + split; [|exists a]; rewrite enn; auto.
          }
          subst.

          assert (hhia: ia >= 1).
          { apply Card_larger_ge_iff.
            eapply Card_trans with
                (B := {n | ~ n == Root /\ presentX a n}).
            apply Card_weak, Card_sym, hia.
            apply Upa; lia.
          }

          assert (hhna: na >= S a).
          { apply Card_larger_ge_iff.
            eapply Card_trans with
                (B := {n | ~ n == Root /\
                           exists x, x < a /\ presentX x n}).
            apply Card_weak, Card_sym, hna.
            apply ih; lia.
          }
          lia.
        }
      }
    Qed.

    Lemma S_: forall n1 n2,
        ~ n1 == Root -> ~ n2 == Root ->
        v (g n1) = v (g n2) -> n1 == n2.
    Proof.
      intros n1 n2 nr1 nr2 hv.
      destruct (T1 _ nr1) as [x1 [hxx1 hx1]].
      destruct (T1 _ nr2) as [x2 [hxx2 hx2]].
      assert (x1 = x2).
      { unfold presentX in hx1, hx2.
        rewrite hv in hx1.
        destruct (@mod_plus_inv (v (g Root)) (v (g n2)) k).
        apply v_ltk.
        apply v_ltk.
        apply H0; auto; lia.
      }
      subst.
      
      eapply Singleton_Num_Card_unique.
      apply Card_antisym_converse.
      apply Card_Same_sym.
      apply T2; eauto. 
      split; [auto|apply hx1].
      split; [auto|apply hx2].
    Qed.


    Lemma last_to_move:
      has_moved_b g' g Root = false ->
      exists n, has_moved_b g' g n = true /\
                has_moved_b g' g (Succ n) = false.
    Proof.
      intros hmr.
      destruct (Step_one_has_moved_true step) as [x hx].
      cut (forall n,
              Between x n Root ->
              (forall nn, Between x nn n ->
                          has_moved_b g' g nn = true) \/
              exists nn, has_moved_b g' g nn = true /\
                         has_moved_b g' g (Succ nn) = false).
      { intros ok.
        destruct (ok _ (Between_self_r _ _)) as [okk|okk]; clear ok.
        specialize (okk _ (Between_self_r _ _)); congruence.
        destruct okk as [nn hnn]; exists nn; auto.
      }
      apply (Between_Succ_ind
               (P := fun n => (forall nn : Node, Between x nn n -> has_moved_b g' g nn = true) \/
                              (exists nn : Node,
                                  has_moved_b g' g nn = true /\
                                  has_moved_b g' g (Succ nn) = false))).
      { intros nn1 nn2 enn.
        apply or_compat.
        apply all_compat; intros X.
        apply impl_compat.
        apply Between_compat; auto.
        tauto.
        apply ex_compat; intros X; tauto.
      }
      { left; intros nn B.
        apply Between_self in B;
          rewrite <- (has_moved_b_compat pg' pg _ _ B); auto.
      }
      { intros n B [ih|ih].
        { case_eq (has_moved_b g' g (Succ n)); intros hm.
          - left; intros nn BB.
            destruct (Between_Pred_r BB) as [BBB|BBB].
            rewrite (has_moved_b_compat pg' pg _ _ BBB); auto.
            apply ih.
            eapply Between_rewrite_r in BBB;
              [|symmetry; apply Pred_Succ]; auto.
          - right; exists n; split; auto.
            apply ih.
            apply Between_self_r.
        }
        { right; auto. }        
      }
    Qed.

    Lemma Z_fully_decreases: Z g' <= N - 2.
    Proof.
      case_eq (has_moved_b g' g Root); intros hmr.
      { assert (nls: ~Leg_state g).
        intros F; apply Leg_state_pot_0 in F; auto;
          [lia|apply pg].
        assert (Zd := Z_decreases ass nls step hmr).
        lia. }
      
      (* root does not move: *)
      assert (vr: v (g Root) = v (g' Root)).
      { apply v_compat.
        rewrite has_moved_false in hmr; symmetry; auto. }
      
      destruct (last_to_move hmr) as [n [hmn hmsn]].
      assert (nr: ~ Root == n).
      { intros F;
          rewrite <- (has_moved_b_compat pg' pg _ _ F) in hmn;
          congruence. }

      case (eq_nat_dec (v (g n)) (v (g Root))); intros hvn.

      (* v (g n) = v (g Root) *)
      assert (vn': forall n', ~ n' == Root -> ~ n' == Succ Root ->
                              v (g Root) <> v (g' n')).
      { intros n' hr' hs'.
        rewrite <- hvn.
        case_eq (has_moved_b g' g n'); intros hmn'.
        { assert (~ Root == n') by (contradict hr'; symmetry; auto).
          assert (hp := has_moved_v_non_Root ass step _ H hmn').
          clear H.
          intros F; rewrite hp in F; apply S_ in F.
          apply Succ_compat in F; rewrite Succ_Pred in F.
          rewrite (has_moved_b_compat pg' pg _ _ F) in hmsn;
            congruence.
          contradict nr; symmetry; auto.
          contradict hs'; apply Pred_injective; rewrite Pred_Succ;
            auto.
        }
        { dup hmn' hhmn'; rewrite has_moved_false in hmn'.
          rewrite (v_compat hmn').
          case (Node_dec n' Root); intros nnr.
          { rewrite (v_compat (pg _ _ nnr)); auto. }
          intros F; apply S_ in F; auto.
          rewrite <- (has_moved_b_compat pg' pg _ _ F) in hhmn';
            congruence.
          contradict nr; symmetry; auto.
        }
      }

      { (* v (g n) = v (g Root) *)
        assert (C: convex g' (v (g' Root))).
        { intros n1 n2 B he.
          case (Node_dec n1 Root); intros h1r.
          { eapply Between_rewrite_r in B; [|symmetry; apply h1r].
            apply Between_self in B.
            apply v_compat, (Step_Proper_l step); symmetry; auto. }
          case (Node_dec n1 (Succ Root)); intros h1s.
          { eapply Between_rewrite_r in B; [|symmetry; apply h1s].
            case (Node_dec n2 Root); intros h2r.
            apply v_compat, (Step_Proper_l step); auto.
            case (Node_dec n2 (Succ Root)); intros h2s.
            rewrite <- he.
            apply v_compat, (Step_Proper_l step);
              rewrite h1s, h2s; auto.
            destruct (Between_Pred_r B); [tauto|].
            eapply Between_rewrite_r in H; [|symmetry; apply Pred_Succ].
            apply Between_self in H; contradict h2r; symmetry; auto.
          }
          specialize (vn' _ h1r h1s).
          lia.
        }
        rewrite <- pot_0_equiv in C; auto.
        2: apply pg'.
        2: eapply step_Assume; eauto.
        lia.
      }

      (* v (g n) <> v (g Root) *)
      assert (vn': forall n', v (g n) <> v (g' n')).
      { intros n'.
        case_eq (has_moved_b g' g n'); intros hmn'.
        { assert (nr': ~ Root == n').
          { intros F; rewrite <- (has_moved_b_compat pg' pg _ _ F)
              in hmn'; congruence. }
          assert (hp := has_moved_v_non_Root ass step _ nr' hmn').
          intros F; rewrite hp in F; apply S_ in F.
          apply Succ_compat in F; rewrite Succ_Pred in F.
          rewrite (has_moved_b_compat pg' pg _ _ F) in hmsn;
            congruence.
          contradict nr; symmetry; auto.
          intros FF; rewrite (v_compat (pg _ _ FF)) in F; tauto.
        }
        { dup hmn' hhmn'; rewrite has_moved_false in hmn'.
          rewrite (v_compat hmn').
          case (Node_dec n' Root); intros nnr.
          { rewrite (v_compat (pg _ _ nnr)); auto. }
          intros F; apply S_ in F; auto.
          rewrite <- (has_moved_b_compat pg' pg _ _ F) in hhmn';
            congruence.
          contradict nr; symmetry; auto.
        }
      }            
      
      { (* v (g n) <> v (g Root) *)          
        destruct (T1 n) as [xn hxn].
        contradict nr; symmetry; auto.
        cut (Z g' <> N - 1). 
        { symmetry in vr.
          assert (nls: ~Leg_state g).
          { intros F.
            apply Leg_state_pot_0 in F; auto.
            2: apply step.
            lia. }
          assert(X := Z_does_not_increase ass nls step). 
          lia. }
        intros F.
        assert (M: Z g' = Z g') by reflexivity.
        apply Z_min in M; auto.
        2: { apply pg'. }
        2: { eapply step_Assume; eauto. }
        destruct M as [_ M2].
        rewrite F in M2.
        destruct hxn as [hxn hxn'].
        specialize (M2 _ hxn).
        rewrite <- convex_b_false in M2.
        2: { apply pg'. }
        contradict M2.
        apply convex_equiv.
        { apply pg'. }
        right; intros nn.
        rewrite <- vr, hxn'.
        apply vn'.
      }
    Qed.
    
  End Z_Fully_Decreases.
  
End Dijkstra_Z.

Unset Implicit Arguments.
Close Scope nat_scope.
