(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- March 2017
*)

(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import ZArith.
From Coq Require Import Lia.
From Coq Require Import Bool.

(** **** Local Imports *)
From Padec Require Import OptionUtil.
From Padec Require Import SetoidUtil.
From Padec Require Import Algorithm.
From Padec Require Import Ring_topology.

(* end show *)

Open Scope nat_scope.
Open Scope signature_scope.
Set Implicit Arguments.

(**
 * Diskstra's first token ring algorithm
  - Proof of self-stabilization and specification
  - Proof of a complexity bound in steps
  - Proof that this complexity bound is exact (by building a worst case
    prefix of execution)
 **)

(**
 * The Algorithm
   
   /contains:/
   Dijkstra first algorithm for token circulation on a rooted ring

   /assume:/ Parameter k>1

   - Definition of the state of a process (ROState and State, equivalence,
     decidability, getter, compatibility)

   - Definition of the algorithm:

     -> function *simple_run* contains the algorithm as written in
        the original paper, using two arguments (the current state of the node
        and the current state of its predecessor);

     -> function *run* calls simple_run with the proper arguments, as
        awaited by PADEC in class Algorithm.
**)

Section Dijkstra_algo.

  Context {Chans: Channels}.

  Class Dijkstra_parameters :=
    {
      (** ** Parameter [k] of the algoritm: node values are incremented 
             modulo k *)
      k: nat;
      H_k: k > 0
    }.

  (** Assuming the Dijkstra parameter and hypothesis: *)
  Context {DP: Dijkstra_parameters}.
  Let H_k := H_k.
  
  (** ** State of a Node in the Algorithm *)
    
  (** *** Read Only Part of a State; local topology information
       It is made of
       - the channel to the predecessor node and
       - a Boolean value meaning that the node is a root in the ring
   *)
    Record Dijkstra_ROState := mkROState {
                                   predc: Channel;
                                   root: bool
                                 }.

    Notation eqROS := (fun (x y: Dijkstra_ROState) =>
                         predc x == predc y /\ root x = root y).

    Instance eqROS_equiv: Equivalence eqROS.
    Proof.
      split; hnf; intuition.
      transitivity (predc y); auto.
      transitivity (root y); auto.
    Qed.

    Global Instance Dijkstra_ROState_setoid: Setoid Dijkstra_ROState :=
      {|equiv := eqROS|}.

    (** *** State of a node
      - a natural counter, v, smaller than k 
      - the local topology read-only part as defined above
     *)

    Record Dijkstra_State := mkState {
                                 vltk: { n: nat | n <= k-1};
                                 Dijkstra_ROpart :> Dijkstra_ROState
                               }.

    Notation eqS := (fun (x y: Dijkstra_State) =>
      vltk x == vltk y /\ Dijkstra_ROpart x == Dijkstra_ROpart y).

    Instance eqS_equiv: Equivalence eqS.
    Proof.
      split; hnf; intuition;
      etransitivity; eauto.
    Qed.

    Global Instance State_setoid :Setoid Dijkstra_State := {|equiv := eqS|}.

    (** *** Getter for [vltk] *)
    Definition v (s: Dijkstra_State): nat := proj1_sig (vltk s).

    Lemma v_ltk: forall (s: Dijkstra_State), v s < k.
    Proof.
      intros [[V Vltk] ro]; unfold v; simpl; lia.
    Qed.

    Lemma v_plus_0: forall (s: Dijkstra_State), (v s + 0) mod k = v s.
    Proof.
      intros s.
      assert (V0: v s + 0 = v s) by lia; rewrite V0; clear V0.
      apply Nat.mod_small, v_ltk.
    Qed.

    (** *** Equality and Compatibility properties on State *)

    Lemma Dijkstra_ROState_dec: Decider (equiv (A := Dijkstra_ROState)).
    Proof.
      intros [p1 r1] [p2 r2]; simpl equiv.
      destruct (Channel_dec p1 p2);[|right;tauto].
      destruct (bool_dec r1 r2);[|right;tauto].
      left;tauto.
    Qed.

    Lemma Dijkstra_State_dec: Decider (equiv (A := Dijkstra_State)).
    Proof.
      intros [n1 ro1] [n2 ro2];simpl equiv.
      destruct (Dijkstra_ROState_dec ro1 ro2) as [h|h];simpl equiv in h;
        [|right;tauto].
      destruct (Nat.eq_dec (proj1_sig n1) (proj1_sig n2));[|right;tauto].
      left;tauto.
    Qed.

    Instance v_compat: Proper fequiv v.
    Proof.  intros [p1 r1] [p2 r2] [hp hr]; intuition. Qed.

    Instance predc_compat_RO: Proper fequiv predc.
    Proof.
      intros [p1 r1] [p2 r2] [hp hr]; simpl; intuition.
    Qed.

    Instance root_compat_RO: Proper fequiv root.
    Proof.
      intros [p1 r1] [p2 r2] [hp hr]; simpl; intuition.
    Qed.

    Instance Dijkstra_ROpart_compat: Proper fequiv Dijkstra_ROpart.
    Proof.
      intros [p1 r1] [p2 r2] [hp hr]; simpl in *; intuition.
    Qed.

    Instance predc_compat: Proper (equiv (A:=Dijkstra_State) ==> equiv) predc.
    Proof.
      intros s1 s2 es;apply predc_compat_RO;apply Dijkstra_ROpart_compat;auto.
    Qed.

    Instance root_compat: Proper (equiv (A:=Dijkstra_State) ==> equiv) root.
    Proof.
      intros [v1 ro1] [v2 ro2]; intro h;
        apply root_compat_RO, Dijkstra_ROpart_compat, h.
    Qed.

    Section Run_definition.

      (** "+1" on natural modulo k *)
      Program Definition Smod (nltk: {n: nat| n <= k-1}):
        {n: nat| n <= k-1} :=  (S (proj1_sig nltk) mod k).
      Next Obligation.
      Proof.
        assert (h: k <> 0) by lia.
        assert (hh := Nat.mod_upper_bound (S nltk) k h).
        lia.
      Qed.

      Global Instance Smod_compat: Proper fequiv Smod.
      Proof.
        intros [n1 hn1] [n2 hn2] en;hnf in *;simpl in *;
        apply f_equal2;lia.
      Qed.

      Lemma Smod_neq:
        k > 1 -> forall (v: {n: nat| n <= k-1}), ~v == (Smod v).
      Proof.
        intros hk.
        assert( H: k <> 0) by lia.
        intros [v hv];unfold Smod;simpl.
        intro h;hnf in h; simpl in h.
        destruct (@le_lt_eq_dec _ _ hv) as [LT|EQ]. 
        * rewrite Nat.mod_small in h by lia.
          lia.
        * subst.
          replace (S (k-1)) with k in h by lia.
          rewrite Nat.mod_same in h by lia.
          lia.
      Qed.

      Lemma Smod_neq_degenerated:
        k = 1 -> forall (v: {n: nat| n <= k-1}), v == (Smod v).
      Proof.
        intros hk.
        intros [v hv];unfold Smod;simpl; hnf; simpl.
        assert (v = 0) by lia.
        subst; rewrite hk.
        symmetry; apply Nat.mod_1_r.
      Qed.
        
    (** ** Run Function for the Algorithm
      - when the root value is equal to the value of its predecessor, 
        its value increments modulo k
      -  when the value of a non-root node is different from that of its 
         predecessor, the node gets the value of its predecessor

      Function simple_run contains the algorithm as written in papers, using 
      two arguments (the current state of the node and the current state of its
      predecessor).

      Function run_algorithm calls simple_run with the arguments, as awaited by
      PADEC in class Algorithm.
     *)

      Definition simple_run (s s_pred: Dijkstra_State): Dijkstra_State :=
         if root s then
           if Nat.eq_dec (v s) (v s_pred) then
             {| vltk := Smod (vltk s); Dijkstra_ROpart := s |}
           else
             s
         else
           if Nat.eq_dec (v s) (v s_pred) then
             s
           else
             {| vltk := vltk s_pred; Dijkstra_ROpart := s |}.

      Definition run_algorithm
                 (s: Dijkstra_State)
                 (peer_state: Channel -> option Dijkstra_State)
        : Dijkstra_State :=
        match peer_state (predc s) with
          None => s (* can't happen *)
        | Some s_pred => simple_run s s_pred
        end.

    End Run_definition.

    (** ** Properties of the Algorithm *)

    (** *** Compatibility of the Run Function *)
    Instance simple_run_compat:
      Proper fequiv simple_run.
    Proof.
      intros [v1 [predc1 root1]] [v2 [predc2 root2]] [hv [hpredc hroot]]
             [v1' ro1'] [v2' ro2'] [hv' hro'].
      unfold simple_run; unfold v in *; simpl in *.
      rewrite hroot, hv, hv'.
      destruct root2; destruct (Nat.eq_dec (proj1_sig v2) (proj1_sig v2'));
        try reflexivity; (split; [| split]; simpl; auto).
      unfold v; simpl.
      apply Smod_compat.
      congruence.
    Qed.

    Instance run_algorithm_compat: Proper fequiv run_algorithm.
    Proof.
      intros s s' es  ps ps' eps.
      unfold run_algorithm.
      assert (h := eps _ _ (predc_compat es)).
      destruct (ps (predc s)); destruct (ps' (predc s')); hnf in h; try tauto.
        apply simple_run_compat; assumption.
      Qed.
      
    (** ** Definition of The Algorithm *)
    Instance Dijkstra_algo: Algorithm :=
      {| State := Dijkstra_State; State_dec:= Dijkstra_State_dec; run := run_algorithm; |}.

    (** *** Stability for the Read-Only Part of the Run Function *)
    Lemma ROpart_stable:
      forall s sn, Dijkstra_ROpart (run_algorithm s sn) =~= Dijkstra_ROpart s.
    Proof.
      intros s sn.
      unfold run_algorithm, simple_run.
      destruct (sn (predc s)).
      2:simpl;auto.
      destruct (root s); destruct (Nat.eq_dec (v s) (v d)); simpl; auto;
        reflexivity.
    Qed.

    Global Instance Dijkstra_Stable:
      Stable_Variable Dijkstra_algo Dijkstra_ROpart.
    Proof.
      esplit.
      apply Dijkstra_ROpart_compat.
      apply ROpart_stable.
    Defined.
    
End Dijkstra_algo.

Unset Implicit Arguments.
Close Scope nat_scope.
Close Scope signature_scope.

