(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- March 2017
 *)

(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import ZArith.
From Coq Require Import Lia.
From Coq Require Import Decidable.

(** **** Local Imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import All_In.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Ring_topology.
From Padec Require Import Dijkstra_algo.
From Padec Require Import Self_Stabilization.

(* end show *)

Open Scope nat_scope.
Open Scope signature_scope.
Set Implicit Arguments.

(**
 * Diskstra's first token ring algorithm
  - Proof of self-stabilization and specification
  - Proof of a complexity bound in steps
  - Proof that this complexity bound is exact (by building a worst case
    prefix of execution)
 **)

(**
* Common Results

   /contains:/
   Common Definitions and Results about Dijkstra Token Ring
   Circulation

   /assumes:/ Parameter k>1; the network is a unidirectional rooted ring
   
   - Definition of the network assumption: it checks
     that the read-only variable of the state of each process exactly
     represents the ring topology defined by the network (channels
     coincide with Pred and Boolean values with the unique root).

   - Predicate Token (on a configuration and a node) holds when the
     node holds a token.  The predicate is proven compatible and
     decidable.  It holds at a node iff the node is enabled.  We also
     prove that for every configuration, there exists at least a node
     that holds a token (i.e., there exists at least one token in the
     ring).  It follows directly that in any configuration, the
     algorithm can perform a step of computation. Hence, no
     configuration is terminal.

   - Legtimate Configurations: configurations with a unique token.  As
     there always exists a token in the network, the legitimate
     configuration predicate, only checks token uniqueness. This
     predicate is compatible and weakly decidable. We characterize
     legitimate configurations (Lemma ls_charac): a configuration g is
     legitimate iff
     - either the root is the unique token holder and g contains only
       one value; or 
     - every node in the slice from the root to the predecessor of the
       token holder has the same value and another value covers all
       remaining nodes.

   - Degenerated rings (1 or 2 nodes): every configuration is legitimate.

The file also contains many utilitaries and intermediary results to
simplify the proofs. The most important ones among them are:
- When a node obtains a token during a step (meaning it did not
  own one before the step), this means that its predecessor moves
  during the step (Lemma token_obtained)
- When a node moves, and then still holds a token, its predecessor
  has also moved in the same step (Lemma two_tokens_move).
**)


(** * Common Definitions and Results about Dijkstra Token Ring Circulation
 **)
Section Common_defs.

  Context {Chans: Channels}
          {Net: Network}
          {RURN: Rooted_Unidirectional_Ring_Network Net}
          {DP: Dijkstra_parameters}.

  Notation Algo := Dijkstra_algo.
  Notation Root := (RURN_root (Net := Net)).
  Notation Ring := (URN_topo (Net := Net)).
  Notation Env := (Env (Net := Net) (Algo := Dijkstra_algo)).

  Notation N := (Ring_size Root).
  Hypothesis HNk: N <= k.

  (** ** Assumption for the Dijkstra algorithm
   Given by Ring_Network_assumption

   Checks that the read-only variable (channel, bool) exactly represents
   the ring topology defined by RURN; channel coincides with Pred and bool 
   with the unique root.
   *)

  Existing Instance Dijkstra_Stable.
#[refine] Instance Dijkstra_RO_assumption: Stable_Assumption (Algo := Dijkstra_algo)
                                                     (Net := Net)
    := {|
      Assume_RO :=
                (fun ro: Node -> Dijkstra_ROState => 
                   Ring_Network_assumption RURN
                                           (fun n: Node => root (ro n))
                                           (fun n: Node => predc (ro n))
                )|}.
  Proof.
    apply Dijkstra_ROpart.
    apply Dijkstra_Stable.
    intros ro1 ro2 ero.
    apply Ring_Network_assumption_compat.
    intros n1 n2 en; apply root_compat_RO.
    apply ero, en.
    intros n1 n2 en; apply predc_compat_RO.
    apply ero, en.
  Defined.    

  (** A simplification lemma: the RUN/run function can be replaced by
   simple_run and the direct use of information from the topology
   using Pred. *)
  Lemma RUN_simpl:
    forall g n, Proper pequiv g -> Assume_RO (ROEnv_part g) ->
                RUN (Algo:=Algo) g n == simple_run (g n) (g (Pred n)).
  Proof using Type.
    intros g n pg [hroot  hpredc ].
    specialize (hpredc n); simpl ROEnv_part in hpredc;
      unfold is_channel in hpredc.
    unfold RUN, Algo, run, run_algorithm.
    assert (local_env g n (predc (g n)) == Some (g (Pred n))).
    { unfold local_env.
      rewrite hpredc.
      reflexivity. }
    change (match local_env g n (predc (g n)) with
            | Some s_pred => simple_run (g n) s_pred
            | None => g n
            end == simple_run (g n) (g (Pred n))).
    destruct (local_env g n (predc (g n))); [|inversion H].
    apply simple_run_compat; [apply pg; reflexivity | inversion H; auto].
  Qed.

  (** ** Token
   This predicate expresses the fact that the node owns a token:
   - either the node is the Root and its value is the same as its predecessor
   - or else its value differs from the one of its predecessor

   Note that this exactly coincides with the fact that the node is 
   enabled (see Lemma [enabled_Token] above).
   *)
  Section Token.

    Definition Token (g: Env) (n: Node): Prop :=
      (Root == n /\  (v  (g n) = v (g (Pred n))))
      \/
      (~Root == n /\  (v (g n) <> v (g (Pred n)))).

    Lemma Token_compat: Proper fequiv Token.
    Proof using Type.
      intros g1 g2 eg n1 n2 en; apply or_compat;
        apply and_compat; try apply not_compat; try rewrite en; try tauto;
          (apply eq_compat;
           apply v_compat,eg;try apply Pred_compat; auto).
    Qed.

    (** Deciding whether a node holds a token or not is decidable! *)
    Lemma Token_dec: forall g n, { Token g n } + { ~Token g n }.
    Proof using Type.
      intros g n; unfold Token; destruct (Node_dec Root n);
        destruct (Nat.eq_dec  (v (g n)) (v (g (Pred n)))); intuition.
    Qed.

    (** A node holds a token iff it is enabled in the algorithm *)
    Lemma enabled_Token:
      N > 1 -> forall (g: Env),
          Proper fequiv g -> Assume g ->
          forall (n: Node), enabled g n <-> Token g n.
    Proof using HNk.
      intros hn g Step_Proper_r assume_g n.
      unfold enabled, enabled_b, RUN, run; simpl;
        unfold run_algorithm, simple_run, Token; simpl.
      destruct assume_g as [hroot hpred].
      unfold is_channel in hpred.
      simpl in *. unfold local_env.
      specialize (hpred n); specialize (hroot n).      
      destruct (peer n (predc (g n))) as [n'|]; simpl in *; try elim hpred.
      case
        (Dijkstra_State_dec
           (if root (g n)
            then
              if Nat.eq_dec (v (g n)) (v (g n'))
              then {| vltk := Smod (vltk (g n)); Dijkstra_ROpart := g n |}
              else g n
            else
              if Nat.eq_dec (v (g n)) (v (g n'))
              then g n
              else {| vltk := vltk (g n'); Dijkstra_ROpart := g n |})
           (g n)); intros h.
      { destruct (root (g n));
          destruct (Nat.eq_dec (v (g n)) (v (g n'))) as [EQ|NEQ].
        - symmetry in h; destruct h as [h _]; simpl vltk in h;
            contradict h; apply Smod_neq; lia.
        - split; try congruence; intros [[hh hhh]|[hh hhh]].
          contradict NEQ; rewrite hhh; symmetry;
            apply v_compat, Step_Proper_r; auto.  
          contradict hh; symmetry; rewrite <- hroot; congruence.
        - split; try congruence; intros [[hh hhh]|[hh hhh]].
          symmetry in hh; rewrite <- hroot in hh; congruence.
          contradict hhh; rewrite EQ; apply v_compat, Step_Proper_r; auto.
        - contradict NEQ; symmetry; apply h.
      }     
      { split; auto; intros _.
        destruct (root (g n));
          destruct (Nat.eq_dec (v (g n)) (v (g n'))) as [EQ|NEQ].
        - left; split; [symmetry; rewrite <- hroot; auto|].
          rewrite EQ; apply v_compat, Step_Proper_r; auto.     
        - contradict h; auto.
        - contradict h; auto.
        - right; split.
          intros F; symmetry in F; rewrite <- hroot in F; congruence.
          intros F; contradict NEQ; rewrite F;
            apply v_compat, Step_Proper_r; symmetry; auto.
      }
    Qed.

    (** In any configuration, there exists at least one token *)
    Lemma Token_exists:
      forall g, Proper fequiv g -> exists n, Token g n.
    Proof using Type.
      intros g Step_Proper_r.
      assert (contra: (forall n, ~Token g n) -> forall n, v (g n) = v (g Root)).
      {
        intro no_token.
        intro n; generalize (all_Between_Pred Root n).
        revert n; pattern (Pred Root) at 1;
          eapply (Ring_Succ_ind _ Root).
        intro n.
        rewrite Between_self.
        intro h; symmetry; apply v_compat, Step_Proper_r, h.
        intros n' hn'.
        intros n hn.
        destruct (Between_Pred_r hn).
        {
          destruct (Node_dec n Root) as [EQ|NEQ].
          apply v_compat, Step_Proper_r, EQ.
          rewrite H in NEQ.
          rewrite (v_compat (Step_Proper_r _ _ H)).
          erewrite <- (hn' n') by apply Between_self_r.
          specialize (no_token (Succ n')).
          unfold Token in no_token; intuition.
          rewrite not_not_iff in H0.
          {
            rewrite H0.
            apply v_compat, Step_Proper_r, Pred_Succ.
            intuition.
          }
          {
            hnf;
              destruct (Nat.eq_dec (v (g (Succ n'))) (v (g (Pred (Succ n')))));
              intuition.
          }
        }
        {
          apply hn'.
          revert H; eapply proj1, Between_compat; try reflexivity.
          apply Pred_Succ.
        }
        Unshelve.
        {
          intros n1 n2 en.
          apply all_compat; intro n.
          apply impl_compat.
          apply Between_compat;auto.
          tauto.
        }
      }
      destruct (All_In_dec_forall_exists all_nodes_prop
                                         (Token_compat Step_Proper_r)
                                         (Token_dec g)) as [ALL|EX]; auto.
      specialize (contra ALL (Pred Root)).
      split with Root; left; split; auto; reflexivity.
    Qed.

    (** It follows directly that in any configuration, the algorithm can 
        perform a step of computation.
     *)
    Lemma Step_exists:
      N > 1 -> forall (g: Env), Assume g -> Proper fequiv g ->
                                exists y, Step y g.
    Proof using HNk.
      intros hn g Assume_g Step_Proper_r.
      destruct Token_exists with g as [nn hnn]; auto.
      rewrite <- enabled_Token in hnn; auto.
      apply enabled_Step with nn;auto.
    Qed.

    Lemma has_moved_true_Token:
      N > 1 -> forall g (ass: Assume g) g' (step: Step g' g) n,
          has_moved_b g' g n = true -> Token g n.
    Proof using HNk.
      intros hn g ass g' step n hm.
      apply has_moved_b_enabled_b in hm; auto.
      apply enabled_Token; auto.
      apply (Step_Proper_r step).
    Qed.

    (** Hence, no configuration is terminal *)
    Lemma no_terminal:
      N > 1 -> forall (g: Env), Proper fequiv g -> Assume g ->
                                ~terminal g. 
    Proof using HNk.
      intros hn g pg ass.
      destruct (Step_exists hn ass pg) as [g' st]; auto.
      apply Step_not_terminal with g'.
      apply st.
    Qed.

    (** *** Constant Slice
     As long as there is no token between the [Root] and a given node [n], 
     the values of [v] between [Root] and [n] are all equal (to the value of, 
     e.g., [Root]).
    *)
    Lemma constant_slice:
      forall (g: Env),
        Proper fequiv g ->
        forall n, ~n == Root ->
                  (forall n', Between (Succ Root) n' n -> ~ Token g n') ->
                  v (g n) = v (g Root).
    Proof.
      intros g Step_Proper_r n.
      pattern n.
      eapply Ring_Succ_ind with (x := Succ Root); clear n.
      {
        intros n1 n2 en; apply impl_compat.
        apply not_compat, equiv_compat; reflexivity || auto.
        apply impl_compat.
        apply all_compat; intro n'; apply impl_compat; reflexivity || auto.
        apply Between_compat; reflexivity || auto.
        apply eq_compat; reflexivity || auto.
        apply eq_equivalence.
        symmetry;
        apply v_compat, Step_Proper_r,en.
      }
      {
        intro non_degen.
        intro h; specialize (h (Succ Root) (Between_self_lr _)).
        unfold Token in h; intuition.
        rewrite (v_compat (Step_Proper_r _ _ (Pred_Succ _))) in H0.
        destruct (Nat.eq_dec (v (g (Succ Root))) (v (g Root))) as [YES|NO];
          intuition.
      }
      {
        intros n hn hh h.
        destruct (Node_dec n Root) as [yes|no].
        {
          specialize (h (Succ Root) (Between_self_l _ _)).
          rewrite (v_compat (Step_Proper_r _ _ (Succ_compat _ _ yes))).
          unfold Token in h; intuition.
          rewrite yes in hh.
          rewrite (v_compat (Step_Proper_r _ _ (Pred_Succ _))) in H.
          destruct (Nat.eq_dec (v (g (Succ Root))) (v (g Root))) as [YES|NO];
            intuition.
        }
        {
          specialize (hn no).
          rewrite <- hn.
          specialize (h (Succ n) (Between_self_r _ _)).
          unfold Token in h; intuition.
          rewrite (v_compat (Step_Proper_r _ _ (Pred_Succ _))) in H0.
          destruct (Nat.eq_dec (v (g (Succ n))) (v (g n))) as [YES|NO];
            intuition.
          intros n' hn'.
          apply h.
          destruct (Between_Succ_r hn') as [hhh|hhh]; auto.
          contradict no; symmetry; eapply Succ_injective; auto.
        }
      }
    Qed.

  End Token.
  
  (** ** Enbled: 
   We provide small Lemmas to simplify the use of run_algorithm about enabled 
   nodes.
   *)
  Section Enabled.

    Variable (g: Env) (pg: Proper fequiv g) (Assume_g: Assume g).
    Hypothesis HN1: N > 1.

    Lemma enabled_Root:
      forall n, Root == n ->
                (enabled g n <-> v (g n) = v (g (Pred n))).
    Proof using Assume_g HN1 HNk pg.
      intros n nr; rewrite enabled_Token; auto.
      unfold Token; intuition.
    Qed.

    Lemma enabled_non_Root:
      forall n, ~Root == n ->
                (enabled g n <-> v (g n) <> v (g (Pred n))).
    Proof using Assume_g HN1 HNk pg.
      intros n nr; rewrite enabled_Token; auto.
      unfold Token; intuition.
    Qed.
    
  End Enabled.

  (** ** Move:
   We provide Lemmas to simplify the use of run_algorithm about moving nodes.
   *)
  Section Move.

    Variable (g: Env) (Assume_g: Assume g) (g': Env) (step: Step g' g).
    
    Lemma has_moved_v_change:
      forall n, has_moved_b g' g n = true <-> v (g n) <> v (g' n).
    Proof using step.
      intros n; split; [intros hm| intros neq].
      { apply has_moved_progress in hm.
        intros eqv; contradict hm.
        split.
        symmetry;apply eqv.
        apply (Stable_Variable_step Dijkstra_Stable); auto.
      }
      {
        case_eq (has_moved_b g' g n); auto; intros nhm; contradict neq.
        apply has_moved_false in nhm; auto.
        symmetry; apply nhm.      
      }
    Qed.

    Lemma has_moved_v_Root:
      forall n, Root == n ->
                has_moved_b g' g n = true ->
                v (g' n) = (v (g (Pred n)) + 1) mod k.
    Proof using Assume_g HNk step.
      intros n nr.
      assert (Step_Proper_r: Proper fequiv g) by apply step.
      assert (pg': Proper fequiv g') by apply (Step_Proper_l step).
      rewrite <- (has_moved_b_compat pg' Step_Proper_r _ _ nr),
      <- (v_compat (pg' _ _ nr)),
      <- (v_compat (Step_Proper_r _ _ (Pred_compat _ _ nr))).
      clear nr n pg'.
      intros hm.

      assert (hhm := has_moved_true _ step hm); auto;
        unfold RUN, run in hhm; simpl in hhm;
          unfold run_algorithm, simple_run in hhm; simpl in hhm;
            destruct hhm as [hv X].
      unfold has_moved_b in hm.
      destruct (State_dec (g' Root) (g Root)) as [d|d];
        try congruence; clear hm.
      destruct Assume_g as [hroot hpred]; simpl in hroot, hpred.
      unfold is_channel in hpred; unfold local_env in *.
      specialize (hpred Root); specialize (hroot Root).
      destruct (root (g Root)).
      2: contradict hroot; intuition.
      destruct (peer Root (predc (g Root))) as [pR |]; [| elim hpred];
        simpl in *.
      destruct (Nat.eq_dec (v (g Root)) (v (g pR))) as [|NEQ].
      - rewrite  (v_compat (Step_Proper_r _ _ hpred)) in e.
        rewrite <- e.
        unfold v; rewrite hv; simpl.
        assert (dd: S (proj1_sig (vltk (g Root))) =
                    (proj1_sig (vltk (g Root)) + 1)) by
            lia; rewrite dd; auto.
      - contradict d; split; auto.
    Qed.        

    Lemma has_moved_v_Root':
      N > 1 -> forall n, Root == n ->
                         has_moved_b g' g n = true ->
                         v (g (Pred n)) <> v (g' n).
    Proof using Assume_g HNk step.
      intros hn n nr hm.
      apply has_moved_v_Root in hm; auto.
      rewrite hm; clear hm.
      assert (d: v (g (Pred n)) + 1 = S (v (g (Pred n)))) by lia; rewrite d.
      apply Smod_neq.
      lia.
    Qed.

    Lemma has_moved_v_non_Root:
      forall n, ~ Root == n ->
                has_moved_b g' g n = true -> v (g' n) = v (g (Pred n)).
    Proof using Assume_g step.
      assert (forall s s_pred s', s' == simple_run s s_pred ->
                                  root s = false ->
                                  ~ s' == s ->
                                  v s' == v s_pred). 
      {
        intros s s_pred s' h' nr ns.
        unfold simple_run in h'.
        rewrite nr in h'.
        case_eq (Nat.eq_dec (v s) (v s_pred)); intros x y;
          rewrite y in h'; clear y.
        - contradict ns; auto.
        - rewrite (v_compat h'); simpl; auto.
      }
      intros n nr hm.
      destruct Assume_g as [hroot hpred]; simpl in hroot, hpred.
      specialize (hpred n); specialize (hroot n).
      apply H with (g n).
      { apply has_moved_true in hm; auto.
        rewrite hm; unfold RUN, run; simpl;
          unfold run_algorithm; simpl.        
        unfold is_channel in hpred; unfold local_env.
        destruct (peer n (predc (g n))) as [pn|]; simpl in *; [|elim hpred];
          simpl; auto.        
        assert (pg: Proper pequiv g) by apply step.
        assert (hp: g pn == g (Pred n)) by (apply pg, hpred).
        split; [| split].
        unfold eqSig; apply v_compat, simple_run_compat; auto.
        apply predc_compat, simple_run_compat; auto. 
        apply root_compat, simple_run_compat; auto. 
      }
      { destruct (root (g n));
        [contradict nr; symmetry; rewrite <- hroot |]; auto.
      }
      { unfold has_moved_b in hm;
        destruct (State_dec (g' n) (g n)) as [d|d];
        try congruence; auto.
      }
    Qed.

    (** When a node moves, and then still holds a token, its predecessor  
      has also moved in the same step.
     *)
    Lemma two_tokens_move:
      N > 1 -> forall n, has_moved_b g' g n = true ->
                         Token g' n ->
                         has_moved_b g' g (Pred n) = true.
    Proof using Assume_g HNk step.
      intros hn n hm tok'.
      assert (pg: Proper fequiv g) by apply step.
      assert (pg': Proper fequiv g') by apply (Step_Proper_l step).
      case_eq (has_moved_b g' g (Pred n)); auto; intros f.
      apply has_moved_false in f; auto.
      destruct f as [f _]. 
      assert (tok: Token g n) by 
          (rewrite <- enabled_Token; unfold enabled; auto;
           apply has_moved_b_enabled_b with g'; auto).
      dup hm hm'.
      apply has_moved_v_change in hm; auto.
      destruct (Node_dec Root n) as [r|r]; destruct tok as [[nr tok]|[nr tok]];
        destruct tok' as [[nr' tok']|[nr' tok']]; try tauto.
      - apply has_moved_v_Root' in hm'; auto.
        contradict hm'.
        unfold v in *.
        rewrite <- tok,tok', f; auto.
      - apply has_moved_v_non_Root in hm'; auto.
        unfold v in *.
        contradict tok'; rewrite f; auto.
    Qed.

    (** When a node obtains a token during a step (meaning it did not
     own one before the step), this means that its predecessor moves
     during the step. *)
    Lemma token_obtained:
      N > 1 -> forall n, ~Token g n -> Token g' n ->
                         has_moved_b g' g (Pred n) = true.
    Proof.
      intros hn n ntok tok'.
      assert (pg: Proper fequiv g) by apply step.
      assert (disabled: ~enabled g n).
      (intros F; rewrite enabled_Token in F; auto; tauto).
      case_eq (has_moved_b g' g n);
        [intros F; 
         apply has_moved_b_enabled_b in F; congruence|
         clear disabled; intros disabled].
      apply has_moved_false in disabled; auto.
      destruct disabled as [veq _].
      case_eq (has_moved_b g' g (Pred n)); [|intros F;
                                             apply has_moved_false in F]; auto.
      destruct F as [pveq _].
      unfold Token in ntok, tok'; contradict ntok.
      unfold v in *; rewrite <- veq, <- pveq; auto.
    Qed.

  End Move.
  
  (** ** Legitimate Configurations
     Configurations with a unique token (see [Leg_State]).
    
     As there always exists a token in the network 
     (see Lemma [Token_exists]), assuming token uniqueness is enough.
   *)
  Section Legitimate_Configuration.
    
    Definition Leg_state (g: Env): Prop :=
      forall n1 n2, Token g n1 -> Token g n2 -> n1 == n2.

    Instance Leg_state_compat: Proper (fequiv ==> iff) Leg_state.
    Proof.
      intros g1 g2 eg.
      unfold Leg_state.
      apply all_compat; intros n1; apply all_compat; intros n2.
      apply impl_compat.
      apply Token_compat; auto.
      apply impl_compat.
      apply Token_compat; auto.
      reflexivity.
    Qed.

    (** Deciding whether a configuration is legitimate or not is weakly 
        decidable *)
    Lemma Leg_state_dec:
      forall g, {Proper fequiv g -> Leg_state g} + {~Leg_state g}.
    Proof.
      intros g.
      pose (P := (fun n1 n2 => Token g n1 -> Token g n2 -> n1 == n2)).
      assert (P_compat: Proper fequiv g -> Proper fequiv P).
      { intros pg n1 n2 en n1' n2' en'; unfold P; split;
          intros h t1 t2.
        - rewrite <- en, <- en'.
          apply h.
          revert t1; apply Token_compat; auto; symmetry; auto.
          revert t2; apply Token_compat; auto; symmetry; auto.
        - rewrite en, en'.
          apply h.
          revert t1; apply Token_compat; auto; symmetry; auto.
          revert t2; apply Token_compat; auto; symmetry; auto.
      }
      assert (P_dec: forall n1 n2, {P n1 n2} + {~P n1 n2}).
      { intros n1 n2.
        destruct (Node_dec n1 n2).
        - left; intros _ _; auto.
        - destruct (Token_dec g n1) as [t1|t1].
          + destruct (Token_dec g n2) as [t2|t2].
            right; intros F; specialize (F t1 t2); auto.
            left; intros tt1 tt2; tauto.
          + left; intros tt1 tt2; tauto.
      }
      destruct (All_In_dec_forall_notforall_2
                  all_nodes_prop all_nodes_prop _ P_dec)
        as [ok|ok]; unfold Leg_state; unfold P in ok.
      - left; intros pg; apply ok, P_compat, pg.
      - right; apply ok.
    Qed.

    (** Charaterization of Legitimate *)
    Lemma about_ls1:
      forall g (pg: Proper pequiv g),
        Leg_state g ->
        exists n, Token g n /\
                  (forall n', Between Root n' (Pred n) ->
                              v (g Root) = v (g n')) /\
                  forall n', Between n n' (Pred Root) ->
                             v (g (Pred Root)) = v (g n').
    Proof.
      intros g pg ls.
      destruct (Token_exists pg) as [n tok].
      exists n; split; auto.
      case (Node_dec (Succ Root) Root); intros degen.
      { (* degenerated ring *)
        split; intros nn _; apply v_compat, pg.
        symmetry; apply Degenerate_ring_Succ; auto.
        apply Succ_injective; rewrite Succ_Pred.
        symmetry; apply Degenerate_ring_Succ; auto.
      }
      split. 
      { 
        generalize (Between_self_r (Succ Root) n).
        pattern n at 1 3.
        apply Ring_Succ_ind with (Succ Root).
        { intros n1 n2 en.
          apply impl_compat.
          apply Between_compat; auto.
          apply all_compat.
          intros x.
          apply impl_compat.
          apply Between_compat; auto.
          apply Pred_compat; auto.
          apply eq_compat; auto.
        }
        { intros _ nn hb.
          eapply Between_rewrite_r in hb;
            [|symmetry; apply Pred_Succ].
          apply Between_self in hb.
          apply v_compat, pg, hb. }
        { intros nn hi hb.        
          case (Node_dec Root nn); intros hnn.
          { intros n' hbb.
            eapply Between_rewrite_r in hbb;
              [|symmetry; apply Pred_Succ].
            eapply Between_rewrite_r in hbb; [|apply hnn].
            apply Between_self in hbb.
            apply v_compat, pg, hbb.        
          }
          case (Node_dec n nn); intros hn.
          { eapply Between_rewrite_r in hb;
              [| symmetry; apply hn].          
            contradict hb; apply Between_not_Succ_m.
            contradict hnn; apply Succ_injective; auto.
          }
          intros n' hbb.
          eapply Between_rewrite_r in hbb;
            [|symmetry; apply Pred_Succ].
          case (Node_dec n' nn); intros hn'.
          { clear hbb.
            apply Between_Pred_m in hb.
            destruct hb as [hb|hb];
              [contradict hnn; apply Succ_injective; auto|].
            eapply Between_rewrite_m in hb;
              [|symmetry; apply Pred_Succ].
            specialize (hi hb (Pred nn) (Between_self_r _ _)).
            rewrite hi, (v_compat (pg _ _ hn')).
            case (Token_dec g nn); intros toknn.
            contradict hn; apply ls; auto.
            unfold Token in toknn; intuition.
          }
          { apply hi.
            apply Between_Pred_m in hb; destruct hb as [hb|hb].
            contradict hnn; apply Succ_injective; auto.
            eapply Between_rewrite_m in hb;
              [|symmetry; apply Pred_Succ]; auto.          
            apply Between_Pred_r in hbb;
              destruct hbb as [hbb|hbb]; [tauto|auto].
          }
        }
      }
      { generalize (Between_self_l n (Pred Root)).
        pattern n at 2 3.
        apply Ring_Pred_ind with (Pred Root).
        { intros n1 n2 hn.
          apply impl_compat.
          apply Between_compat; auto.
          apply all_compat; intros x.
          apply impl_compat.
          apply Between_compat; auto.
          apply eq_compat; auto.
        }
        { intros _ n' hb; apply Between_self in hb.
          apply v_compat, pg, hb.
        }  
        { intros nn hi hb.
          case (Node_dec Root nn); intros hnn.
          { intros n' hbb.
            apply Pred_compat in hnn.
            eapply Between_rewrite_l in hbb;
              [|apply hnn ].
            apply Between_self in hbb.
            apply v_compat, pg, hbb.        
          }
          case (Node_dec n nn); intros hn.
          { eapply Between_rewrite_l in hb;
              [| symmetry; apply hn].          
            contradict hb; apply Between_not_Pred_m.
            contradict hnn; apply Pred_injective; auto.
          }          
          apply Between_Succ_m in hb.
          destruct hb as [hb|hb];
            [contradict hnn; symmetry; apply Pred_injective;
             auto|].
          eapply Between_rewrite_m in hb;
            [| symmetry; apply Succ_Pred].
          specialize (hi hb).
          intros n' hbb.
          case (Node_dec (Pred nn) n'); intros hn'.
          { specialize (hi nn (Between_self_l _ _)).
            rewrite hi, <- (v_compat (pg _ _ hn')).
            symmetry.
            case (Token_dec g nn); intros toknn.
            contradict hn; apply ls; auto.
            unfold Token in toknn; intuition.                  
          }
          { apply hi.
            apply Between_Succ_l in hbb; destruct hbb as [hbb|hbb]; [tauto|].
            eapply Between_rewrite_l in hbb;
              [| symmetry; apply Succ_Pred]; auto.
          }        
        }
      }
    Qed.

    Lemma about_ls2:
      forall g (pg: Proper pequiv g),
        (exists n, Token g n /\
                   (forall n', Between Root n' (Pred n) ->
                               v (g Root) = v (g n')) /\
                   forall n', Between n n' (Pred Root) ->
                              v (g (Pred Root)) = v (g n'))
        -> Leg_state g.
    Proof.
      intros g pg [n [tok [val1 val2]]].
      cut (forall nn, Token g nn -> n == nn).
      { intros ok n1 n2 tok1 tok2.
        rewrite <- (ok _ tok1), <- (ok _ tok2); auto.
      }
      intros nn toknn.
      (* if root <> n, two values in the ring:
       - either nn and (Pred nn) share the same value 
         (either val1 or val2) 
       if root == n, a unique value in the ring
       - val1 only
       *)
      case (Node_dec n Root); intros hnr.
      { assert (vnn: v (g Root) = v (g nn)).
        apply val1.
        symmetry in hnr; apply (Between_rewrite_l _ hnr). 
        apply all_Between_Pred.
        assert (vpnn: v (g Root) = v (g (Pred nn))).
        apply val1.
        symmetry in hnr; apply (Between_rewrite_l _ hnr). 
        apply all_Between_Pred.
        rewrite vnn in vpnn; clear vnn.
        rewrite hnr; clear hnr.
        unfold Token in toknn; intuition.
      }
      assert (xpn := val1 (Pred n) (Between_self_r _ _)).
      assert (xn := val2 n (Between_self_l _ _)).
      case (Node_dec Root nn); intros hr.
      { apply (Token_compat pg _ _ hr) in toknn.
        unfold Token in *; intuition. }
      destruct (Between_alt_m Root nn n) as [B|[B|B]].
      { contradict hnr; symmetry; auto. }
      { apply Between_Pred_r in B.
        destruct B as [B|B]; symmetry; auto.      
        assert (x1 := val1 nn B).
        apply Between_Pred_m in B; destruct B as [B|B]; [tauto|].
        assert (x2 := val1 _ B). 
        unfold Token in toknn; intuition.
      }
      { apply Between_Pred_r in B.
        destruct B as [B|B]; [symmetry in B; tauto|].      
        assert (x1 := val2 _ B).
        apply Between_Pred_m in B; destruct B as [B|B]; [tauto|].
        assert (x2 := val2 _ B). 
        unfold Token in toknn; intuition.
      }
    Qed.

    Lemma ls_charac:
      forall g (pg: Proper pequiv g),
        Leg_state g <->
        (exists n, Token g n /\
                   (forall n', Between Root n' (Pred n) ->
                               v (g Root) = v (g n')) /\
                   forall n', Between n n' (Pred Root) ->
                              v (g (Pred Root)) = v (g n')).
    Proof.
      intros g pg; split; [apply about_ls1|apply about_ls2];
        auto.
    Qed.

    (** In a legitimate configuration, the node preceding the token holder
        has the same value as the root.
        (This is due to the [convex_part] Lemma and to the fact that the token 
        holder is unique.)
     *)
    Lemma legit_convex:
      forall g, Proper fequiv g -> Leg_state g ->
                forall n, Token g n ->  v (g (Pred n)) = v (g Root).
    Proof.
      intros g pg lg n tn.
      dup lg ls; rewrite ls_charac in ls; auto.
      destruct ls as [n' [t' [X _]]].
      rewrite X with (Pred n').
      apply v_compat, pg, Pred_compat, lg; auto.
      apply Between_self_r.
    Qed.      
                                                     
  End Legitimate_Configuration.

  Section Degenerated_Rings.

    (** ** Particular case: the ring contains a unique node
      The root has a token.
      Every configuration is legitimate, hence the stabilization time
      is 0 
     *)
    Lemma Single_Node_has_token:
      N = 1 -> forall g n, Proper pequiv g -> Token g n.
    Proof using Type.
      intros H1 g n pg; unfold Token.
      left; split.
      eapply Degenerate_ring_one; apply H1.
      apply v_compat, pg.
      eapply Degenerate_ring_one; apply H1.
    Qed.
    
    Lemma Single_Node_Legitimate:
      N = 1 -> forall g, Proper pequiv g -> Leg_state g.
    Proof.
      intros h1 g pg.
      assert (all_Root: forall n, n == Root).
      { intros n. eapply Degenerate_ring_one, h1. }
      apply ls_charac; auto.
      exists Root; repeat apply conj;
        [left; split; auto | intros n' _ | intros n' _];
      apply v_compat, pg; (try rewrite all_Root at 1); symmetry; apply all_Root.
    Qed.

    (** ** Particular case: the ring contains only two nodes
      Every configuration is legitimate, hence the stabilization time
      is 0 
     *)
    Lemma Two_Nodes_Legitimate:
      N = 2 -> forall g, Proper pequiv g -> Leg_state g.
    Proof.
      intros h.
      assert (NN: forall n, n == Root \/ n == Succ Root).
      { intros n.
        assert (D := Dist_lt_Ring_size Root Root n).
        rewrite h in D.
        assert (DD: Dist Root n = 0 \/ Dist Root n = 1) by lia.
        destruct DD as [DD|DD]; [left|right].
        - symmetry; apply Dist_refl in DD; auto.
        - destruct (Dist_Succ_l Root n) as [DR|DR]; [|lia].
          symmetry; apply Dist_refl; lia.
      }
      assert (NS: ~ Root == Succ Root).
      { apply Non_Degenerated_size_Succ with Root; lia. }
      assert (NP: Pred Root == Succ Root).
      { specialize (NN (Pred Root)).
        destruct NN as [NN|NN]; auto.
        contradict NS; apply Pred_injective; rewrite Pred_Succ; auto.
      }
      intros g pg; apply ls_charac; auto.
      assert (NPP := NP).
      apply pg, v_compat in NP.
      destruct (eq_nat_dec (v (g Root)) (v (g (Pred Root)))) as [EQ|NE].
      {
        exists Root;split.
        { left; split; auto. }
        split; intros n bn; destruct (NN n) as [hh|hh];
          apply pg, v_compat in hh.
        - symmetry; auto.
        - rewrite EQ, NP; symmetry; auto.
        - rewrite hh; symmetry; auto.
        - rewrite hh; symmetry; auto.
      }
      {
        exists (Succ Root); split.
        { right; split; auto.
          intro hh; apply NE; rewrite NP, hh.
          apply v_compat, pg.
          symmetry; apply Pred_Succ.
        }
        split; intros n bn; destruct (NN n) as [hh|hh];
          apply pg,v_compat in hh.
        - symmetry; auto.
        - contradict NE.
          assert (hn: n == Root).
          {
            symmetry; apply Between_self.
            apply Between_rewrite_r with (Pred (Succ Root)); auto.
            symmetry; apply Pred_Succ.
          }
          apply pg, v_compat in hn.
          rewrite NP, <- hh,hn; auto.
        - assert (hn: n == Succ Root).
          {
            symmetry; apply Between_self.
            apply Between_rewrite_r with (Pred Root); auto.
            symmetry; apply NPP.
          }
          apply pg, v_compat in hn.
          rewrite hn; apply NP.
        - rewrite hh; auto.
      }
    Qed.
    
    (** *** The ring is not degenerated *)
    Lemma Non_degen2:
      N >= 2 -> forall x, ~ Succ x == x.
    Proof.
      intros HN3 x F.
      apply Pred_compat in F.
      rewrite Pred_Succ in F.
      apply Dist_refl in F.      
      assert (D := Dist_Pred_full Root x).
      lia.
    Qed.
    
    Lemma Non_degen3:
      N >= 3 -> forall x, ~ Succ (Succ x) == x.
    Proof.
      intros HN3 x F.
      apply Pred_compat in F.
      rewrite Pred_Succ in F.
      apply Dist_refl in F.      
      assert (D := Dist_Pred_full Root x).
      destruct (Dist_Succ_l x (Pred x)).
      lia.
      apply Dist_refl, Succ_compat in H.
      rewrite Succ_Pred in H.
      contradict H; apply Non_degen2; lia.
    Qed.
    
  End Degenerated_Rings.

  Section Costs.

    Definition Dijkstra_exact_bound (N: nat): nat :=
      3 * (N * (N - 1) / 2) - N - 1.
    
  End Costs.

End Common_defs.

Unset Implicit Arguments.
Close Scope nat_scope.
Close Scope signature_scope.
