(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- March 2017
 *)

(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Bool.
From Coq Require Import ZArith.
From Coq Require Import Lia.

(** **** Local Imports *)
From Padec Require Import Stream.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Exec.
From Padec Require Import Ring_topology.
From Padec Require Import Dijkstra_algo.
From Padec Require Import Dijkstra_common.
From Padec Require Import Dijkstra_Z.
From Padec Require Import Dijkstra_sum_dist.
From Padec Require Import Self_Stabilization.
From Padec Require Import Steps.
From Padec Require Import NatUtils.
From Padec Require Import SetoidUtil.
From Padec Require Import Count.
From Padec Require All_In.
From Padec Require FoldRight.
From Padec Require Import ListUtils.

(* end show *)

Set Implicit Arguments.
Open Scope nat_scope.

(**
 * Diskstra's first token ring algorithm
  - Proof of self-stabilization and specification
  - Proof of a complexity bound in steps
  - Proof that this complexity bound is exact (by building a worst case
    prefix of execution)
 **)

(**
* Complexity Bound

   /contains:/
   Proof of the complexity bound
   - gathers previous results from sum_dist, Z

   /assumes:/ Parameter k>1; the network is a unidirectional rooted
   ring; N <= k

   [Case N >= 4]
   - Definition of the cost using sum_dist and Z
   - Cost 0 means legitimate configuration
   - Cost bound
   - Cost decreasing

   [General case]
 **)

Section Dijkstra_complexity.

  Context {Chans: Channels}
          {Net: Network}
          {RURN: Rooted_Unidirectional_Ring_Network Net}
          {DP: Dijkstra_parameters}.

  Notation Algo := Dijkstra_algo.
  Notation Root := (RURN_root (Net := Net)).
  Notation Ring := (URN_topo (Net := Net)).

  Existing Instance Dijkstra_RO_assumption.

  Notation N := (Ring_size Root).
  Hypothesis HNk: N <= k.

  (** * Proof for non-degenerated cases (N >= 4) *)
  Section Non_Degenerated.
    
    Hypothesis HN4: N >= 4.
    
    Lemma HN1: N > 1. Proof using HN4 HNk. lia. Qed.

    (** ** Cost *)

    (** The cost of a configuration: *)
    Definition cost (g: Env): nat := sum_dist g + (Z g) * N - 2.

    (** sum_dist g + (Z g) * N < 3 implies a legitimate configuration *)
    Lemma cost_0_leg_state:
      forall g, Proper pequiv g -> Assume g ->
                sum_dist g + (Z g) * N < 3 -> Leg_state g.
    Proof.
      intros g pg ass H.
      assert (hz: Z g = 0) by nia.
      assert (hsd: sum_dist g < 3) by nia.
      clear H.
      intros n1 n2 t1 t2.
      case (Node_dec n1 n2); intros hn; auto.
      assert (M := minus2 HNk pg hn t1 t2). 
      rewrite pot_0_equiv in hz; auto.
      unfold convex in hz.
      case (Node_dec n1 Root).
      { intros hnr.
        rewrite (Token_compat pg _ _ hnr) in t1.
        unfold Token in t1; intuition.
        symmetry in H1.
        assert (forall n', v (g n') = v (g Root)).
        { intros n'; apply hz with (Pred Root); auto.
          apply all_Between_Pred. }
        unfold Token in t2.
        rewrite (H n2), (H (Pred n2)) in t2.
        intuition.
        rewrite hnr, H3; auto.
      }
      case (Node_dec n2 Root).
      { intros hnr.
        rewrite (Token_compat pg _ _ hnr) in t2.
        unfold Token in t2; intuition.
        symmetry in H1.
        assert (forall n', v (g n') = v (g Root)).
        { intros n'; apply hz with (Pred Root); auto.
          apply all_Between_Pred. }
        unfold Token in t1.
        rewrite (H n1), (H (Pred n1)) in t1.
        intuition.
      }
      intros h1 h2.
      assert (dd: Dist n1 Root <> Dist n2 Root).
      { intros FF; apply Dist_unique in FF.
        intuition. }
      assert (d1: Dist n1 Root <> 0).
      { intros FF; apply Dist_refl in FF; auto. }
      assert (d2: Dist n2 Root <> 0).
      { intros FF; apply Dist_refl in FF; auto. }
      lia.
      lia.
    Qed.
    
    (** Cost is bounded, using bounds for Z and sum_dist *)
    Lemma cost_bound:
      forall g, Assume g -> Proper pequiv g ->
                Z g <= N - 2 ->
                cost g <= Dijkstra_exact_bound N - 1.
    Proof.
      intros g ass pg hZ.
      unfold Dijkstra_exact_bound.
      
      cut (2 * cost g + 2 * N + 4 <= 3 * N *(N - 1)).
      2: { unfold cost.
           assert (SD := sum_dist_bound HN4 HNk pg). 
           nia.
      }
      intros ok.

      assert (N > 0) by lia.
      destruct (arith_even H) as [k hk]; clear H.
      assert (k >= 3) by nia.
      rewrite hk.
      assert (2 * k / 2 = k).
      { assert (2 * k = k * 2) by nia.
        rewrite H0; clear H0.
        rewrite Nat.div_mul; lia. }
      rewrite H0; clear H0.
      nia.
    Qed.

    (** Cost decreases *)
    Lemma cost_decreases:
      forall g, Assume g -> ~Leg_state g ->
                forall g', Step g' g -> cost g' < cost g.
    Proof.
      intros g ass nls g' step.
      assert (pg: Proper pequiv g) by apply step.
      case (le_gt_dec 3 (sum_dist g + (Z g) * N)); intros X.
      2: { apply cost_0_leg_state in ass; auto; tauto. }
      unfold cost.
      assert (HN1: N > 1) by lia.
      assert (Zd := Z_evolves HNk HN1 ass nls step).
      assert (Sd := sum_dist_evolves HN4 HNk ass step).
      destruct (has_moved_b g' g Root); nia.
    Qed.  

    Lemma Dijkstra_exact_bound_g0:  Dijkstra_exact_bound N > 0.
    Proof.
      assert (N > 0) by lia.
      destruct (arith_even H) as [k hk]; clear H.
      assert (k >= 3) by nia.
      unfold Dijkstra_exact_bound.
      rewrite hk.
      assert (2 * k / 2 = k).
      { assert (2 * k = k * 2) by nia.
        rewrite H0; clear H0.
        rewrite Nat.div_mul; lia. }
      nia.
    Qed.      
    
    Theorem Dijkstra_bound_non_degenerated_cases:
      Steps_complexity (Dijkstra_exact_bound N)
                       (fun e => Leg_state (s_hd e))
                       (fun e => Assume (s_hd e) /\ is_exec e).
    Proof.
      eapply Steps_complexity_Bound_Cost with
          (cost := fun e =>
                     if (Nat.eq_dec (Z (s_hd e)) (N - 1)) then
                       Dijkstra_exact_bound N
                     else cost (s_hd e)).
      { intros g e [ass ise]; split.
        - eapply step_Assume.
          apply ass.
          apply is_exec_inv_1 in ise; auto.
        - apply is_exec_inv_2 in ise; auto. }
      { intros e [ass ise].
        destruct (Leg_state_dec (s_hd e)).
        - left; apply l.
          apply hd_compat; apply is_exec_proper in ise; auto.
        - right; auto. }
      { intros g [ass ise].
        assert (pg := is_exec_proper ise); inversion pg; subst.
        apply is_exec_inv_0 in ise.
        apply no_terminal in ass; auto; [tauto|lia]. }
      (* cost decreases *)
      { intros g e [ass ise] nls.
        assert (step := is_exec_inv_1 ise).
        case (Nat.eq_dec (Z g) (N - 1)); intros HN.
        { assert (ZF := Z_fully_decreases HNk HN1 HN4 ass step HN).
          case (Nat.eq_dec (Z (s_hd e)) (N - 1)); intros HHN; [lia|].
          assert (ass': Assume (s_hd e)) by
              (apply step_Assume with g; auto).
          assert (pg': Proper pequiv (s_hd e)) by apply step.
          assert (CC := cost_bound ass' pg' ZF).
          assert (D0 := Dijkstra_exact_bound_g0).
          lia.
        }
        { case (Nat.eq_dec (Z (s_hd e)) (N - 1)); intros HHN.
          { assert (hz := Z_does_not_increase HNk HN1 ass nls step).
            assert (pg: Proper pequiv g) by apply step.
            assert (hhz := Z_lt_N HNk HN1 ass pg).
            simpl s_hd in hz, hhz.
            lia.
          }
          apply cost_decreases; auto.        
        }
      }
      (* cost 0 means legitimate configuration *)
      { intros e [ass ise].
        case (Nat.eq_dec (Z (s_hd e)) (N - 1)).
        { intros _ y.
          assert (D0 := Dijkstra_exact_bound_g0).
          lia. }
        intros _ c0.
        apply cost_0_leg_state; auto.
        apply hd_compat, (is_exec_proper ise).
        unfold cost in c0; lia.
      }
      (* bounded cost *)
      { intros [g | g e] [ass ise];
        assert (pg := is_exec_proper ise);
        inversion pg; subst.
        { apply is_exec_inv_0 in ise.
          apply no_terminal in ass; auto; [tauto|lia]. }
        case (Nat.eq_dec (Z g) (N - 1)); intros HN.
        { lia. }
        { assert (Z g <= N - 1) by (apply Z_lt_N; auto; lia).
          assert (X: Z g <= N -2) by lia.
          assert (CC := cost_bound ass H2 X).
          lia. }
      }
    Qed.

  End Non_Degenerated.
  
End Dijkstra_complexity.

Unset Implicit Arguments.
Close Scope nat_scope.
