(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- March 2017
 *)

(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Bool.
From Coq Require Import ZArith.
From Coq Require Import Lia.

(** **** Local Imports *)
From Padec Require Import Stream.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Exec.
From Padec Require Import Ring_topology.
From Padec Require Import Dijkstra_algo.
From Padec Require Import Dijkstra_common.
From Padec Require Import Dijkstra_Z.
From Padec Require Import Dijkstra_sum_dist.
From Padec Require Import Self_Stabilization.
From Padec Require Import Steps.
From Padec Require Import NatUtils.
From Padec Require Import SetoidUtil.
From Padec Require Import Count.
From Padec Require All_In.
From Padec Require FoldRight.
From Padec Require Import ListUtils.

(* end show *)

Set Implicit Arguments.
Open Scope nat_scope.

(**
 * Diskstra's first token ring algorithm
  - Proof of self-stabilization and specification
  - Proof of a complexity bound in steps
  - Proof that this complexity bound is exact (by building a worst case
    prefix of execution)
 **)

(**
 * Complexity Bounds for N= 1 to 3 nodes

   /contains:/
   I  - Proof of the complexity bound (0 steps) for N=1 or 2
   II  - Proof of the complexity bound (3 maximum steps) for N=3

   /assumes:/ Parameter k>1; the network is a unidirectional rooted
   ring; N <= k

   Part I: DTRC Complexity for N=1 or 2 is 0 step
   obtained directly since every configuration is legitimate

   Part II: DTRC complexity for N=3 is 3 steps at most

   - Basic facts about the 3-ring (nodes N0(Root)  N1 N2)
   - Elementary characterization of DTRC states for N=3
     * [L3_state] : 3 non-legitimate steps may be performed (but no more)
     * [L2_state] : 2 non-legitimate steps may be performed (but no more)
     * [L1_state] : 1 non-legitimate step  may be performed (but no more)
     * [L0_state] : no non-legitimate step may be performed <-> [Leg_state]
   - These predicates form a partition of DTRC state for N=3
   - Characterization of available transitions:
     * [L3_Step_L2_L0] : A [L3_state] leads to either a [L2_state] or a [L0_state]
     * [L2_Step_L1_L0] : A [L2_state] leads to either a [L1_state] or a [L0_state]
     * [L1_Step_L0]    : A [L1_state] leads to either a [L0_state]
   - Decomposition of possible trace based of the above partition and transition graph

  Detailled transition table: (!) means node moves (.) means no move
  [L3State] == ABA+ (+) means A+1 mod k == B
  - transitions : 3 tokens -> 7 choices
     !!! BAB- !!. BAA !.! BBB !.. BBA .!! AAB .!. AAA ..! ABB
  [L2State] == ABA- (-) means  A+1 mod k <> B
  - transitions : 3 tokens -> 7 choices
     !!! CAB !!. CAA !.! CBB !.. CBA .!! AAB .!. AAA ..! ABB
  [L1State] == ABC
  Transitions: 2 tokens -> 3 choices
     .!! AAB .!. AAC ..! ABB
  [L0State] == [Leg_state]
              /ABB : 1 token -> .!. ABB
  3 subcase  { AAB : 1 token -> ..! AAA
              \AAA : 1 token -> !.. BAA

 **)

(* ** Tools for decidability : should exist already !! *)
(* TODO move elsewhere *)

Notation Decidable A:= ({A} + {~A}) (only parsing).

Lemma or_dec: forall A B, Decidable A -> Decidable B -> Decidable (A \/ B).
Proof using Type.
  intros A B [TTA|NNA] [TTB|NNB]; ((left;tauto)||(right;tauto)).
Qed.
Lemma and_dec: forall A B, Decidable A -> Decidable B -> Decidable (A /\ B).
Proof using Type.
  intros A B [TTA|NNA] [TTB|NNB]; ((left;tauto)||(right;tauto)).
Qed.
Lemma not_dec: forall A, Decidable A -> Decidable (~A).
Proof using Type.
  intros A [TTA|NNA]; ((left;tauto)||(right;tauto)).
Qed.

(* ** Tools for commuting function application with if..then..else :
      should exist already !! *)
(* TODO move elsewhere *)

Lemma ITE_bool_f : forall A B (f:A -> B) (b:bool) a1 a2,
    f (if b then a1 else a2) = if b then (f a1) else (f a2).
Proof using Type.
  intros;case b;reflexivity.
Qed.

Lemma ITE_sumbool_f : forall PA PB A B (f:A -> B) (sb:{PA}+{PB}) a1 a2,
    f (if sb then a1 else a2) = if sb then (f a1) else (f a2).
Proof using Type.
  intros;case sb;reflexivity.
Qed.

(* ** specilized artihmetic results for N=3 *)
(* TODO decide what to do .. maybe should stay here *)

Lemma Smod3: forall p n : nat, 3 <= p -> S n mod p <> n.
Proof using Type.
  intros p n hp.
  rewrite <- Nat.add_1_l.
  intro h.
  assert (hn:n<p).
  {
    rewrite <- h.
    apply Nat.mod_upper_bound ;try  lia.
  }
  assert (hsn:(1+n)<=p) by lia.
  destruct (le_lt_eq_dec hsn) as [LT | EQ].
  - rewrite Nat.mod_small in h by apply LT.
    lia.
  - rewrite EQ,Nat.mod_same in h by lia.
    lia.
Qed.

Lemma SSmod3: forall p n : nat, 3 <= p -> S (S n mod p) mod p <> n.
Proof using Type.
  intros p n hp.
  rewrite <- Nat.add_1_l.
  rewrite Nat.add_mod_idemp_r by lia.
  replace (1+S n) with (2+n) by reflexivity.
  intro h.
  assert (hn:n<p).
  {
    rewrite <- h.
    apply Nat.mod_upper_bound ;try  lia.
  }
  assert (hssn':(2+n)<=1+p) by lia.
  destruct (le_lt_eq_dec hssn') as [LT | EQ].
  - assert (hssn:(2+n)<=p) by lia.
    clear LT.
    destruct (le_lt_eq_dec hssn) as [LT | EQ].
    -- rewrite Nat.mod_small in h by apply LT.
       lia.
    -- rewrite EQ,Nat.mod_same in h by lia.
       lia.
  - rewrite EQ,<- Nat.add_mod_idemp_r,Nat.mod_same,Nat.mod_small in h by lia.
    simpl in h.
    lia.
Qed.

Section Dijkstra_complexity_1_2_3.

  Context {Chans: Channels}
          {Net: Network}
          {RURN: Rooted_Unidirectional_Ring_Network Net}
          {DP: Dijkstra_parameters}.

  Notation Algo := Dijkstra_algo.
  Notation Root := (RURN_root (Net := Net)).
  Notation Ring := (URN_topo (Net := Net)).

  Existing Instance Dijkstra_RO_assumption.

  Notation N := (Ring_size Root).
  Hypothesis HNk: N <= k.

    (** * Proof_ for degenerated case (N = 1 or 2) *)
  Section Degenerated_1_2.

    Hypothesis Nis1or2: N = 1 \/ N = 2.

    Theorem Dijkstra_bound_n_is_1_or_2:
      Steps_complexity 0
                       (fun e => Leg_state (s_hd e))
                       (fun e => Assume (s_hd e) /\ is_exec e).
    Proof using HNk Nis1or2.
      intros e [ass ise].
      exists 0; split; [lia|left].
      assert (pp: Proper pequiv (s_hd e)).
      apply is_exec_s_hd_compat in ise; auto.
      destruct Nis1or2 as [one|two].
      apply Single_Node_Legitimate; auto.
      apply Two_Nodes_Legitimate; auto.
    Qed.
    
  End Degenerated_1_2.
  
  (** * Proof_ for degenerated case (N = 3) *)
  Section Degenerated_3.

    (* ** Main Hypothesis *)
    Hypothesis Nis3: N = 3.

    (* ** Names for the 3 Nodes *)

    Notation N0:=Root (only parsing).
    Definition N1:=Succ Root.
    Definition N2:=Succ N1.

    (* ** Successors for the 3 Nodes *)
    Lemma Succ0: Succ N0 == N1.
    Proof using Type.
      reflexivity.
    Qed.
    Lemma Succ1: Succ N1 == N2.
    Proof using Type.
      reflexivity.
    Qed.
    Lemma Succ2: Succ N2 == N0.
    Proof using Nis3.
      rewrite <- Succ1,<- Succ0.
      etransitivity.
      2:apply  (Succ_nilpotent _ Root).
      rewrite Nis3;reflexivity.
    Qed.

    (* ** Predecessors for the 3 Nodes *)
    Lemma Pred0: Pred N0 == N2.
    Proof using Nis3.
      rewrite <- Succ2,Pred_Succ;reflexivity.
    Qed.
    Lemma Pred1: Pred N1 == N0.
    Proof using Type.
      rewrite <- Succ0,Pred_Succ;reflexivity.
    Qed.
    Lemma Pred2: Pred N2 == N1.
    Proof using Type.
      rewrite <- Succ1,Pred_Succ;reflexivity.
    Qed.

    (* ** The 3 Nodes are distinct *)
    Lemma N0notN1: ~N0==N1.
    Proof using Nis3 HNk.
      eapply Non_Degenerated_size_Succ with Root.
      subst;lia.
    Qed.
    Lemma N1notN2: ~N1==N2.
    Proof using  Nis3 HNk.
      eapply Non_Degenerated_size_Succ with Root.
      subst;lia.
    Qed.
    Lemma N0notN2: ~N0==N2.
    Proof using HNk Nis3.
      rewrite <- Succ2.
      intro h;symmetry in h;revert h.
      eapply Non_Degenerated_size_Succ with Root.
      subst;lia.
    Qed.

    (* ** The 3 Nodes form the whole ring *)
    Lemma Ring3: forall n:Node,n==N0\/n==N1\/n==N2.
    Proof using Nis3.
      intro n.
      assert (h:=traversal_Succ_full_all Root Root n).
      rewrite Nis3 in h.
      simpl traversal_Succ in h.
      do 2 rewrite InA_cons in h.
      rewrite InA_singleton in h.
      rewrite Succ0,Succ1 in h.
      apply h.
    Qed.

    (* ** Distance from the root to the 3 Nodes *)
    Lemma DistN0: Dist Root N0 = 0.
    Proof using Type.
      rewrite Dist_refl;reflexivity.
    Qed.

    Lemma DistN1: Dist Root N1 = 1.
    Proof using HNk Nis3.
      rewrite <- Succ0.
      destruct (Dist_Succ_r Root Root) as [h1|h0].
      - rewrite  h1,DistN0;reflexivity.
      - rewrite Dist_refl,Succ0 in h0.
        elimtype False.
        apply N0notN1,h0.
    Qed.

    Lemma DistN2: Dist Root N2 = 2.
    Proof using HNk Nis3.
      rewrite <- Succ1.
      destruct (Dist_Succ_r Root N1) as [h2|h0].
      - rewrite  h2,DistN1;reflexivity.
      - rewrite Dist_refl,Succ1 in h0.
        elimtype False.
        apply N0notN2,h0.
    Qed.

    (** ** Values of predecessors *)

    Lemma vPred0: forall g, Proper fequiv g -> v (g (Pred N0)) = v (g N2).
    Proof using Nis3.
      intros g pg;apply v_compat,pg,Pred0.
    Qed.

    Lemma vPred1: forall g, Proper fequiv g -> v (g (Pred N1)) = v (g N0).
    Proof using Type.
      intros g pg;apply v_compat,pg,Pred1.
    Qed.

    Lemma vPred2: forall g, Proper fequiv g -> v (g (Pred N2)) = v (g N1).
    Proof using Type.
      intros g pg;apply v_compat,pg,Pred2.
    Qed.

    (** ** Elementary characterization of Tokens presence/absence for each node *)

    Lemma Token_N0: forall g, Proper fequiv g ->
                              (v (g N0) = v (g N2)) -> Token g N0.
    Proof using Nis3.
      intros g pg EQ;left;split;auto.
      rewrite (vPred0 pg);auto.
    Qed.

    Lemma no_Token_N0: forall g, Proper fequiv g ->
                                 (v (g N0) <> v (g N2)) -> ~Token g N0.
    Proof using Nis3.
      intros g pg NE [[hroot eq]|[hnroot neq]];auto.
      rewrite (vPred0 pg) in eq;auto.
    Qed.

    Lemma Token_N1: forall g, Proper fequiv g ->
                              (v (g N0) <> v (g N1)) -> Token g N1.
    Proof using HNk Nis3.
      intros g pg NE;right;split.
      - apply N0notN1.
      - rewrite (vPred1 pg);auto.
    Qed.

    Lemma no_Token_N1: forall g, Proper fequiv g ->
                                 (v (g N0) = v (g N1)) -> ~Token g N1.
    Proof using HNk Nis3.
      intros g pg EQ [[hroot eq]|[hnroot neq]].
      - apply N0notN1;auto.
      - rewrite (vPred1 pg) in neq;auto.
    Qed.

    Lemma Token_N2: forall g, Proper fequiv g ->
                              (v (g N1) <> v (g N2)) -> Token g N2.
    Proof using HNk Nis3.
      intros g pg NE;right;split.
      - apply N0notN2.
      - rewrite (vPred2 pg);auto.
    Qed.

    Lemma no_Token_N2: forall g, Proper fequiv g ->
                                 (v (g N1) = v (g N2)) -> ~Token g N2.
    Proof using HNk Nis3.
      intros g pg EQ [[hroot eq]|[hnroot neq]].
      - apply N0notN2;auto.
      - rewrite (vPred2 pg) in neq;auto.
    Qed.

    (** ** Caracterization of configs by possible exec length (see file header) *)

    (** *  [L3_state] is ABA with (A+1) mod k = B *)

    Definition L3_state (g:Env (Algo:=Dijkstra_algo)) : Prop :=
      v (g N0) = v (g N2) /\
      ~ (v (g N0) = v (g N1)) /\
      S (v (g N0)) mod k = v (g N1).

    (** * [L2_state] is ABA with (A+1) mod k <> B *)

    Definition L2_state (g:Env (Algo:=Dijkstra_algo)):=
      v (g N0) = v (g N2) /\
      ~ (v (g N0) = v (g N1)) /\
      ~ S (v (g N0)) mod k = v (g N1).

    (** * [L1_state] is ABC *)

    Definition L1_state (g:Env (Algo:=Dijkstra_algo)):=
      ~ (v (g N0) = v (g N2)) /\
      ~ (v (g N0) = v (g N1)) /\
      ~ (v (g N1) = v (g N2)).

    (** * [L0_state] subcase : ABB - N1 has the token *)

    Definition L0_state_ABB (g:Env (Algo:=Dijkstra_algo)):=
      ~ (v (g N0) = v (g N2)) /\
      ~ (v (g N0) = v (g N1)) /\
      (v (g N1) = v (g N2)).

    (** * [L0_state] subcase : AAB - N2 has the token *)

    Definition L0_state_AAB (g:Env (Algo:=Dijkstra_algo)):=
      ~ (v (g N0) = v (g N2)) /\
      (v (g N0) = v (g N1)) /\
      ~ (v (g N1) = v (g N2)).

    (** * [L0_state] subcase : AAA - N0 has the token *)

    Definition L0_state_AAA (g:Env (Algo:=Dijkstra_algo)):=
      (v (g N0) = v (g N2)) /\
      (v (g N0) = v (g N1)) /\
      (v (g N1) = v (g N2)).

    (** * [L0_state] is ABB, AAB, or AAA *)

    Definition L0_state (g:Env (Algo:=Dijkstra_algo)):=
      L0_state_ABB g \/ L0_state_AAB g \/ L0_state_AAA g.

    (** * Tactic to prove decidability of formulas with integer equalities *)

    Ltac decidability_tac:=
      repeat (apply or_dec || apply and_dec || apply not_dec || apply eq_nat_dec).

    (** * Decidability results *)
    Lemma L3_dec: forall g, Decidable (L3_state g).
    Proof using Type.
      intro g;decidability_tac.
    Qed.
    Lemma L2_dec: forall g, Decidable (L2_state g).
    Proof using Type.
      intro g;decidability_tac.
    Qed.
    Lemma L1_dec: forall g, Decidable (L1_state g).
    Proof using Type.
      intro g;decidability_tac.
    Qed.
    Lemma L0_dec: forall g, Decidable (L0_state g).
    Proof using Type.
      intro g;decidability_tac.
    Qed.
    Lemma Leg_dec: forall g, Proper pequiv g -> Decidable (Leg_state g).
    Proof using Type.
      intros g pg; destruct (Leg_state_dec g) as [TT|FF];
        [left;apply TT,pg | right;apply FF].
    Qed.

    (** * Tacttic to prove [Proper] for predicates with integer equalities *)

    Ltac compat_tac pg:=
      repeat (apply or_compat||apply and_compat||apply not_compat);
      apply eq_compat;
      repeat (apply v_compat||apply pg||apply (f_equal2 Nat.modulo)||apply (f_equal S));
      reflexivity.

   (** * [Proper] results *)

    Instance L3_compat: Proper fequiv L3_state.
    Proof using Type.
      intros g1 g2 pg;compat_tac pg.
    Qed.
    Instance L2_compat: Proper fequiv L2_state.
    Proof using Type.
      intros g1 g2 pg;compat_tac pg.
    Qed.
    Instance L1_compat: Proper fequiv L1_state.
    Proof using Type.
      intros g1 g2 pg;compat_tac pg.
    Qed.
    Instance L0_compat: Proper fequiv L0_state.
    Proof using Type.
      intros g1 g2 pg;compat_tac pg.
    Qed.

   (** * Tactic to decompose state [g] based on equalities between node values:
       - uses fresh names so we can call it twice in the same proof
       - all known token properties are inferred
       - uses [congruence] to prune absurd cases *)

    Ltac destruct_state g pg:=
      let T0   := fresh "T0_" in
      let T1   := fresh "T1_" in
      let T2   := fresh "T2_" in
      let nT0  := fresh "nT0_" in
      let nT1  := fresh "nT1_" in
      let nT2  := fresh "nT2_" in
      let EQ02 := fresh "EQ02_" in
      let EQ01 := fresh "EQ01_" in
      let EQ12 := fresh "EQ12_" in
      let NE02 := fresh "NE02_" in
      let NE01 := fresh "NE01_" in
      let NE12 := fresh "NE12_" in
      destruct (eq_nat_dec (v (g N0)) (v (g N2))) as [EQ02|NE02];
      [assert (T0:=Token_N0 pg EQ02)|assert (nT0:=no_Token_N0 pg NE02)];
      (destruct (eq_nat_dec (v (g N0)) (v (g N1))) as [EQ01|NE01];
       [assert (nT1:=no_Token_N1 pg EQ01)|assert (T1:=Token_N1 pg NE01)]);
      (destruct (eq_nat_dec (v (g N1)) (v (g N2))) as [EQ12|NE12];
       [assert (nT2:=no_Token_N2 pg EQ12)|assert (T2:=Token_N2 pg NE12)]);
      try congruence.

    (* ** [L0_state] is an elementary characterizatio of legitimate states for N=3 *)

    Lemma Leg_state_L0:
      forall g, Proper pequiv g -> (Leg_state g <-> L0_state g).
    Proof using HNk Nis3.
      intros g pg.
      split.
      {
        intro leg.
        destruct_state g pg.
        - right;right;hnf;tauto.
        - assert (contra:=leg _ _ T1_ T2_);destruct (N1notN2 contra).
        - right;left;hnf;tauto.
        - left;hnf;tauto.
        - assert (contra:=leg _ _ T1_ T2_);destruct (N1notN2 contra).
      }
      {
        assert(pTg:Proper fequiv (Token g)).
        {
          apply Token_compat,pg.
        }
        intros [[NE02 [NE01 EQ12]]|
                [[NE02 [EQ01 NE12]]|
                 [EQ02 [EQ01 EQ12]]]];
          destruct_state g pg; intros na nb Ta Tb;
            destruct (Ring3 na) as [ea|[ea|ea]];
            rewrite ea in Ta|-*;clear ea na;
              destruct (Ring3 nb) as [eb|[eb|eb]];
              rewrite eb in Tb|-*;clear eb nb;
                (tauto||reflexivity).
      }
    Qed.

    (* ** all possible states are either
       [L0_state] or [L1_state] or [L2_state] or [L3_state] *)

    Lemma state_decomp: forall g,
        Proper fequiv g -> (L3_state g \/ L2_state g \/ L1_state g \/ L0_state g).
    Proof using HNk Nis3.
      intros g pg.
      unfold L3_state,L2_state,L1_state,L0_state,
      L0_state_AAA,L0_state_AAB,L0_state_ABB;
        destruct_state g pg;
        destruct (eq_nat_dec (S (v (g Root)) mod k)
                             (v (g N1)));intuition congruence.
    Qed.

    (* ** [L0_state],[L1_state],[L2_state] and [L3_state] are disjoint *)

    Lemma disjoint_L3_L2 : forall g, L3_state g -> L2_state g -> False.
    Proof using Type.
      intros g HA HB;hnf in HA,HB;tauto.
    Qed.

    Lemma disjoint_L3_L1 : forall g, L3_state g -> L1_state g -> False.
    Proof using Type.
      intros g HA HB;hnf in HA,HB;tauto.
    Qed.

    Lemma disjoint_L3_L0 : forall g, L3_state g -> L0_state g -> False.
    Proof using Type.
      intros g HA HB;destruct HB as [HB|[HB|HB]];hnf in HA,HB;tauto.
    Qed.

    Lemma disjoint_L2_L1 : forall g, L2_state g -> L1_state g -> False.
    Proof using Type.
      intros g HA HB;hnf in HA,HB;tauto.
    Qed.

    Lemma disjoint_L2_L0 : forall g, L2_state g -> L0_state g -> False.
    Proof using Type.
      intros g HA HB;destruct HB as [HB|[HB|HB]];hnf in HA,HB;tauto.
    Qed.

    Lemma disjoint_L1_L0 : forall g, L1_state g -> L0_state g -> False.
    Proof using Type.
      intros g HA HB; destruct HB as [HB|[HB|HB]];hnf in HA,HB;tauto.
    Qed.

   (* ** No step is possible if neither N0,N1 or N2 moves *)

    Lemma Step_has_to_move:  forall (g:Env) (g':Env),
        Step (Algo:=Algo) g' g ->
        has_moved_b g' g N0 = false ->
        has_moved_b g' g N1 = false ->
        has_moved_b g' g N2 = false ->
        False.
    Proof using Nis3.
      intros g g' hstep nop0 nop1 nop2.
      destruct (Step_one_has_moved_true hstep) as [NN movNN].
      assert (pg:=Step_proper_r hstep).
      assert (pg':=Step_proper_l hstep).
      destruct (Ring3 NN) as [eNN |[eNN | eNN ]];rewrite eNN in movNN;congruence.
    Qed.

   (** * Tactic to decompose step cases based on who has moved
       - all known token properties are inferred
       - all node value (dis)equalities are inferred
       - the absurd case with no move is pruned
       - [tauto] may be appended to prune moves at disabled nodes *)

    Lemma HN1: N > 1. Proof using HNk Nis3. lia. Qed.

    Ltac destruct_step g g' pg assume hstep :=
      let mov0 := fresh "mov0_" in
      let mov1 := fresh "mov1_" in
      let mov2 := fresh "mov2_" in
      let nop0 := fresh "nop0_" in
      let nop1 := fresh "nop1_" in
      let nop2 := fresh "nop2_" in
      let movT0 := fresh "movT0_" in
      let movT1 := fresh "movT1_" in
      let movT2 := fresh "movT2_" in
      let movN0 := fresh "movN0_" in
      let movN1 := fresh "movN1_" in
      let movN2 := fresh "movN2_" in
      let nopN0 := fresh "nopN0_" in
      let nopN1 := fresh "nopN1_" in
      let nopN2 := fresh "nopN2_" in
        (destruct (Sumbool.sumbool_of_bool (has_moved_b g' g N0)) as [mov0|nop0];
         [ assert(movT0:=has_moved_true_Token HNk HN1 assume hstep _ mov0);
           assert(movN0:=has_moved_v_Root HNk assume hstep _ (reflexivity Root) mov0);
           rewrite (vPred0 pg),Nat.add_1_r in movN0
         | assert (nopN0:= v_compat (proj1 (has_moved_false _  _ N0)  nop0));
           hnf in nopN0 ]);
        (destruct (Sumbool.sumbool_of_bool (has_moved_b g' g N1)) as [mov1|nop1];
         [ assert(movT1:=has_moved_true_Token HNk HN1 assume hstep _ mov1);
           assert(movN1:=has_moved_v_non_Root assume hstep _ (N0notN1) mov1);
           rewrite (vPred1 pg) in movN1
         | assert (nopN1:= v_compat (proj1 (has_moved_false _  _ N1)  nop1));
           hnf in nopN1 ]);
        (destruct (Sumbool.sumbool_of_bool (has_moved_b g' g N2)) as [mov2|nop2];
         [ assert(movT2:=has_moved_true_Token HNk HN1 assume hstep _ mov2);
           assert(movN2:=has_moved_v_non_Root assume hstep _ N0notN2 mov2);
           rewrite (vPred2 pg) in movN2
         | assert (nopN2:= v_compat (proj1 (has_moved_false _  _ N2)  nop2));
           hnf in nopN2 ]);[ .. |destruct  (Step_has_to_move hstep nop0 nop1 nop2)].

    (* * [L3_Step_L2_L0] : A [L3_state] leads to either a [L2_state] or a [L0_state] *)

    Lemma L3_Step_L2_L0: forall (g:Env) (g':Env),
        Assume g -> Step g' g ->
        L3_state g -> L2_state g' \/ L0_state g'.
    Proof using HNk Nis3.
      intros g g' assume hstep [EQ02 [NE01 S02]].
      assert (pg:=Step_proper_r hstep).
      assert (pg':=Step_proper_l hstep).
      unfold L0_state.
      destruct_state g pg.
      destruct_step g g' pg assume hstep.
      - do 0 right;left;hnf.  (* three nodes move -> L2 *)
        repeat (split;try congruence).
        rewrite movN0_,movN1_,<- EQ02.
        apply SSmod3;congruence.
      - do 1 right;left;hnf. (* only nodes N0 and N1 move -> L0 ABB *)
        repeat (split;try congruence).
      - do 3 right;hnf. (* only nodes N0 and N2 move -> L0 AAA *)
        repeat (split;try congruence).
      - do 2 right;left;hnf. (* only N0 moves -> L0 AAB *)
        repeat (split;try congruence).
      - do 2 right;left;hnf. (* only N1 and N2 move -> L0 AAB *)
        repeat (split;try congruence).
      - do 3 right;hnf. (* only N1 moves -> L0 AAA *)
        repeat (split;try congruence).
      - do 1 right;left;hnf. (* only N2 moves -> L0 ABB *)
        repeat (split;try congruence).
    Qed.

    (* * [L2_Step_L1_L0] : A [L2_state] leads to either a [L1_state] or a [L0_state] *)

    Lemma L2_Step_L1_L0: forall (g:Env) (g':Env),
        Assume g -> Step g' g ->
        L2_state g -> L1_state g' \/ L0_state g'.
    Proof using HNk Nis3.
      intros g g' assume hstep [EQ02 [NE01 nS02]].
      assert (pg:=Step_proper_r hstep).
      assert (pg':=Step_proper_l hstep).
      unfold L0_state.
      destruct_state g pg.
      destruct_step g g' pg assume hstep.
      - do 0 right;left;hnf. (* three nodes moves -> L1 *)
       repeat (split;try congruence).
       rewrite movN0_,movN1_,<- EQ02.
       apply Smod3;congruence.
      - do 1 right;left;hnf. (* only N0 and N1 move -> L0 ABB *)
        repeat (split;try congruence).
        -- rewrite movN0_,nopN2_.
           apply Smod3;congruence.
        -- rewrite movN0_,movN1_,<-EQ02_.
          apply Smod3;congruence.
      - do 1 right;left;hnf. (* only N0 and N2 move -> L0 ABB *)
        repeat (split;try congruence).
      - do 0 right;left;hnf. (* only N0 moves -> L1 *)
        repeat (split;try congruence).
        rewrite movN0_,nopN2_.
        apply Smod3;congruence.
      - do 2 right;left;hnf. (* only N1 and N2 move -> L0 AAB *)
        repeat (split;try congruence).
      - do 3 right;hnf. (* only N1 moves -> L0 AAA *)
        repeat (split;try congruence).
      - do 1 right;left;hnf. (* only N2 moves -> L0 ABB *)
        repeat (split;try congruence).
    Qed.

    (* * [L1_Step_L0] : A [L1_state] leads to either a [L0_state] *)

    Lemma L1_Step_L0: forall (g:Env) (g':Env),
        Assume g -> Step g' g ->
        L1_state g -> L0_state g'.
    Proof using HNk Nis3.
      intros g g' assume hstep [EQ02 [NE01 nS02]].
      assert (pg:=Step_proper_r hstep).
      assert (pg':=Step_proper_l hstep).
      unfold L0_state.
      destruct_state g pg.
      destruct_step g g' pg assume hstep;try tauto.
      - do 1 right;left;hnf. (* only N1 and N2 move -> L0 AAB *)
        repeat (split;try congruence).
      - do 1 right;left;hnf. (* only N1 moves -> L0 AAB *)
        repeat (split;try congruence).
      - do 0 right;left;hnf. (* only N2 moves -> L0 ABB *)
        repeat (split;try congruence).
    Qed.

    (* * [exec_no_end] : an [Exec] for DTRC cannot be finite *)
    (* TODO : move elsewhere *)
    Lemma exec_no_end: forall g, Assume g ->
                                 is_exec (s_one g) -> False.
      intros g assume ise.
      assert (pg:=is_exec_s_hd_compat ise).
      assert (h:=is_exec_inv_0 ise).
      assert (hh:=no_terminal HNk HN1 pg assume h).
      apply hh.
    Qed.

    (* ** [Dijkstra_bound_n_is_3] : Main result by state graph exploration
       Let P3 := L3_state \/ L2_state \/ L1_state \/ L0_state
       Let P2 :=             L2_state \/ L1_state \/ L0_state
       Let P1 :=                         L1_state \/ L0_stat
       Let P0 :=                                     L0_state

      We use [Steps_complexity_add] twice.
      The step complexity to reach Leg_state from any state is the sum of:
       - the complexity to reach P2 from P3 (<-> any state)
         1 if L3_state, 0 otherwise
       - the complexity to reach P1 from P2
         1 if L2_state, 0 otherwise
       - the complexity to reach P0 (<-> Leg_states) from P1
         1 if L1_state, 0 otherwise
     *)

    Theorem Dijkstra_bound_n_is_3:
      Steps_complexity 3
                       (fun e => Leg_state (s_hd e))
                       (fun e => Assume (s_hd e) /\ is_exec e).
    Proof using HNk Nis3.
      pose (P0 e:= L0_state (s_hd e)).
      pose (P1 e:= L1_state (s_hd e) \/ P0 e).
      pose (P2 e:= L2_state (s_hd e) \/ P1 e).
      pose (P3 e:= L3_state (s_hd e) \/ P2 e).
      pose (Q e:= Assume (s_hd e) /\ is_exec e).
      assert (Leg_dec: forall e: Exec,
                 Q e -> {Leg_state (s_hd e)} + {~ Leg_state (s_hd e)}).
      {
        intros e [assume ise];
          assert (pg:=is_exec_s_hd_compat ise);
          apply Leg_dec;auto.
      }
      assert (Leg_compat: Proper fequiv (fun e => Leg_state (s_hd e)) ).
      { intros e1 e2 ee;apply Leg_state_compat,hd_compat,ee. }
      assert (P0_dec: forall e: Exec, Q e -> {P0 e} + {~ P0 e}).
      {
        intros e [assume ise];
          assert (pg:=is_exec_s_hd_compat ise);
          apply L0_dec;auto.
      }
      assert (P0_compat: Proper fequiv P0).
      { intros e1 e2 ee;apply L0_compat,hd_compat,ee. }
      assert (P1_dec: forall e: Exec, Q e -> {P1 e} + {~ P1 e}).
      {
        intros e [assume ise];
          assert (pg:=is_exec_s_hd_compat ise);
          apply or_dec;[apply L1_dec|apply P0_dec];hnf;auto.
      }
      assert (P1_compat: Proper fequiv P1).
      { intros e1 e2 ee;apply or_compat;
          [apply L1_compat,hd_compat,ee | apply P0_compat,ee].
      }
      assert (P2_dec: forall e: Exec, Q e -> {P2 e} + {~ P2 e}).
      {
        intros e [assume ise];
          assert (pg:=is_exec_s_hd_compat ise);
          apply or_dec;[apply L2_dec|apply P1_dec];hnf;auto.
      }
      assert (P2_compat: Proper fequiv P2).
      { intros e1 e2 ee;apply or_compat;
          [apply L2_compat,hd_compat,ee | apply P1_compat,ee].
      }
      assert (allways_Q: forall e : Exec, Q e -> Always Q e).
      {
        intros e [assume ise];apply Always_conj;split.
        - apply is_exec_Assume;auto.
        - apply is_exec_Always;auto.
      }
      assert (incl1: forall e : Exec, Q e -> Leg_state (s_hd e) -> P1 e).
      { intros e [assume ise] leg; right.
        assert (pg:=is_exec_s_hd_compat ise);
            rewrite Leg_state_L0 in leg;auto.
      }
      assert (incl2: forall e : Exec, Q e -> P1 e -> P2 e).
      { unfold P2,P1,P0;intros;tauto. }
      assert (Q_proper: forall e : Exec, Q e -> Proper pequiv e).
      { intros e [assume ise];apply (is_exec_proper ise). }
      replace 3 with ((1+1)+1) by reflexivity.
      (apply (@Steps_complexity_add _ _ _ _ P1);auto);
        [apply (@Steps_complexity_add _ _ _ _ P2);auto|].
      {
        hnf.
        intros e [assume ise];
          assert (pg:=is_exec_s_hd_compat ise);
          destruct (state_decomp (is_exec_s_hd_compat ise)) as [hL3|hL210].
        {
          exists 1;split;auto.
          destruct e as [g | g e].
          { destruct (exec_no_end assume ise). }
          right.
          {
            intros [ hL2 | [hL1| hL0]].
            - apply disjoint_L3_L2 with g;auto.
            - apply disjoint_L3_L1 with g;auto.
            - apply disjoint_L3_L0 with g;auto.
          }
          {
            left.
            assert (hstep:=is_exec_inv_1 ise).
            assert (pg':=Step_proper_l hstep).
            apply  L3_Step_L2_L0 in hstep;auto.
            unfold P2,P1,P0;tauto.
          }
        }
        {
          exists 0;split;auto.
          left.
          unfold P2,P1;tauto.
        }
      }
      {
        hnf.
        intros e [[assume ise] hP2];
          assert (pg:=is_exec_s_hd_compat ise);
          destruct (state_decomp (is_exec_s_hd_compat ise)) as [hL3|[hL2|hL10]].
        {
          elimtype False.
          destruct hP2 as [hL2|[hL1|hL0]].
          - apply disjoint_L3_L2 with (s_hd e);auto.
          - apply disjoint_L3_L1 with (s_hd e);auto.
          - apply disjoint_L3_L0 with (s_hd e);auto.
        }
        {
          exists 1;split;auto.
          destruct e as [g | g e].
          { destruct (exec_no_end assume ise). }
          right.
          {
            intros [hL1| hL0].
            - apply disjoint_L2_L1 with g;auto.
            - apply disjoint_L2_L0 with g;auto.
          }
          {
            left.
            assert (hstep:=is_exec_inv_1 ise).
            apply  L2_Step_L1_L0 in hstep;auto.
          }
        }
        {
          exists 0;split;auto.
          left.
          unfold P1;tauto.
        }
      }
      {
        hnf.
        intros e [[assume ise] hP1];
          assert (pg:=is_exec_s_hd_compat ise);
          destruct (state_decomp (is_exec_s_hd_compat ise)) as [hL3|[hL2|[hL1|hL0]]].
        {
          elimtype False.
          destruct hP1 as [hL1|hL0].
          - apply disjoint_L3_L1 with (s_hd e);auto.
          - apply disjoint_L3_L0 with (s_hd e);auto.
        }
        {
          elimtype False.
          destruct hP1 as [hL1|hL0].
          - apply disjoint_L2_L1 with (s_hd e);auto.
          - apply disjoint_L2_L0 with (s_hd e);auto.
        }
        {
          exists 1;split;auto.
          destruct e as [g | g e].
          { destruct (exec_no_end assume ise). }
          right.
          { intros hL0.
            - apply disjoint_L1_L0 with g;auto.
              rewrite Leg_state_L0 in hL0;auto.
          }
          {
            left.
            assert (hstep:=is_exec_inv_1 ise).
            assert (pg':=Step_proper_l hstep).
            apply L1_Step_L0 in hstep;auto.
            rewrite Leg_state_L0;auto.
          }
        }
        {
          exists 0;split;auto.
          left.
          rewrite Leg_state_L0;auto.
        }
      }
    Qed.

  End Degenerated_3.

End Dijkstra_complexity_1_2_3.

Unset Implicit Arguments.
Close Scope nat_scope.
