(**
#<a href="http://wwwverimag.imag.fr/~altisen/PADEC/">
A Framework for Certified SelfStabilization
</br>
PADEC Coq Library
</a>#

 March 2017
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import ZArith.
From Coq Require Import Lia.

(** **** Local Imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import NatUtils.
From Padec Require Import ListUtils.
From Padec Require Import Count.
From Padec Require Import All_In.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Exec.
From Padec Require Import P_Q_Termination.
From Padec Require Import Fairness.
From Padec Require Import Self_Stabilization.
From Padec Require Import Ring_topology.
From Padec Require Import Dijkstra_algo.
From Padec Require Import Dijkstra_common.
From Padec Require Import Dijkstra_Z.
From Padec Require Import Dijkstra_sum_dist.

(* end show *)
Open Scope nat_scope.
Set Implicit Arguments.

(** * Convergence
    i.e. eventually a legitimate configuration is reached
    i.e. eventually there exists a unique token in the ring.
 *)
Section Dijkstra_Convergence.

  Context {Chans: Channels}
          {Net: Network}
          {RURN: Rooted_Unidirectional_Ring_Network Net}
          {DP: Dijkstra_parameters}.

  Notation Algo := Dijkstra_algo.
  Notation Root := (RURN_root (Net := Net)).
  Notation Ring := (URN_topo (Net := Net)).
  Notation Env := (Env (Net := Net) (Algo := Dijkstra_algo)).

  Existing Instance Dijkstra_RO_assumption.

  Notation N := (Ring_size Root).
  Hypothesis HN1: N > 1.

  (** This part of the proof using Type using Type using Type using Type using Type using Type using Type using Type requires the additional property that
   the number of nodes in the ring is smaller than or equal to k.
   *)
  Hypothesis (HNk: N <= k).

  Notation Mset := (MultisetNat.MSetCore.Multiset).
  Notation ltM := (@MultisetNat.MSetOrd.MultisetLT gt).
  Notation eqM := MultisetNat.MSetCore.meq.

  (** Convergence proof is done within nonlegitimate configurations
   and is split into
    convergence to legitimate configurations of steps where the Root moves
    convergence to legitimate configurations of steps where the Root
     does not execute

   In the proof, we only consider configurations which are not legitimate
   and steps between those nonlegitimate configurations.
   *)
  Notation Safe := (fun g => Assume g /\ ~Leg_state g).
  (** safeEnv := {g | Assume g /\ ~ Leg_state g} *)
  Notation safeEnv := (sig Safe).
  Notation getEnv := (fun sg: safeEnv => proj1_sig sg).
  Notation getSafe := (fun sg: safeEnv => proj2_sig sg).
  
  (** Step where the root moves: Trans_with_root is a relation between safeEnv 
   *)
  Definition QTrans_with_root (g' g: safeEnv) :=
    v (getEnv g' Root) <> v (getEnv g Root).

  Definition Trans_with_root := safeQStep QTrans_with_root. 

  (** Step where the root does not move. *)
  Definition QTrans_without_root (g' g: safeEnv) :=
    v (getEnv g' Root) = v (getEnv g Root).

  Definition Trans_without_root := safeQStep QTrans_without_root.

  (** ** Convergence When Root Moves: Z potential decreases at 
    each step the root moves.
   *)
  (** Potential [Z] on safeEnv. *)
  Definition Z_pot :=
    (fun sg: safeEnv => Z (getEnv sg)).
  
  Lemma Z_pot_decreases:
    forall sg sg',
      Trans_with_root sg' sg -> Z_pot sg' < Z_pot sg.
  Proof.
    unfold Trans_with_root, QTrans_with_root,
    safeQStep, safeStep, Z_pot.
    intros [g [ass nls]] [g' [ass' nls']] [step diff].
    apply Z_decreases; auto.
    apply has_moved_v_change; auto.
  Qed.
    
  (** ** Convergence when Root does not move 
   For this part of the proof of convergence, we use 
   the criteria given by the Deshowitz-Manna order expressed 
   in the library [P_Q_Termination].
   *)
  Section Root_does_not_move.

    (** *** Potential when Root does not move 
     for a node: 
       equals 0 when the node is disabled, 
       is equal to the distance from the node to the root 
       otherwise.
    *)
    Definition pot_without_Root (g: Env) (n: Node) :=
      if Token_dec g n then (Dist n Root) else 0.

    Definition Pot_without_Root :=
      (@Pot _ Net Dijkstra_algo Safe
            (fun sg: safeEnv => pot_without_Root (getEnv sg))).

    Instance pot_without_Root_compat:
      Proper fequiv pot_without_Root.
    Proof.
      intros g1 g2 eg n1 n2 en.
      unfold pot_without_Root; simpl; simpl in eg.
      assert (h := Token_compat eg _ _ en).
      hnf in h.
      destruct (Token_dec g1 n1); destruct (Token_dec g2 n2); try tauto.
      apply Dist_compat.
      apply en.
      reflexivity.
    Qed.

    (** *** Common Result to Both Criteria: 
     When a node [n] moves and still holds a token afterwards, 
     this means that there exists a node [n'] such that
          [n'] holds a token in [g]
          has lost it in [g'] and
          [n'] is closer to the Root than [n] (using Succ) 
     This is proven using induction on the Ring traversal (with Succ).
     The node [n'] cannot be the Root (since it moves); in either case,
     the property relies on Lemma [two_tokens_move]
      [forall n, has_moved_b g' g n = true >
                Token g' n > has_moved_b g' g (Pred n) = true.]
     (when a node moves and still holds the token afterwards, this
     means that it predecessor also moves). As the root does not move
     in this context, we have the property.
    *)
    Lemma both_criteria:
      forall g g',
        Assume g -> Step g' g -> v (g' Root) = v (g Root) ->
        forall n, has_moved_b g' g n = true -> Token g' n
                  ->
                  exists n', Token g n' /\ ~Token g' n' /\
                             Dist n' Root > Dist n Root.
    Proof.
      intros g g' ass step hroot n.
      pattern n; eapply Ring_Succ_ind with (x := Root).
       intros n1 n2 en; apply impl_compat.
        apply eq_compat; auto; apply has_moved_b_compat;
          [eapply Step_Proper_l,step | apply step |apply en].
        apply impl_compat.
        apply Token_compat;auto.
        apply (Step_Proper_l step).
        apply ex_compat; intros n';
          apply and_compat; [|apply and_compat].
        apply Token_compat; [apply step | reflexivity].
        apply not_compat; apply Token_compat; [ apply (Step_Proper_l step) |
                                                reflexivity].
        apply gt_compat; auto; apply Dist_compat; auto.
       intros f; apply has_moved_v_change in f; auto; lia. 
       clear n; intros n ih hms toks'.
        assert (pg: Proper fequiv g) by apply step;
          assert (pg': Proper fequiv g') by apply (Step_Proper_l step).
        assert (hm: has_moved_b g' g n = true).
        { rewrite <- (has_moved_b_compat pg' pg _ _ (Pred_Succ n)).
          eapply two_tokens_move; auto.
        }
        assert (nr: ~ Root == n).
        { intros f; symmetry in hroot; contradict hroot;
            apply has_moved_v_change; auto.
          rewrite (has_moved_b_compat pg' pg _ _ f); auto.
        }
        assert (nr': ~ n == Root)
          by (intros f; contradict nr; symmetry; auto).
        rewrite <- Dist_refl in nr'.
        destruct (Token_dec g' n) as [tok'|ntok'].
        + destruct (ih hm tok') as [n' [T [nT d]]].
          exists n'; split; [|split]; auto.
          destruct (Dist_Succ_l n Root); lia.
        + exists n; split; [|split]; auto.
          apply enabled_Token; auto; apply has_moved_b_enabled_b with g';
            auto.
          destruct (Dist_Succ_l n Root); lia.
    Qed.
    
    (** *** Local Criteria: 
     When a node [n], which is not the root, does not hold a token
     in g but obtains one in g', this means that there exists a node 
     [n'] such that
          [n'] holds a token in [g]
          has lost it in [g'] and
          [n'] is closer to the Root than [n] (using Succ) 

     Note that the predecessor of n moves:
       if it loses its token in g', then n' is (Pred n)
       otherwise, we apply [both_criteria] to (Pred n) and obtains
        the result. 
     *)
    Lemma local_criterion_without_Root:
      forall (g g': Env),
        Assume g -> Step g' g -> v (g' Root) = v (g Root) ->
        forall n, ~ Root == n -> ~ Token g n -> Token g' n ->
                  exists n', Token g n' /\ ~Token g' n' /\
                             Dist n' Root > Dist n Root.
    Proof.
      intros g g' assume_g step safe.
      intros n nonRoot ntok tok'.
      assert (hmp := Dijkstra_common.token_obtained
                       HNk assume_g step HN1 ntok tok').
      assert (pg: Proper fequiv g) by apply step.
      assert (nr: Dist (Pred n) Root <> 0).
      {
        intros dp; rewrite Dist_refl in dp; auto; symmetry in dp.
        apply has_moved_v_change in hmp; auto.
        contradict hmp.
        assert (pg': Proper fequiv g') by apply (Step_Proper_l step).
        rewrite <- (v_compat (pg _ _ dp)),
        <- (v_compat (pg' _ _ dp)); auto.
      }
      destruct (Token_dec g' (Pred n)) as [tokp'|ntokp'].
      {
        destruct (both_criteria assume_g step safe hmp
                                tokp') as [n' [tokn' [ntokn' d]]].
        exists n'; split; [| split]; auto.        
        destruct (Dist_Pred_l n Root) as [dp|dp]; lia.
      }
      { assert (tokp: Token g (Pred n))
          by (apply enabled_Token; auto; apply has_moved_b_enabled_b with g';
              auto).
        exists (Pred n); split; [| split]; try tauto.
        destruct (Dist_Pred_l n Root) as [dp|dp]; lia.
      }
    Qed.

    Lemma local_criterion_without_Root':
      forall g g': safeEnv,
        safeStep g' g /\ v (getEnv g' Root) = v (getEnv g Root) ->
        forall n,
          pot_without_Root (getEnv g') n > pot_without_Root (getEnv g) n ->
          exists n',
            pot_without_Root (getEnv g) n' <>
            pot_without_Root (getEnv g') n' /\
            pot_without_Root (getEnv g) n' > pot_without_Root (getEnv g') n.
    Proof.
      unfold safeStep; unfold pot_without_Root.
      intros [g [assume_g nonleg]] [g' [assume_g' nonleg']]
             [step safe] n hgt;simpl in *.
      destruct (Token_dec g' n) as [tok'|ntok'];
        destruct (Token_dec g n) as [tok|ntok]; try lia.
      assert (nonRoot: ~ Root == n)
        by (intros F; symmetry in F; rewrite <- Dist_refl in F; lia);
        clear hgt.
      destruct (local_criterion_without_Root
                  assume_g step safe nonRoot ntok tok')
        as [n' [tokn' [ntok'' d]]].
      exists n'.
      destruct (Token_dec g n'); destruct (Token_dec g' n');
        split; try tauto; lia.
    Qed.

    (** *** Global Criteria 
     There exists a node (different from the root) which loses a
     token from g to g'.

     The step from g to g' provides a node which moves (hence it is 
     not the root and holds a token). If this node loses its token
     in g', we are done, otherwise, we apply [both_criteria] to n
     and obtains the result. 
    *)
    Lemma global_criterion_without_Root:
      forall (g g': Env),
        Assume g -> Step g' g -> v (g' Root) = v (g Root) ->
        exists n, ~ Root == n /\ Token g n /\ ~Token g' n.
    Proof.
      intros g g' assume_g step safe.
      destruct (Step_one_has_moved_true step) as [n hm].
      assert (pg: Proper fequiv g) by apply step.
      assert (nr: ~ Root == n).
      {
        intros f;
          apply has_moved_v_change in hm; auto; contradict hm.
        assert (pg': Proper fequiv g') by apply (Step_Proper_l step).
        rewrite <- (v_compat (pg _ _ f)),
        <- (v_compat (pg' _ _ f)); auto.
      }
      assert (tok: Token g n) by
          (apply enabled_Token; auto; apply has_moved_b_enabled_b with g';
           auto).
      destruct (Token_dec g' n) as [tok'|ntok'].
      { destruct (both_criteria assume_g step safe hm
                                tok') as [n' [tokn' [ntokn' d]]].
        exists n'; split; auto.
        intros f; symmetry in f; rewrite <- Dist_refl in f; lia.
      }
      { exists n; split; auto. }
    Qed.

    Lemma global_criterion_without_Root':
      forall (g g' : safeEnv),
        safeStep g' g /\ v (getEnv g' Root) = v (getEnv g Root) ->
        exists n, pot_without_Root (getEnv g') n <>
                  pot_without_Root (getEnv g) n.
    Proof.
      unfold safeStep; unfold pot_without_Root.
      intros [g [assume_g nonleg]] [g' [assume_g' nonleg']] [step safe];
        simpl in *.
      destruct (global_criterion_without_Root assume_g step safe)
        as [n' [nr [tokn' ntok'']]].
      exists n'.
      destruct (Token_dec g n'); destruct (Token_dec g' n'); auto.
      intros f; contradict nr; symmetry; rewrite <- Dist_refl; lia.
    Qed.

    (** We obtain the relation that corresponds to steps where the root 
     does not move is wellfounded.
     *)
    Lemma Trans_without_root_wf:
      well_founded (safeQStep QTrans_without_root).
    Proof.
      eapply Wf_safeQAlgo_Multiset
        with (fun sg => pot_without_Root (proj1_sig sg)).
      { intros [g sg] [g' sg'] h; simpl;
          apply pot_without_Root_compat, h.
      }
      apply local_criterion_without_Root'.
      apply global_criterion_without_Root'.      
    Qed.
    
  End Root_does_not_move.

  (** Steps with and without Root moves are combined using a
   lexicographic ordering 

   Again, to show that potential in g' is less than or equal to
   potential in g, we rewrite potentials as minimum (using 
   [pot_with_Root_min]) and applies Lemma [is_minimum_le] that says
   that
       if M1 (resp. M2) is minimum of some set A1 (resp. A2) and if for
       every value in A1, there exists a smaller or equal value in A2,
       then M2 is smaller than or equal to M1.

   We first restrict to values smaller than k in M1 
   (i.e. values x such that (convex g ((v (g Root) + x) mod k)) holds).
    If x is 0, we prove that (convex g' (v (g Root))): since the 
     value is copied from the predecessor when a node moves, this 
     cannot break the convex property for (v (g Root)).     

    If x <> 0 then, no node in the ring holds value 
     ((v (g Root) + x) mod k) = ((v (g' Root) + x) mod k).
     As no new value appears in the step, 
     (convex g ((v (g Root) + x) mod k)) holds.
  *)
  Lemma values_do_not_appear_non_Root:
    forall (g g': Env),
      Assume g -> Step g' g ->
      v (g' Root) = v (g Root) -> Z g' <= Z g.
  Proof.
    assert (H_k:=H_k).
    intros g g' ass step safe.
    assert (pg: Proper fequiv g) by apply step.
    assert (pg': Proper fequiv g') by (apply Step_Proper_l with g, step).
    eapply is_minimum_le.
    2: { apply Z_min; auto. auto. auto. }
    2: { apply Z_min; auto. auto. auto. 
         apply step_Assume with g; auto. }
    rewrite safe.
    cut (forall v1, v1 < k ->
                    convex_b g ((v (g Root) + v1) mod k) = true ->
                    exists v2,
                      convex_b g' ((v (g Root) + v2) mod k) = true /\
                      v2 <= v1).
    { intros ok x Cx.
      assert (d: k <> 0) by lia;
        specialize (ok (x mod k) (Nat.mod_upper_bound x _ d)); clear d.
      rewrite Nat.add_mod_idemp_r in ok; [|lia].
      specialize (ok Cx).
      destruct ok as [y [Cy hy]].
      exists y; split; auto.
      transitivity (x mod k); auto; apply Nat.mod_le; lia.
    }
    intros x hhx Cx.
    rewrite <- convex_b_true in Cx; auto.
    destruct (Nat.eq_dec x 0) as [hx|hx].
    { rewrite hx; rewrite hx in Cx; clear x hhx hx.
      exists 0; split; [|lia].
      rewrite <- convex_b_true; auto.
      rewrite v_plus_0; rewrite v_plus_0 in Cx; auto.
      intros n n' B V.
      case_eq (has_moved_b g' g n); intros hmn.
      - assert (nr: ~ Root == n).
        { rewrite has_moved_v_change in hmn; auto.
          contradict hmn.
          rewrite <- (v_compat (pg' _ _ hmn));
            rewrite <- (v_compat (pg _ _ hmn)); lia. }
        eapply has_moved_v_non_Root in hmn; auto.
        rewrite hmn in V.
        case_eq (has_moved_b g' g n'); intros hmn'.
        + assert (nr': ~ Root == n').
          { rewrite has_moved_v_change in hmn'; auto.
            contradict hmn'.
            rewrite <- (v_compat (pg' _ _ hmn'));
              rewrite <- (v_compat (pg _ _ hmn')); lia. }
          eapply has_moved_v_non_Root in hmn'; auto.
          rewrite hmn'.
          destruct (Between_Pred_r B) as [B'|B'].
          rewrite (v_compat (pg _ _ (Pred_compat _ _ B'))); lia.
          destruct (Between_Pred_m B') as [B''|B''].
          tauto.
          specialize (Cx _ _ B''); apply Cx, V.
        + apply has_moved_false in hmn';auto;destruct hmn' as [hmn' _]; auto.
          destruct (Between_Pred_r B) as [B'|B'].
          rewrite <- (v_compat (pg' _ _ B')) in hmn; lia.
          unfold v in *;rewrite hmn'.
          specialize (Cx _ _ B'); apply Cx, V.
       - apply has_moved_false in hmn;auto;destruct hmn as [hmn _]; auto.
        unfold v in V;rewrite hmn in V.
        case_eq (has_moved_b g' g n'); intros hmn'.
        + assert (nr': ~ Root == n').
          { rewrite has_moved_v_change in hmn'; auto.
            contradict hmn'.
            rewrite <- (v_compat (pg' _ _ hmn'));
              rewrite <- (v_compat (pg _ _ hmn')); lia. }
          eapply has_moved_v_non_Root in hmn'; auto.
          rewrite hmn'.
          destruct (Between_Pred_m B) as [B'|B'].
          tauto.
          specialize (Cx _ _ B'); apply Cx, V.
        + apply has_moved_false in hmn';auto;destruct hmn' as [hmn' _]; auto.
          unfold v;rewrite hmn'.
          specialize (Cx _ _ B); apply Cx, V.
    }
    { rewrite convex_equiv in Cx; auto.
      destruct Cx as [[Cx _]|Cx].
      { assert (d: v (g Root) = (v (g Root) + 0) mod k) by
            (rewrite v_plus_0; auto); rewrite d in Cx at 2; clear d.
        assert (d: k <> 0) by lia.
        destruct (mod_plus_inv
                    (v_ltk (g Root))
                    (Nat.mod_upper_bound (v (g Root) + 0) _ d)) as [_ M];
          clear d.
        assert (d: 0 < k) by lia.
        specialize (M _ _ hhx d Cx eq_refl); clear d; lia.
      }
      exists x.
      split; [|lia].
      rewrite <- convex_b_true, convex_equiv; auto.
      right.
      intros n.
      case_eq (has_moved_b g' g n); intros hmn.
       assert (nr: ~ Root == n).
        { rewrite has_moved_v_change in hmn; auto.
          contradict hmn.
          rewrite <- (v_compat (pg' _ _ hmn));
            rewrite <- (v_compat (pg _ _ hmn)); lia. }
        eapply has_moved_v_non_Root in hmn; auto.
        rewrite hmn; apply Cx.
       apply has_moved_false in hmn;auto;destruct hmn as [hmn _]; auto.
        unfold v;rewrite hmn; apply Cx.
    }
  Qed.

  (** *** Step between non legitimate configurations is a wellfounded
   relation
   This is proven splitting the relation into 
     steps where the root moves
     steps where the root does not move.
  Those two set of steps are composed using Lexicographic order:
   the Lexicographic order propery is given Lemma 
    [values_do_not_appear_non_Root];
   the fact that steps where the root moves are wellfounded is 
    given directly by the corresponding potential decreasing;
   the fact that steps where the root does not move are wellfounded 
    is given directly by applying the DershowitzManna criteria given
    by local and global criteria.
  *)
  Lemma Wf_safe:
    well_founded (fun a b: safeEnv => Step (getEnv a) (getEnv b)).
  Proof.
    apply Inclusion.wf_incl with
        (safeQStep (Relation_Operators.union
                      QTrans_with_root QTrans_without_root)).
    {
      intros g' g hg.
      unfold safeQStep, QTrans_with_root, QTrans_without_root.
      split; [apply hg|].
      destruct (Nat.eq_dec (v (proj1_sig g' Root)) (v (proj1_sig g Root)));
        [right | left]; auto.
    }
    eapply Wf_lexico
      with (S1 := fun sg' sg => Z_pot sg' < Z_pot sg)
           (E1 := fun sg' sg => Z_pot sg' = Z_pot sg).
    { intros sg sg' h; apply Z_pot_decreases; auto. }
    { apply Inverse_Image.wf_inverse_image.
      apply lt_wf.
    }
    { intros sg' sg'' sg h1 h2; rewrite h1, h2; auto. }
    { intros sg' sg [gg [hgg hhgg]]; rewrite <- hhgg; auto. }
    { apply Trans_without_root_wf. }
    {
      unfold safeQStep, QTrans_with_root, safeStep.
      intros [g [assg nonleg]] [g' [assg' nonleg']] [step safe];
        simpl in *.
      assert (VDR := values_do_not_appear_non_Root assg step safe).
      unfold Z_pot; simpl.
      lia.
    }
  Qed.

  Lemma Acc_non_leg_step:
    forall (g: Env), Proper fequiv g -> Assume g ->
                     Acc (fun a b => ~ Leg_state b /\ Step a b) g.
  Proof.
    intros g pg assume.
    destruct (Leg_state_dec g) as [leg|non_leg];
      [specialize (leg pg); constructor; intuition|].
    pose (safe_g := (exist (P:=Safe)  (conj assume non_leg))).
    specialize (Wf_safe safe_g).
    replace g with (proj1_sig safe_g) by reflexivity.
    clearbody safe_g; clear g pg non_leg assume.
    change safeEnv in safe_g.
    induction (Wf_safe safe_g) as [[g sg] hrecg hh].
    constructor.
    intros g' [_ step];simpl in *.
    destruct (Leg_state_dec g') as [leg|non_leg];
       [specialize (leg (Step_Proper_l step)); constructor; intuition|].
    assert (assume': Assume_RO (ROEnv_part g')) by
        (apply step_Assume with g; [apply sg | auto]).
    replace g' with (proj1_sig (exist (P:=Safe) (conj assume' non_leg)))
      by reflexivity.
    apply hh; auto.
  Qed.

  (** ** Convergence Theorem (Alternative Proof) **)
  Theorem Dijkstra_convergence_alt:
    convergence Dijkstra_RO_assumption unfair_daemon Leg_state.
  Proof.
    apply Convergence_to_Legitimate.
    { intros g pg _; destruct (Leg_state_dec g) as [LS|LS];
        [left; apply LS | right]; auto.
    }
    { intros g pg ass term; contradict term;
        apply no_terminal;auto.
    }
    { apply Acc_non_leg_step. }
  Qed.

End Dijkstra_Convergence.

Unset Implicit Arguments.
Close Scope nat_scope.
