(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- March 2017
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Lia.

(** **** Local Imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import Stream.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Exec.
From Padec Require Import Ring_topology.
From Padec Require Import Fairness.
From Padec Require Import Self_Stabilization.
From Padec Require Import Dijkstra_algo.
From Padec Require Import Dijkstra_common.
From Padec Require Import Dijkstra_specifications.
From Padec Require Import Dijkstra_complexity.
From Padec Require Import Dijkstra_complexity_degenerated.
From Padec Require Import Steps.

(* end show *)

Open Scope nat_scope.
Set Implicit Arguments.

(**
 * Diskstra's first token ring algorithm
  - Proof of self-stabilization and specification
  - Proof of a complexity bound in steps
  - Proof that this complexity bound is exact (by building a worst case
    prefix of execution)
 **)

(**
* Specifications Proved

   /contains:/
   Proof that the algorithm is self-stabilizing wrt token ring
   circulation

   /assumes:/ Parameter k>1; the network is a unidirectional rooted
   ring; N <= k

   - Closure: The result is direct from the result shown in
   Dijkstra_common, in particular, Lemmas token_obtained and
   two_tokens_move.

   - Convergence: The result is direct from complexity.

   - Specification: As we have (Always Leg_State), (Always is_exec)
     and (Always Assume_RO), we have to prove that for every node, for
     every execution beginning in Assume_RO and Leg_State, it
     eventually reaches a configuration where the node holds the
     token.
     This proof is done by induction on the Ring traversal (using
     Succ), starting from the token holder.  The step of induction
     relies on Lemma token_passes_to_Succ: after a step, a node that
     move transmits its token to its successor.
 **)

Section Dijkstra_spec_proved.

  Context {Chans: Channels}
          {Net: Network}
          {RURN: Rooted_Unidirectional_Ring_Network Net}
          {DP: Dijkstra_parameters}.

  Notation Algo := Dijkstra_algo.
  Notation Root := (RURN_root (Net := Net)).
  Notation Ring := (URN_topo (Net := Net)).
  Notation Env := (Env (Net := Net) (Algo := Dijkstra_algo)).

  Existing Instance Dijkstra_RO_assumption.

  Notation N := (Ring_size Root).
  Hypothesis HNk: N <= k.

  (** ** Closure 
      i.e. Step preserves legitimate configurations
   *)
  Lemma Dijkstra_closure_small:
    forall g g', Assume g -> Leg_state g -> Step g' g -> Leg_state g'.
  Proof using HNk.
    intros g g' ass ls step.

    (* degenerated case (1 node) *)
    case (Compare_dec.le_gt_dec (Ring_size Root) 1); intros HN.
    { assert (R := Ring_size_pos Root).
      apply Single_Node_Legitimate; [lia|lia| apply step].
    }
    
    (* N >= 2 *)
    intros n1 n2 t1' t2'.
    case (Token_dec g n1); case (Token_dec g n2); intros t2 t1.
    { apply ls; auto. }
    { apply (token_obtained HNk ass step) in t2'; auto.
      assert (tt2 := has_moved_true_Token HNk HN ass step _ t2').
      assert (h1: n1 == Pred n2) by (apply ls; auto).
      assert (hn: Token g' (Pred n2))
        by (revert t1';
            apply Token_compat; [apply step| symmetry; apply h1]). 
      assert (tt := two_tokens_move HNk ass step HN t2' hn).
      apply has_moved_true_Token in t2';
        apply has_moved_true_Token in tt; auto.
      assert (C: n2 == Pred n2)
        by (apply Pred_injective; apply ls; auto).
      apply Succ_compat in C; rewrite Succ_Pred in C.
      contradict C.
      apply Non_degen2; lia.
    }
    { apply (token_obtained HNk ass step) in t1'; auto.
      assert (tt2 := has_moved_true_Token HNk HN ass step _ t1').
      assert (h1: n2 == Pred n1) by (apply ls; auto).
      assert (hn: Token g' (Pred n1))
        by (revert t2';
            apply Token_compat; [apply step| symmetry; apply h1]). 
      assert (tt := two_tokens_move HNk ass step HN t1' hn).
      apply has_moved_true_Token in t1';
        apply has_moved_true_Token in tt; auto.
      assert (C: n1 == Pred n1)
        by (apply Pred_injective; apply ls; auto).
      apply Succ_compat in C; rewrite Succ_Pred in C.
      contradict C.
      apply Non_degen2; lia.
    }
    { apply (token_obtained HNk ass step) in t2;
        apply (token_obtained HNk ass step) in t1; auto.
      apply has_moved_true_Token in t1;
        apply has_moved_true_Token in t2; auto.
      apply Pred_injective, ls; auto.
    }
  Qed.
    
  (** ** Closure Lemma *)
  Lemma Dijkstra_closure: closure Dijkstra_RO_assumption
                                  unfair_daemon Leg_state.
  Proof using HNk.
    unfold closure.
    cofix h.
    intros [g | g e]; intros ass ise uf leg.
    left; auto.
    right; auto.
    apply h.
    apply step_Assume with g.
    apply ass.
    inversion ise; auto.
    inversion ise; auto.
    tauto.
    apply Dijkstra_closure_small with g; auto.
    inversion ise; auto.
  Qed.
  
  (** ** Convergence: eventually a legitimate configuration is 
    reached
    Obtained from complexity *)

  Lemma Dijkstra_convergence:
    convergence Dijkstra_RO_assumption unfair_daemon Leg_state.
  Proof using HNk.
    intros e ass ise daem.
    assert (tt := Ring_size_pos Root).
    destruct (Compare_dec.le_gt_dec (Ring_size Root) 2).
    { eapply Steps_complexity_reach.
      apply Dijkstra_bound_n_is_1_or_2; auto; lia.
      split; auto. }    
    destruct (Compare_dec.le_gt_dec (Ring_size Root) 3).
    { eapply Steps_complexity_reach.
      apply Dijkstra_bound_n_is_3; auto; lia.
      split; auto. }        
    { eapply Steps_complexity_reach.
      apply Dijkstra_bound_non_degenerated_cases; auto; lia.
      split; auto. }        
  Qed.

  Section Convergence_alternative_proof.
    
  End Convergence_alternative_proof.
    
  (** ** Specification *)
    (**

    As we have (Always Leg_State), (Always is_exec) and 
    (Always Assume_RO), we have to prove that for every node, 
    for every execution beginning in Assume_RO and Leg_State, it
    eventually reaches a configuration where the node holds the token.
    This proof is done by induction on the Ring traversal (using Succ), 
    starting from the token holder.
    The step of induction relies on Lemma [token_passes_to_Succ]:
    after a step, a node that move transmits its token to its
    successor.
  **)

  Lemma token_passes_to_Succ:
    forall g g', Assume g -> Leg_state g -> Step g' g ->
                 forall n, Token g n -> Token g' (Succ n).
  Proof using HNk.
    intros g g' ass ls step.
    assert (ls' := Dijkstra_closure_small ass ls step).
    assert (pg: Proper pequiv g) by apply step.
    assert (pg': Proper pequiv g') by apply step.
    intros n h.

    (* degenerated case (1 node) *)
    case (Compare_dec.le_gt_dec (Ring_size Root) 1); intros HN.
    { assert (R := Ring_size_pos Root).
      assert (RR: Ring_size Root = 1) by lia; clear R HN.
      destruct (Token_exists pg').
      rewrite <- (Token_compat
                    pg' _ _ (Degenerate_ring_one RR x (Succ n))); auto.
    }

    (* at least two nodes *)
    destruct (Step_one_has_moved_true step) as [nn hm].
    assert (hh: n == nn).
    { apply ls; auto.
      apply (has_moved_true_Token HNk HN ass step); auto. }
    rewrite <- hh in hm; clear hh nn.

    assert (Non: forall x, ~ x == n -> has_moved_b g' g x = false).
    { intros x hx.
      case_eq (has_moved_b g' g x); intros hhx; auto.
      contradict hx.
      apply has_moved_true_Token in hhx; auto. }

    assert (CC: v (g (Succ n)) = v (g' (Succ n))).
    { apply v_compat.
      symmetry; rewrite <- (has_moved_false g' g (Succ n)).
      apply Non.
      intros F; symmetry in F; contradict F;
        eapply Non_Degenerated_size_Succ; eauto. }
    
    unfold Token.
    rewrite <- CC.
    rewrite (v_compat (pg' _ _ (Pred_Succ n))).

    case (Node_dec Root (Succ n)); intros hr.
    { left; split; auto.
      assert (V: ~ Root == n).
      { contradict hr.
        rewrite hr.
        eapply Non_Degenerated_size_Succ; eauto. }

      assert (VV := has_moved_v_non_Root ass step _ V hm).
      { rewrite VV.
        dup ls lls.
        rewrite ls_charac in lls; auto.
        destruct lls as [nn [tnn [ttnn _]]].
        rewrite <- (v_compat (pg _ _ hr)).
        apply ttnn.
        assert (n == nn) by (apply ls; auto).
        eapply Between_rewrite_r.
        eapply Pred_compat; symmetry; apply H.
        apply Between_self_r.
      }
    }
    { right; split; auto.
      intros F; rewrite F in CC.
      assert (Token g (Succ n)).
      unfold Token.
      right; split; auto.
      intros FF; rewrite FF in F.
      rewrite (v_compat (pg _ _ (Pred_Succ n))) in F.
      apply has_moved_v_change in hm; tauto.
      assert (NN: n == Succ n) by (apply ls; auto).
      contradict NN; eapply Non_Degenerated_size_Succ; eauto.
    }
  Qed.
  
  Lemma Dijkstra_spec_ok:
    forall (e: Exec) (assume: Assume (s_hd e))
           (ise: is_exec e) (legit: Leg_state (s_hd e)),
    forall n, Always (Eventually (fun e => Token (s_hd e) n)) e.
  Proof using HNk.
    intros e assume ise legit n.
    case (Compare_dec.le_gt_dec N 1); intros HN.
    (* N = 1 *)
    { assert (hn: N = 1).
      { assert (hp := Ring_size_pos Root). lia. }
      clear legit assume.
      revert ise; revert e; cofix h.
      intros [g | g e] ise.
      left; left; apply Single_Node_has_token; auto.
      apply is_exec_s_hd_compat in ise; auto.
      right; [left|].
      apply Single_Node_has_token; auto.
      apply is_exec_s_hd_compat in ise; auto.
      apply h.
      apply is_exec_inv_2 in ise; auto.
    }
    
    (* N > 1 *)
    assert (h1: Always (fun e_ => Leg_state (s_hd e_)) e).
    { apply Dijkstra_closure; auto.      
      unfold Fairness.unfair_daemon; auto.
    }
    assert (h2: Always is_exec e) by (apply is_exec_Always; auto).
    assert (h3: Always (fun e => Assume_RO (ROEnv_part (s_hd e))) e).
    {
      eapply RO_is_exec_Always; auto.
      apply (Assume_RO_compat (Stable_Assumption := Dijkstra_RO_assumption)). 
    }
    revert h1; apply Always_MP.
    revert h2; apply Always_MP.
    revert h3; apply Always_impl_P.
    
    clear e ise assume legit.
    intros e assume ise legit.
    destruct (Token_exists 
                (is_exec_s_hd_compat ise)) as [nn tok].
    pattern n.
    apply Ring_Succ_ind with nn; clear n.
    * intros n1 n2 en; simpl.
      assert (pe := is_exec_proper ise).
      eapply Eventually_compat;[ |apply pe].
      intros s1 s2 es.
      apply Token_compat; auto.
      apply hd_compat, es.
    * left; apply tok.
    * intro n; clear nn tok.
      apply AE_tool.
      assert (h1: Always (fun e_ => Leg_state (s_hd e_)) e).
      { apply Dijkstra_closure; auto.
        unfold Fairness.unfair_daemon; auto.
      }
      assert (h2: Always is_exec e) by (apply is_exec_Always; auto).
      assert (h3: Always (fun e => Assume_RO (ROEnv_part (s_hd e))) e).
      {
        eapply RO_is_exec_Always; auto.
        apply (Assume_RO_compat (Stable_Assumption :=
                                   Dijkstra_RO_assumption)). 
      }
      revert h1; apply Always_MP.
      revert h2; apply Always_MP.
      revert h3; apply Always_impl_P.
      clear e assume ise legit.
      intros e assume ise legit hn.
      assert (pe := is_exec_proper ise).
      destruct e as [g | g e].
      { destruct (is_exec_inv_0 ise) as [_ hh].
        destruct (Step_exists HNk HN assume (hd_compat pe)) as [g' hhh].
        elimtype False; eapply hh, hhh.
      }
      right; left.
      assert (hh := is_exec_inv_1 ise).
      simpl in hn, legit.
      apply token_passes_to_Succ with g; auto.
  Qed.

  (** * Functional Specification of the Algorithm. *)
  Theorem Dijkstra_self_stab_prop:
    self_stabilization Dijkstra_RO_assumption unfair_daemon
                       Token_Circulation_Spec.
  Proof using HNk.
    split with Leg_state.
    split; [| split].
    - apply Dijkstra_closure.
    - apply Dijkstra_convergence; apply k_ring_size.
    - intros e assume he legit; split with Token; split.
      + cut (Always (fun e => Leg_state (s_hd e)) e).
        * unfold Leg_state; auto.
        * apply Dijkstra_closure; auto.
      + apply Dijkstra_spec_ok;auto.
  Qed.

End Dijkstra_spec_proved.

Unset Implicit Arguments.
Close Scope nat_scope.
