(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- March 2017
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Bool.
From Coq Require Import Lia.
From Coq Require Import Lia.

(** **** Local Imports *)
From Padec Require Import Stream.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Exec.
From Padec Require Import Ring_topology.
From Padec Require Import Dijkstra_algo.
From Padec Require Import Dijkstra_common.
From Padec Require Import Self_Stabilization.
From Padec Require Import Steps.
From Padec Require Import NatUtils.
From Padec Require Import SetoidUtil.
From Padec Require Import Count.
From Padec Require All_In.
From Padec Require FoldRight.
From Padec Require Import ListUtils.

(* end show *)

Set Implicit Arguments.
Open Scope nat_scope.

(**
 * Diskstra's first token ring algorithm
  - Proof of self-stabilization and specification
  - Proof of a complexity bound in steps
  - Proof that this complexity bound is exact (by building a worst case
    prefix of execution)
 **)

(**
* Potential sum_dist 

   /contains:/
   Definitions and results about sum_dist

   /assumes:/ Parameter k>1; the network is a unidirectional rooted
   ring;

   - sum_dist: sum of distances to the root of every token holder
   - sum_dist increases by at most N-1 when the root moves and
     decreases otherwise.
   - sum_dist is naturally upper bounded by the sum of the distances
     of every node to the root.
 **)

Section Dijkstra_sum_dist.

  Context {Chans: Channels}
          {Net: Network}
          {RURN: Rooted_Unidirectional_Ring_Network Net}
          {DP: Dijkstra_parameters}.

  Notation Algo := Dijkstra_algo.
  Notation Root := (RURN_root (Net := Net)).
  Notation Ring := (URN_topo (Net := Net)).

  Existing Instance Dijkstra_RO_assumption.

  (** The ring contains at least 3 nodes *)
  Notation N := (Ring_size Root).
  Hypothesis HN4: N >= 4.
  Hypothesis HNk: N <= k.

  (** ** Definitions for the sum of distances to the root of every
      token holder. 
      We define [sum_dist] which is the complete sum and then
      [sd] which is a partial sum computed on a given slice of the
      ring.
   *)

  (** accumulation function *)
  Notation f_dtok g :=
    (fun n acc => acc + if Token_dec g n then Dist n Root else 0).

  Lemma f_dtok_compat: Proper pequiv (fun g => f_dtok g).
  Proof.
    intros g1 g2 pg n1 n2 en a1 a2 aa.
    rewrite aa.    
    assert (T := Token_compat pg _ _ en).
    destruct (Token_dec g1 n1) as [t1|t1];
      destruct (Token_dec g2 n2) as [t2|t2].
    rewrite en; auto.
    contradict t1; rewrite T; auto.
    contradict t1; rewrite T; auto.
    auto.
  Qed.

  (** sum_dist: sum of distances to the root of every
      token holder. *)
  Definition sum_dist := (fun g => Ring_fold_Pred Root 0 (f_dtok g)).    
  (* fold_right (f_dtok g) 0 (traversal_Pred N Root) *)

  Instance sum_dist_compat: Proper pequiv sum_dist.
  Proof.
    intros g1 g2 eg; unfold sum_dist.
    erewrite Ring_fold_Pred_compat.
    2: reflexivity.
    2: reflexivity.
    2: apply (f_dtok_compat eg).
    auto.
  Qed.

  (** sd: partial sum computed on a given slice of the
      ring *)
  Definition sd := (fun g p q => Between_fold
                                   0 (f_dtok g) p q).
  
  Instance sd_compat:
    Proper (pequiv ==> equiv ==> equiv ==> eq) sd.
  Proof.
    intros g1 g2 eg n1 n2 en m1 m2 em.
    change (sd g1 n1 m1 =~= sd g2 n2 m2).
    unfold sd.
    apply Between_fold_compat; auto.
    reflexivity.
    apply f_dtok_compat; auto.
  Qed.

  (** **** Tools about [sd] 
      to reduce or split the slice on which sd is computed *)
  Lemma sd_Succ:
    forall g p q,
      sd g p q = (if Token_dec g p then Dist p Root else 0) +
                 if Node_dec p q then 0 else  sd g (Succ p) q.
  Proof.
    intros g p q.
    unfold sd, Between_fold.
    assert (T := traversal_Between_def p q).
    destruct (equiv_dec p q) as [n|n];
      destruct (Node_dec p q) as [n'|n']; try tauto.
    { rewrite T; simpl.
      lia.
    }
    rewrite T.
    simpl fold_right; lia.
  Qed.
  
  Lemma sd_Pred:
    forall g (pg: Proper pequiv g) p q,
      sd g p q = (if Token_dec g q then Dist q Root else 0) +
                 if Node_dec p q then 0 else sd g p (Pred q).
  Proof.
    intros g pg p q.
    unfold sd, Between_fold.
    assert (T := traversal_Between_def_rev p q).
    destruct (equiv_dec p q) as [n|n];
      destruct (Node_dec p q) as [n'|n']; try tauto.
    { rewrite app_nil_l in T.
      erewrite FoldRight.fold_right_compat.
      2: apply f_dtok_compat, pg.
      2: reflexivity.
      2: apply T.
      simpl; lia.
    }
    rewrite <- rev_involutive in T.
    rewrite rev_unit in T.
    erewrite FoldRight.fold_right_compat.
    3: reflexivity.
    3: apply T.
    2: apply f_dtok_compat, pg.
    rewrite fold_left_rev_right.
    
    assert (dum: fold_left
                   (fun x y => x + (if Token_dec g y then Dist y Root else 0))
                   (q :: rev (traversal_Between p (Pred q))) 0
                 =
                 fold_left (fun x y =>
                              x + (if Token_dec g y then Dist y Root else 0))
                           (rev (traversal_Between p (Pred q)))
                           (if Token_dec g q then Dist q Root else 0)).
    simpl; auto.
    rewrite dum; clear dum.
    rewrite <- fold_left_rev_right.
    erewrite FoldRight.fold_right_compat.
    3: reflexivity.
    2: apply f_dtok_compat, pg.
    2: rewrite rev_involutive; reflexivity.

    assert(K: forall init l, fold_right (f_dtok g) init l =
                             init + fold_right (f_dtok g) 0 l).
    { intros init; induction l; simpl.
      lia.
      rewrite IHl; lia. }
    rewrite K; auto.
  Qed.
  
  Lemma sd_cut:
    forall g (pg: Proper pequiv g) p q,
    forall n, Between p n q -> ~ n == q ->
              sd g p q = sd g p n + sd g (Succ n) q. 
  Proof.
    intros g pg p q.
    assert (fc: Proper pequiv (f_dtok g)).
    { apply f_dtok_compat; auto. }
    intros n B neq.
    assert (npq: ~ p == q).
    { contradict neq.
      eapply Between_rewrite_r in B; [|apply neq].
      apply Between_self in B.
      rewrite <- B; auto.
    }
    revert B npq neq; revert n.
    apply (Between_Succ_ind
             (P := (fun nn => ~ p == q -> ~ nn == q ->
                              sd g p q = sd g p nn +
                                         sd g (Succ nn) q))).
    { intros a b ab; rewrite ab; auto. }
    { intros pq _.
      unfold sd at 2; unfold Between_fold.
      simpl.
      rewrite Dist_self; simpl.
      rewrite sd_Succ.
      destruct (Node_dec p q); [tauto|auto].
    }
    { intros nn B h N1 N2.
      case (Node_dec nn q); intros hq.
      eapply Between_rewrite_m in B; [|symmetry; apply hq].  
      apply Between_not_Pred_l in B; auto; tauto.
      rewrite h; auto.
      clear h.

      destruct (Node_dec p (Succ nn)) as [hp|hp].
      { eapply Between_rewrite_l in B; [|symmetry; apply hp].
        contradict B; apply Between_not_Succ_r.
        intros F.
        apply Succ_compat in F.
        rewrite Succ_Pred in F; tauto.
      }

      assert (S := sd_Succ g (Succ nn) q).
      destruct (Node_dec (Succ nn) q); [tauto|].
      rewrite S; clear S.
      assert (S := sd_Pred pg p (Succ nn)).
      destruct (Node_dec p (Succ nn)); [tauto|].
      rewrite Pred_Succ in S.
      rewrite S; clear S.
      lia.
    }
  Qed. 

  Lemma sd_one:
    forall g x, sd g x x = if Token_dec g x then Dist x Root else 0.
  Proof.
    intros g x; unfold sd, Between_fold; simpl.
    rewrite Dist_self; simpl; auto.
  Qed.
  
  (** decomposition for sd that exihibits the value of a given 
     node in the slice. *)
  Lemma sd_decomp:
    forall g (pg: Proper pequiv g),
    forall a b x,
      Between a x b ->
      sd g a b =
      (if Node_dec x a then 0 else sd g a (Pred x)) +
      (if Token_dec g x then Dist x Root else 0) +
      (if Node_dec x b then 0 else sd g (Succ x) b).
  Proof.
    intros g pg a b x B.
    case (Node_dec x b); intros xb.
    { rewrite <- xb.
      rewrite sd_Pred; auto.
      destruct (Node_dec x a); destruct (Node_dec a x); auto;
        try (symmetry in e; tauto); lia. }
    { rewrite sd_cut with (n := x); auto.
      case (Node_dec x a); intros xa.
      { rewrite <- xa, sd_one; lia. }
      { rewrite sd_Pred; auto.
        destruct (Node_dec a x); auto;
          try (symmetry in e; tauto); lia. }
    }
  Qed.
  
  (** *** Relation between [sd] and [sum_dist] *)
  Lemma sum_dist_sd:
    forall g (pg: Proper pequiv g) x, sum_dist g = sd g x (Pred x).
  Proof.
    intros g pg x; unfold sum_dist, sd.
    erewrite <- (fold_Between_Ring 0).
    2: reflexivity.
    2: apply f_dtok_compat; auto.
    unfold Ring_fold_Pred, Ring_fold_Succ.
    eapply fold_right_equivlistA_restr; auto.
    3: apply f_dtok_compat; auto.
    apply setoid_equiv.
    apply eq_equivalence.
    { intros xx y xy.
      assert (D: ~ xx == y -> ~ y == xx).
      { intros F; contradict F; symmetry; auto. }
      apply D, xy.
    }
    { intros xx y xy a b ab; split; intros h.
      rewrite <- ab, <- xy; auto.
      rewrite ab, xy; auto.
    }
    { intros xx y z _; lia. }
    apply traversal_Pred_full_no_repeat.
    apply traversal_Succ_full_no_repeat.
    { apply NoDupA_altdef.
      apply traversal_Pred_full_no_repeat.
    }
    { rewrite traversal_Pred_Succ_equiv.
      split; intros _; eapply traversal_Succ_full_all.
    }
  Qed.

  Lemma sum_dist_sd_Root:
    forall g (pg: Proper pequiv g),
      sum_dist g = sd g (Succ Root) (Pred Root).
  Proof.
    intros g pg.
    rewrite (sum_dist_sd pg Root).
    rewrite sd_Succ.
    rewrite Dist_self.
    case (Node_dec Root (Pred Root)); intros hn.
    { apply Succ_compat in hn.
      rewrite Succ_Pred in hn.
      contradict hn; apply Non_degen2; lia. }
    destruct (Token_dec g Root); lia.
  Qed.

  (** *** Tools for sd (at most) decreasing *)
  
  (** When no node obtains a token in some slice, then the sum
      of the slice does not increase 
      (proof by induction on the list of the nodes in the slice) *)
  Lemma sd_no_increase:
    forall g g',
    forall a b,
      (forall n, Between a n b -> Token g' n -> Token g n)
      -> sd g' a b <= sd g a b.
  Proof.
    intros g g' a b nogain; unfold sd, Between_fold.
    assert (R := incl_refl (traversal_Between a b)).
    pose (tb := (traversal_Between a b)).
    fold tb.
    fold tb in R; unfold tb at 2 in R.
    clearbody tb.      
    induction tb as [| x tb ih]; simpl.
    lia.
    specialize (ih (ListUtil.incl_cons_l_incl R)).
    apply ListUtil.incl_cons_l_in in R.
    eapply In_InA in R; [|apply setoid_equiv].
    apply traversal_Between_Between in R.
    specialize (nogain _ R).
    destruct (Token_dec g' x); destruct (Token_dec g x);
      intuition; lia.
  Qed.

  (** When no node obtains a token, but at least one loses a token,
      sd decreases *)
  Lemma sd_decrease:
    forall g (pg: Proper pequiv g) g' (pg': Proper pequiv g'),
    forall a b,
      (forall n, Between a n b -> Token g' n -> Token g n) ->
      (exists n, ~ n == Root /\
                 Between a n b /\ Token g n /\ ~ Token g' n) ->
      sd g' a b < sd g a b.
  Proof.
    intros g pg g' pg' a b nogain [x [nr [Bx [tokx tokx']]]].
    rewrite (sd_decomp pg Bx); rewrite (sd_decomp pg' Bx).
    destruct (Token_dec g x); [|tauto].
    destruct (Token_dec g' x); [tauto|].
    assert (Dist x Root > 0).
    { cut (~ Dist x Root = 0); [lia|].
      contradict nr; rewrite <- Dist_refl; auto. }
    assert (Sb: ~ x == b -> sd g' (Succ x) b <= sd g (Succ x) b).
    { intros xb.
      apply sd_no_increase; auto.
      intros nn B; apply nogain.
      apply (Between_trans3 Bx).
      destruct (Between_Pred_l B) as [bb|bb].
      rewrite Pred_Succ in bb; tauto.
      revert bb; apply Between_compat; auto; symmetry;
        apply Pred_Succ.
    }
    assert (Sa: ~ x == a -> sd g' a (Pred x) <= sd g a (Pred x)).
    { intros xa.
      apply sd_no_increase; auto.
      intros nn B; apply nogain.
      apply (Between_trans2 Bx).
      destruct (Between_Succ_r B) as [bb|bb].
      rewrite Succ_Pred in bb; symmetry in bb; tauto.
      revert bb; apply Between_compat; auto; symmetry;
        apply Succ_Pred.
    }
    case (Node_dec x a); intros xa; [|specialize (Sa xa)];
      case (Node_dec x b); intros xb; [|specialize (Sb xb)|
                                       |specialize (Sb xb)];
      lia.
  Qed.

  (** *)
  Lemma Dist_increases:
    forall a b n, ~ a == Pred b -> Between a n b -> ~ n == b ->
                  ~ n == a -> Dist b a < Dist n a.
  Proof.
    intros a b n apb.
    case (Node_dec a b); intros ab.
    { intros B nb.
      eapply Between_rewrite_l in B; [|symmetry; apply ab].
      apply Between_self in B.
      contradict nb; symmetry; auto.
    }
    apply (Between_Pred_ind
             (P := fun n => ~ n == b -> ~ n == a -> Dist b a < Dist n a)). 
    { intros n1 n2 en; rewrite en; auto. }
    { intros F; contradict F; auto. } 
    clear n; intros n B ih neq.
    case (Node_dec n b); intros nb.
    { intros _; rewrite nb.
      destruct (Dist_Pred_l b a) as [D|D].
      lia.
      apply Dist_refl in D.
      symmetry in D; tauto.
    }
    case (Node_dec n a); intros na.
    { eapply Between_rewrite_m in B; [| symmetry; apply na].
      contradict B; apply Between_not_Succ_r; auto.
    }      
    specialize (ih nb na).
    intros pna.
    destruct (Dist_Pred_l n a) as [D|D].
    lia.
    apply Dist_refl in D; tauto.
  Qed.

  (** In a slice, where no node obtains a token but the last one, 
      if some node inbetween actually loses a token, then sd 
      decrease. *)
  Lemma sd_decrease2:
    forall g (pg: Proper pequiv g) g' (pg': Proper pequiv g'),
    forall a b,
      Between Root a (Pred b) ->
      ~ Token g b -> Token g' b ->
      (forall n, Between a n (Pred b) -> Token g' n -> Token g n) ->
      (exists n, ~ n == Root /\
                 Between a n (Pred b) /\ Token g n /\ ~ Token g' n) ->
      sd g' a b < sd g a b.
  Proof.
    intros g pg g' pg' a sb Bab toksb toksb' nogain
           [x [xnr [B [tokx tokx']]]].
    pose (b := Pred sb).
    fold b in Bab, nogain, B.
    assert (dum: sb == Succ b) by
        (unfold b; rewrite Succ_Pred; auto).
    rewrite dum.
    rewrite (Token_compat pg _ _ dum) in toksb.
    rewrite (Token_compat pg' _ _ dum) in toksb'.
    clear dum; clearbody b; clear sb.
    rewrite (sd_Pred pg), (sd_Pred pg').
    destruct (Token_dec g (Succ b)); [tauto|clear n].
    destruct (Token_dec g' (Succ b)); [clear t|tauto].
    case (Node_dec a (Succ b)); intros asb.
    { rewrite <- (Token_compat pg _ _ asb) in toksb.
      rewrite <- (Token_compat pg' _ _ asb) in toksb'.
      specialize (nogain a (Between_self_l _ _)); intuition. }
    rewrite Pred_Succ.
    rewrite (sd_decomp pg B), (sd_decomp pg' B).
    destruct (Token_dec g x); [clear t|tauto].
    destruct (Token_dec g' x); [tauto|clear n].
    (* *)
    assert (D: Dist (Succ b) Root < Dist x Root).
    { assert (nxsb: ~ x == Succ b).
      { intros F;
          rewrite <- (Token_compat pg _ _ F) in toksb;
          tauto. }
      case (Node_dec (Succ b) Root); intros sbr.
      { rewrite sbr, Dist_self.
        cut (~ Dist x Root = 0); [lia|].
        contradict xnr; apply Dist_refl; auto. }
      eapply Dist_increases; auto.
      { rewrite Pred_Succ.
        contradict xnr.
        eapply Between_rewrite_r in Bab; [|apply xnr].
        apply Between_self in Bab.
        eapply Between_rewrite_r in B; [|apply xnr].
        eapply Between_rewrite_l in B; [|apply Bab].
        apply Between_self in B; symmetry; auto. }
      { eapply Between_trans3.
        destruct (Between_Succ_r Bab); [symmetry in H; tauto|apply H].
        destruct (Between_Succ_r B); [tauto|auto]. }
    }
    
    assert (L: ~ x == a -> sd g' a (Pred x) <= sd g a (Pred x)).
    { intros nax; apply sd_no_increase.
      intros n BB; apply nogain.
      eapply Between_trans2.
      apply B.
      destruct (Between_Succ_r BB).
      rewrite Succ_Pred in H; symmetry in H; tauto.
      revert H; apply Between_compat; auto; symmetry; apply Succ_Pred.
    }

    assert (R: ~ x == b -> sd g' (Succ x) b <= sd g (Succ x) b).
    { intros nbx; apply sd_no_increase.
      intros n BB; apply nogain.
      eapply Between_trans3.
      apply B.
      destruct (Between_Pred_l BB).
      rewrite Pred_Succ in H; tauto.
      revert H; apply Between_compat; auto; symmetry; apply Pred_Succ.
    }

    destruct (Node_dec x a) as [| xa]; [|specialize (L xa)];
      destruct (Node_dec x b) as [| xb]; [|specialize (R xb) | |
                                          specialize (R xb)];
      lia.
  Qed.
  
  (** ** LO
      We build the list of non-root nodes that obtain a token 
      during  a step from g to g'.  
      Note that this list is built traversing the node from 
      (Succ Root) to (Pred Root) in the reverse order.
   *)

  (** *** token_obtained
      We say that a node obtains a token during a step from g to g' 
      when it does not hold a token in g, but holds one in g'. *)
  Definition token_obtained g' g :=
    (fun n =>
       if Token_dec g n then false
       else if Token_dec g' n then true
            else false).

  Lemma token_obtained_true:
    forall g' g n, token_obtained g' g n = true <->
                   ~ Token g n /\ Token g' n.
  Proof.
    intros g' g n; unfold token_obtained.
    destruct (Token_dec g n); destruct (Token_dec g' n); intuition.
  Qed.

  Lemma token_obtained_false:
    forall g' g n, token_obtained g' g n = false <->
                   (Token g' n -> Token g n).
  Proof.
    intros g' g n; unfold token_obtained.
    destruct (Token_dec g n); destruct (Token_dec g' n); intuition.
  Qed.

  Instance token_obtained_compat: Proper pequiv token_obtained.
  Proof.
    intros g1 g2 eg q1 q2 eq n1 n2 en.
    unfold token_obtained.
    assert (T := Token_compat eg _ _ en).
    assert (TT := Token_compat eq _ _ en).
    destruct (Token_dec q1 n1) as [t11|t11];
      destruct (Token_dec q2 n2) as [t22|t22]; auto.
    contradict t22; rewrite <- TT; auto.
    contradict t11; rewrite TT; auto.
    destruct (Token_dec g1 n1) as [t1|t1];
      destruct (Token_dec g2 n2) as [t2|t2]; auto.
    contradict t2; rewrite <- T; auto.
    contradict t1; rewrite T; auto.
  Qed.

  (** *** LO *)
  Definition LO g' g :=
    filter (token_obtained g' g)
           (rev (traversal_Between (Succ Root) (Pred Root))).     

  (** Note that for every node x in LO,
      x itself does not move, its predecessor Pred x
      moves, and its successor Succ x either does not move or
      else loses its token (since it copies x's value and this
      value remains constant in the step).  
   *)
  Lemma TO_Pred_moves:
    forall g' g n, Assume g -> Step g' g ->
                   token_obtained g' g n = true ->
                   has_moved_b g' g (Pred n) = true.
  Proof.
    intros g' g n ass step h.
    rewrite token_obtained_true in h.
    apply Dijkstra_common.token_obtained; auto; 
      [lia|apply h|apply h].
  Qed.

  Lemma TO_not_move:
    forall g' g n, Assume g -> Step g' g ->
                   token_obtained g' g n = true ->
                   has_moved_b g' g n = false.
  Proof.
    intros g' g n ass step h.
    rewrite token_obtained_true in h.
    apply not_true_is_false.
    assert (N > 1) by lia.
    assert (hh := has_moved_true_Token HNk H ass step n).
    intuition.
  Qed.
  
  Lemma TO_not_Succ:
    forall g' g n, Assume g -> Step g' g ->
                   ~ Root == Succ n ->
                   token_obtained g' g n = true ->
                   has_moved_b g' g (Succ n) = false \/
                   ~ Token g' (Succ n).
  Proof.
    intros g' g n ass step hr h.
    rewrite token_obtained_true in h.
    case_eq (has_moved_b g' g (Succ n)); intros hm; [right|left]; auto.
    unfold Token; intuition.
    {
      apply has_moved_v_non_Root in hm; auto.
      contradict H3.
      rewrite hm.
      assert (pg: Proper pequiv g) by apply step.
      assert (pg': Proper pequiv g') by apply step.
      rewrite (v_compat (pg _ _ (Pred_Succ _))).
      rewrite (v_compat (pg' _ _ (Pred_Succ _))).
      apply v_compat; symmetry. 
      assert (X := has_moved_false g' g n); apply X; clear X.
      apply not_true_is_false.
      assert (N > 1) by lia.
      assert (hh := has_moved_true_Token HNk H2 ass step n).
      intuition.
    }
  Qed.

  (** *** Tools for LO decomposition
     We first define tools for LO decomposition. We consider the case
     when LO contains at least one element and then the case of
     at least two (consecutive) elements.
   *)

  (** **** If LO contains at least an element:
      LO can be decomposed as l ++ x ++ l' where x is a node and 
      l and l' are lists.
      Then, we have: 
       - token_obtained x
       - Between (Succ Root) x (Pred Root)
       - ~ x == Root
       - x == Succ Root -> l' == nil
       - x == Pred Root -> l == nil
      Those results come directly from the definitions and the 
      assumption N>=3. 
   *)
  Lemma LO_decomp_tokO_:
    forall g' g (pg': Proper pequiv g') (pg: Proper pequiv g),
    forall a b l x l',
      filter (token_obtained g' g)
             (rev (traversal_Between a b)) == l ++ x :: l' ->
      token_obtained g' g x = true.
  Proof.
    intros g' g pg' pg a b l x l' hlo.
    eapply proj2.
    apply filter_InA.
    apply token_obtained_compat; auto.
    rewrite hlo.
    apply InA_app_iff.
    right; apply InA_cons; left; auto.
  Qed.

  Lemma LO_decomp_tokO:
    forall g' g (pg': Proper pequiv g') (pg: Proper pequiv g),
    forall l x l', LO g' g == l ++ x :: l' ->
                   token_obtained g' g x = true.
  Proof.
    intros g' g pg' pg l x l' hlo.
    unfold LO in hlo; eapply LO_decomp_tokO_; auto; apply hlo.
  Qed.

  Lemma LO_decomp_B:
    forall g' g (pg': Proper pequiv g') (pg: Proper pequiv g),
    forall l x l', LO g' g == l ++ x :: l' ->
                   Between (Succ Root) x (Pred Root).
  Proof.
    intros g' g pg' pg l x l' hlo.
    apply traversal_Between_Between.
    apply InA_rev.
    eapply proj1.
    apply filter_InA.
    apply (token_obtained_compat pg' pg). 
    unfold LO in hlo; rewrite hlo.
    apply InA_app_iff.
    right; apply InA_cons; left; auto.
  Qed.

  Lemma LO_decomp_notRoot:
    forall g' g (pg': Proper pequiv g') (pg: Proper pequiv g),
    forall l x l', LO g' g == l ++ x :: l' -> ~ x == Root.
  Proof.
    intros g' g pg' pg l x l' hlo.
    assert (X := LO_decomp_B pg' pg _ _ _ hlo).
    apply Between_alt_m_bis in X.
    contradict X.
    eapply Between_rewrite_l; [apply Succ_Pred|].
    eapply Between_rewrite_r; [apply Pred_Succ|].
    apply Between_self; symmetry; auto.
    { intros F; apply Succ_injective in F; contradict F;
        apply Non_degen2; lia. }
    { intros F; rewrite Succ_Pred in F; 
        contradict F; apply Non_degen2; lia. }
  Qed.
  
  (** As there is no duplicate in a traversal list of node,
     if LO is split into 3 parts (l, x, l') then l and l' are the
     results of the filtering operation on the list before x, resp.
     after x.
   *)
  Lemma LO_decomp_decomp:
    forall g' g (pg': Proper pequiv g') (pg: Proper pequiv g),
    forall a b l x l',
      filter (token_obtained g' g)
             (rev (traversal_Between a b)) == l ++ x :: l' ->
      filter (token_obtained g' g)
             (rev (if Node_dec x b then nil else
                     traversal_Between (Succ x) b)) == l /\
      filter (token_obtained g' g)
             (rev (if Node_dec x a then nil else
                     traversal_Between a (Pred x))) == l'.
  Proof.
    intros g' g pg' pg a b l x l' hlo.
    eapply ListUtils.filter_app_preserve.
    apply token_obtained_compat; auto.
    4: apply hlo.
    { rewrite <- ListUtils.rev_app_cons.
      apply eqlistA_rev.
      assert (hh: ~ Succ Root == Root) by (apply Non_degen2; lia).
      assert (hhh: Between a x b).
      { apply traversal_Between_Between.
        apply InA_rev.
        eapply proj1.
        apply filter_InA.
        apply (token_obtained_compat pg' pg). 
        rewrite hlo.
        apply InA_app_iff.
        right; apply InA_cons; left; auto.
      }
      assert (D := decomp_list _ hh hhh); clear hh hhh.
      revert D; case (Node_dec x a); case (equiv_dec x a); try tauto;
        case (Node_dec x b); case (equiv_dec x b); tauto.
    }
    { rewrite InA_rev.
      case (Node_dec x b ); intros xb.
      rewrite InA_nil; tauto.
      rewrite traversal_Between_Between.
      apply Between_not_Succ_r; auto. }
    { rewrite InA_rev.
      case (Node_dec x a); intros xa.
      rewrite InA_nil; tauto.
      rewrite traversal_Between_Between.
      apply Between_not_Pred_l; intros F; symmetry in F; auto. }
  Qed.      

  (** Particular case of the considered element being Succ Root
      to prove this result, we directly apply LO_decomp_decomp *)
  Lemma LO_decomp_first:
    forall g' g (pg': Proper pequiv g') (pg: Proper pequiv g),
    forall l x l', LO g' g == l ++ x :: l' -> x == Succ Root ->
                   l' == nil.
  Proof.
    intros g' g pg' pg l x l' hlo hn.
    unfold LO in hlo; apply LO_decomp_decomp in hlo; auto.
    destruct (Node_dec x (Succ Root)); [|tauto].
    destruct hlo as [_ hlo]; rewrite hlo; simpl; reflexivity.
  Qed.

  (** Particular case of the considered element being Pred Root
      to prove this result, we directly apply LO_decomp_decomp *)
  Lemma LO_decomp_last:
    forall g' g (pg': Proper pequiv g') (pg: Proper pequiv g),
    forall l x l', LO g' g == l ++ x :: l' -> x == Pred Root -> l == nil.
  Proof.
    intros g' g pg' pg l x l' hlo hn.
    unfold LO in hlo; apply LO_decomp_decomp in hlo; auto.
    destruct (Node_dec x (Pred Root)); [|tauto].
    destruct hlo as [hlo _]; rewrite hlo; simpl; reflexivity.
  Qed.

  (** **** If LO contains at least two elements:
      LO can be decomposed as l ++ x :: y :: l' where x and y are
      nodes and l and l' are lists.
      - ~ x == y (since a traversal contains no replicate)
      - ~ x == Succ y (two consecutive nodes in the ring cannot 
         be in the list, since the predecessor of a node in LO 
         actually moves in the step (hence has a token in g),
	 see Lemma token_obtained.)
      - ~ x == Succ Root
      - ~ y == Pred Root
      - Between (Succ Root) y (Pred x)
      - Between (Succ y) x (Pred Root)
      The 4 last results come directly from the definitions and 
      the assumption N>=3.
   *)
  Lemma LO_decomp_successive_:
    forall g' g (ass: Assume g) (step: Step g' g),
    forall a b x y l ll,
      filter (token_obtained g' g)
             (rev (traversal_Between a b)) == l ++ x :: y :: ll ->
      ~ x == Succ y.
  Proof.
    intros g' g ass step a b x y l ll hlo hsxy.
    assert (pg' := Step_Proper_l step).
    assert (pg := Step_Proper_r step).
    
    assert (hx: token_obtained g' g x = true).
    { eapply LO_decomp_tokO_; auto.
      rewrite hlo; reflexivity. }
    assert (hy: token_obtained g' g y = true).
    { eapply LO_decomp_tokO_; auto.
      rewrite hlo, <- ListUtil.list_app_first_last; reflexivity. }
    apply token_obtained_true in hx; destruct hx as [hx hhx].
    apply token_obtained_true in hy; destruct hy as [hy hhy].
    rewrite (Token_compat pg _ _ hsxy) in hx.
    rewrite (Token_compat pg' _ _ hsxy) in hhx.
    eapply Dijkstra_common.token_obtained in hx; eauto.
    2: lia.
    apply has_moved_b_enabled_b in hx; auto.
    rewrite Pred_Succ in hx.
    rewrite <- enabled_Token in hy; auto.
    lia.
  Qed.

  Lemma LO_decomp_successive:
    forall g' g (ass: Assume g) (step: Step g' g),
    forall x y l ll, LO g' g == l ++ x :: (y :: ll) ->
                     ~ x == Succ y.
  Proof.
    intros g' g ass step x y l ll hlo.
    unfold LO in hlo; apply LO_decomp_successive_ in hlo;
      auto.
  Qed.
  
  Lemma LO_decomp_successive_first:
    forall g' g (pg': Proper pequiv g') (pg: Proper pequiv g),
    forall x y l ll, LO g' g == l ++ x :: (y :: ll) ->
                     ~ x == Succ Root.
  Proof.
    intros g' g pg' pg x y l ll hlo.
    intros F.
    eapply LO_decomp_first in hlo; auto.
    inversion hlo.
  Qed.

  Lemma LO_decomp_successive_last:
    forall g' g (pg': Proper pequiv g') (pg: Proper pequiv g),
    forall x y l ll, LO g' g == l ++ x :: (y :: ll) ->
                     ~ y == Pred Root.
  Proof.
    intros g' g pg' pg x y l ll hlo.
    rewrite <- ListUtil.list_app_first_last in hlo. 
    intros F.
    apply LO_decomp_last in hlo; auto.
    symmetry in hlo.
    assert (X  := app_cons_not_nil l nil x).
    contradict hlo; contradict X.
    destruct (l ++ x :: nil); [auto|inversion X].
  Qed.

  (** Again, as there is no duplicate in a traversal list of nodes,
     if LO is split into 4 parts (l, x, y, l') then l and l' are the
     results of the filtering operation on the list before x, resp.
     after y, and the filtering operation on the list after x and 
     before y returns an empty list.
   *)
  Lemma LO_decomp_successive_decomp:
    forall g' g (ass: Assume g) (step: Step g' g),
    forall a b l x y l',
      filter (token_obtained g' g)
             (rev (traversal_Between a b)) == l ++ x :: y :: l' ->
      filter (token_obtained g' g)
             (rev (if Node_dec x b then nil else
                     traversal_Between (Succ x) b)) == l /\
      filter (token_obtained g' g)
             (rev (if Node_dec y a then nil else
                     traversal_Between a (Pred y))) == l' /\
      filter (token_obtained g' g)
             (rev (traversal_Between (Succ y) (Pred x))) == nil.
  Proof.
    intros g' g ass step a b l x y l' hlo.
    dup hlo hllo.
    assert (pg' := Step_Proper_l step).
    assert (pg := Step_Proper_r step).
    assert (foo := token_obtained_compat pg' pg).
    eapply LO_decomp_decomp in hlo; auto.
    split; [apply hlo|].
    destruct hlo as [_ hlo].
    destruct (Node_dec x a) as [xa|xa].
    { simpl in hlo; inversion hlo. }
    rewrite <- app_nil_l in hlo.
    eapply LO_decomp_decomp in hlo; auto.
    split; [apply hlo|].
    destruct hlo as [hlo _].
    destruct (Node_dec y (Pred x)) as [xy|xy].
    { apply Succ_compat in xy.
      rewrite Succ_Pred in xy.
      symmetry in xy; contradict xy;
        eapply LO_decomp_successive_; eauto.
    }
    apply hlo.
  Qed.
  
  Lemma LO_decomp_successive_1:
    forall g' g (pg': Proper pequiv g') (pg: Proper pequiv g),
    forall x y l ll,
      LO g' g == l ++ x :: y :: ll -> ~ x == y.
  Proof.
    intros g' g pg' pg x y l ll h.
    cut (NoDupA equiv (LO g' g)).
    { intros ok.
      rewrite h in ok.
      apply NoDupA_swap in ok; [|apply setoid_equiv].
      rewrite ListUtils.NoDupA_cons_iff in ok.
      destruct ok as [ok _].
      contradict ok.
      apply InA_app_iff.
      right; apply InA_cons; auto.
    }
    clear h x y ll l.
    unfold LO.
    apply ListUtils.filter_NoDupA.
    apply token_obtained_compat; auto.
    clear pg' g' pg g.
    apply NoDupA_rev; [apply setoid_equiv|].
    assert (X := traversal_Succ_full_no_repeat Root (Succ Root)).
    assert (T := traversal_Between_def_rev (Succ Root) Root).
    destruct (equiv_dec (Succ Root) Root) as [nn|_].
    { contradict nn; apply Non_degen2; lia. }
    unfold traversal_Between at 1 in T.
    assert (D: 1 + Dist (Succ Root) Root = N).
    { rewrite Dist_Succ_full with (x := Root).
      lia. }
    rewrite D in T; clear D.
    rewrite T in X; clear T.
    apply ListUtils.NoDupA_app_alt in X.
    apply X.
  Qed.
  
  (** Order between the two consecutive elements of LO
      to prove this result, we directly apply LO_decomp_decomp *)
  Lemma LO_decomp_successive_Bl:
    forall g' g (ass: Assume g) (step: Step g' g),
    forall x y l ll, LO g' g == l ++ x :: (y :: ll) ->
                     Between (Succ Root) y (Pred x). 
  Proof.
    intros g' g ass step x y l ll hlo.
    assert (pg' := Step_Proper_l step).
    assert (pg := Step_Proper_r step).
    assert (foo := token_obtained_compat pg' pg).
    apply traversal_Between_Between.      
    apply InA_rev.
    eapply proj1.
    apply filter_InA; [apply foo|].
    unfold LO in hlo; apply LO_decomp_decomp in hlo; auto.
    destruct hlo as [_ hlo].
    destruct (Node_dec x (Succ Root)) as [xsr|xsr].
    { simpl in hlo; inversion hlo. }
    rewrite hlo, InA_cons; left; auto.
  Qed.

  (** Order between the two consecutive elements of LO
      to prove this result, we directly apply LO_decomp_decomp *)
  Lemma LO_decomp_successive_Br:
    forall g' g (pg': Proper pequiv g') (pg: Proper pequiv g),
    forall x y l ll, LO g' g == l ++ x :: (y :: ll) ->
                     Between (Succ y) x (Pred Root). 
  Proof.
    intros g' g pg' pg x y l ll hlo.
    assert (foo := token_obtained_compat pg' pg).
    apply traversal_Between_Between.
    apply InA_rev.
    eapply proj1.
    apply filter_InA; [apply foo|].
    unfold LO in hlo;
      rewrite <- ListUtil.list_app_first_last in hlo.
    apply LO_decomp_decomp in hlo; auto.
    destruct hlo as [hlo _].
    destruct (Node_dec y (Pred Root)) as [xsr|xsr].
    { simpl in hlo. 
      assert (T := app_cons_not_nil l nil x).
      contradict T; symmetry; rewrite <- ListUtils.l_eqlistA_nil.
      symmetry in hlo; apply hlo. }
    rewrite hlo, InA_app_iff; right; rewrite InA_cons; left; auto.
  Qed.

  Lemma LO_decomp_successive_Bl':
    forall g' g (ass: Assume g) (step: Step g' g),
    forall x y l ll, LO g' g == l ++ x :: (y :: ll) ->
                     Between Root (Succ y) (Pred x).
  Proof.
    intros g' g ass step x y l ll hlo.
    dup hlo hllo.
    apply LO_decomp_successive_Bl in hlo; auto.
    destruct (Between_Pred_l hlo) as [B|B].
    rewrite Pred_Succ in B.
    apply Succ_compat in B; rewrite Succ_Pred in B; symmetry in B.
    contradict B; apply LO_decomp_successive_first in hllo; auto; apply step.
    eapply Between_rewrite_l in B; [| symmetry; apply Pred_Succ]. 
    destruct (Between_Succ_m B) as [BB|BB].
    apply LO_decomp_successive in hllo; auto.
    rewrite BB, Succ_Pred in hllo; contradict hllo; auto.
    auto.
  Qed.

  (** *** 3) Remark: in any slice of the ring between two 
      consecutive elements of LO (resp in the first slice, before
      flo) (resp in the last slice after llo), no node obtains a
      token, hence sd computed on this slice does not increase.
   *)

  (** **** Results for Remark 3                          *)
  Lemma _no_TO_2cases:
    forall g g' (ass: Assume g) (step: Step g' g),
      let flo := match LO g' g with nil => Root | a::l => a end in
      ~ flo == Pred Root ->
      forall n, Between (Succ flo) n (Pred Root) ->
                Token g' n -> Token g n.
  Proof.
    intros g g' ass step flo nf.
    assert (pg := Step_Proper_r step).
    assert (pg' := Step_Proper_l step).
    unfold flo in *; clear flo.
    case_eq (LO g' g).
    { intros hlo .
      unfold LO in hlo.
      intros n B.
      case_eq (token_obtained g' g n); intros tok.
      { apply traversal_Between_Between in B.
        rewrite <- InA_rev in B.
        eapply conj in B; [apply and_comm in B|].
        eapply filter_InA in B.
        rewrite hlo, InA_nil in B; tauto.
        apply (token_obtained_compat pg' pg).
        auto. }
      { apply token_obtained_false; auto. }
    }
    { intros ll lo hlo.
      rewrite hlo in nf.
      rewrite <- app_nil_l in hlo.
      assert (hhlo: LO g' g == nil ++ ll :: lo) by (rewrite hlo; auto);
        clear hlo.
      unfold LO in hhlo.
      eapply LO_decomp_decomp in hhlo; auto.
      destruct hhlo as [hlo _].
      destruct (Node_dec ll (Pred Root)) as [t|t]; [tauto|].
      intros n B.
      case_eq (token_obtained g' g n); intros tok.
      { apply traversal_Between_Between in B.
        rewrite <- InA_rev in B.
        eapply conj in B; [apply and_comm in B|].
        eapply filter_InA in B.
        rewrite hlo, InA_nil in B; tauto.
        apply (token_obtained_compat pg' pg).
        auto. }
      { apply token_obtained_false; auto. }
    }
  Qed.

  Lemma LO_nil_no_TO:
    forall g g' (ass: Assume g) (step: Step g' g),
      LO g' g = nil ->
      forall n, Between (Succ Root) n (Pred Root) ->
                Token g' n -> Token g n.
  Proof.
    intros g g' ass step hlo n B; apply _no_TO_2cases; auto.
    { rewrite hlo.
      intros F; apply Succ_compat in F; rewrite Succ_Pred in F.
      contradict F; apply Non_degen2; lia.
    }
    { rewrite hlo; auto. }
  Qed.

  Lemma fist_slice_no_TO:
    forall g g' (ass: Assume g) (step: Step g' g) flo l,
      LO g' g = flo :: l -> ~ flo == Pred Root ->
      forall n, Between (Succ flo) n (Pred Root) ->
                Token g' n -> Token g n.
  Proof.
    intros g g' ass step flo l hlo nf n B; apply _no_TO_2cases;
      auto; rewrite hlo; auto.
  Qed.      

  Lemma last_slice_no_TO:
    forall g' g (ass: Assume g) (step: Step g' g) llo l,
      LO g' g == l ++ llo :: nil -> ~ Succ Root == llo ->
      forall n, Between (Succ Root) n (Pred llo) ->
                Token g' n -> Token g n.
  Proof.
    intros g' g ass step x l Rx srx.
    assert (pg' := Step_Proper_l step).
    assert (pg := Step_Proper_r step).
    unfold LO in Rx; apply LO_decomp_decomp in Rx; auto.
    destruct Rx as [_ Rx].
    destruct (Node_dec x (Succ Root)) as [t|t];
      [symmetry in t; tauto|].
    { intros n B.
      case_eq (token_obtained g' g n); intros tok.
      { apply traversal_Between_Between in B.
        rewrite <- InA_rev in B.
        eapply conj in B; [apply and_comm in B|].
        eapply filter_InA in B.
        rewrite Rx, InA_nil in B; tauto.
        apply (token_obtained_compat pg' pg).
        auto. }
      { apply token_obtained_false; auto. }
    }
  Qed.

  Lemma mid_slice_no_TO:
    forall g' g (ass: Assume g) (step: Step g' g) x y l ll,
      LO g' g == l ++ x :: y :: ll ->
      forall n, Between (Succ y) n (Pred x) ->
                Token g' n -> Token g n.
  Proof.
    intros g' g ass step x y l ll Rx.
    assert (pg' := Step_Proper_l step).
    assert (pg := Step_Proper_r step).
    unfold LO in Rx; apply LO_decomp_successive_decomp in Rx; auto.
    destruct Rx as [_ [_ Rx]].
    { intros n B.
      case_eq (token_obtained g' g n); intros tok.
      { apply traversal_Between_Between in B.
        rewrite <- InA_rev in B.
        eapply conj in B; [apply and_comm in B|].
        eapply filter_InA in B.
        rewrite Rx, InA_nil in B; tauto.
        apply (token_obtained_compat pg' pg).
        auto. }
      { apply token_obtained_false; auto. }
    }
  Qed.

  (** *** Lemma witness

      The proof is mainly an induction on LO. It mainly 
      relies the idea that for every node n in LO (and
      even if LO is empty) we can find a witness node which
      loses a token and is farther than n from the root.

      Lemma witness: in a slice [m - n] of nodes, if n moves, 
      then either m moves and
      holds a token in g' or there exits a node l between m and 
      n which loses its token.

      Lemma witness is proven by induction on m using an
      induction schema of Between.  For the base case, (the
      property has to be proven for n), n
      moves. Hence, either it loses its token or keeps it; both 
      cases prove the base case.  At each induction step, either
      we already found the witness node or the current node moves
      and keeps the token. By two_tokens_moves, we obtain that the 
      predecessor of the current node also moves, hence as in the
      base case, both cases are proven whatever be the fact that 
      the node keeps its token or not.
   *)
  Lemma witness:
    forall g' g (ass: Assume g) (step: Step g' g),
    forall m n,
      has_moved_b g' g n = true ->
      (Token g' m /\ has_moved_b g' g m = true) \/
      (exists x, Between m x n /\ Token g x /\ ~ Token g' x).
  Proof.
    intros g' g ass step.
    assert (pg' := Step_Proper_l step).
    assert (pg := Step_Proper_r step).
    apply (Between_all_ind
             (P := fun m n =>
                     has_moved_b g' g n = true ->
                     (Token g' m /\ has_moved_b g' g m = true) \/
                       (exists x, Between m x n /\ Token g x /\ ~ Token g' x))).

    { intros m1 m2 em n1 n2 en.
      eapply impl_compat.
      rewrite en; tauto.
      apply or_compat.
      apply and_compat.
      apply Token_compat; auto.
      apply eq_compat; auto.
      apply has_moved_b_compat; auto.
      apply ex_compat.
      intros nn.
      apply and_compat.
      apply Between_compat; auto.
      apply and_compat; tauto.
    }
    
    { intros m hm.
      case (Token_dec g' m); intros t.
      - left; split; auto.
      - right; exists m; split; [|split]; auto.
        apply Between_self_l.
        eapply has_moved_true_Token; eauto.
        lia.
    }
    intros m n hmn n' B ih hm.
    specialize (ih hm).
    destruct ih as [ih|ih].
    { destruct ih as [ih ihh].
      apply two_tokens_move in ihh; auto.
      2: lia.
      case (Token_dec g' (Pred n')); intros tokn.
      - left; split; auto.
      - right; exists (Pred n'); split; [|split]; auto.
        apply Between_self_l.
        apply has_moved_true_Token in ihh; auto.
        lia.
    }
    { right.
      destruct ih as [nn [hnn [hhnn hhhnn]]].
      exists nn.
      split; [|split]; auto.
      destruct (Between_Pred_l hnn); auto.
      { apply Succ_compat in H; rewrite Succ_Pred in H.
        eapply Between_rewrite_m in B; [|symmetry; apply H].
        contradict B.
        apply Between_not_Succ_m; auto.
        contradict hmn.
        apply Succ_injective in hmn; auto.
      }      
    }
  Qed.

  (** *** Case: LO is nil. 
      We already have "<=" using the remark above. During the step,
      - If the root moves, we are done.
      - Otherwise, as we consider a step,
        a non-node n moves: we apply Lemma witness to the slice
        [Succ Root - n]: as the root does not move, we exhibit
        a witness node that makes sum_dist decrease.
   *)
  Lemma sd_decr_LO_nil_RM:
    forall g g' (ass: Assume g) (step: Step g' g),
      LO g' g = nil ->
      has_moved_b g' g Root = true -> sum_dist g' <= sum_dist g.
  Proof.
    intros g g' ass step lonil hm.
    assert (pg := Step_Proper_r step).
    assert (pg' := Step_Proper_l step).
    rewrite sum_dist_sd_Root; auto.
    rewrite sum_dist_sd_Root; auto.
    apply sd_no_increase.
    apply LO_nil_no_TO; auto.
  Qed.
  
  Lemma sd_decr_LO_nil:
    forall g g' (ass: Assume g) (step: Step g' g),
      LO g' g = nil ->
      has_moved_b g' g Root = false -> sum_dist g' < sum_dist g.
  Proof.
    intros g g' ass step lonil hm.
    assert (pg := Step_Proper_r step).
    assert (pg' := Step_Proper_l step).
    rewrite sum_dist_sd_Root; auto.
    rewrite sum_dist_sd_Root; auto.
    apply sd_decrease; auto.
    apply LO_nil_no_TO; auto.

    { destruct (Step_one_has_moved_true step) as [n hmn].    
      destruct (witness ass step (Succ Root) _ hmn) as [[hx W]| [x [W WW]]].
      { apply two_tokens_move in W; auto.
        rewrite Pred_Succ in W; congruence.
        lia.
      }
      { assert (hx: ~ x == Root).
        { intros F.
          eapply Between_rewrite_m in W; [|symmetry in F; apply F].
          contradict W; apply Between_not_Succ_r.
          intros FF; rewrite <- FF in hmn; congruence.
        }
        exists x; split; [|split]; auto.
        assert (B := all_Between_Succ Root x).
        destruct (Between_Pred_r B); [tauto|auto].
      }
    }
  Qed.

  (** *** The rest of the proof is done by an induction on LO. It
      relies on the idea that for every node n in LO
      we can find a witness node which loses a token and is farther 
      than n from the root.
   *)

  (** For; the first slice of nodes, [Succ flo - Pred Root]
      we have <=.
   *)
  Lemma sd_first_slice:
    forall g g' (ass: Assume g) (step: Step g' g) flo l,
      LO g' g = flo :: l -> ~ flo == Pred Root ->
      sd g' (Succ flo) (Pred Root) <= sd g (Succ flo) (Pred Root).
  Proof.
    intros g g' ass step flo l hlo hflo.
    apply sd_no_increase.
    apply fist_slice_no_TO with l; auto.
  Qed.

  (** For every slice of nodes [Succ y - x], since
      Pred x moves, and Succ y either does not
      move or else loses its token, Lemma witness applied on 
      [Succ y - Pred x] exhibit a witness node which is
      farther than x the root and makes sd decrease on the slice.
   *)
  Lemma sd_mid_slice:
    forall g g' (ass: Assume g) (step: Step g' g) l x y ll,
      LO g' g == l ++ x :: y :: ll ->
      sd g' (Succ y) x < sd g (Succ y) x.
  Proof.
    intros g g' ass step l x y ll hlo.
    assert (pg := Step_Proper_r step).
    assert (pg' := Step_Proper_l step).      
    assert (tokx := LO_decomp_tokO pg' pg _ _ _ hlo).
    assert (pt := TO_Pred_moves _ ass step tokx).
    apply token_obtained_true in tokx; destruct tokx as [tokx tokx'].
    apply sd_decrease2; auto. 
    { apply LO_decomp_successive_Bl' in hlo; auto. }
    { eapply mid_slice_no_TO; eauto. }
    { assert (toky: token_obtained g' g y = true).
      { eapply LO_decomp_tokO; auto.
        rewrite hlo.
        transitivity ((l++x::nil) ++ y :: ll); [|reflexivity].
        rewrite ListUtil.list_app_first_last; auto.
      }
      apply TO_not_Succ in toky; auto.
      2: { apply LO_decomp_successive_last in hlo; auto.
           contradict hlo; rewrite hlo.
           rewrite Pred_Succ; auto. }
      destruct (witness ass step (Succ y) _ pt) as [[hx W]| [xx [W WW]]].
      destruct toky as [toky|toky]; [congruence|tauto].
      exists xx; split; [|split]; auto.
      {  dup hlo hllo.
         apply LO_decomp_successive_Bl' in hlo; auto.
         intros F; eapply Between_rewrite_m in W; [|symmetry; apply F].
         unfold Between in hlo, W.
         rewrite <- W in hlo.
         assert (D := Dist_add_Ring_size (Succ y) Root Root).
         destruct D as [D|D].
         - apply LO_decomp_successive_last in hllo; auto.
           contradict hllo; rewrite <- D, Pred_Succ; auto.
         - assert (DD := Dist_lt_Ring_size Root Root (Pred x)).
           lia.
      }
    }
  Qed.
  
  (** *** For the last slice [Succ Root - llo], if Succ Root 
      moves and still holds a token, then the root moves. 
      This is the case with no witness when sd increases by at most 
      N - 1 = Dist (Succ Root) Root in the slice. Otherwise, 
      Lemma witness applies to  
      [Succ Root - Pred llo] and exhibits a witness node which is
      farther than llo from the root and makes sd decrease on the
      slice.
   *)
  Lemma sd_last_slice:
    forall g g' (ass: Assume g) (step: Step g' g) l llo,
      LO g' g == l ++ llo :: nil ->
      if has_moved_b g' g Root then
        sd g' (Succ Root) llo <= sd g (Succ Root) llo + N - 1
      else
        sd g' (Succ Root) llo < sd g (Succ Root) llo.
  Proof.
    intros g g' ass step l llo hlo.
    assert (pg := Step_Proper_r step).
    assert (pg' := Step_Proper_l step).      

    assert (tt := LO_decomp_tokO pg' pg _ _ _ hlo).
    assert (pt := TO_Pred_moves _ ass step tt).
    apply token_obtained_true in tt.
    destruct tt as [tokllo tokllo'].
    
    case (Node_dec llo (Succ Root)); intros hllo.
    { rewrite hllo, Pred_Succ in pt.
      rewrite pt.
      erewrite (sd_compat pg); [|reflexivity| apply hllo].
      erewrite (sd_compat pg'); [|reflexivity|apply hllo].
      rewrite sd_one.
      rewrite sd_one.
      rewrite (Token_compat pg _ _ hllo) in tokllo.
      rewrite (Token_compat pg' _ _ hllo) in tokllo'.
      destruct (Token_dec g (Succ Root));
        destruct (Token_dec g' (Succ Root)); try tauto; clear n t.
      assert (D := Dist_Pred_full Root (Succ Root)).
      rewrite Pred_Succ in D; lia.
    }

    destruct (witness ass step (Succ Root) _ pt) as [[hx W]| [xx [W WW]]].
    { apply two_tokens_move in W; auto.
      rewrite Pred_Succ in W.
      rewrite W.
      rewrite (sd_Pred pg _ _); rewrite (sd_Pred pg' _ _).
      destruct (Token_dec g' llo); [|tauto]; clear t.
      destruct (Token_dec g llo); [tauto|]; clear n.
      cut (sd g' (Succ Root) (Pred llo) <= sd g (Succ Root) (Pred llo)).
      { intros ok.
        assert (D := Dist_lt_Ring_size Root llo Root).
        destruct (Node_dec (Succ Root) llo); lia.
      }

      apply sd_no_increase.
      eapply last_slice_no_TO; eauto. 
      contradict hllo; rewrite hllo; auto.
      lia.
    }        
    { cut (sd g' (Succ Root) llo < sd g (Succ Root) llo).
      { case (has_moved_b g' g Root); lia. }
      
      apply sd_decrease2; auto.
      { assert (B := Between_self_l Root (Pred llo)).
        destruct (Between_Succ_m B) as [BB|BB]; auto.
        rewrite BB, Succ_Pred in hllo; contradict hllo; auto.
      }
      { eapply last_slice_no_TO; eauto.
        contradict hllo; rewrite hllo; auto. }
      { exists xx; split; [|split]; auto.
        dup hlo hhlo.
        apply LO_decomp_B in hlo; auto.
        intros F; eapply Between_rewrite_m in W; [|symmetry; apply F].
        destruct (Between_Succ_r W).
        contradict hllo; rewrite H, Succ_Pred; auto.
        eapply Between_rewrite_r in H; [|symmetry; apply Succ_Pred].
        destruct (Between_Succ_r hlo).
        rewrite Succ_Pred in H0; 
          contradict H0; apply Non_degen2; lia.
        eapply Between_rewrite_r in H0; [|symmetry; apply Succ_Pred].
        unfold Between in H0, H.
        rewrite <- H in H0.
        assert (D := Dist_add_Ring_size Root llo Root).
        destruct D as [D|D].
        - apply LO_decomp_notRoot in hhlo; auto.
          symmetry in D; tauto.
        - assert (DD := Dist_lt_Ring_size Root (Succ Root) Root).
          lia.
      }              
    }
  Qed.
  
  (** ** Results on all the slices:
      Assuming LO not nil.
      We compute sd on the slice from Succ Root to flo. It either
      increases by at most N-1 when the root moves and decreases
      otherwise.

      The proof is done by induction on LO, proving that sd computed
      from sd to every element in LO follows the above property.
      Induction step uses Lemmas sd_mid_slice and sd_last_slice.
   *)
  Lemma sd_mid_last_slices:
    forall g g' (ass: Assume g) (step: Step g' g) flo l,
      LO g' g = flo :: l ->
      if has_moved_b g' g Root then
        sd g' (Succ Root) flo <= sd g (Succ Root) flo + N - 1
      else
        sd g' (Succ Root) flo < sd g (Succ Root) flo.
  Proof.
    intros g g' ass step flo l hlo.
    assert (pg := Step_Proper_r step).
    assert (pg' := Step_Proper_l step).
    assert (ass' := step_Assume ass step).

    cut (forall n, InA equiv n (LO g' g) ->
                   if has_moved_b g' g Root then
                     sd g' (Succ Root) n <= sd g (Succ Root) n + N - 1
                   else
                     sd g' (Succ Root) n < sd g (Succ Root) n).
    { intros ok; apply ok.
      rewrite hlo, InA_cons; left; auto. }
    clear flo hlo.
    
    pose (lo := LO g' g).

    (******** induction on the list LO ******************)
    fold lo.
    assert (R: exists l1, LO g' g == l1 ++ lo).
    { exists nil; auto. }
    induction lo as [| x lo ih]; simpl.
    { (* base *)
      intros n ina; apply InA_nil in ina; tauto. }

    (************** induction step ***********************)
    destruct R as [l1 Rx].
    assert (R: exists l1, LO g' g == l1 ++ lo).
    { exists (l1 ++ x :: nil).
      rewrite Rx.
      rewrite ListUtil.list_app_first_last; auto.
    }
    specialize (ih R).
    clear R.
    intros n ina.
    rewrite InA_cons in ina; destruct ina as [ina|ina].
    2: { apply ih; auto. }

    erewrite (sd_compat pg); [|reflexivity|apply ina].
    erewrite (sd_compat pg'); [|reflexivity|apply ina].
    clear n ina.

    destruct lo as [| y lo].
    { eapply sd_last_slice; eauto. }
    { assert (II: InA equiv y (y :: lo)).
      { rewrite InA_cons; left; auto. }
      specialize (ih _ II); clear II.
      assert (B: Between (Succ Root) y x).
      { assert (B := LO_decomp_successive_Bl ass step _ _ _ _ Rx). 
        destruct (Between_Succ_r B).
        rewrite Succ_Pred in H; symmetry in H; contradict H.
        eapply (LO_decomp_successive_first pg' pg); eauto.
        revert H; apply Between_compat; auto; symmetry; apply Succ_Pred.
      }
      assert (NN: ~ y == x).
      { intros F; symmetry in F; contradict F.
        eapply (LO_decomp_successive_1 pg' pg); eauto. }
      rewrite (sd_cut pg B NN), (sd_cut pg' B NN).        
      assert (SD := sd_mid_slice ass step _ _ _ _ Rx).
      destruct (has_moved_b g' g Root); lia.                
    }
  Qed.

  (** ** Lemma sum_dist_evolves

     Case LO is nil: sd_decr_LO_nil and sd_decr_LO_nil_RM provide 
     the result.

     Case LO not nil: sd_mid_last_slices and sd_first_slice provide 
     the result.
   *)
  Lemma sum_dist_evolves:
    forall g g', Assume g -> Step g' g ->
                 if has_moved_b g' g Root then
                   sum_dist g' <= sum_dist g + N - 1
                 else
                   sum_dist g' < sum_dist g.
  Proof.
    intros g g' ass step.
    assert (pg := Step_Proper_r step).
    assert (pg' := Step_Proper_l step).

    case_eq (LO g' g); [intros hlo | intros flo l hlo].
    { case_eq (has_moved_b g' g Root); intros hmr.
      - cut (sum_dist g' <= sum_dist g); [lia|].
        apply sd_decr_LO_nil_RM; auto.
      - apply sd_decr_LO_nil; auto.
    }
    { rewrite sum_dist_sd_Root; auto.
      rewrite sum_dist_sd_Root; auto.
      case (Node_dec flo (Pred Root)); intros hf.
      { rewrite <- (sd_compat pg _ _ (Equivalence_Reflexive _) _ _ hf).
        rewrite <- (sd_compat pg' _ _ (Equivalence_Reflexive _) _ _ hf).
        eapply sd_mid_last_slices; eauto.
      }
      assert (hhlo: LO g' g == nil ++ flo :: l).
      { rewrite hlo; rewrite app_nil_l; auto. }
      rewrite (sd_cut pg (LO_decomp_B pg' pg _ _ _ hhlo) hf). 
      rewrite (sd_cut pg' (LO_decomp_B pg' pg _ _ _ hhlo) hf). 
      clear hhlo.
      assert (SD1 := sd_mid_last_slices ass step hlo).
      assert (SD2 := sd_first_slice ass step hlo hf).
      destruct (has_moved_b g' g Root); lia.
    }
  Qed.      

  (** ** Bound on sum_dist *)

  (** sum_dist is maximal when every node holds a token,
      maximum given by the sum of distances 
     (induction on the slice on which sum_dist is computed)
   *)
  Lemma sum_dist_bound:
    forall g, Proper pequiv g -> 2 * sum_dist g <= N * (N - 1).
  Proof.
    intros g pg.
    rewrite sum_dist_sd_Root; auto.
    unfold sd, Between_fold.
    
    unfold Ring_size.

    cut (forall n,
            Between (Succ Root) n (Pred Root) ->
            2 * fold_right
                  (f_dtok g) 0
                  (traversal_Between n (Pred Root)) <=
            S (Dist n Root) * (S (Dist n Root) - 1)).
    { intros ok.
      apply ok.
      apply Between_self_l. }
    apply (Between_Pred_ind
             (P := fun n =>
                     2 *fold_right
                          (f_dtok g) 0
                          (traversal_Between n (Pred Root)) <=
                     S (Dist n Root) * (S (Dist n Root) - 1))).
    { intros n1 n2 en.
      apply le_compat.
      assert (fold_right (f_dtok g) 0 (traversal_Between n1 (Pred Root)) =
              fold_right (f_dtok g) 0 (traversal_Between n2 (Pred Root))).
      eapply FoldRight.fold_right_compat.
      apply f_dtok_compat, pg.
      auto.
      apply traversal_Between_compat; auto.
      rewrite H; auto.
      rewrite en; auto.
    }        
    { simpl.
      rewrite Dist_self; simpl.
      destruct (Token_dec g (Pred Root)); nia.
    }
    intros n B ih.
    rewrite (traversal_Between_def (Pred n) (Pred Root)).
    case (equiv_dec (Pred n) (Pred Root)); intros hpn.
    { apply Pred_injective in hpn.
      eapply Between_rewrite_m in B; [|symmetry; apply hpn].
      contradict B; apply Between_not_Pred_l.
      apply Non_degen3; lia.
    }
    destruct (Dist_Succ_l (Pred n) Root) as [D|D].
    2: { apply Dist_refl in D.
         apply Succ_compat in D; rewrite Succ_Pred in D.
         eapply Between_rewrite_m in B; [|symmetry; apply D].
         contradict B; apply Between_not_Succ_r.
         intros F; apply Succ_compat in F; rewrite Succ_Pred in F.
         contradict F; apply Non_degen3; lia. }
    rewrite Succ_Pred in D.
    rewrite D.
    assert (fold_right (f_dtok g) 0
                       (Pred n :: traversal_Between (Succ (Pred n))
                             (Pred Root)) =
            fold_right (f_dtok g) 0
                       (traversal_Between (Succ (Pred n)) (Pred Root)) +
            if Token_dec g (Pred n) then Dist (Pred n) Root else 0).
    { simpl; destruct (Token_dec g (Pred n)); lia. }
    rewrite H; clear H.
    assert (fold_right (f_dtok g) 0
                       (traversal_Between (Succ (Pred n)) (Pred Root)) =
            fold_right (f_dtok g) 0
                       (traversal_Between n (Pred Root))).
    { eapply FoldRight.fold_right_compat.
      apply f_dtok_compat, pg.
      lia.
      apply traversal_Between_compat; auto.
      rewrite Succ_Pred; auto.
    }
    rewrite H; clear H.
    destruct (Token_dec g (Pred n)); nia. 
  Qed.

  (** ** Technical lemma to handle particular cases for small values
     of N, exhbit the weight of two token holders in sum_dist 

    Induction on the list of nodes (given by (traversal_Pred N Root))
    if the liste contains both n1 and n2 then res >= Dist n1 Root + Dist n2 Root
    else if the list contains n1 then res >= Dist n1 Root
    else if the list contains n2 then res >= Dist n2 Root
   *)
  Lemma minus2:
    forall g (pg: Proper pequiv g)
           n1 n2,
      ~ n1 == n2 -> Token g n1 -> Token g n2 ->
      sum_dist g >= Dist n1 Root + Dist n2 Root.
  Proof.
    intros g pg n1 n2 hn tok1 tok2.    
    cut ((InA equiv n1 (traversal_Pred N Root) /\
          InA equiv n2 (traversal_Pred N Root) /\
          sum_dist g >= Dist n1 Root + Dist n2 Root) \/
         (InA equiv n1 (traversal_Pred N Root) /\
          ~InA equiv n2 (traversal_Pred N Root) /\
          sum_dist g >= Dist n1 Root) \/
         (~InA equiv n1 (traversal_Pred N Root) /\
          InA equiv n2 (traversal_Pred N Root) /\
          sum_dist g >= Dist n2 Root)
         \/ (~InA equiv n1 (traversal_Pred N Root) /\
             ~InA equiv n2 (traversal_Pred N Root)) ).
    { assert (h1 := traversal_Pred_full_all Root Root n1).
      assert (h2 := traversal_Pred_full_all Root Root n2).
      intros [[ _ [ _ ok]]|]; try tauto. 
    }
    clear HN4. 
    unfold sum_dist, Ring_fold_Pred.
    induction (traversal_Pred N Root) as [| n l ih].
    { right; right; right.
      repeat rewrite InA_nil; tauto.
    }    
    case (Node_dec n1 n); intros h1; case (Node_dec n2 n); intros h2.
    { contradict hn; rewrite h1, h2; auto. }
    { destruct ih as [ih|[ih|[ih|ih]]]. 
      { left; split; [| split].
        rewrite InA_cons; intuition. 
        rewrite InA_cons; intuition. 
        simpl; lia.        
      }
      { right; left; split; [|split].
        rewrite InA_cons; intuition.
        rewrite InA_cons; intuition.
        simpl; lia.
      }
      { left; split; [|split].
        rewrite InA_cons; intuition.
        rewrite InA_cons; intuition.
        simpl.
        destruct (Token_dec g n) as [|t].
        rewrite h1; lia.
        contradict t; revert tok1; apply Token_compat; symmetry; auto.        
      }
      { right; left; split; [|split].
        rewrite InA_cons; intuition.
        rewrite InA_cons; intuition.
        simpl.
        destruct (Token_dec g n) as [|t].
        rewrite h1; lia.
        contradict t; revert tok1; apply Token_compat; symmetry; auto.        
      }
    }
    { destruct ih as [ih|[ih|[ih|ih]]].
      { left; split; [| split].
        rewrite InA_cons; intuition. 
        rewrite InA_cons; intuition. 
        simpl; lia.        
      }
      { left; split; [|split].
        rewrite InA_cons; intuition.
        rewrite InA_cons; intuition.
        simpl.
        destruct (Token_dec g n) as [|t].
        rewrite h2; lia.
        contradict t; revert tok2; apply Token_compat; symmetry; auto.        
      }
      { right; right; left; split; [|split].
        rewrite InA_cons; intuition.
        rewrite InA_cons; intuition.
        simpl; lia.
      }
      { right; right; left; split; [|split].
        rewrite InA_cons; intuition.
        rewrite InA_cons; intuition.
        simpl.
        destruct (Token_dec g n) as [|t].
        rewrite h2; lia.
        contradict t; revert tok2; apply Token_compat; symmetry; auto.        
      }
    } 
    { destruct ih as [ih|[ih|[ih|ih]]]. 
      { left; split; [| split].
        rewrite InA_cons; intuition. 
        rewrite InA_cons; intuition. 
        simpl; lia.        
      }
      { right; left; split; [| split].
        rewrite InA_cons; intuition. 
        rewrite InA_cons; intuition. 
        simpl; lia.        
      }
      { right; right; left; split; [| split].
        rewrite InA_cons; intuition. 
        rewrite InA_cons; intuition. 
        simpl; lia.        
      }
      { right; right; right; split.
        rewrite InA_cons; intuition. 
        rewrite InA_cons; intuition. 
      }
    }
  Qed.
  
  
End Dijkstra_sum_dist.

Unset Implicit Arguments.
Close Scope nat_scope.
