
(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- March 2017
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Bool.
From Coq Require Import ZArith.
From Coq Require Import Lia.
From Coq Require Import ArithRing.

(** **** Local Imports *)
From Padec Require Import Stream.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Exec.
From Padec Require Import Ring_topology.
From Padec Require Import Dijkstra_algo.
From Padec Require Import Dijkstra_common.
From Padec Require Import Dijkstra_specifications.
From Padec Require Import Self_Stabilization.
From Padec Require Import Dijkstra_specifications_proved.
From Padec Require Import Steps.
From Padec Require Import NatUtils.
From Padec Require Import SetoidUtil.
From Padec Require Import Count.
From Padec Require All_In.
From Padec Require FoldRight.
From Padec Require Import ListUtils.
(* end show *)

Set Implicit Arguments.
Open Scope nat_scope.

(**
 * Diskstra's first token ring algorithm
  - Proof of self-stabilization and specification
  - Proof of a complexity bound in steps
  - Proof that this complexity bound is exact (by building a worst case
    prefix of execution)
 **)

(**
 * Worst Case Stabilization Time

   /assumes:/ Parameter k>1; the network is a unidirectional rooted
   ring; N <= k; N >= 4

   /contains:/
   Builds a prefix of execution which reaches a legitimate
   configuration using exactly the number of steps announced by the
   bound in Dijkstra_complexity
 **)

Section Simple.

  (** [simple_env] is the list of values for all nodes in a ring
      (see [mkEnv] which turns a [simple_env] into an [Env])
      First value is the counter at Root,
      followed by the values of counters at its Succ nodes. *)

  Definition simple_env := list nat.

  (** [simple_step se n] performs a step for node number n
      (= power Succ n Root) in simple env se.
      It is only correct for a node n that is enabled, i.e.
      holds a token).
      Note that a step, here, executes only one node.
   *)

  Fixpoint simple_step_Succ (ilast: nat) (se: simple_env) (n: nat):
    simple_env :=
    match se with
      nil => ilast:: nil
    | i::q =>
      match n with
        0 => ilast :: ilast :: q
      | S m => ilast :: simple_step_Succ i q m
      end
    end.

  Definition simple_step (se: simple_env) (n: nat): simple_env :=
    match se with
      nil => se
    | iroot::q =>
      match n with
        0 => S iroot :: q
      | S m => simple_step_Succ iroot q m
      end
    end.

  (** Properties for simple_step_Succ *)
  Lemma simple_step_Succ_spec:
    forall i q n, n<length q ->
                  nth (S n) (simple_step_Succ i q n) 0 =
                  nth n (simple_step_Succ i q n) 0.
  Proof.
    intros i q n;revert i q;induction n as [|n hn]; intros i q;destruct q;simpl;
      try (intros;lia).
    intro h;apply hn;lia.
  Qed.

  Lemma simple_step_Succ_spec2:
    forall i q n, n<length q ->
                  nth (S n) (simple_step_Succ i q n) 0 = nth n (i::q) 0.
  Proof.
    intros i q n;revert i q;induction n as [|n hn]; intros i q;destruct q;simpl;
      try (intros;lia).
    intro h;apply hn;lia.
  Qed.

  Lemma simple_step_Succ_spec3:
    forall i q n p, p<>S n ->
                    nth p (simple_step_Succ i q n) 0 = nth p (i::q) 0.
  Proof.
    intros i q n p;revert i q n;induction p as [|p hp]; intros i q n;
      destruct q;destruct n;simpl;try (intros;lia).
    destruct p;auto||congruence.
    intro h;apply hp;congruence.
  Qed.

  (** Properties for simple_step *)
  Lemma simple_step_spec:
    forall se n, nth (S n) (simple_step se 0) 0 = nth (S n) se 0.
  Proof.
    intros [|iroot q] n;simpl;auto.
  Qed.

  Lemma simple_step_spec2:
    forall se n p, p<>n -> nth p (simple_step se n) 0 = nth p se 0.
  Proof.
    intros [|iroot q] [|n] [|p];simpl;try (intros;lia).
    intros _;destruct q as [|i q];destruct n as [|n];reflexivity.
    intros.
    apply simple_step_Succ_spec3;auto.
  Qed.

  (** [simple_token se n] states that a Token is present at position n
    (node power Succ n Root is enabled) *)

  Definition simple_token (se: simple_env) (n: nat): Prop :=
    n < length se /\
    match n with
      0 => List.hd 0 se = List.last se 0 /\ S (List.hd 0 se) < length se
    | S m => List.nth (S m) se 0 <> List.nth m se 0
    end.

  Lemma simple_token_lt_N: forall se n,
      simple_token se n -> n<length se.
  Proof.
    intros se n [h _];auto.
  Qed.

  (** [simple_env_valid N se] states that se represents a
    valid environment on a size N ring:
    it contains N values, all less than N  *)

  Definition simple_env_valid (N: nat) (se: simple_env): Prop :=
    length se = N /\ Forall (gt N) se.

  (** Given a valid simple env and a node with a token, the simple
    env then produced by simple step is still valid *)
  Lemma simple_step_valid: forall N se n,
      simple_env_valid N se -> simple_token se n ->
      simple_env_valid N (simple_step se n).
  Proof.
    intros N se n.
    destruct n as [| n];
      destruct se as [| iroot q];
      unfold simple_env_valid,simple_step,simple_token.
    try (simpl;lia).
    {
      intros [h1 h2] [h3 h4];split;auto.
      inversion h2;subst.
      right;auto.
      simpl length.
      apply h4.
    }
    {
      intros [h1 h2] h3;split;auto.
    }
    {
      intros [h1 h2] h3;split;auto.
      {
        rewrite <- h1.
        clear - n.
        revert n iroot;induction q as [|i q hq];simpl;auto.
        destruct n as [ |n].
        tauto.
        intro iroot;simpl length.
        rewrite (hq n i).
        auto.
      }
      {
        clear - h2.
        revert n iroot h2;induction q as [|i q hq];simpl;auto.
        destruct n as [ |n].
        {
          intros ii h;inversion h;subst;
            inversion H2;subst.
          right;[|right];auto.
        }
        {
          intros ii h;inversion h;subst;
            inversion H2;subst.
          right;auto.
        }
      }
    }
  Qed.

  (** simple_sched: central scheduler presented by the list of
    nodes to be executed at each step *)

  Definition simple_sched := list nat.

  (** From a simple env se and a simple sched tr,
    [run_simple_sched se tr] builds the simple env which is reached
    after the execution of tr from se. *)
  Fixpoint run_simple_sched (se: simple_env) (tr: simple_sched): simple_env :=
    match tr with
    | nil => se
    | n :: q => run_simple_sched (simple_step se n) q
    end.

  Lemma run_simple_sched_app (se:simple_env) (tr1 tr2:simple_sched) :
    run_simple_sched se (tr1 ++ tr2) =
    run_simple_sched (run_simple_sched se tr1) tr2.
  Proof.
    revert se;induction tr1 as [ | x q hq];simpl;intro se;auto.
  Qed.

  (** Illegitimate simple env (two token exist). *)
  Definition simple_non_leg (se: simple_env): Prop :=
    exists i, exists j, simple_token se i /\ i <> j /\
                        simple_token se j.

  (** [simple_sched_has_tokens se tr] checks that all steps (from se
    and obtained using tr) are enabled *)
  Fixpoint simple_sched_has_tokens (se: simple_env) (tr: simple_sched): Prop :=
    match tr with
      nil => True
    | i :: q => simple_token se i /\
                simple_sched_has_tokens (simple_step se i) q
    end.

  (** [simple_sched_valid se tr] checks that all steps (from se and
    obtained using tr) are enabled and the final env is illegitimate
    i.e., it checks [simple_sched_has_tokens se tr] and that the
    simple env reached after executing tr from se is not legitimate.
   *)
  Fixpoint simple_sched_valid (se: simple_env) (tr: simple_sched): Prop :=
    match tr with
      nil => simple_non_leg se
    | i :: q => simple_token se i /\
                simple_sched_valid (simple_step se i) q
    end.

  Lemma simple_sched_valid_alt (se:simple_env) (tr:simple_sched) :
    simple_sched_valid se tr <->
    (simple_sched_has_tokens se tr /\ simple_non_leg (run_simple_sched se tr)).
  Proof.
    revert se;induction tr as [|x q hq];intro se;simpl;try tauto.
    specialize (hq (simple_step se x));tauto.
  Qed.

  Lemma simple_sched_app_valid (se:simple_env) (tr1 tr2:simple_sched) :
    simple_sched_has_tokens se tr1 ->
    simple_sched_valid (run_simple_sched se tr1) tr2 ->
    simple_sched_valid se (tr1 ++ tr2).
  Proof.
    revert se;induction tr1 as [|x q hq];intro se;simpl;auto.
    intros [h1 h2] h3;split;auto.
  Qed.

  Lemma simple_sched_app_tokens (se:simple_env) (tr1 tr2:simple_sched) :
    simple_sched_has_tokens se tr1 ->
    simple_sched_has_tokens (run_simple_sched se tr1) tr2 ->
    simple_sched_has_tokens se (tr1 ++ tr2).
  Proof.
    revert se;induction tr1 as [|x q hq];intro se;simpl;auto.
    intros [h1 h2] h3;split;auto.
  Qed.

  (*
  Eval compute in (stutter 0 (0::1::2::3::4::5::nil)).
  (* 0 :: 0 :: 1 :: 2 :: 3 :: 4 :: nil *)
  Eval compute in (stutter 1 (0::1::2::3::4::5::nil)).
  (* 0 :: 1 :: 1 :: 2 :: 3 :: 4 :: nil *)
  Eval compute in (stutter 2 (0::1::2::3::4::5::nil)).
  (* 0 :: 1 :: 2 :: 2 :: 3 :: 4 :: nil *)
   *)

  Lemma stutter_token_le: forall se j n,
      0 < n -> n <= j ->
      simple_token se n ->
      simple_token (stutter j se) n.
  Proof.
    intros se j n hn hj.
    destruct n as [|n];try lia;simpl.
    unfold simple_token.
    intros [h1 h2];split.
    rewrite stutter_length;auto.
    intro h;apply h2;clear h2.
    rewrite <- stutter_nth_le in h by lia.
    rewrite <- stutter_nth_le in h by lia.
    assumption.
  Qed.

  Lemma stutter_token_gt: forall se j n,
      j + 2 <= n -> n < length se ->
      simple_token se (n - 1) ->
      simple_token (stutter j se) n.
  Proof.
    intros se j n hj hn.
    destruct n as [|n];try lia;simpl.
    rewrite PeanoNat.Nat.sub_0_r.
    unfold simple_token.
    intros [h1 h2];split.
    rewrite stutter_length;lia.
    destruct n as [|n];try lia;simpl.
    intro h;apply h2;clear h2.
    rewrite <- stutter_nth_gt in h by lia.
    rewrite <- stutter_nth_gt in h;auto;try lia.
    simpl in h;rewrite PeanoNat.Nat.sub_0_r in h;auto.
  Qed.

  Lemma stutter_step:
    forall se j, 0 < j -> j < length se ->
                 simple_step (stutter j se) j =
                 stutter (j-1) se.
  Proof.
    intros se j h1 h2.
    destruct j as [|j];try lia.
    destruct se as [|x q];simpl in h2;try lia.
    simpl.
    rewrite PeanoNat.Nat.sub_0_r.
    clear h1.
    assert (h:j < length q) by lia;clear h2.
    revert x j h;induction q as [|y q hq];intros x j h;
      destruct j;simpl in *;try lia;auto.
    apply (f_equal (cons x)).
    apply hq;lia.
  Qed.

  Section Simple_Worst_Case.

    (** N represents the number of nodes in the ring. *)
    Variable N: nat.

    (** ** building blocks for simple_envs and simple execs. *)

    (** initial simple env for worst case execution
      For N=6, my_se = 0 :: 4 :: 3 :: 2 :: 1 :: 0 :: nil
     *)
    Definition my_se: simple_env := 0 :: dc 0 (N-1).

    (** simple scheduler for worst case execution
      For N=6, my_sched =
       0 :: 5 :: 4 :: 3 :: 2 :: 1 ::
       0 :: 5 :: 4 :: 3 :: 2 :: 1 ::
       0 :: 5 :: 4 :: 3 :: 2 :: 1 ::
       0 :: 5 :: 4 :: 3 :: 2 :: 1 ::
       0 :: 5 :: 4 :: 3 :: 2 :: 1 ::

            5 :: 4 :: 3 :: 2 ::
            5 :: 4 :: 3 :: nil
     *)
    Definition my_sched: simple_sched := 0 :: repl (dc 0 N) (N-2) ++
                                           flatmap (fun p => dc p (N-p))
                                           (uc 1 (N-3)) .

    Hypothesis hN4: N>=4.

    Lemma Nsimpl: exists n, N = 4+n.
    Proof.
      exists (N-4);lia.
    Qed.

    (** my_se has size N *)
    Lemma my_se_length: length my_se = N.
    Proof.
      unfold my_se.
      simpl.
      rewrite dc_length.
      lia.
    Qed.

    (** and my_sched has size Dijkstra_exact_bound N - 1 *)
    Lemma my_sched_length: length my_sched = Dijkstra_exact_bound N - 1.
    Proof.
      unfold my_sched.
      simpl.
      rewrite <- PeanoNat.Nat.add_1_r.
      rewrite app_length,repl_length,dc_length;simpl.
      rewrite flatmap_length.
      erewrite map_ext;[|intro a;eapply dc_length].
      replace (map (Init.Nat.sub N) (uc 1 (N - 3))) with
          (dc 3 (N-3)).
      2:{
        replace 3 with (N-(N-3)) at 1 by lia.
        assert(h:N-3<=N) by lia.
        revert h;
          generalize (N-3) as k;clear hN4;intro k.
        replace N with (N-0) at 1 2 by lia.
        change  (uc 1) with (uc (1+0)).
        assert (hh:0<=N)by lia.
        revert hh;generalize 0 at 1 2 3 5 as j.
        induction k as [|k hk];intros j hj h.
        - simpl;auto.
        - simpl.
          apply (f_equal2 cons).
          lia.
          destruct (Compare_dec.le_lt_eq_dec hj) as [LT|EQ].
          {
            replace (N - j - S k) with (N - S j -k) by lia.
            replace (S (j+1)) with (1+(S j)) by lia.
            apply hk;lia.
          }
          {
            subst.
            lia.
          }
      }
      unfold Dijkstra_exact_bound.
      rewrite <- PeanoNat.Nat.sub_add_distr.
      change (1+1) with 2.
      symmetry.
      rewrite <- triangle_pn.
      replace (3*triangle(N-1)) with
          (2*triangle (N-1) + triangle(N-1)).
      2:ring_simplify;reflexivity.
      rewrite -> PeanoNat.Nat.add_sub_swap by
          (assert (tmp:=triangle_le (N-1));lia).
      rewrite <- PeanoNat.Nat.add_sub_assoc by
          (assert (tmp:=triangle_le (N-1));lia).
      rewrite triangle_half_rectangle_pn.
      rewrite <- NPeano.Nat.add_assoc.
      apply f_equal2;[nia|].
      {
        symmetry.
        apply Minus.plus_minus.
        ring_simplify.
        induction N as [|n hn];try lia.
        destruct (Compare_dec.le_lt_eq_dec hN4) as [LT|EQ].
        2:rewrite <- EQ;reflexivity.
        replace (S n - 1) with (S (n-1)) by lia.
        simpl triangle.
        rewrite hn by lia.
        replace (S n - 3) with (S (n-3)) by lia.
        simpl.
        lia.
      }
    Qed.

    Lemma my_se_valid: simple_env_valid N my_se.
    Proof.
      split.
      {
        unfold my_se;simpl length.
        rewrite dc_length.
        lia.
      }
      {
        right;auto.
        lia.
        apply Forall_impl with (gt (0+(N-1))).
        intro;lia.
        apply dc_gt.
      }
    Qed.

    (** simple env definitions at various steps of the execution *)
    Definition se_NC0 iroot:= dc 1 iroot ++ dc (iroot-1) (N-iroot).
    Definition se_NC iroot j:= stutter j (se_NC0 iroot).
    Definition se_C0 lastconvex := rep (N-1) lastconvex ++ dc
                                       lastconvex (N-lastconvex).
    Definition se_C lastconvex j :=  stutter j (se_C0 lastconvex).

    Lemma se_NC0_length: forall iroot, iroot < N -> length (se_NC0 iroot) = N.
    Proof.
      intros;unfold se_NC0.
      rewrite app_length,dc_length,dc_length.
      lia.
    Qed.

    Lemma se_NC_length:
      forall iroot j, iroot < N -> length (se_NC iroot j) = N.
    Proof.
      intros;unfold se_NC.
      rewrite stutter_length;apply se_NC0_length;auto.
    Qed.

    Lemma se_C0_length: forall lastconvex, lastconvex < N ->
                                           length (se_C0 lastconvex) = N.
    Proof.
      intros;unfold se_C0.
      rewrite app_length,rep_length,dc_length.
      lia.
    Qed.

    Lemma se_C_length: forall lastconvex j, lastconvex < N ->
                                            length (se_C lastconvex j) = N.
    Proof.
      intros;unfold se_C.
      rewrite stutter_length;apply se_C0_length;auto.
    Qed.

    Lemma my_se_step: simple_step my_se 0 = se_NC0 1.
    Proof.
      reflexivity.
    Qed.

    Lemma se_NC0_NC: forall i, i < N ->
                               (se_NC0 i) = se_NC i (N-1).
    Proof.
      intros i hi;rewrite <- (stutter_last_id (se_NC0 i) (j:=N-1));auto.
      rewrite se_NC0_length;auto.
    Qed.

    Lemma se_NC_middle_step:
      forall i j, i < N -> 0 < j -> j < N ->
                  simple_step (se_NC i j) j = se_NC i (j-1).
    Proof.
      intros i j hi h1 h2.
      unfold se_NC.
      apply stutter_step;auto.
      rewrite se_NC0_length;try lia.
    Qed.

    Lemma se_NC_last_step: forall i,
        0 < i -> i<=N-2 ->
        simple_step (se_NC i 0) 0 = se_NC0 (i+1).
    Proof.
      intros [|i] hi1 hi2;try lia.
      replace (S i + 1) with (S (S i)) by lia.
      transitivity (S (S i) :: removelast (se_NC0 (S i)));[reflexivity|].
      unfold se_NC0.
      rewrite removelast_app.
      2:{
        intro h;apply (f_equal (length (A:=nat))) in h.
        rewrite dc_length in h;simpl in h;lia.
      }
      simpl dc.
      simpl app.
      repeat apply f_equal.
      rewrite dc_removelast by lia.
      apply f_equal2;lia.
    Qed.

    Lemma se_NC0_big_step:
      forall i, 0 < i -> i <= N-2 ->
                run_simple_sched (se_NC0 i) (dc 0 N) = se_NC0 (i+1).
    Proof.
      intros i h1 h2.
      rewrite se_NC0_NC;try lia.
      cut (forall j, j<=N-1 ->
                     run_simple_sched (se_NC i j) (dc 0 (S j)) = se_NC0 (i+1)).
      {
        intro h;replace N with (S (N-1)) at 2 by lia.
        apply h;lia.
      }
      intro j;induction j  as [|j hrecj];intro hj;simpl.
      apply se_NC_last_step;auto.
      rewrite <- hrecj by lia.
      simpl.
      apply f_equal2;auto.
      apply f_equal2;auto.
      rewrite se_NC_middle_step by lia.
      apply f_equal.
      lia.
    Qed.

    Lemma se_NC_all_steps:
      run_simple_sched (se_NC0 1) (repl (dc 0 N) (N-2)) = se_NC0 (N-1).
    Proof.
      cut (forall j, j <= N-2 ->
                     run_simple_sched (se_NC0 (N-1-j)) (repl (dc 0 N) j) =
                     se_NC0 (N-1)).
      {
        intro h;replace 1 with ((N-1)- (N-2)) at 1 by lia.
        apply h;lia.
      }
      intro j;induction j as [|j hrecj];intro hj;simpl.
      - rewrite PeanoNat.Nat.sub_0_r;reflexivity.
      - rewrite run_simple_sched_app.
        rewrite se_NC0_big_step by lia.
        replace (N - 1 - S j + 1) with (N-1-j) by lia.
        apply hrecj.
        lia.
    Qed.

    Lemma se_NC_C_step: simple_step (se_NC0 (N-1)) (N-1) = se_C 0 (N-2).
    Proof.
      rewrite <- (stutter_last_id (se_NC0 (N-1)) (j:=N-1)).
      2:{ rewrite se_NC0_length;lia||reflexivity. }
      rewrite stutter_step;try lia.
      2:{ rewrite se_NC0_length;lia||reflexivity. }
      unfold se_NC0,se_C,se_C0.
      simpl.
      replace (N-(N-1)) with 1 by lia.
      replace (N-1-1) with (N-2) by lia.
      simpl.
      rewrite PeanoNat.Nat.sub_0_r.
      rewrite PeanoNat.Nat.add_0_r.
      rewrite (app_removelast_last 0 (l:=dc 0 N)).
      2:{
        intro h;apply (f_equal (length (A:=nat))) in h.
        rewrite dc_length in h;simpl in h;lia.
      }
      rewrite dc_removelast;try lia.
      rewrite dc_last;try lia.
      simpl.
      cut (forall a b i (l:list nat),
              S i = length l -> stutter i (l++a::nil) = stutter i (l++b::nil)).
      {
        intro h;apply h;rewrite dc_length;lia.
      }
      intros a b i l;revert i;induction l as [|x q hq];
        intros i hi;simpl in hi;try lia.
      destruct i as [|i].
      {
        destruct q as [|y qq];simpl in hi;reflexivity||lia.
      }
      simpl;apply f_equal,hq;lia.
    Qed.

    Lemma se_C0_C: forall i,
        i<N -> se_C0 i = se_C i (N-1).
    Proof.
      intros i hi;rewrite <- (stutter_last_id (se_C0 i) (j:=N-1));try lia.
      reflexivity.
      rewrite se_C0_length;auto.
    Qed.

    Lemma se_C_middle_step: forall lc j,
        lc < j -> j < N ->
        simple_step (se_C lc j) j = se_C lc (j-1).
    Proof.
      intros lc j h1 h2.
      unfold se_C.
      apply stutter_step;try lia.
      rewrite se_C0_length;try lia.
    Qed.

    Lemma se_C_C0: forall lc,
        lc < N-1 -> se_C lc lc = se_C0 (lc+1).
    Proof.
      unfold se_C, se_C0;intros lc h.
      rewrite stutter_app_r;rewrite rep_length;try lia.
      replace (lc-lc) with 0 by lia.
      replace (N - lc) with (S (N - (lc + 1))) by lia.
      rewrite rep_app,<- app_assoc.
      apply f_equal.
      simpl;apply f_equal2;try lia.
      replace (N - (lc + 1)) with (S  (N - (lc + 1) -1)) at 1 by lia.
      simpl dc.
      cbv iota beta.
      replace (N - (lc + 1)) with (S  (N - (lc + 1) -1)) at 1 by lia.
      rewrite dc_removelast by lia.
      replace (N - (lc + 1)) with (S  (N - (lc + 1) -1)) at 3 by lia.
      simpl;apply f_equal2;lia|| reflexivity.
    Qed.

    Lemma se_NC_C_big_step:
      run_simple_sched (se_NC0 (N-1)) (dc 1 (N-1)) = se_C0 1.
    Proof.
      replace (N-1) with (S (N-2)) at 2 by lia.
      simpl dc.
      simpl run_simple_sched.
      replace (S (N-2)) with (N-1) by lia.
      rewrite se_NC_C_step.
      cut (forall j, j<N ->  run_simple_sched (se_C 0 j) (dc 1 j) = se_C0 1).
      {
        intro h;apply h;lia.
      }
      intro j;induction j  as [|j hrecj];intro hj;simpl.
      apply se_C_C0;try lia.
      rewrite <- hrecj by lia.
      rewrite se_C_middle_step by lia.
      apply f_equal2;auto.
      apply f_equal2;auto.
      lia.
    Qed.

    Lemma se_C0_big_step: forall lc,
        0 <= lc -> lc < N-1 ->
        run_simple_sched (se_C0 lc) (dc (lc+1) (N-(lc+1))) = se_C0 (lc+1).
    Proof.
      intros lc h1 h2.
      rewrite se_C0_C;try lia.
      cut (forall j , lc<=j -> j<N ->
                      run_simple_sched (se_C lc j) (dc (lc+1) (S j-(lc+1))) =
                      se_C0 (lc+1)).
      {
        intro h;replace N with (S (N-1)) at 2 by lia.
        apply h;lia.
      }
      intro j;induction j  as [|j hrecj];intros hj1 hj2;try lia.
      {
        destruct (Compare_dec.le_lt_eq_dec hj1);try lia.
        subst;simpl.
        apply se_C_C0;try lia.
      }
      {
        destruct (Compare_dec.le_lt_eq_dec hj1);try lia.
        {
          rewrite <- hrecj by lia.
          replace (S (S j) - (lc + 1)) with (S (j - lc)) by lia.
          simpl run_simple_sched at 1.
          replace (lc + 1 + (j - lc)) with (S j) by lia.
          rewrite se_C_middle_step by lia.
          repeat (apply f_equal2;try lia).
        }
        {
          subst.
          replace (S (S j) - (S j + 1)) with 0 by lia.
          simpl.
          rewrite se_C_C0 by lia.
          apply f_equal;lia.
        }
      }
    Qed.

    Lemma se_C_all_steps:
      run_simple_sched (se_NC0 (N-1)) (flatmap (fun p => dc p (N-p))
                                               (uc 1 (N-3))) =
      (se_C0 (N-3)).
    Proof.
      replace (N-3) with (S (N-4)) at 1 by lia.
      simpl.
      rewrite run_simple_sched_app.
      rewrite se_NC_C_big_step.
      cut (forall lc j,
              lc + j = N-3 ->
              run_simple_sched (se_C0 lc)
                               (flatmap (fun p : nat => dc p (N - p))
                                        (uc (lc+1) j)) = se_C0 (N-3)).
      {
        intro h;apply h;lia.
      }
      intros lc j;revert lc;induction j as [|j hrecj];intros lc hj.
      {
        rewrite PeanoNat.Nat.add_0_r in hj.
        simpl;subst;auto.
      }
      {
        simpl.
        rewrite run_simple_sched_app.
        rewrite se_C0_big_step;try lia.
        apply hrecj;lia.
      }
    Qed.

    Lemma my_se_token: simple_token my_se 0.
    Proof.
      unfold my_se,simple_token.
      simpl length;rewrite dc_length.
      simpl hd.
      intuition try lia.
      simpl last.
      rewrite dc_last by lia.
      destruct Nsimpl as [n hn];subst;simpl.
      reflexivity.
    Qed.

    Lemma se_NC0_tokens: forall i j,
        1<=i -> i<N ->
        1<=j -> j<N ->
        simple_token (se_NC0 i) j.
    Proof.
      intros i j h1 h2 h3 h4.
      destruct j as [|j];try lia;clear h3.
      hnf.
      rewrite se_NC0_length;auto.
      split;auto.
      unfold se_NC0.
      destruct (Compare_dec.lt_eq_lt_dec (S j) i) as [[LT | EQ]|GT ].
      {
        rewrite app_nth1;try rewrite dc_length;[|lia].
        rewrite app_nth1;try rewrite dc_length;[|lia].
        rewrite dc_nth;[|lia].
        rewrite dc_nth;[|lia].
        nia.
      }
      {
        rewrite app_nth2;try rewrite dc_length;[|lia].
        rewrite app_nth1;try rewrite dc_length;[|lia].
        subst.
        rewrite dc_nth;[|lia].
        rewrite dc_nth;[|lia].
        nia.
      }
      {
        rewrite app_nth2;try rewrite dc_length;[|lia].
        rewrite app_nth2;try rewrite dc_length;[|lia].
        rewrite dc_nth;[|lia].
        rewrite dc_nth;[|lia].
        nia.
      }
    Qed.

    Lemma se_NC_root_token : forall i, 0<i -> i<=N-2 ->
                                       simple_token (se_NC i 0) 0.
    Proof.
      intros i h1 h2.
      unfold se_NC,se_NC0.
      simpl dc;simpl app.
      destruct i as [|i];try lia.
      clear h1.
      change (simple_token
                (S i :: removelast ( (S i :: dc 1 i) ++
                                                     dc (i - 0) (N - S i))) 0).
      rewrite removelast_app.
      2: {
        intro hh;apply (f_equal (length (A:=nat))) in hh;
        rewrite dc_length in hh;
        simpl in hh;lia.
      }
      unfold simple_token.
      repeat apply conj;try (simpl;lia).
      {
        rewrite dc_removelast by lia.
        simpl hd.
        replace (i-0+1) with (S i) by lia.
        rewrite last_cons by (simpl;congruence).
        rewrite last_app.
        2: {
          intro hh;apply (f_equal (length (A:=nat))) in hh;
          rewrite dc_length in hh;
          simpl in hh;lia.
        }
        rewrite dc_last by lia.
        lia.
      }
      {
        simpl hd.
        simpl length.
        rewrite app_length.
        repeat rewrite dc_length.
        rewrite dc_removelast by lia.
        hnf;simpl.
        simpl.
        simpl stutter.
        simpl simple_token.
        rewrite dc_length by lia.
        lia.
      }
    Qed.

    Lemma se_C0_tokens : forall lc n,
        lc < n -> n < N ->
        simple_token (se_C0 lc) n.
    Proof.
      intros lc j h1 h2.
      destruct j as [|j];try lia.
      split.
      rewrite se_C0_length;lia.
      unfold se_C0.
      rewrite app_nth2;try rewrite rep_length;[|lia].
      rewrite app_nth2;try rewrite rep_length;[|lia].
      rewrite dc_nth;[|lia].
      rewrite dc_nth;[|lia].
      nia.
    Qed.

    Lemma se_NC0_has_tokens_all_steps :
      simple_sched_has_tokens (se_NC0 1) (repl (dc 0 N) (N - 2)).
    Proof.
      cut (forall i j, i + j = N - 2 ->
                       simple_sched_has_tokens
                         (se_NC0 (S i)) (repl (dc 0 N) j)).
      { intro h;apply h;lia. }
      intros i j;revert i;induction j as [|j hrecj];intros i hj;simpl;auto.
      apply simple_sched_app_tokens.
      2: { rewrite se_NC0_big_step by lia;apply hrecj;fold plus;lia. }
      rewrite se_NC0_NC by lia.
      clear hrecj.
      assert (h:i<=N-3) by lia;clear j hj.
      cut (forall j, 0<j -> j<=N ->
                     simple_sched_has_tokens (se_NC (S i) (j-1)) (dc 0 j)).
      intro hh;apply hh;lia.
      induction j as [|j hrecj];try lia.
      destruct j as [|j];intros _ hj.
      {
        simpl;split;auto.
        clear hrecj hj.
        (* lemma starts here *)
        apply se_NC_root_token;lia.
      }
      split;simpl plus.
      2: { rewrite se_NC_middle_step by lia;apply hrecj;lia. }
      apply stutter_token_le;try lia.
      apply se_NC0_tokens;lia.
    Qed.

    Lemma se_C0_has_tokens_all_steps :
      simple_sched_has_tokens (se_NC0 (N-1)) (flatmap (fun p => dc p (N-p))
                                                      (uc 1 (N-3))).
    Proof.
      replace (N-3) with (S (N-4)) by lia.
      simpl flatmap.
      apply simple_sched_app_tokens.
      {
        replace (N-1) with (S (N-2)) at 2 by lia.
        split.
        {
          apply se_NC0_tokens; lia.
        }
        {
          simpl plus;replace (S (N-2)) with (N-1) by lia.
          rewrite se_NC_C_step.
          cut (forall j, j<N -> simple_sched_has_tokens (se_C 0 j) (dc 1 j)).
          { intro h;apply h;lia. }
          intro j;induction j  as [|j hrecj];intro hj;simpl;try constructor.
          {
            apply stutter_token_le;try lia.
            apply se_C0_tokens;try lia.
          }
          {
            rewrite se_C_middle_step by lia.
            replace (S j -1) with j by lia.
            apply hrecj;lia.
          }
        }
      }
      {
        rewrite se_NC_C_big_step by lia.
        cut (forall lc j,
                lc + j = N-3 ->
                simple_sched_has_tokens (se_C0 lc)
                                        (flatmap (fun p : nat => dc p (N - p))
                                                 (uc (lc+1) j))).
        { intro h;apply h;lia. }
        intros lc j; revert lc; induction j as [|j hrecj];intros lc h;
          try constructor.
        simpl flatmap;apply simple_sched_app_tokens.
        2:{ rewrite se_C0_big_step by lia;apply hrecj;lia. }
        assert (hh: lc <= N-4) by lia.
        clear j hrecj h.
        rewrite se_C0_C by lia.
        cut (forall j , lc<=j -> j<N ->
                        simple_sched_has_tokens (se_C lc j)
                                                (dc (lc+1) (S j - (lc+1)))).
        {
          intro h;replace N with (S (N-1)) at 2 by lia.
          apply h;lia.
        }
        intro j;induction j  as [|j hrecj];intros hj1 hj2;try lia.
        {
          destruct (Compare_dec.le_lt_eq_dec hj1);try lia.
          subst;simpl.
          constructor.
        }
        {
          destruct (Compare_dec.le_lt_eq_dec hj1);try lia.
          {
            replace (S (S j) - (lc + 1)) with (S (j - lc)) by lia.
            split.
            {
              replace (lc + 1 + (j - lc)) with (S j) by lia.
              apply stutter_token_le;try lia.
              apply se_C0_tokens;try lia.
            }
            {
              replace (lc + 1 + (j - lc)) with (S j) by lia.
              rewrite se_C_middle_step by lia.
              replace (S j - 1) with j by lia.
              replace (S j - (lc + 1)) with (j - lc) in hrecj by lia.
              apply hrecj;lia.
            }
          }
          {
            subst.
            replace (S (S j) - (S j + 1)) with 0 by lia.
            simpl.
            constructor.
          }
        }
      }
    Qed.

    Lemma last_se_nonleg: simple_non_leg (se_C0 (N-3)).
    Proof.
      exists (2+(N-3));exists (1+(N-3));split;[|split;[lia|]];
        unfold simple_token;simpl;
        rewrite se_C0_length;try lia;split;try lia;
          unfold se_C0;rewrite app_nth2;rewrite rep_length;try lia;
            rewrite app_nth2;rewrite rep_length;try lia;
              replace (N - (N - 3)) with 3 by lia;
              try replace ((N-3) - (N-3)) with 0 by lia;
              try replace (S (N-3) - (N-3)) with 1 by lia;
              try replace (S (S (N-3)) - (N-3)) with 2 by lia;simpl;lia.
    Qed.

    (** my_sched is a valid scheduler (i.e. at each step the node
      which is executed holds a token and the execution ends at a
      illegitimate configuration). *)
    Lemma my_sched_valid : simple_sched_valid my_se my_sched.
    Proof.
      split.
      apply my_se_token.
      apply simple_sched_app_valid;
        rewrite my_se_step.
      apply se_NC0_has_tokens_all_steps.
      rewrite se_NC_all_steps.
      rewrite simple_sched_valid_alt;split.
      apply se_C0_has_tokens_all_steps.
      rewrite se_C_all_steps.
      apply last_se_nonleg.
    Qed.

  End Simple_Worst_Case.

  (*
  Eval compute in my_se 5.
  Eval compute in my_sched 5.
   *)

  (*
  Eval compute in se_NC0 6 3.
  Eval compute in se_NC 6 3 2.
  Eval compute in se_C0 6 3.
  Eval compute in se_C 6 3 2.
   *)
  (*
  Eval compute in
      run_simple_sched (my_se 6) (my_sched 6).

  Lemma test4: simple_sched_valid (my_se 4) (my_sched 4).
  Proof.
    rewrite simple_sched_valid_alt.
    split.
    compute.
    intuition lia.
    exists 2;exists 3;compute.
    lia.
  Qed.

  Lemma test8: simple_sched_valid (my_se 8) (my_sched 8).
  Proof.
    rewrite simple_sched_valid_alt.
    split.
    compute;intuition lia.
    exists 6;exists 7;compute;lia.
  Qed.

  Lemma test16: simple_sched_valid (my_se 16) (my_sched 16).
  Proof.
    rewrite simple_sched_valid_alt.
    split.
    compute;intuition lia.
    exists 14;exists 15;compute;lia.
  Qed.
   *)

  Lemma worst_case:
    forall n, n>=4 -> exists se, length se = n /\ simple_env_valid n se /\
                                 exists st,
                                   (length st) = (Dijkstra_exact_bound n - 1)
                                   /\ simple_sched_valid se st.
  Proof.
    intros n hn.
    exists (my_se n);split.
    apply my_se_length.
    lia.
    split.
    apply my_se_valid.
    lia.
    exists (my_sched n).
    split.
    apply my_sched_length.
    lia.
    apply my_sched_valid;auto.
  Qed.

End Simple.

Section Dijkstra_wc.

  Context {Chans: Channels}
          {Net: Network}
          {RURN: Rooted_Unidirectional_Ring_Network Net}
          {DP: Dijkstra_parameters}.

  Notation Algo := Dijkstra_algo.
  Notation Root := (RURN_root (Net := Net)).
  Notation Ring := (URN_topo (Net := Net)).

  Existing Instance Dijkstra_RO_assumption.

  Notation N := (Ring_size Root).

  (*Hypothesis k_ring_size: N = k.*)
  Hypothesis HNk: N <= k.

  (** [mkEnv se] builds an Env from a simple env
   (invalid or absent values are set to 0)  *)

  Definition ROEnv :=
    fun n => mkROState (URN_pred_channel n)
                       (if Node_dec n Root then true else false).

  Definition mkEnv (se: simple_env): Env (Algo := Dijkstra_algo) :=
    fun n => mkState (ltk_inj k (List.nth (Dist Root n) se 0)) (ROEnv n).

  Instance mkEnv_compat:Proper fequiv mkEnv.
  Proof.
    intros l1 l2  el n1 n2 en;simpl;repeat split.
    {
      symmetry in en,el.
      apply ltk_inj_compat.
      apply Partial_Setoid.
      apply nth_compat;hnf;auto.
      apply Dist_compat;auto.
    }
    {
      apply URN_pred_channel_compat,en.
    }
    {
      destruct (Node_dec n1 Root) as [EQ1|NE1];
        destruct (Node_dec n2 Root) as [EQ2 |NE2];auto.
      {
        contradict NE2.
        rewrite <- en;apply EQ1.
      }
      {
        contradict NE1.
        rewrite en;apply EQ2.
      }
    }
  Qed.

  (** Link between the value of an Env and the list access *)
  Lemma mkEnv_value: forall se n,
      simple_env_valid N se -> v (mkEnv se n) = (List.nth (Dist Root n) se 0).
  Proof.
    intros se n [val1 val2].
    unfold mkEnv;unfold v;simpl.
    rewrite proj1_sig_ltk_inj.
    auto.
    rewrite Forall_forall in val2.
    assert (In (nth (Dist Root n) se 0) se).
    { apply nth_In.
      rewrite val1.
      apply Dist_lt_Ring_size. }
    apply val2 in H.
    lia.
  Qed.

  Lemma mkEnv_ROPart: forall se n, Dijkstra_ROpart (mkEnv se n) = ROEnv n.
    reflexivity.
  Qed.

  (** simple token actually specifies Token for simple env *)
  Lemma simple_token_Token: forall se n,
      simple_env_valid N se ->
      simple_token se (Dist Root n) -> Token (mkEnv se) n.
  Proof.
    unfold Token.
    intros se n [h1 h2] [h3 h4];repeat rewrite mkEnv_value;try (split;auto).
    destruct se as [|iroot q];try inversion h1.
    destruct (Node_dec Root n) as [h|h];[left|right];split;auto.
    {
      replace (Dist Root n) with 0 in * by
          (rewrite <- Dist_refl in h;auto).
      destruct h4 as [h4 h4'].
      simpl hd in h4.
      simpl nth at 1.
      rewrite h4 at 1.
      replace (Dist Root (Pred n)) with (length q).
      2:{
        simpl.
        rewrite H0.
        rewrite Dist_Pred,Pred_Succ,<- h;lia.
      }
      rewrite ListUtil.last_nth.
      replace (length (iroot :: q) - 1) with (length q) by (simpl;lia).
      reflexivity.
    }
    {
      rewrite <- Dist_refl in h.
      destruct (Dist_Pred_r Root n);try tauto.
      rewrite H;simpl nth at 1.
      destruct  (Dist Root n) as [|nn];try lia.
      eapply (f_equal pred) in H;simpl in H.
      rewrite <- H.
      apply h4.
    }
  Qed.

  Lemma Assume_RO_mkEnv: forall se, Assume_RO (ROEnv_part (mkEnv se)).
  Proof.
    intro se;unfold mkEnv;simpl;split;intro n.
    {
      destruct (Node_dec n Root)as [EQ |NE];simpl;intuition.
    }
    {
      apply URN_channels;reflexivity.
    }
  Qed.

  (** simple step actually builds an env that is a Step-successor *)

  Lemma simple_step_Step:
    forall se n, simple_env_valid N se -> simple_token se n ->
                 Step (mkEnv (simple_step se n)) (mkEnv se).
  Proof.
    intros se n hse htok.
    assert (hN:= simple_token_lt_N htok).
    rewrite (proj1 hse) in hN.
    split;try (apply mkEnv_compat;reflexivity).
    {
      intro h.
      assert (nn_compat: Proper equiv (power Succ n Root)).
      {
        apply (power_compat Succ_compat);auto.
      }
      specialize (h _ _ nn_compat).
      apply v_compat in h.
      do 2 rewrite mkEnv_value in h;try apply simple_step_valid;auto.
      assert (hh: (Dist Root (power Succ n Root)) = n).
      {
        transitivity (n mod N).
        rewrite <- power_Succ_Dist_mod.
        apply (power_compat Succ_compat);auto.
        apply Nat.mod_small;auto.
      }
      rewrite hh in h.
      clear nn_compat hh.
      revert h.
      destruct hse as [val1 val2].
      destruct se as [|iroot q].
      simpl in val1;intros;lia.
      destruct n as [|n];simpl;try (intro;lia).
      {
        rewrite simple_step_Succ_spec2.
        generalize H_k.
        hnf in htok.
        intros h1 h2;apply htok.
        symmetry;apply h2.
        simpl in val1.
        lia.
      }
    }
    {
      intro nn.
      destruct (Nat.eq_dec (Dist Root nn) n) as [EQ|NE];[left|right].
      {
        rewrite RUN_simpl;
          [|apply mkEnv_compat;reflexivity|apply Assume_RO_mkEnv].
        assert (val:=hse).
        assert (vals:=simple_step_valid val htok).
        destruct hse as [val1 val2].
        destruct se as [|iroot q].
        simpl in val1;intros;lia.
        destruct n as [|n].
        {
          assert (hroot:root (mkEnv (iroot :: q) nn)=true).
          {
            apply (proj1 (Assume_RO_mkEnv (iroot::q))).
            apply Dist_refl in EQ;symmetry;assumption.
          }
          unfold simple_run.
          rewrite hroot.
          simpl in htok.
          rewrite mkEnv_value,EQ;[|split;auto].
          rewrite mkEnv_value;[|split;auto].
          simpl nth at 1.
          destruct (Nat.eq_dec
                      iroot (nth (Dist Root (Pred nn)) (iroot :: q) 0))
            as [EQQ|NEE].
          {
            split;[|reflexivity].
            simpl simple_step.
            simpl vltk at 2.
            rewrite EQ.
            generalize (mkEnv_value nn vals).
            simpl simple_step.
            unfold v;intro h;hnf.
            rewrite h.
            rewrite EQ;simpl.
            destruct htok as [toka [tokb tokc]].
            simpl in tokc.
            simpl in val1.
            rewrite proj1_sig_ltk_inj;try lia.
            rewrite Nat.mod_small;lia.
          }
          {
            destruct htok as [toka [tokb tokc]].
            simpl hd in tokb.
            rewrite tokb in NEE.
            contradict NEE.
            simpl in val1.
            apply Nat.succ_inj in val1.
            rewrite ListUtil.last_nth at 1.
            rewrite Dist_Pred in val1.
            rewrite Pred_Succ in val1.
            rewrite Dist_refl in EQ.
            rewrite EQ in val1 at 2.
            rewrite <- val1.
            destruct q;simpl;auto;
              rewrite Nat.sub_0_r;auto.
          }
        }
        {
          assert (hroot:root (mkEnv (iroot :: q) nn)=false).
          {
            rewrite <- not_true_iff_false.
            rewrite (proj1 (Assume_RO_mkEnv (iroot::q)) nn).
            intro h;rewrite h in EQ.
            rewrite Dist_self in EQ;lia.
          }
          unfold simple_run.
          rewrite hroot.
          rewrite (mkEnv_value);auto.
          rewrite (mkEnv_value);auto.
          rewrite EQ.
          destruct (Dist_Pred_r Root nn) as [hnn|hnn];try lia.
          rewrite EQ in hnn.
          apply Nat.succ_inj in hnn.
          rewrite <- hnn.
          destruct htok as [toka tokb].
          simpl in toka.
          simpl nth in tokb at 1.
          destruct (Nat.eq_dec (nth (S n) (iroot :: q) 0)
                               (nth n (iroot :: q) 0)).
          {
            simpl nth in e at 1.
            rewrite e in tokb.
            contradict tokb.
            reflexivity.
          }
          {
            split; [| reflexivity ].
            change ( v (mkEnv (simple_step_Succ iroot q n) nn) ==
                     v (mkEnv (iroot :: q) (Pred nn)) ).
            rewrite (mkEnv_value);[|auto].
            rewrite (mkEnv_value);[|auto].
            rewrite EQ;rewrite <- hnn.
            apply simple_step_Succ_spec2.
            simpl length in val1.
            apply Nat.succ_inj in val1.
            rewrite val1.
            apply lt_S_n.
            apply hN.
          }
        }
      }
      {
        assert (h:=simple_step_spec2 se NE).
        rewrite <- mkEnv_value in h at 1;[|apply simple_step_valid;auto].
        rewrite <- mkEnv_value in h;[|assumption].
        split;try assumption.
        reflexivity.
      }
    }
  Qed.

  Lemma simple_sched_valid_Non_leg : forall se sched,
      simple_env_valid N se -> simple_sched_valid se sched ->
      ~ Leg_state (mkEnv se).
  Proof.
    intros se sched hse hsched;revert se hse hsched.
    induction sched as [|n sched hrec];intros se hse hsched.
    {
      intro h;destruct hsched as [i [j [hi [hij hj]]]];contradict hij.
      hnf in h.
      assert (hN:=proj1 hse).
      assert (li: i < N) by (rewrite <- hN;apply simple_token_lt_N,hi).
      assert (lj: j < N) by (rewrite <- hN;apply simple_token_lt_N,hj).
      pose (ni:=power Succ i Root).
      pose (nj:=power Succ j Root).
      assert (hhi:Dist Root ni = i).
      {
        rewrite <- (Nat.mod_small _ _ li).
        rewrite <- power_Succ_Dist_mod;reflexivity.
      }
      assert (hhj:Dist Root nj = j).
      {
        rewrite <- (Nat.mod_small _ _ lj).
        rewrite <- power_Succ_Dist_mod;reflexivity.
      }
      rewrite <- hhi in hi.
      rewrite <- hhj in hj.
      apply simple_token_Token in hi;auto.
      apply simple_token_Token in hj;auto.
      specialize (h ni nj hi hj).
      rewrite <- hhi.
      rewrite <- hhj.
      apply Dist_compat;auto.
    }
    {
      destruct hsched as [htok hst].
      specialize (hrec (simple_step se n) (simple_step_valid hse htok) hst).
      assert (h:=simple_step_Step hse htok).
      intro hh;contradict hrec.
      pose (tr:= s_cons  (mkEnv se) (sync_exec (mkEnv (simple_step se n)))).
      assert (ise:is_exec tr).
      {
        right.
        - rewrite  sync_exec_hd;apply h.
        - apply sync_exec_is_exec.
          apply mkEnv_compat.
          reflexivity.
      }
      assert (htr := Dijkstra_closure HNk (e:=tr) (Assume_RO_mkEnv se) ise I hh).
      unfold tr in htr.
      inversion htr;subst.
      apply Always_P_P in H2.
      rewrite sync_exec_hd in H2.
      apply H2.
    }
  Qed.

  (** For every simple env se and scheduler st, as far as they are
    valid, we can build an execution with first configuration se and
    such that at each step one node executes as specified by st. As we
    already prove convergence towards legitimate, this execution
    converges, but in more steps than the number of steps specified by
    st (since st is satisfied meaning that the last configuration reached
    using st is not legitimate.  *)
  Lemma simple_sched_valid_min_steps : forall se st,
      simple_env_valid N se ->
      simple_sched_valid se st ->
      exists exec,
        ((s_hd exec = mkEnv se) /\ is_exec exec)  /\
        exists nn,
          length st < nn /\
          At_N_steps nn (fun e => Leg_state (s_hd e)) exec.
  Proof.
    intros se st hse hst.
    assert (hnl:= simple_sched_valid_Non_leg _ hse hst).
    revert se hnl hse hst.
    induction st as [|n st hrec].
    {
      intros se hnl hse [i [j [ti [ne tj]]]].
      pose (myexec := sync_exec (mkEnv se)).
      assert (phd: Proper pequiv (mkEnv se)) by (apply mkEnv_compat;auto).
      assert (pmyexec: Proper pequiv myexec) by apply (sync_exec_compat phd).
      assert (is_hd: s_hd myexec = mkEnv se) by apply sync_exec_hd.
      assert (assume: Assume (s_hd myexec))
        by (rewrite is_hd;apply Assume_RO_mkEnv).
      assert (ise:is_exec myexec) by (apply sync_exec_is_exec;auto).
      assert (hNk:N<=k) by lia.
      split with myexec;split;auto.
      assert (hconv:=Dijkstra_convergence hNk assume ise I).
      assert (wd:forall e : Exec,
                 (Proper pequiv e ->
                  (fun suf : Exec => Leg_state (s_hd suf)) e) \/
                 ~ (fun suf : Exec => Leg_state (s_hd suf)) e).
      {
        intro e.
        destruct (Leg_state_dec (s_hd e));[left|right];auto.
        intro pe;apply l,hd_compat,pe.
      }
      destruct (Reachable_Steps wd pmyexec hconv) as [[|n] hn].
      {
        inversion hn ;subst.
        rewrite is_hd in H;tauto.
      }
      {
        exists (S n);split;auto.
        simpl;lia.
      }
    }
    {
      intros se hnl hse [h1 h2].
      assert (hstep:Step (mkEnv (simple_step se n)) (mkEnv se)).
      {
        apply simple_step_Step;auto.
      }
      assert (hses:=simple_step_valid hse h1).
      assert (hnls:= simple_sched_valid_Non_leg st hses h2).
      destruct (hrec (simple_step se n) hnls hses h2) as
          [myexec [[h3 h4] [nn [hn1 hn2]]]].
      exists (s_cons (mkEnv se) myexec).
      split;[split;[reflexivity|right;try rewrite h3;auto]|].
      {
        exists (S nn).
        split;[simpl;lia|].
        right;auto.
      }
    }
  Qed.

  Theorem Dijkstra_min_bound:
    N >= 4 -> forall bound, bound < Dijkstra_exact_bound N ->
                            ~ Steps_complexity bound
                              (fun e => Leg_state (s_hd e))
                              (fun e => Assume (s_hd e) /\ is_exec e).
  Proof.
    intros hN bnd hbnd ncomp.
    destruct (worst_case hN) as [se [lse [vse [ssched [lsched vsched]]]]].
    assert (hh:=simple_sched_valid_min_steps _ vse vsched).
    destruct hh as [exec [hQ [nn [lst2 atn]]]].
    rewrite lsched in lst2.
    assert (pe:=is_exec_proper (proj2 hQ)).
    assert (hbnd2:bnd<nn) by lia.
    eapply Non_Leg_Steps_min_bound;try eapply ncomp; eauto.
    destruct hQ as [hQ1 hQ2].
    split;[rewrite hQ1;apply Assume_RO_mkEnv|apply hQ2].
  Qed.

End Dijkstra_wc.

Unset Implicit Arguments.
Close Scope nat_scope.
