(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- March 2017
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Bool.
From Coq Require Import ZArith.

(** **** Local Imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import OptionUtil.
From Padec Require Import ZUtils.
From Padec Require Import ListUtils.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Spanning_Tree.
From Padec Require Import FoldRight.
From Padec Require Import KClustering_specification.

(* end show *)
Open Scope Z_scope.
Open Scope signature_scope.
Set Implicit Arguments.

(** * Algorithm for KClustering: 
  computation of alpha, parent in cluster and cluster head

  usage:

  forall Channels, 
  exists an Algorithm (the one built here) (uses Channels)
  forall Network using Channels,
     KClustering_algo meets its specification defined on
                   (some) Network and Channels.
 *)

(** * K-Clustering Algorithm *)
Section KClustering_algo.
  
  (** ** Channel interface definitions 
   The algorithm requires totally ordered channels
   *)  
  Context {Chans: Channels}.

  (* Channels should be totally ordered, to be able to select the
   minimum channel in some set *)
  Class OrderedChannels :=
    {
      ltC: relation Channel;
      ltC_dec: Decider ltC;
      ltC_compat:> Proper fequiv ltC;
      ltC_trans: forall x y z, ltC x y -> ltC y z -> ltC x z;
      ltC_strict: forall x y, ltC x y -> ~ x == y;
      ltC_total: forall x y, x == y \/ ltC x y \/ ltC y x
    }.  
  Notation minC := (min _ ltC_dec).
  (**  Definition minC x y := (if ltC_dec x y then x else y).*)

  (** ** Algorithm definitions *)

  (** The algorithm also requires identities on nodes: to refer the
   cluster head of the node, refer the identity of this cluster head
   *)
  Class KC_algo_tools :=
    {
      OrdChans: OrderedChannels;
      IdNode: Type;
      Id_setoid:> Setoid IdNode;
      Id_dec: Decider (equiv (A := IdNode))
    }.

  Context {KCT: KC_algo_tools}.
  Existing Instance OrdChans.
  
  (** *** Read-only state of a node for the algorithm *)  
  Record KCROState :=
    {
      Par: option Channel;
      id: IdNode;
      
      KCpeers_reverse: list (Channel * Channel);
      KCpr_assume: lpeers_reverse_assume KCpeers_reverse
    }.

  Definition KCpeers :=
    (fun ro: KCROState => llpeers (KCpeers_reverse ro)).
  Definition KCreverse :=
    (fun ro: KCROState => llreverse (KCpeers_reverse ro)).
  
  Instance KCreverse_local_compat:
    forall ro, Proper fequiv (KCreverse ro).
  Proof.
    intros ro.
    apply llreverse_compat; auto.
  Qed.

  Definition eqROS :=
    (fun (x y: KCROState) =>
       Par x == Par y /\ id x == id y /\
       equivlistA equiv (KCpeers_reverse x) (KCpeers_reverse y)).
  
  Global Instance eqROS_equiv: Equivalence eqROS.
  Proof.
    unfold eqROS; split; hnf; intuition.
    transitivity (Par y); auto.
    transitivity (id y); auto.
    transitivity (KCpeers_reverse y); auto.      
  Qed.

  Global Instance KCROState_setoid: Setoid KCROState :=
    {| equiv := eqROS |}.


  (** *** State of a node for the algorithm *)  
  Lemma KCROState_dec: Decider (equiv (A := KCROState)).
  Proof.
    intros [p1 i1 r1] [p2 i2 r2]; simpl equiv; unfold eqROS;
      simpl equiv.
    destruct (eqoptionA_dec Channel_dec p1 p2);[|right;tauto].
    destruct (Id_dec i1 i2);[|right;tauto].
    assert (Decider (equiv (A := Channel * Channel))).
    { apply RelProd_dec; apply Channel_dec. }
    destruct (equivlistA_dec X r1 r2);[|right;tauto].
    left; auto.      
  Qed.

  Record KCState :=
    mkState {
        alpha: Z;
        parCl: option Channel; (** None means p itself *)
        hd: IdNode;
        KCROpart:> KCROState
      }.
  
  (** Equality on states (decidable equivalence) *)
  Definition eqS :=
    (fun (x y: KCState) =>
       alpha x == alpha y /\ hd x == hd y /\
       parCl x == parCl y /\ KCROpart x == KCROpart y).

  Global Instance eqS_equiv: Equivalence eqS.
  Proof.
    unfold eqS; split; hnf; intuition;
      etransitivity; eauto.
  Qed.
  
  Global Instance KCState_setoid: Setoid KCState := {| equiv := eqS |}.

  Lemma KCState_dec: Decider eqS.
  Proof.
    intros [xa xp xh xro] [ya yp yh yro]; simpl equiv;
      unfold eqS; simpl equiv.
    destruct (KCROState_dec xro yro) as [h|h]; simpl equiv in h;
      [|right;tauto].
    destruct (Z.eq_dec xa ya); [|right; tauto].
    destruct (Id_dec xh yh); [|right; tauto].
    destruct (eqoptionA_dec Channel_dec xp yp); [|right; tauto].
    left; auto.
  Qed.

  (** **** Compatibility on states *)
  Global Instance alpha_compat: Proper  fequiv alpha.
  Proof. intros x y; simpl; unfold eqS; intuition. Qed.

  Global Instance Par_compat: Proper fequiv Par.
  Proof. intros x y; simpl; unfold eqS, eqROS; intuition. Qed.

  Global Instance id_compat: Proper fequiv id.
  Proof. intros x y; simpl; unfold eqS, eqROS; intuition. Qed.

  Global Instance parCl_compat: Proper fequiv parCl.
  Proof. intros x y; simpl; unfold eqS; intuition. Qed.

  Global Instance hd_compat: Proper fequiv hd.
  Proof. intros x y; simpl; unfold eqS; intuition. Qed.

  Instance KCROpart_compat: Proper fequiv KCROpart.
  Proof. intros x y; simpl; unfold eqS; intuition. Qed.

  Instance KCpeers_reverse_compat:
    Proper (equiv ==> equivlistA equiv) KCpeers_reverse.
  Proof. intros x y; simpl; unfold eqS, eqROS; intuition. Qed.
  
  Instance KCpeers_compat: Proper (equiv ==> equivlistA equiv) KCpeers.
  Proof.
    intros x y; simpl; unfold eqS, eqROS; intuition.
    apply llpeers_compat_equivlist; auto.
  Qed.

  Lemma KCreverse_compat:
    forall sx sy: KCROState,
      sx == sy ->
      forall c, InA equiv c (KCpeers sx) -> KCreverse sx c == KCreverse sy c.
  Proof.
    intros x y; simpl; unfold eqS, eqROS; intuition.
    apply llreverse_compat_equivlist; auto.
    apply x.
  Qed.

  (** ** The Parameter K itself (from K-Clustering) *)
  Context {KCP: KClustering_parameters}.

  (** ** Functions used for the definition of run *)
  Section Run_definition.

    Variable self_state: KCState.
    Variable peer_state: Channel -> option KCState.

    (** Access to children in the spanning tree (made of Par) *)
    Definition lPar :=
      (fun c => option_rect (fun _ => option Channel)
                            (fun s: KCState => Par (KCROpart s)) None
                            (peer_state c)).

    Notation children :=
      (Local_Children (KCpeers self_state) lPar          
                      (KCreverse self_state)).
    
    (** Access to states of the children in the spanning tree (made of
    Par) -- same order as children *)
    Definition ChildrenStates: list KCState := map_filter peer_state children.

    (** Access to values of alpha in the states of the children in the
    spanning tree (made of Par) -- same order as children *)
    Definition ChildrenAlphas: list Z := List.map alpha ChildrenStates.

    Definition _IsShort (_alpha: Z): bool := _alpha <? k.
    Definition _IsTall (_alpha: Z): bool := _alpha >=? k.

    Global Instance _IsShort_compat: Proper fequiv _IsShort.
    Proof. intros a b ab; rewrite ab; reflexivity. Qed.

    Global Instance _IsTall_compat: Proper fequiv _IsTall.
    Proof. intros a b ab; rewrite ab; reflexivity. Qed.

    (** Access to short values of alpha among children *)
    Definition ShortAlphas: list Z := List.filter _IsShort ChildrenAlphas.

    (** Access to tall values of alpha among children *)
    Definition TallAlphas: list Z  := List.filter _IsTall ChildrenAlphas.

    (** Maximum value of alpha among the short children *)
    Definition MaxAShort: Z := List.fold_right Z.max (-1) ShortAlphas.
    (** Minimum value of alpha among the tall children *)
    Definition MinATall: Z  := List.fold_right Z.min (2 * k + 1) TallAlphas.

    (** New value for alpha *)
    Definition NewAlpha: Z :=
      if Z_le_gt_dec (MaxAShort + MinATall + 2) (2 * k) then
        MinATall + 1
      else
        MaxAShort + 1.

    (** Does this channel corresponds to the tall child with minimum
    alpha value? *)
    Definition isMinATalls (c: Channel): bool := 
      option_rect (fun _ => bool)
                  (fun qs => (alpha qs) =? MinATall) false
                  (peer_state c).

    (** Computes the minimum channel among the ones for which
    isMinATalls is true, i.e. among the tall children with minimum
    alpha value. *)
    Definition MinIDMinATall: option Channel :=
      match (List.filter isMinATalls children) with
      | nil => None (* p *)
      | hd::tl => Some (List.fold_right minC hd tl)
      end.

    (** New value for parCl *)
    Definition NewparCl: option Channel :=
      if (alpha self_state) <? k then (Par self_state)
      else if (alpha self_state) =? k then None (* p *)
           else MinIDMinATall.
    
    (** Is this node a cluster head (according to alpha and Par)? *)
    Definition IsClusterHead: bool := 
      ((alpha self_state) =? k) 
      || 
      (_IsShort (alpha self_state) && 
                dec2b (A := option Channel) (eqoptionA_dec Channel_dec)
                (Par self_state) None).

    (** New value for head *)
    Definition NewHead: IdNode :=
      if (IsClusterHead) then (id self_state) 
      else match (parCl self_state) with
           | None => (hd self_state)
           | Some pcl => 
             match (peer_state pcl) with
             | None => (hd self_state)
             | Some ps => (hd ps)
             end
           end.

    (** Alpha action enabled *)
    Definition alpha_enabled_b: bool := 
      match (Z.eq_dec (alpha self_state) NewAlpha) with
      | left _ => false (* = *)
      | right _ => true (* != *)
      end.
    Definition alpha_enabled: Prop := (alpha_enabled_b = true).

    (** ParCl action enabled *)
    Definition parCl_enabled_b: bool := 
      match (eqoptionA_dec Channel_dec (parCl self_state) (NewparCl)) with
      | left _ => false (* = *)
      | right _ => true (* != *)
      end.
    Definition parCl_enabled: Prop := (parCl_enabled_b = true).

    (** Head action enabled *)
    Definition hd_enabled_b: bool := 
      match (Id_dec (hd self_state) (NewHead)) with
      | left _ => false (* = *)
      | right _ => true (* != *)
      end.
    Definition hd_enabled: Prop := (hd_enabled_b = true).

    (** run_algorithm *)
    Definition run_algorithm: KCState :=
      if alpha_enabled_b then 
        mkState NewAlpha (parCl self_state) (hd self_state)
                (KCROpart self_state)
      else if parCl_enabled_b then
             mkState (alpha self_state) NewparCl (hd self_state)  
                     (KCROpart self_state)
           else if hd_enabled_b then 
                  mkState (alpha self_state) (parCl self_state)
                          NewHead (KCROpart self_state)
                else self_state.

  End Run_definition.

  (** ** Locality and Compatibility of definitions used in run_algorithm *)
  Section Run_definition_Locality.
    
    Instance lPar_compat: Proper (fequiv ==> equiv ==> equiv) lPar.
    Proof.
      intros p1 p2 ep c1 c2 ec; unfold lPar.
      apply option_rect_compat; auto.
      intros s1 s2 es; apply Par_compat, es.
    Qed.

    Lemma lPar_locality:
      forall ns1 ns2 c, ns1 c == ns2 c -> lPar ns1 c == lPar ns2 c.
    Proof.
      intros ns1 ns2 c h; unfold lPar.
      destruct (ns1 c); destruct (ns2 c); simpl; auto;
        simpl in h; try tauto.
      apply Par_compat, h.
    Qed.
    
    Global Instance ChildrenStates_compat: 
      Proper (equiv ==> fequiv ==> equivlistA equiv) ChildrenStates.
    Proof.
      intros s1 s2 es lc1 lc2 eqlc; unfold ChildrenStates.
      eapply map_filter_compat; auto.
      apply Local_Children_locality; auto.
      apply lPar_compat; auto.
      apply lPar_compat; auto.
      apply KCreverse_local_compat.
      apply KCreverse_local_compat.
      apply KCpeers_compat, es.
      intros c i; split.
      apply lPar_compat; auto.
      apply KCreverse_compat; auto.
      apply es.
    Qed.
    
    Lemma ChildrenStates_locality:
      forall (s1 s2: KCState) ns1 ns2,
        s1 == s2 ->
        Proper fequiv ns1 -> Proper fequiv ns2 ->
        (forall c, InA equiv c (KCpeers s1) -> ns1 c == ns2 c) ->
        equivlistA equiv (ChildrenStates s1 ns1) (ChildrenStates s2 ns2).
    Proof. 
      intros s1 s2 ns1 ns2 es p1 p2 h; unfold ChildrenStates.
      eapply map_filter_eq_locality; auto.
      { apply Local_Children_locality; auto.
        apply lPar_compat; auto.
        apply lPar_compat; auto.
        apply KCreverse_local_compat; auto.
        apply KCreverse_local_compat; auto.
        apply KCpeers_compat, es.
        intros c i; split.
        apply lPar_locality; auto.
        apply KCreverse_compat, i.
        apply es.
      }
      { intros c i; apply h.
        unfold Local_Children in i.
        eapply filter_InA_locality.
        2: apply i.
        intros c1 c2 ec ii.
        apply Local_IsChild_compat; auto.
        reflexivity.
        apply lPar_compat; auto.
        apply KCreverse_local_compat.
      }
    Qed.

    Global Instance ChildrenAlphas_compat: 
      Proper (equiv ==> fequiv ==> equivlistA equiv) ChildrenAlphas.
    Proof.
      intros s1 s2 es lc1 lc2 eqlc; unfold ChildrenAlphas.
      apply map_compat_equiv; [ apply alpha_compat |
                                apply ChildrenStates_compat; auto].
    Qed.

    Lemma map_eq_locality A {SA: Setoid A} B {SB: Setoid B}:
      forall (f1 f2: A -> B) (l1 l2: list A),
        equivlistA equiv l1 l2 ->
        Proper fequiv f1 -> Proper fequiv f2 ->
        (forall a, InA equiv a l1 -> f1 a == f2 a) ->
        equivlistA equiv (map f1 l1) (map f2 l2).
    Proof.
      intros until l2; intros el p1 p2 h.
      intros x; split; intros hx.
      { apply InA_map_iff in hx; auto.
        destruct hx as [a [ha haa]].
        apply InA_map_iff; auto.
        exists a; split.
        rewrite <- ha; symmetry; apply h; auto.
        rewrite <- el; auto.
      }
      { apply InA_map_iff in hx; auto.
        destruct hx as [a [ha haa]].
        apply InA_map_iff; auto.
        exists a; split.
        rewrite <- ha; apply h; auto.
        rewrite el; auto.
        rewrite el; auto.
      }
    Qed.


    Lemma ChildrenAlphas_locality:
      forall (s1 s2: KCState) ns1 ns2,
        s1 == s2 ->
        Proper fequiv ns1 -> Proper fequiv ns2 ->
        (forall c, InA equiv c (KCpeers s1) -> ns1 c == ns2 c) ->
        equivlistA equiv (ChildrenAlphas s1 ns1) (ChildrenAlphas s2 ns2).
    Proof.
      intros s1 s2 ns1 ns2 es p1 p2 h; unfold ChildrenAlphas.
      eapply map_eq_locality; auto.
      2: apply alpha_compat.
      2: apply alpha_compat.
      apply ChildrenStates_locality; auto.
    Qed.
    
    Global Instance ShortAlphas_compat:
      Proper (equiv ==> fequiv ==> equivlistA equiv) ShortAlphas.
    Proof. 
      intros s1 s2 es lc1 lc2 eqlc; unfold ShortAlphas.
      apply filter_compat_equiv; [apply _IsShort_compat|
                                  apply ChildrenAlphas_compat; auto].
    Qed.

    Lemma ShortAlphas_locality:
      forall (s1 s2: KCState) ns1 ns2,
        s1 == s2 ->
        Proper fequiv ns1 -> Proper fequiv ns2 ->
        (forall c, InA equiv c (KCpeers s1) -> ns1 c == ns2 c) ->
        equivlistA equiv (ShortAlphas s1 ns1) (ShortAlphas s2 ns2).
    Proof.
      intros s1 s2 ns1 ns2 es p1 p2 h; unfold ShortAlphas.
      eapply filter_eq_locality; auto.
      2: apply _IsShort_compat.
      2: apply _IsShort_compat.
      apply ChildrenAlphas_locality; auto.
    Qed.

    Global Instance TallAlphas_compat:
      Proper (equiv ==> fequiv ==> equivlistA equiv) TallAlphas.
    Proof. 
      intros s1 s2 es lc1 lc2 eqlc; unfold TallAlphas.
      apply filter_compat_equiv; [apply _IsTall_compat|
                                  apply ChildrenAlphas_compat; auto].
    Qed.

    Lemma TallAlphas_locality:
      forall (s1 s2: KCState) ns1 ns2,
        s1 == s2 ->
        Proper fequiv ns1 -> Proper fequiv ns2 ->
        (forall c, InA equiv c (KCpeers s1) -> ns1 c == ns2 c) ->
        equivlistA equiv (TallAlphas s1 ns1) (TallAlphas s2 ns2).
    Proof.
      intros s1 s2 ns1 ns2 es p1 p2 h; unfold TallAlphas.
      eapply filter_eq_locality; auto.
      2: apply _IsTall_compat.
      2: apply _IsTall_compat.
      apply ChildrenAlphas_locality; auto.
    Qed.
    
    Global Instance MaxAShort_compat:
      Proper (equiv ==> fequiv ==> equiv) MaxAShort.
    Proof. 
      intros s1 s2 es lc1 lc2 eqlc; unfold MaxAShort.
      apply fold_right_max_compat; auto.
      apply ShortAlphas_compat; auto.
    Qed.

    Lemma MaxAShort_locality:
      forall (s1 s2: KCState) ns1 ns2,
        s1 == s2 ->
        Proper fequiv ns1 -> Proper fequiv ns2 ->
        (forall c, InA equiv c (KCpeers s1) -> ns1 c == ns2 c) ->
        MaxAShort s1 ns1 = MaxAShort s2 ns2.
    Proof.
      intros s1 s2 ns1 ns2 es p1 p2 h; unfold MaxAShort.
      apply FoldRight.fold_right_max_compat; auto.
      apply ShortAlphas_locality; auto.
    Qed.

    Global Instance MinATall_compat:
      Proper (equiv ==> fequiv ==> equiv) MinATall.
    Proof. 
      intros s1 s2 es lc1 lc2 eqlc; unfold MinATall.
      apply fold_right_min_compat; auto.
      apply TallAlphas_compat; auto.
    Qed.

    Lemma MinATall_locality:
      forall (s1 s2: KCState) ns1 ns2,
        s1 == s2 ->
        Proper fequiv ns1 -> Proper fequiv ns2 ->
        (forall c, InA equiv c (KCpeers s1) -> ns1 c == ns2 c) ->
        MinATall s1 ns1 = MinATall s2 ns2.
    Proof.
      intros s1 s2 ns1 ns2 es p1 p2 h; unfold MinATall.
      apply FoldRight.fold_right_min_compat; auto.
      apply TallAlphas_locality; auto.
    Qed.

    Global Instance NewAlpha_compat:
      Proper (equiv ==> fequiv ==> equiv) NewAlpha.
    Proof.
      intros s1 s2 es lc1 lc2 eqlc; unfold NewAlpha.
      rewrite (MaxAShort_compat es eqlc).
      rewrite (MinATall_compat es eqlc).
      reflexivity.
    Qed.

    Lemma NewAlpha_locality:
      forall (s1 s2: KCState) ns1 ns2,
        s1 == s2 ->
        Proper fequiv ns1 -> Proper fequiv ns2 ->
        (forall c, InA equiv c (KCpeers s1) -> ns1 c == ns2 c) ->
        NewAlpha s1 ns1 = NewAlpha s2 ns2.
    Proof.
      intros s1 s2 ns1 ns2 es p1 p2 h; unfold NewAlpha.
      rewrite (MinATall_locality es p1 p2 h).
      rewrite (MaxAShort_locality es p1 p2 h).
      reflexivity.
    Qed.

    Global Instance isMinATalls_compat:
      Proper (equiv ==> fequiv ==> equiv ==> equiv) isMinATalls.
    Proof.
      intros s1 s2 es lc1 lc2 eqlc c1 c2 eqc; unfold isMinATalls.
      apply option_rect_compat; auto.
      intros ss1 ss2 eqs; rewrite eqs, (MinATall_compat es eqlc);
        auto.
    Qed.

    Lemma isMinATalls_locality:
      forall (s1 s2: KCState) ns1 ns2,
        s1 == s2 ->
        Proper fequiv ns1 -> Proper fequiv ns2 ->
        (forall c, InA equiv c (KCpeers s1) -> ns1 c == ns2 c) ->
        forall c, InA equiv c (KCpeers s1) ->
                  isMinATalls s1 ns1 c = isMinATalls s2 ns2 c.
    Proof.
      intros s1 s2 ns1 ns2 es p1 p2 h c i.
      change (isMinATalls s1 ns1 c == isMinATalls s2 ns2 c).
      unfold isMinATalls.      
      apply option_rect_compat; auto.
      rewrite (MinATall_locality es p1 p2 h); intros x y xy;
        rewrite xy; auto.
    Qed.

    Global Instance MinIDMinATall_compat:
      Proper (equiv ==> fequiv ==> equiv) MinIDMinATall.
    Proof.
      intros s1 s2 es lc1 lc2 eqlc; unfold MinIDMinATall.
      assert
        (C: equivlistA
              equiv
              (filter (isMinATalls s1 lc1)
                      (Local_Children (KCpeers s1) (lPar lc1) (KCreverse s1)))
              (filter (isMinATalls s2 lc2)
                      (Local_Children (KCpeers s2) (lPar lc2) (KCreverse s2)))).
      { apply filter_compat_equiv.
        apply isMinATalls_compat; auto.
        apply Local_Children_locality; auto.
        apply lPar_compat; auto.
        apply lPar_compat; auto.
        apply KCreverse_local_compat.
        apply KCreverse_local_compat.
        apply KCpeers_compat, es.
        intros c i; split.
        apply lPar_compat; auto.
        apply KCreverse_compat; auto.
        apply es.
      }
      destruct (filter (isMinATalls s1 lc1)
                       (Local_Children (KCpeers s1)
                                       (lPar lc1) (KCreverse s1)));
        destruct (filter (isMinATalls s2 lc2)
                         (Local_Children (KCpeers s2)
                                         (lPar lc2) (KCreverse s2))); auto.
      symmetry in C; apply equivlistA_cons_nil in C;
        [tauto| apply setoid_equiv].
      apply equivlistA_cons_nil in C; [tauto| apply setoid_equiv].
      simpl.
      eapply fold_right_min_cons; try apply OrdChans; auto.
      Unshelve. apply Channel_dec.      
    Qed.

    Lemma MinIDMinATall_locality:
      forall (s1 s2: KCState) ns1 ns2,
        s1 == s2 ->
        Proper fequiv ns1 -> Proper fequiv ns2 ->
        (forall c, InA equiv c (KCpeers s1) -> ns1 c == ns2 c) ->
        MinIDMinATall s1 ns1 == MinIDMinATall s2 ns2.
    Proof.
      intros s1 s2 lc1 lc2 es p1 p2 h; unfold MinIDMinATall.
      assert
        (C: equivlistA
              equiv
              (filter (isMinATalls s1 lc1)
                      (Local_Children (KCpeers s1) (lPar lc1) (KCreverse s1)))
              (filter (isMinATalls s2 lc2)
                      (Local_Children (KCpeers s2) (lPar lc2) (KCreverse s2)))).
      { apply filter_eq_locality; auto.
        { apply Local_Children_locality; auto.
          apply lPar_compat; auto.
          apply lPar_compat; auto.
          apply KCreverse_local_compat.
          apply KCreverse_local_compat.
          apply KCpeers_compat, es.
          intros c i; split.
          apply lPar_locality; auto.
          apply KCreverse_compat; auto.
          apply es.
        }
        apply isMinATalls_compat; auto.
        apply isMinATalls_compat; auto.
        intros a i.
        apply isMinATalls_locality; auto.
        { unfold Local_Children in i.
          eapply filter_InA_locality.
          2: apply i.
          intros c1 c2 ec ii.
          apply Local_IsChild_compat; auto.
          reflexivity.
          apply lPar_compat; auto.
          apply KCreverse_local_compat.
        }
      }
      destruct (filter (isMinATalls s1 lc1)
                       (Local_Children (KCpeers s1) (lPar lc1) (KCreverse s1)));
        destruct (filter (isMinATalls s2 lc2)
                         (Local_Children (KCpeers s2) (lPar lc2)
                                         (KCreverse s2))); auto.
      symmetry in C; apply equivlistA_cons_nil in C;
        [tauto| apply setoid_equiv].
      apply equivlistA_cons_nil in C; [tauto| apply setoid_equiv].
      simpl.
      eapply fold_right_min_cons; try apply OrdChans; auto.
      Unshelve. apply Channel_dec.      
    Qed.      

    Global Instance NewparCl_compat:
      Proper (equiv ==> fequiv ==> equiv) NewparCl.
    Proof.
      intros s1 s2 eqs lc1 lc2 eqlc; unfold NewparCl.
      rewrite eqs; case (alpha s2 <? k).
      - apply Par_compat, eqs.
      - rewrite eqs; case (alpha s2 =? k);
          [reflexivity| apply MinIDMinATall_compat]; auto. 
    Qed.

    Lemma NewParCl_locality:
      forall (s1 s2: KCState) ns1 ns2,
        s1 == s2 ->
        Proper fequiv ns1 -> Proper fequiv ns2 ->
        (forall c, InA equiv c (KCpeers s1) -> ns1 c == ns2 c) ->
        NewparCl s1 ns1 == NewparCl s2 ns2.
    Proof.
      intros s1 s2 lc1 lc2 es p1 p2 h; unfold NewparCl.
      rewrite es; destruct (alpha s2 <? k).
      apply Par_compat, es.
      rewrite es; destruct (alpha s2 =? k); auto.
      apply MinIDMinATall_locality; auto.
    Qed.
    
    Global Instance IsClusterHead_compat: Proper fequiv IsClusterHead.
    Proof.
      intros s1 s2 eqs; unfold IsClusterHead. 
      rewrite (alpha_compat eqs).
      assert (Par s1 == Par s2) by (apply Par_compat, eqs).
      rewrite H. 
      reflexivity.
    Qed.    

    Global Instance NewHead_compat: Proper fequiv NewHead.
    Proof.
      intros s1 s2 eqs p1 p2 eqp; unfold NewHead.
      rewrite eqs.
      case (IsClusterHead s2).
      - apply id_compat, eqs. 
      - generalize (parCl_compat eqs);
          destruct (parCl s1); destruct (parCl s2);
            intros F; simpl in F; try inversion F; [| rewrite eqs; auto].
        generalize (eqp _ _ F); destruct (p1 c); destruct (p2 c0);
          intros FF; simpl in FF; try inversion FF; [| rewrite eqs; auto].
        apply H0.
    Qed.

    Global Instance alpha_enabled_b_compat:
      Proper fequiv alpha_enabled_b.
    Proof.
      intros s1 s2 eqs lc1 lc2 eqlc; unfold alpha_enabled_b.
      rewrite (alpha_compat eqs).
      rewrite (NewAlpha_compat eqs eqlc).
      reflexivity.
    Qed.

    Global Instance alpha_enabled_compat:
      Proper fequiv alpha_enabled.
    Proof.
      intros s1 s2 eqs lc1 lc2 eqlc; unfold alpha_enabled.
      rewrite (alpha_enabled_b_compat eqs eqlc).
      intuition.
    Qed.

    Global Instance parCl_enabled_b_compat:
      Proper fequiv parCl_enabled_b.
    Proof.
      intros s1 s2 eqs lc1 lc2 eqlc; unfold parCl_enabled_b.
      generalize (parCl_compat eqs);
        destruct (eqoptionA_dec Channel_dec (parCl s1));
        case (eqoptionA_dec Channel_dec (parCl s2)); intros; try congruence.
      contradict n; rewrite <- H, e; apply NewparCl_compat; auto.
      contradict n; rewrite H, e; symmetry; apply NewparCl_compat; auto.
    Qed.

    Global Instance parCl_enabled_compat: 
      Proper fequiv parCl_enabled.
    Proof.
      intros s1 s2 eqs lc1 lc2 eqlc; unfold parCl_enabled.
      rewrite (parCl_enabled_b_compat eqs eqlc).
      auto.
    Qed.

    Global Instance hd_enabled_b_compat: Proper fequiv hd_enabled_b.
    Proof.
      intros s1 s2 eqs lc1 lc2 eqlc; unfold hd_enabled_b.
      generalize (hd_compat eqs); destruct (Id_dec (hd s1) (NewHead s1 lc1));
        destruct (Id_dec (hd s2) (NewHead s2 lc2)); intros F; try congruence.
      contradict n; rewrite <- F, e; apply NewHead_compat; auto.
      contradict n; rewrite F, e; symmetry; apply NewHead_compat; auto.
    Qed.

    Global Instance hd_enabled_compat: Proper fequiv hd_enabled.
    Proof.
      intros s1 s2 eqs lc1 lc2 eqlc; unfold hd_enabled.
      rewrite (hd_enabled_b_compat eqs eqlc).
      auto.
    Qed.

    Global Instance run_algorithm_compat:
      Proper fequiv run_algorithm.
    Proof.
      intros s1 s2 eqs lc1 lc2 eqlc; unfold run_algorithm.
      rewrite <- (alpha_enabled_b_compat eqs eqlc).
      rewrite <- (parCl_enabled_b_compat eqs eqlc).
      rewrite <- (hd_enabled_b_compat eqs eqlc).
      case (alpha_enabled_b s1 lc1).
      { split; [simpl alpha; apply NewAlpha_compat; auto|].
        split; [simpl hd; rewrite eqs; auto|].
        split; [simpl parCl; rewrite eqs; auto|].
        simpl KCROpart; apply eqs.
      }
      case (parCl_enabled_b s1 lc1).
      { split; [simpl alpha; rewrite eqs; auto|].
        split; [simpl hd; rewrite eqs; auto|].
        split; [simpl parCl; apply NewparCl_compat; auto|].
        simpl KCROpart; apply eqs.
      }      
      case (hd_enabled_b s1); auto.
      { split; [simpl alpha; rewrite eqs; auto|].
        split; [simpl hd; apply NewHead_compat; auto|].
        split; [simpl parCl; rewrite eqs; auto|].
        simpl KCROpart; apply eqs.
      }      
    Qed.

  End Run_definition_Locality.
  
  Lemma ROpart_stable:
    forall s sn, KCROpart (run_algorithm s sn) =~= KCROpart s.
  Proof.
    intros s sn; unfold run_algorithm.
    destruct (alpha_enabled_b s sn).
    reflexivity.
    destruct (parCl_enabled_b s sn).
    reflexivity.
    destruct (hd_enabled_b s sn); reflexivity.
  Qed.


  (** * The algorihtm!! *)
  Global Instance KC_algo: Algorithm.
  Proof.
    esplit.
    apply KCState_dec.
    apply run_algorithm_compat.
  Defined.

  Global Instance KC_Stable: Stable_Variable KC_algo KCROpart.
  Proof.
    esplit.
    apply KCROpart_compat.
    apply ROpart_stable.
  Defined.

End KClustering_algo.

Section Projections.

  Context {Chans: Channels}
          {KCP: KClustering_parameters} {KCT: KC_algo_tools}
          {Net: Symmetric_Network}.

  Notation Env := (Env (Algo := KC_algo)).

  Definition Par_part := (fun (g: Env) n => Par (g n)).

  Definition alpha_part := (fun (g: Env) n => alpha (g n)).

  Definition id_part := (fun  (g: Env) n => id (g n)).

  Instance Par_part_compat: Proper fequiv Par_part.
  Proof. intros g1 g2 eg n1 n2 en; apply Par_compat, eg, en. Qed.

  Instance alpha_part_compat: Proper fequiv alpha_part.
  Proof. intros g1 g2 eg n1 n2 en; apply alpha_compat, eg, en. Qed.

  Instance id_part_compat: Proper fequiv id_part.
  Proof. intros g1 g2 eg n1 n2 en; apply id_compat, eg, en. Qed.

End Projections.

Unset Implicit Arguments.
Close Scope Z_scope.
Close Scope signature_scope.
