(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import ZArith.
From Coq Require Import Lia.
From Coq Require Import Bool.
From Coq Require Import Inclusion.
From Coq Require Import RelationPairs.

(** **** Local Imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import OptionUtil.
From Padec Require Import ZUtils.
From Padec Require Import ListUtils.
From Padec Require Import RelationA.
From Padec Require Import FoldRight.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Spanning_Tree.
From Padec Require Import KClustering_specification.
From Padec Require Import KClustering_algo.
From Padec Require Import KClustering_def_termination.
From Padec Require Import P_Q_Termination.

(* end show *)
Open Scope Z_scope.
Open Scope signature_scope.
Set Implicit Arguments.


(** * Termination for KClustering Algorithm: Alpha Steps Only

 This part shows termination of the algorithm when considering alpha
 safe steps only.  Actually, it shows safe inclusion required by the
 rest of the proof:

         [alpha_safe_incl := (forall g g', Alpha_SafeStep g' g -> ltM
 (Pot_alpha g') (Pot_alpha g)] *)

(** ** Roadmap of the proof:

1) Define Potential of a node: 0 if not enabled, else tree depth
signature: node->env -> nat

NB: depth for root == 1

2) Proof: if a node becomes enabled, one of its direct child has moved

3) Proof (down path induction + 2)): if a node becomes enabled, one of its 
   descedant has become disabled

assumption: downpaths are finite (to be proved elsewhere)

4) Proof: there is always a node whose activation has changed
(potential of this node passed from >0 to 0)
(induction on down path)

5) Using 3) and multiset theorem (P_Q_Termination.v) conclude that the 
   algorithm terminates

=> Use results from P_Q_Termination
     - proof of local criterion: see 3)
     - proof of global criterion: see 4)

 **)

Section KClustering_Alpha_Termination.

  (** ** Channel and network *)  
  Context {Chans: Channels} {Net: Symmetric_Network}
          {KCP: KClustering_parameters} {KCT: KC_algo_tools}.

  Existing Instance KC_Stable.
  
  Notation Env := (Env (Algo := KC_algo)).

  Let k_pos := k_pos.

  Variable (root: Node).

  Section Potential_Definition.
    (** ** Potential of a node: 
     0 if not enabled, else depth of the node in the spanning tree.

   Note that this definition requires that the spanning tree
   exists. This is the reason why termination cannot be shown without
   the assumption on safe configurations.  
     *)

    Variable (g: Env) (pg: Proper pequiv g)
             (span_tree: spanning_tree root (Par_part g) /\
                         Symmetric_Network_assumption Net (fun n => KCpeers (g n)) (fun n => KCreverse (g n))).

    (** Local (of a node) alpha-potential is defined using the depth
     of the node in the spanning tree Note that this definition
     requires that the spanning tree exists indeed.  This is the
     reason why termination can not be shown without the assumption of
     a safe environment.  *)


    (** **** depth in the tree 
<<
    Program Fixpoint depth (env: Env) (n: N)
            (W: Acc (is_tree_parent_of env) n) {struct W}: nat :=
      Datatypes.S
        match W with
            Acc_intro h =>
            match (tree_parent env n) as op return 
                  (eqoptionA eqN op (tree_parent env n)) -> nat
            with
              | None   => fun _ => 0%nat
              | Some p => fun e => (@depth env p (h p e))
            end (Equivalence_Reflexive (tree_parent env n))
        end. 
>>
     *)
    Definition depth_  (par:Node -> option Channel) (n: Node)
               (W: Acc (is_tree_parent_of par) n): nat :=
      @Fix_F Node (is_tree_parent_of par) (fun _ => nat)
             (fun (x: Node)
                  (f_rec: forall y: Node, is_tree_parent_of par y x -> nat) => 
                Datatypes.S
                  (match (tree_parent par x) as op return 
                         (op == tree_parent par x) -> nat
                   with
                   | None   => fun _ => 0%nat
                   | Some y => fun e => (f_rec y e)
                   end (Equivalence_Reflexive (tree_parent par x)))
             ) n W.

    Definition depth (n: Node): nat := depth_ (WF_par (Par_part_compat pg) (proj1 span_tree) n).
    
    (* Induction scheme for depth: Acc_ind_dep

   How to prove some property P on (depth n) by induction?

   Goal P (depth n)

   unfold depth.
   generalize (is_tree_parent_wf n); intros W. 
   induction W as [n W IH] using Acc_ind_dep.

     --- This creates two hypothesis
     W:  forall y, y tp n -> Acc tp y
     IH: forall y, y tp n -> P (depth y)

     --- (tp stands for is_tree_parent_of)
     *)

    (** Alpha-Potential of a node: 0 if not alpha-enabled, else tree depth *)
    Definition potential (n: Node): nat := 
      if alpha_enabled_bb g n then depth n
      else 0%nat.

  End Potential_Definition.

  Section Potential_Compatibility.

    Lemma depth__compat:
      (Proper (respectful_dep_param (equiv (A:=nat))
                                    ((Node -> option Channel)::Node::nil)
                                    (pequiv,(equiv,tt))
                                    (inl (fun par n =>  Acc (is_tree_parent_of par) n)::nil)) depth_).
    Proof.
      intros g1 g2 eg n1 n2 en A1 A2.
      revert n2 A2 en; induction A1 as [n1 A1 IH1] using Acc_ind_dep.
      simpl.
      destruct A2 as (A2).
      intros Hn.
      simpl.
      apply f_equal.
      generalize (tree_parent_compat eg _ _ Hn).
      revert  A1 A2 IH1 Hn.
      unfold is_tree_parent_of.
      destruct (tree_parent g1 n1) as [ pp1 |];
        destruct (tree_parent g2 n2) as [ pp2 |];
        intros A1 A2 IH1 Hn tp;
        try elim tp;simpl in *; auto.
    Qed.

    Lemma depth_compat:
      Proper (respectful_dep_param
                (pequiv (A:=Node -> nat )) 
                (Env::nil)
                (pequiv,tt)
                (inl (fun g => Proper fequiv g)
                     ::inl (fun g => spanning_tree root (Par_part g)
                                     /\
                                     Symmetric_Network_assumption
                                       Net (fun n => KCpeers (g n))
                                       (fun n => KCreverse (g n))
                           )
                     ::nil)) depth.
    Proof.
      repeat (hnf;intros).
      apply depth__compat; simpl;auto;apply Par_part_compat;auto.
    Qed.  

    Lemma potential_compat:
      Proper (respectful_dep_param
                (pequiv (A:=Node -> nat )) 
                (Env::nil)
                (pequiv,tt)
                (inl (fun g => Proper fequiv g)
                     ::inl (fun g => spanning_tree root (Par_part g)
                                     /\
                                     Symmetric_Network_assumption
                                       Net (fun n => KCpeers (g n))
                                       (fun n => KCreverse (g n))
                           )
                     ::nil)) potential.
    Proof.
      repeat (hnf;intros).
      unfold potential.
      replace eq with (equiv (A:=nat)) by reflexivity.
      eapply ITE_compat1.
      apply alpha_enabled_bb_compat;auto.
      apply depth__compat;simpl;auto;apply Par_part_compat;auto.
      auto.
    Qed.

  End Potential_Compatibility.

  Section Depth_Properties.

    Variable (g: Env) (pg: Proper pequiv g)
             (span_tree: spanning_tree root (Par_part g) /\
                         Symmetric_Network_assumption
                           Net (fun n => KCpeers (g n))
                           (fun n => KCreverse (g n))).
    
    Lemma depth_gt_0: forall (n: Node), gt (depth pg span_tree n) 0. 
    Proof.
      intros n. 
      unfold depth, depth_.
      rewrite <- Fix_F_eq.
      auto with arith.
    Qed.

    (** Any node which is down-linked to another has smaller depth
    than the first one. *)
    Lemma deeper_depth_step:
      forall (n: Node) (child: Node),
        is_tree_parent_of (Par_part g) n child -> 
        depth pg span_tree child = Datatypes.S (depth pg span_tree n).            
    Proof.
      intros n child stp.
      unfold depth.
      remember (WF_par (Par_part_compat pg) (proj1 span_tree) n)as An.
      remember (WF_par (Par_part_compat pg) (proj1 span_tree) child) as Ac.
      clear HeqAn HeqAc.
      
      destruct Ac as (Ac).
      simpl; apply f_equal.
      revert stp Ac. 
      unfold is_tree_parent_of.      
      case (tree_parent (Par_part g) child). 
      2: intros []. 
      intros.
      apply depth__compat;simpl;auto.
      apply Par_part_compat,pg.
      symmetry; apply stp.
    Qed.

    Lemma deeper_depth:
      forall (p: list Node) (n: Node) (d: Node),
        ~ n == d -> directed_tree_path (Par_part g) n p d ->
        (depth pg span_tree n < depth pg span_tree d)%nat.
    Proof.
      intros p.
      induction p as [| child p IH]; intros n d Hneq Hpath; simpl in Hpath.
      elim (Hneq Hpath).

      destruct Hpath as (stp, Hpath).
      assert (step := deeper_depth_step n child stp).
      case (Node_dec child d); intro.
      erewrite depth_compat in step; [| simpl;apply pg|apply e].
      rewrite step; auto.

      specialize (IH child d n0 Hpath).
      rewrite step in IH; lia.
    Qed.

  End Depth_Properties.

  (** ** Results for one alpha step of execution *)  
  Section With_Step.

    (** The alpha step *)
    Variable (g g': Env)
             (pg: Proper pequiv g) (pg': Proper pequiv g')
             (step: Step g' g)
             (alpha_step: alpha_step_b g' g = true)
             (span_tree: spanning_tree root (Par_part g)
                         /\
                         Symmetric_Network_assumption
                           Net (fun n => KCpeers (g n))
                           (fun n => KCreverse (g n)))
             (span_tree': spanning_tree root (Par_part g')
                          /\
                          Symmetric_Network_assumption
                            Net (fun n => KCpeers (g' n))
                            (fun n => KCreverse (g' n))).

    Lemma Local_Children_Unchanged:
      forall n,
        equivlistA equiv (Local_Children (KCpeers (g n)) (lPar (local_env g n))
                                         (KCreverse (g n)))
                   (Local_Children (KCpeers (g' n)) (lPar (local_env g' n))
                                   (KCreverse (g' n))).
    Proof.
      intros n; apply Local_Children_locality.
      apply lPar_compat, (local_env_compat pg _ _ (Equivalence_Reflexive n)).
      apply lPar_compat, (local_env_compat pg' _ _ (Equivalence_Reflexive n)).
      apply KCreverse_local_compat.
      apply KCreverse_local_compat.
      symmetry; apply KCpeers_Stable; auto.
      intros c i.
      split.
      {  unfold lPar, local_env.
         destruct (peer n c); simpl.
         symmetry; apply Par_compat, (Stable_Variable_step KC_Stable);
           auto.
         auto.
      }
      symmetry; apply KCreverse_Stable; auto.
    Qed.
    
    
    (** *** Any new variables only depends on children alpha-moves *)

    (** New Alphas of Children only depends on children alpha-moves *)
    Lemma ChildrenAlphas_only_depends_on_children:
      forall (n: Node), 
        (forall (child: Node), 
            is_tree_parent_of (Par_part g) n child 
            -> 
            alpha (g child) = alpha (g' child))
        -> equivlistA
             equiv (ChildrenAlphas (g n) (local_env g n))
             (ChildrenAlphas (g' n) (local_env g' n)).
    Proof.
      intros n Hchild; unfold ChildrenAlphas, ChildrenStates. 
      pose (LC := Local_Children (KCpeers (g n)) (lPar (local_env g n))
                                 (KCreverse (g n))).
      pose (LC' := Local_Children (KCpeers (g' n)) (lPar (local_env g' n))
                                  (KCreverse (g' n))).
      assert (eLC: equivlistA equiv LC LC')
        by (unfold LC, LC'; apply Local_Children_Unchanged).
      fold LC LC'.

      assert (tmp: equivlistA
                     equiv (map (alpha) (map_filter (local_env g' n) LC))
                     (map (alpha) (map_filter (local_env g' n) LC'))).
      { eapply map_compat_equiv.
        apply alpha_compat.
        eapply map_filter_eq_locality; auto.
        apply (local_env_compat pg' _ _ (Equivalence_Reflexive n)).
        apply (local_env_compat pg' _ _ (Equivalence_Reflexive n)).
      }
      rewrite <- tmp; clear tmp LC' eLC.

      assert (OK: forall c, InA equiv c LC ->
                            exists m, is_channel n c m /\
                                      is_tree_parent_of (Par_part g) n m).
      { intros. apply Children_inv; auto.
        apply Par_part_compat, step.
        rewrite Children_equiv; [|apply Par_part_compat, step].
        eapply InA_compat; [apply setoid_equiv|reflexivity|..].
        2: apply H.
        unfold LC.
        apply Local_Children_locality; auto.
        { intros cc ccc ecc.
          unfold Par_part.
          generalize (peer_compat _ _ (Equivalence_Reflexive n) _ _ ecc).
          destruct (peer n cc); destruct (peer n ccc);
            simpl; auto; try tauto.
          intros h; apply Par_compat, pg, h. 
        }
        { apply lPar_compat.
          apply (local_env_compat pg _ _ (Equivalence_Reflexive n)).
        }
        { apply reverse_compat; auto. }
        { apply KCreverse_local_compat; auto. }
        { destruct (proj2 span_tree) as [symm _]; symmetry; apply symm. }
        intros cc i; split.
        { unfold lPar, local_env, Par_part.
          rewrite peers_spec_peer in i; destruct i as [Rc hRc].
          unfold is_channel in hRc.
          destruct (peer n cc); simpl in hRc; [|tauto].
          simpl; reflexivity.
        }          
        { destruct (proj2 span_tree) as [_ symm]; symmetry; apply symm; auto. }
      }
      apply eqlistA_equivlistA.
      apply setoid_equiv.
      induction (Local_Children (KCpeers (g n)) (lPar (local_env g n))
                                (KCreverse (g n))) as [| c l IH]; intros.

      simpl; reflexivity.
      simpl.
      unfold local_env.
      case_eq (peer n c); simpl.
      2: intros _; apply IH; intros cc Hcc;
        apply OK; apply InA_cons; right; assumption. 
      intros nc Hnc.
      apply eqlistA_cons.
      2: apply IH; intros cc Hcc; apply OK; apply InA_cons; right; assumption. 
      apply Hchild.
      specialize (OK c).
      assert (x: InA equiv c (c::l)).
      apply InA_cons; left; reflexivity.
      specialize (OK x); clear x.
      destruct OK as [m [isc stp]].
      unfold is_channel in isc.
      rewrite Hnc in isc; simpl in isc.
      rewrite (is_tree_parent_of_compat
                 (Par_part_compat pg)  _ _ (Equivalence_Reflexive n) _ _ isc);
        assumption.
    Qed.

    Lemma MinATall_only_depends_on_children:
      forall (n: Node), 
        (forall (child: Node), 
            is_tree_parent_of (Par_part g) n child 
            -> alpha (g child) = alpha (g' child)) ->
        MinATall (g n) (local_env g n) = MinATall (g' n) (local_env g' n).
    Proof.
      intros n Hchild.
      unfold MinATall; apply fold_right_min_compat; auto.
      unfold TallAlphas; apply filter_compat_equiv; auto.
      apply _IsTall_compat.
      apply ChildrenAlphas_only_depends_on_children; auto.
    Qed.

    (** New Alphas only depends on children alpha-moves *)
    Lemma NewAlpha_only_depends_on_children:
      forall (n: Node), 
        (forall (child: Node), 
            is_tree_parent_of (Par_part g) n child 
            -> alpha (g child) = alpha (g' child))
        -> NewAlpha (g n) (local_env g n) =
           NewAlpha (g' n) (local_env g' n).
    Proof.
      intros n Hchild.
      unfold NewAlpha.
      assert (max: MaxAShort (g n) (local_env g n) =
                   MaxAShort (g' n) (local_env g' n)).
      { unfold MaxAShort; apply fold_right_max_compat; auto.
        unfold ShortAlphas; apply filter_compat_equiv; auto.
        apply _IsShort_compat.
        apply ChildrenAlphas_only_depends_on_children; auto.
      }
      rewrite MinATall_only_depends_on_children, max; auto.
    Qed.

    (** The states after firing an alpha-RUN only depends on children moves *)
    Lemma alpha_RUN_only_depends_on_children: 
      forall (n: Node), 
        eqS (g n) (g' n) ->
        alpha_enabled_bb g n = true ->
        (forall (child: Node), 
            is_tree_parent_of (Par_part g) n child 
            -> alpha (g child) = alpha (g' child))
        -> 
        (RUN g n) == (RUN g' n).
    Proof.
      intros n Heqenv Halpha Hchild.
      unfold RUN, Algorithm.run; simpl; unfold run_algorithm.

      assert (Halpha': alpha_enabled_bb g' n = true).
      unfold alpha_enabled_bb, alpha_enabled_b.
      rewrite <- (NewAlpha_only_depends_on_children n Hchild).
      assert (eqS_equiv := eqS_equiv).
      rewrite <- (alpha_compat Heqenv).
      unfold alpha_enabled_bb, alpha_enabled_b in Halpha.
      apply Halpha.

      case_eq (alpha_enabled_b (g n) (local_env g n)); 
        [intros _ | intros hf; unfold alpha_enabled_bb in Halpha;
                    rewrite hf in Halpha; contradict Halpha;
                    apply diff_false_true].
      case_eq (alpha_enabled_b (g' n) (local_env g' n)); 
        [intros _ | intros hf; unfold alpha_enabled_bb in Halpha';
                    rewrite hf in Halpha'; contradict Halpha';
                    apply diff_false_true].
      rewrite <- (NewAlpha_only_depends_on_children n Hchild).
      unfold KClustering_algo.eqS; simpl.
      split; [| split; [| split]]; auto.
      apply (hd_compat Heqenv).
      apply (parCl_compat Heqenv).
      apply Heqenv.
    Qed.

    Lemma alpha_step_effect:
      forall n, alpha (g' n) = NewAlpha (g n) (local_env g n)
                \/ alpha (g' n) = alpha (g n).
    Proof.
      intros n.
      destruct step as [_ _ dif r].
      specialize (r n); destruct r as [r|r].
      unfold RUN, run in r; simpl in r; unfold run_algorithm in r.
      destruct (alpha_enabled_b (g n) (local_env g n)).
      rewrite (alpha_compat r); simpl; left; reflexivity.
      destruct (parCl_enabled_b (g n) (local_env g n)).
      rewrite (alpha_compat r); simpl; right; reflexivity.
      destruct (hd_enabled_b (g n) (local_env g n)).
      rewrite (alpha_compat r); simpl; right; reflexivity.
      right; apply alpha_compat, r.
      right; apply alpha_compat, r.
    Qed.

    (** When a node that was disabled becomes enabled, one of its
    child has actually moved. *)
    Lemma enabled_child_has_moved:
      forall (n: Node) (H: alpha_enabled_bb g n = false)
             (H': alpha_enabled_bb g' n = true),
      exists (child: Node),
        (is_tree_parent_of (Par_part g) n child) /\
        (alpha (g child)) <> (alpha (g' child)).
    Proof.
      intros n Hdisabled Henabled.

      assert (Hmoved:
                forall child, is_tree_parent_of (Par_part g) n child ->
                              alpha (g child) <> alpha (g' child) ->
                              InA equiv child (map_filter (peer n) (peers n))).
      intros; apply (parent_is_peer_prop)
                with (root := root) (Par := (Par_part g));
      auto; apply span_tree.

      assert (PLUS: 
                exists child,
                  InA equiv child (map_filter (peer n) (peers n))
                  /\ (is_tree_parent_of (Par_part g) n child) 
                  /\ (alpha (g child) <> alpha (g' child))).
      2: destruct PLUS as [c ok]; exists c; apply ok.
      
      revert Hmoved;
        induction (map_filter (peer n) (peers n)) as [| p lpeers Hlpeers]; 
        intro Hmoved.

      
      assert (h: (NewAlpha (g n) (local_env g n)) <>
                 (NewAlpha (g' n) (local_env g' n))).
      { unfold alpha_enabled_bb, alpha_enabled_b in Henabled, Hdisabled.
        revert Hdisabled; case (Z.eq_dec (alpha (g n))
                                         (NewAlpha (g n) (local_env g n))).
        2: intros _ F; contradict F; apply diff_true_false. 
        intros D _.
        revert Henabled; case (Z.eq_dec (alpha (g' n))
                                        (NewAlpha (g' n) (local_env g' n))).
        intros _ F; contradict F; apply diff_false_true.
        intros E _.
        rewrite <- D.
        destruct (alpha_step_effect n) as [S|S].
        rewrite <- D in S.
        rewrite <- S; assumption.
        rewrite <- S; assumption.
      }
      contradict h.
      apply NewAlpha_only_depends_on_children .
      intros child istp.
      specialize (Hmoved child istp).
      case_eq (alpha (g child) =? alpha (g' child)); intros A.
      rewrite Z.eqb_eq in A.
      apply A.
      rewrite Z.eqb_neq in A.
      specialize (Hmoved A).
      rewrite InA_nil in Hmoved.
      elim Hmoved.

      assert (Hib: 
                (~ (Some n) == (tree_parent (Par_part g) p)
                 \/ alpha (g p) = alpha (g' p)) ->
                forall child : Node,
                  is_tree_parent_of (Par_part g) n child ->
                  alpha (g child) <> alpha (g' child) ->
                  InA equiv child lpeers).
      intros.
      specialize (Hmoved child H0 H1).
      rewrite InA_cons in Hmoved.
      destruct Hmoved; [|assumption]; unfold is_tree_parent_of in H0.
      
      destruct H as [H|H].
      contradict H.
      rewrite H0.
      apply tree_parent_compat. 
      apply Par_part_compat. 
      apply step.
      assumption.
      contradict H1.
      rewrite (alpha_compat (pg _ _ H2)), 
      (alpha_compat (pg' _ _ H2)); assumption.
      
      case_eq (alpha (g p) =? alpha (g' p)); intro m.
      rewrite Z.eqb_eq in m.
      assert (tmp: ~ (Some n) == (tree_parent (Par_part g) p) \/
                   alpha (g p) = alpha (g' p)).
      right; assumption.
      specialize (Hib tmp); clear tmp.
      specialize (Hlpeers Hib).
      destruct Hlpeers as (child, Hlpeers).
      destruct Hlpeers as (Hlpeers, Hlpeers').
      exists child.
      split.      
      rewrite InA_cons; right; assumption.
      assumption.
      
      destruct (eqoptionA_dec Node_dec (Some n) (tree_parent (Par_part g) p)).
      rewrite Z.eqb_neq in m.
      exists p.
      split.
      rewrite InA_cons; left; reflexivity.
      split; assumption; assumption.
      assert (tmp: ~ (Some n) == (tree_parent (Par_part g) p) \/
                   alpha (g p) = alpha (g' p)).
      left; assumption.
      specialize (Hib tmp); clear tmp.
      specialize (Hlpeers Hib).
      destruct Hlpeers as (child, Hlpeers).
      destruct Hlpeers as (Hlpeers, Hlpeers').
      exists child.
      split.      
      rewrite InA_cons; right; assumption.
      apply Hlpeers'.
    Qed.

    (** When a node has actually moved while none of its children
     moved, it is no more enabled at next configuration. *)
    Lemma node_has_moved_but_no_child:
      forall n, alpha (g n) <> alpha (g' n) 
                -> (forall child, is_tree_parent_of (Par_part g) n child
                                  -> alpha (g child) = alpha (g' child)) 
                -> alpha_enabled_bb g' n = false.
    Proof.
      intros n Hmov Hchild.
      
      unfold alpha_enabled_bb, alpha_enabled_b.
      (*        rewrite not_true_iff_false.*)
      assert (T: alpha (g' n) = NewAlpha (g' n) (local_env g' n)).
      2: case (Z.eq_dec (alpha (g' n))
                        (NewAlpha (g' n) (local_env g' n)));
        intros T'; [reflexivity| contradict T'; assumption].
      destruct (alpha_step_effect n) as [S|S]; [| contradict Hmov; 
                                                  symmetry; assumption].
      rewrite S.
      apply (NewAlpha_only_depends_on_children); 
        assumption.
    Qed.      
    
    (** When a node has moved, but is still enabled at next
     configuration, it has a child that actually also moved. *)
    Lemma node_has_moved:
      forall (n: Node),
        alpha (g n) <> alpha (g' n) -> alpha_enabled_bb g' n = true ->
        exists (child: Node),
          (is_tree_parent_of (Par_part g) n child) /\
          (alpha (g child) <> alpha (g' child)).
    Proof.      
      intros n Hm Henabled.

      assert (Hmoved:
                forall child, is_tree_parent_of (Par_part g) n child ->
                              alpha (g child) <> alpha (g' child) ->
                              InA equiv child (map_filter (peer n) (peers n))).
      intros; apply (parent_is_peer_prop)
                with (root := root) (Par := (Par_part g));
      auto; apply span_tree.

      assert (PLUS: 
                exists child,
                  InA equiv child (map_filter (peer n) (peers n))
                  /\ (is_tree_parent_of (Par_part g) n child) 
                  /\ (alpha (g child) <> alpha (g' child))).
      2: destruct PLUS as [c ok]; exists c; apply ok.
      
      revert Hmoved;
        induction (map_filter (peer n) (peers n)) as [| p lpeers Hlpeers]; 
        intro Hmoved.
      (***)
      contradict Henabled.
      rewrite not_true_iff_false.
      apply node_has_moved_but_no_child. 
      apply Hm.
      intros child Hc.
      specialize (Hmoved child Hc).
      case (Z.eq_dec (alpha (g child)) (alpha (g' child))); intros C.
      assumption.
      specialize (Hmoved C).
      apply InA_nil in Hmoved.
      elim Hmoved.

      assert (Hib: 
                (~eqoptionA equiv (Some n) (tree_parent (Par_part g) p)
                 \/ alpha (g p) = alpha (g' p)) ->
                forall child : Node,
                  is_tree_parent_of (Par_part g) n child ->
                  alpha (g child) <> alpha (g' child) ->
                  InA equiv child lpeers).
      intros.
      specialize (Hmoved child H0 H1).
      rewrite InA_cons in Hmoved.
      destruct Hmoved; [|assumption]; unfold is_tree_parent_of in H0.

      destruct H as [H|H].
      contradict H.
      rewrite H0.
      apply tree_parent_compat. 
      apply Par_part_compat.
      apply step.
      assumption.
      contradict H1.
      rewrite (alpha_compat (pg _ _ H2)), 
      (alpha_compat (pg' _ _ H2)); assumption.

      case (Z.eq_dec (alpha (g p)) (alpha (g' p))); intro m.
      assert (tmp: ~ (Some n) == (tree_parent (Par_part g) p) \/
                   alpha (g p) = alpha (g' p)).
      right; assumption.
      specialize (Hib tmp); clear tmp.
      specialize (Hlpeers Hib).
      destruct Hlpeers as (child, Hlpeers).
      destruct Hlpeers as (Hlpeers, Hlpeers').
      exists child.
      split.      
      rewrite InA_cons; right; assumption.
      assumption.

      destruct (eqoptionA_dec Node_dec (Some n) (tree_parent (Par_part g) p)).
      exists p.
      split.
      rewrite InA_cons; left; reflexivity.
      split; assumption; assumption.
      assert (tmp: ~ (Some n) == (tree_parent (Par_part g) p) \/
                   alpha (g p) = alpha (g' p)).
      left; assumption.
      specialize (Hib tmp); clear tmp.
      specialize (Hlpeers Hib).
      destruct Hlpeers as (child, Hlpeers).
      destruct Hlpeers as (Hlpeers, Hlpeers').
      exists child.
      split.      
      rewrite InA_cons; right; assumption.
      apply Hlpeers'.
    Qed.

    (** When a node actually moved, it is down-linked (in the spanning
     tree) to a node which was enabled and became disabled. *)
    Lemma moving_node_has_disabled_desc:
      forall (n: Node),
        alpha (g n) <> alpha (g' n) -> 
        exists (descendant: Node), 
          (exists p, directed_tree_path (Par_part g) n p descendant) /\
          alpha_enabled_bb g descendant = true /\
          alpha_enabled_bb g' descendant = false.
    Proof.
      intros n Hmov.
      assert (W := WF_child (Par_part_compat pg) (proj1 span_tree) n).
      induction W as [n Hacc Hn].
      unfold flip in *.

      case_eq (alpha_enabled_bb g' n); intro Hen.

      2:{
        exists n.
        split.
        exists nil; simpl; reflexivity.

        split.
        destruct (alpha_step_effect n) as [S|S];
          [| contradict Hmov; symmetry; assumption].
        unfold alpha_enabled_bb, alpha_enabled_b.
        case (Z.eq_dec (alpha (g n))
                       (NewAlpha (g n) (local_env g n))); intros A.
        contradict Hmov; rewrite S, A; reflexivity.
        reflexivity.
        assumption.
      }
      assert (Hchild := node_has_moved _ Hmov Hen).
      destruct Hchild as (child, Hc).
      destruct Hc as (Hc, Hc').
      specialize (Hn child Hc Hc').
      destruct Hn as (desc, Hd).
      destruct Hd as (Hpath, Hd).
      exists desc.
      split.
      2: exact Hd.
      destruct Hpath as (p, Hpath).
      exists (child::p).
      simpl.
      split; auto.
    Qed.

    (** When a node that was disabled becomes enabled at next
     configuration, it is down-linked (in the spanning tree) to a node
     which was enabled and becomes disabled. *)
    Lemma new_enabled_node_has_disabled_descendant:
      forall (n: Node),
        alpha_enabled_bb g n = false -> 
        alpha_enabled_bb g' n = true -> 
        exists (descendant: Node), 
          (exists p, directed_tree_path (Par_part g) n p descendant) /\ 
          alpha_enabled_bb g descendant = true /\ 
          alpha_enabled_bb g' descendant = false.
    Proof.      
      intros n Hdisabled Henabled.
      assert (Hc:= enabled_child_has_moved _  Hdisabled Henabled).
      destruct Hc as (n', Hc).
      destruct Hc as (Hn', Hmov).
      specialize (moving_node_has_disabled_desc _ Hmov); intro.
      destruct H as (desc, Hd).
      destruct Hd as (Hpath, Hd).
      exists desc.
      split.
      2: apply Hd.
      destruct Hpath as (p, Hpath).
      exists (n'::p).
      simpl.
      split; auto.
    Qed.      

    (** *** 4) Proof: 
     there is always a node that becomes disabled after a step
    (potential of this node passed from >0 to 0) (induction on down path)
     *)

    (** Within a valid step, there exists a node that was enabled but
    becomes disabled *)
    Lemma one_disabled: 
      exists (n: Node), alpha_enabled_bb g n = true /\
                        alpha_enabled_bb g' n = false.
    Proof.      
      (*(progress: exists n, alpha n <> alpha' n).
        destruct progress as [n mov].*)
      unfold alpha_step_b in alpha_step.
      rewrite existsb_exists in alpha_step.
      destruct alpha_step as [x [_ Hx]]; clear alpha_step.
      rewrite andb_true_iff in Hx.
      destruct Hx as [Hx Hx'].
      assert (ok: (alpha (g x)) <> (alpha (g' x))).
      2: destruct (moving_node_has_disabled_desc _ ok) 
        as [d Hd]; exists d; apply Hd.

      unfold has_moved_b in Hx'.
      revert Hx'; case (State_dec (g' x) (g x)); [congruence|].
      intros S _.
      destruct step as [_ _ _ Hs].
      specialize (Hs x).
      destruct Hs as [Hs|Hs]; [| tauto].
      unfold RUN, run in Hs; simpl in Hs; unfold run_algorithm in Hs. 
      unfold alpha_enabled_bb in Hx.
      rewrite Hx in Hs.
      unfold alpha_enabled_b in Hx.
      case_eq (Z.eq_dec (alpha (g x))
                        (NewAlpha (g x) (local_env g x))); intros X XX; rewrite XX in Hx; clear XX.
      contradict Hx; apply diff_false_true.
      rewrite Hs; simpl; auto.
    Qed.

    (** *** 5) Using 3) and multiset theorem (Termination.v, Wf_Algo_Multiset) 
    conclude that the algorithm terminates
          - proof of local criterion: see 3)
          - proof of global criterion: see 4)
     *)

    (** *** Termination Criteria *)
    Section Potential_termination_criteria.

      Lemma depth_preserved_step:
        forall n, depth pg span_tree n = depth pg' span_tree' n.
      Proof.
        intros n.
        apply depth__compat;simpl;auto.
        assert (ro := Stable_Variable_step KC_Stable step).
        unfold Par_part.
        symmetry; intros n1 n2 en; apply (ro _ _ en).
      Qed.

      Lemma diff_pot:
        forall (n: Node), 
          potential pg' span_tree' n <> potential pg span_tree n
          <->
          (alpha_enabled_bb g n = false <-> alpha_enabled_bb g' n = true).
      Proof.
        intro n.
        split; unfold potential, alpha_enabled_bb, alpha_enabled_b;
          destruct (Z.eq_dec (alpha (g' n)) (NewAlpha (g' n) (local_env g' n)));
          destruct (Z.eq_dec (alpha (g n)) (NewAlpha (g n) (local_env g n)));
          intros X.
        contradict X; reflexivity.
        split; intros Y; symmetry; assumption.
        split; intros _; reflexivity.
        contradict X.
        symmetry; apply depth_preserved_step.
        destruct X. 
        assert (x: false = false).
        reflexivity.
        specialize (H x); contradict H; apply diff_false_true.
        assert (tmp := depth_gt_0 pg span_tree n); lia.
        assert (tmp := depth_gt_0 pg' span_tree' n); lia.
        destruct X.
        assert (x: true = true).
        reflexivity.
        specialize (H0 x); contradict H0; apply diff_true_false.
      Qed.

      (** Local Criterion *)
      Lemma alpha_local_criterion:
        forall (n: Node), 
          gt (potential pg' span_tree' n) (potential pg span_tree n)
          -> exists (n': Node), 
            potential pg span_tree n' <> potential pg' span_tree' n'
            /\
            gt (potential pg span_tree n') (potential pg' span_tree' n).
      Proof.
        intros n Hpot_n_.
        case (eq_nat_dec (potential pg span_tree n)
                         (potential pg' span_tree' n)); 
          intros Hpot_n.
        rewrite Hpot_n in Hpot_n_.
        contradict Hpot_n_; apply gt_irrefl. 
        clear Hpot_n_.
        destruct (le_gt_dec (potential pg span_tree n)
                            (potential pg' span_tree' n)) 
          as [ le | gt ].

        (*gt*) 
        2:{ exists n; split; [apply Hpot_n | apply gt]. }

        (*lt*)      
        assert (pot_nul: potential pg span_tree n = 0%nat).
        revert Hpot_n le.
        unfold potential.
        destruct (alpha_enabled_bb g n); destruct (alpha_enabled_bb g' n).
        intros F; contradict F; apply depth_preserved_step.
        intros _ F; contradict F; apply lt_not_le; apply depth_gt_0.
        intros _ _; reflexivity. 
        intros _ _; reflexivity. 

        assert (Hen: potential pg' span_tree' n <>
                     potential pg span_tree n); auto.
        rewrite diff_pot in Hen.

        assert (Hdisabled: alpha_enabled_bb g n = false).
        revert pot_nul.
        unfold potential, alpha_enabled_bb, alpha_enabled_b.
        destruct (Z.eq_dec (alpha (g n)) (NewAlpha (g n) (local_env g n))).
        intros _; reflexivity.
        intros F; symmetry in F; contradict F;apply Nat.lt_neq;
          apply depth_gt_0.

        generalize Hdisabled; intro Henabled; rewrite Hen in Henabled.
        assert (Hdesc := new_enabled_node_has_disabled_descendant 
                           _ Hdisabled  Henabled).
        destruct Hdesc as (d, (Hp, Hd)).
        exists d.
        generalize Hd; intro tmp.
        destruct tmp.
        assert (neq: ~ n == d).
        intro.      
        symmetry in H1.
        rewrite (alpha_enabled_bb_compat pg _ _ H1) in H. 
        contradict Hdisabled; rewrite H; apply diff_true_false.
        clear H H0.

        destruct Hd as (Hd, Hd').

        unfold potential. (*; unfold alpha_enabled_bb, alpha_enabled_b in *.*)
        rewrite Hd, Hd', Henabled.
        split.
        intros F; symmetry in F; contradict F;
          apply Nat.lt_neq; apply depth_gt_0.
        destruct Hp as [p Hp].
        rewrite <- depth_preserved_step.
        apply deeper_depth with p; assumption.
      Qed.

      (** ** Global Criterion *)
      Lemma alpha_global_criterion:
        exists n, potential pg' span_tree' n <> potential pg span_tree n.
      Proof.
        assert (one := one_disabled).  
        destruct one as (n, (one, one')).
        exists n.
        rewrite  diff_pot.
        rewrite one, one'; intuition.
      Qed.

    End Potential_termination_criteria.

  End With_Step.


  (** *** Alpha-Termination Proof *)
  Section Alpha_Termination.

    Definition alpha_potential 
               (sg: sig (Pspanning_tree root)) (n: Node): nat :=
      potential (proj1 (proj2_sig sg)) (proj2 (proj2_sig sg)) n.

    Global Instance alpha_potential_compat:
      Proper fequiv alpha_potential.
    Proof.
      intros e1 e2 eqe n1 n2 eqn.
      unfold alpha_potential.
      apply potential_compat.
      apply eqe.
      apply eqn.
    Qed.
    
    (** Alpha-Termination *)    
    Notation ltM := (@MultisetNat.MSetOrd.MultisetLT gt).
    Notation Pot_alpha := (Pot alpha_potential). 

    Theorem alpha_safe_inclusion:
      inclusion (Alpha_SafeStep (root := root)) (ltM @@ Pot_alpha).
    Proof.
      apply Step_decreases_pot.  
      apply alpha_potential_compat.

      (* local criterion *)
      intros sg sg' step n Hgt.
      unfold safeQStep, safeStep in step.
      destruct step as [step alpha_en]. 
      unfold alpha_potential in *.
      destruct sg as [g [pg span_tree]]; 
        destruct sg' as [g' [pg' span_tree']]; simpl in *.
      apply alpha_local_criterion; assumption.

      (* global criterion *)
      intros sg sg' step.
      unfold safeQStep, safeStep in step.
      destruct step as [step alpha_en]. 
      unfold alpha_potential in *.
      destruct sg as [g [pg span_tree]]; 
        destruct sg' as [g' [pg' span_tree']]; simpl in *.
      apply alpha_global_criterion; assumption.
    Qed.

  End Alpha_Termination.

End KClustering_Alpha_Termination.

Close Scope Z_scope.
Close Scope signature_scope.
Unset Implicit Arguments.
