(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** **** Global Import **)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import ZArith.
From Coq Require Import Lia.
From Coq Require Import Bool.

(** **** Local Import **)
From Padec Require Import SetoidUtil.
From Padec Require Import OptionUtil.
From Padec Require Import ZUtils.
From Padec Require Import ListUtils.
From Padec Require Import RelationA.
From Padec Require Import FoldRight.
From Padec Require Import WellfoundedUtil.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Spanning_Tree.
From Padec Require Import KClustering_algo.
From Padec Require Import KClustering_specification.
From Padec Require Import KClustering_def_termination.
From Padec Require Import KClustering_hd_termination.
From Padec Require Import KClustering_correctness_alpha.

(* end show *)
Open Scope Z_scope.
Set Implicit Arguments.

(** * Correctness for a k-clustering algorithm **)
Section KClustering_Correctness.

  (* ********************************************************)
  (** Channel interface definitions *)  
  Context {Chans: Channels} {Net: Symmetric_Network}
          {KCP: KClustering_parameters} {KCT: KC_algo_tools}.

  Existing Instance KC_Stable.
  Existing Instance OrdChans.
  
  Notation Env := (Env (Algo := KC_algo)).

  Let k_pos := k_pos.

  (** Instantitiation of the functions from the specification **)
  Definition clusterHeadID (s: State): IdNode := (hd s).
  Definition clusterParent (s: State): option Channel := (parCl s).
  Definition kCluster (g: Env) (h: Node) (n: Node): Prop := 
    id (g h) == clusterHeadID (g n).
  Notation clusterHead g h := (kCluster g h h).
  Notation is_cluster_parent := (is_cluster_parent clusterParent).

  Instance clusterHeadID_compat: Proper fequiv clusterHeadID.
  Proof.
    intros s1 s2 es; unfold clusterHeadID.
    apply hd_compat; assumption.
  Qed.

  Instance clusterParent_compat: Proper fequiv clusterParent.
  Proof.
    intros s1 s2 es; unfold clusterParent.
    eapply (parCl_compat ); eauto.
  Qed.

  Instance kCluster_compat: Proper fequiv kCluster.
  Proof.
    intros g1 g2 eg n1 n2 en nn1 nn2 enn; unfold kCluster.
    rewrite (id_compat (KCROpart_compat (eg _ _ en))). 
    rewrite (clusterHeadID_compat (eg _ _ enn)). 
    reflexivity. 
  Qed.

  Section With_Terminal.

    Variable (g: Env) (term: terminal g).

    (** The state of each node is given by the fact that the node is
    disabled. *)
    Lemma state:
      forall n, 
        alpha (g n) = NewAlpha (g n) (local_env g n) /\
        parCl (g n) == NewparCl (g n) (local_env g n) /\
        hd (g n) == NewHead (g n) (local_env g n).
    Proof.
      intros n. 
      assert (D := terminal_all_disabled term).
      specialize (D n).
      unfold enabled, enabled_b in D.
      destruct (State_dec (RUN g n) (g n)) as [S|S]; [|congruence];
        clear D.
      unfold RUN, run in S; simpl in S;
        unfold run_algorithm in S.
      unfold alpha_enabled_b, parCl_enabled_b, hd_enabled_b in S.
      revert S; case (Z.eq_dec (alpha (g n))
                               (NewAlpha (g n) (local_env g n))); intros A S. 
      2: contradict A; symmetry; rewrite <- (alpha_compat S);
        reflexivity.
      revert S; case (eqoptionA_dec Channel_dec (parCl (g n))
                                    (NewparCl (g n) (local_env g n))); intros P S.
      2: contradict P; symmetry; rewrite <- (parCl_compat S);
        reflexivity.
      revert S; case (Id_dec (hd (g n))
                             (NewHead (g n) (local_env g n))); intros I S.
      2: contradict I; symmetry; rewrite <- (hd_compat S);
        reflexivity.
      split; [|split]; auto.
    Qed.

    Variable (root: Node)
             (span: spanning_tree root (Par_part g))
             (sym: Symmetric_Network_assumption
                     Net (fun n => KCpeers (g n)) (fun n => KCreverse (g n)))
             (uniqueID: forall n1 n2, id (g n1) == id (g n2) -> n1 == n2).

    Instance is_cluster_parent_compat: Proper fequiv is_cluster_parent.
    Proof.
      intros g1 g2 eg a a' ea b b' eb; unfold is_cluster_parent.
      generalize (clusterParent_compat (eg _ _ eb)).
      destruct (clusterParent (g1 b));
        destruct (clusterParent (g2 b')); intros F; simpl in F;
          try inversion F; auto.
      apply equiv_compat.
      apply peer_compat; auto.
      apply Some_compat; auto.
    Qed.   

    (** Cluster edges are included in the set of edges of the kDom-graph. *)
    Lemma inclu:
      inclusion (is_cluster_parent g) (is_kDom_edge g).
    Proof.
      intros b a K.
      unfold is_kDom_edge, is_tree_parent_of, tree_parent; simpl.
      unfold is_cluster_parent, clusterParent in K.

      revert K; case_eq (parCl (g a)); simpl; [| tauto]. 
      intros c hc K.

      generalize (proj1 (proj2 (state a))); unfold NewparCl.
      rewrite hc.
      case_eq (alpha (g a) ?= k); intros Ab.
      assert (tmp: (alpha (g a) <? k) = false).
      rewrite Z.compare_eq_iff in Ab; rewrite Z.ltb_nlt; lia.
      rewrite tmp.
      clear tmp.
      assert (tmp: (alpha (g a) =? k) = true).
      rewrite Z.compare_eq_iff in Ab; rewrite Z.eqb_eq; lia.
      rewrite tmp.
      clear tmp.
      intros F; elim F.
      assert (tmp: (alpha (g a) <? k) = true).
      rewrite Z.compare_lt_iff in Ab; rewrite Z.ltb_lt; lia.
      rewrite tmp.
      clear tmp.
      unfold Par_part; destruct (Par (g a)); simpl.
      intros C.
      rewrite (peer_compat _ _  (Equivalence_Reflexive a) _ _ C) in K.
      revert K; destruct (peer a c0).
      simpl.
      intros H; symmetry; assumption.
      intros F; elim F.
      tauto.

      assert (tmp: (alpha (g a) <? k) = false).
      rewrite Z.compare_gt_iff in Ab; rewrite Z.ltb_nlt; lia.
      rewrite tmp.
      clear tmp.
      assert (tmp: (alpha (g a) =? k) = false).
      rewrite Z.compare_gt_iff in Ab; rewrite Z.eqb_neq; lia.
      rewrite tmp.
      clear tmp.
      intros M.
      assert (pg: Proper pequiv g).
      apply term.
      unfold option_rect.
      assert (YY := same_par_pcl a b pg sym).
      assert (tmp: pcl_chain g b a).
      { unfold pcl_chain, get_pcl.
        rewrite hc.
        revert K; case_eq (peer a c).
        2: simpl; tauto.
        intros n hn; rewrite <- hn; clear n hn.
        intros K.
        destruct (state a) as [A [P _]].
        unfold alpha_enabled_bb, alpha_enabled_b.
        revert A; destruct (Z.eq_dec (alpha (g a))
                                     (NewAlpha (g a) (local_env g a)));
        [| tauto]; intros _.
        unfold pcl_enabled_bb, parCl_enabled_b.
        revert P; destruct (eqoptionA_dec Channel_dec (parCl (g a))
                                          (NewparCl (g a) (local_env g a)));
        [| tauto]; intros _.
        assert (dum: negb false && negb false = true)
          by auto with bool; rewrite dum; clear dum.
        symmetry; auto.
      }
      specialize (YY tmp).
      destruct YY as [[FF _] | [ _ [ _ OK]]].
      contradict FF.

      assert (tmpp: (alpha (g a) <? k) = false).
      rewrite Z.compare_gt_iff in Ab; rewrite Z.ltb_nlt; lia.
      rewrite tmpp; apply diff_false_true.

      generalize OK; intros OK'.
      unfold Par_part, is_tree_parent_of, tree_parent in OK.
      revert OK; case_eq (Par (g b)); simpl; [| tauto].

      intros c' hc'; case_eq (peer b c'); simpl; [| tauto].
      intros n hn hhn.
      split; [assumption|].
      assert  (hp: (Some a) == (peer b c')).
      rewrite hn; simpl; assumption.
      clear hn hhn n.

      destruct (state a) as [A [_ _]].
      unfold NewAlpha in A.
      unfold MinIDMinATall, isMinATalls in M.
      remember (MinATall (g a) (local_env g a)) as min.
      remember (MaxAShort (g a) (local_env g a)) as max.

      assert (X: alpha (g b) = min).
      { revert M; case_eq (filter
                             (fun c : Channel =>
                                option_rect (fun _ : option KCState => bool)
                                            (fun qs : KCState => alpha qs =? min) false
                                            (local_env g a c))
                             (Local_Children (KCpeers (g a)) (lPar (local_env g a))
                                             (KCreverse (g a)))).
        intros _ F; elim F.
        intros cc ll F M.
        simpl in M; apply (fold_right_min_InA
                             (A_dec := Channel_dec) _ ltC_dec c) in M.
        rewrite <- F in M.
        rewrite filter_InA in M.
        2: intros c1 c2 ec; unfold local_env;
          generalize (peer_compat _ _ (Equivalence_Reflexive a) _ _ ec);
          destruct (peer a c1); destruct (peer a c2); simpl; try tauto;
            intros en; rewrite (alpha_compat (pg _ _ en)); reflexivity.

        destruct M as [MM M].
        unfold local_env in M;
          revert K M; destruct (peer a c); simpl; try tauto.
        intros en an.
        rewrite Z.eqb_eq in an.
        rewrite <- an.
        eapply (alpha_compat ); apply pg; symmetry; apply en.
      }
      rewrite A, X; clear X.
      assert (tmpp: alpha (g a) > k).
      rewrite Z.compare_gt_iff in Ab; lia.
      rewrite A in tmpp; clear A.
      assert (MAX := MaxAShort_range g a).
      rewrite Heqmax; rewrite Heqmax in tmpp; clear Heqmax.
      unfold MaxAShort in *.
      destruct (fold_right_Zmax_dec (-1)  
                                    (ShortAlphas (g a) (local_env g a))
               ) as [eqa | ina]; unfold Z.eq in *.
      rewrite eqa; rewrite eqa in MAX, tmpp. 
      revert tmpp; destruct (Z_le_gt_dec (-1 + min + 2) (2 * k)); intros; lia.
      unfold ShortAlphas in *.
      rewrite filter_InA in ina.
      2: intros x y exy; rewrite exy; reflexivity.
      destruct ina as [ina2 tall].
      unfold _IsShort in tall.
      rewrite <- Zlt_is_lt_bool in tall.
      unfold _IsShort in *.
      remember (fold_right Z.max (-1)
                           (filter (fun _alpha : Z => _alpha <? k)
                                   (ChildrenAlphas (g a) (local_env g a)))) as FR.
      destruct (Z_le_gt_dec (FR + min + 2) (2 * k)); intros; lia.
    Qed.

    Lemma is_cluster_parent_wf:
      well_founded (is_cluster_parent g).
    Proof.
      apply Inclusion.wf_incl with (pcl_chain g).
      2: apply pcl_chain_wf with root; try assumption.
      2: apply term.
      intros x y; unfold pcl_chain, get_pcl, is_cluster_parent, clusterParent.
      destruct (state y) as [A [P _]].
      unfold alpha_enabled_bb, alpha_enabled_b.
      revert A; destruct (Z.eq_dec (alpha (g y))
                                   (NewAlpha (g y) (local_env g y)));
      [| tauto]; intros _.
      unfold pcl_enabled_bb, parCl_enabled_b.
      revert P; destruct (eqoptionA_dec Channel_dec (parCl (g y))
                                        (NewparCl (g y) (local_env g y)));
      [| tauto]; intros _.
      assert (dum: negb false && negb false = true)
        by auto with bool; rewrite dum; clear dum.

      destruct (parCl (g y)); [|tauto].
      destruct (peer y c); symmetry; tauto.
    Qed.

    Notation cluster_path := (cluster_path clusterParent g).
    Instance cluster_path_compat: Proper fequiv cluster_path.
    Proof.
      apply chain_compat.
      apply is_cluster_parent_compat.
      apply term.
    Qed.

    Notation cluster_path_ :=
      (agreed_cluster_path clusterHeadID clusterParent g).

    Instance cluster_path__compat: Proper fequiv cluster_path_.
    Proof.
      intros h1 h2 eh p1 p2 ep n1 n2 en; unfold agreed_cluster_path.
      assert (eqS_equiv := eqS_equiv).
      assert (pg: Proper pequiv g). apply term.
      rewrite (cluster_path_compat _ _ eh ep _ _ en).
      specialize (eqlistA_equivlistA _ ep) ; intros ep'.
      split; intros H; split. 
      apply H. 
      2: apply H.
      intros q ina.
      erewrite clusterHeadID_compat.
      2: symmetry; apply (pg _ _ eh).
      apply H.
      rewrite (InA_compat _ (Equivalence_Reflexive q) ep'); assumption.
      intros q ina.
      erewrite clusterHeadID_compat.
      2: apply (pg _ _ eh).
      apply H.
      rewrite <- (InA_compat _ (Equivalence_Reflexive q) ep'); assumption.
    Qed.

    Lemma ok_length:
      forall p b e, cluster_path_ b p e -> Z_of_nat (length p ) <= k.
    Proof.
      intros p b e path.
      eapply any_is_linked; try assumption.
      apply term.
      apply span.
      unfold cluster_path in path; unfold is_kDom_path.
      apply sym.
      intros n; unfold alpha_enabled_bb, alpha_enabled_b.
      destruct (state n) as [A _]; revert A;
        destruct (Z.eq_dec (alpha (g n)) (NewAlpha (g n) (local_env g n)));
        [ auto| tauto].
      eapply chain_incl.
      2: apply path.
      apply inclu.
    Qed.

    (** Every node in a cluster path designates the same cluster head. *)
    Lemma same_hd:
      forall p b e, cluster_path b p e -> (hd (g b)) == (hd (g e)).
    Proof.
      assert (pg: Proper pequiv g).
      apply term.
      induction p; simpl.

      intros b e eqbe.
      apply (hd_compat ); apply pg; apply eqbe.

      intros b e [PR CP].
      specialize (IHp _ _ CP). 

      unfold is_cluster_parent, clusterParent in PR.
      revert PR; case_eq (parCl (g a)); simpl; try tauto; intros c eqc PR.
      destruct (state a) as [_ [pcl hd]].
      eapply transitivity. 
      2: apply IHp.
      rewrite hd.
      unfold NewHead.

      assert (NC: IsClusterHead (g a) = false).
      unfold NewparCl in pcl.
      unfold IsClusterHead, _IsShort.
      rewrite orb_false_iff.
      rewrite eqc in pcl.
      revert pcl; case_eq (alpha (g a) <? k).
      intros X Y.
      split.
      rewrite Z.eqb_neq; rewrite Z.ltb_lt in X; lia.
      simpl; rewrite dec2b_spec_false.
      symmetry in Y; rewrite Y; tauto.
      intros X; case_eq (alpha (g a) =? k); simpl; try tauto.

      rewrite NC, eqc.
      unfold local_env.
      revert PR; destruct (peer a c); simpl; try tauto.
      intros en; apply (hd_compat ); apply pg;
        symmetry; apply en.
    Qed.

    Lemma chain_TC:
      forall (p: list Node), p <> nil ->
                             forall (x y: Node), 
                               cluster_path x p y ->
                               transitive_closure (is_cluster_parent g) x y.
    Proof.
      induction p; intros nnil.
      contradict nnil; reflexivity.
      clear nnil.
      intros x y [PT CP].
      case_eq (p); [intros P | intros l b P].
      rewrite P in CP; simpl in CP.
      left.
      assert (pg: Proper pequiv g).
      apply term.
      rewrite <- (is_cluster_parent_compat
                    pg _ _ (Equivalence_Reflexive x) _ _ CP); assumption.
      right with a.
      left; assumption.
      apply IHp.
      rewrite P; intros F; symmetry in F; contradict F; apply nil_cons; clear P.
      assumption.
    Qed.

    Lemma minidminatall_not_none:
      forall p, alpha (g p) > k -> 
                exists c pp, Some c == MinIDMinATall (g p) (local_env g p) /\
                             peer p c == Some pp.
    Proof.
      assert (pg: Proper pequiv g).
      apply term.
      intros p GT. 
      destruct (state p) as [A _].
      unfold MinIDMinATall.
      case_eq (filter (isMinATalls (g p) (local_env g p))
                      (Local_Children (KCpeers (g p)) (lPar (local_env g p))
                                      (KCreverse (g p)))).

      (* nil *)
      intros F; rewrite (filter_nil ) in F.

      contradict F.
      rewrite Bool.not_false_iff_true.
      rewrite existsb_exists_InA.
      2: eapply isMinATalls_compat.
      2: reflexivity. 
      2: unfold local_env; intros c1 c2 ec;
        generalize (peer_compat _ _ (Equivalence_Reflexive p) _ _ ec);
        destruct (peer p c1); destruct (peer p c2); simpl; try tauto; 
          intros en; apply pg; assumption.

      assert (LE := new_alpha_0_2k g p).
      rewrite <- A in LE.
      unfold NewAlpha in A.

      assert (MIN := MinATall_range g p). 
      assert (MAX := MaxAShort_range g p). 

      destruct (Z_le_gt_dec
                  (MaxAShort (g p) (local_env g p) +
                   MinATall (g p) (local_env g p) + 2) 
                  (2 * k)). 
      2: lia.

      unfold MinATall in *.
      destruct (fold_right_Zmin_InA
                  (2*k+1) (TallAlphas (g p) (local_env g p))); unfold Z.eq in *.
      rewrite H in A; lia.

      unfold TallAlphas in H; rewrite filter_InA in H.
      2: intros a b e; unfold _IsTall; rewrite e; reflexivity. 
      destruct H as [H _].
      
      assert (TT := ChildrenAlphas_inv pg sym _ H).
      destruct TT as [pp [c [ischan [ap tp]]]].
      assert (X := InA_Children (Par_part_compat pg) p pp).
      rewrite X in tp; clear X.
      destruct tp as [cc [ischan' ina']].
      exists cc.
      split.
      assert (tmp:
                equivlistA
                  equiv
                  (Local_Children (peers p)
                                  (fun c0 => match peer p c0 with
                                             | Some x => Par_part g x
                                             | None => None
                                             end) (reverse p))
                  (Local_Children (KCpeers (g p))
                                  (lPar (local_env g p))
                                  (KCreverse (g p)))).
      {
        eapply Local_Children_locality.
        intros c1 c2 ec.
        generalize (peer_compat _ _  (Equivalence_Reflexive p) _ _ ec);
          destruct (peer p c1); destruct (peer p c2); simpl; try tauto.
        intros en; apply Par_part_compat; [apply pg | apply en].
        apply lPar_compat, (local_env_compat pg _ _ (Equivalence_Reflexive p)).
        apply reverse_compat; auto.
        apply KCreverse_local_compat.
        symmetry; apply (proj1 sym).
        intros ccc i; split.
        unfold lPar, local_env, Par_part; destruct (peer p ccc); simpl.
        apply Par_compat, pg; auto.
        tauto.
        symmetry; apply (proj2 sym); auto.
      }
      rewrite <- (@InA_compat _ _ _  _ _ (Equivalence_Reflexive cc) _ _ tmp);
        clear tmp.
      assert (tmp := (Children_equiv p (Par_part_compat pg))).
      apply eqlistA_equivlistA in tmp.
      rewrite <- (@InA_compat _ _ _ _ _ (Equivalence_Reflexive cc) _ _ tmp);
        assumption.
      apply setoid_equiv.
      
      unfold isMinATalls. 
      unfold local_env at 2.
      unfold is_channel in ischan'.
      revert ischan'; destruct (peer p cc); [| intros F; elim F].
      simpl; intros enpp.
      rewrite Z.eqb_eq.
      rewrite (alpha_compat (pg _ _ enpp)), ap.
      unfold MinATall, TallAlphas; reflexivity.

      (* non nil *)
      intros c l ecl.
      remember (fold_right
                  (min (ltC (Chans:=Chans)) (ltC_dec (Chans:=Chans))) c l)
        as fr.
      exists fr.

      assert (tmp := fold_right_min_InA
                       (A_dec := Channel_dec) _ ltC_dec fr c l).     
      assert (tmpp: fr == fold_right (min (ltC (Chans:=Chans))
                                          (ltC_dec (Chans:=Chans))) c l).
      { rewrite Heqfr; reflexivity. }
      specialize (tmp tmpp); clear tmpp.
      rewrite <- ecl in tmp.
      rewrite filter_InA_locality in tmp.
      destruct tmp as [I G].
      unfold Local_Children in I.
      rewrite filter_InA_locality in I.
      destruct I as [I L].
      apply (proj1 sym) in I.
      rewrite peers_spec_peer in I.
      unfold is_channel in I.
      destruct I as [pp Hpp].
      exists pp.
      split; auto.

      { intros c1 c2 ec i.
        etransitivity.
        apply Local_IsChild_locality; auto.
        apply Local_IsChild_compat; auto.
        apply KCpeers_compat, (pg _ _ (Equivalence_Reflexive p)).
        apply lPar_compat, (local_env_compat pg _ _ (Equivalence_Reflexive p)).
        apply KCreverse_local_compat; auto.
      }
      { intros c1 c2 ec i.
        apply isMinATalls_compat; auto.
        apply (local_env_compat pg _ _ (Equivalence_Reflexive p)).
      }
    Qed.

    (* None parCl is self-head *)
    Lemma non_pcl_self_head:
      forall p, parCl (g p) == None -> id (g p) == hd (g p).
    Proof.
      intros p N.
      destruct (state p) as [_ [pcl hd]].
      rewrite N in pcl. 
      rewrite hd.
      unfold NewHead.
      revert pcl.
      unfold NewparCl, IsClusterHead, _IsShort.

      case_eq (alpha (g p) <?k).
      intros LE ROOT.
      rewrite <- ROOT.
      assert (tmp: (dec2b (A := option Channel)
                          (eqoptionA_dec Channel_dec) None None) = true).
      apply dec2b_spec_true; reflexivity.
      rewrite tmp; simpl; clear tmp.
      rewrite orb_true_r.
      reflexivity.

      case_eq (alpha (g p) =?k).
      intros EQ GE _; simpl.
      reflexivity.
      intros NE GE F; contradict F.
      assert (tmpp: alpha (g p) > k).
      rewrite Z.eqb_neq in NE.
      rewrite Z.ltb_ge in GE.
      lia.
      destruct (minidminatall_not_none _ tmpp) as [c [n [H _]]].
      rewrite <- H.
      auto.
    Qed.
    

    (** Every node in a (non nil) cluster path has a different ids
    from another. *)
    Lemma diff_id_in_cluster_path:
      forall p b e, cluster_path b p e 
                    -> p <> nil -> ~ (id (g b)) == (id (g e)).
    Proof.
      intros p b e CP nonnil ID.
      apply uniqueID in ID.
      assert (no_loop := wf_rel_has_non_reflexive_transitive_closure
                           is_cluster_parent_wf).
      specialize (no_loop b).
      contradict no_loop.
      apply chain_TC with p.
      assumption.
      rewrite (cluster_path_compat
                 _ _ (Equivalence_Reflexive b) (Equivalence_Reflexive p)
                 _ _ ID); 
        assumption.
    Qed.


    Lemma parCl_is_a_node:
      forall p c, parCl (g p) = Some c -> exists pp, (peer p c) == (Some pp).
    Proof.
      intros p c epc.
      destruct (state p) as [_ [P _]].
      unfold NewparCl in P.
      rewrite epc in P.
      case_eq (alpha (g p) <?k); intros GT; rewrite GT in P.
      assert (PP := @parent_is_peer _ _ _ _ span).
      unfold Par_part in PP; symmetry in P.
      specialize (PP _ _ P).
      unfold is_channel in PP.
      apply PP.
      case_eq (alpha (g p) =?k); intros LE; rewrite LE in P.
      inversion P.
      assert (tmpp: alpha (g p) > k).
      rewrite Z.eqb_neq in LE.
      rewrite Z.ltb_ge in GT.
      lia.
      destruct (minidminatall_not_none _ tmpp) as [cc [n [H HH]]].
      rewrite <- P in H; simpl in H.
      exists n.
      rewrite <- (peer_compat _ _ (Equivalence_Reflexive p) _ _ H).
      assumption.
    Qed.

    (** Every none parCl has same hd and id *)
    Lemma hd_id_same:
      forall n,  (parCl (g n)) == None <-> (id (g n)) == (hd (g n)).
    Proof.
      intros p. 
      split.

      (* -> *)
      apply non_pcl_self_head.

      (* <- *)
      intros H.
      assert (pg: Proper pequiv g).
      apply term.
      assert (tmp: (forall x y : Node,
                       {(is_cluster_parent g) x y} +
                       {~ (is_cluster_parent g) x y})).
      intros x y.
      unfold is_cluster_parent.
      destruct (clusterParent (g y)); [|right; tauto];
        destruct (eqoptionA_dec Node_dec (peer y c) (Some x));
        [left|right];assumption.
      destruct (wf_finite_path all_nodes_prop  tmp 
                               (is_cluster_parent_compat pg)
                               is_cluster_parent_wf p) 
        as [path [h [C CH]]]; clear tmp.
      assert (tmp:  (parCl (g h)) == None).
      unfold is_cluster_parent, clusterParent in CH.
      revert CH; case_eq (parCl (g h)).
      intros c yc S.
      destruct (parCl_is_a_node _ yc) as [x N]. 
      specialize (S x); contradict S;  rewrite N; auto.
      reflexivity.

      assert (C' := C).
      revert C; case_eq path; simpl.
      intros C eq.
      rewrite <- tmp.
      eapply (parCl_compat );apply pg; symmetry; assumption.
      intros q p' Hp [_ ch].
      assert (D := diff_id_in_cluster_path).
      unfold cluster_path in D.
      specialize (D _ _ _ C').
      assert (tmpp: path <> nil).
      rewrite Hp.
      intros c; symmetry in c; contradict c; apply nil_cons.
      specialize (D tmpp); clear tmpp.
      contradict D.
      assert (OCH := non_pcl_self_head h tmp).
      assert (S := same_hd path h p C').
      rewrite OCH, S; symmetry; apply H.
    Qed.

    Lemma pcl_not_none:
      forall p, kDominator g p = true <-> (parCl (g p)) == None.
    Proof.
      intros p.
      unfold kDominator, IsRoot, _IsShort, alpha_part, Par_part.
      destruct (state p) as [_ [P _]].
      unfold NewparCl in P.
      rewrite orb_true_iff, andb_true_iff, dec2b_spec_true.
      split; intros H.
      (* -> *)
      destruct H as [H| [H HH]].
      assert (tmp: alpha (g p) <? k = false).
      rewrite Z.eqb_eq in H; rewrite Z.ltb_ge; lia.
      rewrite tmp, H in P.
      apply P.
      rewrite H in P.
      rewrite P; apply HH.
      
      (* <- *)
      case_eq (alpha (g p) <? k); intros LT; rewrite LT in P.
      right; split; [reflexivity|
                     rewrite <- P; apply H].
      case_eq (alpha (g p) =? k); intros EQ; rewrite EQ in P.
      left; reflexivity.
      rewrite P in H; contradict H.
      rewrite Z.eqb_neq in EQ.
      rewrite Z.ltb_ge in LT.
      assert (GT: alpha (g p) > k).
      lia.
      destruct (minidminatall_not_none _ GT) as [x [_ [Hx _]]].
      rewrite <- Hx.
      tauto.
    Qed.

    Lemma clusterHead_is_kDominator:
      forall (n: Node), kDominator g n = true <-> clusterHead g n.
    Proof.
      intros n.
      rewrite pcl_not_none.
      unfold kCluster, clusterHeadID.
      apply hd_id_same.
    Qed.

    (******************************************************)
    (******************************************************)
    (******************************************************)

    Definition get_cluster_parent (n: Node): option Node :=
      match (parCl (g n)) with 
      | None => None
      | Some c => (peer n c)
      end.

    Instance get_cluster_parent_compat: Proper fequiv get_cluster_parent.
    Proof.
      intros a b eab; unfold get_cluster_parent.
      assert (pg: Proper pequiv g).
      apply term.
      generalize (parCl_compat (pg _ _ eab));
        destruct (parCl (g a)); destruct (parCl (g b)); simpl; 
          try tauto.
      intros ec; apply peer_compat; assumption.
    Qed.
    Definition is_cp x y := (Some x) == (get_cluster_parent y).

    Lemma is_cp_equiv:
      forall x y, is_cp x y <-> (is_cluster_parent  g) x y.
    Proof.
      intros x y; unfold is_cp, is_cluster_parent, get_cluster_parent, clusterParent.
      destruct (parCl (g y)); split; auto; intros; symmetry; assumption.
    Qed.

    Instance is_cp_compat: Proper fequiv is_cp.
    Proof.
      intros a a' ea b b' eb.
      rewrite is_cp_equiv. rewrite is_cp_equiv.
      apply is_cluster_parent_compat; try assumption.
      apply term.
    Qed.

    Lemma is_cp_wf: well_founded is_cp.
    Proof.
      eapply Inclusion.wf_incl.  
      2: apply is_cluster_parent_wf.
      intros x y H; apply is_cp_equiv; assumption.
    Qed.      

    Definition build_cluster_path (n: Node) 
               (W: Acc is_cp n): list Node :=
      @Fix_F Node is_cp (fun _ => list Node)
             (fun (x : Node)
                  (f_rec : forall y : Node, is_cp y x -> list Node) =>
                (match get_cluster_parent x as x0 
                       return (x0 == get_cluster_parent x -> list Node) 
                 with
                 | Some px => fun heq => (f_rec px heq)
                 | None => fun _ => nil
                 end (Equivalence_Reflexive (get_cluster_parent x)))
                  ++ x::nil) n W.

    Definition build_cp (n: Node): list Node :=
      build_cluster_path (is_cp_wf n).

    Instance build_cp_compat: Proper (equiv ==> eqlistA equiv) build_cp.
    Proof.
      assert (pg: Proper pequiv g).
      apply term.
      intros a b eab.
      unfold build_cp.
      remember (is_cp_wf a) as A; clear HeqA.
      remember (is_cp_wf b) as B; clear HeqB.
      revert b B eab.
      induction A as [a A IH] using Acc_ind_dep.
      simpl.
      destruct B as (B).
      intros eab.
      simpl.
      generalize (get_cluster_parent_compat a b eab).
      revert  A B IH eab. 
      unfold is_cp.
      destruct (get_cluster_parent a) as [ p1 |];
        destruct (get_cluster_parent b) as [ p2 |];
        intros A B IH eab tp.
      simpl in tp.
      specialize (IH _ (Equivalence_Reflexive (Some p1)) 
                     p2 (B p2 (Equivalence_Reflexive (Some p2))) tp).
      apply app_eqlistA_compat.
      assumption.
      apply eqlistA_cons; [assumption|reflexivity].
      inversion tp.
      inversion tp.
      apply app_eqlistA_compat.
      reflexivity.
      apply eqlistA_cons; [assumption|reflexivity].
    Qed.

    Lemma build_cp_is_cluster_path:
      forall n, 
        match (build_cp n) with 
        | nil => False
        | hd::tl => cluster_path hd tl n
        end.
    Proof.
      intros n.
      unfold cluster_path, build_cp.
      remember (is_cp_wf n) as A; clear HeqA.
      induction A as [n A IH] using Acc_ind_dep; simpl.
      revert  A IH; unfold is_cp.

      case_eq (get_cluster_parent n). intros pn Hpn. 
      intros A IH.
      change (match build_cluster_path
                      (A pn (Equivalence_Reflexive (Some pn))) ++ n :: nil with
              | nil => False
              | hd :: tl => chain (is_cluster_parent g) hd tl n
              end).
      case_eq (build_cluster_path
                 (A pn (Equivalence_Reflexive (Some pn))) ++ n::nil); intros B.
      assert (X := app_cons_not_nil
                     (build_cluster_path
                        (A pn (Equivalence_Reflexive (Some pn)))) nil n); auto. 
      specialize (IH _ (Equivalence_Reflexive (Some pn))).

      case_eq (build_cluster_path
                 (A pn (Equivalence_Reflexive (Some pn)))); intros BB.
      rewrite BB in IH; elim IH.
      intros lBB HBB lB H.      
      rewrite HBB in IH.
      rewrite <- app_comm_cons in H.
      inversion H; clear H.
      rewrite H1 in IH; clear H1 H2.
      apply chain_app.
      apply is_cluster_parent_compat.
      apply term.
      exists pn.
      split; [apply IH | simpl].
      split; [ | reflexivity].
      rewrite <- is_cp_equiv; unfold is_cp; rewrite Hpn; reflexivity.

      intros _ _ _.
      case_eq (nil ++ n::nil).
      intros F.
      assert (X :=app_cons_not_nil nil nil n).
      auto.
      intros nn ll.
      intros H.
      rewrite app_nil_l in H.
      inversion H; simpl; reflexivity.
    Qed.

    Lemma build_cp_head:
      forall n, 
        match (build_cp n) with 
        | nil => False
        | hd::_ => clusterHead g hd
        end.
    Proof.
      intros n.
      unfold build_cp, kCluster, clusterHeadID.
      remember (is_cp_wf n) as A; clear HeqA.
      induction A as [n A IH] using Acc_ind_dep; simpl.
      revert  A IH; unfold is_cp.

      case_eq (get_cluster_parent n).
      intros pn Hpn A IH.
      change (match
                 build_cluster_path
                   (A pn (Equivalence_Reflexive (Some pn))) ++ n :: nil
               with
               | nil => False
               | hdd :: _ => id (g hdd) == hd (g hdd)
               end).
      case_eq (build_cluster_path
                 (A pn (Equivalence_Reflexive (Some pn))) ++ n::nil); intros B.
      assert (X := app_cons_not_nil
                     (build_cluster_path
                        (A pn (Equivalence_Reflexive (Some pn)))) nil n).
      auto.
      specialize (IH _ (Equivalence_Reflexive (Some pn))).
      case_eq (build_cluster_path
                 (A pn (Equivalence_Reflexive (Some pn)))); intros BB.
      rewrite BB in IH; elim IH.
      intros lBB HBB lB H.
      rewrite HBB in IH.
      rewrite <- app_comm_cons in H.
      inversion H; clear H.
      rewrite <- H1; assumption.

      intros G _ _.
      case_eq (nil ++ n::nil).
      intros F.
      assert (X :=app_cons_not_nil nil nil n).
      auto.
      intros nn ll.
      intros H.
      rewrite app_nil_l in H.
      inversion H; simpl; rewrite H1 in G.
      apply non_pcl_self_head.
      unfold get_cluster_parent in G. 
      case_eq (parCl (g nn)).
      intros c pc; rewrite pc in G.
      destruct (parCl_is_a_node nn pc) as [x FF].
      rewrite G in FF.
      elim FF.
      intros _; reflexivity.
    Qed.

    Lemma build_cp_head_id:
      forall b e, 
        kCluster g b e ->
        match (build_cp e) with 
        | nil => False
        | hdd::_ => hdd == b
        end.
    Proof.
      assert (pg: Proper pequiv g).
      apply term.
      intros b e kc.
      assert (H := build_cp_head e).
      assert (C := build_cp_is_cluster_path e).
      case_eq (build_cp e).
      intros FF; rewrite FF in C; elim C.
      intros h l bhl.
      rewrite bhl in C, H.
      assert (S := same_hd _ _ _ C).
      unfold kCluster, clusterHeadID in *.
      apply (uniqueID).
      rewrite kc, H; apply S.
    Qed.

    Lemma build_cluster_path_pr:
      forall b e,  kCluster g b e -> 
                   (exists p, cluster_path b p e) /\ clusterHead g b.
    Proof.
      intros b e kc.
      assert (P := build_cp_is_cluster_path e).
      assert (H := build_cp_head e).
      assert (I := build_cp_head_id _ _ kc).
      destruct (build_cp e).
      tauto.
      split.
      exists l; eapply cluster_path_compat.
      symmetry; apply I.
      reflexivity.
      reflexivity.
      apply P.
      eapply kCluster_compat.
      apply term.
      symmetry; apply I.
      symmetry; apply I.
      apply H.
    Qed.

    Lemma clusterHeadID_OK_OK:
      clusterHeadID_OK kCluster id clusterHeadID g.
    Proof.
      unfold clusterHeadID_OK, kCluster, clusterHeadID.
      tauto.
    Qed.

    Lemma cluster_path'_cluster_path:
      forall b p e, cluster_path b p e <-> cluster_path_ b p e.
    Proof.
      assert (pg: Proper pequiv g).
      apply term.
      intros b p_be e; split; intros P; [|apply P].
      unfold cluster_path; repeat split; try assumption.
      intros q ina; unfold clusterHeadID.
      apply InA_split in ina.
      destruct ina as [p_bt [q' [p_ue [eqq ep_be]]]].
      apply same_hd with (p_bt ++ q'::nil).
      unfold cluster_path in *.
      assert (comp := is_cluster_parent_compat pg).
      rewrite ep_be in P; apply chain_app; apply chain_app in P; try assumption.
      destruct P as [t [ep_bt ep_te]].
      simpl in ep_te; destruct ep_te as [etq _].
      exists t.
      split; [apply ep_bt|simpl; split; [apply etq | symmetry; apply eqq]].
    Qed.

    Lemma cluster_path_OK_OK:
      cluster_path_OK kCluster clusterHeadID clusterParent g.
    Proof.
      unfold cluster_path_OK.
      intros h ch n kCl.
      destruct (build_cluster_path_pr _ _ kCl)as [[p Hp] Hh].
      exists p; split.
      apply cluster_path'_cluster_path; assumption.
      assert (pg: Proper pequiv g).
      apply term.
      eapply any_is_linked; try assumption.
      apply pg.
      apply span.
      apply sym.
      { intros nn; destruct (state nn) as [A _].
        unfold alpha_enabled_bb, alpha_enabled_b.
        destruct (Z.eq_dec (alpha (g nn)) (NewAlpha (g nn) (local_env g nn)));
          [auto|tauto].
      }
      eapply RelationA.chain_incl.
      2: unfold cluster_path in Hp; apply Hp.
      apply inclu.
    Qed.

    Lemma kCluster_partition_OK:
      partition_OK kCluster g.    
    Proof.
      unfold partition_OK.
      intros n.
      assert (pg: Proper pequiv g).
      apply term.
      assert (H := build_cp_head n).
      assert (C := build_cp_is_cluster_path n).
      case_eq (build_cp n).
      intros FF; rewrite FF in C; elim C.
      intros h l bhl.
      rewrite bhl in C, H.
      assert (S := same_hd _ _ _ C).
      unfold kCluster, clusterHeadID in H.
      split.
      exists h; unfold kCluster, clusterHeadID; rewrite <- S; apply H.
      intros b b'; unfold kCluster, clusterHeadID; intros kc kc'.
      apply (uniqueID).
      rewrite kc, kc'; reflexivity.
    Qed.


    Lemma kClustering_ID_SPEC_OK:
      kClustering_ID_SPEC kCluster id clusterHeadID clusterParent g.
    Proof.
      split.
      apply clusterHeadID_OK_OK.
      apply cluster_path_OK_OK.
    Qed.

  End With_Terminal.

End KClustering_Correctness.

Close Scope Z_scope.
Unset Implicit Arguments.
