(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
*)
(* begin show *)

(** Global imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import ZArith.
From Coq Require Import Lia.
From Coq Require Import Bool.

(** Local imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import OptionUtil.
From Padec Require Import ZUtils.
From Padec Require Import ListUtils.
From Padec Require Import RelationA.
From Padec Require Import FoldRight.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Spanning_Tree.
From Padec Require Import KClustering_specification.
From Padec Require Import KClustering_algo.
From Padec Require Import KClustering_def_termination.

(* end show *)
Open Scope signature_scope.
Open Scope Z_scope.
Set Implicit Arguments.

(** * Correctness of the K-Clustering algorithm: alpha-variables

  Once a terminal configuration is reached, we prove that the
 alpha-variables draw a k-dominating set *)
Section KClustering_Correctness_Alpha.

  (* ********************************************************)
  (** Channel interface definitions *)  
  Context {Chans: Channels} {Net: Symmetric_Network}
          {KCP: KClustering_parameters} {KCT: KC_algo_tools}.

  Existing Instance KC_Stable.
           
  Notation Env := (Env (Algo := KC_algo)).

  Let k_pos := k_pos.


  (** ** Definition for kdom-graph *)
  Section KDom_Graph.

    (** ** kDomSet definitions *)
    Definition is_kDom_edge (g: Env) (x y: Node): Prop := 
      match alpha (g y) ?= k with
        | Lt => (is_tree_parent_of (Par_part g) x y)
        | Eq => False
        | Gt => (is_tree_parent_of (Par_part g) y x) /\ 
                alpha (g x) = alpha (g y) - 1
      end.

    Global Instance is_kDom_edge_compat: Proper fequiv is_kDom_edge.
    Proof.
      intros g1 g2 eg n1 n2 en n'1 n'2 en'; unfold is_kDom_edge.
      rewrite (alpha_compat (eg _ _ en')). 
      case (alpha (g2 n'2) ?=k).
      reflexivity.
      apply is_tree_parent_of_compat; try assumption; apply Par_part_compat;
        assumption.
      rewrite (eg _ _ en), (eg _ _ en').
      apply and_compat; [apply is_tree_parent_of_compat|]; auto; try tauto.
      apply Par_part_compat; auto.
    Qed.

  End KDom_Graph.

  (** ** alpha is always between 0 and 2k *)
  Section Range_of_alpha_Preserved.

    Lemma MinATall_range: 
      forall (g: Env) (n: Node),
        k <= MinATall (g n) (local_env g n) <= 2*k+1.
    Proof.
      intros g n.
      unfold MinATall. 
      destruct (fold_right_Zmin_dec (2*k+1) 
                                    (TallAlphas (g n) (local_env g n)))
        as [eqa | ina];
        unfold Z.eq in *.
      rewrite eqa; lia.

      split.
      unfold TallAlphas in ina.
      rewrite filter_InA in ina.
      2: intros a b ab; unfold _IsTall; rewrite ab; reflexivity.
      destruct ina as [ina2 tall].
      unfold _IsTall in tall.
      rewrite Z.geb_le in tall.
      assumption.

      apply fold_right_Zmin_eqA_leA.
      reflexivity.
    Qed.

    Lemma MaxAShort_range: 
      forall (g: Env) (n: Node),
        -1 <= MaxAShort (g n) (local_env g n) <= k-1.
    Proof.
      intros g n.
      unfold MaxAShort.
      destruct (fold_right_Zmax_dec (-1) (ShortAlphas (g n) (local_env g n)))
        as [eqa | ina];
        unfold Z.eq in *.
      rewrite eqa; lia.
      split.
      rewrite <- Z.ge_le_iff.
      apply fold_right_Zmax_eqA_leA.
      reflexivity.
      unfold ShortAlphas in ina.
      rewrite filter_InA in ina.
      2: intros a b ab; unfold _IsShort; rewrite ab; reflexivity.
      destruct ina as [ina2 tall].
      unfold _IsShort in tall.
      rewrite <- Zlt_is_lt_bool in tall.
      unfold ShortAlphas, _IsShort; lia.
    Qed.
    
    Lemma new_alpha_0_2k : 
      forall (g: Env) (n: Node), 
        0 <= NewAlpha (g n) (local_env g n) <= 2*k.
    Proof.
      intros g n.
      unfold NewAlpha.
      assert (tmp := MinATall_range g n).
      assert (tmp2 := MaxAShort_range g n).
      destruct (Z_le_gt_dec
         (MaxAShort (g n) (local_env g n) +
          MinATall (g n) (local_env g n) + 2) 
         (2 * k)).
      lia.
      lia.
    Qed.

  End Range_of_alpha_Preserved.

  (* *************************************************************)
  (** ** Assume that the computation of alpha provides no new value *)
  (** relation on the values of alpha between children and parent  *)
  Section Alpha_child_parent.
    
    (** Assumptions on Inputs *)
    Variable (g: Env) (pg: Proper pequiv g)
             (symm: Symmetric_Network_assumption
                      Net (fun n => KCpeers (g n))
                      (fun n => KCreverse (g n))).

    Lemma ChildrenAlphas_inv: 
      forall n a,
        InA equiv a (ChildrenAlphas (g n) (local_env g n) ) ->
        exists m c, is_channel n c m /\ 
                    alpha (g m) = a /\ 
                    is_tree_parent_of (Par_part g) n m.
    Proof.
      intros n a ina.
      unfold ChildrenAlphas in ina.
      rewrite InA_map_iff in ina.
      2: apply alpha_compat.
      destruct ina as [s [eqa ina]].
      unfold ChildrenStates in ina.
      rewrite InA_map_filter_iff in ina.
      2: intros c1 c2 ec;
        apply (local_env_compat pg _ _ (Equivalence_Reflexive n) _ _ ec). 
      destruct ina as [c [ina eqao]].
      assert (X: equivlistA
                   equiv
                   (Local_Children (peers n)
                                   (fun c0 : Channel => match peer n c0 with
                                                        | Some x => Par_part g x
                                                        | None => None
                                                        end) (reverse n))
                   (Local_Children (KCpeers (g n))
                                   (lPar (local_env g n)) (KCreverse (g n)))).
      {
        eapply Local_Children_locality.
        intros c1 c2 ec.
        generalize (peer_compat _ _  (Equivalence_Reflexive n) _ _ ec);
          destruct (peer n c1); destruct (peer n c2); simpl; try tauto.
        intros en; apply Par_part_compat; [apply pg | apply en].
        apply lPar_compat, (local_env_compat pg _ _ (Equivalence_Reflexive n)).
        apply reverse_compat; auto.
        apply KCreverse_local_compat.
        symmetry; apply (proj1 symm).
        intros cc i; split.
        unfold lPar, local_env, Par_part; destruct (peer n cc); simpl.
        apply Par_compat, pg; auto.
        tauto.
        symmetry; apply (proj2 symm); auto.
      }
      assert (tmp: InA equiv c (Children (Par_part g) n)).
      {
        assert (tmp := (Children_equiv n (Par_part_compat pg))).
        apply eqlistA_equivlistA in tmp.
        2: apply setoid_equiv.
        rewrite (InA_compat _ (Equivalence_Reflexive c) tmp); clear tmp.
        revert ina; apply InA_compat; auto; apply setoid_equiv.
      }
      apply (Children_inv (Par_part_compat pg)) in tmp.
      destruct tmp as [m [ischan ispar]].      
      exists m.
      exists c.
      split; auto.
      split; auto.
      rewrite <- eqa.
      eapply alpha_compat.
      unfold local_env in eqao.
      unfold is_channel in ischan.
      destruct (peer n c); [| elim ischan].
      simpl in eqao.
      symmetry; rewrite <- eqao.
      simpl in ischan; apply pg; assumption.
    Qed.

    Lemma NewAlpha_inv: 
      forall (n: Node),
        alpha (g n) = NewAlpha (g n) (local_env g n) -> alpha (g n) > 0 ->
        exists m, is_tree_parent_of (Par_part g) n m /\ alpha (g n) =
                                                        alpha (g m) + 1.
    Proof.
      intros n same pos.
      unfold NewAlpha in same.
      remember (MaxAShort (g n) (local_env g n)) as max.
      remember (MinATall (g n) (local_env g n)) as min.
      revert same; case_eq (Z_le_gt_dec (max + min + 2) (2 * k));
      intros le'  eq' same.
      unfold MinATall in Heqmin.
      remember (TallAlphas (g n) (local_env g n)) as tas.
      remember (2 * k + 1) as x0.
      assert (fdec := fold_right_Zmin_dec x0 tas).
      destruct fdec as [fdec | fdec].
      rewrite fdec in Heqmin.
      assert (le'' := le').
      rewrite Heqmin in le''.
      rewrite Heqx0 in le''.
      assert (X := MaxAShort_range g n).

      lia.
      rewrite <- Heqmin in fdec.
      rewrite Heqtas in fdec.
      unfold TallAlphas in fdec; rewrite filter_InA in fdec.
      destruct fdec as [child tall].
      clear eq'.
      
      specialize (ChildrenAlphas_inv _ child); try assumption.
      clear child; intros child.
      destruct child as [m [path [ischan [amin stp]]]].
      exists m.
      split.
      assumption.
      rewrite amin; assumption.
      intros x y exy; rewrite exy; reflexivity. 

      unfold MaxAShort in Heqmax.
      remember (ShortAlphas (g n) (local_env g n)) as sas.
      remember (-1) as x0.
      assert (fdec := fold_right_Zmax_dec x0 sas).
      destruct fdec as [fdec | fdec].
      rewrite fdec in Heqmax.
      rewrite Heqmax in same.
      rewrite Heqx0 in same.
      simpl in same.
      rewrite same in pos.
      lia.
      rewrite <- Heqmax in fdec.
      rewrite Heqsas in fdec.
      unfold ShortAlphas in fdec.
      rewrite filter_InA in fdec.
      destruct fdec as [child short].


      specialize (ChildrenAlphas_inv _ child); try assumption.
      clear child; intros child.
      destruct child as [m [path [ischan [amax stp]]]].
      exists m.
      split.
      assumption.
      rewrite same, amax.
      reflexivity.
      intros x y exy; rewrite exy; reflexivity.
    Qed.
    
  End Alpha_child_parent.

  (* ************************************************************)
  (** ** Correctness *)
  Section Final_kDomSet.

    (** Assumptions on Inputs *)
    Variable (g: Env) (pg: Proper pequiv g).

    Variable (root: Node)
             (span_tree: spanning_tree root (Par_part g))
             (sym: Symmetric_Network_assumption
                     Net (fun n => KCpeers (g n)) (fun n => KCreverse (g n))).

    Variable (final_alpha: 
                forall (n: Node), alpha_enabled_bb g n = false).

    Lemma same: 
      forall (n: Node), alpha (g n) = NewAlpha (g n) (local_env g n). 
    Proof.
      intros n. 
      specialize (final_alpha n).
      unfold alpha_enabled_bb, alpha_enabled_b in final_alpha.
      revert final_alpha; 
        case (Z.eq_dec (alpha (g n))
                      (NewAlpha (g n) (local_env g n))); intros X final.
      rewrite X; reflexivity.
      contradict final; apply diff_true_false.
    Qed.

    Definition is_kDom_path := chain (is_kDom_edge g).

    Notation eqCob := (dec2b (A := option Channel) (eqoptionA_dec Channel_dec)).
    Definition IsRoot (g: Env) (p: Node) := eqCob (Par_part g p) None.

    Lemma IsRoot_is_root_equiv:
      forall n1 n2 (en: n1 == n2) g1 g2 (eg: g1 =~= g2),
        is_root (Par_part g1) n1 <-> IsRoot g2 n2 = true.
    Proof.
      intros n1 n2 en g1 g2 eg.
      rewrite (is_root_compat (Par_part_compat eg) _ _ en).
      unfold is_root, IsRoot, dec2b.
      destruct (eqoptionA_dec Channel_dec (Par_part g2 n2) None) as [EQ|NE].
      intuition.
      intuition.
    Qed.
      
    Global Instance IsRoot_compat: Proper fequiv IsRoot.
    Proof.
      intros g1 g2 eg n1 n2 en; unfold IsRoot, dec2b.
        generalize (Par_part_compat eg _ _ en).
        destruct (Par_part g1 n1); destruct (Par_part g2 n2); simpl.
        2: tauto.
        2: tauto.
        2: reflexivity.
        intros. 
        case_eq (eqoptionA_dec Channel_dec (Some c) None); 
          case_eq (eqoptionA_dec Channel_dec (Some c0) None); intros.
        reflexivity.
        elim e.
        elim e.
        reflexivity.
      Qed.

      Definition kDominator (g: Env) (p: Node) := 
        orb (Z.eqb (alpha (g p)) k) (andb
                                       (_IsShort (alpha (g p))) (IsRoot g p)).

      Global Instance kDominator_compat: Proper fequiv kDominator.
      Proof.
        intros g1 g2 eg n1 n2 en.
        unfold kDominator. 
        rewrite (alpha_compat (eg _ _ en)).
        rewrite (IsRoot_compat eg _ _ en).
        reflexivity.
      Qed.
      
      (** *** Tall nodes *)
    Lemma tall_is_linked:
      forall (n: Node) (i: nat),
        alpha (g n) = k + Z_of_nat i ->
        (exists (r: Node), 
          kDominator g r = true /\
          exists (p: list Node), (is_kDom_path r p n))
        /\
        forall r p, is_kDom_path r p n -> (length p <= i)%nat.
    Proof.
      intros n i. 
      revert n.
      induction i as [ | j Hrec]; intros n eqz.
      simpl in eqz;rewrite Zplus_0_r in eqz.
      split.
      exists n;  split;
        [
          unfold kDominator;
          simpl in eqz;
          rewrite <- Z.eqb_eq in eqz;
          rewrite eqz;
          simpl;
          reflexivity
        |
        exists nil;
          simpl;
          reflexivity
        ].
      (* /\ *)
      intros r p.
      remember (rev p) as q.
      destruct q.
      assert (tmp:= rev_involutive p).
      rewrite <- Heqq in tmp; simpl in tmp.
      intros _; rewrite <- tmp; simpl; lia.
      assert (tmp := rev_unit (rev q) n0).
      rewrite rev_involutive, Heqq in tmp.
      rewrite <- rev_intro in tmp.
      rewrite <- tmp; unfold is_kDom_path; simpl.
      rewrite chain_app; simpl.
      2: apply is_kDom_edge_compat; apply pg. 
      intros [y [C [F en]]].
      rewrite (is_kDom_edge_compat pg _ _ (Equivalence_Reflexive y) _ _ en)
        in F.
      unfold is_kDom_edge in F.
      assert (tmpp: alpha (g n) ?= k = Eq).
      rewrite Z.compare_eq_iff; assumption.
      rewrite tmpp in F; tauto. 

      (* induction step *)
      split.
      assert (pos: alpha (g n) > 0).
      rewrite eqz; lia.
      assert (na := NewAlpha_inv pg sym n (same n) pos).
      destruct na as [m [path eqz']].
      rewrite eqz in eqz'.
      rewrite Nat2Z.inj_succ in eqz'.
      assert (h: alpha (g m) = k + Z.of_nat j).
      lia.
      destruct (Hrec _ h) as [[r [dom_set [p path']]] _]; clear Hrec.

      exists r.
      split.
      assumption.
      exists (p ++ (n :: nil)).
      apply chain_app.
      apply is_kDom_edge_compat; assumption.
      exists m.
      split.
      assumption.
      simpl.
      split; [| reflexivity].
      unfold is_kDom_edge.
      rewrite Nat2Z.inj_succ in eqz.
      assert (tmp: alpha (g n) ?= k = Gt).
      2: rewrite tmp; split; [assumption | rewrite eqz; lia]. 
      rewrite Z.compare_gt_iff; lia.
      (* /\ *)
      intros rr pp.
      remember (rev pp) as q.
      destruct q.
      rewrite rev_intro, rev_involutive in Heqq.
      simpl in Heqq. 
      intros _; rewrite <- Heqq; simpl; lia.
      assert (tmp := rev_unit (rev q) n0).
      rewrite rev_involutive, Heqq in tmp.
      rewrite <- rev_intro in tmp.
      rewrite <- tmp; unfold is_kDom_path; simpl.
      rewrite chain_app; simpl.
      2: apply is_kDom_edge_compat; apply pg. 
      intros [y [C [F en]]].
      rewrite (is_kDom_edge_compat pg _ _ (Equivalence_Reflexive y) _ _ en)
        in F.
      unfold is_kDom_edge in F.
      assert (tmpp: alpha (g n) ?= k = Gt).
      rewrite Z.compare_gt_iff; rewrite eqz; rewrite Nat2Z.inj_succ; lia.
      rewrite tmpp in F; clear tmpp. 
      destruct F as [_ ay].
      assert (tmpp: alpha (g y) = k + Z.of_nat j).
      rewrite ay, eqz, Nat2Z.inj_succ; lia.
      specialize (Hrec _ tmpp); clear tmpp.
      rewrite app_length; simpl.
      unfold is_kDom_path in Hrec.
      destruct Hrec as [_ Hrec]; specialize (Hrec _ _ C).
      lia.
    Qed.

    Lemma tall_k: 
      forall (n: Node), _IsTall (alpha (g n)) = true ->
                        exists (i: nat), alpha (g n) = k + Z.of_nat i.
    Proof.
      intros n tall.
      unfold _IsTall in tall.
      rewrite Z.geb_le in tall.
      apply ZUtils.le_plus in tall.
      destruct tall as [i p].
      exists i.
      assumption.
    Qed.
    
    Lemma short_k: 
      forall (n: Node), 
        _IsShort (alpha (g n)) = true ->
        exists (i: nat), (Z.of_nat i < k)%Z /\ alpha (g n) = k - 1 - Z.of_nat i.
    Proof.
      intros n short.
      unfold _IsShort in short.
      rewrite Z.ltb_lt in short.
      rewrite <- Z.gt_lt_iff in short.
      apply ZUtils.gt_plus in short.
      destruct short as [i p].
      exists i.
      split; [|assumption].
      rewrite (same n) in p.
      assert (h := new_alpha_0_2k g n).
      lia.
    Qed.

    Lemma short_or_tall:
      forall (n: Node), {_IsShort (alpha (g n)) = true} +
                        {_IsTall (alpha (g n)) = true}.
    Proof.
      intros n.
      unfold _IsShort.
      unfold _IsTall.
      destruct (Z_lt_le_dec ((g n).(alpha)) k) as [h | h].
      left; apply Z.ltb_lt; assumption.
      right; rewrite Z.geb_le; assumption.
    Qed.

    Lemma alpha_monotonic_small:
      forall (n m: Node), 
        is_tree_parent_of (Par_part g) m n ->
        _IsShort (alpha (g n)) = true -> alpha (g m) >= alpha (g n) + 1.
    Proof.
      intros n m nm n_short.
      destruct (short_or_tall m) as [m_short | m_tall].

      (* m short *)
      assert (same := same m).
      unfold NewAlpha in same.
      remember (MaxAShort (g m) (local_env g m)) as max.
      remember (MinATall (g m) (local_env g m)) as min.      
      assert (ina: InA
                     equiv (alpha (g n))
                     (filter _IsShort (ChildrenAlphas (g m) (local_env g m)))). 
      rewrite filter_InA.
      2: intros a b ab; unfold _IsShort; rewrite ab; reflexivity.
      split.
      2: assumption.
      rewrite (InA_Children (Par_part_compat pg) m n) in nm.
      unfold ChildrenAlphas, ChildrenStates.
      rewrite InA_map_iff.
      exists (g n).
      split; [reflexivity|].
      rewrite InA_map_filter_iff.
      destruct nm as [c [ischan ina]].
      exists c.
      assert (X: equivlistA
                   equiv
                   (Local_Children (peers m)
                                   (fun c0 : Channel => match peer m c0 with
                                                        | Some x => Par_part g x
                                                        | None => None
                                                        end) (reverse m))
                   (Local_Children (KCpeers (g m))
                                   (lPar (local_env g m)) (KCreverse (g m)))).
      {
        eapply Local_Children_locality.
        intros c1 c2 ec.
        generalize (peer_compat _ _  (Equivalence_Reflexive m) _ _ ec);
          destruct (peer m c1); destruct (peer m c2); simpl; try tauto.
        intros en; apply Par_part_compat; [apply pg | apply en].
        apply lPar_compat, (local_env_compat pg _ _ (Equivalence_Reflexive m)).
        apply reverse_compat; auto.
        apply KCreverse_local_compat.
        symmetry; apply (proj1 sym).
        intros cc i; split.
        unfold lPar, local_env, Par_part; destruct (peer m cc); simpl.
        apply Par_compat, pg; auto.
        tauto.
        symmetry; apply (proj2 sym); auto.
      }
      rewrite <- X, <- Children_equiv .
      split; auto.
      unfold local_env; unfold is_channel in ischan.
      rewrite ischan; simpl; reflexivity.
      apply Par_part_compat, pg.
      apply (local_env_compat pg _ _ (Equivalence_Reflexive m)).
      apply alpha_compat.
      
      destruct (Z_le_gt_dec (max + min + 2) (2 * k)).      
      assert (min_range := MinATall_range g m).
      unfold _IsShort in m_short.
      rewrite Z.ltb_lt in m_short.
      lia.
      rewrite same.
      assert (tmp: max >= alpha (g n)).
      rewrite Heqmax.
      apply fold_right_Zmax_InA_leA.
      assumption.
      lia.

      (* m tall *)
      unfold _IsShort in n_short.
      unfold _IsTall in m_tall.
      rewrite Z.ltb_lt in n_short.
      rewrite Z.geb_le in m_tall.
      lia.
    Qed.
    
    Lemma short_is_linked:
      forall (i: nat) (p: Node),
        alpha (g p) = k - Z.of_nat i ->
        (exists (kdom: Node),
           kDominator g kdom = true /\
           exists (path: list Node), (is_kDom_path kdom path p))
        /\
        forall kdom path, is_kDom_path kdom path p -> (length path <= i)%nat.
    Proof.
      assert (tmp := well_founded_ind lt_wf).
      specialize 
        (tmp (fun i => 
                forall n,
                  alpha (g n) = k - Z.of_nat i ->
                  (exists r,
                    kDominator g r = true /\
                    (exists p, (is_kDom_path r p n)))
                  /\
                  forall r p, is_kDom_path r p n -> (length p <= i)%nat
        )).
      apply tmp; clear tmp.
      intros i Hrec p eqa.
      destruct i as [ | i].

      (* base case *)
      apply tall_is_linked.
      simpl; simpl in eqa.
      lia.

      (* inductive case *)
      rewrite Nat2Z.inj_succ in eqa.
      assert (i_pos: 0 <= Z.of_nat i).
      lia.
      assert (p_short: _IsShort (alpha (g p)) = true).
      unfold _IsShort.
      rewrite Z.ltb_lt.
      lia.
      case_eq (IsRoot g p).
      
      (* p_root *)
      intros p_root.
      split.
      exists p; split.
      unfold kDominator.
      rewrite orb_true_iff; right; rewrite andb_true_iff; split; assumption.
      exists nil; simpl; reflexivity.
      (* /\ *)
      assert (kDom: kDominator g p = true).
      unfold kDominator.
      rewrite p_short, p_root, orb_comm; simpl; reflexivity.
      intros kdom path.
      remember (rev path) as qpath.
      destruct qpath.
      assert (tmp:= rev_involutive path).
      rewrite <- Heqqpath in tmp; simpl in tmp.
      intros _; rewrite <- tmp; simpl; lia.
      assert (tmp := rev_unit (rev qpath) n).
      rewrite rev_involutive, Heqqpath in tmp.
      rewrite <- rev_intro in tmp.
      rewrite <- tmp; unfold is_kDom_path; simpl.
      rewrite chain_app; simpl.
      2: apply is_kDom_edge_compat; apply pg. 
      intros [y [C [F en]]].
      rewrite (is_kDom_edge_compat pg _ _ (Equivalence_Reflexive y) _ _ en)
        in F.
      unfold is_kDom_edge in F.
      unfold _IsShort in p_short.      
      assert (tmpp: alpha (g p) ?= k = Lt).
      rewrite Z.compare_lt_iff; rewrite Z.ltb_lt in p_short; assumption. 
      rewrite tmpp in F; clear tmpp. 
      unfold IsRoot in p_root.
      rewrite dec2b_spec_true in p_root.
      unfold is_tree_parent_of, tree_parent in F.
      revert F; revert p_root; unfold Par_part; destruct (Par (g p)); simpl; try tauto.

      (* p_not_root *)
      intros p_not_root.
      split.
      assert (not_root: ~is_root (Par_part g) p).
      rewrite (IsRoot_is_root_equiv _ _ (Equivalence_Reflexive p) pg),
      p_not_root; apply diff_false_true.
      destruct (non_root_has_parent span_tree p not_root)
        as [q q_is_tree_parent_of_p].
      assert (q_is_kDom_edge_p: (is_kDom_edge g q p)).
      unfold is_kDom_edge.
      apply Zcompare_elim.
      intros f; lia.
      intros _; assumption.
      intros f; lia.
      assert (q_in_dom_set: 
                exists kdom,
                  kDominator g kdom = true /\
                  (exists path', (is_kDom_path kdom path' q))).

      (* kdom >:: path' ::> q *)
      assert (tmp := 
                alpha_monotonic_small p q q_is_tree_parent_of_p p_short).
      destruct (short_or_tall q) as [q_short | q_tall].
      (* q short *)
      assert (tmp' := short_k q q_short).
      destruct tmp' as [j eq'].
      remember (j + 1)%nat as j'.
      assert (eq'': alpha (g q) = (k) - Z.of_nat j').
      rewrite Heqj', Nat2Z.inj_add.
      unfold Z.sub in *.
      rewrite Z.opp_add_distr.
      simpl.
      subst.
      lia.
      assert (lt': (j' < Datatypes.S i)%nat).
      lia.
      eapply Hrec.
      apply lt'.
      apply eq''.

      (* m_tall *)
      destruct (tall_k _ q_tall) as [ii H]. 
      apply tall_is_linked with ii; assumption.
      
      (* kdom >:: path' ::> q /\ q :> p -> rkdom >:: path ::> p *)
      destruct q_in_dom_set as [kdom [CH [path' Hpath']]].   
      exists kdom.
      split; [ assumption |].
      exists (path' ++ (p :: nil)).
      unfold is_kDom_path.
      rewrite chain_app; [  | apply is_kDom_edge_compat; assumption].
      exists q.
      split; [ assumption | simpl].
      split; [ assumption | reflexivity].

      (* /\ *)
      assert (not_root: ~is_root (Par_part g) p).      
      rewrite (IsRoot_is_root_equiv _ _ (Equivalence_Reflexive p) pg),
      p_not_root; apply diff_false_true.

      intros B PATH.
      remember (rev PATH) as RPATH.
      destruct RPATH as [| pp RPATH].

      rewrite rev_intro, rev_involutive in HeqRPATH.
      simpl in HeqRPATH. 
      intros _; rewrite <- HeqRPATH; simpl; lia.

      assert (tmp := rev_unit (rev RPATH) pp).
      rewrite rev_involutive, HeqRPATH in tmp.
      rewrite <- rev_intro in tmp.
      rewrite <- tmp; unfold is_kDom_path; simpl.
      rewrite chain_app; simpl.
      2: apply is_kDom_edge_compat; apply pg. 

      intros [q [C [F en]]].
      rewrite (is_kDom_edge_compat pg _ _ (Equivalence_Reflexive q) _ _ en)
        in F.
      rewrite app_length; simpl.
      clear HeqRPATH tmp en pp.
      unfold is_kDom_edge in F.
      assert (tmpp: alpha (g p) ?= k = Lt).
      rewrite Z.compare_lt_iff; rewrite eqa; lia. 
      rewrite tmpp in F; clear tmpp. 

      destruct (short_or_tall q) as [q_short | q_tall].
     
      (* q_short *)
      {
        assert (X := alpha_monotonic_small _ _ F p_short).
        rewrite eqa in X.
        unfold _IsShort in q_short; rewrite Z.ltb_lt in q_short.
        assert (tmp: ( Z.to_nat (k - (alpha (g q))) < Datatypes.S i)%nat ).
        lia.
        specialize (Hrec _ tmp q); clear tmp.      
        assert (tmp: alpha (g q) = k - Z.of_nat (Z.to_nat (k - alpha (g q)))).
        rewrite Z2Nat.id.
        2: lia.
        lia.
        specialize (Hrec tmp); clear tmp.
        unfold is_kDom_path in Hrec.
        destruct Hrec as [_ Hrec]; specialize (Hrec _ _ C).
        lia.
      }
      (* tall *)
      { 
        assert (same := same q).
        unfold NewAlpha in same.
        remember (MaxAShort (g q) (local_env g q)) as max.
        remember (MinATall (g q) (local_env g q)) as min.
        destruct (Z_le_gt_dec (max + min + 2) 
                            (2 * (k))) as [le' | gt'].
      (* max + min + 2 <= 2 * k *)
        assert (ina: InA
                       equiv (alpha (g p))
                       (filter _IsShort (ChildrenAlphas (g q) (local_env g q)))). 
        rewrite filter_InA.
        2: intros a b ab; unfold _IsShort; rewrite ab; reflexivity.
        split.
        2: assumption.
        rewrite (InA_Children (Par_part_compat pg) q p) in F.
        unfold ChildrenAlphas, ChildrenStates.
        rewrite InA_map_iff.
        exists (g p).
        split; [reflexivity|].
        rewrite InA_map_filter_iff.
        destruct F as [c [ischan ina]].
        exists c.
        split.
        assert (X: equivlistA
                     equiv
                     (Local_Children (peers q)
                                     (fun c0 : Channel => match peer q c0 with
                                                          | Some x => Par_part g x
                                                          | None => None
                                                          end) (reverse q))
                     (Local_Children (KCpeers (g q))
                                     (lPar (local_env g q)) (KCreverse (g q)))).
        {
          eapply Local_Children_locality.
          intros c1 c2 ec.
          generalize (peer_compat _ _  (Equivalence_Reflexive q) _ _ ec);
            destruct (peer q c1); destruct (peer q c2); simpl; try tauto.
          intros en; apply Par_part_compat; [apply pg | apply en].
          apply lPar_compat, (local_env_compat pg _ _ (Equivalence_Reflexive q)).
          apply reverse_compat; auto.
          apply KCreverse_local_compat.
          symmetry; apply (proj1 sym).
          intros cc ii; split.
          unfold lPar, local_env, Par_part; destruct (peer q cc); simpl.
          apply Par_compat, pg; auto.
          tauto.
          symmetry; apply (proj2 sym); auto.
        }
        rewrite <- (InA_compat _ (Equivalence_Reflexive c) X); clear X.
        assert (tmp := (Children_equiv q (Par_part_compat pg))).
        apply eqlistA_equivlistA in tmp.
        rewrite <- (InA_compat _ (Equivalence_Reflexive c) tmp); assumption.
        apply setoid_equiv.
        unfold local_env.
        unfold is_channel in ischan.
        rewrite ischan; simpl; apply pg; reflexivity.
        intros c1 c2 ec; unfold local_env;
          generalize (peer_compat _ _ (Equivalence_Reflexive q) _ _ ec);
          destruct (peer q c1); destruct (peer q c2); simpl; try tauto;
          intros en; apply pg; assumption.
        apply (alpha_compat ).

        assert (K := fold_right_Zmax_InA_leA (-1) ina). 
        unfold MaxAShort, ShortAlphas in Heqmax;
          rewrite <- Heqmax in K.
        rewrite eqa in K.
        unfold _IsTall in q_tall; rewrite Z.geb_le in q_tall.
        unfold _IsShort in p_short; rewrite Z.ltb_lt in p_short.
        rewrite eqa in p_short.
        assert (tmp: exists j, alpha (g q) = k + Z.of_nat j).
        specialize (ZUtils.le_plus q_tall); auto.
        destruct tmp as [j Hj]. 
        assert (OK := tall_is_linked _ _ Hj). 
        destruct OK as [_ OK]; specialize (OK B (rev RPATH)).
        unfold is_kDom_path in OK; specialize (OK C).
        lia.

        (* max + min + 2 > 2 * k *)
        assert (eq': alpha (g q) = k).
        unfold _IsTall in q_tall.
        rewrite Z.geb_le in q_tall.
        assert (range_max := MaxAShort_range g q). 
        lia.
        destruct RPATH as [| x RPATH]; simpl in C; [simpl; lia|].
        rewrite chain_app in C.
        2: apply is_kDom_edge_compat; apply pg. 
        destruct C as [qq [CC [FF en]]]; simpl in en.
        rewrite (is_kDom_edge_compat pg _ _ (Equivalence_Reflexive qq) _ _ en)
          in FF. 
        unfold is_kDom_edge in FF.
        rewrite <- Z.compare_eq_iff in eq'; rewrite eq' in FF; elim FF.
      }
    Qed.

    Lemma any_is_linked:
      forall (n: Node), 
        (exists (r: Node), kDominator g r = true /\
                           exists (p: list Node), (is_kDom_path r p n))
        /\
        forall r p, is_kDom_path r p n -> Z.of_nat (length p) <= k.
    Proof.
      intros n.
      destruct (short_or_tall n) as [n_short | n_tall].
      (*short *)
      destruct (short_k n n_short) as [i [Hi HHi]].
      assert (tmp: exists ii, alpha (g n) = k - Z.of_nat ii).
      exists (Datatypes.S i); zify; lia.
      destruct tmp as [ii Hii].
      destruct (short_is_linked _ _ Hii) as [[r [kdom [p Hp]]] F].      
      split.
      exists r; split; [apply kdom | exists p; apply Hp].
      intros rr pp krp; specialize (F _ _ krp); lia.
      (* tall *)
      destruct (tall_k n n_tall) as [i Hi].
      destruct (tall_is_linked _ _ Hi) as [[r [kdom [p Hp]]] F].
      split.
      exists r; split; [apply kdom | exists p; apply Hp].
      assert (same := same n).
      assert (A := new_alpha_0_2k g n).
      intros rr pp krp; specialize (F _ _ krp); lia.
    Qed.

  End Final_kDomSet.

End KClustering_Correctness_Alpha.

Close Scope signature_scope.
Close Scope Z_scope.
Unset Implicit Arguments.
