(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
*)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import ZArith.
From Coq Require Import Lia.
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import RelationPairs.

(** **** Local Imports *)
From Padec Require Import NatUtils.
From Padec Require Import OptionUtil.
From Padec Require Import SetoidUtil.
From Padec Require Import Count.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Spanning_Tree.
From Padec Require Import KClustering_specification.
From Padec Require Import KClustering_algo.
From Padec Require Import KClustering_def_termination.
From Padec Require Import KClustering_correctness_alpha.

(* end show *)
Open Scope Z_scope.
Set Implicit Arguments.

(** * How Many Elements in the cluster?  Count the number of elements
 in the cluster (cluster heads) built by the algorithm.
 
 We prove bounds on it.  *)
Section KClustering_count.

  (* ********************************************************)
  (** Channel interface definitions *)  
  Context {Chans: Channels} {Net: Symmetric_Network}
          {KCP: KClustering_parameters} {KCT: KC_algo_tools}.

  Existing Instance KC_Stable.
  Existing Instance OrdChans.
           
  Notation Env := (Env (Algo := KC_algo)).

  Let k_pos := k_pos.

  (** Assumptions on Inputs *)
  Variable (g: Env) (pg: Proper pequiv g).
  
  Variable (root: Node)
           (span_tree: spanning_tree root (Par_part g))
           (sym: Symmetric_Network_assumption
                   Net (fun n => KCpeers (g n)) (fun n => KCreverse (g n))).

  Variable (final_alpha: 
              forall (n: Node), alpha_enabled_bb g n = false).

  (** *** Dominating Heads 
   Two kinds: 
              - the regular one (with alpha equals k)
              - (maybe) the root 
   *)
  Definition DomHeads := { n: Node | kDominator g n = true }.

  (** Regular Dominating Heads are nodes such that alpha = k *) 
  Definition RegDomHead p := (alpha (g p)) = k.

  Definition RegDomHeads := { n: Node | RegDomHead n }.

  Instance RegDomHead_compat: Proper fequiv RegDomHead.
  Proof.
    intros n1 n2 eqn.
    unfold RegDomHead.
    rewrite (alpha_compat (pg _ _ eqn)).
    reflexivity. 
  Qed.

  (** Regular Nodes are those below a tall node: 
      They cannot be in the dominating set of the short root (if short) *)

  Inductive HasTallAncestor: Node -> Prop :=
  | HTA_self: forall n, (alpha (g n)) >= k -> HasTallAncestor n
  | HTA_parent:  
      forall m n, is_tree_parent_of (Par_part g) m n ->
                  HasTallAncestor m -> HasTallAncestor n.

  Definition RegNode := { n: Node | HasTallAncestor n }.

  Instance HasTallAncestor_compat: Proper fequiv HasTallAncestor.
  Proof.
    intros n1 n2 en;split.
    intro w;revert n2 en;induction w;intros n2 en;[left|right with m].
    replace (alpha (g n2)) with (alpha (g n)).
    assumption.
    eapply (alpha_compat ).
    apply pg; assumption.
    revert H;apply is_tree_parent_of_compat.
    apply Par_part_compat; apply pg; reflexivity.
    reflexivity.
    symmetry;assumption.
    assumption.
    
    intro w;revert n1 en;induction w;intros n1 en;[left|right with m].
    replace (alpha (g n1)) with (alpha (g n)).
    assumption.
    eapply (alpha_compat ).
    symmetry;apply pg; assumption.
    revert H;apply is_tree_parent_of_compat.
    apply Par_part_compat; apply pg; reflexivity.
    reflexivity.
    assumption.
    assumption.
  Qed.

  (** Relation between a node and its closest ancestor s.t. alpha=k *)
  Inductive RegHeadAncestor: Node -> Node -> Prop :=
  | RHA_self: forall n h, n == h -> (alpha (g h)) = k -> RegHeadAncestor n h
  | RHA_parent:  
      forall m n h, is_tree_parent_of (Par_part g) m n -> (alpha (g n)) <> k -> 
                    RegHeadAncestor m h -> RegHeadAncestor n h.

  Instance RegHeadAncestor_compat: 
    Proper fequiv RegHeadAncestor.
  Proof.
    intros n1 n2 en a1 a2 ea;split.
    intro w;revert n2 en a2 ea;induction w;intros n2 en a2 ea;[left|right with m].
    rewrite <- en, <- ea;assumption.
    rewrite <- H0.
    eapply (alpha_compat ).
    symmetry; apply pg; assumption.
    revert H;apply is_tree_parent_of_compat.
    apply Par_part_compat; apply pg. 
    reflexivity.
    symmetry;assumption.
    intro e;apply H0.
    rewrite <- e.
    eapply (alpha_compat ).
    apply pg; assumption.
    apply IHw.
    reflexivity.
    assumption.

    intro w;revert n1 en a1 ea;induction w;intros n1 en a1 ea;
      [left|right with m].
    rewrite en, ea;assumption.
    rewrite <- H0.
    eapply (alpha_compat ).
    apply pg; assumption.
    revert H;apply is_tree_parent_of_compat.
    apply Par_part_compat; apply pg.
    reflexivity.
    assumption.
    intro e;apply H0.
    rewrite <- e.
    eapply (alpha_compat ).
    symmetry;apply pg;assumption.
    apply IHw.
    reflexivity.
    assumption.
  Qed.

  Section with_cardinals.

    Variable (card_N: nat)
             (Hcard_N: Num_Card Same Node card_N).

    Variable (card_DH: nat)
             (Hcard_DH: Num_Card Same DomHeads card_DH).

    Variable (card_RDH: nat)
             (Hcard_RDH: Num_Card Same RegDomHeads card_RDH).


    Variable (card_RN: nat)
             (Hcard_RN: Num_Card Same RegNode card_RN).

    (** [Rcount] is an existence witness for a the cardinality result *)
    Definition Rcount 
               (cpl: ltN (Datatypes.S (Z.to_nat k)) * RegDomHeads) 
               (rn: RegNode): Prop := 
      Z.to_nat (alpha (g (proj1_sig rn))) = proj1_sig (fst cpl) /\ 
      RegHeadAncestor (proj1_sig rn) (proj1_sig (snd cpl)).

    (** [count_ok] every small index is present in every dominating set *)
    Theorem simple_counting: 
      (card_RN >= (Datatypes.S (Z.to_nat k)) * card_RDH)%nat.
    Proof.
      unfold ge.
      rewrite <- Card_smaller_le_iff.
      apply Card_trans with (ltN (Datatypes.S (Z.to_nat k)) *
                             ltN card_RDH)%type _.
      apply Card_weak;apply Card_sym;simpl sym_Card_Prop;
        apply Card_as_many_prod.
      apply Card_trans
        with (ltN (Datatypes.S (Z.to_nat k)) * RegDomHeads)%type _.
      apply Card_sym;apply Card_weak;eapply Card_prod.
      apply Card_refl;apply eqltN_equiv.
      apply Hcard_RDH.
      apply Card_trans with RegNode _.
      2:apply Hcard_RN.

      simpl.
      split with Rcount.
      intros [[alpha1 halpha1] [ch1 hch1]] [[alpha2 halpha2] [ch2 hch2]] 
             [halpha hch];
        do 2 red in halpha;
        red in hch;
        simpl in halpha,hch.
      intros [n1 h1] [n2 h2] hn.
      red in hn;simpl in hn.
      unfold Rcount;simpl.
      apply Morphisms_Prop.and_iff_morphism.
      apply PER_morphism.
      apply Equivalence_PER.
      apply eq_equivalence.
      apply f_equal.
      eapply (alpha_compat ).
      apply pg; assumption.
      assumption.
      apply RegHeadAncestor_compat.
      assumption.
      apply hch.
      
      intros [[alpha1 halpha1] [ch1 hch1]] [n1 h1] [h11 h21].
      intros [[alpha2 halpha2] [ch2 hch2]] [n2 h2] [h12 h22]. 
      intro en;red in en;simpl in en.
      simpl in *.
      split.
      compute.
      rewrite <- h11, <- h12.
      apply f_equal.
      apply alpha_compat, pg; auto.
      assert (enn: n1 == n2) by apply en; clear en; rename enn into en.
      red;simpl.
      red;simpl.
      clear h1 h11 hch1 h2 h12 hch2.
      revert n2 ch2 en h22.
      induction h21;
        intros n2 ch2 en h22; 
        rewrite <- en in h22;
        inversion h22.
      rewrite H in H1.
      assumption.
      elim H2;rewrite <- H0.
      eapply (alpha_compat ).
      apply pg;assumption.
      elim H0;rewrite <- H2.
      eapply (alpha_compat ).
      apply pg; assumption.
      eapply IHh21 with m0.    
      apply (is_tree_parent_unique
               (Par_part_compat pg) _ _ (Equivalence_Reflexive n)_ _ H H1).
      assumption.

      intros [[alpha1 halpha1] [ch1 hch1]] [[alpha2 halpha2] [ch2 hch2]] 
             [halpha hch].
      compute in halpha;do 2 red in hch;simpl in hch.
      subst; rename alpha2 into a.
      generalize halpha2;revert halpha1.
      replace a with ((Z.to_nat k) -(Z.to_nat k  - a) )%nat.
      2:lia.    
      clear halpha2.
      generalize (Z.to_nat k - a)%nat (le_minus (Z.to_nat k) a).
      clear a.
      intro n;induction n as [|i Hi].
      rewrite <- minus_n_O.
      assert (hhch1:HasTallAncestor ch1).
      left.
      rewrite hch1.
      lia.
      assert (hhch2:HasTallAncestor ch2).
      left.
      rewrite hch2.
      lia.
      intros _ ha1 ha2;
        esplit with (exist hhch1);
        esplit with (exist hhch2).

      repeat split;simpl;red in hch1,hch2.
      rewrite hch1.
      reflexivity.
      left.
      reflexivity.
      assumption.
      rewrite hch2.
      reflexivity.
      left.
      reflexivity.
      assumption.
      apply hch.
      intro h.
      assert (hh:  (i <= Z.to_nat k)%nat).
      lia.
      specialize (Hi hh).
      intros halpha1 halpha2.
      assert (hhh:  (Z.to_nat k - i < Datatypes.S (Z.to_nat k))%nat).
      lia.
      specialize (Hi hhh hhh). 
      destruct Hi as [[b1 hb1] [[b2 hb2] [[h1a h1b] [[h2a h2b] eb]]]].
      red in eb;simpl in eb.
      unfold Rcount in *;simpl in *.
      destruct (Z_le_gt_dec (alpha (g b1)) 0) as [LE | GT].
      destruct (alpha (g b1)) as [| neg | pos];simpl in h1a,LE.
      lia.
      compute in LE.
      elim LE.
      reflexivity.
      lia.
      destruct (NewAlpha_inv pg sym _ (same g final_alpha b1) GT)
        as [son [hson eson]].
      assert(hhson: HasTallAncestor son).
      right with b1;assumption.
      split with (exist hhson).
      split with (exist hhson).
      rewrite Zplus_comm in eson.
      assert (eson':=Zplus_minus_eq _ _ _ eson).
      repeat split;simpl;try rewrite eson'.

      rewrite Z2Nat.inj_sub;try lia.

      right with b1;auto.
      intro e;rewrite eson,e,Zplus_comm,Z2Nat.inj_add in h1a;try lia.
      clear - h1a.
      revert h1a.
      simpl;unfold Pos.to_nat;simpl.
      intro;lia. 
      
      right with b2;auto.
      revert hson;eapply proj1.
      apply is_tree_parent_of_compat;try assumption.
      apply Par_part_compat; apply pg.
      reflexivity.

      intro e;rewrite eson,e,Zplus_comm,Z2Nat.inj_add in h1a;try lia.

      reflexivity.
    Qed.

    (** [split_counting_cases]
       if root is tall:
              -  all dominating heads are regular
              -  all nodes are regular (have a tall ancestor) 
       if root is short:
              -  all dominating heads are regular except the root
              -  at least one node is singular (namely the root) 
     *)
    Lemma split_counting_cases: 
      (( card_DH=card_RDH /\ card_N = card_RN ) \/
       ( card_DH = Datatypes.S card_RDH /\ card_N >= Datatypes.S card_RN))%nat.
    Proof.
      destruct span_tree. 
      destruct (Z.geb_spec (alpha (g root)) k) as [hroot| hroot];
        [left|right];split.
      
      (* tall.1 *)
      { eapply (Subset_Num_Card_iff_eq (A := Node)); eauto.
        intros n1 n2 en; unfold kDominator,RegDomHead.
        rewrite en.
        rewrite Bool.orb_true_iff.
        rewrite Bool.andb_true_iff.
        rewrite Z.eqb_eq.
        simpl; intuition.
        rewrite <- (IsRoot_is_root_equiv) in H1; eauto.
        specialize (is_root_unique _ _ is_root_root H1).
        unfold _IsShort in H.
        rewrite  Z.ltb_lt in H.
        rewrite <- en, <- is_root_unique in H; lia.
      }
      (* tall.2 *)
      { eapply (Subset_Num_Card_iff_eq (A := Node)); eauto.
        eapply Card_trans with Node _.
        apply Trivial_subset_Num_Card.
        apply Hcard_N.
        intros _ n _. 
        split;[intros _| split].
        generalize (WF_par (Par_part_compat pg) span_tree  n).
        intro wf;induction wf.
        
        case_equiv (Par (g x)) ;[intro c|].
        
        intro hx. specialize (parent_is_peer _ _ hx).
        destruct parent_is_peer as  [n hcn].
        assert (hn: is_tree_parent_of (Par_part g) n x). 
        unfold is_tree_parent_of,tree_parent.
        unfold is_channel in hcn.
        rewrite <- hcn.
        replace (peer x c)
          with (option_rect
                  (fun _ : option Channel => option Node)
                  (peer x) None (Some c)).
        2:reflexivity.
        apply option_rect_compat;try reflexivity.
        intros c1 c2 ec;apply peer_compat;try reflexivity.
        assumption.
        symmetry.
        assumption.
        
        specialize (H0 _ hn).
        right with n.
        apply hn.
        tauto.
      
        left.
        apply Z.le_ge. 
        assert (hxroot: x == root).
        apply is_root_unique;try assumption.
        rewrite -> hroot.
        assert (h:alpha  (g x) = alpha (g root)).
        apply alpha_compat,pg; auto.
        rewrite h.
        apply Z.le_refl.
      }
      (* case root is short *)

      (* short.1 *)
      { replace (Datatypes.S card_RDH) with (1+card_RDH)%nat. 
        2:lia.
        eapply (Subset_Num_Card_iff_eq (A := Node)); eauto.
        eapply Disjoint_union_Num_Card.
        apply is_root_compat.
        eapply Par_part_compat,pg.
        intros n1 n2 en.
        eapply RegDomHead_compat. 
        apply en.
        apply Singleton_Num_Card with root.
        apply is_root_root.
        auto.
        apply Hcard_RDH.
        intros x h1x h2x.
        simpl in h2x.
        red in h1x,h2x.
        specialize (is_root_unique _ _ is_root_root h1x). 
        rewrite <- is_root_unique in h2x; lia.
        
        intros n1 n2 en.
        rewrite or_comm.
        unfold kDominator.
        rewrite Bool.orb_true_iff.
        apply Morphisms_Prop.or_iff_morphism.
        unfold RegDomHead.
        rewrite Z.eqb_eq.
        apply PER_morphism.
        apply Equivalence_PER.
        apply eq_equivalence.
        apply alpha_compat,pg; auto.
        reflexivity.
        rewrite Bool.andb_true_iff.
        rewrite <- (IsRoot_is_root_equiv); eauto.
        split.
        intros [_ h];revert h;eapply proj1;apply is_root_compat.
        apply Par_part_compat, pg. 
        assumption.
        intro h.
        assert (h1:is_root (Par_part g) n1).
        revert h;eapply proj1;apply is_root_compat.
        apply Par_part_compat,pg.
        symmetry;assumption.
        split;[|apply h1].
        specialize (is_root_unique _ _ is_root_root h1).
        unfold _IsShort.
        rewrite Z.ltb_lt.
        rewrite <- is_root_unique; auto.
      }
      (* short.2 *)
      { rewrite <- Card_larger_ge_iff.
        apply Card_trans with Node _.
        apply Card_weak. 
        apply Card_sym.
        apply Hcard_N.
        eapply Card_trans
          with (sig (fun x : Node => is_root (Par_part g) x \/
                                     HasTallAncestor x)) _.
        apply Card_sym.
        simpl sym_Card_Prop. 
        eapply Subset_Num_Card.
        apply Union_compat;
          [apply is_root_compat|apply HasTallAncestor_compat]; auto.
        apply Par_part_compat, pg.
        apply Card_weak.
        replace (Datatypes.S card_RN) with (1+card_RN)%nat. 
        2:lia.
        eapply Disjoint_union_Num_Card.
        apply (is_root_compat (Par_part_compat pg)). 
        apply HasTallAncestor_compat.
        apply Singleton_Num_Card with root.
        apply setoid_equiv.
        apply is_root_root.
        intros root' hroot'.
        apply is_root_unique;assumption.
        apply Hcard_RN.
        intros root' hroot'.
        specialize (is_root_unique _ _ hroot' is_root_root).
        rewrite is_root_unique.
        intro h;inversion h.
        lia.
        red in is_root_root,H.
        unfold tree_parent in H.
        destruct (Par_part g root) as [x|];assumption.
      }
    Qed.

    (* main_result *)

    Theorem counting_theorem_assuming_cardinals_exist :  
      (card_N - 1 >= (Datatypes.S (Z.to_nat k)) * (card_DH - 1))%nat.
    Proof.
      destruct split_counting_cases as [[HCH HN]|[HCH HN]].
      rewrite mult_minus_distr_l.
      unfold ge.
      etransitivity;[apply minus_le_compat_l|apply minus_le_compat_r].
      lia.
      rewrite HCH.
      rewrite HN.
      apply simple_counting.
      rewrite HCH.
      unfold ge in *|-*.
      simpl minus.
      rewrite <- minus_n_O.
      etransitivity;[apply simple_counting|].
      lia.
    Qed.

  End with_cardinals.

  (* [main_theorem] : Main counting Property *)

  Theorem counting_theorem_ 
          (card_N: nat)
          (Hcard_N: Num_Card Same Node card_N)
          (card_DH: nat)
          (Hcard_DH: Num_Card Same DomHeads card_DH) :
    (card_N - 1 >= (Datatypes.S (Z.to_nat k)) * (card_DH - 1))%nat.
  Proof.
    assert (tmp1: Proper fequiv (fun n: Node => RegDomHead n)).
    intros n1 n2 en;apply RegDomHead_compat, en.
    assert (tmp2: (forall a: Node,
                     (fun n: Node => RegDomHead n) a \/
                     ~ (fun n: Node => RegDomHead n) a)).
    intro a;destruct (Z.eq_dec (alpha (g a)) k);[left|right];assumption.
    assert (tmp3:
              (forall a1 a2: Node, a1 == a2 \/ ~ a1 == a2)).
    intros a1 a2;destruct (Node_dec a1 a2);[left|right];assumption.

    destruct (Classical_subset_Num_Card
                tmp3 (fun n: Node => RegDomHead n) tmp1 tmp2 all_nodes)
      as [card_RDH Hcard_RDH].
    intros a _; apply all_nodes_prop.
    clear tmp1 tmp2.
    assert (tmp2: (forall a: Node,
                     (fun n: Node => HasTallAncestor n) a \/
                     ~ (fun n: Node => HasTallAncestor n) a)).
    intro n;induction (WF_par (Par_part_compat pg) span_tree n) as [n _ hn].
    destruct (Z_lt_ge_dec (alpha (g n)) k) as [SHORT|TALL].
    2:left;left;apply TALL.
    destruct (father_dec  span_tree n) as [[m hm]|hroot].
    specialize (hn m hm).
    destruct hn as [YES|NO].
    left;right with m;assumption.
    right.
    intro hn;apply NO.
    inversion hn;subst.
    lia.

    assert (h:=is_tree_parent_unique
                 (Par_part_compat pg) _ _  (Equivalence_Reflexive _) _  _ hm H).

    rewrite h.
    assumption.

    right.
    clear hn;intro hn;inversion hn;subst.
    lia.
    generalize hroot H.
    unfold is_tree_parent_of,tree_parent,is_root.
    destruct (Par_part g n) as [x|];simpl;tauto.

    destruct (Classical_subset_Num_Card
                tmp3 (fun n: Node => HasTallAncestor n)
                HasTallAncestor_compat tmp2 all_nodes) as [card_RN Hcard_RN].
    intros a _; apply all_nodes_prop.
    apply counting_theorem_assuming_cardinals_exist with card_RDH card_RN;auto.
  Qed.

  Theorem counting_theorem:
    exists (card_N: nat) (card_DH: nat), 
      Num_Card Same Node card_N /\ 
      Num_Card Same DomHeads card_DH /\ 
      (card_N - 1 >= (Datatypes.S (Z.to_nat k)) * (card_DH - 1))%nat.
  Proof.
    assert (tmp3:
              (forall a1 a2: Node, a1 == a2 \/ ~ a1 == a2)).
    intros a1 a2;destruct (Node_dec a1 a2);[left|right];assumption.

    destruct ( Classical_set_Num_Card tmp3 _ all_nodes_prop) 
      as [n Hn].

    assert (kdom_compat: Proper fequiv 
                                (fun n: Node => (kDominator g n = true))).
    intros n1 n2 en; rewrite (kDominator_compat pg _ _ en); 
    split; trivial.
    assert (kdom_classical: 
              forall n: Node, (fun n: Node => (kDominator g n = true)) n \/ 
                           ~(fun n: Node => (kDominator g n = true)) n).
    intros n1;    
      destruct (Bool.bool_dec (kDominator g n1) true); auto.
    destruct (Classical_subset_Num_Card tmp3 _ kdom_compat 
                                        kdom_classical all_nodes) as [dh Hdh].
    intros a _; apply all_nodes_prop.
    exists n.
    exists dh.
    split; [|split]; try assumption.
    apply counting_theorem_; assumption.
  Qed.

End KClustering_count.

Close Scope Z_scope.
Unset Implicit Arguments.




