(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidClass.
From Coq Require Import SetoidList.
From Coq Require Import ZArith.
From Coq Require Import Bool.
From Coq Require Import Union.
From Coq Require Import Relations.

(** **** Local Imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import OptionUtil.
From Padec Require Import ZUtils.
From Padec Require Import ListUtils.
From Padec Require Import RelationA.
From Padec Require Import FoldRight.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import KClustering_algo.
From Padec Require Import KClustering_specification.
From Padec Require Import Spanning_Tree.
From Padec Require Import P_Q_Termination.

(* end show *)
Open Scope signature_scope.
Open Scope Z_scope.
Set Implicit Arguments.

(** roadmap 

1) show termination for each king of safe steps:
   -  Step_decreasing Alpha_SafeStep alpha_potential all_nodes.
   -  Step_decreasing Pcl_SafeStep pcl_potential all_nodes.
   -  Step_decreasing Head_SafeStep hd_pot all_nodes.

2) use union_lex_wf3

 **)
Section KClustering_DEF_termination.

  (** ** Channel and network *)  
  Context {Chans: Channels} {Net: Symmetric_Network}
          {KCP: KClustering_parameters} {KCT: KC_algo_tools}.
  
  Variable (root: Node).
  Notation Env := (Env (Algo := KC_algo)).

  (** ** Safe configurations and safe steps:
      configurations with spanning tree assumed
   *)
  Section SafeEnv.
    
    Definition Pspanning_tree (g: Env): Prop :=
      Proper pequiv g /\ spanning_tree root (Par_part g)
      /\ Symmetric_Network_assumption
           Net (fun n => KCpeers (g n)) (fun n => KCreverse (g n)).

    Global Instance Pspanning_tree_compat: Proper fequiv Pspanning_tree.
    Proof.
      intros env1 env2 E.
      split; intros [P [sp S]]; split; auto; try proper_; split.
      eapply spanning_tree_compat; eauto.
      symmetry; apply Par_part_compat, E.
      eapply Symmetric_Network_assumption_compat_equiv; eauto.
      symmetry; intros n1 n2 en; apply KCpeers_compat, E; auto.
      intros n c i; simpl.
      apply KCreverse_compat; auto.
      symmetry; apply E; auto.
      eapply spanning_tree_compat; eauto.
      apply Par_part_compat, E.
      eapply Symmetric_Network_assumption_compat_equiv; eauto.
      intros n1 n2 en; apply KCpeers_compat, E; auto.
      intros n c i; simpl.
      apply KCreverse_compat; auto.
      apply E; auto.
    Qed.
    
    (** safe configurations are stable *)

    Lemma KCpeers_Stable:
      forall g g', Step g' g ->
                   (SetoidClass.equiv ==>
                                      equivlistA (A:=Channel) SetoidClass.equiv)
                     (*equiv ==> equivlistA equiv*) (fun n => KCpeers (g' n))
                                                    (fun n => KCpeers (g n)).
    Proof.
      intros g g' step n1 n2 en.
      apply KCpeers_compat.
      apply (Stable_Variable_step KC_Stable); auto.
    Qed.

    Lemma KCreverse_Stable:
      forall g g',
        Step g' g ->
        forall n1 n2 c1 c2,
          n1 == n2 -> c1 == c2 ->
          InA (SetoidClass.equiv (A:=Channel)) c1 (KCpeers (g n1))
          -> KCreverse (g' n2) c2 == KCreverse (g n1) c1.
    Proof.
      intros g g' step n1 n2 c1 c2 en ec i.
      transitivity (KCreverse (g n1) c2).
      apply KCreverse_compat.
      rewrite (KCROpart_compat ((Step_proper_r step) _ _ en)).
      apply (Stable_Variable_step KC_Stable); auto.
      rewrite <- ec.
      symmetry in en; rewrite (KCpeers_Stable step _ _ en); auto.
      apply KCreverse_local_compat.
      symmetry; auto.
    Qed.

    Lemma spanning_tree_stable:
      forall (g: Env) (g': Env),
        Pspanning_tree g -> Step g' g -> Pspanning_tree g'.
    Proof.
      intros g g' [pg [span sym]] trans; split.
      apply Step_Proper_l with g; auto.
      split.
      eapply spanning_tree_compat; eauto.
      unfold Par_part; intros n1 n2 en; apply Par_compat.
      apply (Stable_Variable_step KC_Stable); auto.
      eapply Symmetric_Network_assumption_compat_equiv.
      apply KCpeers_Stable.
      apply trans.
      intros n c i; simpl.
      apply KCreverse_Stable.
      apply trans.
      reflexivity.
      reflexivity.
      assert (K := KCpeers_Stable trans _ _ (Equivalence_Reflexive n));
        simpl in K.  
      rewrite <- K; auto.
      auto.
    Qed.    
    
  End SafeEnv.

  Notation getEnv := proj1_sig.

  (** ** Alpha Steps: a safe step in which at least one node has
   executed an alpha action. Note that other nodes may also have
   executed other kind of actions (parCl and Head actions).  *)
  Section Alpha_SafeSteps.

    Definition alpha_enabled_bb (g: Env) (n: Node): bool :=
      alpha_enabled_b (g n) (local_env g n).

    Global Instance alpha_enabled_bb_compat: Proper fequiv alpha_enabled_bb.
    Proof.
      intros e1 e2 eqe n1 n2 eqn; unfold alpha_enabled_bb.
      apply alpha_enabled_b_compat.
      apply eqe; auto.
      apply (local_env_compat eqe _ _ eqn).
    Qed.

    (** Steps restricted to the ones such that:
           at least one node in the step executes an alpha action *)
    Definition QTrans_alpha (sg' sg: sig Pspanning_tree): Prop := 
      exists (p: Node), alpha_enabled_bb (getEnv sg) p = true /\ 
                        has_moved_b (getEnv sg') (getEnv sg) p = true.

    (** Boolean version of QTrans_alpha *)
    Definition alpha_step_b (g' g: Env): bool :=
      existsb 
        (fun (n: Node) => 
           andb (alpha_enabled_bb g n) (has_moved_b g' g n))
        all_nodes. 

    (** Equivalence between both definitions *)
    Lemma QTrans_alpha_alt: 
      forall sg' sg, QTrans_alpha sg' sg <-> 
                     alpha_step_b (getEnv sg') (getEnv sg) = true.
    Proof.
      intros; unfold QTrans_alpha, alpha_step_b; rewrite existsb_exists_InA.
      { split; intros [x H]; exists x. 
        split; [apply all_nodes_prop| rewrite andb_true_iff; split;
                                      destruct H; assumption].
        destruct H as [_ AM]; rewrite andb_true_iff in AM; split; apply AM. 
      }
      { intros n1 n2 en; erewrite alpha_enabled_bb_compat; eauto.
        erewrite has_moved_b_compat; eauto;
          destruct sg as [g [pg s]]; destruct sg' as [g' [pg' s']]; auto.
        destruct sg as [g [pg s]]; auto.
      }
    Qed.

    (** Alpha Safe Steps *)
    Definition Alpha_Safe :=
      (fun sg' sg: sig Pspanning_tree =>
         alpha_step_b (getEnv sg') (getEnv sg) = true).
    Definition Alpha_SafeStep := safeQStep Alpha_Safe.

  End Alpha_SafeSteps.

  (** ** Parent Cluster Steps: No node has executed an alpha action,
   at least one has executed a parCl action. Note that other nodes may
   also have executed a head action.  *)
  Section Pcl_SafeSteps.
    
    Definition pcl_enabled_bb (g: Env) (n: Node): bool :=
      parCl_enabled_b (g n) (local_env g n).

    Global Instance pcl_enabled_bb_compat: Proper fequiv pcl_enabled_bb.
    Proof.
      intros e1 e2 eqe n1 n2 eqn; unfold pcl_enabled_bb.
      apply parCl_enabled_b_compat. 
      apply eqe; auto.
      apply (local_env_compat eqe _ _ eqn).
    Qed.

    (** Steps restricted to the ones such that:
     - not an alpha safe step (i.e., no node executes an alpha action)
     - at least node executes a parCl action
     *)
    Definition QTrans_pcl (sg' sg: sig Pspanning_tree): Prop :=
      ~ QTrans_alpha sg' sg /\ 
      exists (p: Node), pcl_enabled_bb (getEnv sg) p = true /\ 
                        has_moved_b (getEnv sg') (getEnv sg) p = true. 

    (** Boolean version of QTrans_pcl *)
    Definition pcl_step_b (g' g: Env): bool :=
      andb
        (forallb
           (fun n: Node => orb (negb (alpha_enabled_bb g n)) 
                               (negb (has_moved_b g' g n)))
           all_nodes)
        (existsb 
           (fun (n: Node) => 
              andb (pcl_enabled_bb g n) (has_moved_b g' g n))
           all_nodes).

    (** Equivalence between both definitions *)
    Lemma QTrans_pcl_alt:
      forall sg' sg, QTrans_pcl sg' sg <->
                     pcl_step_b (getEnv sg') (getEnv sg) = true.
    Proof.
      intros; unfold QTrans_pcl, pcl_step_b; rewrite QTrans_alpha_alt; 
        rewrite andb_true_iff; rewrite existsb_exists_InA.
      {
        unfold alpha_step_b; split; intros [A [p [P H]]].
        - erewrite existsb_forallb in A.
          rewrite not_true_iff_false, negb_false_iff in A.
          split; [apply A|].
          + exists p; split; [apply all_nodes_prop | rewrite andb_true_iff].
            split; assumption.
          + intros a; simpl.
            rewrite negb_orb, negb_involutive; rewrite negb_involutive;
              reflexivity.
        - split.
          + erewrite existsb_forallb.
            rewrite not_true_iff_false, negb_false_iff;apply A.
            intros a; simpl.
            rewrite negb_orb, negb_involutive; rewrite negb_involutive;
              reflexivity.
          + exists p; rewrite andb_true_iff in H; split; apply H.
      }
      { intros n1 n2 en; erewrite pcl_enabled_bb_compat; eauto.
        erewrite has_moved_b_compat; eauto;
          destruct sg as [g [pg s]]; destruct sg' as [g' [pg' s']]; auto.
        destruct sg as [g [pg s]]; auto.
      }
    Qed.
    
    (** Pcl Safe Step *)
    Definition Pcl_Safe :=
      (fun sg' sg: sig Pspanning_tree =>
         pcl_step_b (getEnv sg') (getEnv sg) = true).
    Definition Pcl_SafeStep := safeQStep Pcl_Safe.

  End Pcl_SafeSteps.

  (** ** Head Cluster Steps: No node has executed either an alpha
   action or an parCl action. At least one has executed a head action.
   *)
  Section Head_SafeSteps.

    Definition hd_enabled_bb (g: Env) (n: Node): bool :=
      (hd_enabled_b (g n) (local_env g n)).

    Global Instance hd_enabled_bb_compat: Proper fequiv hd_enabled_bb.
    Proof.
      intros e1 e2 eqe n1 n2 eqn; unfold hd_enabled_bb.
      apply hd_enabled_b_compat.
      apply eqe; auto.
      apply (local_env_compat eqe _ _ eqn).
    Qed.

    (** Steps that are neither alpha safe steps nor Plc safe steps.
     i.e., no node executes an alpha action nor a pcl action.
     *)
    Definition QTrans_hd (sg' sg: sig Pspanning_tree): Prop :=
      ~ QTrans_alpha sg' sg /\ ~ QTrans_pcl sg' sg.

    (** Boolean version of QTrans_hd *)
    Definition hd_step_b (g' g: Env): bool :=
      (forallb
         (fun n: Node => 
            orb (negb (orb (alpha_enabled_bb g n)
                           (pcl_enabled_bb g n)))
                (negb (has_moved_b g' g n)))
         all_nodes).

    (** Equivalence between both definitions *)
    Lemma QTrans_hd_alt:
      forall sg' sg, QTrans_hd sg' sg <->
                     hd_step_b (getEnv sg') (getEnv sg) = true.
    Proof.
      intros sg' sg; 
        unfold QTrans_hd.
      rewrite QTrans_pcl_alt, QTrans_alpha_alt.

      repeat rewrite not_true_iff_false.
      unfold alpha_step_b, hd_step_b, pcl_step_b.

      split.
      intros [H1 H2]. 
      rewrite andb_false_iff in H2.
      rewrite existsb_forallb
        with (notf := (fun n : Node =>
                         negb (alpha_enabled_bb (getEnv sg) n) ||
                         negb (has_moved_b (getEnv sg') (getEnv sg) n))), 
                                   negb_false_iff in H1.
      2: intros a; apply negb_sym; symmetry; apply negb_andb.
      rewrite H1 in H2.
      destruct H2 as [H2|H2].
      contradict H2; apply diff_true_false.
      rewrite existsb_forallb
        with (notf := (fun n : Node =>
                         negb (pcl_enabled_bb (getEnv sg) n) ||
                         negb (has_moved_b (getEnv sg') (getEnv sg) n))),
                                   negb_false_iff in H2
      .      2: intros a; apply negb_sym; symmetry; apply negb_andb.
      rewrite forallb_forall in *.
      intros x inx; specialize (H1 _ inx); specialize (H2 _ inx).     
      rewrite orb_true_iff in H1, H2.
      destruct H1 as [H1|H1]; rewrite negb_true_iff in H1;
        rewrite H1; simpl.
      2: intuition.
      destruct H2 as [H2|H2]; rewrite H2; intuition.

      intros H.
      rewrite forallb_forall in H.
      split.
      rewrite existsb_forallb
        with (notf := (fun n : Node =>
                         negb (alpha_enabled_bb (getEnv sg) n) ||
                         negb (has_moved_b (getEnv sg') (getEnv sg) n))) .
      2: intros a; apply negb_sym; symmetry; apply negb_andb.
      rewrite negb_false_iff.
      rewrite forallb_forall; intros x inx.
      specialize (H x inx).
      rewrite orb_true_iff in H.
      destruct H as [H|H]; [|rewrite H; intuition].
      rewrite negb_orb, andb_true_iff in H.
      destruct H as [H _]; rewrite H; intuition.
      rewrite andb_false_iff.
      right. rewrite <- negb_true_iff, existsb_forallb
               with (notf := (fun n : Node =>
                                negb(pcl_enabled_bb
                                       (getEnv sg) n) ||
                                negb(has_moved_b (getEnv sg') (getEnv sg) n))).
      2: intros a; apply negb_sym; symmetry; apply negb_andb.
      rewrite negb_involutive, forallb_forall.
      intros x inx; specialize (H x inx).
      rewrite orb_true_iff, negb_orb, andb_true_iff in H.
      destruct H as [[_ H] | H]; rewrite H; intuition.
    Qed.

    (** Head safe step *)
    Definition Head_Safe :=
      (fun sg' sg: sig Pspanning_tree =>
         hd_step_b (getEnv sg') (getEnv sg) = true).
    Definition Head_SafeStep := safeQStep Head_Safe.

  End Head_SafeSteps.

  (** ** Safe Steps as union of Alpha, Cluster Parent, Cluster Head
   Steps: the relation SafeStep is exactly the union of relations
   Alpha_SafeStep, Pcl_SafeStep and Head_SafeStep.  *)
  Section Split_SafeStep.

    Lemma Trans_union:
      relation_equivalence 
        (safeStep (safe := Pspanning_tree))
        (safeQStep (safe := Pspanning_tree)
                   (union Alpha_Safe
                          (union Pcl_Safe Head_Safe))).
    Proof.
      intros sx sy. 
      unfold safeQStep, union, Head_Safe, Pcl_Safe, Alpha_Safe;
        simpl.
      split; [|intuition].
      intros H; split; auto.
      rewrite <- QTrans_alpha_alt, <- QTrans_pcl_alt, 
      <- QTrans_hd_alt.
      unfold QTrans_hd.
      case_eq (alpha_step_b (getEnv sx) (getEnv sy)); intros A.
      rewrite <- QTrans_alpha_alt in A; left; auto.
      case_eq (pcl_step_b (getEnv sx) (getEnv sy)); intros P.
      rewrite <- QTrans_pcl_alt in P; right; left; auto. 
      rewrite <- not_true_iff_false in A, P.
      rewrite <- QTrans_alpha_alt in A; 
        rewrite <- QTrans_pcl_alt in P; right; right; split; 
          [assumption |assumption].
    Qed.

  End Split_SafeStep.

End KClustering_DEF_termination.

Close Scope Z_scope.
Close Scope signature_scope.
Unset Implicit Arguments.
