(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import ZArith.
From Coq Require Import Lia.
From Coq Require Import Bool.
From Coq Require Import Inclusion.
From Coq Require Import RelationPairs.

(** **** Local Imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import OptionUtil.
From Padec Require Import ZUtils.
From Padec Require Import ListUtils.
From Padec Require Import RelationA.
From Padec Require Import FoldRight.
From Padec Require Import All_In.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Spanning_Tree. 
From Padec Require Import KClustering_specification.
From Padec Require Import KClustering_algo.
From Padec Require Import KClustering_def_termination.
From Padec Require Import KClustering_alpha_termination.
From Padec Require Import KClustering_pcl_termination.
From Padec Require Import P_Q_Termination.

(* end show *)
Open Scope Z_scope.
Open Scope signature_scope.
Set Implicit Arguments.

(** Termination for KClustering Algorithm: Head Steps Only

 This part shows termination of the algorithm when considering Head
 safe steps only.  Actually, it shows safe inclusion required by the
 rest of the proof:

         [hd_safe_incl := (forall g g', Head_SafeStep g' g -> ltM
         (Pot_hd g') (Pot_hd g)]

 *)
Section KClustering_Hd_Termination.

  (** ** Channel and network *)  
  Context {Chans: Channels} {Net: Symmetric_Network}
          {KCP: KClustering_parameters} {KCT: KC_algo_tools}.

  Variable (root: Node).
  Existing Instance KC_Stable.
  
  Notation Env := (Env (Algo := KC_algo)).

  Let k_pos := k_pos.
  
  (** ** Roadmap 

    - define the well founded relation which corresponds
      to the chain of pcl pointers when alpha_disabled and pcl_disabled
      (show that it is actually well founded)

    - define dist_hd (structural induction on the above well founded
      order)

    - define potential ( = N+1 - dist_hd )

    - prove criteria *)

  (** ** Chain Of Pcl Pointers Defines the well founded relation which
    corresponds to the chain of pcl pointers when alpha_disabled and
    pcl_disabled (show that it is actually well founded) *)
  Section Chain_Of_Pcl_Pointers.

    Definition get_pcl (g: Env) (n: Node): option Node :=
      match parCl (g n) with 
        None => None 
      | Some p =>
        match (peer n p) with None => None
                         | Some pn => if andb (negb (alpha_enabled_bb g n))
                                              (negb (pcl_enabled_bb g n)) 
                                      then (peer n p)
                                      else None
        end
      end.

    Instance get_pcl_compat: Proper fequiv get_pcl.
    Proof.
      intros env1 env2 eqe n1 n2 eqn.
      unfold get_pcl.
      generalize (parCl_compat (eqe _ _ eqn));
        case_eq (parCl (env1 n1)); 
        case_eq (parCl (env2 n2)).
      intros c2 eqc2 c1 eqc1 eqcc.
      simpl in eqcc.
      generalize (peer_compat _ _ eqn _ _ eqcc);
        destruct (peer n1 c1); destruct (peer n2 c2); intros.
      simpl in H.
      rewrite (alpha_enabled_bb_compat eqe _ _ eqn).
      rewrite (pcl_enabled_bb_compat eqe _ _ eqn).
      destruct (negb (alpha_enabled_bb env2 n2) &&
                     negb (pcl_enabled_bb env2 n2)).
      simpl; assumption.
      reflexivity.
      inversion H.
      inversion H.
      reflexivity.
      intros a b c FF; inversion FF.
      intros a b c FF; inversion FF.
      intros; reflexivity.
    Qed.

    (** ** Well founded order *)
    Definition pcl_chain (g: Env): relation Node :=
      fun pn n => Some pn == get_pcl g n.

    Instance pcl_chain_compat: Proper fequiv pcl_chain.
    Proof.
      intros g1 g2 eg n1 n2 en n'1 n'2 en'.
      unfold pcl_chain.
      rewrite en.
      rewrite (get_pcl_compat eg _ _ en').
      reflexivity.
    Qed.

    Lemma pcl_chain_dec:
      forall g n1 n2, {pcl_chain g n1 n2} + {~pcl_chain g n1 n2}.
    Proof.
      intros env n1 n2.
      unfold pcl_chain, get_pcl.
      destruct (parCl (env n2)).
      destruct (peer n2 c); simpl.
      destruct ((negb (alpha_enabled_bb env n2) &&
                      negb (pcl_enabled_bb env n2))).
      apply Node_dec.
      auto.
      auto.
      auto.
    Qed.

    Lemma pcl_unique:
      forall env1 env2 n1 n2 p1 p2,
        env1 =~= env2 -> n1 == n2 ->
        pcl_chain env1 p1 n1 -> pcl_chain env2 p2 n2 -> p1 == p2.
    Proof.
      intros env1 env2 n1 n2 p1 p2 HE HN HP1 HP2.
      unfold pcl_chain in HP1, HP2.
      rewrite <- (get_pcl_compat HE _ _ HN) in HP2.
      rewrite <- HP2 in HP1.
      apply HP1.
    Qed.

    Lemma ap_disabled:
      forall n pn env, pcl_chain env pn n 
                       -> alpha_enabled_bb env n = false /\
                          pcl_enabled_bb env n = false.
    Proof.
      intros n pn env hpn.
      unfold pcl_chain, get_pcl in hpn.
      case_eq (parCl (env n)); 
        [intros c pcln| intros pcln]; rewrite pcln in hpn;
          [|inversion hpn].
      case_eq (peer n c); 
        [intros ppn hppn| intros hppn]; rewrite hppn in hpn;
          [|inversion hpn].
      case_eq (negb (alpha_enabled_bb env n) && negb (pcl_enabled_bb env n)); 
        intros ap; rewrite ap in hpn; [|inversion hpn].
      rewrite andb_true_iff in ap.      
      destruct ap as [ap1 ap2].
      rewrite negb_true_iff in ap1, ap2.
      repeat split; assumption.
    Qed.

    Lemma _noclusterhead:
      forall n pn env, pcl_chain env pn n 
                       -> IsClusterHead (env n) = false.
    Proof.
      (* not cluster head *)
      intros n pn env ch.
      unfold KClustering_algo.IsClusterHead,_IsShort.
      
      assert (ap := ap_disabled n pn env ch).
      destruct ap as [pp ap].
      unfold pcl_chain, get_pcl in ch.
      
      rewrite pp, ap in ch. simpl in ch.
      case_eq (parCl (env n)); [intros c pcl| intros pcl]; 
        rewrite pcl in ch; [| elim ch].
      case_eq (peer n c); [intros cc pcc| intros pcc]; 
        rewrite pcc in ch; [| elim ch].

      unfold pcl_enabled_bb, parCl_enabled_b in ap.
      destruct (eqoptionA_dec Channel_dec (parCl (env n))
                              (NewparCl (env n) (local_env env n))) 
        as [eqpcl|]; [| contradict ap; apply diff_true_false].
      unfold NewparCl in eqpcl.
      rewrite pcl in eqpcl.

      rewrite orb_false_iff.
      case_eq (alpha (env n) <? k); intro A; [|auto].
      rewrite A in eqpcl.
      split.
      rewrite Z.ltb_lt in A.
      rewrite Z.eqb_neq.
      intros GG.
      rewrite GG in A.
      contradict A.
      apply Z.lt_irrefl.
      rewrite andb_false_iff.
      right.
      rewrite <- eqpcl.
      unfold dec2b.
      case (eqoptionA_dec Channel_dec (Some c) None );
        simpl; intuition.
      rewrite A in eqpcl.
      
      case_eq (alpha (env n) =? k); intro AA;
        [ rewrite AA in eqpcl; inversion eqpcl |].

      split;[reflexivity|].
      rewrite andb_false_iff.
      left.
      reflexivity.
    Qed.

    Lemma noclusterhead_:
      forall n env, IsClusterHead (env n) = true
                    -> forall pn, ~pcl_chain env pn n.
    Proof.
      (* not cluster head *)
      intros n env CH pn.
      intros FF.
      specialize (_noclusterhead n pn env FF); intros.
      rewrite H in CH.
      contradict CH. apply diff_false_true.
    Qed.

    Lemma MinIDMinATall_IsChild:
      forall n (env: Env) (c: Channel),
        Symmetric_Network_assumption
          Net (fun n => KCpeers (env n)) (fun n => KCreverse (env n))
        -> Proper pequiv env ->
        Some c == MinIDMinATall (env n) (local_env env n) ->
        InA equiv c (Children (Par_part env) n).
    Proof.
      intros n g c sym pg eqmin.
      unfold MinIDMinATall in eqmin.
      revert eqmin;
        case_eq (filter (isMinATalls (g n) (local_env g n))
                        (Local_Children
                           (KCpeers (g n))
                           (lPar (local_env g n)) (KCreverse (g n)))).
      intros E eqmin; elim eqmin.
      intros cc l E eqmin.
      assert (FR := fold_right_min_InA
                      (A_dec := Channel_dec) _ ltC_dec c cc l eqmin).
      rewrite <- E in FR.
      apply filter_InA in FR.
      2:     apply (isMinATalls_compat ).
      2: apply pg; reflexivity.
      2: apply (local_env_compat pg _ _ (Equivalence_Reflexive n)). 
      destruct FR as [FR _].
      unfold Local_Children in FR.
      rewrite filter_InA_locality in FR.
      unfold Children.      
      rewrite filter_InA.
      assert (dum: InA equiv c (peers n)) by apply sym, FR.
      split; auto.
      unfold IsChild.
      destruct FR as [_ FR]; rewrite <- FR.
      apply Local_IsChild_locality; auto.
      unfold lPar, local_env, Par_part; destruct (peer n c); simpl; auto.
      reflexivity.
      symmetry; apply (proj2 sym _ _ dum).
      apply IsChild_compat; auto; apply Par_part_compat; auto.
      intros x y xy i.
      apply Local_IsChild_compat; auto.
      apply KCpeers_compat, pg; auto.
      apply lPar_compat, (local_env_compat pg _ _ (Equivalence_Reflexive n)).
      apply KCreverse_local_compat.
    Qed.

    Lemma same_par_pcl:
      forall n pn (env: Env),
        Proper pequiv env ->
        Symmetric_Network_assumption
          Net (fun n => KCpeers (env n)) (fun n => KCreverse (env n)) ->
        pcl_chain env pn n
        -> 
        ((alpha (env n) <? k ) = true /\ 
         (is_tree_parent_of (Par_part env) pn n))
        \/
        ((alpha (env n) <? k) = false /\
         (alpha (env n) =? k) = false /\
         (is_tree_parent_of (Par_part env) n pn)).
    Proof.
      intros n pn env env_proper sym pcl.
      unfold Par_part.
      specialize (ap_disabled n pn env pcl) ; intros AP.
      destruct AP as [ A P].
      unfold pcl_chain, get_pcl in pcl.
      rewrite A, P in pcl; simpl in pcl.
      revert pcl; case_eq (parCl (env n));  [intros c|]; 
        intros pcl ppcl; [| elim ppcl].
      revert ppcl; case_eq (peer n c); [ intros pnn |];
        intros pe eqppn ; [|elim eqppn].

      unfold pcl_enabled_bb, parCl_enabled_b in P.
      destruct (eqoptionA_dec Channel_dec (parCl (env n))
                              (NewparCl  (env n) (local_env env n)))
        as [ eqpcl | _ ]; 
        [ clear P | contradict P; apply diff_true_false].
      unfold NewparCl in eqpcl.
      rewrite pcl in eqpcl.
      case_eq (alpha (env n) <? k); intro HA;
        rewrite HA in eqpcl.
      left. repeat split.
      unfold is_tree_parent_of, tree_parent; simpl.
      unfold option_rect; revert eqpcl;
        destruct (Par (env n)); [| simpl; tauto].
      intros X; simpl in X.
      assert (XX := peer_compat _ _ (Equivalence_Reflexive n) _ _ X).
      rewrite pe in XX.
      revert XX; destruct (peer n c0); [|simpl; tauto].
      simpl; intros.
      rewrite <- XX; assumption. 

      case_eq (alpha (env n) =? k); intro HAA;
        rewrite HAA in eqpcl; [inversion eqpcl |].
      assert (cc :=  MinIDMinATall_IsChild n c sym env_proper eqpcl).
      unfold Children in cc.
      rewrite filter_InA in cc.
      2: apply IsChild_compat;
        [apply Par_part_compat; apply env_proper| reflexivity].
      right; split; [reflexivity|split; [reflexivity|]].
      unfold IsChild in cc.
      destruct cc as [cc' cc].
      unfold Local_IsChild, Par_part in cc.
      rewrite pe in cc.
      unfold dec2b, option_rect, Par_part in cc.
      unfold Par_part, is_tree_parent_of, tree_parent, option_rect.
      revert cc;
        generalize (Par_compat (KCROpart_compat (env_proper _ _ eqppn)));
        case_eq (Par (env pn)); case_eq (Par (env pnn)).
      2: intros _ x _ F; elim F. 
      2: intros x _ _ F; elim F. 
      2: congruence. 
      intros cnn Hcnn cn Hcn eqcn.
      case (Channel_dec cnn (reverse n c)).
      2: intros _ F;  contradict F; apply diff_false_true. 
      intros R _.
      simpl in eqcn; rewrite <- eqcn in R.
      rewrite R.
      assert (X := reverse_spec' n c pn).
      unfold is_channel in X; symmetry; apply X.
      rewrite pe; simpl; symmetry; apply eqppn.
    Qed.

    Lemma preserved_ge_alpha:
      forall env n pn, Proper pequiv env ->
                       pcl_chain env pn n -> (alpha (env n) >=? k) = true
                       -> (alpha (env pn) >=? k) = true.
    Proof.
      intros env n pn env_proper hpn HA.
      assert (diffk: (alpha (env n)) =? k = false).
      specialize (_noclusterhead n pn env hpn).
      unfold KClustering_algo.IsClusterHead; intros ch.
      rewrite orb_false_iff in ch; apply ch.
      assert (HAA: (alpha (env n)) <? k = false).
      rewrite Z.geb_le in HA.
      rewrite Z.eqb_neq in diffk.
      rewrite Z.ltb_nlt.
      lia.    

      assert (same := ap_disabled _ _ _ hpn).
      destruct same as [ A P].
      unfold alpha_enabled_bb, alpha_enabled_b, NewAlpha in A.
      revert A;
        case (Z.eq_dec (alpha (env n))
                       (if Z_le_gt_dec
                             (MaxAShort (env n) (local_env env n) +
                              MinATall (env n) (local_env env n) + 2)
                             (2 * k)
                        then MinATall (env n) (local_env env n) + 1
                        else MaxAShort (env n) (local_env env n) + 1));
        [intros A _ | intros _ FF; contradict FF; apply diff_true_false].
      revert A; case (Z_le_gt_dec
                        (MaxAShort (env n) (local_env env n) +
                         MinATall (env n) (local_env env n) + 2) 
                        (2 * k)); intros cond A.

      2:{
        assert (range: -1 <= MaxAShort (env n) (local_env env n) <= k-1).
        unfold MaxAShort.
        destruct (fold_right_Zmax_dec
                    (-1) (ShortAlphas (env n) (local_env env n)))
          as [eqa | ina];
          unfold Z.eq in *.
        rewrite eqa; lia.
        split.
        rewrite <- Z.ge_le_iff.
        apply fold_right_Zmax_eqA_leA.
        reflexivity.
        unfold ShortAlphas in ina.
        rewrite filter_InA in ina.
        2: intros x y exy; rewrite exy; reflexivity.
        destruct ina as [ina2 tall].
        unfold _IsShort in tall.
        rewrite <- Zlt_is_lt_bool in tall.
        unfold ShortAlphas, _IsShort.
        clear ina2 cond A.
        lia.
        
        assert (x := Zplus_eq_compat _ _ _ _ A (Equivalence_Reflexive (-1))).
        assert (AA: alpha (env n) + - 1 <= k -1).
        rewrite A.
        lia.
        clear x range.
        rewrite Z.geb_le in HA.
        rewrite Z.eqb_neq in diffk.
        contradict diffk.
        lia.
      }
      
      clear cond.
      
      unfold pcl_enabled_bb, parCl_enabled_b, NewparCl in P.
      rewrite HAA, diffk in P.
      revert P; case (eqoptionA_dec Channel_dec (parCl (env n))
                                    (MinIDMinATall (env n) (local_env env n)));
      [ intros eqpcl _ | intros _ FF; contradict FF; apply diff_true_false].
      unfold pcl_chain, get_pcl in hpn.
      revert hpn; case_eq (parCl (env n)); [|intros _ FF; inversion FF]. 
      intros c.
      case_eq (peer n c); [|intros _ _ FF; inversion FF]. 
      intros ppn.
      destruct (negb (alpha_enabled_bb env n) && negb (pcl_enabled_bb env n)); 
        [|intros _ _ FF; inversion FF].
      intros pe pc hpn; simpl in hpn.
      rewrite pc in eqpcl.

      assert (youpi: alpha (env pn) = MinATall (env n) (local_env env n)).
      unfold MinIDMinATall, isMinATalls in eqpcl.
      revert eqpcl;
        case_eq (filter
                   (fun c =>
                      option_rect (fun _ : option KCState => bool)
                                  (fun qs => alpha qs =?
                                             MinATall (env n)
                                                      (local_env env n)) false
                                  (local_env env n c))
                   (Local_Children (KCpeers (env n))
                                   (lPar (local_env env n))
                                   (KCreverse (env n))));              
        [intros eqpcl FF; inversion FF|].
      intros cf lf F fr.
      simpl in fr.
      unfold local_env in F.
      assert (inac: InA equiv c (cf::lf)).
      { eapply (fold_right_min_InA (A_dec := Channel_dec) _ ltC_dec); auto. 
        apply fr. } 
      rewrite <- F in inac.
      rewrite filter_InA in inac.
      destruct inac as [ _ inac].
      rewrite pe in inac; simpl in inac.
      unfold local_env.
      rewrite Z.eqb_eq in inac.
      rewrite (alpha_compat (env_proper _ _ hpn)).
      auto.
      
      clear F fr inac.
      intros c1 c2 eqc.
      generalize  (peer_compat _ _ (Equivalence_Reflexive n) _ _ eqc);
        destruct (peer n c1); destruct (peer n c2); simpl; intros.
      rewrite (alpha_compat (env_proper  _ _ H)); reflexivity.
      elim H. elim H. reflexivity.

      assert (range: k <= MinATall (env n) (local_env env n) <= 2*k+1).
      unfold MinATall, TallAlphas.
      destruct (fold_right_Zmin_dec (2 * k + 1) 
                                    (filter _IsTall
                                            (ChildrenAlphas
                                               (env n) (local_env env n))))
        as [eqa | ina]; unfold Z.eq in *.
      rewrite eqa; lia.
      split.
      rewrite filter_InA in ina.
      2: intros x y exy; rewrite exy; reflexivity.
      destruct ina as [ina2 tall].
      unfold _IsTall in tall.
      rewrite Z.geb_le in tall.
      assumption.
      apply fold_right_Zmin_eqA_leA.
      reflexivity.

      rewrite Z.geb_le.
      rewrite youpi.
      destruct range. apply H. 
    Qed.

    Lemma Acc_Tall:
      forall (n: Node) (env: Env),
        Proper pequiv env ->
        Symmetric_Network_assumption
          Net (fun n => KCpeers (env n)) (fun n => KCreverse (env n)) ->

        (alpha (env n) >=? k) = true ->
        Acc (flip (is_tree_parent_of (Par_part env))) n ->
        Acc (pcl_chain env) n.
    Proof.
      intros n env env_proper sym HA acc_ct.
      induction acc_ct  as [ n Hct IH].
      apply Acc_intro; intros pn hpn.
      apply IH. 
      unfold flip.

      remember  (get_pcl env n) as spn. 
      case_eq spn. intros ppn eqpn.
      assert (hppn: pcl_chain env ppn n).
      unfold pcl_chain. rewrite <- eqpn, Heqspn; reflexivity.
      assert (ppcl := same_par_pcl n ppn env_proper sym hppn).
      rewrite Z.geb_le in HA.
      assert (tmp: (alpha (env n) <? k) = false ).
      rewrite Z.ltb_ge; assumption.
      destruct ppcl as [ FF | same];
        [rewrite tmp in FF;
         destruct FF as [ FF _];
         contradict FF; apply diff_false_true|].
      destruct same as [_ [ _ same]].
      assert (noy := pcl_unique
                       _ _ _ _ env_proper (Equivalence_Reflexive n) hppn hpn).
      eapply is_tree_parent_of_compat. 
      4: apply same.
      apply Par_part_compat; apply env_proper; reflexivity.
      reflexivity.
      symmetry; assumption.

      intros nson.      
      unfold pcl_chain in hpn.
      rewrite <- Heqspn, nson in hpn.
      inversion hpn.
      apply (preserved_ge_alpha _ _ env_proper hpn HA). 
    Qed.

    Lemma Acc_Short:
      forall (n: Node) (env: Env),
        Proper pequiv env ->
        Symmetric_Network_assumption
          Net (fun n => KCpeers (env n)) (fun n => KCreverse (env n)) ->
        (alpha (env n) <=? k) = true ->
        Acc (is_tree_parent_of (Par_part env)) n ->
        well_founded (flip(is_tree_parent_of (Par_part env))) ->
        Acc (pcl_chain env) n.
    Proof.
      intros n env env_proper sym HA acc_cc acc_ct.
      induction acc_cc  as [ n Hct IH].
      apply Acc_intro; intros pn hpn.
      case_eq (alpha (env pn) >=? k); intros HAA.
      apply (Acc_Tall env_proper sym HAA (acc_ct pn)).     
      apply IH. 

      remember  (get_pcl env n) as spn. 
      case_eq spn. intros ppn eqpn.
      assert (hppn: pcl_chain env ppn n).
      unfold pcl_chain. rewrite <- eqpn, Heqspn; reflexivity.
      assert (ppcl := same_par_pcl n ppn env_proper sym hppn).
      destruct ppcl as [ same | FF]. 
      2:lia.
      destruct same as [_ same].
      assert (noy := pcl_unique
                       _ _ _ _ env_proper (Equivalence_Reflexive n) hpn hppn).
      eapply is_tree_parent_of_compat. 
      4: apply same.
      apply Par_part_compat; apply env_proper; reflexivity.
      assumption. 
      reflexivity.

      intros nson.
      unfold pcl_chain in hpn.
      rewrite <- Heqspn, nson in hpn.      
      inversion hpn.
      lia.
    Qed.

    (** well founded *)
    Lemma pcl_chain_wf: 
      forall (env: Env), 
        Symmetric_Network_assumption
          Net (fun n => KCpeers (env n)) (fun n => KCreverse (env n)) ->
        Proper pequiv env -> spanning_tree root (Par_part env) 
        -> well_founded (pcl_chain env).
    Proof.
      intros env sym env_proper span.
      intros n.
      case_eq (alpha (env n) >=? k); intro HAAA.
      apply Acc_Tall; auto.
      apply (WF_child (Par_part_compat env_proper) span).
      assert (HAA: alpha (env n) <=? k = true).
      lia.
      apply Acc_Short; try assumption.
      apply (WF_par (Par_part_compat env_proper) span).
      apply (WF_child (Par_part_compat env_proper) span).
    Qed.

  End Chain_Of_Pcl_Pointers.

  (** ** Head Potentila
    - define dist_hd (structural induction on the above well founded order)
    - define potential ( = N+1 - dist_hd )
   *)
  Section Hd_Potential.


    (** ** Head-Distance *)
    Fixpoint dist_hd (env: Env) (n: Node) 
             (W: Acc (pcl_chain env) n): Z  :=
      match get_pcl env n as x0 
            return (x0 == (get_pcl env n) -> Z) 
      with
      | Some px => fun heq => 1 + dist_hd (Acc_inv W px heq)
      | None => fun _ => 0
      end (Equivalence_Reflexive (get_pcl env n)).
    
    Lemma dist_hd_compat:
      forall env1 env2 n1 n2
             (A1: Acc (pcl_chain env1) n1)
             (A2: Acc (pcl_chain env2) n2),
        env1 =~= env2 -> n1 == n2 -> dist_hd A1 = dist_hd A2.
    Proof.
      intros env1 env2 n1 n2 A1 A2 He Hn.
      revert n2 A2 Hn.
      induction A1 as [n1 A1 IH1] using Acc_ind_dep.
      destruct A2 as (A2).
      intros Hn.
      generalize (get_pcl_compat He n1 n2 Hn).
      revert  A1 A2 IH1 Hn. 
      unfold pcl_chain.
      simpl; destruct (get_pcl env1 n1) as [ p1 |];
        destruct (get_pcl env2 n2) as [ p2 |];
        intros A1 A2 IH1 Hn tp.
      shelve.
      inversion tp.
      inversion tp.
      reflexivity.
      Unshelve.
      change (1+dist_hd (A1 p1 (Equivalence_Reflexive (R:=equiv) (Some p1))) =
              1+dist_hd (A2 p2 (Equivalence_Reflexive (R:=equiv) (Some p2)))).
      simpl in tp.
      apply f_equal2;auto.
    Qed.

    Lemma dist_hd_cluster_head: 
      forall (n: Node) (env: Env) (A: Acc (pcl_chain env) n)
             (PE: Proper pequiv env), 
        IsClusterHead (env n) = true -> dist_hd A = 0.
    Proof.
      intros n env A PE CH.
      assert (tmp := noclusterhead_ n env CH). 
      unfold pcl_chain in tmp.
      assert (tmp': None == (get_pcl env n)).
      revert tmp; destruct (get_pcl env n).
      intros tmp; specialize (tmp n0); contradict tmp; reflexivity.
      intros _; reflexivity.
      destruct A as (A).
      clear tmp CH PE.
      revert A tmp'. unfold pcl_chain.
      simpl.
      destruct (get_pcl env n); intros. 
      elim tmp'.
      reflexivity.
    Qed.

    Lemma dist_hd_pcl: 
      forall n pn env (An: Acc (pcl_chain env) n) (Apn: Acc (pcl_chain env) pn)
             (PE: Proper pequiv env), 
        pcl_chain env pn n -> dist_hd An = 1+ dist_hd Apn.
    Proof.
      intros n pn env An Apn PE pcl.
      destruct An as (An).
      simpl.
      revert pcl Apn An. 
      unfold pcl_chain.
      destruct (get_pcl env n); intros.
      2: inversion pcl.
      simpl in pcl.
      rewrite (dist_hd_compat
                 Apn (An n0 (Equivalence_Reflexive (Some n0))) PE pcl). 
      reflexivity.
    Qed.

    Lemma hd_pcl_chain_wf: 
      forall (env: sig (Pspanning_tree root)),
        well_founded (pcl_chain (proj1_sig env)).
    Proof.
      intros [env [s p]]; simpl.
      apply pcl_chain_wf; auto; apply p.
    Qed.
    
    Definition dist_hd' (env: sig (Pspanning_tree root)) (n: Node): Z :=
      dist_hd (hd_pcl_chain_wf env n).

    Instance dist_hd'_compat: Proper fequiv dist_hd'.
    Proof.
      intros [e1 sp1] [e2 sp2] eqe n1 n2 eqn.
      unfold dist_hd'.
      apply dist_hd_compat.
      simpl.
      intros a b eab.
      apply eqe; assumption.
      assumption.
    Qed.

    (** ** Head Potential *)
    (* permettre d'utiliser l'ordre inverse dans P_Q_termination
     ou instancier NN par le nombre de noeuds + 1, p.ex *)
    Variable (NN: Z)
             (HNN: forall (env: sig (Pspanning_tree root))
                          (n: Node), NN > dist_hd' env n).

    Definition hd_potential (env: sig (Pspanning_tree root)) (n: Node): Z :=
      if andb 
           (andb (negb (alpha_enabled_bb (proj1_sig env) n))
                 (negb (pcl_enabled_bb (proj1_sig env) n)))
           (hd_enabled_bb (proj1_sig env) n) then NN - dist_hd' env n
      else  0.

    Instance hd_potential_compat: Proper fequiv hd_potential.
    Proof.
      intros es1 es2 eqe n1 n2 eqn.
      unfold hd_potential. simpl. 
      rewrite (dist_hd'_compat  eqe  _ _  eqn).
      destruct es1 as [e1 s1]; destruct es2 as [e2 s2].
      simpl.
      rewrite (hd_enabled_bb_compat eqe _ _ eqn).
      rewrite (alpha_enabled_bb_compat eqe _ _ eqn).
      rewrite (pcl_enabled_bb_compat eqe _ _ eqn).
      reflexivity.
    Qed.

    Lemma hd_potential_pos:
      forall env n, hd_potential env n >= 0.
      intros.
      unfold hd_potential.
      case (negb (alpha_enabled_bb (proj1_sig env) n) &&
                 negb (pcl_enabled_bb (proj1_sig env) n) &&
                 (hd_enabled_bb (proj1_sig env) n)).
      2: lia.
      specialize (HNN env n).
      lia.
    Qed.
    
  End Hd_Potential.

  (** ** Results with a Head Step *)
  Section With_Hd_Step.

    Variable (g g': Env)
             (step: Step g' g)
             (hd_step: hd_step_b g' g = true)
             (span_tree_env: Pspanning_tree root g)
             (span_tree_env': Pspanning_tree root g').

    
    Variable (NN: Z)
             (HNN: forall (env: sig (Pspanning_tree root)) (n: Node),
                 NN > dist_hd' env n).

    (** ** Head Global Cirteria *)
    Lemma hd_no_alpha_pcl_enabled_when_move:
      forall n, has_moved_b g' g n = true
                -> alpha_enabled_bb g n = false /\
                   pcl_enabled_bb g n = false.
    Proof.
      intros n H.
      unfold hd_step_b in hd_step.
      rewrite forallb_forall in hd_step.
      generalize (all_nodes_prop n); intros ain.
      rewrite InA_alt in ain.
      destruct ain as [n' [eqn in']].
      specialize (hd_step _ in').
      rewrite orb_true_iff in hd_step.
      destruct hd_step as [hd_step'|hd_step'].
      2:{
        rewrite negb_true_iff in hd_step'.
        erewrite has_moved_b_compat in hd_step'.
        3: eapply step.
        2: apply Step_Proper_l with g; apply step.
        2: symmetry; apply eqn.
        rewrite hd_step' in H; contradict H; apply diff_false_true.
      }
      rewrite negb_orb, andb_true_iff in hd_step'.
      destruct hd_step' as [X XX].
      rewrite negb_true_iff in X, XX.
      erewrite (alpha_enabled_bb_compat).
      2: apply step.
      2: apply eqn.
      split; [assumption|].
      erewrite (pcl_enabled_bb_compat ).
      2: apply step.
      2: apply eqn.
      assumption.
    Qed.

    Lemma hd_enabled_when_move:
      forall n, has_moved_b g' g n = true
                -> hd_enabled_bb g n = true.
    Proof.
      intros n sds.
      assert (AP := hd_no_alpha_pcl_enabled_when_move n sds).
      destruct AP as [A P].
      unfold has_moved_b in sds.
      destruct (State_dec (g' n) (g n)) as [S|S]; [congruence|]; clear sds.
      destruct step as [_ _ _ sds].
      unfold RUN, run in sds; simpl in sds.
      unfold run_algorithm in sds.
      specialize (sds n).
      unfold alpha_enabled_bb in A; rewrite A in sds.
      unfold pcl_enabled_bb in P; rewrite P in sds.
      unfold hd_enabled_bb;
        revert sds; case (hd_enabled_b (g n) (local_env g n)); 
          simpl; intros.
      reflexivity.
      tauto.
    Qed.

    Lemma hd_alpha_pcl_unchanged:
      forall n, alpha (g n) = alpha (g' n) /\ (parCl (g n)) == (parCl (g' n)).
    Proof.
      intros n.
      case_eq (has_moved_b g' g n); intros hm.
      { destruct step as [pg pg' _ r].
        dup hm hm'.
        unfold has_moved_b in hm'.
        destruct (State_dec (g' n) (g n)) as [S|S]; [congruence|]; clear hm'.
        specialize (r n); destruct r as [r|r]; [| tauto].
        assert (no := hd_no_alpha_pcl_enabled_when_move n hm).
        unfold RUN, run in r; simpl in r; unfold run_algorithm in r.
        unfold alpha_enabled_bb, pcl_enabled_bb in no; destruct no as [no1 no2].
        rewrite no1, no2 in r.
        destruct (hd_enabled_b (g n) (local_env g n)); [| tauto].
        rewrite r; simpl; split; reflexivity.
      }
      { unfold has_moved_b in hm.
        destruct (State_dec (g' n) (g n)) as [S|S]; [|congruence].
        split; [apply alpha_compat | apply parCl_compat];
          symmetry; auto.
      }          
    Qed.
    
    Lemma cluster_head_unchanged:
      forall n, IsClusterHead (g n) = IsClusterHead (g' n).
    Proof.
      intros n.
      unfold KClustering_algo.IsClusterHead, _IsShort. 
      assert (APC := hd_alpha_pcl_unchanged n); destruct APC as [A _].
      assert (P: Par (g n) == Par (g' n)).
      { apply Par_compat.
        symmetry; apply (Stable_Variable_step KC_Stable); auto.
      }
      rewrite A, P; reflexivity.
    Qed.

    Lemma exists_hd_move:
      exists n, has_moved_b g' g n = true /\ 
                hd_enabled_bb g n = true.
    Proof.
      destruct step as [pg pg' dif r].
      destruct (All_In_dec_forall_exists all_nodes_prop
                                         (P := fun n => ~ g' n == g n)).
      intros n1 n2 en; rewrite en; reflexivity.
      intros nn; destruct (State_dec (g' nn) (g nn));
        [right|left]; auto.
      { contradict dif.
        intros n1 n2 en; rewrite en; clear n1 en.
        specialize (n n2).
        destruct (State_dec (g' n2) (g n2)); tauto.
      }          
      destruct e as [n hn].
      exists n.
      assert (M: has_moved_b g' g n = true).
      { unfold has_moved_b.
        destruct (State_dec (g' n) (g n)); tauto.
      }
      split; auto.
      apply hd_enabled_when_move; assumption.
    Qed.

    Lemma hd_still_enabled:
      forall n, hd_enabled_bb g n = true ->
                hd_enabled_bb g' n = true -> has_moved_b g' g n = true
                -> exists pn, pcl_chain g pn n /\
                              hd_enabled_bb g pn = true /\
                              has_moved_b g' g pn = true.
    Proof.
      intros n en en' mov.
      destruct step as [pg pg' dif r].
      destruct (hd_no_alpha_pcl_enabled_when_move n mov) as [ A P].
      unfold has_moved_b in mov;
        specialize (r n); destruct (State_dec (g' n) (g n));
          [congruence|].
      destruct r; [| tauto]; clear mov.
      rename H into mov.
      unfold RUN, run in mov; simpl in mov;
        unfold run_algorithm in mov.
      unfold alpha_enabled_bb in A; unfold pcl_enabled_bb in P; 
        unfold hd_enabled_bb in en;
        rewrite A, P, en in mov; simpl in mov.
      rename mov into E.
      
      assert (hdnew: ~ (hd (g n)) == (hd (g' n))).
      { rewrite (hd_compat E); simpl.
        unfold hd_enabled_bb, hd_enabled_b in en; revert en; 
          destruct (Id_dec (hd (g n)) (NewHead (g n) (local_env g n)));
          [congruence|]; auto.
      }
      assert (CH: IsClusterHead (g n) = false).
      { rewrite <- not_true_iff_false.
        unfold hd_enabled_bb, hd_enabled_b, NewHead in en'.
        unfold hd_enabled_b, NewHead in en.
        unfold NewHead in E.
        intros CH.
        rewrite CH in en, E.
        rewrite (cluster_head_unchanged n) in CH.      
        rewrite CH in en'.
        revert en en'; case (Id_dec (hd (g n)) (id (g n)));
          case (Id_dec (hd (g' n)) (id (g' n))); try congruence.
        intros F; contradict F; rewrite (hd_compat E);
          rewrite <- (id_compat (KCROpart_compat E)); simpl; reflexivity.
      }
      assert (HD := hd_compat E); simpl in HD.
      unfold NewHead, local_env in HD; rewrite CH in HD.
      assert (EX: get_pcl g n <> None).
      { intros NEX. unfold get_pcl in NEX.
        unfold alpha_enabled_bb, pcl_enabled_bb in NEX;
          rewrite A, P in NEX; simpl in NEX.
        revert HD NEX; case_eq (parCl (g n)). 
        2: intros _ F; contradict hdnew; symmetry; assumption.
        intros c EX; destruct (peer n c);
          [intros _ FF; inversion FF|
           simpl; intros FF; contradict hdnew; symmetry; assumption].
      }        
      case_eq (get_pcl g n);
        [intros pn hpn |
         intros FF; rewrite FF in EX; contradict EX; reflexivity].

      exists pn. split.
      unfold pcl_chain; rewrite hpn; reflexivity.
      clear EX.      

      (* *)
      unfold get_pcl in hpn.
      unfold alpha_enabled_bb, pcl_enabled_bb in hpn; 
        rewrite A, P in hpn; simpl in hpn.
      case_eq (parCl (g n)).
      2: intros FF; rewrite FF in hpn; inversion hpn.
      intros c eqc.
      rewrite eqc in hpn, HD.
      case_eq (peer n c).
      2: intros FF; rewrite FF in hpn; inversion hpn.
      intros ppn eqppn; rewrite eqppn in hpn, HD; simpl in hpn, HD.
      rewrite hpn in eqppn.
      inversion hpn.
      rewrite H0 in HD.
      clear ppn hpn H0.
      
      generalize en'; intros en''.
      unfold hd_enabled_bb, hd_enabled_b in en''. 
      unfold NewHead in en''.
      
      rewrite (cluster_head_unchanged n) in CH; rewrite CH in en''.
      unfold local_env in en''.
      assert (EN'': (if Id_dec (hd (g' n))
                               match parCl (g' n) with
                               | Some Par =>
                                 match option_map g' (peer n Par) with
                                 | Some ps => hd ps
                                 | None => hd (g' n)
                                 end
                               | None => hd (g' n)
                               end
                     then false
                     else true) = true); auto; clear en''.
      revert EN''; case (Id_dec (hd (g' n))
                                match parCl (g' n) with
                                | Some Par =>
                                  match option_map g' (peer n Par) with
                                  | Some ps => hd ps
                                  | None => hd (g' n)
                                  end
                                | None => hd (g' n)
                                end); intros EE.
      intros FF; contradict FF; apply diff_false_true.
      intros _.
      assert (PC: (parCl (g n)) == (parCl (g' n))).
      rewrite <- (parCl_compat E); simpl.
      eapply (parCl_compat ); apply (Step_Proper_l step); 
        reflexivity.
      rewrite eqc in PC.
      revert EE PC; case_eq (parCl (g' n)).
      2: intros _ FF; contradict FF; apply (hd_compat ).
      2: apply (Step_Proper_l step); reflexivity.
      intros cc eqpc EE eqcc; simpl in eqcc.
      generalize (peer_compat _ _ (Equivalence_Reflexive n) _ _ eqcc); 
        intros tmp.
      rewrite eqppn in tmp.
      revert EE tmp; case_eq (peer n cc); simpl.
      2: intros _ _ FF; elim FF.
      intros ppn eqccppn EE eqpppn.
      rewrite <- (hd_compat ((Step_Proper_l step) _ _ eqpppn)) in EE.
      clear eqpppn eqccppn ppn.
      clear eqcc eqpc cc.
      rewrite HD in EE.

      (* *)
      split; [|
              unfold has_moved_b;
              destruct (State_dec (g' pn) (g pn)); auto;
              contradict EE; apply hd_compat; symmetry; auto].
      destruct step as [_ _ _ r].
      specialize (r pn).
      destruct r as [r|r].
      2: contradict EE; apply hd_compat; symmetry; auto.
      rename r into D.
      unfold RUN, run in D; simpl in D; unfold run_algorithm in D.
      case_eq (alpha_enabled_b (g pn) (local_env g pn));
        intros AA; rewrite AA in D.
      { contradict EE.
        rewrite (hd_compat  D); auto. }
      case_eq (parCl_enabled_b (g pn) (local_env g pn));
        intros PP; rewrite PP in D.
      { contradict EE.
        rewrite (hd_compat D); auto. }
      case_eq (hd_enabled_b (g pn) (local_env g pn));
        intros HH; rewrite HH in D.
      2: contradict EE; rewrite (hd_compat D); auto.        
      apply HH.
    Qed.

    Lemma exists_hd_move_disabled:
      exists n, hd_enabled_bb g n = true /\
                hd_enabled_bb g' n = false /\
                has_moved_b g' g n = true.
    Proof.
      destruct exists_hd_move as [n [ds en]].
      assert (pg: Proper pequiv g).
      apply step.
      assert (span: spanning_tree root (Par_part g)).
      apply span_tree_env.
      induction (pcl_chain_wf (proj2 (proj2 span_tree_env)) pg span n)
        as [n hpcl IH].
      case_eq (hd_enabled_bb g' n); intros hdn';
        [| exists n; repeat split; assumption].
      destruct (hd_still_enabled n en hdn' ds) as [pn [hpn [hdpn pnmov]]].        
      apply IH with pn; assumption.
    Qed.

    Lemma alpha_enabled_unchanged_hd:
      forall n, alpha_enabled_bb g n = alpha_enabled_bb g' n.
    Proof.
      intros n.
      unfold alpha_enabled_bb, alpha_enabled_b.
      assert (pg: Proper pequiv g).
      apply step.
      assert (pg': Proper pequiv g').
      apply (Step_Proper_l step).
      case (Z.eq_dec (alpha (g n))
                     (NewAlpha (g n) (local_env g n)));
        case (Z.eq_dec (alpha (g' n))
                       (NewAlpha (g' n) (local_env g' n))); intros Z' Z.
      reflexivity. 
      3: reflexivity.
      contradict Z'.
      assert (X := hd_alpha_pcl_unchanged).
      rewrite <- (proj1 (X n)), Z.
      apply NewAlpha_only_depends_on_children with root; try assumption.
      apply span_tree_env.
      apply span_tree_env'.      

      intros child _; apply X. 
      contradict Z.
      assert (X := hd_alpha_pcl_unchanged).
      rewrite (proj1 (X n)), Z'.
      symmetry; apply NewAlpha_only_depends_on_children with root;
        try assumption.
      apply span_tree_env.
      apply span_tree_env'.      
      intros child _; apply X. 
    Qed.

    Lemma pcl_enabled_unchanged:
      forall n, pcl_enabled_bb g n = pcl_enabled_bb g' n.
    Proof.
      intros n.
      unfold pcl_enabled_bb, parCl_enabled_b.
      generalize (hd_alpha_pcl_unchanged n); intros [ _ p ].
      assert (OK: (NewparCl (g n) (local_env g n)) ==
                  (NewparCl (g' n) (local_env g' n))).
      apply pcl_NewParCl_unchanged_ with root; try assumption.
      apply step.
      apply step.
      apply span_tree_env.
      apply span_tree_env'.
      intros nn.
      apply hd_alpha_pcl_unchanged.
      revert OK; 
        case (eqoptionA_dec Channel_dec (parCl (g n))
                            (NewparCl (g n) (local_env g n))); 
        case (eqoptionA_dec Channel_dec (parCl (g' n))
                            (NewparCl (g' n) (local_env g' n)));
        intros.
      reflexivity.
      contradict n0; rewrite <- p, <- OK; apply e.
      contradict n0; rewrite p, OK; apply e.
      reflexivity.
    Qed.

    Lemma Sg: Pspanning_tree root g.
    Proof.
      split.
      apply step.
      apply span_tree_env. 
    Qed.
    Lemma Sg': Pspanning_tree root g'.
    Proof.
      split.
      apply (Step_Proper_l step).
      apply spanning_tree_stable with g.
      apply Sg.
      assumption.
    Qed.

    Notation sg := (exist Sg).  
    Notation sg' := (exist Sg').  

    (** *)
    Section Global_criterium.

      Lemma global_hd_criterium:
        exists n, hd_potential NN sg n <> hd_potential NN sg' n. 
      Proof.
        assert (mov := exists_hd_move_disabled).
        destruct mov as [n [en [dis dn]]]. 
        exists n.
        unfold hd_potential.
        destruct (hd_no_alpha_pcl_enabled_when_move n dn) as [A P].
        simpl.
        rewrite <- (alpha_enabled_unchanged_hd n), 
        <- (pcl_enabled_unchanged n), A, P , en, dis.
        simpl.
        generalize (HNN sg n).
        lia.
      Qed.

    End Global_criterium.

    (** ** Head Local Criteria *)
    Lemma get_pcl_unchanged:
      forall n, (get_pcl g n) == (get_pcl g' n).
    Proof.
      intros n.
      unfold get_pcl.
      destruct (hd_alpha_pcl_unchanged n) as [ A P ].
      revert P; destruct (parCl (g n)); destruct (parCl (g' n)).
      2: intros FF; inversion FF.
      2: intros FF; inversion FF.
      2: intros; reflexivity.
      simpl; intros pcl.
      generalize (peer_compat _ _ (Equivalence_Reflexive n) _ _ pcl);
        destruct (peer n c); destruct (peer n c0).
      2: intros FF; inversion FF.
      2: intros FF; inversion FF.
      2: intros; reflexivity.
      simpl; intros nn.
      rewrite (alpha_enabled_unchanged_hd n),  (pcl_enabled_unchanged n).
      destruct ( negb (alpha_enabled_bb g' n) && negb (pcl_enabled_bb g' n));
        simpl; auto.
    Qed.

    Lemma dist_hd_unchanged:
      forall n, dist_hd (hd_pcl_chain_wf sg n) =
                dist_hd (hd_pcl_chain_wf sg' n).
    Proof.
      intros n.
      generalize (hd_pcl_chain_wf sg  n); intros W1. 
      generalize (hd_pcl_chain_wf sg' n); intros W2. 
      induction W1 as [n W1 IH1] using Acc_ind_dep.
      destruct W2 as (W2).

      assert(G := get_pcl_unchanged n); revert G.
      revert W1 W2 IH1.
      unfold pcl_chain.
      simpl.
      destruct (get_pcl g n) as [ p |];
        destruct (get_pcl g' n) as [ p' |];
        intros W1 W2 IH1 G.
      2: inversion G.
      2: inversion G.
      2: reflexivity.
      simpl in G.

      specialize (IH1 _ (Equivalence_Reflexive (R:=equiv) (Some p)) (W2 p G)) .
      assert (X: dist_hd (W2 p G) =
                 dist_hd (W2 p' (Equivalence_Reflexive (R:=equiv) (Some p')))).
      apply dist_hd_compat.
      intros a b eab; apply (Step_Proper_l step); assumption.
      assumption.
      rewrite X in IH1; clear X.
      
      revert IH1; 
        destruct (dist_hd (W1 p (Equivalence_Reflexive (Some p))));
        destruct (dist_hd (W2 p' (Equivalence_Reflexive (Some p'))));
        intros; try inversion IH1; reflexivity.
    Qed.

    Lemma dist_hd'_unchanged:
      forall n, dist_hd' sg n = dist_hd' sg' n.
      intros n.
      unfold dist_hd'; simpl.
      rewrite dist_hd_unchanged.
      reflexivity.
    Qed.

    Lemma hd_dont_move:
      forall n, hd_enabled_bb g n = false -> eqS (g n) (g' n).
    Proof.
      intros n dis.

      case_eq (has_moved_b g' g n); intros hm.
      2: unfold has_moved_b in hm;
        symmetry; destruct (State_dec (g' n) (g n));
          [auto | congruence].

      destruct step as [pg pg' dif r].
      specialize (r n); destruct r as [r|r]; [| symmetry; auto].

      destruct (hd_no_alpha_pcl_enabled_when_move n) as [A H]; auto.

      unfold RUN, run in r; simpl in r;
        unfold run_algorithm in r.
      unfold alpha_enabled_bb in A; rewrite A in r.
      unfold pcl_enabled_bb in H; rewrite H in r.
      unfold hd_enabled_bb in dis; rewrite dis in r.
      symmetry; tauto.
    Qed.

    (** *)
    Section Local_criterium.

      Lemma local_crit_next_steps:
        forall n, hd_enabled_bb g n = true -> hd_enabled_bb g' n = true ->
                  has_moved_b g' g n = true ->
                  exists n', hd_enabled_bb g n' = true /\
                             hd_enabled_bb g' n' = false /\
                             has_moved_b g' g n' = true /\
                             dist_hd' sg n' < dist_hd' sg' n.
      Proof.
        intros n.
        unfold dist_hd'.
        generalize (hd_pcl_chain_wf sg n); intros W1. 
        generalize (hd_pcl_chain_wf sg' n); intros W2. 
        revert W2.
        induction W1 as [n W1 IH1] using Acc_ind_dep.
        destruct W2 as (W2).          
        intros H H' ds.
        assert (DONE := hd_still_enabled n H H' ds).
        assert (pg: Proper pequiv g).
        apply step.
        destruct DONE as [pn [pcl [HPN dpn]]].
        assert (distpn :=
                  dist_hd_pcl (hd_pcl_chain_wf sg n) (W1 pn pcl) pg pcl).
        rewrite (dist_hd_unchanged) in distpn.
        assert (tmp' :=
                  (dist_hd_compat (W1 pn pcl)
                                  (hd_pcl_chain_wf sg pn)
                                  pg (Equivalence_Reflexive pn))). 
        rewrite tmp' in distpn.
        clear tmp'.
        rewrite (dist_hd_unchanged) in distpn.

        case_eq (hd_enabled_bb g' pn); intros HPN'.
        (* hd_enabled_bb env' pn = true: use IH *)
        assert (tmp: pcl_chain (proj1_sig sg') pn n).
        unfold proj1_sig; simpl.
        unfold pcl_chain.
        rewrite <- get_pcl_unchanged.
        unfold pcl_chain in pcl.
        assumption.
        specialize (IH1 pn pcl (W2 pn tmp) HPN HPN' dpn).
        destruct IH1 as [n' [HN [HN' O]]].
        exists n'.
        repeat split; try assumption.
        apply O.
        apply transitivity with (dist_hd (W2 pn tmp)).
        apply O.
        assert (pg': Proper pequiv g').
        apply (Step_Proper_l step).
        rewrite ((dist_hd_compat (Acc_intro n W2)
                                 (hd_pcl_chain_wf sg' n)
                                 pg' (Equivalence_Reflexive n))). 
        assert (tmp' :=
                  (dist_hd_compat
                     (W2 pn tmp)
                     (hd_pcl_chain_wf sg' pn) pg' (Equivalence_Reflexive pn))).
        rewrite tmp'.
        rewrite distpn.
        lia.

        (* hd_enabled_bb env' pn = false: exists pn *)
        exists pn.
        repeat split; try assumption.
        assert (pg': Proper pequiv g').
        apply (Step_Proper_l step).
        assert (tmp' :=
                  (dist_hd_compat (Acc_intro n W2)
                                  (hd_pcl_chain_wf sg' n) pg'
                                  (Equivalence_Reflexive n))). 
        rewrite tmp'; clear tmp'.
        rewrite dist_hd_unchanged.
        rewrite distpn.
        lia.
      Qed.

      Lemma local_hd_criterium:
        forall n,
          (hd_potential NN sg n) < (hd_potential NN sg' n) -> 
          exists n', (hd_potential NN sg n') > (hd_potential NN sg' n)
                     /\
                     (hd_potential NN sg n') <> (hd_potential NN sg' n').
      Proof.
        intros n.
        unfold hd_potential, dist_hd'; simpl.

        case_eq (alpha_enabled_bb g n); intros AE;
          rewrite (alpha_enabled_unchanged_hd n) in AE; rewrite AE.
        intros FF; simpl in FF; contradict FF; lia.

        case_eq (pcl_enabled_bb g n); intros PE;
          rewrite (pcl_enabled_unchanged n) in PE; rewrite PE.
        intros FF; simpl in FF; contradict FF; lia.

        simpl negb; simpl andb.
        case_eq (hd_enabled_bb g n); intros H.

        (* hd_enabled_bb g n = true *)
        case_eq (hd_enabled_bb g' n); intros H'.
        intros FF.
        rewrite dist_hd_unchanged in FF.
        contradict FF. change (   ~
                                    NN - dist_hd (hd_pcl_chain_wf sg' n) <
                                  NN - dist_hd (hd_pcl_chain_wf sg' n)
                              );
                         lia.
        intros FF.
        specialize (HNN sg n).
        unfold dist_hd' in HNN.
        contradict FF. change ( ~ NN - dist_hd (hd_pcl_chain_wf sg n) < 0);
                         lia.

        (* hd_enabled_bb g n = false *)
        case_eq (hd_enabled_bb g' n); intros H'.
        2: intros FF; contradict FF; lia.
        intros _.

        assert (NM := hd_dont_move n H).

        unfold hd_enabled_bb, hd_enabled_b in H', H.
        revert H';
          case (Id_dec (hd (g' n))
                       (NewHead (g' n) (local_env g' n))).
        intros _ FF; contradict FF; apply diff_false_true.
        revert H;
          case (Id_dec (hd (g n))
                       (NewHead (g n) (local_env g n))).
        2: intros _ FF; contradict FF; apply diff_true_false.
        intros H _ H' _.
        rewrite (hd_compat NM) in H.
        rewrite H in H'.
        unfold NewHead in H'.

        (*CHFALSE*)
        assert (CH: IsClusterHead (g' n) = false).
        { revert H'; case_eq (IsClusterHead (g n)); intros CH; 
            rewrite (cluster_head_unchanged n) in CH; rewrite CH.
          - intros nh.
            contradict nh.
            apply id_compat, KCROpart_compat; assumption.
          - intros _; reflexivity.
        }
        rewrite CH in H'.
        rewrite <- (cluster_head_unchanged n) in CH; rewrite CH in H'.
        
        destruct NM as [A [Hd [Pc RO]]]. 
        revert Pc H'; case_eq (parCl (g n)); case_eq (parCl (g' n)).
        2: intros x y z FF; elim FF.
        2: intros x y z FF; elim FF.
        2: intros _ _ _ FF; contradict FF; assumption.
        intros c pc cc pcc eqc; simpl in eqc.
        unfold local_env.
        generalize (peer_compat _ _ (Equivalence_Reflexive n) _ _ eqc);
          case_eq (peer n c); case_eq (peer n cc).
        2: intros x y z FF; inversion FF.
        2: intros x y z FF; inversion FF.
        2: intros x xx xxx FF; contradict FF; simpl; assumption.
        simpl; intros pn epn ppn eppn eqpn nh; simpl in nh, eqpn.
        assert (pg: Proper pequiv g) by apply step.
        rewrite (hd_compat (pg _ _ eqpn)) in nh.
        assert (V: has_moved_b g' g ppn = true).
        { unfold has_moved_b;
            destruct (State_dec (g' ppn) (g ppn)); auto.
          contradict nh; apply hd_compat; symmetry; auto.
        }
        assert (dok: g' ppn == RUN g ppn).
        { destruct step as [_ _ _ r].
          specialize (r ppn); destruct r as [r|r]; auto.
          contradict nh; apply hd_compat; symmetry; auto.
        }
        assert (VV: has_moved_b g' g pn = true).
        { erewrite has_moved_b_compat.
          apply V.
          apply Step_Proper_l with g; assumption. 
          apply pg.
          apply eqpn.
        }
        assert (ok: pcl_chain g ppn n).
        { unfold pcl_chain, get_pcl.
          rewrite pcc, epn.
          rewrite <- (alpha_enabled_unchanged_hd n) in AE; rewrite AE.
          rewrite <- (pcl_enabled_unchanged n) in PE; rewrite PE.
          simpl; symmetry; assumption.
        }
        assert (ok': pcl_chain g' pn n).
        { unfold pcl_chain.
          rewrite <- (get_pcl_unchanged n).
          rewrite eqpn; apply ok.
        }
        assert (Hpn := hd_enabled_when_move _ VV). 

        case_eq (hd_enabled_bb g' pn); intros Hpn'.
        (* hd_enabled_bb g' pn = true *)
        { destruct (local_crit_next_steps pn Hpn Hpn' VV)
            as [n' [Hn' [H'n' [D DD]]]].
          exists n'.

          destruct (hd_no_alpha_pcl_enabled_when_move _ D) as [A' P']. 
          split.
          rewrite Hn', A', P'.
          simpl negb; simpl andb.
          rewrite dist_hd_unchanged.
          assert (pg': Proper pequiv g').
          apply (Step_Proper_l step).
          change (NN - dist_hd (hd_pcl_chain_wf sg' n') >
                  NN - dist_hd (hd_pcl_chain_wf sg' n)).
          rewrite (dist_hd_pcl (hd_pcl_chain_wf sg' n) (hd_pcl_chain_wf sg' pn)
                               pg' ok').
          unfold dist_hd', proj1_sig in DD.
          rewrite dist_hd_unchanged in DD.
          revert DD.
          change (dist_hd (hd_pcl_chain_wf sg' n') <
                  dist_hd (hd_pcl_chain_wf sg' pn) ->
                  NN - dist_hd (hd_pcl_chain_wf sg' n') >
                  NN - (1 + dist_hd (hd_pcl_chain_wf sg' pn))
                 ).
          lia.

          rewrite A', P', Hn', H'n'.
          rewrite alpha_enabled_unchanged_hd in A'; rewrite A'.
          rewrite pcl_enabled_unchanged in P'; rewrite P'.
          simpl negb; simpl andb.
          rewrite <- dist_hd_unchanged.
          specialize (HNN sg n').
          unfold dist_hd' in HNN. revert HNN. 
          change (   NN > dist_hd (hd_pcl_chain_wf sg n') ->
                     NN - dist_hd (hd_pcl_chain_wf sg n') <> 0);        lia.
        }
        (* hd_enabled_bb g' pn = false *)
        { exists pn.
          destruct (hd_no_alpha_pcl_enabled_when_move ppn V) as [AA PP].
          rewrite <- (alpha_enabled_bb_compat pg _ _ eqpn) in AA.
          rewrite <- (pcl_enabled_bb_compat pg _ _ eqpn) in PP.
          rewrite AA, PP; simpl.
          rewrite Hpn, Hpn'.
          rewrite alpha_enabled_unchanged_hd in AA; rewrite AA.
          rewrite pcl_enabled_unchanged in PP; rewrite PP.
          simpl negb; simpl andb.
          rewrite <- dist_hd_unchanged.
          rewrite <- (pcl_chain_compat
                        pg _ _ eqpn _ _ (Equivalence_Reflexive n)) in ok.
          assert (X :=
                    dist_hd_pcl (hd_pcl_chain_wf sg n) (hd_pcl_chain_wf sg pn)
                                pg ok).
          rewrite X.
          specialize (HNN sg pn).
          unfold dist_hd' in HNN.
          revert X.
          change (dist_hd (hd_pcl_chain_wf sg n) =
                  1 + dist_hd (hd_pcl_chain_wf sg pn) ->
                  NN - dist_hd (hd_pcl_chain_wf sg pn) >
                  NN - (1 + dist_hd (hd_pcl_chain_wf sg pn)) /\
                  NN - dist_hd (hd_pcl_chain_wf sg pn) <> 0);
            lia.
        }
      Qed.

    End Local_criterium.

  End With_Hd_Step.
  
  Section Hd_Termination.

    Variable (NN: Z)
             (HNN: forall (env: sig (Pspanning_tree root)) (n: Node),
                 NN > dist_hd' env n).

    Definition hd_pot (sg: sig (Pspanning_tree root)) (n: Node) :=
      Z.to_nat (hd_potential NN sg n).
    Instance hd_pot_compat: Proper fequiv hd_pot.
    Proof.
      intros e1 e2 eqe n1 n2 eqn.
      unfold hd_pot.
      rewrite (hd_potential_compat NN eqe _ _ eqn).
      reflexivity.
    Qed.

    Notation ltM := (@MultisetNat.MSetOrd.MultisetLT gt).
    Notation Pot_hd := (Pot hd_pot).

    (** ** Head Termination *)
    Lemma hd_safe_inclusion:
      inclusion (Head_SafeStep (root := root)) (ltM @@ Pot_hd).
    Proof.

      assert (poteq: forall (sg sg': sig (Pspanning_tree root)) n
                            (step: Step (proj1_sig sg') (proj1_sig sg)),
                 hd_potential NN (exist (Sg step (proj2_sig sg))) n =
                 hd_potential NN sg n).
      { intros [g span] [g' span'] n step; simpl.
        apply hd_potential_compat; [| reflexivity].
        apply span.
      }
      assert (poteq': forall (sg sg': sig (Pspanning_tree root)) n
                             (step: Step (proj1_sig sg') (proj1_sig sg)),
                 hd_potential NN (exist (Sg' step (proj2_sig sg))) n
                 = hd_potential NN sg' n).
      { intros [g span] [g' span'] n step; simpl.
        apply hd_potential_compat; [| reflexivity].
        apply (Step_Proper_l step).
      }
      apply Step_decreases_pot.  

      apply hd_pot_compat.

      (* local criterion *)
      { intros sg sg' [step hdstep] n Hgt.
        unfold hd_pot in *.
        assert (XX :=
                  local_hd_criterium
                    step hdstep (proj2_sig sg) (proj2_sig sg') HNN n).
        assert (tmp: hd_potential
                       NN (exist (Sg step (proj2_sig sg))) n <
                     hd_potential
                       NN (exist (Sg' step (proj2_sig sg))) n).
        rewrite (poteq _ _ n step); rewrite (poteq' _ _ n step).
        rewrite  Z2Nat.inj_lt.  apply Hgt. 
        rewrite <- Z.ge_le_iff; apply hd_potential_pos; apply HNN.
        rewrite <- Z.ge_le_iff; apply hd_potential_pos; apply HNN.
        specialize (XX tmp); clear tmp.
        destruct XX as [n' XX].
        exists n'.
        destruct XX as [XX1 XX2].
        rewrite (poteq _ _ n' step) in XX1, XX2;
          rewrite (poteq' _ _ n step) in XX1;
          rewrite (poteq' _ _ n' step) in XX2.
        split.
        intros F; contradict XX2; apply Z2Nat.inj.   
        rewrite <- Z.ge_le_iff; apply hd_potential_pos; apply HNN.
        rewrite <- Z.ge_le_iff; apply hd_potential_pos; apply HNN.
        apply F.
        apply Z.gt_lt in XX1; rewrite Z2Nat.inj_lt in XX1. 
        lia.
        rewrite <- Z.ge_le_iff; apply hd_potential_pos; apply HNN.
        rewrite <- Z.ge_le_iff; apply hd_potential_pos; apply HNN.
      }
      (* global criterion *)
      { intros sg sg' [step hdstep].
        unfold hd_pot.
        destruct (global_hd_criterium step hdstep (proj2_sig sg)
                                      (proj2_sig sg') HNN) as [n XX].
        exists n.
        rewrite (poteq _ _ n step), (poteq' _ _ n step) in XX.
        intros F; contradict XX.
        apply Z2Nat.inj.
        rewrite <- Z.ge_le_iff; apply hd_potential_pos; apply HNN.
        rewrite <- Z.ge_le_iff; apply hd_potential_pos; apply HNN.
        symmetry; apply F.
      }
    Qed.

  End Hd_Termination.

End KClustering_Hd_Termination.

Close Scope signature_scope.
Close Scope Z_scope.
Unset Implicit Arguments.
