(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import ZArith.
From Coq Require Import Lia.
From Coq Require Import Bool.
From Coq Require Import Inclusion.
From Coq Require Import RelationPairs.

(** **** Local Imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import OptionUtil.
From Padec Require Import ZUtils.
From Padec Require Import ListUtils.
From Padec Require Import RelationA.
From Padec Require Import FoldRight.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Spanning_Tree.
From Padec Require Import KClustering_specification.
From Padec Require Import KClustering_algo.
From Padec Require Import KClustering_def_termination.
From Padec Require Import KClustering_alpha_termination.
From Padec Require Import P_Q_Termination.

(* end show *)
Open Scope signature_scope.
Open Scope Z_scope.
Set Implicit Arguments.

(** * Termination for KClustering Algorithm: Parent Cluster Steps Only

 This part shows termination of the algorithm when considering Pcl
 safe steps only.  Actually, it shows safe inclusion required by the
 rest of the proof:

         [pcl_safe_incl := (forall g g', Pcl_SafeStep g' g -> ltM
 (Pot_pcl g') (Pot_pcl g)] *)

Section KClustering_Pcl_Termination.

  (** ** Channel and network *)  
  Context {Chans: Channels} {Net: Symmetric_Network}
          {KCP: KClustering_parameters} {KCT: KC_algo_tools}.

  Existing Instance KC_Stable.
  
  Notation Env := (Env (Algo := KC_algo)).

  Let k_pos := k_pos.

  Variable (root: Node).

  (** ** PCL Potential *)
  Definition pcl_potential (sg: sig (Pspanning_tree root)) (n: Node): nat :=
    if pcl_enabled_bb (proj1_sig sg) n then 1%nat else 0%nat.

  Global Instance pcl_potential_compat: Proper fequiv pcl_potential.
  Proof.
    intros [e1 s1] [e2 s2] eqe n1 n2 eqn.
    unfold pcl_potential; simpl. 
    rewrite (pcl_enabled_bb_compat eqe _ _ eqn).
    reflexivity.
  Qed.

  (** ** Results with a PCL step *)
  Section With_Pcl_Step.

    Variable (g g': Env)
             (pg: Proper pequiv g) (pg': Proper pequiv g')
             (step: Step g' g)
             (pcl_step: pcl_step_b g' g = true)
             (span_tree: spanning_tree root (Par_part g) /\
                         Symmetric_Network_assumption
                           Net (fun n => KCpeers (g n))
                           (fun n => KCreverse (g n)))
             (span_tree': spanning_tree root (Par_part g') /\
                          Symmetric_Network_assumption
                            Net (fun n => KCpeers (g' n))
                            (fun n => KCreverse (g' n))).

    Lemma pcl_no_alpha_enabled_when_move:
      forall n,  has_moved_b g' g n = true -> alpha_enabled_bb g n = false.
    Proof.
      intros n H.
      apply andb_prop in pcl_step.
      destruct pcl_step as [nota _]; clear pcl_step.
      rewrite forallb_forall in nota.
      specialize (all_nodes_prop n); intro n_in.
      rewrite InA_alt in n_in.
      destruct n_in as [n' [eqn in']].
      specialize (nota _ in').
      erewrite has_moved_b_compat in H.
      2: apply pg'.
      2: apply pg.
      2: apply eqn.
      erewrite alpha_enabled_bb_compat.
      2: apply pg.
      2: apply eqn.
      rewrite H in nota; simpl in nota; intuition.
      rewrite orb_true_iff in nota.
      destruct nota as [N|F]; 
        [rewrite <- negb_true_iff; assumption|
         contradict F; apply diff_false_true]. 
    Qed.

    Lemma exists_pcl_move:
      exists n, has_moved_b g' g n = true /\ pcl_enabled_bb g n = true.
    Proof.
      apply andb_prop in pcl_step.
      destruct pcl_step as [_ pstep]; clear pcl_step.
      rewrite existsb_exists in pstep.
      destruct pstep as [n [ _ ed]].
      exists n.
      rewrite andb_true_iff in ed; destruct ed.
      split; assumption.
    Qed.

    Lemma pcl_alpha_unchanged:
      forall n, alpha (g n) = alpha (g' n).
    Proof.
      intros n.
      case_eq (has_moved_b g' g n); intros M.
      assert (no := pcl_no_alpha_enabled_when_move _ M).
      unfold has_moved_b in M.
      revert M; destruct (State_dec (g' n) (g n)) as [M|M];
        try congruence; intros _.
      destruct step as [_ _ dif Hs].
      specialize (Hs n); destruct Hs as [Hs|Hs]; [| tauto].
      unfold RUN, run in Hs; simpl in Hs; unfold run_algorithm in Hs.
      unfold alpha_enabled_bb in no;  rewrite no in Hs; simpl in Hs. 
      destruct (parCl_enabled_b (g n) (local_env g n)).
      rewrite Hs; simpl; auto.
      destruct (hd_enabled_b (g n) (local_env g n)).
      rewrite Hs; simpl; auto.
      tauto.

      unfold has_moved_b in M.
      revert M; destruct (State_dec (g' n) (g n)) as [M|M];
        try congruence; intros _.
      destruct step as [_ _ dif Hs].
      specialize (Hs n); destruct Hs as [Hs|Hs].
      2: apply alpha_compat; symmetry; auto.
      apply alpha_compat; symmetry; auto.
    Qed.
    
    Lemma pcl_NewParCl_unchanged_:
      (forall n, alpha (g n) = alpha (g' n)) ->
      forall n, 
        NewparCl (g n) (local_env g n) == NewparCl (g' n) (local_env g' n).
    Proof.
      intros pa n.
      unfold NewparCl.
      rewrite <- (pa n).
      assert (eqP: forall n n', n == n' -> (Par (g n)) == (Par (g' n'))).
      intros nn nn' enn.
      symmetry in enn.
      assert (ro := Stable_Variable_step _ step _ _ enn).
      apply Par_compat; symmetry; apply ro.

      case (alpha (g n) <? k).
      apply eqP; reflexivity.
      case (alpha (g n) =? k).
      reflexivity.

      assert (ok := MinATall_only_depends_on_children
                      pg pg' step span_tree span_tree' n).
      assert (tmp: forall child : Node,
                 is_tree_parent_of (Par_part g) n child ->
                 alpha (g child) = alpha (g' child)).
      intros; apply pa.
      specialize (ok tmp); clear tmp.

      unfold MinIDMinATall. 
      
      assert (XX: equivlistA
                    equiv (filter (isMinATalls (g n) (local_env g n))
                                  (Local_Children
                                     (KCpeers (g n))
                                     (lPar (local_env g n)) (KCreverse (g n))))
                    (filter (isMinATalls (g' n) (local_env g' n))
                            (Local_Children
                               (KCpeers (g' n))
                               (lPar (local_env g' n)) (KCreverse (g' n))))).
      {
        apply filter_compat_equiv.
        intros c1 c2 eqc; unfold isMinATalls.
        rewrite <- ok.
        unfold local_env at 2 4.
        generalize (peer_compat _ _ (Equivalence_Reflexive n) _ _ eqc);
          destruct (peer n c1); destruct (peer n c2); simpl.
        2: intros F; elim F.
        2: intros F; elim F.
        2: tauto.
        intros en.
        rewrite <- pa, (alpha_compat (pg _ _ en)).
        reflexivity.
        apply Local_Children_Unchanged; auto.
      }

      destruct ((filter (isMinATalls (g n) (local_env g n))
                        (Local_Children
                           (KCpeers (g n))
                           (lPar (local_env g n)) (KCreverse (g n)))));
        destruct ((filter (isMinATalls (g' n) (local_env g' n))
                          (Local_Children (KCpeers (g' n))
                                          (lPar (local_env g' n))
                                          (KCreverse (g' n))))); auto.
      symmetry in XX; apply equivlistA_cons_nil in XX;
        [tauto| apply setoid_equiv]. 
      apply equivlistA_cons_nil in XX; [tauto| apply setoid_equiv]. 

      simpl.
      eapply fold_right_min_cons; try apply OrdChans; auto.
      Unshelve. apply Channel_dec.      
    Qed.

    Lemma pcl_NewParCl_unchanged:
      forall n, 
        NewparCl (g n) (local_env g n) == NewparCl (g' n) (local_env g' n).
    Proof.
      apply pcl_NewParCl_unchanged_.
      apply pcl_alpha_unchanged.
    Qed.

    Lemma alpha_enabled_unchanged_pcl:
      forall n, alpha_enabled_bb g n = alpha_enabled_bb g' n.
    Proof.
      intros n.
      unfold alpha_enabled_bb, alpha_enabled_b.
      rewrite pcl_alpha_unchanged.
      erewrite NewAlpha_only_depends_on_children; eauto.
      intros child _; apply pcl_alpha_unchanged.
    Qed.

    Lemma pcl_becomes_disabled:
      forall n, pcl_enabled_bb g n = true 
                ->  has_moved_b g' g n = true
                -> pcl_enabled_bb g' n = false.
    Proof.
      intros n pcl R.
      assert (na: alpha_enabled_bb g n = false). 
      apply pcl_no_alpha_enabled_when_move; assumption.
      
      unfold pcl_enabled_bb, parCl_enabled_b; simpl.
      case (eqoptionA_dec Channel_dec (parCl (g' n))
                          (NewparCl (g' n) (local_env g' n))).
      intros; reflexivity.
      rewrite <- (pcl_NewParCl_unchanged n).
      intros F; contradict F.
      unfold has_moved_b in R.
      revert R; destruct (State_dec (g' n) (g n)) as [M|M];
        try congruence; intros _.
      destruct step as [_ _ dif Hs].
      specialize (Hs n); destruct Hs as [RR|Hs]; [| tauto].
      unfold RUN, run in RR; simpl in RR;
        unfold run_algorithm in RR. 
      unfold alpha_enabled_bb in na; unfold pcl_enabled_bb in pcl;
        revert RR na pcl; destruct (alpha_enabled_b (g n) (local_env g n)).
      intros _ F; contradict F; apply diff_true_false.
      destruct (parCl_enabled_b (g n) (local_env g n)); [| congruence].
      intros H _ _; rewrite H; simpl; reflexivity.
    Qed.
    
    Lemma pcl_no_move:
      forall n, pcl_enabled_bb g n = false -> parCl (g' n) == parCl (g n).
    Proof.
      intros n npcl.
      destruct step as [_ _ dir r].
      specialize (r n).
      destruct r as [r|r].
      { unfold RUN, run in r; simpl in r; unfold run_algorithm in r. 
        revert r; destruct (alpha_enabled_b (g n) (local_env g n)).
        intros P; rewrite (parCl_compat P); simpl; reflexivity.
        unfold pcl_enabled_bb in npcl; rewrite npcl.
        destruct (hd_enabled_b (g n) (local_env g n));
          intros P; rewrite P; simpl; reflexivity.
      }
      rewrite r; reflexivity.
    Qed.

    Lemma pcl_stays_disabled:
      forall n, pcl_enabled_bb g n = false -> pcl_enabled_bb g' n = false.
    Proof.
      intros n npcl.
      unfold pcl_enabled_bb, parCl_enabled_b.
      case ( eqoptionA_dec Channel_dec (parCl (g' n))
                           (NewparCl (g' n) (local_env g' n))); intros np; auto.
      rewrite <- (pcl_NewParCl_unchanged n) in np.
      contradict np.
      rewrite (pcl_no_move n npcl).
      unfold pcl_enabled_bb, parCl_enabled_b in npcl.
      destruct (eqoptionA_dec Channel_dec (parCl (g n))
                              (NewparCl (g n) (local_env g n)));
        auto.
      contradict npcl; congruence. 
    Qed.

  End With_Pcl_Step.

  (** ** PCL Criteria *)
  Section Pcl_Termination_Criteria.

    Variable (sg sg': sig (Pspanning_tree root))
             (step: Pcl_SafeStep sg' sg).

    Lemma Pcl_step_b: pcl_step_b (proj1_sig sg') (proj1_sig sg) = true.
    Proof. apply step. Qed.
    Lemma Pcl_step: Step (proj1_sig sg') (proj1_sig sg).
    Proof. apply step. Qed.

    Lemma global_pcl_criterium:
      exists n, pcl_potential sg n <> pcl_potential sg' n.
    Proof.
      destruct (exists_pcl_move _ _ Pcl_step_b) as [n [ds pcl]].
      exists n.
      unfold pcl_potential.
      rewrite pcl.
      assert (x := pcl_becomes_disabled
                     (proj1 (proj2_sig sg)) (proj1 (proj2_sig sg'))
                     Pcl_step Pcl_step_b
                     (proj2 (proj2_sig sg))
                     (proj2 (proj2_sig sg'))
                     n pcl).
      rewrite x; auto.
    Qed.  

    Lemma local_pcl_criterium:
      forall n,
        (pcl_potential sg n) <> (pcl_potential sg' n) -> 
        gt (pcl_potential sg n) (pcl_potential sg' n).
    Proof.
      unfold pcl_potential.
      intros n.
      case_eq (pcl_enabled_bb (proj1_sig sg) n); 
        case_eq (pcl_enabled_bb (proj1_sig sg') n); intros pcl' pcl H.
      contradict H; reflexivity.
      auto with arith.
      assert (x := pcl_stays_disabled
                     (proj1 (proj2_sig sg)) (proj1 (proj2_sig sg'))
                     Pcl_step Pcl_step_b
                     (proj2 (proj2_sig sg))
                     (proj2 (proj2_sig sg'))
                     n pcl).
      contradict pcl'; congruence.
      lia.
    Qed.

  End Pcl_Termination_Criteria.

  (** ** PCL Termination *)
  Section Pcl_Termination.

    Notation ltM := (@MultisetNat.MSetOrd.MultisetLT gt).
    Notation Pot_pcl := (Pot pcl_potential).

    Lemma pcl_safe_inclusion:
      inclusion (Pcl_SafeStep (root := root)) (ltM @@ Pot_pcl).
    Proof.
      apply Step_decreases_pot.  
      apply pcl_potential_compat.

      (* local criterion *)
      intros sg sg' step n Hgt.
      case (eq_nat_dec (pcl_potential sg n) (pcl_potential sg' n)); intros Hneq.
      rewrite Hneq in Hgt.
      contradict Hgt; apply gt_irrefl. 
      unfold safeStep in step.
      generalize step; intros Htrans'; destruct Htrans'.
      contradict Hgt.
      assert (Hneq': pcl_potential sg n <> pcl_potential sg' n).
      apply Hneq. clear Hneq.

      assert (XX := local_pcl_criterium step n). 
      specialize (XX Hneq').
      apply lt_asym. apply XX.

      (* global criterion *)
      intros sg sg' step.
      assert (XX := global_pcl_criterium step). 
      destruct XX as [n XX].
      exists n.
      auto.
    Qed.

  End Pcl_Termination.

End KClustering_Pcl_Termination.

Close Scope signature_scope.
Close Scope Z_scope.
Unset Implicit Arguments.

