(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** **** Global Import **)
From Coq Require Import ZArith.
From Coq Require Import Lia.
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Bool.
From Coq Require Import RelationPairs.
From Coq Require Import Wellfounded.Union.
From Coq Require Import Morphisms_Prop.

(** **** Local Import **)
From Padec Require Import SetoidUtil.
From Padec Require Import OptionUtil.
From Padec Require Import ListUtils.
From Padec Require Import RelationA.
From Padec Require Import Count.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Fairness.
From Padec Require Import Self_Stabilization.
From Padec Require Import KClustering_specification.
From Padec Require Import KClustering_algo.
From Padec Require Import Spanning_Tree.
From Padec Require Import KClustering_def_termination.
From Padec Require Import P_Q_Termination.
From Padec Require Import KClustering_alpha_termination.
From Padec Require Import KClustering_pcl_termination.
From Padec Require Import KClustering_hd_termination.
From Padec Require Import KClustering_correctness_alpha.
From Padec Require Import KClustering_correctness.
From Padec Require Import KClustering_count.

(* end show *)
Open Scope Z_scope.
Set Implicit Arguments.

(** * Proof of the algorithm **)
Section KClustering_Specification_Proved.

  Context {Chans: Channels} {Net: Symmetric_Network}
          {KCP: KClustering_parameters} {KCT: KC_algo_tools}.

  Existing Instance KC_Stable.
  Existing Instance OrdChans.
           
  Notation Env := (Env (Algo := KC_algo)).

  Let k_pos := k_pos.

  (** ** Input Assumptions:
   Finite symmetric topology and Spanning tree assumption *)  
  Definition assumptions_RO ro: Prop :=
    (forall n, equivlistA equiv (KCpeers_reverse (ro n))
                          (build_symlinks Net n)) /\
    Spanning_Tree_Network_assumption net (fun n => Par (ro n))
    /\ forall n1 n2, id (ro n1) == id (ro n2) 
                     -> n1 == n2.

  Instance assumptions_RO_compat: Proper fequiv assumptions_RO.
  Proof.
    intros ro1 ro2 ero; unfold assumptions_RO.
    apply and_compat; try tauto.
    apply all_compat.
    intros x;
      rewrite (KCpeers_reverse_compat (ero _ _ (Equivalence_Reflexive x)));
      tauto.
    apply and_compat.
    apply Spanning_Tree_Network_assumption_compat; auto.
    intros n1 n2 en; apply Par_compat, ero; auto.
    split; intros H n1 n2.
    rewrite <- (id_compat (ero _ _ (Equivalence_Reflexive n1))).
    symmetry in ero; rewrite (id_compat (ero _ _ (Equivalence_Reflexive n2))).
    auto.
    rewrite (id_compat (ero _ _ (Equivalence_Reflexive n1))).
    symmetry in ero;
      rewrite <- (id_compat (ero _ _ (Equivalence_Reflexive n2))).
    auto.
  Qed.
    
  Lemma assume_ensures_Symmetric_Network:
    forall roenv,
      assumptions_RO roenv ->
      Symmetric_Network_assumption
        Net (fun n => KCpeers (roenv n))
        (fun n => KCreverse (roenv n)).
  Proof.
    intros ro sym.
    apply build_links_ensures_Symmetric_Network_Assumption_equivlist.
    apply sym.
  Qed.

  Instance KC_Assumption:
    Stable_Assumption (Algo := KC_algo) (Net := net).
  Proof.
    esplit.
    apply assumptions_RO_compat.
  Defined.


  Section Termination.

    Section With_root_sym.

      Variable (root: Node).

      Notation ltM := (@MultisetNat.MSetOrd.MultisetLT gt).

      (** ** Global Termination *)
      Notation SafeStep := (safeStep (safe := Pspanning_tree root)).
      Lemma KClustering_WF: well_founded SafeStep.
      Proof.
        assert (HNN: exists (N: nat), Num_Card Same Node N).
        { eapply Classical_set_Num_Card.
          intros a1 a2;destruct (Node_dec a1 a2);[left|right];assumption.
          apply all_nodes_prop.
        }
        destruct HNN as [N HN].
        assert (HNN: forall (env : sig (Pspanning_tree root)) (n : Node),
                   (Z_of_nat N) + 1 > dist_hd' env n).
        {
          intros g n. 
          
          assert (pR:Proper fequiv (pcl_chain (proj1_sig g))).
          unfold pcl_chain. intros pn1 pn2 hpn nn1 nn2 hnn.
          apply equiv_compat; auto.
          apply get_pcl_compat;auto.
          apply (proj1(proj2_sig g)).
          
          pose (P:=fun _:Node => True).
          assert (P_compat: Proper fequiv P).  
          hnf;intuition.
          assert (P_dec:forall n,{P n}+{~P n}).
          unfold P;intuition.
          assert (PN : Num_Card Same (sig P) N).
          eapply Card_trans with Node _.
          apply Trivial_subset_Num_Card.
          apply HN.
          assert (pcl_P : forall n0 l, chain (pcl_chain (proj1_sig g)) n0 l n
                                       -> P n0).
          unfold P;intuition.  
          (* assert (pcl_NP : forall n0, ~P n0 -> exists l,chain eqN
          (pcl_chain (getEnv g)) n l n0). *)
          (*       unfold P;intuition. *)
          clearbody P;clear HN.
          revert N n P P_compat P_dec pcl_P (* pcl_NP *) PN.
          induction N as [|N IHN];intros n P P_compat P_dec pcl_P
                                         (* pcl_NP *) PN.
          
          (* cas N=0 *)
          
          destruct PN as [ [I Ip Iu It] [S Sp Su St] ];
            unfold left_unique, left_total in *. 
          specialize (pcl_P n nil (Equivalence_Reflexive n)).
          specialize (It (exist pcl_P)
                         (exist pcl_P) (Equivalence_Reflexive n) ).
          destruct It as [[b0 Hb0] _]; lia.

          (* cas sucesseur *)
          
          pose (P':= (fun n0 => P n0 /\ ~ n == n0)).

          assert (P'_compat:Proper fequiv P').
          intros n1 n2 en;apply Morphisms_Prop.and_iff_morphism.
          apply P_compat;auto.
          rewrite en;reflexivity.

          assert (P'_dec:forall n0 : Node, {P' n0} + {~ P' n0}).
          intros n1;unfold P';destruct (P_dec n1) as [Pyes|Pno].
          destruct (Node_dec n n1) as [eyes|eno].
          right;tauto.
          left;tauto.
          right;tauto.
          unfold dist_hd'.
          rewrite (dist_hd_compat
                     (hd_pcl_chain_wf g n)
                     (Acc_intro n (fun y hy => (hd_pcl_chain_wf g y)))).
          2:destruct g as [g [pg hg]];simpl;intros;apply pg;auto.
          2:apply (Equivalence_Reflexive n).

          simpl dist_hd.
          case_equiv (get_pcl (proj1_sig g) n);[intro n0|];
            intro E;symmetry in E.      
          2:lia.
          change (pcl_chain (proj1_sig g) n0 n) in E.
          change (Z.of_nat (S N) + 1 >  1+ dist_hd' g n0  ).
          rewrite Zplus_comm.
          apply Zplus_gt_compat_l.
          rewrite Nat2Z.inj_succ.
          rewrite <- Z.add_1_r.


          apply IHN with P';clear IHN;try assumption.


          (* chain -> P' *)

          intros n1 l hn1;split.
          apply (pcl_P n1 (l++(n::nil))). 
          clear - n n1 n0 l hn1 E.
          revert n1 hn1.
          induction l as [|n1 q hq];
            intros n2 hn2; simpl in hn2 |- *.
          split;[|reflexivity].
          unfold pcl_chain.
          rewrite hn2.
          apply E.
          intuition.
          intro e.
          assert (hn:chain (pcl_chain (proj1_sig g)) n l n0).
          revert hn1;apply chain_compat;auto;try reflexivity.      
          clear hn1.
          assert (hh: chain (pcl_chain (proj1_sig g)) n (l++(n::nil)) n).
          apply chain_app;auto.
          split with n0;intuition.
          simpl;intuition.
          assert (h:=
                    WellfoundedUtil.wf_rel_has_non_reflexive_transitive_closure
                      (hd_pcl_chain_wf g)).
          apply (h n);clear h.
          rewrite transitive_closure_iff_exists_non_empty_chain.
          exists l.
          revert hh.
          apply chain_cons_non_empty.

          (* Cardinal *)
          clear E n0.
          specialize (pcl_P n nil (Equivalence_Reflexive n)).
          assert (count_n := Singleton_Num_Card
                               (fun n0:Node => n == n0) n
                               (Equivalence_Reflexive n) (fun x h => h)).
          assert (hh1:forall a : Node, P' a \/ ~ P' a).
          intro a; destruct (P'_dec a);[left|right];assumption.
          assert (hh2:forall a1 a2 : Node, a1 == a2 \/ ~ a1 == a2).
          intros a1 a2; destruct (Node_dec a1 a2);[left|right];assumption.

          destruct (@Classical_subset_Num_Card
                      _ _ hh2 P' P'_compat hh1 all_nodes) as [NN HNN].
          intros a _; apply all_nodes_prop.
          assert (n_compat:Proper fequiv (fun n0 : Node => n == n0)).
          intros n1 n2 hn12;rewrite hn12;reflexivity.
          assert (disjoint:forall x : Node,
                     P' x -> (fun n0 : Node => n == n0) x -> False).
          unfold P';intuition.
          assert (hh:= @Disjoint_union_Num_Card
                         Node _ _ P'_compat _ n_compat _ HNN _ count_n
                         disjoint).
          assert (hequiv: 
                    (fun x : Node => P' x \/
                                     (fun n0 : Node => n == n0) x) =~= P).
          intros n1 n2 hn12;unfold P';simpl;rewrite <- hn12;revert pcl_P;
            destruct (Node_dec n n1) as [e|ne];[rewrite e|];
              generalize (Equivalence_Reflexive (R:=equiv) n1);tauto.
          assert (hhh:= @Inj_iff
                          _ _ (fun x : Node => P' x \/
                                               (fun n0 : Node => n == n0) x)
                          P P_compat hequiv).
          assert (e:(NN+1=S N)%nat). 
          rewrite <-  Card_same_eq_iff.
          eapply Card_trans with {x : Node | P' x \/
                                             (fun n0 : Node => n == n0) x} _;
            auto.
          eapply Card_sym.
          apply hh.
          eapply Card_trans with {x : Node | P x} _; auto.
          rewrite NatUtil.plus_1_S in e.
          injection e;clear e;intro e.
          rewrite e in HNN.
          apply HNN.
        }
        
        (* Union *)
        Notation Trans_union := (@Trans_union _ _ _ _ root).
        rewrite (well_founded_morphism _ _ Trans_union).         
        eapply union_lex_wf3.
        apply alpha_safe_inclusion; assumption.
        apply pcl_safe_inclusion; assumption.
        apply Inclusion.wf_incl
          with (ltM @@ Pot (hd_pot (Z.of_nat N+1)))%signature.
        apply hd_safe_inclusion.
        apply HNN.
        apply (Inverse_Image.wf_inverse_image
                 _ _ ltM (Pot (hd_pot (Z.of_nat N+1)))).
        apply MultisetNat.MSetOrd.mOrd_wf.
        cbv; clear; intros; congruence.
        apply lt_wf.
      
        (* Priorities *)
        (* ** alpha potential unchanged when pcl or hd step *)
        { intros sg sg' STEP.
          assert (step: Step (proj1_sig sg') (proj1_sig sg)).
          destruct STEP as [STEP|STEP];
            [unfold  Pcl_Safe in STEP| unfold Head_Safe in STEP]; 
            destruct STEP as [STEP _]; apply STEP. 
          assert (U: alpha_potential sg =~= alpha_potential sg').
          2:{
            left; unfold P_Q_Termination.Pot.
            apply eq_meq.
            eapply map_compat.
            apply U.
            reflexivity.
          }
          (* U *)
          intros n1 n2 en.
          assert (dum: Proper pequiv sg)
            by apply step; rewrite en; clear dum.
          unfold alpha_potential, potential.
          rewrite <- depth_preserved_step
            with  
              (pg:= (proj1 (proj2_sig sg))) (span_tree:= (proj2 (proj2_sig sg)))
              (pg':= (proj1 (proj2_sig sg')))
              (span_tree':= (proj2 (proj2_sig sg'))); try assumption. 
          destruct STEP as [STEP|STEP].
          rewrite alpha_enabled_unchanged_pcl
            with (root:=root) (g:=(proj1_sig sg))(g':=(proj1_sig sg'));
            try assumption.
          reflexivity.
          apply (proj1 (proj2_sig sg)).
          apply (proj1 (proj2_sig sg')).
          apply STEP.
          apply (proj2_sig sg).
          apply (proj2_sig sg').
          rewrite alpha_enabled_unchanged_hd
            with (root:=root)(g:=(proj1_sig sg))(g':=(proj1_sig sg'));
            try assumption.
          reflexivity.
          apply STEP.
          apply (proj2_sig sg).
          apply (proj2_sig sg').
        }
        (* ** pcl potential unchanged when hd step *)
        { intros sg sg' STEP.
          assert (step: Step (proj1_sig sg') (proj1_sig sg)).
          unfold Head_SafeStep in STEP;
            destruct STEP as [STEP _]; apply STEP. 
          assert (U: pcl_potential sg =~= pcl_potential sg').
          2:{
            unfold P_Q_Termination.Pot.
            left; apply eq_meq.
            eapply map_compat.
            apply U.
            reflexivity.
          }
          (* U *)
          intros n1 n2 en.
          assert (dum: Proper pequiv sg)
            by apply step; rewrite en; clear dum.
          unfold pcl_potential.
          rewrite pcl_enabled_unchanged
            with (root:=root)(g:=(proj1_sig sg)) (g':=(proj1_sig sg'));
            try assumption.
          reflexivity.
          apply STEP.
          apply (proj2_sig sg).
          apply (proj2_sig sg').
        }
      Qed.

    End With_root_sym.

    Lemma KClustering_termination:
      termination KC_Assumption. 
    Proof.
      intros g assume pg.
      dup assume ass; destruct ass as [sym [[root span] unique]].
      eapply Acc_Algo_Multiset. 
      2: apply KClustering_WF.
      apply spanning_tree_stable.
      split; [| split]; auto.
      apply span.
      apply assume_ensures_Symmetric_Network; auto.
    Qed.

  End Termination.

  (** ** Correctness *) 
  Section Correctness.

    Section Counting.
      
      Lemma counting_OK_OK:
        forall (g: Env), Assume g -> terminal g -> counting_OK kCluster g. 
      Proof.
        intros g assume term.
        assert (symm := assume_ensures_Symmetric_Network assume). 
        destruct assume as [sym [[root span] unique]].
        assert (pg: Proper pequiv g) by apply term.
        assert (disabled_alpha: forall n, alpha_enabled_bb g n = false).
        { intros n.
          unfold alpha_enabled_bb, alpha_enabled_b.
          destruct (state term n) as [A _]; revert A.
          destruct (Z.eq_dec (alpha (g n)) (NewAlpha (g n) (local_env g n)));
            [auto|tauto].
        }
        destruct (counting_theorem pg span symm disabled_alpha)
          as [N [D [HN [HD const]]]]; auto.
        exists N.
        exists D.
        split; auto.
        split; auto.
        eapply Inj_iff2; auto.
        intros n n' en; apply kCluster_compat; auto.
        2: apply HD.
        intros n n' en.
        rewrite <- clusterHead_is_kDominator; try assumption.
        apply eq_compat; auto.
        apply kDominator_compat; try assumption.
        apply span. 
      Qed.

    End Counting.

    (** The specification is proven for: *)
    Notation kClustering_SPEC_strong :=
      (kClustering_SPEC_strong kCluster
                               (** ids in the state *)
                               id
                               (** the local information clusterHeadID
                               provides the identifier of the
                               clusterhead *) 
                               clusterHeadID
                               (** the local information clusterParent
                               provides the pointer to build a route
                               to the clusterhead*)
                               clusterParent).

    Lemma kClustering_partial_correctness:
      P_correctness KC_Assumption kClustering_SPEC_strong. 
    Proof.
      unfold P_correctness.
      intros g assume term.
      assert (symm := assume_ensures_Symmetric_Network assume). 
      assert (A := assume).
      destruct A as [sym [[root span] unique]].
      rewrite kClustering_SPEC_strong_alt.
      split; split.
      apply clusterHeadID_OK_OK; auto.
      eapply cluster_path_OK_OK with root; auto. 
      eapply kCluster_partition_OK with root; assumption.
      apply counting_OK_OK; assumption.
    Qed.

  End Correctness.

  (** ** The algorithm is self-stabilizing and silent *) 
  Section Silent_Self_Stabilization.

    (** Self Stabilization is proven for: *)
    Notation kClustering_SPEC_strong :=
      (kClustering_SPEC_strong kCluster
                               (** ids in the state *)
                               id
                               (** the local information clusterHeadID
                               provides the identifier of the
                               clusterhead *) 
                               clusterHeadID
                               (** the local information clusterParent
                               provides the pointer to build a route
                               to the clusterhead*)
                               clusterParent).

    Theorem kclustering_silent_self_stabilization:
      silence KC_Assumption unfair_daemon /\
      self_stabilization KC_Assumption unfair_daemon
                (only_one kClustering_SPEC_strong). 
    Proof.
      cut (silence KC_Assumption unfair_daemon).
      { intros S; split; auto.
        apply silent_self_stabilization; auto.
        apply kClustering_partial_correctness.
      }
      apply convergence_silence.
      apply termination_is_silence.
      apply KClustering_termination.
    Qed.

  End Silent_Self_Stabilization.

End KClustering_Specification_Proved.

Close Scope Z_scope.
Unset Implicit Arguments.
