(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import ZArith.
From Coq Require Import Lia.

(** **** Local Imports *)
From Padec Require Import OptionUtil.
From Padec Require Import ZUtils.
From Padec Require Import ListUtils.
From Padec Require Import SetoidUtil.
From Padec Require Import FoldRight.
From Padec Require Import Tree_topology.
From Padec Require Import Algorithm.

(* end show *)
Open Scope Z_scope.
Open Scope signature_scope.
Set Implicit Arguments.

(** * Definition of alpha and its computation *)
Section KDomSet_Algo.

  Context {Chans: Channels}.

  Class KDomSet_parameters :=
    {
      (** ** The parameter K itself (from K-Dominating-Set) *)   
      k: Z;
      k_pos: k > 0
    }.
  Context {KDP: KDomSet_parameters}.
  Let k_pos := k_pos.

  (** ** State of a Node in the Algorithm *)
  
  (** *** Read Only Part of a State
       It stores local information about the tree topology:
       - list of children
   *)
  Record KDom_ROState := mkROState {
                             is_root: bool;
                             children_c: list Channel;
                           }.

  Notation eqROS := (fun (x y: KDom_ROState) =>
                       is_root x == is_root y /\
                       equivlistA equiv (children_c x) (children_c y)).

  Instance eqROS_equiv: Equivalence eqROS.
  Proof.
    split; hnf; intuition.
    etransitivity; eauto.
    etransitivity; eauto.
  Qed.

  Global Instance KDom_ROState_setoid: Setoid KDom_ROState :=
    {| equiv := eqROS |}.

  (** *** State of a Node
      - the alpha value (an integer)
      - the read-only part defined above
   *)

  Record KDom_State := mkState {
                           alpha: Z;
                           KDom_ROpart:> KDom_ROState;
                         }.

  Notation eqS := (fun (x y: KDom_State) =>
                     alpha x == alpha y /\
                     KDom_ROpart x == KDom_ROpart y).

  Instance eqS_equiv: Equivalence eqS.
  Proof.
    split; hnf; intuition;
      etransitivity; eauto.
  Qed.

  Global Instance KDom_State_setoid: Setoid KDom_State :=
    {| equiv := eqS |}.

  (** *** Equality and Compatibility properties on State *)

  Lemma KDom_ROState_dec: Decider (equiv (A := KDom_ROState)).
  Proof.
    intros [i1 c1] [i2 c2]; simpl equiv.
    destruct (Bool.bool_dec i1 i2); [|right; intuition].
    destruct (equivlistA_dec Channel_dec c1 c2);
      [left|right]; intuition.
  Qed.

  Lemma KDom_State_dec: Decider (equiv (A := KDom_State)).
  Proof.
    intros [a1 ro1] [a2 ro2]; simpl equiv.
    destruct (KDom_ROState_dec ro1 ro2) as [h|h];
      simpl equiv in h; [| right; tauto].
    destruct (Z.eq_dec a1 a2);[left|right]; tauto. 
  Qed.

  Instance alpha_compat: Proper fequiv alpha.
  Proof.  intros [a1 r1] [a2 r2] [ha hr]; intuition. Qed.

  Instance is_root_compat_RO: Proper fequiv is_root.
  Proof.
    intros [i1 c1] [i2 c2] [hi hc]; simpl; intuition.
  Qed.

  Instance children_c_compat_RO:
    Proper (equiv ==> equivlistA equiv) children_c.
  Proof.
    intros [i1 c1] [i2 c2] [hi hc]; simpl; intuition.
  Qed.

  Instance KDom_ROpart_compat: Proper fequiv KDom_ROpart.
  Proof.
    intros [p1] [p2] hr; simpl in *; intuition.
  Qed.

  Instance is_root_compat:
    Proper (equiv (A := KDom_State) ==> equiv) is_root.
  Proof.
    intros s1 s2 es; apply is_root_compat_RO;
      apply KDom_ROpart_compat; auto.
  Qed.

  Instance children_c_compat:
    Proper (equiv (A := KDom_State) ==> equivlistA equiv) children_c.
  Proof.
    intros s1 s2 es; apply children_c_compat_RO;
      apply KDom_ROpart_compat; auto.
  Qed.

  Section Run_definition.

    (** ** Run Function for the Algorithm
     *)

    Definition _IsShort _alpha := (_alpha <? k).

    Instance _IsShort_compat: Proper fequiv _IsShort.
    Proof. intros a b ab; rewrite ab; reflexivity. Qed.
    
    Definition _IsTall _alpha  := (_alpha >=? k).

    Instance _IsTall_compat: Proper fequiv _IsTall.
    Proof. intros a b ab; rewrite ab; reflexivity. Qed.

    Definition min_Atall l :=
      List.fold_right Z.min (2 * k + 1) (List.filter _IsTall l).

    Instance min_Atall_compat:
      Proper (equivlistA equiv ==> equiv) min_Atall.
    Proof.
      intros l1 l2 el; unfold min_Atall.
      apply fold_right_min_compat; auto.
      apply filter_compat_equiv; auto.
      apply _IsTall_compat.
    Qed.
    
    Definition max_Ashort l :=
      List.fold_right Z.max (-1) (List.filter _IsShort l).

    Instance max_Ashort_compat:
      Proper (equivlistA equiv ==> equiv) max_Ashort.
    Proof.
      intros l1 l2 el; unfold min_Atall.
      apply fold_right_max_compat; auto.
      apply filter_compat_equiv; auto.
      apply _IsShort_compat.
    Qed.

    Definition new_alpha (alpha_children: list Z): Z :=
      let max := max_Ashort alpha_children in
      let min := min_Atall alpha_children in
      if Z_le_gt_dec (max + min + 2) (2 * k)
      then min + 1 else max + 1.

    Instance new_alpha_compat:
      Proper (equivlistA equiv ==> equiv) new_alpha.
    Proof.
      intros l1 l2 el; unfold new_alpha.
      case (Z_le_gt_dec (max_Ashort l1 + min_Atall l1 + 2) (2 * k));
        case (Z_le_gt_dec (max_Ashort l2 + min_Atall l2 + 2) (2 * k)); intros h1 h2.
      rewrite el; auto.
      rewrite el in h2; lia.
      rewrite el in h2; lia.
      rewrite el; auto.
    Qed.

    Definition alphas (s_children: list KDom_State): list Z := map alpha s_children.

    Instance alphas_compat: Proper (equivlistA equiv ==> equivlistA equiv) alphas.
    Proof.
      intros l l' el.
      unfold alphas. 
      apply map_compat_equiv; auto.
      apply alpha_compat.
    Qed.

    Definition children_states (s: KDom_ROState)
               (peer_state: Channel -> option KDom_State): list KDom_State :=
      map_filter peer_state (children_c s).

    Instance children_states_compat: Proper (equiv ==> fequiv ==> equivlistA equiv) children_states.
    Proof.
      intros s s' es l l' el; unfold children_states.
      apply map_filter_compat; auto.
      apply children_c_compat_RO; auto.
    Qed.

    Lemma children_states_depends:
      forall s1 s2,
        equivlistA equiv (children_c s1) (children_c s2) ->
        forall ps1 ps2,
          Proper pequiv ps1 -> Proper pequiv ps2 ->
          (forall c, InA equiv c (children_c s1) -> ps1 c == ps2 c) ->
          equivlistA equiv (children_states s1 ps1) (children_states s2 ps2).
    Proof.
      intros s1 s2 es ps1 ps2 eps; unfold children_states.
      apply map_filter_eq_locality; auto.
    Qed.      
      
    Definition run_algorithm
               (s: KDom_State)
               (peer_state: Channel -> option KDom_State)
      : KDom_State :=
      {| alpha := new_alpha (alphas (children_states s peer_state));
         KDom_ROpart := s |}.


  End Run_definition.

  (** ** Properties of the Algorithm *)

  (** *** Compatibility of the Run Function *)

  Instance run_algorithm_compat: Proper fequiv run_algorithm.
  Proof.
    intros s s' es  ps ps' eps.
    unfold run_algorithm; split.
    apply new_alpha_compat, alphas_compat; auto.
    apply map_filter_compat; auto.
    apply children_c_compat; auto.
    simpl; split; [apply is_root_compat | apply children_c_compat]; auto.
  Qed.

  (** *** Stability for the Read-Only Part of the Run Function *)
  Lemma ROpart_stable:
    forall s sn, KDom_ROpart (run_algorithm s sn) =~= KDom_ROpart s.
  Proof.
    intros [a c] sn.
    unfold run_algorithm; simpl.
    split; auto.
    reflexivity.
  Qed.

  (** ** Definition of The Algorithm *)
  Instance KDom_algo: Algorithm :=
    {| State := KDom_State; State_dec := KDom_State_dec; run := run_algorithm; |}.

  Global Instance KDom_Stable: Stable_Variable KDom_algo KDom_ROpart.
  Proof.
    esplit.
    apply KDom_ROpart_compat.
    apply ROpart_stable.
  Defined.
  
  Section Short_Or_Tall.

    Lemma short_or_tall: forall alpha,
      {_IsShort alpha = true} + {_IsTall alpha = true}.
    Proof.
      intros alpha; unfold _IsShort, _IsTall.
      destruct (Z_lt_le_dec alpha k) as [h | h];[left|right];lia.
    Qed.

  End Short_Or_Tall.

  (** ** alpha is always between 0 and 2k *)
  Section Range_of_alpha.

    Lemma new_alpha_0_2k:
      forall alphas, 0 <= new_alpha alphas <= 2 * k.
    Proof.
      intros alphas; unfold new_alpha.
      assert (MIN: k <= min_Atall alphas <= 2*k+1).
      { unfold min_Atall.
        destruct (fold_right_Zmin_dec (2 * k + 1) 
                                      (filter _IsTall alphas)) as [eqa | ina]; unfold Z.eq in *.
        - rewrite eqa; lia. 
        - rewrite filter_InA in ina.
          2: apply _IsTall_compat.
          destruct ina as [ina2 tall].
          unfold _IsTall in tall.
          rewrite Z.geb_le in tall; auto.
          split; auto.
          apply fold_right_Zmin_eqA_leA; auto.
      }
      assert (MAX: -1 <=  max_Ashort alphas <= k-1).
      {
        unfold max_Ashort.
        destruct (fold_right_Zmax_dec (-1) (filter _IsShort alphas)) as [eqa | ina];
          unfold Z.eq in *.
        - rewrite eqa; lia.
        - rewrite <- Z.ge_le_iff.
          split; auto.
          apply fold_right_Zmax_eqA_leA.
          reflexivity.
          rewrite filter_InA in ina.
          2: apply _IsShort_compat.
          destruct ina as [ina2 tall].
          unfold _IsShort in tall.
          rewrite <- Zlt_is_lt_bool in tall.
          unfold _IsShort.
          lia.
      }
      destruct
      (Z_le_gt_dec (max_Ashort alphas + min_Atall alphas + 2)
                   (2 * k)); lia.
    Qed.

  End Range_of_alpha.

End KDomSet_Algo.

(** *** Common definitions and results ****************** *)
From Padec Require Import RelModel.
(** ***************************************************** *)

Section KDom_Assume.

  Context {Chans: Channels}.
  
  Context {Net: Network}
          {DTN: Down_Tree_Network Net}.

  Context {KDP: KDomSet_parameters}.
  Let k_pos := k_pos.

  Global Instance KDom_RO_assumption:
    Stable_Assumption (Algo := KDom_algo) (Net := Net).
  Proof.
    esplit.
    intros ro1 ro2 ero.
    apply Down_Tree_Network_assumption_compat. 
    intros n1 n2 en; apply children_c_compat_RO, ero, en.
    intros n1 n2 en; apply is_root_compat_RO, ero, en.
  Defined.
  
  Notation Env := (Env (Net := Net) (Algo := KDom_algo)).
  
  Lemma children_c_equiv: forall (env: Env) (n: Node), 
      Assume env ->
      equivlistA equiv (children_c (env n)) (DTN_children n).
  Proof.
    intros env n [henv _].
    apply (henv _ _ (Equivalence_Reflexive n)).
  Qed.
  
  Lemma is_root_equiv: forall (env: Env) n, 
      Assume env -> is_root (env n) == DTN_root n.
  Proof.
    intros env n [_ henv].
    apply (henv _ _ (Equivalence_Reflexive n)).
  Qed.
  
  Notation alphas_ :=
    (fun (env: Env) n => alphas (children_states (env n) (local_env env n))).
  
  Notation new_alpha_ := (fun env n => new_alpha (alphas_ env n)).

  Section KDom_Graph.
    
    (** ** kDomSet definitions *)
    Definition is_kDom_edge (env: Env) (x y: Node): Prop := 
      match alpha (env y) ?= k with
      | Lt => (isParent x y)
      | Eq => False
      | Gt => (isParent y x) /\ alpha (env x) = alpha (env y) - 1
      end.
    
    Global Instance is_kDom_edge_compat: Proper fequiv is_kDom_edge.
    Proof.
      intros e1 e2 ee n1 n2 en n'1 n'2 en'; unfold is_kDom_edge.
      case_eq (alpha (e1 n'1) ?= k);
        rewrite (alpha_compat (ee _ _ en'));
        intros I; rewrite I; auto.
      rewrite <- (isParent_compat _ _ en _ _ en'); auto.
      rewrite <- (isParent_compat _ _ en' _ _ en);
        rewrite (alpha_compat (ee _ _ en)); auto.
    Qed.

    Definition kDominator (env: Env) (p: Node) := 
      orb (Z.eqb (alpha (env p)) k)
          (andb (_IsShort (alpha (env p))) (is_root (env p))).
    
    Global Instance kDominator_compat: Proper fequiv kDominator.
    Proof.
      intros e1 e2 ee p1 p2 ep; unfold kDominator. 
      rewrite (alpha_compat (ee _ _ ep)).
      rewrite (is_root_compat (ee _ _ ep)).
      auto.
    Qed.
    
  End KDom_Graph.
  
  (* *************************************************************)
  (** ** Assume that the computation of alpha provides no new value *)
  (** relation on the values of alpha between children and parent  *)
  Section Alpha_child_parent.
    
    (** Assumptions on Inputs *)
    Variable (env: Env)
             (env_compat: Proper pequiv env)
             (hassume: Assume env).
    
    (** Local properties of alpha, based on is_tree_parent *)
    Lemma InA_alphas: 
      forall (n: Node) (a: Z), 
        (exists (m: Node), isParent n m /\ alpha (env m) = a)
        <-> InA equiv a (alphas_ env n).
    Proof.
      intros n a.
      split.
      { intros [m [nm eqa]].
        unfold alphas.
        apply InA_map_iff; [apply alpha_compat|].
        exists (env m).
        split; auto.
        apply InA_map_filter_iff.
        intros aa b ab; rewrite ab; auto.
        unfold isParent in nm.
        apply DTN_children_ok in nm.
        destruct nm as [c [inc incc]].
        exists c; split.
        rewrite children_c_equiv; auto.
        unfold local_env; unfold is_channel in incc;
          rewrite incc; auto.
      }
      { intros ina.
        unfold alphas in ina.
        apply InA_map_iff in ina; [|apply alpha_compat].
        destruct ina as [sm [hsm insm]].
        unfold children_states in insm.
        apply InA_map_filter_iff in insm.
        2: intros aa b ab; rewrite ab; auto.
        destruct insm as [c [hc hcc]].
        unfold local_env in hcc.
        case_eq (peer n c); [intros m|]; intros hm; rewrite hm in hcc.
        2: inversion hcc.
        exists m; split.
        { apply DTN_children_ok.
          exists c; split; [|unfold is_channel; rewrite hm]; auto.
          rewrite children_c_equiv in hc; auto. }
        { simpl option_map in hcc.
          inversion hcc.
          rewrite <- hsm, H; auto. }          
      }
    Qed.
      
    Lemma new_alpha_inv: 
      forall (n: Node),
        alpha (env n) = new_alpha_ env n ->
        alpha (env n) > 0 ->
        exists m, isParent n m /\ alpha (env n) = alpha (env m) + 1.
    Proof.
      intros n same pos.
      unfold new_alpha in same.
      remember (max_Ashort (alphas_ env n)) as max.
      remember (min_Atall (alphas_ env n)) as min.
      revert same; case_eq (Z_le_gt_dec (max + min + 2) (2 * k));
        intros le'  eq' same.
      {
        unfold min_Atall in Heqmin.
        remember (filter _IsTall (alphas_ env n)) as tas.
        remember (2 * k + 1) as x0.
        destruct (fold_right_Zmin_dec x0 tas) as [fdec | fdec];
          simpl in fdec.
        - rewrite fdec in Heqmin.
          assert (le'' := le').
          rewrite Heqmin in le''.
          rewrite Heqx0 in le''.
          assert (X: -1 <= max <= k-1).
          rewrite Heqmax.
          unfold max_Ashort in Heqmax.
          destruct (fold_right_Zmax_dec (-1)  (filter _IsShort (alphas_ env n)))  as [eqa | ina].
          + simpl in eqa; lia.
          + split.
            rewrite <- Z.ge_le_iff.
            apply fold_right_Zmax_eqA_leA.
            reflexivity.
            rewrite filter_InA in ina.
            2: apply _IsShort_compat.
            destruct ina as [ina2 tall].
            unfold _IsShort in tall.
            rewrite <- Zlt_is_lt_bool in tall.
            unfold max_Ashort, _IsShort.
            lia.
          + lia.
            
        - rewrite <- Heqmin in fdec.
          rewrite Heqtas in fdec.
          rewrite filter_InA in fdec.
          2: apply _IsTall_compat.
          destruct fdec as [child tall].         
          rewrite <- InA_alphas in child; try assumption.
          destruct child as [m [path eqz]].
          exists m.
          split; auto; lia.
      }
      {
        unfold max_Ashort in Heqmax.
        remember (filter _IsShort (alphas_ env n)) as sas.
        remember (-1) as x0.
        assert (fdec := fold_right_Zmax_dec x0 sas).
        destruct fdec as [fdec | fdec]; simpl in fdec.
        - rewrite fdec in Heqmax.
          rewrite Heqmax in same.
          rewrite Heqx0 in same.
          simpl in same.
          rewrite same in pos.
          lia.
        - rewrite <- Heqmax in fdec.
          rewrite Heqsas in fdec.
          rewrite filter_InA in fdec.
          2: apply _IsShort_compat.
          destruct fdec as [child short].
          rewrite <- InA_alphas in child; try assumption.
          destruct child as [m [path eqz]].
          exists m.
          split; auto; lia.
      }
    Qed.
    
  End Alpha_child_parent.
  
End KDom_Assume.


Close Scope signature_scope.
Close Scope Z_scope.
Unset Implicit Arguments.
