(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Bool.
From Coq Require Import ZArith.

(** **** Local Imports *)
From Padec Require Import OptionUtil.
From Padec Require Import ZUtils.
From Padec Require Import ListUtils.
From Padec Require Import SetoidUtil.
From Padec Require Import FoldRight.
From Padec Require Import Algorithm.
From Padec Require Import Spanning_Tree.

(* end show *)
Open Scope Z_scope.
Open Scope signature_scope.
Set Implicit Arguments.

(** * K Dominating Set: the Algoritm 
 
  Algorithm for a process p:
    Global Input (shared by every process): k positive number
 
    Input: p.children contains a list of channels 
       assumption: the set of every p.children (for every node p) shapes a
                   downward tree (i.e. the tree is directed, from root to 
                   leaves).
 
    Variables: p.alpha integer number
 
    ----- Algorithm in itself (for any node p)
    Nb: for readability, we write any channel c in p.children as a node q, 
        namely q is the node which is pointed by c. 
    Functions:  MaxAShort(p) = max( { q.alpha | q in p.children | q.alpha < k }
                                    U { -1 } ) 
                MinATall(p) = min ( { q.alpha | q in children | q.alpha >= k }
                                    U { 2k + 1 } )
                NewAlpha(p) = if MaxAShort(p) + MinATall(p) + 2 <= 2k 
                              then MinATall(p) + 1
                              else MaxAShort(p) + 1
    Rule:  p.alpha <> NewAlpha(p) --> p.alpha(p) = NewAlpha(p) 
 *)

(** * Definition of alpha and its computation *)
Section KDomSet_Algo.
  
  Context {Chans: Channels}.

  Variable (k: Z).

  (** State of node for the algorithm *)
  Record KDomROState :=
    mkROState {
        (** children of the process, implemented as a list. *)
        _children: list Channel
      }.

  Notation eqROS :=
    (fun (x y: KDomROState) =>
       equivlistA equiv (_children x) (_children y)).

  Global Instance eqROS_equiv: Equivalence eqROS.
  Proof.
    split; hnf; intuition;etransitivity;eauto.
  Qed.

  Global Instance KDomROState_setoid: Setoid KDomROState :=
    {| equiv := eqROS |}.

  Record KDomState :=
    mkState {
        alpha: Z;
        (** read-only part of the state *)
        KDomROpart:> KDomROState
      }.

  Notation eqS :=
    (fun (x y: KDomState) =>
       alpha x == alpha y /\ KDomROpart x == KDomROpart y).

  Global Instance eqS_equiv: Equivalence eqS.
  Proof.
    split; hnf; intuition;
      etransitivity; eauto.
  Qed.

  Global Instance KDomState_setoid: Setoid KDomState :=
    {| equiv := eqS |}.

  Lemma KDomROState_dec: Decider (equiv (A := KDomROState)).
  Proof.
    intros [c1] [c2]; simpl equiv.
    destruct (equivlistA_dec Channel_dec c1 c2);
      [left; auto|right; tauto].
  Qed.

  Lemma KDomState_dec: Decider (equiv (A := KDomState)).
  Proof.
    intros [a1 ro1] [a2 ro2]; simpl.
    destruct (Z.eq_dec a1 a2); [|right; tauto].
    destruct (KDomROState_dec ro1 ro2);
      [left;auto|right;simpl in n; intuition].
  Qed.

  (** **** Compatibility on states *)
  Global Instance alpha_compat: Proper fequiv alpha.
  Proof. intros x y [h _];auto. Qed.
  
  Global Instance _children_compat:
    Proper (equiv ==> equivlistA equiv) _children.
  Proof. intros x y h;auto. Qed.

  Global Instance KDomROpart_compat: Proper fequiv KDomROpart.
  Proof. intros x y exy; apply exy. Qed.
  
  (** ** Function run *)
  (* MaxAShort(p) = max( { q.alpha | q in p.children | p.alpha < k } U { -1 } ) *)
  Definition MaxAShort (children_alphas: list Z): Z :=
    List.fold_right
      Z.max (-1) (List.filter (fun a => a <? k)
                              children_alphas).
  
  Global Instance MaxAShort_compat: 
    Proper (equivlistA equiv ==> equiv) MaxAShort.
  Proof.
    intros c1 c2 ec; simpl.
    eapply fold_right_max_compat; auto.
    eapply filter_compat_equiv; auto.
    intros a1 a2 ea; rewrite ea; reflexivity.
  Qed.
  
  (*MinATall(p) = min ( { q.alpha | q in children | q.alpha >= k } U { 2k + 1 } ) *)
  Definition MinATall (children_alphas: list Z): Z :=
    List.fold_right
      Z.min (2 * k + 1) (List.filter (fun a => a >=? k)
                                     children_alphas).
  
  Global Instance MinATall_compat: 
    Proper (equivlistA equiv ==> equiv) MinATall.
  Proof.
    intros c1 c2 ec; simpl.
    eapply fold_right_min_compat; auto.
    eapply filter_compat_equiv; auto.
    intros a1 a2 ea; rewrite ea; reflexivity.
  Qed.

  (*NewAlpha(p) = if MaxAShort(p) + MinATall(p) + 2 <= 2k 
                              then MinATall(p) + 1
                              else MaxAShort(p) + 1*)
    Definition NewAlpha (children_alphas: list Z): Z :=
      let short := MaxAShort children_alphas in
      let tall := MinATall children_alphas in
      if Z_le_gt_dec (short + tall + 2) (2 * k)
      then tall + 1 else short + 1.
    
    Global Instance NewAlpha_compat: 
      Proper (equivlistA equiv ==> equiv) NewAlpha.
    Proof.
      intros c1 c2 ec; unfold NewAlpha.
      rewrite (MaxAShort_compat ec), (MinATall_compat ec). 
      reflexivity.
    Qed.

    Definition alpha_list
               (state: KDomState)
               (peer_state: Channel -> option KDomState):
      list Z :=
      map_filter (fun c =>
                    match peer_state c with
                    | None => None (* should never occur *)
                    | Some s => Some (alpha s) end)
                 (_children state).

    Global Instance alpha_list_compat:
      Proper (equiv ==> pequiv ==> equivlistA equiv) alpha_list.
    Proof.
      intros s1 s2 es p1 p2 ep.
      eapply map_filter_compat; auto.
      { intros c1 c2 ec.
        assert (dum := ep _ _ ec).
        destruct (p1 c1); destruct (p2 c2); try tauto.
        apply Some_compat, alpha_compat; auto. }
      apply _children_compat; auto.
      apply es.
    Qed.
    
    Definition run (state: KDomState)
               (peer_state: Channel -> option KDomState): KDomState :=
      {| alpha := NewAlpha (alpha_list state peer_state);
         KDomROpart := KDomROpart state |}.

    Global Instance run_compat: Proper fequiv run.
    Proof.
      intros s1 s2 es ps1 ps2 eps; split.
      apply NewAlpha_compat, alpha_list_compat; auto.
      simpl; apply KDomROpart_compat; auto.
    Qed.      
    
    (****************** OLD **************************
     *)

    (*
     Variable (alpha: Node -> Z).
 
     Lemma short_or_tall:
@@ -363,119 +225,30 @@ Section KDomSet_Alpha.
       left; apply Z.ltb_lt; assumption.
       right; rewrite Z.geb_Bool.le; apply Z.Bool.le_le; assumption.
     Qed.
     *)
    
    Lemma ROpart_stable: forall s sn, KDomROpart (run s sn) =~= KDomROpart s.
    Proof.
      intros s sn; unfold run; simpl; apply _children_compat; auto.
    Qed.
    
    (** ** The Algorihtm *)
    Global Instance KDomSet_algo: Algorithm.
    Proof.
      esplit.
      apply KDomState_dec.
      apply run_compat.
    Defined.
    
    Global Instance KDom_Stable:
      Stable_Variable KDomSet_algo KDomROpart.
    Proof.
      esplit.
      apply KDomROpart_compat.
      apply ROpart_stable.
    Defined.
    
End KDomSet_Algo.

Close Scope signature_scope.
Close Scope Z_scope.
Unset Implicit Arguments.
