(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
*)
(* begin show *)

(** Global imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import ZArith.
From Coq Require Import Lia.
From Coq Require Import Bool.

(** Local imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import OptionUtil.
From Padec Require Import ZUtils.
From Padec Require Import ListUtils.
From Padec Require Import RelationA.
From Padec Require Import FoldRight.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Tree_topology.
From Padec Require Import KDomSet_algo_tree_topology.
From Padec Require Import KDomSet_specification_tree_topology.
From Padec Require Import Self_Stabilization.

(* end show *)
Open Scope Z_scope.
Open Scope signature_scope.
Set Implicit Arguments.

(** * Correctness of the K-Dominating Set algorithm

  Once a terminal configuration is reached, we prove that the variables draw a k-dominating set
 *)
Section KDomSet_Correctness.

  (* ********************************************************)
  Context {Chans: Channels}
          {Net: Network}
          {DTN: Down_Tree_Network Net}
          {KDP: KDomSet_parameters}.
  Let k_pos := k_pos.
  
  Notation Tree := (DTN_topo (Net := Net)). 
  Context {SPAN: is_spanning _ Tree}.

  Notation Algo := KDom_algo.
  Notation Env := (Env (Net := Net) (Algo := KDom_algo)).

  Existing Instance KDom_Stable.
  Existing Instance KDom_RO_assumption.

  Notation alphas_ :=
    (fun (env: Env) n => alphas (children_states (env n) (local_env env n))).
  
  Notation new_alpha_ := (fun env n => new_alpha (alphas_ env n)).

  (* ************************************************************)
  (** ** Correctness *)
  Section Final_kDomSet.

    (** Assumptions on Inputs *)
    Variable (env: Env)
             (env_compat: Proper pequiv env)
             (hassume: Assume env).

    Variable (final: forall (n: Node), enabled_b env n = false).

    Lemma same: forall (n: Node), alpha (env n) = new_alpha_ env n.
    Proof.
      intros n. 
      specialize (@final n).
      unfold enabled_b, RUN, run in final; simpl in final;
        unfold run_algorithm in final.
      revert final; case (KDom_State_dec
                            {| alpha := new_alpha_ env n;
                               KDom_ROpart := env n |} (env n)).
      intros [h1 _]; simpl in h1; auto.
      congruence.
    Qed.

    Definition is_kDom_path := chain (is_kDom_edge env).

    Lemma tall_has_path_to_root:
      forall (n: Node) (i: nat),
        alpha (env n) = k + Z_of_nat i ->
        exists (r: Node), 
          kDominator env r = true /\
          exists (p: list Node), (is_kDom_path r p n).
    Proof.
      intros n i. 
      revert n.
      induction i as [ | j Hrec]; intros n eqz.
      simpl in eqz;rewrite Zplus_0_r in eqz.
      exists n;
        split;
        [
          unfold kDominator;
          simpl in eqz;
          rewrite <- Z.eqb_eq in eqz;
          rewrite eqz;
          simpl;
          reflexivity
        |
        exists nil;
          simpl;
          reflexivity
        ].
      assert (na := new_alpha_inv env_compat hassume n).
      assert (pos: alpha (env n) > 0).
      rewrite eqz. 
      lia.
      apply na in pos.
      2: apply same.
      clear na.
      destruct pos as [m [path eqz']].
      rewrite eqz in eqz'.
      rewrite Nat2Z.inj_succ in eqz'.
      unfold Z.succ in eqz'.
      assert (h: eq (alpha (env m)) (k + Z.of_nat j)).
      lia.
      apply Hrec in h.
      clear Hrec.
      destruct h as [r [dom_set [p path' ]]].
      exists r.
      split.
      assumption.
      exists (p ++ (n :: nil)).
      apply chain_app.
      apply is_kDom_edge_compat; assumption.
      exists m.
      split.
      assumption.
      simpl.
      split; [| reflexivity].
      unfold is_kDom_edge.
      rewrite Nat2Z.inj_succ in eqz.
      assert (eqa: alpha (env n) = alpha (env m) + 1).
      lia.
      assert (pos': Z.of_nat j + Z.of_nat 1 > Z.of_nat 0).
      rewrite <- Nat2Z.inj_add.
      rewrite <- Nat2Z.inj_gt.
      rewrite plus_comm.
      simpl.
      apply gt_Sn_O.
      simpl in pos'.
      assert (gt': alpha (env n) > k).
      lia.
      apply Zgt_compare in gt'.
      destruct (alpha (env n) ?= k); try elim gt'.
      split.
      assumption.
      lia.
    Qed.

    Lemma tall_is_linked:
      forall (n: Node) (i: nat),
        alpha (env n) = k + Z_of_nat i ->
        (exists (r: Node), 
          kDominator env r = true /\
          exists (p: list Node), (is_kDom_path r p n)).
    Proof.
      intros n i. 
      revert n.
      induction i as [ | j Hrec]; intros n eqz.
      simpl in eqz;rewrite Zplus_0_r in eqz.
      exists n;  split;
        [
          unfold kDominator;
          simpl in eqz;
          rewrite <- Z.eqb_eq in eqz;
          rewrite eqz;
          simpl;
          reflexivity
        |
        exists nil;
          simpl;
          reflexivity
        ].

      assert (pos: alpha (env n) > 0).
      rewrite eqz; lia.
      assert (na := new_alpha_inv env_compat hassume n (same n) pos).       destruct na as [m [path eqz']].
      rewrite eqz in eqz'.
      rewrite Nat2Z.inj_succ in eqz'.
      assert (h: alpha (env m) = k + Z.of_nat j).
      lia.
      destruct (Hrec _ h) as [r [dom_set [p path']]]; clear Hrec.

      exists r.
      split.
      assumption.
      exists (p ++ (n :: nil)).
      apply chain_app.
      apply is_kDom_edge_compat; assumption.
      exists m.
      split.
      assumption.
      simpl.
      split; [| reflexivity].
      unfold is_kDom_edge.
      rewrite Nat2Z.inj_succ in eqz.
      assert (tmp: alpha (env n) ?= k = Gt).
      2: rewrite tmp; split; [assumption | rewrite eqz; lia]. 
      rewrite Z.compare_gt_iff; lia.
    Qed.

    Lemma tall_is_kDominated:
      forall (n: Node) (i: nat),
        alpha (env n) = k + Z_of_nat i ->
        forall r p, is_kDom_path r p n -> (length p <= i)%nat.
    Proof.
      intros n i. 
      revert n.
      induction i as [ | j Hrec]; intros n eqz.
      simpl in eqz;rewrite Zplus_0_r in eqz.
      intros r p.
      remember (rev p) as q.
      destruct q.
      assert (tmp:= rev_involutive p).
      rewrite <- Heqq in tmp; simpl in tmp.
      intros _; rewrite <- tmp; simpl; lia.
      assert (tmp := rev_unit (rev q) n0).
      rewrite rev_involutive, Heqq in tmp.
      rewrite <- rev_intro in tmp.
      rewrite <- tmp; unfold is_kDom_path; simpl.
      rewrite chain_app; simpl.
      2: apply is_kDom_edge_compat; auto. 
      intros [y [C [F en]]].
      rewrite en in F.
      unfold is_kDom_edge in F.
      assert (tmpp: alpha (env n) ?= k = Eq).
      rewrite Z.compare_eq_iff; assumption.
      rewrite tmpp in F; tauto. 

      intros rr pp.
      remember (rev pp) as q.
      destruct q.
      rewrite rev_intro, rev_involutive in Heqq.
      simpl in Heqq. 
      intros _; rewrite <- Heqq; simpl; lia.
      assert (tmp := rev_unit (rev q) n0).
      rewrite rev_involutive, Heqq in tmp.
      rewrite <- rev_intro in tmp.
      rewrite <- tmp; unfold is_kDom_path; simpl.
      rewrite chain_app; simpl.
      2: apply is_kDom_edge_compat; auto. 
      intros [y [C [F en]]].
      rewrite en in F.
      unfold is_kDom_edge in F.
      assert (tmpp: alpha (env n) ?= k = Gt).
      rewrite Z.compare_gt_iff; rewrite eqz; rewrite Nat2Z.inj_succ; lia.
      rewrite tmpp in F; clear tmpp. 
      destruct F as [stp ay].
      assert (tmpp: alpha (env y) = k + Z.of_nat j).
      rewrite ay, eqz, Nat2Z.inj_succ; lia.
      specialize (Hrec _ tmpp); clear tmpp.
      rewrite app_length; simpl.
      unfold is_kDom_path in Hrec.
      specialize (Hrec _ _ C).
      lia.
    Qed.

    Lemma tall_k: 
      forall (n: Node),
        _IsTall (alpha (env n)) = true ->
        exists (i: nat), alpha (env n) = k + Z.of_nat i.
    Proof.
      intros n tall.
      unfold _IsTall in tall.
      rewrite Z.geb_le in tall.
      apply ZUtils.le_plus in tall; auto.
    Qed.
    
    Lemma short_k: 
      forall (n: Node), 
        _IsShort (alpha (env n)) = true ->
        exists (i: nat), (Z.of_nat i < k)%Z /\
                         alpha (env n) = k - 1 - Z.of_nat i.
    Proof.
      intros n short.
      unfold _IsShort in short.
      rewrite Z.ltb_lt, <- Z.gt_lt_iff in short.
      apply ZUtils.gt_plus in short.
      destruct short as [i p].
      exists i; split; [|assumption].
      rewrite (same n) in p.
      assert (h := new_alpha_0_2k (alphas_ env n)). 
      lia.
    Qed.

    Lemma alpha_monotonic_small:
      forall (n m: Node), 
        isParent m n ->
        _IsShort (alpha (env n)) = true ->
        alpha (env m) >= alpha (env n) + 1.
    Proof.
      intros n m nm n_short.
      destruct (short_or_tall (alpha (env m))) as [m_short | m_tall].
      { (* m short *)
        assert (same := same m).
        unfold new_alpha, max_Ashort, min_Atall in same.
        remember (fold_right Z.max (-1)
                             (filter _IsShort (alphas_ env m))) as max.
        remember (fold_right Z.min (2 * k + 1)
                             (filter _IsTall (alphas_ env m))) as min.      
        assert (ina: InA Z.eq (alpha (env n)) (filter _IsShort (alphas_ env m))). 
        { rewrite filter_InA.
          2: apply _IsShort_compat.
          split; auto.
          rewrite <- InA_alphas; auto.
          exists n.
          split; auto.
        }
        
        destruct ((filter _IsShort (alphas_ env m))) as [| short].
        { rewrite InA_nil in ina.
          elim ina. }
        destruct (Z_le_gt_dec (max + min + 2) (2 * k)).      
        { assert (min_range: k <= min <= 2*k+1).
          rewrite Heqmin.
          destruct (fold_right_Zmin_dec (2 * k + 1) 
                                        (filter _IsTall (alphas_ env m))) as [eqa | _ina]; unfold Z.eq in *.
          { rewrite eqa; lia. }
          split.
          rewrite filter_InA in _ina.
          2: apply _IsTall_compat. 
          destruct _ina as [ina2 tall].
          unfold _IsTall in tall.
          rewrite Z.geb_le in tall.
          assumption.
          apply fold_right_Zmin_eqA_leA.
          reflexivity.

          unfold _IsShort in m_short.
          rewrite Z.ltb_lt in m_short.
          lia.
        }
        { rewrite same.
          assert (tmp: max >= alpha (env n)).
          rewrite Heqmax.
          apply fold_right_Zmax_InA_leA.
          assumption.
          lia.
        }
      }
      { (* m tall *)
        unfold _IsShort in n_short.
        unfold _IsTall in m_tall.
        rewrite Z.ltb_lt in n_short.
        rewrite Z.geb_le in m_tall.
        lia.
      }
    Qed.
    
    Lemma short_is_linked:
      forall (i: nat) (n: Node),
        alpha (env n) = k - Z.of_nat i ->
        (exists (r: Node),
           kDominator env r = true /\
           exists (p: list Node), (is_kDom_path r p n)).
    Proof.
      assert (tmp := well_founded_ind lt_wf).
      specialize 
        (tmp (fun i => 
                forall n,
                  alpha (env n) = k - Z.of_nat i ->
                  (exists r,
                    kDominator env r = true /\
                    (exists p, (is_kDom_path r p n))))).
      apply tmp; clear tmp.
      intros i Hrec n eqa.
      destruct i as [ | i].

      (* base case *)
      apply tall_is_linked with 0%nat.
      simpl; simpl in eqa.
      lia.

      (* inductive case *)
      rewrite Nat2Z.inj_succ in eqa.
      assert (i_pos: 0 <= Z.of_nat i).
      lia.
      assert (n_short: _IsShort (alpha (env n)) = true).
      unfold _IsShort.
      rewrite Z.ltb_lt.
      lia.
      case_eq (is_root (env n)).

      (* n_root *)
      intros n_root.
      exists n; split.
      unfold kDominator.
      rewrite orb_true_iff; right; rewrite andb_true_iff; split; assumption.
      exists nil; simpl; reflexivity.

      (* n_not_root *)
      intros n_not_root.
(*      assert (not_root: ~is_root Par n).      
      rewrite (IsRoot_is_root_equiv Par_compat _ _ (Equivalence_Reflexive n)), 
      n_not_root; apply diff_false_true. *)
      assert (m_isParent_n: exists m, Parent n == Some m).
      {
        rewrite is_root_equiv in n_not_root; auto.
        assert (rt := DTN_root_ok n).
        case_eq (Parent n); [intros m m_isParent_n_|].
        2: { intros hm; rewrite hm, n_not_root in rt.
             destruct rt as [rt _].
             assert (x: false = true) by (apply rt; auto).
             congruence. }
        exists m; auto.
      }
      destruct m_isParent_n as [m m_isParent_n].
      assert (m_is_kDom_edge_n: (is_kDom_edge env m n)).
      { unfold is_kDom_edge.
        apply Zcompare_elim.
        intros f; lia.
        intros _; auto.
        intros f; lia.
      }
      assert (m_in_dom_set: 
                exists r,
                  kDominator env r = true /\
                  (exists p, (is_kDom_path r p m))).
      {
        (* r >:: p ::> m *)
        destruct (short_or_tall (alpha (env m))) as [m_short | m_tall].
        assert (tmp := 
                  alpha_monotonic_small n m m_isParent_n n_short).
        (* m short *)
        assert (tmp' := short_k m m_short).
        destruct tmp' as [j eq'].
        remember (j + 1)%nat as j'.
        assert (eq'': alpha (env m) = (k) - Z.of_nat j').
        rewrite Heqj', Nat2Z.inj_add.
        unfold Z.sub in *.
        rewrite Z.opp_add_distr.
        remember (Z.of_nat 1) as one.
        simpl in Heqone.
        subst.
        lia.
        assert (Hrec' := Hrec).
        specialize (Hrec' j').
        assert (lt': (j' < Datatypes.S i)%nat).
        lia.
        specialize (Hrec' lt' m eq'').
        assumption.
        
        (* m_tall *)
        destruct (tall_k _ m_tall) as [ii H]. 
        apply tall_is_linked with ii; assumption.
      }
      { 
        (* r >:: p ::> m /\ m :> n -> r >:: p ::> n *)
        destruct m_in_dom_set as [r [r_ch [p path]]].
        exists r.
        split.
        assumption.
        exists (p ++ (n :: nil)).
        unfold is_kDom_path.
        rewrite chain_app; [| apply is_kDom_edge_compat].
        exists m.
        split.
        assumption.
        simpl.
        split.
        assumption.
        reflexivity.
        assumption.
      }
    Qed.

    Lemma short_is_kDominated:
      forall (i: nat) (n: Node),
        alpha (env n) = k - Z.of_nat i ->
        forall r p, is_kDom_path r p n -> (length p <= i)%nat.
    Proof.
      assert (tmp := well_founded_ind lt_wf).
      specialize 
        (tmp (fun i => 
                forall n,
                  alpha (env n) = k - Z.of_nat i ->
                  forall r p, is_kDom_path r p n -> (length p <= i)%nat)).
      apply tmp.
      clear tmp.
      intros i Hrec n eqa.
      destruct i as [ | i].

      (* base case *)
      apply tall_is_kDominated; simpl in *; lia.

      (* inductive case *)
      rewrite Nat2Z.inj_succ in eqa.
      unfold Z.succ in eqa.
      assert (i_pos: 0 <= Z.of_nat i).
      lia.
      assert (n_short: _IsShort (alpha (env n)) = true).
      unfold _IsShort; rewrite Z.ltb_lt; lia.
      case_eq (is_root (env n)).

      (* n_root *)
      intros n_root.
      assert (kdom: kDominator env n = true).
      unfold kDominator.
      rewrite n_short, n_root, orb_comm; simpl; reflexivity.
      intros r p.
      remember (rev p) as q.
      destruct q.
      assert (tmp:= rev_involutive p).
      rewrite <- Heqq in tmp; simpl in tmp.
      intros _; rewrite <- tmp; simpl; lia.
      assert (tmp := rev_unit (rev q) n0).
      rewrite rev_involutive, Heqq in tmp.
      rewrite <- rev_intro in tmp.
      rewrite <- tmp; unfold is_kDom_path; simpl.
      rewrite chain_app; simpl.
      2: apply is_kDom_edge_compat; auto. 
      intros [y [C [F en]]].
      rewrite en in F.
      unfold is_kDom_edge in F.
      unfold _IsShort in n_short.      
      assert (tmpp: alpha (env n) ?= k = Lt).
      rewrite Z.compare_lt_iff; rewrite Z.ltb_lt in n_short; assumption. 
      rewrite tmpp in F; clear tmpp. 
      rewrite is_root_equiv in n_root; auto.
      rewrite <- DTN_root_ok in n_root.
      unfold isParent in F; rewrite F in n_root; inversion n_root.

      (* n_not_root *)
      intros n_not_root rr pp.
      remember (rev pp) as q.
      destruct q.
      rewrite rev_intro, rev_involutive in Heqq.
      simpl in Heqq. 
      intros _; rewrite <- Heqq; simpl; lia.
      assert (tmp := rev_unit (rev q) n0).
      rewrite rev_involutive, Heqq in tmp.
      rewrite <- rev_intro in tmp.
      rewrite <- tmp; unfold is_kDom_path; simpl.
      rewrite chain_app; simpl.
      2: apply is_kDom_edge_compat; auto. 

      intros [y [C [F en]]].
      rewrite en in F.
      rewrite app_length; simpl.
      clear Heqq tmp en n0.
      unfold is_kDom_edge in F.
      assert (tmpp: alpha (env n) ?= k = Lt).
      rewrite Z.compare_lt_iff; rewrite eqa; lia. 
      rewrite tmpp in F; clear tmpp. 

      destruct (short_or_tall (alpha (env y))) as [y_short | y_tall].
      
      (* short *)
      assert (X := alpha_monotonic_small _ _ F n_short).
      rewrite eqa in X.
      rewrite Z.sub_add_distr, Z.sub_add in X.
      unfold _IsShort in y_short; rewrite Z.ltb_lt in y_short.
      assert (tmp: ( Z.to_nat (k - (alpha (env y))) < S i)%nat ).
      lia.
      specialize (Hrec _ tmp y); clear tmp.      
      assert (tmp: alpha (env y) = k - Z.of_nat (Z.to_nat (k - alpha (env y)))).
      rewrite Z2Nat.id.
      2: lia.
      lia.
      specialize (Hrec tmp); clear tmp.
      unfold is_kDom_path in Hrec.
      specialize (Hrec _ _ C).
      lia.

      (* tall *)
      assert (same := same y).
      unfold new_alpha, max_Ashort, min_Atall in same.
      remember (fold_right Z.max (-1)
                  (filter _IsShort (alphas_ env y))) as max.
      remember (fold_right Z.min (2 * k + 1)
                  (filter _IsTall (alphas_ env y))) as min.            
      destruct (Z_le_gt_dec (max + min + 2) 
                            (2 * (k))) as [le' | gt'].
      (* max + min + 2 <= 2 * k *)
      assert (ina: InA eq (alpha (env n)) (filter _IsShort (alphas_ env y))). 
      rewrite filter_InA.
      2: intros x z exy; rewrite exy; reflexivity.
      split; [| assumption].
      apply InA_alphas; try assumption.
      exists n.
      split; [assumption | reflexivity].
      assert (K := fold_right_Zmax_InA_leA (-1) ina). 
      rewrite eqa in K.
      unfold _IsTall in y_tall; rewrite Z.geb_le in y_tall.
      unfold _IsShort in n_short; rewrite Z.ltb_lt in n_short.
      rewrite eqa in n_short.
      assert (tmp: exists j, alpha (env y) = k + Z.of_nat j).
      specialize (ZUtils.le_plus y_tall); auto.
      destruct tmp as [j Hj]. 
      assert (OK := tall_is_kDominated _ _ Hj rr (rev q)).
      unfold is_kDom_path in OK; specialize (OK C).
      rewrite <- Heqmax in K.
      lia.

      (* max + min + 2 > 2 * k *)
      assert (eq': alpha (env y) = k).
      unfold _IsTall in y_tall.
      rewrite Z.geb_le in y_tall.
      assert (range_max: -1 <= max <= k-1).
      rewrite Heqmax.
      destruct (fold_right_Zmax_dec (-1)  (filter _IsShort (alphas_ env y))) as [eqa_ | ina_];
        unfold Z.eq in *.
      rewrite eqa_; lia.
      split.
      rewrite <- Z.ge_le_iff.
      apply fold_right_Zmax_eqA_leA.
      reflexivity.
      rewrite filter_InA in ina_.
      2: intros x yy exy; rewrite exy; reflexivity.
      destruct ina_ as [ina2 tall].
      unfold _IsShort in tall.
      rewrite <- Zlt_is_lt_bool in tall.
      unfold _IsShort.
      lia.
      lia.
      destruct q as [| x q]; simpl in C; [simpl; lia|].
      rewrite chain_app in C.
      2: apply is_kDom_edge_compat; auto. 
      destruct C as [yy [CC [FF en]]]; simpl in en.
      rewrite en in FF.
      unfold is_kDom_edge in FF.
      rewrite <- Z.compare_eq_iff in eq'; rewrite eq' in FF; elim FF.
    Qed.

    Lemma any_is_linked:
      forall (n: Node), exists (r: Node),
        kDominator env r = true /\
        exists (p: list Node), (is_kDom_path r p n).
    Proof.
      intros n.
      destruct (short_or_tall (alpha (env n))) as [n_short | n_tall].
      (*short *)
      destruct (short_k n n_short) as [i [Hi HHi]].
      assert (tmp: exists ii, alpha (env n) = k - Z.of_nat ii).
      exists (S i); zify; lia.
      destruct tmp as [ii Hii].
      destruct (short_is_linked _ _ Hii) as [r [kdom [p Hp]]].      
      exists r; split; [apply kdom | exists p; apply Hp].
      (* tall *)
      destruct (tall_k n n_tall) as [i Hi].
      destruct (tall_is_linked _ _ Hi) as [r [kdom [p Hp]]].      
      exists r; split; [apply kdom | exists p; apply Hp].
    Qed.

    Lemma any_is_kDominated:
      forall (n: Node),
      forall r p, is_kDom_path r p n -> Z.of_nat (length p) <= k.
    Proof.
      intros n.
      destruct (short_or_tall (alpha (env n))) as [n_short | n_tall].
      (*short *)
      {
        destruct (short_k n n_short) as [i [Hi HHi]].
        assert (tmp: exists ii, alpha (env n) = k - Z.of_nat ii).
        exists (S i); zify; lia.
        destruct tmp as [ii Hii].
        intros r p Hp.
        assert (OK := short_is_kDominated _ _ Hii _ _ Hp).  
        lia.
      }
      (* tall *)
      {
        destruct (tall_k n n_tall) as [i Hi].
        intros r p Hp.
        assert (OK := tall_is_kDominated _ _ Hi _ _ Hp).
        assert (same := same n).
        assert (A := new_alpha_0_2k (alphas_ env n)). 
        hnf in same,A.
        lia.
      }
    Qed.

    (** Lemma that proves correctness:
     Any node is linked to dom_set head through a path of length smaller than k
     *)
    Definition IsInkDomSet (k: Z) (root: Node) (n: Node): Prop :=
      kDominator env root = true /\
      exists p, (is_kDom_path root p n) /\ (Z_of_nat (length p) <= k)%Z.

    Lemma any_is_in_kDom_set: 
      forall (n: Node), exists (r: Node), IsInkDomSet k r n.
    Proof.
      intros.
      unfold IsInkDomSet.
      destruct (any_is_linked n) as [r [kdom [p Hp]]].
      exists r; split; [apply kdom | exists p; split; [ apply Hp|]].
      apply (any_is_kDominated _ _ _ Hp).
    Qed.

  End Final_kDomSet.


    (** ** Correctness *)
    Section Correctness.
      
    Definition kDominator_ (g: Env) (n: Node): Prop  := 
      kDominator g n = true.

    Theorem kdom_set_at_terminal:
      forall (g: Env), 
        Assume g -> terminal g -> kDOM kDominator_ g (k := k).
    Proof.
      intros g ass term n.
      assert (final: forall n, enabled_b g n = false).
      { intros nn.
        apply not_true_is_false; intros F.
        eapply terminal_all_disabled; eauto.
      }      
      destruct term as [pg _].
      assert (OK := 
                any_is_in_kDom_set pg ass final n).
      destruct OK as (d, (kdom,(p,(okp,lp)))).
      exists d.
      unfold kDominator_; split; [ apply kdom | exists p].
      split; [| exact lp].
      clear lp kdom.
      eapply RelationA.chain_incl.
      2: apply okp.
      clear okp.
      intros n1 n2.
      unfold is_kDom_edge, is_neighbor_of.

      intro H;      
        destruct (alpha (g n2) ?= k); [ elim H | |];
          unfold isParent in H; rewrite DTN_children_ok in H.
      destruct H as [c [_ par]]; exists c; left; auto.
      destruct H as [[c [_ par]] _]; exists c; right; auto.
    Qed.

    Lemma kDOM_partial_correctness: 
      P_correctness KDom_RO_assumption (kDOM kDominator_ (k := k)).
    Proof.
      unfold Self_Stabilization.P_correctness; 
      apply kdom_set_at_terminal.
    Qed.
    
  End Correctness.

End KDomSet_Correctness.

Close Scope Z_scope.
Close Scope signature_scope.
Unset Implicit Arguments.
