(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
*)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import ZArith.
From Coq Require Import Lia.
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import RelationPairs.

(** **** Local Imports *)
From Padec Require Import NatUtils.
From Padec Require Import OptionUtil.
From Padec Require Import SetoidUtil.
From Padec Require Import Count.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Tree_topology.
From Padec Require Import KDomSet_algo_tree_topology.

(* end show *)
Open Scope Z_scope.
Open Scope signature_scope.
Set Implicit Arguments.

(** * How Many Elements in the k-Hop-Dominating Set?
 Count the number of elements in the dominating set (dominating heads) built by the algorithm.
 
 We prove bounds on it.
 *)
Section KDomSet_count.

  (* ********************************************************)
  (** Channel and Network definitions *)
  (* ********************************************************)
  Context {Chans: Channels}
          {Net: Network}
          {DTN: Down_Tree_Network Net}
          {KDP: KDomSet_parameters}.
  Let k_pos := k_pos.
  
  Notation Tree := (DTN_topo (Net := Net)). 
  Context {SPAN: is_spanning _ Tree}.

  Notation Algo := KDom_algo.
  Notation Env := (Env (Net := Net) (Algo := KDom_algo)).

  Existing Instance KDom_Stable.
  Existing Instance KDom_RO_assumption.

  Existing Instance alpha_compat.
  Existing Instance isParent_compat.
  Existing Instance isRoot_compat.
  
  (** Assumptions on Inputs *)
  Variable (env: Env) (env_compat: Proper pequiv env)
           (hassume: Assume env)
           (final: terminal env).

  (** *** Dominating Heads 
   Two kinds: 
              - the regular one (with alpha equals k)
              - (maybe) the root 
   *)
  Notation kDominator := (kDominator env).
  Definition DomHeads := { n: Node | kDominator n = true }.
  
  Notation eqDH := (equiv (A := DomHeads)).

  (** Regular Dominating Heads are nodes such that alpha = k *) 
  Definition RegDomHead p := alpha (env p) = k.

  Definition RegDomHeads := { n: Node | RegDomHead n }.

  Notation eqRDH := (equiv (A := RegDomHeads)).

  Global Instance RegDomHead_compat: Proper fequiv RegDomHead.
  Proof.
    intros n1 n2 eqn; unfold RegDomHead.
    rewrite eqn; reflexivity.
  Qed.

  (** Regular Nodes are those below a tall node: 
      They cannot be in the dominating set of the short root (if short) *)

  Inductive HasTallAncestor: Node -> Prop :=
  | HTA_self: forall n, alpha (env n) >= k -> HasTallAncestor n
  | HTA_parent:  
      forall m n, isParent m n ->
                  HasTallAncestor m -> HasTallAncestor n.

  Definition RegNode := { n: Node | HasTallAncestor n }.

  Notation eqRN := (equiv (A := RegNode)).

  Global Instance HasTallAncestor_compat: Proper fequiv HasTallAncestor.
  Proof.
    intros n1 n2 en;split.
    { intro w;revert n2 en;induction w;intros n2 en;[left|right with m].
      - rewrite <- en; auto.
      - rewrite <- en; auto.
      - apply IHw; auto.
    }
    { intro w;revert n1 en;induction w;intros n1 en;[left|right with m].
      - rewrite en; auto.
      - rewrite en; auto.
      - apply IHw; auto.
    }
  Qed.
  
  (** Relation between a node and its closest ancestor s.t. alpha=k *)
  Inductive RegHeadAncestor: Node -> Node -> Prop :=
  | RHA_self: forall n h, n == h -> alpha (env h) = k -> RegHeadAncestor n h
  | RHA_parent:  
      forall m n h, isParent m n -> alpha (env n) <> k -> 
                    RegHeadAncestor m h -> RegHeadAncestor n h.

  Global Instance RegHeadAncestor_compat: 
    Proper fequiv RegHeadAncestor.
  Proof.
    intros n1 n2 en a1 a2 ea;split.
    { intro w;revert n2 en a2 ea;induction w;intros n2 en a2 ea;[left|right with m].
      - rewrite <- en, <- ea; auto.
      - rewrite <- ea; auto.
      - rewrite <- en; auto.
      - rewrite <- en; auto.
      - apply IHw; auto.
    }
    { intro w;revert n1 en a1 ea;induction w;intros n1 en a1 ea;[left|right with m].
      - rewrite en, ea; auto.
      - rewrite ea; auto.
      - rewrite en; auto.
      - rewrite en; auto.
      - apply IHw; auto.
    }
  Qed.
  
  Section with_cardinals.

    Variable (card_N: nat)
             (Hcard_N: Num_Card Same Node card_N).

    Variable (card_DH: nat)
             (Hcard_DH: Num_Card Same DomHeads card_DH).

    Variable (card_RDH: nat)
             (Hcard_RDH: Num_Card Same RegDomHeads card_RDH).

    Variable (card_RN: nat)
             (Hcard_RN: Num_Card Same RegNode card_RN).

    (** [Rcount] is an existence witness for a the cardinality result *)
    Definition Rcount 
               (cpl: ltN (Datatypes.S (Z.to_nat k)) * RegDomHeads) 
               (rn: RegNode): Prop := 
      Z.to_nat (alpha (env (proj1_sig rn))) = proj1_sig (fst cpl) /\ 
      RegHeadAncestor (proj1_sig rn) (proj1_sig (snd cpl)).

    (** [count_ok] every small index is present in every dominating set *)
    Theorem simple_counting: 
      (card_RN >= (Datatypes.S (Z.to_nat k)) * card_RDH)%nat.
    Proof.
      unfold ge.
      rewrite <- Card_smaller_le_iff.
      apply Card_trans with (ltN (Datatypes.S (Z.to_nat k)) * ltN card_RDH)%type _.
      apply Card_weak;apply Card_sym;simpl sym_Card_Prop;apply Card_as_many_prod.
      apply Card_trans with (ltN (Datatypes.S (Z.to_nat k)) * RegDomHeads)%type _.
      apply Card_sym;apply Card_weak;eapply Card_prod.
      apply Card_refl;apply eqltN_equiv.
      apply Hcard_RDH.
      apply Card_trans with RegNode _.
      2:apply Hcard_RN.

      simpl.
      split with Rcount.
      { intros [[alpha1 halpha1] [ch1 hch1]] [[alpha2 halpha2]
                                                [ch2 hch2]] 
               [halpha hch];
        do 2 red in halpha;
        red in hch;
        simpl in halpha,hch.
        intros [n1 h1] [n2 h2] hn.
        red in hn;simpl in hn.
        unfold Rcount;simpl.
        apply Morphisms_Prop.and_iff_morphism.
        apply PER_morphism.
        apply Equivalence_PER.
        apply eq_equivalence.
        apply f_equal.
        apply alpha_compat.
        { assert (n1 == n2) by apply hn.
          rewrite H; auto. }
        assumption.
        apply RegHeadAncestor_compat.
        assumption.
        apply hch.
      }
      intros [[alpha1 halpha1] [ch1 hch1]] [n1 h1] [h11 h21].
      intros [[alpha2 halpha2] [ch2 hch2]] [n2 h2] [h12 h22]. 
      intro en;red in en;simpl in en.
      simpl in *.
      split.
      compute.
      rewrite <- h11, <- h12.
      apply f_equal.
      apply alpha_compat.
        { assert (n1 == n2) by apply en.
          rewrite H; auto. }
      assert (enn: n1 == n2) by apply en; clear en; rename enn into en.
      red;simpl.
      red;simpl.
      clear h1 h11 hch1 h2 h12 hch2.
      revert n2 ch2 en h22.
      induction h21;
        intros n2 ch2 en h22; 
        rewrite <- en in h22;
        inversion h22.
      rewrite H in H1.
      assumption.
      elim H2;rewrite <- H0.
      apply alpha_compat.
      rewrite H; auto. 
      elim H0;rewrite <- H2.
      apply alpha_compat.
      rewrite H1; auto. 
      eapply IHh21 with m0.    
      eapply isParent_unique; eauto.
      assumption.

      intros [[alpha1 halpha1] [ch1 hch1]] [[alpha2 halpha2] [ch2 hch2]] 
             [halpha hch].
      compute in halpha;do 2 red in hch;simpl in hch.
      subst; rename alpha2 into a.
      generalize halpha2;revert halpha1.
      replace a with ((Z.to_nat k) -(Z.to_nat k  - a) )%nat.
      2:lia.    
      clear halpha2.
      generalize (Z.to_nat k - a)%nat (le_minus (Z.to_nat k) a).
      clear a.
      intro n;induction n as [|i Hi].
      rewrite <- minus_n_O.
      assert (hhch1:HasTallAncestor ch1).
      left.
      rewrite hch1.
      lia.
      assert (hhch2:HasTallAncestor ch2).
      left.
      rewrite hch2.
      lia.
      intros _ ha1 ha2.
      esplit with (exist hhch1);
        esplit with (exist hhch2).

      repeat split;simpl;red in hch1,hch2.
      rewrite hch1.
      reflexivity.
      left.
      reflexivity.
      assumption.
      rewrite hch2.
      reflexivity.
      left.
      reflexivity.
      assumption.
      apply hch.
      intro h.
      assert (hh:  (i <= Z.to_nat k)%nat).
      lia.
      specialize (Hi hh).
      intros halpha1 halpha2.
      assert (hhh:  (Z.to_nat k - i < Datatypes.S (Z.to_nat k))%nat).
      lia.
      specialize (Hi hhh hhh). 
      destruct Hi as [[b1 hb1] [[b2 hb2] [[h1a h1b] [[h2a h2b] eb]]]].
      red in eb;simpl in eb.
      unfold Rcount in *;simpl in *.
      destruct (Z_le_gt_dec (alpha (env b1)) 0) as [LE | GT].
      destruct (alpha (env b1)) as [| neg | pos];simpl in h1a,LE.
      lia.
      compute in LE.
      elim LE.
      reflexivity.
      lia.

      destruct (new_alpha_inv env_compat hassume b1)
        as [son [hson eson]]; auto. 
      { rewrite terminal_def_equiv in final.
        unfold terminal_alt in final.
        specialize (final _ _ (Equivalence_Reflexive b1)).
        unfold RUN, run in final; simpl in final; unfold run_algorithm.
        symmetry; apply final.
      }
      
      assert(hhson: HasTallAncestor son).
      right with b1;assumption.
      split with (exist hhson).
      split with (exist hhson).
      rewrite Zplus_comm in eson.
      assert (eson':=Zplus_minus_eq _ _ _ eson).
      repeat split;simpl;try rewrite eson'.

      rewrite Z2Nat.inj_sub;try lia.
 
      right with b1;auto.
      intro e;rewrite eson,e,Zplus_comm,Z2Nat.inj_add in h1a;try lia.

      lia.

      right with b2;auto.
      assert (B: b1 == b2) by apply eb.
      rewrite <- B; auto.

      intro e;rewrite eson,e,Zplus_comm,Z2Nat.inj_add in h1a;try lia.

      reflexivity.
    Qed.

    (** [split_counting_cases]
       if root is tall:
              -  all dominating heads are regular
              -  all nodes are regular (have a tall ancestor) 
       if root is short:
              -  all dominating heads are regular except the root
              -  at least one node is singular (namely the root) 
     *)
    Lemma split_counting_cases: 
      forall x: Node, 
        (( card_DH=card_RDH /\ card_N = card_RN ) \/
         ( card_DH = Datatypes.S card_RDH /\ card_N >= Datatypes.S card_RN))%nat.
    Proof.
      intros x; remember (proj1_sig (getRoot x)) as root.
      destruct (Z.geb_spec (alpha (env root)) k) as [hroot| hroot];
        [left|right];split.
      
      (* tall.1 *)
      {
        eapply (@Subset_Num_Card_iff_eq Node); eauto.
        intros n1 n2 en; unfold kDominator,RegDomHead.
        rewrite <- (env_compat _ _ en).
        rewrite Bool.orb_true_iff.
        rewrite Bool.andb_true_iff.
        rewrite Z.eqb_eq.
        simpl; intuition.
        rewrite is_root_equiv, <- DTN_root_ok in H1.
        eapply uniq_root in H1.
        2: apply (proj2_sig (getRoot x)).
        rewrite <- Heqroot in H1.
        rewrite <- H1 in H.
        rewrite <- H1.      
        unfold _IsShort in H.
        rewrite  Z.ltb_lt in H.
        lia.
        assumption.        
      }
      (* tall.2 *)
      {
        eapply (@Subset_Num_Card_iff_eq Node); eauto.
        eapply Card_trans with Node _.
        apply Trivial_subset_Num_Card.
        apply Hcard_N.
        intros _ n _. 
        split;[intros _| split].
        { apply subtree_down_induction with root.
          apply HasTallAncestor_compat.
          left; lia.
          intros nn _ hh c chi.
          right with nn; auto.
          apply in_children; auto.
          apply Dist_LCA_root.
          rewrite Heqroot; destruct (getRoot x); auto.
        }
      }
      (* case root is short *)

      (* short.1 *)
      { replace (Datatypes.S card_RDH) with (1+card_RDH)%nat. 
        2: lia.
        eapply (@Subset_Num_Card_iff_eq Node); eauto.
        eapply Disjoint_union_Num_Card.
        apply isRoot_compat.
        eapply RegDomHead_compat. 
        { apply Singleton_Num_Card with root.
          rewrite Heqroot; destruct (getRoot x); auto.
          intros a2 hr; apply uniq_root; auto.
          rewrite Heqroot; destruct (getRoot x); auto.
        }
        apply Hcard_RDH.
        { intros xx h1x h2x.
          simpl in h2x.
          red in h1x,h2x.
          apply uniq_root with root _ in h1x.
          2: rewrite Heqroot; destruct (getRoot x); auto.
          rewrite <- h1x in h2x; lia.
        }
        { intros n1 n2 en.
          rewrite or_comm.
          unfold kDominator.
          rewrite Bool.orb_true_iff.
          apply Morphisms_Prop.or_iff_morphism.
          unfold RegDomHead.
          rewrite Z.eqb_eq.
          apply PER_morphism.
          apply Equivalence_PER.
          apply eq_equivalence.
          apply alpha_compat, env_compat, en.
          reflexivity.
          rewrite Bool.andb_true_iff.
          rewrite <- en.
          split.
          intros [_ h].        
          rewrite is_root_equiv, <- DTN_root_ok in h; auto; apply h.
          intro h.
          split.
          2: rewrite is_root_equiv, <- DTN_root_ok; auto; apply h.
          apply uniq_root with root _ in h.
          2: rewrite Heqroot; destruct (getRoot x); auto.
          rewrite <- h; unfold _IsShort.
          rewrite Z.ltb_lt; lia.
        }
      }
      (* short.2 *)
      { rewrite <- Card_larger_ge_iff.
        apply Card_trans with Node _.
        { apply Card_weak. 
          apply Card_sym.
          apply Hcard_N.
        }
        eapply Card_trans with (sig (fun x : Node => isRoot x \/ HasTallAncestor x)) _.
        { apply Card_sym.
          simpl sym_Card_Prop. 
          eapply Subset_Num_Card.
          apply Union_compat; [apply isRoot_compat|apply HasTallAncestor_compat]; auto.
        }
        apply Card_weak.
        replace (Datatypes.S card_RN) with (1+card_RN)%nat. 
        2:lia.
        eapply Disjoint_union_Num_Card.
        apply isRoot_compat.
        apply HasTallAncestor_compat.
        { apply Singleton_Num_Card with root.
          rewrite Heqroot; destruct (getRoot x); auto.
          intros a2 hr; apply uniq_root; auto.
          rewrite Heqroot; destruct (getRoot x); auto.
        }
        apply Hcard_RN.
        { intros root' hroot'.
          apply uniq_root with root _ in hroot'.
          2: rewrite Heqroot; destruct (getRoot x); auto.
          rewrite <- hroot'.
          intro h;inversion h; subst.
          lia.
          assert (R := proj2_sig (getRoot x)).
          simpl in R; unfold isRoot in R; unfold isParent in H.
          rewrite H in R; inversion R.
        }
      }
    Qed.

    (* main_result *)

    Theorem counting_theorem_assuming_cardinals_exist :  
      forall x: Node, 
        (card_N - 1 >= (Datatypes.S (Z.to_nat k)) * (card_DH - 1))%nat.
    Proof.
      intros x.
      destruct split_counting_cases as [[HCH HN]|[HCH HN]].
      apply x.
      rewrite mult_minus_distr_l.
      unfold ge.
      etransitivity;[apply minus_le_compat_l|apply minus_le_compat_r].
      lia.
      rewrite HCH;rewrite HN.
      apply simple_counting.
      rewrite HCH.
      unfold ge in *|-*.
      simpl minus.
      rewrite <- minus_n_O.
      etransitivity;[apply simple_counting|].
      lia.
    Qed.

  End with_cardinals.

  (* [main_theorem] : Main counting Property *)

  Theorem counting_theorem_ 
          (card_N: nat)
          (Hcard_N: Num_Card Same Node card_N)
          (card_DH: nat)
          (Hcard_DH: Num_Card Same DomHeads card_DH) :
    forall x: Node, 
      (card_N - 1 >= (Datatypes.S (Z.to_nat k)) * (card_DH - 1))%nat.
  Proof.
    intros x; remember (proj1_sig (getRoot x)) as root.
    assert (tmp1: Proper fequiv (fun n: Node => RegDomHead n)).
    { intros n1 n2 en;apply RegDomHead_compat, en. }
    assert (tmp2: (forall a: Node,
                     (fun n: Node => RegDomHead n) a \/
                     ~ (fun n: Node => RegDomHead n) a)).
    { intro a;destruct (Z.eq_dec (alpha (env a)) k);
        [left|right];assumption. }
    assert (tmp3:
              (forall a1 a2: Node, a1 == a2 \/ ~ a1 == a2)).
    { intros a1 a2;destruct (Node_dec a1 a2);
        [left|right];assumption. }

    destruct (Classical_subset_Num_Card tmp3 (fun n: Node => RegDomHead n) tmp1 tmp2 all_nodes) as [card_RDH Hcard_RDH].
    { intros a _; apply all_nodes_prop. }
    clear tmp1 tmp2.

    assert (tmp2: (forall a: Node,
                      (fun n: Node => HasTallAncestor n) a \/
                      ~ (fun n: Node => HasTallAncestor n) a)).
    { intros a.
      eapply subtree_down_induction with (rT := root) (n := a).
      { intros n1 n2 en; rewrite en; auto. }
      { case (Z_lt_ge_dec (alpha (env root)) k); intros ha.
        - right; intros F; inversion F; subst.
          lia.
          assert (R := proj2_sig (getRoot x)).
          simpl in R; unfold isRoot in R; unfold isParent in H.
          rewrite H in R; inversion R.
        - left; left; auto.
      }
      { intros n _ [hn|hn] c chi; [left|].
        - right with n; auto.
          apply in_children; auto.
        - case (Z_lt_ge_dec (alpha (env c)) k); intros ha.
          + right; intros F; inversion F; subst.
            lia.
            apply in_children in chi; auto.
            apply isParent_unique with n _ _ in H; auto.
            rewrite <- H in H0; tauto.
          + left; left; auto.
      }
      apply Dist_LCA_root.
      rewrite Heqroot; destruct (getRoot x); auto.
    }
    destruct (Classical_subset_Num_Card
                tmp3
                (fun n: Node => HasTallAncestor n)
                HasTallAncestor_compat tmp2 all_nodes)
      as [card_RN Hcard_RN].
    intros a _; apply all_nodes_prop.
    apply counting_theorem_assuming_cardinals_exist
      with card_RDH card_RN; auto.
  Qed.

  Theorem counting_theorem_non_empty:
    forall x: Node,
    exists (card_N: nat) (card_DH: nat), 
      Num_Card Same Node card_N /\ Num_Card Same DomHeads card_DH /\ 
      (card_N - 1 >= (Datatypes.S (Z.to_nat k)) * (card_DH - 1))%nat.
  Proof.
    assert (tmp3:
              (forall a1 a2: Node, a1 == a2 \/ ~ a1 == a2)).
    intros a1 a2;destruct (Node_dec a1 a2);[left|right];assumption.

    destruct ( Classical_set_Num_Card tmp3 _ all_nodes_prop) 
      as [n Hn].

    assert (kdom_compat: Proper fequiv
                                (fun n: Node => (kDominator n = true))).
    { intros n1 n2 en; rewrite en; auto. }
    
    assert (kdom_classical: 
              forall n: Node, (fun n: Node => (kDominator n = true)) n \/ 
                           ~(fun n: Node => (kDominator n = true)) n).
    { intros n1;    
        destruct (Bool.bool_dec (kDominator n1) true); auto. }
    
    destruct (Classical_subset_Num_Card
                tmp3 _ kdom_compat 
                kdom_classical all_nodes) as [dh Hdh].
    { intros a _; apply all_nodes_prop. }
    intros x.
    exists n.
    exists dh.
    split; [|split]; try assumption.
    apply counting_theorem_; assumption.
  Qed.

  Theorem counting_theorem:
    exists (card_N: nat) (card_DH: nat), 
      Num_Card Same Node card_N /\ Num_Card Same DomHeads card_DH /\ 
      (card_N - 1 >= (Datatypes.S (Z.to_nat k)) * (card_DH - 1))%nat.
  Proof.
    assert (A := all_nodes_prop).
    destruct all_nodes as [| a all].
    2: { apply counting_theorem_non_empty, a. }

    assert (tmp3:
              (forall a1 a2: Node, a1 == a2 \/ ~ a1 == a2)).
    { intros a1 a2;destruct (Node_dec a1 a2);
        [left|right];assumption. }
    assert (kdom_compat: Proper fequiv
                                (fun n: Node => (kDominator n = true))).
    { intros n1 n2 en; rewrite en; auto. }
    
    assert (kdom_classical: 
              forall n: Node, (fun n: Node => (kDominator n = true)) n \/ 
                           ~(fun n: Node => (kDominator n = true)) n).
    { intros n1;    
        destruct (Bool.bool_dec (kDominator n1) true); auto. }
      
    assert (N: Num_Card Same Node 0).
    { 
      destruct ( Classical_set_Num_Card tmp3 _ all_nodes_prop) 
        as [n Hn].
      assert (T: n = 0%nat).
      eapply (@Subset_Num_Card_iff_eq Node).
      eapply (@Card_trans _ _ _ Node).
      apply (@Trivial_subset_Num_Card Node).
      apply Hn.
      apply Empty_subset_Num_Card.
      intros n1.
      specialize (A n1); rewrite InA_nil in A; tauto.
      rewrite <- T; auto.
    } 

    exists 0%nat; exists 0%nat; split; [|split]; auto.

    {
      destruct (Classical_subset_Num_Card
                  tmp3 _ kdom_compat 
                  kdom_classical all_nodes) as [dh Hdh].
      { intros a _; apply all_nodes_prop. }
      assert (T: dh = 0%nat).
      cut (le dh 0); [lia|].
      eapply (@Subset_Num_Card_impl_le Node).
      apply Hdh.
      eapply (@Card_trans _ _ _ Node).
      apply (@Trivial_subset_Num_Card Node).
      apply N.
      intros n1 n2 en h; tauto.
      rewrite <- T; auto.
    }
    lia.    
  Qed.
  

End KDomSet_count.

Close Scope Z_scope.
Close Scope signature_scope.
Unset Implicit Arguments.




