(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import ZArith.
From Coq Require Import Lia.
From Coq Require Import Bool.

(** **** Local Imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import OptionUtil.
From Padec Require Import ZUtils.
From Padec Require Import ListUtils.
From Padec Require Import FoldRight.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Self_Stabilization.
From Padec Require Import P_Q_Termination.
From Padec Require Import Tree_topology.
From Padec Require Import KDomSet_algo_tree_topology.

(* end show *)
Open Scope Z_scope.
Open Scope signature_scope.
Set Implicit Arguments.

(**
 *  Termination for K-Dominating Set Algorithm
 *)

(** Roadmap 

1) Define Potential of a node: 0 if not enabled, else tree depth
signature: node->env -> nat

NB: depth for root == 1

2) Proof: if a node becomes enabled, one of its child has moved

3) Proof (down path induction + 2)): if a node becomes enabled, one of its 
   descendant has become disabled

assumption: downpaths are finite (to be proved elsewhere)

4) Proof: there is always a node whose activation has changed
(potential of this node passed from >0 to 0)
(induction on down path)

5) Using 3) and multiset theorem (P_Q_Termination.v) concludes that the 
   algorithm terminates

=> Use results from P_Q_Termination
     - proof of local criterion: see 3)
     - proof of global criterion: see 4)

 **)
Section KDomSet_Termination.

  (* ********************************************************)
  (** Channel and Network definitions *)
  (* ********************************************************)
  Context {Chans: Channels}
          {Net: Network}
          {DTN: Down_Tree_Network Net}
          {KDP: KDomSet_parameters}.
  Let k_pos := k_pos.
  
  Notation Tree := (DTN_topo (Net := Net)). 
  Context {SPAN: is_spanning _ Tree}.

  Notation Algo := KDom_algo.
  Notation Env := (Env (Net := Net) (Algo := KDom_algo)).

  Existing Instance KDom_Stable.
  Existing Instance KDom_RO_assumption.

  Notation alphas_ :=
    (fun (env: Env) n => alphas (children_states (env n) (local_env env n))).
  
  Notation new_alpha_ := (fun env n => new_alpha (alphas_ env n)).

  Section Potential_Definition.

    (** ** Potential of a node: 
     0 if not enabled, else depth of the node in the spanning tree.
     *)

    Definition potential (env: Env) (n: Node): nat :=
      if enabled_b env n then 1 + Depth n else 0.

    Instance potential_compat: Proper pequiv potential.
    Proof.
      intros e1 e2 ee n1 n2 en; unfold potential.
      apply ITE_compat1; auto.
      apply enabled_b_compat; auto.
      apply f_equal, Depth_compat; auto.
    Qed.

  End Potential_Definition.
    
  (** ** Results for one step of execution *)
  Section With_Step.

    Variable (env env': Env)
             (step: Step env' env)
             (hassume: Assume env).

    (** Read Only are constant values *)
    Lemma children_c_same:
      forall n, equivlistA equiv (children_c (env n))
                           (children_c (env' n)).
    Proof.
      intros n.
      apply children_c_compat_RO.
      rewrite (Stable_Variable_step KDom_Stable step _ _ (Equivalence_Reflexive n)).
      reflexivity.
    Qed.

    Lemma is_root_same:
      forall n, is_root (env n) == is_root (env' n).
    Proof.
      intros n.
      apply is_root_compat_RO.
      rewrite (Stable_Variable_step KDom_Stable step _ _ (Equivalence_Reflexive n)).
      reflexivity.
    Qed.

    Lemma alpha_same:
      forall n, env n == env' n <-> alpha (env n) = alpha (env' n).
    Proof.
      intros n; split; simpl.
      tauto.
      intros a; split; [|split]; auto.
      apply is_root_same.
      apply children_c_same.
    Qed.

    Lemma alpha_change:
      forall n, ~ env n == env' n <-> alpha (env n) <> alpha (env' n).
    Proof.
      intros n; split; simpl.
      intros G F; contradict G.
      apply alpha_same; auto.
      intros G F; contradict G.
      apply alpha_same; apply F.
    Qed.

    (** New alphas only depends on children moves *)
    Lemma alphas_only_depends_on_children:
        forall (n: Node),
          (forall (child: Node), isParent n child -> env child = env' child)
          -> equivlistA equiv (alphas_ env n) (alphas_ env' n).
      Proof.
        intros n Hchild.
        unfold alphas.
        apply map_compat_equiv.
        apply alpha_compat.
        apply children_states_depends.
        apply children_c_same.
        { intros x y xy.
          assert (X := Step_Proper_r step).
          rewrite xy; auto. }
        { intros x y xy.
          assert (X := Step_Proper_l step).
          rewrite xy; auto. }
        { intros c hc.
          unfold local_env.
          case_eq (peer n c); [| auto]; intros child hch; simpl option_map.
          rewrite Hchild; auto.
          apply DTN_children_ok.
          exists c; unfold is_channel; split.
          rewrite children_c_equiv in hc; auto.
          rewrite hch; auto.
        }
      Qed.          
      
      (** New alpha only depends on children moves *)
      Lemma new_alpha_only_depends_on_children:
        forall (n: Node),
          (forall (child: Node), isParent n child -> env child = env' child)
          -> new_alpha_ env n = new_alpha_ env' n.
      Proof.
        intros n Hchild.
        apply new_alpha_compat.
        apply alphas_only_depends_on_children; auto.
      Qed.        
        
      (** When a node that was disabled becomes enabled, one of its child has actually moved. *)
      Lemma enabled_child_has_moved:
        forall (n: Node),
          enabled_b env n = false -> enabled_b env' n = true ->
          exists (child: Node),
            isParent n child /\ ~ env child == env' child.
      Proof.
        intros n Hdis Hen.
        assert (hr: ~ (RUN env n) == (RUN env' n)).
        { assert (hen: RUN env n == env n).
          { assert (dum := enabled_prop env n).
            unfold enabled in dum.
            rewrite Hdis in dum.
            destruct (State_dec (RUN env n) (env n)); auto.
            rewrite <- dum in n0; try congruence. }
          assert (hen': RUN env n == env' n).
          { destruct (Step_correct step n) as [st|st]; symmetry; auto.
            rewrite hen; auto. }
          rewrite hen'.
          unfold enabled_b in Hen.
          destruct (State_dec (RUN env' n) (env' n)); try congruence; intuition.
        }
        clear Hdis Hen.
        unfold RUN, run in hr; simpl in hr.
        assert (C: ~equivlistA equiv
                    (children_states (env n) (local_env env n))
                    (children_states (env' n) (local_env env' n))).
        { intros F; contradict hr; split; [|split]; auto.
          2: apply is_root_same.
          2: apply children_c_same.
          apply new_alpha_compat, alphas_compat; auto. }
        clear hr.
        unfold children_states in C.
        apply map_filter_eq_locality_converse in C; auto.
        2: apply KDom_State_dec.
        2: apply children_c_same. 
        2: { intros x y xy.
             assert (X := Step_Proper_r step).
             rewrite xy; auto. }
        2: { intros x y xy.
             assert (X := Step_Proper_l step).
             rewrite xy; auto. }
        destruct C as [c_child [ina nloc]].
        rewrite children_c_equiv in ina; auto.
        dup ina ina'.
        apply DTN_channels_ok in ina.
        destruct ina as [child hch].
        exists child; split.
        apply DTN_children_ok.
        exists c_child; auto.
        intros A; contradict nloc.
        unfold is_channel in hch; unfold local_env.
        assert (Pl := Step_Proper_l step); 
          assert (Pr := Step_Proper_r step); rewrite hch;
            simpl option_map; clear Pl Pr.
        rewrite A; auto.
      Qed.

      (** When a node has actually moved while none of its children moved, 
     it is no more enabled at next configuration. *)
      Lemma node_has_moved_but_no_child:
        forall n, ~ env n == env' n ->
                  (forall child, isParent n child ->
                                 env child = env' child) ->
                  enabled_b env' n = false. 
      Proof.
        intros n Hmov Hchild.
        destruct (Step_correct step n) as [st|st].
        2: { symmetry in st; tauto. }
        clear Hmov.
        apply alphas_only_depends_on_children in Hchild.
        unfold enabled_b.
        case (State_dec (RUN env' n) (env' n)); auto.
        intros se; contradict se.        
        rewrite st.
        unfold RUN, run; simpl; unfold run_algorithm.
        split; [symmetry|split; symmetry;
                 [apply is_root_same|apply children_c_same]].
        apply new_alpha_compat; auto.
      Qed.
      
      (** When a node has moved, but is still enabled at next configuration, 
     it has a child that actually also moved. *)
      Lemma node_has_moved:
        forall n,
          ~ env n == env' n -> enabled_b env' n = true ->
          exists child,
            isParent n child /\ ~ env child == env' child.
      Proof.
        intros n Hm Hen.
        destruct (Step_correct step n) as [st|st].
        2: { symmetry in st; tauto. }
        clear Hm.
        unfold enabled_b in Hen.
        case_eq (State_dec (RUN env' n) (env' n));
          intros e dum; rewrite dum in Hen; clear dum;
            [congruence|clear Hen].
        rewrite st in e; clear st.
        unfold RUN, run in e; simpl in e; unfold run_algorithm in e.
        assert (C: ~equivlistA equiv
                    (children_states (env n) (local_env env n))
                    (children_states (env' n) (local_env env' n))).
        { intros F; contradict e; split; [|split]; auto; symmetry.
          2: apply is_root_same.
          2: apply children_c_same.
          apply new_alpha_compat, alphas_compat; auto. }
        clear e.
        unfold children_states in C.
        apply map_filter_eq_locality_converse in C; auto.
        2: apply KDom_State_dec.
        2: apply children_c_same.
        2: { intros x y xy.
             assert (X := Step_Proper_r step).
             rewrite xy; auto. }
        2: { intros x y xy.
             assert (X := Step_Proper_l step).
             rewrite xy; auto. }
        destruct C as [c_child [ina nloc]].
        rewrite children_c_equiv in ina; auto.
        dup ina ina'.
        apply DTN_channels_ok in ina.
        destruct ina as [child hch].
        exists child; split.
        apply DTN_children_ok.
        exists c_child; auto.
        intros A; contradict nloc.
        unfold is_channel in hch; unfold local_env.
        assert (Pl := Step_Proper_l step); 
          assert (Pr := Step_Proper_r step); rewrite hch;
            simpl option_map; clear Pl Pr.
        rewrite A; auto.
      Qed.
      
      (** When a node actually moved, it is down-linked (in the spanning tree) to a node 
     which was enabled and became disabled. *)
      Lemma moving_node_has_disabled_desc:
        forall (n: Node),
          ~ env n == env' n -> 
          exists (descendant: Node), 
            isAncestor n descendant /\
            enabled_b env descendant = true /\ 
            enabled_b env' descendant = false.
      Proof.        
        intros n Hmov.
        apply path_down_induction
          with (P := fun n => ~ env n == env' n)
               (Q := fun d => enabled_b env d = true /\
                              enabled_b env' d = false); auto.
        { intros n1 n2 en.
          assert (x := Step_Proper_r step).
          assert (y := Step_Proper_l step).
          rewrite en; auto.
        }
        { intros n1 n2 en.
          assert (x := Step_Proper_r step).
          assert (y := Step_Proper_l step).
          rewrite en; auto.
        }
        { intros a; destruct (enabled_b env a); destruct (enabled_b env' a).
          right; intros [_ x]; congruence.
          left; auto.
          right; intros [x _]; congruence.
          right; intros [x _]; congruence.
        }
        intros m hm hen.
        clear Hmov.
        assert (hen': enabled_b env m = false \/ enabled_b env' m = true).
        { revert hen; destruct (enabled_b env m); destruct (enabled_b env' m); auto. }
        clear hen; destruct hen' as [hen|hen].
        { rewrite <- not_true_iff_false in hen; contradict hen.
          unfold enabled_b.
          case (State_dec (RUN env m) (env m)); auto.
          intros r; destruct (Step_correct step m);
            contradict hm.
          rewrite <- r; symmetry; auto.
          rewrite H; auto.
        }
        { apply node_has_moved in hm; auto.
          destruct hm as [ch [par hch]].
          exists ch; split; [|left]; auto.
          apply in_children; auto.
        }
      Qed.      

     (** When a node that was disabled becomes enabled at next configuration, it is
     down-linked (in the spanning tree) to a node which was enabled and becomes disabled. *)
      Lemma new_enabled_node_has_disabled_descendant:
        forall (n: Node),
          enabled_b env n = false -> enabled_b env' n = true ->
          exists (descendant: Node), 
            isAncestor n descendant /\
            enabled_b env descendant = true /\ 
            enabled_b env' descendant = false.
      Proof.      
        intros n Hdisabled Henabled.
        assert (Hc:= enabled_child_has_moved _ Hdisabled Henabled).
        destruct Hc as (n', Hc).
        destruct Hc as (Hn', Hmov).
        specialize (moving_node_has_disabled_desc Hmov); intro.
        destruct H as (desc, Hd).
        destruct Hd as (Hpath, Hd).
        exists desc.
        split.
        2: apply Hd.
        apply isAncestor_trans with n'; auto.
        apply isAncestor_Parent; auto.
      Qed.      

      (** *** 4) Proof: 
     there is always a node that becomes disabled after a step
    (potential of this node passed from >0 to 0) (induction on down path)
       *)

      (** Within a valid step, there exists a node that was enabled but becomes disabled *)
      Lemma one_disabled: 
        exists (n: Node), enabled_b env n = true /\ enabled_b env' n = false.
      Proof.
        destruct (Step_one_has_moved_true step) as [n mov].
        apply has_moved_progress in mov.
        assert (mov': ~env n == env' n).
        { contradict mov; symmetry; auto. }
        assert (Hdesc := moving_node_has_disabled_desc mov').
        destruct Hdesc as (d, Hd).
        destruct Hd as (Hd', Hd).
        exists d. exact Hd.
      Qed.

      (** *** 5) Using 3) and multiset theorem (Termination.v, Wf_Algo_Multiset) conclude that the 
    algorithm terminates
          - proof of local criterion: see 3)
          - proof of global criterion: see 4)
       *)

      (** *** Termination Criteria *)
      Section Potential_termination_criteria.

        Lemma diff_pot:
          forall n, potential env' n <> potential env n
                    <->
                    (enabled_b env n = false <-> enabled_b env' n = true).
        Proof.
          intro n.
          split; unfold potential;
            destruct (enabled_b env' n); destruct (enabled_b env n); try lia; intros; try split; auto.
        Qed.

        (** Local Criterion *)
        Lemma kHopDom_local_criterion:
          forall (n: Node), 
            gt (potential env' n) (potential env n)
            -> exists (n': Node),
                 potential env n' <> potential env' n'
                 /\
                 gt (potential env n') (potential env' n).
        Proof.
          intros n Hpot_n_.
          case (eq_nat_dec (potential env n) (potential env' n)); 
            intros Hpot_n.
          rewrite Hpot_n in Hpot_n_.
          contradict Hpot_n_; apply gt_irrefl. 
          clear Hpot_n_.
          destruct (le_gt_dec (potential env n) (potential env' n)) 
            as [ le | gt ].

      (*gt*) 
      2: { exists n; split; [apply Hpot_n | apply gt]. }

      (*lt*)      
      assert (pot_nul: potential env n = 0%nat).
      revert Hpot_n le.
      unfold potential.
      destruct (enabled_b env n); destruct (enabled_b env' n); intros;
        lia.

      assert (Hen: potential env' n <> potential env n); auto.
      rewrite diff_pot in Hen.

      assert (Hdisabled: enabled_b env n = false). 
      revert pot_nul.
      unfold potential.
      destruct (enabled_b env n); intros; auto; lia.

      generalize Hdisabled; intro Henabled; rewrite Hen in Henabled.
      assert (Hdesc := new_enabled_node_has_disabled_descendant 
                         _ Hdisabled  Henabled).
      destruct Hdesc as (d, (Hp, Hd)).
      exists d.
      generalize Hd; intro tmp.
      destruct tmp.
      assert (neq: ~(n == d)).
      intro.      
      symmetry in H1.
      assert (pr := Step_Proper_r step).
      rewrite H1 in H.
      rewrite H in Hdisabled; congruence.
      clear H H0.

      destruct Hd as (Hd, Hd').

      unfold potential; rewrite Hd, Hd', Henabled.
      split; [lia|].
      assert (D := deeper_Depth _ _ neq Hp).
      lia.
    Qed.

        (** ** Global Criterion *)
        Lemma kHopDom_global_criterion:
          exists n, potential env' n <> potential env n.
        Proof.
          assert (one := one_disabled).  
          destruct one as (n, (one, one')).
          exists n.
          rewrite  diff_pot.
          rewrite one, one'.
          split; congruence.
        Qed.

      End Potential_termination_criteria.

  End With_Step.
  
  Section Termination.
    
    Lemma kDOM_termination: termination KDom_RO_assumption. 
    Proof.
      intros env hassume penv.
      apply Acc_Algo_Multiset with Assume; auto.
        
      { (* stable *)
        intros g g' hh step.
        revert hh; apply Assume_RO_compat.
        eapply (Stable_Variable_step KDom_Stable); auto.
      }

      { (* well founded *)
        apply NoQTrans_Wf_Multiset with (fun g' g => True);
          [| intros; tauto].
        pose (pot := (fun sg : {g : Env | Assume g} =>
                        potential (proj1_sig sg))). 
        apply Wf_safeQAlgo_Multiset with pot.
        { (* pot_compat *)
          unfold pot.
          intros [g1 sg1] [g2 sg2] e; simpl.
          apply potential_compat; auto.
        }
        { (* local criterion *)
          intros [g1 sg1] [g2 sg2] [step _]; unfold pot; simpl.
          apply kHopDom_local_criterion; auto.
        }
        { (* global criterion *)
            intros [g1 sg1] [g2 sg2] [step _]; unfold pot; simpl.
            apply kHopDom_global_criterion; auto.
        }
      }
    Qed.

  End Termination.

End KDomSet_Termination.

Close Scope signature_scope.
Close Scope Z_scope.
Unset Implicit Arguments.
