(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Oct 2016
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import Relations.
From Coq Require Import RelationPairs.
From Coq Require Import Bool.
From Coq Require Import Wellfounded.
From Coq Require Import SetoidClass.

(** **** Local Imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import ListUtils.
From Padec Require Import OptionUtil.
From Padec Require Import WellfoundedUtil.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Self_Stabilization.
From Padec Require Import Stream.
From Padec Require Import Simulation.
From Padec Require Import Exec.
From Padec Require Import Fairness.

(* end show *)
Open Scope signature_scope.
Set Implicit Arguments.

Section Composition.

  Context {Chans: Channels} {Net: Network}.

  (** ** Algorithm 1 
   Definitions and Notations.
   *)
  Variable (Algo1: Algorithm).
  Notation S1 := (@State Chans Algo1).
  Notation Env1:= (Env (Algo := Algo1)).
  Notation Exec1 := (Exec (Algo:=Algo1)).

  Variable Assumption1: Stable_Assumption (Algo := Algo1).
  Existing Instance Assumption1.
  Notation Assume1 := (Assume (Stable_Assumption := Assumption1)).
  Notation Assume1_RO := (Assume_RO (Stable_Assumption := Assumption1)).
  Instance Assume1_RO_compat: Proper fequiv Assume1_RO
    := (Assume_RO_compat (Stable_Assumption := Assumption1)).
  Notation RO1 := (ROType (Stable_Assumption := Assumption1)).
  Notation RO1_part := (getRO (Stable_Assumption := Assumption1)).

  Instance RO1_Stable: Stable_Variable Algo1 RO1_part
    := (ROstable (Stable_Assumption := Assumption1)).
  Instance RO1_part_compat: Proper fequiv RO1_part
    := (GetVar_compat (Stable_Variable := RO1_Stable)).
  Instance RO1_Psetoid: PartialSetoid RO1
    := (VarType_setoid (Stable_Variable := RO1_Stable)).
  Notation RO1_part_stable := (GetVar_stable (Stable_Variable := RO1_Stable)).
  
  (** ** Algorithm 2 
   Definitions and Notations.
   *)
  Variable (Algo2: Algorithm).
  Notation S2 := (@State Chans Algo2).
  Notation Env2:= (Env (Algo := Algo2)).
  Notation Exec2 := (Exec (Algo:=Algo2)).

  Variable Assumption2: Stable_Assumption (Algo := Algo2).
  Existing Instance Assumption2.
  Notation Assume2 := (Assume (Stable_Assumption := Assumption2)).
  Notation Assume2_RO := (Assume_RO (Stable_Assumption := Assumption2)).
  Instance Assume2_RO_compat: Proper fequiv Assume2_RO
    := (Assume_RO_compat (Stable_Assumption := Assumption2)).
  Notation RO2 := (ROType (Stable_Assumption := Assumption2)).
  Notation RO2_part := (getRO (Stable_Assumption := Assumption2)).

  Instance RO2_Stable: Stable_Variable Algo2 RO2_part
    := (ROstable (Stable_Assumption := Assumption2)).
  Instance RO2_part_compat: Proper fequiv RO2_part
    := (GetVar_compat (Stable_Variable := RO2_Stable)).
  Instance RO2_Psetoid: PartialSetoid RO2
    := (VarType_setoid (Stable_Variable := RO2_Stable)).
  Notation RO2_part_stable := (GetVar_stable (Stable_Variable := RO2_Stable)).
  
  (** ** Algorithm 3 (Composite Algorithm)

   Definitions and Notations.

   We build the third algorithm as the composition of Algo1 and Algo2
   as follows.

   S3 is the state type for Algo3.

      - Algo3 is given the values of variables of Algo1 (i.e. in some
        state of type S1) using function read1: S3 -> S1

      - Algo3 can modify the values of variables of Algo1 (i.e. in
        some state of type S1) using function write1: S3 -> S1 -> S3 -
        same for Algo2, read2 and write2. 

 *)

  Class Composition_State: Type:=
    {
      S3: Type;
      setoid: Setoid S3;
      S3_dec: Decider (equiv (A:=S3));

      read1: S3 -> S1;
      write1: S3 -> S1 -> S3;
      read2: S3 -> S2;
      write2: S3 -> S2 -> S3;

      read1_compat: Proper fequiv read1;
      write1_compat: Proper fequiv write1;
      read2_compat: Proper fequiv read2;
      write2_compat: Proper fequiv write2;

      (** Guarantees that write1 actually writes the S1 part of the
          state, since the projection of the modified state on the
          result allows to get back what was written. (same for S2) *)
      stabRW1: forall (s1: S1) (s3: S3), read1 (write1 s3 s1) == s1;
      stabRW2: forall (s2: S2) (s3: S3), read2 (write2 s3 s2) == s2;

      (** Apart from the read-only part of S2, write2 cannot modified
          the S1 part of the state.  The idea behind this constraint
          is that Algo2 cannot modify the (S2-writtable part of the)
          data computed by Algo1.  (Algo2 cannot either modify the
          (S2-read-only part) because it is read-only).  *)
      stabRO12: forall (s2: S2) (s3: S3),
          RO2_part s2 =~= RO2_part (read2 s3)
          -> read1 (write2 s3 s2) == read1 s3
    }.

  Context `{CS: Composition_State}.

  Existing Instance setoid.

  Existing Instance read1_compat.
  Existing Instance write1_compat.
  Existing Instance read2_compat.
  Existing Instance write2_compat.

  (** projection on S1 of a configuration *)
  Notation lenv1 := (fun sn: Channel -> option S3 =>
                       (fun c => option_map (read1 (Composition_State:=CS)) (sn c))).

  (** projection on S2 of a configuration *)
  Notation lenv2 := (fun sn: Channel -> option S3 =>
                       (fun c => option_map (read2 (Composition_State:=CS)) (sn c))).

  Instance lenv1_compat: Proper fequiv lenv1.
  Proof.
    intros sn1 sn2 esn c1 c2 ec.
    rewrite (esn _ _ ec); reflexivity.
  Qed.

  Instance lenv2_compat: Proper fequiv lenv2.
  Proof.
    intros sn1 sn2 esn c1 c2 ec.
    rewrite (esn _ _ ec); reflexivity.
  Qed.

  (** ** RUN function of Algo3

   Now, Algo3 is the hierarchical composition of Algo1 and Algo2: for
   any node, p, it is enabled either if Algo1 or Algo2 is enabled The
   function run on p returns

   - the result of Algo1 on p, if Algo1 is enabled on p

   - otherwise, the result of Algo2 on p otherwise.

   Namely, at each process, Algo1 has priority over Algo2.  

   *)
  Definition composition_run (s: S3) (sn: Channel -> option S3):
    S3 :=
    if State_dec (run (read1 s) (lenv1 sn)) (read1 s)
    then
      if State_dec (run (read2 s) (lenv2 sn)) (read2 s)
      then s
      else write2 s (run (read2 s) (lenv2 sn))
    else write1 s (run (read1 s) (lenv1 sn)).

  Instance composition_run_compat: Proper fequiv composition_run.
  Proof.
    intros s1 s2 S l1 l2 L; unfold composition_run.
    assert (L1 := lenv1_compat L).
    assert (L2 := lenv2_compat L).
    assert (Rd1 := read1_compat _ _ S).
    assert (Rd2 := read2_compat _ _ S).
    assert (R1 := run_compat _ _ Rd1 L1).
    assert (R2 := run_compat _ _ Rd2 L2).
    destruct (State_dec (run (read1 s1) (lenv1 l1)) (read1 s1));
      destruct (State_dec (run (read1 s2) (lenv1 l2)) (read1 s2)).
    { destruct (State_dec (run (read2 s1) (lenv2 l1)) (read2 s1));
        destruct (State_dec (run (read2 s2) (lenv2 l2)) (read2 s2)).
      { apply S. }
      { contradict n; rewrite <- R2, <- Rd2; auto. }
      { contradict n; rewrite R2, Rd2; auto. }
      { apply write2_compat; auto. }
    }
    { contradict n; rewrite <- R1, <- Rd1; auto. }
    { contradict n; rewrite R1, Rd1; auto. }
    { apply write1_compat; auto. }
  Qed.

  (** ** Definition of Algo3
   Definition of read-only variables and Assumption
   *)
  Instance Collateral_composition: Algorithm :=
    {| State_dec := S3_dec; run_compat := composition_run_compat |}.

  Notation Algo3 := Collateral_composition.
  Notation Env3 := (Env (Algo := Algo3)).
  Notation Exec3 := (Exec (Algo:=Algo3)).

  (** The read-only part of Algo3 is made of the read-only variables of Algo1 *)
  Let RO3 := RO1.
  Instance RO3_Psetoid: PartialSetoid RO3 := RO1_Psetoid.
  Notation RO3_part := (fun s: S3 => RO1_part (read1 s)).
  Instance RO3_part_compat: Proper fequiv RO3_part.
  Proof.
    intros s1 s2 es; apply RO1_part_compat, read1_compat; auto.
  Qed.
  
  Lemma RO3_part_stable: forall s sn,
      RO3_part (run s sn) =~= RO3_part s.
  Proof.
    intros s sn.
    unfold run; simpl; unfold composition_run.
    destruct (State_dec (run (read1 s) (lenv1 sn)) (read1 s)).
    destruct (State_dec (run (read2 s) (lenv2 sn)) (read2 s)).
    apply RO1_part_compat,read1_compat; reflexivity.
    - rewrite stabRO12; auto.
      apply RO1_part_compat; auto.
      apply RO2_part_stable.
    - rewrite stabRW1.
      apply RO1_part_stable; auto.
  Qed.

  Instance RO3_Stable: Stable_Variable Algo3 RO3_part.
  Proof.
    esplit.
    apply RO3_part_compat.
    apply RO3_part_stable.
  Defined.

  (** The read-only Assumption for Algo3 is given by the read-only
  assumption of Algo1 *)
  Definition Assume3_RO: (Node -> RO3) -> Prop :=
    (fun r => Assume1_RO (fun n => (r n))).
  
  Instance Assume3_RO_compat: Proper fequiv Assume3_RO.
  Proof.
    intros r1 r2 er; apply Assume1_RO_compat; auto.
  Qed.
  
  Instance Assumption3: Stable_Assumption (Algo := Algo3).
  Proof.
    esplit.
    apply Assume3_RO_compat.
  Defined.

  Notation Assume3 := (Assume (Stable_Assumption := Assumption3)).
  
  (** ** Projections from 3 to 1 *)
  Definition envread1 (g: Env3): Env1 := fun n => read1 (g n).

  Lemma envread1_compat: Proper fequiv envread1.
  Proof.
    intros g k H1; unfold envread1.
    intros m n H2; apply read1_compat.
    apply H1; apply H2.
  Qed.
  
  Lemma envread1_prop:
    forall g1 g2 n, g1 =~= g2 ->
                    (local_env (envread1 g1) n)
                      =~=
                      fun c => option_map read1 (local_env g2 n c) .
  Proof.
    intros g1 g2 n eg c1 c2 ec.
    unfold local_env,envread1.
    assert (h:peer n c1 == peer n c2).
    { apply peer_compat;auto. }    
    destruct (peer n c1);destruct (peer n c2);simpl in *;tauto || auto.
    apply read1_compat;auto.
  Qed.

  Lemma Assume_31:
    forall g, Proper pequiv g ->
              (Assume3 g <-> Assume1 (envread1 g)).
  Proof.
    unfold Assume, Assume_RO, ROEnv_part, envread1 in *; simpl in *.
    intros g pg. 
    unfold Assume3_RO.
    apply Assume1_RO_compat.
    intros n1 n2 en.
    apply RO1_part_compat, read1_compat, pg; auto.
  Qed.

  (** ** Projections from 3 to 2 *)
  Definition envread2 (g: Env3): Env2 := fun n => read2 (g n).

  Lemma envread2_compat: Proper fequiv envread2.
  Proof.
    intros g k H1; unfold envread2.
    intros m n H2; apply read2_compat.
    apply H1; apply H2.
  Qed.
  
  Lemma envread2_prop:
    forall g1 g2 n, g1 =~= g2 ->
                    (local_env (envread2 g1) n)
                      =~=
                      fun c => option_map read2 (local_env g2 n c) .
  Proof.
    intros g1 g2 n eg c1 c2 ec.
    unfold local_env,envread1.
    assert (h:peer n c1 == peer n c2).
    { apply peer_compat;auto. }    
    destruct (peer n c1);destruct (peer n c2);simpl in *;tauto || auto.
    apply read2_compat;auto.
  Qed.
  
  (** ** Technical Tools about RUN for Algo1, Algo2 and Algo3 *)

  Lemma run1_rewrite:
    forall g g': Env3,
      g =~= g' ->
      forall n n', n == n' ->
                   run (read1 (g n)) (lenv1 (local_env g n))
                   == RUN (envread1 g') n'.
  Proof.
    intros g g' eg n n' en.
    apply run_compat.
    apply read1_compat, eg, en.
    intros c c' ec.
    unfold local_env.
    generalize (peer_compat _ _ en _ _ ec);
      destruct (peer n c); destruct (peer n' c');
        simpl; try tauto.
    intros H; apply read1_compat, eg, H.
  Qed.

  Lemma run2_rewrite:
    forall g g': Env3,
      g =~= g' -> forall n n', n == n' ->
                               run (read2 (g n)) (lenv2 (local_env g n))
                               == RUN (envread2 g') n'.
  Proof.
    intros g g' eg n n' en.
    apply run_compat.
    apply read2_compat; apply eg; apply en.
    intros c c' ec.
    unfold local_env, envread2.
    generalize (peer_compat _ _ en _ _ ec);
      destruct (peer n c); destruct (peer n' c');
        simpl; try tauto.
    intros H; apply read2_compat, eg, H.
  Qed.

  (** Alternative definition for RUN3 *)
  Definition composition_RUN_alt_def (g: Env3) (n: Node): S3 :=
    if enabled_b (envread1 g) n then
      write1 (g n) (RUN (envread1 g) n)
    else if enabled_b (envread2 g) n then
           write2 (g n) (RUN (envread2 g) n)
         else (g n).

  Lemma composition_RUN_alt:
    fequiv RUN composition_RUN_alt_def.
  Proof.
    intros g' g eg n' n en.
    rewrite (RUN_compat eg _ _ en).
    assert (pg: Proper pequiv g) by proper_.
    clear g' eg n' en.    
    unfold RUN,composition_RUN_alt_def,enabled_b, run;
      simpl; unfold composition_run.
    destruct (State_dec
                (run (read1 (g n)) (lenv1 (local_env g n))) 
                (read1 (g n))) as [h1| h1];
      rewrite <- envread1_prop in h1;try apply pg;
        destruct (State_dec
                    (run (read2 (g n)) (lenv2 (local_env g n))) 
                    (read2 (g n))) as [h2| h2];
        rewrite <- envread2_prop in h2;try apply pg;
          destruct (State_dec (RUN (envread1 g) n) (envread1 g n));try tauto;
            destruct (State_dec (RUN (envread2 g) n) (envread2 g n));try tauto.
    auto.
    apply write2_compat ; auto;unfold RUN;rewrite envread2_prop;
      auto;reflexivity.
    apply write1_compat ; auto;unfold RUN;rewrite envread1_prop;auto;
      reflexivity.
    apply write1_compat ; auto;unfold RUN;rewrite envread1_prop;auto;
      reflexivity.
  Qed.

  Lemma RUN_31_enabled:
    forall (g: Env3) (n: Node),
      Proper fequiv g ->
      enabled_b (envread1 g) n = true ->
      read1 (RUN g n) == (RUN (envread1 g) n).
  Proof.
    intros g n pg en1.
    rewrite (composition_RUN_alt pg _ _ (Equivalence_Reflexive n)).
    unfold composition_RUN_alt_def.
    rewrite en1.
    apply stabRW1.
  Qed.

  Lemma RUN_31_disabled:
    forall (g: Env3) (n: Node),
      Proper fequiv g ->
      enabled_b (envread1 g) n = false ->
      read1 (RUN g n) == (envread1 g) n.
  Proof.
    intros g n pg dis1.
    rewrite (composition_RUN_alt pg _ _ (Equivalence_Reflexive n)).
    unfold composition_RUN_alt_def.
    rewrite dis1.
    case_eq (enabled_b (envread2 g) n);[intro en2|intro dis2].
    apply stabRO12.
    apply RO2_part_stable.
    reflexivity.
  Qed.

  Lemma RUN_32_enabled:
    forall (g: Env3) (n: Node),
      Proper fequiv g ->
      enabled_b (envread1 g) n = false ->
      enabled_b (envread2 g) n = true -> 
      read2 (RUN g n) == (RUN (envread2 g) n).
  Proof.
    intros g n pg dis1 en2.
    rewrite (composition_RUN_alt pg _ _ (Equivalence_Reflexive n)).
    unfold composition_RUN_alt_def.
    rewrite dis1,en2.
    apply stabRW2.
  Qed.

  Lemma RUN_32_disabled:
    forall (g: Env3) (n: Node),
      Proper fequiv g ->
      enabled_b (envread1 g) n = false ->
      enabled_b (envread2 g) n = false -> 
      read2 (RUN g n) == (envread2 g) n.
  Proof.
    intros g n pg dis1 dis2.
    rewrite (composition_RUN_alt pg _ _ (Equivalence_Reflexive n)).
    unfold composition_RUN_alt_def.
    rewrite dis1,dis2.
    reflexivity.
  Qed.

  Lemma RUN_3_disabled:
    forall (g: Env3) (n: Node),
      Proper fequiv g ->
      enabled_b (envread1 g) n = false ->
      enabled_b (envread2 g) n = false -> 
      RUN g n == g n.
  Proof.
    intros g n pg dis1 dis2.
    rewrite (composition_RUN_alt pg _ _ (Equivalence_Reflexive n)).
    unfold composition_RUN_alt_def.
    rewrite dis1,dis2.
    reflexivity.
  Qed.

  (** ** Technical Tools about enabled_b for Algo1, Algo2, Algo3 *)
  Lemma enabled_12_3:
    forall g n, Proper pequiv g ->
                (enabled_b g n = true <-> enabled_b (envread1 g) n = true
                                          \/
                                          enabled_b (envread2 g) n = true).
  Proof.
    intros g n pg; split.
    { intros en3; unfold enabled_b in en3.
      destruct (State_dec (RUN g n) (g n)) as [e3|e3];
        [congruence|]; clear en3.
      rewrite (composition_RUN_alt pg _ _ (Equivalence_Reflexive n)) in e3.
      unfold composition_RUN_alt_def in e3.
      destruct (enabled_b (envread1 g) n); [left; auto|].
      destruct (enabled_b (envread2 g) n); [right; auto|].
      contradict e3; auto.
    }
    { intros dis; unfold enabled_b.
      destruct (State_dec (RUN g n) (g n)) as [e3|e3];
        [| auto].
      rewrite (composition_RUN_alt pg _ _ (Equivalence_Reflexive n)) in e3;
        unfold composition_RUN_alt_def in e3.
      case_eq (enabled_b (envread1 g) n); intros E1;
        rewrite E1 in dis, e3.
      - unfold enabled_b in E1.
        destruct (State_dec (RUN (envread1 g) n) (envread1 g n)) as [S1|S1];
          [congruence|].
        contradict S1.
        unfold envread1; rewrite <- e3.
        rewrite stabRW1.
        reflexivity.
      - destruct dis as [|dis]; [congruence|].
        rewrite dis in e3.
        unfold enabled_b in dis.
        destruct (State_dec (RUN (envread2 g) n) (envread2 g n)) as [|S2];
          [congruence|].
        contradict S2.
        unfold envread2; rewrite <- e3.
        rewrite stabRW2.
        reflexivity.
    }
  Qed.

  Lemma enabled_12_3_alt:
    forall g n, Proper pequiv g ->
                enabled_b g n = enabled_b (envread1 g) n
                                || enabled_b (envread2 g) n.
  Proof.
    intros g n pg; assert (ok := enabled_12_3 n pg).
    destruct (enabled_b g n);
      destruct (enabled_b (envread1 g) n);
      destruct (enabled_b (envread2 g) n); 
      try congruence; intuition; auto with bool.
  Qed.
  
  Lemma enabled_13:
    forall g3 n, Proper fequiv g3 ->
                 enabled_b (envread1 g3) n = true -> enabled_b g3 n = true.
  Proof.
    intros g3 n pg E1; apply enabled_12_3; auto.
  Qed.

  Lemma enabled_23:
    forall g3 n,
      Proper fequiv g3 ->
      enabled_b (envread2 g3) n = true ->
      enabled_b g3 n = true.
  Proof.
    intros g3 n pg E1; apply enabled_12_3; auto.
  Qed.

  Lemma enabled_32:
    forall g3 n,
      Proper fequiv g3 ->
      enabled_b (envread1 g3) n = false ->
      enabled_b g3 n = true ->
      enabled_b (envread2 g3) n = true.
  Proof.
    intros g3 n p D1 E2.
    apply enabled_12_3 in E2; auto.
    rewrite D1 in E2.
    intuition; auto with bool.
  Qed.

  Lemma disabled_12_3:
    forall g n, Proper pequiv g ->
                enabled_b g n = false <-> enabled_b (envread1 g) n = false
                                          /\
                                          enabled_b (envread2 g) n = false.
  Proof.
    intros g n pg.
    assert (ok := enabled_12_3_alt n pg).
    rewrite ok; clear ok.
    destruct (enabled_b g n);
      destruct (enabled_b (envread1 g) n);
      destruct (enabled_b (envread2 g) n); 
      try congruence; intuition; auto with bool.
  Qed.  
  
  Lemma Step_disabled1_unchanged1:
    forall (g' g: Env3), Step g' g ->
                         forall n, enabled_b (envread1 g) n = false ->
                                   read1 (g' n) == read1 (g n).
  Proof.
    intros g' g step n dis1.
    destruct step as [pg pg' _ r].
    destruct (r n) as [R1|R1]; rewrite R1.
    - apply RUN_31_disabled; auto.
    - reflexivity.
  Qed.

  (** ** Technical Tools about has_moved_b for Algo1, Algo2, Algo3 *)
  Lemma has_moved_12_3_all:
    forall g' g n,
      Step g' g ->
      has_moved_b g' g n = true <->
      enabled_b (envread1 g) n = true /\
      has_moved_b (envread1 g') (envread1 g) n = true
      \/
      enabled_b (envread1 g) n = false /\
      has_moved_b (envread1 g') (envread1 g) n = false /\
      enabled_b (envread2 g) n = true /\
      has_moved_b (envread2 g') (envread2 g) n = true.
  Proof.
    intros g' g n step.
    assert (E := has_moved_b_enabled_b step n).
    destruct step as [pg pg' _ r]; specialize (r n).
    split.
    { intros hm; specialize (E hm); revert hm.
      unfold has_moved_b.
      destruct (State_dec (g' n) (g n)) as [|S]; [congruence|]; intros _.
      destruct r as [r|r]; [| tauto].
      case_eq (enabled_b (envread1 g) n); intros E1.
      - assert (R := RUN_31_enabled _ _ E1).
        destruct (State_dec (envread1 g' n) (envread1 g n)) as [S1|S1];
          [|left; split; auto].
        unfold enabled_b in E1.
        destruct (State_dec (RUN (envread1 g) n) (envread1 g n));
          [congruence|].
        contradict n0; rewrite <- R, <- r; apply S1.
      - rewrite enabled_12_3 in E; auto.
        destruct E as [|E2]; [congruence|rewrite E2].
        assert (R1 := RUN_31_disabled _ _ E1).
        assert (R2 := RUN_32_enabled _ _ E1 E2).
        destruct (State_dec (envread1 g' n) (envread1 g n)) as [SS1|SS1].
        2: contradict SS1; unfold envread1; rewrite <- R1, r; auto.
        destruct (State_dec (envread2 g' n) (envread2 g n)) as [S2|S2];
          [|right; auto].
        unfold enabled_b in E2.
        destruct (State_dec (RUN (envread2 g) n) (envread2 g n));
          [congruence|].
        contradict n0; rewrite <- R2, <- r; apply S2.
    }
    { clear E.
      unfold has_moved_b.
      destruct (State_dec (g' n) (g n)) as [|S]; [|auto].
      intros [[H1 H2]|[_ [_ [_ H4]]]].
      { destruct (State_dec (envread1 g' n) (envread1 g n))
          as [S1|S1]; [congruence|]; clear H2.
        contradict S1; unfold envread1; apply read1_compat; auto.
      }
      { destruct (State_dec (envread2 g' n) (envread2 g n))
          as [S1|S1]; [congruence|]; clear H4.
        contradict S1; unfold envread2; apply read2_compat; auto.
      }
    }
  Qed.

  Lemma has_moved_12_3_true:
    forall g' g n,
      Step g' g ->
      has_moved_b g' g n = true <->
      has_moved_b (envread1 g') (envread1 g) n = true \/
      has_moved_b (envread2 g') (envread2 g) n = true.
  Proof.
    intros g' g n step.
    assert (E := has_moved_b_enabled_b step n).
    destruct step as [pg pg' _ r]; specialize (r n).
    split.
    { intros hm; specialize (E hm); revert hm.
      unfold has_moved_b.
      destruct (State_dec (g' n) (g n)) as [|S]; [congruence|]; intros _.
      destruct (State_dec (envread1 g' n) (envread1 g n)) as [S1|S1];
        [|left; auto].
      destruct (State_dec (envread2 g' n) (envread2 g n)) as [S2|S2];
        [|right; auto].
      destruct r as [r|r]; [| tauto].
      rewrite enabled_12_3 in E; auto.
      case_eq (enabled_b (envread1 g) n); intros E1.
      - assert (R := RUN_31_enabled _ _ E1).
        unfold enabled_b in E1.
        destruct (State_dec (RUN (envread1 g) n) (envread1 g n));
          [congruence|].
        contradict n0; rewrite <- R, <- r; apply S1.
      - rewrite E1 in E; destruct E as [| E]; [congruence|].
        assert (R := RUN_32_enabled _ _ E1 E).
        unfold enabled_b in E.
        destruct (State_dec (RUN (envread2 g) n) (envread2 g n));
          [congruence|].
        contradict n0; rewrite <- R, <- r; apply S2.
    }
    { clear E.
      unfold has_moved_b.
      destruct (State_dec (g' n) (g n)) as [|S]; [|auto].
      intros [H|H].
      { destruct (State_dec (envread1 g' n) (envread1 g n))
          as [S1|S1]; [congruence|]; clear H.
        contradict S1; unfold envread1; apply read1_compat; auto.
      }
      { destruct (State_dec (envread2 g' n) (envread2 g n))
          as [S1|S1]; [congruence|]; clear H.
        contradict S1; unfold envread2; apply read2_compat; auto.
      }
    }
  Qed.

  Lemma has_moved_12_3_true_alt:
    forall g' g n, Step g' g ->
                   has_moved_b g' g n =
                   has_moved_b (envread1 g') (envread1 g) n
                   || has_moved_b (envread2 g') (envread2 g) n.
  Proof.
    intros g' g n step; assert (ok := has_moved_12_3_true n step).
    destruct (has_moved_b g' g n);
      destruct (has_moved_b (envread1 g') (envread1 g) n);
      destruct (has_moved_b (envread2 g') (envread2 g) n); 
      try congruence; intuition; auto with bool.
  Qed.

  Lemma has_moved_12_3_false:
    forall g' g n, Step g' g ->
                   has_moved_b g' g n = false <->
                   has_moved_b (envread1 g') (envread1 g) n = false
                   /\ has_moved_b (envread2 g') (envread2 g) n = false.
  Proof.
    intros g' g n step; assert (ok := has_moved_12_3_true n step).
    destruct (has_moved_b g' g n);
      destruct (has_moved_b (envread1 g') (envread1 g) n);
      destruct (has_moved_b (envread2 g') (envread2 g) n); 
      try congruence; intuition; auto with bool.
  Qed.
  
  Lemma has_moved_13:
    forall g3' g3 n,
      has_moved_b (envread1 g3') (envread1 g3) n = true ->
      has_moved_b g3' g3 n = true.
  Proof.
    intros g3' g3 n.
    unfold has_moved_b.
    destruct (State_dec (envread1 g3' n) (envread1 g3 n)) as [|S1];
      [congruence|]; intros _.
    destruct (State_dec (g3' n) (g3 n)) as [S3|]; [| auto].
    contradict S1.
    unfold envread1; apply read1_compat; auto.
  Qed.

  Lemma has_moved_31:
    forall g3' g3 n,
      Step g3' g3 ->
      enabled_b (envread1 g3) n = true ->
      has_moved_b g3' g3 n = true ->
      has_moved_b (envread1 g3') (envread1 g3) n = true.
  Proof.
    intros g3' g3 n step en hm.
    rewrite has_moved_12_3_all in hm; auto.
    rewrite en in hm.
    destruct hm as [[_ hm]|[hm _]]; [auto|congruence].
  Qed.

  Lemma has_moved_23:
    forall g3' g3 n,
      has_moved_b (envread2 g3') (envread2 g3) n = true ->
      has_moved_b g3' g3 n = true.
  Proof.
    intros g3' g3 n.
    unfold has_moved_b.
    destruct (State_dec (envread2 g3' n) (envread2 g3 n)) as [|S2];
      [congruence|]; intros _.
    destruct (State_dec (g3' n) (g3 n)) as [S3|]; [| auto].
    contradict S2.
    unfold envread2; apply read2_compat; auto.
  Qed.

  Lemma has_moved_32:
    forall g3' g3 n,
      Step g3' g3 -> 
      enabled_b (envread1 g3) n = false ->
      has_moved_b g3' g3 n = true ->
      has_moved_b (envread2 g3') (envread2 g3) n = true.
  Proof.      
    intros g3' g3 n step en hm.
    rewrite has_moved_12_3_all in hm; auto.
    rewrite en in hm.
    destruct hm as [[hm _]|[_ hm]]; [congruence|apply hm].
  Qed.

  (** ** Projection of an execution from Algo3 to Algo1 *)
  CoFixpoint execread1 (e3: Exec3): Exec1 :=
    match e3 with
    | s_one g3 => s_one (envread1 g3)
    | s_cons g3 e3' => s_cons (envread1 g3) (execread1 e3')
    end.

  Local Definition tmp e3 :=
    match e3 with
    | s_one g3 => s_one (envread1 g3)
    | s_cons g3 e3' => s_cons (envread1 g3) (execread1 e3')
    end.
  
  Lemma execread1_unfold:
    forall e3,
      execread1 e3 =
      match e3 with
      | s_one g3 => s_one (envread1 g3)
      | s_cons g3 e3' => s_cons (envread1 g3) (execread1 e3')
      end.
  Proof.
    assert (forall e, execread1 e = tmp e).
    intros e.
    rewrite (@Stream_decomposition_lemma _) with (execread1 e).
    destruct e; simpl; reflexivity.
    intros e; rewrite H at 1; unfold tmp.
    reflexivity.
  Qed.

  Lemma execread1_fold_cons:
    forall g3 e3, execread1 (s_cons g3 e3) =
                  (s_cons (envread1 g3) (execread1 e3)).
  Proof.
    intros g3 e3; rewrite execread1_unfold at 1; reflexivity.
  Qed.

  Lemma execread1_fold_one:
    forall g3, execread1 (s_one g3) = (s_one (envread1 g3)).
  Proof.
    intros g3; rewrite execread1_unfold at 1; reflexivity.
  Qed.

  Instance execread1_compat: Proper fequiv execread1.
  Proof.
    cofix execread1_compat.
    intros e3 e3' E;
      destruct e3 as [g3 |g3 e3];
      destruct e3' as [g3' |g3' e3']; inversion E; subst.
    - rewrite execread1_fold_one.
      rewrite execread1_fold_one.
      left.
      apply envread1_compat; assumption.
    - rewrite execread1_fold_cons.
      rewrite execread1_fold_cons.
      right.
      apply envread1_compat; assumption.
      apply execread1_compat; assumption.
  Qed.

  Lemma s_hd_read1: forall e3,
      envread1 (s_hd e3) = s_hd (execread1 e3).
  Proof.
    intros e3.
    rewrite execread1_unfold.
    destruct e3; simpl; reflexivity.
  Qed.

  (** ** Projection of an execution from Algo3 to Algo2 *)
  CoFixpoint execread2 (e3: Exec3): Exec2 :=
    match e3 with
    | s_one g3 => s_one (envread2 g3)
    | s_cons g3 e3' => s_cons (envread2 g3) (execread2 e3')
    end.

  Local Definition tmpp e3 :=
    match e3 with
    | s_one g3 => s_one (envread2 g3)
    | s_cons g3 e3' => s_cons (envread2 g3) (execread2 e3')
    end.
  
  Lemma execread2_unfold:
    forall e3,
      execread2 e3 =
      match e3 with
      | s_one g3 => s_one (envread2 g3)
      | s_cons g3 e3' => s_cons (envread2 g3) (execread2 e3')
      end.
  Proof.
    assert (forall e, execread2 e = tmpp e).
    intros e.
    rewrite (@Stream_decomposition_lemma _) with (execread2 e).
    destruct e; simpl; reflexivity.
    intros e; rewrite H at 1; unfold tmpp.
    reflexivity.
  Qed.

  Lemma execread2_fold_cons:
    forall g3 e3, execread2 (s_cons g3 e3) =
                  (s_cons (envread2 g3) (execread2 e3)).
  Proof.
    intros g3 e3; rewrite execread2_unfold; reflexivity.
  Qed.

  Lemma execread2_fold_one:
    forall g3, execread2 (s_one g3) = (s_one (envread2 g3)).
  Proof.
    intros g3; rewrite execread2_unfold; reflexivity.
  Qed.

  Instance execread2_compat: Proper fequiv execread2.
  Proof.
    cofix execread2_compat.
    intros e3 e3' E;
      destruct e3 as [g3 |g3 e3];
      destruct e3' as [g3' |g3' e3']; inversion E; subst.
    - rewrite execread2_fold_one.
      rewrite execread2_fold_one.
      left.
      apply envread2_compat; assumption.
    - rewrite execread2_fold_cons.
      rewrite execread2_fold_cons.
      right.
      apply envread2_compat; assumption.
      apply execread2_compat; assumption.
  Qed.

  Lemma s_hd_read2: forall e3,
      envread2 (s_hd e3) = s_hd (execread2 e3).
  Proof.
    intros e3.
    rewrite execread2_unfold.
    destruct e3; simpl; reflexivity.
  Qed.

  (** ** Terminal Configurations: *)

  (** A configuration is terminal for Algo3 iff its
   projections on Algo1 and Algo2 are also terminal.
   *)
  Lemma terminal_31:
    forall (g: Env3), terminal g -> terminal (envread1 g).
  Proof.
    intros g term.
    assert (pg: Proper pequiv g) by apply term.
    apply terminal_all_disabled_reverse; auto.
    apply envread1_compat, pg.
    intros n.
    apply terminal_all_disabled with (n := n) in term.
    unfold enabled in term; rewrite enabled_12_3_alt in term; auto.
    unfold enabled.
    auto with bool.
  Qed.

  Lemma terminal_32:
    forall (g: Env3), terminal g -> terminal (envread2 g).
  Proof.
    intros g term.
    assert (pg: Proper pequiv g) by apply term.
    apply terminal_all_disabled_reverse; auto.
    apply envread2_compat, pg.
    intros n.
    apply terminal_all_disabled with (n := n) in term.
    unfold enabled in term; rewrite enabled_12_3_alt in term; auto.
    unfold enabled.
    auto with bool.
  Qed.

  Lemma terminal_12_3:
    forall g, Proper fequiv g ->
              terminal (envread1 g) /\ terminal (envread2 g) ->
              terminal g.
  Proof.
    intros g pg [T1 T2].
    apply terminal_all_disabled_reverse; auto.
    intros n.
    apply terminal_all_disabled with (n := n) in T1.
    apply terminal_all_disabled with (n := n) in T2.
    unfold enabled; rewrite enabled_12_3_alt; auto.
    unfold enabled in T1, T2.
    intros E; rewrite orb_true_iff in E; intuition; auto with bool.
  Qed.


  (** Once Algo1 has reached a terminal configuration, it remains in
   the same terminal configuration *)
  Lemma terminal1_continued':
    forall g' g,
      Step g' g -> terminal (envread1 g) -> envread1 g =~= envread1 g'.
  Proof.
    intros g' g step term n1 n2 en.
    assert (pg: Proper pequiv g) by apply step.
    unfold envread1; rewrite (pg _ _ en); clear n1 en.
    symmetry; apply Step_disabled1_unchanged1; auto.        
    eapply terminal_all_disabled in term.
    unfold enabled in term; rewrite not_true_iff_false in term;
      apply term.
  Qed.

  Lemma terminal1_continued:
    forall g' g,
      Step g' g ->
      terminal (envread1 g) -> terminal (envread1 g').
  Proof.
    intros g' g step term.
    eapply terminal_compat.
    2: apply term.
    symmetry; apply terminal1_continued'.
    apply step.
    apply term.
  Qed.

  Lemma terminal1_continued'_bis_:
    forall e3 g1,
      is_exec e3 -> terminal (envread1 (s_hd e3)) ->
      g1 =~= envread1 (s_hd e3) ->
      Always (fun e => s_hd e =~= g1) (execread1 e3).
  Proof.
    cofix terminal1_continued'_bis_.
    destruct e3 as [g3 | g3 e3]; intros g1 ise term eg;
      rewrite execread1_unfold.
    - apply always_one; simpl.
      rewrite eg; simpl.
      apply envread1_compat.
      change (fequiv (s_hd (s_one g3)) (s_hd (s_one g3))).
      apply (is_exec_s_hd_compat (Algo:=Algo3)),ise.
    - apply always_cons; simpl.
      rewrite eg; simpl.
      apply envread1_compat.
      change (fequiv (s_hd (s_cons g3 e3)) (s_hd (s_cons g3 e3))).
      apply (is_exec_s_hd_compat (Algo:=Algo3)),ise.
      apply terminal1_continued'_bis_.
      apply is_exec_inv_2 with g3, ise.
      apply terminal1_continued with g3.
      apply is_exec_inv_1, ise.
      apply term.
      rewrite eg; simpl.
      apply terminal1_continued'.
      apply is_exec_inv_1, ise.
      apply term.
  Qed.

  Lemma terminal1_continued'_bis:
    forall e3,
      is_exec e3 -> terminal (envread1 (s_hd e3)) ->
      Always (fun e => s_hd e =~= envread1 (s_hd e3)) (execread1 e3).
  Proof.
    intros e3 ise term.
    eapply Always_impl_P.
    2: apply terminal1_continued'_bis_; try assumption.
    2: apply term.
    simpl.
    auto.
  Qed.

  Lemma terminal1_continued_bis:
    forall e3,
      is_exec e3 -> terminal (envread1 (s_hd e3)) ->
      Always (fun e => terminal (envread1 (s_hd e))) e3.
  Proof.
    cofix terminal1_continued_bis.
    destruct e3 as [g3 | g3 e3]; intros ise term.
    - apply always_one; assumption.
    - apply always_cons.
      * assumption.
      * apply terminal1_continued_bis.
        apply is_exec_inv_2 with g3, ise.
        apply terminal1_continued with g3.
        apply is_exec_inv_1, ise.
        apply term.
  Qed.

  Lemma terminal1_continued_bis'':
    forall e3,
      is_exec e3 ->
      Eventually (fun e => terminal (envread1 (s_hd e))) e3 ->
      Eventually (Always (fun e => terminal (envread1 (s_hd e)))) e3.
  Proof.
    intros e3 ise E.
    induction E.
    - apply eventually_now.
      apply terminal1_continued_bis; assumption.
    - apply eventually_later.
      apply IHE.
      apply is_exec_inv_2 with a, ise.
  Qed.
  
  (** ** Technical Tools about EN, AN, NEUT for Algo1, Algo2, Algo3 *)
  
  Lemma EN_1_3: forall e3 n,
      Proper fequiv (s_hd e3) ->
      EN n (execread1 e3) ->
      EN n e3.
  Proof.
    intros e3 n pg.
    unfold EN, EN_b.
    rewrite <- s_hd_read1.
    apply enabled_13; assumption.
  Qed.

  Lemma EN_2_3: forall e3 n,
      Proper fequiv (s_hd e3) ->
      EN n (execread2 e3) ->
      EN n e3.
  Proof.
    intros e3 n pg.
    unfold EN, EN_b.
    rewrite <- s_hd_read2.
    apply enabled_23; assumption.
  Qed.

  Lemma ACT_1_3: forall g3 e3 n,
      Step (s_hd e3) g3 ->
      ACT n (execread1 (s_cons g3 e3)) ->
      ACT n (s_cons g3 e3).
  Proof.
    intros g3 e3 n step.
    rewrite execread1_unfold.
    unfold ACT, ACT_b.
    rewrite <- s_hd_read1.
    intros hm1; eapply has_moved_13; assumption.
  Qed.

  Lemma ACT_3_1:
    forall g e n,
      Step (s_hd e) g ->
      EN n (execread1 (s_cons g e)) ->
      ACT n (s_cons g e) ->
      ACT n (execread1 (s_cons g e)).
  Proof.
    intros g e n step.
    unfold EN, EN_b, ACT, ACT_b.
    rewrite <- s_hd_read1.
    intros en h.
    rewrite execread1_unfold.
    rewrite <- s_hd_read1.
    apply has_moved_31; auto.
  Qed.

  Lemma NEUT_3_1':
    forall g e n,
      Step (s_hd e) g ->
      EN n  (execread1 (s_cons g e)) ->
      NEUT n (s_cons g e) ->
      NEUT n (execread1 (s_cons g e)).
  Proof.
    intros g e n step.
    unfold EN, EN_b, NEUT, NEUT_b.
    rewrite <- s_hd_read1.
    intros en.
    rewrite execread1_unfold.
    intros nt.
    rewrite andb_true_iff in *.
    destruct nt as [en3 nt]; rewrite andb_true_iff in nt.
    destruct nt as [nhm nen].
    simpl in en.
    rewrite <- s_hd_read1.
    split; [| rewrite andb_true_iff; split].
    assumption.

    assert (X := has_moved_13).
    specialize (X (s_hd e) g n).
    case_eq (has_moved_b (envread1 (s_hd e)) (envread1 g) n);
      intros F.
    apply X in F.
    rewrite F in nhm; assumption.
    auto.

    assert (X := enabled_13 n (Step_Proper_l step)).
    case_eq (enabled_b (envread1 (s_hd e)) n); intros F.
    apply X in F.
    rewrite F in nen; assumption.
    auto.
  Qed.

  Lemma NEUT_3_1:
    forall e n,
      is_exec e ->
      EN n (execread1 e) ->
      NEUT n e ->
      NEUT n (execread1 e).
  Proof.
    intros [g | g e]; rewrite execread1_unfold; simpl; intros n ise en neut.
    - unfold NEUT, NEUT_b in neut; simpl in neut;
        auto with bool.
    - apply NEUT_3_1'.
      inversion ise;try tauto; assumption.
      rewrite execread1_unfold; assumption.
      assumption.
  Qed.

  Lemma EN_AN_1_3:
    forall e n,
      is_exec e ->
      EN n (execread1 e) -> AN n e ->
      AN n (execread1 e).
  Proof.
    intros e n ise.
    unfold AN.
    intros en1 an3.
    destruct e as [g | g e].
    - unfold ACT, NEUT in *; simpl in *; tauto.
    - destruct an3 as [a3|n3].
      left; apply ACT_3_1; auto.
      apply is_exec_inv_1; auto.
      right; apply NEUT_3_1; assumption.
  Qed.

  Lemma E_AN_3_1:
    forall e3 n,
      is_exec e3 ->
      EN n (execread1 e3) ->
      Eventually (AN n) e3 ->
      Eventually (AN n) (execread1 e3).
  Proof.
    intros e3 n ise en E.
    induction E as [e | g e E IH].

    apply eventually_now.
    apply EN_AN_1_3; assumption.

    rewrite execread1_unfold.
    case_eq (EN_b n (execread1 e)); intros enn.
    apply eventually_later.
    apply IH.
    apply is_exec_inv_2 with g; auto.
    apply enn.
    apply eventually_now.
    apply en_progress.
    apply en.
    unfold EN; rewrite enn; auto.
  Qed.

  Lemma ACT_3_2:
    forall e n, is_exec e ->
                enabled_b (envread1 (s_hd e)) n = false ->
                ACT n e ->
                ACT n (execread2 e).
  Proof.
    intros e n p D.
    unfold ACT, ACT_b.
    destruct e as [g | g e]; [intros F; contradict F; apply diff_false_true|].
    intros h.
    rewrite execread2_unfold.
    rewrite <- s_hd_read2.
    apply has_moved_32; auto.
    apply is_exec_inv_1; auto.
  Qed.

  Lemma NEUT_3_2:
    forall e n, is_exec e -> terminal (envread1 (s_hd e)) ->
                NEUT n e ->
                NEUT n (execread2 e).
  Proof.
    intros e n ise term.
    unfold NEUT, NEUT_b.
    destruct e as [g | g e]; [intros F; contradict F; apply diff_false_true|].
    rewrite execread2_unfold.
    intros nt.
    rewrite andb_true_iff in *.
    destruct nt as [en3 nt]; rewrite andb_true_iff in nt.
    destruct nt as [nhm nen].
    rewrite <- s_hd_read2.
    assert (p: Proper fequiv (s_hd e)).
    { apply (is_exec_s_hd_compat (Algo := Algo3)).
      apply is_exec_inv_2 in ise; auto.
    }
    split; [| rewrite andb_true_iff; split].
    { rewrite negb_true_iff in nen, nhm.
      rewrite enabled_12_3 in en3.
      eapply (terminal_all_disabled) in term.
      unfold enabled in term.
      rewrite not_true_iff_false in term.
      rewrite term in en3.
      destruct en3; [congruence|auto].
      apply is_exec_proper in ise.
      ssplit ise; apply ise.
    }
    { rewrite negb_true_iff in *.        
      rewrite has_moved_12_3_false in nhm.
      apply nhm.
      apply is_exec_inv_1 in ise; auto.
    }
    { rewrite negb_true_iff in *.        
      rewrite disabled_12_3 in nen; auto.
      apply nen.
    }
  Qed.

  Lemma term_EN_AN_3_2:
    forall e n, is_exec e ->
                terminal (envread1 (s_hd e)) ->
                EN n (execread2 e) ->
                AN n e ->
                AN n (execread2 e).
  Proof.
    intros e n ise term en an3.
    unfold AN in *.
    destruct e as [g | g e].
    rewrite execread2_unfold.
    unfold ACT, NEUT in *; simpl in *; assumption.
    destruct an3 as [a3|nt3].
    left; apply ACT_3_2; auto.
    eapply terminal_all_disabled in term.
    unfold enabled in term; rewrite not_true_iff_false in term.
    apply term.
    right; apply NEUT_3_2; assumption.
  Qed.
  
  Lemma E_AN_3_2:
    forall e3 n,
      is_exec e3 ->
      terminal (envread1 (s_hd e3)) ->
      EN n (execread2 e3) ->
      Eventually (AN n) e3 ->
      Eventually (AN n) (execread2 e3).
  Proof.
    intros e3 n ise term en E.
    induction E as [e | g e E IH].

    apply eventually_now.
    apply term_EN_AN_3_2; assumption.

    rewrite execread2_unfold.
    case_eq (EN_b n (execread2 e)); intros enn.
    apply eventually_later.
    apply IH.
    apply is_exec_inv_2 with g, ise.
    rewrite s_hd_read1 in term; simpl in term.
    apply terminal1_continued with g.
    apply is_exec_inv_1, ise.
    assumption.
    assumption.

    apply eventually_now.
    apply en_progress.
    apply en.
    unfold EN; rewrite enn.
    apply diff_false_true.
  Qed.

  (** ** Once Algo1 has reached a terminal configuration, the
     projection of an execution on Algo2 is actually an execution of
     Algo2.  *)
  Lemma Step_32:
    forall g' g,        
      Step g' g ->
      (forall n, enabled_b (envread1 g) n = false \/
                 has_moved_b g' g n = false) ->
      Step (envread2 g') (envread2 g).
  Proof.
    intros g' g step ht.
    dup step step'.
    destruct step as [pg pg' _ r].
    split.
    { apply envread2_compat, pg. }
    { apply envread2_compat, pg'. }
    { destruct (Step_one_has_moved_true step') as [n hm].
      assert (E3 := has_moved_b_enabled_b step' _ hm).
      intros F; specialize (F _ _ (Equivalence_Reflexive n));
        contradict F.
      rewrite <- has_moved_progress.
      apply has_moved_32; auto.
      destruct (ht n); [auto|congruence].
    }
    { 
      intros n; specialize (r n); destruct r as [r|r].
      2: right; unfold envread2; apply read2_compat; auto.
      destruct (ht n).
      2: {
        rewrite has_moved_12_3_false in H; auto.
        destruct H as [_ H].
        apply has_moved_false in H.
        right; auto.
      }
      case_eq (enabled_b (envread2 g) n); intros E2.
      + apply RUN_32_enabled in E2; auto.
        unfold envread2; rewrite r; left; apply E2.
      + apply RUN_32_disabled in E2; auto.
        unfold envread2; rewrite r; right; apply E2.
    }
  Qed.

  Lemma is_exec_3_2:
    forall e3, is_exec e3 -> terminal (envread1 (s_hd e3)) ->
               is_exec (execread2 e3).
  Proof.
    cofix is_exec_3_2.
    destruct e3 as [g3 | g3 e3]; rewrite execread2_unfold;
      rewrite s_hd_read1; simpl;
        intros ise term.
    { (* s_one *)
      left.
      apply terminal_32.
      apply is_exec_inv_0, ise.
    }
    { (* s_cons *)
      right.
      * rewrite <- s_hd_read2.
        apply Step_32; auto.
        apply is_exec_inv_1,ise.
        intros n; left; apply terminal_all_disabled with _ n in term; auto.
        unfold enabled in term; rewrite not_true_iff_false in term;
          auto.
      * apply is_exec_3_2.
        apply is_exec_inv_2 with g3,ise.
        apply terminal1_continued with g3;auto.
        apply is_exec_inv_1,ise.
    }
  Qed.

  Lemma Step_31:
    forall g' g, Step g' g ->
                 forall n, enabled_b (envread1 g) n = true ->
                           has_moved_b g' g n = true ->
                           Step (envread1 g') (envread1 g).
  Proof.
    intros g' g step n en hm.
    dup step step'.
    destruct step as [pg pg' _ r].
    split.
    { apply envread1_compat, pg. }
    { apply envread1_compat, pg'. }
    {        
      intros F; specialize (F _ _ (Equivalence_Reflexive n));
        contradict F.
      rewrite <- has_moved_progress.
      apply has_moved_31; auto.
    }
    { clear step' en hm.
      intros nn; specialize (r nn); destruct r as [r|r].
      2: right; unfold envread1; apply read1_compat; auto.
      case_eq (enabled_b (envread1 g) nn); intros E1.
      + apply RUN_31_enabled in E1; auto.
        unfold envread1; rewrite r; left; apply E1.
      + apply RUN_31_disabled in E1; auto.
        unfold envread1; rewrite r; right; apply E1.
    }
  Qed.
    
  (* ****************************************************** *)
  (** ** Checking the Specication of Algo3
   *)
  Section Spec_OK.

    (** Algo1 is silent, hence its spec can be reduced to a 
     predicate SPEC1 on Env1  *)
    Variable (SPEC1: Env1 -> Prop)
             (P_correct1: P_correctness Assumption1 SPEC1).

    (** Algo2 is self-stabilizing for its specification SPEC2,
     under the daemon LS2, with legitimate states LS2 *)
    Variable (SPEC2: Exec2 -> Prop)
             (LS2: Env2 -> Prop)
             (DF2: Exec2 -> Prop)
             (spec_ok2: spec_ok Assumption2 DF2 LS2 SPEC2).

    (** Assumption between 1 and 2: once Algo1 has stabilized
      into SPEC1, Assume2 should hold. *)
    Variable (SPEC1_Assume2:
                forall (g: Env3),
                  Proper fequiv g -> Assume3 g -> SPEC1 (envread1 g) ->
                  Assume2 (envread2 g)).

    (** Algo3: an execution meets the specification for Algo3, SPEC3,
     if its projection on Algo2 meets SPEC2 and if it verifies SPEC1
     on each of its configurations. *)
    Definition Compo_SPEC SPEC_env1 SPEC2 :=
      fun e => Always (fun ex => SPEC_env1 (envread1 (s_hd ex))) e /\
               SPEC2 (execread2 e).
    Notation SPEC3 := (Compo_SPEC SPEC1 SPEC2).

    Notation LS3 := (fun g3 => terminal (envread1 g3) /\ LS2 (envread2 g3)).
    
    Theorem composition_spec_ok:
      forall DF3: Exec3 -> Prop,
        (forall e, is_exec e -> DF3 e ->
                   terminal (s_hd (execread1 e))->
                   DF2 (execread2 e)) ->
        spec_ok Assumption3 DF3 LS3 SPEC3.
    Proof.
      intros DF3 hDF3.
      intros e ass ise dfe [hls2 ht1].
      cut (SPEC1 (envread1 (s_hd e))).
      { intros sp1; split.
        - eapply Always_impl_P.
          2: { eapply Always_conj.
               split.
               apply terminal1_continued_bis; auto.
               eapply RO_is_exec_Always.
               apply Assume1_RO_compat.
               apply ise.
               apply ass.
          }
          intros ee htee.
          apply P_correct1; auto; apply htee.
        - apply spec_ok2.
          rewrite <- s_hd_read2.
          apply SPEC1_Assume2; auto.
          eapply (is_exec_s_hd_compat (Algo := Algo3)); auto.
          apply is_exec_3_2; auto. 
          apply hDF3; auto.
          rewrite <- s_hd_read1; auto.
          rewrite <- s_hd_read2; auto.
      }
      apply P_correct1; auto.
    Qed.
    
  End Spec_OK.

  (* ****************************************************** *)
  (** ** Checking the Partial Correctness of Algo3 in case Algo2 is
   silent *)
  Section Partial_Correctness.

    (** Algo1 is silent, hence its spec can be reduced to a predicate
     SPEC1 on Env1 *)
    Variable (SPEC1: Env1 -> Prop)
             (P_correct1: P_correctness Assumption1 SPEC1).

    (** Assumption between 1 and 2: once Algo1 has stabilized into
      SPEC1, Assume2 should hold. *)
    Variable (SPEC1_Assume2:
                forall (g: Env3),
                  Proper fequiv g -> Assume3 g -> SPEC1 (envread1 g) ->
                  Assume2 (envread2 g)).

    (** Algo2 is silent: *)
    Variable (SPEC2: Env2 -> Prop)
             (P_correct2: P_correctness Assumption2 SPEC2).

    (** Algo3 *)
    Notation SPEC3 := (fun g => SPEC1 (envread1 g) /\
                                SPEC2 (envread2 g)).
    
    Theorem composition_P_correctness:
      P_correctness Assumption3 SPEC3.
    Proof.
      intros g ass term.
      assert (pg: Proper fequiv g) by apply term.
      apply Assume_31 in ass; auto.
      assert (P1 := P_correct1 ass (terminal_31 term)).
      split; auto.
      apply P_correct2.
      apply SPEC1_Assume2; auto.
      apply terminal_32; auto.
    Qed.

  End Partial_Correctness.

  (* ****************************************************** *)
  (** ** Checking the Closure of Algo3
   *)
  Section Closure.

    (** Legitimate states for Algo1: terminal states *)
    (** Legitimate states for Algo2: *)
    Variable (LS2: Env2 -> Prop) (LS2_compat: Proper pequiv LS2).

    Notation LS3 := (fun g3 => terminal (envread1 g3) /\ LS2 (envread2 g3)).

    (** Assumption between 1 and 2: once Algo1 has stabilized Assume2
      should hold. *)
    Variable (Term1_Assume2:
                forall (g: Env3),
                  Proper fequiv g -> Assume3 g -> terminal (envread1 g) ->
                  Assume2 (envread2 g)).

    (** Daemon fairness *)
    Variable (DF2: Exec2 -> Prop) (DF3: Exec3 -> Prop)
             (hDF3: forall e, is_exec e -> DF3 e ->
                              terminal (s_hd (execread1 e))->
                              DF2 (execread2 e)).

    Theorem composition_closure:
      closure Assumption2 DF2 LS2 -> closure Assumption3 DF3 LS3.
    Proof.
      intros cl2.
      intros e ass ise dfe [term ls].
      apply Always_conj.
      split.
      { apply terminal1_continued_bis; auto. }
      { cut (Always (fun s => LS2 (s_hd s)) (execread2 e)).
        { clear; revert e; cofix h.
          intros [g | g e] A.
          - left.
            rewrite s_hd_read2.
            apply Always_P_P in A; auto.
          - right.
            rewrite s_hd_read2.
            apply Always_P_P in A; auto.
            rewrite execread2_fold_cons in A.
            apply h; apply Always_inv_2 in A; auto.
        }
        apply cl2.
        - rewrite <- s_hd_read2. 
          apply Term1_Assume2; auto.
          apply hd_compat; apply Exec.is_exec_proper in ise; auto.
        - apply is_exec_3_2; auto.
        - apply hDF3; auto.
          rewrite <- s_hd_read1; auto. 
        - rewrite <- s_hd_read2; auto.
      }
    Qed.
    
  End Closure.
  
  (* ****************************************************** *)
  (** ** Termination: *)

  (** Termination of Composition requires additionnal hypothesis.

    Discussion from J. Lefrere internship report:

    A good argument for additional hypotheses is to prevent the
    asynchronous daemon from postponing the execution of A1 steps
    indefinitely. This may be possible if that at some point of the
    program, there are some rules in A2 that enable divergent
    behaviors.

- (A) One way of preventing this from happening, is to assume that
    from any configuration of the first algorithm, the second one will
    stabilize in a finite time. This will imply that, in a finite
    time, any daemon avoiding to execute A1 steps will run out of
    options and thus will be forced to make A1 progress. Such
    hypothesis is sufficient to ensure the stabilization of an
    algorithm.

- (B) That is not the only solution, one other is to change the
    execution model by adding a fairness assumption to the
    scheduler. For example, a weakly fair scheduler will have to
    eventually activate any rule that stays enabled for a sufficiently
    long time. This is what will make A1 progress in this solution.

Theses two solutions are not comparable, there may be cases where will
be applicable and not the other.

   *)

  (** *** Termination: Case A *)
  Section A.

    
    Variable (Hterm1: (termination (Algo := Algo1) Assumption1)).
    Variable (Hterm2: forall (g: Env2),
                 Proper fequiv g ->
                 Acc (Step (Algo := Algo2)) g).
    
    Lemma Hterm2_alt: well_founded (Step (Algo := Algo2)).
    Proof.
      intros g.
      apply Acc_intro.
      intros y step.
      apply Hterm2.
      apply (Step_Proper_l step).
    Qed.

    (** ****

     We show first termination for the lexicographic order:
     this order is a relation on pairs (Env1, Env2) built by:

        - lp1: Step (Algo1) g1' g1 -> lexp (g1,_) (g1',_)
        - lp2: eqE1 g1' g1 -> Step (Algo2) g2' g2
                           -> lexp (g1,g2) (g1',g2')

     ****

     We show Acc for this order (see Lemma term_lexico_pairs_Acc):
         - the first order (on Algo1) is Acc
           (this is directly obtained from the assumption Hterm1)
         - the second order (on Algo2) is well_founded
           (this is obtained from Lemma Hterm2_alt)

         - other properties about composition and transitivity:

              - forall g' g2 g1: Env1, Step g' g1 -> eqE1 g1 g2 -> Step g' g2
                (compatibility of Step)

              - transitive eqE1

     *)

    Lemma term_lexico_pairs:
      forall g, Assume3 g -> Proper fequiv g ->
                Acc  (lexico_pairs fequiv (Step (Algo := Algo1))
                                   (Step (Algo := Algo2)))
                     (envread1 g, envread2 g).
    Proof.
      intros g ass pg.
      apply lexico_pairs_Acc.
      { apply pequiv_prf.       (* transitive fequiv *) }
      { intros a b u.           (* inclusion *)
        destruct u as [z [ v t ]].
        eapply Step_compat.
        3: apply v.
        assert (pa: Proper fequiv z).
        apply v.
        apply (Step_Proper_l v).
        symmetry; apply t.
      }
      apply Hterm2_alt.
      apply Hterm1.
      apply Assume_31; auto.
      apply envread1_compat, pg.
    Qed.

    (** ****

     We then need to show that Step is included into this order
     in some sense:
       - inclusion Step (lexp @@ F)

       - we fix F := fun g3 => (envread1 g3, envread2 g3)

     See Lemma incl_lexp.

     We introduce a Boolean predicate [step1_b]
     that allows to decide wether a step from Algo3 is to take
     into account for Algo1 or for Algo2 (remember that within
     a step, some nodes may have performed a step from Algo1
     while others from Algo2).

       - A step from Algo3 has to be considered for convergence
         in Algo1, as far as some configuration of Algo1
         evolves, hence as far as at least a node has actually
         executed an action of Algo1. This is exactly what
         [step1_b] expresses: it is true when, during a step of
          Algo3, there exists a node which is enabled for some
         action of Algo1 and actually moves.

       - Any other step of Algo3 (ie when step1_b is false, ie
         no node executes an action of Algo1) is considered in
         the convergence of Algo2.

     The proof of Lemma [incl_lexp] is divided on two cases. For a
     step of Algo3 between g and g', we split the proof into the case when
     (step1_b g' g) is true and the case when it is false.

          - In case (step1_b g' g) is true,
            Lemma Step1_is_Step_Algo1 proves that for a step of Algo3
            between g and g' (we note d3 the corresponding Diff),
            there is a step of Algo1 between (envread1 g) and (envread1 g').

          - In case (step1_b g' g) is false, we have to prove that

             - eqE1 (envread1 g') (envread1 g)
               which is shown by Lema Step_disabled1_unchanged1, and that

             - there is a step of Algo2 between (envread2 g) and (envread2 g').
               This is shown by Lemma Step2_is_Step_Algo2.

     *)

    Definition step1_b (g' g: Env3): bool :=
      existsb
        (fun (n: Node) =>
           andb (enabled_b (envread1 g) n)
                (has_moved_b g' g n))
        all_nodes.

    Lemma step1_b_true:
      forall g' g, Proper fequiv g -> Proper fequiv g' ->
                   (step1_b g' g = true <->
                    exists n, enabled_b (envread1 g) n = true /\
                              has_moved_b g' g n = true).
    Proof.
      intros g' g pg pg'.
      unfold step1_b.
      rewrite existsb_exists.
      split; intros s1.
      destruct s1 as [x [_ M]].
      exists x.
      rewrite andb_true_iff in M.
      split; apply M.
      destruct s1 as [x M].
      specialize (all_nodes_prop x); intros inx.
      rewrite InA_alt in inx.
      destruct inx as [x' [ex inx']].
      exists x'.
      split.
      apply inx'.
      rewrite andb_true_iff.
      rewrite <- (enabled_b_compat (envread1_compat pg) _ _ ex).
      rewrite <- (has_moved_b_compat pg' pg _ _ ex).
      assumption.
    Qed.

    Lemma step1_b_false:
      forall g' g, Proper fequiv g -> Proper fequiv g' ->
                   (step1_b g' g = false <->
                    forall n, enabled_b (envread1 g) n = false \/
                              has_moved_b g' g n = false).
    Proof.
      intros g' g pg pg'.
      unfold step1_b.
      erewrite existsb_forallb.
      rewrite negb_false_iff.
      rewrite forallb_forall.
      split; intros s2 x.
      specialize (all_nodes_prop x); intros inx.
      rewrite InA_alt in inx.
      destruct inx as [x' [ex inx']].
      specialize (s2 _ inx').
      rewrite <- negb_true_iff.
      rewrite or_comm.
      rewrite <- negb_true_iff.
      rewrite or_comm.
      rewrite <- orb_true_iff.
      rewrite (enabled_b_compat (envread1_compat pg) _ _ ex).
      rewrite (has_moved_b_compat pg' pg _ _ ex).
      apply s2.
      intros _; simpl; specialize (s2 x).
      rewrite orb_true_iff.
      rewrite negb_true_iff.
      rewrite or_comm.
      rewrite negb_true_iff.
      rewrite or_comm.
      apply s2.

      intros a; simpl.
      rewrite negb_orb.
      rewrite negb_involutive.
      rewrite negb_involutive.
      reflexivity.
    Qed.

    Lemma Step1_is_Step_Algo1:
      forall g' g, step1_b g' g = true -> Step g' g ->
                   Step (envread1 g') (envread1 g).
    Proof.
      intros g' g s1 step.
      split.
      { apply envread1_compat, step. }
      { apply envread1_compat, step. }
      { rewrite step1_b_true in s1; try apply step.
        destruct s1 as [n [en1 hm1]].
        rewrite has_moved_12_3_all in hm1; auto.
        rewrite en1 in hm1; destruct hm1 as [[_ hm1]|[hm1 _]];
          [|congruence].
        unfold has_moved_b in hm1.
        destruct (State_dec (envread1 g' n) (envread1 g n));
          [congruence|auto].
      }
      { intros n.
        destruct step as [pg pg' _ r].
        destruct (r n) as [rn|rn]; clear r.
        2: right; unfold envread1; apply read1_compat; auto.
        case_eq (enabled_b (envread1 g) n); intros E1.
        - left; apply RUN_31_enabled in E1; auto.
          rewrite <- rn in E1; apply E1.
        - right; apply RUN_31_disabled in E1; auto.
          rewrite <- rn in E1; apply E1.
      }
    Qed.

    Lemma Step2_is_Step_Algo2:
      forall g' g, step1_b g' g = false -> Step g' g ->
                   Step (envread2 g') (envread2 g).
    Proof.
      intros g' g s2 step.
      rewrite step1_b_false in s2; auto; try apply step.
      dup step step'.
      destruct step as [pg pg' _ r].
      split.
      { apply envread2_compat, pg. }
      { apply envread2_compat, pg'. }
      { destruct (Step_one_has_moved_true step') as [n hm].
        assert (E3 := has_moved_b_enabled_b step' _ hm); clear step'.
        destruct (s2 n) as [D1|]; [|congruence]; clear s2.
        intros F; specialize (F _ _ (Equivalence_Reflexive n));
          contradict F.
        unfold has_moved_b in hm.
        destruct (State_dec (g' n) (g n)) as [|S]; [congruence|]; clear hm.
        destruct (r n) as [rn|]; [| tauto]; clear r.
        apply enabled_12_3 in E3; auto.
        rewrite D1 in E3; destruct E3 as [| E3]; [congruence|].
        unfold envread2; rewrite rn.
        rewrite RUN_32_enabled; auto.
        unfold enabled_b in E3.
        destruct (State_dec (RUN (envread2 g) n) (envread2 g n));
          [congruence|auto].
      }
      { clear step'.
        intros n; specialize (r n); destruct r as [r|r].
        2: right; unfold envread2; apply read2_compat; auto.
        specialize (s2 n); destruct s2 as [s2|s2].
        - case_eq (enabled_b (envread2 g) n); intros E2.
          + apply RUN_32_enabled in E2; auto.
            unfold envread2; rewrite r; left; apply E2.
          + apply RUN_32_disabled in E2; auto.
            unfold envread2; rewrite r; right; apply E2.
        - unfold has_moved_b in s2.
          destruct (State_dec (g' n) (g n)); [| congruence]; clear s2.
          right; unfold envread2; apply read2_compat; auto.
      }
    Qed.

    Lemma incl_lexico_pairs:
      inclusion (Step (Algo := Algo3))
                (RelCompFun
                   (lexico_pairs fequiv (Step (Algo := Algo1))
                                 (Step (Algo := Algo2)))
                   (fun (g: Env3) => (envread1 g, envread2 g))).
    Proof.
      intros g' g step.
      unfold "@@".
      case_eq (step1_b g' g); intros step1.
      (* Step Algo1 *)
      left.
      apply Step1_is_Step_Algo1; assumption.
      right.

      (* eqE1 *)
      intros n n' en; unfold envread1.
      assert (pg: Proper fequiv g).
      apply step.
      rewrite <- (read1_compat _ _ (pg  _ _ en)).
      clear n' en.
      rewrite (step1_b_false pg (Step_Proper_l step)) in step1.
      specialize (step1 n).
      destruct step1 as [E | N].
      apply Step_disabled1_unchanged1; assumption.
      apply read1_compat.
      apply (has_moved_false g' g n); auto. 
      (* Step Algo2 *)
      apply Step2_is_Step_Algo2; assumption.
    Qed.

    Theorem termination_A: termination Assumption3.
    Proof.
      intros g ass pg.
      eapply Acc_incl.
      apply incl_lexico_pairs.
      apply (Inverse_Image.Acc_inverse_image
               _ _ (lexico_pairs fequiv (Step (Algo := Algo1))
                                 (Step (Algo := Algo2)))
               (fun (g: Env3) => (envread1 g, envread2 g))).
      apply term_lexico_pairs.
      apply ass.
      apply pg.
    Qed.

  End A.
 
  (* ****************************************************** *)
  (* ****************************************************** *)


  (** *** Termination: Case B *)
  Section B.

    (** Algo1 is silent under weakly fair daemon *)
    Variable (Hterm1: silence Assumption1 weakly_fair).

    (** Algo2 converges to some legitimate states LS2, under weakly
    faire daemon *)
    Variable (LS2: Env2 -> Prop)
             (LS2_compat: Proper pequiv LS2)
             (Conv2: convergence Assumption2 weakly_fair LS2).

    (** Assumption between 1 and 2: once Algo1 has stabilized, Assume2
    should hold. *)
    Variable (assume_32:
                forall g3, Proper fequiv g3 -> Assume3 g3 ->
                           terminal (envread1 g3) ->
                           Assume2 (envread2 g3)).

    (** Legitimate States for Algo3 *)
    Notation LS3 := (fun g3 => terminal (envread1 g3) /\ LS2 (envread2 g3)).

    (** *** Weakly Fair from 3 to 1 *)
    Lemma WF_exec_3_1:
      forall (e3: Exec3),
        is_exec e3 ->
        weakly_fair e3 ->
        weakly_fair (execread1 e3).
    Proof.
      unfold weakly_fair.
      intros e3 ise A n.
      specialize (A n).
      revert A ise. revert e3.

      cofix WF_exec_3_1.

      destruct e3 as [g3 | g3 e3]; intros A ise;
        rewrite execread1_unfold in *.

      { (* s_one *)
        assert (X: finite_exec (s_one (envread1 g3))).
        left.
        tauto.
        apply always_one.
        eapply WF_finite_exec.
        left.
        apply terminal_31.
        apply is_exec_inv_0; auto.
        assumption.
      }
      { (* s_cons *)
        inversion A; subst; clear A.
        apply always_cons.
        clear WF_exec_3_1.
        intros en.
        assert (X:= @E_AN_3_1 (s_cons g3 e3) n).
        rewrite execread1_unfold in X; apply X.
        auto.
        apply en.
        apply H1.
        apply EN_1_3.
        apply (is_exec_s_hd_compat (Algo:=Algo3)),ise.
        apply en.
        apply WF_exec_3_1; [| apply is_exec_inv_2 with g3]; assumption.
      }
    Qed.

    (** *** Weakly Fair from 3 to 2 *)
    Lemma WF_exec_3_2:
      forall (e3: Exec3),
        is_exec e3 ->
        terminal (envread1 (s_hd e3)) ->
        weakly_fair e3 ->
        weakly_fair (execread2 e3).
    Proof.
      unfold weakly_fair.
      intros e3 ise term A n.
      specialize (A n).
      revert A ise term. revert e3.

      cofix WF_exec_3_2.

      destruct e3 as [g3 | g3 e3]; intros A ise term;
        rewrite execread2_unfold in *.

      { (* s_one *)
        assert (X: finite_exec (s_one (envread2 g3))).
        left; tauto.
        apply always_one.
        eapply WF_finite_exec.
        left.
        apply terminal_32.
        apply is_exec_inv_0, ise.
        assumption.
      }
      { (* s_cons *)
        inversion A; subst; clear A.
        simpl in term.
        apply always_cons.
        clear WF_exec_3_2.
        intros en.
        assert (X:= E_AN_3_2).
        specialize (X _ n ise term).
        rewrite execread2_unfold in X; apply X; auto.
        apply H1.
        unfold EN, EN_b.
        apply enabled_23.
        apply (is_exec_s_hd_compat (Algo:=Algo3)),ise.
        unfold EN, EN_b; apply en.
        
        apply WF_exec_3_2.
        assumption.
        apply is_exec_inv_2 with g3, ise.
        apply terminal1_continued with g3.
        apply is_exec_inv_1, ise.
        apply term.
      }
    Qed.
    
    Section With_Squeeze.

      Lemma move1_en_an_neq:
        forall g3 e3 n,
          Step (s_hd e3) g3 ->
          enabled_b (envread1 g3) n = true ->
          AN n (s_cons g3 e3) ->
          ~ (envread1 g3) =~= (envread1 (s_hd e3)).
      Proof.
        intros g3 e3 n step en1 an.
        destruct an as [en | neut].
        - intros F;
            specialize (F _ _ (Equivalence_Reflexive n));
            rewrite s_hd_read1 in F; contradict F.
          apply ACT_mov.
          rewrite <- execread1_fold_cons.
          apply ACT_3_1; simpl; auto.
        - rewrite s_hd_read1.
          eapply NEUT_mov.
          rewrite <- execread1_fold_cons.
          apply NEUT_3_1'; auto.
          unfold EN, EN_b.
          rewrite <- s_hd_read1; simpl.
          apply en1.
          apply neut.
      Qed.

      Lemma move1_term_en_neq:
        forall g3' g3 n,
          Step g3' g3 ->
          terminal g3' ->
          enabled_b (envread1 g3) n = true ->
          ~(envread1 g3') =~= (envread1 g3).
      Proof.
        intros g3' g3 n step term en1 F.
        apply terminal_31 in term.
        rewrite (terminal_compat F) in term.
        eapply terminal_all_disabled with (n:=n) in term .
        unfold enabled in term; rewrite en1 in term; congruence.
      Qed.

      Lemma execread1_all_compat:
        forall e3, is_exec e3 -> Proper pequiv (execread1 e3).
      Proof.
        cofix execread1_all_compat.
        destruct e3 as [g3 | g3 e3]; intros ise;
          rewrite execread1_unfold; subst.
        - left; apply envread1_compat.
          change (fequiv (s_hd (s_one g3)) (s_hd (s_one g3))).
          apply (is_exec_s_hd_compat (Algo:=Algo3)),ise.
        - right.
          apply envread1_compat.
          change (fequiv (s_hd (s_cons g3 e3)) (s_hd (s_cons g3 e3))).
          apply (is_exec_s_hd_compat (Algo:=Algo3)),ise.
          apply execread1_all_compat,is_exec_inv_2 with g3,ise.
      Qed.

      Lemma non_terminal1_Skippable:
        forall g3 e3, is_exec (s_cons g3 e3) ->
                      weakly_fair  (s_cons g3 e3) ->
                      ~ terminal (envread1 g3) ->
                      Skippable (envread1 g3)
                                (s_cons (envread1 g3) (execread1 e3)).
      Proof.
        intros g3 e3 ise wf nt.
        assert (pg1: Proper fequiv (envread1 g3)).
        { apply envread1_compat.
          change (fequiv (s_hd (s_cons g3 e3)) (s_hd (s_cons g3 e3))).
          apply (is_exec_s_hd_compat (Algo:=Algo3)),ise.
        }
        destruct (terminal_dec_bis pg1) as [[n en]|]; [| tauto].
        specialize (wf n).
        apply Always_P_P in wf.
        assert (tmp: EN  n (execread1 (s_cons g3 e3))).
        unfold EN, EN_b; assumption.
        assert (X := EN_1_3 _ (is_exec_s_hd_compat ise) tmp).
        specialize (wf X).

        clear X tmp pg1 nt.

        right. (* on peut pas faire l'induction sinon... *)
        remember (envread1 g3) as g1.
        assert (eg1: g1 =~= (envread1 g3)).
        {
          rewrite Heqg1.
          apply envread1_compat.
          change (fequiv (s_hd (s_cons g3 e3)) (s_hd (s_cons g3 e3))).
          apply (is_exec_s_hd_compat (Algo:=Algo3)),ise.
        }
        clear Heqg1.
        remember (s_cons g3 e3) as ge3.
        revert g3 e3 eg1 Heqge3;
          induction wf as [ge3 E | gg3 ge3 E IH];
          intros g3 e3 eg1 Heqge3.
        {
          rewrite Heqge3 in E, ise.
          unfold enabled in en;
            rewrite (enabled_b_compat eg1 _ _ (Equivalence_Reflexive n)) in en.
          assert (st := is_exec_inv_1 ise).
          assert (isee := is_exec_inv_2 ise).
          assert (ne := (move1_en_an_neq st en E)).
          rewrite s_hd_read1 in ne.
          assert (tmp: Proper pequiv
                              (execread1 e3)).
          apply execread1_all_compat.
          assumption.
          assert (tmpp: Proper fequiv (envread1 g3)).
          apply envread1_compat; apply st.
          destruct (Skippable_if_move' _ tmpp tmp ne) as [nok|ok].
          { contradict ne.
            inversion nok; subst.
            apply Always_P_P in H2.
            symmetry; apply H2.
          }
          {
            eapply Acc_per_morphism';[..| apply ok];eauto.
            apply Rskip_compat,eg1.
            intros x y sk px.
            unfold Rskip in sk.
            specialize (sk px).
            destruct sk as [sk | [sk1 sk2] ].
            inversion sk; subst.
            proper_.
            proper_.
            intros x y sk px.
            unfold Rskip in sk.
            specialize (sk px).
            destruct sk as [sk | [sk1 sk2] ].
            inversion sk; subst.
            proper_.
            proper_.
            right.
            apply eg1.
            apply execread1_all_compat; assumption.
          }
        } {
          assert (pg1: Proper fequiv g1).
          { unfold Proper; rewrite eg1; apply envread1_compat.
            rewrite Heqge3 in ise.
            change (fequiv (s_hd (s_cons g3 e3)) (s_hd (s_cons g3 e3))).
            apply (is_exec_s_hd_compat (Algo:=Algo3)),ise.
          }
          destruct e3 as [fe3 | fe3 ne3].
          - assert (isee := is_exec_inv_2 ise).
            assert (st := is_exec_inv_1 ise).
            inversion Heqge3; subst.
            assert (term := is_exec_inv_0 isee).
            unfold enabled in en;
              rewrite (enabled_b_compat eg1 _ _ (Equivalence_Reflexive n))
              in en.
            assert (ok := move1_term_en_neq _ st term en).
            assert (X := @Skippable_if_move' _ _ _ pg1 (execread1 (s_one fe3))).
            assert (pp: Proper pequiv (execread1 (s_one fe3))).
            rewrite execread1_unfold; left; apply envread1_compat; apply term.
            specialize (X pp); clear pp.
            assert (ppp: ~ g1 =~= (s_hd (execread1 (s_one fe3)))).
            rewrite execread1_unfold; simpl; intros F; contradict ok;
              symmetry. rewrite <- eg1; auto.
            specialize (X ppp); clear ppp.
            destruct X as [nok|X].
            * inversion nok.
              apply Always_P_P in H2.
              rewrite execread1_unfold in H2; simpl in H2;
                rewrite eg1 in H2; contradiction.
            * apply X.

          - case (Env_dec g1 (envread1 fe3)).

            * intros e.
              specialize (e pg1).
              assert (pp: Proper fequiv (envread1 fe3)).
              { rewrite Heqge3 in ise.
                apply envread1_compat.
                apply is_exec_inv_2 in ise.
                apply is_exec_inv_1 in ise.
                apply ise.
              }
              specialize (e pp).
              inversion Heqge3.
              assert (isee := is_exec_inv_2 ise).
              specialize (IH isee fe3 ne3 e H1).
              apply Acc_intro.
              intros e1 sk.
              unfold Rskip in sk.
              assert (p: Proper pequiv
                                (s_cons g1 (execread1 (s_cons fe3 ne3)))).
              unfold Proper; right.
              rewrite e at 2; assumption.
              apply execread1_all_compat; subst; auto.
              specialize (sk p).
              destruct sk as [sk | [sk1 sk2]].
            + eapply Acc_per_morphism';[..|apply IH].
              apply Rskip_compat.
              apply pg1.
              intros x y skk px.
              unfold Rskip in skk.
              specialize (skk px).
              destruct skk as [skk|[skk1 skk2]].
              inversion skk; subst.
              unfold Proper; rewrite H5 at 2; assumption.
              unfold Proper; rewrite skk1 at 2; assumption.
              intros x y skk px.
              unfold Rskip in skk.
              specialize (skk px).
              destruct skk as [skk|[skk1 skk2]].
              inversion skk; subst.
              proper_.
              proper_.
              inversion sk; subst.
              rewrite H6.
              rewrite execread1_unfold at 1.
              right.
              symmetry; assumption.
              apply execread1_all_compat.
              apply is_exec_inv_2 with fe3,isee.
            + inversion sk2.
              * assert (pe3: Proper pequiv (execread1 (s_cons fe3 ne3))).
                { apply execread1_all_compat.
                  rewrite Heqge3 in ise.
                  apply is_exec_inv_2 with g3,ise.
                }
                assert (ok := @Skippable_if_move'
                                _ _ _ pg1 (execread1 (s_cons fe3 ne3)) pe3).
                rewrite <- s_hd_read1 in ok; simpl in ok.
                intros ne; specialize (ok ne).
                destruct ok as [nok|ok].
            + inversion nok.
              apply Always_P_P in H2; simpl in H2.
              symmetry in H2; contradiction.
            + assumption.
        }
      Qed.

      Lemma exec_can_collapse:
        forall e3, is_exec e3 -> weakly_fair e3 ->
                   Squeezable (execread1 e3).
      Proof.
        cofix exec_can_collapse.
        destruct e3 as [g3 | g3 e3]; intros ise wf;
          rewrite execread1_unfold.
        {
          left; unfold Skippable; simpl.
          left.
          apply always_one; simpl.
          apply envread1_compat.
          inversion ise; subst.
          apply H0.
        }{
          right; simpl.
          - case (terminal_dec (envread1 g3)).
            * intros pt; left.
              apply always_cons.
              simpl.
              apply envread1_compat.
              change (fequiv (s_hd (s_cons g3 e3)) (s_hd (s_cons g3 e3))).
              apply (is_exec_s_hd_compat (Algo:=Algo3)),ise.
              eapply Always_impl_P.
              2: apply terminal1_continued'_bis.
              simpl.
              intros ee Hee.
              rewrite Hee.
              symmetry; apply terminal1_continued'.
              apply is_exec_inv_1,ise.
              apply pt.
              apply envread1_compat.
              change (fequiv (s_hd (s_cons g3 e3)) (s_hd (s_cons g3 e3))).
              apply (is_exec_s_hd_compat (Algo:=Algo3)),ise.
              apply is_exec_inv_2 with g3,ise.
              apply terminal1_continued with g3.
              apply is_exec_inv_1,ise.
              apply pt.
              apply envread1_compat.
              change (fequiv (s_hd (s_cons g3 e3)) (s_hd (s_cons g3 e3))).
              apply (is_exec_s_hd_compat (Algo:=Algo3)),ise.
            * intros nt.
              apply non_terminal1_Skippable; assumption.
          - apply exec_can_collapse.
            apply is_exec_inv_2 with g3,ise.
            intros n; specialize (wf n); inversion wf; assumption.
        }
      Qed.

      
      Lemma is_constant_terminal:
        forall e3 g1, is_exec e3 -> weakly_fair e3 ->
                      is_constant g1 (execread1 e3) -> terminal g1.
      Proof.
        cut (forall e3 g1, is_exec e3 -> weakly_fair e3 ->
                           ~ terminal g1 ->
                           ~ is_constant g1 (execread1 e3)).
        { intros ok e3 g1 ise wf i.
          specialize (ok _ g1 ise wf).
          destruct (terminal_dec g1).
          - apply t.
            apply is_constant_Proper_l with (execread1 e3).
            apply i.
          - specialize (ok n).
            contradiction.
        }{
          intros e3 g1 ise wf nt i.

          assert (pf: Proper fequiv g1).
          apply is_constant_Proper_l with (execread1 e3).
          apply i.

          assert (e: (envread1 (s_hd e3)) =~= g1).
          rewrite s_hd_read1.
          apply is_constant_hd; apply i.

          destruct (terminal_dec_bis pf) as [[n en]|t]; [|tauto].
          specialize (wf n).
          apply Always_P_P in wf.
          unfold EN, EN_b in wf.
          unfold enabled in en;
            rewrite <- (enabled_b_compat e _ _ (Equivalence_Reflexive n)) in en.
          specialize (wf (enabled_13 _ (is_exec_s_hd_compat ise) en)).

          clear nt pf.
          induction wf.
          - destruct s as [g | g s].
            + apply AN_one in H; tauto.
            + assert (x: (envread1 g) =~= (envread1 (s_hd s))).
              rewrite execread1_fold_cons in i.
              apply is_constant_inv in i.
              simpl in e; rewrite e.
              rewrite s_hd_read1; symmetry; apply is_constant_hd; apply i.
              contradict x.
              eapply move1_en_an_neq with n.
              apply is_exec_inv_1,ise.
              simpl in en; assumption.
              assumption.
          - rewrite execread1_fold_cons in i.
            apply is_constant_inv in i.
            assert (x: (envread1 (s_hd s)) =~= (envread1 a)).
            rewrite s_hd_read1.
            simpl in e; rewrite e.
            apply is_constant_hd; apply i.

            case_eq (enabled_b (envread1 (s_hd s)) n);
              intros ens.
            + apply IHwf.
              apply is_exec_inv_2 with a,ise.
              eapply is_constant_compat;[..|apply i].
              rewrite <- e at 1; apply e.
              apply execread1_all_compat.
              apply is_exec_inv_2 with a,ise.
              simpl in e; rewrite x; apply e.
              apply ens.
            + simpl in en.
              rewrite (enabled_b_compat
                         x _ _ (Equivalence_Reflexive n)) in ens.
              rewrite en in ens.
              congruence.
        }
      Qed.

      Lemma Step_31_neq:
        forall g3' g3,
          Step g3' g3 ->
          ~ (envread1 g3') =~= (envread1 g3) ->
          Step (envread1 g3') (envread1 g3).
      Proof.
        intros g3' g3 step neq.
        apply Step1_is_Step_Algo1.
        2: apply step.
        rewrite <- not_false_iff_true.
        intros F.
        contradict neq.
        assert (pg3: Proper fequiv g3).
        apply step.
        rewrite step1_b_false in F.
        2: apply pg3.
        2: apply (Step_Proper_l step).
        intros n n' en.
        rewrite <- (envread1_compat pg3 _ _ en).
        clear n' en.
        specialize (F n).
        destruct F as [F|F].
        - unfold envread1.
          apply Step_disabled1_unchanged1; assumption.
        - unfold envread1.
          apply read1_compat.
          rewrite has_moved_false in F; auto.
      Qed.

      Lemma suffix_13:
        forall e1 e3, Proper pequiv e3 -> is_suffix (execread1 e3) e1 ->
                      exists e3', is_suffix e3 e3' /\ (execread1 e3') =~= e1.
      Proof.
        intros e1 e3 p3 suf.
        remember (execread1 e3) as pi.
        assert (Hpi: pi =~= (execread1 e3)).
        rewrite Heqpi; apply execread1_compat; apply p3.
        clear Heqpi.
        revert e3 p3 Hpi; induction suf as [pi H| xi pi H IH]; intros e3 p3 Hpi.
        {
          exists e3.
          split.
          left; apply p3.
          rewrite <- Hpi; apply H.
        }
        {
          destruct e3 as [g3 | g3 e3]; rewrite execread1_unfold in Hpi;
            inversion Hpi; subst.
          rename p3 into pg3.
          assert (p3: Proper pequiv e3).
          inversion pg3; assumption.
          destruct (IH e3 p3 H5) as [e3' [suf E]].
          exists e3'.
          split.
          - right; apply suf.
          - apply E.
        }
      Qed.
      
      Lemma is_exec_simu':
        forall (e3: Exec3) (e1: Exec1),
          is_exec e3 ->
          Always (fun e =>
                    forall g1, is_constant g1 (execread1 e)
                               -> terminal g1) e3 ->
          Simulation e1 (execread1 e3) ->
          Always moves e1 -> is_exec e1.
      Proof.
        cofix is_exec_simu'.
        intros [g3 | g3 e3] [g1 | g1 e1] ise ICT sim A.
        { (* (s_one g3) (s_one g1) *)
          rewrite execread1_fold_one in sim.
          inversion sim; subst.
          left.
          apply (terminal_compat H1).
          apply terminal_31.
          inversion ise; assumption.
        }
        { (* (s_one g3) (s_cons g1 e1) *)
          rewrite execread1_fold_one in sim.
          inversion sim.
        }
        { (* (s_cons g3 e3) (s_one g1) *)
          left.
          rewrite execread1_fold_cons in sim.
          inversion sim; subst.
          apply (terminal_compat H2).
          apply Always_inv_2 in ICT; apply Always_P_P in ICT.
          apply ICT.
          revert H3; apply is_constant_compat.
          symmetry; auto.
          apply execread1_compat; apply is_exec_inv_2 in ise;
            apply is_exec_proper in ise; auto.
        }
        { (* (s_cons g3 e3) (s_cons g1 e1) *)
          rewrite execread1_fold_cons in sim.
          inversion sim; subst.
          rename s2 into e1'; clear sim.
          assert (pg1: Proper fequiv (envread1 g3)).
          unfold Proper; rewrite <- H3 at 1; apply H3.
          assert (pe13: Proper pequiv (execread1 e3)).
          apply execread1_all_compat.
          apply is_exec_inv_2 with g3,ise.

          assert (x: is_cons_star (envread1 g3) e1' (execread1 e3)).
          eapply is_cons_star_compat;[..| apply H5].
          symmetry; assumption.
          eapply Simulation_Proper_r; eauto.
          apply pe13.

          assert (pe1' := Simulation_Proper_r H4).
          assert (SUF := @cons_star_suffix_bis
                           _ _ (envread1 g3) (execread1 e3)
                           e1' pg1 pe13 
                           (is_cons_star_suffix H5) x).
          rewrite <- execread1_fold_cons in SUF.
          apply suffix_13 in SUF.
          2: apply (is_exec_proper (Algo:=Algo3)); assumption.
          destruct SUF as [e3'' [SUF E]].
          destruct e3'' as [g3' | g3' e3'];
            rewrite execread1_unfold in E; inversion E; subst.
          rewrite <- H3 in H2.

          assert (EX := suffix_is_exec ise SUF).
          right.

          - assert (X: (envread1 (s_hd e3')) =~= (s_hd e1)).
            rewrite s_hd_read1.
            rewrite (Simulation_hd H4).
            apply hd_compat; assumption.
            rewrite <- (Step_compat X H2).
            apply Step_31_neq.
            apply is_exec_inv_1,EX.
            intros F.
            apply Always_P_P in A; unfold moves in A;apply A.
            rewrite <- X,F,H2;proper_.

          - apply is_exec_simu' with e3'.
            apply is_exec_inv_2 with g3',EX.
            apply Always_inv_2 with g3'.
            eapply Always_suffix; auto.
            { intros EE EEE EEEE; split.
              - intros H G IC.
                apply H.
                assert (PG: G =~= G).
                { rewrite <- (is_constant_hd IC).
                  apply hd_compat, execread1_compat.
                  rewrite <- EEEE at 1; auto.
                }
                revert IC; apply is_constant_compat; auto.
                apply execread1_compat; auto.
              - intros H G IC.
                apply H.
                assert (PG: G =~= G).
                { rewrite <- (is_constant_hd IC).
                  apply hd_compat, execread1_compat.
                  rewrite EEEE at 1; symmetry; auto.
                }
                revert IC; apply is_constant_compat; auto.
                apply execread1_compat; symmetry; auto.
            }
            apply ICT.
            apply SUF. 
            eapply Simulation_compat.
            3: apply H4.
            apply Simulation_Proper_l in H4; auto.
            assumption.
            inversion A; assumption.
        }
      Qed.

      (* ****************************** *)
      (* ****************************** *)        
      Lemma is_exec_simu:
        forall (e3: Exec3) (e1: Exec1),
          is_exec e3 -> weakly_fair e3 ->
          Simulation e1 (execread1 e3) -> Always moves  e1 ->
          is_exec e1.
      Proof.
        intros e3 e1 ise3 wf3 sim m.
        apply is_exec_alt.
        split.
        { revert m.
          apply Always_MP.
          apply Stepwise_Always_Simulation_monotonic with (execread1 e3); auto.
          { split; [|split].
            tauto.
            intros a x h F; symmetry in F; contradict F; apply Step_progress;
              auto.
            intros a b x y ab xy; apply Step_compat; auto.
          }
          clear sim wf3 e1.
          revert ise3; revert e3; cofix h.
          intros [g3 | g3 e3] ise; rewrite execread1_unfold.
          { left; tauto. }
          { right.
            intros m.
            apply is_exec_inv_1 in ise.
            rewrite <- s_hd_read1.
            apply Step_31_neq; auto.
            intros F; unfold moves in m; contradict m.
            rewrite <- s_hd_read1; symmetry; auto.
            apply h.
            apply is_exec_inv_2 in ise; auto.
          }
        }
        { clear m.
          revert sim wf3 ise3; revert e1 e3.
          cofix h.
          intros [g1 | g1 e1] [g3 | g3 e3] sim wf ise;
            rewrite execread1_unfold in sim.
          { left.
            apply Simulation_hd in sim; simpl in sim; rewrite sim.
            apply terminal_31.
            apply is_exec_inv_0; auto.
          }
          { inversion_clear sim.
            left.
            apply is_constant_terminal with e3; auto.
            apply is_exec_inv_2 in ise; auto.
            apply weakly_fair_inv in wf; auto.
          }
          { inversion sim. }
          { apply Simulation_inv_cons_star in sim.
            destruct sim as [s3 [h3 c3]].
            right; auto.
            apply is_cons_star_suffix in c3.
            apply suffix_13 in c3.
            2: { apply is_exec_proper in ise; inversion ise; auto. }
            destruct c3 as [suf3 [hsuf hex3]].
            apply h with suf3; auto.
            { assert (dum: Proper pequiv e1)
                by (apply Simulation_Proper_l with s3; auto).
              revert h3; apply Simulation_compat; auto. }
            apply weakly_fair_inv in wf.
            apply suffix_weakly_fair with e3; auto.
            apply is_exec_inv_2 in ise.
            apply suffix_is_exec with e3; auto.
          }            
        }
      Qed.
      
      (* ****************************** *)
      (* ****************************** *)

      Lemma collapse_is_exec:
        forall e3 (ise: is_exec e3) (wf: weakly_fair e3),
          is_exec (@Squeeze
                     _ _ Env_dec
                     _ (execread1_all_compat ise)
                     (exec_can_collapse ise wf)).
      Proof.
        intros e3 ise wf.
        eapply is_exec_simu.
        apply ise.
        apply wf.
        apply Simulation_Squeeze.
        apply Squeeze_no_repeat.
      Qed.

      Lemma weakly_fair_monotonic:
        Simulation_monotonic (A := Env1) weakly_fair.
      Proof.
        intros X Y sim wfY n.
        apply Always_Simulation_monotonic with Y; auto.
        { intros x y xy.
          apply impl_compat; rewrite xy; tauto. }
        apply Simulation_Impl.
        { apply EN_compat; auto. }
        { apply Eventually_compat, AN_compat; auto. }
        apply Instant_Simulation_comonotonic.
        { unfold EN, EN_b; intros a b h.
          rewrite h; tauto. }
        apply Eventually_Simulation_monotonic.
        apply Stepwise_Simulation_monotonic.
        { split; [|split].
          - intuition;apply AN_one.
          - intros a sa [h|h].
            + apply ACT_mov in h; auto.
            + apply NEUT_mov in h; auto.
          - intros a b sa sb ab hda.
            assert (d1: has_moved_b (s_hd sa) a n =
                        has_moved_b (s_hd sb) b n)
              by (apply has_moved_b_compat; auto). 
            assert (d2: enabled_b (s_hd sa) n =
                        enabled_b (s_hd sb) n)
              by (apply enabled_b_compat; auto). 
            assert (d3: enabled_b a n =
                        enabled_b b n)
              by (apply enabled_b_compat; auto). 
            unfold AN, ACT, ACT_b, NEUT, NEUT_b; rewrite d1, d2, d3.
            tauto.
        }
      Qed.
      
      Lemma weakly_fair_simu:
        forall (e3: Exec3) (e1: Exec1),
          is_exec e3 -> weakly_fair e3 ->
          Simulation e1 (execread1 e3) ->
          weakly_fair e1.
      Proof.
        intros e3 e1 ise wf3 sim.
        apply weakly_fair_monotonic with (execread1 e3); auto.
        apply WF_exec_3_1; auto.
      Qed.

      Lemma collapse_weakly_fair:
        forall e3 (ise: is_exec e3) (wf: weakly_fair e3),
          weakly_fair (@Squeeze
                         _ _ Env_dec _
                         (execread1_all_compat ise)
                         (exec_can_collapse ise wf)).
      Proof.
        intros e3 ise wf.
        eapply weakly_fair_simu.
        apply ise.
        apply wf.
        apply Simulation_Squeeze.
      Qed.

    End With_Squeeze.

    Lemma E_term1:
      forall e3,
        Assume3 (s_hd e3) -> is_exec e3 -> weakly_fair e3 ->
        Eventually (fun e => terminal (envread1 (s_hd e))) e3.
    Proof.
      intros e3 ass ise wf.
      assert (WF := collapse_weakly_fair ise wf).
      assert (ISE := collapse_is_exec ise wf).
      remember (@Squeeze _ _ Env_dec _
                         (execread1_all_compat ise)
                         (exec_can_collapse ise wf)) as Ce3.

      assert (Fin: finite_exec Ce3).
      { apply Hterm1; auto.
        rewrite HeqCe3, Squeeze_hd.
        rewrite <- s_hd_read1; apply Assume_31; auto.
        apply hd_compat, (is_exec_proper ise).
      }
      assert (X := finite_exec_E_terminal ISE Fin).
      assert (Y := @Simulation_Squeeze _ _ Env_dec (execread1 e3)
                                       (execread1_all_compat ise)
                                       (exec_can_collapse ise wf)).
      rewrite <- HeqCe3 in Y.

      cut (Eventually  (fun e => terminal (s_hd e)) (execread1 e3)).
      { intros E.
        clear - E.
        remember (execread1 e3) as e1.
        revert Heqe1. revert e3.
        induction E.
        - intros e3 He.
          apply eventually_now.
          rewrite s_hd_read1, <- He; assumption.
        - intros [g3 | g3 e3] He;
            rewrite execread1_unfold in He; inversion He; clear He.
          apply eventually_later.
          apply IHE; assumption.
      }
      {
        eapply Eventually_Simulation_comonotonic;eauto.
        intros a b ab; apply terminal_compat, hd_compat, ab.
        apply Instant_Simulation_comonotonic.
        intros a b h; rewrite h; tauto.
      }
    Qed.

    Lemma assume_suffix:
      forall e, Assume3 (s_hd e) ->  is_exec e ->
                forall e_suf, is_suffix e e_suf -> Assume3 (s_hd e_suf).
    Proof.
      intros e ass ise e_suf suf.
      induction suf.
      - revert ass; apply Assume_compat.
        symmetry; apply hd_compat; auto.
      - apply IHsuf.
        revert ass; unfold Assume3.
        apply Assume_RO_compat.
        unfold ROEnv_part; apply Stable_Variable_step.
        apply is_exec_inv_1; simpl; auto.
        apply is_exec_inv_2 with a,ise.
    Qed.

    Theorem composition_convergence: convergence Assumption3 weakly_fair LS3.
    Proof.
      intros e3 ass ise wf.
      apply Eventually_conj.
      apply (is_exec_proper ise).
      apply E_term1; auto.
      { intros s' hs' ts'.
        apply terminal1_continued_bis; auto.
        apply suffix_is_exec with e3; auto. }
      { intros s' hs' ts'.
        cut (Eventually (fun suf : Exec2 => LS2 (s_hd suf)) (execread2 s')).
        { remember (execread2 s') as e2.
          assert (ps': Proper pequiv s').
          { apply is_suffix_Proper with e3; auto. }
          clear hs' ts'.
          intros h; revert ps' Heqe2; revert s'.
          induction h as [s hs | g s hs ih].
          { intros s' ps' hs'; left.
            rewrite s_hd_read2, <- hs'; auto. }
          { intros [g' | g' s'] ps' hs'; rewrite execread2_unfold in hs';
              inversion hs'; subst.
            right.
            apply ih.
            inversion ps'; auto.
            inversion hs'; auto. }
        }
        apply Conv2.                  
        rewrite <- s_hd_read2.
        { apply assume_32; auto.
          apply hd_compat, is_suffix_Proper with e3; auto.
          apply assume_suffix with e3; auto. }
        { apply is_exec_3_2; auto.
          apply suffix_is_exec with e3; auto. }
        { apply WF_exec_3_2; auto.
          apply suffix_is_exec with e3; auto.
          apply suffix_weakly_fair with e3; auto. }
      }
    Qed.         

  End B.

  Section Results.
   
    Theorem Compo_B_Self_Stab (SPEC_env1: Env1 -> Prop) SPEC2:
      (forall g, Proper fequiv g -> SPEC_env1 (envread1 g) ->
                 Assume2 (envread2 g)) ->
      silence Assumption1 weakly_fair ->
      self_stabilization Assumption1 weakly_fair (only_one SPEC_env1)  ->
      self_stabilization Assumption2 weakly_fair SPEC2 ->
      self_stabilization Assumption3 weakly_fair (Compo_SPEC SPEC_env1 SPEC2).
    Proof.
      intros hassume sil1 [LS1 ss1] [LS2 ss2].
      assert (hassume': forall g3,
                 Proper fequiv g3 -> Assume3 g3 -> terminal (envread1 g3) ->
                 Assume2 (envread2 g3)).
      { intros g pg ass tg; apply hassume; auto.
        apply (silent_terminal_SPEC
                 (Daemon_fairness := weakly_fair)); auto.
        intros gg asgg tgg;
          apply weakly_fair_finite_exec; auto; left; auto.
        split with LS1; auto.
      }        
      split with (fun g3 => terminal (envread1 g3) /\ LS2 (envread2 g3)).
      split; [|split].
      { apply composition_closure with weakly_fair; auto.        
        2: apply ss2.
        cofix h.
        intros [g | g e] ise wf term n; dup wf wf'; specialize (wf n).
        - rewrite execread2_unfold.
          left.        
          apply Always_P_P in wf.
          intros en.
          rewrite <- execread2_fold_one.
          apply E_AN_3_2; auto.
          apply wf.
          rewrite <- execread2_fold_one in en.
          apply EN_2_3 in en; auto.
          apply hd_compat; apply is_exec_proper in ise; auto.
        - rewrite execread2_fold_cons. 
          right.
          { 
            intros en.
            rewrite <- execread2_fold_cons. 
            apply E_AN_3_2; auto.
            apply Always_P_P in wf; apply wf.
            rewrite <- execread2_fold_cons in en. 
            apply EN_2_3; auto.
            apply hd_compat; apply is_exec_proper in ise; auto.
          }
          { apply h; auto.
            apply is_exec_inv_2 in ise; auto.
            apply weakly_fair_inv in wf'; auto.
            rewrite <- s_hd_read1.
            apply terminal1_continued with g.
            apply is_exec_inv_1 in ise; auto.
            rewrite <- s_hd_read1 in term; auto.
          }
      }
      { apply composition_convergence; auto.
        apply ss2. }
      { apply composition_spec_ok with weakly_fair; auto.
        apply (silent_terminal_SPEC
                 (Daemon_fairness := weakly_fair)); auto.
        intros gg asgg tgg;
          apply weakly_fair_finite_exec; auto; left; auto.
        split with LS1; auto.
        apply ss2. 
        intros e ie wf te; apply WF_exec_3_2; auto.
        rewrite s_hd_read1; auto.
      }
    Qed.

    Theorem Compo_Silent_Silent_B_Silent:
      (forall g, Proper fequiv g ->
                 Assume3 g -> terminal (envread1 g) ->
                 Assume2 (envread2 g)) ->
      silence Assumption1 weakly_fair ->
      silence Assumption2 weakly_fair ->
      silence Assumption3 weakly_fair.
    Proof.
      intros hassume sil1 sil2.
      apply convergence_silence in sil2.
      apply convergence_silence.
      intros e a i w.
      assert (C := composition_convergence sil1 sil2 hassume a i w).
      clear a w.
      induction C as [s h| g s h ih].
      - left; apply terminal_12_3; auto.
        apply hd_compat, (is_exec_proper i).
      - right; apply ih, is_exec_inv_2 with g, i.
    Qed.
    
  End Results.
  

End Composition.


Close Scope signature_scope.
Unset Implicit Arguments.
