(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- June 2017
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Bool.
From Coq Require Import Lia.
From Coq Require Import RelationPairs.

(** **** Local Imports *)
From Padec Require Import WellfoundedUtil.
From Padec Require Import SetoidUtil.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Stream.
From Padec Require Import Simulation.
From Padec Require Import Stream_Length.
From Padec Require Import Exec.
From Padec Require Import ListUtils.
From Padec Require Import BoolSet.
From Padec Require Import Self_Stabilization.
From Padec Require Import Composition.
From Padec Require Import Fairness.
From Padec Require Import Steps.
From Padec Require Import NatUtils.

(* end show *)
Open Scope nat_scope.
Open Scope signature_scope.

Section Tools_For_Composition_Steps_0.
  
  (* dans Stream *)
  Lemma Always_True: forall A (s: @Stream A), Always (fun _ => True) s.
  Proof.
    intros A.
    cofix h.
    intros [a | a s]; [left|right]; try tauto.
    apply h.
  Qed.

  Lemma length_P_equiv:
    forall A (P Q: @Stream A -> Prop) s n,
      WUntil (fun s => P s <-> Q s) (fun s => P s /\ Q s) s ->
      (* pas oblige d'avoir Always ... TODO *)
      length_until_P P s n  <-> length_until_P Q s n.
  Proof.
    intros until n.
    intros hu; split; intros hl.
    { induction hl as [s hl |a s n hl nP ih].
      - left; inversion hu; subst; tauto.
      - right; inversion hu; subst; try tauto.
    }
    { induction hl as [s hl |a s n hl nP ih].
      - left; inversion hu; subst; tauto.
      - right; inversion hu; subst; try tauto.
    }
  Qed.
  
(****************************************************************)  
  (* dans Stream *)
  Lemma Always_conj:
    forall A (P Q: @Stream A -> Prop) s, 
      Always P s -> Always Q s -> Always (fun s => P s /\ Q s) s.
  Proof.
    intros A P Q; cofix h.
    intros [a | a s] hp hq; inversion hp; inversion hq;
      [left|right]; auto.
  Qed.    

  Lemma Forever_P_P:
    forall (A: Type) P (s: @Stream A), Forever P s -> P s.
  Proof. intros A P [a | a s] f; inversion f; auto. Qed.

  Lemma Forever_forall:
    forall (A: Type) P, (forall (s: @Stream A), P s) -> forall s, Forever P s.
  Proof.
    intros A P h.
    cofix hh.
    intros [a | a s]; [left|right].
    apply h.
    apply h.
    apply hh.
  Qed.
  
End Tools_For_Composition_Steps_0.

Section Tools_For_Composition_Steps.

  Context {Chans: Channels} {Algo: Algorithm}
          {Net: Network}.

  
  (** *** Has_Moved1 e3
    true iff the first two configurations of e3 exist and
             are linked by a step of Algo1 *)
 
  Definition Has_Moved1 (e: Exec): bool :=
    (existsb (fun n => ACT_b n e) all_nodes). 

  Instance Has_Moved1_compat: Proper fequiv Has_Moved1.
  Proof.
    intros e1 e2 ee; unfold Has_Moved1.
    apply existsb_compat with equiv.
    2: reflexivity.
    intros n1 n2 en; apply ACT_b_compat; auto.
  Qed.

  Definition Term := (fun e: Exec => terminal (s_hd e)).

  Lemma Has_Moved1_not_terminal:
    forall e1, is_exec e1 ->
               Has_Moved1 e1 = true -> Term e1 -> False.
  Proof.
    intros e ise HM T.
    unfold Has_Moved1 in HM; rewrite existsb_exists_InA in HM.
    2: intros n1 n2 enn; apply ACT_b_compat;
      [apply enn| apply is_exec_proper in ise; auto].
    destruct HM as [n [_ act]].
    apply ACT_b_EN_b in act; auto.
    unfold EN_b in act.
    contradict act; rewrite Bool.not_true_iff_false.
    apply terminal_all_disabled with (n := n) in T;
      unfold enabled in T; rewrite not_true_iff_false in T; auto.
  Qed.

  Lemma Has_Moved1_progress:
    forall e1, Proper pequiv e1 -> Has_Moved1 e1 = true ->
               exists g1 e1', e1 =~= s_cons g1 e1' /\ ~ (s_hd e1') =~= g1. 
  Proof.
    intros e1 pe1 hm.
    unfold Has_Moved1 in hm; rewrite existsb_exists_InA in hm.
    2: intros n1 n2 enn; apply ACT_b_compat; [apply enn| apply pe1].
    destruct hm as [n [_ act]].
    unfold ACT_b in act.
    destruct e1 as [g1 | g1 e1]; [contradict act; auto with bool|].
    exists g1; exists e1; split; auto.
    apply has_moved_progress in act; auto.
  Qed.
  
  Lemma Has_Moved1_progress':
    forall g1 e1, Proper pequiv (s_cons g1 e1) -> Has_Moved1 (s_cons g1 e1) = true ->
                  ~ s_hd e1 =~= g1. 
  Proof.
    intros g1 e1 p hm; destruct (Has_Moved1_progress _ _ hm) as [gg1 [ee1 [h1 h2]]].
    contradict h2.
    ssplit h1; destruct h1 as [h1 hh1].
    rewrite <- h1, <- h2; apply hd_compat; symmetry; auto.
  Qed.
  
  Lemma Has_not_Moved1_same1:
    forall g e, Step (s_hd e) g -> Proper pequiv (s_cons g e) ->
                Has_Moved1 (s_cons g e) = false -> g =~= s_hd e.
  Proof.
    intros g e st p hm.
    unfold Has_Moved1 in hm.
    erewrite existsb_forallb with
        (notf := fun n => negb (ACT_b n (s_cons g e))) in hm.
    2: intros n; rewrite negb_involutive; auto.
    rewrite negb_false_iff in hm.
    rewrite forallb_forall_InA in hm.
    2: intros n1 n2 en; apply f_equal, ACT_b_compat; auto.
    intros nn n en.
    ssplit p; destruct p as [pg pe].
    rewrite (pg _ _ en); clear nn en.
    specialize (hm n (all_nodes_prop n)).
    rewrite negb_true_iff in hm.
    unfold ACT_b in hm.
    apply has_moved_false in hm; auto.
    symmetry; auto.
  Qed.

  Lemma is_exec_Has_Moved1:
    forall g e, is_exec (s_cons g e) ->
                Has_Moved1 (s_cons g e) = true.
  Proof.
    intros g e ise.
    dup ise ise'.
    apply is_exec_inv_1 in ise'.
    apply Step_one_has_moved_true in ise'.
    destruct ise' as [n hm].
    unfold Has_Moved1; rewrite existsb_exists_InA.
    2: intros nn nnn en; apply ACT_b_compat; auto.
    2: apply is_exec_proper in ise; auto.
    exists n; split; [apply all_nodes_prop|].
    unfold ACT_b; auto.
  Qed.

  Lemma Term_wd: Weak_Dec pequiv Term.
  Proof.
    intros.
    intros e.
    destruct (terminal_dec (s_hd e)); [left|right; auto].
    intros pe; apply t; apply hd_compat, pe.
  Qed.

  Lemma Term_compat: Proper pequiv Term.
  Proof.
    intros.
    intros x y xy.
    apply terminal_compat, hd_compat, xy.
  Qed.

  (* dans Exec *)
  Definition endS := (fun s:Exec => match s with | s_one _ => True
                                            | s_cons _ _ => False end).
  Definition finiteT := (Finally endS).
  Lemma finite_finiteT:
    forall e: Exec, finiteT e -> finite e.
  Proof. intros e; apply Finally_Eventually. Qed.

  Lemma finiteT_finite:
    forall e: Exec, Proper pequiv e -> finite e -> finiteT e.
  Proof.
    intros e; apply Eventually_Dec_Finally'.
    intros [g | g ee].
    - left; tauto.
    - right; tauto.
  Qed.

  Lemma is_exec_finite:
    forall e (ise: is_exec e), Finally Term e -> finiteT e.
  Proof.
    intros e ise ft; apply finiteT_finite.
    apply is_exec_proper in ise; auto.
    induction ft.
    - destruct s as [g | g e].
      left; auto.
      apply is_exec_inv_1 in ise.
      apply Step_not_terminal in ise.
      simpl in p; tauto.
    - right; apply IHft.
      apply is_exec_inv_2 in ise; auto.
  Qed.

  Lemma is_exec_finite':
    forall e (ise: is_exec e), finiteT e -> Finally Term e.
  Proof.
    intros e ise ft; induction ft.
    - destruct s as [g | g e].
      left; apply is_exec_inv_0 in ise; auto.
      unfold endS in p; tauto.
    - right.
      apply is_exec_inv_1 in ise.
      apply Step_not_terminal in ise.
      simpl; auto.
      apply IHft.
      apply is_exec_inv_2 in ise; auto.
  Qed.

  Lemma is_constant_dec_finite:
    forall (e: Exec) (a: Env),
      Proper pequiv e -> Proper pequiv a ->
      finiteT e -> {is_constant a e} + {~is_constant a e}.
  Proof.
    intros e a pe pa fin.
    induction fin.
    - destruct s as [g | g e]; [|tauto].
      destruct (Env_dec g a).
      + left; left; simpl; apply p0; auto.
        ssplit pe; auto.
      + right; intros F; contradict n.        
        inversion F; auto.
    - destruct (Env_dec a0 a).
      + ssplit pe; destruct pe as [pa0 pe].
        specialize (IHfin pe).
        destruct IHfin.
        * left; right.
          simpl; apply p; auto.
          apply i.
        * right; intros F; contradict n0.
          inversion F; auto.
      + right; intros F; contradict n0.
        inversion F; auto.
  Qed.

  Lemma finiteT_non_constant_Acc_RSkip:
    forall (e: Exec) (a: Env), 
      Proper pequiv e -> finiteT e ->
      ~ is_constant a e -> Acc (Rskip a) e.
  Proof.
    intros e a pe fin nc.
    induction fin.
    - unfold endS in p; destruct s as [g | g e]; [|tauto].
      apply Acc_intro.
      intros s' rs.
      specialize (rs pe); destruct rs as [rs|[_ rs]].
      + ssplit rs.
      + contradict nc; left; simpl; symmetry.
        ssplit rs; apply rs.
    - apply Acc_intro.
      intros s' rs.
      specialize (rs pe).
      destruct rs as [rs|[_ rs]]; [|ssplit rs].
      eapply Acc_per_morphism.
      { apply Rskip_compat.
        ssplit rs; destruct rs as [rs _]; rewrite rs; auto. 
      }
      { intros x y rxy px; specialize (rxy px).
        destruct rxy as [rxy | [rxy1 rxy2]].
        { destruct x as [ax | ax x]; [ssplit rxy |].
          ssplit rxy; destruct rxy as [_ rxy].
          unfold Proper; rewrite rxy at 1; symmetry; auto.
        }
        unfold Proper; rewrite rxy1 at 1; symmetry; auto.
      }
      { ssplit rs; apply rs. }
      { apply IHfin.
        ssplit pe; apply pe.
        { intros F; contradict nc.
          right.
          simpl; ssplit rs; symmetry; apply rs.
          revert F; apply is_constant_compat.
          ssplit rs; destruct rs as [rs _]; rewrite rs; auto. 
          ssplit rs; destruct rs as [_ rs]; rewrite <- rs; auto.
        }      
      }
  Qed.

  Lemma collapsible_finite:
    forall e: Exec, Proper pequiv e -> finiteT e -> Squeezable e.
  Proof.
    intros e pe fin; induction fin.
    { unfold endS in p; destruct s as [a | a s]; [| tauto].
      left; simpl; left; left; simpl; ssplit pe; apply pe.
    }
    { right; simpl.
      2: apply IHfin; ssplit pe; apply pe.
      unfold Skippable.
      destruct (is_constant_dec_finite (s_cons a s) a); auto.
      ssplit pe; apply pe.
      right; auto.
      right; apply finiteT_non_constant_Acc_RSkip; auto.
      right; auto.
    }
  Qed.

  Definition Move :=
    (fun e: Exec => match e with
                    | s_one _ => false
                    | s_cons g e =>
                      if Env_dec (s_hd e) g then false
                      else true
                    end).
  Lemma Move_compat: Proper fequiv Move.
  Proof.
    unfold Move.
    intros [x | x sx] [y | y sy] xy; ssplit xy; auto; destruct xy as [xy1 xy2].
    destruct (Env_dec (s_hd sx) x); destruct (Env_dec (s_hd sy) y); auto.
    contradict n; rewrite <- xy1.
    transitivity (s_hd sx).
    apply hd_compat; symmetry; auto.
    apply p.
    unfold Proper; apply hd_compat; rewrite xy2 at 2; auto.
    unfold Proper; rewrite xy1 at 2; auto.
    contradict n; rewrite xy1.
    transitivity (s_hd sy).
    apply hd_compat; auto.
    apply p.
    unfold Proper; apply hd_compat; rewrite <- xy2 at 2; symmetry; auto.
    unfold Proper; rewrite <- xy1 at 1; auto.
  Qed.

  Lemma Nb_Steps_rewrite:
    forall (e: Exec) (fe: finiteT e) P Q,
      Always (fun e => P e = Q e) e -> Nb_Steps P fe = Nb_Steps Q fe.
  Proof.
    intros e fe P Q al; induction fe; simpl; auto.
    case_eq (P (s_cons a s)); intros hp; case_eq (Q (s_cons a s)); intros hq.
    - apply f_equal, IHfe, Always_inv_2 with a; auto.
    - apply Always_P_P in al.
      rewrite hp, hq in al; contradict al; auto with bool.
    - apply Always_P_P in al.
      rewrite hp, hq in al; contradict al; auto with bool.
    - apply IHfe, Always_inv_2 with a; auto.
  Qed.
  
  Lemma star_Move:
    forall (e: Exec) (pe: Proper pequiv e) (fe: finiteT e) a
           s (ps: Proper pequiv s) (fs: finiteT s),
      is_cons_star a s e ->
            (~s_hd e =~= s_hd s -> S (Nb_Steps Move fs) =
                                   Nb_Steps Move fe) /\
            (s_hd e =~= s_hd s -> Nb_Steps Move fs =
                                  Nb_Steps Move fe).
  Proof.
    intros e pe fe.
    intros a s ps fs star.
    revert ps fs; induction star; intros p1 f1.
    { split; intros hd.
      - contradict hd; apply hd_compat; symmetry; auto.
      - apply Nb_Steps_compat; auto.
        unfold Proper; apply Move_compat.
        intros [x | x sx] [y | y sy] xy; ssplit xy; simpl; tauto.
    }
    split; intros n.
    { ssplit pe; destruct pe as [pb p2].
      assert (f2: finiteT s2) by
          (inversion fe; unfold endS in *; [tauto|auto]).
      specialize (IHstar p2 f2 p1 f1).
      destruct (Env_dec (s_hd s2) (s_hd s1)) as [h12 | h12]. 
      + specialize (h12 (hd_compat p2) (hd_compat p1)).
        destruct IHstar as [_ ih]; specialize (ih h12).
        rewrite ih.
        inversion fe; [unfold endS in *; tauto|]; subst.
        change (~ (fun s => match s with
                            | s_one _ => True
                            | s_cons _ _ => False
                            end) (s_cons b s2)) in H1.
        transitivity (Nb_Steps Move (finally_not_yet _ H1 X)).
        simpl.
        case (Env_dec (s_hd s2) b); intros s2b.
        * contradict n; simpl s_hd.
          rewrite <- h12.
          symmetry; apply s2b; auto; unfold Proper;
            rewrite h12 at 2; auto.
        * apply f_equal.
          apply Nb_Steps_compat; auto.
          apply Move_compat.
          intros [x | x sx] [y | y sy] xy; simpl; try tauto.
          ssplit xy.
          symmetry in xy; ssplit xy.
        * apply Nb_Steps_compat; auto.
          apply Move_compat.
          intros [x | x sx] [y | y sy] xy; simpl; try tauto.
          ssplit xy.
          symmetry in xy; ssplit xy.
          right; auto.
          
      + destruct IHstar as [ih _]; specialize (ih h12).
        rewrite ih.
        inversion fe; [unfold endS in *; tauto|]; subst.
        change (~ (fun s => match s with
                            | s_one _ => True
                            | s_cons _ _ => False
                            end) (s_cons b s2)) in H1.
        transitivity (Nb_Steps Move (finally_not_yet _ H1 X)).
        simpl.
        destruct (Env_dec (s_hd s2) b).
        * apply Nb_Steps_compat; auto.
          apply Move_compat.
          intros [x | x sx] [y | y sy] xy; simpl; try tauto.
          ssplit xy.
          symmetry in xy; ssplit xy.
        * contradict h12.
          apply hd_compat.
          symmetry; apply is_cons_star_epsilon with a; auto.
        * apply Nb_Steps_compat; auto.
          apply Move_compat.
          intros [x | x sx] [y | y sy] xy; simpl; try tauto.
          ssplit xy.
          symmetry in xy; ssplit xy.
          right; auto.
    }
    { ssplit pe; destruct pe as [pb p2].
      assert (f2: finiteT s2)
        by (inversion fe; [unfold endS in *; tauto|auto]).
      specialize (IHstar p2 f2 p1 f1).
      destruct (Env_dec (s_hd s2) (s_hd s1)) as [h12 | h12].
      + specialize (h12 (hd_compat p2) (hd_compat p1)).
        destruct IHstar as [_ ih]; specialize (ih h12).
        rewrite ih.
        inversion fe; [unfold endS in *; tauto|]; subst.
        change (~ (fun s => match s with
                            | s_one _ => True
                            | s_cons _ _ => False
                            end) (s_cons b s2)) in H1.
        transitivity (Nb_Steps Move (finally_not_yet _ H1 X)).
        simpl.
        destruct (Env_dec (s_hd s2) b).
        * apply Nb_Steps_compat; auto.
          apply Move_compat.
          intros [x | x sx] [y | y sy] xy; simpl; try tauto.
          ssplit xy.
          symmetry in xy; ssplit xy.
        * contradict n0.
          rewrite n; auto.
        * apply Nb_Steps_compat; auto.
          apply Move_compat.
          intros [x | x sx] [y | y sy] xy; simpl; try tauto.
          ssplit xy.
          symmetry in xy; ssplit xy.
          right; auto.
      + dup h12 h12'.
        contradict h12'; apply hd_compat; symmetry.
        apply is_cons_star_epsilon with a; auto.
        simpl s_hd in n. symmetry in n; rewrite <- H in n;
                           rewrite n in h12.
        intros F; contradict h12; symmetry; auto.
    }
  Qed.

  (* dans RelModel *)
  Lemma step_progress:
    forall (g g': Env), Step g g' -> ~ g =~= g'.
  Proof.
    intros g g' st; destruct (Step_one_has_moved_true st).
    apply has_moved_progress in H.
    contradict H.
    apply H; auto.
  Qed.

  Lemma is_cons_star_finiteT:
    forall (e: Exec) (pe: Proper pequiv e) a
           s (ps: Proper pequiv s),
      finiteT e -> is_cons_star a s e -> finiteT s.
  Proof.
    intros e pe a s ps fe star; apply finiteT_finite; auto.
    apply finite_finiteT in fe; auto.
    induction star.
    - revert fe; apply finite_compat; auto.
    - apply IHstar; auto.
      ssplit pe; apply pe.
      apply finite_inv with b; auto.
  Qed.
      
  Lemma same_steps_simulation:
    forall (e: Exec) s (sim: Simulation s e)
           (fe: finiteT e) (fs: finiteT s),
      Proper pequiv e -> is_exec s -> Nb_Steps Move fs == Nb_Steps Move fe.
  Proof.
    intros e s sim fe fs pe ise.
    assert (ps: Proper pequiv s) by (apply is_exec_proper in ise; auto).
    (* induction sur s *)
    revert sim fe pe; revert e;
      induction fs as [s Ts| gs s nTgs fTs ih];
      intros e sim fe pe; simpl.
    { destruct s as [a | a s]; [| unfold endS in *; tauto].
      assert (C := Simulation_inv_one sim).
      clear - e C fe pe.
      induction fe; simpl; auto.
      inversion C; subst; simpl; auto.
      case (Env_dec (s_hd s) a0); intros aa.
      apply IHfe; auto.
      ssplit pe; apply pe.
      contradict aa.
      simpl in H1; rewrite H1.
      apply Always_P_P in H2; auto.
    }
    { ssplit ps; destruct ps as [_ ps].
      destruct fe as [e Te| ge e nTge fe]; simpl;
        [destruct e as [gg | gg e]; [inversion sim|tauto]|].
      inversion sim; subst.
      case (Env_dec (s_hd s) gs); intros sgs.
      - apply is_exec_inv_1 in ise.
        apply step_progress in ise.        
        contradict ise.
        apply sgs.
        apply hd_compat; auto.
        unfold Proper; rewrite H3 at 2; auto.
      - ssplit pe; destruct pe as [_ pe].
        assert (fs2: finiteT s2).
        { apply is_cons_star_finiteT with e gs; auto.
          eapply Simulation_Proper_r; eauto.
        }
        assert (ps2 := Simulation_Proper_r H4).
        destruct (star_Move _ pe fe _ _ ps2 fs2 H5) as [sm1 sm2].
        case (Env_dec (s_hd e) ge); intros ege.
        + transitivity (S (Nb_Steps Move fs2)).
          apply f_equal, ih; auto.
          apply is_exec_inv_2 in ise; auto.
          apply sm1.
          contradict sgs.
          rewrite  (Simulation_hd H4), <- sgs, H3.
          apply ege.
          apply hd_compat, pe.
          unfold Proper; rewrite <- H3 at 1; auto.
        + apply f_equal.
          transitivity (Nb_Steps Move fs2).
          apply ih; auto.
          apply is_exec_inv_2 in ise; auto.
          apply sm2.          
          apply hd_compat.
          symmetry; apply (is_cons_star_epsilon H5).
          contradict ege.
          rewrite <- H3; symmetry; auto.
    }
  Qed.

  Inductive Nb_Steps_eq (P: Exec -> bool) (T: Exec -> Prop):
            nat -> Exec -> Prop :=
  | at_T_eq: forall e, T e -> Nb_Steps_eq P T 0 e
  | add_step1: forall n g e, ~ T (s_cons g e) -> Nb_Steps_eq P T n e ->
                             P (s_cons g e) = true -> Nb_Steps_eq P T (S n) (s_cons g e)
  | add_step2: forall n g e, ~ T (s_cons g e) -> Nb_Steps_eq P T n e ->
                             P (s_cons g e) = false -> Nb_Steps_eq P T n (s_cons g e).

  Instance Nb_Steps_eq_compat: Proper fequiv Nb_Steps_eq.
  Proof.
    intros P1 P2 eP T1 T2 eT nn n en.
    rewrite en; clear nn en.
    revert eT eP; revert P1 P2 T1 T2.
    cut (forall P1 P2 T1 T2 e1 e2,
            P1 =~= P2 -> T1 =~= T2 ->
            e1 =~= e2 -> Nb_Steps_eq P1 T1 n e1 ->
            Nb_Steps_eq P2 T2 n e2).
    { intros ok P1 P2 T1 T2 eT eP e1 e2 ee; split; intros h.
      eapply ok  with (e1 := e1); eauto.
      apply ok with (P1 := P2) (T1 := T2) (e1 := e2); auto;
        symmetry; auto.
    }
    { intros P1 P2 T1 T2 e1 e2 eP eT ee nb.
      revert ee; revert e2; induction nb; intros e2 ee.
      - apply at_T_eq.
        revert H; apply eT, ee. 
      - destruct e2 as [gg | gg e2]; dup ee ee'; ssplit ee'.
        apply add_step1.
        contradict H; revert H; apply eT; right; apply ee'.
        apply IHnb, ee'.
        rewrite <- H0; symmetry; apply eP; right; apply ee'.
      - destruct e2 as [gg | gg e2]; dup ee ee'; ssplit ee'.
        apply add_step2.
        contradict H; revert H; apply eT; right; apply ee'.
        apply IHnb, ee'.
        rewrite <- H0; symmetry; apply eP; right; apply ee'.
    }
  Qed.

  Lemma Nb_Steps_equiv':
    forall P T n e (fe: Finally T e), Nb_Steps_eq P T n e <-> Nb_Steps P fe = n.
  Proof.
    intros P T n e; split; intro h.
    { revert h; revert n; induction fe as [e ht | g e ht fe ih]; intros n st; simpl.
      { inversion st; subst; auto; tauto. }
      { destruct n as [| n].
        - inversion st; subst; [tauto|].
          rewrite H4.
          apply ih; auto.
        - inversion st; subst; rewrite H4.
          + apply f_equal, ih; auto.
          + apply ih; auto.
      }
    }
    { revert h; revert n; induction fe as [e ht | g e ht fe ih]; intros n st; simpl in st.
      { subst; apply at_T_eq; auto. }
      { destruct n as [| n].
        { case_eq (P (s_cons g e)); intros hp; rewrite hp in st.
          - lia.
          - apply add_step2; auto.
        }
        { case_eq (P (s_cons g e)); intros hp; rewrite hp in st.
          - apply add_step1; auto.
          - apply add_step2; auto.
        }
      }
    }
  Qed.

  Lemma Nb_Steps_eq_Finally:
    forall P T e n, Forever (fun s : Stream => {T s} + {~ T s}) e ->
                    Proper pequiv e -> Nb_Steps_eq P T n e ->
                    Finally T e.
  Proof.
    intros P T e n fdec pe nst.
    eapply Eventually_Dec_Finally''; auto.
    induction nst.
    - left; auto.
    - right.
      apply IHnst.
      inversion fdec; auto.
      ssplit pe; apply pe. 
    - right.
      apply IHnst.
      inversion fdec; auto.
      ssplit pe; apply pe. 
  Qed.
  
  Lemma Nb_Steps_eq_le:
    forall P T n e (pe: Proper pequiv e)
           (fdec: Forever (fun s : Stream => {T s} + {~ T s}) e),
      Nb_Steps_le P T n e <->
      exists k, k <= n /\ Nb_Steps_eq P T k e.
  Proof.
    intros P T.

    intros n e pe fdec; split; intros h.
    { induction h as [ n e ht| n g e st hp ih | n g e st hp ih].
      - exists 0; split; [lia|apply at_T_eq; auto].
      - inversion fdec; subst.
        destruct H0 as [t|t].
        exists 0; split; [lia|apply at_T_eq,t; auto].
        destruct hp as [k [hk hhk]]; auto.
        ssplit pe; apply pe.
        exists (S k); split; [lia|apply add_step1; auto].
      - inversion fdec; subst.
        destruct H0 as [t|t].
        exists 0; split; [lia|apply at_T_eq,t; auto].
        destruct hp as [k [hk hhk]]; auto.
        ssplit pe; apply pe.
        exists k; split; [lia|apply add_step2; auto].
    }
    { destruct h as [k [hk hhk]].
      revert hk; revert n;
        induction hhk as [ k ht| k g e ht st ih hp | k g e ht st ih hp];
        intros n hk.
      - apply at_T; auto.
      - destruct n as [|n]; [lia|].
        apply add_stepP; auto.
        apply ih; [ssplit pe; apply pe|inversion fdec; auto | lia].
      - apply add_stepNP; auto.
        apply ih; [ssplit pe; apply pe | inversion fdec; auto | lia].
    }
  Qed.

  Lemma Nb_Steps_exist:
    forall P T e, Finally T e -> exists n, Nb_Steps_eq P T n e .
  Proof.
    intros P T e f.
    exists (Nb_Steps P f).
    rewrite Nb_Steps_equiv'; reflexivity.
  Qed.
    
  Lemma Nb_Steps_eq_unique:
    forall P T e n1 n2, Nb_Steps_eq P T n1 e -> Nb_Steps_eq P T n2 e -> n1 = n2.
  Proof.
    intros P T e n1 n2 st1; revert n2; induction st1; intros n2 st2.
    { inversion st2; subst; auto; tauto. }
    { inversion st2; [tauto| | ]; subst.
      apply f_equal, IHst1; auto.
      rewrite H0 in H6; contradict H6; auto with bool.
    }
    { inversion st2; [tauto| | ]; subst.
      rewrite H0 in H6; contradict H6; auto with bool.
      apply IHst1; auto.
    }
  Qed.
      
End Tools_For_Composition_Steps.
  
Section Composition_Steps.

  Context {Chans : Channels} {Net: Network}.

  (** 1 *)
  Context (Algo1: Algorithm).
  
  Notation run1 := (@run Algo1).
  Notation S1 := (@State Algo1).
  Notation Env1 := (@Env Chans Net Algo1).
  Notation Exec1 := (@Exec Chans Net Algo1).

  Variable Assumption1: Stable_Assumption (Algo := Algo1).
  Existing Instance Assumption1.

  Notation RO1 := (ROType (Stable_Assumption := Assumption1)).
  Notation RO1_part := (getRO (Stable_Assumption := Assumption1)).

  (** 2 *)
  Context (Algo2 : Algorithm).
  
  Notation run2 := (@run Algo2).
  Notation S2 := (@State Algo2).
  Notation Env2 := (@Env Chans Net Algo2).
  Notation Exec2 := (@Exec Chans Net Algo2).
  
  Variable Assumption2: Stable_Assumption (Algo := Algo2).
  Existing Instance Assumption2.

  Notation RO2 := (ROType (Stable_Assumption := Assumption2)).
  Notation RO2_part := (getRO (Stable_Assumption := Assumption2)).

  (** 3: compo *)
  Context {CS: Composition_State Algo1 Assumption2}.
  Definition Algo3 := (Collateral_composition (Assumption2 := Assumption2)).
  Notation run3 := (@run Algo3).

  Existing Instance Algo3.
 
  Notation S3 := (S3 (Assumption2 := Assumption2)).
  Notation Env3 := (@Env Chans Net Algo3).
  Notation Exec3 := (@Exec Chans Net Algo3).
  
  Notation read1 := (read1 (Assumption2 := Assumption2)).
  Notation write1 := (write1 (Assumption2 := Assumption2)).
  Notation read2 := (read2 (Assumption2 := Assumption2)).
  Notation write2 := (write2 (Assumption2 := Assumption2)).

  Existing Instance read1_compat.
  Existing Instance write1_compat.
  Existing Instance read2_compat.
  Existing Instance write2_compat.

  Notation envread1 := (envread1 (Assumption2 := Assumption2)).
  Notation envread2 := (envread2 (Assumption2 := Assumption2)).
  Notation execread1 := (execread1 (Assumption2 := Assumption2)).
  Notation execread2 := (execread2 (Assumption2 := Assumption2)).
  
  Notation Assume1 := (Assume (Stable_Assumption := Assumption1)).
  Notation Assume2 := (Assume (Stable_Assumption := Assumption2)).

  Instance Assumption3: Stable_Assumption (Algo := Algo3) :=
    Assumption3 Assumption1 (Assumption2 := Assumption2).

  Notation Assume3 := (Assume (Stable_Assumption := Assumption3)).

  Context (N1: nat).

  Hypothesis steps1:
    Steps_complexity (Algo := Algo1) N1 Term
                     (fun e1 => Assume1 (s_hd e1) /\ is_exec e1).

  Context (N2: nat).

  Hypothesis steps2:
    Steps_complexity (Algo := Algo2) N2 Term is_exec.

  
  Hypothesis Terminal1_Assume2 :
    forall g3, Proper fequiv g3 -> Assume1 (envread1 g3) ->
               terminal (envread1 g3) -> (Assume2 (envread2 g3)).

  Lemma Assume_31: forall g3, Assume3 g3 = Assume1 (envread1 g3).
  Proof.
    intros; reflexivity.
  Qed.
    

  (**************************************************)
  (**tools ******************************************)
  
  Lemma Has_not_Moved1_same:
    forall g e, is_exec (s_cons g e) ->
                Has_Moved1 (execread1 (s_cons g e)) = false ->
                  envread1 g =~= envread1 (s_hd e).
  Proof.
    intros g e ise hm.
    assert (st := is_exec_inv_1 ise).
    unfold Has_Moved1 in hm.
    erewrite existsb_forallb with
        (notf := fun n => negb (ACT_b n (execread1 (s_cons g e)))) in hm.
    2: intros n; rewrite negb_involutive; auto.
    rewrite negb_false_iff in hm.
    rewrite forallb_forall_InA in hm.
    2: intros n1 n2 en; apply f_equal, ACT_b_compat; [auto|apply execread1_compat].
    2: apply is_exec_proper in ise; auto.
    intros nn n en.
    cut (envread1 g nn == envread1 g n).
    2: apply envread1_compat; [apply st| auto].
    intros tmp; rewrite tmp; clear tmp nn en.
    unfold envread1; symmetry.
    case (bool_dec (enabled_b (envread1 g) n) false); intros en.
    apply Step_disabled1_unchanged1; auto.
    rewrite not_false_iff_true in en.    
    specialize (hm n (all_nodes_prop n)).
    rewrite negb_true_iff, execread1_unfold in hm.
    unfold ACT_b in hm.
    {
      unfold has_moved_b in hm.
      destruct (State_dec (s_hd (execread1 e) n) (envread1 g n)) as [S|S]; [|congruence].
      rewrite <- s_hd_read1 in S; apply S.
    }
  Qed.

  (****************************************************************)
  Lemma s_hd_read2: forall e3,
      envread2 (s_hd e3) = s_hd (execread2 e3).
  Proof.
    intros e3.
    rewrite execread2_unfold.
    destruct e3; simpl; reflexivity.
  Qed.

  (****************************************************************)
  Lemma At_N_steps_3_2:
    forall n e3,
      Assume3 (s_hd e3) -> is_exec e3 ->
      terminal (envread1 (s_hd e3)) ->
      At_N_steps n Term (execread2 e3) -> At_N_steps n Term e3.
  Proof.
    induction n as [| n ih]; intros e3 ass ise term st2.
    - inversion st2.
      left; apply terminal_12_3.
      apply is_exec_proper in ise; auto; apply hd_compat; auto.
      split; [auto|].
      rewrite s_hd_read2; auto.
    - destruct e3 as [g3 | g3 e3];
        rewrite execread2_unfold in st2; inversion st2; subst.
      right.
      + intros F; contradict H2.
        simpl; apply terminal_32, F.
      + apply ih; auto.
        { revert ass; unfold Assume; apply Assume_RO_compat.
          unfold ROEnv_part; apply Stable_Variable_step.
          apply is_exec_inv_1 in ise; auto.
        }
        { apply is_exec_inv_2 with g3, ise. }
        { apply Always_P_P with
              (P := fun e3 => terminal (envread1 (s_hd e3))).
          apply Always_inv_2 with g3.
          apply terminal1_continued_bis; auto.
        }
  Qed.

    
  (** Number of steps of Algo1 executed by the execution (of Algo3) *)
  Notation term1 := (fun e3 => terminal (envread1 (s_hd e3))).
  Notation hasmoved1 := (fun e3 => Has_Moved1 (execread1 e3)).
  Notation Nb_steps1 := (@Nb_Steps _ _ _  hasmoved1 term1).
  Notation Nb_steps1_le := (@Nb_Steps_le _ _ _ hasmoved1 term1).

  Lemma term1_wd: Weak_Dec pequiv term1.
  Proof.
    intros e; rewrite s_hd_read1.
    destruct (Term_wd (execread1 e)); [left|right]; auto.
    intros p; apply t; apply execread1_compat, p.
  Qed.

  Lemma term1_compat: Proper pequiv term1.
  Proof.
    intros x y xy; rewrite s_hd_read1; rewrite s_hd_read1.
    apply Term_compat; apply execread1_compat; auto.
  Qed.

  Lemma Always_term1:
    forall e (ise: is_exec e),
      Always
        (fun ee =>
           match ee with
           | s_one _ => True
           | s_cons g e1 =>
             hasmoved1 (s_cons g e1) = false -> term1 e1 -> term1 (s_cons g e1)
           end) e.
  Proof.
    cofix h.
    intros [g | g e] ise.
    - left; simpl; auto.
    - right.
      intros hm; apply Has_not_Moved1_same in hm; auto.
      simpl; apply terminal_compat; auto.
      apply h; apply is_exec_inv_2 in ise; auto.
  Qed.      
      
  Lemma Nb_steps1_cons:
    forall n g e, is_exec (s_cons g e) ->
                  Nb_steps1_le n (s_cons g e) -> Nb_steps1_le n e.
  Proof.
    intros n g e ise nst.
    assert (pge: Proper pequiv (s_cons g e))
      by (apply is_exec_proper in ise; auto).
    assert (f := Nb_Steps_Finally term1_wd pge nst).
    rewrite (Nb_Steps_equiv _ f) in nst.
    inversion f; subst.
    - apply Nb_Steps_more with 0; auto.
      apply term1_wd.
      ssplit pge; apply pge.
      constructor 1.
      apply terminal1_continued with g.
      apply is_exec_inv_1; auto.
      apply H.
      lia.
    - assert (Nb_steps1 f ==
              Nb_steps1 (finally_not_yet _ H0 X)).
      { apply Nb_Steps_compat; auto.
        intros E EE EEE;
          apply Has_Moved1_compat, execread1_compat, EEE.
        apply term1_compat.
      }
      rewrite H in nst.
      simpl in nst.
      rewrite (Nb_Steps_equiv _ X).
      change ((if Has_Moved1 (execread1 (s_cons g e))
               then S (Nb_steps1 X)
               else Nb_steps1 X) <= n) in nst.
      destruct (Has_Moved1 (execread1 (s_cons g e)));
        change (Nb_steps1 (e:=e) X <= n); lia.
  Qed.

  Lemma finally_term:
    forall e3: Exec3, Assume3 (s_hd e3) -> is_exec e3 -> Finally Term e3.
  Proof.
    intros e3 ass ise.
    apply Eventually_Dec_Finally'.
    { intros x; destruct (Term_wd x); [left|right]; auto. }
    { apply is_exec_proper in ise; auto. }
    eapply Convergence_to_Legitimate'.
    { intros.
      destruct (terminal_dec g); [left; apply t|right]; auto. }
    tauto.
    2: auto.    
    2: apply ass.
    2: reflexivity.
    { eapply Morphisms_Prop.Acc_rel_morphism.
      intros x y; simpl; split; intros h.
      destruct h as [_ h2]; apply h2.
      split; auto; apply Step_not_terminal with x; auto.
      reflexivity.
      eapply termination_A; auto.
      { intros g ass1 pg.
        apply Steps_complexity_Acc_term with
            (Q := fun g => Assume1 g) (N := N1); auto.
        intros gg ggg assg step; revert assg.
        unfold Assume; apply Assume_RO_compat.
        unfold ROEnv_part; apply Stable_Variable_step; auto.
      }
      { intros g ass2.
        apply Steps_complexity_Acc_term with (Q := fun g => True) (N := N2); auto.
        revert steps2.
        apply Steps_complexity_compat.
        reflexivity.
        apply Term_compat.
        intros e ee hee; split.
        intros [_ h]; rewrite <- hee; auto.
        intros h; split; [tauto|]; rewrite hee; auto.
        intros e [_ h]; apply is_exec_proper, h.
        intros e h; apply is_exec_proper, h.
      }
      apply hd_compat, is_exec_proper; auto.
    }
  Qed.

  Lemma finally_term1:
    forall e3, Assume3 (s_hd e3) -> is_exec e3 -> Finally term1 e3.
  Proof.
    intros e3 ass ise.
    apply Eventually_Dec_Finally'.
    { apply term1_wd. }
    { apply is_exec_proper in ise; auto. }
    apply Eventually_impl_P with Term.
    intros s t; apply terminal_31; auto.
    apply Finally_Eventually.
    apply finally_term; auto.
  Qed.
  
  (** *** build_proj2_sync e3
    from a stream of configurations of Algo3, 
    builds a stream of configurations of Algo2, such that:
    - until reaching a step of Algo1, it is the projection execread2
    - after this step, it is the synchronous execution of Algo2
   *)   
  CoFixpoint build_proj2_sync (e3: Exec3): Exec2 :=
    match e3 with
    | s_one g3 => s_one (envread2 g3)
    | s_cons g3 e3' =>
      if Has_Moved1 (execread1 e3) then sync_exec (envread2 g3)
      else s_cons (envread2 g3) (build_proj2_sync e3')
    end.

  Lemma build_proj2_sync_unfold:
    forall e3, build_proj2_sync e3 =
               match e3 with
               | s_one g3 => s_one (envread2 g3)
               | s_cons g3 e3' =>
                 if Has_Moved1 (execread1 e3) then
                   sync_exec (envread2 g3)
                 else s_cons (envread2 g3) (build_proj2_sync e3')
              end.
  Proof.
    intros e.
    rewrite Stream_decomposition_lemma with (build_proj2_sync e).
    unfold Stream_decompose; simpl.
    destruct e as [g | g e].
    * reflexivity.
    * destruct (Has_Moved1 (execread1 (s_cons g e))).
      + destruct (sync_exec (envread2 g)); reflexivity.
      + reflexivity.
  Qed.

  Instance build_proj2_sync_compat:
    Proper fequiv build_proj2_sync.
  Proof.
    cofix h.
    intros [g1 | g1 e1] [g2 | g2 e2] ee; ssplit ee;
      rewrite build_proj2_sync_unfold.

    + rewrite build_proj2_sync_unfold.
      left.
      apply envread2_compat; auto.

    + rewrite (build_proj2_sync_unfold (s_cons g2 e2)).
      case_eq (Has_Moved1 (execread1 (s_cons g1 e1)));
        case_eq (Has_Moved1 (execread1 (s_cons g2 e2))).

      * intros _ _; apply sync_exec_compat.
        apply envread2_compat; apply ee.
      * intros F G; contradict F; rewrite Bool.not_false_iff_true,
                                  <- G.
        apply Has_Moved1_compat, execread1_compat; right; symmetry;
          apply ee.
      * intros F G; contradict F; rewrite Bool.not_true_iff_false,
                                  <- G.
        apply Has_Moved1_compat, execread1_compat; right; symmetry;
          apply ee.
      * intros _ _; right.
        apply envread2_compat, ee.
        apply h, ee.
  Qed.
        
  Lemma build_proj2_sync_hd:
    forall e3, s_hd (build_proj2_sync e3) = envread2 (s_hd e3).
  Proof.
    intros [g | g e]; rewrite build_proj2_sync_unfold.
    - reflexivity.
    - destruct (Has_Moved1 (execread1 (s_cons g e))).
      * rewrite sync_exec_hd; reflexivity.
      * reflexivity.
  Qed.

  Lemma build_proj2_sync_is_exec:
    forall e3, is_exec e3 -> is_exec (build_proj2_sync e3).
  Proof.
    cofix h.
    intros [g3 | g3 e3] ise; rewrite build_proj2_sync_unfold.
    * left; apply is_exec_inv_0 in ise; apply terminal_32; auto.
    * case_eq (Has_Moved1 (execread1 (s_cons g3 e3))); intros HM.
      + apply sync_exec_is_exec.
        apply envread2_compat.
        apply is_exec_proper in ise.
        ssplit ise; apply ise.
      + right.
        - rewrite build_proj2_sync_hd.
          apply Step2_is_Step_Algo2.
          2: apply is_exec_inv_1, ise.
          unfold step1_b.
          unfold Has_Moved1 in HM.
          rewrite <- HM. 
          apply existsb_compat with equiv.
          2: reflexivity.
          intros n1 n2 en.
          erewrite ACT_b_compat.
          2: symmetry; apply en.
          2: apply execread1_compat,
             (is_exec_proper (Algo:=Algo3)), ise.
          clear n2 en.
          rewrite execread1_unfold.
          unfold ACT_b.
          case_eq ((enabled_b (envread1 g3) n1)); intros EN;
            case_eq (has_moved_b (s_hd e3) g3 n1); intros HM3;
              case_eq (has_moved_b (s_hd (execread1 e3))
                                   (envread1 g3) n1); intros HM1;
                auto;
                rewrite <- s_hd_read1 in HM1.
          { contradict HM1; rewrite Bool.not_false_iff_true.
            apply has_moved_31; auto.
            apply is_exec_inv_1 in ise; apply ise.
          }
          { contradict HM3; rewrite Bool.not_false_iff_true.
            apply has_moved_13; auto.
          }
          { assert (step: Step (s_hd e3) g3) by
                (apply is_exec_inv_1 in ise; auto).
            assert (H := has_moved_12_3_all n1 step).
            rewrite H, EN in HM3; clear H.
            destruct HM3 as [[X _] | [_ [HM3 _ ]]]; [congruence|].
            contradict HM1.
            rewrite not_true_iff_false; auto.
          }
          { assert (step: Step (s_hd e3) g3) by
                (apply is_exec_inv_1 in ise; auto).
            assert (H := has_moved_12_3_false n1 step).
            rewrite H in HM3.
            contradict HM1.
            rewrite not_true_iff_false; apply HM3.
          }
          
        - apply h.
          apply is_exec_inv_2 in ise; apply ise.
  Qed.

  Lemma enabled_31:
    forall g3 n,
      Proper pequiv g3 ->
      terminal (envread2 g3) ->
      enabled_b g3 n = true ->
      enabled_b (envread1 g3) n = true.
  Proof.
    intros g3 n p term.
    assert (D := terminal_all_disabled term); specialize (D n).
    unfold enabled_b at 1.
    destruct (State_dec (RUN g3 n) (g3 n)); [congruence|]; intros _.
    assert (X := composition_RUN_alt p _ _ (Equivalence_Reflexive n)).
    unfold composition_RUN_alt_def in X.
    unfold enabled in D; rewrite not_true_iff_false in D; rewrite D in X.
    destruct (enabled_b (envread1 g3) n); [auto | tauto].
  Qed.

  Lemma has_moved_31:
    forall g3' g3 n,
      Step g3' g3 ->
      terminal (envread2 g3) ->
      has_moved_b g3' g3 n = true ->
      has_moved_b (envread1 g3') (envread1 g3) n = true.
  Proof.
    intros g' g n step term hm.
    assert (D := terminal_all_disabled term); specialize (D n).
    unfold enabled in D; rewrite not_true_iff_false in D.
    rewrite has_moved_12_3_all in hm; auto.
    rewrite D in hm.
    destruct hm as [hm|[_ [_ [hm_ ]]]]; [|congruence].
    apply hm.
  Qed.

  Lemma Step_1_2:
    forall g e , Proper pequiv (s_cons g e) ->
                 Step (s_hd e) g -> terminal (envread2 g) ->
                 hasmoved1 (s_cons g e) = true.
  Proof.
    intros g e pge st term.
    destruct (Step_one_has_moved_true st) as [n hm].
    apply has_moved_31 in hm; auto.
    unfold Has_Moved1; rewrite existsb_exists_InA.
    exists n; split; [apply all_nodes_prop; auto |].
    rewrite execread1_unfold; unfold ACT_b.
    rewrite <- s_hd_read1; auto.
    intros nn nnn en; apply ACT_b_compat; auto.
    apply execread1_compat, pge.
  Qed.

    
  Lemma proj2_sync_steps:
    forall e3 n,
      is_exec e3 -> ~ term1 e3 ->
      At_N_steps n Term (build_proj2_sync e3) ->
      exists n', n' <= n /\
                 At_N_steps
                   n'
                   (fun ee3 => Has_Moved1 (execread1 ee3) = true)
                   e3.
  Proof.
    intros e n.
    revert e; induction n as [| n ih]; intros e ise nt st;
      assert (pe := is_exec_proper ise).
    { exists 0.
      split; [lia|].
      left.
      inversion st as [_1 term|]; subst.
      unfold Term in term;
        rewrite build_proj2_sync_hd in term.
      case_eq (Has_Moved1 (execread1 e)); intros hm; [auto|].
      clear st.
      destruct e as [g | g e].
      * simpl; apply is_exec_inv_0 in ise.
        contradict nt; apply terminal_31; auto.
      * apply is_exec_inv_1, Step_1_2 in ise; auto.
        rewrite ise in hm; contradict hm; auto with bool.
    }        
    {
      destruct e as [g | g e];
        rewrite build_proj2_sync_unfold in st;
        inversion st as [| ee gg _3 ntt stt heegg]; subst.
      case_eq (Has_Moved1 (execread1 (s_cons g e)));
        intros hm; rewrite hm in heegg,st.
      + exists 0; split; [lia|].
        left; auto.
      +  inversion heegg; subst; clear heegg.
         ssplit pe; destruct pe as [pg pe].
         assert (nttt: ~ terminal (envread1 (s_hd e))).
         { apply Has_not_Moved1_same in hm; auto.
           contradict nt; simpl.
           revert nt; apply terminal_compat; auto.
         }
         apply is_exec_inv_2 in ise.
         specialize (ih _ ise nttt stt).
         destruct ih as [n' [hn' huntil']].
         exists (S n').
         split; [lia|].
         right; auto.
         intros F; rewrite hm in F; auto with bool.
    }
  Qed.

  Lemma constant_term1:
    forall e3 g1,
      Assume3 (s_hd e3) -> is_exec e3 ->
      is_constant g1 (execread1 e3) -> terminal g1.
  Proof.
    intros e3 g1 ass ise.
    assert (F := finally_term1 _ ass ise).
    revert g1; induction F as [e3 T | g3 e3 NT F ih].
    { intros g1 ic.
      rewrite <- (is_constant_hd ic).
      rewrite <- s_hd_read1; auto.
    }
    { intros g1 ic.
      rewrite execread1_unfold in ic; apply is_constant_inv in ic.
      apply ih; auto.
      revert ass; unfold Assume; apply Assume_RO_compat.
      unfold ROEnv_part; apply Stable_Variable_step.
      apply is_exec_inv_1 in ise; auto.
      apply is_exec_inv_2 in ise; auto.
    }
  Qed.

  Lemma execread1_finite:
    forall e, finiteT e -> finiteT (execread1 e).
  Proof.
    intros e fin.
    induction fin.
    - destruct s as [g | g e]; [| unfold endS in *; tauto].
      rewrite execread1_unfold.
      left; tauto.
    - rewrite execread1_unfold.
      right.
      tauto.
      auto.
  Qed.  

  Lemma At_N_steps_0:
    forall P (e: Exec1), At_N_steps 0 P e -> P e.
  Proof. intros P e af; inversion af; auto. Qed.

  Lemma At_SN_steps_one:
    forall P n (g: Env1), At_N_steps (S n) P (s_one g) -> False.
  Proof. intros P n g af; inversion af; auto. Qed.

  Lemma At_SN_steps_cons:
    forall P n (g: Env1) e, At_N_steps (S n) P (s_cons g e) ->
                            At_N_steps n P e.
  Proof. intros P n g e af; inversion af; auto. Qed.

  Lemma At_N_steps_Finally:
    forall N P (e: Exec1), At_N_steps N P e -> Finally P e.
  Proof.
    intros n P; induction n as [| n ih]; intros e af.
    - apply At_N_steps_0 in af; left; auto.
    - destruct e as [g | g e].
      apply At_SN_steps_one in af; tauto.
      right.
      apply After_SN_steps_notP in af; auto.
      apply ih.
      apply At_SN_steps_cons in af; auto.
  Qed.

  Lemma bounded_step1:
    forall e, Assume3 (s_hd e) -> is_exec e -> Nb_steps1_le N1 e.
  Proof.
    intros e ass ise.

    cut (Nb_Steps_le Has_Moved1
                     (fun e => terminal (s_hd e)) N1 (execread1 e)).
    {
      pose (e1 := execread1 e).
      assert (h: e1 =~= e1) by (apply execread1_compat;
                                apply is_exec_proper in ise; auto).
      fold e1.
      unfold e1 at 2 in h.
      clearbody e1; clear steps1.
      intros nst; revert ise ass h; revert e;
        induction nst as [ N1 e1 T | N1 g1 e1 nst ih hm |
                           N1 g1 e1 nst ih hm];
        intros e ise ass h.
      { apply at_T.
        rewrite s_hd_read1.
        revert T; apply terminal_compat, hd_compat; symmetry; auto.
      }
      { destruct e as [g | g e]; dup h h';
          rewrite execread1_unfold in h'; ssplit h'.
        apply add_stepP.
        apply ih.
        apply is_exec_inv_2 in ise; auto.
        revert ass; unfold Assume; apply Assume_RO_compat.
        unfold ROEnv_part; apply Stable_Variable_step.
        apply is_exec_inv_1 in ise; auto.
        apply h'.
        rewrite <- hm; apply Has_Moved1_compat; symmetry; auto.
      }
      { destruct e as [g | g e]; dup h h';
          rewrite execread1_unfold in h'; ssplit h'.
        apply add_stepNP.
        apply ih.
        apply is_exec_inv_2 in ise; auto.
        revert ass; unfold Assume; apply Assume_RO_compat.
        unfold ROEnv_part; apply Stable_Variable_step.
        apply is_exec_inv_1 in ise; auto.
        apply h'.
        rewrite <- hm; apply Has_Moved1_compat; symmetry; auto.
      }
    }

    cut (Nb_Steps_le Has_Moved1 endS N1 (execread1 e)).
    {
      clear - e ise; pose (e1 := execread1 e).
      assert (he: e1 =~= execread1 e) by
          (unfold e1; apply execread1_compat; apply is_exec_proper in ise; auto).
      intros nst; fold e1 in nst; clearbody e1.
      revert ise he; revert e.
      induction nst.
      { destruct e as [g | g ee]; [|unfold endS in *; tauto].
        intros e ise he.
        destruct e as [gg | gg e]; rewrite execread1_unfold in he; ssplit he.
        rewrite execread1_unfold.
        apply at_T; simpl.
        apply terminal_31.
        apply is_exec_inv_0 in ise; auto.
      }
      { intros [gg | gg ee] ise he; rewrite execread1_unfold in he; ssplit he.
        rewrite execread1_unfold.
        constructor 2.
        apply IHnst; auto.
        apply is_exec_inv_2 in ise; auto.
        apply he.
        rewrite <- H; apply Has_Moved1_compat.
        right; symmetry; apply he.
      }
      { intros [gg | gg ee] ise he; rewrite execread1_unfold in he; ssplit he.
        rewrite execread1_unfold; constructor 3.
        apply IHnst; auto.
        apply is_exec_inv_2 in ise; auto.
        apply he.
        rewrite <- H; apply Has_Moved1_compat.
        right; symmetry; apply he.
      }
    }

    assert (pe: Proper pequiv e) by (apply is_exec_proper in ise; auto).

    assert (fe1: finiteT (execread1 e)).
    { apply execread1_finite.
      apply is_exec_finite; auto.
      apply finally_term; auto.
    }
    assert (e_can_collapse: Squeezable (execread1 e)).
    {
      apply collapsible_finite.
      { apply execread1_compat; apply is_exec_proper in ise; auto. }
      auto.
    }

    assert (pe1: Proper pequiv (execread1 e)) by (apply execread1_compat, pe).

    pose (ce1 := (@Squeeze _ _ Env_dec _ pe1 e_can_collapse)).
    assert (ise1: is_exec ce1).
    { 
      eapply is_exec_simu'; auto.
      apply ise.
      apply Always_MP with (fun e => is_exec e /\ Assume3 (s_hd e)).
      apply Forall_Always.
      intros ee3 [isee assee] g1 ic.
      apply constant_term1 with ee3; auto.
      apply Always_conj.
      apply is_exec_Always; auto.
      eapply RO_is_exec_Always; auto.
      apply Assume_RO_compat.
      apply Simulation_Squeeze.
      apply Squeeze_no_repeat.
    }

    destruct (steps1 ce1) as [n1 [hn1 stn1]].
    { split; auto.
      unfold ce1; rewrite Squeeze_hd.
      rewrite <- s_hd_read1.
      apply ass.
    }

    assert (x: Simulation ce1 (execread1 e))
      by apply Simulation_Squeeze. 

    assert (hhn1: At_N_steps n1 endS ce1).
    {
      clear - ce1 ise1 n1 stn1.
      induction stn1.
      { destruct s as [g | g s].
        left; unfold endS in *; tauto.
        apply is_exec_inv_1' in ise1; tauto.
      }
      { right.
        tauto.
        apply IHstn1; apply is_exec_inv_2 in ise1; auto.
      }
    }
    clear stn1.

    assert (F := At_N_steps_Finally _ _ _ hhn1).
    assert (h: Nb_Steps Has_Moved1 F = n1).
    {
      clearbody ce1; clear - ce1 ise1 F hhn1 n1.
      revert hhn1; revert n1; induction F; simpl; intros n1 hhn1.
      - destruct s as [g | g e]; [| unfold endS in *; tauto].
        inversion hhn1; lia.
      - dup ise1 ise1'.
        apply is_exec_Has_Moved1 in ise1'.
        rewrite ise1'.
        destruct n1 as [| n1]; [inversion hhn1; tauto|].
        apply f_equal, IHF.
        apply is_exec_inv_2 in ise1; auto.
        inversion hhn1; auto.
    }
    clear hhn1.
    
    assert (al1: Always (fun e => Has_Moved1 e = Move e) ce1).
    {
      clearbody ce1.
      clear - ce1 ise1.
      revert ise1; revert ce1; cofix h; intros ce1 ise1.
      destruct ce1 as [g1 | g1 ce1].
      - left.
        unfold Has_Moved1, Move, ACT_b.
        rewrite <- not_true_iff_false; intros F.
        rewrite existsb_exists in F; destruct F as [x [_ hh]].
        contradict hh; auto with bool.
      - right.
        { unfold Move; dup ise1 hm.
          apply is_exec_Has_Moved1 in hm; auto; rewrite hm.
          apply Has_Moved1_progress' in hm.
          2: apply is_exec_proper in ise1; auto.
          destruct (Env_dec (s_hd ce1) g1); auto.
          contradict hm; apply p.
          apply hd_compat; apply is_exec_proper in ise1; ssplit ise1; apply ise1.
          apply is_exec_proper in ise1; ssplit ise1; apply ise1.
        }
        apply h; apply is_exec_inv_2 in ise1; auto.
    }

    assert (al2: Always (fun e => Has_Moved1 e = Move e) (execread1 e)).
    {
      clear - e ise.
      revert ise; revert e; cofix h; intros e ise.
      destruct e as [g | g e]; rewrite execread1_unfold.
      - left.
        unfold Has_Moved1, Move, ACT_b.
        rewrite <- not_true_iff_false; intros F.
        rewrite existsb_exists in F; destruct F as [x [_ hh]].
        contradict hh; auto with bool.
      - right.
        { unfold Move. 
          case_eq (Has_Moved1 (execread1 (s_cons g e))); intros hm.
          - rewrite execread1_unfold in hm; rewrite hm; 
              apply Has_Moved1_progress' in hm.
            2: apply is_exec_proper in ise; ssplit ise; right;
              [apply envread1_compat | apply execread1_compat]; apply ise.
            destruct (Env_dec (s_hd (execread1 e)) (envread1 g)); auto.          
            contradict hm; apply p; apply is_exec_proper in ise; ssplit ise;
              [apply hd_compat, execread1_compat | apply envread1_compat]; apply ise.
          - dup hm hm'.
            apply Has_not_Moved1_same in hm; auto.
            rewrite execread1_unfold in hm'; rewrite hm'.
            destruct (Env_dec (s_hd (execread1 e)) (envread1 g)); auto.          
            contradict n; rewrite <- s_hd_read1; symmetry; auto.
        }
        apply h; apply is_exec_inv_2 in ise; auto.
    }

    assert (NN: Nb_Steps Has_Moved1 F = Nb_Steps Has_Moved1 fe1).
    {
      etransitivity.
      apply (Nb_Steps_rewrite _ F Has_Moved1 Move al1).
      symmetry; etransitivity.
      apply (Nb_Steps_rewrite _ fe1 Has_Moved1 Move al2).
      symmetry; apply same_steps_simulation; auto.
    }

    rewrite (Nb_Steps_equiv _ fe1); lia.
  Qed. 

  Lemma big_step: 
    forall k e3, Assume3 (s_hd e3) -> is_exec e3 ->
                 Nb_steps1_le (S k) e3 ->
                 exists n, n <= S N2 /\
                           At_N_steps n (Nb_steps1_le k) e3.
  Proof.
    intros K e3 ass ise nbst.    
    assert (pe3: Proper pequiv e3) by (apply is_exec_proper in ise; auto).

     assert (FT: forall e, Proper pequiv e ->
                          Forever (fun s => {terminal (envread1 (s_hd s))} +
                                            {~ terminal (envread1 (s_hd s))}) e).
    { intros e pe; assert (x := Forever_forall _ _ term1_wd e).
      clear - e pe x.
      revert e pe x; cofix h.
      intros [g | g e]; intros pe x.
      - left.
        apply Forever_P_P in x; destruct x as [x|x];
          [left; apply x; auto | right; auto].
      - right.
        apply Forever_P_P in x; destruct x as [x|x];
          [left; apply x; auto | right; auto].
        apply h.
        ssplit pe; apply pe.
        inversion x; auto.
    }      

    apply Nb_Steps_eq_le in nbst; auto.
    destruct nbst as [k [hk hhk]].
    case (Compare_dec.le_lt_eq_dec hk); intros hK.
    {
      exists 0; split; [lia|left].
      apply Nb_Steps_eq_le; auto.
      exists k; split; [lia|auto].
    }
    subst; clear hk.

    assert (f2: exists n2, n2 <= N2 /\
                           At_N_steps n2 (fun e => hasmoved1 e = true) e3).
    { clear FT.
      destruct (steps2 _ (build_proj2_sync_is_exec _ ise))
        as [n2 [hn stn2]].
      assert (nt: ~ term1 e3).
      { clear n2 hn stn2.
        pose (sk := S K).
        fold sk in hhk.
        assert (hsk: sk > 0) by (unfold sk; lia); clearbody sk; clear K ass.
        induction hhk as [e3 t | sk g3 e3 nt st ih hm |
                          sk g3 e3 nt st ih hm].
        - lia.
        - destruct sk as [| sk].
          { inversion st; subst; auto. }
          assert (pos: S sk > 0) by lia.
          assert (isee: is_exec e3) by (apply is_exec_inv_2 in ise; auto).
          assert (pee: Proper pequiv e3) by (ssplit pe3; apply pe3).
          specialize (ih isee pee pos); clear isee pee pos.
          contradict ih.
          eapply terminal1_continued.
          apply is_exec_inv_1 in ise; apply ise.
          simpl in ih; auto.
        - destruct sk as [| sk].
          { inversion st; subst; auto. }
          assert (pos: S sk > 0) by lia.
          assert (isee: is_exec e3) by (apply is_exec_inv_2 in ise; auto).
          assert (pee: Proper pequiv e3) by (ssplit pe3; apply pe3).
          specialize (ih isee pee pos); clear isee pee pos.
          contradict ih.
          eapply terminal1_continued.
          apply is_exec_inv_1 in ise; apply ise.
          simpl in ih; auto.
      }
      destruct (proj2_sync_steps _ _ ise nt stn2) as [n'2 [hn'2 stn'2]].
      exists n'2; split; [lia|auto].
    }
    destruct f2 as [n2 [hn2 stn2]].
    exists (S n2); split; [lia|].
    clear - e3 hhk pe3 n2 stn2 FT.
    induction stn2 as [e3 hm | g3 e3 n2 hm stn2 ih].
    {
      destruct e3 as [g3 | g3 e3].
      { rewrite execread1_unfold in hm.
        unfold Has_Moved1, ACT_b in hm.
        rewrite existsb_exists in hm.
        destruct hm as [x [_ F]]; contradict F; auto with bool.
      }
      { inversion hhk; subst.
        clear H4 H2.
        right.
        - intros F; rewrite Nb_Steps_eq_le in F; auto.
          destruct F as [k [hk stk]].
          assert (k = S K) by (eapply Nb_Steps_eq_unique; eauto).
          lia.
        - left.
          apply Nb_Steps_eq_le;
            [ssplit pe3; apply pe3 | apply FT; ssplit pe3; apply pe3|].
          exists K; split; [lia | auto]. 
        - rewrite hm in H4; contradict H4; auto with bool.
      }
    }
    { right.
      { clear ih n2 stn2.
        intros F; rewrite Nb_Steps_eq_le in F; auto.
        destruct F as [k [hk stk]].
        assert (k = S K) by (eapply Nb_Steps_eq_unique; eauto).
        lia.
      }
      { apply ih.
        inversion hhk; subst; auto.
        contradict hm; auto.
        ssplit pe3; apply pe3.
      }
    }
  Qed.

  (* RODAMAP *)

  (** Main Result 
   An execution of Algo3 has the following shape:
     ->->...steps2 (at most N2) ... -> step1 -> ... steps2 (at most N2) -> step1 -> ...

    (at most N1 step1) -> (terminal1 is reached) -> at most N2 step2 (terminal2 is reached)

   step1 = at least one node executes Algo1
   step2 = every node that moves executes Algo2
   *)
  Theorem Collateral_step_complexity:
    Steps_complexity (Algo := Algo3) (N1 * (S N2) + N2)
                     Term (fun e3 => Assume3 (s_hd e3) /\ is_exec e3).
  Proof.
    apply Steps_complexity_add with term1.
    { intros e3 [_ ise]; destruct (term1_wd e3); [left|right]; auto.
      apply t; apply is_exec_proper in ise; auto.
    }
    { intros e3 [_ ise]; destruct (Term_wd e3); [left|right]; auto.
      apply t; apply is_exec_proper in ise; auto.
    }
    { apply term1_compat. }
    { apply Term_compat. }
    { intros e [ass ise]; apply Always_conj.
      eapply RO_is_exec_Always; auto.
      apply Assume_RO_compat.
      apply is_exec_Always; auto.
    }
    { intros e [_ ise]; apply is_exec_proper; auto. }
    { intros e _; apply terminal_31. }
    {
      rewrite <- sum_prod.
      pose (ass_exec := fun e3: Exec3 => Assume3 (s_hd e3) /\ is_exec e3).
      fold ass_exec.
      eapply Steps_complexity_monotonic with
          (Q := fun e : Exec3 => ass_exec e /\ Nb_steps1_le N1 e)
          (P := Nb_steps1_le 0).
      { intros e3 [ass ise]; split; [split|]; auto.
        apply bounded_step1; auto.
      }
      { cofix h.
        intros [g | g e] [[ass ise] nbst].
        - left; split; [split|]; auto.
        - right.
          split; [split|]; auto.
          apply h.
          split; [split|].
          { revert ass; unfold Assume; apply Assume_RO_compat.
            unfold ROEnv_part; apply Stable_Variable_step.
            apply is_exec_inv_1; auto.
          }
          { apply is_exec_inv_2 with g; auto. }
          { apply Nb_steps1_cons with g; auto. }
      }
      { intros e [[ass ise] nst] nst0.
        change ((fun ee => terminal (envread1 (s_hd ee))) e).
        eapply Nb_Steps_0.
        apply term1_wd.
        apply is_exec_proper in ise; auto.
        2: apply nst0.
        apply Always_term1; auto.
      }
      { intros e [[_ ise] _].
        destruct (terminal_dec (envread1 (s_hd e))).
        left; apply t; apply envread1_compat, hd_compat.
        apply is_exec_proper in ise; auto.
        right; auto.        
      }
      { apply (Steps_complexity_induction
                 (Q := ass_exec) (Pot := Nb_steps1_le)).
        { intros n e ise; apply Nb_Steps_le_wd.
          apply finally_term1, ise.
          apply ise.
        }
        apply Nb_Steps_le_compat.
        { intros x y xy.
          apply Has_Moved1_compat, execread1_compat, xy. }
        { apply term1_compat. }
        { intros e [ass ise]; apply Always_conj.
          eapply RO_is_exec_Always; auto.
          apply Assume_RO_compat.
          apply is_exec_Always; auto.
        }
        { intros e [_ ise]; apply is_exec_proper in ise; auto. }
        { intros e [ass ise] n0 n.
          apply Nb_Steps_more with 0; auto.
          apply is_exec_proper in ise; auto.
          apply term1_wd.
          apply is_exec_proper in ise; auto.
          lia.
        }
        { intros k e [[ass ise] nS].
          apply big_step; auto.
        }
      }
    }
    { intros e3 [[ass3 ise3] term1].
      destruct (steps2 (execread2 e3)) as [n [hn hhn]].
      - simpl; apply is_exec_3_2; auto.
      - exists n; split; auto.        
        apply At_N_steps_3_2; auto.
    }        
  Qed.
  
  (****************************************************************)  
  (****************************************************************)  
  (****************************************************************)  
  (****************************************************************)  
  (****************************************************************)  
  (****************************************************************)  
  (****************************************************************) 
  
End Composition_Steps.

Close Scope signature_scope.
Close Scope nat_scope.

(* 

silent -> forall exec, finite number of rounds TODO


   *)
