(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
*)
(* begin show *)

(** **** Global imports **)
From Coq Require Import Arith.
From Coq Require Import NArith.
From Coq Require Import Setoid.
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import RelationPairs.
From Coq Require Import Morphisms.
From Coq Require Import List.
From Coq Require Import Lia.

(** **** Local imports **)
From Padec Require Import NatUtils.
From Padec Require Import Stream.
From Padec Require Import Algorithm.
From Padec Require Import OptionUtil.
From Padec Require Import ListUtils.
From Padec Require Import FoldRight.
From Padec Require Import SetoidUtil.
From Padec Require Import NatBuild.
From Padec Require Import RelModel.
From Padec Require Import P_Q_Termination.
From Padec Require Import Self_Stabilization.
From Padec Require Import Fairness.
From Padec Require Import Exec.

(* end show *)
Open Scope signature_scope.
Open Scope nat_scope.
Set Implicit Arguments.

Section Decr_Generic_Net.

  Context {Chans: Channels}
          {Net: Network}.

  (** * (Very Simple) Example *)

  (** The state of the nodes are natural numbers. *)

  (** This algorithm does not use neighborhood:
    At each step, the node decreases its state. *)

  Program Instance Decr: Algorithm :=
    {|
      State_setoid := Setoid_nat;
      State_dec := eq_nat_dec;
      run := fun n _  => pred n;
      (*  from Library Coq.Init.Datatype: 
          unit is a singleton datatype with sole inhabitant tt *)
    |}.

  (* begin hide *)
  Next Obligation.
    repeat intro;apply f_equal;auto.
  Qed.
  (* end hide *)

End Decr_Generic_Net.

Section Decr.

  (** Nodes are also natural numbers, labeled from [0] to [K]-1. *)
  Context {Chans:Channels}.

  Variable K: N.

  (** The network is made of [K] nodes, with no edge. *)
  Program Instance NetK: Network:=
    {| Node_dec := ltK_dec (K := K);
       peer := fun _ _ => None;
       link := fun _ _ => None;
       all_nodes := Kenum K
    |}.

  (* begin hide *)
  Next Obligation.
    tauto.
  Qed.

  Next Obligation.
    intros [n1 hn1] [n2 hn2] en c1 c2 ec.
    hnf; trivial.
  Qed.

  Next Obligation.
    repeat intro;hnf.
    tauto.
  Qed.

  Next Obligation.
      apply ltK_in_Kenum.
  Defined.
  (* end hide *)

  (** Configurations *)
  Notation env := (@Env _ NetK Decr).

  (** Terminal Configuration to reach *)
  Definition target (g: env ) := forall n , g n = 0.

  (** [target] is indeed a terminal configuration, i.e. no more valid
      step can be executed from it. *)
  Theorem final: forall e, target e <-> terminal e.
  Proof.
    intros e.
    rewrite terminal_def_equiv.
    unfold terminal_alt.
    split; intros h. intros s1 s2 es.
    simpl.
    rewrite (h s1), (h s2); reflexivity.
    intros n.
    specialize (h _ _ (Equivalence_Reflexive n)).
    simpl in h.
    lia.
  Qed.

  (** Potential is defined as the sum of the states on all nodes. *)
  Definition potential g :=
    List.fold_right plus 0 (List.map g (@all_nodes _ NetK)).

  (** Any valid step decreases the global potential *)
  Theorem potential_decreases:
    forall (g g': env), Step g' g -> potential g' < potential g.
  Proof.
    intros g g' [pg pg' ngg' hgg'].
    assert (hrel: forall n, g' n <= g n).
    {
      intros n;destruct (hgg' n) as [h|h];rewrite h;simpl;lia.
    }
    clear hgg'.
    assert (hgg':forall l, fold_right Init.Nat.add 0 (List.map g' l) <=
                           fold_right Init.Nat.add 0 (List.map g l)).
    {
      intro n.
      apply fold_right_compat with le.
      repeat intro;lia.
      lia.
      apply (map_compat (eqA:=eq)).
      intros n1 n2 en;rewrite en;auto.
      reflexivity.
    }
    cut (potential g' <> potential g).
    {
      specialize (hgg' all_nodes).
      unfold potential;simpl in *;lia.
    }
    intro h;apply ngg';clear ngg'.
    intros n_ n en.
    rewrite en;clear n_ en.
    generalize (ltK_in_Kenum (K := K) n).
    revert h n.
    unfold potential;simpl all_nodes.
    induction (Kenum K) as [|j q hq];simpl.
    {
      intros _ n h;rewrite InA_nil in h;elim h.
    }
    {
      intros hh n h.
      destruct (le_lt_eq_dec (hrel j)) as [LT|EQ].
      {
        specialize (hgg' q);simpl in *;lia.
      }
      {
        rewrite InA_cons in h;destruct h as [h|h].
        rewrite h;apply EQ.
        apply hq;auto;simpl in *;lia.
      }
    }
  Qed.

  (** Steps of this algorithm on this network are well founded *)
  Lemma WF_decr: well_founded (@Step _ NetK Decr).
  Proof.
    assert (H := Inverse_Image.wf_inverse_image env nat lt potential lt_wf).
    eapply Inclusion.wf_incl; [| apply H].
    intros g' g h; auto using potential_decreases.
  Qed.

  Program Instance Null_Assumption: Stable_Assumption (Algo:= Decr) :=
    {|
      getRO:= fun _ => tt;
      Assume_RO:= fun _ => True;
    |}.

  (* begin hide *)
  Next Obligation.
    esplit;repeat intro;reflexivity.
  Qed.

  Next Obligation.
    esplit;repeat intro;reflexivity.
  Qed.

  (* end hide *)

  Lemma SS_decr:
    self_stabilization Null_Assumption unfair_daemon
                       (fun e => target (Stream.s_hd e)).
    split with target;repeat apply conj.
    {
      unfold closure.
      cofix h; intros [g | g e] ass ise uf t.
      left; auto.
      apply final in t.
      apply is_exec_inv_1, Step_not_terminal in ise.
      tauto.
    }
    {
      intros e.
      remember (s_hd e) as g; revert Heqg; revert e.
      assert (wf := WF_decr g).
      induction wf.
      intros [g | g e] shd ass ise uf.
      left.
      apply is_exec_inv_0 in ise.
      apply final; auto.
      right.
      apply H0 with (s_hd e); auto.
      simpl in shd; rewrite shd; inversion ise; auto.
      inversion ise; auto.
    }
    {
      hnf;firstorder.
    }
  Qed.

End Decr.

Close Scope signature_scope.
Close Scope nat_scope.
Unset Implicit Arguments.