(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidClass.
From Coq Require Import SetoidList.
From Coq Require Import Bool.

(** **** Local Imports *)
From Padec Require Import Stream.
From Padec Require Import WellfoundedUtil.
From Padec Require Import SetoidUtil.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.

(* end show *)
Open Scope signature_scope.
Set Implicit Arguments.

Section Exec.

  Context {Chans: Channels} {Net: Network} {Algo: Algorithm}.

  (** ** Execution
   An execution is a maximal sequence of steps of the algorithm.
   *)

  (** *** Type Exec:
  Exec = finite or infinite sequence of configurations (Env)   *)

  Definition Exec: Type := Stream (A := Env).
  Definition finite_exec := (finite (A := Env)).
  Definition infinite_exec := (infinite (A := Env)).

  (** is_exec:
   a maximal sequence (i.e. infinite unless it reaches a [terminal]
   configuration) such that any two consecutive configurations in
   the sequence actually form a [Step] of the algorithm *)

  Definition is_exec: Exec -> Prop :=
    fun e => Always (fun e => match e with
                              | s_one g => terminal g
                              | s_cons g e => Step (s_hd e) g end) e.

  Lemma is_exec_inv_0: forall g, is_exec (s_one g) -> terminal g.
  Proof.
    intros g h; inversion h; auto.
  Qed.

  Lemma is_exec_inv_1:
    forall g e, is_exec (s_cons g e) -> Step (s_hd e) g.
  Proof.
    intros g e ise; inversion ise; auto.
  Qed.

  Lemma is_exec_inv_1':
    forall g e, is_exec (s_cons g e) -> ~ terminal g.
  Proof.
    intros g e ise; eapply Step_not_terminal.
    apply is_exec_inv_1, ise.
  Qed.

  Lemma is_exec_inv_2: forall g e, is_exec (s_cons g e) -> is_exec e.
  Proof.
    intros g e ise; inversion ise; auto.
  Qed.

  Global Instance is_exec_compat: Proper fequiv is_exec.
  Proof.
    apply Always_compat.
    intros [g1 | g1 e1] [g2 | g2 e2] ee; ssplit ee.
    rewrite ee; auto.
    destruct ee as [ee1 ee2].
    apply Step_compat; auto.
    apply hd_compat; auto.
  Qed.

  Global Instance is_exec_s_hd_compat: forall e, is_exec e ->
                                                 Proper pequiv (s_hd e).
  Proof.
    intros e is_e; destruct e as [g | g e]; simpl s_hd.
    apply is_exec_inv_0 in is_e; apply is_e.
    apply is_exec_inv_1 in is_e; apply is_e.
  Qed.

  Global Instance is_exec_proper: forall e, is_exec e -> Proper pequiv e.
  Proof.
    cofix h.
    intros e is_e; destruct e as [g | g e].
    constructor 1; apply is_exec_inv_0 in is_e; apply is_e.
    constructor 2.
    apply is_exec_inv_1 in is_e; apply is_e.
    apply h; apply is_exec_inv_2 in is_e; apply is_e.
  Qed.

  Lemma proper_Always_hd_proper: forall e,
      Proper pequiv e -> Always (fun e => Proper pequiv (s_hd e)) e.
  Proof.
    cofix h; intros [g | g e] hh.
    left; apply hd_compat; auto.
    right.
    apply hd_compat; auto.
    apply h; inversion hh; auto.
  Qed.

  (** is_exec_alt : aternative is_exec as a conjuction of simpler properties *)

  Notation T := (fun e => match e with s_one g => terminal g
                                  | s_cons _ _ => True end).
  Notation S := (fun e => match e with s_one g => True
                                  | s_cons a s => Step (s_hd s) a end).
  Lemma is_exec_alt:
    forall e, is_exec e <-> Always S e /\ Always T e.
  Proof.
    intros e.
    cut (Proper pequiv e -> (is_exec e <-> Always S e /\ Always T e)).
    { intros ok.
      split; intros h.
      specialize (ok (is_exec_proper h)).
      apply ok; auto.
      assert (Proper pequiv e).
      clear ok.
      revert h; revert e; cofix h.
      intros [g | g e] [h1 h2].
      inversion h2; subst.
      left; apply H0.
      right.
      inversion h1; subst.
      apply H1.
      apply h.
      split; [inversion h1|inversion h2]; auto.
      apply ok; auto.
    }
    rewrite <- Always_conj.
    intros pe; apply Always_compat; auto.
    intros [g1 | g1 e1] [g2 | g2 e2] h; inversion_clear h.
    rewrite H; simpl; tauto.
    rewrite (Step_compat (hd_compat H0) H); simpl; tauto.
  Qed.

  Lemma is_exec_Always:
    forall e: Exec, is_exec e -> Always is_exec e.
  Proof.
    cofix h.
    intros [g | g e] ise.
    left; auto.
    right; auto.
    apply h, is_exec_inv_2 with g; auto.
  Qed.

  Lemma finite_exec_E_terminal:
    forall e, is_exec e -> finite_exec e ->
              Eventually (fun e => terminal (s_hd e)) e.
  Proof.
    intros e ise fin.
    induction fin.
    {
      apply eventually_now.
      destruct s as [g | g e]; [|tauto].
      apply is_exec_inv_0; auto.
    }{
      apply eventually_later.
      apply IHfin.
      apply is_exec_inv_2 with a; auto.
    }
  Qed.

  Lemma finite_exec_E_terminal_converse:
    forall e, is_exec e -> Eventually (fun e => terminal (s_hd e)) e ->
              finite_exec e.
  Proof.
    intros e ise E.
    induction E as [[g | g e] now | g e later ih].
    { left; auto. }
    { contradict now; apply is_exec_inv_1' with e, ise. }
    { right; apply ih; eapply is_exec_inv_2, ise. }
  Qed.

  Lemma RO_is_exec_Always:
    forall (VarType: Type) (getVar: State -> VarType)
           (SV: Stable_Variable _ getVar),
    forall (P: (Node -> VarType) -> Prop) (P_compat: Proper fequiv P) e,
      is_exec e -> P (fun n => getVar ((s_hd e) n)) ->
      Always (fun e => P (fun n => getVar ((s_hd e) n))) e.
  Proof.
    intros until SV; intros P P_compat.
    cofix h; intros [g | g e] ise hP.
    left; auto.
    right; auto.
    apply h.
    apply is_exec_inv_2 in ise; auto.
    eapply step_assume; eauto.
    apply is_exec_inv_1 in ise; auto.
  Qed.

  Lemma is_exec_When: forall P e (fe: Finally P e), is_exec e ->
                                                    is_exec (When fe).
  Proof.
    intros P e fe ise.
    induction fe; simpl.
    assumption.
    apply IHfe.
    eapply is_exec_inv_2; apply ise.
  Qed.

  Lemma is_exec_Assume:
    forall {SA: Stable_Assumption} e,
      Assume (s_hd e) -> is_exec e ->
      Always (fun e => Assume (s_hd e)) e.
  Proof.
    intros SA.
    cofix h.
    intros [g | g e] ass ise.
    - left; auto.
    - right; auto.
      apply h.
      apply step_Assume with g; auto.
      apply is_exec_inv_1 in ise; auto.
      apply is_exec_inv_2 in ise; auto.
  Qed.      

  Lemma suffix_Assume:
    forall {SA: Stable_Assumption} (e: Exec),
      is_exec e -> Assume (s_hd e) ->
      forall e_suf,
        is_suffix e e_suf -> Assume (s_hd e_suf).
  Proof.
    intros SA e ise ass e_suf hsuf.
    eapply is_exec_Assume in ise; auto.
    eapply Always_suffix in ise.
    3: apply hsuf.
    apply Always_P_P in ise; auto.
    { intros x y xy.
      apply Assume_compat, Stream.hd_compat; auto. }
  Qed.

  (** * Enabled / Activated / Neutralized *)
  Section Enabled_Activated_Neutralized.

    (** Node is enabled at the beginning of execution: *)
    Definition EN_b (n: Node) (e: Exec): bool := enabled_b (s_hd e) n.
    Definition EN (n: Node) (e: Exec): Prop := EN_b n e = true.

    (** Node is activated at the beginning of execution: *)
    Definition ACT_b (n: Node) (e: Exec): bool :=
      match e with
        s_one _ => false
      | s_cons g e' => has_moved_b (s_hd e') g n
      end.
    Definition ACT (n: Node) (e: Exec): Prop := ACT_b n e = true.

    (** Node is neutralized at the beginning of execution: *)
    Definition NEUT_b (n: Node) (e: Exec): bool :=
      match e with
        s_one _ => false
      | s_cons g e' =>
        andb (enabled_b g n)
             (andb (negb (has_moved_b (s_hd e') g n))
                   (negb (enabled_b (s_hd e') n)))
      end.

    Definition NEUT (n: Node) (e: Exec): Prop := NEUT_b n e = true.

    (** Node is activated or neutralized at the beginning of execution: *)
    Definition AN (n: Node) (e: Exec): Prop := ACT n e \/ NEUT n e.
    Definition AN_b (n: Node) (e: Exec): bool := ACT_b n e || NEUT_b n e.

    Lemma AN_alt: forall n e, AN n e <-> AN_b n e = true.
    Proof.
      intros n e.
      unfold AN, ACT, NEUT, AN_b.
      rewrite orb_true_iff; tauto.
    Qed.

    Definition act_neut_b e n := AN_b n e.

    (** Some direct properties (from the model) *)
    Lemma act_en: forall (n: Node) (e: Exec), is_exec e -> ACT n e -> EN n e.
    Proof.
      unfold ACT, EN, ACT_b, EN_b; intros n e ise a.
      destruct e as [g | g e]; [congruence|]; simpl.
      apply has_moved_b_enabled_b with (s_hd e); auto.
      apply is_exec_inv_1; auto.
    Qed.

    Lemma neut_en: forall (n: Node) (e: Exec), NEUT n e -> EN n e.
    Proof.
      unfold NEUT, EN, NEUT_b, EN_b; intros n e a.
      destruct e as [g | g e]; [congruence|]; simpl.
      rewrite andb_true_iff in a.
      tauto.
    Qed.

    Lemma AN_one: forall n g, ~AN n (s_one g).
    Proof.
      intros n g; unfold AN, ACT, NEUT, ACT_b, NEUT_b; simpl.
      intuition.
    Qed.

    Lemma AN_en: forall n e, is_exec e -> AN n e -> EN n e.
    Proof.
      intros n e ise an; unfold AN in an; destruct an as [A|N].
      apply act_en; auto.
      apply neut_en; assumption.
    Qed.

    Lemma en_progress: forall n e g,
        EN n (s_cons g e) -> ~EN n e -> AN n (s_cons g e).
    Proof.
      unfold AN, EN, EN_b, NEUT, NEUT_b, ACT, ACT_b;
        intros n e g en enn; simpl in *.
      destruct (has_moved_b (s_hd e) g n).
      - left; auto.
      - right.
        rewrite andb_true_iff; split; [assumption | rewrite andb_true_iff].
        split; [rewrite negb_true_iff; auto| apply eq_true_not_negb; auto].
    Qed.

    Lemma en_terminal: forall (g: Env), terminal g ->
                                        forall (n: Node), ~EN n (s_one g).
    Proof.
      intros g term n; unfold EN, EN_b; simpl.
      eapply terminal_all_disabled with (n:=n) in term.
      unfold enabled in term; auto.
    Qed.

    Lemma ACT_mov:
      forall n g e, ACT n (s_cons g e) ->
                    ~( g n == s_hd e n).
    Proof.
      intros n g e act.
      unfold ACT, ACT_b in act; simpl in act.
      intros F; symmetry in F; revert F;
        apply has_moved_progress; assumption.
    Qed.

    Lemma NEUT_mov:
      forall n g e, NEUT n (s_cons g e) -> ~( g =~= s_hd e ).
    Proof.
      intros n g e neut.
      unfold NEUT, NEUT_b in neut; simpl in neut.
      rewrite andb_true_iff in neut; destruct neut as [en neut].
      rewrite andb_true_iff in neut; destruct neut as [_ nen].
      rewrite negb_true_iff in nen.
      intros F.
      unfold enabled_b in en, nen.
      destruct (State_dec (RUN g n) (g n)); [contradict en; congruence|].
      destruct (State_dec (RUN (s_hd e) n) (s_hd e n)); [|contradict nen;
                                                          congruence].
      contradict n0; rewrite (F _ _ (Equivalence_Reflexive n)), <- e0.
      apply RUN_compat; auto.
    Qed.

    Lemma AN_mov:
      forall n g e, AN n (s_cons g e) -> ~(g =~= s_hd e).
    Proof.
      intros n g e; unfold AN; intros [H|H].
      apply ACT_mov in H.
      intros F; contradict H; apply F; reflexivity.
      apply NEUT_mov with n; assumption.
    Qed.

    (** ** Compatibility *)
    Global Instance EN_b_compat: Proper fequiv EN_b.
    Proof.
      intros n1 n2 en e1 e2 ee; unfold EN_b.
      apply enabled_b_compat; [apply hd_compat, ee | apply en].
    Qed.

    Global Instance EN_compat: Proper fequiv EN.
    Proof.
      intros n1 n2 en e1 e2 ee; unfold EN.
      rewrite EN_b_compat; eauto.
    Qed.

    Global Instance ACT_b_compat: Proper fequiv ACT_b.
    Proof.
      intros n1 n2 en e1 e2 ee; unfold ACT_b.
      destruct e1 as [g1 | g1 e1]; destruct e2 as [g2 | g2 e2];
        ssplit ee; auto.
      apply has_moved_b_compat; auto.
      apply hd_compat, ee.
      apply ee.
    Qed.

    Global Instance ACT_compat: Proper fequiv ACT.
    Proof.
      intros n1 n2 en e1 e2 ee; unfold ACT.
      rewrite ACT_b_compat; eauto.
    Qed.

    Global Instance NEUT_b_compat: Proper fequiv NEUT_b.
    Proof.
      intros n1 n2 en e1 e2 ee; unfold NEUT_b.
      destruct e1 as [g1 | g1 e1]; destruct e2 as [g2 | g2 e2];
        ssplit ee; auto.
      destruct ee as [eg ee].
      assert (E1: enabled_b g1 n1 = enabled_b g2 n2)
        by (apply enabled_b_compat; auto).
      assert (E2: enabled_b (s_hd e1) n1 = enabled_b (s_hd e2) n2)
        by (apply enabled_b_compat; auto; apply hd_compat; auto).
      assert (E3: has_moved_b (s_hd e1) g1 n1 = has_moved_b (s_hd e2) g2 n2)
        by (apply has_moved_b_compat; auto; apply hd_compat; auto).
      rewrite E1, E2, E3; reflexivity.
    Qed.

    Global Instance NEUT_compat: Proper fequiv NEUT.
    Proof.
      intros n1 n2 en e1 e2 ee; unfold NEUT.
      rewrite NEUT_b_compat; eauto.
    Qed.

    Global Instance AN_b_compat: Proper fequiv AN_b.
    Proof.
      intros n1 n2 en e1 e2 ee; unfold AN_b.
      apply f_equal2.
      apply ACT_b_compat;auto.
      apply NEUT_b_compat;auto.
    Qed.

    Global Instance AN_compat: Proper fequiv AN.
    Proof.
      intros n1 n2 en e1 e2 ee; unfold AN.
      apply or_compat; [apply ACT_compat|apply NEUT_compat]; auto.
    Qed.

    Lemma E_AN_not_constant:
      forall e n, Eventually (AN n) e -> forall g, ~ is_constant g e.
    Proof.
      intros e n E g C.
      induction E as [e now|gg e later ih].
      { destruct e as [x | x e].
        - unfold AN, ACT, NEUT in now; simpl in now.
          destruct now as [H|H]; congruence.
        - apply AN_mov in now.
          contradict now.
          inversion C; subst; simpl s_hd in H1.
          apply Always_P_P in H2.
          rewrite H1; symmetry; assumption.
      }{ apply ih; apply is_constant_inv with gg; apply C. }
    Qed.

    Lemma AN_act_neut_b: forall e n, act_neut_b e n = true <-> AN n e.
    Proof.
      intros e n;apply orb_true_iff.
    Qed.

    Lemma EN_enabled_b: forall e n, enabled_b (s_hd e) n = true <-> EN n e.
    Proof.
      intros e n; reflexivity.
    Qed.

    Global Instance act_neut_b_compat: Proper fequiv act_neut_b.
    Proof.
      intros e1 e2 ee n1 n2 en; unfold act_neut_b;
        apply AN_b_compat; assumption.
    Qed.

    Lemma act_neut_b_enabled_b: forall e n, is_exec e ->
                                            act_neut_b e n = true ->
                                            enabled_b (s_hd e) n = true.
    Proof.
      intros e n ise H; rewrite AN_act_neut_b in H.
      apply AN_en in H; unfold EN, EN_b in H; auto.
    Qed.

    Lemma ACT_b_EN_b: forall n e, is_exec e ->
                                  ACT_b n e = true ->
                                  EN_b n e = true.
    Proof.
      intros n [g | g e] ise; unfold ACT, ACT_b, EN, EN_b;
        try congruence; simpl.
      apply has_moved_b_enabled_b.
      apply is_exec_inv_1,ise.
    Qed.

    Lemma ACT_EN: forall n e, is_exec e -> ACT n e -> EN n e.
    Proof.
      unfold ACT, EN; apply ACT_b_EN_b.
    Qed.

    Lemma act_neut_contd:
      forall g (e: Exec) n,
        enabled_b g n = true ->
        act_neut_b (s_cons g e) n = false ->
        enabled_b (s_hd e) n = true.
    Proof.
      intros g e n en an.
      unfold act_neut_b, AN_b in an.
      apply orb_false_elim in an.
      unfold ACT_b, NEUT_b in an.
      destruct an as [hm an].
      apply andb_false_elim in an.
      destruct an as [an|an]; [congruence|].
      apply andb_false_elim in an.
      destruct an as [an|an]; rewrite negb_false_iff in an;
        congruence.
    Qed.

    Lemma act_b_true:
      forall (g: Env) e p,
        Step (s_hd e) g ->
        ACT_b p (s_cons g e) = true -> RUN g p == s_hd e p.
    Proof.
      intros g e p st an; unfold ACT_b in an.
      apply has_moved_true in an; auto; symmetry; auto.
    Qed.

    Lemma neut_b_true:
      forall (g: Env) e p,
        Step (s_hd e) g ->
        NEUT_b p (s_cons g e) = true -> RUN (s_hd e) p == s_hd e p.
    Proof.
      intros g e p st an; unfold NEUT_b in an.
      apply andb_prop in an; destruct an as [_ an].
      apply andb_prop in an; destruct an as [_ an].
      rewrite negb_true_iff in an.
      apply enabled_b_false in an; auto.
    Qed.

  End Enabled_Activated_Neutralized.

  Section Synchronous.

    CoFixpoint sync_exec (g: Env): Exec :=
      if terminal_dec g then s_one g
      else s_cons g (sync_exec (RUN g)).

    Lemma sync_exec_unfold:
      forall g, sync_exec g =
                if terminal_dec g then s_one g
                else s_cons g (sync_exec (RUN g)).
    Proof.
      intros g.
      rewrite Stream_decomposition_lemma with (sync_exec g).
      unfold Stream_decompose; simpl.
      destruct (terminal_dec g); reflexivity.
    Qed.

    Lemma sync_exec_hd: forall g, (s_hd (sync_exec g)) = g.
    Proof.
      intros g.
      rewrite sync_exec_unfold.
      destruct (terminal_dec g); simpl; reflexivity.
    Qed.

    Global Instance sync_exec_compat: Proper fequiv sync_exec.
    Proof.
      intros g1 g2 eg; revert g1 g2 eg; cofix h.
      intros g1 g2 eg; rewrite (sync_exec_unfold g1);
        rewrite (sync_exec_unfold g2).
      destruct (terminal_dec g1); destruct (terminal_dec g2).
      - left; auto.
      - contradict n; rewrite <- (terminal_compat eg); apply t;
          unfold Proper; rewrite eg at 2; auto.
      - contradict n; rewrite (terminal_compat eg); apply t;
          unfold Proper; rewrite <- eg at 1; auto.
      - right; auto.
        apply h, RUN_compat, eg.
    Qed.

    Definition sync_step (g' g: Env): Prop :=
      forall n, enabled_b g n = true -> has_moved_b g' g n = true.

    Definition synchronous_exec: Exec -> Prop :=
      Always (fun e => match e with
                       | s_one _ => True
                       | s_cons g e => sync_step (s_hd e) g end).

    Lemma synchronous_exec_sync_exec:
      forall g, Proper pequiv g -> synchronous_exec (sync_exec g).
    Proof.
      cofix h.
      intros g pg.
      rewrite sync_exec_unfold.
      destruct (terminal_dec g).
      left; tauto.
      right.
      2: apply h.
      rewrite sync_exec_hd.
      intros n_ en.
      unfold has_moved_b.
      assumption.
      apply RUN_compat,pg.
    Qed.

    Lemma sync_exec_is_exec:
      forall g, Proper pequiv g -> is_exec (sync_exec g).
    Proof.
      cofix h.
      intros g pg.
      rewrite sync_exec_unfold.
      destruct (terminal_dec g).
      * left; apply t; auto.
      * right.
        rewrite sync_exec_hd.
        apply RUN_Step; auto.
        apply h.
        apply RUN_compat,pg.
    Qed.

  End Synchronous.

  Section Sequential.

    Definition at_most_one_enabled (g: Env): Prop :=
      forall n n', enabled_b g n = true -> enabled_b g n' = true -> n == n'.

    Definition sequential_exec: Exec -> Prop :=
      Always (fun e => at_most_one_enabled (s_hd e)).

  End Sequential.

End Exec.

Close Scope signature_scope.
Unset Implicit Arguments.
