(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Bool.

(** **** Local Imports *)
From Padec Require Import WellfoundedUtil.
From Padec Require Import SetoidUtil.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Stream.
From Padec Require Import Exec.

(* end show *)
Open Scope signature_scope.
Set Implicit Arguments.

Section Fairness.

  Context {Chans: Channels}
          {Net: Network} {Algo: Algorithm}.

  (** * Unfair Daemon: the most permissive daemon *)
  Definition unfair_daemon (e: Exec): Prop := True.

  (** * Weakly Fair Execution *)
  Section Weakly_Fair.

    (** “Weakly fair” means that there is no infinite suffix of
        execution in which a process p is continuously enabled without
        ever being selected by the daemon.

        In a weakly fair execution, processes are fairly activated,
        i.e., if a process is continuously enabled, then it eventually
        executes one of its enabled actions.
     *)

    (** ** Definition *)
    Definition weakly_fair (e: Exec): Prop :=
      forall (n: Node),
        Always (fun e' => EN n e' -> Eventually (AN n) e') e.

    Lemma weakly_fair_inv:
      forall g e, weakly_fair (s_cons g e) -> weakly_fair e.
    Proof.
      intros g e wf n.
      specialize (wf n).
      inversion wf; assumption.
    Qed.

    Lemma weakly_fair_Always:
      forall e, weakly_fair e -> Always weakly_fair e.
    Proof.
      cofix h.
      intros [g | g e] wf.
      - left; auto.
      - right; auto.
        apply h.
        apply weakly_fair_inv with g; auto.
    Qed.

    (** ** Finite Executions are weakly fair *)
    Lemma WF_finite_exec':
      forall (e: Exec), is_exec e -> finite_exec e ->
                        forall (n: Node),
                          Always (Eventually (EN n)) e -> False.
    Proof.
      intros e ise fin n A.
      induction fin as [ [g | g e] he | g e];
        inversion ise; inversion A; subst; try tauto.
      - inversion H2; subst.
        apply en_terminal with g n in H0; tauto.
    Qed.

    Lemma WF_finite_exec:
      forall (e: Exec), is_exec e -> finite_exec e ->
                        forall (n: Node), EN n e -> Eventually (AN n) e.
    Proof.
      intros e ise fin n en.
      induction fin as [ [g | g e] he | g e he]; inversion ise; subst;
        try tauto.
      - apply en_terminal with g n in H0; tauto.
      - case_eq (EN_b n e); intro enn.
        * apply eventually_later; apply IHhe; assumption.
        * apply eventually_now; apply en_progress;
            [assumption | unfold EN; rewrite enn; apply diff_false_true].
    Qed.

    Lemma WF_finite_exec'':
      forall (e: Exec), is_exec e -> finite_exec e ->
                        forall (n: Node),
                          ~Always (fun e' : Exec => EN n e' /\ ~ AN n e') e.
    Proof.
      intros e ise fin n A.
      induction fin as [ [g | g e] he | g e he term];
        inversion ise; inversion A; subst;
          try tauto.
      apply en_terminal with g n in H0; tauto.
    Qed.

    Lemma weakly_fair_finite_exec:
      forall (e: Exec), is_exec e -> finite_exec e -> weakly_fair e.
    Proof.
      intros e ise fin n.
      induction fin as [ [g | g e] he | g e he term];
        inversion ise; subst; try tauto.
      apply always_one.
      apply WF_finite_exec; [ apply ise | left ]; tauto.
      apply always_cons.
      apply WF_finite_exec; [ apply ise | right; assumption ].
      apply term; assumption.
    Qed.

    Lemma en_E_A:
      forall (e: Exec), is_exec e ->
                        forall (n: Node), EN n e -> WUntil (EN n) (AN n) e.
    Proof.
      intros e ise n.
      revert ise. revert e.
      cofix en_E_A.
      destruct e as [g | g e]; intros ise en;
        inversion ise; subst;try (simpl in *;tauto).

      apply en_terminal with g n in H0; tauto.
      case_eq (EN_b n e); intros enc.
      apply wuntil_later; [|apply en_E_A]; assumption.
      apply wuntil_now.
      apply en_progress; [assumption|
                          unfold EN; rewrite enc; apply diff_false_true].
    Qed.

    (** ** Necessary conditions for weakly fair *)
    Lemma WF_NC1:
      forall (e: Exec), weakly_fair e ->
                        forall (n: Node), Always (Eventually (EN n)) e ->
                                          Always (Eventually (AN n)) e.
    Proof.
      intros e H n; specialize (H n).
      apply AE_tool3; assumption.
    Qed.
    (** Note that the converse cannot be proven:
   It would require:
   -   EN n e -> Eventually (AN n) e \/ Always (EN n) e
   namely, decide wether becomes AN forever or remains EN forever
   -- it seems like halting problem...

   Actually, we have:
   - EN n e -> (EN n) UNTIL (AN n) e
   but no more...
     *)

    Lemma WF_NC2:
      forall (e: Exec), weakly_fair e ->
                        forall (n: Node),
                          ~Eventually (Always (fun e' =>
                                                 EN n e' /\ ~ AN n e')) e.
    Proof.
      intros e H n; specialize (H n).
      apply AE_tool4.
      2: assumption.
      unfold EN.
      intros ee; case_eq (EN_b n ee); auto.
    Qed.
    (** We do not succeed in proving the converse.
   It would require:
   -   ~Eventually (Always (fun e' : Exec => EN n e' /\ ~ AN n e')) e
       -> EN n e -> Eventually (AN n) e
     *)

    (** ** Compatibility *)
    Global Instance weakly_fair_compat: Proper fequiv weakly_fair.
    Proof.
      intros e1 e2 ee; unfold weakly_fair.
      apply all_compat; intros n.
      apply Always_compat; auto.
      clear e1 e2 ee; intros e1 e2 ee.
      apply impl_compat.
      apply EN_compat; auto.
      apply Eventually_compat; auto.
      apply AN_compat; auto.
    Qed.

  End Weakly_Fair.

  (** * Suffix of Execution *)
  Section Suffix.

    (** ** Properties *)
    Lemma suffix_is_exec:
      forall (e: Exec), is_exec e ->
                        forall (e_suf: Exec),
                          is_suffix e e_suf -> is_exec e_suf.
    Proof.
      intros e ise e_suf suf.
      induction suf as [e s_now| g e s_later IH].
      eapply is_exec_compat.
      symmetry; apply s_now.
      apply ise.
      apply IH.
      apply is_exec_inv_2 with g; auto.
    Qed.

    Lemma suffix_weakly_fair:
      forall (e: Exec), weakly_fair e ->
                        forall (e_suf: Exec),
                          is_suffix e e_suf -> weakly_fair e_suf.
    Proof.
      intros e wf e_suf suf.
      induction suf as [e s_now| g e s_later IH].
      { eapply weakly_fair_compat.
        symmetry in s_now; apply s_now.
        apply wf. }
      { apply IH.
        intros n; specialize (wf n).
        inversion wf; assumption. }
    Qed.

    Lemma suffix_finite_exec:
      forall (e e_suf: Exec) (is_suf: is_suffix e e_suf),
        finite_exec e <-> finite_exec e_suf.
    Proof.
      intros e e_suf suf.
      split; intros H.
      { induction suf as [e s_now| g e s_later IH].
        - symmetry in s_now; eapply finite_compat; eauto.
        - apply IH, finite_inv with g; auto.
      }{ induction suf as [e s_now| g e s_later IH].
         - eapply finite_compat; eauto.
         - right; apply  IH.
       }
    Qed.

  End Suffix.

  (* * Strongly Fair Execution *)
  Section Strongly_Fair.

    Definition strongly_fair (e: Exec): Prop :=
      forall (n: Node), Always (Eventually (EN n)) e ->
                        Always (Eventually (ACT n)) e.

  End Strongly_Fair.

  Section Weakly_Fair_Unfair.

    Lemma weakly_fair_unfair:
      forall e, weakly_fair e -> unfair_daemon e.
    Proof.
      unfold unfair_daemon; intuition.
    Qed.
    
  End Weakly_Fair_Unfair.
  
End Fairness.

Close Scope signature_scope.
Unset Implicit Arguments.
