(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** **** Global imports **)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.

(** **** Local imports **)
From Padec Require Import OptionUtil.
From Padec Require Import Stream.
From Padec Require Import SetoidUtil.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Exec.

(* end show *)
Open Scope signature_scope.
Set Implicit Arguments.

(** * Functional Model

  Semantics of the model (Algorithm, Network, Channels) using
  functions. Compared to the relational model semantics, it
  additionnaly uses a daemon.

  Looking at an execution, next step from a configuration is computed,
  using the <<run>> function applied on nodes selected by the daemon.

  Relational and Functional semantics are proven equivalent using the
  daemon.

  We work on executions. The daemon, in this case, in a coninductive
  type <<Daemon>> (each element represents the selected nodes at the
  current step).

  We require the daemon to be sound (namely to select next state of
  enabled nodes only) and to achieve progress (at least one node
  should be selected if possible).

  And we prove,

  - soundness: the daemon is equipped with a function that computes an
    execution from an initial configuration. We prove that the result
    is actually an execution from the algorithm.

  - completeness: we prove that any execution of the algorithm is
    actually produce using a daemon (namely that there exists a daemon
    which produces an equivalent execution. This latter proof builds
    the daemon step by step.

 **)
Section FunModel.

  Context {Chans: Channels} {Net: Network} {Algo: Algorithm}.

  (** ** Daemon on executions *)
  Section Daemon_Exec.

    (** a <<Daemon>> is given by <<(Daemon_Step select next)>>.

    Daemon <<next>> is the daemon for next step.

    The role of function <<select: Diff -> Diff>> is to represent the
    enabled nodes selected by the daemon.
    From a difference built by the algorithm, <<d>>, <<(select d)>>
    is another difference: for node <<p>> such that
    - <<(d p)>> is <<None>>, <<p>> is disabled and <<(select d)>> is
      also <<None>>;
    - <<(d p)>> is <<Some s>>, where <<s>> is the next state of <<p>>,
          - <<p>> can be chosen by the daemon: in this case, <<(select d)>> equals <<Some s>>, namely <<(d p)>>;
          - <<p>> is not chosen by the daemon: in this case, <<(select d)>> equals <<None>>.

    We call this property: soundness of the select function. We
    require, futhermore, that the daemon actually selects an enabled
    nodes if possible: this is the progress property of the select
    function. Those constraints on the function <<select>> are
    expressed in the predicate <<sound_progress>>

    The fact that the daemon satisfies <<sound_progress>> at each step
    is expressed in the predicate <<is_Daemon>>.
     *)
    CoInductive Daemon: Type :=
      Daemon_Step:
        (Env -> Env -> Env) -> (* selects the enabled nodes *)
        (Env -> Env -> Daemon) -> (* Daemon for next step *)
        Daemon.

    (** equality (partial equivalence relation) on daemon *)
    CoInductive eqDaemon: relation Daemon :=
      eqDaemon_Step:
        forall f1 f2,
          fequiv f1 f2 ->
          forall dm1 dm2, (fequiv ==> fequiv ==> eqDaemon) dm1 dm2 ->
                          eqDaemon (Daemon_Step f1 dm1) (Daemon_Step f2 dm2).

    Global Instance eqDaemon_per: PER eqDaemon.
    Proof.
      split.

      unfold Symmetric.
      cofix h.
      intros D1 D2 eD.
      destruct D1; destruct D2; inversion eD; subst.
      apply eqDaemon_Step;auto.
      symmetry;auto.
      intros f1 f2 ef g1 g2 eg.
      apply h,H4;symmetry;assumption.

      unfold Transitive.
      cofix h.
      intros D1 D2 D3 e12 e23.
      destruct D1;  destruct D2;  inversion D3;
        inversion e12; inversion e23;subst.
      apply eqDaemon_Step.
      transitivity e0;auto.
      intros f1 f2 ef g1 g2 eg.
      apply h with (d0 f1 g1); auto.
      apply H4;auto.
      apply H9;auto.
    Qed.

    Global Instance Daemon_PSetoid: PartialSetoid Daemon
      := @Build_PartialSetoid _ eqDaemon _.

    Definition daemon_decompose (d: Daemon): Daemon :=
      match d with Daemon_Step select next => Daemon_Step select next end.

    Lemma daemon_decomposition_lemma:
      forall (d: Daemon), d = daemon_decompose d.
    Proof.
      intros d.
      destruct d; simpl; reflexivity.
    Qed.

    (** constraints on the select function of the daemon:
     first line expresses soundness and second line progress.
     *)
    Definition sound_progress (select: Env -> Env -> Env): Prop :=
      forall g1 g2, Proper fequiv g1 -> Proper fequiv g2 ->
                    (forall n : Node, select g1 g2 n == g2 n \/ select g1 g2 n == g1 n)
                    /\
                    (select g1 g2 =~= g1 -> (g1 =~= g2)). (* progress *)

    Instance sound_progress_compat:
      Proper fequiv sound_progress.
    Proof.
      intros sel1 sel2 esel; unfold sound_progress.
      apply all_compat; intros g1;apply all_compat;intros g2;apply impl_iff_compat_ext.
      reflexivity.
      intro pg1.
      apply impl_iff_compat_ext.
      reflexivity.
      intro pg2.
      apply and_compat.
      apply all_compat; intros n;apply or_compat;apply equiv_compat;reflexivity || apply esel;auto.
      apply impl_compat;try reflexivity.
      apply pequiv_compat;try apply esel;auto.
    Qed.

    (** Characterization of a daemon: we require that at each step,
     the select function is compatible and verifies soundness and
     progress.*)
    CoInductive is_Daemon: Daemon -> Prop :=
      is_Daemon_Step:
        forall (select: Env -> Env -> Env) (next: Env -> Env -> Daemon),
          Proper fequiv select ->
          sound_progress select ->
          Proper fequiv next ->
          (forall g1 g2, Proper fequiv g1 -> Proper fequiv g2 -> is_Daemon (next g1 g2) ) ->
          is_Daemon (Daemon_Step select next).

    Global Instance is_Daemon_compat: Proper (eqDaemon ==> iff) is_Daemon.
    Proof.
      assert (tmp: forall d1 d2: Daemon,
                 eqDaemon d1 d2 -> is_Daemon d1 -> is_Daemon d2).
      { cofix h.
        intros [s1 n1] [s2 n2] ed X;
          inversion ed; inversion X; subst.
        apply is_Daemon_Step.
        proper_.
        eapply sound_progress_compat; [symmetry; eauto|]; auto.
        proper_.
        intros g1 g2 pg1 pg2;specialize (H10 g1 g2 pg1 pg2).
        eapply h;eauto.
        apply H4;eassumption.
      }
      intros d1 d2; split; apply tmp; [| symmetry]; auto.
    Qed.

    (** Daemons are compatible. *)
    Lemma is_Daemon_proper:
      forall (daem: Daemon), is_Daemon daem -> Proper eqDaemon daem.
    Proof.
      cofix h.
      intros daem id.
      destruct daem.
      apply eqDaemon_Step.
      inversion id; assumption.
      intros g1 g2 eg g1' g2' eg'.
      inversion id; subst;eauto.
      apply H3;auto.
    Qed.

    (** *** Synchronous Daemon

    Example of daemon: synchronous daemon choses every enabled nodes, at each step.
     *)
    Definition sync_select (g1 g2 : Env): Env := g2.

    Global Instance sync_select_compat: Proper fequiv sync_select.
    Proof.
      intros _g _g' _eg g g' eg.
      unfold sync_select; assumption.
    Qed.

    Lemma sync_sound_progress: sound_progress sync_select.
    Proof.
      intros g1 g2 pg1 pg2.
      unfold sync_select in *.
      split.
      left;auto.
      symmetry;auto.
    Qed.

    CoFixpoint d_sync: Daemon := Daemon_Step sync_select (fun _ _ => d_sync).

    Lemma d_sync_unfold: d_sync = Daemon_Step sync_select (fun _ _ => d_sync).
    Proof.
      rewrite (daemon_decomposition_lemma d_sync) at 1.
      simpl; reflexivity.
    Qed.

    Lemma d_sync_compat: Proper pequiv d_sync.
      hnf.
      cofix d_sync_compat.
      rewrite d_sync_unfold.
      apply eqDaemon_Step.
      apply sync_select_compat.
      intros _1 _2 _e _3 _4 _ee.
      apply d_sync_compat.
    Qed.

    Lemma d_sync_is_daemon: is_Daemon d_sync.
    Proof.
      cofix H.
      rewrite d_sync_unfold.
      apply is_Daemon_Step.
      apply sync_select_compat.
      apply sync_sound_progress.
      intros _1 _2 _e _3 _4 _ee;apply d_sync_compat.
      intros;apply H.
    Qed.

    (** ** Soundness (execution) *)

    (** Execution built from a daemon and an initial configuration:

      the select function of the daemon applies to <<(RUN init)>>;
      next configuration is then obtained with <<diff_eval>> on the
      result.
     *)
    CoFixpoint build_exec (init: Env) (daem: Daemon): Exec :=
      match daem with
      | Daemon_Step select next =>
        if (terminal_dec init)
        then s_one init
        else s_cons init (build_exec (select init (RUN init)) (next init (RUN init)))
      end.

    Lemma build_exec_unfold:
      forall  (init: Env) (daem: Daemon),
        build_exec init daem =
        match daem with
        | Daemon_Step select next =>
          if (terminal_dec init)
          then s_one init
          else s_cons init (build_exec (select init (RUN init)) (next init (RUN init)))
        end.
    Proof.
      pose (tmp init daem :=match daem with
                            | Daemon_Step select next =>
                              if (terminal_dec init)
                              then s_one init
                              else s_cons init (build_exec (select init (RUN init)) (next init (RUN init)))
                            end).

      assert (forall init daem, build_exec init daem = tmp init daem).
      intros init daem.
      rewrite Stream_decomposition_lemma with (build_exec init daem); simpl.
      unfold tmp; destruct daem.
      destruct (terminal_dec init); reflexivity.
      intros init daem; rewrite H; unfold tmp; reflexivity.
    Qed.

    (** Any execution (<<Exec>>) built using a daemon is actually
     an exection (<<is_exec>> from relational stemantics) *)
    Theorem computational_soundness:
      forall daem, is_Daemon daem ->
                   forall init, Proper pequiv init ->
                                is_exec (build_exec init daem).
    Proof.
      intros daem is_daem init init_compat;
        revert daem is_daem init init_compat.
      cofix h.
      intros daem is_daem init init_compat.
      rewrite build_exec_unfold.
      destruct daem; simpl.
      inversion is_daem; subst.
      case (terminal_dec init).
      {
        intros term.
        left; apply term; assumption.
      }
      intros.
      right.
      shelve.
      apply h;auto.
      apply H4; [|apply RUN_compat]; auto.
      apply H1; [|apply RUN_compat]; auto.
      Unshelve.
      assert (tmp: Step (e init (RUN init)) init).
      {
        split;auto.
        apply H1; [|apply RUN_compat]; auto.
        shelve.
        apply H2;try apply RUN_compat;apply init_compat.
        Unshelve.
        intro hh;apply n;clear n;unfold terminal;split.
        proper_.
        intros gg [h1 h2 h3 h4].
        apply H2 in hh;try apply RUN_compat;try proper_.
        apply h3.
        intros n1 n2 en.
        rewrite <- en.
        destruct (h4 n1) as [hhh | hhh];rewrite hhh;symmetry;auto.
      }
      {
        simpl.
        destruct d.
        case (terminal_dec (e init (RUN init)));auto.
      }
    Qed.

    (** ** Completeness (execution) *)

    (** To prove completeness, we have to build a daemon from an
     execution: we build the select function as follows. Note that
     the result has to verify soundness and progress.

     If the parameter <<d>> of the select function does not come from
     the execution of the algorithm (if it is not <<(RUN (s_hd e))>>,
     then we return <<d>> to achieve soundness (and progress).

     In either case, when the execution is over, the select function
     selects nothing and returns <<None>> for any parameter. When the
     execution is going on,
          - if <<(d p)>> is None, then the node is disabled: we return
          <<None>> also for soundness
          - if <<(d p)>> is not None and equals the state of <<p>> in
          the next configuration of the execution, <<p>> may have
          moved or may have produced the same state but not moved.
               - if it has moved, we return <<(d p)>> to achieve
                 soundness and progress
               - otherwise, it has not moved, but its local algorithm
                 in the current state computes exactly this same
                 state: in this case, the daemon can select it or not
                 without changing the computation. As we have no mean
                 to distinguish, in our model, this case, from the
                 previous one, we also return <<(d p)>>.
          - in the last case, (<<(d p)>> is not None, but differs from
          next state), the node has not moved and we return <<None>>.
     *)
    Definition build_select (e: Exec): Env -> Env -> Env :=
      fun g1 g2 =>
        match e with
        | s_one _ => g2
        | s_cons g e' =>
          if Env_dec g1 g then
            if Env_dec g2 (RUN g1) then
              s_hd e'
            else
              g2
          else g2
        end.

    Lemma build_select_compat: Proper fequiv build_select.
    Proof.
      intros [g1 | g1 e1] [g2 | g2 e2] ee;ssplit ee;auto;
        intros gg1 gg2 egg ggg1 ggg2 eggg;simpl;auto.
      destruct ee as [eg ee].
      destruct (Env_dec gg1 g1);destruct (Env_dec gg2 g2);[
        destruct (Env_dec ggg1 (RUN gg1));destruct (Env_dec ggg2 (RUN gg2)) |..];
      intros n1 n2 en;simpl.
      - eapply (hd_compat ee);auto.
      - contradict n.
        rewrite <- eggg,<- egg.
        apply p1;try apply RUN_compat;proper_.
      - contradict n.
        rewrite eggg,egg.
        apply p1;try apply RUN_compat;proper_.
      - apply eggg,en.
      - contradict n.
        rewrite <- egg,<- eg.
        apply p;proper_.
      - contradict n.
        rewrite egg,eg.
        apply p;proper_.
      - apply eggg,en.
    Qed.

    Lemma build_select_sound_progress:
      forall e (is_e: is_exec e), sound_progress (build_select e).
    Proof.
      intros e is_e.
      intros g1 g2 pg1 pg2; split; unfold build_select.

      { (* sound *)
        intros n.
        destruct e as [g|g e].
        {
          left;auto.
        }
        destruct ( Env_dec g1 g).
        shelve.
        { left;auto. }
        Unshelve.
        destruct (Env_dec g2 (RUN g1)).
        shelve.
        { left;auto. }
        Unshelve.
        assert (pg := is_exec_s_hd_compat is_e).
        specialize (p pg1 pg).
        specialize (p0 pg2 (RUN_compat pg1)).
        assert (hstep:=  is_exec_inv_1 is_e).
        destruct hstep as [ _ _ _ h].
        specialize (h n);destruct h as [h|h].
        left;rewrite h,(p0 n n (Equivalence_Reflexive _)).
        apply RUN_compat;auto;symmetry;apply p.
        right;rewrite h;symmetry;apply p;reflexivity.
      }
      {
        destruct e as [g|g e].
        auto.
        destruct ( Env_dec g1 g);auto.
        destruct (Env_dec g2 (RUN g1));auto.
        assert (pg := is_exec_s_hd_compat is_e).
        specialize (p pg1 pg).
        specialize (p0 pg2 (RUN_compat pg1)).
        assert (hstep:=  is_exec_inv_1 is_e).
        destruct hstep as [ _ _ h _].
        intros hh;contradict h.
        auto.
      }
    Qed.

    (** The daemon built from an execution simply uses
     <<build_select>>.*)
    CoFixpoint build_daemon (e: Exec): Daemon :=
      match e with
      | s_one _ => Daemon_Step (build_select e) (fun _ _ => build_daemon e)
      | s_cons _ e' => Daemon_Step (build_select e) (fun _ _ => build_daemon e')
      end.

    Local Definition tmpp e :=
      match e with
      | s_one _ => Daemon_Step (build_select e)  (fun _ _ => build_daemon e)
      | s_cons _ e' => Daemon_Step (build_select e)  (fun _ _ => build_daemon e')
      end.

    Lemma build_daemon_unfold:
      forall e,
        build_daemon e =
        match e with
        | s_one _ => Daemon_Step (build_select e) (fun _ _ => build_daemon e)
        | s_cons _ e' => Daemon_Step (build_select e) (fun _ _ => build_daemon e')
        end.
    Proof.
      assert (forall e, build_daemon e = tmpp e).
      intros e.
      rewrite daemon_decomposition_lemma with (build_daemon e).
      destruct e; simpl; reflexivity.
      intros e; rewrite H at 1; unfold tmpp.
      reflexivity.
    Qed.

    Lemma build_daemon_compat: Proper pequiv build_daemon.
      hnf.
      cofix build_daemon_compat.
      intros x y hxy;rewrite (build_daemon_unfold x),(build_daemon_unfold y).
      destruct x as [gx |gx x]; destruct y as [gy |gy y]; ssplit hxy;
        apply eqDaemon_Step;simpl;repeat intro;
          apply build_select_compat || apply build_daemon_compat;(left || right ||idtac);intuition.
    Qed.

    (** The daemon built from <<build_daemon>> is actually a daemon. *)
    Lemma build_daemon_is_Daemon:
      forall e, is_exec e -> is_Daemon (build_daemon e).
    Proof.
      cofix h.
      intros [g | g e] is_e; rewrite build_daemon_unfold; simpl.
      - constructor.
        apply build_select_compat, is_exec_proper; auto.
        apply build_select_sound_progress; auto.
        repeat intro.
        assert (pg := is_exec_s_hd_compat is_e).
        apply build_daemon_compat;left;auto.
        auto.
      - constructor.
        apply build_select_compat, is_exec_proper; auto.
        apply build_select_sound_progress; auto.
        repeat intro;auto.
        assert (pg := is_exec_proper is_e).
        ssplit pg;destruct pg.
        apply build_daemon_compat;auto.
        repeat intro;apply h; apply is_exec_inv_2 with g; auto.
    Qed.

    (** Completeness: the proof uses <<build_daemon>> and shows
     that the daemon obtained from an execution <<e>> computes an
     execution with <<build_exec>> which corresponds to the initial
     execution <<e>>.
     *)
    Theorem computational_completeness:
      forall (init: Env) (init_compat: Proper pequiv init) (e: Exec),
        is_exec e -> pequiv (s_hd e) init ->
        exists daemon, is_Daemon daemon /\
                       e =~= build_exec init daemon.
    Proof.
      intros init init_compat e is_e fst.
      exists (build_daemon e).
      split.
      apply (build_daemon_is_Daemon is_e).

      revert init init_compat e is_e fst; cofix H; intros.
      rewrite build_daemon_unfold with e.
      rewrite build_exec_unfold.

      destruct e as [g | g e]; inversion is_e; subst; simpl in *;try tauto.
      (* s_one g *)
      {
        case (terminal_dec init).
        * intros _; left; assumption.
        * intros F; contradict F.
          rewrite <- (terminal_compat fst); assumption.
      }
      {
        case (terminal_dec init); intros term.
        specialize (term init_compat).
        destruct term as [_ term].
        specialize (term (s_hd e)).
        contradict term.
        assert (X := hd_compat (is_exec_proper H3)).
        rewrite <- (Step_compat X fst);
          assumption; clear X.

        (* s_cons g e *)
        right.
        assumption.
        apply H;clear H;auto.
        apply (build_select_compat (is_exec_proper is_e) );auto.
        apply RUN_compat;auto.

        case (Env_dec init g);intro hh.
        case (Env_dec (RUN init) (RUN init));intro hhh.
        apply (hd_compat (is_exec_proper (is_exec_inv_2 is_e))).
        contradict hhh;apply RUN_compat;auto.
        symmetry in fst;elim (hh fst).
      }
    Qed.

  End Daemon_Exec.

(* * Adversarial Daemons:

   The type Daemon is simply the result of the daemon, as may be
   observed.

   Is would be interesting to be able to build an adversarial daemon
   which could be adaptative, next step would be based on the
   beginning of the execution.

 *)

End FunModel.

Close Scope signature_scope.
Unset Implicit Arguments.
