(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import Setoid.
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import RelationPairs.
From Coq Require Import Lt.
From Coq Require Import NArith.
From Coq Require Import Lia.

(** **** Local Imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import Algorithm.
From Padec Require Import NatUtils.
From Padec Require Import ListUtils.

(* end show *)
Open Scope signature_scope.
Open Scope nat_scope.
Set Implicit Arguments.

(** * Build a network given as adjacency lists; nodes are numbers from
0 up to N

   Connections are given as a list of lists of option ltk: - channel
   is rank of peers in peer list - reply_to: - find peer: if None
   return 0 else - find self in peer's peers: return index if found 0
   if not found *)
Section make_net.

  Variable (matrix: list (list N)). (* adjascence list *)

  Definition K := match length matrix with O => 1%N | n => N.of_nat n end.
  (* number of nodes *)

  Lemma Kpos: (0<K)%N.
  Proof.
    unfold K.
    destruct (length matrix) as [|M];
      reflexivity.
  Qed.

  Definition node_matrix := 
    map (map_filter (ltK_inj K)) matrix.

  Fixpoint channel_error n (l: list (ltK K)) i:= 
    match l with 
      nil => None
    | n'::q => if (N.eq_dec n (proj1_sig n')) then Some i 
               else channel_error n q (S i)
    end.

  Lemma  channel_error_prop: forall n l i,
      channel_error (proj1_sig n) (l: list (ltK K)) 0 == Some i ->
      nth_error l i == Some n.
  Proof.
    intros n l i;replace i with (i+0) at 1 by lia.
    generalize 0%nat as k.
    revert n i; induction l  as [ | x q hq];intros n i k h.
    { destruct i;simpl in *;tauto. }
    { simpl in h;destruct i;simpl in h;
        destruct (N.eq_dec (proj1_sig n) (proj1_sig x)) as [yes|no];
        simpl;simpl in h.
      {
        congruence.
      }
      {
        elimtype False;clear hq .
        assert (hh:k<S k) by lia.
        revert hh h;generalize (S k) as kk.
        induction q as [|xx q hq];simpl;intros kk h;auto.
        destruct ( N.eq_dec (proj1_sig n) (proj1_sig xx));simpl in h.
        simpl;intro hh; lia.
        simpl;apply hq;lia.
      }
      {
        elimtype False;lia.
      }
      {
        apply hq with (S k).
        rewrite PeanoNat.Nat.add_succ_r;apply h.
      }
    }
  Qed.

  Definition row (n: N) := 
    nth (N.to_nat n) node_matrix nil.

  Definition peer_ (n: N) (c:nat) := 
    let l := row n in
    match nth_error l c with 
      Some n' => 
      if (N.eq_dec n (proj1_sig n')) then None else
        match channel_error (proj1_sig n') l 0 with 
          Some c' =>if (PeanoNat.Nat.eq_dec c c') then Some n' else None
        | None => None end
    | None => None
    end.

  Lemma peer_prop: forall n c n',
    peer_ n c == Some n' ->
      nth_error (row n) c == Some n' /\
      channel_error (proj1_sig n') (row n) 0 == Some c /\
      ~ n == (proj1_sig n').
  Proof.
    intros n c n'.
    unfold peer_.
    destruct (nth_error (row n) c) as [n1|];simpl;try tauto.
    destruct (N.eq_dec n (proj1_sig n1)) as [yes|no];simpl;try tauto.
    case_eqs (equiv (A:=option nat)) (channel_error (proj1_sig n1) (row n) 0);
      simpl;try tauto.
    intros nn enn.
    destruct (PeanoNat.Nat.eq_dec c nn);simpl;try tauto.
    subst;split;auto.
    rewrite <- enn.
    destruct n1;destruct n';hnf in H;simpl in H;simpl;subst;split.
    reflexivity.
    assumption.
  Qed.

  Definition link_ (n: N) (n': N) :=
    if  N.eq_dec n' n then None else channel_error n' (row n) 0.
  
  Definition peers_ n l :=
    map_filter
      (fun n' => if N.eq_dec (proj1_sig n') n then None else
                   channel_error (proj1_sig n') l 0) l.

  Definition reply_to_ n c:=
    match peer_ n c with
      None => 0 
    | Some n' => 
      match channel_error n (row (proj1_sig n')) 0 with
        None => 0
      | Some c' => c' 
      end end.

  Instance Chans: Channels := {| Channel := nat; Channel_dec := PeanoNat.Nat.eq_dec |}.

  Program Definition Net: Network (Chans:=Chans) :=
    {|
      Node := ltK K;
      Node_dec := ltK_dec (K := K);
      peer:=peer_;
      link:=link_;
      all_nodes:=Kenum K;
      all_nodes_prop := ltK_in_Kenum (K := K)
    |}.
  (* begin hide *)
  Obligation 1. (* link_peer_compat *)
  (*  destruct n1 as [n1 ltn1];destruct n2 as [n2 ltn2];simpl. *)
  split.
  {
    unfold link_.
    destruct N.eq_dec;simpl;try tauto.
    intro h_;assert (h:=channel_error_prop _ _ _ h_).
    unfold peer_.
    destruct nth_error as [n2'|].
    hnf in h;rewrite h.
    destruct N.eq_dec;try congruence.
    destruct channel_error as [n3|].
    destruct PeanoNat.Nat.eq_dec.
    apply h.
    contradict n4;auto.
    elim h_.
    elim h.
  }
  {
    intro h;apply peer_prop in h.
    destruct h as [h1 [h2 h3 ]].
    unfold link_.
    destruct N.eq_dec.
    contradict h3;auto.
    hnf;auto.
    auto.
  }
  Qed.

  (* Obligation of Net: Proper peer_ *)
  Obligation 3.
  intros [n1 h1] [n2 h2] en;hnf in en;simpl in en.
  revert h1;rewrite en.
  intros;reflexivity.
  Qed.

  (*Obligation of Net: proper link_ *)
  Obligation 2.
  intros [a ha] [b hb] ab c d cd; simpl.
  hnf in ab; simpl in ab.
  rewrite ab, cd.
  reflexivity.
  Qed.
  (* end hide *)
End make_net.

(* Check  (Net ((1::nil)::(2::nil)::(0::nil)::nil))%N. *)

Close Scope signature_scope.
Close Scope nat_scope.
Unset Implicit Arguments.
