(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** **** Global imports **)
From Coq Require Import Arith.
From Coq Require Import List.
From Coq Require Import Relations.
From Coq Require Import Wellfounded.
From Coq Require Import Setoid.
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Basics.
From Coq Require Import RelationPairs.
From Coq Require Import Lia.

(** **** Local imports **)
From CoLoR Require MultisetNat.
(** *)
From Padec Require Import SetoidUtil.
From Padec Require Import ListUtils.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import WellfoundedUtil.

(* end show *)
Open Scope signature_scope.
Set Implicit Arguments.

(** * Core Results for Proving Termination

   We show a sufficient condition based on some potential
   function. Each node in a given configuration provides a local
   potential. Usual termination proofs are based on some global
   potential based on the local ones. For instance, potentials can be
   natural numbers, the global potential can be the sum of the local
   potentials and the argument for termination might be based on the
   fact that the global potential strictly decreases each step of the
   algorithm.

   Instead, we build the global potential as the multiset containing
   the local potential of every nodes. Our sufficient condition for
   termination uses the Dershowitz-Manna order on finite multi-sets
   and defines two criteria:

   - _Local criterion_ expresses that given a step, for any node, if
     its local potential has not decreased, there exists another node
     (typically one of its neighbors) for which the potential actually
     decreased.

    - _Global criterion_ exhibits, at any step, a node whose potential
      decreased.

   The result can be used for any [Step]s of the model, or for
   subrelations on it. We allow conditionned transitions such that:

      1) the algorithm is initialized in some stable set of
      configurations.  - in this set, every configuations should be
      proper (mandatory) - there might be no more constraints (proper
      self-stabilization) - but in case of composed algorithms, it can
      also express the fact that a first algorithm has stabilized
      (into this stable set) and that a second algorithm composed to
      the first one stabilizes provided that the first one stabilized
      in this set.  This condition might be necessary to be able to
      define potentials.

      2) we may need to study a particular classes of transitions
        (characterized by the firing condition).  This might be useful
        to cut the termination proof in different classes of
        converging transitions.

 **)

Section P_Q_Termination.

  Context {Chans: Channels}
          {Net: Network}
          {Algo: Algorithm}.

  Section Multiset_Criterion.

    (* ** Safe Configurations *)
    Variable (safe: Env -> Prop).

    Notation safeEnv := (sig safe).

    Notation getEnv := (@proj1_sig Env safe).
    Notation getSafe := (@proj2_sig Env safe).

    (** Assumptions on [safe] configurations in which the algorithm
        is initialized.
        - should be stable against [valid_diff]
        - should be compatible
     **)
    Variable (safe_compat: Proper fequiv safe)
             (stable_safe: forall (g g': Env), safe g -> Step g' g -> safe g').

    (* ** Conditionned Steps *)

    (** Safe Steps (from [safeEnv] to [safeEnv])
        NB: transition from g to g' *)
    Definition safeStep (g' g: safeEnv): Prop := Step (getEnv g') (getEnv g).

    Global Instance safeStep_compat: Proper fequiv safeStep.
    Proof.
      intros [g sg] [g' sg'] eg [gg sgg] [gg' sgg'] egg.
      simpl in eg, egg; unfold eqSig in eg, egg; simpl in eg, egg.
      unfold safeStep; simpl; apply Step_compat; auto.
    Qed.
    
    (* Conditionned Steps using [QTrans] predicate *)
    Variable (QTrans: safeEnv -> safeEnv -> Prop).
    
    Definition safeQStep (g' g: safeEnv): Prop := safeStep g' g /\ QTrans g' g.

    Instance safeQStep_compat: Proper fequiv QTrans -> Proper fequiv safeQStep.
    Proof.
      intros hQ g g' eg gg gg' egg; unfold safeQStep.
      apply and_compat; [ apply safeStep_compat | apply hQ ]; auto.
    Qed.

    (** ** [Mnat] is the type of potential values endowed with
        - equality (decidable equivalence): [eqP]
        - strict order: [gtP], compatible w.r.t. [eqP]

        It stands for natural numbers in multisets.
     **)
    Notation Mnat (*: Type*) := MultisetNat.MSetCore.Sid.Eq.A.

    Notation eqP (*: relation Mnat*) := MultisetNat.MSetCore.Sid.Eq.eqA.
    Notation eqP_dec (*: Decider eqP*) := MultisetNat.MSetCore.Sid.eqA_dec.
    Instance eqP_equiv: Equivalence eqP :=
      MultisetNat.MSetCore.Sid.Eq.eqA_Equivalence.

    Notation gtP (*: relation Mnat*) := gt.

    Instance gtP_so: StrictOrder gtP.
    (* begin hide *)
    Proof.
      split; cbv; intros; lia.
    Qed.
    (* end hide *)

    Instance gtP_compat: Proper (eqP ==> eqP ==> iff) gtP.
    (* begin hide *)
    Proof.
      cbv; intros; split; intros; lia.
    Qed.
    (* end hide *)

    Lemma classical_gtP: forall x y, gtP x y \/ ~ gtP x y.
    Proof.
      intros x y; destruct (gt_dec x y); auto.
    Qed.

    (** ** [pot] is the potential function for local nodes
        It should be compatible.
        It only needs to be defined on the set of safe configurations.
     **)
    Variable (pot: safeEnv -> Node -> Mnat).
    Context {pot_compat: Proper fequiv pot}.

    (** ** Global Potential **)
    Definition Pot (g: safeEnv): list nat := List.map (pot g) all_nodes.

    (** ** Local Criterion (see above) **)
    Hypothesis local_crit:
      forall (g g': safeEnv),
        safeQStep g' g ->
        forall (n: Node),
          gtP (pot g' n) (pot g n) ->
          exists (n': Node),
            ~ (pot g n') == (pot g' n') /\ gtP (pot g n') (pot g' n).

    (** ** Global Criterion (see above) **)
    Hypothesis global_crit:
      forall (g g': safeEnv),
        safeQStep g' g -> exists (n: Node), ~ (pot g' n) == (pot g n).

    (** ** Core Result: safeQStep decreases Pot *)

    (** *** Tools on multisets *)
    Notation Mset := (MultisetNat.MSetCore.Multiset).
    Notation gtM := (@MultisetNat.MSetOrd.MultisetGT gt).
    Notation ltM := (@MultisetNat.MSetOrd.MultisetLT gt).
    Notation eqM := MultisetNat.MSetCore.meq.
    Notation Mempty := MultisetNat.MSetCore.empty.
    Notation Munion := MultisetNat.MSetCore.union.
    Notation Msingleton := MultisetNat.MSetCore.singleton.
    Notation MIn := MultisetNat.MSetOrd.MSet.member.

    Lemma cons_union:
      forall (x: nat) (q: Mset), eqM (x::q) (Munion (Msingleton x) q).
    Proof.
      intros x q; reflexivity.
    Qed.

    Lemma nil_empty: @eqM nil Mempty.
    Proof.
      reflexivity.
    Qed.

    Lemma eq_meq:
      forall l1 l2 : Mset, l1 == l2 -> eqM l2 l1.
    Proof.
      simpl.
      induction l1; induction l2; simpl.
      intros _; reflexivity.
      intros F; inversion F.
      intros F; inversion F.
      intros X; inversion X; subst.
      rewrite cons_union.
      rewrite cons_union.
      apply MultisetNat.MSetOrd.MSet.union_morph.
      reflexivity.
      apply IHl1.
      assumption.
    Qed.

    Lemma MIn_empty: forall x, MIn x Mempty <-> False.
    Proof.
      split; [|tauto].
      intro h; apply MultisetNat.MSetOrd.MSet.not_empty with Mempty x.
      assumption.
      reflexivity.
    Qed.

    Lemma MIn_union:
      forall x P Q, MIn x (Munion P Q) <-> MIn x P \/ MIn x Q.
    Proof.
      intros x P Q.
      unfold MIn. rewrite MultisetNat.MSetCore.union_mult.
      intuition.
      assert(forall a b, (a + b > 0)%nat
                         -> (a > 0)%nat \/ (b > 0)%nat).
      intros [|a] [|b]; intuition.
      apply H0. exact H.
    Qed.

    Lemma MIn_InA: forall x MS, MIn x MS <-> InA equiv x MS.
    Proof.
      induction MS as [|y MS hrec].
      split; intro h.
      rewrite nil_empty in h.
      rewrite MIn_empty in h.
      elim h.
      cbv in h; rewrite InA_nil in h.
      elim h.
      unfold Mnat, MultisetNat.MSetCore.Sid.Eq.A,MultisetNat.Nat.A.
      rewrite InA_cons.
      rewrite cons_union.
      rewrite MIn_union.
      apply Morphisms_Prop.or_iff_morphism.
      split.
      apply MultisetNat.MSetOrd.MSet.member_singleton.
      intro h; rewrite h.
      apply MultisetNat.MSetOrd.MSet.singleton_member.
      apply hrec.
    Qed.

    (** *** Core Theorem
     safe_incl := (forall g g', safeQStep g' g -> ltM (Pot g') (Pot g)
     *)
    Lemma Step_decreases_pot:
      inclusion safeQStep (ltM @@ Pot)%signature.
    Proof.
      intros g' g HsafeTrans.
      specialize (local_crit HsafeTrans).
      specialize (global_crit HsafeTrans).

      destruct HsafeTrans as [step_g HQ].
      dup step_g step.
      destruct step as [pg pg' dif r]. 
      hnf.
      pose (test n :=
              if eqP_dec (pot g' n) (pot g n)
              then true else false).
      pose (test_not n :=
              if eqP_dec (pot g' n) (pot g n)
              then false else true).
      pose (X := List.map (pot g)
                          (List.filter test_not all_nodes)).
      pose (Y := List.map (pot g')
                          (List.filter test_not all_nodes)).
      pose (Z := List.map (pot g) (List.filter test all_nodes)).
      assert (ptest_not: Proper fequiv test_not).
      {
        intros n1 n2 en12; unfold test_not.
        destruct (eqP_dec (pot g' n1) (pot g n1))
          as [e1|ne1];
          destruct (eqP_dec (pot g' n2) (pot g n2))
          as [e2|ne2];try reflexivity.

        rewrite (pot_compat pg n1 n2 en12) in e1.
        rewrite (pot_compat pg' n1 n2 en12) in e1.
        elim ne2.
        exact e1.

        rewrite <- (pot_compat pg n1 n2 en12) in e2.
        rewrite <- (pot_compat pg' n1 n2 en12) in e2.
        elim ne1.
        exact e2.
      }
      split with X Y Z;
        unfold Pot in *.

      (* X not empty *)

      intro h.
      destruct global_crit as [n hn].
      apply hn; clear hn.
      generalize (all_nodes_prop n); revert h n.
      unfold X; clear X Y Z.
      induction all_nodes as [|n0 l hl]; simpl; intros hX n hn.
      rewrite InA_nil in hn.
      elim hn.
      rewrite InA_cons in hn.
      unfold test_not in hX at 1.
      destruct (eqP_dec (pot g' n0) (pot g n0)) as [e|ne].
      destruct hn as [hn|hn].
      rewrite <- (pot_compat pg n n0 hn) in e.
      rewrite <- (pot_compat pg' n n0 hn) in e.
      exact e.

      specialize (r n).
      apply hl; auto.
      simpl in hX.
      unfold Mnat, MultisetNat.MSetCore.Sid.Eq.A, MultisetNat.Nat.A
        in *.

      (*assert (E := MultisetNat.MSetOrd.MSet.MultisetSetoid).*)
      assert (C := (cons_union (pot g n0) (map (pot g) (filter test_not l)))).
      symmetry in hX.
      assert (T := MultisetNat.MSetOrd.MSet.meq_trans hX C).
      (*      rewrite cons_union in hX.*)
      elimtype False.
      eapply MultisetNat.MSetOrd.MSet.union_notempty.
      eapply MultisetNat.MSetTheory.singleton_notempty.
      symmetry; apply hX.

      (* union Z X *)

      unfold Z, X.
      clear X Y Z.
      induction all_nodes as [|n l hl]; simpl.
      reflexivity.
      unfold test at 1, test_not at 1.
      destruct (eqP_dec (pot g' n) (pot g n))
        as [e|ne];
        simpl;
        unfold Mnat, MultisetNat.MSetCore.Sid.Eq.A, MultisetNat.Nat.A;
        do 2 rewrite cons_union.
      apply MultisetNat.MSetOrd.MSet.union_morph.
      reflexivity.
      apply hl.
      unfold Munion.
      rewrite MultisetNat.MSetOrd.MSet.union_assoc.
      rewrite (MultisetNat.MSetOrd.MSet.union_comm
                 (map (B := nat) (pot g) (filter test l))).
      rewrite <- MultisetNat.MSetOrd.MSet.union_assoc.
      apply MultisetNat.MSetOrd.MSet.union_morph.
      reflexivity.
      apply hl.

      (* union Z Y *)

      unfold Z, Y.
      clear X Y Z.
      induction all_nodes as [|n l hl]; simpl.
      reflexivity.
      unfold test at 1,test_not at 1.
      destruct (eqP_dec (pot g' n) (pot g n))
        as [e|ne];
        simpl;
        unfold Mnat, MultisetNat.MSetCore.Sid.Eq.A, MultisetNat.Nat.A;
        do 2 rewrite cons_union.
      apply MultisetNat.MSetOrd.MSet.union_morph.
      apply MultisetNat.MSetOrd.MSet.singleton_morph.
      apply e.
      apply hl.
      unfold Munion.
      rewrite MultisetNat.MSetOrd.MSet.union_assoc.
      rewrite (MultisetNat.MSetOrd.MSet.union_comm
                 (map (B := nat) (pot g) (filter test l))).
      rewrite <- MultisetNat.MSetOrd.MSet.union_assoc.
      apply MultisetNat.MSetOrd.MSet.union_morph.
      reflexivity.
      apply hl.

      (* forall y in Y exists x in X st x > y *)

      unfold X, Y.
      intros y hy.
      fold MIn in hy.
      rewrite MIn_InA in hy.
      rewrite InA_map_iff in hy.
      2: apply pot_compat; apply pg'.
      destruct hy as [n [en hn]].
      rewrite <- en.
      clear y en.
      destruct (classical_gtP (pot g n) (pot g' n))
        as [ok|ko].

      (* if we are smaller -> ok *)

      split with (pot g n); [|assumption].
      fold MIn.
      rewrite MIn_InA.
      rewrite InA_map_iff.
      2: apply pot_compat;apply pg.
      split with n; split.
      reflexivity.
      assumption.

      (* else we are smaller than someone else *)

      destruct (@local_crit n) as [n' [hne hlt]].
      rewrite filter_InA in hn; try assumption.
      specialize (proj2 hn);clear hn;intro hn.
      unfold test_not in hn.
      destruct (eqP_dec (pot g' n) (pot g n)) as [e|ne].
      clear - hn; discriminate hn.
      (**)
      assert (tmp := nat_total_order (pot g' n) (pot g n)).
      assert (tmp': (pot g' n) <> (pot g n)). congruence.
      specialize (tmp tmp').
      destruct tmp.
      contradict ko. apply H.
      apply H.
      split with (pot g n'); [|assumption].
      rewrite MIn_InA.
      rewrite InA_map_iff.
      2: apply pot_compat; apply pg.
      split with n'; split.
      reflexivity.
      rewrite filter_InA; try assumption.
      split.
      apply all_nodes_prop.
      unfold test_not.
      destruct (eqP_dec (pot g' n') (pot g n')) as [e|ne].
      symmetry in e.
      elim (hne e).
      reflexivity.
    Qed.

    (** ** The main theorem: applied to prove that algorithm terminates!
        Conditionned transitions are well founded.
     **)
    Lemma Wf_safeQAlgo_Multiset: well_founded safeQStep.
    Proof.
      apply Inclusion.wf_incl with (ltM @@ Pot)%signature.
      apply Step_decreases_pot.
      apply (Inverse_Image.wf_inverse_image _ _ ltM Pot).
      apply MultisetNat.MSetOrd.mOrd_wf.
      cbv; clear; intros; congruence.
      apply lt_wf.
    Qed.

    Section No_QTrans.

      (** Once proven that safe transitions are well founded, it can be
    easily deduced that any execution terminates provided that it
    starts in a safe configuration.
       *)
      Lemma Acc_Algo_Multiset:
        well_founded safeStep -> forall g: Env, safe g -> Acc Step g.
      Proof.
        intros H g sg.
        (*        assert (h := Wf_safeQAlgo_Multiset (exist _ _ sg)).*)
        specialize (H (exist sg)).
        change (Acc Step (getEnv (exist sg))).
        clear - safe_compat stable_safe H.
        induction H as [[g_ sg_] _ hrech] using Acc_ind_dep.
        constructor.
        intros y t.
        assert (h: safe y).
        simpl in t.
        apply (stable_safe sg_ t).
        specialize (hrech (exist h)).
        apply hrech.
        unfold safeStep.
        apply t.
      Qed.

      Lemma NoQTrans_Wf_Multiset:
        well_founded safeQStep -> (forall x x', QTrans x x') -> well_founded safeStep.
      Proof.
        intros wf hq.
        apply Inclusion.wf_incl with safeQStep.
        intros x x' hs.
        unfold safeQStep.
        split; [assumption | apply hq].
        assumption.
      Qed.

    End No_QTrans.

  End Multiset_Criterion.
  
  (** * Lexicographic orders on safeEnv *)
  Section Lexico.

    Variable (safe: Env -> Prop).
    Notation safeEnv := (sig safe).

    Section Lexico_core.

      (** Trans1 has priority on Trans 2 *)
      Variable QTrans1: relation safeEnv.
      Variable QTrans2: relation safeEnv.
      Notation step1 := (safeQStep QTrans1).
      Notation step2 := (safeQStep QTrans2).

      (** Relations abstracting step1 *)
      Variable (S1: relation safeEnv)
               (HS1: inclusion step1 S1) (WF1: well_founded S1)
               (E1: relation safeEnv) (TE1: Transitive E1)
               (CE1: forall g' g, (exists gg, S1 g' gg /\ E1 gg g)
                                  -> S1 g' g).     

      Variable (WF2: well_founded step2).
      
      Variable (Disjoint: forall e e',
                   step2 e' e -> E1 e' e \/ S1 e' e).
      
      Lemma Wf_lexico:
        well_founded (safeQStep (union QTrans1 QTrans2)).
      Proof.
        apply Inclusion.wf_incl with (union step1 step2).
        { intros g' g; unfold union, safeQStep; intuition. }
        apply Inclusion.wf_incl with
            (lexico_pairs E1 S1 step2 @@ (fun g => (g,g))).
        { intros g' g [U|U];
            unfold "@@"; apply lexico_pairs_intro.
          - left; auto.
          - destruct (Disjoint U) as [D|D].
            * right; split; auto.
            * left; auto.              
        }
        { apply Inverse_Image.wf_inverse_image
            with (f := (fun g => (g,g))).
          apply lexico_pairs_wf; auto.
        }
      Qed.
      
    End Lexico_core.

    Import MultisetNat.MSetOrd.
    Notation Mnat (*: Type*) := MultisetNat.MSetCore.Sid.Eq.A.
    Notation ltM := (@MultisetNat.MSetOrd.MultisetLT gt).
    Notation eqM := MultisetNat.MSetCore.meq.

    Lemma ltM_wf: well_founded ltM.
    Proof.
      apply MultisetNat.MSetOrd.mOrd_wf.
      cbv; clear; intros; congruence.
      apply lt_wf.
    Qed.

    Lemma ltM_comp_wf:
      forall A (f: A -> MultisetNat.MSetCore.Multiset),
        well_founded (ltM @@ f).
    Proof.
      intros A f.
      apply Inverse_Image.wf_inverse_image with (f := f).
      apply ltM_wf.
    Qed.      
    
    Instance ltM_compat: Proper (eqM ==> eqM ==> iff) ltM.
    Proof.
      intros a b eab x y exy.
      unfold ltM, transp.
      apply MultisetGT_morph; [apply exy|apply eab].
    Qed.

    Lemma ltM_comp_trans:
      forall A (f: A -> MultisetNat.MSetCore.Multiset),
        Transitive (eqM @@ f).
    Proof.
      intros A f.
      apply RelCompFun_Transitive;
        [split| apply Equivalence_Transitive].
    Qed.      

    Lemma ltM_eqM_comp_commut:
      forall A (f: A -> MultisetNat.MSetCore.Multiset),
      forall a b, (exists c, ltM (f a) (f c) /\ eqM (f c) (f b)) ->
                  ltM (f a) (f b).
    Proof.
      intros A f a b [c [lc ec]].
      rewrite <- ec; auto.
    Qed.
    
    (** * Union of two classes of converging transitions
     composed using priorities, i.e. lexicographic order.
     *)
    Section Lexico_MultiSet2.

      Variable (pot1: safeEnv -> Node -> Mnat).
      Notation Pot1 := (@Pot safe pot1).

      (** Trans1 has priority on Trans 2 *)
      Variable QTrans1: relation safeEnv.
      Variable QTrans2: relation safeEnv.
      Notation Trans1 := (safeQStep QTrans1).
      Notation Trans2 := (safeQStep QTrans2).
      Variable (Hterm1: inclusion Trans1 (ltM @@ Pot1)%signature).
      Variable (Hterm2: well_founded Trans2). 
      Variable (Hdisjoint1: forall e e', Trans2 e' e ->  eqM (Pot1 e') (Pot1 e) \/ ltM (Pot1 e') (Pot1 e)).

      Lemma union_lex_wf2:
        well_founded (safeQStep (union QTrans1 QTrans2)).
      Proof.
        eapply Wf_lexico with (E1 := eqM@@Pot1) (S1 := ltM@@Pot1);
          unfold "@@" in *; auto.
        - apply ltM_comp_wf. 
        - apply ltM_comp_trans.
        - apply ltM_eqM_comp_commut.
      Qed.

    End Lexico_MultiSet2.

    (** * Union of three classes of converging transitions
     composed using priorities, i.e. lexicographic order.
     *)
    Section Lexico_MultiSet3.

      Variable (pot1 pot2: safeEnv -> Node -> Mnat).
      Notation Pot1 := (@Pot safe pot1).
      Notation Pot2 := (@Pot safe pot2).

      Variable QTrans1 QTrans2 QTrans3: relation safeEnv.
      Notation Trans1 := (safeQStep QTrans1).
      Notation Trans2 := (safeQStep QTrans2).
      Notation Trans3 := (safeQStep QTrans3).
      Variable (Hterm1: inclusion Trans1 (ltM @@ Pot1)%signature).
      Variable (Hterm2: inclusion Trans2 (ltM @@ Pot2)%signature).
      Variable (Hterm3: well_founded Trans3).
      Variable (Hdisjoint1: forall e e', Trans2 e' e \/ Trans3 e' e -> eqM (Pot1 e') (Pot1 e) \/ ltM (Pot1 e') (Pot1 e)).
      Variable (Hdisjoint2: forall e e', Trans3 e' e ->  eqM (Pot2 e') (Pot2 e) \/ ltM (Pot2 e') (Pot2 e)).

      Lemma union_lex_wf3:
        well_founded (safeQStep (union QTrans1
                                       (union QTrans2 QTrans3))).
      Proof.
        apply union_lex_wf2 with pot1; auto.
        apply union_lex_wf2 with pot2; auto.
        intros g g' st; apply Hdisjoint1.
        destruct st as [st [Qst|Qst]]; [left|right]; split; auto.
      Qed.        

    End Lexico_MultiSet3.

    (** * Union of four classes of converging transitions
     composed using priorities, i.e. lexicographic order.
     *)
    Section Lexico_MultiSet4.

      Variable (pot1 pot2 pot3: safeEnv -> Node -> Mnat).
      Notation Pot1 := (@Pot safe pot1).
      Notation Pot2 := (@Pot safe pot2).
      Notation Pot3 := (@Pot safe pot3).

      Variable (QTrans1 QTrans2 QTrans3 QTrans4: relation safeEnv).
      Notation Trans1 := (safeQStep QTrans1).
      Notation Trans2 := (safeQStep QTrans2).
      Notation Trans3 := (safeQStep QTrans3).
      Notation Trans4 := (safeQStep QTrans4).
      
      Variable (Hterm1: inclusion Trans1 (ltM @@ Pot1)%signature).
      Variable (Hterm2: inclusion Trans2 (ltM @@ Pot2)%signature).
      Variable (Hterm3: inclusion Trans3 (ltM @@ Pot3)%signature).
      Variable (Hterm4: well_founded Trans4).

      Variable (Hdisjoint1:
                  forall e e', Trans2 e' e \/ Trans3 e' e \/ Trans4 e' e ->
                               eqM (Pot1 e') (Pot1 e) \/
                               ltM (Pot1 e') (Pot1 e)).
      Variable (Hdisjoint2:
                  forall e e', Trans3 e' e \/ Trans4 e' e ->
                               eqM (Pot2 e') (Pot2 e) \/
                               ltM (Pot2 e') (Pot2 e)).
      Variable (Hdisjoint3:
                  forall e e', Trans4 e' e ->
                               eqM (Pot3 e') (Pot3 e) \/
                               ltM (Pot3 e') (Pot3 e)).

      Lemma union_lex_wf4:
        well_founded
          (safeQStep (union QTrans1 (union QTrans2 (union QTrans3 QTrans4)))).
      Proof.
        apply union_lex_wf3 with pot1 pot2; auto.
        apply union_lex_wf2 with pot3; auto.
        - intros g g' st; apply Hdisjoint1.
          destruct st as [s2 | [ st [Qst3|Qst4]]];
            [left; auto |right; left| right; right]; split; auto.
        - intros g g' st; apply Hdisjoint2.
          destruct st as [st [Qst|Qst]]; [left|right]; split; auto.
      Qed.

    End Lexico_MultiSet4.
  End Lexico.

End P_Q_Termination.


Unset Implicit Arguments.
Close Scope signature_scope.