(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016 - updated (change-run) April 2018
 *)
(* begin show *)

(** **** Global imports **)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Bool.

(** **** Local imports **)
From Padec Require Import SetoidUtil.
From Padec Require Import OptionUtil.
From Padec Require Import ListUtils.
From Padec Require Import ZUtils.
From Padec Require Import All_In.
From Padec Require Import Algorithm.

(* end show *)
Open Scope signature_scope.
Set Implicit Arguments.


(** * Relational Model

  Semantics of the model (Algorithm, Network, Channels) using
  relations.  Mainly, we express the fact that a configuration [g] can
  be followed by another one [g'] in an execution, i.e. that [g] ->
  [g'] is a (valid) step of the execution.

  The relational semantics does not use any notion of daemon
  (scheduler) but consider all possible valid steps of the algorithm.

 *)
Section RelModel.

  Context {Chans: Channels} {Net: Network} {Algo: Algorithm}.

  (** ** Environment
   a.k.a. configuration (global state of the system) *)
  Definition Env := Node -> State.

  (** Equivalence of Environments is (weakly-)decidable *)

  Program Definition Env_dec (g g': Env):
    {Proper pequiv g -> Proper pequiv g' -> g =~= g' } + {~ (g =~= g')}
    (* begin hide *)
    :=
      let test := forallb (fun n => if State_dec (g n) (g' n)
                                    then true else false)
                          all_nodes in
      match Bool.bool_dec test true with
      | left h => left (fun pf => _)
      | right h => right (fun hh => h _)
      end.
  Obligation 1.
  intros n1 n2 en; clear Heq_anonymous.
  rewrite <- en; clear en n2.
  rewrite forallb_forall_InA in h.
  specialize (h _ (all_nodes_prop n1)).
  destruct (State_dec (g n1) (g' n1)); congruence.
  clear n1 h; intros n1 n2 en.
  destruct (State_dec (g n1) (g' n1));
    destruct (State_dec (g n2) (g' n2)); try reflexivity.
  contradict n; rewrite <- en; auto.
  contradict n; rewrite en; auto.
  Qed.
  Obligation 2.
  clear Heq_anonymous h.
  rewrite forallb_forall_InA.
  intros x _; destruct (State_dec (g x) (g' x)); auto.
  intros n1 n2 en.
  destruct (State_dec (g n1) (g' n1));
    destruct (State_dec (g n2) (g' n2)); try reflexivity;
      contradict n; apply hh; reflexivity.
  Qed.

  (** Function which represents the local env of a node into a given
      configuration.

    Precisely, given a node [n] and a configuration [env] and a channel [c],
    <<(local_env env n c)>> returns
      - [None] if [c] is not a channel from [n]
      - and otherwise, <<Some s>> with [s] the state of the neighbor of [n]
        reached via [c], in [env]

    This relation
          - is #<a href="Padec.Model.Algorithm.html##compatibility">#compatible#</a># and
            also
          - provides the same information as neighbourhood_state_rel, above
   *)
  Definition local_env (g: Env) (n: Node) (c: Channel): option State
    := option_map g (peer n c).

  Global Instance local_env_compat: Proper fequiv local_env.
  Proof.
    intros g1 g2 eg n1 n2 en c1 c2 ec; unfold local_env.
    eapply (option_map_compat).
    apply eg.
    apply peer_compat; auto.
  Qed.

  (** ** RUN: running the algorithm (function <<run>>) on straightforward
     parameters
     The function is compatible.
   *)
  Definition RUN (env: Env) (n: Node): State :=
    run (env n) (local_env env n).

  Global Instance RUN_compat: Proper fequiv RUN.
  Proof.
    intros env1 env2 He n1 n2 Hn; unfold RUN.
    apply run_compat.
    apply He; auto.
    apply local_env_compat; auto.
  Qed.

  (** ** Steps of the algorithm

    During a step, several nodes may update their own state at the
    same time.  *)

  (** ** Valid step
    A step is called _valid_ when
         - at least one node actually moves and
         - every update corresponds to the algorithm execution

    This relation
      - is #<a href="Padec.Model.Algorithm.html##compatibility">#compatible#</a>#

    NB: step from g to g'
   *)

  Record Step (g':Env) (g: Env): Prop :=
    {

      (** * these 2 coercions are ambiguous in coq 8.15 *)
      Step_proper_r:> Proper pequiv g;
      (* Step_proper_l:> Proper pequiv g'; *)
      Step_proper_l: Proper pequiv g';
      Step_progress: ~ g' =~= g;
      Step_correct: forall n, g' n == RUN g n \/ g' n == g n
    }.

  Global Instance Step_compat: Proper fequiv Step.
  Proof.
    intros e1 e1' eq1 e2 e2' eq2; split.
    { intros [pg1 pg'1 dif1 r1]; split.
      - proper_.
      - proper_.
      - intros F; contradict dif1; rewrite eq1, eq2; auto.
      - intros n; specialize (r1 n); destruct r1 as [r1|r1].
        left; rewrite <- (eq1 _ _ (Equivalence_Reflexive n));
          rewrite <- (RUN_compat eq2 _ _ (Equivalence_Reflexive n)); auto.
        right; rewrite <- (eq1 _ _ (Equivalence_Reflexive n)),
               <- (eq2 _ _ (Equivalence_Reflexive n)); auto.
    }
    { intros [pg1 pg'1 dif1 r1]; split.
      - proper_.
      - proper_.
      - intros F; contradict dif1; rewrite <- eq1, <- eq2; auto.
      - intros n; specialize (r1 n); destruct r1 as [r1|r1].
        left; rewrite (eq1 _ _ (Equivalence_Reflexive n));
          rewrite (RUN_compat eq2 _ _ (Equivalence_Reflexive n)); auto.
        right; rewrite (eq1 _ _ (Equivalence_Reflexive n)),
               (eq2 _ _ (Equivalence_Reflexive n)); auto.
    }
  Qed.

  Global Instance Step_Proper_r:
    forall (g' g: Env), Step g' g -> Proper fequiv g.
  Proof. intros g' g h; apply h. Qed.

  Global Instance Step_Proper_l:
    forall (g' g: Env), Step g' g -> Proper fequiv g'.
  Proof. intros g' g h; apply h. Qed.

  (** ** Enabled
     A node is said to be enabled in the current configuration when it can
     execute some action.
   *)
  Definition enabled_b (env: Env) (n: Node): bool :=
    if State_dec (RUN env n) (env n) then false else true.

  Global Instance enabled_b_compat: Proper fequiv enabled_b.
  Proof.
    intros g1 g2 eg m1 m2 em; unfold enabled_b.
    destruct (State_dec (RUN g1 m1) (g1 m1));
      destruct (State_dec (RUN g2 m2) (g2 m2)); auto.
    contradict n; rewrite <- (eg _ _ em), <- e;
      apply RUN_compat; symmetry; auto.
    contradict n; rewrite (eg _ _ em), <- e; apply RUN_compat; auto.
  Qed.

    Lemma enabled_b_true:
    forall (g: Env) p, enabled_b g p = true -> ~ RUN g p == g p.
  Proof.
    intros g p hen; unfold enabled_b in hen.
    destruct (State_dec (RUN g p) (g p)); [congruence|auto].
  Qed.

  Lemma enabled_b_false:
    forall (g: Env) p, enabled_b g p = false -> RUN g p == g p.
  Proof.
    intros g p hen; unfold enabled_b in hen.
    destruct (State_dec (RUN g p) (g p)); [auto|congruence].
  Qed.

  Definition enabled (env: Env) (n: Node): Prop := (enabled_b env n = true).

  Global Instance enabled_compat: Proper fequiv enabled.
  Proof.
    intros g1 g2 eg m1 m2 em; unfold enabled.
    rewrite enabled_b_compat; [reflexivity| |]; auto.
  Qed.

  Lemma enabled_prop:
    forall (env: Env) (n: Node), enabled env n <-> ~ RUN env n == env n.
  Proof.
    intros env n; unfold enabled, enabled_b.
    case (State_dec (RUN env n) (env n)); intros.
    split; intros F; contradict F; auto.
    split; auto.
  Qed.

  (** If a node is enabled then the system may performe a step. *)
  Lemma enabled_Step:
    forall (env: Env) (n: Node), Proper pequiv env ->
                                 enabled env n -> exists env', Step env' env.
  Proof.
    intros env n env_compat H.
    exists (RUN env).
    split; auto.
    apply RUN_compat; auto.
    apply enabled_prop in H.
    contradict H; apply H; auto.
  Qed.

  (** [has_moved_b] tests if a node has moved between two environments *)
  Definition has_moved_b (g' g: Env) (p: Node): bool :=
    if State_dec (g' p) (g p) then false else true.

  Global Instance has_moved_b_compat: Proper fequiv has_moved_b.
  Proof.
    intros g'1 g'2 eg' g1 g2 eg m1 m2 em; unfold has_moved_b.
    destruct (State_dec (g'1 m1) (g1 m1));
      destruct (State_dec (g'2 m2) (g2 m2)); auto.
    contradict n; rewrite <- (eg _ _ em), <- (eg' _ _ em); auto.
    contradict n; rewrite (eg _ _ em), (eg' _ _ em); auto.
  Qed.

  Lemma has_moved_progress:
    forall g' g p, has_moved_b g' g p = true <-> ~ g' p == g p.
  Proof.
    intros g' g p; unfold has_moved_b.
    destruct (State_dec (g' p) (g p)); split; intros; auto with bool.
  Qed.

  Lemma has_moved_false:
    forall g' g p, has_moved_b g' g p = false <-> g' p == g p.
  Proof.
    intros g' g p; unfold has_moved_b.
    destruct (State_dec (g' p) (g p)); split; intros; auto.
    contradict H; auto with bool.
    tauto.
  Qed.

  Lemma eqE_has_moved_b_false:
    forall g g', g =~= g' -> forall n, has_moved_b g g' n = false.
  Proof.
    intros g g' h n; specialize (h n n (Equivalence_Reflexive n)).
    case_eq (has_moved_b g g' n); [intro etrue | auto].
    apply has_moved_progress in etrue; tauto.
  Qed.

  Lemma has_moved_true:
    forall g' g p, Step g' g -> has_moved_b g' g p = true ->
                   g' p == RUN g p.
  Proof.
    intros g' g p [pg pg' dif r]; unfold has_moved_b.
    destruct (State_dec (g' p) (g p)); intros; [congruence|].
    specialize (r p); destruct r as [r|r]; auto; tauto.
  Qed.

  Lemma Step_one_has_moved_true:
    forall g g', Step g' g -> exists n, has_moved_b g' g n = true.
  Proof.
    intros g g' [pg pg' dif _].
    destruct (All_In_dec_forall_exists all_nodes_prop
                                       (P := fun n => ~ g' n == g n))
             as [N | [e he]].
    { intros n nn en; rewrite en; auto. }
    { intros a.
      destruct (State_dec (g' a) (g a)); [right | left]; auto. }
    - contradict dif; intros nn nnn en; rewrite en; clear nn en.
      specialize (N nnn).
      destruct (State_dec (g' nnn) (g nnn)); [| contradict N]; auto.
    - exists e; rewrite has_moved_progress; auto.
  Qed.

  Lemma has_moved_b_enabled_b:
    forall (g' g: Env), Step g' g ->
                        forall n,
                          has_moved_b g' g n = true -> enabled_b g n = true.
  Proof.
    intros g' g [pg pg' _ r] n hm.
    unfold has_moved_b in hm; unfold enabled_b.
    destruct (State_dec (g' n) (g n)) as [hgg'|hgg']; [contradict hm; auto|].
    specialize (r n); destruct r as [r|r]; [| tauto].
    destruct (State_dec (RUN g n) (g n)) as [hr|hr]; auto.
    contradict hgg'; rewrite <- hr; auto.
  Qed.

  Section with_Stable_Var.

    (** ** Read-Only States and Configurations

     - The read-only part of a configuration is always kept unchanged,
     whatever be the step.
     - This is, henceforth, also true for valid traces
     - Any true property on the read only part of the initial
     configuration keeps being true all along the trace.
     *)

    Variable VarType: Type.
    Variable getVar: State -> VarType.
    Variable SV: Stable_Variable _ getVar.

    Notation EnvVar (*: Type*) := (Node -> VarType).

    Notation getVarPart (*: Env -> EnvVar*) :=
      (fun env: Env => fun n => getVar (env n)).

    Instance getVarPart_compat: Proper fequiv getVarPart.
    Proof.
      intros g g' eg; intros n1 n2 en.
      apply GetVar_compat, eg, en.
    Qed.

    Lemma Stable_Variable_RUN:
      forall (env: Env) (n: Node), getVar (RUN env n) =~= getVar (env n).
    Proof. intros until n; apply GetVar_stable. Qed.

    Lemma Stable_Variable_step:
      forall (env' env: Env),
        Step env' env -> getVarPart env' =~= getVarPart env.
    Proof.
      intros until env; intros [pg pg' _ r] n1 n2 en.
      rewrite en; clear en n1.
      specialize (r n2); destruct r as [r|r].
      rewrite r; apply Stable_Variable_RUN.
      apply GetVar_compat; auto.
    Qed.

    Lemma step_assume:
      forall (g g': Env) P,
        Proper (pequiv ==> iff) P ->
        P (getVarPart g) -> Step g' g -> P (getVarPart g').
    Proof.
      intros until P; intros P_compat pg st.
      eapply P_compat.
      2: apply pg.
      apply Stable_Variable_step; auto.
    Qed.

  End with_Stable_Var.

  (** ** Assumptions on inputs of the algorithm *)
  Class Stable_Assumption: Type :=
    {
      ROType: Type;
      getRO: State -> ROType;
      ROstable:> Stable_Variable _ getRO;
      Assume_RO: (Node -> ROType) -> Prop;
      Assume_RO_compat: Proper fequiv Assume_RO;
      
      ROEnv := Node -> ROType;
      ROEnv_part := (fun env: Env => fun n => getRO (env n));
      Assume := (fun g: Env => Assume_RO (ROEnv_part g));
    }.
  
  Global Instance ROEnv_part_compat {SA: Stable_Assumption}:
    Proper pequiv ROEnv_part.
  Proof.
    intros ro1 ro2 ero; unfold ROEnv_part.
    intros n1 n2 en; apply GetVar_compat; auto.
  Qed.
  
  Global Instance Assume_compat {SA: Stable_Assumption}:
    Proper pequiv Assume.
  Proof.
    intros g1 g2 eg.
    unfold Assume.
    apply Assume_RO_compat, ROEnv_part_compat; auto.
  Qed.

  (** Build a Trivial assumption on the same stable variable as an exisitng one *)
  (** Should not be an instance -> loop *)
  Definition Trivial_Assumption (StA:Stable_Assumption) : Stable_Assumption :=
    @Build_Stable_Assumption
      _ _ ROstable (fun _ => True) (fun _ _ _ => conj (fun _ => I) (fun _ => I)).
  
  Lemma step_Assume:      
    forall {SA: Stable_Assumption} (g g': Env),
      Assume g -> Step g' g -> Assume g'.
  Proof.
    intros [RO get stab assRO assRO_c roe groe Assume] g g'.  
    unfold RelModel.Assume, Assume_RO.
    intros ass stp.
    apply (step_assume stab assRO_c ass stp).
  Qed.

  (** *** Terminal configuration
   A terminal configuration is a compatible configuration from which 
   no more step is allowed *)
  Definition terminal (g: Env): Prop :=
    Proper pequiv g /\ forall g', ~Step g' g.

  Definition terminal_alt (g: Env): Prop := RUN g =~= g.

  Lemma terminal_def_equiv:
    forall (g: Env), terminal g <-> terminal_alt g.
  Proof.
    intros g; unfold terminal, terminal_alt; split.
    { intros [pg term].
      destruct (Env_dec (RUN g) g) as [r|r].
      - apply r; auto; apply RUN_compat; auto.
      - specialize (term (RUN g)); contradict term.
        split; auto.
        apply RUN_compat; auto.
    }
    { intros r; split.
      - proper_.
      - intros g' [pg pg' dif rr].
        contradict dif; intros nn n en; rewrite en; clear nn en.
        destruct (rr n) as [rrr|rrr];
          [rewrite <- (r _ _ (Equivalence_Reflexive n))|];auto.
    }
  Qed.

  Global Instance terminal_alt_compat: Proper fequiv terminal_alt.
  Proof.
    intros g1 g2 eg; unfold terminal_alt.
    split; intros H.
    rewrite <- eg; auto.
    rewrite eg; auto.
  Qed.

  Global Instance terminal_compat: Proper fequiv terminal.
  Proof.
    intros g1 g2 eg.
    do 2 rewrite terminal_def_equiv.
    apply terminal_alt_compat; assumption.
  Qed.

  Lemma terminal_alt_dec:
    forall g, { Proper pequiv g -> terminal_alt g } + { ~ terminal_alt g }.
  Proof.
    intros g; unfold terminal_alt.
    destruct (Env_dec (RUN g) g) as [x|]; [left|right]; auto.
    intros pg; apply x; auto.
    apply RUN_compat; auto.
  Qed.

  Lemma terminal_dec:
    forall g, { Proper pequiv g -> terminal g } + { ~ terminal g }.
  Proof.
    intros g.
    destruct (terminal_alt_dec g) as [T | T].
    - left; intros p; specialize (T p);
        rewrite <- terminal_def_equiv in T; assumption.
    - right; intros F; rewrite <- terminal_def_equiv in T; contradiction.
  Qed.

  Lemma terminal_all_disabled:
    forall g, terminal g -> forall n, ~enabled g n.
  Proof.
    intros g term; rewrite terminal_def_equiv in term;
      unfold terminal_alt in term.
    intros n; rewrite enabled_prop; auto.
  Qed.

  Lemma terminal_all_disabled_reverse:
    forall g, Proper pequiv g -> (forall n, ~enabled g n) -> terminal g.
  Proof.
    intros g pg H.
    rewrite terminal_def_equiv; unfold terminal_alt.
    intros nn n en; rewrite en; clear en nn.
    specialize (H n).
    rewrite enabled_prop in H.
    destruct (State_dec (RUN g n) (g n)); tauto.
  Qed.

  Lemma terminal_dec_bis:
    forall (g: Env), Proper pequiv g -> { n | enabled g n} + { terminal g }.
  Proof.
    intros g pg.
    cut ({n | enabled g n} + {forall n, ~ enabled g n}).
    { intros [ok|ok]; [left; auto| right].
      rewrite terminal_def_equiv; intros nn n en; rewrite en; clear nn en.
      specialize (ok n); rewrite enabled_prop in ok.
      destruct (State_dec (RUN g n) (g n)); tauto.
    }
    apply (All_In_dec_forall_ex all_nodes_prop).
    - apply enabled_compat, pg.
    - intros a; unfold enabled; destruct (enabled_b g a); [left|right]; auto.
  Qed.

  Lemma Step_not_terminal:
    forall g g', Step g' g -> ~ terminal g.
  Proof.
    intros g' g step [_ term].
    specialize (term g).
    tauto.
  Qed.

  Lemma RUN_Step: forall g, Proper pequiv g -> ~terminal g -> Step (RUN g) g.
  Proof.
    intros g pg nterm; split; auto.
    - apply RUN_compat, pg.
    - intros F; contradict nterm; rewrite terminal_def_equiv; auto.
  Qed.

End RelModel.

Close Scope signature_scope.
Unset Implicit Arguments.
