
(**
*Certified Round Complexity of Self-Stabilizing Algorithms*
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Bool.
From Coq Require Import Lia.

(** **** Local Imports *)
From Padec Require Import Algorithm.
From Padec Require Import BoolSet.
From Padec Require Import SetoidUtil.
From Padec Require Import RelModel. (* TODO: Ambiguous paths: [@Step_proper_l] : Step >-> Proper *)
From Padec Require Import Stream.
From Padec Require Import Exec.
From Padec Require Import ListUtils.
From Padec Require Import Fairness.

(* end show *)
Open Scope nat_scope.
Open Scope signature_scope.
Set Implicit Arguments.

(** * Rounds *)
(** In this section, we give tools to compute time complexity of
   an algorithm in terms of rounds.
 *)
Section Rounds.

  Context {Chans: Channels} {Net: Network} {Algo: Algorithm}
          {SA: Stable_Assumption}.

  Notation zerob := (Nat.eqb 0).
  Notation not_zerob := (fun n => negb (Nat.eqb 0 n)).

  (** ** Tools 
   The definitions for rounds use 
     - sets of nodes that we model as Boolean set (see BoolSet)
     - activated nodes (i.e., nodes that execute during a step) and
       neutralized nodes (i.e., nodes that becomes disabled during
       a step without executing).
   We first introduce tools about Boolean node sets and activated and
   neutralized nodes.
   *)
  
  (** *** Tools to describe Node set as Boolean sets 
   This type will be used to represent the set of unsatisfied 
   nodes while computing a round.
   A set of unsatisfied nodes is initialized by the set of
   enabled nodes. It then decreases at each step by the nodes
   which are activated or neutralized during the step.
   *)
  Notation NodeSet := (BoolSet (A:=Node)).
  Notation eqNodeSet := (eqBoolSet (eqA := (equiv (A:=Node)))).
  Notation leNodeSet := (leBoolSet (eqA := (equiv (A:=Node)))).
  Notation is_empty := (fun S => EMPTY all_nodes S).

  Lemma eqNodeSet_dec: forall A B,
      Proper pequiv A -> Proper pequiv B ->
      {eqNodeSet A B}+{~eqNodeSet A B}.
  Proof.
    intros A B pA pB.
    eapply eqBoolSet_dec; auto.
    apply all_nodes_prop.
  Qed.
  
  Definition UNSAT_init g := enabled_b g.
  Instance UNSAT_init_compat: Proper pequiv UNSAT_init.
  Proof. apply enabled_b_compat. Qed.

  Definition UNSAT_update U e := diffBoolSet U (act_neut_b e).
  Instance UNSAT_update_compat: Proper pequiv UNSAT_update.
  Proof.
    intros U V UV g f gf.
    apply diffBoolSet_compat; auto.
    apply act_neut_b_compat; auto.
  Qed.

  Lemma UNSAT_shrinks (g: Env) (e: Exec) (U U': NodeSet)
        (pU: Proper fequiv U) (pg: Proper fequiv g)
        (pe: Proper pequiv e):
    eqNodeSet U' (UNSAT_update U (s_cons g e)) ->
    forall n,  U' n = true -> U n = true. 
  Proof.
    intros eU' w.
    specialize (eU' _ _ (Equivalence_Reflexive w)).
    rewrite eU'.
    unfold diffBoolSet, andBoolSet.
    apply andb_prop.
  Qed.

  Lemma UNSAT_enabled (g: Env) (e: Exec) (U U': NodeSet)
        (pU: Proper fequiv U) (pg: Proper fequiv g)
        (pe: Proper pequiv e):
    eqNodeSet U' (UNSAT_update U (s_cons g e)) -> 
    leNodeSet U (UNSAT_init g) ->
    leNodeSet U' (UNSAT_init (s_hd e)).
  Proof.
    unfold UNSAT_update, UNSAT_init.
    intros eU' lU.
    intros  w w_ ew.
    case_eq (U' w); intro hw; [|reflexivity].
    simpl.
    rewrite <- (enabled_b_compat (hd_compat pe) _ _ ew).
    specialize (eU' _ _ (Equivalence_Reflexive w)).
    unfold diffBoolSet, andBoolSet, negBoolSet in eU'.
    rewrite hw in eU'; symmetry in eU'.
    rewrite andb_true_iff in eU'.
    destruct eU' as [hw1 eU'].
    specialize (lU _ _ (Equivalence_Reflexive w)).
    rewrite hw1 in lU; simpl in lU.
    rewrite negb_true_iff in eU'.
    unfold act_neut_b,  AN_b, ACT_b, NEUT_b in eU'.
    rewrite orb_false_iff in eU'.
    destruct eU' as [mov2 eU'].
    rewrite lU in eU'.
    simpl in eU'.
    rewrite mov2 in eU'.
    simpl in eU'.
    rewrite negb_false_iff in eU'.
    assumption.
  Qed.

  Lemma UNSAT_update_incr:
    forall U U' e,
      Proper pequiv e -> leNodeSet U U' ->
      leNodeSet (UNSAT_update U e) (UNSAT_update U' e).
  Proof.
    intros U U' e pe hle.
    apply diffBoolSet_le_compat; auto.
    unfold flip.
    apply eqBoolSet_leBoolSet.
    apply act_neut_b_compat; auto.
  Qed.

  Lemma UNSAT_are_enabled:
    forall U (pu: Proper pequiv U) g (pg: Proper pequiv g)
           e (pe: Proper pequiv e),
      leNodeSet U (UNSAT_init (s_hd (s_cons g e))) ->
      leNodeSet (UNSAT_update U (s_cons g e))
                (UNSAT_init (s_hd e)).
  Proof.
    intros U pu g pg e pe leu p q pq.
    rewrite pq.
    specialize (leu _ _ pq); rewrite pq in leu.
    clear pq; clear p.
    case_eq (UNSAT_update U (s_cons g e) q); intros hq;
      unfold Bool.Bool.le; auto.
    unfold UNSAT_update, diffBoolSet, andBoolSet, negBoolSet
      in hq.
    assert (hu: U q = true).
    { case_eq (U q); intros hh; auto.
      rewrite hh in hq; tauto. }
    rewrite hu in leu.
    unfold Bool.Bool.le in leu.
    assert (hh: act_neut_b (s_cons g e) q = false).
    { case_eq (act_neut_b (s_cons g e) q); intros hh; auto.
      rewrite hh, hu in hq; unfold negb, "&&" in hq.
      congruence. } 
    clear hq hu.
    unfold UNSAT_init in leu; unfold UNSAT_init.
    apply act_neut_contd with g; auto.
  Qed.
  
  Instance negb_compat: Proper fequiv negb.
  Proof. intros a b ab; rewrite ab; auto. Qed.
   
  Lemma diff_not_empty:
    forall (s s': NodeSet),
      Proper eqNodeSet s -> Proper eqNodeSet s' ->
      ~ eqNodeSet (diffBoolSet s s') empty_BoolSet
      -> ~ eqNodeSet s empty_BoolSet.
  Proof.
    intros s s' ps ps' neq; contradict neq.
    rewrite neq; apply empty_diff.
  Qed.

  Lemma le_empty:
    forall U U', leNodeSet U U' ->
                 eqNodeSet U' empty_BoolSet ->
                 eqNodeSet U empty_BoolSet.
  Proof.
    intros U U' hle h'.
    intros a b ab; unfold empty_BoolSet; simpl.
    specialize (h' _ _ ab); unfold empty_BoolSet in h'; simpl in h'.
    assert (pa: Proper pequiv a) by proper_.
    specialize (hle _ _ pa); simpl in hle.
    rewrite h' in hle.
    revert hle; unfold Bool.Bool.le.
    case_eq (U a); congruence.
  Qed.

  Lemma is_empty_true:
    forall S, Proper pequiv S ->
              is_empty S = true <-> eqNodeSet S empty_BoolSet.
  Proof.
    intros S h; erewrite EMPTY_empty.
    2: apply setoid_refl.
    2: apply all_nodes_prop.
    2: apply h.
    tauto.
  Qed.

  Lemma is_empty_false:
    forall S, Proper pequiv S ->
              is_empty S = false <-> ~ eqNodeSet S empty_BoolSet.
  Proof.
    intros S h.
    assert (T := is_empty_true h).
    split; intros hh.
    intros F; apply T in F; congruence.
    rewrite <- T in hh.
    apply not_true_is_false in hh; auto.
  Qed.

  Lemma le_U_enabled:
    forall e U p, 
      U p = true -> leNodeSet U (UNSAT_init (s_hd e)) ->
      enabled_b (s_hd e) p = true.
  Proof.
    intros e U p hu hle.
    specialize (hle _ _ (Equivalence_Reflexive p)).
    unfold leNodeSet, UNSAT_init, Bool.Bool.le in hle.
    rewrite hu in hle; auto.
  Qed.

  Lemma empty_act_neut:
    forall e U p, 
      U p = true -> eqNodeSet (UNSAT_update U e) empty_BoolSet
      -> act_neut_b e p = true.
  Proof.
    intros e U p hu emp.
    specialize (emp _ _ (Equivalence_Reflexive p)).
    unfold empty_BoolSet, UNSAT_update, diffBoolSet,
    andBoolSet, negBoolSet in emp.
    destruct (act_neut_b e p); auto.
    rewrite hu in emp.
    unfold "&&", negb in emp.
    congruence.
  Qed.

  Lemma in_diff_act_neut:
    forall e U p,
      U p = true -> UNSAT_update U e p = false ->
        act_neut_b e p = true.
  Proof.
    intros e U p hu huu.
    unfold UNSAT_update, diffBoolSet, andBoolSet,
    negBoolSet, "&&", negb in huu.
    rewrite hu in huu.
    destruct (act_neut_b e p); auto.
  Qed.

  Lemma in_diff_not_act_neut:
    forall e U p,
      U p = true -> UNSAT_update U e p = true ->
        act_neut_b e p = false.
  Proof.
    intros e U p hu huu.
    unfold UNSAT_update, diffBoolSet, andBoolSet,
    negBoolSet, "&&", negb in huu.
    rewrite hu in huu.
    destruct (act_neut_b e p); auto.
  Qed.

  (** Round Definition *)
  
  CoInductive at_most_rounds_aux (P: Exec -> Prop):
    NodeSet -> nat -> Exec -> Prop :=
  | rnd_here: forall U n e, P e -> at_most_rounds_aux P U n e
  | rnd_in: forall U n g e,
      at_most_rounds_aux P (UNSAT_update U (s_cons g e)) n e ->
      n > 0 ->
      ~eqNodeSet (UNSAT_update U (s_cons g e)) empty_BoolSet ->                        
      at_most_rounds_aux P U n (s_cons g e)
  | rnd_chge: forall U n g e,
      at_most_rounds_aux P (UNSAT_init (s_hd e)) n e ->
      eqNodeSet (UNSAT_update U (s_cons g e)) empty_BoolSet ->                        
      at_most_rounds_aux P U (S n) (s_cons g e).

  Definition at_most_rounds (P: Exec -> Prop) (n: nat) (e: Exec): Prop :=
    at_most_rounds_aux P (UNSAT_init (s_hd e)) n e.
  
  Lemma at_most_rounds_aux_incr:
    forall P U n e, at_most_rounds_aux P U n e ->
                    forall n', n' >= n ->
                               at_most_rounds_aux P U n' e.
  Proof.
    intros P.
    cofix h.
    intros U n e am n' hn'.
    inversion am; subst.
    { apply rnd_here; auto. }
    { apply rnd_in; auto.
      apply h with n; auto.
      lia.
    }
    { destruct n' as [|n']; [lia|].
      apply rnd_chge; auto.
      apply h with n0; auto.
      lia.
    }
  Qed.

  Lemma at_most_rounds_incr:
    forall P n e, at_most_rounds P n e ->
                  forall n', n' >= n -> at_most_rounds P n' e.
  Proof.
    unfold at_most_rounds.
    intros P n e ha n' hn'.
    apply at_most_rounds_aux_incr with n; auto.
  Qed.

  Instance at_most_rounds_aux_compat:
    Proper pequiv at_most_rounds_aux.
  Proof.
    cut (forall p q u v n x y,
            p =~= q -> u =~= v -> x =~= y ->
            at_most_rounds_aux p u n x -> at_most_rounds_aux q v n y).
    { intros ok p q pq u v uv n m nm x y xy.
      simpl in nm; subst.
      split; apply ok; auto; symmetry; auto. }
    cofix h.
    intros p q u v n x y pq uv xy hh.
    inversion hh; subst.
    - apply rnd_here.
      revert H; apply pq, xy; auto.
    - inversion xy; subst.
      apply rnd_in; auto.
      apply h with p (UNSAT_update u (s_cons g e)) e; auto.
      apply UNSAT_update_compat; auto.
      contradict H1; rewrite <- H1.
      apply UNSAT_update_compat; auto.
    - inversion xy; subst.
      apply rnd_chge; auto.
      apply h with p (UNSAT_init (s_hd e)) e; auto.
      apply UNSAT_init_compat; auto.
      apply hd_compat; auto.
      rewrite <- H0.
      apply UNSAT_update_compat; symmetry; auto.
  Qed.

  Instance at_most_rounds_compat: Proper pequiv at_most_rounds.
  Proof.
    intros p q pq n m nm x y xy.
    apply at_most_rounds_aux_compat; auto.
    apply UNSAT_init_compat, hd_compat; auto.
  Qed.

  Lemma at_most_rounds_equiv_context:
    forall (P1: Exec -> Prop) (P2: Exec -> Prop)
           Q e (pe: Proper pequiv e) k,
      (Always Q) e -> Always (fun e => P1 e -> P2 e) e ->
      at_most_rounds P1 k e -> at_most_rounds P2 k e.
  Proof.
    intros P1 P2 Q e pe k hq ha.
    cut (forall U, Proper pequiv U ->
                   at_most_rounds_aux P1 U k e ->
                   at_most_rounds_aux P2 U k e).
    { intros ok; apply ok.
      apply UNSAT_init_compat, hd_compat; auto. }
    revert pe hq ha; revert e k;
      cofix h; intros e k pe hq ha U pu hr; inversion hr; subst.
    { apply rnd_here.
      apply Always_P_P in ha; auto. }    
    destruct k as [|k]; [lia|].
    { apply rnd_in; auto.
      apply h; auto.
      inversion pe; auto.
      inversion hq; auto.
      inversion ha; auto.
      apply UNSAT_update_compat; auto.
    }
    { apply rnd_chge; auto.
      apply h; auto.
      inversion pe; auto.
      inversion hq; auto.
      inversion ha; auto.
      apply UNSAT_init_compat, hd_compat; auto.
      inversion pe; auto.
    }
  Qed.

  Lemma at_most_rounds_cov:
    forall (P1: Exec -> Prop) (P2: Exec -> Prop) 
           e (pe: Proper pequiv e) k
           (impl2: Always (fun e => P1 e -> P2 e) e),
      at_most_rounds P1 k e -> at_most_rounds P2 k e.
  Proof.
    intros P1 P2 e pe ha.
    apply at_most_rounds_equiv_context with (fun _ => True); auto. 
    clear; revert e; cofix h; intros [g | g e].
    left; auto.
    right; auto.
  Qed.
    
  (** Function Round Definition   
     as long as one can compute the reachability of the 
     predicate. *)

  (** count_rounds_aux assumes Finally P e 
   (i.e. reachability of P can be computed); 
   it uses a set of unsatisfied nodes. 
   - Either Finally tells that P is reached, hence one round
   - Or the set of unsatisfied is refreshed by removing moving 
     or neutralized nodes. If this new set is empty, this 
     starts a new round (add 1 and initialize the new set of 
     unsatisfied with the enabled nodes of the current 
     configuration) otherwise we go on the current round.
   *)
  Fixpoint count_rounds_aux (P: Exec -> Prop) (g: Env) (e: Exec)
           (U: NodeSet) (F: Finally P e): nat :=
    match F with
    | finally_now _ _ _ => 1
    (* end of final (partial) round because P e *)
    | @finally_not_yet _ _ g' e' _ F' =>
      if is_empty (UNSAT_update U (s_cons g e))
      then 1 + count_rounds_aux g' (UNSAT_init g') F'
      else count_rounds_aux g' (UNSAT_update U (s_cons g e)) F'
    end.

  (** count_rounds provides the number of rounds to reach P.
   It compute the 0 case separately. Otherwise it uses 
   count_rounds_aux and initialize the new set of unsatisfied 
   with the enabled nodes of the initial configuration.
   *)
  Definition count_rounds
             (P: Exec -> Prop) (P_dec: Weak_Dec pequiv P) 
             (e: Exec) (pe: Proper pequiv e)
             (he: Eventually P e): nat :=
    match (Eventually_Dec_Finally' P_dec pe he) in
          Finally _ e_ return  nat with
    | finally_now _ _ _ => 0 (* zero round because P e *)
    | finally_not_yet g' e' F' =>
      count_rounds_aux g' (UNSAT_init g') F'
    end.

  Lemma count_rounds_aux_S: forall P g e U (F: Finally P e), 
      count_rounds_aux g U F <> 0.
  Proof.
    intros P g e U F; revert g U.
    induction F; intros g U; simpl; [lia|].
    destruct (is_empty
                    (UNSAT_update U (s_cons g (s_cons a s)))).
    lia.
    apply IHF.
  Qed.

  (** *** Progress Towards P *) 
  (** Simplfied assumption about an execution converging towards
   a goal P; if not yet reached, then there should be some 
   enabled node(s). *)
  Definition progress_P (P: Exec -> Prop) (e: Exec): Prop :=
    Eventually P e ->
    Until (fun e => ~ eqNodeSet (UNSAT_init (s_hd e))
                      empty_BoolSet) P e.
  
  Lemma progress_P_is_exec:
    forall P (P_dec: Weak_Dec pequiv P) e,
      is_exec e -> progress_P P e.
  Proof.
    intros P P_dec e ise ev; unfold UNSAT_init.
    induction ev.
    { left; auto. }
    { destruct (P_dec (s_cons a s)).
      { left; apply p, is_exec_proper, ise. }
      right.
      2: { apply IHev.
           apply is_exec_inv_2 in ise; auto. }
      { assert (pge := is_exec_proper ise).
        ssplit pge; destruct pge as [pg pe].
        simpl; intros F; rewrite <- EMPTY_empty in F; auto.
        2: apply all_nodes_prop.
        2: apply enabled_b_compat; auto. 
        unfold EMPTY in F.
        rewrite forallb_forall_InA in F.
        2: { intros aa b ab; apply negb_compat.
             apply enabled_b_compat; auto. }
        apply is_exec_inv_1 in ise; dup ise st.
        apply Step_one_has_moved_true in st.
        destruct st as [p hm].
        specialize (F _ (all_nodes_prop p)).
        apply has_moved_b_enabled_b in hm; auto.
        rewrite hm, negb_true_iff in F; congruence. }
    }
  Qed.

  Lemma progress_P_contd:
    forall P e g, progress_P P (s_cons g e)
                  -> ~ P (s_cons g e) -> progress_P P e.
  Proof.
    intros P e g; unfold progress_P; intros pP np ev.
    assert (dum: Eventually P (s_cons g e)) by
        (apply eventually_later, ev).
    specialize (pP dum); clear dum ev.
    inversion pP; subst; tauto. 
  Qed.

  Lemma progress_P_now:
    forall P e, progress_P P e -> Eventually P e -> ~ P e ->
                ~ eqNodeSet (UNSAT_init (s_hd e)) empty_BoolSet.
  Proof.
    intros P e; unfold progress_P; intros pP ev np.
    specialize (pP ev).
    inversion pP; subst; tauto.
  Qed.
  
  Theorem round_functional_equivalence:
    forall P (P_dec: Weak_Dec pequiv P)
           (P_compat: Proper pequiv P)
           e (pe: Proper pequiv e) (he: Eventually P e)
           (pP: progress_P P e) n,
      at_most_rounds P n e <->
      exists n', n' <= n /\ count_rounds P_dec pe he = n'.
  Proof using Type.
    intros P P_dec P_compat e pe he pP n.
    unfold count_rounds.
    generalize (Eventually_Dec_Finally' P_dec pe he);
      intro F; clear he.
    destruct F as [ e p_e| g e nPge F].
    { split; intros h.
      exists 0; split; lia.
      apply rnd_here; auto. }
    unfold at_most_rounds; simpl.
    assert (Nemp: ~eqNodeSet (UNSAT_init (s_hd (s_cons g e))) empty_BoolSet).
    { apply progress_P_now with P; auto.
      right.
      apply Finally_Eventually; auto. }
    simpl in Nemp.
    assert (pU: Proper pequiv (UNSAT_init g)).
    { apply UNSAT_init_compat; inversion pe; auto. }
    pose (U := UNSAT_init g).
    fold U; fold U in Nemp, pU.
    clearbody U.
    revert Nemp pU; revert U.
    revert pe pP nPge; revert g n.
    induction F as [ e p_e| gg e nPgge F ih]; simpl;
      split; intros h.
    { exists 1; split; [|lia].
      destruct n as [|n]; [|lia].
      inversion h; subst.
      tauto.
      lia.
    }
    { destruct n as [|n].
      { destruct h as [n' [hn' hhn']]; lia. }
      clear h.
      case_eq (EMPTY all_nodes
                     (UNSAT_update U (s_cons g e))).
      { intros emp.
        rewrite is_empty_true in emp.
        2: { apply UNSAT_update_compat; auto. }
        eapply rnd_chge; auto.
        apply rnd_here; auto. }
      { intros emp.
        rewrite is_empty_false in emp.
        2: { apply UNSAT_update_compat; auto. }
        eapply rnd_in; auto.
        apply rnd_here; auto. 
        lia. }        
    }
    { inversion h; subst.
      - tauto.
      - apply ih in H1; auto.
        destruct H1 as [n' [hn' hc]].
        exists n'; split; [lia|].
        case_eq (EMPTY all_nodes
                       (UNSAT_update U (s_cons g (s_cons gg e)))).
        { intros emp.
          rewrite is_empty_true in emp.
          2: { apply UNSAT_update_compat; auto. }
          tauto. }
        intros _; auto.
        inversion pe; auto.
        apply progress_P_contd with g; auto.
        apply UNSAT_update_compat; auto.
      - destruct n0 as [|n0].
        { inversion H3; subst; [tauto|lia]. }
        apply ih in H3; auto.
        destruct H3 as [n' [hn' hc]].
        exists (S n'); split; [lia|].
        case_eq (EMPTY all_nodes
                       (UNSAT_update U (s_cons g (s_cons gg e)))).
        { intros emp.
          rewrite is_empty_true in emp.
          2: { apply UNSAT_update_compat; auto. }
          simpl in hc; lia.
        }
        { intros emp.
          rewrite is_empty_false in emp.
          2: { apply UNSAT_update_compat; auto. }
          tauto. }
        inversion pe; auto.
        apply progress_P_contd with g; auto.
        apply progress_P_now with P; auto.
        apply progress_P_contd with g; auto.
        right; apply Finally_Eventually; auto.
        eapply UNSAT_init_compat; simpl.
        inversion pe; inversion H7; auto.
    }
    { destruct h as [n' [hn' hc]].
      revert hc;
        case_eq (EMPTY all_nodes
                       (UNSAT_update U (s_cons g (s_cons gg e)))).
      { intros emp hc.
        rewrite is_empty_true in emp.
        2: { apply UNSAT_update_compat; auto. }
        destruct n as [|n]; [lia|].
        apply rnd_chge; auto.
        apply ih; auto.
        inversion pe; auto.
        apply progress_P_contd with g; auto.
        apply progress_P_now with P; auto.
        apply progress_P_contd with g; auto.
        right; apply Finally_Eventually; auto.
        eapply UNSAT_init_compat; simpl.
        inversion pe; inversion H4; auto.
        destruct n' as [|n']; [lia|].
        exists n'; split; [lia|auto].
      }
      { intros emp hc.
        rewrite is_empty_false in emp.
        2: { apply UNSAT_update_compat; auto. }
        apply rnd_in; auto.
        apply ih; auto.
        inversion pe; auto.
        apply progress_P_contd with g; auto.
        apply UNSAT_update_compat; auto.
        exists n'; split; [lia|auto].
        { destruct n as [|n]; [|lia].
          assert (X: n' <> 0).
          rewrite <- hc; apply count_rounds_aux_S.
          lia.
        }
      }
    }
  Qed.

  Lemma round_functional_equivalence':
    forall P (P_dec: Weak_Dec pequiv P)
           (P_compat: Proper pequiv P)
           e (pe: Proper pequiv e) (he: Eventually P e)
           (pP: progress_P P e) n',
      count_rounds P_dec pe he = n' ->
        forall n, n' <= n -> at_most_rounds P n e.
  Proof.
    intros P P_dec P_compat e pe he pP n' h n hn.
    eapply round_functional_equivalence; auto.
    exists n'; split; auto.
    apply h.
  Qed.
  
  (** Induction Schema - Usual Proof Pattern to prove Round Complexity *)

  Lemma rnd_glue0:
    forall P (pc: Proper pequiv P) e (pe: Proper pequiv e)
           U U' (pu': Proper pequiv U'),
      leNodeSet U U' ->
      at_most_rounds_aux P U 1 e ->
      at_most_rounds_aux P U' 1 e.
  Proof. 
    intros P pc; cofix h.
    intros e pe U U' pu' hle am.
    inversion am; subst.
    { apply rnd_here; auto. }
    { apply rnd_in; auto.
      apply h with (UNSAT_update U (s_cons g e0)); auto.
      inversion pe; auto.
      apply UNSAT_update_compat; auto.
      apply UNSAT_update_incr; auto.
      contradict H1.
      apply le_empty with (UNSAT_update U' (s_cons g e0)); auto.
      apply UNSAT_update_incr; auto.
    }
    case_eq (EMPTY all_nodes
                   (UNSAT_update U' (s_cons g e0))).
    { intros emp.
      rewrite is_empty_true in emp.
      2: { apply UNSAT_update_compat; auto. }
      apply rnd_chge; auto.
    }
    { intros emp.
      rewrite is_empty_false in emp.
      2: { apply UNSAT_update_compat; auto. }
      apply rnd_in; auto.
      apply h with (UNSAT_update U (s_cons g e0)); auto.
      inversion pe; auto.
      apply UNSAT_update_compat; auto.
      apply UNSAT_update_incr; auto.
      inversion H0; subst.
      2: { lia. }
      apply at_most_rounds_aux_incr with 0; auto.
      apply rnd_here; auto.
    }
  Qed.

  Lemma rnd_glue:
    forall P (pc: Proper pequiv P) e (pe: Proper pequiv e)
           U (pu: Proper pequiv U) U' (pu': Proper pequiv U'),
      leNodeSet U U' -> leNodeSet U' (UNSAT_init (s_hd e)) ->
      at_most_rounds_aux P U' 1 e ->
      at_most_rounds_aux P U 2 e.
  Proof.
    intros P pc; cofix h.
    intros e pe U pu U' pu' hle hlei am.
    inversion am; subst.
    { apply at_most_rounds_aux_incr with 0;
        [apply rnd_here; auto|lia]. }
    case_eq (EMPTY all_nodes
                   (UNSAT_update U (s_cons g e0))).
    { intros emp.
      rewrite is_empty_true in emp.
      2: { apply UNSAT_update_compat; auto. }
      apply rnd_chge; auto.
      apply rnd_glue0 with (UNSAT_update U' (s_cons g e0)); auto.
      inversion pe; auto.
      apply UNSAT_init_compat, hd_compat; inversion pe; auto.
      apply UNSAT_are_enabled; auto.
      inversion pe; auto.
      inversion pe; auto.
    }
    { intros emp.
      rewrite is_empty_false in emp.
      2: { apply UNSAT_update_compat; auto. }
      apply rnd_in; auto.
      apply h with (UNSAT_update U' (s_cons g e0)); auto.
      inversion pe; auto.
      apply UNSAT_update_compat; auto.
      apply UNSAT_update_compat; auto.
      apply UNSAT_update_incr; auto.
      apply UNSAT_are_enabled; auto.
      inversion pe; auto.
      inversion pe; auto.
    }
    { apply rnd_chge; auto.
      apply at_most_rounds_aux_incr with 0; auto.
      apply le_empty with (UNSAT_update U' (s_cons g e0)); auto.
      apply UNSAT_update_incr; auto.
    }
  Qed.
  
  Lemma schema_round_step:
    forall (P1 P2: Exec -> Prop) (P1c: Proper pequiv P1)
           (P2c: Proper pequiv P2)
           (n: nat) (e: Exec) (pe: Proper pequiv e),
      Always (fun e => P1 e -> at_most_rounds P2 1 e) e ->
      at_most_rounds P1 n e -> at_most_rounds P2 (S n) e.
  Proof.
    intros P1 P2 P1c P2c [|n] e pe.
    { intros ha; apply Always_P_P in ha.
      intros am; inversion am; subst.
      auto.
      lia.
    }
    pose (n' := S n).
    assert (hn: n' > 0) by lia.
    fold n'; clearbody n'; clear n.
    unfold at_most_rounds.
    pose (U := UNSAT_init (s_hd e)).
    assert (pU: Proper pequiv U).
    { apply UNSAT_init_compat; inversion pe; auto. }
    assert (pUU: leNodeSet U U).
    { apply eqBoolSet_leBoolSet.
      apply UNSAT_init_compat, hd_compat, pe. }
    unfold U at 2 in pUU; fold U.
    clearbody U.
    revert pe pU pUU hn; revert e U n'.
    cofix h.
    intros e U n pe pU pUU hn ha am.
    inversion am; subst.
    { apply Always_P_P in ha.
      apply ha in H.
      apply at_most_rounds_aux_incr with 2; auto; [|lia].
      apply rnd_glue with (UNSAT_init (s_hd e)); auto.
      apply UNSAT_init_compat, hd_compat; auto.
      apply eqBoolSet_leBoolSet.
      apply UNSAT_init_compat, hd_compat; auto.
    }
    { apply rnd_in; auto.
      apply h; auto.
      inversion pe; auto.
      apply UNSAT_update_compat; auto.
      apply UNSAT_enabled with g U; auto.
      inversion pe; auto.
      inversion pe; auto.
      apply UNSAT_update_compat; auto.
      apply Always_inv_2 in ha; auto.
    }
    { apply rnd_chge; auto.
      destruct n0 as [|n0].
      { inversion H; subst.
        2: { lia. }
        apply Always_inv_2 in ha; apply Always_P_P in ha.
        auto.
      }
      apply h; auto.
      inversion pe; auto.
      apply UNSAT_init_compat, hd_compat; inversion pe; auto.
      apply eqBoolSet_leBoolSet.
      apply UNSAT_init_compat, hd_compat; inversion pe; auto.
      lia.
      apply Always_inv_2 in ha; auto.
    }
  Qed.
  
  Lemma schema_round_induction:
    forall (P: nat -> Exec -> Prop) (Pc: Proper pequiv P)
           (B: nat) (e: Exec) (pe: Proper pequiv e),
      P 0 e -> (* base case *)
      Always (fun e => forall k,
                  k < B -> P k e -> (* induction step along e*)
                  at_most_rounds (P (S k)) 1 e) e ->
      at_most_rounds (P B) B e.
  Proof.
    intros P Pc B e pe h0 hind.
    induction B as [| B ih].
    { apply rnd_here; auto. }
    { apply schema_round_step with (P1 := P B); auto.
      apply Pc; simpl; auto.
      apply Pc; simpl; auto.
      { revert hind; apply Always_impl_P.
        intros ee ha hp; apply ha; auto. }
      apply ih; auto.
      { revert hind; apply Always_impl_P.
        intros ee ha k hk hp; apply ha; auto. }
    }
  Qed.

  Lemma at_most_rounds_forall:
    forall (P: Node -> Exec -> Prop) k e (pe: Proper pequiv e),
      (forall p,
          at_most_rounds (Always (P p)) k e) <->
      at_most_rounds (fun e => forall p, Always (P p) e) k e.
  Proof.
    intros P k e pe; split.

    {
      unfold at_most_rounds.
      pose (U := UNSAT_init (s_hd e)).
      assert (pu: Proper pequiv U).
      { apply UNSAT_init_compat, hd_compat, pe. }
      fold U.
      clearbody U.
      revert pe pu; revert U e k.
      cofix h; intros U [g | g e] k pe pu hp.
      { apply rnd_here.
        intros p.
        specialize(hp p).
        inversion hp; subst; auto.
      }
      destruct k as [|k].
      { apply rnd_here.
        intros p; specialize (hp p).
        inversion hp; subst; [auto|lia].
      }
      case_eq (EMPTY all_nodes
                     (UNSAT_update U (s_cons g e))).
      { intros emp.
        rewrite is_empty_true in emp.
        2: { apply UNSAT_update_compat; auto. }
        apply rnd_chge; auto.
        apply h; auto.
        inversion pe; auto.
        { apply UNSAT_init_compat, hd_compat.
          inversion pe; auto. }
        intros p; specialize (hp p).
        inversion hp; subst.
        apply rnd_here.
        inversion H; auto.
        tauto.
        auto.
      }
      { intros emp.
        rewrite is_empty_false in emp.
        2: { apply UNSAT_update_compat; auto. }
        apply rnd_in; auto.
        apply h; auto.
        inversion pe; auto.
        { apply UNSAT_update_compat; auto. }
        intros p; specialize (hp p).
        inversion hp; subst.
        apply rnd_here.
        inversion H; auto.
        auto.
        tauto.
        lia.
      }
    }
    {
      unfold at_most_rounds.
      pose (U := UNSAT_init (s_hd e)).
      assert (pu: Proper pequiv U).
      { apply UNSAT_init_compat, hd_compat, pe. }
      fold U.
      clearbody U.
      intros hh p.
      revert pe pu hh; revert U e k.
      cofix h; intros U [g | g e] k pe pu hp.
      { apply rnd_here.
        inversion hp; auto.
      }
      destruct k as [|k].
      { apply rnd_here.
        inversion hp; subst; [auto|lia].
      }
      case_eq (EMPTY all_nodes
                     (UNSAT_update U (s_cons g e))).
      { intros emp.
        rewrite is_empty_true in emp.
        2: { apply UNSAT_update_compat; auto. }
        apply rnd_chge; auto.
        apply h.
        inversion pe; auto.
        { apply UNSAT_init_compat, hd_compat.
          inversion pe; auto. }
        inversion hp; subst; auto.
        { apply rnd_here.
          intros pp; specialize (H pp); inversion H; auto. }
        { tauto. }
      }
      { intros emp.
        rewrite is_empty_false in emp.
        2: { apply UNSAT_update_compat; auto. }
        apply rnd_in; auto.
        apply h.
        inversion pe; auto.
        { apply UNSAT_update_compat; auto. }
        inversion hp; subst; auto.
        { apply rnd_here.
          intros pp; specialize (H pp); inversion H; auto. }
        { tauto. }
        lia.
      }
    }
  Qed.

  Lemma at_most_rounds_forall':
    forall (P: Node -> Exec -> Prop) (pc: Proper pequiv P)
           k e (pe: Proper pequiv e),
      (forall p,
          at_most_rounds (Always (P p)) k e) <->
      at_most_rounds (Always (fun e => forall p, (P p) e)) k e.
  Proof.
    intros P pc k e pe.
    rewrite at_most_rounds_forall; auto.
    apply at_most_rounds_compat; auto.
    { intros x y xy.
      rewrite  Always_forall.
      apply Always_compat; auto.
      clear x y xy; intros x y xy.
      split; intros h q.
      rewrite <- xy; auto.
      rewrite xy; auto.
    }
  Qed.

  Definition Round_complexity (Q P: Exec -> Prop) (k: nat) :=
    forall exec,  Q exec -> at_most_rounds P k exec.

  (** Compatibility for Round_complexity requires more assumption: *)
  Lemma Round_complexity_compat: 
    forall n1 n2 (en: n1 == n2)
           P1 P2 (eP: fequiv P1 P2)
           (Q1 Q2: Exec -> Prop) (eQ: fequiv Q1 Q2)
           (hp1: forall e, Q1 e -> Proper pequiv e)
           (hp2: forall e, Q2 e -> Proper pequiv e),
      Round_complexity Q1 P1 n1 == Round_complexity Q2 P2 n2.
  Proof.
    intros n nn en P1 P2 eP Q1 Q2 eQ hp1 hp2.
    rewrite <- en; clear nn en.
    split; intros h e qe; specialize (h e).
    { rewrite <- (eQ _ _ (hp2 _ qe)) in qe.
      specialize (h qe).
      revert h; apply at_most_rounds_compat; symmetry; auto.
      apply hp1; auto.
    }
    { rewrite (eQ _ _ (hp1 _ qe)) in qe.
      specialize (h qe).
      revert h; apply at_most_rounds_compat; auto.
      apply hp2; auto.
    }
  Qed.

  (** Compatibility for Round_complexity with implies: *)
  Lemma Round_complexity_compat_cov:
    forall n1 n2 (en: n1 == n2)
           (P1: Exec -> Prop) (P1c: Proper pequiv P1)
           (P2: Exec -> Prop) (P2c: Proper pequiv P2)
           (Q1: Exec -> Prop) (Q1c: Proper pequiv Q1)
           (Q2: Exec -> Prop) (Q2c: Proper pequiv Q2)
           (eQ: (pequiv ==> impl) Q2 Q1)
           (eP: forall e, Q2 e -> Always (fun e => P1 e -> P2 e) e)
           (hp2: forall e, Q2 e -> Proper pequiv e),
      Round_complexity Q1 P1 n1 -> Round_complexity Q2 P2 n2.
  Proof.
    intros n nn en P1 P1c P2 P2c Q1 Q1c Q2 Q2c eQ eP hp2 (*d2*).
    rewrite <- en; clear nn en.
    intros h e qe; specialize (h e).
    specialize (hp2 _ qe).
    specialize (eQ _ _ hp2).
    specialize (h (eQ qe)); revert h.
    apply at_most_rounds_cov; auto.
  Qed.    

  Lemma at_most_rounds_scheme_per_node:
    forall e (ise: is_exec e) (ass: Assume (s_hd e))
           (P: Node -> Exec -> Prop) p
           (cnd_here: enabled_b (s_hd e) p = false -> P p e)
           (cnd_act:
              Always (fun e =>
                        act_neut_b e p = true ->
                        match e with
                          s_one _ => True
                        | s_cons g e' => P p e' end) e) k,
      at_most_rounds (P p) (S k) e.
  Proof.
    intros e ise ass P p cnd_here cnd_act k.

    case_eq (enabled_b (s_hd e) p); intros hen.
    2: { apply rnd_here; auto. }

    cut (forall U,
            Proper pequiv U -> U p = true ->
            leNodeSet U (UNSAT_init (s_hd e)) ->
            at_most_rounds_aux (P p) U (S k) e).
    { intros ok; apply ok; auto.
      apply UNSAT_init_compat, Stream.hd_compat; auto.
      apply is_exec_proper in ise; auto.
      unfold UNSAT_init.
      apply eqBoolSet_leBoolSet, UNSAT_init_compat, Stream.hd_compat.
      apply is_exec_proper in ise; auto.
    }
    clear hen cnd_here.
    
    revert ise ass cnd_act; revert e.
    cofix h; intros e ise ass cnd_act U pu hr hle.
    assert (pe: Proper pequiv e).
    { apply is_exec_proper in ise; auto. }
    destruct e as [g | g e].
    { apply is_exec_inv_0 in ise.
      apply terminal_all_disabled with (n := p) in ise.
      contradict ise.
      apply le_U_enabled with (p := p) in hle; auto.
    }
    
    assert (isee: is_exec e).
    { apply is_exec_inv_2 in ise; auto. }
    assert (asss: Assume (s_hd e)).
    { eapply step_Assume.
      apply ass.
      apply is_exec_inv_1 in ise; auto. }
    
    case_eq (EMPTY all_nodes (UNSAT_update U (s_cons g e))).
    { intros emp.
      rewrite is_empty_true in emp.
      2: { apply UNSAT_update_compat; auto. }
      apply rnd_chge; auto.
      apply rnd_here.
      apply Always_P_P in cnd_act.
      apply cnd_act.
      apply empty_act_neut with U; auto.
    }
    { intros emp.
      rewrite is_empty_false in emp.
      2: { apply UNSAT_update_compat; auto. }
      apply rnd_in; auto.
      case_eq (UNSAT_update U (s_cons g e) p); intros hrr.
      { apply h; auto.
        inversion cnd_act; auto.
        apply UNSAT_update_compat; auto.
        apply UNSAT_are_enabled; auto.
        inversion pe; auto.
        inversion pe; auto.
      }
      { apply rnd_here.
        apply Always_P_P in cnd_act.
        apply cnd_act.
        apply in_diff_act_neut with U; auto.
      }
      lia.
    }
  Qed.

  (** Results with weakly fair daemon 


   weakly_fair e -> forall U included in (enabled e), Eventually (U becomes empty)
    nb: this requires to define a new Eventually property, with accumulator
    how to prove this???  coind prop ---> ind prop

   prove that: wf e -> forall node n, Eventually (n is not in U)
    uses the Eventually part in the weakly_fair definition -- ok

   lemma above should be provable using induction on Eventually (U becomes empty)

   *)

  Lemma UNSAT_update_le:
    forall U (pu: Proper pequiv U) e (pe: Proper pequiv e),
      leNodeSet (UNSAT_update U e) U.
  Proof.
    intros U pu e pe x y xy.
    rewrite xy; clear xy x.
    unfold UNSAT_update, diffBoolSet, andBoolSet, "&&",
    Bool.Bool.le, negBoolSet, negb.
    destruct (U y); auto. 
    destruct (if act_neut_b e y then false else true); auto.
  Qed.

  Lemma UNSAT_update_empty:
    forall U e, eqNodeSet U empty_BoolSet ->
                eqNodeSet (UNSAT_update U e) empty_BoolSet.
  Proof.
    intros U e h x y xy.
    unfold empty_BoolSet, UNSAT_update, diffBoolSet,
    andBoolSet, negBoolSet, "&&".
    specialize (h _ _ xy); unfold empty_BoolSet in h.
    rewrite h; auto.
  Qed.
    
  CoFixpoint build_pair_U_e U e :=
    match e with
    | s_one g => s_one (U, g)
    | s_cons g e => s_cons (U, g) (build_pair_U_e (UNSAT_update U (s_cons g e)) e)
    end.

  Lemma build_pair_U_e_s_hd:
    forall U e, s_hd (build_pair_U_e U e) = (U, s_hd e).
  Proof.
    intros U [g | g e]; simpl; auto.
  Qed.

  Lemma build_pair_U_e_s_one:
    forall U g, build_pair_U_e U (s_one g) = s_one (U, g).
  Proof.
    intros U g.
    rewrite Stream_decomposition_lemma at 1; simpl.
    auto.
  Qed.

  Lemma build_pair_U_e_s_cons:
    forall U g e, build_pair_U_e U (s_cons g e) =
                  s_cons (U, g)
                         (build_pair_U_e (UNSAT_update U (s_cons g e)) e).
  Proof.
    intros U g e.
    rewrite Stream_decomposition_lemma at 1; simpl.
    auto.
  Qed.

  Instance build_pair_U_e_compat: Proper pequiv build_pair_U_e.
  Proof.
    intros U V UV e f ef;apply  Stream_decompose_inv.
    revert  U V UV e f ef;cofix hx;intros U V UV e f ef.
    destruct ef  as [ x y xy | x y e f xy ef]; simpl.
    { left.
      split; unfold RelationPairs.RelCompFun; simpl; auto.
    }
    { right.
      split; unfold RelationPairs.RelCompFun; simpl; auto.
      (* PC : must rewrite in hypothesis to preserve guardedness *)
      assert (UU: UNSAT_update U (s_cons x e) =~=
                                 UNSAT_update V (s_cons y f)).
      { apply UNSAT_update_compat; auto.
        right; auto. }
      specialize (hx _ _ UU _ _ ef).
      do 2 rewrite <- Stream_decomposition_lemma in hx.
      apply hx; auto.
    }
  Qed.    

  Lemma build_pair_U_e_not_increasing_1:
    forall U (pU: Proper pequiv U) e (pe: Proper pequiv e) n,
      U n = false ->
      Always (fun x => fst (s_hd x) n = false)
             (build_pair_U_e U e).
  Proof.
    cofix h.
    intros U pU [g | g e] pe n hn.
    { rewrite build_pair_U_e_s_one.
      left; simpl; auto.
    }
    { rewrite build_pair_U_e_s_cons.
      right; simpl; auto.
      apply h.
      { apply UNSAT_update_compat; auto. }
      { inversion pe; auto. }
      assert (hh := UNSAT_update_le pU pe).
      specialize (hh _ _ (Equivalence_Reflexive n)).
      unfold Bool.Bool.le in hh; rewrite hn in hh.
      destruct (UNSAT_update U (s_cons g e) n);
        congruence.
    }    
  Qed.

  Lemma build_pair_U_e_not_increasing:
    forall U (pU: Proper pequiv U)
           e (pe: Proper pequiv e)
           (hU: leNodeSet U (UNSAT_init (s_hd e))),
    forall n,
      Always
        (fun x =>
           fst (s_hd x) n = false ->
           Always (fun x' => fst (s_hd x') n = false) x)
        (build_pair_U_e U e).
  Proof.
    cofix h.
    intros U pU [g | g e] pe hu n.
    { rewrite build_pair_U_e_s_one.
      left; simpl.
      intros hn; left; simpl; auto.
    }
    { rewrite build_pair_U_e_s_cons.
      right; simpl.
      intros hn; right; simpl; auto.
      { apply build_pair_U_e_not_increasing_1.
        apply UNSAT_update_compat; auto.
        inversion pe; auto.
        assert (hh := UNSAT_update_le pU pe).
        specialize (hh _ _ (Equivalence_Reflexive n)).
        unfold Bool.Bool.le in hh; rewrite hn in hh.
        destruct (UNSAT_update U (s_cons g e) n);
          congruence.
      }
      { apply h.
        apply UNSAT_update_compat; auto.
        inversion pe; auto.
        apply UNSAT_are_enabled; auto;
          inversion pe; auto.
      }
    }
  Qed.
 
  Lemma Eventually_forall:
    forall A (PSA: PartialSetoid A)
           (x: Stream (A := A)) (px: Proper pequiv x)
           (P: Node -> Stream (A := A) -> Prop) (pP: Proper pequiv P), 
      (forall n, Eventually (P n) x) ->
      (forall n, Always (fun x' => P n x' -> Always (P n) x') x) ->
      Eventually (fun x' => forall n, P n x') x.
  Proof.
    intros A PSA x px P pP he ha.
    cut (Eventually (fun x' => forall n, InA equiv n all_nodes -> P n x') x).
    { clear he ha.
      intros ok; induction ok.
      - left; intros n; apply H.
        apply all_nodes_prop.
      - right; apply IHok.
        inversion px; auto. }
    induction all_nodes as [| a all ih].
    { left; intros n h. 
      rewrite InA_nil in h; tauto. }
    { specialize (he a).
      induction he.
      { induction ih.
        - left; intros n hn; rewrite InA_cons in hn; destruct hn as [hn|hn].
          + rewrite hn; auto. 
          + auto.
        - right.
          apply IHih.
          + inversion px; auto.
          + specialize (ha a); apply Always_P_P in ha.
            specialize (ha H).
            apply Always_inv_2, Always_P_P in ha; auto.
          + intros n.
            specialize (ha n).
            apply Always_inv_2 in ha; auto.
      }
      { right.
        apply IHhe.
        - inversion px; auto.
        - intros n.
          specialize (ha n).
          apply Always_inv_2 in ha; auto.
        - inversion ih; subst.
          + left; intros n hn.
            specialize (H _ hn).
            specialize (ha n).
            apply Always_P_P in ha.
            specialize (ha H).
            apply Always_inv_2, Always_P_P in ha; auto.
          + auto.
      }
    }
  Qed.
    
  Lemma weakly_fair_enabled_becomes_empty:
    forall e (pe: Proper pequiv e),
      weakly_fair e ->
      forall U, Proper pequiv U ->
                leNodeSet U (UNSAT_init (s_hd e)) ->
                Eventually (fun x => eqNodeSet (fst (s_hd x)) empty_BoolSet)
                           (build_pair_U_e U e).
  Proof.
    intros e pe wf U pU hU.
    cut (Eventually (fun x =>
                       forall n, (fst (s_hd x)) n = false)
                    (build_pair_U_e U e)).
    { intros ok; induction ok.
      - left; intros x y xy.
        rewrite H; auto.
      - right; auto.
    }
    eapply Eventually_forall.
    { apply build_pair_U_e_compat; auto. }
    { intros n m nm x y xy.
      pose (g := s_hd x).
      assert (g =~= s_hd y).
      unfold g; apply hd_compat; auto.
      fold g.
      rewrite <- H.
      destruct g as [UU s].
      simpl.
      apply eq_compat; auto.
      destruct H as [H _].
      unfold RelationPairs.RelCompFun in H; simpl in H.
      assert (Proper pequiv UU) by proper_.
      rewrite nm; auto.
    }
    2: { intros n.
         apply build_pair_U_e_not_increasing; auto. }
    { intros n.
      specialize (wf n).
      case_eq (EN_b n e); intros en.
      { apply Always_P_P in wf.
        specialize (wf en).
        clear en; revert hU pU; revert U.
        induction wf; intros U hU pU.
        - destruct s as [g | g s].
          { unfold AN, ACT, NEUT, ACT_b, NEUT_b in H.
            destruct H; congruence. }
          rewrite build_pair_U_e_s_cons; simpl.
          right; left.
          rewrite build_pair_U_e_s_hd; simpl.
          rewrite <- AN_act_neut_b in H.
          unfold UNSAT_update, diffBoolSet, negBoolSet,
          andBoolSet, "&&"; rewrite H; destruct (U n); auto.
        - rewrite build_pair_U_e_s_cons.
          right; apply IHwf.
          inversion pe; auto.
          apply UNSAT_are_enabled; auto;
            inversion pe; auto.
          apply UNSAT_update_compat; auto.
      }
      { left.
        rewrite build_pair_U_e_s_hd; simpl.
        unfold UNSAT_init in hU.
        specialize (hU _ _ (Equivalence_Reflexive n)).
        unfold EN_b in en; rewrite en in hU.
        destruct (U n); auto.
      }
    }
  Qed.
    
  Lemma at_most_rounds_weakly_fair:
    forall e, Proper pequiv e -> weakly_fair e ->
              forall P n,
                at_most_rounds P n e -> Eventually P e.
  Proof.
    intros e pe wf P n.
    unfold at_most_rounds.
    pose (U := UNSAT_init (s_hd e)).
    assert (pU: Proper fequiv U).
    { unfold U; apply UNSAT_init_compat, hd_compat; auto. }
    assert (leU: leNodeSet U (UNSAT_init (s_hd e))).
    { unfold U; apply eqBoolSet_leBoolSet.
      fold U; auto. }
    fold U.
    clearbody U.
    revert wf pe pU leU; revert e U.
    induction n as [|n ih]; intros e U wf pe pU leU hr.
    { inversion hr; subst.
      left; auto.
      lia.
    }
    { dup wf emp.
      apply weakly_fair_enabled_becomes_empty with (U := U) in emp; auto.
      remember (build_pair_U_e U e) as bp.
      revert Heqbp hr wf pe pU leU; revert U e; induction emp;
        intros U e hbp hr wf pe pU leU.
      { rewrite hbp in H; rewrite build_pair_U_e_s_hd in H; simpl in H.
        inversion hr; subst.
        - left; auto.
        - contradict H2.
          apply UNSAT_update_empty; auto.
        - right; eapply ih.
          apply weakly_fair_inv in wf; auto.
          inversion pe; auto.
          3: apply H1.
          apply UNSAT_init_compat, hd_compat; inversion pe; auto.
          apply eqBoolSet_leBoolSet.
          apply UNSAT_init_compat, hd_compat; inversion pe; auto.
      }
      { inversion hr; subst.
        { left; auto. }
        { right.
          apply IHemp with (UNSAT_update U (s_cons g e0)); auto.
          rewrite build_pair_U_e_s_cons in hbp.
          inversion hbp; auto.
          apply weakly_fair_inv in wf; auto.
          inversion pe; auto.
          apply UNSAT_update_compat; auto.
          apply UNSAT_are_enabled; auto;
            inversion pe; auto.
        }
        { right.
          eapply ih.
          apply weakly_fair_inv in wf; auto.
          inversion pe; auto.
          3: apply H0.
          apply UNSAT_init_compat, hd_compat; inversion pe; auto.
          apply eqBoolSet_leBoolSet.
          apply UNSAT_init_compat, hd_compat; inversion pe; auto.
        }
      }
    }
  Qed.
      
  
End Rounds.  

Close Scope nat_scope.
Close Scope signature_scope.
Unset Implicit Arguments.
