(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.

(** **** Local Imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Stream.
From Padec Require Import Exec.
From Padec Require Import Fairness.

(* end show *)
Set Implicit Arguments.
Open Scope signature_scope.

(** * Specifications for self-stabilizing algorithms and
    silent self-stabilizing algorithms **)

Section Self_Stabilization.

  Context {Chans: Channels} {Net: Network} {Algo: Algorithm}.

  (** Assumption made on read-only variables of the algorithm *)
  Hypothesis assume: Stable_Assumption.

  (** In the sequel, we use:
   - [Daemon_fairness: Exec -> Prop] = specification of the Daemon Fairness
   (e.g. unfair daemon, weakly fair daemon, synchronous daemon ...)
   - [SPEC: Exec -> Prop] = specification of the algorithm
   - [LEG_STATE: Env -> Prop] = legitimate states for stelf-stabilization
   *)

  (** Self-stabilization: Closure w.r.t. legitimate states
  Definition closure (LEG_STATE: Env -> Prop): Prop :=
    forall (env env': Env),
      Assume env -> LEG_STATE env -> Step env' env -> LEG_STATE env'.
   *)
  Definition closure (Daemon_fairness: Exec -> Prop)
             (LEG_STATE: Env -> Prop): Prop :=
    forall (e: Exec),
      Assume (s_hd e) -> is_exec e -> Daemon_fairness e ->
      LEG_STATE (s_hd e) -> Always (fun e => LEG_STATE (s_hd e)) e.

  (** Self-stabilization: Convergence to legitimate states *)
  Definition convergence (Daemon_fairness: Exec -> Prop)
             (LEG_STATE: Env -> Prop): Prop :=
    forall (e: Exec),
      Assume (s_hd e) -> is_exec e -> Daemon_fairness e ->
      Eventually (fun suf: Exec => LEG_STATE (s_hd suf)) e.

  (** Self-stabilization: Meets specification when initialized in
       legitimate states *)
  Definition spec_ok (Daemon_fairness: Exec -> Prop)
             (LEG_STATE: Env -> Prop) (SPEC: Exec -> Prop): Prop :=
    forall (e: Exec),
      Assume (s_hd e) -> is_exec e -> Daemon_fairness e ->
      LEG_STATE (s_hd e) -> SPEC e.

  (** Self-stabilization: full property *)
  Definition self_stab (Daemon_fairness: Exec -> Prop)
             (LEG_STATE: Env -> Prop) (SPEC: Exec -> Prop): Prop :=
    closure Daemon_fairness LEG_STATE /\ convergence Daemon_fairness LEG_STATE
    /\ spec_ok Daemon_fairness LEG_STATE SPEC.

  Definition self_stabilization (Daemon_fairness: Exec -> Prop)
             (SPEC: Exec -> Prop): Prop :=
    exists (LEG_STATE: Env -> Prop), self_stab Daemon_fairness LEG_STATE SPEC.

  Lemma self_stabilization_alt:
    forall Daemon_fairness SPEC LEG_STATE,
      self_stab Daemon_fairness LEG_STATE SPEC <->
      forall e, Assume (s_hd e) -> is_exec e -> Daemon_fairness e ->
                ( LEG_STATE (s_hd e) ->
                  Always (fun e => LEG_STATE (s_hd e)) e )
                (* closure *)
                /\
                Eventually (fun suf: Exec => LEG_STATE (s_hd suf)) e
                (* convergence *)
                /\
                ( LEG_STATE (s_hd e) -> SPEC e ) (* specification *) .
  Proof.
    intros; unfold self_stab, closure, convergence, spec_ok; intuition;
      apply H; auto.
  Qed.

  Section Silence.
    (** ** Self-stabilization for Silent Algorithm *)

    (** Silent Algorithm: an algorithm for which any execution is finite
     is equivalent to: an algorithm converges to terminal configuration
     *)
    Definition silence (Daemon_fairness: Exec -> Prop): Prop :=
      forall (e: Exec), Assume (s_hd e) -> is_exec e ->
                        Daemon_fairness e -> finite_exec e.

    Lemma convergence_silence (Daemon_fairness: Exec -> Prop):
      convergence Daemon_fairness terminal <-> silence Daemon_fairness.
    Proof.
      split; intros h e ass ise df; specialize (h e ass ise df).
      - apply finite_exec_E_terminal_converse; auto.
      - apply finite_exec_E_terminal; auto.
    Qed.

    Lemma self_stabilization_for_silence (Daemon_fairness: Exec -> Prop)
          (SPEC: Exec -> Prop):
      silence Daemon_fairness ->
      ( self_stabilization Daemon_fairness SPEC <->
        spec_ok Daemon_fairness (fun g => terminal g) SPEC ).
    Proof.
      unfold silence, self_stabilization, self_stab.
      intros silent.
      split.
      (* -> *)
      { intros [LS [clos [conv spok]]] ex ass is_e df_e term.
        apply spok; try assumption.
        induction (conv ex) as [| g e IH]; auto.
        destruct term as [_ term].
        specialize (term (s_hd e)).
        contradict term; apply is_exec_inv_1,is_e.
      }
      (* <- *)
      { intros spok.
        exists (fun e: Env => terminal e).
        split; [|split].
        - intros [g | g e] ass ise daem term.
          left; auto.
          apply is_exec_inv_1' in ise; tauto.
        - rewrite convergence_silence; auto.
        - apply spok.
      }
    Qed.

    Lemma silence_unfair_weakly_fair:
      silence unfair_daemon ->
      forall e, Assume (s_hd e) -> is_exec e ->
                unfair_daemon e -> weakly_fair e.
    Proof.
      intros _t sil e ass ise uf.
      apply weakly_fair_finite_exec; auto.
    Qed.
    
    (** ** Partial Correctness
     Any terminal configuration satisfies the specification.
     *)
    Definition P_correctness (SPEC_env: Env -> Prop): Prop :=
      forall (g: Env), Assume g -> terminal g -> SPEC_env g.

    Definition only_one (SPEC_env: Env -> Prop) (ex: Exec): Prop :=
      match ex with
      | s_one g => SPEC_env g
      | s_cons _ _ => False
      end.

    (** ** Self-Stabilization for Silent Algorithm *)
    Theorem silent_self_stabilization
            (Daemon_fairness: Exec -> Prop) (SPEC_env: Env -> Prop):
      silence Daemon_fairness -> P_correctness SPEC_env ->
      self_stabilization Daemon_fairness (only_one SPEC_env).
    Proof.
      intros PC Term.
      rewrite self_stabilization_for_silence; auto.
      intros [g | g e] ass is_e df_e t_e; simpl in *.
      - apply Term; auto.
      - contradict t_e; simpl; eapply Step_not_terminal;
          apply is_exec_inv_1, is_e.
    Qed.

    Lemma silent_terminal_SPEC
          (Daemon_fairness: Exec -> Prop) (SPEC_env: Env -> Prop):
      (forall g, Assume g -> terminal g -> Daemon_fairness (s_one g))
      -> silence Daemon_fairness ->
      self_stabilization Daemon_fairness (only_one SPEC_env) ->
      P_correctness SPEC_env.
    Proof.
      intros p_daem sil [LS ss].
      intros g ass tg.
      cut (only_one SPEC_env (s_one g)).
      { simpl; tauto. }
      apply ss; auto.
      left; auto.
      destruct ss as [_ [cv _]].
      specialize (cv (s_one g)).
      assert (a: Assume (s_hd (s_one g))) by apply ass.
      assert (i: is_exec (s_one g)) by (left; auto).
      specialize (cv a i (p_daem _ ass tg)).
      inversion cv; auto.
    Qed.

  End Silence.

  (** ** Proving convergence to a set of legitimate configurations can be done
      using Acc and well-foundedness. Namely, if from any configuration with
      assumption met the transitive closure of Step started from a non
      legitimate configuration is Acc (well-founded) then the execution
      starting from this configuration converges to Leg_State.

       Note that this property is made to be used with the weakest daemon
       fairness assumption, i.e. the unfair daemon (True).
   *)
  Section Convergence.

    Variable (Leg_State: Env -> Prop)
             (Leg_State_compat: Proper fequiv Leg_State)
             (Leg_State_dec:
                forall g, Proper pequiv g -> Assume g ->
                          {Leg_State g} + {~Leg_State g}).

    Hypothesis (term_is_leg:
                  forall g, Proper pequiv g -> Assume g ->
                            terminal g -> Leg_State g).

    Lemma Convergence_to_Legitimate':
      forall (g: Env),
        Acc (fun g' g => ~Leg_State g /\ Step g' g) g ->
        forall (e: Exec) (ise: is_exec e) (assume: Assume (s_hd e)),
          s_hd e = g -> Eventually (fun suf: Exec => Leg_State (s_hd suf)) e.
    Proof.
      intros g acc.
      induction acc using Acc_ind_dep.
      intros [g | g e] ise ass shd; subst.
      left; simpl.
      apply term_is_leg; auto.
      apply is_exec_proper in ise; ssplit ise; auto.
      inversion ise; auto.
      assert (pg: Proper pequiv g) by
          (apply is_exec_proper in ise; ssplit ise; apply ise).
      destruct (Leg_State_dec pg); auto.
      left; auto.
      right.
      eapply H; auto.
      split; auto.
      inversion ise; auto.
      inversion ise; auto.
      unfold Assume, ROEnv_part; eapply step_assume.
      apply Assume_RO_compat.
      apply ass.
      inversion ise; auto.
    Qed.

    Lemma Convergence_to_Legitimate:
      (forall (g: Env) (pg: Proper pequiv g) (assume: Assume g),
          Acc (fun g' g => ~Leg_State g /\ Step g' g) g) ->
      convergence unfair_daemon Leg_State.
    Proof.
      intros ok e ass ise df.
      apply Convergence_to_Legitimate' with (s_hd e); auto.
      apply ok; auto.
      apply is_exec_s_hd_compat, ise.
    Qed.

  (** Note on [Acc]:
<<
  Inductive Acc (A : Type) (R : A -> A -> Prop) (x : A) : Prop :=
        Acc_intro : (forall y : A, R y x -> Acc R y) -> Acc R x

  Acc ~Leg_State/\Step g0: given by
    forall g1, ~Leg_State g0 /\ Step g1 g0 (g0->g1) -> Acc g1
    ... becomes true iff no configuration g satisfies: 
             ~Leg_State gn /\ Step g gn
        namely when gn is Leg_State or
                    gn is a terminal configuration 
                    (and hence Leg_State, due to [term_is_leg]

  Namely, any execution from g0 reaches Leg_State.
>>
   *)
  End Convergence.

  (** ** Termination

     Any configuration that is reachable from an initial configuration
     [g] satisfying the assumptions can be reached within a finite
     number of steps

     (i.e., using [Step] a finite number of times)

     Note that this property is made to be used with the weakest daemon 
     fairness assumption, i.e.
     the unfair daemon (True).
   *)
  Definition termination: Prop :=
    forall (g: Env), Assume g -> Proper pequiv g -> Acc Step g.

  Lemma termination_is_silence: termination -> convergence unfair_daemon terminal.
  Proof.
    intros term.
    apply Convergence_to_Legitimate with (Leg_State := terminal); auto.
    - intros g pg _; destruct (terminal_dec g); [left | right]; auto.
    - intros g pg ass; specialize (term _ ass pg).
      clear ass.
      induction term as [g _ ih].
      apply Acc_intro; intros g' [nterm step].
      apply ih; auto.
      eapply Step_Proper_l, step.
  Qed.

End Self_Stabilization.

Close Scope signature_scope.
Unset Implicit Arguments.
