(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- June 2016
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Lia.

(** **** Local Imports *)
From Padec Require Import ZUtils.
From Padec Require Import NatUtils.
From Padec Require Import SetoidUtil.
From Padec Require Import Algorithm.
From Padec Require Import RelModel.
From Padec Require Import Stream.
From Padec Require Import Stream_Length.
From Padec Require Import Exec.
From Padec Require Import BoolSet.
From Padec Require Import Self_Stabilization.

(* end show *)
Open Scope nat_scope.
Open Scope signature_scope.
Set Implicit Arguments.

Section At_Steps.

  Context {Chans: Channels} {Net: Network} {Algo: Algorithm}.

  (** * Steps *)
  (** In this section, we give tools to compute time complexity in steps of
   an algorithm.
   In particular, we compute the number of steps required to reach some 
   predicate [P]. 
   *)
  Definition At_N_steps N: (Exec -> Prop) -> (Exec -> Prop) :=
    fun P e => length_until_P (A := Env) P e N.

  Instance At_N_steps_compat: Proper fequiv At_N_steps.
  Proof.
    unfold At_N_steps; intros N1 N2 eN P1 P2 eP e1 e2 ee.
    apply length_until_P_compat; auto.
  Qed.

  Lemma After_SN_steps_notP:
    forall N P (e: Exec), At_N_steps (S N) P e -> ~P e.
  Proof.
    intros until e. intros af; inversion af; auto.
  Qed.

  Lemma At_N_steps_Eventually:
    forall N P e, At_N_steps N P e -> Eventually P e.
  Proof.
    intros n P; induction n as [| n ih]; intros e af;
      inversion af; subst.
    - left; auto.
    - right.
      apply ih; auto.
  Qed.

  Lemma Reachable_Steps:
    forall (P: Exec -> Prop),
      (forall e, (Proper pequiv e -> P e) \/ ~ P e) ->
      forall e, Proper pequiv e -> Eventually P e ->
                exists N, At_N_steps N P e.
  Proof. apply Stream_Length.Reachable_Steps; auto. Qed.

  Lemma At_N_steps_unique:
    forall e P N1 N2,
      Proper pequiv e ->
      At_N_steps N1 P e -> At_N_steps N2 P e -> N1 = N2.
  Proof.
    intros e P n1.
    revert e;
      induction n1 as [| n1 ih]; intros e [| n2] pe af1 af2.
    - auto.
    - inversion af1; apply After_SN_steps_notP in af2; tauto.
    - inversion af2; apply After_SN_steps_notP in af1; tauto.
    - destruct e as [g | g e];
        inversion af1; inversion af2; subst; try tauto.
      apply f_equal, ih with e.
      ssplit pe; apply pe.
      assumption.
      assumption.
  Qed.

  Fixpoint count_steps_aux (P: Exec -> Prop) (e: Exec)
           (F: Finally P e): nat :=
    match F in Finally _ _ return  nat with
    | finally_now _ _ _ => 0
    | finally_not_yet _ _ F' => S (count_steps_aux F')
    end.

  Instance count_steps_aux_compat:
    Proper (respectful_dep_param
              (equiv (A:=nat))
              ((Exec -> Prop)::Exec::nil)
              (fequiv,(pequiv,tt))
              (inr Finally::nil)) (@count_steps_aux).
  Proof.
    intros P1 P2 hP e1 e2 he F1 F2.
    revert e2 he F2.
    induction F1 as [e1 hP1| g1 e1 nP F1 ih]; destruct F2 as [| g2 e2 nP2 F2];
      simpl in *; specialize (hP  _ _ he); auto; try tauto.
    apply f_equal, ih.
    inversion he; auto.
  Qed.

  Definition count_steps (P: Exec -> Prop) (e: Exec)
             (P_dec: Weak_Dec pequiv P) (pe: Proper pequiv e)
             (he: Eventually P e): nat :=
    count_steps_aux (Eventually_Dec_Finally' P_dec pe he).

  Instance count_steps_compat:
    Proper (respectful_dep_param
              (equiv (A:=nat))
              ((Exec -> Prop)::Exec::nil)
              (fequiv,(pequiv,tt))
              (inr (fun P e => Weak_Dec pequiv P)
                   ::inl (fun P e => Proper pequiv e)
                   ::inl Eventually :: nil))          (@count_steps).
  Proof.
    intros P1 P2 hP e1 e2 he P1_dec pe1 F1 P2_dec pec2 F2.
    apply count_steps_aux_compat; auto.
  Qed.

  Theorem steps_equivalence:
    forall (P: Exec -> Prop) (P_dec: Weak_Dec pequiv P),
    forall e (pe: Proper pequiv e) (he: Eventually P e) N,
      count_steps P_dec pe he = N <-> At_N_steps N P e.
  Proof.
    intros P P_dec e pe he n; split; intros h.
    { unfold count_steps in h.
      revert n h; induction (Eventually_Dec_Finally' P_dec pe he).
      - intros n h; simpl in h; subst.
        left; auto.
      - intros [| N] h.
        + simpl in h; inversion h.
        + right; auto.
          apply IHf.
          ssplit pe; apply pe.
          apply Finally_Eventually; assumption.
          inversion h; auto.
    }
    { unfold count_steps.
      revert h. revert n.
      induction (Eventually_Dec_Finally' P_dec pe he); simpl.
      - intros [| n] h; auto.
        inversion h; subst; tauto.
      - intros [| nn] h; inversion h; subst.
        tauto.
        apply f_equal, IHf.
        ssplit pe; apply pe.
        apply Finally_Eventually; assumption.
        assumption.
    }
  Qed.

  Lemma Finite_exec_Steps:
    forall e (ise: is_exec e),
      finite_exec e <->
      exists N, At_N_steps N (fun e => terminal (s_hd e)) e.
  Proof.
    intros e ise; split.
    { intros fin.
      eapply Reachable_Steps.
      intros ee.
      destruct (terminal_dec (s_hd ee)) as [T|T];
        [left; intros p; apply T, hd_compat, p|right]; auto.
      apply is_exec_proper; auto.
      apply finite_exec_E_terminal; assumption.
    }
    { intros [N st].
      induction st as [ e| g e N nterm st ih].
      destruct e as [g | g e];
        [left; auto| apply is_exec_inv_1' in ise; tauto].
      right; apply ih.
      apply is_exec_inv_2 with g; auto.
    }
  Qed.

  Lemma at_steps_concat:
    forall (P Q: Exec -> Prop) e,
      Always (fun e => Q e -> P e) e ->
      forall np, At_N_steps np P e ->
                 forall nq,
                   (forall w: Finally P e, At_N_steps nq Q (When w)) ->
                   At_N_steps (np + nq) Q e.
  Proof.
    intros P Q e al np hnp nq hnq.
    revert hnq hnp al; revert nq e;
      induction np as [| np ih]; intros nq e hnp hnq al.
    - inversion hnq; subst.
      specialize (hnp (finally_now _ _ H)); simpl in hnp.
      apply hnp.
    - destruct e as [g | g e]; inversion hnq; subst; try tauto.
      assert (tmp: S np + nq = S (np + nq)) by lia; rewrite tmp; clear tmp.
      right.
      { apply Always_P_P in al; tauto. }
      { apply ih.
        intros w.
        specialize (hnp (finally_not_yet _ H2 w));
          simpl in hnp; apply hnp.
        assumption.
        eapply Always_inv_2, al.
      }
  Qed.

  Lemma steps_induction:
    forall (Q: Exec -> Prop) (Pot: nat -> Exec -> Prop)
           (Q_means_Always: forall e, Q e -> Always Q e)
           (m: nat -> nat),
      (forall e, Q e -> Pot 0 e -> forall n, Pot n e) ->
      (forall e k, Q e -> Pot (S k) e ->
                   At_N_steps (m (S k)) (Pot k) e) ->
      forall n e, Q e -> Pot n e ->
                  At_N_steps (sum m n) (Pot 0) e.
  Proof.
    intros Q Pot Q_means_Always m incl ind_step.
    induction n as [| n ih].
    - intros e qe p0; left; apply p0.
    - intros e qe pS.
      simpl.
      eapply at_steps_concat with (P := Pot n).
      { apply Always_impl_P with Q.
        2: apply Q_means_Always, qe.
        intros s qs ps; apply incl; auto.
      }
      { apply ind_step; auto. }
      { intros w; apply ih.
        - apply Always_P_P, Always_When.
          apply Q_means_Always, qe.
        - apply When_P.
      }
  Qed.

  Lemma steps_induction_le:
    forall (Q: Exec -> Prop) (Pot: nat -> Exec -> Prop)
           (**)
           (Pot_wd: forall n e, Q e -> {Pot n e} + { ~Pot n e})
           (Pot_compat: Proper fequiv Pot)
           (**)
           (Q_means_Always: forall e, Q e -> Always Q e)
           (Q_means_compat: forall e, Q e -> Proper pequiv e)
           (m: nat -> nat),
      (forall e, Q e -> Pot 0 e -> forall n, Pot n e) ->
      (forall e k, Q e -> Pot (S k) e ->
                   exists n, n <= m (S k) /\
                             At_N_steps n (Pot k) e) ->
      forall N e, Q e -> Pot N e ->
                  exists n, n <= sum m N /\
                            At_N_steps n (Pot 0) e.
  Proof.
    intros Q Pot Pot_wd Pot_compat Q_means_Always q_means_compat
           m incl ind_step.
    induction N as [| N ih].
    - intros e qe p0; exists 0; split; [auto | left; apply p0].
    - intros e qe pS; simpl.
      destruct (ind_step e N) as [nS [hnS af]]; auto.
      clear ind_step.
      assert (w: Finally (Pot N) e).
      {
        eapply Eventually_Dec_Finally''; auto.
        2: apply At_N_steps_Eventually with nS; auto.
        apply Forever_forall_Type with Q.
        apply Pot_wd.
        apply Q_means_Always; auto.
      }
      destruct (ih (When w)) as [nN [hnN afN]].
      { apply Always_P_P, Always_When, Q_means_Always, qe. }
      { apply When_P. }
      exists (nS + nN); split; [lia|].
      eapply at_steps_concat with (P := Pot N); auto.
      { apply Always_impl_P with Q.
        2: apply Q_means_Always, qe.
        intros s qs ps; apply incl; auto.
      }
      { intros ww.
        revert afN; apply At_N_steps_compat; auto.
        apply When_compat.
        apply Pot_compat; auto.
        apply q_means_compat, qe.
      }
  Qed.

  (* Generic step complexity :
     we bound the step complexity (until P is achieved)
     of the set of streams satisfying Q
   a typical P is (terminal) or a set of legitimate configurations
   a typical Q is is_exec /\ initial assumptions /\ scheduling assumptions
   *)
  Definition Steps_complexity N (P Q: Exec -> Prop) :=
    forall exec, Q exec -> exists n, n <= N /\
                                     At_N_steps n P exec.

  (** Compatibility requires more assumptions: *)
  Lemma Steps_complexity_compat:
    forall n1 n2 (en: n1 == n2)
           P1 P2 (eP: fequiv P1 P2)
           (Q1 Q2: Exec -> Prop) (eQ: fequiv Q1 Q2)
           (hp1: forall e, Q1 e -> Proper pequiv e)
           (hp2: forall e, Q2 e -> Proper pequiv e),
      Steps_complexity n1 P1 Q1 == Steps_complexity n2 P2 Q2.
  Proof.
    intros n nn en P1 P2 eP Q1 Q2 eQ hp1 hp2.
    rewrite <- en; clear nn en.
    split; intros h e qe; specialize (h e).
    { rewrite <- (eQ _ _ (hp2 _ qe)) in qe.
      specialize (h qe).
      destruct h as [n' [len st]].
      exists n'.
      split; auto.
      eapply At_N_steps_compat.
      reflexivity.
      symmetry; apply eP.
      apply hp1, qe.
      apply st.
    }
    { rewrite (eQ _ _ (hp1 _ qe)) in qe.
      specialize (h qe).
      destruct h as [n' [len st]].
      exists n'.
      split; auto.
      eapply At_N_steps_compat.
      reflexivity.
      apply eP.
      apply hp2, qe.
      apply st.
    }
  Qed.

  Lemma Steps_complexity_reach:
    forall N P Q, Steps_complexity N P Q ->
                  forall e, Q e -> Eventually P e.
  Proof.
    intros N P Q st e Qe.
    specialize (st _ Qe).
    destruct st as [n [_ stt]].
    apply Reachable_steps_converse with n; auto.
  Qed.

  Lemma At_N_steps_monotonic:
    forall (Q P P': Exec -> Prop)
           (inclP: forall e, Always Q e -> P e -> P' e)
           n e,
      Always Q e -> Always (fun e => P' e \/ ~P' e) e ->
      At_N_steps n P e ->
      exists n', n' <= n /\ At_N_steps n' P' e.
  Proof.
    intros Q P P' inclP n e hq wd hn.
    induction hn as [ e hp | g e N nterm st ih].
    { exists 0; split; [lia|].
      left; apply inclP; auto.
    }
    { destruct (Always_P_P wd) as [hp'|hp'].
      - exists 0; split; [lia|].
        left; auto.
      - apply Always_inv_2 in hq.
        apply Always_inv_2 in wd.
        destruct (ih hq wd) as [n' [hn' st']].
        exists (S n'); split; [lia|].
        right; auto.
    }
  Qed.

  Lemma Steps_complexity_monotonic:
    forall
      (Q Q': Exec -> Prop) (inclQ: forall e, Q' e -> Q e)
      (Q_means_Always: forall e, Q e -> Always Q e)
      (P P': Exec -> Prop) (inclP: forall e, Q e -> P e -> P' e)
      (clas: forall e, Q e -> P' e \/ ~P' e)
      N,
      Steps_complexity N P Q -> Steps_complexity N P' Q'.
  Proof.
    intros Q Q' inclQ alQ P P' inclP clas N hN.
    intros e hq.
    assert (hqq: Q e).
    { apply inclQ, hq. }
    destruct (hN e hqq) as [n [hn stN]].
    apply At_N_steps_monotonic with (Q := Q) (P' := P') in stN.
    - destruct stN as [n' [hn' stn']].
      exists n'; split; [lia| auto].
    - intros ee al pee; apply inclP; [apply Always_P_P|]; auto.
    - auto.
    - apply Always_MP with Q.
      apply Forall_Always, clas.
      apply alQ, hqq.
  Qed.

  Lemma Steps_complexity_Acc:
    forall N (P Q: Env -> Prop),
      (forall g g', Q g -> Step g' g -> Q g') ->
      Steps_complexity N (fun e => P (s_hd e))
                       (fun e => Q (s_hd e) /\ is_exec e) ->
      forall g, Q g -> Acc (fun g' g => ~P g /\ Step g' g) g.
  Proof.
    intros N P Q Q_stable stN.
    pose (Pot := fun n g =>
                   forall e, s_hd e = g ->
                             is_exec e ->
                             exists n' : nat,
                               n' <= n /\
                               At_N_steps n' (fun e => P (s_hd e)) e).
    assert (PotN: forall g, Q g -> Pot N g).
    { intros hq g e hg ise.
      apply stN; split; auto.
     rewrite hg; auto.
    }
    clear stN.

    intros g Qg; specialize (PotN _ Qg).
    revert PotN Qg; revert g; induction N as [| N ih].
    { intros g h0 qg.
      apply Acc_intro; intros g' [nP st].
      specialize (h0 (sync_exec g)).
      destruct h0 as [n0 [hn0 st0]].
      apply sync_exec_hd.
      apply   sync_exec_is_exec, st.
      assert (tmp: n0 = 0) by lia; subst.
      inversion st0; subst.
      rewrite sync_exec_hd in H; tauto.
    }
    { intros g hN Qg.
      apply Acc_intro; intros g' [nP st].
      apply ih.
      - intros e he ise.
        specialize (hN (s_cons g e)).
        destruct hN as [n' [hn' stn']].
        simpl; reflexivity.
        right.
        rewrite he; auto.
        apply ise.
        inversion stn'; subst.
        + apply Step_not_terminal in st; simpl in H; tauto.
        + exists n; split; [lia|]; auto.
      - apply Q_stable with g; auto.
    }
  Qed.

  Lemma Steps_complexity_Acc_term:
    forall N (Q: Env -> Prop),
      (forall g g', Q g -> Step g' g -> Q g') ->
      Steps_complexity N (fun e => terminal (s_hd e))
                       (fun e => Q (s_hd e) /\ is_exec e) ->
      forall g, Q g -> Acc (fun g' g => Step g' g) g.
  Proof.
    intros N Q Q_stable stN.
    intros g qg.
    eapply Morphisms_Prop.Acc_rel_morphism.
    3: apply Steps_complexity_Acc with (P := fun g => terminal g)
                                       (Q := Q) (N := N); auto.
    3: apply qg.
    2: reflexivity.
    clear qg g.
    intros g g'; simpl; split.
    intros st; split; auto; apply Step_not_terminal with g; auto.
    intros [nt st]; auto.
  Qed.

  Lemma Steps_complexity_induction:
    forall (Q: Exec -> Prop) (Pot: nat -> Exec -> Prop)
           (**)
           (Pot_wd: forall n e, Q e -> {Pot n e} + { ~Pot n e})
           (Pot_compat: Proper fequiv Pot)
           (**)
           (Q_means_Always: forall e, Q e -> Always Q e)
           (Q_means_compat: forall e, Q e -> Proper pequiv e)
           (m: nat -> nat),
      (forall e, Q e -> Pot 0 e -> forall n, Pot n e) ->
      (forall k, Steps_complexity (m (S k)) (Pot k)
                                  (fun e => Q e /\ Pot (S k) e)) ->
      forall n, Steps_complexity (sum m n) (Pot 0)
                                 (fun e => Q e /\ Pot n e).
  Proof.
    intros until m; intros incl ind_step n e [eq P0e].
    eapply steps_induction_le; auto.
    assumption.
    clear e eq P0e n.
    intros e k qe PSke.
    apply ind_step; split; auto.
  Qed.

  Lemma Steps_complexity_add:
    forall (Q P1 P2: Exec -> Prop)
           (P1_wd: forall e, Q e -> {P1 e} + { ~P1 e})
           (P2_wd: forall e, Q e -> {P2 e} + { ~P2 e})
           (P1_compat: Proper fequiv P1)
           (P2_compat: Proper fequiv P2)
           (Q_means_Always: forall e, Q e -> Always Q e)
           (Q_means_compat: forall e, Q e -> Proper pequiv e)
           n1 n2,
      (forall e, Q e -> P2 e -> P1 e) ->
      Steps_complexity n1 P1 Q ->
      Steps_complexity n2 P2 (fun e => Q e /\ P1 e) ->
      Steps_complexity (n1 + n2) P2 Q.
  Proof.
    intros until n2; intros incl h1 h2.
    pose (Pot := fun n =>
                   match n with 0 => P2
                           | S nn => match nn with 0 => P1
                                              | S _ => fun _ => True
                                     end end).
    assert (wd: forall n e, Q e -> {Pot n e} + {~Pot n e}).
    { intros [|[|n]]; simpl; auto. }
    assert (pc: Proper fequiv Pot).
    { intros [|[|n]] [|[|nn]]; simpl; try lia; intros; auto.
      subst; simpl; intros e ee _; tauto.
    }
    pose (m := fun n =>
                 match n with 0 => 0
                         | S nn => match nn with 0 => n2
                                            | S _ => n1 end end).
    assert (incl_: forall e, Q e -> Pot 0 e -> forall n, Pot n e).
    { simpl; intros ee qee p2 [|[|n]]; simpl; auto. }
    assert (step: forall k,
               Steps_complexity (m (S k)) (Pot k)
                                (fun e => Q e /\ Pot (S k) e)).
    { intros [|[|k]]; simpl.
      apply h2.
      intros e [qe _].
      destruct (h1 _ qe) as [x [hx hhx]].
      exists x; split; [lia|]; auto.
      intros ee [_ p1ee]; exists 0; split; [lia|left; auto].
    }
    assert (res := Steps_complexity_induction (Q := Q) (Pot := Pot)
           wd pc Q_means_Always Q_means_compat m incl_ step 2).
    simpl in res.
    intros e qe.
    destruct (res e) as [x [hx hhx]].
    split; auto.
    exists x; split; [lia|]; auto.
  Qed.

  (** Execution and cost:

    we endow an execution with a natural number called its cost.
    The number of steps of the execution before reaching some desirable 
    property P is upper bounded by its cost provided that:
    - the cost decreases when not in P
    - at cost 0, P is reached

    We proceed by induction on the cost of the head of the execution.
    - [cost_0] provides the base case.
    - [cost_decreases] provides the induction step.
  *)
  Lemma exec_bounded_length:
    forall (Q: Exec -> Prop) (Q_preserved: forall g e, Q (s_cons g e) ->
                                                       Q e)
           (P: Exec -> Prop) (P_clas: forall e, Q e -> P e \/ ~ P e)
           (P_Q: forall g, Q (s_one g) -> P (s_one g)),
    forall (cost: Exec -> nat)
           (cost_decreases: forall g e,
               Q (s_cons g e) -> ~ P (s_cons g e) ->
               cost e < cost (s_cons g e))
           (cost_0: forall e, Q e -> cost e = 0 -> P e),
    forall e, Q e -> exists ne, At_N_steps ne P e /\ ne <= cost e.
  Proof.
    intros Q Q_preserved P P_clas P_Q cost cost_decreases cost_0 e hQ.
    assert (C: cost e <= cost e) by lia.
    revert C.
    pose (C := cost e).
    fold C; unfold C at 1.
    clearbody C.
    revert hQ; revert e.
    induction C as [| C ih].
    { intros e hQ hc.
      exists 0; split; [|lia].
      left.
      apply cost_0; auto; lia.
    }
    { intros [g | g e] hQ hc.
      { exists 0.
        split; [|lia].
        left; auto. }
      case (P_clas _ hQ); intros lP.
      { exists 0; split; [|lia].
        left; auto. }
      { destruct (ih e) as [ne [lne hC]].
        eapply Q_preserved; eauto.
        specialize (cost_decreases _ _ hQ lP);
          lia.
        exists (S ne).
        split; [|lia].
        right; auto.
      }
    }
  Qed.

  Theorem Steps_complexity_Bound_Cost:
    forall (Q: Exec -> Prop)
           (Q_preserved: forall g e, Q (s_cons g e) -> Q e)
           (P: Exec -> Prop)
           (P_clas: forall e, Q e -> P e \/ ~ P e)
           (P_Q: forall g, Q (s_one g) -> P (s_one g)),
    forall (cost: Exec -> nat)
           (cost_decreases: forall g e,
               Q (s_cons g e) -> ~ P (s_cons g e) ->
               cost e < cost (s_cons g e))
           (cost_0: forall e, Q e -> cost e = 0 -> P e),
    forall (N: nat)
           (cost_bound: forall e, Q e -> cost e <= N),
      Steps_complexity N P Q.
  Proof.
    intros.
    intros e hq.
    specialize (cost_bound _ hq).
    eapply exec_bounded_length in hq; eauto.
    destruct hq as [ne [st bst]].
    exists ne; split; auto.
    lia.
  Qed.
  
  (** If we can exhibit an execution starting with N non-legitimate
  configurations, then the algorithm cannot have a step complexity
  less than N *)
  Theorem Non_Leg_Steps_min_bound :
    forall exec N (P Q: Exec -> Prop),
      Proper pequiv exec ->
      Q exec ->
      At_N_steps N P exec ->
      forall n, n<N -> ~ Steps_complexity n P Q.
  Proof.
    intros exec N P Q pe hQ hN.
    intros n hn hsc.
    destruct (hsc exec hQ) as [nn [hln  hnn]].
    assert (hh:=At_N_steps_unique pe hN hnn).
    lia.
  Qed.

End At_Steps.

Section Before_Steps. (* useless?? *)

  Context {Chans: Channels}  {Net: Network} {Algo: Algorithm}.

  Inductive Before_N_steps: nat -> (Exec -> Prop) -> (Exec -> Prop) :=
  | at_P: forall (P: Exec -> Prop) e n, P e -> Before_N_steps n P e
  | add_step: forall (P: Exec -> Prop) g e n,
      Before_N_steps n P e -> Before_N_steps (S n) P (s_cons g e).

  Lemma Before_N_steps_equiv:
    forall P e,
      Always (fun e => P e \/ ~P e) e ->
      forall n, Before_N_steps n P e <->
                exists n', n' <= n /\ At_N_steps n' P e.
  Proof.
    intros P e P_clas n.
    revert P_clas; revert e; induction n as [| n ih]; split.
    - intros ll; exists 0; split; [lia|].
      left; inversion ll; auto.
    - intros [n' [hn' ss]].
      assert (n' = 0) by lia; subst.
      left; inversion ss; auto.
    - intros ll.
      destruct (Always_P_P P_clas) as [hP|hP].
      + exists 0; split; [lia | left]; auto.
      + inversion ll; subst; try tauto.
        rewrite ih in H0.
        destruct H0 as [n' [hn' atn']].
        exists (S n'); split; [lia|right]; auto.
        apply Always_inv_2 in P_clas; auto.
    - intros [n' [hn' ss]].
      inversion ss; subst.
      + left; auto.
      + right.
        rewrite ih.
        exists n0; split; [lia|]; auto.
        apply Always_inv_2 in P_clas; auto.
  Qed.

End Before_Steps.

Section Nb_Steps.

  Context {Chans: Channels} {Net: Network} {Algo: Algorithm}.


  (** Number of steps of with Property P before reaching T *)
  Fixpoint Nb_Steps (P: Exec -> bool) (T: Exec -> Prop)
           (e: Exec) (f: Finally T e): nat :=
    match f in Finally _ _ return nat with
    | finally_now _ _ _ => 0
    | @finally_not_yet _ _ g' e' _ F' =>
      if P (s_cons g' e') then S (Nb_Steps P F')
      else Nb_Steps P F'
    end.


  Instance Nb_Steps_compat:
    Proper (respectful_dep_param
              (equiv (A:=nat))
              ((Exec -> bool)::(Exec -> Prop)::Exec::nil)
              (fequiv,(fequiv,(pequiv,tt)))
              (inr (fun P T e => Finally T e)::nil)) Nb_Steps.
  Proof.
    intros P1 P2 eP T1 T2 eT e1 e2 ee f1 f2; simpl.
    revert e2 ee f2 eP eT; revert P2 T2;
      induction f1 as [e1 t1 | g1 e1 nt1 f1 ih];
      intros P2 T2 e2 ee f2 eP eT;
      destruct f2 as [e2 t2 | g2 e2 nt2 f2].
    - simpl; auto.
    - contradict nt2; revert t1; apply eT, ee.
    - contradict nt1; revert t2; apply eT, ee.
    - simpl.
      assert (hm := eP _ _ ee); rewrite <- hm; clear hm.
      destruct (P1 (s_cons g1 e1)).
      + apply f_equal, ih; auto.
        ssplit ee; apply ee.
      + apply ih; auto.
        ssplit ee; apply ee.
  Qed.

  (* Upper bound on Nb_Steps *)
  Inductive Nb_Steps_le (P: Exec -> bool) (T: Exec -> Prop):
    nat -> Exec -> Prop :=
  | at_T: forall k e, T e -> Nb_Steps_le P T k e
  | add_stepP: forall k g e, Nb_Steps_le P T k e ->
                             P (s_cons g e) = true ->
                             Nb_Steps_le P T (S k) (s_cons g e)
  | add_stepNP: forall k g e, Nb_Steps_le P T k e ->
                              P (s_cons g e) = false ->
                              Nb_Steps_le P T k (s_cons g e).

  Instance Nb_Steps_le_compat: Proper fequiv Nb_Steps_le.
  Proof.
    intros P1 P2 eP T1 T2 eT nn n en.
    rewrite en; clear nn en.
    revert eT eP; revert P1 P2 T1 T2.
    cut (forall P1 P2 T1 T2 e1 e2,
            P1 =~= P2 -> T1 =~= T2 ->
            e1 =~= e2 -> Nb_Steps_le P1 T1 n e1 ->
            Nb_Steps_le P2 T2 n e2).
    { intros ok P1 P2 T1 T2 eT eP e1 e2 ee; split; intros h.
      eapply ok  with (e1 := e1); eauto.
      apply ok with (P1 := P2) (T1 := T2) (e1 := e2); auto;
        symmetry; auto.
    }
    { intros P1 P2 T1 T2 e1 e2 eP eT ee nb.
      revert ee; revert e2; induction nb; intros e2 ee.
      - apply at_T.
        revert H; apply eT, ee.
      - destruct e2 as [gg | gg e2]; dup ee ee'; ssplit ee'.
        apply add_stepP.
        apply IHnb, ee'.
        rewrite <- H; symmetry; apply eP, ee.
      - destruct e2 as [gg | gg e2]; dup ee ee'; ssplit ee'.
        apply add_stepNP.
        apply IHnb, ee'.
        rewrite <- H; symmetry; apply eP, ee.
    }
  Qed.

  Lemma Nb_Steps_Finally:
    forall P T e n, Weak_Dec pequiv T ->
                    Proper pequiv e -> Nb_Steps_le P T n e ->
                    Finally T e.
  Proof.
    intros P T e n wd pe nst.
    apply Eventually_Dec_Finally'; auto.
    induction nst.
    - left; auto.
    - right.
      apply IHnst.
      ssplit pe; apply pe.
    - right.
      apply IHnst.
      ssplit pe; apply pe.
  Qed.

  Lemma Nb_Steps_equiv:
    forall P T e (f: Finally T e),
    forall n, Nb_Steps_le P T n e <-> Nb_Steps P f <= n.
  Proof.
    intros P T e f; induction f as [e t | g e nt f ih]; simpl.
    { intros n; split; intros h; [lia|].
      apply at_T; auto.
    }
    { intros n; split; intros h.
      { inversion h; subst; try tauto;
          case (Bool.bool_dec (P (s_cons g e)) true); intros hm.
        - rewrite hm.
          apply le_n_S.
          apply ih; auto.
        - tauto.
        - rewrite H3 in hm.
          contradict hm; auto with bool.
        - rewrite H3.
          apply ih; auto.
      }
      { case (Bool.bool_dec (P (s_cons g e)) true); intros hm.
        - rewrite hm in h.
          destruct n as [| n]; [lia|].
          apply add_stepP; auto.
          rewrite ih; auto.
          lia.
        - rewrite Bool.not_true_iff_false in hm; rewrite hm in h.
          apply add_stepNP; auto.
          apply ih; auto.
      }
    }
  Qed.

  Lemma Nb_Steps_0:
    forall P T e,  Weak_Dec pequiv T ->
                   Proper pequiv e ->
                   Always ( fun e => match e with
                                     | s_one _ => True
                                     | s_cons g e =>
                                       P (s_cons g e) = false -> T e -> T (s_cons g e) end) e ->
                   Nb_Steps_le P T 0 e ->
                   T e.
  Proof.
    intros P T e wd pe al nst.
    assert (f := Nb_Steps_Finally wd pe nst).
    rewrite (Nb_Steps_equiv _ f) in nst.
    assert (n0: Nb_Steps P f = 0) by lia; clear nst.
    induction f; simpl in n0.
    { auto. }
    case (Bool.bool_dec (P (s_cons a s)) true); intros hm.
    rewrite hm in n0.
    { lia. }
    { rewrite Bool.not_true_iff_false in hm.
      rewrite hm in n0.
      dup al preserve.
      apply Always_P_P in preserve.
      apply preserve; auto.
      apply IHf; auto.
      ssplit pe; apply pe.
      inversion al; auto.
    }
  Qed.

  Lemma Nb_Steps_more:
    forall P T n e, Weak_Dec pequiv T -> Proper pequiv e ->
                    Nb_Steps_le P T n e -> forall n', n' >= n -> Nb_Steps_le P T n' e.
  Proof.
    intros P T n e wd pe nst n' hn'.
    assert (f := Nb_Steps_Finally wd pe nst).
    rewrite (Nb_Steps_equiv _ f) in nst.
    assert (Nb_Steps P f <= n') by lia.
    rewrite Nb_Steps_equiv.
    apply H.
  Qed.

  Lemma Nb_Steps_le_wd:
    forall P T e, Finally T e ->
                  forall n, {Nb_Steps_le P T n e} +
                            {~Nb_Steps_le P T n e}.
  Proof.
    intros P T e f n.
    pose (N := Nb_Steps P f).
    case (Compare_dec.le_dec N n); intros g; unfold N in g;
      rewrite <- Nb_Steps_equiv in g; auto.
  Qed.

End Nb_Steps.

Close Scope nat_scope.
Close Scope signature_scope.
Unset Implicit Arguments.
