(**
*Certified Round Complexity of Self-Stabilizing Algorithms*
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import ZArith.
From Coq Require Import Bool.
From Coq Require Import Lia.

(** **** Local Imports *)
From Padec Require Import OptionUtil.
From Padec Require Import SetoidUtil.
From Padec Require Import RelationA.
From Padec Require Import Count.
From Padec Require Import ListUtils.
From Padec Require Import All_In.


(* end show *)
Open Scope nat_scope.
Open Scope signature_scope.
Set Implicit Arguments.

(** * Graph: 
    Tools for connected graph, path, distance and diameter *)
Section Graph.

  Context {A} `{SA: Setoid A} {DA: Decider (equiv (A:=A))}.

  (** finite set of vertices *)
  Variable (all: list A) (all_here: forall a, InA equiv a all).

  (** Number of vertices 
      Use Count.Classical_set_Num_Card to obtain N *)
  Notation HN_N := (Classical_set_Num_CardT DA _ all_here).
  Definition N: nat := (proj1_sig HN_N).
  Lemma HN: Num_Card Same A N.
  Proof. apply (proj2_sig HN_N). Qed.
  
  (** edges *)
  Variable (G: relation A) (G_compat: Proper pequiv G) (G_dec: Decider G).

  (** undirected graph *)
  Definition is_undirected: Prop :=
    forall x y, G x y -> G y x.
  
  (** path (non empty path): 
   (is_path x p y) means a path exists from x to y, made of the
   consecutive nodes in (x :: p ++ y :: nil). 
   *)
  Definition is_path: A -> list A -> A -> Prop := (non_empty_chain G).

  Lemma hd_app:
    forall l (h: A) p, nil <> l -> hd h l = hd h (l ++ p).
  Proof.
    intros [| a l] h p; simpl; intros H; [contradict H|]; auto.
  Qed.

  Lemma hd_eq:
    forall l (h1 h2: A), nil <> l -> hd h1 l = hd h2 l.
  Proof.
    intros [| a l] h1 h2 nin; simpl; congruence.
  Qed.

  Lemma hd_eq':
    forall l (h1 h2: A), nil <> l -> hd h1 l == hd h2 l.
  Proof.
    intros [| a l] h1 h2 nin; simpl; [congruence|auto].
  Qed.

  Lemma rev_nil:
    forall l: list A, nil = l <-> nil = rev l.
  Proof.
    intros [| a l]; simpl.
    tauto.
    split; intros F; [inversion F|].
    contradict F; apply app_cons_not_nil.
  Qed.
  
  Global Instance hd_compat:
    Proper (equiv ==> eqlistA equiv ==> equiv) (hd (A := A)).
  Proof.
    intros h1 h2 eh [| x1 p1] [| x2 p2] ep; inversion_clear ep;
      simpl; auto.
  Qed.
  
  Global Instance is_path_compat':
    Proper (equiv ==> eqlistA (equiv) ==> pequiv) is_path.
  Proof.
    unfold is_path; apply non_empty_chain_compat; auto.
  Qed.

  Lemma is_path_cons:
    forall p x xx y, is_path x (xx :: p) y -> is_path xx p y.
  Proof. intros p x xx y; unfold is_path; simpl; tauto. Qed.

  Lemma is_path_fst:
    forall p x y, is_path x p y -> G x (hd y p).
  Proof. intros [| xx p] x; unfold is_path; simpl; tauto. Qed.

  Lemma is_path_cat:
    forall p q x y z, is_path x p y -> is_path y q z ->
                      is_path x (p ++ y :: q) z.
  Proof.
    unfold is_path.
    induction p as [| a p ih]; intros q x y z hp hq.
    simpl in hp; simpl; split; auto.
    simpl in hp; destruct hp as [hp hp']; simpl; split; auto.
  Qed.
  
  (** connected graph *)
  Definition is_connected: Prop :=
    forall x y, ~ x == y -> exists p, is_path x p y.

  Lemma is_connected_has_neighbor:
    is_connected -> forall x y, ~ x == y ->
                                forall p, exists q, G p q.
  Proof.
    intros ic x y xy p.
    case (DA x p); intros hp.
    { rewrite hp in xy; clear hp x.
      apply ic in xy; destruct xy as [pp hp].
      destruct pp as [| a pp]; unfold is_path in hp; simpl in hp.
      exists y; apply hp.
      exists a; apply hp.
    }
    { clear xy  y.
      cut (~ p == x); [|contradict hp; symmetry; auto].
      clear hp; intros hp.
      apply ic in hp; destruct hp as [pp hp].
      destruct pp as [| a pp]; unfold is_path in hp; simpl in hp.
      exists x; apply hp.
      exists a; apply hp.
    } 
  Qed.

  (** distance between nodes (relational) *)
  Definition is_dist x y d: Prop :=
    (x == y /\ d = 0) \/ ~ x == y /\
    (exists p, is_path x p y /\ S (length p) = d) /\
    (forall p, is_path x p y -> S (length p) >= d).

  Instance is_dist_compat': Proper pequiv is_dist.
  Proof.
    intros x1 x2 ex y1 y2 ey n1 n2 en.
    apply or_compat; [rewrite en, ex, ey; tauto|].
    apply and_compat; [rewrite ex, ey; tauto|apply and_compat]. 
    apply ex_compat; intros p.
    rewrite ex, ey, en; tauto.
    apply all_compat; intros p.
    rewrite ex, ey, en; tauto.
  Qed.

  Lemma is_dist_unique:
    forall x y d1 d2, is_dist x y d1 -> is_dist x y d2 ->
                      d1 = d2.
  Proof.
    intros x y n1 n2 [[he1 h1]|[hn1 [[p1 [h1 hh1]] hhh1]]]
           [[he2 h2]|[hn2[[p2 [h2 hh2]] hhh2]]].
    subst; auto.
    tauto.
    tauto.
    specialize (hhh2 p1 h1); rewrite hh1 in hhh2.
    specialize (hhh1 p2 h2); rewrite hh2 in hhh1.
    lia.
  Qed.

  (** *)
  Lemma is_dist_neighbors:
    forall r x y nx ny, G x y -> is_dist x r nx ->
                        is_dist y r ny -> nx <= S ny.
  Proof.
    intros r x y nx ny hg [hx|hx] [hy|hy].
    lia.
    { destruct hx as [hx hxx]; subst; lia. }
    { destruct hy as [hy hyy]; subst.
      destruct hx as [_ [_ hf]].
      assert (hh: is_path x nil r).
      simpl; rewrite <- hy; auto.
      specialize (hf _ hh); simpl in hf; lia. }
    { destruct hy as [_ [[py [hpy hly]] hfy]].
      destruct hx as [_ [_ hfx]].
      assert (hh: is_path x (y :: py) r).
      simpl; split; auto.
      specialize (hfx _ hh); simpl in hfx.
      lia.
    }
  Qed.

  Lemma is_dist_neighbors':
    forall r x y nx ny, G x y -> G y x ->
                        is_dist x r nx -> is_dist y r ny ->
                        nx = ny \/ nx = S ny \/ S nx = ny.
  Proof.
    intros r x y nx ny hg hg' hx hy.
    eapply is_dist_neighbors in hg; eauto.    
    eapply is_dist_neighbors in hg'; eauto.
    lia.
  Qed.
  
  Lemma is_dist_neighbors_ex:
    forall r x n, is_dist x r (S n) ->
                  exists y, G x y /\ is_dist y r n.
  Proof.
    intros r x n [hd|hd]; [lia|].
    destruct hd as [_ [[p [he he']] hf]].
    destruct p as [|y p]; simpl in he', he.
    { exists r; split; auto; unfold is_dist;  left; split;
        [auto|lia]. }
    exists y; split; [apply he|].
    case (DA y r); intros hr.
    { left; split; auto.
      assert (hh: is_path x nil r).
      simpl; rewrite <- hr; apply he.
      specialize (hf _ hh); simpl in hf; lia. }
    { right; split; auto; split.
      exists p; split; [apply he|lia].
      intros pp hpp.
      assert (hh: is_path x (y :: pp) r).
      simpl; split; [apply he|]; auto.
      specialize (hf _ hh); simpl in hf.
      lia.      
    }
  Qed.
  
  (** diameter of G (relational) 
   If G is not connected, defines the greatest diameter among
   the diameters of the connected components of G.
   *)
  Definition is_diameter d: Prop := 
    (exists x y, is_dist x y d) /\
    (forall x y n, is_dist x y n -> d >= n).

  Global Instance is_diameter_compat: Proper pequiv is_diameter.
  Proof. intros a b ab; rewrite ab; auto. Qed.
  
  Lemma is_diameter_unique:
    forall d1 d2, is_diameter d1 -> is_diameter d2 -> d1 = d2.
  Proof.
    intros d1 d2 [[x1 [y1 h1]] hh1] [[x2 [y2 h2]] hh2].
    specialize (hh1 _ _ _ h2).
    specialize (hh2 _ _ _ h1).
    lia.
  Qed.
  
  (** *** We prove: diameter <= N *)

  (** simple path *)
  Fixpoint is_simple_b p: bool :=
    match p with
    | nil => true
    | x :: q => if InA_dec DA x q then false else is_simple_b q
    end.

  Definition is_simple p := is_simple_b p = true.

  Global Instance is_simple_b_compat: Proper pequiv is_simple_b.
  Proof.
    intros p1; induction p1 as [| x p1 ih]; intros [| y p2].
    simpl; auto.
    intros h; inversion h.
    intros h; inversion h.
    intros h; inversion_clear h.
    simpl.
    case (InA_dec DA x p1); case (InA_dec DA y p2); auto.
    intros f1 f2; rewrite H, H0 in f2; tauto.
    intros f1 f2; rewrite H, H0 in f2; tauto.
    intros _ _; apply ih; auto.
  Qed.
  
  Global Instance is_simple_compat: Proper pequiv is_simple.
  Proof.
    intros p1 p2 ep; unfold is_simple; rewrite ep; auto.
  Qed.

  Lemma is_simple_inv:
    forall x p, is_simple (x :: p) -> is_simple p.
  Proof.
    intros x p; unfold is_simple; intros hs; inversion hs; subst.
    revert H0; case (InA_dec DA x p); [congruence|auto].
  Qed.
  
  Lemma is_simple_inv':
    forall x p, is_simple (x :: p) -> ~ InA equiv x p.
  Proof.
    intros x p; unfold is_simple; intros hs; inversion hs; subst.
    revert H0; case (InA_dec DA x p); [congruence|auto].
  Qed.
  
  Lemma simple_le_N:
    forall p, is_simple p -> length p <= N.
  Proof.
    intros p hs.
    assert (g: Num_Card Same {a | InA equiv a p} (length p)).
    { 
      induction p as [| x p hi].
      - simpl length.
        eapply Inj_iff2.
        3: apply Empty_subset_Num_Card.
        intros a b ab; repeat rewrite InA_nil; auto.
        intros _ b _; rewrite InA_nil; auto.
      - unfold is_simple in hs, hi; simpl is_simple_b in hs.
        revert hs; case (InA_dec DA x p).
        { intros _ F; congruence. }
        intros nx hp; specialize (hi hp).
        simpl length.
        Open Scope nat_scope. (* ???? why is opening this scope required again??? *)
        assert (h: length p  + 1 = S (length p)) by lia.
        rewrite <- h.
        eapply Inj_iff2.
        3: eapply Disjoint_union_Num_Card.
        6: { eapply Singleton_Num_Card with (PA1 := fun y => x == y).
             reflexivity.
             tauto. }
        5: apply hi.
        2: { intros a b ab; rewrite InA_cons, ab; simpl ; intuition. }
        4: { intros a hh hhh.
             contradict nx.
             rewrite hhh.
             apply hh. }
        { intros a b ab; rewrite ab; simpl; tauto. }
        { intros a b ab; rewrite ab; simpl; tauto. }
        { intros a b ab; rewrite ab; simpl; tauto. }
    }
    eapply Subset_Num_Card_impl_le.
    - apply g.
    - unfold Num_Card.
      eapply Card_trans.
      apply Trivial_subset_Num_Card.
      apply HN.
    - intros a _ _; hnf; tauto.
  Qed.

  Global Instance length_compat: Proper (equiv (A := list A) ==> equiv) (length (A := A)).
  Proof.
    intros p; induction p as [| a p ih]; intros[| b q].
    simpl; auto.
    intros x; inversion x.
    intros x; inversion x.
    intros x; simpl; apply f_equal, ih; inversion x; auto.
  Qed.

  (** build a simple path:
  code that does the job (but Coq refuses it because of induction guard)

  Fixpoint remove_head x q: list A :=
    match q with
    | nil => nil
    | a :: l => if InAb x (a :: l) then l
                else remove_head x l
    end.

  Fixpoint build_simple_path p: list A :=
    match p with
    | nil => nil
    | x :: q => if InA_dec DA x q
                then
                  x :: build_simple_path (remove_head x q)
                else x :: build_simple_path q
    end.

   here a solution that merge the tow above functions:
   *)

  Definition InAb x l := if InA_dec DA x l then true else false.

  Fixpoint bsp (p: list A) (in_remove: bool) (h: A): list A :=
    match p with
    | nil => nil
    | x :: q =>
      if in_remove && InAb h (x :: q) then bsp q true h
      else if InA_dec DA x q then x :: bsp q true x
           else x :: bsp q false x
    end.
  
  Definition build_simple_path p :=
    match p with nil => nil | x :: q => bsp p false x end.
  
  Lemma build_simple_path_is_simple:
    forall p q, q = build_simple_path p -> is_simple q.
  Proof.
    assert (inc: forall p x ir y, InA equiv y (bsp p ir x) ->
                                  InA equiv y p). 
    { induction p as [| x p ih]; intros z ir y iny.
      simpl in iny; rewrite InA_nil in iny; tauto.
      simpl in iny; destruct ir; [rewrite andb_true_l in iny|
                                  rewrite andb_false_l in iny].
      - revert iny; case_eq (InAb z (x :: p)); intros i iny.
        unfold InAb in i; destruct (InA_dec DA z (x :: p)); [|congruence].
        rewrite InA_cons; right; eapply ih; eauto.
        destruct (InA_dec DA x p); rewrite InA_cons in iny; rewrite InA_cons.
        case iny; [auto|right; eapply ih; eauto].
        case iny; [auto|right; eapply ih; eauto].
      - destruct (InA_dec DA x p); rewrite InA_cons in iny; rewrite InA_cons.
        case iny; [auto|right; eapply ih; eauto].
        case iny; [auto|right; eapply ih; eauto].
    }
    
    assert (nin: forall p x, ~ InA equiv x (bsp p true x)).
    { induction p as [| x p hi]; intros y.
      rewrite InA_nil; tauto.
      simpl bsp.
      case_eq (InAb y (x :: p)); intros i. 
      eapply hi; eauto.
      unfold InAb in i; destruct (InA_dec DA y (x :: p)); [congruence|].
      destruct (InA_dec DA x p); 
        rewrite InA_cons in n; rewrite InA_cons; intuition;
          apply inc in H0; tauto.
    }

    cut (forall p ir h, is_simple_b (bsp p ir h) = true).
    { intros ok [| x p] q hq; unfold is_simple; rewrite hq; auto.
      apply ok. }
    induction p as [| x p hi]; intros ir h.
    { destruct ir; auto. }
    { simpl; destruct ir.
      - rewrite andb_true_l.
        case_eq (InAb h (x :: p)); intros i; [apply hi|].
        unfold InAb in i; destruct (InA_dec DA h (x :: p)); [congruence|].
        destruct (InA_dec DA x p); simpl.
        { case (InA_dec DA x (bsp p true x)).
          intros hh; contradict hh; eapply nin.
          intros _; apply hi. }
        { case (InA_dec DA x (bsp p false x)).
          intros hh; apply inc in hh; tauto.
          intros _; apply hi. }
      - rewrite andb_false_l.
        destruct (InA_dec DA x p); simpl.
        { case (InA_dec DA x (bsp p true x)).
          intros hh; contradict hh; apply nin.
          intros _; apply hi. }
        { case (InA_dec DA x (bsp p false x)).
          intros hh; apply inc in hh; tauto.
          intros _; apply hi. } 
    }
  Qed.

  Lemma build_simple_path_length:
    forall p q, q = build_simple_path p -> length q <= length p.
  Proof.
    cut (forall p x ir, length (bsp p ir x) <= length p).
    { intros ok [| a p] q hq; rewrite hq; unfold build_simple_path; auto. }
    induction p as [| a p ih]; intros x ir; simpl.
    lia.
    destruct (ir && InAb x (a :: p)).
    etransitivity; [apply ih|lia].
    destruct (InA_dec (eqA:=equiv) DA a p); simpl; apply le_n_S; auto.
  Qed.
  
  Lemma InA_hd:
    forall p (h:A), p <> nil -> InA equiv (hd h p) p.
  Proof.
    intros [| a p] h H; simpl.
    contradict H; auto.
    rewrite InA_cons; left; auto.
  Qed.

  Lemma bsp_nil:
    forall p ir h, bsp p ir h = nil <->
                   p = nil \/
                   ir = true /\
                   forall x, h == hd x (rev p). 
  Proof.
    induction p as [| x p ih]; intros ir h.
    intuition.
    simpl.
    destruct ir.
    - rewrite andb_true_l.
      case_eq (InAb h (x :: p)); intros i;
        unfold InAb in i; destruct (InA_dec DA h (x :: p)); try congruence;
          clear i.
      + rewrite InA_cons in i0; destruct i0 as [i|i];
          rewrite ih; split; intros [H|H].
        * rewrite H; right; split; auto.
        * destruct H as [_ H]; right; split; auto; simpl.
          intros xx.
          destruct p as [| a p]; simpl; simpl in H; auto.
          rewrite <- hd_app; auto.
          apply app_cons_not_nil.
        * inversion H.
        * destruct p as [| a p]; [left; auto|].
          destruct H as [_ H]; simpl in H. 
          right; split; auto.
          intros xx; specialize (H xx).
          rewrite <- hd_app in H.
          2: apply app_cons_not_nil.
          simpl; auto.
        * rewrite H in i; rewrite InA_nil in i; tauto.
        * right; split; auto.
          intros xx.
          rewrite <- hd_app; [apply H|].
          intros F; rewrite <- rev_nil in F;
            rewrite <- F in i; rewrite InA_nil in i; tauto.
        * inversion H.
        * right; split; auto.
          destruct H as [_ H]; intros xx; specialize (H xx).
          rewrite <- hd_app in H; auto.
          intros F; rewrite <- rev_nil in F;
            rewrite <- F in i; rewrite InA_nil in i; tauto.
      + destruct (InA_dec DA x p); split.
        * intros H; inversion H.
        * intros [H | [_ H]]; [inversion H|].
          contradict n.
          assert (pn: nil <> rev p).
          { intros F; rewrite <- rev_nil in F;
              rewrite <- F in i; rewrite InA_nil in i; tauto. }
          rewrite InA_cons; right.
          specialize (H h); rewrite <- hd_app in H; auto.
          rewrite <- InA_rev.
          rewrite H; apply InA_hd; auto.
        * intros H; inversion H.
        * intros [H | [_ H]]; [inversion H|].
          contradict n.
          assert (t: InA equiv h (rev p ++ x :: nil)).
          { specialize (H h). rewrite H.
            apply InA_hd.
            intros F; symmetry in F; contradict F; apply app_cons_not_nil. }
          rewrite InA_cons.
          rewrite InA_app_iff in t; destruct t as [t|t].
          right; apply InA_rev; auto.
          rewrite InA_cons in t; destruct t as [t|t];
            [left; auto|rewrite InA_nil in t; tauto].
    - rewrite andb_false_l.
      destruct (InA_dec DA x p); split;
        intros H; inversion H; inversion H0; congruence.
  Qed.     

  Lemma build_simple_path_not_nil:
    forall p, p <> nil-> build_simple_path p <> nil.
  Proof.
    intros [| x p] np; [tauto|].
    simpl.
    destruct (InA_dec (eqA:=equiv) DA x p); congruence.
  Qed.
  
  Lemma bsp_first:
    forall p h, hd h p == hd h (build_simple_path p).
  Proof.
    intros [| x p] h; simpl; auto.
    destruct (InA_dec DA x p); simpl; auto.
  Qed.

  (** non empty path with no information about first and last 
   elements *)
  Inductive is_path_: list A -> Prop :=
  | is_path_one: forall x, is_path_ (x :: nil)
  | is_path_step: forall x hp p, G x hp -> is_path_ (hp :: p)
                                 -> is_path_ (x :: hp :: p).

  Lemma is_path__is_path:
    forall p x y,
      is_path_ (x :: p ++ y :: nil) <-> is_path x p y.
  Proof.
    induction p as [| xx p ih]; intros x; simpl; split;
      intros h.
    inversion h; auto.
    right; [auto|left].
    { inversion_clear h.
      split; auto; clear H.
      apply ih; auto. }        
    { right; destruct h as [d h]; auto; clear d.
      apply ih; auto. }
  Qed.

  Instance is_path__compat:
    Proper (eqlistA (equiv) ==> equiv) is_path_.
  Proof.
    cut (forall p1 p2, eqlistA (equiv) p1 p2 -> is_path_ p1 -> is_path_ p2).
    { intros ok.
      intros p1 p2 ep; split; apply ok; [|symmetry in ep]; auto.
    }
    induction p1 as [| a1 p1 ih]; intros [|a2 p2].
    { intros _; intros H; inversion H. }
    { intros H; inversion H. }
    { intros H; inversion H. }
    { intro ee; destruct p2 as [| aa2 p2].
      intros _; left.
      inversion_clear ee.
      destruct p1 as [| aa1 p1]; [inversion H0|].
      intros HH.
      inversion_clear HH.
      inversion_clear H0.
      right.
      revert H1; apply G_compat; symmetry; auto.
      eapply ih; auto.
    }
  Qed.

  Lemma is_path__inv:
    forall x p,  p <> nil -> G x (hd x p) -> is_path_ p ->
                 is_path_ (x :: p).
  Proof.
    intros x [| y p] hp hg hpp; [congruence|right; auto].
  Qed.
  
  Lemma is_path__inv':
    forall x p,  is_path_ (x :: p) ->
                 p = nil \/ G x (hd x p) /\ is_path_ p.
  Proof.
    intros x [| y p] hp; [left|right; split;
                               inversion_clear hp; simpl]; auto.
  Qed.

  Lemma is_path__cat:
    forall p q, is_path_ p -> is_path_ q ->
                (forall h1 h2, G (hd h1 (rev p)) (hd h2 q)) ->
                is_path_ (p  ++ q).
  Proof.
    induction p as [| x p].
    intros q h; inversion h.
    intros [| y q].
    intros _ h; inversion h.
    intros hp hq h.
    simpl.
    destruct p as [| xx p].
    { simpl.
      right; auto.
      specialize (h x y); simpl in h; auto. }
    right.
    2: { apply IHp; auto.
         inversion hp; auto.
         intros h1 h2; specialize (h h1 h2).
         assert (t: hd h1 (rev (x :: xx :: p)) =
                    hd h2 (rev (xx :: p))).
         simpl.
         rewrite <- hd_app.
         2: apply app_cons_not_nil.
         apply hd_eq.
         apply app_cons_not_nil.
         remember (xx :: p) as lp.           
         simpl in h; simpl.
         rewrite <- hd_app in h.
         2: { intros F; rewrite <- rev_nil in F; congruence. }
         auto.
    }
    inversion_clear hp; auto.
  Qed.

  Lemma is_path__cat_conv_l:
    forall p q, p <> nil -> is_path_ (p ++ q) -> is_path_ p.
  Proof.
    induction p as [| x p ih]; [congruence|].
    { intros [| y q] _; simpl; intros hpq.
      rewrite app_nil_r in hpq; auto.
      destruct p as [| xx p]; [left; auto|].
      apply is_path__inv; try congruence.
      simpl; inversion hpq; auto.
      eapply ih; try congruence.
      inversion hpq; eauto.
    }
  Qed.

  Lemma is_path__cat_conv_m:
    forall p q, p <> nil -> q <> nil -> is_path_ (p ++ q) ->
                forall h1 h2, G (hd h1 (rev p)) (hd h2 q).
  Proof.
    induction p as [| x p ih]; [congruence|].
    { intros [| y q]; [congruence|intros _ _]; intros hpq h1 h2.
      destruct p as [| xx p]; simpl in hpq.
      inversion hpq; auto.
      remember (xx :: p) as pp; simpl.
      rewrite <- hd_app.
      2: { intros F; apply rev_nil in F; congruence. }
      eapply (ih (y ::q)); try congruence.
      rewrite Heqpp; inversion hpq; auto.
      Unshelve.
      apply h2.
    }
  Qed.
  
  Lemma is_path__cat_conv_r:
    forall p q, q <> nil -> is_path_ (p ++ q) -> is_path_ q.
  Proof.
    intros p q; revert p; induction q as [| y q ih].
    { intros p nq; congruence. }
    { intros [| x p] _; simpl; intros hpq; auto.
      destruct q as [| yy q]; [left; auto|].
      apply is_path__inv; try congruence.
      2: { eapply ih; try congruence.
           assert (P: (x :: p ++ y :: nil) ++ yy :: q =
                      x :: p ++ y :: yy :: q).
           change (((x :: p) ++ y :: nil) ++ yy :: q =
                   x :: p ++ y :: yy :: q).
           rewrite (ListUtil.list_app_first_last
                      (x :: p) (yy :: q) y). auto.
           rewrite P; auto. }
      { cut (G (hd x (rev (p ++ y :: nil))) (hd x (yy :: q))).
        rewrite rev_app_distr; simpl; auto.
        apply is_path__cat_conv_m.
        intros F; symmetry in F; contradict F; apply app_cons_not_nil.
        congruence.
        apply is_path__inv' in hpq.
        destruct hpq as [hpq|[_ hpq]].
        { symmetry in hpq; contradict hpq; apply app_cons_not_nil. }
        assert (P: p ++ y :: yy :: q = (p ++ y :: nil) ++ yy :: q).
        rewrite (ListUtil.list_app_first_last p (yy :: q) y); auto.
        rewrite <- P; auto.
      }
    }      
  Qed.

  Lemma is_path__cat_conv:
    forall p q x y,
      is_path_ ((p ++ x :: nil) ++ y :: q) ->
      is_path_ (p ++ x :: nil) /\ is_path_ (y :: q) /\
      G x y.
  Proof.
    intros p q x y hpq.
    assert (np: p ++ x :: nil <> nil).
    { intros F; symmetry in F; contradict F; apply app_cons_not_nil. }
    split; [|split].
    eapply is_path__cat_conv_l, hpq; auto.
    eapply is_path__cat_conv_r, hpq; auto; congruence.
    cut (G (hd x (rev (p ++ x :: nil))) (hd x (y :: q))).
    simpl; intros ok.
    rewrite rev_app_distr in ok; simpl; auto.
    eapply is_path__cat_conv_m, hpq; auto; congruence.
  Qed.        
  
  Lemma bsp_first_:
    forall p,
      is_path_ p ->
      forall x b l,
        (InA equiv x p -> bsp p true x = b :: l -> G x b) /\
        (~InA equiv x p -> forall ir, bsp p ir x = b :: l ->
                                      b = hd x p).
  Proof.
    induction p as [| h p ih]; intros hp x b l.
    inversion hp.
    destruct p as [| hh p].

    { split; intros ni.
      - rewrite InA_cons in ni.
        destruct ni as [ni|ni]; [| rewrite InA_nil in ni; tauto].
        simpl.
        unfold InAb; destruct (InA_dec DA x (h :: nil)).
        intros F; inversion F.
        rewrite InA_cons in n; intuition.
      - intros ir; simpl.
        assert (t: InAb x (h :: nil) = false).
        { unfold InAb.
          case (InA_dec DA x (h :: nil)); intros n; auto.
          rewrite InA_cons in n, ni.
          rewrite InA_nil in n, ni.
          intuition; auto.
        }
        rewrite t; rewrite andb_false_r.
        intros hh; inversion hh; auto.
    }

    inversion_clear hp.
    specialize (ih H0).
    split.

    { intros ina hb.
      assert (t: bsp (h :: hh :: p) true x =
                 bsp (hh :: p) true x).
      { remember (hh :: p) as hhp.
        simpl.
        assert (i: InAb x (h :: hhp) = true).
        { unfold InAb.
          case (InA_dec DA x (h :: hhp)).
          intros _; auto.
          intros F; tauto. }
        rewrite i; auto.
      }
      rewrite t in hb; clear t.
      specialize (ih x b l); destruct ih as [ih1 ih2].
      
      rewrite InA_cons in ina; destruct ina as [ina|ina].
      - case (InA_dec DA x (hh :: p)); intros i.
        apply ih1; auto.
        specialize (ih2 i _ hb); simpl in ih2.
        rewrite ih2, ina; auto.

      - apply ih1; auto.
    }

    { intros ina ir hb.
      remember (hh :: p) as hhp.
      simpl in hb.
      assert (t: InAb x (h :: hhp) = false).
      { unfold InAb.
        case (InA_dec DA x (h :: hhp)).
        intros F; tauto.
        intros _; auto. }
      rewrite t, andb_false_r in hb.
      destruct (InA_dec DA h hhp); inversion hb; simpl; auto.
    }
  Qed.    
  
  Lemma build_simple_path_is_path_:
    forall p q,
      q = build_simple_path p -> is_path_ p -> is_path_ q.
  Proof.
    cut (forall p q ir rx, q = bsp p ir rx -> q <> nil ->
                           is_path_ p -> is_path_ q).
    { intros ok [| x p] q.
      simpl; intros h; rewrite h; simpl; auto.
      intros h1 h2; eapply ok; eauto.
      rewrite h1; apply build_simple_path_not_nil; congruence.
    }
    
    induction p as [| x p ih]; intros q ir rx hq nq.
    { simpl in hq.
      rewrite hq; simpl; tauto. }

    destruct p as [| xx p].
    { intros _; simpl in hq; rewrite hq.
      destruct (ir && InAb rx (x :: nil)); simpl;
        [tauto|left].
    }

    intros hp; inversion_clear hp.
    remember (xx :: p) as lp.
    assert (nlp: lp <> nil)
      by (intros F; rewrite F in Heqlp; inversion Heqlp).
    simpl in hq.
    case_eq (ir && InAb rx (x :: lp)); intros hir;
      rewrite hir in hq.
    - eapply ih; eauto.
    - revert hq; case (InA_dec (eqA:=equiv) DA x lp);
        intros ina hq.
      case_eq (bsp lp true x).
      { intros nb; rewrite hq, nb; left. }
      intros b lb hlb.
      rewrite hq,hlb. right.
      assert (t := bsp_first_ H0 x b lb).
      apply t; auto.
      eapply ih.
      symmetry; apply hlb.
      congruence.
      auto.
      rewrite hq.
      assert (t: bsp lp false x <> nil).
      { intros F; apply bsp_nil in F.
        destruct F as [F|[F _]]; [tauto|congruence].
      }
      case_eq (bsp lp false x); [congruence|].
      intros bb lbb hlbb.
      right.
      * assert (tt := bsp_first_ H0 x bb lbb).
        destruct tt as [_ tt].
        specialize (tt ina false hlbb).
        rewrite tt, Heqlp; simpl; auto.
      * eapply ih; eauto.
        congruence.
  Qed.

  Lemma bsp_last:
    forall p h, hd h (rev p) == hd h (rev (build_simple_path p)).
  Proof.
    cut (forall p x,
            (forall h, hd h (rev p) == hd h (rev (bsp p false x))) /\
            (forall h, bsp p true x <> nil ->
                       hd h (rev p) == hd h (rev (bsp p true x)))).
    { intros ok p h.
      unfold build_simple_path.
      destruct p as [| x p].
      simpl; auto.
      destruct p as [| xx p].
      simpl; auto.
      edestruct ok as [okk _].
      rewrite okk.
      reflexivity.
    }

    induction p as [| x p ih].
    { intros hx; split; intros h; simpl; auto. }
    intros hx; split; intros h.

    - destruct p as [| xx p].
      simpl; auto.
      remember (xx :: p) as lp.
      simpl.
      rewrite <- hd_app.
      2: { intros F; apply rev_nil in F; congruence. }
      case (InA_dec DA x lp); intros ina.
      case_eq (bsp lp true x); intros hhb.
      rewrite bsp_nil in hhb.
      destruct hhb as [hhb|[_ hhb]].
      rewrite hhb in ina; rewrite InA_nil in ina; tauto.
      rewrite <- hhb; simpl; auto.
      intros lb hlb.
      rewrite <- hlb; simpl.
      rewrite <- hd_app.
      2: { intros F; apply rev_nil in F; congruence. }
      apply ih.
      congruence.
      assert (hbb: bsp lp false x <> nil).
      { intros F; apply bsp_nil in F.
        destruct F as [F|[F _]]; congruence. }
      simpl.
      rewrite <- hd_app.
      2: { intros F; apply rev_nil in F; congruence. }
      apply ih.
    - intros nnil.
      destruct p as [| xx p].
      simpl; auto.
      unfold InAb; case (InA_dec DA hx (x :: nil)); intros ina.
      rewrite InA_cons in ina; destruct ina as [ina|ina];
        [| rewrite InA_nil in ina; tauto].
      contradict nnil.
      rewrite bsp_nil; right; split; auto.
      simpl; auto.
      remember (xx :: p) as lp.
      simpl.
      rewrite <- hd_app; auto.
      2: { intros F; rewrite <- rev_nil in F; congruence. }
      case_eq (InAb hx (x :: lp)); intros ina;
        simpl in nnil; rewrite ina in nnil.
      * specialize (ih hx); destruct ih as [_ ih]; apply ih; auto.
      * clear nnil.
        case (InA_dec DA x lp); intros inna.
        case_eq (bsp lp true x).
        { intros nnil.
          rewrite bsp_nil in nnil.
          destruct nnil as [nnil|[_ nnil]]; [congruence|].
          rewrite <- nnil; simpl; auto. }
        { intros b lb hlb.
          rewrite <- hlb; simpl.
          rewrite <- hd_app.
          2: { intros F; rewrite <- rev_nil in F; congruence. }
          specialize (ih x); destruct ih as [_ ih]; apply ih;
            congruence.
        }
        { simpl; rewrite <- hd_app.
          2: { intros F; rewrite <- rev_nil in F.
               symmetry in F; rewrite bsp_nil in F.
               destruct F as [F|[F _]]; congruence. }
          specialize (ih x); destruct ih as [ih _]; apply ih.
        }
  Qed.
  
  Lemma path_simple_path:
    forall p x y, is_path x p y -> ~ x == y ->
                  {q | is_simple (x :: q ++ y :: nil)
                       /\ length q <= length p
                       /\ is_path x q y}.
  Proof.
    intros p x y hp nxy.
    remember (build_simple_path (x :: p ++ y :: nil)) as bsp.
    assert (S: is_simple bsp).
    eapply build_simple_path_is_simple; eauto.
    assert (L: length bsp <= length (x :: p ++ y :: nil)).
    eapply build_simple_path_length; eauto.    
    assert (B: {lbsp | bsp == x :: lbsp ++ y :: nil}).
    { assert (f := bsp_first (x :: p ++ y :: nil) x).
      assert (l := bsp_last  (x :: p ++ y :: nil) x).
      rewrite <- Heqbsp in f, l; simpl in f, l.
      rewrite rev_app_distr in l; simpl in l.
      clear -nxy f l.
      assert (bl: length bsp > 1).
      { destruct bsp as [| b [| bb bsp]].
        contradict nxy; symmetry; apply l.
        contradict nxy; rewrite l; simpl; apply f.
        simpl; lia.
      }
      clear nxy.
      revert bl f l; revert x y; induction bsp as [| b bsp ih];
        intros x y bl f l; simpl in f, l, bl.
      lia.
      destruct bsp as [| bb [| bbb lbsp]]; simpl in bl.
      { lia. }
      { exists nil.
        right; [symmetry; auto|].
        simpl; right; [rewrite l; auto| reflexivity]. }
      remember (bb :: bbb :: lbsp) as bsp.
      assert (d: length bsp > 1) by (rewrite Heqbsp; simpl; lia).
      assert (ff: bb == hd bb bsp) by (rewrite Heqbsp; auto).
      assert (ll: y == hd bb (rev bsp)).
      { rewrite <- hd_app in l.
        rewrite l.
        apply hd_eq'.
        intros F; rewrite <- rev_nil in F; rewrite Heqbsp in F; congruence.
        intros F; rewrite <- rev_nil in F; rewrite Heqbsp in F; congruence. }
      specialize (ih bb y d ff ll).
      destruct ih as [llbsp hh].
      exists (bb :: llbsp).
      right; [symmetry; auto|rewrite hh; reflexivity].
    }
    destruct B as [lbsp hlbsp].
    exists lbsp; split.
    - rewrite <- hlbsp; auto.
    - rewrite <- is_path__is_path.
      rewrite <- is_path__is_path in hp.
      apply build_simple_path_is_path_ in Heqbsp; auto.
      rewrite hlbsp in L; simpl in L.
      rewrite app_length, app_length in L; simpl in L.
      rewrite hlbsp in Heqbsp.
      split; [lia|auto].
  Qed.

  Lemma path_simple_path_r:
    forall p x y, is_path x p y ->
                  { q | is_simple (q ++ y :: nil)
                        /\ length q <= length p
                        /\ is_path x q y}.
  Proof.
    intros p x y hp.
    pose (bsp := build_simple_path (p ++ y :: nil)).
    assert (S: is_simple bsp).
    eapply build_simple_path_is_simple; unfold bsp; auto.
    assert (L: length bsp <= length (p ++ y :: nil)).
    eapply build_simple_path_length; eauto.    
    assert (B: { lbsp | bsp == lbsp ++ y :: nil}).
    { assert (l := bsp_last  (p ++ y :: nil) x).
      fold bsp in l; simpl in l.
      rewrite rev_app_distr in l; simpl in l.
      assert (nn: bsp <> nil).
      unfold bsp; apply build_simple_path_not_nil.
      intros F; symmetry in F; contradict F; apply app_cons_not_nil.
      clear -l nn; clearbody bsp.
      induction bsp as [| b bsp ih]; [congruence|].
      destruct bsp as [| bb bsp].
      exists nil; simpl; symmetry; right; [apply l|]; auto.
      destruct ih as [lbsp h].
      { remember (bb :: bsp) as B; simpl in l.
        rewrite l, <- hd_app; auto.
        rewrite HeqB.
        intros F; apply rev_nil in F; congruence. }
      { congruence. }
      { exists (b :: lbsp).
        simpl; right; auto. }
    }
    destruct B as [lbsp hlbsp].
    exists lbsp; split.
    - rewrite <- hlbsp; auto.
    - rewrite <- is_path__is_path.
      split.
      rewrite hlbsp, app_length, app_length in L; simpl in L.
      lia.      
      apply is_path__inv.
      + intros F; symmetry in F; contradict F; apply app_cons_not_nil.
      + rewrite <- hlbsp.     
        unfold bsp; rewrite <- bsp_first.
        destruct p as [| b p]; simpl; apply hp.
      + rewrite <- hlbsp; unfold bsp.
        eapply build_simple_path_is_path_.
        reflexivity.
        apply is_path__is_path in hp.
        apply is_path__inv' in hp; destruct hp as [hp|hp].
        symmetry in hp; contradict hp; apply app_cons_not_nil.
        apply hp.
  Qed.

  Lemma path_simple_path_l:
    forall p x y, is_path x p y ->
                  { q | is_simple (x :: q)
                        /\ length q <= length p
                        /\ is_path x q y}.
  Proof.
    intros p x y hp.
    remember (build_simple_path (x :: p)) as bsp.
    assert (S: is_simple bsp).
    eapply build_simple_path_is_simple; eauto.
    assert (L: length bsp <= length (x :: p)).
    eapply build_simple_path_length; eauto.    
    assert (B: { lbsp | bsp == x :: lbsp}).
    { assert (f := bsp_first (x :: p) x).
      rewrite <- Heqbsp in f; simpl in f.
      destruct bsp as [|b bsp].
      symmetry in Heqbsp; contradict Heqbsp;
        apply build_simple_path_not_nil; congruence.
      exists bsp; right; [symmetry; apply f|reflexivity].
    }
    destruct B as [lbsp hlbsp].
    exists lbsp; split; [|split].
    - rewrite <- hlbsp; auto.
    - rewrite hlbsp in L; simpl in L; lia.
    - apply is_path__is_path in hp.
      rewrite <- is_path__is_path.
      change (is_path_ ((x :: lbsp) ++ y :: nil)).
      apply is_path__cat.
      { apply build_simple_path_is_path_ in Heqbsp.
        rewrite <- hlbsp; auto.
        eapply is_path__cat_conv_l; [congruence|].
        apply hp. }
      { left; auto. }
      { change (is_path_ ((x :: p) ++ y :: nil)) in hp.
        intros h1 h2; eapply (is_path__cat_conv_m) in hp; try congruence.
        rewrite <- hlbsp; simpl;
          assert (l := bsp_last (x :: p) h1).
        rewrite <- Heqbsp in l; rewrite <- l.
        apply hp.
        Unshelve. apply h2.
      }
  Qed.

  Lemma dist_le_N:
    forall x y d, is_dist x y d -> S d <= N.
  Proof.
    intros x y d [[he1 h1]|[hn1 [[p1 [h1 hh1]] hhh1]]].
    { assert (hn: N >= 1).
      assert (Proper equiv x) by (unfold Proper; reflexivity).
      assert (ge := Nonempty_Num_Card _ H).
      apply Card_larger_ge_iff.
      eapply Card_trans.
      apply Card_weak, Card_sym, HN. 
      apply ge.
      lia.
    }
    apply path_simple_path in h1; auto.
    destruct h1 as [q [hq [_ hq']]].
    apply simple_le_N in hq.
    simpl in hq; rewrite app_length in hq; simpl in hq.
    specialize (hhh1 q hq').
    lia.
  Qed.
  
  Lemma diameter_le_N:
    forall d, is_diameter d -> S d <= N.
  Proof.
    intros d [[x [y hd]] b].
    eapply dist_le_N; eauto.
  Qed.
  
  (** ** Decidability:
    - compute if a path exists
    - compute distance
    - compute diameter
   *)
  Lemma is_path_dec:
    forall x p y, {is_path x p y}+{~is_path x p y}.
  Proof.
    intros x p y.
    case (chain_dec DA G_dec x (p ++ y :: nil) y); intros h;
      [left|right].
    rewrite chain_cons_iff_non_empty_chain in h; auto; apply h.
    contradict h.
    rewrite chain_cons_iff_non_empty_chain; auto; apply h.
  Qed.

  Lemma is_simple_dec: forall p, {is_simple p}+{~is_simple p}.
  Proof.
    induction p as [| a p ih]; unfold is_simple; simpl.
    left; auto.
    destruct ih as [ih|ih]; destruct (InA_dec DA a p).
    right; congruence.
    left; auto.
    right; congruence.
    right; contradict ih; apply ih.
  Qed.

  (** returns the list of simple paths made from p 
      + one predecessor (elements in l) 
   *)
  Fixpoint _enum_list (p: list A) (y: A) (l: list A):
    list (list A) :=
    match l with
    | nil => p :: nil
    | a :: ll => if InA_dec DA a p then _enum_list p y ll
                 else if G_dec a (hd y p) then (a :: p) :: _enum_list p y ll
                      else  _enum_list p y ll
    end.

  Instance enum_list_compat: Proper pequiv _enum_list.
  Proof.
    intros p p' ep y y' ey l.
    revert ep ey; revert p p' y y'; induction l as [| a l ih];
      intros p p' y y' ep ey l' el; inversion_clear el; simpl.
    right; auto.
    case (InA_dec DA a p); case (InA_dec DA x' p'); intros ina ina'.
    apply ih; auto.
    2: { contradict ina'; rewrite H, ep; auto. }
    { contradict ina; rewrite <- H, <- ep; auto. }
    case (G_dec a (hd y p)); case (G_dec x' (hd y' p')); intros hg hg'.
    right; auto; apply ih; auto.
    { contradict hg; rewrite <- H, <- ey, <- ep; auto. }
    { contradict hg'; rewrite H, ey, ep; auto. }
    apply ih; auto.
  Qed.

  (** increase s by adding one element - when possible - to any
      any of its paths *)
  Fixpoint _enum_set (s: list (list A)) (y: A): list (list A) :=
    match s with
    | nil => nil
    | p :: ss => _enum_list p y all ++ _enum_set ss y
    end.

  Instance enum_set_compat: Proper pequiv _enum_set.
  Proof.
    intros s; induction s as [| a s ih]; intros [| a' s'] es y y' ey;
      simpl; auto; inversion_clear es.
    apply app_eqlistA_compat.
    apply enum_list_compat; auto.
    apply ih; auto.
  Qed.

  (** all simple paths to y of length at most n *)
  Fixpoint enum_path0 (y: A) (n: nat): list (list A) :=
    match n with
    | 0 => nil :: nil
    | S nn => _enum_set (enum_path0 y nn) y
    end.

  Instance enum_path0_compat: Proper pequiv enum_path0.
  Proof.
    intros y y' ey n; revert ey; revert y y';
      induction n as [| n ih]; intros y y' ey n' en;
        simpl in en; rewrite <- en; simpl; auto.
    apply enum_set_compat; auto.
  Qed.

  (** all simple paths from x to y *)
  Definition enum_path (x y: A): list (list A) :=
    map_filter
      (fun p: list A =>
         match p with nil => None
                 | a :: pp => if DA x a then Some pp else None
         end)
      (enum_path0 y N).

  Instance enum_path_compat: Proper pequiv enum_path.
  Proof.
    intros x x' ex y y' ey.
    apply (map_filter_compat_eqlistA (eqA := equiv (A := list A))).
    intros [| a p] [| a' p'] ep; auto; inversion_clear ep.
    reflexivity.
    case (DA x a); case (DA x' a'); intros h h'; auto.
    contradict h; rewrite <- ex, <- H; auto.
    contradict h'; rewrite ex, H; auto.
    reflexivity.
    apply enum_path0_compat; auto.
  Qed.
  
  Lemma enum_list_is_simple: forall y p l q,
      is_simple p ->
      InA equiv q (_enum_list p y l) -> is_simple q.
  Proof.
    intros y p l q hp; revert q hp; revert p;
      induction l as [| a l ih]; simpl.
    { intros p q hp ina.
      rewrite InA_cons in ina; destruct ina as [ina|ina];
        [| rewrite InA_nil in ina; tauto].
      rewrite ina; auto. }
    intros p q hp.
    case_eq (InA_dec DA a p); intros ina he; eauto.
    case (G_dec a (hd y p)); intros hg inn; eauto.
    rewrite InA_cons in inn; destruct inn as [inn|inn]; [|eauto].
    rewrite inn; unfold is_simple; simpl.
    rewrite he.
    apply hp.
  Qed.
  
  Lemma enum_set_is_simple:
    forall s, (forall p, InA equiv p s -> is_simple p) ->
              forall y p, InA equiv p (_enum_set s y) -> is_simple p.
  Proof.
    induction s as [| q s ih]; simpl.
    { intros _ _ p F; rewrite InA_nil in F; tauto. }
    { intros hs y p ina.
      rewrite InA_app_iff in ina; destruct ina as [ina|ina];
        [|eauto].
      apply enum_list_is_simple in ina; auto. 
      apply hs.
      rewrite InA_cons; left; reflexivity.
    }
  Qed.

  Lemma enum_path0_is_simple: forall y n p,
      InA equiv p (enum_path0 y n) -> is_simple p.
  Proof.
    intros y; induction n as [| n ih]; simpl.
    intros p ina.
    rewrite InA_cons in ina; destruct ina as [ina|ina].
    rewrite ina; unfold is_simple; simpl; auto.
    rewrite InA_nil in ina; tauto.
    intros p ina.
    eapply enum_set_is_simple; eauto.
  Qed.

  Lemma enum_path_is_simple: forall x y p,
      InA equiv p (enum_path x y) -> is_simple (x :: p).
  Proof.
    intros x y p ina; unfold enum_path in ina.
    assert (S := enum_path0_is_simple y N).
    pose (E := enum_path0 y N); fold E in S, ina; clearbody E.
    revert p ina.
    induction E as [| e E ih]; simpl.
    { intros p ina; rewrite InA_nil in ina; tauto. }
    intros p.
    destruct e as [| a e]; [auto|].
    case (DA x a); intros xa; [|auto].
    rewrite InA_cons; intros ina; destruct ina as [ina|ina]; [|auto].
    apply S.
    rewrite InA_cons; left; right; auto.
  Qed.

  Lemma enum_list_is_path:
    forall y p, is_path_ (p ++ y :: nil) ->
                forall l q,
                  InA equiv q (_enum_list p y l) ->
                  is_path_ (q ++ y :: nil).
  Proof.
    intros y p hp; induction l as [| a l ih]; simpl; intros q hq.
    { rewrite InA_cons in hq; destruct hq as [hq|hq];
        [|rewrite InA_nil in hq; tauto].
      rewrite hq; auto. }
    { revert hq; case (InA_dec DA a p); intros ina; auto.
      case (G_dec a (hd y p)); intros hg; auto.
      intros inn; rewrite InA_cons in inn; destruct inn as [inn|inn]; [|auto].
      rewrite inn.
      simpl; apply is_path__inv; auto.
      intros F; symmetry in F; contradict F; apply app_cons_not_nil.
      revert hg; apply G_compat; auto.
      destruct p as [| b pp]; simpl; auto.
    }
  Qed.      

  Lemma enum_set_is_path:
    forall y s, (forall p, InA equiv p s -> is_path_ (p ++ y :: nil)) ->
                forall q,
                  InA equiv q (_enum_set s y) -> is_path_ (q ++ y :: nil).
  Proof.
    intros y; induction s as [|e s ih]; simpl; intros hs q hq.
    { rewrite InA_nil in hq; tauto. }
    { rewrite InA_app_iff in hq; destruct hq as [hq|hq]; [|auto].
      eapply enum_list_is_path.
      2: apply hq.
      apply hs; rewrite InA_cons; left; reflexivity.
    }
  Qed.

  Lemma enum_path0_is_path:
    forall y n p,
      InA equiv p (enum_path0 y n) -> is_path_ (p ++ y :: nil).
  Proof.
    intros y; induction n as [| n ih]; simpl; intros p hp.
    { rewrite InA_cons in hp; destruct hp as [hp|hp].
      rewrite hp; simpl; left.
      rewrite InA_nil in hp; tauto. }
    { eapply enum_set_is_path; eauto. }
  Qed.
  
  Lemma enum_path_is_path: forall x y p,
      InA equiv p (enum_path x y) -> is_path x p y.
  Proof.
    intros x y p ina; unfold enum_path in ina.
    apply is_path__is_path.
    assert (S := enum_path0_is_path y N).
    pose (E := enum_path0 y N); fold E in S, ina; clearbody E.
    revert p ina.
    induction E as [| e E ih]; simpl.
    { intros p ina; rewrite InA_nil in ina; tauto. }
    { intros p.
      destruct e as [| a e]; [auto|].
      case (DA x a); intros xa; [|auto].
      rewrite InA_cons; intros ina; destruct ina as [ina|ina]; [|auto].
      apply (S (x :: p)).
      rewrite InA_cons; left; right; auto.
    }
  Qed.

  Lemma enum_list_all:
    forall l q x y p,
      is_path x p y -> is_simple (x :: p) ->
      (x :: p  == q \/ InA equiv x l /\ p == q) ->
      InA equiv (x :: p) (_enum_list q y l).
  Proof.
    induction l as [| a l ih]; simpl;
      intros q x y p hp hs [ina|[ina inaa]]. 
    { rewrite InA_cons; left; auto. }
    { rewrite InA_nil in ina; tauto. }
    { case (InA_dec DA a q); intros aq; [eauto|].
      case (G_dec a (hd y q)); intros hg; [|eauto].
      rewrite InA_cons; right; eauto.
    }
    { case (InA_dec DA a q); intros aq;
        rewrite InA_cons in ina; destruct ina as [ina|ina].
      - apply is_simple_inv' in hs; rewrite <- inaa, <- ina in aq; tauto.
      - eapply ih; eauto.
      - case (G_dec a (hd y q)); intros hg.
        rewrite InA_cons; left; right; auto.
        contradict hg; rewrite <- ina; apply is_path_fst;
          rewrite <- inaa; auto.
      - case (G_dec a (hd y q)); intros hg;
          [rewrite InA_cons; right|]; apply ih; auto.        
    }
  Qed.
  
  Lemma enum_set_all:
    forall s x y p,
      is_path x p y -> is_simple (x :: p) ->
      (InA equiv (x :: p) s \/ InA equiv p s) ->
      InA equiv (x :: p) (_enum_set s y).
  Proof. 
    induction s as [| e s ih]; simpl; intros x y p hp hs ina.
    { destruct ina as [ina|ina]; rewrite InA_nil in ina; tauto. }
    { rewrite InA_app_iff.

      destruct ina as [ina|ina].
      rewrite InA_cons in ina; destruct ina as [ina|ina].
      - left; eapply enum_list_all; eauto.
      - right; eapply ih; eauto.
      - rewrite InA_cons in ina; destruct ina as [ina|ina].
        + left; apply enum_list_all; auto.
        + right; apply ih; auto.
    }
  Qed.

  Lemma enum_path0_all:
    forall n y x p,
      is_path x p y -> is_simple (x :: p) -> length (x :: p) <= n ->
      InA equiv (x :: p) (enum_path0 y n).
  Proof.
    induction n as [| n ih]; simpl; intros y x p hp hs hl; [lia|].
    eapply enum_set_all; eauto.
    case (le_gt_dec n (length p)); intros hn.
    assert (n = length p) by lia; clear hl hn.
    2: { left; eapply ih; eauto. }
    destruct p as [|xx pp].
    { rewrite H; clear H; simpl.
      right; rewrite InA_cons; left; auto. }
    right; eapply ih; eauto.
    eapply is_path_cons; eauto.
    eapply is_simple_inv; eauto.
    lia. 
  Qed.
  
  Lemma enum_path_all: forall x y p,
      is_path x p y -> is_simple (x :: p) ->
      InA equiv p (enum_path x y).
  Proof.
    intros x y p hp hs; unfold enum_path.
    assert (S := enum_path0_all (n := N) y (x := x) (p := p)
                                hp hs (simple_le_N hs)).
    pose (E := enum_path0 y N); fold E in S; fold E; clearbody E.
    rewrite InA_map_filter_iff. 
    exists (x :: p); split; auto.
    case (DA x x); [|intros F; contradict F]; auto.
    { clear; intros [| xx p] [| y q] pq; auto;
        inversion_clear pq.
      case (DA x xx); case (DA x y); auto; intros h t.
      contradict h; rewrite t; auto.
      contradict t; rewrite h; symmetry; auto. }
  Qed.
  
  Lemma exists_path_dec:
    forall x y, {p | is_path x p y}+{forall p, ~is_path x p y}.
  Proof.
    intros x y.
    pose (E := enum_path x y).
    case_eq E.
    { intros hE.
      right; intros p hp.
      destruct (path_simple_path_l p x y) as [q [hq [_ hq']]]; auto.
      apply enum_path_all in hq'; auto.
      fold E in hq'; rewrite hE, InA_nil in hq'; tauto.
    }
    { intros p lE hE; left; exists p.
      apply enum_path_is_path.
      fold E; rewrite hE; auto. }
  Qed. 

  Lemma exists_path_dec':
    forall x y, {p | is_path x p y}+{~exists p, is_path x p y}.
  Proof.
    intros x y.
    pose (E := enum_path x y).
    case_eq E.
    { intros hE.
      right; intros [p hp].
      destruct (path_simple_path_l p x y) as [q [hq [_ hq']]]; auto.
      apply enum_path_all in hq'; auto.
      fold E in hq'; rewrite hE, InA_nil in hq'; tauto.
    }
    { intros p lE hE; left; exists p.
      apply enum_path_is_path.
      fold E; rewrite hE; auto. }
  Qed. 

  Lemma exists_path_length_dec:
    forall x y k, {p | length p <= k /\ is_path x p y}+
                  {forall p, length p <= k -> ~is_path x p y}.
  Proof.
    intros x y k.
    pose (E := enum_path x y).
    pose (fE := filter (fun p => if le_gt_dec (length p) k then true else false) E).
    assert (Pdum:
              Proper (equiv ==> eq)
                     (fun p =>
                        if le_gt_dec (length p) k then true else false)).
    { intros a b ab; case (le_gt_dec (length a) k);
        case (le_gt_dec (length b) k); auto.
      rewrite ab; lia.
      rewrite ab; lia. }    
    case_eq fE.
    { intros hE.
      right; intros p hp hpp.
      apply path_simple_path_l in hpp.
      destruct hpp as [q [hs [hl hq]]].
      assert (F: InA equiv q fE).
      unfold fE; rewrite filter_InA; auto.
      split.
      apply enum_path_all; auto.
      case (le_gt_dec (length q) k); [auto|intros F; lia].
      rewrite hE in F; rewrite InA_nil in F; tauto.
    }
    { intros p lE hE.
      assert (hp: InA equiv p fE).
      rewrite hE; rewrite InA_cons; left; auto.
      unfold fE in hp; rewrite filter_InA in hp; auto.
      destruct hp as [hp hpp]; revert hpp;
        case (le_gt_dec (length p) k); [intros hpp _|congruence].      
      left; exists p; split; auto.
      apply enum_path_is_path; auto.
    }      
  Qed.
  
  Lemma is_connected_has_neighbor':
    is_connected ->
    forall x y, ~ x == y ->
                forall p, exists q, ~ p == q /\ G p q.
  Proof.
    intros ic x y xy p.
    case (DA x p); intros hp.
    { rewrite hp in xy; clear hp x.     
      dup xy xy'.
      apply ic in xy'; destruct xy' as [pp hp].
      apply path_simple_path in hp; auto.
      destruct hp as [q [hs [_ hip]]].
      destruct q as [| xq q].
      unfold is_path in hip; simpl in hip.
      exists y; split; auto.
      exists xq; split.
      apply is_simple_inv' in hs.
      simpl in hs; rewrite InA_cons in hs.
      tauto.
      unfold is_path in hip; apply hip.
    }
    { clear xy y.
      cut (~ p == x); [|contradict hp; symmetry; auto].
      clear hp; intros hp.
      dup hp hp'.
      apply ic in hp'; destruct hp' as [pp hpp].
      apply path_simple_path in hpp; auto.
      destruct hpp as [q [hs [_ hip]]].
      destruct q as [| xq q].
      unfold is_path in hip; simpl in hip.
      exists x; split; auto.
      exists xq; split.
      apply is_simple_inv' in hs.
      simpl in hs; rewrite InA_cons in hs.
      tauto.
      unfold is_path in hip; apply hip.
    }
  Qed.
  
  (** Minimum of a list *)
  Definition min_list (l: list nat): option nat :=
    match l with nil => None
            | a :: ll => Some (fold_right min a ll)
    end.

  Global Instance min_list_compat: Proper pequiv min_list.
  Proof.
    intros [| a l] [| a' l'] el; inversion_clear el; simpl; auto.
    eapply FoldRight.fold_right_compat; eauto.
    { intros aa aa' ea b b' eb; simpl in ea, eb; rewrite ea, eb; auto. }
  Qed.

  Lemma min_list_None:
    forall l, min_list l = None <-> l = nil.
  Proof.
    unfold min_list; intros [| a l]; [tauto|split;congruence].
  Qed.

  Lemma fr_min1:
    forall a l, InA equiv (fold_right min a l) (a :: l).
  Proof.
    intros a.
    induction l as [| aa l ih]; simpl; rewrite InA_cons.
    { left; auto. }
    case (le_gt_dec aa (fold_right min a l)); intros h.
    - assert (min aa (fold_right min a l) = aa) by (apply min_l; auto).      
      rewrite InA_cons in ih; destruct ih as [ih|ih];
        [simpl in ih; rewrite ih in *|]; right; auto.
    - assert (min aa (fold_right min a l) = fold_right min a l) by (apply min_r; lia).
      rewrite InA_cons in ih; destruct ih as [ih|ih].
      simpl in ih; rewrite ih in *.
      left; auto.
      right; rewrite H; rewrite InA_cons; right; auto.
  Qed.

  Lemma fr_min2:
    forall a l b, InA equiv b (a :: l) -> fold_right min a l <= b.
  Proof.
    intros a l; revert a.
    induction l as [| aa l ih]; intros a b ina.
    { rewrite InA_cons in ina; destruct ina as [ina|ina];
        [|rewrite InA_nil in ina; tauto].
      simpl in *; subst; auto. }
    assert (inna: b == aa \/ InA equiv b (a :: l)).
    { rewrite InA_cons, InA_cons in ina; rewrite InA_cons.
      destruct ina as [ina|[ina|ina]]; intuition. }
    clear ina; destruct inna as [ina|ina].
    simpl; simpl in ina; rewrite ina; apply Nat.le_min_l.
    simpl.
    etransitivity.
    apply Nat.le_min_r.
    apply ih; auto.
  Qed.

  Lemma fr_min3:
    forall m a l, InA equiv m (a :: l) /\
                  (forall b, InA equiv b (a :: l) -> m <= b)
                  -> m = fold_right min a l.
  Proof.
    intros m a l; revert a m; induction l as [| aa l ih]; intros a m [ina hle].
    { rewrite InA_cons in ina; destruct ina as [ina|ina];
        [|rewrite InA_nil in ina; tauto].
      simpl in *; rewrite ina; auto. }
    assert (inna: m == aa \/ InA equiv m (a :: l)).
    { rewrite InA_cons, InA_cons in ina; rewrite InA_cons.
      destruct ina as [ina|[ina|ina]]; intuition. }
    clear ina; destruct inna as [ina|ina]; simpl.
    - rewrite ina; symmetry; apply min_l.
      rewrite <- ina; apply hle.
      rewrite InA_cons, InA_cons; specialize (fr_min1 a l);
        rewrite InA_cons; intros [h|h]; [left|right; right]; auto.
    - case (le_gt_dec aa (fold_right min a l)); intros h.
      + assert (min aa (fold_right min a l) = aa) by (apply min_l; lia).
        rewrite H; clear H.
        assert (m <= aa) by (apply hle; auto).
        assert (aa <= m).
        etransitivity; [apply h| apply fr_min2; auto].
        lia.
      + assert (min aa (fold_right min a l) =
                fold_right min a l) by (apply min_r; lia).
        rewrite H; clear H h.
        apply ih; split; auto.
        intros b; rewrite InA_cons; intros [inna|inna]; apply hle; auto.        
  Qed.  
  
  Lemma min_list_Some:
    forall l m, min_list l = Some m <->
                InA equiv m l /\
                forall a, InA equiv a l -> m <= a.
  Proof.
    unfold min_list; intros [| a l] m.
    split; [intros h; congruence|
            intros [h _]; rewrite InA_nil in h; tauto].
    split.
    - intros h; inversion h; subst; clear h.
      split; [apply fr_min1|intros b; apply fr_min2].
    - intros [h hh]; apply f_equal; symmetry; apply fr_min3; auto.
  Qed.
  
  (** Maximum of a list *)
  Definition max_list (l: list nat): option nat :=
    match l with nil => None
            | a :: ll => Some (fold_right max a ll)
    end.

  Global Instance max_list_compat: Proper pequiv max_list.
  Proof.
    intros [| a l] [| a' l'] el; inversion_clear el; simpl; auto.
    eapply FoldRight.fold_right_compat; eauto.
    { intros aa aa' ea b b' eb; simpl in ea, eb; rewrite ea, eb; auto. }
  Qed.

  Lemma max_list_None:
    forall l, max_list l = None <-> l = nil.
  Proof.
    unfold max_list; intros [| a l]; [tauto|split;congruence].
  Qed.

  Lemma fr_max1:
    forall a l, InA equiv (fold_right max a l) (a :: l).
  Proof.
    intros a.
    induction l as [| aa l ih]; simpl; rewrite InA_cons.
    { left; auto. }
    case (le_gt_dec (fold_right max a l) aa); intros h.
    - assert (max aa (fold_right max a l) = aa) by (apply max_l; auto).
      rewrite InA_cons in ih; destruct ih as [ih|ih];
        [simpl in ih; rewrite ih in *|]; right; auto.
    - assert (max aa (fold_right max a l) = fold_right max a l) by (apply max_r; lia).
      rewrite InA_cons in ih; destruct ih as [ih|ih].
      simpl in ih; rewrite ih in *.
      left; auto.
      right; rewrite H; rewrite InA_cons; right; auto.
  Qed.

  Lemma fr_max2:
    forall a l b, InA equiv b (a :: l) -> fold_right max a l >= b.
  Proof.
    intros a l; revert a.
    induction l as [| aa l ih]; intros a b ina.
    { rewrite InA_cons in ina; destruct ina as [ina|ina];
        [|rewrite InA_nil in ina; tauto].
      simpl in *; subst; auto. }
    assert (inna: b == aa \/ InA equiv b (a :: l)).
    { rewrite InA_cons, InA_cons in ina; rewrite InA_cons.
      destruct ina as [ina|[ina|ina]]; intuition. }
    clear ina; destruct inna as [ina|ina].
    simpl; simpl in ina; rewrite ina; apply Nat.le_max_l.
    simpl.
    etransitivity.
    apply Nat.le_max_r.
    apply ih; auto.
  Qed.

  Lemma fr_max3:
    forall m a l, InA equiv m (a :: l) /\
                  (forall b, InA equiv b (a :: l) -> m >= b)
                  -> m = fold_right max a l.
  Proof.
    intros m a l; revert a m; induction l as [| aa l ih]; intros a m [ina hle].
    { rewrite InA_cons in ina; destruct ina as [ina|ina];
        [|rewrite InA_nil in ina; tauto].
      simpl in *; rewrite ina; auto. }
    assert (inna: m == aa \/ InA equiv m (a :: l)).
    { rewrite InA_cons, InA_cons in ina; rewrite InA_cons.
      destruct ina as [ina|[ina|ina]]; intuition. }
    clear ina; destruct inna as [ina|ina]; simpl.
    - rewrite ina; symmetry; apply max_l.
      rewrite <- ina; apply hle.
      rewrite InA_cons, InA_cons; specialize (fr_max1 a l);
        rewrite InA_cons; intros [h|h]; [left|right; right]; auto.
    - case (le_gt_dec (fold_right max a l) aa); intros h.
      + assert (max aa (fold_right max a l) = aa) by (apply max_l; lia).
        rewrite H; clear H.
        assert (m >= aa) by (apply hle; auto).
        assert (aa >= m).
        etransitivity; [apply h| apply fr_max2; auto].
        lia.
      + assert (max aa (fold_right max a l) =
                fold_right max a l) by (apply max_r; lia).
        rewrite H; clear H h.
        apply ih; split; auto.
        intros b; rewrite InA_cons; intros [inna|inna]; apply hle; auto.        
  Qed.
  
  Lemma max_list_Some:
    forall l m, max_list l = Some m <->
                InA equiv m l /\
                forall a, InA equiv a l -> m >= a.
  Proof.
    unfold max_list; intros [| a l] m.
    split; [intros h; congruence|
            intros [h _]; rewrite InA_nil in h; tauto].
    split.
    - intros h; inversion h; subst; clear h.
      split; [apply fr_max1|intros b; apply fr_max2].
    - intros [h hh]; apply f_equal; symmetry; apply fr_max3; auto.
  Qed.

  (** Distance between x and y (None if no path exists) *)
  Notation length := (length (A := A)).
  Definition compute_dist (x y: A): option nat :=
    if DA x y then Some 0 else
      match min_list (map length (enum_path x y)) with
      | None => None | Some d => Some (S d) end.

  Instance compute_dist_compat: Proper pequiv compute_dist.
  Proof.
    intros x x' ex y y' ey; unfold compute_dist;
      case (DA x y); case (DA x' y'); intros hx hy; auto.
    { contradict hx; rewrite <- ex, <- ey; auto. }
    { contradict hx; rewrite <- ex, <- ey; auto. }
    assert (min_list (map length (enum_path x y)) ==
            min_list (map length (enum_path x' y'))).
    { apply min_list_compat.
      eapply map_compat.
      apply length_compat.
      apply enum_path_compat; auto. }
    revert H; destruct (min_list (map length (enum_path x y)));
      destruct (min_list (map length (enum_path x' y')));
      intros H; inversion_clear H; auto.
  Qed.

  Lemma compute_dist_is_dist:
    forall x y, match compute_dist x y with
                | None => forall d, ~is_dist x y d
                | Some d => is_dist x y d
                end.
  Proof.
    intros x y; unfold compute_dist.
    case (DA x y); intros xy.
    { left; split; auto. }
    pose (E := map length (enum_path x y)).
    fold E.
    case_eq (min_list E).
    - intros d hd; unfold is_dist; right; split; auto.
      rewrite min_list_Some in hd; destruct hd as [ina hd].
      split.
      + unfold E in ina; rewrite InA_map_iff in ina; [|apply length_compat].
        destruct ina as [p [hp inna]].
        exists p; split; [apply enum_path_is_path; auto|
                          simpl in hp; lia].
      + intros p hp; apply le_n_S.
        apply path_simple_path_l in hp; destruct hp as [q [hq [hhq hhhq]]].
        etransitivity; [|apply hhq].
        apply hd.
        unfold E; rewrite InA_map_iff; [|apply length_compat].
        exists q; split; auto; apply enum_path_all; auto.
    - intros hm d hd; rewrite min_list_None in hm;
        unfold is_dist in hd;
        destruct hd as [hd|[_ [[p [hp _]] _]]]; [tauto|].
      apply path_simple_path_l in hp.
      destruct hp as [q [hs [_ hq]]].
      apply enum_path_all in hq; auto.
      unfold E in hm; apply map_eq_nil in hm.
      rewrite hm in hq; rewrite InA_nil in hq; tauto.
  Qed.

  Lemma is_dist_dec:
    forall x y, {d | is_dist x y d}+{forall d, ~is_dist x y d}.
  Proof.
    intros x y.
    assert (hdd := compute_dist_is_dist x y).
    destruct (compute_dist x y).
    left; exists n; auto.
    right; auto.
  Qed.

  Lemma is_dist_dec0:
    forall x y d, {is_dist x y d}+{~is_dist x y d}.
  Proof.
    intros x y d; destruct (is_dist_dec x y) as [[dd hd]|hd].
    case (Nat.eq_dec d dd); intros he.
    subst; left; auto.
    right; intros F; contradict he; eapply is_dist_unique; eauto.
    right; apply hd.
  Qed.

  Lemma dist_exists':
    forall x y, {p | ~x == y /\ is_path x p y} + { x == y } ->
                { d | is_dist x y d }.
  Proof.
    intros x y [[p [_ hp]]|h].
    2: { exists 0; left; auto. }
    assert (hd := compute_dist_is_dist x y).
    revert hd; unfold compute_dist; case (DA x y); intros xy hd.
    { exists 0; auto. }
    revert hd; case_eq (min_list (map length (enum_path x y))).
    - intros d hd hdd; rewrite min_list_Some in hd; destruct hd as [hd hhd].
      exists (S d); right; split; auto; split.
      rewrite InA_map_iff in hd; [|apply length_compat].
      destruct hd as [q [hq inna]].
      exists q; split; [apply enum_path_is_path; auto| simpl in hq;lia].
      intros pp hpp; apply le_n_S.
      apply path_simple_path_l in hpp.
      destruct hpp as [q [hs [hq hq']]].
      apply enum_path_all in hq'; auto. 
      etransitivity; [|apply hq].
      apply hhd; rewrite InA_map_iff; [|apply length_compat].
      exists q; split; auto.
    - intros hm hd; rewrite min_list_None in hm.
      apply path_simple_path_l in hp.
      destruct hp as [q [hs [_ hq']]].
      apply enum_path_all in hq'; auto.
      apply map_eq_nil in hm; rewrite hm, InA_nil in hq'; tauto.
  Qed.

  Lemma dist_exists:
    forall x y p, is_path x p y -> { d | is_dist x y d }.
  Proof.
    intros x y p hp; apply dist_exists'.
    case (DA x y);[right|left; exists p; split]; auto.
  Qed.

  Lemma is_dist_equiv:
    forall x y d, is_dist x y d <-> compute_dist x y == Some d.
  Proof.
    intros x y; generalize (compute_dist_is_dist x y).
    destruct (compute_dist x y); split.
    intros h; cut (n = d); [subst; auto|].
    eapply is_dist_unique; eauto.
    intros h; inversion h; subst; auto.
    intros h; specialize (H d); tauto.
    intros h; inversion h.
  Qed.

  Definition compute_dist' x y p (hp: is_path x p y): nat :=
    proj1_sig (dist_exists x y _ hp).

  Lemma compute_dist'_compat:
    forall x x' (ex: x == x') y y' (ey: y == y') p p'
           (ep: p == p') (hp: is_path x p y)
           (hp': is_path x' p' y'),
      compute_dist' x y p hp = compute_dist' x' y' p' hp'.
  Proof.
    intros; unfold compute_dist'.
    destruct (dist_exists x y p hp) as [d hd].
    destruct (dist_exists x' y' p' hp') as [d' hd'].
    simpl.
    eapply is_dist_unique.
    apply hd.
    rewrite ex, ey; auto.
  Qed.
  
  Lemma is_dist_equiv':
    forall x y p (hp: is_path x p y),
      is_dist x y (compute_dist' x y _ hp).
  Proof.
    intros x y p hp; unfold compute_dist'.
    destruct (dist_exists x y _ hp) as [d hd]; simpl; auto.
  Qed.

  Lemma is_connected' (ic: is_connected):
    forall x y, { p | ~x == y /\ is_path x p y } + { x == y }.
  Proof.
    intros x y; destruct (DA x y).
    right; auto.
    destruct (exists_path_dec x y) as [[p h]|h].
    left; exists p; split; auto.
    left; exists nil.
    specialize (ic x y n).
    destruct ic as [p hp].
    contradict hp; apply h.
  Qed.

  Definition compute_dist'' (ic: is_connected) x y: nat :=
    proj1_sig (dist_exists' x y (is_connected' ic x y)).

  Instance compute_dist''_compat:
    forall ic, Proper (equiv ==> equiv ==> eq) (compute_dist'' ic).
  Proof.
    intros ic x x' ex y y' ey; simpl; unfold compute_dist''.
    destruct (is_connected' ic x y) as [[p [hp hhp]]|I];
      destruct (is_connected' ic x' y') as [[p' [hp' hhp']]|I'].
    2: { contradict hp; rewrite ex, ey; auto. }
    2: { contradict hp'; rewrite <- ex, <- ey; auto. }
    - destruct (dist_exists' x y (inleft (exist (conj hp hhp)))) as [d hd].
      destruct (dist_exists' x' y' (inleft (exist (conj hp' hhp')))) as [d' hd'].
      simpl.
      eapply is_dist_unique.
      apply hd.
      rewrite ex, ey; auto.
    - destruct (dist_exists' x y (inright I)) as [d hd].
      destruct (dist_exists' x' y' (inright I')) as [d' hd'].
      simpl.
      eapply is_dist_unique.
      apply hd.
      rewrite ex, ey; auto.
  Qed.
  
  Lemma is_dist_equiv'' (ic: is_connected):
    forall x y, is_dist x y (compute_dist'' ic x y).
  Proof.
    intros x y; unfold compute_dist''.
    destruct (dist_exists' x y (is_connected' ic x y))
      as [d hd]; simpl; auto. 
  Qed.

  (** ** Compute Diameter *)
  
  Notation map_filter_dist x :=
    (map_filter (fun y => compute_dist x y) all).

  Instance map_filter_dist_compat: Proper pequiv (fun x => map_filter_dist x).
  Proof.
    intros a b ab.
    apply (map_filter_compat_eqlistA (eqA := equiv (A := A))); auto.
    2: { reflexivity. }
    intros x y xy; apply compute_dist_compat; auto.
  Qed.
  
  Definition compute_diameter: option nat :=
    max_list (fold_right
                (fun x l => (map_filter_dist x) ++ l) nil
                all).

  Lemma fr_app:
    forall x fl ll (f: A -> list nat) (f_compat: Proper pequiv f),
      InA equiv x (fold_right (fun y l => f y ++ l) fl ll) <->
      InA equiv x fl \/ exists y, InA equiv y ll /\ InA equiv x (f y).
  Proof.
    intros x fl ll f fc; split; intros hin.
    { induction ll as [| a ll ih]; simpl in hin.
      left; auto.
      rewrite InA_app_iff in hin; destruct hin as [hin|hin].
      right; exists a; split; auto.
      specialize (ih hin); destruct ih as [ih|[y [ih ihh]]].
      left; auto.
      right; exists y; split; auto.
    }
    { induction ll as [| a ll ih]; simpl;
        destruct hin as [hin|hin]; auto.
      destruct hin as [y [hin _]]; rewrite InA_nil in hin; tauto.
      rewrite InA_app_iff; right; auto.
      rewrite InA_app_iff. 
      destruct hin as [y [hhin hin]];
        rewrite InA_cons in hhin; destruct hhin as [hhin|hhin].
      left; rewrite <- hhin; auto.
      right; apply ih.
      right; exists y; auto.
    }
  Qed.

  Lemma fr_app_nil:
    forall fl ll (f: A -> list nat) (f_compat: Proper pequiv f),
      fold_right (fun y l => f y ++ l) fl ll = nil <->
      fl = nil /\ forall x, InA equiv x ll -> f x = nil.
  Proof.    
    intros fl ll f fc; split; intros hn.
    { induction ll as [| a ll ih]; simpl in hn.
      { split; auto.
        intros x hx; rewrite InA_nil in hx; tauto. }
      apply app_eq_nil in hn; destruct hn as [hn hhn].
      specialize (ih hhn); destruct ih as [ih ihh].
      split; auto; intros x inx;
        rewrite InA_cons in inx; destruct inx as [inx|inx].
      rewrite inx; auto.
      auto.
    }
    { induction ll as [| a ll ih]; simpl.
      - apply hn.
      - destruct hn as [hhn hn].
        apply ListUtil.app_nil.
        auto.
        apply ih.
        split; auto.
    }        
  Qed.

  Lemma compute_diameter_is_diameter:
    match compute_diameter with
    | None => forall d, ~is_diameter d
    | Some d => is_diameter d
    end.
  Proof.
    unfold compute_diameter.
    case_eq (max_list
               (fold_right
                  (fun x l => map_filter_dist x ++ l) nil
                  all)).
    intros d hd; rewrite max_list_Some in hd; destruct hd as [hd hd'].
    split.
    { clear hd'.
      apply fr_app in hd.
      2: { intros a b ab; apply map_filter_dist_compat; auto. }
      destruct hd as [hd|hd].
      rewrite InA_nil in hd; tauto.
      destruct hd as [y [_ ina]].
      rewrite InA_map_filter_iff in ina.
      2: { intros a b ab; apply compute_dist_compat; auto. }
      destruct ina as [x [_ hd]].
      rewrite <- is_dist_equiv in hd.
      exists y; exists x; auto. }
    { clear hd.
      intros x y n hn.
      apply hd'.
      apply fr_app.
      { intros a b ab; apply map_filter_dist_compat; auto. }
      right.
      exists x; split; [apply all_here|].
      rewrite InA_map_filter_iff.
      2: { intros a b ab; apply compute_dist_compat; auto. }
      exists y; split; [apply all_here|].
      rewrite <- is_dist_equiv; auto. }
    { intros hm d hd; apply max_list_None in hm.
      apply fr_app_nil in hm.
      destruct hm as [_ hm].
      destruct hd as [[x [y hd]] _].
      specialize (hm x (all_here x)).
      cut (exists y, InA equiv y (map_filter_dist x)).
      { intros [yy ok]; rewrite hm in ok; rewrite InA_nil in ok; tauto. }
      clear hm.
      exists d; rewrite InA_map_filter_iff.
      2: { intros a b ab; apply compute_dist_compat; auto. }
      exists y; split; [apply all_here|].
      apply is_dist_equiv; auto.
      { intros a b ab; apply map_filter_dist_compat; auto. }
    }
  Qed.

  Lemma is_diameter_dec: {d | is_diameter d}+{forall d, ~is_diameter d}.
  Proof.
    generalize (compute_diameter_is_diameter).
    destruct compute_diameter; intros h.
    left; exists n; auto.
    right; auto.
  Qed.

  Lemma is_diameter_dec0:
    forall d, {is_diameter d}+{~is_diameter d}.
  Proof.
    intros d; destruct is_diameter_dec as [[dd h]|]; [|right]; auto.
    case (Nat.eq_dec d dd); intros hh.
    left; rewrite hh; auto.
    right; contradict hh; apply is_diameter_unique; auto.
  Qed.

  Lemma is_diameter_equiv:
    forall d, is_diameter d <-> compute_diameter = Some d.
  Proof.
    intros d.
    generalize (compute_diameter_is_diameter).
    destruct compute_diameter; intros h; split; intros hh.
    assert (n = d) by (apply is_diameter_unique; auto).
    rewrite H; auto.
    inversion hh; subst; auto.
    specialize (h d); contradict h; auto.
    congruence.
  Qed.
  
  Lemma diameter_exists: 
    forall (a: A), {d | is_diameter d}.
  Proof.
    intros a.
    generalize (compute_diameter_is_diameter).
    case_eq compute_diameter. 
    intros; exists n; auto.
    intros h hd.
    unfold compute_diameter in h; rewrite max_list_None in h.
    rewrite fr_app_nil in h.
    destruct h as [_ h].
    specialize (h a (all_here _)).
    cut ({ y | InA equiv y (map_filter_dist a)}).
    { intros [yy ok]; rewrite h in ok; rewrite InA_nil in ok; tauto. }
    clear h.
    exists 0; rewrite InA_map_filter_iff.
    2: { intros aa b ab; apply compute_dist_compat; auto. }
    exists a; split; [apply all_here|].
    apply is_dist_equiv.
    2: { intros aa b ab; apply map_filter_dist_compat; auto. }
    left; split; auto.
  Qed.

  Definition compute_diameter' (a: A): nat :=
    proj1_sig (diameter_exists a).

  Lemma is_diameter_equiv':
    forall a, is_diameter (compute_diameter' a).
  Proof.
    intros a; unfold compute_diameter'.
    destruct (diameter_exists a) as [d hd]; simpl; auto.
  Qed.

  Lemma is_dist_nul:
    forall a b, is_dist a b 0 <-> a == b.
  Proof.
    intros a b; split; unfold is_dist; intros h.
    destruct h as [h|h].
    apply h.
    destruct h as [_ [[p [_ hh]] _]].
    lia.
    left; split; auto.
  Qed.
  
  Lemma is_dist_nul':
    forall a d, is_dist a a d <-> d = 0.
  Proof.
    intros a d; split; unfold is_dist; intros h.
    destruct h as [h|h].
    apply h.
    destruct h as [h _]; contradict h; auto.
    left; split; auto.
  Qed.
  
  Lemma is_dist_one:
    forall a b, ~ a == b -> G a b -> is_dist a b 1.
  Proof.
    intros a b n h; right.
    split; auto; split.
    exists nil; split; simpl; auto.
    intros p hp; lia.
  Qed.

  Lemma is_dist_one':
    forall a b d, ~ a == b -> G a b -> is_dist a b d -> d = 1.
  Proof.
    intros a b d n h hd.
    apply is_dist_one in h; auto.
    eapply is_dist_unique; eauto.
  Qed.

  Lemma is_diameter_pos:
    forall a b d, ~ a == b -> G a b -> is_diameter d -> d > 0.
  Proof.
    intros a b d h hd.
    unfold is_diameter; intros [h1 h2].
    cut (d >= 1); [lia|].
    eapply h2.
    apply is_dist_one; eauto.
  Qed.
  
  Lemma compute_diameter'_pos:
    forall a b, ~ a == b -> G a b -> compute_diameter' a > 0.
  Proof.
    intros a b n h; eapply is_diameter_pos in h; auto.
    2: apply (is_diameter_equiv' a).
    auto.
  Qed.
  
End Graph.

Section Compat_without_G.
  Context {A} `{SA: Setoid A}.

  Global Instance is_path_compat:
    Proper (pequiv ==> equiv ==> eqlistA (equiv) ==> pequiv) (is_path (A := A)).
  Proof.
    unfold is_path; apply non_empty_chain_compat.
  Qed.

  Global Instance is_dist_compat: Proper pequiv (is_dist (A := A)).
  Proof.
    intros g1 g2 eg x1 x2 ex y1 y2 ey n1 n2 en.
    apply or_compat; [rewrite en, ex, ey; tauto|].
    apply and_compat; [rewrite ex, ey; tauto|apply and_compat]. 
    apply ex_compat; intros p.
    apply and_compat.
    apply is_path_compat; auto.
    reflexivity.
    simpl in en; rewrite en; tauto.
    apply all_compat; intros p.
    apply impl_compat.
    apply is_path_compat; auto.
    reflexivity.
    simpl in en; rewrite en; tauto.
  Qed.
  
End Compat_without_G.

Section SubGraph.

  Context {A} `{SA: Setoid A} {DA: Decider (equiv (A:=A))}.

  (** *** Subgraph *)
  Definition is_subgraph (G_sub G: relation A): Prop :=
    inclusion G_sub G.

  Global Instance is_subgraph_compat: Proper pequiv is_subgraph.
  Proof.
    intros G G' eg H H' eh.
    split; etransitivity.
    etransitivity.
    2: apply H0.
    intros a b ab; revert ab; apply eg; auto.
    intros a b ab; revert ab; apply eh; auto.
    etransitivity.
    2: apply H0.
    intros a b ab; revert ab; apply eg; auto.
    intros a b ab; revert ab; apply eh; auto.
  Qed.
  
  Lemma is_path_subgraph:
    forall G G_sub, is_subgraph G_sub G ->
                    forall p x y, is_path G_sub x p y ->
                                  is_path G x p y.
  Proof.
    intros G G_sub hsub.
    induction p as [| a p ih]; intros x y hp; simpl in hp; simpl.
    apply hsub; auto.
    split; [apply hsub|apply ih]; apply hp.
  Qed.

  Lemma is_dist_subgraph:
    forall G G_sub, is_subgraph G_sub G ->
                    forall x y n n_sub,
                      is_dist G x y n -> is_dist G_sub x y n_sub
                      -> n <= n_sub.
  Proof.
    intros G G_sub hsub x y n n_sub
           [[hd hd']|[hd [[p [hd' hd'']] hd''']]]
           [[hd_sub hd_sub']|
            [hd_sub [[p_sub [hd_sub' hd_sub'']] hd_sub''']]];
      try tauto.
    subst; lia.
    cut (S n <= S n_sub); [lia|].
    rewrite <- hd_sub''.
    apply (is_path_subgraph (G := G)) in hd_sub'; auto.
    specialize (hd''' _ hd_sub').
    lia.
  Qed.

  Lemma is_subgraph_dec:
    forall l (finite: forall (a:A), InA equiv a l),
    forall (G' G: relation A),
      Proper pequiv G -> Proper pequiv G' ->
      (forall x y, {G x y} + {~G x y}) ->  
      (forall x y, {G' x y} + {~G' x y}) ->  
      {is_subgraph G' G} + {~is_subgraph G' G}.
  Proof.
    intros lA finite G' G pg pg' gdec g'dec.
    assert (dd: forall x y, {G' x y -> G x y} + {~ (G' x y -> G x y)}).
    { intros x y.
      destruct (g'dec x y); destruct (gdec x y).
      left; auto.
      right; auto.
      left; auto.
      left; tauto.
    }
    assert (a := All_In_dec_forall_notforall_2
                   finite finite _ dd).
    destruct a as [a|a]; [left|right]; [|apply a].
    intros x y; apply a.
    intros l m lm c d cd; apply impl_compat.
    apply pg'; auto.
    apply pg; auto.
  Qed.
  
End SubGraph.

(*
Check (build_simple_path (A := nat) (DA := eq_nat_dec)).
Eval compute in (build_simple_path (A := nat) (DA := eq_nat_dec) (0::0::nil)).
Eval compute in (build_simple_path (A := nat) (DA := eq_nat_dec)
                                   (1::0::1::2::3::0::4::5::0::nil)).

Check (bsp (A := nat) (DA := eq_nat_dec)).
Eval compute in (bsp (A := nat) (DA := eq_nat_dec) (0::0::nil) true 3).
Eval compute in (bsp (A := nat) (DA := eq_nat_dec) (0::0::nil) true 0).
Eval compute in (bsp (A := nat) (DA := eq_nat_dec) (0::0::nil) false 3).
Eval compute in (bsp (A := nat) (DA := eq_nat_dec) (0::0::nil) false 0).
Eval compute in (bsp (A := nat) (DA := eq_nat_dec)
                     (1::0::1::2::3::0::4::5::0::nil) true 1).
Eval compute in (bsp (A := nat) (DA := eq_nat_dec)
                     (1::0::1::2::3::0::4::5::0::nil) false 1).

Eval compute in (bsp (A := nat) (DA := eq_nat_dec)
                     (1::0::1::2::3::0::4::5::0::nil) true 0).

Eval compute in (bsp (A := nat) (DA := eq_nat_dec)
                     (1::0::1::2::3::0::4::5::0::nil) true 2).
Eval compute in (bsp (A := nat) (DA := eq_nat_dec)
                     (1::0::1::2::3::0::4::5::0::nil) false 2).

Eval compute in (bsp (A := nat) (DA := eq_nat_dec)
                     (1::0::1::2::3::0::4::5::0::nil) true 8).
Eval compute in (bsp (A := nat) (DA := eq_nat_dec)
                     (1::0::1::2::3::0::4::5::0::nil) false 8).

Eval compute in (bsp (A := nat) (DA := eq_nat_dec)
                     (3::3::1::1::4::1::6::nil) true 1).
 *)

Section Path_Chain.

  Context {A} `{SA: Setoid A}.
  Variable (G: relation A) (G_compat: Proper pequiv G) (*G_dec: Decider G*).

  (*
  (is_path x p y) means a path exists from x to y, made of the
   consecutive nodes in (x :: p ++ y :: nil). *)
  Lemma is_path_chain:
    forall a b p, is_path G a p b <-> chain G a (p ++ b :: nil) b.
  Proof.
    intros a b p; revert a b; induction p as [| x p ih]; intros a b; simpl.
    - intuition.
    - split; intros [h hh]; split; auto; apply ih; auto.
  Qed.
  
  Lemma chain_is_path:
    forall a b p, chain G a p b <->
                  ( a == b /\ p == nil ) \/
                  exists pp, p == pp ++ b :: nil /\
                             is_path G a pp b.
  Proof.
    intros a b p; split; intros h.
    - assert (p = rev (rev p)).
      symmetry; apply rev_involutive.
      rewrite H in *; clear H.
      destruct (rev p) as [| x rp].
      left; split; auto.
      right.
      simpl in h; simpl.
      assert (x == b).  apply chain_last_eq in h; auto.
      assert (t: is_path G a (rev rp) x).
      apply is_path_chain; auto.
      revert h; apply chain_compat; auto.
      reflexivity.
      exists (rev rp).
      split.
      apply app_eqlistA_compat; [reflexivity|].
      simpl; auto.
      rewrite <- H; auto.
    - destruct h as [[h hh]|[pp [h1 h2]]].
      rewrite hh; simpl; auto.
      apply is_path_chain in h2.
      rewrite h1; auto.
  Qed.
  
End Path_Chain.

Section Tools_for_Distance.

  Context {A} `{SA: Setoid A} {DA: Decider (equiv (A:=A))}.
  Variable (G: relation A) (G_compat: Proper pequiv G) (G_dec: Decider G).

  Hypothesis (Connected: is_connected G).
  
  (** finite set of vertices *)
  Variable (all: list A) (all_here: forall a, InA equiv a all).

  (** * Tools for distance *)
  (** distance: Node -> Node -> nat *)
  Definition dist := (compute_dist'' all_here G_compat
                                     G_dec Connected
                                     (DA := DA)).

  Global Instance dist_compat: Proper pequiv dist.
  Proof. apply compute_dist''_compat. Qed.
  
  Lemma is_dist_dist: forall x y, is_dist G x y (dist x y).
  Proof. intros x y; apply is_dist_equiv''. Qed.
  
  Lemma is_dist_dist': forall (x y: A) n, is_dist G x y n -> dist x y = n.
  Proof. intros x y n hn.
         symmetry; apply (is_dist_unique hn).
         apply is_dist_dist.
  Qed.

  Lemma dist_nil: forall n, dist n n = 0.
  Proof.
    intros n.
    destruct (is_dist_dist n n) as [DD|[DD _]];
      [apply DD|contradict DD; auto].
  Qed.    

  Lemma dist_nil': forall n n', dist n n' = 0 <-> n == n'.
  Proof.
    intros n n'.
    rewrite <- (is_dist_nul G n n').
    split; intros h.
    rewrite <- h; apply is_dist_dist.
    assert (H := is_dist_dist n n').
    eapply is_dist_unique in h; [apply h|apply H].
  Qed.
  
  Lemma dist_neighbors:
    forall (Bidir: forall x y, G x y -> G y x),
    forall r, 
    forall p q, G p q -> dist q r = dist p r \/
                         S (dist q r) = dist p r \/
                         dist q r = S (dist p r).
  Proof.
    intros Bidir r.
    intros p q h; eapply is_dist_neighbors' in h.
    destruct h as [h|[h|h]].
    left; apply h.
    right; right; auto.
    right; left; auto.
    apply G_compat.
    apply Bidir; auto.
    apply is_dist_dist.
    apply is_dist_dist.
  Qed.

  Lemma dist_neighbor_ex:
    forall p r n, dist p r = S n ->
                  exists q, G p q /\ dist q r = n.
  Proof.
    intros p r n hn.
    assert (DD := is_dist_dist p r).
    rewrite hn in DD; eapply is_dist_neighbors_ex in DD.
    2: apply G_compat.
    Unshelve. 2: apply DA. 
    destruct DD as [q [hq hqq]].
    exists q; split; auto.
    apply is_dist_dist'; auto.
  Qed.
  
  (** Tools about diameter *)
  Lemma diam_dist:
    forall (one: A),
      let Diam := compute_diameter'
                    all_here G_compat G_dec one (DA := DA) in
      forall p q, dist p q <= Diam.
  Proof.
    intros one Diam.
    intros p q.
    assert (H: is_diameter G Diam) by apply is_diameter_equiv'.
    unfold is_diameter in H; destruct H as [_ H].
    eapply H.
    apply is_dist_dist.
  Qed.

  Lemma Diam_pos:
    forall one,
      let Diam := compute_diameter'
                    all_here G_compat G_dec one (DA := DA) in
      forall (p q: A), ~ p == q -> Diam > 0.
  Proof.
    intros one Diam p q pq.
    eapply is_connected_has_neighbor' in pq.
    destruct pq as [qq [hqq hg]].
    eapply compute_diameter'_pos. 
    apply hqq.
    apply hg.
    apply G_compat.
    apply Connected.
    Unshelve. apply DA.
  Qed.

End Tools_for_Distance.


Close Scope signature_scope.
Close Scope nat_scope.
Unset Implicit Arguments.
