(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- March 2017
 *)
(* begin show *)

(** Global imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import ZArith.
From Coq Require Import Lia.


(** Local imports *)
From CoLoR Require ListUtil.
(**  *)
From Padec Require Import SetoidUtil.
From Padec Require Import ListUtils.
From Padec Require Import NatUtils.
From Padec Require Import ZUtils.
From Padec Require Import FoldRight.
From Padec Require Import RelModel.
From Padec Require Import Algorithm.
From Padec Require Import Count.

(* end show *)
Set Implicit Arguments.
Open Scope nat_scope.


(** * Ring Topology

  A ring topology on elements of type A is defined using the following 
  functions:
    - a predecessor function Pred to go one step backward on the ring
    - a successor function Succ to go one step forward on the ring
    - a distance function Dist which counts the number of (clockwise) steps
      between any two elements in the ring.

  Convention: 
    - Dist increments (if it can) with Succ
    - Dist decrements (if it can) with Pred 

  Note that the aforementioned functions are total hence every element of A is 
  covered by the structure which is indeed a ring. 

  The definition is setoid-compatible, hence it requires
  - a equivalence on A, considered as the equality on the elements of A and
  - that each function (Pred, Succ, Dist) is compatible w.r.t. this relation.
 *)

Class Ring_Topology {A} `{SA: Setoid A}: Type :=
  mkRing {
      Pred: A -> A;
      Succ: A -> A;
      Dist: A -> A -> nat;
      Pred_compat:> Proper fequiv Pred;
      Succ_compat:> Proper fequiv Succ;
      Dist_compat:> Proper fequiv Dist;
      Pred_Succ_inv: forall x y, Succ x == y <-> x == Pred y;
      Dist_refl: forall x y, Dist x y = 0 <-> x == y;
      Dist_Succ_r: forall x y,
          Dist x (Succ y) = 1 + Dist x y \/ Dist x (Succ y) = 0;
      Dist_Pred_l: forall x y,
          Dist (Pred x) y = 1 + Dist x y \/ Dist (Pred x) y = 0
    }.

Section With_Ring_Topology.

  Context {A} `{Rt: Ring_Topology A}.

  Instance A_equiv: Equivalence equiv := setoid_equiv.

  (** The equality on A is decidable *)
  Lemma equiv_dec: Decider equiv.
  Proof.
    intros x y.
    case_eq (Dist x y).
    - intros d; left; apply Dist_refl; auto.
    - intros n d; right; intros F.
      apply Dist_refl in F; rewrite F in d; lia.
  Qed.

  (** ** Straightforward Properties about Pred, Succ and Dist on a Ring *)
  Lemma Dist_self: forall x, Dist x x = 0.
  Proof.
    intros x; rewrite Dist_refl; reflexivity.
  Qed.

  Lemma Pred_Succ: forall (x: A), Pred (Succ x) == x.
  Proof.
    intro x; symmetry; rewrite <- Pred_Succ_inv; reflexivity.
  Qed.

  Lemma Succ_Pred: forall (x: A), Succ (Pred x) == x.
  Proof.
    intro x; rewrite Pred_Succ_inv; reflexivity.
  Qed.

  Lemma Pred_injective: forall x y, Pred x == Pred y -> x == y.
  Proof.
    intros x y h.
    rewrite <- (Succ_Pred x), <- (Succ_Pred y).
    apply Succ_compat; assumption.
  Qed.

  Lemma Succ_injective: forall x y, Succ x == Succ y -> x == y.
  Proof.
    intros x y h.
    rewrite <- (Pred_Succ x), <- (Pred_Succ y).
    apply Pred_compat; assumption.
  Qed.

  Lemma Dist_Succ_l: forall x y, Dist x y = 1 + Dist (Succ x) y \/
                                 Dist x y = 0.
  Proof.
    intros x y.
    destruct (Dist_Pred_l (Succ x) y) as [h|h];
      rewrite Pred_Succ in h; rewrite h; lia.
  Qed.

  Lemma Dist_Pred_r: forall x y, Dist x y = 1 + Dist x (Pred y) \/
                                 Dist x y = 0.
  Proof.
    intros x y.
    destruct (Dist_Succ_r x (Pred y)) as [h|h];
      rewrite Succ_Pred in h; rewrite h; lia.
  Qed.

  Lemma Dist_Succ: forall x y, Dist x y = Dist (Succ x) (Succ y).
  Proof.
    intros x y.
    destruct (Dist_Succ_l x y) as [h|h].
    rewrite h.
    destruct (Dist_Succ_r (Succ x) y) as [h'|h'].
    rewrite h'.
    auto.
    rewrite Dist_refl in h'.
    cut (equiv x y).
    rewrite <- Dist_refl.
    lia.
    rewrite <- (Pred_Succ x),<- (Pred_Succ y).
    apply Pred_compat,h'.
    rewrite h; rewrite Dist_refl in h.
    assert (h' := Succ_compat _ _ h).
    rewrite <- Dist_refl  in h'; lia.
  Qed.

  Lemma Dist_Pred: forall x y, Dist x y = Dist (Pred x) (Pred y).
  Proof.
    intros x y; rewrite (Dist_Succ (Pred x) (Pred y)).
    do 2 rewrite Succ_Pred.
    reflexivity.
  Qed.

  (** ** Ring size

   The ring size is computed by one ring traversal, going backward or
   equivalently forward.  The equivalence between those two
   definitions is proved below.

   To be able to count, the ring_size function requires to fix a node
   (the origin / 0 for counting) but we show that the result
   (Ring_size) is the same whatever be this node (see
   Ring_size_invariant).  *)
  Definition Ring_size  (x: A): nat := S (Dist (Succ x) x).
  Definition Ring_size' (x: A): nat := S (Dist x (Pred x)).

  Instance Ring_size_compat: Proper pequiv Ring_size.
  Proof.
    intros x1 x2 hx.
    unfold Ring_size.
    apply f_equal, Dist_compat.
    apply Succ_compat, hx.
    apply hx.
  Qed.

  Lemma Ring_size_pos: forall x, Ring_size x > 0.
  Proof.
    intros x.
    unfold Ring_size.
    lia.
  Qed.

  Lemma Ring_size_invariant: forall (x y: A), Ring_size x = Ring_size y.
  Proof.
    intros x y.
    generalize (refl_equal (Dist x y)).
    generalize (Dist x y) at 2.
    intro n; revert x.
    induction n as [| n hn].
    - intros x hxy; rewrite Dist_refl in hxy.
      apply Ring_size_compat, hxy.
    - intros x hx.
      specialize (hn (Succ x)).
      rewrite <- hn.
      {
        clear n y hx hn.
        unfold Ring_size.
        apply f_equal.
        case_eq (Dist (Succ x) x).
        {
          rewrite Dist_refl; intro h.
          do 2 rewrite h.
          symmetry.
          rewrite Dist_refl.
          reflexivity.
        }
        {
          intros n.
          intro hn.
          destruct (Dist_Succ_r (Succ (Succ x)) x) as [hx | hx]; rewrite hx.
          {
            rewrite <- hn.
            transitivity (Dist (Pred (Succ (Succ x))) x).
            rewrite Pred_Succ.
            reflexivity.
            destruct (Dist_Pred_l (Succ (Succ x)) x) as [hxx | hxx];
              rewrite hxx.
            reflexivity.
            elimtype False.
            apply (Nat.neq_0_succ n).
            rewrite <- hxx, <- hn.
            rewrite Pred_Succ.
            reflexivity.
          }
          {
            rewrite <- hn, Dist_refl.
            rewrite Dist_refl in hx.
            assert (h := Pred_compat _ _ hx).
            rewrite Pred_Succ, Pred_Succ in h.
            apply h.
          }
        }
      }
      {
        clear hn.
        destruct (Dist_Pred_l (Succ x) y) as [hxy | hxy];
          rewrite Pred_Succ in hxy; lia.
      }
  Qed.

  Lemma Ring_size_alt: forall (x y: A), Ring_size x = Ring_size' y.
  Proof.
    intros x y.
    transitivity (Ring_size (Pred y)).
    apply Ring_size_invariant.
    apply (f_equal S).
    rewrite Succ_Pred.
    reflexivity.
  Qed.

  Instance Ring_size'_compat: Proper pequiv Ring_size'.
  Proof.
    intros x y exy.
    do 2 rewrite <- (Ring_size_alt x); auto.
  Qed.

  (** ** Interative Tools (power) on Ring Topology *)

  Lemma power_Succ_Pred:
    forall (k: nat) (x y: A),
      power Succ k x == y <-> x == power Pred k y.
  Proof.
    intro k; induction k as [| k hk]; simpl.
    tauto.
    intros; rewrite Pred_Succ_inv, hk.
    rewrite <- power_S.
    reflexivity.
  Qed.

  Lemma Dist_power_Succ:
    forall (k: nat) (x y: A), Dist x y = k -> power Succ k x == y.
  Proof.
    intro k; induction k as [|k hk]; intros x y h; simpl.
    * apply Dist_refl; auto.
    * destruct (Dist_Succ_l x y) as [hh| hh]; try congruence.
      simpl in hh.
      specialize (hk (Succ x) y).
      rewrite <- power_S in hk.
      apply hk.
      apply eq_add_S.
      congruence.
  Qed.

  Lemma Dist_power_Pred:
    forall (k: nat) (x y: A), Dist x y = k -> x == power Pred k y.
  Proof.
    intros; rewrite <- power_Succ_Pred.
    apply Dist_power_Succ; auto.
  Qed.

  Lemma power_Succ_Pred':
    forall (p q: A) d, Pred (power Succ d q) ==
                       power Succ d (Pred q).
  Proof.
    intros p q.
    induction d as [| d ih].
    { simpl; auto. }
    simpl.
    rewrite <- ih.
    rewrite Pred_Succ, Succ_Pred.
    auto.
  Qed.
  
  Lemma power_Pred_Succ':
    forall (p q: A) d, Succ (power Pred d q) ==
                       power Pred d (Succ q).
  Proof.
    intros p q.
    induction d as [| d ih].
    { simpl; auto. }
    simpl.
    rewrite <- ih.
    rewrite Pred_Succ, Succ_Pred.
    auto.
  Qed.

  Lemma Succ_nilpotent: forall x y,
      power Succ (Ring_size y) x == x.
  Proof.
    intros x y; rewrite (Ring_size_invariant y (Pred x)).
    unfold Ring_size.
    simpl.
    rewrite Pred_Succ_inv.
    rewrite Succ_Pred.
    apply Dist_power_Succ.
    reflexivity.
  Qed.

  Lemma Pred_nilpotent: forall x y,
      power Pred (Ring_size y) x == x.
  Proof.
    intros x y; symmetry; rewrite <- power_Succ_Pred.
    apply Succ_nilpotent.
  Qed.

  Lemma Succ_modulo: forall (k: nat) x y,
      power Succ k x == power Succ (k mod (Ring_size y)) x.
  Proof.
    intros k x y; rewrite (Nat.div_mod k (Ring_size y)) at 1.
    2:unfold Ring_size; auto.
    rewrite power_add.
    generalize (k/Ring_size y); intro j.
    induction j as [|j hj].
    - rewrite mult_0_r.
      reflexivity.
    - rewrite <- mult_n_Sm.
      rewrite power_add, Succ_nilpotent.
      assumption.
  Qed.

  Lemma Pred_modulo: forall (k: nat) x y,
      power Pred k x == power Pred (k mod (Ring_size y)) x.
  Proof.
    intros k x y; rewrite (Nat.div_mod k (Ring_size y)) at 1.
    2:unfold Ring_size; auto.
    rewrite power_add.
    generalize (k/Ring_size y); intro j.
    induction j as [|j hj].
    - rewrite mult_0_r.
      reflexivity.
    - rewrite <- mult_n_Sm.
      rewrite power_add, Pred_nilpotent.
      assumption.
  Qed.

  Lemma Dist_unique: forall x1 x2 y1 y2,
      Dist x1 y1 = Dist x2 y2 -> (equiv x1 x2 <-> equiv y1 y2).
  Proof.
    intros x1 x2 y1 y2 h.
    pose (k := Dist x2 y2).
    assert (h1 := Dist_power_Succ _ _ h).
    fold k in h1.
    assert (h2 := Dist_power_Succ _ _ (refl_equal k)).
    clearbody k; clear h.
    split; intro h.
    {
      rewrite <- h1, <- h2.
      apply power_compat; auto; apply Succ_compat.
    }
    {
      rewrite power_Succ_Pred in h1, h2.
      rewrite h1, h2.
      apply power_compat; auto; apply Pred_compat.
    }
  Qed.

  (** ** Inductive Schema on the ring: 
       to prove a property P on every element of the ring, prove that:
       - (base case) it holds at some element x
       - (induction step) if it holds somewhere, let say at element y,
       then it also holds at the successor (Succ y)

       The same Lemma exists with Pred.
   *)
  Lemma Ring_Succ_ind: forall (P: A -> Prop),
      Proper pequiv P ->
      forall x, P x -> (forall y, P y -> P (Succ y)) -> forall z, P z.
  Proof.
    intros P P_compat x hx hsucc z.
    assert (h := Dist_power_Succ x z (refl_equal _)).
    rewrite <- h.
    generalize (Dist x z) as k.
    clear z h.
    induction k as [|k hk]; simpl; auto.
  Qed.

  Lemma Ring_Pred_ind: forall (P: A -> Prop),
      Proper pequiv P ->
      forall x, P x -> (forall y, P y -> P (Pred y)) -> forall z, P z.
  Proof.
    intros P P_compat x hx hpred z.
    assert (h := Dist_power_Pred z x (refl_equal _)).
    rewrite h.
    generalize (Dist z x) as k.
    clear z h.
    induction k as [|k hk]; simpl; auto.
  Qed.

  (** ** Ring made of a unique element (properties for Pred and Succ) *)
  Lemma Degenerate_ring_Succ:
    forall x, Succ x == x <-> forall y, equiv y x.
  Proof.
    split.
    intros h y.
    rewrite <- (Dist_power_Succ x y (refl_equal _)).
    generalize (Dist x y);intro k.
    induction k as [|k hk]; simpl.
    reflexivity.
    rewrite hk.
    apply h.
    intro h; apply h.
  Qed.

  Lemma Degenerate_ring_Pred:
    forall x, x == Pred x <-> forall y, equiv x y.
  Proof.
    intro x; rewrite <- Pred_Succ_inv, Degenerate_ring_Succ.
    intuition; symmetry; auto.
  Qed.

  Lemma Degenerate_ring_one:
    forall x, Ring_size x = 1 -> forall y n, y == n.
  Proof.
    intros x h y n.
    apply Degenerate_ring_Succ.
    rewrite (Ring_size_invariant x n) in h.
    unfold Ring_size in h.
    apply Dist_refl; lia.
  Qed.

  (** Non degenerated ring *)
  Lemma Non_Degenerated_Succ:
    forall x, ~ Succ x == x -> forall y, ~ Succ y == y.
  Proof.
    intros x nx y.
    contradict nx.
    rewrite Degenerate_ring_Succ in nx.
    rewrite nx; symmetry; apply nx; auto.
  Qed.

  Lemma Non_Degenerated_Pred:
    forall x, ~ Pred x == x -> forall y, ~ Pred y == y.
  Proof.
    intros x nx y.
    contradict nx.
    symmetry in nx; rewrite Degenerate_ring_Pred in nx.
    symmetry; rewrite <- nx; apply nx; auto.
  Qed.

  Lemma Non_Degenerated_size_Succ:
    forall x, Ring_size x > 1 -> forall y, ~ y == Succ y.
  Proof.
    intros x h y.
    rewrite (Ring_size_invariant x y) in h.
    unfold Ring_size in h.
    intros F; symmetry in F; rewrite F, Dist_self in h; lia.
  Qed.

  Lemma Non_Degenerated_size_Pred:
    forall x, Ring_size x > 1 -> forall y, ~ y == Pred y.
  Proof.
    intros x h y f.
    apply Succ_compat in f.
    rewrite Succ_Pred in f.
    symmetry in f; contradict f.
    apply Non_Degenerated_size_Succ with x; auto.
  Qed.

  Lemma Dist_add_Ring_size:
    forall x y z, x == y \/ Ring_size z = Dist x y + Dist y x .
  Proof.
    intros x y z.
    pattern y.
    eapply Ring_Succ_ind with x; clear y.
    {
      intros a1 a2 ea; rewrite ea; reflexivity.
    }
    {
      left; reflexivity.
    }
    {
      intros y [h|h].
      {
        rewrite <- h; clear h.
        destruct (Dist_Pred_r x (Succ x)) as [h|h].
        {
          rewrite Pred_Succ, Dist_self in h;simpl in h.
          right; rewrite (Ring_size_invariant z x);
            unfold Ring_size;lia.
        }
        {
          left; rewrite <- Dist_refl; assumption.
        }
      }
      {
        destruct (Dist_Succ_r x y) as [h1|h1].
        2: rewrite Dist_refl in h1; left; assumption.
        destruct (Dist_Succ_l y x) as [h2|h2].
        right; lia.
        {
          rewrite Dist_refl in h2.
          rewrite h2, Dist_self in *.
          unfold Ring_size in h; lia.
        }
      }
    }
  Qed.

  (** ** Between 
    Given 3 elements x y z, Between x y z provides the order of the 
    element x y z on the ring, _following the successors_     

    Between forbids to perform a full traversal, namely, x and z
    should not be the same element of the ring, unless y is also
    equal to x and z.
   *)   
  Definition Between (x y z: A): Prop := Dist x y + Dist y z = Dist x z.

  Instance Between_compat: Proper pequiv Between.
  Proof.
    intros x1 x2 ex y1 y2 ey z1 z2 ez.
    unfold Between.
    rewrite ex, ey, ez.
    reflexivity.
  Qed.

  Lemma Between_dec:
    forall x y z, {Between x y z} + {~Between x y z}.
  Proof.
    intros; unfold Between. 
    destruct (Nat.eq_dec (Dist x y + Dist y z) (Dist x z));
      [left|right]; auto.
  Qed.

  Lemma Between_Succ:
    forall x y z, Between x y z <-> Between (Succ x) (Succ y) (Succ z).
  Proof.
    intros x y z; unfold Between; do 3 rewrite <- Dist_Succ; reflexivity.
  Qed.

  Lemma Between_Pred:
    forall x y z, Between x y z <-> Between (Pred x) (Pred y) (Pred z).
  Proof.
    intros x y z; unfold Between; do 3 rewrite <- Dist_Pred; reflexivity.
  Qed.

  Lemma Between_eq:
    forall x y z, Between x y z -> (x == z <-> (x == y /\ y == z)).
  Proof.
    unfold Between; intros x y z h; do 3 rewrite <- Dist_refl.
    lia.
  Qed.

  Lemma Between_strict:
    forall x y z, Between x y z -> ~x == z \/ x == y /\ y == z.
  Proof.
    intros x y z h.
    destruct (Between_eq h) as [b _].
    destruct (equiv_dec x z); intuition.
  Qed.
  
  Lemma Between_circ:
    forall x y z, Between x y z -> x == y \/ Between y z x.
  Proof.
    intros x y z b; unfold Between in *.
    destruct (Between_strict b) as [bs|]; [| tauto].
    destruct (Dist_add_Ring_size x y x) as [dx|];[left; auto | right].
    destruct (Dist_add_Ring_size x z x) as [dz|];[tauto |].
    lia.
  Qed.

  Lemma Between_eq_l:
    forall x y z, x == y -> Between x y z.
  Proof.
    unfold Between; intros x y z h;
      rewrite h; rewrite h in h; rewrite <- Dist_refl in h.
    lia.
  Qed.

  Lemma Between_eq_r:
    forall x y z, y == z -> Between x y z.
  Proof.
    unfold Between; intros x y z h;
      rewrite h; rewrite h in h; rewrite <- Dist_refl in h.
    lia.
  Qed.

  Lemma Between_self:
    forall x y, Between x y x <-> x == y.
  Proof.
    unfold Between; intros x y.
    rewrite Dist_self; rewrite <- Dist_refl; split.
    lia.
    rewrite Dist_refl; intro e; rewrite e, Dist_self.
    lia.
  Qed.

  Lemma Between_self_l:
    forall x y, Between x x y.
  Proof.
    unfold Between; intros x y; rewrite Dist_self; lia.
  Qed.

  Lemma Between_self_r:
    forall y z, Between y z z.
  Proof.
    unfold Between; intros y z; rewrite Dist_self; lia.
  Qed.

  Lemma Between_self_lr:
    forall x, Between x x x.
  Proof.
    intro x;apply Between_self_l.
  Qed.

  Lemma Between_Succ_l:
    forall x y z, Between x y z ->
                  x == y \/ Between (Succ x) y z.
  Proof.
    unfold Between; intros x y z h; rewrite <- Dist_refl.
    generalize (Dist_Succ_l x y) (Dist_Succ_l x z).
    intuition (try lia).
  Qed.

  Lemma Between_Pred_l:
    forall x y z, Between x y z ->
                  Pred x == z \/ Between (Pred x) y z.
  Proof.
    unfold Between; intros x y z h; rewrite <- Dist_refl.
    generalize (Dist_Pred_l x y) (Dist_Pred_l x z).
    intuition (try lia).
    rewrite Dist_refl in H1.
    rewrite H1 in H.
    lia.
  Qed.

  Lemma Between_Succ_m:
    forall x y z, Between x y z ->
                  y == z \/ Between x (Succ y) z.
  Proof.
    unfold Between; intros x y z h; rewrite <- Dist_refl.
    generalize (Dist_Succ_r x y) (Dist_Succ_l y z).
    intuition (try lia).
    rewrite Dist_refl in H1.
    rewrite <- H1 in H.
    lia.
  Qed.

  Lemma Between_Pred_m:
    forall x y z, Between x y z ->
                  x == y \/ Between x (Pred y) z.
  Proof.
    unfold Between; intros x y z h; rewrite <- Dist_refl.
    generalize (Dist_Pred_r x y) (Dist_Pred_l y z).
    intuition (try lia).
    rewrite Dist_refl in H.
    rewrite H in H1.
    lia.
  Qed.

  Lemma Between_Succ_r:
    forall x y z, Between x y z ->
                  x == Succ z \/ Between x y (Succ z).
  Proof.
    unfold Between; intros x y z h; rewrite <- Dist_refl.
    generalize (Dist_Succ_r y z) (Dist_Succ_r x z).
    intuition (try lia).
    rewrite Dist_refl in H1.
    rewrite <- H1 in H.
    lia.
  Qed.

  Lemma Between_Pred_r:
    forall x y z, Between x y z ->
                  y == z \/ Between x y (Pred z).
  Proof.
    unfold Between; intros x y z h; rewrite <- Dist_refl.
    generalize (Dist_Pred_r y z) (Dist_Pred_r x z).
    intuition (try lia).
  Qed.

  Lemma Between_dist_le_l:
    forall x y z, Dist x y <= Dist x z <-> Between x y z.
  Proof.
    intros x y z; unfold Between; split; try lia.
    pattern z; eapply Ring_Succ_ind with y.
    { intros a1 a2 ea; rewrite ea; reflexivity. }
    { rewrite Dist_self; lia. }
    clear z; intros z h.
    destruct (Dist_Succ_r x z) as [hx|hx]; rewrite hx.
    2: rewrite Nat.le_0_r; intro hxy; rewrite Dist_refl in hxy;
      rewrite hxy in *; rewrite Dist_self; lia.
    destruct (Dist_Succ_r y z) as [hy|hy]; rewrite hy.
    2: rewrite Dist_refl in hy; rewrite <- hy in hx; lia.
    intro hxy; destruct (le_lt_eq_dec hxy) as [LT|EQ].
    lia.
    clear hxy.
    rewrite <- EQ in hx.
    assert (hh := proj1 (Dist_unique _ _ _ _ hx)
                        (Equivalence_Reflexive _)); clear hx EQ.
    rewrite hh, Dist_self in hy.
    lia.
  Qed.

  Lemma Between_dist_le_r:
    forall x y z, Dist y z <= Dist x z <-> Between x y z.
  Proof.
    intros x y z; unfold Between; split; try lia.
    pattern x; eapply Ring_Pred_ind with y.
    { intros a1 a2 ea; rewrite ea; reflexivity. }
    { rewrite Dist_self; lia. }
    clear x; intros x h.
    destruct (Dist_Pred_l x z) as [hz|hz]; rewrite hz.
    2: rewrite Nat.le_0_r; intro hyz; rewrite Dist_refl in hyz;
      rewrite hyz in *; rewrite Dist_self; lia.
    destruct (Dist_Pred_l x y) as [hy|hy]; rewrite hy.
    2: rewrite Dist_refl in hy; rewrite hy in hz; lia.
    intro hyz; destruct (le_lt_eq_dec hyz) as [LT|EQ].
    lia.
    clear hyz.
    rewrite <- EQ in hz.
    assert (hh := proj2 (Dist_unique _ _ _ _ hz)
                        (Equivalence_Reflexive _)); clear hz EQ.
    rewrite hh, Dist_self in hy.
    lia.
  Qed.

  Lemma Between_alt_l:
    forall x y z, Between x y z \/ Between x z y.
  Proof.
    intros x y z;
      rewrite <- (or_compat _ _ (Between_dist_le_l x y z)
                            _ _ (Between_dist_le_l x z y)).
    lia.
  Qed.

  Lemma Between_alt_r:
    forall x y z, Between x y z \/ Between y x z.
  Proof.
    intros x y z;
      rewrite <- (or_compat _ _ (Between_dist_le_r x y z)
                            _ _ (Between_dist_le_r y x z)).
    lia.
  Qed.

  Lemma Between_alt_m:
    forall x y z, x == z \/ Between x y z \/ Between z y x.
  Proof.
    intros x y z.
    destruct (Dist_add_Ring_size x z y) as [ey|hy].
    auto.
    destruct (Dist_add_Ring_size z y y) as [ez|hz].
    right; right; apply Between_eq_l; auto.
    right.
    rewrite <- (or_compat _ _ (Between_dist_le_r x y z)
                          _ _ (Between_dist_le_l z y x)).
    lia.
  Qed.

  Lemma all_Between_Succ: forall x y, Between (Succ x) y x.
  Proof.
    intros x; eapply Ring_Succ_ind with (Succ x).
    hnf; intros; apply Between_compat; auto; reflexivity.
    apply Between_eq_l; reflexivity.
    intros y h; destruct (Between_Succ_m h); auto.
    apply Between_eq_l.
    symmetry; apply Succ_compat, H.
  Qed.

  Lemma all_Between_Pred: forall x y, Between x y (Pred x).
  Proof.
    intros x y; generalize (all_Between_Succ (Pred x) y).
    rewrite Succ_Pred; auto.
  Qed.

  Lemma Between_rewrite_l:
    forall x y, y == x -> forall z t, Between x z t -> Between y z t.
  Proof.
    intros x y xy z t; apply Between_compat; auto.
  Qed.

  Lemma Between_rewrite_m:
    forall x y, y == x -> forall z t, Between z x t -> Between z y t.
  Proof.
    intros x y xy z t; apply Between_compat; auto.
  Qed.

  Lemma Between_rewrite_r:
    forall x y, y == x -> forall z t, Between z t x -> Between z t y.
  Proof.
    intros x y xy z t; apply Between_compat; auto.
  Qed.

  Lemma Dist_lt_Ring_size: forall r x y, Dist x y < Ring_size r.
  Proof.
    intros r x y; rewrite (Ring_size_invariant r y).
    unfold lt.
    apply le_n_S.
    assert (h := all_Between_Succ y x).
    hnf in h.
    lia.
  Qed.

  Lemma Between_not_Succ_m:
    forall x y, ~ x == Succ y -> ~ Between x (Succ y) y.
  Proof.
    intros x y nx B.
    unfold Between in B.
    assert (dr := Dist_Succ_l y y).
    assert (dl := Dist_Succ_r x y).
    destruct dr as [dr|dr]; destruct dl as [dl|dl]; try lia.
    apply Dist_refl in dl; tauto.
    apply Dist_refl in dl; tauto.
  Qed.

  Lemma Between_not_Pred_m:
    forall x y, ~ x == Pred y -> ~ Between y (Pred y) x.
  Proof.
    intros x y nx B.
    unfold Between in B.
    assert (dr := Dist_Pred_l y x).
    assert (dl := Dist_Pred_r y y).
    destruct dr as [dr|dr]; destruct dl as [dl|dl]; try lia.
    apply Dist_refl in dr; symmetry in dr; tauto.
    apply Dist_refl in dr; symmetry in dr; tauto.
  Qed.

  Lemma Between_not_Pred_l:
    forall x y, ~ x == y -> ~ Between x y (Pred y).
  Proof.
    intros x y nx B.
    unfold Between in B.
    assert (dr := Dist_Pred_r x y).
    assert (dl := Dist_Pred_r y y).
    destruct dr as [dr|dr]; destruct dl as [dl|dl]; try lia.
    apply Dist_refl in dr; tauto.
    apply Dist_refl in dr; tauto.
  Qed.

  Lemma Between_not_Succ_r:
    forall x y, ~ x == y -> ~ Between (Succ x) x y.
  Proof.
    intros x y nx B.
    unfold Between in B.
    assert (dr := Dist_Succ_l x y).
    assert (dl := Dist_Succ_l x x).
    destruct dr as [dr|dr]; destruct dl as [dl|dl]; try lia.
    apply Dist_refl in dr; tauto.
    apply Dist_refl in dr; tauto.
  Qed.

  
  Lemma Between_Succ_ind:
    forall (P: A -> Prop) (P_compat: Proper pequiv P),
    forall p q,
      P p ->
      (forall n', Between p n' (Pred q) -> P n' -> P (Succ n')) ->
      forall n, Between p n q -> P n.
  Proof.
    intros P P_compat p q init hi n.
    pattern n.
    apply Ring_Succ_ind with p.
    { intros n1 n2 en.
      apply impl_compat.
      apply Between_compat; auto.
      apply P_compat; auto.
    }
    { auto. }
    { intros nn hhi hb.
      case (equiv_dec (Succ nn) nn); intros deg.
      { (*degenerated ring *)
        rewrite Degenerate_ring_Succ in deg.
        rewrite deg; rewrite deg in init; auto.
      }
      case (equiv_dec p (Succ nn)); intros hp.
      { rewrite <- hp; auto. }
      case (equiv_dec nn q); intros hq.
      { eapply Between_rewrite_r in hb; [| apply hq].
        contradict hb; apply Between_not_Succ_m; auto.
      }                
      apply Between_Pred_m in hb; destruct hb as [hb|hb];
        [tauto|].
      eapply Between_rewrite_m in hb;
        [| symmetry; apply Pred_Succ].
      apply hi; auto.
      apply Between_Pred_r in hb; destruct hb as [hb|hb];
        [tauto|auto].
    }
  Qed.    

  Lemma Dist_Pred_full:
    forall x, forall y, Dist y (Pred y) = Ring_size x - 1.
  Proof.
    intros x y.
    destruct (equiv_dec y (Pred y)) as [D|D].
    { unfold Ring_size.
      rewrite <- D.
      apply Succ_compat in D.
      rewrite Succ_Pred in D.
      assert (Succ x == x).
      rewrite Degenerate_ring_Succ in D.
      rewrite D; symmetry; apply D.
      rewrite H.
      rewrite Dist_self.
      rewrite Dist_self.
      lia.
    }
    destruct (Dist_add_Ring_size y (Pred y) x) as [DD|DD];
      [tauto|].
    destruct (Dist_Pred_l y y) as [P|P].
    rewrite Dist_self in P; lia.
    rewrite Dist_refl in P; symmetry in P; tauto.
  Qed.        
  
  Lemma Dist_Succ_full:
    forall x, forall y, Dist (Succ y) y = Ring_size x - 1.
  Proof.
    intros x y.
    erewrite <- Dist_Pred_full.
    apply Dist_compat.
    reflexivity.
    symmetry; apply Pred_Succ.
  Qed.

  Lemma Between_trans1:
    forall y a b c, Between b a y -> Between (Succ y) a c ->
                    Between b c y.

  Proof.
    intros y a b c C1 C2.
    case (equiv_dec a (Succ y)); intros ax.
    { clear C2; unfold Between in C1.
      rewrite ax in C1.
      rewrite (Dist_Succ_full y) in C1.
      assert (DR := Dist_lt_Ring_size y b y).
      assert (D0: Dist b (Succ y) = 0) by lia.
      apply Dist_refl in D0.
      eapply Between_rewrite_l; [apply D0|].
      apply all_Between_Succ.
    }
    case (equiv_dec a y); intros apx.
    { clear C1; unfold Between in C2. 
      rewrite apx in C2.
      rewrite (Dist_Succ_full y) in C2.
      assert (DR := Dist_lt_Ring_size y (Succ y) c). 
      assert (D0: Dist y c = 0) by lia.
      apply Dist_refl in D0.
      eapply Between_rewrite_r; [apply D0|].
      apply Between_self_r.
    }
    case (equiv_dec (Succ y) c); intros xc.
    { eapply Between_rewrite_r in C2; [|apply xc].
      apply Between_self in C2.
      symmetry in C2; tauto.
    }
    case (equiv_dec b y); intros bpx.
    { eapply Between_rewrite_l in C1;
        [|symmetry in bpx; apply bpx].
      apply Between_self in C1.
      symmetry in C1; tauto.
    }  
    case (equiv_dec c y); intros cpx.
    { eapply Between_rewrite_m; [apply cpx|].
      apply Between_self_r.
    }
    destruct (Between_alt_m b c y) as [h|[h|h]].
    tauto.
    auto.
    unfold Between in C1, C2, h.
    assert (dbx := Dist_add_Ring_size b y y).
    destruct dbx as [dbx|dbx]; [tauto|].
    assert (dax := Dist_add_Ring_size a (Succ y) y).
    destruct dax as [dax|dax]; [tauto|].
    assert (dpx := Dist_Pred_r a (Succ y));
      rewrite Pred_Succ in dpx.
    destruct dpx as [dpx|dpx];
      [| apply Dist_refl in dpx; tauto].
    assert (dpxc := Dist_Pred_l (Succ y) c).
    destruct dpxc as [dpxc|dpxc]; rewrite Pred_Succ in dpxc;
      [| apply Dist_refl in dpxc; symmetry in dpxc; tauto].
    assert (ok: Dist c b = 0) by lia.
    apply Dist_refl in ok.
    eapply Between_rewrite_l; [symmetry in ok; apply ok|].
    apply Between_self_l.
  Qed.

  Lemma Between_Pred_ind:
    forall (P: A -> Prop) (P_compat: Proper pequiv P),
    forall p q,
      P q ->
      (forall n', Between (Succ p) n' q -> P n' -> P (Pred n')) ->
      forall n, Between p n q -> P n.
  Proof.
    intros P P_compat p q init hi n.
    pattern n.
    apply Ring_Pred_ind with q.
    { intros n1 n2 en.
      apply impl_compat.
      apply Between_compat; auto.
      apply P_compat; auto.
    }
    { auto. }
    { intros nn hhi hb.
      case (equiv_dec nn (Pred nn)); intros deg.
      { (*degenerated ring *)
        rewrite Degenerate_ring_Pred in deg.
        rewrite <- deg; rewrite <- deg in init; auto.
      }
      case (equiv_dec (Pred nn) q); intros hp.
      { rewrite hp; auto. }
      case (equiv_dec p nn); intros hq.
      { eapply Between_rewrite_l in hb; [| symmetry; apply hq].
        contradict hb; apply Between_not_Pred_m;
          intros F; symmetry in F; auto.
      }                
      apply Between_Succ_m in hb; destruct hb as [hb|hb];
        [tauto|].
      eapply Between_rewrite_m in hb;
        [| symmetry; apply Succ_Pred].
      apply hi; auto.
      apply Between_Succ_l in hb; destruct hb as [hb|hb];
        [tauto|auto].
    }
  Qed.

  Lemma Between_all_ind:
    forall (P: A -> A -> Prop) (P_compat: Proper pequiv P),
      (forall n, P n n) ->
      (forall m n, ~ m == n ->
                   forall n', Between (Succ m) n' n ->
                              P n' n -> P (Pred n') n) ->
    forall m n, P m n.
  Proof using Type.
    intros P P_compat hbase hind m n.
    case (equiv_dec m n); intros hmn.
    { rewrite hmn; apply hbase. }
    
    cut (forall mm,
            Between m mm n -> P mm n).
    { intros ok; apply ok.
      apply Between_self_l. }

    apply (Between_Pred_ind (P := fun y => P y n)).
    { intros n1 n2 en; apply P_compat; auto. }
    { apply hbase. }
    { intros n' B ih.
      apply hind with m; auto. }
  Qed.


  Lemma Between_trans2:
    forall a b n x,
      Between a n b -> Between a x n -> Between a x b.
  Proof.
    intros a b n x B1 B2.
    case (equiv_dec a n); intros hn.
    { eapply Between_rewrite_r in B2; [| apply hn].
      apply Between_self in B2.
      eapply Between_rewrite_l; [apply B2|].
      apply Between_self_l.
    }
    revert B2; revert x.
    eapply (Between_Pred_ind 
              (P := fun x => Between a x b)).
    { intros n1 n2 en; apply Between_compat; auto. }
    auto.
    intros nn B2 B3.
    destruct (Between_Pred_m B3).
    { eapply Between_rewrite_m in B2; [| apply H].
      contradict B2; apply Between_not_Succ_r.
      auto.
    }
    auto.
  Qed.

  Lemma Between_trans3:
    forall a b n x,
      Between a n b -> Between n x b -> Between a x b.
  Proof.
    intros a b n x B1 B2.
    case (equiv_dec n b); intros hn.
    { eapply Between_rewrite_l in B2; [| symmetry; apply hn].
      apply Between_self in B2.
      eapply Between_rewrite_r; [apply B2|].
      apply Between_self_r.
    }
    revert B2; revert x.
    eapply (Between_Succ_ind 
              (P := fun x => Between a x b)).
    { intros n1 n2 en; apply Between_compat; auto. }
    auto.
    intros nn B2 B3.
    destruct (Between_Succ_m B3).
    { eapply Between_rewrite_m in B2; [| symmetry; apply H].
      contradict B2; apply Between_not_Pred_l.
      auto.
    }
    auto.
  Qed.

  Lemma Between_trans4:
    forall a b n x,
      Between a n b -> Between n x b -> Between a n x.
  Proof.
    intros a b n x B1 B2.
    case (equiv_dec n b); intros hn.
    { eapply Between_rewrite_r in B2; [| apply hn].
      apply Between_self in B2.
      eapply Between_rewrite_r; [symmetry; apply B2|].
      apply Between_self_r.
    }
    revert B2; revert x.
    eapply (Between_Pred_ind 
              (P := fun x => Between a n x)).
    { intros n1 n2 en; apply Between_compat; auto. }
    auto.
    intros nn B2 B3.
    destruct (Between_Pred_r B3).
    { eapply Between_rewrite_m in B2; [| apply H].
      contradict B2; apply Between_not_Succ_r.
      auto.
    }
    auto.    
  Qed.

  Lemma power_Succ_Dist_mod:
    forall (k: nat) x y z,
      power Succ k x == y <-> Dist x y = k mod (Ring_size z).
  Proof.
    intros k x y z; split.
    {
      revert x y z; induction k as [|k hk]; intros x y z h.
      {
        simpl in h; rewrite <- Dist_refl in h.
        simpl.
        lia.
      }
      {
        rewrite power_S in h.
        specialize (hk (Succ x) y z h).
        replace (S k) with (1+k).
        2: lia.
        rewrite Nat.add_mod.
        2: unfold Ring_size; lia.
        {
          destruct (Dist_Succ_l x y) as [hh|hh]; rewrite hh.
          {
            rewrite (Nat.mod_small 1).
            rewrite (Nat.mod_small (1 + k mod Ring_size z)).
            lia.
            rewrite <- hk, <- hh.
            apply Dist_lt_Ring_size.
            apply le_n_S; case_eq (Dist (Succ z) z); intros; try lia.
            rewrite Dist_refl, Degenerate_ring_Succ in H.
            rewrite H in hh at 1.
            rewrite (H y) in hh at 1.
            rewrite Dist_self in hh.
            lia.
          }
          {
            rewrite Dist_refl in hh.
            rewrite <- hh in *.
            clear hh.
            destruct (@le_lt_eq_dec 1 (Ring_size z)).
            unfold Ring_size; lia.
            2: rewrite <- e; reflexivity.
            rewrite (Nat.mod_small 1); auto.
            rewrite <- hk.
            rewrite <- (Nat.mod_same (Ring_size z)) at 1.
            2: unfold Ring_size; lia.
            apply Nat.mod_wd.
            apply Ring_size_invariant.
            reflexivity.
          }
        }
      }
    }
    {
      intro h; assert (hh := Dist_power_Succ _ _ h).
      rewrite (Succ_modulo _ _ z).
      assumption.
    }
  Qed.

  Lemma power_Pred_Dist_mod:
    forall (k: nat) x y z,
      x == power Pred k y <-> Dist x y = k mod (Ring_size z).
  Proof.
    intros k x y z.
    rewrite <- power_Succ_Pred.
    apply power_Succ_Dist_mod; auto.
  Qed.

  Lemma Between_alt_m_bis:
    forall x y z, ~ Succ x == x -> ~ x == Succ z ->
                  Between x y z ->
                  ~ Between (Succ z) y (Pred x). 
  Proof.
    intros x y z nondeg xsz B F.
    assert (ND := Non_Degenerated_Succ _ nondeg); clear nondeg.
    unfold Between in *.
    case (equiv_dec y (Succ z)); intros ysz.
    { clear F.
      destruct (Dist_add_Ring_size y z x)
        as [DD|DD].
      { rewrite DD in ysz; symmetry in ysz; specialize (ND z); tauto. }
      rewrite ysz in DD, B.
      destruct (Dist_Succ_r z z) as [D|D].
      rewrite Dist_self in D.
      assert (h := Dist_lt_Ring_size x x z).
      assert (F: Dist x (Succ z) = 0) by lia.
      apply Dist_refl in F; tauto.
      { apply Dist_refl in D; symmetry in D; specialize (ND z); tauto. }
    }
    case (equiv_dec y (Pred x)); intros ypx.
    { clear F.
      destruct (Dist_add_Ring_size x y x)
        as [DD|DD].
      { rewrite DD in ypx; apply Succ_compat in ypx.
        rewrite Succ_Pred in ypx.
        specialize (ND y); tauto. }
      rewrite ypx in DD, B.    
      destruct (Dist_Pred_l x x) as [D|D].
      rewrite Dist_self in D.
      assert (h := Dist_lt_Ring_size x x z).
      assert (F: Dist (Pred x) z = 0) by lia.
      apply Dist_refl in F; apply Succ_compat in F; rewrite Succ_Pred in F; tauto.
      { apply Dist_refl in D; apply Succ_compat in D; rewrite Succ_Pred in D.
        symmetry in D; specialize (ND x); tauto. }
    }
    assert (xz: ~ x == z).
    { intros FF; rewrite FF in *.
      rewrite Dist_self in B.
      assert (d: Dist z y = 0) by lia.
      rewrite Dist_refl in d; rewrite d in *.
      assert (BB := Between_not_Succ_r ypx).
      unfold Between in BB; tauto.
    }
    destruct (Dist_Pred_l x y) as [AA|AA].
    2:{ apply Dist_refl in AA; symmetry in AA; tauto. }
    destruct (Dist_Succ_r y z) as [C|C].
    2:{ apply Dist_refl in C; tauto. }
    destruct (Dist_add_Ring_size x z x) as [N1|N1]; [tauto|].
    destruct (Dist_Succ_r z (Pred x)) as [zx|zx]; rewrite Succ_Pred in zx.
    2:{ rewrite Dist_refl in zx; symmetry in zx; tauto. }
    destruct (Dist_Pred_l (Succ z) (Pred x)) as [szpx|szpx];
      rewrite Pred_Succ in szpx.
    2:{ apply Dist_refl, Succ_compat in szpx.
        rewrite Succ_Pred in szpx; symmetry in szpx; tauto. }
    destruct (Dist_add_Ring_size (Pred x) y x) as [N2|N2];
      [ symmetry in N2; tauto|].
    destruct (Dist_add_Ring_size y (Succ z) x) as [N3|N3]; [ tauto|].
    lia.
  Qed.

  (** ** List of the ring elements *)
  
  (** Build the list of n elements of the ring, starting from
      element x *)
  Fixpoint traversal_Succ (n: nat) (x: A): list A :=
    match n with | 0 => nil | S p => x :: (traversal_Succ p (Succ x)) end.

  Fixpoint traversal_Pred (n: nat) (x: A): list A :=
    match n with | 0 => nil | S p => x :: (traversal_Pred p (Pred x)) end.

  Lemma traversal_Succ_compat:
    Proper (eq ==> equiv ==> eqlistA equiv) traversal_Succ.
  Proof.
    intros n_ n en; subst.
    induction n as [|n hn]; intros a1 a2 ea; simpl.
    left.
    right; auto.
    apply hn, Succ_compat, ea.
  Qed.

  Lemma traversal_Succ_compat_equiv:
    Proper (eq ==> equiv ==> equivlistA equiv) traversal_Succ.
  Proof.
    hnf; intros; hnf; intros; apply eqlistA_equivlistA; auto using A_equiv.
    apply traversal_Succ_compat; auto.
  Qed.

  Lemma traversal_Pred_compat:
    Proper (eq ==> equiv ==> eqlistA equiv) traversal_Pred.
  Proof.
    intros n_ n en; subst.
    induction n as [|n hn]; intros a1 a2 ea; simpl.
    left.
    right; auto.
    apply hn, Pred_compat, ea.
  Qed.

  Lemma traversal_Pred_compat_equiv:
    Proper (eq ==> equiv ==> equivlistA equiv) traversal_Pred.
  Proof.
    hnf; intros; hnf; intros; apply eqlistA_equivlistA; auto using A_equiv.
    apply traversal_Pred_compat; auto.
  Qed.

  (** *)
  Lemma traversal_Succ_app:
    forall x (k: nat),
      eqlistA equiv (traversal_Succ (S k) x)
              ((traversal_Succ k x)++(power Succ k x)::nil).
  Proof.
    intros x k; revert x; induction  k as [|k Hk]; intro x.
    reflexivity.
    {
      simpl in *.
      right.
      reflexivity.
      specialize (Hk (Succ x)).
      rewrite <- power_S in Hk.
      auto.
    }
  Qed.

  Lemma traversal_Succ_Pred_rev:
    forall x (k: nat),
      eqlistA equiv (traversal_Succ k x)
              (rev (traversal_Pred k (power Succ k (Pred x)))).
  Proof.
    intros x k.
    revert x; induction k as [| k ih].
    { simpl; intros _; reflexivity. }
    { intros x.
      rewrite traversal_Succ_app.
      simpl.
      assert (one: equiv (power Succ k x) (Succ (power Succ k (Pred x)))).
      { clear - SA.
        induction k as [| k ih]; simpl.
        rewrite Succ_Pred; reflexivity.
        apply Succ_compat, ih.
      }
      etransitivity.
      apply eqlistA_app; auto using A_equiv.
      apply eqlistA_app; auto using A_equiv.
      apply rev_eqlistA_compat, traversal_Pred_compat; auto.
      rewrite Pred_Succ; reflexivity.
    }
  Qed.
  
  Lemma traversal_Pred_Succ_rev:
    forall x (k: nat),
      eqlistA equiv (traversal_Pred k x)
              (rev (traversal_Succ k (power Pred k (Succ x)))).
  Proof.
    intros x k.
    rewrite rev_rev_eqlistA; auto.
    rewrite traversal_Succ_Pred_rev    .
    apply rev_eqlistA_compat, traversal_Pred_compat; auto.
    symmetry; rewrite power_Succ_Pred.
    rewrite <- S_power, power_S, Pred_Succ.
    reflexivity.
  Qed.

  Lemma traversal_Succ_Pred_equiv:
    forall x (k: nat),
      equivlistA equiv (traversal_Succ k x)
                 (traversal_Pred k (power Succ k (Pred x))).
  Proof.
    symmetry; rewrite rev_equivlistA; auto.
    symmetry; apply eqlistA_equivlistA, traversal_Succ_Pred_rev;
      auto using A_equiv.
  Qed.
  
  Lemma traversal_Pred_Succ_equiv:
    forall x (k: nat),
      equivlistA equiv (traversal_Pred k x)
                 (traversal_Succ k (power Pred k (Succ x))).
  Proof.
    intros x k.
    symmetry; rewrite rev_equivlistA; auto.
    symmetry; apply eqlistA_equivlistA, traversal_Pred_Succ_rev;
      auto using A_equiv.
  Qed.

  Lemma traversal_Succ_Dist_Between:
    forall x y z,
      InA equiv y (traversal_Succ (S (Dist x z)) x) <-> Between x y z.
  Proof.
    intros x y z; revert y; pattern x; eapply Ring_Pred_ind with z.
    {
      intros z1 z2 ez.
      apply all_compat.
      intro y'; apply iff_compat.
      apply InA_compat; auto using A_equiv.
      apply traversal_Succ_compat_equiv; auto.
      rewrite ez; reflexivity.
      rewrite ez; reflexivity.
    }
    {
      intros y; rewrite Between_self, Dist_self; simpl.
      rewrite InA_cons, InA_nil.
      intuition; symmetry; auto.
    }
    {
      clear x; intros y hrec x.
      destruct (Dist_Pred_l y z) as [h|h]; rewrite h.
      transitivity (InA equiv x (Pred y :: traversal_Succ (S (Dist y z)) y)).
      {
        simpl;
          apply InA_compat_eqlistA; reflexivity || auto.
        right; try reflexivity.
        right; try reflexivity.
        apply Succ_Pred.
        apply traversal_Succ_compat; try reflexivity.
        rewrite Succ_Pred.
        reflexivity.
      }
      {
        rewrite InA_cons.
        rewrite (hrec x).
        split.
        {
          rewrite <- Dist_refl; unfold Between.
          intuition (try lia).
          rewrite Dist_refl in H0; rewrite H0.
          rewrite Dist_self; lia.
          destruct (Dist_Pred_l y x); try lia.
          rewrite Dist_refl in H; rewrite H.
          rewrite Dist_self; auto.
        }
        {
          intro hh.
          destruct (Between_Succ_l hh).
          left; symmetry; auto.
          right; rewrite <- (Succ_Pred y); apply H.
        }
      }
      {
        rewrite Dist_refl in h.
        simpl.
        rewrite h.
        rewrite InA_cons, InA_nil, Between_self.
        intuition (symmetry; auto).
      }
    }
  Qed.

  Lemma traversal_Pred_Dist_Between:
    forall x y z,
      InA equiv y (traversal_Pred (S (Dist x z)) z) <-> Between x y z.
  Proof.
    intros x y z.
    rewrite traversal_Pred_Succ_equiv.
    rewrite <- traversal_Succ_Dist_Between.
    apply InA_compat_eqlistA; reflexivity||auto.
    apply traversal_Succ_compat; reflexivity||auto.
    rewrite power_S.
    rewrite Pred_Succ.
    symmetry; apply Dist_power_Pred.
    reflexivity.
  Qed.

  (** The list traversal_Succ of length Ring_Size (and from any element)
   contains no repeat, but exactly every element in A.
   This is also true for traversal_Pred.
   *)
  Lemma traversal_Succ_full_no_repeat:
    forall x y, NoDupA equiv (traversal_Succ (Ring_size x) y).
  Proof.
    intros y x.
    rewrite (Ring_size_alt y x).
    unfold Ring_size'.
    clear y.
    generalize (Pred x) as y.
    intro y.
    pattern x.
    eapply Ring_Pred_ind with y.
    {
      intros a1 a2 ea.
      apply NoDupA_compat; auto.
      apply traversal_Succ_compat; auto.
      rewrite ea.
      reflexivity.
    }
    {
      rewrite Dist_self.
      compute.
      rewrite NoDupA_cons_iff.
      rewrite InA_nil.
      intuition.
    }
    {
      clear x.
      intros x hrec.
      destruct (Dist_Pred_l x y) as [h|h]; rewrite h.
      {
        simpl.
        rewrite NoDupA_cons_iff.
        split; eauto.
        assert (~InA equiv (Pred x) (traversal_Succ (S (Dist x y)) x)).
        { intro hh.
          rewrite traversal_Succ_Dist_Between in hh.
          hnf in hh.
          lia.
        }
        intro hh; apply H; revert hh.
        eapply proj1.
        apply InA_compat_eqlistA; reflexivity||auto.
        simpl; right.
        apply Succ_Pred.
        apply traversal_Succ_compat; reflexivity||auto.
        apply Succ_compat, Succ_Pred.
        revert hrec; eapply proj1.
        apply NoDupA_compat; auto.
        simpl; right.
        symmetry; apply Succ_Pred.
        apply traversal_Succ_compat; reflexivity||auto.
        symmetry; apply Succ_compat, Succ_Pred.
      }
      {
        simpl.
        rewrite NoDupA_cons_iff.
        rewrite InA_nil.
        intuition.
      }
    }
  Qed.

  Lemma traversal_Succ_full_all:
    forall x y z, InA equiv z (traversal_Succ (Ring_size x) y).
  Proof.
    intros x y z .
    rewrite (Ring_size_alt x y).
    unfold Ring_size'.
    rewrite traversal_Succ_Dist_Between.
    apply all_Between_Pred.
  Qed.

  Lemma traversal_Pred_full_no_repeat:
    forall x y, NoDupA equiv (traversal_Pred (Ring_size x) y).
  Proof.
    intros x y.
    eapply NoDupA_compat.
    apply traversal_Pred_Succ_rev.
    auto.
    apply NoDupA_rev; auto using A_equiv.
    apply traversal_Succ_full_no_repeat.
  Qed.

  Lemma traversal_Pred_full_all:
    forall x y z, InA equiv z (traversal_Pred (Ring_size x) y).
  Proof.
    intros x y z .
    rewrite (Ring_size_invariant x y).
    unfold Ring_size.
    rewrite traversal_Pred_Dist_Between.
    apply all_Between_Succ.
  Qed.

  (** Traversal Between *)
  Definition traversal_Between (p q: A): list A :=
    traversal_Succ (1 + (Dist p q)) p.

  Lemma traversal_Between_def:
    forall p q,
      traversal_Between p q = p :: if equiv_dec p q then nil
                                   else traversal_Between (Succ p) q.
  Proof.
    intros p q.
    case (equiv_dec p q); intros pq.
    { unfold traversal_Between.
      apply Dist_refl in pq; rewrite pq; simpl.
      auto. }
    { unfold traversal_Between.
      assert (D := Dist_Succ_l p q).
      destruct D as [D|D]; [|apply Dist_refl in D; tauto].
      rewrite D; simpl.
      auto.
    }
  Qed.

  Lemma traversal_Between_rev:
    forall p q, traversal_Between p q ==
                rev (traversal_Pred (1 + (Dist p q)) q).
  Proof.
    intros p q.
    unfold traversal_Between.
    rewrite traversal_Pred_Succ_rev.
    rewrite rev_involutive.
    simpl power.
    apply traversal_Succ_compat; auto.
    assert (dp := Dist_power_Pred _ _ (eq_refl (Dist p q))).
    rewrite dp at 1.
    apply Succ_injective.
    rewrite Succ_Pred.
    apply power_Pred_Succ'; auto.
  Qed.

  Lemma traversal_Between_def_rev:
    forall p q,
      traversal_Between p q == (if equiv_dec p q then nil
                                else traversal_Between p (Pred q)) ++ q::nil.
  Proof.
    intros p q.
    case (equiv_dec p q); intros pq.
    { unfold traversal_Between.
      dup pq pq'.
      apply Dist_refl in pq; rewrite pq; simpl.
      right; auto. }
    { rewrite traversal_Between_rev at 1.
      simpl traversal_Pred at 1.
      rewrite <- rev_involutive.
      apply rev_eqlistA_compat.
      rewrite rev_unit.
      right; auto.
      assert (tb := traversal_Between_rev p (Pred q)).
      apply rev_eqlistA_compat in tb.
      rewrite rev_involutive in tb.
      destruct (Dist_Pred_r p q) as [D|D].
      { rewrite <- D in tb.
        rewrite tb.
        reflexivity.
      }
      { apply Dist_refl in D; tauto. }
    }
  Qed.

  Lemma traversal_Between_self:
    forall p, traversal_Between p p = p::nil.
  Proof.
    intros p.
    assert (T := traversal_Between_def p p).
    destruct (equiv_dec p p) as [_|h]; [|contradict h]; auto.
  Qed.
  
  Instance traversal_Between_compat: Proper pequiv traversal_Between.
  Proof.
    intros n1 n2 en m1 m2 em.
    unfold traversal_Between.
    apply traversal_Succ_compat.
    rewrite en, em; auto.
    auto.
  Qed.

  Lemma traversal_Between_Between:
    forall p q n, InA equiv n (traversal_Between p q) <-> Between p n q.
  Proof.
    intros p q n.
    unfold traversal_Between.
    apply traversal_Succ_Dist_Between.
  Qed.

  Lemma traversal_Between_app:
    forall p q,
      ~ q == Pred q -> ~ p == q ->
      forall n, Between p n (Pred q) ->
                traversal_Between p q ==
                traversal_Between p n ++ traversal_Between (Succ n) q.
  Proof.
    intros p q nondeg npq.
    case (equiv_dec p (Pred q)); intros nppq.
    { intros n B.
      eapply Between_rewrite_r in B; [|apply nppq].
      apply Between_self in B.
      rewrite <- B; clear B n.
      rewrite nppq, Succ_Pred.
      rewrite traversal_Between_self.
      rewrite traversal_Between_self.
      assert (T := traversal_Between_def (Pred q) q).
      destruct (equiv_dec (Pred q) q) as [hq|hq];
        [symmetry in hq; tauto|].
      rewrite T; clear T.
      rewrite <- app_comm_cons.
      right; auto.
      rewrite Succ_Pred, traversal_Between_self.
      simpl; auto.      
    }
    
    apply (Between_Pred_ind
             (P := fun n =>
                     traversal_Between p q ==
                     traversal_Between p n ++ traversal_Between (Succ n) q)).
    { intros n1 n2 en; apply equiv_compat; auto.
      rewrite en; auto.
    }
    { assert (T := traversal_Between_def_rev p q).
      destruct (equiv_dec p q) as [F|_]; [tauto|].
      rewrite T; clear T.
      apply app_eqlistA_compat; [reflexivity|].
      unfold traversal_Between; simpl.
      right.
      rewrite Succ_Pred; auto.
      assert (D := Dist_self q).
      rewrite <- Succ_Pred in D at 1 .
      rewrite D; clear D; simpl.
      reflexivity.
    }
    
    intros n B ih.
    
    case (equiv_dec p n); intros hpn.
    { eapply Between_rewrite_m in B; [|apply hpn].
      contradict B.
      apply Between_not_Succ_r.
      auto. }
    case (equiv_dec q n); intros hqn.
    { eapply Between_rewrite_m in B; [|apply hqn].
      contradict B.
      apply Between_not_Pred_l.
      contradict nppq.
      apply Succ_injective; rewrite Succ_Pred; auto.
    }
    
    rewrite ih.
    rewrite Succ_Pred.
    assert (T := traversal_Between_def n q).
    destruct (equiv_dec n q) as [F|_]; [contradict hqn; symmetry; auto|].
    rewrite T; clear T.
    assert (T := traversal_Between_def_rev p n).
    destruct (equiv_dec p n) as [F|_]; [tauto|].
    rewrite T; clear T.
    simpl.
    rewrite ListUtil.list_app_first_last.
    reflexivity.
  Qed.
  
  Lemma traversal_Between_app':
    forall p q,
      ~ q == Pred q -> ~ p == q ->
      forall n, Between (Succ p) n q ->
                traversal_Between p q ==
                traversal_Between p (Pred n) ++ traversal_Between n q.
  Proof.
    intros p q nondeg npq n B.
    assert (T := traversal_Between_app _ nondeg npq).
    specialize (T (Pred n)).
    rewrite Succ_Pred in T.
    apply T.
    apply Between_Succ.
    revert B; apply Between_compat; auto; apply Succ_Pred.
  Qed.

  (** Decomposition of a slice of nodes.

     The result comes directly from traversal_Between_app, 
     traversal_Between_def and traversal_Between_def_rev.
   *)
  Lemma decomp_list:
    forall r, ~ Succ r == r ->
              forall a b x,
                Between a x b ->
                traversal_Between a b ==
                (if equiv_dec x a then nil else traversal_Between a (Pred x))
                  ++
                  x ::
                  (if equiv_dec x b
                   then nil else traversal_Between (Succ x) b).
  Proof.
    intros r ND2 a b x B.
    case (equiv_dec x a); intros xa; case (equiv_dec x b); intros xb.
    { rewrite app_nil_l.
      rewrite <- traversal_Between_self.
      apply traversal_Between_compat; symmetry; auto. }
    { rewrite app_nil_l.
      rewrite  traversal_Between_def.
      right; [symmetry; auto|].
      case (equiv_dec a b); intros ab; [rewrite <- ab in xb; tauto|].
      apply traversal_Between_compat; symmetry; auto.
      apply Succ_compat; auto. }
    { rewrite  traversal_Between_def_rev.
      apply eqlistA_app; [apply SA| |].
      2: right; symmetry; auto.
      case (equiv_dec a b); intros ab; [rewrite <- ab in xb; tauto|].
      apply traversal_Between_compat; symmetry; auto.
      apply Pred_compat; auto. }
    { rewrite traversal_Between_app.
      apply eqlistA_app; [apply SA| |].
      reflexivity.
      rewrite (traversal_Between_compat
                 _ _ (Succ_Pred x) _ _ (Equivalence_Reflexive _)).
      rewrite  traversal_Between_def.
      right; [symmetry|]; auto.
      case (equiv_dec x b); intros ab; [tauto|].       
      reflexivity.
      { intros F; apply Succ_compat in F; rewrite Succ_Pred in F;
          contradict F; apply (Non_Degenerated_Succ r), ND2; lia. }
      { intros F; eapply Between_rewrite_r in B; [| apply F].
        apply Between_self in B; symmetry in B; tauto. }
      { destruct (Between_Pred_r B); [tauto|].
        destruct (Between_Pred_m H); [symmetry in H0; tauto|].
        auto. }
    }
  Qed.
  
  (** Ring_fold_Succ applies f_acc to every element of the ring from a *) 
  Definition Ring_fold_Succ (T: Type) (a: A) (init: T) (f_acc: A -> T -> T) :=
    fold_right f_acc init (traversal_Succ (Ring_size a) a).

  Definition Ring_fold_Pred (T: Type) (a: A) (init: T) (f_acc: A -> T -> T) :=
    fold_right f_acc init (traversal_Pred (Ring_size a) a).

  Global Instance Ring_fold_Pred_compat {T: Type} {PT: PartialSetoid T}:
    Proper pequiv (Ring_fold_Pred (T := T)).
  Proof.
    intros x y xy i1 i2 ii f g fg.
    unfold Ring_fold_Pred.
    eapply fold_right_compat.
    apply fg.
    apply ii.
    apply traversal_Pred_compat.
    apply Ring_size_compat.
    auto.
    auto.
  Qed.

  Global Instance Ring_fold_Succ_compat {T: Type} {PT: PartialSetoid T}:
    Proper pequiv (Ring_fold_Succ (T := T)).
  Proof.
    intros x y xy i1 i2 ii f g fg.
    unfold Ring_fold_Succ.
    eapply fold_right_compat.
    apply fg.
    apply ii.
    apply traversal_Succ_compat.
    apply Ring_size_compat.
    auto.
    auto.
  Qed.
  
  (** ** The number of elements in A (i.e. in the ring) is exactly given 
   by Ring_size *)
  Lemma Ring_elements_card:
    forall x, Num_Card Same A (Ring_size x).
  Proof.
    intro x; split.
    {
      split with (fun y k => Dist x y = proj1_sig k).
      * intros y1 y2 hy [k1 hk1] [k2 hk2] hk; hnf in hk; simpl in *.
        rewrite hy, hk; reflexivity.
      * intros y1 [k1 hk1] h1 y2 [k2 hk2] h2 h; hnf in h; simpl in *.
        rewrite <- h1, <- h2 in h.
        rewrite <- (Dist_unique  _ _ _ _ h).
        reflexivity.
      * intros y1 y2 hy.
        split with (@exist _ _ (Dist x y1) (Dist_lt_Ring_size x x y1)).
        split with (@exist _ _ (Dist x y2) (Dist_lt_Ring_size x x y2)).
        intuition.
        hnf; simpl.
        apply Dist_compat; reflexivity||auto.
    }
    {
      split with (fun k y => Dist x y = proj1_sig k).
      * intros [k1 hk1] [k2 hk2] hk y1 y2 hy; hnf in hk; simpl in *.
        rewrite hy, hk; reflexivity.
      * intros [k1 hk1] y1 h1 [k2 hk2] y2 h2 h; hnf in h; simpl in *.
        hnf; simpl.
        rewrite <- h1, <- h2.
        apply Dist_compat; reflexivity||auto.
      * intros [k1 hk1] [k2 hk2] hk.
        simpl.
        split with (power Succ k1 x).
        split with (power Succ k2 x).
        hnf in hk; simpl in hk.
        {
          intuition.
          * rewrite (proj1 (power_Succ_Dist_mod k1 x (power Succ k1 x) x)).
            apply Nat.mod_small; auto.
            reflexivity.
          * rewrite (proj1 (power_Succ_Dist_mod k2 x (power Succ k2 x) x)).
            apply Nat.mod_small; auto.
            reflexivity.
          * apply power_compat; reflexivity|| auto || apply  Succ_compat.
        }
    }
  Qed.

  Definition Between_fold
             (T: Type) (init: T) (f_acc: A -> T -> T)
             (p q: A): T :=
    fold_right f_acc init (traversal_Between p q).  

  Instance Between_fold_compat: forall (T: Type) {PT: PartialSetoid T},
      Proper pequiv (Between_fold (T := T)).
  Proof.
    intros T PT a b ab f g fg n1 n2 en m1 m2 em.
    unfold Between_fold.
    eapply FoldRight.fold_right_compat.
    apply fg.
    auto.
    apply traversal_Between_compat; auto.
  Qed.
  
  Lemma fold_Between_Ring:
    forall (T: Type) {PT: PartialSetoid T}
           (init: T) (pi: Proper pequiv init)
           (f_acc: A -> T -> T)
           (f_acc_compat: Proper fequiv f_acc),
    forall p, Ring_fold_Succ p init f_acc
                             =~=
                             Between_fold init f_acc p (Pred p).
  Proof.
    intros until p.
    unfold Ring_fold_Succ, Between_fold, traversal_Between.
    eapply FoldRight.fold_right_compat; auto.
    apply f_acc_compat.
    apply traversal_Succ_compat; auto.
    rewrite Dist_Pred_full with (x := p).
    assert (P := Ring_size_pos p).
    lia.
  Qed.

End With_Ring_Topology.

(** * Ring Network
  Given a network, this section provides two definitions:
  - the first one constrains the network to be a unidirectional ring, namely 
    it should have a ring topology (as defined above) on nodes and the Pred
    function should correspond to existing channels
  - the second one adds the constraint of being rooted: the definition simply
    assumes a particular node to be the root.
 *)
Section Ring_Network.

  Variable (Chans: Channels)
           (Net: Network).

  Class Unidirectional_Ring_Network :=
    mkURN {
        URN_topo :> Ring_Topology (A:=Node);
        URN_pred_channel : Node -> Channel;
        URN_pred_channel_compat :> Proper fequiv URN_pred_channel;
        URN_channels: forall x y,
            Pred x == y <-> is_channel x (URN_pred_channel x) y
      }.
  (* Additional hypothesis: channel unicity if we want a ring graph instead
     of a ring subgraph *)

  Class Rooted_Unidirectional_Ring_Network :=
    mkRURN {
        RURN_URN :>  Unidirectional_Ring_Network;
        RURN_root: Node;
      }.

End Ring_Network.

(** * Ring Algorithm

  We define a property which provides the assumption for an algorithm 
  to run on a rooted ring network,i.e., the state of every node should contain 
  - the channel that goes to the predecessor in the ring; this is given by 
    function ring_pred_channel 
  - and whether it is the root; this is given by function ring_is_root. 
 *)
Section Ring_Algorithm.
  
  Definition Ring_Network_assumption {Chans: Channels}
             (Net: Network)
             (RURN: Rooted_Unidirectional_Ring_Network Net)
             (ring_is_root: Node -> bool)
             (ring_pred_channel: Node -> Channel): Prop :=
    (forall n,
        ring_is_root n = true <-> n == RURN_root (Net := Net))
    /\
    (forall n, is_channel n (ring_pred_channel n) (Pred n)).

  Instance Ring_Network_assumption_compat:
    forall {Chans: Channels} Net
           (RURN: Rooted_Unidirectional_Ring_Network Net),
      Proper fequiv (Ring_Network_assumption RURN).
  Proof.
    intros ro1 ro2 ero.
    split; intros [h1 h2]; split; intros n.
    - split; intros h.
      + apply h1.
        rewrite <- h; eapply H; auto.
      + rewrite <- h1 in h.
        rewrite <- h; symmetry; eapply H; auto.
    - eapply is_channel_compat.
      reflexivity.
      2: reflexivity.
      2: apply h2.
      symmetry; apply H0; auto. 
    - split; intros h.
      + apply h1.
        rewrite <- h; symmetry; apply H; auto.
      + rewrite <- h1 in h.
        rewrite <- h; apply H; auto.
    - eapply is_channel_compat.
      reflexivity.
      2: reflexivity.
      2: apply h2.
      apply H0; auto.
  Qed.

End Ring_Algorithm.

(** * Ring instance
  We prove that the class Ring_Topology is inhabited. We build the
  ring on Z/kZ, of size k
 *)
Section Ring_instance.

  Open Scope Z_scope.
  Variable (k: Z) (k_pos: k > 0).
  Notation eqk := (eqk k).

  Instance Zmodk_Setoid :Setoid Z :=
    {| equiv := eqk; setoid_equiv := eqk_equiv k_pos |}.
  
  Program Definition ring: Ring_Topology  :=
    {| Pred := (fun x => x - 1);
       Succ := (fun x => x + 1);
       Dist :=
         (fun x y =>
            if gtk_dec k x y then Z.to_nat(k - x mod k + y mod k)
            else Z.to_nat (y mod k - x mod k)
         )
    |}.
  
  Next Obligation. (* Pred_compat *)
    intros x y exy; rewrite eqk_def_equiv' in *; unfold eqk_alt in *; auto.
    rewrite <- Zminus_mod_idemp_l.
    symmetry; rewrite <- Zminus_mod_idemp_l.
    rewrite exy; auto.
  Qed.

  Next Obligation. (* Succ_compat *)
    intros x y exy; rewrite eqk_def_equiv' in *; unfold eqk_alt in *; auto.
    rewrite <- Zplus_mod_idemp_l.
    symmetry; rewrite <- Zplus_mod_idemp_l.
    rewrite exy; auto.
  Qed.

  Next Obligation. (* Dist_compat *)
    intros x x' ex y y' ey.
    assert (c := gtk_compat k_pos ex ey).
    rewrite eqk_def_equiv' in *; unfold eqk_alt in *; auto.
    rewrite <- ex, <- ey.
    destruct (gtk_dec k x y); destruct (gtk_dec k x' y'); tauto.
  Qed.

  Next Obligation. (* Pred_Succ_inv *)
    split; intros h;
      rewrite eqk_def_equiv' in *; unfold eqk_alt in *; auto.
    { rewrite <- Zminus_mod_idemp_l, <- h.
      rewrite Zminus_mod_idemp_l.
      assert (foo: x + 1 - 1 = x).
      lia.
      rewrite foo; auto. }
    { rewrite <- Zplus_mod_idemp_l, h.
      rewrite Zplus_mod_idemp_l.
      assert (foo: y - 1 + 1 = y).
      lia.
      rewrite foo; auto. }
  Qed.

  Next Obligation. (* Dist_refl *)
    split; intros h;
      rewrite eqk_def_equiv' in *; unfold eqk_alt in *;
        destruct (gtk_dec k x y) as [g|g]; unfold gtk in g; auto.
    { assert (t := gt_diff k_pos x y).
      rewrite <- Z2Nat.inj_0, Z2Nat.inj_iff in h; lia. }
    { rewrite <- Z2Nat.inj_0, Z2Nat.inj_iff in h; lia. }
    { lia. }
    { assert (t: y mod k - x mod k = 0).
      rewrite h; lia.
      rewrite t, Z2Nat.inj_0; auto.
    }
  Qed.

  Next Obligation. (* Dist_Succ_r *)
    assert (X := Z_mod_lt x _ k_pos).
    assert (Y := Z_mod_lt y _ k_pos).
    assert (Y1 := Z_mod_lt (y + 1) _ k_pos).
    destruct (Z.eq_dec (y mod k) (k - 1)).
    { assert (t := mod_plus_1' k_pos _ e).
      destruct (gtk_dec k x (y + 1)) as [g|g];
        destruct (gtk_dec k x y) as [g'|g'];
        unfold gtk in g, g'; rewrite e,t in *; try lia.
(*      left.
        rewrite <- Z2Nat.inj_succ.
        2: assert (tt := gt_diff k_pos x y); lia.
        apply Z2Nat.inj_iff; lia.
      - right.
        rewrite <- Z2Nat.inj_0.
        apply Z2Nat.inj_iff; lia.*)
    }
    { assert (t := mod_plus_1 k_pos _ n).
      destruct (gtk_dec k x (y + 1)) as [g|g];
        destruct (gtk_dec k x y) as [g'|g'];
        unfold gtk in g, g'; rewrite t in *; try lia.
     (* - left.
        rewrite <- Z2Nat.inj_succ.
        2: assert (tt := gt_diff k_pos x y); lia.
        apply Z2Nat.inj_iff; lia.
      - right.
        rewrite <- Z2Nat.inj_0.
        apply Z2Nat.inj_iff; lia.
      - left.
        rewrite <- Z2Nat.inj_succ.
        2: assert (tt := gt_diff k_pos x y); lia.
        apply Z2Nat.inj_iff; lia. *)
    }
  Qed.

  Next Obligation. (* Dist_Pred_l *)
    assert (X := Z_mod_lt x _ k_pos).
    assert (Y := Z_mod_lt y _ k_pos).
    assert (X1 := Z_mod_lt (x - 1) _ k_pos).
    destruct (Z.eq_dec (x mod k) 0).
    { assert (t := mod_minus_1' k_pos _ e).
      destruct (gtk_dec k (x - 1) y) as [g|g];
        destruct (gtk_dec k x y) as [g'|g'];
        unfold gtk in g, g'; rewrite e,t in *; try lia.
     (*  - left.
        rewrite <- Z2Nat.inj_succ.
        2: assert (tt := gt_diff k_pos x y); lia.
        apply Z2Nat.inj_iff; lia.
      - right.
        rewrite <- Z2Nat.inj_0.
        apply Z2Nat.inj_iff; lia. *)
    }
    { assert (t := mod_minus_1 k_pos _ n).
      destruct (gtk_dec k (x - 1) y) as [g|g];
        destruct (gtk_dec k x y) as [g'|g'];
        unfold gtk in g, g'; rewrite t in *; try lia.
      (* - left.
        rewrite <- Z2Nat.inj_succ.
        2: assert (tt := gt_diff k_pos x y); lia.
        apply Z2Nat.inj_iff; lia.
      - right.
        rewrite <- Z2Nat.inj_0.
        apply Z2Nat.inj_iff; lia.
      - left.
        rewrite <- Z2Nat.inj_succ.
        2: assert (tt := gt_diff k_pos x y); lia.
        apply Z2Nat.inj_iff; lia. *)
    }
  Qed.

  Lemma ring_size_k:
    forall x, Ring_size  (Rt:=ring) x = Z.to_nat k.
  Proof.
    intro x; transitivity (Ring_size (Rt:=ring) 0).
    apply Ring_size_invariant.
    unfold Ring_size.
    simpl.
    destruct (Z_le_lt_eq_dec 1%Z k) as [LT|EQ].
    lia.
    destruct (gtk_dec k 1 0) as [GT|NGT]; [unfold gtk in GT|unfold gtk in NGT];try lia.
    {
      rewrite Zmod_1_l; try lia.
      rewrite Zmod_0_l; try lia.
    }
    {
      rewrite Zmod_1_l in  NGT; try lia.
      rewrite Zmod_0_l in NGT; try lia.
    }
    destruct (gtk_dec k 1 0) as [h|h]; unfold gtk in h; subst; try lia.
    rewrite Zmod_1_r in h; try lia.
    rewrite Zmod_0_l in h; try lia.
    reflexivity.
  Qed.

  Close Scope Z_scope.

End Ring_instance.


Unset Implicit Arguments.
Close Scope signature_scope.
Close Scope nat_scope.
