(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** Global imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Bool.

(** Local imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import OptionUtil.
From Padec Require Import ListUtils.
From Padec Require Import RelationA.
From Padec Require Import WellfoundedUtil.
From Padec Require Import FoldRight.
From Padec Require Import ZUtils.
From Padec Require Import Algorithm.

(* end show *)
Open Scope signature_scope.
Set Implicit Arguments.

(** * Spanning Tree:
  Definition + properties
 *)
Section Spanning_Tree.

  (** ** Defined on a network and on parent pointers (channels) *)
  Context {Chans: Channels} {Net: Network}.

  (** ** Definition of a spanning tree over the network *)
  Section Is_Tree_Parent_Definition.

    Variable (Par: Node -> option Channel).

    (** *** Tools for spanning tree definition *)
    Definition tree_parent (x: Node): option Node :=
      option_rect (fun _ => option Node) (peer x) None (Par x).

    Definition is_tree_parent_of (x y: Node): Prop :=
      Some x == tree_parent y.

    Definition directed_tree_path := chain is_tree_parent_of.

    Definition is_tree_parent_or_child_of :=
      symmetric_closure is_tree_parent_of.

    Definition exists_tree_path :=
      reflexive_transitive_closure is_tree_parent_or_child_of.

  End Is_Tree_Parent_Definition.

  Section Is_Tree_Parent_Properties.

    Global Instance tree_parent_compat: Proper fequiv tree_parent.
    Proof.
      intros par1 par2 He x y eqn.
      unfold tree_parent.
      eapply option_rect_compat.
      apply peer_compat, eqn.
      reflexivity.
      apply He, eqn.
    Qed.

    Global Instance is_tree_parent_of_compat:
      Proper fequiv is_tree_parent_of.
    Proof.
      intros par par' He x1 x2 eqx y1 y2 eqy; unfold is_tree_parent_of.
      apply equiv_compat;
        [apply Some_compat|apply tree_parent_compat]; auto.
    Qed.

    Lemma is_tree_parent_of_dec:
      forall (Par: Node -> option Channel), Decider (is_tree_parent_of Par).
    Proof. intros Par x y; apply (eqoptionA_dec Node_dec). Qed.

    Lemma is_tree_parent_unique:
      forall Par1 Par2,
        Par1 =~= Par2 ->
        forall x1 x2, x1 == x2 ->
                      forall y1 y2 ,
                        is_tree_parent_of Par1 y1 x1 ->
                        is_tree_parent_of Par2 y2 x2 ->
                        y1 == y2.
    Proof.
      intros Par1 Par2 epar x1 x2 ex y1 y2 stp1 stp2.
      unfold is_tree_parent_of in *.
      rewrite (tree_parent_compat epar _ _ ex), <- stp2 in stp1.
      apply stp1.
    Qed.

    Global Instance directed_tree_path_compat:
      Proper (fequiv ==> equiv ==> eqlistA equiv ==> equiv ==> iff)
             directed_tree_path.
    Proof.
      intros par1  par2  Hpar  root1 root2 Hroot
             path1 path2 Hpath n1    n2    Hn.
      unfold directed_tree_path.
      apply chain_equiv_equiv; auto.
      apply is_tree_parent_of_compat; auto.
    Qed.

  End Is_Tree_Parent_Properties.

  Section Spanning_Tree_Definition.

    Definition is_root (Par: Node -> option Channel) (n: Node): Prop :=
      Par n == None.

    Global Instance is_root_compat: Proper fequiv is_root.
    Proof.
      intros p1 p2 ep n1 n2 en; unfold is_root.
      apply equiv_compat; auto; reflexivity.
    Qed.

    (** *** Definition of a spanning tree
     The following definition expresses that read-only variables
     [(par: ROEnv)] shape a spanning tree as a subgraph of network [Net].

     First, by construction, each node [p] has exactly at most one
     parent [(par p)].

     Predicate [parent_is_peer] means that any link from [p] to
     [(par p)] is an existing channel in the network.

     Predicates [is_root_root] and [is_root_unique] say that there
     exists a unique node [root] with no parent, i.e., with
     [(par root)] equals [None].

     Finally, predicate [no_loop] assumes that there is no loop
     following [par] pointers, namely, the transitive closure of [par]
     is not reflexive.
     *)
    Class spanning_tree (root: Node) (Par: Node -> option Channel): Prop :=
      {
        parent_is_peer:
          forall (n: Node) (c: Channel), Par n == Some c ->
                                         exists m, is_channel n c m;
        is_root_root: is_root Par root;
        is_root_unique:
          forall (n1 n2: Node), is_root Par n1 -> is_root Par n2 -> n1 == n2;
        no_loop: forall (p: Node),
            ~ transitive_closure (is_tree_parent_of Par) p p
      }.
    (*                                                               *)

    (** *** Direct results on the spanning tree definition *)
    Global Instance spanning_tree_compat: Proper fequiv spanning_tree.
    Proof.
      intros r1 r2 Hr par1 par2 Hp.
      split; split;
        destruct H as (Hpar, Hroot, Hunique, Hloop).
      (* -> *)
      intros n c h; apply Hpar; rewrite <- h; apply Hp; reflexivity.
      (*rewrite <- Hp. rewrite <- Hr; auto. pourquoi ca passe pas ?*)
      revert Hroot; apply is_root_compat; symmetry; auto.
      intros n1 n2 h hh; apply Hunique; rewrite Hp; auto.
      intros p; rewrite <- Hp; auto.
      (* <- *)
      intros n c h; apply Hpar; rewrite <- h; symmetry; apply Hp; reflexivity.
      (*rewrite Hp. rewrite Hr; auto. pourquoi ca passe pas ?*)
      revert Hroot; apply is_root_compat; auto.
      intros n1 n2 h hh; apply Hunique; rewrite <- Hp; auto.
      intros p; rewrite Hp; auto.
    Qed.

    (** the no_loop assumption of span_tree implies that
     is_tree_parent_of is a well-founded order
     (this will be used, later, for induction proof).
     *)
    Lemma WF_par:
      forall (root: Node) (Par: Node -> option Channel),
        Proper fequiv Par -> spanning_tree root Par ->
        well_founded (is_tree_parent_of Par).
    Proof.
      intros root par par_compat span.
      apply (finite_acyclic_wf (A_dec:=Node_dec)) with all_nodes.
      apply all_nodes_prop.
      apply is_tree_parent_of_compat; auto.
      apply is_tree_parent_of_dec.
      apply span.
    Qed.

    (** the no_loop assumption of span_tree implies that
     is_tree_child_of, namely (flip is_tree_parent_of) is a
     well-founded order
     (this will be used, later, for induction proof).
     *)
    Lemma WF_child:
      forall (root: Node) (Par: Node -> option Channel),
        Proper fequiv Par -> spanning_tree root Par ->
        well_founded (flip (is_tree_parent_of Par)).
    Proof.
      intros root par par_compat span.
      rewrite <- (wf_flip_finite (A_dec := Node_dec)).
      apply WF_par with root; assumption.
      apply all_nodes_prop.
      apply is_tree_parent_of_compat; auto.
      apply is_tree_parent_of_dec.
    Qed.

  End Spanning_Tree_Definition.

End Spanning_Tree.

(*
Class Spanning_Tree_Local_Topology {Chans: Channels} :=
  {
    ST_local_topology:> Local_Topology;

    ST_Par: Local_Topo -> option Channel;
    ST_Par_compat:> Proper fequiv ST_Par;
  }.
 *)

Definition Spanning_Tree_Network_assumption {Chans: Channels}
           (Net: Network)
           (tested_par: Node -> option Channel): Prop :=
  exists root, spanning_tree root tested_par.

Global Instance Spanning_Tree_Network_assumption_compat:
  forall {Chans: Channels} Net,
    Proper fequiv (Spanning_Tree_Network_assumption Net).
Proof.
  intros.
  intros t1 t2 et.
  split; intros [root h].
  exists root; revert h; apply spanning_tree_compat; auto.
  intros n1 n2 en; symmetry; apply et; symmetry; auto.
  exists root; revert h; apply spanning_tree_compat; auto.
Qed.

Section Spanning_Tree_Symmetric_Network.

  Context {Chans: Channels} {Net: Symmetric_Network}.

  Section Spanning_Tree_Properties.

    (** Algorithm [father_dec] computes
      - either the tree father with the associated proof
      - or a proof that the node is the root *)
    Lemma father_dec:
      forall (root: Node) (Par: Node -> option Channel),
        spanning_tree root Par ->
        forall (n: Node), {m | is_tree_parent_of Par m n} + {is_root Par n}.
    Proof.
      intros root Par span n.
      unfold is_tree_parent_of, tree_parent, is_root.
      assert (htree: forall c,
                 Par n == Some c -> exists m, is_channel n c m).
      { destruct span as [ h _ _ _]; apply h. }
      unfold is_channel in *.
      case_equiv (Par n);[intro c|];intro h; simpl.
      specialize (htree _ h).
      destruct (peer n c) as [m |];simpl.
      left; split with m; reflexivity.
      right; destruct htree as [_ []].
      right; trivial.
    Qed.

    Lemma non_root_has_parent:
      forall (root: Node) (Par: Node -> option Channel),
        spanning_tree root Par ->
        forall (n: Node), ~is_root Par n ->
                          exists m, (is_tree_parent_of Par m n).
    Proof.
      intros root Par span n not_root.
      unfold is_root in not_root.
      case_eq (Par n);
        [intros c pareq |
         intros h; rewrite h in not_root; elim not_root; reflexivity].
      assert (h := parent_is_peer n c).
      assert (tmp: Par n == Some c) by (rewrite pareq; reflexivity).
      specialize (h tmp); clear tmp.
      destruct h as [m chan].
      exists m.
      unfold is_tree_parent_of, tree_parent; rewrite pareq.
      unfold is_channel in chan.
      symmetry; apply chan.
    Qed.

    Lemma parent_is_peer_prop:
      forall (root: Node) (Par: Node -> option Channel),
        spanning_tree root Par ->
        forall (n child: Node), is_tree_parent_of Par n child
                                -> InA equiv child
                                       (map_filter (peer n) (peers n)).
    (******child --par=c--> n
                     <----c'---    **)
    Proof.
      intros root Par span n child Htree.
      assert (symm := sym_net child n).
      unfold is_tree_parent_of, tree_parent in Htree.
      assert (Hxx: exists c, Some c = Par child).
      { revert Htree; case (Par child); intros.
        exists c; reflexivity.
        elim Htree.
      }
      destruct Hxx as (c, Hxx).
      rewrite <- Hxx in Htree.
      assert (chan: exists c', is_channel n c' child).
      { apply symm.
        exists c.
        symmetry in Htree; apply Htree.
      }
      destruct chan as (c', Hc').
      unfold is_channel in Hc'.
      apply InA_map_filter_iff.
      apply peer_compat; reflexivity.
      exists c'.
      split; auto.
      specialize (peers_spec_peer n); intro h.
      apply h.
      exists child; auto.
    Qed.

    Lemma path_to_root:
      forall (root: Node) (Par: Node -> option Channel),
        Proper fequiv Par ->
        spanning_tree root Par ->
        forall (n: Node), exists (p: list Node),
            (directed_tree_path Par root p n).
    Proof.
      intros root Par ppar span.
      assert (wf_up := WF_par ppar span).
      destruct span as [hpeerpar hroot hunique hnoloop].
      intros n.
      induction (wf_up n) as [w hw hhw].
      { case_equiv (Par w);[intro c|];intro h.
        destruct (hpeerpar _ _ h) as [x hx].
        assert (hxw: is_tree_parent_of Par x w).
        { unfold is_tree_parent_of, tree_parent.
          red in hx.
          rewrite <- hx.
          rewrite h.
          reflexivity.
        }
        specialize (hhw x hxw).
        destruct hhw as [p hp].
        split with (p ++ w::nil).
        unfold directed_tree_path; rewrite chain_app.
        exists x; simpl chain; repeat split; auto.
        apply is_tree_parent_of_compat; auto.
        exists nil; apply hunique; auto.
      }
    Qed.

  End Spanning_Tree_Properties.

  (** Children defined using local information only *)
  Section Local_Children.

    Definition Local_IsChild (peers: list Channel)
               (peer_Par: Channel -> option Channel)
               (reply: Channel -> Channel)
               (c: Channel): bool :=
      option_rect (fun _ => bool)
                  (* Some s => *) (fun s => dec2b Channel_dec s (reply c))
                  (* None => *) false
                  (* match *) (peer_Par c) (* with *).

    Definition Local_Children (peers: list Channel)
               (peer_Par: Channel -> option Channel)
               (reply: Channel -> Channel)
      := List.filter (Local_IsChild peers peer_Par reply) peers.

    Global Instance Local_IsChild_compat:
      Proper (equivlistA equiv ==> fequiv) Local_IsChild.
    Proof.
      intros peers1 peers2 eqpeers par1 par2 eqpar
             reply_to1 reply_to2 eqreply_to c1 c2 eqc; unfold Local_IsChild.
      apply option_rect_compat; auto; try reflexivity.
      intros x y eq; rewrite eq, (eqreply_to _ _ eqc); reflexivity.
    Qed.

    Global Instance Local_Children_compat:
      Proper (equivlistA equiv ==> fequiv ==> fequiv
                         ==> equivlistA equiv) Local_Children.
    Proof.
      intros peers1 peers2 eqpeers par1 par2 eqpar reply1 reply2 eqreply c;
        unfold Local_Children.
      apply InA_compat; [apply Channel_setoid| reflexivity|].
      apply filter_compat_equiv; auto.
      apply Local_IsChild_compat; auto.
    Qed.

    Lemma Local_IsChild_locality:
      forall ps1 ps2 lPar1 lPar2 rev1 rev2,
      forall c, InA equiv c ps1 -> lPar1 c == lPar2 c ->
                rev1 c == rev2 c ->
                Local_IsChild ps1 lPar1 rev1 c =
                Local_IsChild ps2 lPar2 rev2 c.
    Proof.
      intros ps1 ps2 lPar1 lPar2 rev1 rev2 c i hlp hp .
      unfold Local_IsChild.
      destruct (lPar1 c); destruct (lPar2 c); simpl in hlp;
        try tauto.
      change (option_rect (fun _ : option Channel => bool)
                          (fun s : Channel =>
                             dec2b Channel_dec s (rev1 c)) false
                          (Some c0) ==
              option_rect (fun _ : option Channel => bool)
                          (fun s : Channel =>
                             dec2b Channel_dec s (rev2 c)) false
                          (Some c1)).
      eapply option_rect_compat; auto.
      intros s1 s2 es.
      rewrite es, hp; auto.
    Qed.

    Lemma Local_Children_locality:
      forall ps1 ps2 lPar1 lPar2 rev1 rev2,
        Proper fequiv lPar1 -> Proper fequiv lPar2 ->
        Proper fequiv rev1 -> Proper fequiv rev2 ->
        equivlistA equiv ps1 ps2 ->
        (forall c, InA equiv c ps1 -> lPar1 c == lPar2 c /\
                                      rev1 c == rev2 c) ->
        equivlistA equiv (Local_Children ps1 lPar1 rev1)
                   (Local_Children ps2 lPar2 rev2).
    Proof.
      intros ps1 ps2 lPar1 lPar2 rev1 rev2
             pp1 pp2 pr1 pr2 hps hind.
      unfold Local_Children; apply filter_eq_locality; auto.
      apply Local_IsChild_compat; auto; proper_.
      apply Local_IsChild_compat; auto; proper_.
      intros a i; apply Local_IsChild_locality; auto; apply hind;
        auto.
    Qed.

  End Local_Children.

  Section Children.

    Definition IsChild (Par: Node -> option Channel)
               (n: Node) (c: Channel): bool :=
      Local_IsChild (peers n)
                    (fun c => match (peer n c) with None => None
                                               | Some x => Par x end)
                    (reverse n) c.

    Lemma Is_Child_is_tree_parent_of:
      forall Par n m,
        Proper fequiv Par ->
        ((exists c, Par m == Some c /\
                    IsChild Par n (reverse m c) = true /\
                    is_channel m c n)
         <-> is_tree_parent_of Par n m).
    Proof.
      intros Par n m Par_compat.
      assert (sym := sym_net n m).
      split.
      { (* -> *)
        intros [c [ec [_ hm]]].
        assert (tmp: exists cc, is_channel m cc n) by (exists c; auto).
        apply sym in tmp.
        destruct tmp as [cr icr].
        assert (tmp: is_channel m c n /\ is_channel n cr m) by (split; auto).
        assert (ccr := reverse_spec _ _ _ _ tmp); clear tmp.
        rewrite <- ccr in icr; clear cr ccr.
        assert (R := reverse_involutive _ _ _ hm).
        unfold is_tree_parent_of, tree_parent;
          unfold is_channel in *.
        rewrite <- hm, ec; simpl; reflexivity.
      }
      { (* <- *)
        unfold is_tree_parent_of, tree_parent;
          case_eq (Par m);[| simpl; intros _ F; elim F].
        intros c ec nm.
        symmetry in nm.
        assert (mcn: is_channel m c n) by (simpl in nm; apply nm).
        assert (nc'm := reverse_spec' m c n mcn).
        exists c.
        split; [ reflexivity | split; [| assumption]].
        unfold IsChild, Local_IsChild; unfold is_channel in nc'm.
        assert (R := reverse_involutive _ _ _ mcn).
        destruct (peer n (reverse m c)); [|elim nc'm].
        simpl in nc'm; specialize (Par_compat _ _ nc'm);
          rewrite ec in Par_compat.
        destruct (Par n0); [|elim Par_compat; tauto]; simpl in *.
        rewrite dec2b_spec_true; rewrite Par_compat; symmetry; auto.
      }
    Qed.

    Global Instance IsChild_compat: Proper fequiv IsChild.
    Proof.
      intros par1 par2 eqpar n1 n2 en c1 c2 eqc; unfold IsChild.
      apply Local_IsChild_compat; auto.
      apply peers_compat_equiv; assumption.
      intros cc1 cc2 ecc.
      generalize (peer_compat _ _ en _ _ ecc);
        destruct (peer n1 cc1); destruct (peer n2 cc2); intros h;
          simpl in h; try (elim h; reflexivity); auto.
      apply reverse_compat; assumption.
    Qed.

    Definition Children
               (Par: Node -> option Channel) (n: Node): list Channel :=
      List.filter (IsChild Par n) (peers n).

    Lemma Children_equiv:
      forall Par n,
        Proper fequiv Par ->
        eqlistA equiv (Children Par n)
                (Local_Children (peers n)
                                (fun c: Channel =>
                                   match (peer n c) with None => None
                                                    | Some x => Par x end)
                                (reverse n)).
    Proof.
      intros Par n Par_compat.
      unfold Children, Local_Children, IsChild.
      apply filter_compat; auto.
      intros c1 c2 eqc.
      apply Local_IsChild_compat; auto.
      apply peers_compat_equiv; reflexivity.
      intros cc1 cc2 eqcc.
      generalize (peer_compat _ _ (Equivalence_Reflexive n) _ _ eqcc);
        destruct (peer n cc1); destruct (peer n cc2); intros h;
          simpl in h; try (elim h; reflexivity); auto.
      apply reverse_compat; reflexivity.
      reflexivity.
    Qed.

    Global Instance Children_compat:
      Proper (fequiv ==> equiv ==> equivlistA equiv) Children.
    Proof.
      intros par1 par2 eqpar n1 n2 en.
      rewrite Children_equiv; [rewrite Children_equiv|]; try proper_.
      apply Local_Children_compat.
      apply peers_compat_equiv; assumption.
      intros cc1 cc2 eqcc.
      generalize (peer_compat _ _ en _ _ eqcc);
        destruct (peer n1 cc1); destruct (peer n2 cc2); intros h;
          simpl in h; try (elim h; reflexivity); auto.
      apply reverse_compat; auto.
    Qed.

    Lemma Children_inv:
      forall (Par: Node -> option Channel),
        Proper fequiv Par ->
        forall (n: Node) (c: Channel),
          InA equiv c (Children Par n) ->
          exists m, is_channel n c m /\ is_tree_parent_of Par n m.
    Proof.
      intros Par Par_compat n c ina.
      assert (symm := sym_net).
      unfold Children in ina.
      rewrite filter_InA in ina; blast.
      2: apply IsChild_compat; [apply Par_compat | reflexivity].
      destruct ina  as [ina child].
      rewrite peers_spec_peer in ina.
      destruct ina as [m hm].
      exists m.
      split; [assumption|].
      apply Is_Child_is_tree_parent_of; try assumption.
      assert (tmp: exists cc, is_channel n cc m) by (exists c; assumption).
      apply (symm n m) in tmp.
      destruct tmp as [cr ecr].
      exists cr.
      assert (tmp: is_channel m cr n /\ is_channel n c m)
        by (split; assumption).
      assert (P := reverse_spec _ _ _ _ tmp); clear tmp.
      split; [|split; [|assumption]].
      2: rewrite P; auto.
      unfold IsChild, Local_IsChild in child.
      revert hm child; unfold is_channel;
        destruct (peer n c); [|intros F; elim F].
      simpl; intros T X; rewrite <- T; clear T.
      destruct (Par n0); simpl in X; [| congruence].
      rewrite <- P, reverse_involutive in X; auto.
      rewrite dec2b_spec_true in X; simpl; auto.
    Qed.

    Lemma InA_Children:
      forall (Par: Node -> option Channel),
        Proper fequiv Par ->
        forall (n m: Node),
          is_tree_parent_of Par n m <->
          exists (c: Channel), is_channel n c m /\
                               InA equiv c (Children Par n).
    Proof.
      intros Par Par_compat n m.
      assert (Hsym := sym_net).
      split.
      { (* -> *)
        intros nm.
        rewrite <- Is_Child_is_tree_parent_of in nm; auto.
        destruct nm as [c [ec [ic ncm]]].
        assert (tmp: exists cc, is_channel m cc n) by (exists c; auto).
        apply Hsym in tmp.
        destruct tmp as [cc ecc].
        exists cc; split; [assumption|].
        assert (nc'm := reverse_spec' m c n ncm).
        assert (T: reverse m c == cc).
        { apply reverse_spec with n; split; assumption. }
        unfold Children; rewrite filter_InA.
        2: apply IsChild_compat; [apply Par_compat | reflexivity].
        split.
        rewrite peers_spec_peer; exists m; assumption.
        rewrite <- T; auto.
      }
      { (* <- *)
        intros [c [ncm ina]].
        apply Children_inv in ina; auto.
        destruct ina as [m' [ncm' stp]].
        assert (d: m == m') by (eapply is_channel_unique; eauto).
        rewrite d; auto.
      }
    Qed.

  End Children.

End Spanning_Tree_Symmetric_Network.

Unset Implicit Arguments.
Close Scope signature_scope.
