(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- March 2017
- first author: Benjamin Collet
*)
(* begin show *)

(** Global imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import ZArith.
From Coq Require Import Lia.

(** Local imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import OptionUtil.
From Padec Require Import NatUtils.
From Padec Require Import ListUtils.
From Padec Require Import FoldRight.
From Padec Require Import RelModel.
From Padec Require Import Algorithm.
From Padec Require Import RelationA.

(* end show *)
Open Scope nat_scope.
Open Scope signature.
  
Global Instance proj1_sig_compat {A} `{SA: Setoid A} (P: A -> Prop): Proper pequiv (@proj1_sig _ P).
Proof.
  intros [a ha] [b hb] ab; simpl; compute in ab; auto.
Qed.

Lemma app_cons_equivlist {A: Type} (eqA: relation A):
  forall a l1 l2, equivlistA eqA (l1 ++ a::l2) (a:: l1 ++ l2).
Proof.
  intros.
  intros x.
  rewrite InA_cons, InA_app_iff, InA_app_iff, InA_cons.
  tauto.
Qed.

Lemma fold_right_map {A: Type} {SA: Setoid A}:
  forall (f: A -> A -> A) (f_compat: Proper pequiv f)
         (g: A -> A)
         (l: list A) x
         (commute1: forall a, f (g a) x == g (f a x))
         (commute2: forall a b, f (g a) (g b) == g (f a b)),
    l <> nil -> fold_right f x (map g l) == g (fold_right f x l).
Proof.
  intros f f_compat g l x commute1 commute2.
  induction l as [| a l ih].
  intros F; contradict F; reflexivity.
  intros _; simpl.
  destruct l as [| b l].
  simpl; auto.
  rewrite ih; auto.
  intros F; inversion F.
Qed.

Lemma fold_right_max_ge:
  forall l x, fold_right max x l >= x.
Proof.
  intros l x.
  induction l as [| a l ih]; simpl.
  lia.
  assert (X := Nat.le_max_r a (fold_right max x l)).
  lia.
Qed.

Lemma fold_right_max_ge2:
  forall l x, l <> nil -> forall y, In y l -> fold_right max x l >= y.
Proof.
  intros l x nl.
  induction l as [| a l ih]; simpl.
  contradict nl; auto.
  intros y [hy|hy].
  rewrite hy.
  assert (T := Nat.le_max_l y (fold_right max x l)); lia.
  cut (fold_right max x l >= y).
  { intros h; assert (T := Nat.le_max_r a (fold_right max x l)); lia. }
  apply ih; auto.
  contradict hy; rewrite hy; apply in_nil.
Qed.

Lemma fold_right_max_exact:
  forall l a, l <> nil -> In a l ->
              (forall b, In b l -> b <= a) ->
              a = fold_right max 0 l.
Proof.
  induction l as [|b l ih].
  intros a F; contradict F; reflexivity.
  intros a N ina mx.
  simpl.
  destruct l as [|bb l].
  { simpl.
    simpl in ina; destruct ina as [ina|ina]; [|tauto].
    rewrite ina, Nat.max_0_r; auto. }
  case (Nat.eq_dec a b); intros ab.
  { rewrite <- ab.
    symmetry; apply Nat.max_l.
    apply mx.
    assert (dum: forall x y : nat, {Init.Nat.max x y == x} + {Init.Nat.max x y == y}).
    { intros x y; destruct (le_gt_dec x y).
      right; apply Nat.max_r; auto.
      left; apply Nat.max_l; lia.
    }
    destruct (fold_right_f_dec_restr (f := max) dum 0 (l := bb::l)); clear dum.
    { intros [| y] _; simpl; auto. }
    inversion e.
    rewrite InA_alt in i.
    destruct i as [ff [hff iff]].
    rewrite hff.
    right; auto.
  }
  { rewrite <- (ih a).
    symmetry; apply Nat.max_r, mx; simpl; left; auto.
    intros F; inversion F.
    simpl in ina; destruct ina as [ina|ina]; [lia|auto].
    intros bbb inb; apply mx.
    simpl; simpl in inb; right; auto.
  }
Qed.  
  
Section Max_All_In.

  Variable (A: Type) (SA: PartialSetoid A)
           (eqA_dec: Decider (pequiv (A := A)))
           (all: list A) (all_prop: forall a, InA pequiv a all)
           (f: A -> nat) (f_compat: Proper pequiv f).

  Lemma arg_max: all <> nil -> { max | forall y, f max >= f y}.
  Proof.
    intros n_nil.
    cut ({max | forall y, InA pequiv y all -> f max >= f y}).
    { intros [max ok].
      exists max; auto.
    }
    clear all_prop.
    induction all as [| a all_ ih].
    congruence.
    clear n_nil.
    case_eq all_.
    { intros hnil; exists a; intros y hin.
      rewrite InA_cons in hin.
      destruct hin.
      rewrite H; auto.
      rewrite InA_nil in H; tauto.
    }
    intros x y z; assert (n_nil: all_ <> nil) by congruence; rewrite <- z; clear x y z.
    destruct ih as [max hmax]; auto.
    case (le_gt_dec (f a) (f max)); intros c.
    - exists max.
      intros y hy.
      case (eqA_dec y a); intros hay.
      + rewrite hay; lia.
      + apply hmax.
        rewrite InA_cons in hy.
        destruct hy; [tauto|auto].
    - exists a.
      intros y hy.
      case (eqA_dec y a); intros hay.
      + rewrite hay; lia.
      + cut (f max >= f y).
        lia.
        rewrite InA_cons in hy.
        destruct hy; [tauto|auto].
  Qed.
  
End Max_All_In.

Section Power_option.

  Variable (A: Type).  

  Definition opt_bind (f: A -> option A): option A -> option A :=
    fun x => match x with
               None => None
             | Some x => f x
             end.

  Global Instance opt_bind_compat eqA:
    Proper ((eqA ==> eqoptionA eqA) ==> eqoptionA eqA ==> eqoptionA eqA) opt_bind.
  Proof.
    intros f1 f2 ef.
    hnf.
    intros opt_x1 opt_x2 ex.
    destruct opt_x1; destruct opt_x2; simpl.
    - apply ef.
      apply ex.
    - inversion ex.
    - inversion ex.
    - tauto.
  Qed.
        
  Definition power_opt (f: A -> option A) (n: nat): A -> option A :=
    fun x => power (opt_bind f) n (Some x).

  Global Instance power_opt_compat eqA:
    Proper ((eqA ==> eqoptionA eqA) ==> @eq nat ==> eqA ==> eqoptionA eqA) power_opt.
  Proof.
    intros f1 f2 ef _n n en. rewrite en. clear en _n.
    hnf.
    intros x1 x2 ex.
    unfold power_opt.
    apply power_compat.
    - apply opt_bind_compat.
      apply ef.
    - reflexivity.
    - apply ex.    
  Qed.
                                                                               
  Lemma power_opt_0: forall f x, power_opt f 0 x = Some x.
  Proof.
    intros.
    unfold power_opt.
    simpl.
    reflexivity.
  Qed.

  Lemma power_opt_None: forall f n, power (opt_bind f) n None = None.
  Proof.
    induction n.
    - unfold power.
      reflexivity.
    - simpl.
      rewrite IHn.
      unfold opt_bind.
      reflexivity.
  Qed.

  Lemma S_power_opt: forall f n x,
      power_opt f (S n) x = (opt_bind f) (power_opt f n x).
  Proof.
    reflexivity.
  Qed.

  Lemma power_opt_S: forall f n x,
      power_opt f (S n) x = opt_bind (power_opt f n) (f x).
  Proof.
    intros f n x.
    assert (opt_bind_redef: f x = opt_bind f (Some x)). tauto.
    unfold power_opt.
    case_eq (f x).
    {
      intros a HsomeA.
      unfold opt_bind at 2.
      rewrite opt_bind_redef in HsomeA.
      rewrite <- HsomeA.
      apply power_S.
    }
    {
      intros Hnone.
      unfold opt_bind at 2.
      rewrite power_S.
      rewrite <- opt_bind_redef.
      rewrite Hnone.
      apply power_opt_None.
    }
  Qed.

  Lemma power_opt_add: forall f n p x,
      power_opt f (n + p) x = opt_bind (power_opt f n) (power_opt f p x).
  Proof.
    intros.
    induction n.
    {
      unfold power_opt at 2.
      simpl.
      unfold opt_bind.
      case_eq (power_opt f p x); reflexivity.
    }
    {
      simpl.
      rewrite S_power_opt.
      unfold opt_bind at 2.
      case_eq (power_opt f p x).
      {
        intros.
        rewrite H in IHn.
        rewrite IHn.
        apply S_power_opt.
      }
      {
        intros.
        rewrite H in IHn.
        unfold opt_bind in IHn.
        rewrite IHn.
        unfold opt_bind.
        reflexivity.
      }
    }
  Qed.
  
  Lemma power_opt_mult: forall f n p x,
      power_opt f (n * p) x = power_opt (power_opt f n) p x.
  Proof.
    intros f n p x.
    induction p.
    {
      rewrite mult_0_r.
      rewrite ?power_opt_0.
      reflexivity.
    }
    {
      rewrite mult_succ_r.
      rewrite S_power_opt.
      rewrite Nat.add_comm.
      rewrite power_opt_add.
      rewrite IHp.
      reflexivity.
    }
  Qed.
  
End Power_option.


(** * Tree Topology
  A tree topology on elements of type A is defined as:
    begin given
    - a parent function Parent:
      it represents going one step closer to the root of the tree
    - a distance to lowest common ancestor function Dist_LCA:
      it counts the number of steps to get to a common ancestor
   
  This definition is setoid-compatible, hence it requires
  - a relation on A, considered as the equality on the elements of A and
  - that each function (Parent, Dist_LCA) is compatible wrt this relation.
*)
Class Tree_Topology {A} `{SA:Setoid A}: Type :=
  mkTree {
      Parent: A -> option A;
      Dist_LCA: A -> A -> nat;  (* Distance to Lowest Common Ancestor *)
      Parent_compat:> Proper fequiv Parent;
      Dist_LCA_compat:> Proper fequiv Dist_LCA;
      
      isParent x y := Parent y == Some x;
      isRoot x := Parent x == None;
      isAncestor x y := Dist_LCA x y = 0;

      Dist_LCA_refl: forall x y,
          x == y <-> Dist_LCA x y = 0 /\ Dist_LCA y x = 0;
      Dist_LCA_root: forall x y, isRoot x -> isAncestor x y;
      Dist_LCA_Parent: forall x Px,
          isParent Px x <-> Dist_LCA x Px = 1 /\ Dist_LCA Px x = 0;
      Dist_LCA_Parent_l: forall x y Px,
          isParent Px x -> Dist_LCA Px y = pred (Dist_LCA x y);
      Dist_LCA_Parent_r: forall x y Py,
          isParent Py y -> ~(isAncestor y x) -> Dist_LCA x y = Dist_LCA x Py;
      Dist_LCA_Parent_r_ancestor: forall x y Py,
          isParent Py y -> isAncestor y x -> Dist_LCA x y + 1 = Dist_LCA x Py;
      
      all_nodes: list A;
      all_nodes_prop: forall n, InA equiv n all_nodes;
    }.

Section With_Tree_Topology.
  
  Context {A} `{Tt: Tree_Topology A}.

  Instance A_equiv: Equivalence equiv := setoid_equiv.

  Lemma eqA_dec: Decider equiv.
  Proof.
    intros x y.
    case_eq (Dist_LCA x y + Dist_LCA y x).
    {
      intros d.
      left.
      apply Dist_LCA_refl.
      lia.
    }
    {
      intros n d.
      right.
      intuition.
      apply Dist_LCA_refl in H.
      destruct H as [dxy dyx].
      rewrite dxy in d.
      rewrite dyx in d.
      assert (d': 0 + 0 = S n -> False).
      inversion d.
      apply d' in d.
      exact d.
    }
  Qed.
  
  Instance isParent_compat: Proper pequiv (@isParent A SA Tt).
  Proof.
    intros x y xy z t zt; unfold isParent; rewrite xy, zt; auto.
  Qed.

  Instance isRoot_compat: Proper pequiv isRoot.
  Proof.
    intros x y xy; unfold isRoot; rewrite xy; auto.
  Qed.

  Instance isAncestor_compat: Proper pequiv (@isAncestor A SA Tt).
  Proof.
    intros x y xy z t zt; unfold isAncestor; rewrite xy, zt; auto.
  Qed.

  Lemma Dist_LCA_self: forall x, Dist_LCA x x = 0.
  Proof.
    intros x.
    apply Dist_LCA_refl.
    reflexivity.
  Qed.

  Lemma Dist_LCA_Parent_l_not_ancestor: forall x y Px,
      isParent Px x -> ~isAncestor x y -> Dist_LCA x y = Dist_LCA Px y + 1.
  Proof.
    intros x y Px hPx nanc; unfold isAncestor in nanc.
    rewrite (Dist_LCA_Parent_l x y Px hPx); lia.
  Qed.

  Lemma isParent_dec: forall x y, {isParent x y} + {~isParent x y}.
  Proof.
    intros x y; unfold isParent; destruct (Parent y).
    destruct (eqA_dec a x); [left | right]; auto.
    right; auto.
  Qed.

  Lemma has_Parent_dec: forall x, {y | Parent x == Some y} + {Parent x == None}.
  Proof.
    intros x; destruct (Parent x).
    left; exists a; auto.
    right; auto.
  Qed.

  Lemma isParent_unique:
    forall x y z, isParent x y -> isParent z y -> x == z.
  Proof.
    intros x y z pxy pzy; unfold isParent in *.
    rewrite pzy in pxy.
    symmetry in pxy; apply pxy.
  Qed.

  Lemma isParent_notsym: forall x y, isParent x y -> ~isParent y x.
  Proof.
    intros x y xy yx.
    rewrite Dist_LCA_Parent in xy, yx; lia.  
  Qed.

  Lemma isParent_notid: forall x y, isParent x y -> ~ x == y.
  Proof.
    intros x y h F.
    rewrite <- F in h.
    dup h hh.
    apply isParent_notsym in h; tauto.
  Qed.
  
  Lemma isParent_notsym_isAncestor: forall x y, isParent x y -> ~isAncestor y x.
  Proof.
    intros x y xy yx.
    rewrite Dist_LCA_Parent in xy; unfold isAncestor in yx; lia.
  Qed.

  Lemma isAncestor_dec: forall x y, {isAncestor x y} + {~isAncestor x y}.
  Proof.
    intros x y; unfold isAncestor.
    apply Nat.eq_dec.
  Qed.    
    
  Lemma isAncestor_inherit: forall x Px y, isParent Px x -> isAncestor x y -> isAncestor Px y.
  Proof.
    unfold isAncestor.
    intros x Px y h hh.
    rewrite (Dist_LCA_Parent_l x y Px), hh; auto.
  Qed.  

  Lemma isAncestor_down_inherit:
    forall x y Py, isParent Py y -> isAncestor x y -> ~x == y ->
                   isAncestor x Py.
  Proof.
    unfold isAncestor.
    intros x y Py h hh n.
    assert (hhh := Dist_LCA_Parent_r x y Py h).
    rewrite hh in hhh; symmetry; apply hhh.
    intros F; contradict n; apply Dist_LCA_refl; auto.
  Qed.  

  Lemma isAncestor_Parent: forall x Px, isParent Px x -> isAncestor Px x.
  Proof.
    unfold isAncestor.
    intros x Px h.
    apply Dist_LCA_Parent; auto.
  Qed.  

  Lemma isAncestor_refl: forall x, isAncestor x x.
  Proof.
    unfold isAncestor; intros x; apply Dist_LCA_refl; auto.
  Qed.

  Lemma isAncestor_antisym: forall x y,
      isAncestor x y -> isAncestor y x -> x == y.
  Proof.
    intros x y; rewrite Dist_LCA_refl; auto.
  Qed.
  
  Lemma isAncestor_trans: forall x y z, isAncestor x y -> isAncestor y z -> isAncestor x z.
  Proof.
    unfold isAncestor; intros x y z hxy hyz.
    remember (Dist_LCA z y) as dzy; rename Heqdzy into hdzy.
    revert hyz hdzy; revert z; induction dzy;
      intros z hdyz hdzy.
    { assert (h: z == y) by (apply Dist_LCA_refl; auto).
      rewrite h; auto. }
    case (eqA_dec z y); intros hzy.
    { rewrite hzy; auto. }
    case (has_Parent_dec z).
    2: {
        intros r.
        assert (h: y == z).
        rewrite Dist_LCA_refl.
        split; [|apply Dist_LCA_root]; auto.
        rewrite h in hxy; auto.
    }
    intros [Pz hPz].
    rewrite (Dist_LCA_Parent_r x z Pz hPz).
    apply IHdzy.
    { rewrite (Dist_LCA_Parent_r y z Pz hPz) in hdyz; auto.
      contradict hzy; apply Dist_LCA_refl; auto.
    }
    { rewrite (Dist_LCA_Parent_l z y Pz hPz).
      rewrite <- hdzy; simpl; auto.
    }

    { contradict hzy; rename hzy into hzx.
      clear IHdzy dzy hdzy Pz hPz.
      unfold isAncestor in hzx.
      remember (Dist_LCA x z) as dxz; rename Heqdxz into hdxz.
      revert hxy hzx hdxz; revert x; induction dxz;
        intros x hxy hdxz hzx.
      - assert (h1: x == z) by (apply Dist_LCA_refl; auto).
        rewrite <- h1 in hdyz; rewrite <- h1.
        apply Dist_LCA_refl; auto.
      - case (has_Parent_dec x).
        2: { intros F;
             apply Dist_LCA_root with (y := z) in F.
             unfold isAncestor in F; lia.
        }
        intros [Px hPx].
        apply IHdxz with Px.
        apply isAncestor_inherit with x; auto.
        apply isAncestor_down_inherit with x; auto.
        intros F; rewrite F, Dist_LCA_self in hzx; lia.
        assert (hh := Dist_LCA_Parent_l x z Px hPx).
        rewrite <- hzx in hh; lia.
    }
  Qed.
  
  Lemma uniq_root: forall x y, isRoot x -> isRoot y -> x == y.
  Proof.
    intros x y Hx Hy.
    assert (Hd: Dist_LCA x y = 0 /\ Dist_LCA y x = 0).
    {
      split; apply Dist_LCA_root; assumption.
    }
    apply Dist_LCA_refl.
    exact Hd.
  Qed.
  
  Lemma Parent_non_id: forall x y, isParent x y -> ~ x == y.
  Proof.
    intros x y h F.
    apply Dist_LCA_Parent in h.
    apply Dist_LCA_refl in F.
    lia.
  Qed.

  Lemma have_common_ancestor: forall x y, { a | isAncestor a x /\ isAncestor a y}.
  Proof.
    intros a b.
    remember (Dist_LCA a b) as dab; rename Heqdab into hdab.
    remember (Dist_LCA b a) as dba; rename Heqdba into hdba.
    revert dba hdba hdab; revert a b; induction dab; intros a b dba hdba hdab.
    { 
      exists a; unfold isAncestor; split; auto.
      apply Dist_LCA_self.
    }
    {
      case (has_Parent_dec a).
      - intros [Pa hPa].
        case (isAncestor_dec a b); intros Aab.
        exists a; split; auto; apply Dist_LCA_refl; auto.
        assert (D := Dist_LCA_Parent_r _ _ _ hPa Aab).
        rewrite D in hdba; clear D.
        specialize (IHdab _ _ _ hdba).
        assert (h: dab = Dist_LCA Pa b).
        { rewrite (Dist_LCA_Parent_l a b Pa hPa); lia. }
        specialize (IHdab h); clear h.
        destruct IHdab as [AA [ha hb]].
        exists AA; split; auto.
        eapply isAncestor_trans.
        apply ha.
        apply isAncestor_Parent; auto.
      - intros r; exists a; split.
        apply isAncestor_refl.
        apply Dist_LCA_root; auto.
    }         
  Qed.
  
  Definition emptyTree := all_nodes = nil.
  Lemma not_emptyTree: forall x: A, ~emptyTree.
  Proof.
    intros x.
    assert (h := all_nodes_prop x).
    case_eq all_nodes.
    intros hh; rewrite hh in h; rewrite InA_nil in h; tauto.
    congruence.
  Qed.    
    
  Lemma emptyTree_root: { r | isRoot r } + { emptyTree }.
  Proof.
    case_eq all_nodes.
    { intros n; right; auto. }
    intros a all_n h_all; left.
    assert( arg_max: { max | forall y, Dist_LCA a max >= Dist_LCA a y }).
    { eapply arg_max.
      apply eqA_dec.
      apply all_nodes_prop.
      intros x y xy; rewrite xy; auto.
      congruence.
    }
    destruct arg_max as [max hmax].
    remember (Dist_LCA max a) as dma; rename Heqdma into hdma.
    clear all_n h_all.
    revert hmax hdma; revert a max; induction dma; intros a max hdma hmax.
    {
      exists max.
      case (has_Parent_dec max).
      2: tauto.
      intros [pm hpm].
      symmetry in hmax.
      assert (h := Dist_LCA_Parent_r_ancestor a max pm hpm hmax).
      specialize (hdma pm).
      lia.      
    }
    {
      case (has_Parent_dec max).
      2: exists max; auto.
      intros [pm hpm].
      assert (h := Dist_LCA_Parent_l max a pm hpm).
      rewrite <- hmax in h; simpl in h.
      edestruct IHdma as [r ir].
      3: exists r; auto.
      2: symmetry; apply h.
      intros y; specialize (hdma y).
      case (isAncestor_dec max a); intros Aab.
      - assert (hh := Dist_LCA_Parent_r_ancestor a max pm hpm).
        rewrite <- hh; auto.
        lia.
      - assert (hh := Dist_LCA_Parent_r a max pm hpm).
        rewrite <- hh; auto.
    }
  Qed.

  Definition getRoot (x: A): {r: A | isRoot r}.
  Proof.
    destruct emptyTree_root; auto.
    contradict e; apply not_emptyTree; auto.
  Qed.

  Lemma getRoot_same: forall x y, getRoot x == getRoot y.
  Proof.
    intros x y.
    destruct (getRoot x); destruct (getRoot y).
    assert (ir := uniq_root _ _ i i0).
    compute; auto.
  Qed.

  Instance getRoot_compat: Proper pequiv getRoot.
  Proof.
    intros x y _.
    apply getRoot_same.
  Qed.
    
  (*  Variable (root: A) (Hroot: isRoot root). *)

  Definition isChild x y := Parent x == Some y.
  Definition isSibling x y := Parent x == Parent y.

  Definition isChild_b :=
    (fun x y => if eqoptionA_dec eqA_dec (Parent x) (Some y)
                then true else false).
  Definition Children x :=
    filter (fun cx => isChild_b cx x) all_nodes.

  Definition isLeaf x := equivlistA equiv (Children x) nil.

  Instance isChild_compat: Proper pequiv isChild.
  Proof.
    intros x y xy z t zt; unfold isChild; rewrite xy, zt; auto.
  Qed.

  Instance isSibling_compat: Proper pequiv isSibling.
  Proof.
    intros x y xy z t zt; unfold isSibling; rewrite xy, zt; auto.
  Qed.

  Instance isChild_b_compat: Proper pequiv isChild_b.
  Proof.
    intros x y xy z t zt; unfold isChild_b.
    case (eqoptionA_dec eqA_dec (Parent x) (Some z));
    case (eqoptionA_dec eqA_dec (Parent y) (Some t));
    rewrite xy, zt; congruence.
  Qed.
  
  Instance Children_compat:
    Proper (equiv ==> equivlistA equiv) Children.
  Proof.
    intros x y xy; unfold Children.
    apply filter_compat_equiv; [|reflexivity].
    intros cx1 cx2 ecx. rewrite xy, ecx; auto.
  Qed.

  Instance isLeaf_compat: Proper pequiv isLeaf.
  Proof.
    intros x y xy; unfold isLeaf; rewrite xy; auto.
  Qed.

  Lemma isChild_b_isChild:
    forall x y, isChild_b x y = true <-> isChild x y.
  Proof.
    intros x y. unfold isChild_b, isChild.
    destruct (eqoptionA_dec eqA_dec (Parent x) (Some y)).
    tauto.
    split; intros; [congruence|tauto].
  Qed.

  Lemma isChild_b_isChild_false:
    forall x y, isChild_b x y = false <-> ~isChild x y.
  Proof.
    intros x y. unfold isChild_b, isChild.
    destruct (eqoptionA_dec eqA_dec (Parent x) (Some y)).
    intuition.   
    tauto.
  Qed.

  Lemma isChild_b_isParent:
    forall x y, isChild_b x y = true <-> isParent y x.
  Proof.
    intros x y. unfold isChild_b, isParent.
    destruct (eqoptionA_dec eqA_dec (Parent x) (Some y)).
    tauto.
    split; intros; [congruence|tauto].
  Qed.

  Lemma isChild_isParent:
    forall x y, isChild x y <-> isParent y x.
  Proof.
    intros x y. unfold isChild, isParent; tauto.
  Qed.

  Lemma in_children:
    forall x Px, isParent Px x <-> InA equiv x (Children Px).
  Proof.
    intros x Px; split; intros hp.
    { apply filter_InA.
      intros cx cxx ecx; rewrite ecx; auto.
      split; [apply all_nodes_prop|auto].
      apply isChild_b_isParent; auto. }
    { unfold Children in hp.
      apply filter_InA in hp.
      2: intros cx cxx ecx; rewrite ecx; auto.
      apply isChild_b_isParent, hp.
    }
  Qed.

  Lemma childrent_not_nil:
    forall x Px, isParent Px x -> ~ equivlistA equiv (Children Px) nil.
  Proof.
    intros x Px hp.
    rewrite in_children in hp.
    destruct (Children Px).
    apply InA_nil in hp; tauto.
    apply equivlistA_cons_nil.
    apply setoid_equiv.
  Qed.
    
  (* ???
    fold_right (fun y z => max y z)
              (map (fun y => Dist_LCA y x)
                   (filter (fun y => beq_nat (Dist_LCA x y) 0) all_nodes))
              0. *)
  (*Fixpoint Height x :=
<    fold_right (fun x1 x2 => max x1 x2)
               0
               (map (fun y => Height y) (Children x)).*)
  
  Lemma exists_child:
    forall x y, Dist_LCA y x > 0 -> Dist_LCA x y = 0 ->
                ~ equivlistA equiv (Children x) nil.
  Proof.
      intros x y dp dn.
      remember (Dist_LCA y x) as dyx; rename Heqdyx into hdyx.
      revert hdyx dn dp; revert y; induction dyx; intros y hdyx dn dp. 
      - lia.
      - case (has_Parent_dec y).
        { intros [Py hPy].
          case (dyx) as [| ddyx].
          { assert (P: isParent x y) by (apply Dist_LCA_Parent; auto).
            apply childrent_not_nil with y; auto.
          }
          { assert (ddd := Dist_LCA_Parent_l y x Py hPy).
            rewrite <- hdyx in ddd; simpl in ddd.
            apply IHdyx with Py; auto.
            assert (dd := Dist_LCA_Parent_r x y Py hPy).
            rewrite <- (Dist_LCA_Parent_r x y Py hPy); auto.
            unfold isAncestor; lia.
            lia.
          }
        }
        { intros F; apply Dist_LCA_root with (y := x) in F.
          unfold isAncestor in F; lia.
        }
  Qed.

  Lemma two_children:
    forall x y z, isChild y x -> isChild z x -> ~ y == z ->
                  ~isAncestor y z.
  Proof.
    intros x y z chy chz neq anc.
    contradict neq.
    apply Dist_LCA_refl; split; auto.
    rewrite isChild_isParent in chy, chz.
    assert (X := Dist_LCA_Parent_r_ancestor z y x chy anc).
    rewrite Dist_LCA_Parent in chz.
    lia.
  Qed.
  
  Lemma Leaf_isAncestor:
    forall x y, isLeaf x -> isAncestor x y -> x == y.
  Proof.
    intros x y l a.
    case_eq (Dist_LCA y x).
    { intros aa.
      apply Dist_LCA_refl; auto. }
    { intros n aa.
      contradict l.
      apply exists_child with y; auto; lia.
    }
  Qed.

  Lemma nonLeaf:
    forall x, ~isLeaf x -> exists c, isParent x c.
  Proof.
    intros x nl.
    unfold isLeaf in nl.
    case_eq (Children x).
    intros hc; rewrite hc in nl; contradict nl; reflexivity.
    intros c lc hc.
    exists c.
    apply in_children.
    rewrite hc, InA_cons; left; auto.
  Qed.

  Definition isAncestor_b x y :=
    if (Nat.eq_dec 0 (Dist_LCA x y)) then true else false.

  Instance isAncestor_b_compat: Proper pequiv isAncestor_b.
  Proof.
    intros x y xy z t zt; unfold isAncestor_b.
    case (Nat.eq_dec 0 (Dist_LCA x z));
      case (Nat.eq_dec 0 (Dist_LCA y t)); rewrite xy, zt; congruence.
  Qed.    

  Lemma isAncestor_b_isAncestor:
    forall x y, isAncestor_b x y = true <-> isAncestor x y.
  Proof.
    intros x y; unfold isAncestor, isAncestor_b.
    destruct (Nat.eq_dec 0 (Dist_LCA x y)); split; congruence.
  Qed.

  Lemma isAncestor_b_isAncestor_false:
    forall x y, isAncestor_b x y = false <-> ~isAncestor x y.
  Proof.
    intros x y; unfold isAncestor, isAncestor_b.
    destruct (Nat.eq_dec 0 (Dist_LCA x y)); split; congruence.
  Qed.

  Lemma isAncestor_b_isParent:
    forall x y, isParent x y -> isAncestor_b x y = true.
  Proof.
    intros x y h; apply isAncestor_b_isAncestor.
    apply isAncestor_Parent; auto.
  Qed.
  
  Definition Height x :=
    fold_right max
               0
               (map (fun y => Dist_LCA y x)
                    (filter (fun y => isAncestor_b x y) all_nodes)).

  Instance Height_compat: Proper pequiv Height.
  Proof.
    intros x y xy; unfold Height.
    eapply fold_right_compat.
    intros a b ab c d cd; rewrite ab, cd; auto.
    auto.
    eapply map_compat.
    intros a b ab; rewrite xy, ab; auto.
    apply filter_compat.
    intros vy yy yyy; rewrite xy, yyy; auto.
    reflexivity.
  Qed.

  Lemma max0: forall a b, max a b = 0 -> a = 0 /\ b = 0.
  Proof.
    induction a.
    intros b hb; split; auto.
    intros [|b] hb; simpl in hb; auto.
    lia.
  Qed.    
    
  Lemma Height_leaf: forall x, isLeaf x <-> Height x = 0.
  Proof.
    intros x; split; intros h.
    { unfold Height.
      induction all_nodes as [| a all ih]; auto.
      case_eq (isAncestor_b x a);
        intros anc; simpl; rewrite anc; auto.
      { apply isAncestor_b_isAncestor in anc;
          eapply Leaf_isAncestor in h; eauto.
        rewrite Dist_LCA_refl in h.
        destruct h as [_ Hdax].
        simpl.
        rewrite Hdax; auto. }
    }
    { unfold isLeaf, Children.
      unfold Height in h.
      induction all_nodes as [| a all ih]; simpl.
      reflexivity.
      simpl in h.
      case_eq (isAncestor_b x a); intros anc; rewrite anc in h; simpl in h.
      { apply max0 in h.
        destruct h as [h hh].       
        specialize (ih hh).
        assert (i: isChild_b a x = false).
        { apply isChild_b_isChild_false.
          intros i. rewrite isChild_isParent in i.
          apply Parent_non_id in i; contradict i.
          apply Dist_LCA_refl; split; auto.
          apply isAncestor_b_isAncestor; auto. }
        rewrite i; auto.
      }
      { specialize (ih h).
        assert (c: isChild_b a x = false).
        { rewrite <- Bool.not_true_iff_false;
            rewrite <- Bool.not_true_iff_false in anc.
          contradict anc.          
          apply isChild_b_isParent in anc.
          apply isAncestor_b_isParent; auto. 
        }
        rewrite c; auto.
      }
    }
  Qed.

  Lemma noAncestor:
    forall x y, isAncestor x y ->
                forall a, ~isAncestor x a -> ~isAncestor y a.
  Proof.
    intros x y Axy a Axa.
    contradict Axa.
    apply isAncestor_trans with y; auto.
  Qed.

  Lemma ancestor_child_exists:
    forall x a, ~ x == a -> isAncestor x a -> exists c, isParent x c /\ isAncestor c a.
  Proof.
    intros x a xa Axa.
    remember (Dist_LCA a x) as dax.
    revert Heqdax Axa xa; revert a; induction dax as [| dax ih]; intros a hdax Axa xa.
    contradict xa; apply Dist_LCA_refl; split; auto.
    case (has_Parent_dec a).
    { intros [Pa hPa].
      case (eqA_dec x Pa); intros xPa.
      { exists a; split.
        rewrite xPa; auto.
        apply isAncestor_refl.
      }
      { destruct (ih Pa) as [c [hc hhc]]; auto.
        rewrite Dist_LCA_Parent_l_not_ancestor with a x Pa in hdax; auto.
        lia.
        contradict xa; apply Dist_LCA_refl; split; auto.    
        apply isAncestor_down_inherit with a; auto.
        exists c; split; auto.
        apply isAncestor_trans with Pa; auto.
        apply isAncestor_Parent; auto.
      }
    }
    { intros hr.
      contradict xa.
      apply Dist_LCA_refl; split; auto.
      apply Dist_LCA_root; auto.
    }
  Qed.
                
  Lemma Height_isParent:
    forall x y, isParent x y -> Height x >= 1 + Height y.
  Proof.
    intros x c par.
    unfold Height.
    pose (fun x => (filter (fun y => isAncestor_b x y) all_nodes)) as lAnc.
    
    fold (lAnc x).
    fold (Height c).

    assert (X: forall a, isAncestor x a -> ~ isAncestor c a -> Dist_LCA a x = Dist_LCA a c).
    { intros a AA nA.
      symmetry; apply Dist_LCA_Parent_r; auto. }
    
    assert (Y: forall a, isAncestor c a -> Dist_LCA a x = 1 + Dist_LCA a c).
    { intros a AA.
      erewrite <- Dist_LCA_Parent_r_ancestor; eauto; lia. }
    
    assert (L: equivlistA equiv (lAnc x)
                          (filter (fun a => andb (isAncestor_b x a) (negb (isAncestor_b c a))) all_nodes
                                  ++
                                  filter (fun a => isAncestor_b c a) all_nodes)).
    { unfold lAnc.
      clear lAnc.
      induction all_nodes as [| a all ih]; simpl.
      reflexivity.
      case_eq (isAncestor_b x a); intros Axa.
      case_eq (isAncestor_b c a); intros Aca.
      { assert ((true && negb true)%bool = false) by auto.
        rewrite H; clear H.
        rewrite ih.
        intros xx;
          rewrite InA_app_iff, InA_cons, InA_cons, InA_app_iff.
        tauto.
      }
      { assert ((true && negb false)%bool = true) by auto.
        rewrite H; clear H.
        rewrite ih.
        intros xx;
          rewrite InA_app_iff, InA_cons, InA_cons, InA_app_iff.
        tauto.        
      }
      { assert ((false && negb (isAncestor_b c a))%bool = false) by auto.
        rewrite H; clear H.
        rewrite ih.
        assert (Aca: isAncestor_b c a = false).
        apply isAncestor_b_isAncestor_false.
        apply noAncestor with x.
        apply isAncestor_Parent; auto.
        apply isAncestor_b_isAncestor_false; auto.
        rewrite Aca; clear Aca.
        rewrite <- ih; reflexivity.
      }
    }

    assert (F: fold_right max 0 (map (fun y => Dist_LCA y x) (lAnc x))
               ==
               fold_right max 0 (map (fun y => Dist_LCA y x)
                                     ((filter (fun a => andb (isAncestor_b x a) (negb (isAncestor_b c a))) all_nodes)
                                        ++
                                        filter (fun a => isAncestor_b c a) all_nodes))).
    { eapply fold_right_compat_equivlist; auto.
      intros aa ab aab t v tv; rewrite aab, tv; auto.
      intros aa ab abc.
      rewrite Nat.max_comm at 1.
      assert (Init.Nat.max aa abc = Init.Nat.max abc aa) by (rewrite Nat.max_comm; auto).
      rewrite H; rewrite Nat.max_assoc; auto.
      intros aa ab.
      rewrite Nat.max_assoc, Nat.max_id; auto.
      intros a b; apply Nat.eq_dec.
      apply map_compat_equiv; auto.
      intros aa ab aaab; rewrite aaab; auto.
    }
    rewrite F; clear F L.
    rewrite map_app, fold_right_app.

    assert (M1: map (fun a => Dist_LCA a x) (filter (fun a => isAncestor_b c a) all_nodes)
                ==
                map (fun a => 1 + Dist_LCA a c) (filter (fun a => isAncestor_b c a) all_nodes)).
    { clear lAnc.
      induction all_nodes as [| a all ih]; simpl.
      reflexivity.
      case_eq (isAncestor_b c a); intros Aca; simpl.
      { right.
        apply Y.
        apply isAncestor_b_isAncestor; auto.
        apply ih.
      }
      apply ih.
    }
    assert (M2: map (fun a => Dist_LCA a x) (filter (fun a => (isAncestor_b x a && negb (isAncestor_b c a))%bool) all_nodes)
                ==
                map (fun a => Dist_LCA a c) (filter (fun a => (isAncestor_b x a && negb (isAncestor_b c a))%bool) all_nodes)).
    { clear lAnc M1.
      induction all_nodes as [| a all ih]; simpl.
      reflexivity.
      case_eq ((isAncestor_b x a && negb (isAncestor_b c a))%bool); intros AA; simpl.
      {
        right.
        apply andb_prop in AA.
        destruct AA as [Axa Aca].
        rewrite Bool.negb_true_iff in Aca.
        apply X.
        apply isAncestor_b_isAncestor; auto.
        apply isAncestor_b_isAncestor_false; auto.
        apply ih.
      }
      apply ih.
    }
    rewrite M1, M2; clear M1 M2.

    assert (F: (fold_right Init.Nat.max 0
                           (map (fun a : A => 1 + Dist_LCA a c)
                                (filter (fun a : A => isAncestor_b c a) all_nodes)))
               = 1 + Height c).
    {
      rewrite <- map_map.
      rewrite fold_right_map; auto.
      intros j i ji k l kl; rewrite ji, kl; auto.
      intros a; rewrite Nat.max_0_r, Nat.max_0_r; auto.
      intros F; apply map_eq_nil, filter_nil in F.
      contradict F; rewrite Bool.not_false_iff_true.
      apply existsb_exists_InA.
      intros aa aaa aaaa; rewrite aaaa; auto.
      exists c; split; [apply all_nodes_prop|].
      apply isAncestor_b_isAncestor, Dist_LCA_self.
    }
    rewrite F; clear F.
    apply fold_right_max_ge.
  Qed.
  
  Lemma Height_isParent_exists:
    forall x, ~isLeaf x -> exists y, isParent x y /\ Height x = 1 + Height y.
  Proof.
    intros x nL.
    unfold Height.

    assert (dum: forall x y : nat, {Init.Nat.max x y == x} + {Init.Nat.max x y == y}).
    { intros xx y; destruct (le_gt_dec xx y).
      right; apply Nat.max_r; auto.
      left; apply Nat.max_l; lia.
    }
    destruct (fold_right_f_dec_restr (f := max) dum 0 (l := map (fun y : A => Dist_LCA y x) (filter (fun y : A => isAncestor_b x y) all_nodes))).
    { intros [| y] _; simpl; auto. }
    { apply map_eq_nil, filter_nil in e.
      contradict e; rewrite Bool.not_false_iff_true, existsb_exists_InA.
      apply nonLeaf in nL; destruct nL as [c hc].
      exists c; split; [apply all_nodes_prop|].
      apply isAncestor_b_isParent; auto.
      intros l m lm; rewrite lm; auto. }
    apply InA_map_iff in i.
    2: intros yy yyy yyyy; rewrite yyyy; auto.
    fold (Height x). fold (Height x) in i.
    destruct i as [l [hl inl]].
    destruct (ancestor_child_exists x l) as [c [hc hhc]].
    { contradict nL; rewrite nL, Dist_LCA_self in hl.
      apply Height_leaf; simpl in hl; rewrite nL; lia. }
    {
      apply filter_InA in inl.
      2: intros yy yyy yyyy; rewrite yyyy; auto.
      apply isAncestor_b_isAncestor, inl. }
    exists c; split; auto.
    rewrite <- Dist_LCA_Parent_r_ancestor with _ c _ in hl; auto.
    fold (Height c).
    rewrite <- hl.
    cut (Dist_LCA l c = Height c); [lia|].
    apply fold_right_max_exact.
    { intros F; apply map_eq_nil, filter_nil in F.
      contradict F; rewrite Bool.not_false_iff_true, existsb_exists_InA.
      2: intros m mm mmm; rewrite mmm; auto.
      exists l; split; [apply all_nodes_prop|apply isAncestor_b_isAncestor]; auto.
    }
    { cut (InA equiv (Dist_LCA l c)
               (map (fun y : A => Dist_LCA y c) (filter (fun y : A => isAncestor_b c y) all_nodes))).
      { rewrite InA_alt; intros [v [hhv hv]].
        rewrite hhv; auto. }
      rewrite InA_map_iff.
      2: intros m mm mmm; rewrite mmm; auto.
      exists l; split; auto.
      apply filter_InA.
      intros m mm mmm; rewrite mmm; auto.
      split; [apply all_nodes_prop|apply isAncestor_b_isAncestor]; auto.
    }
    { intros d hd.
      apply (In_InA (eqA:=equiv (A:=nat))) in hd; [| apply Setoid_nat].
      assert (hh := Height_isParent _ _ hc).
      rewrite <- hl in hh.
      assert (hhh: Dist_LCA l c >= Height c) by lia; clear hh.
      transitivity (Height c); [|lia].
      unfold Height.
      apply fold_right_max_ge2; auto.
      intros F; rewrite F, InA_nil in hd; tauto.
      rewrite InA_alt in hd.
      destruct hd as [dd [hdd hhdd]].
      rewrite hdd; auto.
    }
  Qed.

  Lemma Height_alt:
    forall x, Height x =
              fold_right max
                         0
                         (map (fun y => 1 + Height y) (Children x)).
  Proof.
    intros x.
    case_eq (Children x).
    { simpl; intros hlc.
      apply Height_leaf; auto.
      unfold isLeaf; rewrite hlc; reflexivity.
    }
    intros c lc hlc.
    rewrite <- hlc.
    apply fold_right_max_exact.
    rewrite hlc; simpl; intros F; inversion F.
    {
      destruct (Height_isParent_exists x) as [cc [pcc hcc]].
      { unfold isLeaf; rewrite hlc.
        apply equivlistA_cons_nil.
        apply setoid_equiv.
      }
      cut (InA equiv (Height x) (map (fun y : A => 1 + Height y) (Children x))).
      { intros h; rewrite InA_alt in h.
        destruct h as [y [hy hhy]]; rewrite hy; auto. }
      rewrite InA_map_iff.
      2: intros l ll lll; rewrite lll; auto.
      exists cc; split; [simpl; lia|].
      apply in_children; auto.
    }
    intros b inb.
    apply (In_InA (eqA:=equiv (A:=nat))) in inb; [| apply Setoid_nat].
    rewrite InA_map_iff in inb.
    2: intros aa bb aabb; rewrite aabb; auto.
    destruct inb as [cc [hcc incc]].
    rewrite <- hcc.
    cut (Height x >= 1 + Height cc); [lia|].
    apply Height_isParent.
    apply in_children; auto.
  Qed.

  Definition Depth (x: A): nat := Dist_LCA x (proj1_sig (getRoot x)).

  Instance Depth_compat: Proper pequiv Depth.
  Proof.
    intros x1 x2 ex; unfold Depth.
    apply Dist_LCA_compat; auto.
    apply proj1_sig_compat, getRoot_same.
  Qed.

  (** Any node which is down-linked to another has smaller depth than the first one. *)
  Lemma deeper_Depth_step:
    forall a child, isParent a child -> Depth child = Depth a + 1.
  Proof.
    intros a child par; unfold Depth.
    rewrite (Dist_LCA_Parent_l_not_ancestor child _ a par).
    rewrite (getRoot_same a child); auto.
    assert (R := proj2_sig (getRoot child)).
    intros F; assert (X: child == proj1_sig (getRoot child)).
    { assert (D := Dist_LCA_root (proj1_sig (getRoot child)) child R).
      apply Dist_LCA_refl.
      split; auto. }
    unfold isParent in par; unfold isRoot in R.
    rewrite <- X in R; rewrite R in par.
    inversion par.
  Qed.

  Lemma deeper_Depth:
    forall a b, ~ a == b -> isAncestor a b -> Depth a < Depth b.
  Proof.
    intros a b neq anc.
    remember (Depth b) as db.
    revert anc neq Heqdb; revert b; induction db as [| db ih]; intros b hdb neq anc.
    { contradict neq.
      apply Dist_LCA_refl; split; auto.
      apply Dist_LCA_root.
      assert (D := Dist_LCA_root _ b (proj2_sig (getRoot b))).
      assert (X: b == (proj1_sig (getRoot b))).
      2: rewrite X; apply (proj2_sig (getRoot b)).
      apply Dist_LCA_refl; split; auto.
    }
    { case (has_Parent_dec b).
      { intros [Pb hPb].
        case (eqA_dec a Pb); intros heq.
        { rewrite deeper_Depth_step with (a := a) in anc.
          lia.
          rewrite heq; auto.
        }
        unfold Depth in anc.
        rewrite Dist_LCA_Parent_l_not_ancestor with (Px := Pb) in anc; auto.
        specialize (ih Pb (isAncestor_down_inherit _ _ _ hPb hdb neq) heq). 
        assert (db = Depth Pb) by (unfold Depth; rewrite <- (getRoot_same b); lia).
        specialize (ih H); clear H.
        lia.
        {
          intros F; assert (hrt: b == proj1_sig (getRoot b)).
          apply Dist_LCA_refl; split; auto.
          apply Dist_LCA_root.
          apply proj2_sig.
          assert (isRoot b) by (rewrite hrt; apply proj2_sig).
          unfold isRoot in H; rewrite hPb in H; congruence.
        }
      }
      { intros rt; assert (hrt: b == proj1_sig (getRoot b)).
        apply uniq_root; auto.
        apply proj2_sig.
        rewrite hrt in anc.
        assert (F := Dist_LCA_root _ (proj1_sig (getRoot b)) (proj2_sig (getRoot b))).
        unfold Depth in anc; unfold isAncestor in F.
        rewrite (getRoot_same (proj1_sig (getRoot b)) b) in anc.
        lia.
      }
    }
  Qed.

  (** Induction Schema 
    in a subtree T: T is identified by its root rT, 
                    a node n is in T iff isAncestor rT n
   *)
(* TODO
  Lemma subtree_up_induction:
    forall (P: A -> Prop) (P_compat: Proper pequiv P) (rT: A),
      (forall l, isAncestor rT l -> isLeaf l -> P l) ->
      (forall n, isAncestor rT n ->
                 (forall c, InA equiv c (Children n) -> P c) -> P n) ->
      forall n, isAncestor rT n -> P n.
 *)
  
  Lemma subtree_down_induction:
    forall (P: A -> Prop) (P_compat: Proper pequiv P) (rT: A),
      P rT ->
      (forall n, isAncestor rT n ->
                 P n -> forall c, InA equiv c (Children n) -> P c) ->
      forall n, isAncestor rT n -> P n.
  Proof.
    intros P P_compat rT hrT ind goal hgoal.
      remember (Dist_LCA goal rT) as D; rename HeqD into hD.
      revert hD hgoal; revert goal; induction D; intros goal hD hgoal.
      { assert (h: rT == goal).
        apply Dist_LCA_refl; split; auto.
        rewrite <- h; auto.
      }
      case (eqA_dec rT goal); intros heq.
      { rewrite <- heq; auto. }
      case (has_Parent_dec goal).
      { intros [Pgoal hPgoal].
        apply ind with Pgoal.
        apply isAncestor_down_inherit with goal; auto.
        2: apply in_children; auto.
        apply IHD.
        assert (ddd := Dist_LCA_Parent_l goal rT Pgoal hPgoal).
        rewrite <- hD in ddd; simpl in ddd; auto.
        apply isAncestor_down_inherit with goal; auto.
      }
      { intros N.
        assert (h: rT == goal).
        apply Dist_LCA_root with (y := rT) in N.
        apply Dist_LCA_refl; split; auto.
        tauto.
      }
  Qed.
  
(* TODO
  Lemma tree_up_induction:
    forall (P: A -> Prop) (P_compat: Proper pequiv P),
      (forall l, isLeaf l -> P l) ->
      (forall n, (forall c, InA equiv c (Children n) -> P c) -> P n) ->
      forall n, P n.
  Proof.
    intros.
    destruct (getRoot n) as [r hr].
    apply subtree_up_induction with r; auto.
    apply Dist_LCA_root; auto.
  Qed.
 *)
  
  Lemma tree_down_induction:
    forall (P: A -> Prop) (P_compat: Proper pequiv P) (rT: A),
      isRoot rT -> P rT ->
      (forall n, P n -> forall c, InA equiv c (Children n) -> P c) ->
      forall n, P n.
  Proof.
    intros.
    apply subtree_down_induction with rT; auto.
    intros; apply H1 with n0; auto.
    apply Dist_LCA_root; auto.
  Qed.

(* TODO
  Lemma path_up_induction:
    forall (P: A -> Prop) (P_compat: Proper pequiv P) (n: A),
      forall d, isAncestor n d -> P d ->
                (forall nn, isAncestor nn d ->
                            forall c, InA equiv c (Children nn) ->
                                      P c -> P nn) ->
                P n.
 *)
  
  Lemma path_down_induction:
    forall (P: A -> Prop) (P_compat: Proper pequiv P)
           (Q: A -> Prop) (Q_compat: Proper pequiv Q) (Q_classical: forall a, {Q a} + {~Q a}),
      (forall m, P m -> ~ Q m ->
                 exists c, InA equiv c (Children m) /\
                           (P c \/ Q c)) ->
      forall n, P n -> exists d, isAncestor n d /\ Q d.
  Proof.
    intros until n.
    rename H into ind.
    intros hpn; destruct (Q_classical n) as [hq|hq]; auto.
    exists n; split; auto; apply Dist_LCA_self.
    revert hpn hq.
    remember (Height n) as hn.
    assert (heqn: Height n <= hn) by lia.
    clear Heqhn; revert heqn; revert n; induction hn as [| hn ih]; intros n hhn hpn hqn.
    { destruct (ind _ hpn hqn) as [c [hc _]].     
      assert (heqn: Height n = 0) by lia.   
      apply Height_leaf in heqn.
      unfold isLeaf in heqn.
      rewrite heqn in hc.
      rewrite InA_nil in hc; tauto.
    }
    { destruct (ind _ hpn hqn) as [c [hc hpqc]].
      assert (pp: isParent n c) by (rewrite in_children; auto).
      destruct (Q_classical c) as [hq|hq].
      { exists c; split; auto; apply isAncestor_Parent; auto. }
      destruct hpqc as [hp|hhq]; [| tauto].
      destruct (ih c) as [d [hd qd]]; auto.
      apply Height_isParent in pp; lia.
      exists d; split; auto.
      eapply isAncestor_trans; eauto.
      apply isAncestor_Parent; auto.
    }
  Qed.

  (* TODO
  Lemma path_down_induction2:
    forall (P: A -> Prop) (P_compat: Proper pequiv P) (n: A),
      forall a, isAncestor a n -> P a ->
                (forall nn, isAncestor nn n -> P nn ->
                            forall c, InA equiv c (Children nn) ->
                                      isAncestor c n -> P c) ->
                P n.
   *)
  
  Lemma tr_d:
    forall p q,
      transitive_closure (fun x y => Some x == Parent y) p q
      -> Dist_LCA q p > 0.
  Proof.
    intros p q T.
    rewrite <- transitive_closure_left_iff_transitive_closure in T.
    induction T.
    { symmetry in H; apply Dist_LCA_Parent in H.
      destruct H as [H _]; lia.
    }      
    symmetry in H.
    case (isAncestor_dec y z); intros a.
    - assert (X := Dist_LCA_Parent_r_ancestor z y x H a).
      lia.
    - assert (Y := Dist_LCA_Parent_r z y x H a).
      lia.
  Qed.   
    
  Lemma no_loop:
    forall p,
      ~ transitive_closure (fun x y => Some x == Parent y) p p.
  Proof.
    intros p F.
    apply tr_d in F.
    case (Nat.eq_dec (Dist_LCA p p) 0); intros ff; [lia|].
    contradict ff.
    assert (dum: p == p) by auto.
    apply (Dist_LCA_refl p p) in dum.
    tauto.
  Qed.
  
End With_Tree_Topology.

(** * Tree Network
  Given a network, this section provides the definition that constrains the network to be endowed 
  with a rooted spanning tree, namely it should have a tree topology (as defined above) on nodes 
  and the Parent function should correspond to channels
  The tree is _bidirectionnal_.
 *)

Set Implicit Arguments.

Section Tree_Network.

  Variable (Chans: Channels).

  Definition is_spanning (Net: Network)
             (topo: Tree_Topology (A := Node)): Prop :=
    equivlistA equiv (Algorithm.all_nodes (Network := Net))
               (@all_nodes _ _ topo).
    
  Class Up_Tree_Network (Net: Network) :=
    mkUTN {
        UTN_topo:> Tree_Topology (A := Node);

        UTN_parent: Node -> option Channel;
        UTN_parent_compat:> Proper fequiv UTN_parent;        

        UTN_channels_ok:
          forall x c, UTN_parent x == Some c ->
                      exists y, is_channel x c y;

        UTN_parent_ok:
          forall x y, Parent x == Some y <->
                      exists c, UTN_parent x == Some c /\
                                is_channel x c y;
        
        UTN_root_ok:
          forall x, Parent x == None <-> UTN_parent x == None
      }.

  Class Down_Tree_Network (Net: Network) :=
    mkDTN {
        DTN_topo:> Tree_Topology (A := Node);

        DTN_root: Node -> bool;
        DTN_root_compat:> Proper fequiv DTN_root;

        DTN_children: Node -> list Channel;
        DTN_children_compat:> Proper fequiv DTN_children;

        DTN_channels_ok:
          forall c y, InA equiv c (DTN_children y) ->
                      exists x, is_channel y c x;
              
        DTN_children_ok:
          forall x y, Parent x == Some y <->
                      exists c, InA equiv c (DTN_children y) /\
                                is_channel y c x;

        DTN_root_ok:
          forall x, Parent x == None <-> DTN_root x = true
      }.

    Class Tree_Network (Net: Symmetric_Network) :=
    mkTN {
        TN_topo:> Tree_Topology (A := Node);


        TN_parent: Node -> option Channel;
        TN_parent_compat:> Proper fequiv TN_parent;        

        TN_children: Node -> list Channel;
        TN_children_compat:> Proper fequiv TN_children;

        TN_channels_parent_ok:
          forall x c, TN_parent x == Some c ->
                      exists y, is_channel x c y;

        TN_channels_children_ok:
          forall c y, InA equiv c (TN_children y) ->
                      exists x, is_channel y c x;

        TN_parent_children_ok:
          forall x y,
            Parent x == Some y <->
            exists c, TN_parent x == Some c /\
                      InA equiv (reverse x c) (TN_children y) /\
                      is_channel x c y;
        (* is_channel y (reverse x c) x ensured by Symmetric_Network *)

        TN_root_ok:
          forall x, Parent x == None <-> TN_parent x == None
      }.

End Tree_Network.

(** * Tree Algorithm
  We define what is an algorithm that computes on a spanning tree network: the state of every node 
  should contain the channels defined by the spanning rooted tree network.
  Those information should be read-only.
*)
Section Tree_Algorithm.

  Definition Up_Tree_Network_assumption {Chans: Channels}
             (Net: Network)
             (UTN: Up_Tree_Network Net)
             (tree_parent: Node -> option Channel): Prop := 
    tree_parent =~= UTN_parent (Net := Net).

  Definition Down_Tree_Network_assumption {Chans: Channels}
             (Net: Network)
             (DTN: Down_Tree_Network Net)
             (tree_children: Node -> list Channel)
             (tree_root: Node -> bool): Prop := 
    (equiv ==> equivlistA equiv) tree_children (DTN_children (Net := Net)) /\
    tree_root =~= DTN_root (Net := Net).

  Definition Tree_Network_assumption {Chans: Channels}
             (Net: Symmetric_Network)
             (TN: Tree_Network Net)
             (tree_parent: Node -> option Channel)
             (tree_children: Node -> list Channel): Prop := 
    tree_parent =~= TN_parent (Net := Net) /\
    (equiv ==> equivlistA equiv) tree_children (TN_children (Net := Net)).

  Instance Up_Tree_Network_assumption_compat:
    forall {Chans: Channels} Net
           (UTN: Up_Tree_Network Net),
      Proper fequiv (Up_Tree_Network_assumption UTN).
  Proof.
    intros.
    intros ro1 ro2 ero; unfold Up_Tree_Network_assumption.
    split; intros h; rewrite <- h; [symmetry|]; auto.
  Qed.

  Instance Down_Tree_Network_assumption_compat:
    forall {Chans: Channels} Net
           (DTN: Down_Tree_Network Net),
      Proper ((equiv ==> equivlistA equiv) ==> fequiv ==> equiv)
             (Down_Tree_Network_assumption DTN).
  Proof.
    intros.
    intros c1 c2 ec i1 i2 ei; unfold Down_Tree_Network_assumption.
    split; intros [h1 h2]; split;    
     try rewrite <- h1; try rewrite <- h2; auto; symmetry; auto.
  Qed.

  Instance Tree_Network_assumption_compat:
    forall {Chans: Channels} Net
           (TN: Tree_Network Net),
      Proper (fequiv ==> (equiv ==> equivlistA equiv) ==> equiv)
             (Tree_Network_assumption TN).
  Proof.
    intros.
    intros c1 c2 ec i1 i2 ei; unfold Tree_Network_assumption.
    split; intros [h1 h2]; split;    
     try rewrite <- h1; try rewrite <- h2; auto; symmetry; auto.
  Qed.

End Tree_Algorithm.

Unset Implicit Arguments.


Close Scope nat_scope.
Close Scope signature.
