(**
*Certified Round Complexity of Self-Stabilizing Algorithms*
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import ZArith.
From Coq Require Import Bool.
From Coq Require Import Lia.

(** **** Local Imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import RelationA.
From Padec Require Import All_In.
From Padec Require Import WellfoundedUtil.
From Padec Require Import Graph_Diameter.
From Padec Require Import ListUtils.

(* end show *)
Open Scope nat_scope.
Open Scope signature_scope.
Set Implicit Arguments.

(** * DAG: directed acyclic graph *)
Section DAG.

  Context {A} `{SA: Setoid A} {DA: Decider (equiv (A:=A))}.

  Variable (D: relation A) (D_compat: Proper pequiv D)
           (D_dec: Decider D).
  
  (** is_dag: directed acyclic graph: no cycle *)
  Definition is_dag: Prop :=
    (forall x, ~ transitive_closure D x x).

  Definition is_root x: Prop := forall y, ~ D x y.

  Notation is_simple := (is_simple (DA := DA)).
  Notation is_simple_b := (is_simple_b (DA := DA)).

  (** finite set of vertices *)
  Variable (all: list A) (all_here: forall a, InA equiv a all).

  (** at least one element *)
  Variable (one: A).

  (** a dag of at least one element contains a root *)
  Lemma exists_root_in_dag: is_dag -> exists x, is_root x.
  Proof.
    intros id; unfold is_dag in id;
      apply (finite_acyclic_wf (A_dec := DA) (lA := all)) in id;
      auto.
    erewrite wf_flip_finite in id; auto.
    Unshelve.
    2: apply DA.
    assert (fd_dec: Decider (flip D)).
    { intros x y.
      destruct (D_dec y x); [left|right]; auto. }
    assert (fd_comp: Proper pequiv (flip D)).
    { intros x y xy a b ab; rewrite xy, ab; auto. }
    destruct (wf_finite_path all_here fd_dec fd_comp id one)
      as [p [r [_ hr]]].
    exists r; unfold is_root; intros y.
    specialize (hr y); apply hr.
  Qed.

  Lemma unique_root_in_connected_dag:
    is_dag -> is_connected D ->
    forall r1 r2, is_root r1 -> is_root r2 -> r1 == r2.
  Proof.
    intros id ic r1 r2 h1 h2.
    case (DA r1 r2); auto; intros ne.
    specialize (ic r1 r2 ne).
    destruct ic as [p hp]. 
    destruct p as [| x p]; simpl in hp.
    specialize (h1 r2); tauto.
    specialize (h1 x); tauto.
  Qed.

  Lemma chain_flip_nil:
    forall x y, chain D x nil y -> chain (flip D) y nil x.
  Proof.
    intros x y; simpl; intros h; symmetry; auto.
  Qed.

  Lemma chain_flip:
    forall (R: relation A) (rc: Proper pequiv R),
    forall p x y yy, RelationA.chain R x (p ++ yy :: nil) y ->
                     RelationA.chain (flip R) y (rev (x :: p)) x.
  Proof.
    induction p as [| a p ih]; intros x y yy; simpl.
    intros [h1 h2]; split; auto; rewrite <- h2; apply h1.
    intros [h1 h2].
    rewrite chain_app.
    2: apply flip_proper_rel; auto. 
    apply ih in h2; simpl in h2.
    exists a; split; auto; simpl.
    split; auto; apply h1.
  Qed.
  
  Lemma path_to_root_in_dag:
    is_dag ->
    forall x, exists p r, is_root r /\ (r == x \/
                                        is_path D x p r).
  Proof.
    intros id x; unfold is_dag in id;
      apply (finite_acyclic_wf (A_dec := DA) (lA := all)) in id;
      auto.
    erewrite wf_flip_finite in id; auto.
    Unshelve.
    2: apply DA.
    assert (fd_dec: Decider (flip D)).
    { intros xx y.
      destruct (D_dec y xx); [left|right]; auto. }
    assert (fd_comp: Proper pequiv (flip D)).
    { intros xx y xy a b ab; rewrite xy, ab; auto. }
    destruct (wf_finite_path all_here fd_dec fd_comp id x)
      as [p [r [hp hr]]].
    case_eq (rev p).
    intros hrp; symmetry in hrp; rewrite <- rev_nil in hrp; subst.
    exists nil; exists r.
    split; [unfold is_root; apply hr|left; apply hp].
    intros b rp hrp.
    apply (f_equal (rev (A := A))) in hrp.
    rewrite rev_involutive in hrp; simpl in hrp.
    rewrite hrp in hp.
    exists rp; exists r; split;
      [unfold is_root; apply hr|right].
    apply chain_flip in hp; auto.
    rewrite <- chain_cons_non_empty.
    revert hp; apply chain_compat; auto.
    simpl; rewrite rev_involutive; reflexivity.
  Qed.    

  Lemma is_path_app:
    forall p p' x y a, is_path D x (p ++ a :: p') y <->
                       is_path D x p a /\ is_path D a p' y.
  Proof.
    induction p as [| b p ih]; intros p' x y a; simpl.
    tauto.
    split; intros [h1 h2].
    rewrite and_assoc.
    split; [|apply ih]; auto.
    destruct h1 as [h1 h1'].
    split; [|apply ih; split]; auto.
  Qed.
  
  Lemma is_simple_witness:
    forall p, is_simple_b p = false -> exists x l l' l'',
        p == l ++ x :: l' ++ x :: l''.
  Proof.
    induction p as [| a p ih]; intros hs; simpl in hs.
    congruence.
    revert hs; case (InA_dec DA a p).
    { clear ih; intros ina _.
      exists a; exists nil; simpl.
      revert ina; induction p as [| b p ih]; intros ina.
      rewrite InA_nil in ina; tauto.
      rewrite InA_cons in ina; destruct ina as [ina|ina].
      - exists nil; exists p; simpl;
          right; auto; right; [symmetry; auto|reflexivity].
      - apply ih in ina; clear ih; destruct ina as [l' [l'' h]].
        exists (b :: l'); exists l''; right; auto.
        change (eqlistA equiv (b :: p) (b :: l' ++ a :: l'')).
        right; [|inversion h]; auto.
    }
    { intros _ h; destruct (ih h) as [y [t [t' [t'' hh]]]].
      clear ih h.
      exists y; exists (a :: t); exists t'; exists t''.
      change (a :: p == a :: t ++ y :: t' ++ y :: t'').
      right; auto.
    }
  Qed.

  Lemma eqlistA_app_end:
    forall l l' x y,
      l ++ x :: nil == l' ++ y :: nil -> x == y /\ l == l'.
  Proof.
    intros l l' x y h.
    apply rev_eqlistA_compat in h;
      rewrite rev_app_distr in h;
      rewrite rev_app_distr in h; simpl in h.
    inversion_clear h.
    split; auto.
    apply rev_eqlistA_compat in H0;
      rewrite rev_involutive in H0;
      rewrite rev_involutive in H0; auto.
  Qed.
  
  Lemma is_path_simple_in_dag:
    is_dag ->
    forall p x y, is_path D x p y ->
                  is_simple (x :: p ++ y :: nil).
  Proof.
    intros id p x y hp; unfold is_simple.
    case_eq (is_simple_b (x :: p ++ y :: nil));
      auto; intro hs.
    cut (exists a loop, is_path D a loop a).
    { intros [a [loop ok]].
      specialize (id a).
      contradict id; apply transitive_closure_iff_exists_non_empty_chain.
      exists loop; auto. }
    
    apply is_simple_witness in hs.
    destruct hs as [a [l [loop [l' h]]]].
    exists a; exists loop.

    case_eq (rev l');
      [intros rn; symmetry in rn; rewrite <- rev_nil in rn;
       rewrite <- rn in h; clear rn|
       intros  c rl rn;
       apply (f_equal (rev (A := A))) in rn;
       rewrite rev_involutive in rn; simpl in rn;
       rewrite rn in h; clear rn];
      destruct l as [| b l]; simpl in h.
    { inversion_clear h.
      apply eqlistA_app_end in H0.
      revert hp; apply is_path_compat; auto; symmetry; auto;
        apply H0.
    }
    { inversion_clear h.     
      assert (dum: l ++ a :: loop ++ a :: nil = (l ++ a :: loop) ++ a :: nil).
      { rewrite app_comm_cons; apply app_assoc. }
      rewrite dum in H0; clear dum.
      apply eqlistA_app_end in H0.
      destruct H0 as [h1 h2]; rewrite h2 in hp.
      apply is_path_app in hp.
      rewrite <- h1 at 2; apply hp.
    }
    { inversion_clear h.
      assert (dum: loop ++ a :: rev rl ++ c :: nil = (loop ++ a :: rev rl) ++ c :: nil).
      { rewrite app_comm_cons; apply app_assoc. }
      rewrite dum in H0; clear dum.
      apply eqlistA_app_end in H0.
      destruct H0 as [h1 h2]; rewrite h2 in hp.
      apply is_path_app in hp.
      rewrite <- H at 1; apply hp.
    }
    { inversion_clear h.
      assert (dum: l ++ a :: loop ++ a :: rev rl ++ c :: nil =
                   (l ++ a :: loop ++ a :: rev rl) ++ c :: nil).
      { rewrite app_comm_cons; rewrite app_comm_cons.
        rewrite app_assoc; rewrite app_assoc .
        rewrite app_comm_cons; rewrite app_assoc; auto.
      }
      rewrite dum in H0; clear dum.
      apply eqlistA_app_end in H0.
      destruct H0 as [h1 h2]; rewrite h2 in hp.
      apply is_path_app in hp.
      destruct hp as [_ hp]; apply is_path_app in hp.
      apply hp.
    }
  Qed.

  Lemma no_cycle_in_dag:
    is_dag -> forall x p, ~ is_path D x p x.
  Proof.
    intros id x p h.
    specialize (id x); contradict id.
    rewrite transitive_closure_iff_exists_non_empty_chain.
    exists p; apply h.
  Qed.

  Lemma get_parent: is_dag -> forall n, ~ is_root n -> exists m, D n m.
  Proof.
    intros id n nr.
    apply path_to_root_in_dag with n in id.
    destruct id as [p [r [nnr [h|h]]]].
    { contradict nr; unfold is_root in *.
      intros y; rewrite <- h; auto. }
    { destruct p as [| x p]; simpl in h.
      exists r; auto.
      exists x; apply h. }
  Qed.

  Lemma WF_par: is_dag -> well_founded D.
  Proof.
    intros id.
    apply (finite_acyclic_wf (A_dec:= DA)) with all; auto.
  Qed.

  Lemma WF_child: is_dag -> well_founded (flip D).
  Proof.
    rewrite <- (wf_flip_finite (A_dec := DA)); auto.
    apply WF_par; auto.
  Qed.
  
End DAG.

(** * Tree: tools to describe a rooted tree, 
   directed from leaves to the root.
   - spanning tree
   - BFS spanning tree *)
Section Tree.
  
  Context {A} `{SA: Setoid A} {DA: Decider (equiv (A:=A))}.

  (** finite set of vertices *)
  Variable (all: list A) (all_here: forall a, InA equiv a all).
  
  (** relation to describe a tree *)
  Variable (T: relation A) (T_compat: Proper pequiv T) (T_dec: Decider T).

  Variable (Root: A).
  
  Definition single_parent: Prop :=
    forall x y1 y2, T x y1 -> T x y2 -> y1 == y2.

  Definition all_path_to_root: Prop :=
    forall x, ~ x == Root -> exists p, is_path T x p Root.

  Definition is_tree: Prop :=
    is_dag T /\ single_parent /\ all_path_to_root.
  
  Lemma no_path_from_root:
    is_root T Root -> forall x p, ~ is_path T Root p x.
  Proof.
    intros hr x [| y p] F; simpl in F; [|destruct F as [F _]];
      contradict F; apply hr. 
  Qed.

  Lemma Root_is_root:
    is_dag T -> all_path_to_root ->
    forall p, p == Root <-> is_root T p.
  Proof.
    intros id ap.
    intros x; split; intros hx.
    { intros y.
      case (DA y Root); intros hy.
      - eapply no_cycle_in_dag with (x := y) (p := nil) in id.
        contradict id; simpl; rewrite hy, <- hx at 1; auto.
      - specialize (ap _ hy); destruct ap as [p hp].
        eapply no_cycle_in_dag with (x := x) (p := y :: p) in id.
        contradict id; simpl; split; auto.
        rewrite hx; auto.
    }
    { case (DA x Root); intros hxx; auto.
      specialize (ap _ hxx); destruct ap as [[| a p] hp]; simpl in hp.
      specialize (hx Root); tauto.
      specialize (hx a); tauto.
    }
  Qed.

  Lemma Root_is_root':
    is_dag T -> all_path_to_root -> is_root T Root.
  Proof.
    intros; rewrite <- Root_is_root; auto.
  Qed.

  Lemma is_root_unique:
    is_dag T -> all_path_to_root ->
    forall r, is_root T r -> r == Root.
  Proof.
    intros id ap r nr.
    assert (nR: is_root T Root) by (apply Root_is_root; auto).
    case (DA r Root); intros hr; auto.
    specialize (ap _ hr); destruct ap as [[| a p] hp].
    specialize (nr Root); contradict nr; apply hp.
    specialize (nr a); contradict nr; simpl in hp; apply hp.
  Qed.

  Lemma is_spanning_build_path:
    all_path_to_root -> forall n, ~ n == Root ->
                                  {p | is_path T n p Root }.
  Proof.
    intros ap n hr.
    case (exists_path_dec all_here T_compat T_dec n Root (DA := DA)).
    tauto.
    intros c.
    specialize (ap _ hr).
    contradict hr.
    destruct ap as [pt hpt].
    specialize (c pt); tauto.
  Qed.
  
  Lemma is_path_unique_in_tree:
    is_dag T -> single_parent ->
    forall x y p1 p2, is_path T x p1 y -> is_path T x p2 y ->
                      p1 == p2.
  Proof.
    intros id hroot x y p1 p2 hp1 hp2.
    revert hp1 hp2; revert x y p2.
    induction p1 as [| x1 p1 ih].
    { intros x y [| x2 p2] hp1 hp2; simpl in hp1, hp2.
      auto.
      destruct hp2 as [hp2 hp2'].
      specialize (hroot _ _ _ hp1 hp2).
      rewrite <- hroot in hp2'; specialize (id y); contradict id.
      rewrite transitive_closure_iff_exists_non_empty_chain.
      exists p2; auto.
    }
    intros x y [| x2 p2] hp1 hp2; simpl in hp1, hp2.
    { destruct hp1 as [hp1 hp1'].
      specialize (hroot _ _ _ hp1 hp2).
      rewrite hroot in hp1'; specialize (id y); contradict id.
      rewrite transitive_closure_iff_exists_non_empty_chain.
      exists p1; auto.
    }
    { destruct hp1 as [hp1 hp1']; destruct hp2 as [hp2 hp2'].
      specialize (hroot _ _ _ hp1 hp2).
      right; auto.
      eapply ih; eauto.
      rewrite hroot; auto.
    }
  Qed.

  Lemma is_path_dist_in_tree:
    is_dag T -> single_parent ->
    forall p x y, is_path T x p y ->
                  forall d, S (length p) = d <-> is_dist T x y d.
  Proof.
    intros id sp p x y hp d; split; intros h.
    { unfold is_dist.
      right.
      repeat split.
      { contradict hp; rewrite hp.
        apply no_cycle_in_dag; auto. }
      exists p; split; auto.
      intros q hq.
      assert (S (length q) = d).
      rewrite <- h; apply f_equal, length_compat.
      eapply is_path_unique_in_tree; eauto.
      lia. }
    { destruct h as [[h _] | [_[h _]]].
      contradict hp; rewrite h.
      apply no_cycle_in_dag; auto.
      destruct h as [q [hq hd]].
      rewrite <- hd; apply f_equal, length_compat.
      eapply is_path_unique_in_tree; eauto.
    }
  Qed.

  (** Leaf *)
  Definition is_leaf x := forall y, ~ T y x.

  Instance is_leaf_compat: Proper pequiv is_leaf.
  
  Proof.
    intros x y xy; apply all_compat; intros a.
    apply not_compat, T_compat; auto.
  Qed.

  Lemma is_leaf_dec: forall x, {is_leaf x}+{~ is_leaf x}.
  Proof.
    intros x.
    assert(tc: Proper pequiv (fun y => T y x)).
    { intros a b ab; apply T_compat; auto. }
    destruct (All_In_dec_forall_ex all_here tc) as [[a ha]|]; auto.
    { right; contradict ha; auto. }
  Qed.

  (** Children *)
  Definition children x :=
    filter (fun y => if T_dec y x then true else false) all.

  Instance fun_T_dec_compat:
    Proper fequiv
           (fun x y => if T_dec y x then true else false).
  Proof.
    intros x y xy a b ab.
    destruct (T_dec a x); destruct (T_dec b y); auto.
    contradict n; rewrite <- ab, <- xy; auto.
    contradict n; rewrite ab, xy; auto.
  Qed.

  Instance children_compat: Proper pequiv children.
  Proof.
    intros x y xy; apply filter_compat.
    apply fun_T_dec_compat; auto.
    reflexivity.
  Qed.

  Lemma children_prop:
    forall x y, InA equiv y (children x) <-> T y x.
  Proof.
    intros x y; unfold children; rewrite filter_InA.
    case (T_dec y x); intros h; split; auto.
    intros [_ hh]; congruence.
    apply fun_T_dec_compat; auto.
  Qed.
  
  Lemma is_leaf_no_children:
    forall x, is_leaf x <-> children x = nil.
  Proof.
    intros x.
    unfold children; rewrite filter_nil'.
    rewrite forallb_forall_InA.
    split; intros h.
    intros y _; rewrite Bool.negb_true_iff.
    specialize (h y); destruct (T_dec y x); tauto.
    intros y; specialize (h y (all_here y));
      rewrite Bool.negb_true_iff in h.
    destruct (T_dec y x); congruence.
    { intros a b ab.
      apply negb_sym; rewrite negb_involutive.
      apply fun_T_dec_compat; symmetry; auto. }
  Qed.    

  (** Subtree *)
  Definition in_subtree r x :=
    r == x \/ exists p, is_path T x p r.

  Lemma in_subtree_dec: forall r x, {in_subtree r x}+{~ in_subtree r x}.
  Proof.
    intros r x.
    destruct (DA r x).
    left; left; auto.
    destruct (exists_path_dec' all_here T_compat T_dec x r (DA := DA))
      as [[p hp]|h].
    left; right; exists p; auto.
    right; contradict h; destruct h as [h|h]; tauto.
  Qed.
  
  Instance in_subtree_compat: Proper pequiv in_subtree.
  Proof.
    intros a b ab x y xy; apply or_compat.
    rewrite ab, xy; tauto.
    apply ex_compat; intros z.
    apply is_path_compat; auto.
    reflexivity.
  Qed.

  Definition subtree_elements (x: A): list A :=
    filter (fun l => if in_subtree_dec x l then true else false) all.
  
  Instance subtree_elements_compat: Proper pequiv subtree_elements.
  Proof.
    intros x y xy.
    apply filter_compat; [|reflexivity].
    intros l m lm.
    destruct (in_subtree_dec x l); destruct (in_subtree_dec y m); auto.
    contradict n; rewrite <- xy, <- lm; auto.
    contradict n; rewrite xy, lm; auto.
  Qed.

End Tree.


Section SpanningTree.
  Context {A} `{SA: Setoid A} {DA: Decider (equiv (A:=A))}.

  (** finite set of vertices *)
  Variable (all: list A) (all_here: forall a, InA equiv a all).

  (** relation to describe a tree *)
  Variable (T: relation A) (T_compat: Proper pequiv T) (T_dec: Decider T).

  (** relation to describe the graph *)
  Variable (G: relation A) (G_compat: Proper pequiv G) (G_dec: Decider G).

  Variable (Root: A).  

  Definition is_spanning_tree: Prop :=
    is_subgraph T G /\ is_tree T Root.

  Definition is_BFS: Prop :=
    forall x d, is_dist T x Root d -> is_dist G x Root d.

End SpanningTree.

Section Compat.

  Context {A} `{SA: Setoid A}.

  Instance single_parent_compat:
    Proper pequiv (single_parent (A := A)).
  Proof.
    intros a b ab; apply all_compat; intros x.
    apply all_compat; intros y.
    apply all_compat; intros z.
    apply impl_compat.
    apply ab; auto.
    apply impl_compat; [apply ab; auto|tauto].
  Qed.

  Instance all_path_to_root_compat:
    Proper pequiv (all_path_to_root (A := A)).
  Proof.
    intros t1 t2 tt r1 r2 rr; apply all_compat; intros x.
    apply impl_compat.
    rewrite rr; tauto.
    apply ex_compat; intros y.
    apply is_path_compat; auto.
    reflexivity.
  Qed.
  
  Instance is_dag_compat:
    Proper pequiv (is_dag (A := A)).
  Proof.
    intros a b ab; apply all_compat; intros x.
    apply not_compat, transitive_closure_compat; auto.
  Qed.
  
  Instance is_tree_compat:
    Proper pequiv (is_tree (A := A)).
  Proof.
    intros a b ab r1 r2 er; apply and_compat.
    apply is_dag_compat; auto.
    apply and_compat.
    apply single_parent_compat; auto.
    apply all_path_to_root_compat; auto.
  Qed.
  
  Instance is_spanning_tree_compat:
    Proper pequiv (is_spanning_tree (A := A)).
  Proof.
    intros a b ab c d cd v f vf; apply and_compat.
    apply is_subgraph_compat; auto.
    apply is_tree_compat; auto.
  Qed.

  Instance is_root_compat: Proper pequiv (is_root (A := A)).
  Proof.
    intros a b ab c d cd; apply all_compat; intros x.
    apply not_compat, ab; auto.
  Qed.
  
  Instance is_BFS_compat: Proper pequiv (is_BFS (A := A)).
  Proof.
    intros G H eg T TT et r1 r2 er; unfold is_BFS.
    apply all_compat; intros x.
    apply all_compat; intros y.
    apply impl_compat.
    apply is_dist_compat; auto.
    apply is_dist_compat; auto.
  Qed.

End Compat.

Section Dec_finite.

  Context {A} `{SA: Setoid A} {DA: Decider (equiv (A:=A))}.
  Variable (all_A: list A)
           (all_A_prop: forall a, InA equiv a all_A).

  Lemma is_root_dec:
    forall p (r: A),
      Proper pequiv p -> (forall x, {p r x} + {~ p r x}) ->
      {is_root p r} + {~ is_root p r}.
  Proof.
    intros p r pp pdec.
    destruct (All_In_dec_forall_ex all_A_prop (pp _ _ (Equivalence_Reflexive r))) as [[X his_root]|his_root]; auto.
    right. intros hit. 
    specialize (hit X); tauto.
  Qed.
  
  Lemma is_dag_dec:
    forall (G: relation A),
      Proper pequiv G -> (forall x y, {G x y} + {~G x y}) ->
      {is_dag G} + {~is_dag G}.
  Proof.
    intros G pg gdec.
    assert (D: forall x,
               {~ RelationA.transitive_closure G x x} +
               {~~RelationA.transitive_closure G x x}).
    intros x.
    destruct (WellfoundedUtil.finite_transitive_closure_dec
                all_A_prop gdec pg (A_dec := DA) x x);
      [right|left]; auto.
    destruct (All_In_dec_forall_notforall all_A_prop _ D).
    left; intros x; apply n.
    { intros a b ab.
      apply not_compat.
      apply RelationA.transitive_closure_compat; auto. }
    right; auto.
  Qed.

  Lemma single_parent_dec:
    forall (G: relation A),
      Proper pequiv G ->  (forall x y, {G x y} + {~G x y}) ->
      {single_parent G} + {~single_parent G}.
  Proof.
    intros G pg gdec.
    assert (X: forall x,
               {forall a b, G x a -> G x b -> a == b}+
               {~(forall a b, G x a -> G x b -> a == b)}).
    { intros x.
      assert (D: forall a b, {G x a -> G x b -> a == b} +
                             {~ (G x a -> G x b -> a == b)}).
      { intros a b; destruct (gdec x a); destruct (gdec x b); destruct (DA a b); try (left; tauto).
        right; auto.
      }
      destruct (All_In_dec_forall_notforall_2
                  all_A_prop all_A_prop _ D); [left|right]; auto.
      apply e.
      { intros a b ab l m lm.
        apply impl_compat; [apply pg;auto|].
        apply impl_compat; [apply pg; auto|].
        rewrite ab, lm; tauto.
      }
    }
    destruct (All_In_dec_forall_notforall all_A_prop _ X).
    left; intros a b c; apply e.
    { intros l m lm; apply all_compat; intros x; apply all_compat; intros y.
      apply impl_compat; [apply pg;auto|].
      apply impl_compat; [apply pg; auto|tauto].
    }
    right; auto.
  Qed.

  Lemma all_path_to_root_dec:
    forall (G: relation A) (Root: A),
      Proper pequiv G ->  (forall x y, {G x y} + {~G x y}) ->
      {all_path_to_root G Root} + {~ all_path_to_root G Root}.
  Proof.
    intros G r pg gdec.
    
    unfold all_path_to_root.
    cut (forall x, 
            {~ x == r -> exists p, is_path G x p r} +
            {~ (~ x == r -> exists p, is_path G x p r)}).
    { intros P.
      destruct (All_In_dec_forall_notforall all_A_prop _ P).
      left; apply e.
      { intros x y xy.
        apply impl_compat.
        rewrite xy; tauto.
        apply ex_compat; intros l.
        rewrite xy; tauto. }
      right; intuition.
    }
    intros x.
    destruct (exists_path_dec' all_A_prop pg gdec x r
                               (DA := DA)); [left|].
    intros _; destruct s as [l hl]; exists l; auto.
    destruct (DA x r); intuition.
  Qed.

  Lemma is_tree_dec:
    forall (G: relation A) (Root: A),
      Proper pequiv G ->  (forall x y, {G x y} + {~G x y}) ->
      {is_tree G Root} + {~ is_tree G Root}.
  Proof.
    intros G Root pg gdec; unfold is_tree.
    destruct (is_dag_dec pg gdec); [|right; intuition].
    destruct (single_parent_dec pg gdec); [|right; intuition].
    destruct (all_path_to_root_dec Root pg gdec); [|right; intuition].
    left; auto.
  Qed. 

  Lemma is_spanning_tree_dec:
    forall (G G': relation A),
      Proper pequiv G ->  (forall x y, {G x y} + {~G x y}) ->
      Proper pequiv G' ->  (forall x y, {G' x y} + {~G' x y}) ->
      forall p, {is_spanning_tree G G' p} + {~ is_spanning_tree G G' p}.
  Proof.
    intros G G' pg gdec pg' g'dec p; unfold is_spanning_tree.
    destruct (is_subgraph_dec all_A_prop pg' pg g'dec gdec); [|right; intuition].
    destruct (is_tree_dec p pg gdec); [|right; intuition].
    left; auto.
  Qed.

  Lemma is_BFS_dec:
    forall (G G': relation A),
      Proper pequiv G ->  (forall x y, {G x y} + {~G x y}) ->
      Proper pequiv G' ->  (forall x y, {G' x y} + {~G' x y}) ->
      forall p, {is_BFS G' G p} + { ~ is_BFS G' G p}.
  Proof.
    intros G G' pg gdec pg' g'dec p.
    unfold is_BFS.
    assert (D: forall x, {forall d, is_dist G' x p d -> is_dist G x p d} +
                         {~forall d, is_dist G' x p d -> is_dist G x p d}).
    { intros x.
      destruct (is_dist_dec (DA := DA) all_A_prop pg' g'dec x p) as [hg'|hg'].
      destruct (is_dist_dec (DA := DA) all_A_prop pg gdec x p) as [hg|hg].
      { destruct hg as [d hd]; destruct hg' as [d' hd'].
        destruct (Nat.eq_dec d d'); subst.
        left; intros dd hdd. 
        apply is_dist_unique with (d2 := d') in hdd; subst; auto.
        right; intros f; contradict n.
        apply is_dist_unique with G x p; auto. }
      { destruct hg' as [dd hdd].
        right; intros f; specialize (f _ hdd).
        specialize (hg dd); tauto. }
      { left; intros dd hdd.
        specialize (hg' dd); tauto. }
    }      
    destruct (All_In_dec_forall_notforall all_A_prop _ D).
    left; apply i.
    { intros x y xy; apply all_compat; intros d.
      rewrite xy; tauto. }
    right; auto.
  Qed.

End Dec_finite.

(** Tools for height *)
Section Height.

  Context {A} `{SA: Setoid A} {DA: Decider (equiv (A:=A))}.

  (** finite set of vertices *)
  Variable (all: list A) (all_here: forall a, InA equiv a all).

  (** relation to describe a tree *)
  Variable (T: relation A) (T_compat: Proper pequiv T) (T_dec: Decider T).

  Variable (Root: A) (IT: is_tree T Root).
  
  (** height of a subtree of root a *)
  Definition is_height (a: A) (h: nat): Prop :=
    (is_leaf T a /\ h = 0) \/    
    (forall b p, is_path T b p a -> h >= S (length p)) /\
    exists b p, is_path T b p a /\ h = S (length p).
  
  (** Maximum on a non-empty list *)
  Definition max_list' (a: nat) (l: list nat): nat :=
    fold_right max a l.

  Global Instance max_list'_compat: Proper pequiv max_list'.
  Proof.
    intros a a' ea l l' el; unfold max_list'.
    eapply FoldRight.fold_right_compat; eauto.
    { intros aa aa' eaa b b' eb; simpl in ea, eb; rewrite eaa, eb; auto. }
    apply el.
  Qed.

  Lemma max_list'_prop:
    forall a l m, max_list' a l = m <->
                  InA equiv m (a :: l) /\
                  forall x, InA equiv x (a :: l) -> m >= x.
  Proof.
    unfold max_list'; intros a l m.
    split.
    - intros h; subst. 
      split; [apply fr_max1|intros b; apply fr_max2].
    - intros [h hh]; symmetry; apply fr_max3; auto.
  Qed.

  (** Height *)

  Definition height: nat :=
    compute_diameter' all_here T_compat T_dec Root (DA := DA).
  

End Height.

Close Scope signature_scope.
Close Scope nat_scope.
Unset Implicit Arguments.
