(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- June 2016
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Bool.

(** **** Local Imports *)
From Padec Require Import SetoidUtil.

(* end show *)
Open Scope signature_scope.
Set Implicit Arguments.

(** * Bool Set

  set represented by a Boolean function

  + tools on BoolSet

 **)

Section BoolSet.

  Context {A: Type}.

  Definition BoolSet: Type := A -> bool.

  Definition empty_BoolSet: BoolSet := (fun _ => false).
  Definition full_BoolSet: BoolSet := (fun _ => true).

  Definition andBoolSet: BoolSet -> BoolSet -> BoolSet :=
    (fun s s' => (fun a => andb (s a) (s' a))).

  Definition orBoolSet: BoolSet -> BoolSet -> BoolSet :=
    (fun s s' => (fun a => orb (s a) (s' a))).

  Definition negBoolSet: BoolSet -> BoolSet :=
    (fun s => (fun a => negb (s a))).

  Definition diffBoolSet: BoolSet -> BoolSet -> BoolSet :=
    (fun s s' => andBoolSet s (negBoolSet s')).

  Section with_eqA.

    Context {eqA: relation A}.

    Definition eqBoolSet := (eqA ==> @eq bool).
    Definition leBoolSet := (eqA ==> Bool.le).

    Global Instance eqBoolSet_per `{PER _ eqA}: PER eqBoolSet.
    Proof.
      apply respectful_per; auto.
      apply Equivalence_PER, eq_equivalence.
    Qed.

    Global Instance eqBoolSet_compat `{PER _ eqA}: Proper (eqBoolSet ==> eqBoolSet ==> equiv) eqBoolSet.
    Proof.
      intros a b ab c d cd.
      split; intros h.
      intros x y xy.
      erewrite <- cd; [|apply xy].  
      erewrite <- ab; [apply h|]; symmetry; apply xy. 
      intros x y xy.
      erewrite ab; [|apply xy].
      erewrite cd; [apply h|]; symmetry; apply xy. 
    Qed.

    Global Instance leb_trans: Transitive Bool.le.
    Proof. intros [|] [|] [|];simpl in *;tauto||congruence. Qed.

    Global Instance leb_refl: Reflexive Bool.le.
    Proof. intros [|];simpl in *;tauto||congruence. Qed.

    Global Instance leBoolSet_trans `{PER _ eqA}: Transitive leBoolSet.
    Proof.
      intros x y z h1 h2 a b eab.
      apply leb_trans with (y a).      
      assert (pa:Proper eqA a) by proper_.
      apply h1, pa.
      apply h2, eab.
    Qed.

    Lemma eq_leb: forall b1 b2, b1 = b2 -> Bool.le b1 b2.
    Proof.
      intros b1 b2; case b1; case b2; cbv; auto; congruence.
    Qed.

    Lemma eqBoolSet_leBoolSet: forall U1 U2, eqBoolSet U1 U2 -> leBoolSet U1 U2.
    Proof.
      intros U1 U2 eU e1 e2 e.
      apply eq_leb, eU, e.
    Qed.

    Global Instance full_BoolSet_compat: Proper (eqBoolSet) full_BoolSet.
    Proof.
      intros s1 s2 es;reflexivity.
    Qed.

    Global Instance full_BoolSet_le_compat: Proper (leBoolSet) full_BoolSet.
    Proof.
      intros s1 s2 es;simpl;reflexivity.
    Qed.

    Global Instance empty_BoolSet_compat: Proper (eqBoolSet) empty_BoolSet.
    Proof.
      intros s1 s2 es;reflexivity.
    Qed.

    Global Instance empty_BoolSet_le_compat: Proper (leBoolSet) empty_BoolSet.
    Proof.
      intros s1 s2 es;simpl;reflexivity.
    Qed.

    Global Instance andBoolSet_compat:
      Proper (eqBoolSet ==> eqBoolSet ==> eqBoolSet) andBoolSet.
    Proof.
      intros s1 s2 es s1' s2' es' n1 n2 en; unfold andBoolSet; simpl.
      rewrite (es _ _ en); rewrite (es' _ _ en); reflexivity.
    Qed.

    Global Instance andBoolSet_le_compat:
      Proper (leBoolSet ==> leBoolSet ==> leBoolSet) andBoolSet.
    Proof.
      intros s1 s2 es s1' s2' es' n1 n2 en; unfold andBoolSet; simpl.
      specialize (es _ _ en);     specialize (es' _ _ en).
      destruct (s1 n1); destruct (s2 n2);
        destruct (s1' n1); destruct (s2' n2); simpl; auto.
    Qed.

    Global Instance orBoolSet_compat:
      Proper (eqBoolSet ==> eqBoolSet ==> eqBoolSet) orBoolSet.
    Proof.
      intros s1 s2 es s1' s2' es' n1 n2 en; unfold orBoolSet; simpl.
      rewrite (es _ _ en);     rewrite (es' _ _ en); auto.
    Qed.

    Global Instance orBoolSet_le_compat:
      Proper (leBoolSet ==> leBoolSet ==> leBoolSet) orBoolSet.
    Proof.
      intros s1 s2 es s1' s2' es' n1 n2 en; unfold orBoolSet; simpl.
      specialize (es _ _ en);     specialize (es' _ _ en).
      destruct (s1 n1); destruct (s2 n2);
        destruct (s1' n1); destruct (s2' n2); simpl; auto.
    Qed.

    Global Instance negBoolSet_compat:
      Proper (eqBoolSet ==> eqBoolSet) negBoolSet.
    Proof.
      intros s1 s2 es n1 n2 en; unfold negBoolSet; simpl.
      specialize (es _ _ en).
      destruct (s1 n1); destruct (s2 n2); simpl; auto.
    Qed.

    Global Instance negBoolSet_le_compat `{Symmetric _ eqA}:
      Proper (flip leBoolSet ==> leBoolSet) negBoolSet.
    Proof.
      intros s1 s2 es n1 n2 en; unfold negBoolSet; simpl.
      symmetry in en; specialize (es _ _ en).
      destruct (s1 n1); destruct (s2 n2); simpl; auto.
    Qed.

    Global Instance diffBoolSet_compat:
      Proper (eqBoolSet ==> eqBoolSet ==> eqBoolSet) diffBoolSet.
    Proof.
      intros s1 s2 es s1' s2' es'; unfold diffBoolSet.
      apply andBoolSet_compat; [|apply negBoolSet_compat]; assumption.
    Qed.

    Global Instance diffBoolSet_le_compat `{Symmetric _ eqA}:
      Proper (leBoolSet ==> flip leBoolSet ==> leBoolSet) diffBoolSet.
    Proof.
      intros sym s1 s2 es s1' s2' es'; unfold diffBoolSet.
      apply andBoolSet_le_compat; [| apply negBoolSet_le_compat]; assumption.
    Qed.

    Section Tools.

      Lemma le_BoolSet_impl:
        forall s s', leBoolSet s s' ->
                     forall a a', eqA a a' ->
                                  s a = true -> s' a' = true.
      Proof.
        intros s s' ls a a' ea ha;specialize (ls _ _ ea);rewrite ha in ls.
        destruct (s' a');auto.
      Qed.

      Lemma eq_le_iff:
        Symmetric eqA ->
        forall s s', eqBoolSet s s' <-> leBoolSet s s' /\ leBoolSet s' s.
      Proof.
        intros sym s1 s2; split; intros H.
        split; intros n1 n2 en; [|symmetry in en]; specialize (H _ _ en);
          [destruct (s1 n1); destruct (s2 n2) | 
           destruct (s1 n2); destruct (s2 n1)]; simpl; auto.
        destruct H as [H1 H2].
        intros n1 n2 en; specialize (H1 _ _ en);
          symmetry in en; specialize (H2 _ _ en). 
        destruct (s1 n1); destruct (s2 n2); simpl; auto.
      Qed.

      Lemma neg_neg:
        forall s, Proper eqBoolSet s -> eqBoolSet (negBoolSet (negBoolSet s)) s.
      Proof.
        intros s ps n1 n2 en; unfold negBoolSet.
        rewrite (ps _ _ en); apply negb_involutive. 
      Qed.
      
      Lemma and_le: forall s s',
          Proper eqBoolSet s -> leBoolSet (andBoolSet s s') s.
      Proof.
        intros s1 s2 ps1 n1 n2 en; unfold andBoolSet.
        rewrite (ps1 _ _ en).        
        destruct (s1 n2); destruct (s2 n1); simpl; auto.
      Qed.
      
      Lemma le_or:
        forall s s', Proper eqBoolSet s -> leBoolSet s (orBoolSet s s').
      Proof.
        intros s1 s2 ps1 n1 n2 en; unfold orBoolSet.
        rewrite (ps1 _ _ en).        
        destruct (s1 n2); destruct (s2 n1); simpl; auto.       
      Qed.
      
      Lemma diff_le:
        forall s s', Proper eqBoolSet s -> leBoolSet (diffBoolSet s s') s.
      Proof.
        intros s1 s2 ps1 n1 n2 en; unfold diffBoolSet, andBoolSet, negBoolSet.
        rewrite (ps1 _ _ en).        
        destruct (s1 n2); destruct (s2 n1); simpl; auto.       
      Qed.

      Lemma diff_empty_le:
        forall s s', Proper eqBoolSet s' -> eqBoolSet (diffBoolSet s s') empty_BoolSet -> leBoolSet s s'.
      Proof.
        intros s1 s2; unfold diffBoolSet, andBoolSet, negBoolSet, empty_BoolSet.
        intros e2 es n1 n2 en; specialize (es _ _ en); specialize (e2 _ _ en);
          rewrite andb_false_iff in es.
        destruct (s2 n1); destruct (s1 n1); destruct (s2 n2); simpl; auto.
        destruct es as [es|es]; [symmetry|
                                 rewrite negb_false_iff in es];auto.
      Qed.
      
      Lemma negb_empty:
        eqBoolSet full_BoolSet (negBoolSet empty_BoolSet).
      Proof.
        intros n1 n2 en; unfold empty_BoolSet, full_BoolSet, negBoolSet;
          simpl; auto.
      Qed.
      
      Lemma negb_full:
        eqBoolSet empty_BoolSet (negBoolSet full_BoolSet).
      Proof.
        intros n1 n2 en; unfold empty_BoolSet, full_BoolSet, negBoolSet;
          simpl; auto.
      Qed.
      
      Lemma empty_and:
        forall s, eqBoolSet (andBoolSet s empty_BoolSet) empty_BoolSet.
      Proof.
        intros s n1 n2 en; unfold empty_BoolSet, andBoolSet; simpl;
          apply andb_false_r.
      Qed.
      
      Lemma empty_or:
        forall s, Proper eqBoolSet s -> eqBoolSet (orBoolSet s empty_BoolSet) s.
      Proof.
        intros s ps n1 n2 en; unfold orBoolSet, empty_BoolSet.
        rewrite (ps _ _ en).
        apply orb_false_r.
      Qed.
      
      Lemma empty_diff:
        forall s, eqBoolSet  (diffBoolSet empty_BoolSet s) empty_BoolSet.
      Proof.
        intros s n1 n2 en; unfold diffBoolSet, andBoolSet, empty_BoolSet.
        apply andb_false_l; auto.
      Qed.
      
      Lemma diff_empty:
        forall s, Proper eqBoolSet s ->
                  eqBoolSet  (diffBoolSet s empty_BoolSet) s.
      Proof.
        intros s ps n1 n2 en;
          unfold diffBoolSet, andBoolSet, empty_BoolSet, negBoolSet.
        rewrite (ps _ _ en); simpl; apply andb_true_r.
      Qed.

      Lemma full_and:
        forall s, Proper eqBoolSet s -> eqBoolSet (andBoolSet s full_BoolSet) s.
      Proof.
        intros s ps n1 n2 en; unfold andBoolSet, full_BoolSet.
        rewrite (ps _ _ en).
        apply andb_true_r.
      Qed.
      
      Lemma full_or:
        forall s, eqBoolSet (orBoolSet s full_BoolSet) full_BoolSet.
      Proof.
        intros s n1 n2 en; unfold orBoolSet, full_BoolSet.
        apply orb_true_r.
      Qed.
      
      Lemma diff_full:
        forall s, eqBoolSet  (diffBoolSet s full_BoolSet) empty_BoolSet.
      Proof.
        intros s n1 n2 en;
          unfold diffBoolSet, full_BoolSet, empty_BoolSet,
          andBoolSet, negBoolSet; simpl.
        apply andb_false_r.
      Qed.

      Lemma idem_and:
        Symmetric eqA -> Reflexive eqA ->
        forall s s', eqBoolSet s s' -> eqBoolSet (andBoolSet s s') s.
      Proof.
        intros sym refl s s' es n1 n2 en; unfold andBoolSet.        
        symmetry in en; rewrite (es _ _ en); rewrite (es _ _ (refl n1)).
        rewrite andb_diag; auto.
      Qed.
      
      Lemma idem_or:
        Symmetric eqA -> Reflexive eqA ->
        forall s s', eqBoolSet s s' -> eqBoolSet (orBoolSet s s') s.
      Proof.
        intros sym refl s1 s2 es n1 n2 en; unfold orBoolSet.
        symmetry in en; rewrite (es _ _ en); rewrite (es _ _ (refl n1)).
        rewrite orb_diag; auto.
      Qed.

      Lemma idem_diff:
        Reflexive eqA ->
        forall s s', eqBoolSet s s' ->
                     eqBoolSet (diffBoolSet s s') empty_BoolSet.
      Proof.
        intros refl s1 s2 es n1 n2 en;
          unfold diffBoolSet, andBoolSet, negBoolSet, empty_BoolSet.
        rewrite <- (es _ _ (refl n1)).
        rewrite andb_negb_r; auto.
      Qed.

    End Tools.
    
    Section with_Finite_Set.

      Lemma epsilon_BoolSet:
        forall (lA: list A),
          (forall a, InA eqA a lA) ->
          forall BS: BoolSet, Proper (eqBoolSet) BS ->
                              ~ eqBoolSet BS (fun _ => false) ->
                              { a: A | BS a = true }.
      Proof.
        intros lA all_A NS pNS.
        assert (hh:forall l: list A,
                   ~Forall (fun n => NS n=false) l -> {n:A| NS n=true}).
        {
          induction l as [|n l hl];simpl;intro h.
          contradict h.
          apply Forall_nil.
          case_eq (NS n).
          split with n;auto.
          intro en; apply hl.
          intro hh;apply h.
          apply Forall_cons;auto.
        }
        {
          intro h; apply (hh lA).
          rewrite Forall_forall.
          intro hhh;apply h.
          intros n1 n2 en.
          destruct (proj1 (InA_alt _ _ _)(all_A n1)) as [n3 [e3 h3]].
          rewrite e3.
          apply hhh;auto.
        }
      Qed.

      Lemma eqBoolSet_dec:
        Reflexive eqA ->
        forall (lA: list A), (forall a, InA eqA a lA) ->
                             forall BS1 BS2,
                               Proper eqBoolSet BS1 -> Proper eqBoolSet BS2 ->
                               { eqBoolSet BS1 BS2 } + { ~eqBoolSet BS1 BS2 }.
      Proof.
        intros refl lA all_A NS1 NS2 pNS1 pNS2.
        case_eq (forallb (fun n => eqb (NS1 n) (NS2 n)) lA);intro h.
        {
          left.
          rewrite forallb_forall in h.
          intros  n1 n2 en;rewrite en.
          assert (hh:=all_A n2).
          rewrite InA_alt in hh.
          destruct hh as [n [e hh]].
          specialize (h _ hh).
          apply eqb_eq.
          rewrite e;rewrite h;auto.
          split.
        }
        {
          rewrite <- not_true_iff_false in h.
          rewrite forallb_forall in h.
          right;intro hh;apply h;clear h all_A.
          intro x;induction lA as [| n nodes hrec];simpl.
          contradiction.
          intros [ hhh | hhh].
          specialize (hh   _ _ (refl x)).
          rewrite hh.
          destruct (NS2 x);auto.
          auto.
        }
      Qed.

      Definition EMPTY (lA: list A) (BS: BoolSet): bool :=
        (forallb (fun a => negb (BS a)) lA).

      Lemma EMPTY_empty:
        Reflexive eqA ->
        forall (lA: list A),
          (forall a, InA eqA a lA) ->
          forall BS, Proper eqBoolSet BS ->
                     EMPTY lA BS = true <-> eqBoolSet BS empty_BoolSet.
      Proof.
        intros refl lA all_A en en_compat;unfold EMPTY, empty_BoolSet;
          rewrite forallb_forall.
        transitivity (forall x: A, InA eqA x lA -> negb (en x) = true).
        split.
        {
          intuition.
          rewrite InA_alt in H0.
          destruct H0 as [y [exy hy]].
          specialize (H y hy).
          rewrite exy.
          assumption.
        }
        { intuition. }
        transitivity (forall x: A, negb (en x) = true).
        firstorder.
        split;intuition.
        intros x y.
        specialize (H x).
        destruct (en x).
        compute in H.
        congruence.
        congruence.
        rewrite (H x x);
        reflexivity.
      Qed.

      Lemma epsilon_BoolSet_:
        Reflexive eqA ->
        forall (lA: list A),
          (forall a, InA eqA a lA) ->
          forall BS: BoolSet, Proper (eqBoolSet) BS ->
                              EMPTY lA BS = false ->
                              { a: A | BS a = true }.
      Proof.
        intros refl lA all BS ps em.
        apply epsilon_BoolSet with lA; try assumption.
        intros F; contradict em.
        rewrite not_false_iff_true.
        rewrite EMPTY_empty; assumption.
      Qed.

    End with_Finite_Set.

  End with_eqA.

  Global Instance BoolSet_PartialSetoid `{PartialSetoid A}:
    PartialSetoid BoolSet :=
    @Build_PartialSetoid _ (pequiv ==> @eq bool) _.
  
End BoolSet.

Close Scope signature_scope.
Unset Implicit Arguments.
