(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** **** Global Imports *)
Require Import Relations.
Require Import Setoid.
Require Import SetoidList.
Require Import SetoidClass.
Require Import Zcompare.
Require Import Basics.
Require Import Arith.
Require Import List.
Require Import RelationPairs.
Require Import Morphisms.
Require Import Coq.Numbers.Natural.Peano.NPeano.
Require Import Lia.

(** **** Local Imports *)
Require Import NatUtils.
Require Import OptionUtil.
Require Import SetoidUtil.

(* end show *)
Unset Implicit Arguments.
Open Scope nat_scope.

(** Counting the number of elements in sets *)

(** ** Setoid injection *)

(** Definitions of tools to order setoids by Cardinality

      - setoids A/RA and B/RB are compared (A has less equivalence
        classes than B)

      - we define Injection as the existence of an injection relation
        [inj_R] between A/RA and B/RB that satisfies two properties:

            - at_most_one_left: [inj_R] transports the relation pairs
              from B to A.

            - at_least_one_right: [inj_R] relates every relation pair
              in A to at least one relation pair in B *)

(** Relation R transports pairs in RB to pairs in RA :

   For any class b in B/RB there is at most one pair a in A/RA st R a
 b *)

Definition left_unique
           A `{SA:Setoid A}
           B `{SB:Setoid B}
           (R: A -> B -> Prop) :=
  forall (a1: A) (b1: B) (h1: R a1 b1)
         (a2: A) (b2: B) (h2: R a2 b2),
    (b1 == b2 -> a1 == a2).

(** Relation R relates each pair in RA to at least one pair in RB *)

Definition left_total
           A `{SA:Setoid A}
           B `{SB:Setoid B}
           (R: A -> B -> Prop) :=
  forall (a1 a2: A), a1 == a2 ->
                     exists (b1 b2: B),
                       R a1 b1 /\ R a2 b2 /\ b1 == b2.

Inductive Inj
          A `{SA:Setoid A}
          B `{SB:Setoid B} : Prop:=
  mkInj:
    forall
      (inj_R:  A -> B -> Prop)
      (inj_proper: Proper fequiv inj_R)
      (inj_left_unique: left_unique A B inj_R)
      (inj_left_total: left_total A B inj_R),
      Inj A B.

(** * Inj facts *)

(** [Inj] is reflexive *)
Lemma Inj_refl:
  forall {A: Type} `{SA: Setoid A},
    Inj A A.
Proof.
  intros A SA; try split;split with equiv.
  apply equiv_compat.
  intros a b ab c d cd h.
  rewrite ab, cd;assumption.
  intros a b ab; exists a, b; repeat split; auto; apply HA. 
Qed.

(* [Inj] is transitive *)
Lemma Inj_trans: forall
    {A: Type} `{SA: Setoid A}
    (B: Type) `{SB: Setoid B}
    {C: Type} `{SC: Setoid C},
    Inj A B ->
    Inj B C -> Inj A C.
Proof.
  intros A SA B SB C SC.
  intros [R1 PR1 TR1 FR1] [R2 PR2 TR2 FR2].
  split with (fun a c  => exists b , R1 a b /\ R2 b c).
  intros a1 a2 ha c1 c2 hc; split;
    intros [b [hab hbc]]; split with b; split;
      specialize (PR1 _ _ ha _ _ (Equivalence_Reflexive b));
      specialize (PR2 _ _ (Equivalence_Reflexive b) _ _ hc);
      simpl equiv in PR1,PR2;
      intuition.
  intros a1 c1 [b1 [hab1 hbc1]] a2 c2 [b2 [hab2 hbc2]] hc;
    eauto.
  intros a1 a2 ha.
  destruct (FR1 _ _ ha) as [b1 [b2 [hab1 [hab2 hb]]]].
  destruct (FR2 _ _ hb) as [c1 [c2 [hbc1 [hbc2 hc]]]].
  clear PR1 TR1 FR1 PR2 TR2 FR2.
  firstorder.
Qed.

(** ** Modelling Cardinality ordering *)
(** [Card] with parameter Smaller, Same or Larger (see [Card_Prop])
allows to express A/RA having as many, less or more classes then B/RB
*)

(** ** Card_Prop : enumeration of the three basic *)
Inductive Card_Prop: Type := Smaller | Same | Larger.

(** symmetric values for Card_Prop *)
Definition sym_Card_Prop (prop: Card_Prop): Card_Prop :=
  match prop with
  | Smaller => Larger
  | Same => Same
  | Larger => Smaller
  end.

Definition Card (prop: Card_Prop)
           (A : Type) {SA: Setoid A}
           (B : Type) {SB: Setoid B} : Prop:=
  match prop with
    Smaller => Inj A B
  | Same =>  Inj A B /\ Inj B A
  | Larger => Inj B A
  end.

(** Card Same implies any relation *)
Lemma Card_weak: forall (prop: Card_Prop)
                        {A : Type} `{SA: Setoid A}
                        {B : Type} `{SB: Setoid B},
    Card Same A B -> Card prop A B.
Proof.
  intros [ | | ] A SA B SB ;
    intros [[R1 PR1 FR1 TR1] [R2 PR2 FR2 TR2]];
    try split;(split with R1;auto) || (split with R2;auto).
Qed.

(** [Card] is reflexive *)
Lemma Card_refl:
  forall (prop: Card_Prop) {A : Type} `{SA: Setoid A},
    Card prop A A.
Proof.
  intros [ | | ] A SA;simpl;[|split|];apply Inj_refl;assumption.
Qed.

(** [Card] is symmetric when using opposite property *)
Lemma Card_sym:
  forall (prop: Card_Prop)
         {A : Type} `{SA: Setoid A}
         {B : Type} `{SB: Setoid B},
    Card (sym_Card_Prop prop) A B -> Card prop B A.
Proof.
  intros [ | | ] A B RA RB;simpl;intuition.
Qed.

(** Idem but with iff *)
Lemma Card_sym_iff:
  forall (prop: Card_Prop)
         {A : Type} `{SA: Setoid A}
         {B : Type} `{SB: Setoid B},
    Card (sym_Card_Prop prop) B A <-> Card prop A B.
Proof.
  split.
  apply Card_sym; assumption.
  intro h; destruct prop as [ | | ]; apply (Card_sym);
    simpl sym_Card_Prop;auto.
Qed.

(** [Card] is transitive *)
Lemma Card_trans:
  forall (prop: Card_Prop)
         {A : Type} `{SA: Setoid A}
         (B : Type) `{SB: Setoid B}
         {C : Type} `{SC: Setoid C},
    Card prop A B ->
    Card prop B C -> Card prop A C.
Proof.
  intros  [ | | ] A SA B SB C SC H1 H2;simpl in *;try split;
    eapply (Inj_trans B);intuition.
Qed.

(* Would be nice but Universe Inconsistency

Global Instance Card_setoid: Setoid (sigT (@Setoid)).
 *)

(** Larger and Smaller (using the same counting relation)
      implies Same *)
Lemma Card_antisym:
  forall {A : Type} `{SA: Setoid A}
         {B : Type} `{SB: Setoid B},
    Card Smaller A B -> Card Larger A B ->
    Card Same A B.
Proof.
  intros A SA B SB;simpl;intuition.
Qed.

(** Same implies Larger and Smaller *)
Lemma Card_antisym_converse:
  forall {A : Type} `{SA: Setoid A}
         {B : Type} `{SB: Setoid B},
    Card Same A B -> Card Smaller A B /\ Card Larger A B.
Proof.
  intros A SA B SB [].
  split; assumption.
Qed.

(** Same implies symmetric *)
Lemma Card_Same_sym:
  forall {A : Type} `{SA: Setoid A}
         {B : Type} `{SB: Setoid B},
    Card Same A B -> Card Same B A.
Proof.
  intros A B RA RB same.
  unfold Card in *.
  split; apply same.
Qed.

(** ** Counting fundamental lemma using [ltN]
         ([Inj] captures finite Cardinality ordering) *)

Lemma Inj_le_iff:
  forall (m n: nat), Inj (ltN m) (ltN n) <-> m <= n.
Proof.
  intros m n; split.
  2:{
    (* <- *)
    intros H. induction H as [ | m' n Hrec].
    apply Inj_refl.
    apply Inj_trans with (ltN m') _.
    apply Hrec.
    
    split with (fun (x: ltN m')
                    (y: ltN (S m')) => @proj1_sig _ _ x = @proj1_sig _ _ y).
    intros [a1 Ha1] [a2 Ha2] ea [b1 Hb1] [b2 Hb2] eb; simpl.
    compute in ea, eb. split; congruence.
    
    intros [a1 Ha1] [b1 Hb1] e1 [a2 Ha2] [b2 Hb2] e2; compute.
    compute in e1, e2. intro; congruence.
    
    intros [a1 Ha1] [a2 Ha2] ea; compute.
    compute in ea.
    exists (@exist _ (fun n => n < S m') a1 (lt_S Ha1)).
    exists (@exist _ (fun n => n < S m') a2 (lt_S Ha2)).
    split; [|split]; congruence.
  }
  (* -> *)
  revert n.
  induction m as [| m Hrecm].
  intros; lia.
  intros [| n].
  intros [R PR TR  FR ].
  specialize
    (FR (@exist _ (fun x=> x<S m) 0 (lt_0_Sn m ))
        (@exist _ (fun x=> x<S m) 0 (lt_0_Sn m ))
        (refl_equal 0)).
  destruct FR as [[x h] _].
  elim (lt_n_0 h).

  intros [R PR TR  FR ].
  apply le_n_S.
  apply Hrecm;clear Hrecm.
  split with (fun x y =>
                (R (@ltSN_inj m x) (@ltSN_inj n y)) \/
                (R (@ltSN_inj m x) (@ltSN_max n)
                 /\ R (@ltSN_max m) (@ltSN_inj n y))).


  (* Proper *)
  intros aa1 aa2 ea bb1 bb2 eb.
  apply or_compat.
  apply PR; auto.
  apply and_compat.
  apply PR; auto.
  apply PR; auto.

  (* at_most_one_left *)
  intros aa1 bb1 e1 aa2 bb2 e2.
  destruct e1 as [ H1 | [H11 H12]];
    destruct e2 as [ H2 | [H21 H22]].
  apply (TR _ _ H1 _ _ H2).

  intros H.
  assert (hh := ltSN_inj_compat H).
  specialize (TR _ _ H1 _ _ H22 hh).
  elim (ltSN_inj_not_max TR).

  intros H.
  assert (hh := ltSN_inj_compat H).
  specialize (TR _ _ H12 _ _ H2 hh).
  symmetry in TR.
  elim (ltSN_inj_not_max TR).

  intros H.
  assert (hh := ltSN_inj_compat H).
  specialize (TR _ _ H11 _ _ H21).
  apply TR.
  reflexivity.

  (* func_left *)
  intros aa1 aa2 ea.

  assert (hh := ltSN_inj_compat ea).
  assert (FR' := FR _ _ hh).
  destruct FR' as [ b1 [ b2 [ Hr1 [ Hr2 He ]]]].
  destruct (ltSN_inj_max_dec b1) as [ [b1' Hb1] | Hb1];
    destruct (ltSN_inj_max_dec b2) as [ [b2' Hb2] | Hb2].
  exists b1'. exists b2'.
  split.
  left;
    apply (PR _ _ (Equivalence_Reflexive (ltSN_inj aa1)) _ _ Hb1);
    assumption.
  split.
  left;
    apply (PR _ _ (Equivalence_Reflexive (ltSN_inj aa2)) _ _ Hb2);
    assumption.
  rewrite Hb1 in He; rewrite Hb2 in He; apply He.

  rewrite Hb1, Hb2 in He;
    elim (ltSN_inj_not_max He).

  rewrite Hb1, Hb2 in He;
    symmetry in He;
    elim (ltSN_inj_not_max He).

  (* b1 = b2 = n*)
  rewrite Hb1 in Hr1.
  rewrite Hb2 in Hr2.
  clear Hb1 Hb2 b1 b2 He.
  assert (FR'' := FR _ _ (Equivalence_Reflexive (ltSN_max m))).
  destruct FR'' as [ bb1 [ bb2 [ Hr1' [ Hr2' eb ]]]].

  destruct (ltSN_inj_max_dec bb1) as [[bb1' hbb1]|hbb1];
    destruct (ltSN_inj_max_dec bb2) as [[bb2' hbb2]|hbb2];
    rewrite hbb1 in Hr1',eb;rewrite hbb2 in Hr2',eb.
  split with bb1'.
  split with bb2'.
  auto.

  elim (ltSN_inj_not_max eb).
  symmetry in eb; elim (ltSN_inj_not_max eb).

  specialize (TR _ _ Hr1 _ _ Hr1' (Equivalence_Reflexive _)).
  elim (ltSN_inj_not_max TR).
Qed.

(** Correspondence between ltN size and Card ordering *)

Lemma Card_smaller_le_iff:
  forall (m n: nat), Card Smaller (ltN m) (ltN n) <-> m <= n.
Proof.
  apply Inj_le_iff.
Qed.

Lemma Card_larger_ge_iff:
  forall (m n: nat), Card Larger (ltN m) (ltN n) <-> m >= n.
Proof.
  intros m n;apply Inj_le_iff.
Qed.

Lemma Card_same_eq_iff:
  forall (m n: nat), Card Same (ltN m) (ltN n) <-> m = n.
Proof.
  intros m n.
  transitivity (m<=n /\ n<=m); [|lia].
  simpl Card.
  apply Morphisms_Prop.and_iff_morphism;apply Inj_le_iff.
Qed.


(** For instance, (Num_Card RA Same n) means that RA contains exactly
n elements. *)
Definition Num_Card prop (A:Type) {SA:Setoid A} n := Card prop A (ltN n).

(** ** Cartesian Products *)

(** Cartesian product is monotonic w.r.t. Cardinality *)

Lemma Inj_prod:
  forall {A1 : Type} `{SA1: Setoid A1}
         {A2 : Type} `{SA2: Setoid A2}
         {B1 : Type} `{SB1: Setoid B1}
         {B2 : Type} `{SB2: Setoid B2},
    Inj A1 A2 ->  Inj B1 B2 -> Inj (A1*B1) (A2*B2).
Proof.
  intros A1 SA1 A2 SA2 B1 SB1 B2 SB2 HA HB.
  destruct HA as [RA PRA TRA FRA].
  destruct HB as [RB PRB TRB FRB].
  split with (fun ab1 ab2 =>
                RA (fst ab1) (fst ab2) /\ RB (snd ab1) (snd ab2)).

  (* proper *)
  intros ab1 ab1' [ eqA1 eqB1 ] ab2 ab2' [ eqA2 eqB2 ].
  apply and_compat.
  apply (PRA _ _ eqA1 _ _ eqA2).
  apply (PRB _ _ eqB1 _ _ eqB2).

  (* at_most_one_left *)
  intros ab1 ab2 [eqA eqB ] ab1' ab2' [eqA' eqB' ].
  intros [];split.
  apply (TRA _ _ eqA _ _ eqA');assumption.
  apply (TRB _ _ eqB _ _ eqB');assumption.

  (* func_left *)
  intros ab1 ab1' [eqA1 eqB1].
  specialize (FRA _ _ eqA1). destruct FRA as [a2 [a2' [eqa2 [eqa2' eqa]]]].
  specialize (FRB _ _ eqB1). destruct FRB as [b2 [b2' [eqb2 [eqb2' eqb]]]].
  exists (a2, b2). exists (a2', b2').
  intuition.
  split;assumption.
Qed.

(** *** Counting elements in product sets *)
Lemma Card_prod:
  forall prop
         {A1 : Type} `{SA1: Setoid A1}
         {A2 : Type} `{SA2: Setoid A2}
         {B1 : Type} `{SB1: Setoid B1}
         {B2 : Type} `{SB2: Setoid B2},
    Card prop A1 A2 ->  Card prop B1 B2 ->
    Card prop (A1*B1) (A2*B2).
Proof.
  intros [ | | ] A1 A2 B1 B2 RA1 RA2 RB1 RB2 ;try split;apply Inj_prod;
    simpl Card in *;intuition.
Qed.

(** ** Encoding from  (ltN m * ltN n) to (ltN (m * n))
   Unique and compatible. *)
Program Definition encode (m n: nat) (ab: ltN m * ltN n): ltN (m * n)
  := let (a, b) := ab in m * b + a.
Obligation 1.
Proof.
  destruct a as [ a Ha ]; destruct b as [ b Hb ].
  simpl.
  assert (Hbe:=lt_le_S Hb).
  assert (Hmult:=mult_le_compat_l _ _ m Hbe).
  specialize (plus_le_lt_compat _ _ _ _ Hmult Ha).
  rewrite (mult_succ_r m b).
  lia.
Qed.

Lemma encode_compat:
  forall (m n: nat),
    Proper fequiv (encode m n).
Proof.
  intros m n.
  intros [[m1 Hm1] [n1 Hn1]] [[m2 Hm2] [n2 Hn2]] [eqm eqn].
  compute in eqm, eqn.
  compute; fold plus; fold mult.
  rewrite eqm, eqn; reflexivity.
Qed.

(** [encode] is injective *)
Lemma encode_unique:
  forall m n ab ab',
    (encode m n ab) == (encode m n ab') ->
    ab == ab'.
Proof.
  intros m n [[a Ha] [b Hb]] [[a' Ha'] [b' Hb']] H.
  compute in H; fold plus in *; fold mult in *.
  compute.
  apply and_comm.
  apply (NPeano.Nat.div_mod_unique m b b' a a' Ha Ha' H).
Qed.

(** ** Decoding from (ltN (m * n)) to (ltN m * ltN n)
   Unique and compatible. *)
Program Definition decode (m n: nat) (Hm: m > 0)
        (ab: ltN (m * n)): (ltN m * ltN n) :=
  (modulo (proj1_sig ab) m, div (proj1_sig ab) m).
Obligation 1.
Proof.
  apply mod_bound_pos.
  lia. assumption.
Qed.

Obligation 2.
Proof.
  destruct ab as [ab Hab].
  simpl.
  assert (H := div_mod ab m).
  assert (HH: m <> 0).
  lia.
  specialize (H HH).
  rewrite H in Hab.
  clear HH.
  assert (HH: m * NPeano.div ab m < m * n).
  lia.
  rewrite <- NPeano.Nat.mul_lt_mono_pos_l in HH; auto.
Qed.

Lemma decode_compat:
  forall (m n: nat) (Hm: m > 0),
    Proper fequiv (decode m n Hm).
Proof.
  intros m n Hm [ab Hab] [ab' Hab'] eqab.
  hnf in eqab;simpl in eqab.
  split;unfold decode;hnf;simpl;apply f_equal2;reflexivity|| assumption.
Qed.

(** [decode] o [encode] is identity *)
Lemma decode_encode:
  forall (m n: nat) (Hm: m > 0) (a: ltN m) (b: ltN n),
    (decode _ _ Hm (encode _ _ (a,b))) == (a, b).
Proof.
  intros m n Hm [a Ha] [b Hb].
  split;unfold decode;hnf;simpl.
  symmetry;
    apply Nat.mod_unique with b;lia.
  symmetry;
    apply Nat.div_unique with a;lia.
Qed.

(** [encode] o [decode] is identity *)
Lemma encode_decode:
  forall (m n: nat) (Hm: m > 0) (ab: ltN (m * n)),
    (encode _ _ (decode _ _ Hm ab)) == ab.
Proof.
  intros m n Hm [ab Hab].
  unfold encode;hnf;simpl.
  assert (m <> 0). lia.
  symmetry. apply (div_mod ab m H).
Qed.

(** *** Count elements in (ltN (m * n)) and in (ltN n * ltN m) *)
Lemma Inj_prod_left:
  forall m n , Inj (ltN m * ltN n) (ltN (m * n)).
Proof.
  intros m n.

  split with (fun (ab: (ltN m * ltN n))
                  (card: ltN (m * n))
              => (encode m n ab) == card).
  (* proper *)
  intros a_b ab [eqa eqb] c c' eqc.
  simpl in eqa, eqb.
  simpl.
  rewrite (encode_compat _ _ _ _ (conj eqa eqb)), eqc;
    reflexivity.

  (* at_most_one_left *)
  intros a_b ab eqab a_b' ab' eqab'.
  simpl.
  simpl in eqab, eqab'.
  rewrite <- eqab, <- eqab'.
  intros AB.
  apply (encode_unique m n _ _ AB).

  (* func_left *)
  intros ab ab' Hab.
  exists (encode _ _ ab).
  exists (encode _ _ ab').
  simpl.
  repeat split; try reflexivity.
  apply encode_compat; auto.
Qed.

Lemma Inj_prod_right_non_zero:
  forall m n , m > 0 ->
               Inj (ltN (m * n)) (ltN m * ltN n).
Proof.
  intros m n Hm.
  split with (fun
                 (card: ltN (m * n))
                 (ab: (ltN m * ltN n))
               =>  encode m n ab == card).

  (* proper (<-) *)
  intros  c c' eqc a_b ab [eqa eqb].
  simpl in eqa, eqb.
  simpl.
  rewrite (encode_compat _ _ _ _ (conj eqa eqb)), eqc;
    reflexivity.

  (* at_most_one_left (<-) *)
  intros a_b ab eqab a_b' ab' eqab'.
  simpl.
  simpl in eqab, eqab'.
  rewrite <- eqab, <- eqab'.
  intros AB.
  apply (encode_compat _ _ _ _ AB).

  (* func_right *)
  intros ab ab' Hab.
  exists (decode _ _ Hm ab).
  exists (decode _ _ Hm ab').
  split.
  apply encode_decode.
  split.
  apply encode_decode.
  specialize (decode_compat  _ _ Hm _ _ Hab); intro.
  destruct H as [H1 H2].
  split;assumption.
Qed.

Lemma Inj_prod_right_zero:
  forall n, Inj (ltN (0 * n)) (ltN 0 * ltN n).
Proof.
  intro n;
    split with (fun x y => False).
  repeat intro;reflexivity.
  intros [a b] ab [].
  intros [a b] ab.
  elimtype False.
  lia.
Qed.

Lemma Card_as_many_prod:
  forall m n ,
    Card Same (ltN n * ltN m) (ltN (n * m)).
Proof.
  intros m n.
  split.
  apply Inj_prod_left.
  case (zerop n); intros.
  rewrite e.
  apply Inj_prod_right_zero.
  apply Inj_prod_right_non_zero.
  lia.
Qed.

Lemma Num_Card_prod:
  forall {A : Type} `{SA: Setoid A}
         {B : Type} `{SB: Setoid B},
  forall prop n m,
    Num_Card prop A n -> Num_Card prop B m -> Num_Card prop (A*B) (n * m).
Proof.
  unfold Num_Card.
  intros A SA B SB prop n m Hn Hm.
  eapply Card_trans with (ltN n * ltN m)%type _.
  apply Card_prod; assumption.
  apply Card_weak.
  apply Card_as_many_prod.
Qed.

(** *** Disjoint sum of relations *)
Definition coprod_rel (A B: Type) (RA: relation A) (RB: relation B):
  relation (A + B) :=
  fun s1 s2 => match s1,s2 with
                 inl a1,inl a2 => RA a1 a2
               | inr b1,inr b2 => RB b1 b2
               | _ ,_ => False
               end.

Arguments coprod_rel [A B].

Infix "+" := coprod_rel: signature_scope.

Global Instance coprod_rel_equiv:
  forall A B (RA: relation A) (RB: relation B),
    Equivalence RA -> Equivalence RB -> Equivalence (RA + RB).
Proof.
  intros A B RA RB RA_equiv RB_equiv.
  split.
  intros [a|b];simpl;reflexivity.
  intros [a1|b1] [a2|b2] h;simpl in h|-*;((symmetry;apply h) || elim h).
  intros [a1|b1] [a2|b2] [a3|b3] h12 h23;simpl in h12,h23|-*;simpl;
    try (elim h12 || elim h23).
  transitivity a2;assumption.
  transitivity b2;assumption.
Qed.

Global Instance coproduct_setoid
         {A : Type} `{SA: Setoid A}
         {B : Type} `{SB: Setoid B} : Setoid (A+B) :=
  {|equiv:=coprod_rel equiv equiv|}.

Lemma Inj_coprod:
  forall {A1 : Type} `{SA1: Setoid A1}
         {A2 : Type} `{SA2: Setoid A2}
         {B1 : Type} `{SB1: Setoid B1}
         {B2 : Type} `{SB2: Setoid B2},
    Inj A1 A2 ->  Inj B1 B2 -> Inj (A1+B1) (A2+B2).
Proof.
  intros A1 SA1 A2 SA2 B1 SB1 B2 SB2 hA hB.
  destruct hA as [RA PRA TRA FRA].
  destruct hB as [RB PRB TRB FRB].
  split with (fun s1 s2 => match s1,s2 with
                             inl a1,inl a2 => RA a1 a2
                           | inr b1,inr b2 => RB b1 b2
                           | _ ,_ => False
                           end).

  (* proper *)
  intros [a1|b1] [a1'|b1'] eq1 [a2|b2] [a2'|b2'] eq2;
    simpl in eq1,eq2|-*;try solve [elim eq1|elim eq2| reflexivity].
  apply PRA;assumption.
  apply PRB;assumption.

  (* at_most_one_left *)
  intros [a1|b1] [a2|b2] h [a1'|b1'] [a2'|b2'] h';
    simpl in h,h'|-*;try solve [intros []| elim h|elim h'| reflexivity].
  apply TRA;assumption.
  apply TRB;assumption.

  (* func_left *)
  intros [a1|b1] [a1'|b1'] eq1;simpl in eq1;try elim eq1.
  specialize (FRA _ _ eq1). destruct FRA as [a2 [a2' [eqa2 [eqa2' eqa]]]].
  exists (inl a2);exists (inl a2');simpl;intuition.
  specialize (FRB _ _ eq1). destruct FRB as [b2 [b2' [eqb2 [eqb2' eqb]]]].
  exists (inr b2);exists (inr b2');simpl;intuition.
Qed.

Lemma Card_coprod:
  forall prop
         {A1 : Type} `{SA1: Setoid A1}
         {A2 : Type} `{SA2: Setoid A2}
         {B1 : Type} `{SB1: Setoid B1}
         {B2 : Type} `{SB2: Setoid B2},
    Card prop A1 A2 ->  Card prop B1 B2 ->
    Card prop (A1 + B1) (A2 + B2).
Proof.
  intros [ | | ] A1 SA1 A2 SA2 B1 SB1 B2 SB2;try split;apply Inj_coprod;
    simpl in *;tauto.
Qed.

(** *** Encoding / Decoding from (ltN m + ltN n) to ltN (m + n) *)
Program Definition coprod_encode (m n: nat) (x: ltN m + ltN n): ltN (m + n) :=
  match x with inl (@exist _ _ m' h) => m'+ n | inr (@exist _ _ n' h) => n' end.
Next Obligation.
Proof.
  lia.
Qed.
Next Obligation.
Proof.
  lia.
Qed.

Program Definition coprod_decode (m n: nat) (x: ltN (m+n)): ltN m + ltN n :=
  let (k,hk):= x in
  match le_gt_dec n k return ltN m + ltN n  with
    left h => inl (k-n) | right h => inr k end.
(* begin hide *)
Next Obligation.
  lia.
Qed.
(* end hide *)

Global Instance coprod_encode_compat (m n: nat):
  Proper fequiv (coprod_encode m n).
Proof.
  intros [[a1 h1]|[b1 h1]] [[a2 h2]|[b2 h2]] h;compute in *;fold plus in *;
    elim h;reflexivity.
Qed.

Global Instance coprod_decode_compat (m n: nat):
  Proper fequiv (coprod_decode m n).
Proof.
  intros [x1 h1] [x2 h2] h;compute in h;subst;simpl.
  destruct (le_gt_dec n x2) as [LE | GT];compute;fold minus;lia.
Qed.

Lemma coprod_encode_unique:
  forall n m x x',
    (coprod_encode m n x) == (coprod_encode m n x') ->
    x == x'.
Proof.
  intros n m [[a Ha]|[b Hb]] [[a' Ha']|[b' Hb']] H;
    compute in H; fold plus in *;compute;
      lia.
Qed.

Lemma coprod_decode_encode:
  forall m n x,
    (coprod_decode m n (coprod_encode m n x)) == x.
Proof.
  intros m n [[a Ha]|[b Hb]];simpl.
  destruct (le_gt_dec n (a+n)) as [LE | GT];compute;fold minus;fold plus;lia.
  destruct (le_gt_dec n b) as [LE | GT];compute;fold minus;fold plus;lia.
Qed.

Lemma coprod_encode_decode:
  forall m n x,
    (coprod_encode m n (coprod_decode m n x)) == x.
Proof.
  intros m n [k hk];simpl.
  destruct (le_gt_dec n k) as [LE | GT];compute;fold minus;fold plus;lia.
Qed.

(** *** Count elements in (ltN (m + n)) and in (ltN n + ltN m) *)

Lemma Inj_coprod_left:
  forall (m n: nat),Inj (ltN m + ltN n) (ltN (m + n)).
Proof.
  intros m n.
  split with (fun (ab: (ltN m + ltN n))
                  (card: ltN (m + n))
              => (coprod_encode m n ab) == card).
  (* proper *)
  intros a_b ab eqab c c' eqc.
  apply equiv_compat;try assumption.
  apply coprod_encode_compat;reflexivity|| assumption.

  (* at_most_one_left *)
  intros a_b ab eqab a_b' ab' eqab'.
  rewrite <- eqab, <- eqab'.
  apply coprod_encode_unique.

  (* func_left *)
  intros ab ab' Hab.
  exists (coprod_encode _ _ ab).
  exists (coprod_encode _ _ ab').
  repeat split; try reflexivity.
  apply coprod_encode_compat; auto.
Qed.

Lemma Inj_coprod_right:
  forall (m n: nat), Inj (ltN (m + n)) (ltN m + ltN n).
Proof.
  intros m n.
  split with (fun (card: ltN (m + n))
                  (ab: (ltN m + ltN n))
              => (coprod_encode m n ab) == card).
  (* proper *)
  intros a_b ab eqab c c' eqc.
  apply equiv_compat;try reflexivity.
  apply coprod_encode_compat;assumption || reflexivity.
  assumption.

  (* at_most_one_left *)
  intros a_b ab eqab a_b' ab' eqab'.
  rewrite <- eqab, <- eqab'.
  apply coprod_encode_compat.

  (* func_right *)
  intros ab ab' Hab.
  exists (coprod_decode _ _ ab).
  exists (coprod_decode _ _ ab').
  split.
  apply coprod_encode_decode.
  split.
  apply coprod_encode_decode.
  apply coprod_decode_compat.
  assumption.
Qed.

Lemma Card_as_many_coprod:
  forall (m n: nat),Card Same (ltN m + ltN n) (ltN (m + n)).
Proof.
  intros m n;split.
  apply Inj_coprod_left.
  apply Inj_coprod_right.
Qed.

Lemma Num_Card_prop_coprod:
  forall {A : Type} `{SA: Setoid A}
         {B : Type} `{SB: Setoid B},
  forall prop n m,
    Num_Card prop A n -> Num_Card prop B m -> Num_Card prop (A + B) (n + m).
Proof.
  unfold Num_Card.
  intros A SA B SB prop n m Hn Hm.
  apply Card_trans with (ltN n + ltN m)%type _.
  apply Card_coprod; assumption.
  apply Card_weak.
  apply Card_as_many_coprod.
Qed.


Infix "+" := coprod_rel: signature_scope.

Lemma Inj_unit_l: Inj unit (ltN 1).
  split with (fun  _ _ => True);cbv.
  cbv;tauto.
  destruct a1,a2;reflexivity.
  do 2 esplit with (@exist _ _ 0 _);cbv.
  intuition.
  Unshelve.
  hnf;lia.
  hnf;lia.
Qed.

Lemma Inj_unit_r: Inj (ltN 1) unit.
  split with (fun  _ _ => True);cbv.
  cbv;tauto.
  intros [[ | n1] hn1] [] [] [[ | n2] hn2] [] [];lia.
  do 2 split with tt;intuition.
Qed.

Lemma Num_Card_prop_unit:
  forall prop, Num_Card prop unit 1.
Proof.
  unfold Num_Card.
  intros [ | | ];try split;simpl;auto using Inj_unit_r,Inj_unit_l.
Qed.

Lemma Inj_empty_l: Inj Empty_set (ltN 0).
  split with (fun  _ _ => True);cbv;intros [].
Qed.

Lemma Inj_empty_r: Inj (ltN 0) Empty_set.
  split with (fun  _ _ => True);cbv;intros [n h];elimtype False;lia.
Qed.

Lemma Num_Card_prop_empty:
  forall prop, Num_Card prop Empty_set 0.
Proof.
  unfold Num_Card.
  intros [ | | ];try split;simpl;auto using Inj_empty_r,Inj_empty_l.
Qed.

(** * Subsets *)
Section Subsets.

  Context {A} `{SA:Setoid A}.

  (** A subset has Smaller Cardinality *)
  Lemma Subset_Num_Card: forall P:A -> Prop, Proper fequiv P ->
                                             Card Smaller (sig P) A.
    intros P P_compat.
    split with (fun ta a => a == (proj1_sig ta)).
    intros [ta1 ha1] [ta2 ha2] hta a1 a2 ha;compute in hta;simpl.
    apply equiv_compat;assumption.
    intros [ta1 ha1] a1 h1 [ta2 ha2] a2 h2;simpl in *.
    eapply proj1;apply equiv_compat;assumption.
    intros [ta1 ha1] [ta2 ha2] hta;simpl in *.
    split with ta1;split with ta2.
    hnf in hta.
    intuition.
  Qed.

  (** The set itself is one of its subset, but of same Cardinality *)
  Lemma Trivial_subset_Num_Card: Card Same (sig (fun _:A => True)) A.
    split.
    apply Subset_Num_Card.
    split;constructor.

    split with (fun a ta => a == (proj1_sig ta)).
    intros a1 a2 ha [ta1 []] [ta2 []] hta;simpl in *.
    apply equiv_compat;assumption.
    intros a1 [ta1 []] h1 a2 [ta2 []] h2;simpl in *.
    eapply proj2,equiv_compat;assumption.
    intros a1 a2 ha;simpl in *.
    split with (@exist _ _ a1 I);split with  (@exist _ _ a2 I).
    compute.
    intuition.
  Qed.

  (** The trivially empty subset contains 0 element... *)
  Lemma Empty_subset_Num_Card:
    Num_Card Same (sig (fun _:A => False)) 0.
    split;split with (fun _ _ => False).
    intros [a []].
    intros [a []].
    intros [a []].
    intros [n h];elimtype False;lia.
    intros [n h];elimtype False;lia.
    intros [n h];elimtype False;lia.
  Qed.

  Variables (PA1: A -> Prop) (PA1_compat: Proper fequiv PA1).
  Variables (PA2: A -> Prop) (PA2_compat: Proper fequiv PA2).

  (** SA1 stands for: { x | PA1 x } *)
  Notation SA1 := (sig PA1).
  (** SA2 stands for: { x | PA2 x } *)
  Notation SA2 := (sig PA2).

  (** relation RA restricted between SA1 and SA2 *)
  Notation RRA  := (fun (a1: SA1) (a2: SA2) =>
                      (proj1_sig a1) == (proj1_sig a2)).

  (** If for all RA-related elements, x, y,
   we have (PA1 x) implies (PA2 y), then
   the subset built from PA1 (RSA1) is
   smaller than the one from PA2 (RSA2). *)
  Lemma Inj_impl: (equiv ==> impl)%signature PA1 PA2 -> Card Smaller SA1 SA2.
    intro h;split with RRA.
    intros [a11 h11] [a12 h12] h1 [a21 h21] [a22 h22] h2;compute in h1,h2;simpl.
    rewrite h1,h2;reflexivity.
    intros [a11 h11] [a21 h21] h1 [a12 h12] [a22 h22] h2;simpl in *;compute.
    rewrite h1,h2.
    trivial.
    intros [a11 h11] [a12 h12] h1.
    split with  (@exist _ PA2 _ (h _ _ (Equivalence_Reflexive _) h11)).
    split with (@exist _ PA2 _ (h _ _ (Equivalence_Reflexive _) h12)).
    repeat split;reflexivity || assumption.
  Qed.
    
  (** same with Larger *)
  Lemma Inj_impl2: (equiv ==> flip impl)%signature PA1 PA2 -> Card Larger SA1 SA2.
    intro h;split with (Basics.flip RRA).
    intros [a11 h11] [a12 h12] h1 [a21 h21] [a22 h22] h2;compute in h1,h2;unfold flip;simpl.
    rewrite h1,h2;reflexivity.
    intros [a11 h11] [a21 h21] h1 [a12 h12] [a22 h22] h2;simpl in *;compute.
    rewrite h1,h2.
    trivial.
    intros [a11 h11] [a12 h12] h1.
    split with (@exist _ PA1 _ (h _ _ (Equivalence_Reflexive _) h11)).
    split with (@exist _ PA1 _ (h _ _ (Equivalence_Reflexive _) h12)).
    compute in h1 |-*.
    repeat split;reflexivity || assumption.
  Qed.

  (** If for all RA-related elements, x, y, we have (PA1 x) <-> (PA2
   y), then the subset built from PA1 (RSA1) has same Num_Card than
   the one from PA2 (RSA2). *)
  Lemma Inj_iff: PA1 =~= PA2 -> Card Same SA1 SA2.
    intro h;split;[apply Inj_impl|apply Inj_impl2];compute in *;firstorder.
  Qed.

  Lemma Inj_iff2: PA1 =~= PA2 ->
                  forall n, Num_Card Same SA1 n -> Num_Card Same SA2 n.
  Proof.
    intros sameR n hn.
    assert (T := Inj_iff sameR).
    unfold Num_Card in *.
    apply Card_trans with SA1 _.
    apply Card_Same_sym; apply T.
    apply hn.
  Qed.
    
  (** n1 and n2 are Num_Card of RSA1 and RSA2 *)
  Variable (n1: nat) (hn1: Num_Card Same SA1 n1).
  Variable (n2: nat) (hn2: Num_Card Same SA2 n2).

  (** If PA1 implies PA2 through RA-related elements, then n1 is
  smaller than or equal n2 *)
  Lemma Subset_Num_Card_impl_le: (equiv ==> impl)%signature PA1 PA2 -> n1 <= n2.
  Proof.
    intro h;apply Card_smaller_le_iff.
    apply Card_trans with SA1 _.
    apply Card_weak;apply Card_sym;auto.
    apply Card_trans with SA2 _.
    apply Inj_impl;assumption.
    apply Card_weak;apply hn2.
  Qed.

  (** same for greater *)
  Lemma Subset_Num_Card_flipimpl_ge:
    (equiv ==> flip impl)%signature PA1 PA2 -> n1 >= n2.
  Proof.
    intro h;apply Card_larger_ge_iff.
    apply Card_trans with SA1 _.
    apply Card_weak;apply Card_sym;auto.
    apply Card_trans with SA2 _.
    simpl;apply Inj_impl2;assumption.
    apply Card_weak;apply hn2.
  Qed.

  (** same for equality *)
  Lemma Subset_Num_Card_iff_eq: PA1 =~= PA2 -> n1 = n2.
  Proof.
    intro h;apply le_antisym.
    apply Subset_Num_Card_impl_le.
    compute in h|-*.
    intros x y hh;apply h;auto.
    apply Subset_Num_Card_flipimpl_ge.
    compute in h|-*.
    intros x y hh;apply h;auto.
  Qed.
        
  (** A non-empty setoid contains at least 1
   element... *)
  Lemma Nonempty_Num_Card: forall a, Proper equiv a -> Num_Card Larger A 1.
  Proof.
    intros a a_compat.
    split with (fun _ x => x == a).
    intros _ _ _ a1 a2 ha.
    rewrite ha.
    reflexivity.
    intros [[ |x1] hx1] _ _ [[|x2] hx2] _ _ _;compute;lia.
    intros _ _ _;do 2 split with a;repeat (exact a_compat || split).
  Qed.

  (** A singleton contains exactly one element. *)
  Lemma Singleton_Num_Card:
    forall a1, PA1 a1 -> (forall a2, PA1 a2 -> a1 == a2) -> Num_Card Same SA1 1.
  Proof.
    intros a ha1 ha2.
    split.
    split with (fun x _ => (proj1_sig x) == a).
    intros a1 a2 ha _ _ _.
    apply equiv_compat;reflexivity||assumption.
    intros [y1 hy1] [[ |x1] hx1] _ [y2 hy2] [[ |x2] hx2] _;compute;try lia.
    intros _;transitivity a;[symmetry |];apply ha2;auto.
    intros y1 y2 hy;do 2 split with (ltSN_max  0).
    split.
    symmetry.
    apply ha2.
    apply (proj2_sig y1).
    split.
    symmetry.
    apply ha2.
    apply (proj2_sig y2).
    reflexivity.

    split with (fun _ x => (proj1_sig x) == a).
    intros _ _ _ a1 a2 ha.
    apply equiv_compat;reflexivity||assumption.
    intros [[ |x1] hx1] [y1 hy1] _ [[ |x2] hx2] [y2 hy2];compute;intros;try lia.
    assert (a_compat:Proper equiv a).
    apply ha2;apply ha1.
    intros _ _ _;do 2 split with (@exist _ PA1 a ha1);repeat (exact a_compat || split).
  Qed.

  (** A singleton contains exactly one element (converse). *)
  Lemma Singleton_Num_Card_unique:
    Num_Card Smaller SA1 1 -> forall a1 a2, PA1 a1 -> PA1 a2 -> a1 == a2.
  Proof.
    intros c1 a1 a2 p1 p2.
    assert (exists _: ltN 1, True).
    { assert (dum: 0 < 1) by lia.
      exists (@exist _ (fun n => n < 1) _ dum); tauto. }
    destruct H as [zero _].
    assert (all_zero: forall b, b == zero).
    { destruct b as [b hb]; destruct zero as [zero hz].
      simpl; unfold eqSig; simpl.
      destruct b as [| b]; [|lia].
      destruct zero as [| b]; [|lia].
      lia. }
      
    destruct c1 as [f1 pf1 t1u t1l].
    assert (x2 := t1l (exist p1) (exist p1) (Equivalence_Reflexive _)). 
    destruct x2 as [b1 [_ [hf1 _]]].
    assert (x3 := t1l (exist p2) (exist p2) (Equivalence_Reflexive _)). 
    destruct x3 as [b2 [_ [hf2 _]]].

    rewrite all_zero in hf1, hf2.
    assert (x1 := t1u _ _ hf1 _ _ hf2 (Equivalence_Reflexive _)).
    apply x1.
  Qed.
  
  (** Union of disjoints sets *)
  Section Disjoint.

    Hypothesis Disjoint: forall x, PA1 x -> PA2 x -> False.

    Notation PA12 := (fun x => PA1 x \/ PA2 x).
    Notation SA12 := {x: A | PA12 x}.

    Global Instance Union_compat: Proper fequiv PA12.
    Proof.
      intros a1 a2 h12.
      apply Morphisms_Prop.or_iff_morphism.
      apply PA1_compat;assumption.
      apply PA2_compat;assumption.
    Qed.

    Lemma Disjoint_union_as_many_coproduct: Card Same SA12 (SA1 +  SA2)%type.
    Proof.
      split.
      split with (fun a a12 => match a12 with
                                 inl a1 => (proj1_sig a) == (proj1_sig a1)
                               | inr a2 => (proj1_sig a) == (proj1_sig a2) end).

      intros [a ha] [a' ha'] ea [[a1 ha1]|[a2 ha2]]
             [[a1' ha1']|[a2' ha2']] ea12;
        compute in ea,ea12;try elim ea12;simpl.
      rewrite ea,ea12.
      reflexivity.
      rewrite ea,ea12.
      reflexivity.
      intros [a ha] [[a1 ha1]|[a2 ha2]] ea [a' ha'] [[a1' ha1']|[a2' ha2']] ea';
        compute in ea,ea';try (elim ea||elim ea');simpl;unfold eqSig;simpl;
          rewrite ea,ea';
          try reflexivity;try (intros[]);intro h;
            (rewrite h in ha1||rewrite h in ha2);
            eauto.

      intros [a [ha1|ha2]] [a' [ha1'|ha2']] ea;compute in ea.
      split with (inl (exist ha1)).
      split with (inl (exist ha1')).
      compute.
      compute in ea.
      intuition.

      elim (Disjoint a').
      rewrite ea in ha1;assumption.
      assumption.
      elim (Disjoint a').
      assumption.
      rewrite ea in ha2;assumption.

      split with (inr (exist ha2)).
      split with (inr (exist ha2')).
      compute.
      compute in ea.
      intuition.

      split with (fun a12 a => match a12 with
                                 inl a1 => (proj1_sig a) == (proj1_sig a1)
                               | inr a2 => (proj1_sig a) == (proj1_sig a2) end).

      intros [[a1 ha1]|[a2 ha2]]
             [[a1' ha1']|[a2' ha2']] ea12 [a ha] [a' ha'] ea;
        compute in ea,ea12;try elim ea12;simpl.
      rewrite ea,ea12.
      reflexivity.
      rewrite ea,ea12.
      reflexivity.
      intros [[a1 ha1]|[a2 ha2]]
             [a ha] ea [[a1' ha1']|[a2' ha2']] [a' ha'] ea';
        compute in ea,ea';try (elim ea||elim ea');simpl;
          unfold eqSig;simpl;rewrite ea,ea';
            try reflexivity;try (intros[]);intro h;
              (rewrite h in ha1||rewrite h in ha2);
            eauto.

      intros [[a1 ha1]|[a2 ha2]] [[a1' ha1']|[a2' ha2']] ea12;
        compute in ea12;try elim ea12;simpl.
      split with (@exist _ _ a1  (or_introl ha1)).
      split with (@exist _ _ a1' (or_introl ha1')).
      compute;intuition.
      split with (@exist _ _ a2  (or_intror ha2)).
      split with (@exist _ _ a2' (or_intror ha2')).
      compute;intuition.
    Qed.

    Lemma Disjoint_union_Num_Card: Num_Card Same SA12 (n1 + n2).
    Proof.
      apply Card_trans with (SA1+SA2)%type _;
        try repeat first  [apply eqSig_equiv|apply coprod_rel_equiv]; auto.
      apply Disjoint_union_as_many_coproduct.
      apply Card_trans with (ltN n1+ltN n2)%type _;
        try repeat first  [apply eqSig_equiv|apply coprod_rel_equiv];
        auto using eq_equivalence.
      apply Card_coprod;
        try repeat first  [apply eqSig_equiv|apply coprod_rel_equiv|assumption].
      apply Card_as_many_coprod.
    Qed.

  End Disjoint.

End Subsets.


Section Classical_subsets.

  Context {A} `{SA:Setoid A}.

  Variable A_classical: forall a1 a2:A, a1 == a2 \/ ~ a1 == a2.

  Lemma InA_classical : forall a l, InA equiv a l \/ ~InA equiv a l.
  Proof.
    intro a.
    induction l as [|x q hq].
    right;intro h;inversion h.
    destruct hq as [YES| NO].
    left;right;assumption.
    destruct (A_classical a x) as [YES|NONO].
    left;left;assumption.
    right;intro h.
    inversion h;subst.
    apply NONO;assumption.
    apply NO;assumption.
  Qed.

  (** Predicate P defines the subset *)
  Section With_P.

    Variable (P: A -> Prop)
             (P_compat: Proper fequiv P)
             (P_classical: forall (a: A), P a \/ ~P a).

    Notation PL := (fun l a => P a /\ InA equiv a l).

    (** relation RA restricted to SA1 *)
    Notation LSA l := (sig (PL l)).

    Instance PL_compat: Proper (equivlistA equiv ==> fequiv) PL.
    Proof.
      intros l1 l2 hl a1 a2 ha.
      apply and_compat.
      apply P_compat;assumption.
      apply InA_compat;try assumption.
      apply setoid_equiv.
    Qed.

    Lemma listP_Num_Card: forall l:list A, exists n,  Num_Card Same (LSA l) n.
    Proof.
      induction l as [|a q hq].
      split with 0.

      eapply Card_trans with (sig (fun _ => False)) _.
      2:apply Empty_subset_Num_Card.
      apply Inj_iff.
      compute;intuition.
      compute;intuition.
      inversion H2.
      destruct hq as [n hn].
      destruct (InA_classical a q) as [OLD|NEW];[split with n|].
      {
        eapply Card_trans with (LSA q) _.
        2:apply hn.
        apply Inj_iff.
        apply PL_compat;reflexivity.
        apply PL_compat.
        intro x;split;rewrite InA_cons;try tauto.
        intros [h1|h2];try rewrite h1;tauto.
      }
      {
        destruct (P_classical a) as [YES|NO];[split with (n+1)|split with n].
        {
          unfold Num_Card.
          eapply Card_trans with (sig (fun x => PL q x \/ a==x))%type _.
          2:apply Disjoint_union_Num_Card.
          4: apply hn.
          apply Inj_iff.
          intros a1 a2 ha.
          apply or_compat.
          apply PL_compat;assumption||reflexivity.
          apply equiv_compat;assumption||reflexivity.
          intros a1 a2 ha.
          rewrite InA_cons.
          split;rewrite ha in *;intuition.
          rewrite H0 in YES;assumption.
          apply PL_compat;reflexivity.
          apply equiv_compat;reflexivity.

          apply Singleton_Num_Card with a;try assumption.
          reflexivity.
          firstorder.
          intros x [_ hx] hax.
          rewrite <- hax in hx.
          auto.
        }
        {
          eapply Card_trans with (LSA q) _.
          2:apply hn.
          apply Inj_iff;try assumption.
          apply PL_compat;reflexivity.

          intros a1 a2 ha.
          rewrite InA_cons.
          split;rewrite ha in *;intuition.
          contradict NO.
          rewrite H in H0.
          assumption.
        }
      }
    Qed.

    (** The set itself is one of its subset, but of same Cardinality *)

    Lemma Classical_subset_Num_Card:
      forall lA: list A, (forall a:A, P a -> InA equiv a lA) ->
                         exists cardP, Num_Card Same (sig P) cardP.
    Proof.
      intros lA A_enum.
      destruct (listP_Num_Card lA) as [cardP HcardP].
      split with cardP.
      eapply Card_trans with (LSA lA) _.
      2:apply HcardP.
      apply Inj_iff;try assumption.
      intros a1 a2 ha.
      apply PL_compat;assumption||reflexivity.
      intros a1 a2 ha.
      rewrite ha.
      hnf; intuition.
    Qed.

  End With_P.

  Lemma Classical_set_Num_Card:
    forall lA: list A, (forall a:A, InA equiv a lA) ->
                       exists cardL, Num_Card Same A cardL.
  Proof.
    intros lA inA.
    destruct (Classical_subset_Num_Card (fun _ => True)) with (lA := lA) as [n Hn].
    hnf;reflexivity.
    left;tauto.
    intuition.
    exists n.
    unfold Num_Card in *.
    eapply Card_trans with (sig (fun _ => True)) _.
    2: apply Hn.
    apply Card_sym; unfold sym_Card_Prop; apply Trivial_subset_Num_Card.
  Qed.

End Classical_subsets.

Section Decidable_subsets.

  Context {A} `{SA:Setoid A}.

  Variable (DA: forall a b, {a == b} + {~a == b}).

  (** Predicate P defines the subset *)
  Section With_P.

    Variable (P: A -> Prop)
             (P_compat: Proper fequiv P)
             (P_dec: forall (a: A), {P a} + {~P a}).

    Notation PL := (fun l a => P a /\ InA equiv a l).

    (** relation RA restricted to SA1 *)
    Notation LSA l := (sig (PL l)).

    Instance PL_dec_compat: Proper (equivlistA equiv ==> fequiv) PL.
    Proof.
      intros l1 l2 hl a1 a2 ha.
      apply and_compat.
      apply P_compat;assumption.
      apply InA_compat;try assumption.
      apply setoid_equiv.
    Qed.

    Lemma listP_Num_CardT:
      forall l:list A, { n |  Num_Card Same (LSA l) n}.
    Proof.
      induction l as [|a q hq].
      split with 0.

      eapply Card_trans with (sig (fun _ => False)) _.
      2:apply Empty_subset_Num_Card.
      apply Inj_iff.
      compute;intuition.
      compute;intuition.
      inversion H2.
      destruct hq as [n hn].
      destruct (InA_dec DA a q) as [OLD|NEW];[split with n|].
      {
        eapply Card_trans with (LSA q) _.
        2:apply hn.
        apply Inj_iff.
        apply PL_dec_compat;reflexivity.
        apply PL_dec_compat.
        intro x;split;rewrite InA_cons;try tauto.
        intros [h1|h2];try rewrite h1;tauto.
      }
      {
        destruct (P_dec a) as [YES|NO];[split with (n+1)|split with n].
        {
          unfold Num_Card.
          eapply Card_trans with (sig (fun x => PL q x \/ a==x))%type _.
          2:apply Disjoint_union_Num_Card.
          4: apply hn.
          apply Inj_iff.
          intros a1 a2 ha.
          apply or_compat.
          apply PL_compat;assumption||reflexivity.
          apply equiv_compat;assumption||reflexivity.
          intros a1 a2 ha.
          rewrite InA_cons.
          split;rewrite ha in *;intuition.
          rewrite H0 in YES;assumption.
          apply PL_dec_compat;reflexivity.
          apply equiv_compat;reflexivity.

          apply Singleton_Num_Card with a;try assumption.
          reflexivity.
          firstorder.
          intros x [_ hx] hax.
          rewrite <- hax in hx.
          auto.
        }
        {
          eapply Card_trans with (LSA q) _.
          2:apply hn.
          apply Inj_iff;try assumption.
          apply PL_dec_compat;reflexivity.

          intros a1 a2 ha.
          rewrite InA_cons.
          split;rewrite ha in *;intuition.
          contradict NO.
          rewrite H in H0.
          assumption.
        }
      }
    Qed.

    (** The set itself is one of its subset, but of same Cardinality *)

    (*listP_Num_CardT*)
    Lemma Classical_subset_Num_CardT:
      forall lA: list A, (forall a:A, P a -> InA equiv a lA) ->
                         { cardP | Num_Card Same (sig P) cardP}.
    Proof.
      intros lA A_enum.
      destruct (listP_Num_CardT lA) as [cardP HcardP].
      split with cardP.
      eapply Card_trans with (LSA lA) _.
      2:apply HcardP.
      apply Inj_iff;try assumption.
      intros a1 a2 ha.
      apply PL_compat;assumption||reflexivity.
      intros a1 a2 ha.
      rewrite ha.
      hnf; intuition.
    Qed.

  End With_P.

  Lemma Classical_set_Num_CardT:
    forall lA: list A, (forall a:A, InA equiv a lA) ->
                       { cardL | Num_Card Same A cardL}.
  Proof.
    intros lA inA.
    destruct (Classical_subset_Num_CardT (fun _ => True)) with (lA := lA) as [n Hn].
    hnf;reflexivity.
    left;tauto.
    intuition.
    exists n.
    unfold Num_Card in *.
    eapply Card_trans with (sig (fun _ => True)) _.
    2: apply Hn.
    apply Card_sym; unfold sym_Card_Prop; apply Trivial_subset_Num_Card.
  Qed.

End Decidable_subsets.

Require Import setoid_ring.Ring.

Section Triangles_and_Trapezoids.

  Variable a b:nat.

  Hypothesis ab_interval: a<=b.

  Notation trapezoid a b cpl :=
    (a<= fst cpl /\ fst cpl <b /\ snd cpl < fst cpl).
  Definition Trapezoid:= {cpl| trapezoid a b cpl}.

  Let count:=((b-a)*(a+(b-1))/2).

  Lemma count_Trapezoid:  Num_Card Same Trapezoid count.
  Proof.
    pose (k:=b-a).
    assert (Hb:b=k+a).
    unfold k;lia.
    unfold count,Trapezoid;fold k;clear count.
    rewrite <- pred_of_minus.
    clearbody k.
    subst.
    clear ab_interval.
    induction k.
    {
      simpl mult;simpl div.
      apply Inj_iff2 with (fun cpl => False);auto.
      * intros x y exy;rewrite exy; subst;reflexivity.
      * intros x y exy;inversion exy;subst;hnf;try lia.
      * apply Empty_subset_Num_Card.
    }
    {
      replace (S k * (a + pred (S k + a)) / 2)
        with ((k+a)+(k * (a + pred (k + a)))/2).
      shelve.
      rewrite plus_comm.
      repeat rewrite <- PeanoNat.Nat.div_add;try lia.
      apply f_equal2;try lia.
      destruct k;simpl;ring.
      Unshelve.
      pose (Pk cpl:= fst cpl = k+a /\ snd cpl < fst cpl).
      {
        eapply Inj_iff2 with (fun cpl => Pk cpl \/ trapezoid a (k + a) cpl).
        intros x y exy; rewrite exy;subst;reflexivity.
        intros x y [efst esnd];hnf in *;unfold Pk;intuition lia.
        {
          apply Disjoint_union_Num_Card.
          intros x y [efst esnd]; unfold Pk;hnf in *;intuition.
          intros x y [efst esnd]; unfold Pk;hnf in *;intuition.
          3:intros [x y];unfold Pk;lia.
          2:apply IHk.
          split.
          {
            split with (fun cpl n => snd (proj1_sig cpl) = proj1_sig n).
            * intros [[x1 y1] hc1] [[x2 y2] hc2] [ex ey] [n1 hn1] [n2 hn2] en;
                hnf in *;simpl in *;subst;auto.
            * intros [[x1 y1] hc1] [n1 hn1] h1 [[x2 y2] hc2] [n2 hn2] h2 h;
                hnf in *;destruct hc1;destruct hc2;simpl in * ;subst;auto.
            * intros [[x1 y1] [hc1 hcc1]] [[x2 y2] [hc2 hcc2]] [ex ey];
                hnf in *;simpl in *;subst;auto.
              split with (@exist _ _ y2 hcc1);split with (@exist _ _ y2 hcc1);
                simpl;intuition.
          }
          {
            split with (fun n cpl => snd (proj1_sig cpl) = proj1_sig n).
            intros [n1 hn1] [n2 hn2] en [[x1 y1] hc1] [[x2 y2] hc2] [ex ey];
              hnf in *;simpl in *;subst;auto.
            intros  [n1 hn1] [[x1 y1] hc1] h1 [n2 hn2] [[x2 y2] hc2] h2 h;
              hnf in *;destruct hc1;destruct hc2;simpl in * ;inversion h;
                subst;auto.
            intros [n1 hn1] [n2 hn2] en;hnf in *;simpl in *;subst;auto.
            split with (@exist _ Pk (k+a,n2) (conj (refl_equal _) hn2));
              split with (@exist _ Pk (k+a,n2) (conj (refl_equal _) hn2)).
            simpl;intuition.
          }
        }
      }
    }
  Qed.

End  Triangles_and_Trapezoids.

Unset Implicit Arguments.
Close Scope nat_scope.