(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** **** Global imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import RelationClasses.
From Coq Require Import Lia.
From CoLoR Require Import ListUtil.
Global Existing Instance relation_equivalence_equivalence. 

(** **** Local imports **)
From Padec Require Import SetoidUtil.
From Padec Require Import OptionUtil.

(* end show *)
Open Scope signature_scope.
Set Implicit Arguments.

(** * Setoid and Partial Setoid of Lists *)

(**
  Two equalities exist on lists:
  - eqlistA: point-wise equality
  - equivlistA: two lists are equal iff they contain the same 
                elements, up to equivalence on those elements
  Both equalities preserve the property of their base-equality (i.e.
  equivalence or PER).

  We chose to define eqlistA as the base equality on list and to
  instantiate a setoid/partial setoid on it. Namely, in further 
  developments, [equiv] or [pequiv] on list refers to the point-wise
  equality. [equivlistA] has to be called explicitely.
 *)
Global Instance eqlistA_per {A} {eqA} {eqA_per: @PER A eqA}:
  PER (eqlistA eqA).
Proof.
  split.
  intros l1 l2 el.
  induction el as [|a1 a2 l1 l2 ea el hrec].
  left.
  right; [symmetry|]; assumption.
  intros l1 l2 l3 el; revert l3.
  induction el as [|a1 a2 l1 l2 ea el hrec].
  trivial.
  intros [|a3 l3] h; inversion_clear h.
  right; [transitivity a2|apply hrec]; assumption.
Qed.

Global Instance List_Setoid {A} `{Setoid A}: Setoid (list A) :=
  {| equiv := eqlistA equiv |}.

Global Instance List_Partial_Setoid {A} `{PartialSetoid A}: PartialSetoid (list A) :=
  {| pequiv := eqlistA pequiv |}.

Global Instance eqlistA_eq_is_eq (A: Type):
  subrelation (eqlistA (@eq A)) (@eq (list A)).
Proof.
  intros x y h.
  induction h as [| x y qx qy hxy hq];
    congruence.
Qed.

(** * InA                                                          *)
Section INA.

  (** InA_compat provides compatibility of InA wrt equivlistA,
   InA_compat_eqlistA, below, provides compatibility wrt eqlistA *)
  Global Instance  InA_compat_eqlistA {A} `{Setoid A}:
    Proper (equiv ==> eqlistA equiv ==> iff) (InA equiv).
  Proof.
    intros a b ab l ll el.
    apply eqlistA_equivlistA in el.
    2: apply H.
    apply InA_compat; auto.
    apply H.
  Qed.

  Lemma InA_app_app A (eqA:relation A): forall x l,
      Proper (eqlistA eqA) l -> InA eqA x l ->
      exists a b, eqlistA eqA (a ++ (x :: nil) ++ b) l.
  Proof.
    intros x l pl ina.
    induction pl as [ | a1 a2 l1 l2 ea el Hrec];simpl in *.
    {
      rewrite InA_nil in ina.
      elim ina.
    }
    {
      rewrite InA_cons in ina.
      destruct ina as [inh | int].
      {
        exists nil.
        exists l1.
        rewrite app_nil_l.
        right;auto.
      }
      {
        specialize (Hrec int).
        destruct Hrec as [a [b h]].
        exists (a1 :: a),b. 
        apply eqlistA_cons; auto.
      }
    }
  Qed.
  
  Lemma InA_app (A:Type) {SA:Setoid A}: forall (x:A) a b,
      InA equiv x (a ++ b) <-> ((InA equiv x a) \/ (InA equiv x b)).
  Proof.
    intros x a b.
    revert b.
    induction a; intros b.
    simpl.
    rewrite InA_nil.
    assert (False_or: forall P Q, ~ P -> (P \/ Q <-> Q)).
    tauto.
    rewrite False_or;auto.
    reflexivity.
    auto.
    simpl.
    do 2 rewrite InA_cons.
    rewrite or_assoc.
    rewrite IHa.
    reflexivity.
  Qed.

  Lemma list_app_preserve {A} `{Setoid A}:
    forall (x: A) a1 b1 a2 b2,
      ~ InA equiv x a1 -> ~ InA equiv x b1 ->
      a1 ++ x :: b1 == a2 ++ x :: b2 -> a1 == a2 /\ b1 == b2.
  Proof.
    intros x a1 b1 a2 b2 in1 in2 hl.
    cut (a1 == a2).
    { intros ok; rewrite ok in hl.
      split; auto.
      clear ok a1 in1.
      induction a2 as [| y a2 ih].
      simpl in hl.
      inversion hl; auto.
      apply ih; simpl in hl; inversion hl; auto.
    }
    revert in1 in2 hl; revert a1 b1 b2.
    induction a2 as [| y a2 ih]; intros a1 b1 b2 in1 in2 hl. 
    { destruct a1 as [| z a1].
      reflexivity.
      simpl in hl; inversion_clear hl.
      contradict in1; rewrite InA_cons; left; symmetry; assumption.
    }
    { destruct a1 as [| z a1].
      simpl in hl.
      inversion_clear hl.
      contradict in2; rewrite H1.
      apply InA_app_iff; right; apply InA_cons; left; reflexivity.
      right.
      inversion hl; assumption.
      eapply ih.
      3:{ simpl in hl; inversion_clear hl.
          apply H1. }
      contradict in1; apply InA_cons; right; assumption.
      assumption.
    }
  Qed.
  
End INA.

(** * NoDupA                                                       *)
Section NoDupA.

  Lemma NoDupA_nil_iff: forall (A: Type) (eqA: A -> A -> Prop),
    NoDupA eqA nil <-> True.
  Proof.
    intuition.
  Qed.

  Lemma NoDupA_cons_iff: forall (A: Type) (eqA: A -> A -> Prop) x q,
      NoDupA eqA (x::q) <-> ~InA eqA x q /\ NoDupA eqA q.
  Proof.
    split.
    intro h;inversion h;subst;auto.
    intros [h1 h2];apply NoDupA_cons;auto.
  Qed.

  Global Instance NoDupA_compat {A} `{Setoid A}:
    Proper (eqlistA equiv ==> iff) (NoDupA equiv).
  Proof.
    compute;intros.
    change (NoDupA equiv x <-> NoDupA equiv y).
    induction H0.
    reflexivity.
    do 2 rewrite NoDupA_cons_iff.
    apply and_compat.
    simpl.
    apply not_compat.
    apply InA_compat_eqlistA;auto.
    auto.
  Qed.

  Lemma NoDupA_app_alt {A} `{Setoid A}:
    forall l1 l2: list A, NoDupA equiv (l1++l2) ->
                          NoDupA equiv l1 /\ NoDupA equiv l2.
  Proof.
    induction l1 as [| x l1 ih].
    { intros l2 h.
      split.
      apply NoDupA_nil.
      simpl in h; auto.
    }
    { intros l2 h.
      assert (x::l1 ++ l2 == x :: (l1 ++ l2)) by (simpl; reflexivity).
      rewrite H0 in h; clear H0.
      rewrite NoDupA_cons_iff in h.
      destruct h as [h1 h2].
      specialize (ih _ h2).
      split; [|apply ih].
      rewrite NoDupA_cons_iff.
      split; [|apply ih].
      contradict h1.
      rewrite InA_app_iff; left; auto.
    }
  Qed.

End NoDupA.

(** ** Tools about ForallA                                         *)
Section ForallA.

  Context {A: Type} (eqA: relation A).
  Context {eqA_refl: Reflexive eqA}.

  Section FF.

    Context (P: A -> Prop) {P_compat: Proper (eqA ==> iff) P}.

    Lemma Forall_forallA:
      forall (l : list A),
        Forall P l <-> (forall (x : A), InA eqA x l -> P x).
    Proof.
      intros; rewrite Forall_forall.
      setoid_rewrite InA_alt.
      firstorder.
    Qed.

    Definition Forall_hd := Forall_inv.

    Lemma Forall_tl:
      forall a l, Forall P (a :: l) -> Forall P l.
    Proof.
      intros a l H.
      inversion H; congruence.
    Qed.

  End FF.

  Global Instance Forall_compat :
    Proper ((eqA ==> iff) ==> (eqlistA eqA) ==> iff) (@Forall A).
  Proof.
    intros p1 p2 ep l1 l2 el.
    assert (h1:Proper (eqA ==> iff) p1).
    intros x y e.
    etransitivity.
    apply ep.
    apply e.
    symmetry.
    apply ep.
    apply eqA_refl.
    assert (h2: Proper (eqA ==> iff) p2).
    intros x y e.
    etransitivity.
    symmetry.
    apply ep.
    apply eqA_refl.
    apply ep.
    apply e.
    induction el as [|x1 x2 l1 l2 eh el Hrec]; simpl.
    split; left.
    split; intro hh; inversion hh as [| x l hx hl]; right.
    rewrite <- (ep _ _ eh); assumption.
    rewrite <- Hrec; assumption.
    rewrite -> (ep _ _ eh); assumption.
    rewrite -> Hrec; assumption.
  Qed.

  Global Instance Forall_compat_equiv :
    Proper ((eqA ==> iff) ==> (equivlistA eqA) ==> iff) (@Forall A).
  Proof.
    intros p1 p2 ep l1 l2 el.
    assert (h1: Proper (eqA ==> iff) p1).
    intros x y e.
    etransitivity.
    apply ep.
    apply e.
    symmetry.
    apply ep.
    apply eqA_refl.
    assert (h2: Proper (eqA ==> iff) p2).
    intros x y e.
    etransitivity.
    symmetry.
    apply ep.
    apply eqA_refl.
    apply ep.
    apply e.
    rewrite (Forall_forallA); [|assumption].
    rewrite (Forall_forallA); [|assumption].
    apply all_compat; intro a.
    apply impl_compat. 
    apply el.
    apply ep.
    reflexivity.
  Qed.

  Lemma forall_inv_2:
    forall (P: A -> Prop) (a: A) (l: list A),
      Forall P (a :: l) -> Forall P l.
  Proof.
    intros P a l h.
    inversion h.
    subst.
    assumption.
  Qed.

  Lemma Forall_head_and_tail:
    forall (h: A) t p, Forall p (h :: t) <-> p h /\ Forall p t.
  Proof.
    split.
    intro fa.
    split.
    apply (Forall_inv fa).
    apply (forall_inv_2 fa).
    intros [a fa].
    apply (Forall_cons h a fa).
  Qed.

  Lemma forall_app:
    forall (p: A -> Prop) l1 l2,
      Forall p (l1 ++ l2) <-> (Forall p l1) /\ (Forall p l2).
  Proof.
    intros p l1 l2.
    do 3 rewrite Forall_forall.
    split.
    intros fa.
    split.
    intros x xin.
    apply fa.
    apply in_or_app.
    left.
    assumption.
    intros x xin.
    apply fa.
    apply in_or_app.
    right.
    assumption.
    intros [fa1 fa2] x xin.
    rewrite in_app_iff in xin.
    destruct xin as [xin1 | xin2].
    apply fa1.
    assumption.
    apply fa2.
    assumption.
  Qed.

End ForallA.

(** ** Tools about map compatibility                               *)
Section map_compat.

  Context {A: Type} (eqA: relation A).
  Context {B: Type} (eqB: relation B).

  Global Instance map_compat:
    Proper ((eqA ==> eqB) ==> eqlistA eqA ==> eqlistA eqB)
           (@List.map A B).
  Proof.
    intros f1 f2 ef l1 l2 el.
    induction el as [|a1 a2 l1 l2 ea el hrec]; simpl.
    auto.
    right.
    apply ef; apply ea.
    apply hrec.
  Qed.

  Lemma map_eq_locality :
    forall (eqA_refl: Reflexive eqA)
           (eqB_PER: PER eqB)
           (f1: A -> B) (p1: Proper (eqA ==> eqB) f1)
           (f2: A -> B) (p2: Proper (eqA ==> eqB) f2)
           (l1 l2: list A) (el: eqlistA eqA l1 l2),
      (forall a, InA eqA a l1 -> eqB (f1 a) (f2 a)) ->
      eqlistA eqB (map f1 l1) (map f2 l2).
  Proof.
    intros eqA_refl eqB_PER f1 p1 f2 p2.
    induction l1 as [| a1 l1 ih]; destruct l2 as [| a2 l2]; simpl;
      intros; auto.
    inversion el.
    inversion el.
    inversion el; subst.
    right.
    rewrite <- H3; auto. 
    apply ih; auto.
  Qed.
  
End map_compat.

Section map_compat_equivlist.

  Context {A: Type} `{SA:Setoid A}.
  Context {B: Type} `{SB:Setoid B}.

  Global Instance map_compat_equiv :
    Proper (fequiv ==> equivlistA equiv ==> equivlistA equiv)
           (@List.map A B).
  Proof.
    intros f1 f2 ef l1 l2 el.
    unfold equivlistA in *.
    intro b.
    split.
    rewrite InA_alt in *.
    intros [b1 [eb1 h1]].
    rewrite in_map_iff in h1.
    destruct h1 as [a1 [ea1 h1]].
    rewrite InA_alt.
    specialize (el a1).
    destruct el as [el _].
    rewrite InA_alt in el.
    assert (h2: InA equiv a1 l2).
    apply el.
    exists a1.
    split; [reflexivity|assumption].
    rewrite InA_alt in h2.
    destruct h2 as [a2 [ea h2]].
    exists (f2 a2).
    split.
    rewrite <- ea1 in eb1.
    rewrite eb1.
    apply ef,ea.
    rewrite in_map_iff.
    exists a2.
    split; auto.

    rewrite InA_alt in *.
    intros [b2 [eb2 h2]].
    rewrite in_map_iff in h2.
    destruct h2 as [a2 [ea2 h2]].
    rewrite InA_alt.
    specialize (el a2).
    destruct el as [_ el].
    rewrite InA_alt in el.
    assert (h1: InA equiv a2 l1).
    apply el.
    exists a2.
    split; [reflexivity|trivial].
    rewrite InA_alt in h1.
    destruct h1 as [a1 [ea h1]].
    exists (f1 a1).
    split.
    rewrite <- ea2 in eb2.
    rewrite eb2.
    symmetry.
    apply ef.
    symmetry; apply ea.
    rewrite in_map_iff.
    exists a1.
    split; auto.
  Qed.

  Lemma InA_map_iff:
    forall (f: A -> B) {f_compat: Proper fequiv f}
           (l: list A) (b: B),
      InA equiv b (map f l) <->
      (exists a : A, (f a) == b /\ InA equiv a l).
  Proof.
    intros f hf l b.
    rewrite InA_alt.
    split.
    intros [b' [eb hb]].
    rewrite in_map_iff in hb.
    destruct hb as [a [e ha]].
    split with a.
    rewrite e.
    split.
    symmetry; assumption.
    apply In_InA; [apply SA|auto].
    intros [a [e ha]].
    rewrite InA_alt in ha.
    destruct ha as [a' [ea ha]].
    split with (f a').
    split.
    rewrite <- e.
    apply hf; assumption.
    rewrite in_map_iff.
    split with a'.
    auto.
  Qed.

End map_compat_equivlist.

(** * Existsb Forallb *)
Section Existsb_Forallb.

  Variable (A: Type).
  Variable (f: A -> bool)
           (notf: A -> bool)
           (H: forall a, f a = negb (notf a)).

  Lemma existsb_forallb: 
    forall l, existsb f l = negb (forallb notf l).
    induction l as [| a l IH].
    simpl; reflexivity.
    simpl.
    rewrite Bool.negb_andb.
    rewrite IH.
    rewrite (H a).
    reflexivity.
  Qed.

  Lemma forallb_existsb: 
    forall l, forallb f l = negb (existsb notf l).
    induction l as [| a l IH].
    simpl; reflexivity.
    simpl.
    rewrite Bool.negb_orb.
    rewrite IH.
    rewrite (H a).
    reflexivity.
  Qed.

End Existsb_Forallb.

(** ** Tools about filter *)
Section filter.

  Context {A: Type} `{SA:Setoid A}.

  Global Instance filter_compat :
    Proper (fequiv ==> eqlistA equiv ==> eqlistA equiv)
           (filter (A:=A)).
  Proof.
    intros f1 f2 ef l1 l2 el.
    induction el as  [|a1 a2 l1 l2 ea el]; simpl.
    left.
    rewrite (ef _ _ ea).
    destruct (f2 a2).
    right; auto.
    auto.
  Qed.

  Global Instance filter_compat_equiv :
    Proper (fequiv ==> equivlistA equiv ==> equivlistA equiv)
           (filter (A:=A)).
  Proof.
    intros f1 f2 ef l1 l2 el a.
    assert (pf1: Proper fequiv f1) by proper_.
    assert (pf2: Proper fequiv f2) by proper_.
    repeat rewrite (filter_InA _  _ _ ).
    unfold equivlistA in el.
    apply and_compat.
    apply el.
    rewrite (ef _ _ (Equivalence_Reflexive _)).
    reflexivity.
  Qed.

  Lemma filter_nil:
    forall (f: A -> bool) (l: list A), filter f l = nil
                                       <-> existsb f l = false.
  Proof.
    intros f l.
    split; induction l; simpl; try tauto; intros H.
    rewrite Bool.orb_false_iff.
    case (Bool.bool_dec (f a) true); intros F.
    rewrite F in H.
    symmetry in H; contradict H; apply nil_cons.
    split; rewrite Bool.not_true_iff_false in F.
    apply F.
    apply IHl.
    rewrite F in H; apply H.

    rewrite Bool.orb_false_iff in H.
    destruct H as [HH H].
    case (Bool.bool_dec (f a) true); intros F.
    contradict F; rewrite HH; apply Bool.diff_false_true.
    rewrite Bool.not_true_iff_false in F; rewrite F.
    apply IHl; apply H.
  Qed.

  Lemma filter_nil':
    forall (f: A -> bool) (l: list A),
      filter f l = nil <-> forallb (fun a => negb (f a)) l = true.
  Proof.
    intros f l; rewrite filter_nil.
    erewrite forallb_existsb.
    2: reflexivity.
    rewrite Bool.negb_true_iff; tauto.
  Qed.

  Lemma filter_eq_locality:
    forall (f1 f2: A -> bool) (l1 l2: list A),
      equivlistA equiv l1 l2 ->
      Proper fequiv f1 -> Proper fequiv f2 ->
      (forall a, InA equiv a l1 -> f1 a = f2 a) ->
      equivlistA equiv (filter f1 l1) (filter f2 l2).
  Proof.
    intros until l2; intros el p1 p2 h.
    intros x; split; intros hx.
    { rewrite filter_InA; auto.
      rewrite filter_InA in hx; auto; destruct hx as [hx hxx].
      split.
      rewrite <- el; auto.
      rewrite <- hxx; symmetry; apply h; auto.
    }
    { rewrite filter_InA; auto.
      rewrite filter_InA in hx; auto; destruct hx as [hx hxx].
      split.
      rewrite el; auto.
      rewrite <- hxx; apply h; auto.
      rewrite el; auto.
    }
  Qed.

  Lemma filter_InA_locality:
    forall (f: A -> bool),
    forall (l: list A),
      (forall x y, x == y -> InA equiv x l -> f x = f y) ->
      forall x,
        InA equiv x (filter f l) <-> InA equiv x l /\ f x = true.
  Proof.
    intros.
    induction l as [| a l ih]; simpl.
    split; rewrite InA_nil; tauto.
    split; rewrite InA_cons; intros h.
    { case_eq (f a); intros fa; rewrite fa in h.
      - rewrite InA_cons in h; destruct h as [h|h].
        + split; [left; auto|].
          rewrite <- fa; apply H; auto.
        + rewrite ih in h; auto.
          split; [right|]; apply h.
      - rewrite ih in h; auto.
        split; [right|]; apply h.
    }
    { case_eq (f a); intros fa.
      - rewrite InA_cons; destruct h as [[h|h] hh].
        + left; auto.
        + right; apply ih; auto.
      - apply ih; auto.
        destruct h as [[h|h] hh]; split; auto.
        contradict hh; rewrite Bool.not_true_iff_false, <- fa.
        apply H; auto.
    }
  Qed.

  Lemma filter_NoDupA:
    forall (f: A -> bool) (f_c: Proper pequiv f)
           (l: list A),
      NoDupA equiv l -> NoDupA equiv (filter f l).
  Proof.
    intros f f_c; induction l as [| x l ih]; simpl.
    { intros _.
      apply NoDupA_nil. }
    { intros h.
      inversion_clear h.
      destruct (f x).
      - apply NoDupA_cons.
        rewrite filter_InA; intuition.
        apply ih; auto.
      - apply ih; auto.
    }
  Qed.

  Lemma filter_app_preserve:
    forall (f: A -> bool) (pf: Proper pequiv f),
    forall l l1 x l2 (hl: l == l1 ++ x :: l2)
           (inx1: ~ InA equiv x l1) (inx2: ~ InA equiv x l2) lf1 lf2,
      filter f l == lf1 ++ x :: lf2 ->
      filter f l1 == lf1 /\ filter f l2 == lf2.
  Proof.
    intros f pf l l1 x l2 hl inx1 inx2 lf1 lf2 hf.
    rewrite hl in hf; clear hl l.
    rewrite ListUtil.filter_app in hf.
    simpl filter at 2 in hf.
    destruct (f x).
    { apply list_app_preserve with x; auto.
      contradict inx1; apply filter_InA in inx1; auto; apply inx1.
      contradict inx2; apply filter_InA in inx2; auto; apply inx2.
    }
    { rewrite <- ListUtil.filter_app in hf.
      assert (InA equiv x (filter f (l1 ++ l2))).
      rewrite hf.
      apply InA_app_iff; right; apply InA_cons; left; auto.
      apply filter_InA in H; auto.
      destruct H as [H _].
      apply InA_app_iff in H; intuition.
    }
  Qed.
  
End filter.

(** ** Tools about map_filter

 map_filter transforms a list into another, by applying to each
 element of the first list a Some/None function (the resulting element
 is added to the new list if it is not None).  
 *)

(** ** Tools about RestrictA

 RestrictA is a subrelation of equality, conditionned using a
 compatible predicate.  
 *)
Section RESTRICTA.

  Definition RestrictA A (eqA:relation A) (P: A -> Prop) a1 a2 :=
    eqA a1 a2 /\ P a1.

  Instance RestrictA_per {A: Type} `{SA:PartialSetoid A}
           {P: A -> Prop} {P_proper: Proper fequiv P}: PER (RestrictA pequiv P).
  Proof.
    split.
    intros a1 a2 [e h]; split.
    symmetry; assumption.
    rewrite <- e; assumption.
    intros a1 a2 a3 [e12 h] [e23 _]; split.
    transitivity a2; assumption.
    assumption.
  Qed.

  Instance RestrictA_per_s {A: Type} `{SA:Setoid A}
           {P: A -> Prop} {P_proper: Proper fequiv P}: PER (RestrictA equiv P).
  Proof.
    split.
    intros a1 a2 [e h]; split.
    symmetry; assumption.
    rewrite <- e; assumption.
    intros a1 a2 a3 [e12 h] [e23 _]; split.
    transitivity a2; assumption.
    assumption.
  Qed.

End RESTRICTA.

(** ** Tools about ConsistentA

 A list of pairs is consistent whenever for any two pairs in the list,
 if their first parts are equal, their second parts are also equal.
 *)
Section ConsistentA.

  Context {A: Type} `{SA:Setoid A}.
  Context {B: Type} `{SB:Setoid B}.

  Definition ConsistentA (l: list  (A*B)) :=
    forall a1 a2 b1 b2, InA equiv (a1,b1) l ->
                        InA equiv (a2,b2) l ->
                        a1 == a2 -> b1 == b2.

  Global Instance ConsistentA_compat:
    Proper (equivlistA equiv ==> iff) ConsistentA .
  Proof.
    intros l1 l2 el; split; intros h a1 a2 b1 b2 h1 h2 ea;
      apply h with a1 a2.
    rewrite (el (a1, b1)); assumption.
    rewrite (el (a2, b2)); assumption.
    assumption.
    rewrite <- (el (a1, b1)); assumption.
    rewrite <- (el (a2, b2)); assumption.
    assumption.
  Qed.

End ConsistentA.


(** ** Tools about assocA

 assocA searches into a list of pairs a first part and returns its
 second part (None if not found).  
 *)
Section AssocA.

  Context {A: Type} `{SA: Setoid A}.
  Context `{Adec: Decider equiv}.

  Context {B: Type} `{SB: Setoid B}.


  Fixpoint assocA (x: A) (l: list (A * B)): option B :=
    match l with
    | nil => None
    | (y, v)::q => if Adec x y then Some v else
                     assocA x q end.

  Global Instance assocA_compat:
    Proper fequiv  assocA.
  Proof.
    intros a1 a2 ea l1 l2 el.
    induction el as [|[k1 b1] [k2 b2] l1 l2 [ek eb] el hrec]; simpl.
    auto.
    cbv in ek,eb.
    destruct (Adec a1 k1) as [e1|ne1];
      destruct (Adec a2 k2) as [e2|ne2].
    apply eb.
    elim ne2.
    rewrite <- ea.
    rewrite e1.
    apply ek.
    elim ne1.
    rewrite ea.
    rewrite e2.
    symmetry;
      apply ek.
    apply hrec.
  Qed.

  Lemma assocA_none:
    forall (l: list(A * B)) (a: A),
      (assocA a l) == None <->
      forall (b: B), ~ InA equiv (a, b) l.
  Proof.
    intros until a; split.
    induction l as [|[a0 b0] l Hl]; simpl.
    intros; rewrite InA_nil.
    auto.
    destruct (Adec a a0) as [E |NE]; simpl.
    unfold value; congruence.
    intro h; specialize (Hl h).
    intro b; rewrite InA_cons; intros [H|H].
    apply NE.
    eapply proj1.
    apply H.
    apply (Hl _ H).
    induction l as [|[a0 b0] l Hl]; simpl.
    reflexivity.
    destruct (Adec a a0) as [E |NE]; simpl.
    intro h; specialize (h b0);
      elim h; auto.
    intro h; apply Hl.
    intro b; specialize (h b).
    auto.
  Qed.

  Lemma assocA_InA:
    forall (l: list(A*B))  (a: A) (b: B),
      (assocA a l) == (Some b) ->
      InA equiv (a, b) l.
  Proof.
    induction l as [| [a' b'] t]; simpl.
    intros; elim H.
    intros a b; destruct (Adec a a') as [e|ne]; simpl.
    intro h; symmetry in h; rewrite InA_cons; left; split; auto.
    intro h; right; auto.
  Qed.

  Lemma InA_AssocA:
    forall (l: list(A*B))  (a: A) (b: B),
      ConsistentA l ->
      (InA equiv (a, b) l <->
       (assocA a l) == (Some b)).
  Proof.
    intros l a b h; split.
    2: auto using assocA_InA.
    revert a b h.
    induction l as [| [a' b'] t hrec]; simpl.
    intros a b _ H;rewrite InA_nil in H; elim H.
    intros a b; destruct (Adec a a') as [e|ne]; simpl.
    intros hC hI; eapply hC.
    apply InA_cons; left; reflexivity.
    apply hI.
    symmetry; apply e.
    intros hC hI; apply hrec.
    intros a1 a2 b1 b2 h1 h2 ea; eapply hC; eauto.
    rewrite InA_cons in hI.
    destruct hI as  [[ea _] | hI].
    elim (ne ea).
    exact hI.
  Qed.

  Global Instance assocA_compat2:
    Proper (equiv ==> RestrictA (equivlistA equiv) ConsistentA ==> equiv)
           assocA.
  Proof.
    intros a1 a2 ea l1 l2 [el hl1].
    assert (hl2 := proj1 (ConsistentA_compat el) hl1).
    case_equiv (assocA a1 l1); [intro b|]; intro h;symmetry.
    rewrite <- InA_AssocA in h; auto.
    rewrite <- InA_AssocA ; auto.
    assert (hh: (a1, b) == (a2, b)).
    split; [assumption|compute; reflexivity].
    rewrite hh in h.
    rewrite (el (a2, b)) in h.
    exact h.
    rewrite assocA_none.
    rewrite assocA_none in h.
    intros b hb; specialize (h b); apply h.
    assert (hh:(a1, b) == (a2, b)).
    split; [assumption|compute; reflexivity].
    rewrite <- hh in hb.
    rewrite <- (el (a1, b)) in hb.
    exact hb.
  Qed.

  Lemma assoc_dec:
    forall (eqB_dec: Decider (equiv (A := B))) (a: A) l,
      {b: B | InA equiv (a, b) l}
      +
      {forall b, ~ (InA equiv (a, b) l)}.
  Proof.
    intros eqB_dec a l.
    induction l as [ | ab2 t Hrec].
    right.
    intros b ina.
    rewrite InA_nil in ina.
    elim ina.
    destruct ab2 as [a2 b2].
    destruct (Adec a a2).
    left.
    assert (proof: InA equiv (a, b2) ((a2, b2) :: t)).
    apply InA_cons_hd.
    hnf.
    unfold RelationPairs.RelCompFun; simpl.
    split.
    rewrite e.
    reflexivity.
    reflexivity.
    apply (@exist B (fun b: B => InA equiv
                                     (a, b) ((a2, b2) :: t)) b2 proof).
    destruct Hrec as [ina | nina].
    left.
    destruct ina as [term proof].
    assert (new_proof :=
              InA_cons_tl (a2, b2) proof).
    apply (@exist _ (fun b: B => InA equiv (pair a b)
                                     (cons (pair a2 b2) t)) term new_proof).
    right.
    intros b ina.
    rewrite InA_cons in ina.
    destruct ina as [inah | inat].
    cbv in inah.
    apply n.
    apply (proj1 inah).
    apply (nina b).
    assumption.
  Qed.

End AssocA.

(** * Tools about rev on a list *)
Section Rev_List.

  Lemma rev_equivlistA {A} `{SA:Setoid A}: 
    forall l: list A, equivlistA equiv l (rev l).
  Proof.
    intros l.
    destruct SA as [eqA eqA_equiv].
    induction l as [| b l ih]; simpl.
    reflexivity.
    intros bb; split; intros h.
    - rewrite InA_cons in h; rewrite InA_app_iff; destruct h as [h|h].
      + right; rewrite InA_cons; left; auto.
      + left; eapply InA_compat; auto.
        reflexivity.
        symmetry; apply ih.
        apply h.
    - rewrite InA_cons; rewrite InA_app_iff in h; destruct h as [h|h].
      + right; eapply InA_compat; auto.
        reflexivity.
        apply ih.
        apply h.
      + left; rewrite InA_cons in h; destruct h as [h|h]; auto.
        apply InA_nil in h; tauto.
  Qed.

  Lemma rev_rev_eqlistA {A} `{SA:Setoid A}:
    forall (l l': list A),
      eqlistA equiv l (rev l') <-> eqlistA equiv (rev l) l'.
  Proof.
    destruct SA as [eqA eqA_equiv].
    cut (forall l l', eqlistA eqA l (rev l') -> eqlistA eqA (rev l) l').
    { intros ok l l'; split; intros h; [|symmetry];
        apply ok; [|symmetry]; auto. }
    { intros l l'; intros h.
      assert (ok := eqlistA_rev_app h (Equivalence_Reflexive nil)).
      do 2 rewrite <- app_nil_end in ok.
      rewrite ok.
      rewrite rev_involutive.
      reflexivity.
    }
  Qed.

  Lemma rev_app_cons {A} `{SA:Setoid A}:
    forall (x: A) (l1 l2: list A),
      rev (l1 ++ x :: l2) == rev l2 ++ x :: rev l1.
  Proof.
    intros l1 l2 x.
    apply rev_rev_eqlistA.
    rewrite rev_app_distr. 
    rewrite rev_involutive.
    simpl; rewrite rev_involutive.
    rewrite ListUtil.list_app_first_last.
    reflexivity.
  Qed.

End Rev_List.

Section Existsb_InA.

  Lemma existsb_exists_InA {A: Type} `{SA: Setoid A}:
    forall (f : A -> bool) (f_compat: Proper fequiv f) 
           (l : list A),
      existsb f l = true <-> (exists x : A, InA equiv x l /\ f x = true).
  Proof.
    intros; induction l; simpl; split; intros; try congruence.
    destruct H as [x [ina _]].
    rewrite InA_nil in ina; tauto.
    rewrite Bool.orb_true_iff in H; destruct H as [F | E].
    exists a; simpl; split;[|assumption].
    rewrite InA_cons; left; reflexivity.
    rewrite IHl in E; clear IHl.
    destruct E as [x [I F]].
    exists x; split; [| assumption].
    rewrite InA_cons; right; assumption.
    rewrite Bool.orb_true_iff.
    destruct H as [x [I F]].
    rewrite InA_cons in I; destruct I as [I|I].
    left; rewrite <- (f_compat _ _ I); assumption.
    right; apply IHl.
    exists x; split; assumption.
  Qed.

End Existsb_InA.

Section Forallb_InA.

  Lemma forallb_forall_InA {A: Type} `{SA: Setoid A}:
    forall (f : A -> bool) (f_compat: Proper fequiv f) 
           (l : list A),
      forallb f l = true <-> (forall x : A, InA equiv x l -> f x = true).
  Proof.
    intros; induction l; simpl; split; intros; try congruence.
    rewrite InA_nil in H0; tauto.
    rewrite InA_cons in H0.
    rewrite Bool.andb_true_iff in H; destruct H as [H H'].
    destruct H0 as [H0|H0].
    erewrite f_compat.
    apply H.
    assumption.
    destruct IHl as [IHl _]; specialize (IHl H').
    apply IHl; assumption.
    rewrite Bool.andb_true_iff.
    split.
    apply H; rewrite InA_cons; left; reflexivity.
    apply IHl; intros x inx; apply H; rewrite InA_cons; right; assumption.
  Qed.

End Forallb_InA.

Section Existsb_Forallb_compat.

  Global Instance existsb_compat:
    forall A (eqA: relation A),
      Proper ( (eqA ==> eq) ==> (eqlistA eqA) ==> eq) (existsb (A:=A)).
  Proof.
    intros A eqA f g efg lf lg eqlfg.
    revert lg eqlfg; induction lf as [| a lf IH]; destruct lg as [| b lg];
      intros eqlfg; inversion_clear eqlfg; subst; simpl; auto.
    rename H into eab; rename H0 into elfg.
    specialize (efg _ _ eab); rewrite efg.
    specialize (IH _ elfg); rewrite IH.
    auto.
  Qed.

  Global Instance forallb_compat:
    forall A (eqA: relation A),
      Proper ( (eqA ==> eq) ==> (eqlistA eqA) ==> eq) (forallb (A:=A)).
  Proof.
    intros A eqA f g efg lf lg eqlfg.
    revert lg eqlfg; induction lf as [| a lf IH]; destruct lg as [| b lg];
      intros eqlfg; inversion_clear eqlfg; subst; simpl; auto.
    rename H into eab; rename H0 into elfg.
    specialize (efg _ _ eab); rewrite efg.
    specialize (IH _ elfg); rewrite IH.
    auto.
  Qed.

End Existsb_Forallb_compat.

Section Other_forallb.

  Lemma forallb_compat1:
    forall (A: Type) (eqA: A -> A -> Prop) (refl: Reflexive eqA)
           (f1 f2: A -> bool) (EF: (eqA ==> eq) f1 f2) 
           (l : list A),
      forallb f1 l = forallb f2 l.
  Proof.
    intros; induction l; simpl.
    - auto.
    - rewrite (EF _ _ (refl a)), IHl; reflexivity.
  Qed.

  Global Instance forallb_compat_equiv:
    forall (A : Type) (SA : Setoid A),
      Proper ((equiv ==> eq) ==> equivlistA equiv ==> eq) (forallb (A:=A)).
  Proof.
    intros A eqA f1 f2 ef l1 l2 el. 
    cut (forallb f1 l1 = true <-> forallb f2 l2 =true).
    destruct (forallb f1 l1);destruct (forallb f2 l2);intuition congruence.
    rewrite forallb_forall_InA;try proper_.
    rewrite forallb_forall_InA;try proper_.
    apply all_compat.
    intro a.
    apply impl_compat.
    apply el.
    apply eq_compat;auto.
  Qed.

End Other_forallb.

(** * Decidability on lists *)
Section Decidability_on_Lists.

  Lemma eqlistA_dec:
    forall {A: Type} {eqA: relation A},
      Decider eqA -> Decider (eqlistA eqA).
  Proof.
    intros A eqA dec.
    intros l1.
    induction l1 as [| h1 l1 IH1];  intros [| h2 l2]. 
    left; auto.
    right; intros FF; inversion FF.
    right; intros FF; inversion FF.
    specialize (IH1 l2).
    case (dec h1 h2); intros H; destruct IH1 as [L|L].
    left; apply eqlistA_cons; assumption.
    right; intros FF; inversion FF; auto.
    right; intros FF; inversion FF; auto.
    right; intros FF; inversion FF; auto.
  Qed.

  Lemma equivlistA_dec:
    forall {A: Type} {SA: Setoid A},
      Decider (equiv (A := A)) -> Decider (equivlistA equiv).
  Proof.
    intros A SA eqA_dec.
    intros l1 l2.
    pose (inl1 := fun x => if InA_dec eqA_dec x l1 then true else false).
    pose (inl2 := fun x => if InA_dec eqA_dec x l2 then true else false).
    assert (inl1_compat: Proper (equiv ==> eq) inl1).
    {
      intros a1 a2 ea.
      unfold inl1.
      destruct (InA_dec eqA_dec a1 l1);
        destruct (InA_dec eqA_dec a2 l1);
        try rewrite ea in *;tauto || auto.
    }
    assert (inl2_compat: Proper (equiv ==> eq) inl2).
    {
      intros a1 a2 ea.
      unfold inl2.
      destruct (InA_dec eqA_dec a1 l2);destruct (InA_dec eqA_dec a2 l2);
        try rewrite ea in *;tauto || auto.
    }
    pose (cond1 := forallb inl2 l1).
    pose (cond2 := forallb inl1 l2).
    case_eq cond1; intros h1.
    case_eq cond2; intros h2.
    {
      left.
      unfold cond1 in h1; erewrite forallb_forall_InA in h1; auto.
      unfold cond2 in h2; rewrite forallb_forall_InA in h2; auto.
      intros x.
      unfold inl1 in h2; unfold inl2 in h1.
      split; intros H.
      - apply h1 in H.
        destruct (InA_dec (eqA:=equiv) eqA_dec x l2); congruence.
      - apply h2 in H.
        destruct (InA_dec (eqA:=equiv) eqA_dec x l1); congruence.
    }
    {
      clear h1.
      right; intros F.
      unfold cond2 in h2.
      rewrite (forallb_existsb inl1 (fun x => negb (inl1 x))) in h2.
      rewrite Bool.negb_false_iff in h2.
      rewrite existsb_exists_InA in h2.
      shelve.
      intros a1 a2 ea;apply f_equal.
      apply inl1_compat;auto.
      intros;symmetry;apply Bool.negb_involutive.
      Unshelve.
      destruct h2 as [x [h2 h1]].
      rewrite Bool.negb_true_iff in h1.
      unfold inl1 in h1.
      destruct (InA_dec (eqA:=equiv) eqA_dec x l1) as [yes|no];try congruence.
      contradict no;rewrite <- (F x) in h2;auto.
    }
    {
      right; intros F.
      unfold cond1 in h1.
      rewrite (forallb_existsb inl2 (fun x => negb (inl2 x))) in h1.
      rewrite Bool.negb_false_iff in h1.
      rewrite existsb_exists_InA in h1.
      shelve.
      intros a1 a2 ea;apply f_equal.
      apply inl2_compat;auto.
      intros;symmetry;apply Bool.negb_involutive.
      Unshelve.
      destruct h1 as [x [h1 h2]].
      rewrite Bool.negb_true_iff in h2.
      unfold inl2 in h2.
      destruct (InA_dec (eqA:=equiv) eqA_dec x l2) as [yes|no];try congruence.
      contradict no;rewrite (F x) in h1;auto.
    }
  Qed.

End Decidability_on_Lists.

Section Rev.

  Variable (A: Type).

  Lemma rev_intro:
    forall l l': list A, l = l' <-> rev l = rev l'.
  Proof.
    induction l as [| a l H].
    intros l'; simpl.
    split; intros X.
    rewrite <- X; simpl; reflexivity.
    rewrite <- rev_involutive.
    rewrite <- X; simpl; reflexivity.

    intros [| a' l']; simpl.
    split. 
    intros F; inversion F.
    intros F; symmetry in F; contradict F; apply app_cons_not_nil.
    split; intros X.
    inversion X.
    reflexivity.
    destruct (app_inj_tail _ _ _ _ X) as [Y YY].
    rewrite YY.
    rewrite <- H in Y.
    rewrite Y.
    reflexivity.
  Qed. 
  
  Definition list_rev_destruct:
    forall eqA (eqA_refl: Reflexive eqA) (l:list A),
      {first:list A  & {last:A | eqlistA eqA l (first++ last::nil) }}
      + {eqlistA eqA l nil}.
  Proof.
    intros eqA eqA_refl l;induction l as [|a l Hl];simpl.
    right;left.
    left.
    destruct Hl as [[next [last Hcons]]| Hnil ].
    split with (a::next).
    exists last.
    simpl app.
    right;auto.
    exists nil.
    exists a.
    right;auto.
  Defined.

  Global Instance removelast_eqlistA_compat :
    forall eqA, Proper (eqlistA eqA ==> eqlistA eqA) (@removelast A).
  Proof.
    intros eqA l1 l2 hl;induction hl as [| a1 a2 l1 l2 ha hl hrecl];
      [left|destruct hl;[left|right;assumption]].
  Qed.

End Rev.

Section map_filter.

  Context {A: Type} `{SA: Setoid A}.
  Context {B: Type} `{SB: Setoid B}.

  Section Fixed_fopt.

    Context (fopt: A -> option B).

    Context {fopt_compat: Proper fequiv fopt}.

    Fixpoint map_filter (l: list A) :=
      match l with
        nil => nil
      | a :: q =>
        match fopt a with
          None =>  map_filter q
        | Some b => b::map_filter q
        end end.

    Lemma InA_map_filter_iff:
      forall (l: list A) (b: B),
        InA equiv b (map_filter l) <->
        (exists a: A, InA equiv a l /\ (fopt a) == (Some b)).
    Proof.
      intros until b.
      split.

      induction l as [ | a l Hl]; simpl.
      rewrite InA_nil.
      intro h; elim h.

      case_eq (fopt a).

      intros b0 eb0.
      rewrite InA_cons.
      intros [e | h].

      (* * b0 = b *)
      exists a.
      rewrite eb0.
      rewrite InA_cons.
      simpl.
      split. left; reflexivity. symmetry; assumption.

      (* * In b (map_filter l) *)
      specialize (Hl h).
      firstorder.

      (* option_f = None *)
      intros Ha Hb.
      specialize (Hl Hb).
      firstorder.

      (* <- *)
      induction l as [ | a l Hl]; simpl.
      firstorder.
      rewrite InA_nil in H; elim H.

      intros [a0 [H1 H2]].
      rewrite InA_cons in H1.
      destruct H1 as [H1 | H1].
      rewrite H1 in H2.
      clear a0 H1.
      destruct (fopt a); simpl in H2.
      2: elim H2.
      rewrite <- H2.
      rewrite InA_cons.
      left; reflexivity.
      case_eq (fopt a).
      intros; rewrite InA_cons; right.
      firstorder.
      firstorder.
    Qed.

  End Fixed_fopt.

  Global Instance map_filter_compat:
    Proper (fequiv ==> equivlistA equiv ==> equivlistA equiv) map_filter.
  Proof.
    intros f1 f2 ef l1 l2 el b.
    repeat rewrite InA_map_filter_iff.
    apply ex_compat.
    intro a.
    apply and_compat.
    apply el.
    rewrite (ef a a); reflexivity.
    intros x y eq0.
    rewrite <- (ef _ _ eq0).
    symmetry.
    apply ef.
    reflexivity.
    intros x y eq0.
    rewrite (ef _ _ eq0).
    symmetry.
    apply ef.
    reflexivity.
  Qed.

  Lemma map_filter_eq_locality:
    forall (f1 f2: A -> option B) (l1 l2: list A),
      equivlistA equiv l1 l2 ->
      Proper fequiv f1 -> Proper fequiv f2 ->
      (forall a, InA equiv a l1 -> f1 a == f2 a) ->
      equivlistA equiv (map_filter f1 l1) (map_filter f2 l2).
  Proof.
    intros until l2; intros el p1 p2 h.
    intros x; split; intros hx.
    { apply InA_map_filter_iff in hx; auto.
      destruct hx as [a [ha haa]].
      apply InA_map_filter_iff; auto.
      exists a; split.
      rewrite <- el; auto.
      rewrite <- haa; symmetry; apply h; auto.
    }
    { apply InA_map_filter_iff in hx; auto.
      destruct hx as [a [ha haa]].
      apply InA_map_filter_iff; auto.
      exists a; split.
      rewrite el; auto.
      rewrite <- haa; apply h; auto.
      rewrite el; auto.
    }
  Qed.

  Lemma map_filter_eq_locality_converse:
    Decider (equiv (A := B)) ->
    forall (f1 f2: A -> option B) (l1 l2: list A),
      equivlistA equiv l1 l2 ->
      Proper fequiv f1 -> Proper fequiv f2 ->
      ~ equivlistA equiv (map_filter f1 l1) (map_filter f2 l2) -> 
      exists a, InA equiv a l1 /\ ~ f1 a == f2 a.
  Proof.
    intros eqB_dec.
    intros until l2; intros el p1 p2 h.
    assert (X: ~ (forall a : A, InA equiv a l1 -> f1 a == f2 a)).
    intros F; eapply map_filter_eq_locality in F; eauto.
    clear h el l2.
    induction l1 as [| a1 l1 ih].
    contradict X.
    intros a; rewrite InA_nil; tauto.
    case (eqoptionA_dec eqB_dec (f1 a1) (f2 a1)); intros op.
    2: { exists a1; split; auto. }
    assert (XX: ~ (forall a : A, InA equiv a l1 -> f1 a == f2 a)).
    { intros F; contradict X.
      intros a; rewrite InA_cons; intros [eqa|ina].
      rewrite eqa; auto.
      apply F; auto. }
    specialize (ih XX); clear XX X.
    destruct ih as [a [ina fa]].
    exists a; split; [rewrite InA_cons; right|]; auto.
  Qed.    
  
End map_filter.

Global Instance map_filter_compat_eqlistA:
  forall A B (eqA: relation A) (eqB: relation B),
    Proper ((eqA ==> eqoptionA eqB)
              ==> eqlistA eqA ==> eqlistA eqB) map_filter.
Proof.
  intros A B eqA eqB f1 f2 ef l1 l2 el.
  induction el; simpl.
  left.
  generalize (ef _ _ H).
  destruct (f1 x); destruct (f2 x'); simpl.
  intros X; right; assumption.
  intros F; elim F.
  intros F; elim F.
  intros _; assumption.
Qed.

(** * Tools about eqlistA *)
Section EqlistA_Utils.

  Variable (A : Type) (eqA : relation A).

  Lemma l_eqlistA_nil : forall l, eqlistA eqA l nil <-> l = nil.
  Proof.
    intros l.
    split.
    intros eq.
    apply eqlistA_length in eq.
    simpl in eq.
    destruct l;simpl in eq;congruence.
    intro h;rewrite h;hnf;auto.
  Qed.

  Lemma l_eqlistA_cons : forall l1 h2 t2, eqlistA eqA l1 (h2 :: t2) ->
                                          exists h1 t1, l1 = h1 :: t1.
  Proof.
    intros l1 h2 t2 eql.
    inversion eql as [ | h1' h2' t1' t2' eqh eqt]; subst.
    exists h1'.
    exists t1'.
    reflexivity.
  Qed.

  Lemma cons_eqlistA_cons :
    forall h1 t1 h2 t2, eqlistA eqA (h1 :: t1) (h2 :: t2) <->
                        eqA h1 h2 /\ eqlistA eqA t1 t2.
  Proof.
    intros h1 t1 h2 t2.
    split.
    intros eql.
    inversion eql as [ | h1' h2' t1' t2' eqh eqt]; subst.
    split; assumption.
    intros [eqh eqt].
    apply eqlistA_cons; assumption.
  Qed.

End EqlistA_Utils.

Global Instance nth_compat (A:Type) (SA:PartialSetoid A) : Proper fequiv (nth (A:=A)).
Proof.
  intros n_ n en.
  hnf in en;rewrite en;clear n_ en.
  induction n as [|n hn];intros l1 l2 el def1 def2 edef;inversion el;subst;simpl;auto.
  apply hn;auto.
Qed.

Ltac lsplit pl :=
  hnf in pl;
  match (type of pl) with
  | (eqlistA ?R (cons ?a1 ?l1) (cons ?a2 ?l2)) =>
    let X := fresh "X" in
    assert (X:=proj1 (cons_eqlistA_cons R a1 l1 a2 l2) pl);
    clear pl;rename X into pl
  | (eqlistA ?R nil nil) => clear pl
  | (eqlistA ?R nil (cons _ _)) =>
    inversion pl
  | (eqlistA ?R (cons _ _) nil) =>
    inversion pl
  end.

(** * Stutter : copy list with stuttering at element n *)

Section Stutter.

  Fixpoint stutter A n (l:list A):=
    match l with
      nil => nil
    | i :: q =>
      match n with
        0 => i :: removelast (i::q)
      | S n => i :: stutter n q end end.

  Lemma removelast_length: forall A (l:list A), length (removelast l) = (length l) - 1.
  Proof.
    intros a l; induction l ;auto.
    simpl.
    destruct l.
    simpl;lia.
    transitivity (S (length (removelast (a1 :: l))));[reflexivity|].
    rewrite IHl.
    simpl;lia.
  Qed.

  Lemma stutter_length: forall A n l, length (stutter (A:=A) n l) = length l.
    intros A n;induction n.
    {
      intros [ | a l].
      reflexivity.
      unfold stutter.
      transitivity (S (length (a :: l) -1)).
      rewrite <- removelast_length;reflexivity.
      simpl;lia.
    }
    {
      intros [ | a l].
      reflexivity.
      simpl.
      apply f_equal,IHn.
    }
  Qed.

  Lemma removelast_Forall: forall A P (l:list A), Forall P l ->
                                                  Forall P (removelast l).
  Proof.
    intros A P l; induction l ;auto.
    simpl.
    destruct l.
    trivial.
    intro H;inversion H;subst;right;auto.
  Qed.

  Lemma stutter_Forall: forall A P n l,
      Forall P l -> Forall P (stutter (A:=A) n l).
    intros A P n;induction n.
    {
      intros [ | a l].
      trivial.
      intro H;inversion H;subst.
      unfold stutter.
      apply removelast_Forall in H.
      right;auto.
    }
    {
      intros [ | a l].
      trivial.
      intro H;inversion H;subst;right;auto.
    }
  Qed.

  Lemma stutter_nth_le: forall A a (l:list A) i j,
      i <= j ->
      nth i l a = nth i (stutter j l) a.
  Proof.
    intros A a l i j hj;revert i j hj.
    induction l as [|x q hrecl];simpl.
    destruct i;destruct j;simpl;auto.
    destruct i;destruct j;simpl;auto.
    intros;lia.
    intros hj;apply hrecl;lia.
  Qed.

  Lemma stutter_nth_gt: forall A a (l:list A) i j,
      j <= i - 1 -> i < length l ->
      nth (i - 1) l a = nth i (stutter j l) a.
  Proof.
    intros A a l i j hi hj;revert i j hi hj.
    induction l as [|x q hrecl];intros i j hi hj.
    destruct i;destruct j;simpl in *;auto;try lia.
    destruct i;destruct j;simpl in hj;auto;try lia.
    2: {
      destruct i;try lia.
      replace (S (S i) - 1) with (S (S i - 1)) in * by lia.
      apply hrecl;lia.
    }
    simpl;rewrite PeanoNat.Nat.sub_0_r; revert hj;clear hi hrecl.
    revert x i;induction q as [|y q hq];simpl;intros x i;intros;try lia.
    destruct i as [|i].
    destruct q; simpl;simpl in hj;try lia;auto.
    apply hq; lia.
  Qed.

  Lemma stutter_nth_gt2: forall A a (l:list A) i j,
      j <= i -> i < length l - 1 ->
      nth i l a = nth (1+i) (stutter j l) a.
  Proof.
    intros A a l i j hi hj;revert i j hi hj.
    induction l as [|x q hrecl];simpl;intros i j hi hj.
    destruct i;destruct j;simpl;auto.
    destruct i;destruct j;simpl;auto;intros;try lia.
    destruct q;simpl in hj;lia || reflexivity.
    2: {
      simpl;apply hrecl;lia.
    }
    rewrite PeanoNat.Nat.sub_0_r in hj; revert hj;clear hi hrecl.
    revert x i;induction q as [|y q hq];simpl;intros x i;intros;try lia.
    destruct i as [|i].
    destruct q; simpl;simpl in hj;try lia;auto.
    apply hq; lia.
  Qed.

  Lemma stutter_last_id:
    forall A (l:list A) j, j = length l - 1 ->
                           stutter j l = l.
  Proof.
    intros A l j h.
    subst.
    destruct l as [|x q];simpl;auto.
    rewrite PeanoNat.Nat.sub_0_r.
    revert x;induction q as [|y q hq];intros x;simpl;auto.
    apply (f_equal (cons x)).
    apply hq.
  Qed.

  Lemma stutter_app_r:
    forall A (l1 l2:list A) j, j >= length l1 ->
                               stutter j (l1++l2) = l1 ++ stutter (j - length l1) l2.
  Proof.
    intros A l1 l2.
    induction l1 as [|x q hq];intros j h;simpl;auto.
    rewrite PeanoNat.Nat.sub_0_r.
    reflexivity.
    destruct j as [|j];simpl in h;try lia.
    simpl.
    apply (f_equal (cons x)).
    apply hq.
    lia.
  Qed.

End Stutter.

(** * Tools for building lists of natural numbers (goins up, down, with repeated patterns) *)

Section List_builders.

  (* list counting up from min to min+len-1 *)

  Fixpoint uc min len :=
    match len with 0 => nil | S n => min :: uc (min+1) n end.

  Lemma uc_length: forall min len, length (uc min len) = len.
  Proof.
    intros min len; revert min.
    induction len;simpl;auto.
  Qed.

  (* list counting down from min+len-1 to min *)

  Fixpoint dc min len :=
    match len with 0 => nil | S n => min+n :: dc min n end.

  Lemma dc_length: forall min len, length (dc min len) = len.
  Proof.
    intros min len; revert min.
    induction len;simpl;auto.
  Qed.

  (* list made of [len] copies of [cte] *)

  Fixpoint rep cte len : list nat :=
    match len with 0 => nil | S n => cte :: rep cte n end.

  Lemma rep_length: forall cte len, length (rep cte len) = len.
  Proof.
    intros cte len; revert cte.
    induction len;simpl;auto.
  Qed.

  (* list made of [len] copies of list [lcte] *)

  Fixpoint repl A (lcte:list A) len : list A :=
    match len with 0 => nil | S n => lcte ++ repl lcte n end.

  Lemma repl_length: forall A lcte len, length (repl (A:=A) lcte len) = len * length lcte.
  Proof.
    intros A lcte len.
    induction len;simpl;auto.
    rewrite app_length;lia.
  Qed.

  Fixpoint flatmap A B (f:A -> list B) l : list B :=
    match l with nil => nil | x::q  => f x ++ flatmap f q end.

  Lemma flatmap_length:
    forall A B f l, length (flatmap (A:=A) (B:=B) f l) =
                    fold_right plus 0 (List.map (fun a => length (f a)) l).
  Proof.
    intros A B f l.
    induction l;simpl;auto.
    rewrite app_length;lia.
  Qed.


  Lemma dc_le: forall min len, Forall (le min) (dc min len).
  Proof.
    intros min len; revert min; induction len as [|len hlen]; simpl;try right;auto.
    lia.
  Qed.

  Lemma dc_gt: forall min len, Forall (gt (min+len)) (dc min len).
  Proof.
    intros min len.
    replace min with (min + len - len) at 2 by lia.
    assert (h:min + len >= len) by lia.
    revert h;
      generalize (min + len) as max.
    clear min.
    induction len as [|len hlen]; simpl;try right;auto.
    lia.
    replace (max - S len) with ((max - 1) - len) by lia.
    apply Forall_impl with (gt (max-1)).
    intro;lia.
    apply hlen.
    lia.
  Qed.

  Lemma dc_nth: forall min len x i, i < len -> nth i (dc min len) x = min + len - i - 1.
  Proof.
    intros min len x;
      induction len as [|len hlen];intros i hi.
    simpl;lia.
    destruct i as [|i].
    simpl;lia.
    simpl dc;simpl nth.
    rewrite hlen;lia.
  Qed.

  Lemma dc_last: forall min len x, len > 0 -> last (dc min len) x = min.
  Proof.
    intros min len x;
      destruct len as [|len];intro h;try lia.
    induction len as [|len hlen].
    simpl;lia.
    rewrite <- hlen at 2 by lia.
    reflexivity.
  Qed.

  Lemma dc_removelast: forall min len,
      len > 0 ->
      removelast (dc min len) = dc (min+1) (len-1).
  Proof.
    intros min len;
      destruct len as [|len];intro h;try lia;clear h.
    induction len as [|len hlen].
    reflexivity.
    simpl dc at 2.
    replace (S len - 1) with len in hlen by lia.
    rewrite <- hlen.
    simpl;apply f_equal2;auto.
    lia.
  Qed.

  Lemma uc_le: forall min len, Forall (le min) (uc min len).
  Proof.
    intros min len; revert min; induction len; simpl;try right;auto.
    apply Forall_impl with (le (min + 1)).
    intro;lia.
    apply IHlen.
  Qed.

  Lemma uc_gt: forall min len, Forall (gt (min+len)) (uc min len).
  Proof.
    intros min len; revert min; induction len as [|len hlen]; simpl;try right;auto.
    lia.
    replace (min + S len) with ((min + 1) + len) by lia.
    apply hlen.
  Qed.

  Lemma rep_app: forall x n p, rep x (n+p)=rep x n ++ rep x p.
  Proof.
    intros x n p;induction n;simpl;auto.
    apply f_equal;auto.
  Qed.

  Lemma last_cons: forall A x d (l:list A), l <> nil -> last (x::l) d = last l d.
  Proof.
    intros A x d [|y q];simpl;auto.
    congruence.
  Qed.

  Lemma last_app: forall A d (l1 l2:list A), l2 <> nil -> last (l1++l2) d=last l2 d.
  Proof.
    intros A d l1 l2 hl2.
    induction l1 as [|x q hq].
    {
      simpl app;reflexivity.
    }
    {
      simpl app;rewrite last_cons.
      induction l2 as [|y qq hqq];try congruence.
      destruct q as [|z qq].
      assumption.
      simpl app;congruence.
    }
  Qed.

End List_builders.

Section App.

  Context {A: Type} `{SA: PartialSetoid A}.

  Lemma app_proper_l:
    forall (a b: list A), Proper pequiv (a ++ b) -> Proper pequiv a.
  Proof.
    induction a as [| x a ih]; intros b h.
    left.
    change (Proper pequiv (x :: a ++ b)) in h.
    right.
    inversion h; auto.
    apply ih with b.
    inversion h; auto.
  Qed.

  Lemma app_proper_r:
    forall (a b: list A), Proper pequiv (a ++ b) -> Proper pequiv b.
  Proof.
    induction a as [| x a ih]; intros b h.
    rewrite app_nil_l in h; auto.
    apply ih.
    change (Proper pequiv (x :: a ++ b)) in h.
    inversion h; auto.
  Qed.

  Lemma app_cut:
    forall (l1 l2 m1 m2: list A),
      l1 ++ l2 =~= m1 ++ m2 ->
      (exists x, l1 =~= m1 ++ x /\ m2 =~= x ++ l2) \/
      (exists y, m1 =~= l1 ++ y /\ l2 =~= y ++ m2).
  Proof.
    induction l1 as [| a1 l1 ih]; intros l2 m1 m2 h.

    { rewrite app_nil_l in h.
      right; exists m1; split; auto.
      rewrite app_nil_l; eapply app_proper_l; proper_.
    }      

    destruct m1 as [| b1 m1].

    { rewrite app_nil_l in h.
      destruct m2 as [|b2 m2]; inversion h; subst.
      assert (dum: l1 ++ l2 =~= nil ++ m2) by auto.
      rewrite app_nil_l in dum.
      left; exists (a1 :: l1); split.
      rewrite app_nil_l; right; auto.
      eapply app_proper_l; proper_. 
      cut (b2 :: m2 =~= a1 :: (l1 ++ l2)).
      auto.
      right; symmetry; auto.
    }

    assert (dum: a1 :: (l1 ++ l2) =~= b1 :: (m1 ++ m2)) by auto.
    clear h; inversion dum; subst.  
    specialize (ih _ _ _ H4); destruct ih as [[x [ih ihh]]|[y [ih ihh]]].
    - left; exists x; split; auto.
      cut (a1 :: l1 =~= b1 :: (m1 ++ x)); [|right]; auto.
    - right; exists y; split; auto.
      cut (b1 :: m1 =~= a1 :: (l1 ++ y)); [|right; [symmetry|]]; auto.
  Qed.

End App.

Close Scope signature_scope.
Unset Implicit Arguments.
