(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import Setoid.
From Coq Require Import RelationPairs.
From Coq Require Import Lt.
From Coq Require Import Lia.
From Coq Require Import ArithRing.
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import BinInt.
From Coq Require Import Nat.
From Coq Require Import Lia.

(** **** Local Imports *)
From Padec Require Import OptionUtil.
From Padec Require Import SetoidUtil.
From Padec Require Import ListUtils.
From Padec Require FoldRight.

(* end show *)
Open Scope nat_scope.
Open Scope signature_scope.
Set Implicit Arguments.

(** * Natural numbers up to some value [N] 
  N is of type [nat]
 *)
Section Nat_lt_n.

  Variable (N: nat).

  (** *** Type of natural numbers from 0 to N-1 *)
  Definition ltN := {n: nat | n < N }.

  Definition ltN_dec: Decider (equiv (A:=ltN)) :=
    fun x y => Peano_dec.eq_nat_dec (proj1_sig x) (proj1_sig y).

  (** *** Enumeration of all numbers in ltN *)
  Program Fixpoint enum_rec (k: nat) (h: k <= N) l: list {m: nat | m < N} :=
    match k with
    | O => l
    | S m => enum_rec (le_S_n _ _ (le_S h)) (exist h :: l)
    end.

  Definition enum: list {m: nat | m < N} := enum_rec (PeanoNat.Nat.le_refl N) nil.

  Lemma enum_rec_InA: 
    forall k (h: k <= N), forall x, forall l,
          InA equiv x l ->
          InA equiv x (enum_rec h l).
    intros k h x l.
    revert h l.
    induction k as [| k hk].
    intros; assumption.
    intros; apply hk.
    apply InA_cons.
    right; assumption.
  Qed.

  Lemma enum_rec_enum: forall k (h: k <= N),
      forall x, forall l, proj1_sig x < k -> InA equiv x (enum_rec h l).
    intros k h [n hn].
    induction k as [|k hk].
    simpl;intros;lia.
    simpl in *.
    intros l hnk.
    destruct (le_lt_or_eq (le_S_n _ _ hnk)) as [hlt | heq].
    auto.
    subst.
    apply enum_rec_InA.
    left.
    reflexivity.
  Qed.

  Lemma ltN_in_enum: forall x: {m: nat | m < N}, InA equiv x enum.
    intros.
    unfold enum.
    apply enum_rec_enum.
    apply (proj2_sig x).
  Qed.

End Nat_lt_n.

(** ** Additional Facts about [ltN n] defined as { p:nat | p < n } *)

(** n is of type (ltN (S n)), (it is the greatest number in (ltN (S n)) *)
Program Definition ltSN_max (n: nat): ltN (S n) := n.

(** transform a number x (in ltN) into the same number in (ltN (S n)) *)
Definition ltSN_inj: forall (n: nat), ltN n -> ltN (S n) :=
  fun n x =>
    @exist _ (fun x => x < S n) (proj1_sig x) (lt_S (proj2_sig x)).

(** no number from ltSN_inj can be equal to max *)
Lemma ltSN_inj_not_max:
  forall (n: nat) (k: ltN n),
    ~ (ltSN_inj k) == (ltSN_max n).
  (********)
  compute.
  intros n [ k Hk ].
  lia.
Qed.

(** being the max (n) or coming from inj is decidable *)
Definition ltSN_inj_max_dec:
  forall (n: nat) (k: ltN (S n)),
    { k': ltN n | k == (ltSN_inj k') }
    +
    { k == (ltSN_max n) }.
  (********)
  intros n [k hk]; simpl.
  destruct (Peano_dec.eq_nat_dec k n) as [E | NE];
    [right; assumption | left].
  assert (hkn: k < n) by lia.
  split with (@exist _ (fun x => x < n) _ hkn); reflexivity.
Qed.

(** ltSN_inj is compatible **)
Global Instance ltSN_inj_compat:
  forall (n: nat), Proper fequiv (@ltSN_inj n).
(* begin hide *******)
intros n [k1 hk1] [k2 hk2]; exact (fun n => n).
Qed.
(* end hide *)

(** * Natural numbers up to some value [K] 
  K is of type [N]
 *)
Section N_lt_n.

  Import NArith.

  Open Scope N_scope.

  Variable (K: N).

  (** *** Type of natural numbers from 0 to N-1 *)  
  Definition ltK := {n: N | n < K }.

  Definition ltK_dec: Decider (equiv (A:=ltK)) :=
    fun x y => N.eq_dec (proj1_sig x) (proj1_sig y).
  Import BinNat.N.

  (** *** Enumeration of all numbers in ltN *)
  Program Definition Kenum_rec: forall n: N, n <= K -> list ltK -> list ltK :=
    peano_rect (fun k => k<=K -> list ltK -> list ltK) (fun h l => l) 
               (fun m F h l => F _ (@exist _ (fun k => k < K) m _ :: l)). 
  (* begin hide *)
  Next Obligation.
    rewrite <- add_1_r in h.
    rewrite <- h.
    apply le_add_r.
  Qed.

  Next Obligation.
    rewrite <- le_succ_l.
    rewrite <- h.
    reflexivity.
  Qed.
  (* end hide *)

  Definition Kenum:= Kenum_rec  (le_refl K) nil.

  Lemma Kenum_rec_InA: forall k (h: k <= K), forall x, forall l, 
          InA equiv x l -> InA equiv x (Kenum_rec h l).
    intros k h x l.
    revert h l.
    induction k as [| k hk] using peano_ind.
    intros; assumption.
    unfold Kenum_rec.
    intros hhk l hl;rewrite peano_rect_succ.
    simpl;apply hk.
    apply InA_cons.
    right;assumption.
  Qed.
  
  Lemma Kenum_rec_enum: forall k (h: k <= K), forall x, forall l, 
          proj1_sig x < k -> InA equiv x (Kenum_rec h l).
    intros k h [n hn].
    induction k as [|k hk] using peano_ind;simpl in *.
    intros l hh;elim (nlt_0_r _ hh).
    intros l hnk.
    rewrite <- le_succ_l in hnk.
    rewrite <- succ_le_mono in hnk.
    rewrite lt_eq_cases in hnk.
    destruct hnk as [hlt | heq].
    unfold Kenum_rec;rewrite peano_rect_succ;auto.
    subst.
    unfold Kenum_rec;rewrite peano_rect_succ;simpl.
    apply Kenum_rec_InA.
    rewrite InA_cons.
    left.
    reflexivity.
  Qed.

  Lemma ltK_in_Kenum: forall x: {m: N | m < K}, InA equiv x Kenum.
    intros.
    unfold Kenum.
    apply Kenum_rec_enum.
    apply (proj2_sig x).
  Qed.

  Definition ltK_inj (n:N): option ltK.
    destruct (ZArith_dec.Z_lt_dec (Z.of_N n) (Z.of_N K)) as [h|_].
    apply Some;
      split with n.
    apply (proj2 (Znat.N2Z.inj_lt _ _) h).
    apply None.
  Defined.

  Close Scope N_scope.

End N_lt_n.

(** * Power Composition *)
Section Power_composition.

  Variable (A: Type).

  Fixpoint power (f: A -> A) (n: nat): A -> A :=
    match n with
    | 0%nat => (fun x => x)
    | S n => (fun x => f (power f n x))
    end.

  Lemma S_power: forall f n x, power f (S n) x = f (power f n x).
  Proof.
    reflexivity.
  Qed.

  Lemma power_S: forall f n x,
      power f (S n) x = (power f n (f x)).
  Proof.
    intros f n;
      induction n as [| n hn]; auto;intro x.
    simpl;apply f_equal,hn.
  Qed.

  Lemma power_add: forall f n p x,
      power f (n+p) x = (power f n (power f p x)).
  Proof.
    intros f n;
      induction n as [| n hn]; auto;intros x p.
    simpl;apply f_equal,hn.
  Qed.

  Lemma power_mult: forall f n p x,
      power f (n*p) x = (power (power f p) n x).
  Proof.
    intros f n;
      induction n as [| n hn]; auto;intros x p.
    simpl.
    rewrite power_add.
    apply f_equal,hn.
  Qed.

  Global Instance power_compat eqA:
    Proper ((eqA ==> eqA) ==> eq ==> eqA ==> eqA) power.
  Proof.
    intros f1 f2 ef n_ n en;subst.
    induction n as [| n hn];simpl.
    intros a1 a2 ea;apply ea.
    intros a1 a2 ea.
    apply ef,hn,ea.
  Qed.

End Power_composition.

(** * Sum *)
Section Sum.
  Open Scope nat_scope.

  (** sum f x = sum_{i = 1 .. x} f i *)
  Fixpoint sum (f: nat -> nat) (x: nat): nat :=
    match x with
    | O => O
    | S x' => (f x) + (sum f x')
    end.

  Lemma sum_prod: forall x y, sum (fun _ => y) x = x * y.
  Proof.
    induction x as [| x ih]; intros y; simpl.
    - reflexivity.
    - rewrite ih; reflexivity.
  Qed.

  Close Scope nat_scope.

End Sum.

(** * Modulo *)
Section Mod.
  Open Scope nat_scope.
  Lemma mod_plus_inv':
    forall (a b c: nat),
      a < c -> b < c ->
      exists x, x < c /\ (a + x) mod c = b /\
                forall x', x' < c -> x' <> x ->
                           (a + x') mod c <> b.
  Proof.
    intros a b c ha hb.
    case (Compare_dec.le_gt_dec a b); intros hab.
    {
      assert (D := Minus.le_plus_minus_r hab).
      exists (b - a).
      rewrite D; split; [lia|]; split.
      apply PeanoNat.Nat.mod_small; auto.
      intros x' hx' hn; intros F; contradict hn.
      assert (X := PeanoNat.Nat.div_mod (a + x') c).
      assert (d: c <> 0) by  lia; specialize (X d); clear d.
      pose (q := ((a + x') / c)); fold q in X; clearbody q.
      rewrite F in X.
      destruct q; [lia |].
      assert (c * S q >= c).
      { rewrite <- mult_n_Sm.
        cut (c * q >= 0); [lia |].
        apply PeanoNat.Nat.mul_nonneg_nonneg; lia.
      }
      lia.
    }
    {
      assert (hhab: a <= c + b) by lia.
      assert (D := Minus.le_plus_minus_r hhab).
      exists (c + b - a).
      rewrite D; split; [lia|].
      split.
      rewrite <- PeanoNat.Nat.add_mod_idemp_l; [|lia].
      rewrite PeanoNat.Nat.mod_same; [|lia].
      apply PeanoNat.Nat.mod_small; auto.
      intros x' hx' hn; intros F; contradict hn.
      assert (X := PeanoNat.Nat.div_mod (a + x') c).
      assert (d: c <> 0) by  lia; specialize (X d); clear d.
      pose (q := ((a + x') / c)); fold q in X; clearbody q.
      rewrite F in X.
      destruct q; [lia |destruct q].
      { rewrite PeanoNat.Nat.mul_1_r in X; lia. }
      { rewrite <- mult_n_Sm in X.
        assert (c * S q >= c).
        { rewrite <- mult_n_Sm.
          cut (c * q >= 0); [lia |].
          apply PeanoNat.Nat.mul_nonneg_nonneg; lia.
        }
        lia.
      }
    }
  Qed.

  Lemma mod_plus_inv:
    forall (a b c: nat),
      a < c -> b < c ->
      (exists x, x < c /\ (a + x) mod c = b)
      /\
      forall x1 x2, x1 < c -> x2 < c ->
                    (a + x1) mod c = b ->
                    (a + x2) mod c = b -> x1 = x2.
  Proof.
    intros a b c ha hb.
    destruct (mod_plus_inv' ha hb) as [x [hx [hhx hhhx]]].
    split; [exists x; auto|].
    intros x1 x2 h1 h2 hh1 hh2.
    destruct (PeanoNat.Nat.eq_dec x1 x); [|specialize (hhhx _ h1 n); tauto].
    destruct (PeanoNat.Nat.eq_dec x2 x); [|specialize (hhhx _ h2 n); tauto].
    lia.
  Qed.

  Lemma arith_even:
    forall n, n > 0 -> exists k, n * (n - 1) = 2 * k.
  Proof.
    induction n as [| n ih].
    lia.
    intros pos.
    case (PeanoNat.Nat.eq_dec n 0); intros hn.
    exists 0; nia.
    assert (n > 0) by lia.
    specialize (ih H); clear H; destruct ih as [k hk].
    (* S n * (S n - 1)  =   (n+1)* n = n*n + n = n*n + n - n + n
         = n*(n-1) + 2n = 2*k + 2n = 2*(k+n) *)
    exists (k + n).
    nia.
  Qed.

  Close Scope nat_scope.

End Mod.

(** * Minimum *)
Section Min.

  Definition is_minimum (f: nat -> bool) (m: nat): Prop :=
    f m = true /\  forall m',  m' < m -> f m' = false.

  Lemma is_minimum_unique:
    forall f min1 min2, is_minimum f min1 -> is_minimum f min2 ->
                        min1 = min2.
  Proof.
    intros f min1 min2 [M1 h1] [M2 h2].
    destruct (Compare_dec.le_gt_dec min1 min2); [destruct (PeanoNat.Nat.eq_dec min1 min2)|].
    - auto.
    - assert (x: min1 < min2) by lia.
      specialize (h2 _ x); congruence.
    - assert (x: min2 < min1) by lia.
      specialize (h1 _ x); congruence.
  Qed.

  Lemma is_minimum_lt:
    forall f1 min1 f2 min2,
      (forall v1, f1 v1 = true -> exists v2, f2 v2 = true /\ v2 < v1)
      -> is_minimum f1 min1 -> is_minimum f2 min2 -> min2 < min1.
  Proof.
    intros f1 min1 f2 min2 h [M1 M1'] [M2 M2'].
    specialize (h min1 M1); destruct h as [v2 [V2 V2']].  
    assert (hh: min2 <= v2).
    { destruct (Compare_dec.le_gt_dec min2 v2); auto.
      assert (d: v2 < min2) by lia; specialize (M2' _ d); congruence.
    }  
    lia.
  Qed.
  
  Lemma is_minimum_le:
    forall f1 min1 f2 min2,
      (forall v1, f1 v1 = true -> exists v2, f2 v2 = true /\ v2 <= v1)
      -> is_minimum f1 min1 -> is_minimum f2 min2 -> min2 <= min1.
  Proof.
    intros f1 min1 f2 min2 h [M1 M1'] [M2 M2'].
    specialize (h min1 M1); destruct h as [v2 [V2 V2']].  
    assert (hh: min2 <= v2).
    { destruct (Compare_dec.le_gt_dec min2 v2); auto.
      assert (d: v2 < min2) by lia; specialize (M2' _ d); congruence.
    }  
    lia.
  Qed.

  (** ** Pick min {x | f x=true /\ x < n}, return min_so_far if none found *)
  Fixpoint min_below_ (f: nat -> bool) (min_so_far n: nat): nat :=
    match n with
    | 0 => min_so_far
    | S n'=> if f n'
             then min_below_ f n' n'
             else min_below_ f min_so_far n'
    end.

  Lemma min_below__spec:
    forall f msf n,
      (min_below_ f msf n = msf /\ forall x, x < n -> f x = false) \/
      (min_below_ f msf n < n   /\ is_minimum f (min_below_ f msf n)).
  Proof.
    intros f msf n; revert msf; induction n as [| n hn]; simpl.
    { intros msf; left; split; auto.
      intros x hx; lia.
    }
    { intros msf.
      case_eq (f n); intros hf.
      - specialize (hn n); destruct hn as [[hn1 hn2]|[hn1 hn2]].
        + right; split; rewrite hn1; [lia|].
          split; auto.
        + right; split; auto; lia.
      - specialize (hn msf); destruct hn as [[hn1 hn2]|[hn1 hn2]].
        + left; split; auto.
          intros x hx; case (PeanoNat.Nat.eq_dec x n); intros hnx.
          * rewrite hnx; auto.
          * apply hn2; lia.
        + right; split; auto; lia.
    }
  Qed.

  Definition min_below (f: nat -> bool) (n: nat): nat := min_below_ f n n.

  Lemma min_below_spec:
    forall f n,
      (min_below f n = n /\ forall x, x < n -> f x = false) \/
      (min_below f n < n   /\ is_minimum f (min_below f n)).
  Proof. unfold min_below; intros f n; apply min_below__spec. Qed.

End Min.

(** ** Pick argmin: 

   - i.e, pick an element a in a list such that 
     a = argmin { f a, a in l } 
   - where f: A -> nat and l: list A

 *)
Section Argmin.
  Variable (A: Type) (f: A -> nat).
  
  Context {SA: Setoid A}.
  Variable (f_compat: Proper pequiv f).
  Variable (DA: Decider (equiv (A := A))).

  Definition argmin_nat (x y: A): A :=
    if Compare_dec.le_gt_dec (f x) (f y) then x else y.
  
  Definition argmin (l: list A): option A :=
    match l with
    | nil => None
    | a :: l => Some (fold_right argmin_nat a l)
    end.

  Lemma argmin_nat_spec1:
    forall x y, f (argmin_nat x y) <= f x.
  Proof.
    intros x y; unfold argmin_nat.
    case (Compare_dec.le_gt_dec (f x) (f y)); intros h1; lia.
  Qed.
  
  Lemma argmin_nat_spec2:
    forall x y, f (argmin_nat x y) <= f y.
  Proof.
    intros x y; unfold argmin_nat.
    case (Compare_dec.le_gt_dec (f x) (f y)); intros h1; lia.
  Qed.
  
  Lemma argmin_nat_spec3:
    forall x y, {argmin_nat x y == x} + {argmin_nat x y == y}.
  Proof.
    intros x y; unfold argmin_nat.
    destruct (Compare_dec.le_gt_dec (f x) (f y)); [left|right]; auto.
  Qed.

  Lemma argmin_nat_sym:
    forall x y, f x <> f y -> argmin_nat x y = argmin_nat y x.
  Proof.
    intros x y xy; unfold argmin_nat.
    case (Compare_dec.le_gt_dec (f x) (f y)); case (Compare_dec.le_gt_dec (f y) (f x));
      intros h1 h2; try lia; auto.
  Qed.
    
  Lemma argmin_nat_assoc:
    forall x y z, argmin_nat x (argmin_nat y z) =
                  argmin_nat (argmin_nat x y) z.
  Proof.
    intros x y z; unfold argmin_nat.
    case (Compare_dec.le_gt_dec (f x) (f y)); case (Compare_dec.le_gt_dec (f y) (f z));
      intros h1 h2; auto.
    case (Compare_dec.le_gt_dec (f x) (f y)); case (Compare_dec.le_gt_dec (f x) (f z));
      intros h3 h4; auto; lia.
    case (Compare_dec.le_gt_dec (f x) (f y)); intros h3; auto; lia.
    case (Compare_dec.le_gt_dec (f x) (f z)); intros h3; auto; lia.
  Qed.
  
  Instance argmin_nat_compat: Proper pequiv argmin_nat.
  Proof.
    intros a b ab x y xy; unfold argmin_nat.
    case (Compare_dec.le_gt_dec (f a) (f x));
      case (Compare_dec.le_gt_dec (f b) (f y)); intros h h';
        auto; rewrite xy, ab in h'; lia.
  Qed.
    
  Instance argmin_compat: Proper pequiv argmin.
  Proof.
    intros l l' el; unfold argmin.
    destruct l; destruct l'; auto; inversion_clear el.
    simpl; eapply FoldRight.fold_right_compat; auto.
    apply argmin_nat_compat.
    auto.
  Qed.
    
  (* argmin_compat: TODO; with equivlistA if NoDupA (map f l) *)
        
  Definition is_minimum_list (l: list nat) (m: nat): Prop :=
    InA equiv m l /\ forall n, InA equiv n l -> m <= n.

  Instance is_minimum_list_compat: Proper pequiv is_minimum_list.
  Proof.
    intros l1 l2 el m1 m2 em; unfold is_minimum_list.
    apply and_compat; [apply InA_compat_eqlistA; auto| apply all_compat].
    intros x; apply impl_compat.
    apply InA_compat_eqlistA; auto.
    rewrite em; tauto.
  Qed.

  Instance is_minimum_list_compat_equivlistA:
    Proper ((equivlistA (equiv (A := nat))) ==> pequiv) is_minimum_list.
  Proof.
    intros l1 l2 el m1 m2 em; unfold is_minimum_list.
    apply and_compat; [apply InA_compat; auto; apply setoid_equiv
                      | apply all_compat].
    intros x; apply impl_compat.
    apply InA_compat; auto; apply setoid_equiv.
    rewrite em; tauto.
  Qed.

  Definition inclA l l': Prop :=
    forall x, InA equiv x l -> InA equiv x l'.

  Lemma filter_inclA:
    forall (g: A -> bool) (l: list A), inclA (filter g l) l.
  Proof.
    intros g; induction l as [| a l ih]; simpl.
    intros x; rewrite InA_nil; tauto.
    intros x; destruct (g a); intros h; rewrite InA_cons.
    rewrite InA_cons in h; destruct h as [h|h].
    left; auto.
    right; apply ih; auto.
    right; apply ih; auto.
  Qed.

  Lemma argmin_min1:
    forall l a, argmin l == Some a -> InA equiv a l.
  Proof.
    intros l aa; unfold argmin; destruct l as [| a ll];
      intros h.
    { inversion h. }
    simpl in h.
    rewrite <- h; clear h.
    edestruct (FoldRight.fold_right_f_dec (f := argmin_nat)).
    apply argmin_nat_spec3.
    rewrite e; auto. 
    auto. 
  Qed.
    
  Lemma fold_right_min0:
    forall a l, f (fold_right argmin_nat a l) <= f a.
  Proof.
    intros a l; revert a; induction l as [| b l ih];
      intros a; simpl; auto.
    etransitivity.
    2: apply ih.
    apply argmin_nat_spec2.
  Qed.

  Lemma fold_right_min:
    forall x a l, InA equiv x l ->
                  f (fold_right argmin_nat a l) <= f x.
  Proof.
    intros x a l; revert x a; induction l as [| b l ih]; simpl;
      intros x a h.
    { rewrite InA_nil in h; tauto. }
    rewrite InA_cons in h; destruct h as [h|h].
    { rewrite h; apply argmin_nat_spec1. }
    { etransitivity.
      apply argmin_nat_spec2.
      apply ih; auto. }
  Qed.

  Lemma argmin_nat_same:
    forall x y, f x = f y -> argmin_nat x y == x.
  Proof.
    intros x y h; unfold argmin_nat.
    case (Compare_dec.le_gt_dec (f x) (f y)); auto.
    intros hh; lia.
  Qed.

  Lemma argmin_nat_same':
    forall x, argmin_nat x x == x.
  Proof.
    intros x; unfold argmin_nat.
    case (Compare_dec.le_gt_dec (f x) (f x)); auto.
  Qed.

  Lemma fold_right_min':
    forall a l, ~ InA equiv a l ->
                fold_right argmin_nat a l == a ->
                forall x, InA equiv x l -> f a < f x.
  Proof.
    intros a l nin; induction l as [| b l ih];
      intros hfr x inx.
    rewrite InA_nil in inx; tauto.
    rewrite InA_cons in inx; destruct inx as [inx|inx].
    {  simpl in hfr.
       case (PeanoNat.Nat.eq_dec (f b) (f (fold_right argmin_nat a l)));
         intros he.
       assert (TT: argmin_nat b (fold_right argmin_nat a l) == b).
       apply argmin_nat_same; auto.
       rewrite hfr in TT.
       contradict nin; auto.
       destruct (argmin_nat_spec3 b (fold_right argmin_nat a l)).
       rewrite hfr in e.
       contradict nin; auto.
       assert (TT := argmin_nat_spec1 b (fold_right argmin_nat a l)).
       rewrite hfr in TT; rewrite <- inx in TT.
       rewrite <- inx in he.
       rewrite hfr in e; rewrite <- e in he; lia.
    }
    { apply ih; auto.
      simpl in hfr.
      destruct (argmin_nat_spec3 b (fold_right argmin_nat a l)).
      rewrite hfr in e.
      contradict nin; auto.
      rewrite <- e; auto.
    }
  Qed.

  Lemma argmin_nat_l:
    forall a b, f a <= f b -> argmin_nat a b == a.
  Proof.
    intros a b h; unfold argmin_nat.
    case (Compare_dec.le_gt_dec (f a) (f b)); [auto|intros hh; lia].
  Qed.

  Lemma fold_right_min'':
    forall a b l1 l2, 
      ~ InA equiv a l1 ->
      fold_right argmin_nat b (l1 ++ a :: l2) == a ->
      forall n : A, InA equiv n l1 -> f a < f n.
  Proof.
    intros a b l1 l2; induction l1 as [| x l1 ih]; intros hh h.
    intros n; rewrite InA_nil; tauto.
    intros n inn; rewrite InA_cons in inn; destruct inn as [inn|inn].
    { simpl in h.
      case (Compare_dec.le_gt_dec (f n) (f a)); intros R; [|lia].
      rewrite inn in R.
      cut (a == x); [intros ok; contradict hh; auto|].
      apply argmin_nat_l in R.
      rewrite <- R, <- h.
      rewrite argmin_nat_assoc, argmin_nat_same'; auto.
    }
    { apply ih; auto.
      simpl in h.
      destruct (argmin_nat_spec3 x (fold_right argmin_nat b (l1 ++ a :: l2)) ).
      rewrite e in h; contradict hh; symmetry in h; auto.
      rewrite <- e, h; auto.
    }
  Qed.
    
  Lemma argmin_nat_r:
    forall a b, f b < f a -> argmin_nat a b == b.
  Proof.
    intros a b h; unfold argmin_nat.
    case (Compare_dec.le_gt_dec (f a) (f b)); [intros hh; lia|auto].
  Qed.
  
  Lemma fold_right_min'_conv:
    forall a l, ~ InA equiv a l ->
                (forall x, InA equiv x l -> f a < f x) ->
                fold_right argmin_nat a l == a.
  Proof.
    intros a l; induction l as [| b l ih]; auto.
    intros nin h; simpl.
    rewrite ih; auto.
    apply argmin_nat_r.
    cut (f a < f b); [lia|auto].
  Qed.

  Lemma fold_right_min_other:
    forall a b l1 l2,
      f b <= f a ->
      (forall n, InA equiv n l1 -> f n > f b) ->
      (forall n, InA equiv n l2 -> f b <= f n) ->
      fold_right argmin_nat a (l1 ++ b :: l2) == b.
  Proof.
    intros a b l1 l2 ha h1 h2.
    induction l1 as [| x l1 ih].
    { rewrite app_nil_l.
      induction l2 as [| x l2 ih]; simpl; auto.
      apply argmin_nat_l; auto.
      destruct (argmin_nat_spec3
                  x (fold_right argmin_nat a l2)) as [D|D];
        rewrite D; clear D; auto.
      apply argmin_nat_l; auto.
    }
    { simpl.
      rewrite ih; auto.
      apply argmin_nat_r, h1; auto. 
    }
  Qed.

  Lemma argmin_None:
    forall l, argmin l = None <-> l = nil.
  Proof.
    intros [| a l]; unfold argmin.
    tauto.
    split; intros h; inversion h.
  Qed.

  Lemma argmin_Some:
    forall b l a,
      argmin (b :: l) == Some a <->
      (a == b /\ ~ InA equiv a l /\
       forall n, InA equiv n l -> f a < f n)
      \/
      (f a <= f b /\
       exists l1 l2, l == l1 ++ a :: l2 /\ ~ InA equiv a l1 /\
                     (forall n, InA equiv n l1 -> f a < f n) /\
                     (forall n, InA equiv n l2 -> f a <= f n)).
  Proof.
    intros b l a.
    unfold argmin.
    case (InA_dec DA a l); intros ina.
    { split; intros h; simpl in h.
      right.
      split.
      rewrite <- h; apply fold_right_min0.
      assert (exists l1 l2, l == l1 ++ a :: l2 /\ ~ InA equiv a l1).
      { clear h b; induction l as [| b l ih].
        rewrite InA_nil in ina; tauto.
        rewrite InA_cons in ina; destruct ina as [ina|ina].
        exists nil; exists l; simpl.
        split; [right; [symmetry; auto|reflexivity]|
                rewrite InA_nil; tauto].
        case (DA a b); intros ha.
        exists nil; exists l; split.
        simpl; right; [symmetry; auto|reflexivity].
        rewrite InA_nil; tauto.
        specialize (ih ina); destruct ih as [l1 [l2 [h hh]]].
        exists (b :: l1); exists l2.
        split.
        simpl; right; auto.
        contradict hh; rewrite InA_cons in hh; destruct hh; [tauto|auto].
      }
      { destruct H as [l1 [l2 [hl hll]]].
        exists l1; exists l2; split; [|split]; auto.
        rewrite hl in h; clear hl.
        split.
        eapply fold_right_min''; auto; apply h.
        intros n hn; rewrite <- h; apply fold_right_min.
        rewrite InA_app; right; right; auto.
      }
      { destruct h as [[ _ [h _]]|
                       [ha [l1 [l2 [hl [inn [hf hff]]]]]]]; [tauto|].
        simpl; rewrite hl; apply fold_right_min_other; auto.
      }
    }
    { split; intros h.
      assert (a == b).
      simpl in h; rewrite <- h.
      destruct (FoldRight.fold_right_f_dec argmin_nat_spec3 b l); auto.
      rewrite h in i; tauto.
      left; split; [| split]; auto.
      apply fold_right_min'; auto.
      rewrite H at 1; auto.
      destruct h as [h|[_ [l1 [l2 [hl _]]]]].
      2: { contradict ina; rewrite hl.
           rewrite InA_app; auto. }
      destruct h as [ha [_ hf]].
      apply fold_right_min'_conv in hf; auto.
      simpl; rewrite <- ha; auto.
    }
  Qed.
    
  Lemma argmin_min2:
    forall l a, argmin l == Some a ->
                forall n, InA equiv n l -> f a <= f n.
  Proof.
    intros l a; unfold argmin; destruct l as [| b l].
    simpl; tauto.
    intros h; simpl in h.
    intros x ina.
    rewrite <- h; clear h.
    rewrite InA_cons in ina; destruct ina as [ina|ina].
    rewrite ina; apply fold_right_min0.
    apply fold_right_min; auto.
  Qed.

  (*
  Lemma argmin_Some:
    forall l a, argmin l == Some a ->
                InA equiv a l /\                
                forall n, InA equiv n l -> f a <= f n.
  Proof.
    intros l a h; split.
    apply argmin_min1; auto; rewrite h; auto.
    apply argmin_min2; auto; rewrite h; auto.
  Qed.

                                                     
  
  Lemma argmin_le_first:
    forall a l m, argmin (a :: l) == Some m -> f m <= f a.
  Proof.
    intros a l m h; apply argmin_Some in h.
    destruct h as [_ h]; apply h; auto.
  Qed.
  
  Lemma argmin_first:
    forall a l, ~ InA equiv a l -> argmin (a :: l) == Some a ->
                forall x, InA equiv x l -> f a < f x.
  Proof.
    intros a l nin ha x inx.
    unfold argmin in ha; simpl in ha.
    eapply fold_right_min'.
    2: apply ha.
    auto.
    auto.    
  Qed.    

  Lemma argmin_first':
    forall a l, ~ InA equiv a l -> 
                (forall x, InA equiv x l -> f a < f x) ->
                argmin (a :: l) == Some a.
  Proof.
    intros a l nin h.
    unfold argmin; simpl.
    apply fold_right_min'_conv; auto.
  Qed.    
*)    


End Argmin.

Section Range.

  (** List of consecutive values from 0 to n (n is excluded) *)
  Fixpoint range (n: nat): list nat :=
    match n with
      0 => nil
    | S m => m :: range m
    end.

  (** Specification *)
  Lemma In_range: forall n x, In x (range n) <-> x < n.
    intro n; induction n as [|n hn]; split; simpl;
      intuition try lia.
    rewrite hn in H0; lia.
    destruct (Compare_dec.le_lt_eq_dec H); intuition try lia.
    assert (hh: x < n) by lia.
    rewrite <- hn in hh; right; auto.
  Qed.

End Range.


(** * Triangular numbers  triangle n = sigma(i,i=0..n) *)

Section Triangular_Numbers.

  Open Scope nat_scope.
  
  Fixpoint triangle n :=
    match n with
      0 => 0
    | S n => S n + triangle n
    end.

  Lemma triangle_0 : triangle 0 = 0.
  Proof.
    reflexivity.
  Qed.

  Lemma triangle_S : forall n, triangle (S n) = S n + triangle n.
  Proof.
    reflexivity.
  Qed.

  Lemma triangle_half_rectangle:
    forall n, 2*triangle n = n*(n+1).
  Proof.
    intro n; induction n as [|n].
    - reflexivity.
    - simpl triangle.
      ring_simplify.
      rewrite IHn;clear IHn.
      ring_simplify;reflexivity.
  Qed.

  Lemma triangle_half_rectangle_pn:
    forall n,  2*triangle (n-1) = n*(n-1).
  Proof.
    intro n; rewrite triangle_half_rectangle.
    destruct n;simpl;ring_simplify.
    reflexivity.
    rewrite PeanoNat.Nat.sub_0_r;reflexivity.
  Qed.

  Lemma triangle_n : forall n, triangle n = n*(n+1)/2.
  Proof.
    intro n.
    assert (h:=triangle_half_rectangle n).
    apply PeanoNat.Nat.div_unique_exact;lia.
  Qed.

  Lemma triangle_pn: forall n, triangle (n-1) = n*(n-1)/2.
  Proof.
    destruct n.
    reflexivity.
    rewrite PeanoNat.Nat.mul_comm.
    simpl PeanoNat.Nat.sub.
    rewrite PeanoNat.Nat.sub_0_r.
    rewrite (triangle_n n).
    rewrite PeanoNat.Nat.add_comm.
    reflexivity.
  Qed.

  Lemma triangle_le: forall n, n<= triangle n.
  Proof.
    destruct n;simpl;lia.
  Qed.

  Close Scope nat_scope.

End Triangular_Numbers.

Section ltk_inj.

  Variable k: nat.

  Lemma k_non_0: 0<=k-1.
  Proof.
    lia.
  Qed.

  (** ltk_inj turns a nat into a bounded integer if possible 
      (else truncate sit to 0) *)

  Definition ltk_inj i : {i | i <= k-1} :=
    match Compare_dec.le_dec i (k-1) with
      left GT => @exist _ _ i GT
    | right _ => @exist _ _ 0 k_non_0
    end.

  Instance ltk_inj_compat:Proper fequiv ltk_inj.
  Proof.
    intros x1 x2 ex.
    hnf in ex;rewrite ex;reflexivity.
  Qed.

  Lemma  proj1_sig_ltk_inj: forall n, n<k -> proj1_sig (ltk_inj n) = n.
    intros n hn; unfold ltk_inj.
    destruct (Compare_dec.le_dec n (k - 1)) as [LE|GT];simpl;try lia.
  Qed.

End ltk_inj.

Close Scope nat_scope.
Close Scope signature_scope.
Unset Implicit Arguments.
