(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** **** Global imports **)
From Coq Require Import SetoidList.
From Coq Require Import ZArith.
From Coq Require Import Bool.
From Coq Require Import BoolEq.
From Coq Require Import Morphisms.
From Coq Require Import RelationPairs.
From Coq Require Import Setoid.
From Coq Require Import SetoidClass.

(** **** Local imports **)
From Padec Require Import SetoidUtil.
From Padec Require Import ListUtils.

Open Scope Z_scope.

(* end show *)
Open Scope signature_scope.
Set Implicit Arguments.


(** (Strict) Partial Order Relation *)
Class POR {A} eqA `{equ: Equivalence A eqA} (R: relation A): Prop :=
  {
    POR_Antisymmetric :> Antisymmetric A eqA R | 4 ;
    POR_Transitive :> Transitive R | 4
  }.

Section With_A.

  Context {A} `{SA: Setoid A}.

  Section RelationA.

    Definition subrelationA (R1 R2: relation A): Prop :=
      subrelation R1 R2 /\
      Proper fequiv R1 /\ Proper fequiv R2.

    Global Instance subrelationA_POR: POR subrelationA.
    Proof.
      firstorder.
    Qed.

    Lemma relationA_PER_POR: subrelation fequiv subrelationA.
    Proof.
      unfold subrelationA, subrelation; simpl.
      intros PA PB epb; intuition; try proper_.
      revert H; eapply proj1, epb; reflexivity.
    Qed.

  End RelationA.

  (***************)
  (** ** Closure *)
  (***************)
  Section Closure.

    Context
      (P: relation A -> Prop)
      `{P_compat: Proper _ (fequiv ==> iff) P}
      (C: relation A -> relation A)
      `{C_compat: Proper _ (fequiv ==> fequiv) C}
    .

    (****************************)
    (** **** Closure hypotheses *)
    (****************************)
    Section Closure_Hypotheses.

      Definition closure_superrelation :=
        forall R, Proper fequiv R -> subrelationA R (C R).

      Definition closure_minimal :=
        forall R1 R2, subrelationA R1 R2 ->
                      P R2 ->
                      subrelationA (C R1) R2.

      Definition closure_valid :=
        forall R, Proper fequiv R -> P (C R).

    End Closure_Hypotheses.

    Context
      (superrelation: closure_superrelation)
      (minimal: closure_minimal)
      (valid: closure_valid)
    .

    (****************************)
    (** **** Closure properties *)
    (****************************)
    Section Closure_Monotone.

      Instance closure_monotone: Proper (subrelationA ==> subrelationA) C.
      Proof.
        unfold Proper.
        unfold respectful.
        intros R1 R2 R1_sub_R2.
        apply minimal; blast.
        transitivity R2.
        assumption.
        apply superrelation.
        apply (proj2 (proj2 R1_sub_R2)).
        apply valid.
        destruct R1_sub_R2 as [_ [_ R2_compat]].
        assumption.
      Qed.

    End Closure_Monotone.

    Section Closure_Of_Valid_Is_Subrelation.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
        (R_valid: P R)
      .

      Lemma closure_of_valid_subrelation: subrelation (C R) R.
      Proof.
        apply minimal; blast.
        apply relationA_PER_POR; blast.
      Qed.

    End Closure_Of_Valid_Is_Subrelation.

    Section Closure_Of_Valid_Is_Eq.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
        (R_valid: P R)
      .

      Lemma closure_of_valid_indentity: fequiv R (C R).
      Proof.
        assert (sub := closure_of_valid_subrelation R_valid).
        assert (super := superrelation R_compat).
        cbv in sub.
        cbv in super.
        cbv.
        intros x1 x2 eqx y1 y2 eqy.
        rewrite eqx.
        rewrite eqy.
        split.
        intros Rx1x2.
        apply super.
        assumption.
        intros Cx2y2.
        apply sub; blast.
      Qed.

    End Closure_Of_Valid_Is_Eq.

  End Closure.

  (*************************)
  (** ** Reflexive closure *)
  (*************************)
  Section Reflexive_Closure.

    (********************)
    (** **** Definition *)
    (********************)
    Section Reflexive_Closure_Definition.
      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
      .

      Definition reflexive_closure x y := x == y \/ R x y.

      Global Instance reflexive_closure_compat: Proper fequiv reflexive_closure.
      Proof.
        unfold Proper.
        unfold respectful.
        assert (half_lemma: forall x1 x2: A,
                   x1 == x2 ->
                   forall y1 y2: A,
                     y1 == y2 -> (reflexive_closure x1 y1 ->
                                  reflexive_closure x2 y2)
               ).
        intros x1 x2 eqx y1 y2 eqy Sx1y1.
        unfold reflexive_closure in *.
        destruct Sx1y1 as [Ex1y1 | Rx1y1];
          [left | right]; rewrite <- eqx; rewrite <- eqy; assumption.
        split; apply half_lemma; blast.
      Qed.


    End Reflexive_Closure_Definition.

    (****************************)
    (** **** Closure hypotheses *)
    (****************************)
    Section Reflexive_Closure_Superrelation.

      Lemma reflexive_closure_superrelation:
        closure_superrelation reflexive_closure.
      Proof.
        unfold closure_superrelation.
        intros R R_compat.
        unfold subrelationA.
        split.
        cbv.
        intros x y Rxy.
        right.
        assumption.
        split.
        assumption.
        apply reflexive_closure_compat; blast.
      Qed.

    End Reflexive_Closure_Superrelation.

    Section Reflexive_Closure_Minimal.

      Lemma reflexive_closure_minimal:
        closure_minimal Reflexive reflexive_closure.
      Proof.
        unfold closure_minimal.
        intros R1 R2 sub R2_refl.
        unfold subrelationA.
        destruct sub as [sub [R1_compat R2_compat]].
        split.
        cbv.
        intros x y [Exy | R1xy].
        rewrite Exy.
        reflexivity.
        apply sub.
        assumption.
        split.
        apply reflexive_closure_compat; blast.
        assumption.
      Qed.

    End Reflexive_Closure_Minimal.

    Section Reflexive_Closure_Valid.

      Lemma reflexive_closure_valid: closure_valid Reflexive reflexive_closure.
      Proof.
        unfold closure_valid.
        intros R R_compat.
        unfold Reflexive.
        intros x.
        unfold reflexive_closure.
        left.
        reflexivity.
      Qed.

    End Reflexive_Closure_Valid.

    (****************************)
    (** **** Closure properties *)
    (****************************)
    Section Reflexive_Closure_Monotone.

      Instance reflexive_closure_monotone:
        Proper (subrelationA ==> subrelationA) reflexive_closure.
      Proof.
        eapply closure_monotone.
        apply reflexive_closure_superrelation.
        apply reflexive_closure_minimal.
        apply reflexive_closure_valid.
      Qed.

    End Reflexive_Closure_Monotone.

    (* TODO add other properties *)

    (**************************************)
    (** **** Reflexive_Closure_Properties *)
    (**************************************)
    Section Reflexive_Closure_Reflexive.

      Context
        (R : relation A)
        `{R_compat : Proper _ fequiv R}.

      Instance reflexive_closure_reflexive : Reflexive (reflexive_closure R).
      Proof.
        apply reflexive_closure_valid; blast.
      Qed.


    End Reflexive_Closure_Reflexive.

    Section Reflexive_Closure_Transitive.

      Context
        (R : relation A)
        `{R_compat : Proper _ fequiv R}
        `{R_trans : Transitive _ R}
      .

      Instance reflexive_closure_transitive: Transitive (reflexive_closure R).
      Proof.
        unfold Transitive in *.
        intros x y z Sxy Syz.
        unfold reflexive_closure in *.
        destruct Sxy as [Exy | Rxy];
          destruct Syz as [Eyz | Ryz].
        left; transitivity y; assumption.
        right; rewrite Exy; assumption.
        right; rewrite <- Eyz; assumption.
        right; apply (R_trans Rxy); assumption.
      Qed.


    End Reflexive_Closure_Transitive.

    Section Reflexive_Closure_Symmetric.

      Context
        (R : relation A)
        `{R_compat : Proper _ fequiv R}
        `{R_symm : Symmetric _ R}
      .

      Global Instance reflexive_closure_symmetric: Symmetric (reflexive_closure R).
      Proof.
        unfold Symmetric in *.
        intros x y Sxy.
        unfold reflexive_closure in *.
        destruct Sxy as [Exy | Rxy].
        left; symmetry; assumption.
        right; apply R_symm; assumption.
      Qed.

    End Reflexive_Closure_Symmetric.

    Section Reflexive_Closure_Dec.
      Context
        (eqA_dec: Decider (equiv (A := A)))
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
        (R_dec: Decider R)
      .

      Lemma reflexive_closure_dec: Decider (reflexive_closure R).
      Proof.
        unfold Decider.
        intros x y.
        unfold reflexive_closure.
        destruct (eqA_dec x y) as [eq | neq].
        left; left; assumption.
        destruct (R_dec x y) as [r | nr].
        left; right; assumption.
        right; intros [eq | r]; [apply neq; assumption | apply nr; assumption].
      Qed.

    End Reflexive_Closure_Dec.

  End Reflexive_Closure.

  (***************************)
  (** *** Transitive closure *)
  (***************************)
  Section Transitive_Closure.

    (********************)
    (** **** Definition *)
    (********************)
    Section Transitive_Closure_Definition.

      Inductive transitive_closure (R: relation A): relation A :=
      | transitive_closure_step (x y: A): R x y -> transitive_closure R x y
      | transitive_closure_transitivity (x y z: A):
          transitive_closure R x y -> transitive_closure R y z ->
          transitive_closure R x z
      .

      Global Instance transitive_closure_compat :
        Proper fequiv transitive_closure.
      Proof.
        intros R1 R2 Rc x1 x2 eqx y1 y2 eqy.
        split; [intros Sx1y1 | intros Sx2y2].

        revert x2 y2 eqx eqy.
        induction Sx1y1 as [x y Rxy| x y z Sxy Sx2y2 Syz Sy2z2];
          intros x2 y2 eqx eqy.
        left;  rewrite <- (Rc _ _ eqx _ _ eqy); assumption.
        specialize (Sx2y2 x2 y eqx (Equivalence_Reflexive y)).
        specialize (Sy2z2 y y2 (reflexivity y) eqy).
        apply (transitive_closure_transitivity Sx2y2 Sy2z2).

        revert x1 y1 eqx eqy.
        induction Sx2y2 as [x y Rxy| x y z Sxy Sx1y1 Syz Sy1z1];
          intros x2 y2 eqx eqy.
        left;  rewrite (Rc _ _ eqx _ _ eqy); assumption.
        specialize (Sx1y1 x2 y eqx (Equivalence_Reflexive y)).
        specialize (Sy1z1 y y2 (reflexivity y) eqy).
        apply (transitive_closure_transitivity Sx1y1 Sy1z1).
      Qed.

    End Transitive_Closure_Definition.

    Section Transitive_Closure_Left_Definition.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
      .

      Inductive transitive_closure_left: relation A :=
      | transitive_closure_left_step (x y: A):
          R x y -> transitive_closure_left x y
      | transitive_closure_left_transitivity (x y z: A):
          R x y -> transitive_closure_left y z -> transitive_closure_left x z
      .

      Instance transitive_closure_left_compat:
        Proper fequiv transitive_closure_left.
      Proof.
        unfold Proper.
        unfold fequiv.
        unfold respectful.
        assert (half_lemma: forall x1 x2: A,
                   x1 == x2 ->
                   forall z1 z2: A,
                     z1 == z2 ->
                     (transitive_closure_left x1 z1 ->
                      transitive_closure_left x2 z2)
               ).
        intros x1 x2 eqx z1 z2 eqz Cx1y1.
        revert x2 eqx z2 eqz.
        induction Cx1y1 as [x1 z1 Rx1z1 | x1 y1 z1 Rx1y1 Ry1z1 Hrec];
          intros x2 eqx z2 eqz.
        left.
        rewrite <- eqx.
        rewrite <- eqz.
        assumption.
        rewrite eqx in Rx1y1.
        apply (transitive_closure_left_transitivity Rx1y1).
        apply Hrec.
        reflexivity.
        assumption.
        split; apply half_lemma; blast.
      Qed.

    End Transitive_Closure_Left_Definition.

    Section Transitive_Closure_Left_Iff_Transitive_Closure.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
      .

      Lemma transitive_closure_left_iff_transitive_closure:
        forall x y,
          (transitive_closure_left R) x y <-> (transitive_closure R) x y
      .
      Proof.
        assert (left_to_right: forall x y,
                   (transitive_closure_left R) x y -> (transitive_closure R) x y
               ).
        intros x z.
        intros Cxz.
        induction Cxz as [x z Rxz | x y z Rxy Cyz Hrec].
        left; assumption.
        apply (transitive_closure_transitivity (y:=y)).
        left; assumption.
        assumption.
        intros x z.
        split.
        apply left_to_right.
        intros Cxz.
        induction Cxz as [x z Rxz | x y z Cxy C'xy Cyz C'yz].
        left; assumption.
        induction C'xy as [x y Rxy | x t y Rxt C'tz Hrec].
        apply (transitive_closure_left_transitivity x Rxy C'yz).
        apply (transitive_closure_left_transitivity x Rxt).
        apply Hrec; blast.
      Qed.

    End Transitive_Closure_Left_Iff_Transitive_Closure.

    Section Transitive_Closure_Right_Definition.
      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
      .

      Inductive transitive_closure_right: relation A :=
      | transitive_closure_right_step (x y: A): R x y ->
                                                transitive_closure_right x y
      | transitive_closure_right_transitivity (x y z: A):
          transitive_closure_right x y -> R y z -> transitive_closure_right x z
      .

      Global Instance transitive_closure_right_compat:
        Proper fequiv transitive_closure_right.
      Proof.
        unfold Proper.
        unfold fequiv.
        unfold respectful.
        assert (half_lemma: forall x1 x2: A,
                   x1 == x2 ->
                   forall z1 z2: A,
                     z1 == z2 ->
                     (transitive_closure_right x1 z1 ->
                      transitive_closure_right x2 z2)
               ).
        intros x1 x2 eqx z1 z2 eqz Cx1z1.
        revert z2 eqz.
        induction Cx1z1 as [x1 z1 Rx1z1 | x1 y1 z1 Cx1y1 Ry1z1]; intros z2 eqz.
        left; rewrite <- eqx; rewrite <- eqz; assumption.
        apply (transitive_closure_right_transitivity (y := y1)).
        apply Ry1z1; blast.
        rewrite <- eqz; assumption.
        split; apply half_lemma; blast.
      Qed.

    End Transitive_Closure_Right_Definition.

    Section Transitive_Closure_Right_Iff_Transitive_Closure.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
      .

      Lemma transitive_closure_right_iff_transitive_closure:
        forall x y,
          (transitive_closure_right R) x y <-> (transitive_closure R) x y
      .
      Proof.
        assert (left_to_right: forall x y,
                   (transitive_closure_right R) x y -> (transitive_closure R) x y
               ).
        intros x z.
        intros Cxz.
        induction Cxz as [x z Rxz | x y z Cxy Ryz C'xy].
        left; assumption.
        apply (transitive_closure_transitivity (y := y)); blast.
        left; assumption.
        intros x z.
        split.
        apply left_to_right.
        intros Cxz.
        induction Cxz as [x z Rxz | x y z Cxy C'xy Cyz C'yz].
        left; assumption.
        induction C'yz as [y z Ryz | y t z Ryt Hrec Rtz].
        apply (transitive_closure_right_transitivity (y := y)); assumption.
        apply (transitive_closure_right_transitivity (y := t)).
        apply Hrec; blast.
        assumption.
      Qed.

    End Transitive_Closure_Right_Iff_Transitive_Closure.

    (****************************)
    (** **** Closure hypotheses *)
    (****************************)
    Section Transitive_Closure_Superrelation.

      Lemma transitive_closure_superrelation:
        closure_superrelation transitive_closure.
      Proof.
        unfold closure_superrelation.
        intros R R_compat.
        split.
        unfold subrelation.
        unfold predicate_implication.
        unfold pointwise_lifting.
        unfold impl.
        intros x y Rxy.
        left.
        assumption.
        split.
        assumption.
        apply transitive_closure_compat;
          assumption.
      Qed.

    End Transitive_Closure_Superrelation.

    Section Transitive_Closure_Minimal.

      Lemma transitive_closure_minimal:
        closure_minimal Transitive transitive_closure.
      Proof.
        intros R1 R2 [sub [R1_compat R2_compat]] trans_R2.
        split.
        cbv.
        intros x z Cxz.
        induction Cxz as [x z R1xz | x y z Cxy R2xy Cyz R2yz].
        apply sub; assumption.
        transitivity y; assumption.
        split.
        apply transitive_closure_compat;
          assumption.
        assumption.
      Qed.

    End Transitive_Closure_Minimal.

    Section Transitive_Closure_Valid.

      Lemma transitive_closure_valid:
        closure_valid Transitive transitive_closure.
      Proof.
        unfold closure_valid.
        intros R R_compat.
        unfold Transitive.
        intros x y z Sxy Syz.
        apply (transitive_closure_transitivity Sxy Syz).
      Qed.

    End Transitive_Closure_Valid.

    (****************************)
    (** **** Closure properties *)
    (****************************)

    Section Transitive_Closure_Monotone.

      Instance transitive_closure_monotone:
        Proper (subrelationA ==> subrelationA) transitive_closure.
      Proof.
        eapply closure_monotone.
        apply transitive_closure_superrelation.
        apply transitive_closure_minimal.
        apply transitive_closure_valid.
      Qed.

    End Transitive_Closure_Monotone.

    (***************************************)
    (** **** Transitive closure properties *)
    (***************************************)
    Section Transitive_Closure_Reflexive.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
        `{R_refl: Reflexive _ R}
      .

      Instance transitive_closure_reflexive: Reflexive (transitive_closure R).
      Proof.
        unfold Reflexive.
        intros x.
        left.
        reflexivity.
      Qed.


    End Transitive_Closure_Reflexive.

    (* TODO: forall x, in_cycle x \/ R x x <-> 
                       Reflexive (transitive_closure R) *)

    Section Transitive_Closure_Transitive.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
      .

      Instance transitive_closure_transitive: Transitive (transitive_closure R).
      Proof.
        apply transitive_closure_valid; blast.
      Qed.


    End Transitive_Closure_Transitive.

    Section Transitive_Closure_Symmetric.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
        `{R_symm: Symmetric _ R}
      .
      Instance transitive_closure_symmetric: Symmetric (transitive_closure R).
      Proof.
        unfold Symmetric in *.
        intros x z Cxz.
        induction Cxz as [x z R1xz | x y z Cxy Cyx Cyz Czy].
        left; apply R_symm; assumption.
        apply (transitive_closure_transitivity Czy Cyx).
      Qed.

    End Transitive_Closure_Symmetric.

  End Transitive_Closure.

  (**************************)
  (** *** Symmetric closure *)
  (**************************)
  Section Symmetric_Closure.

    (********************)
    (** **** Definition *)
    (********************)
    Section Symmetric_Closure_Definition.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
      .

      Definition symmetric_closure x y := R x y \/ R y x.

      Global Instance symmetric_closure_compat: Proper fequiv symmetric_closure.
      Proof.
        unfold Proper.
        unfold respectful.
        assert (half_lemma: forall x1 x2: A,
                   x1 == x2 ->
                   forall y1 y2: A,
                     y1 == y2 -> (symmetric_closure x1 y1 ->
                                  symmetric_closure x2 y2)
               ).
        intros x1 x2 eqx y1 y2 eqy.
        unfold symmetric_closure.
        intros h.
        rewrite eqx in h.
        rewrite eqy in h.
        assumption.
        split; apply half_lemma; blast.
      Qed.


    End Symmetric_Closure_Definition.

    (****************************)
    (** **** Closure hypotheses *)
    (****************************)
    Section Symmetric_Closure_Superrelation.

      Lemma symmetric_closure_superrelation:
        closure_superrelation symmetric_closure.
      Proof.
        unfold closure_superrelation.
        intros R R_compat.
        split.
        cbv.
        intros x y Rxy.
        left; assumption.
        split.
        assumption.
        apply symmetric_closure_compat;
          assumption.
      Qed.

    End Symmetric_Closure_Superrelation.

    Section Symmetric_Closure_Minimal.

      Lemma symmetric_closure_minimal:
        closure_minimal Symmetric symmetric_closure.
      Proof.
        unfold closure_minimal.
        intros R1 R2 sub R2_symm.
        split.
        cbv.
        intros x y [R1xy | R1yx].
        apply sub; assumption.
        symmetry; apply sub; assumption.
        destruct sub as [_ [R1_compat R2_compat]].
        split.
        apply symmetric_closure_compat;
          assumption.
        assumption.
      Qed.

    End Symmetric_Closure_Minimal.

    Section Symmetric_Closure_Valid.

      Lemma symmetric_closure_valid: closure_valid Symmetric symmetric_closure.
      Proof.
        unfold closure_valid.
        intros R R_compat.
        unfold Symmetric.
        intros x y Cxy.
        unfold symmetric_closure in *.
        destruct Cxy as [Rxy | Ryx]; tauto.
      Qed.

    End Symmetric_Closure_Valid.

    (****************************)
    (** **** Closure properties *)
    (****************************)

    Section Symmetric_Closure_Monotone.

      Instance symmetric_closure_monotone:
        Proper (subrelationA ==> subrelationA) symmetric_closure.
      Proof.
        eapply closure_monotone.
        apply symmetric_closure_superrelation.
        apply symmetric_closure_minimal.
        apply symmetric_closure_valid.
      Qed.

    End Symmetric_Closure_Monotone.

    (**************************************)
    (** **** Symmetric closure properties *)
    (**************************************)
    Section Symmetric_Closure_Reflexive.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
        `{R_refl: Reflexive _ R}
      .

      Instance symmetric_closure_reflexive: Reflexive (symmetric_closure R).
      Proof.
        unfold Reflexive.
        intros x.
        unfold symmetric_closure.
        left.
        apply R_refl.
      Qed.

    End Symmetric_Closure_Reflexive.

    Section Symmetric_Closure_Transitive.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
      .

      (* TODO: (Transitive (symmetric_closure R)) <-> 
               Symmetric R /\ Transitive R (not sure it's true) *)

      Definition symmetric_closure_transitive:
        False -> Transitive (symmetric_closure R).
      Proof.
        intros false.
        elim false.
      Qed.

    End Symmetric_Closure_Transitive.

    Section Symmetric_Closure_Symmetric.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
      .

      Instance symmetric_closure_symmetric: Symmetric (symmetric_closure R).
      Proof.
        apply symmetric_closure_valid; blast.
      Qed.

    End Symmetric_Closure_Symmetric.

    Section Symmetric_Closure_Dec.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
        (R_dec: Decider R)
      .

      Lemma symmetric_closure_dec: Decider (symmetric_closure R).
      Proof.
        unfold Decider.
        intros x y.
        unfold symmetric_closure.
        destruct (R_dec x y) as [Rxy | nRxy].
        left; left; assumption.
        destruct (R_dec y x) as [Ryx | nRyx].
        left; right; assumption.
        right; intros [Rxy | Ryx]; [apply nRxy | apply nRyx]; assumption.
      Qed.

    End Symmetric_Closure_Dec.

  End Symmetric_Closure.

  (**************************************)
  (** *** Reflexive transitive closure *)
  (**************************************)
  Section Reflexive_Transitive_Closure.

    (********************)
    (** **** Definition *)
    (********************)
    Section Reflexive_Transitive_Closure_Definition.

      Context  (R: relation A) `{R_compat: Proper _ fequiv R}.

      Definition reflexive_transitive_closure :=
        reflexive_closure (transitive_closure R).

      Instance reflexive_transitive_closure_compat:
        Proper fequiv reflexive_transitive_closure.
      Proof.
        unfold reflexive_transitive_closure.
        unfold Proper.
        intros x1 x2 eqx y1 y2 eqy.
        rewrite eqx.
        rewrite eqy.
        reflexivity.
      Qed.

    End Reflexive_Transitive_Closure_Definition.

    (****************************)
    (** **** Closure hypotheses *)
    (****************************)
    Section Reflexive_Transitive_Closure_Superrelation.

      Lemma reflexive_transitive_closure_superrelation:
        closure_superrelation reflexive_transitive_closure.
      Proof.
        unfold closure_superrelation.
        intros R R_compat.
        unfold reflexive_transitive_closure.
        transitivity (transitive_closure R).
        apply transitive_closure_superrelation;
          assumption.
        apply reflexive_closure_superrelation.
        apply transitive_closure_compat;
          assumption.
      Qed.

    End Reflexive_Transitive_Closure_Superrelation.

    Section Reflexive_Transitive_Closure_Minimal.

      Lemma reflexive_transitive_closure_minimal:
        closure_minimal PreOrder reflexive_transitive_closure.
      Proof.
        unfold closure_minimal.
        intros R1 R2 sub [R2_relf R2_trans].
        unfold reflexive_transitive_closure.
        apply reflexive_closure_minimal; blast.
        apply transitive_closure_minimal; blast.
      Qed.

    End Reflexive_Transitive_Closure_Minimal.

    Section Reflexive_Transitive_Closure_Valid.

      Lemma reflexive_transitive_closure_valid:
        closure_valid PreOrder reflexive_transitive_closure.
      Proof.
        unfold closure_valid.
        intros R R_compat.
        unfold reflexive_transitive_closure.
        split.
        apply reflexive_closure_valid.
        apply transitive_closure_compat;
          assumption.
        apply reflexive_closure_transitive.
        apply transitive_closure_compat;
          assumption.
        apply transitive_closure_valid.
        assumption.
      Qed.

    End Reflexive_Transitive_Closure_Valid.

    (****************************)
    (** **** Closure properties *)
    (****************************)
    Section Reflexive_Transitive_Closure_Monotone.

      Instance reflexive_transitive_closure_monotone:
        Proper (subrelationA ==> subrelationA) reflexive_transitive_closure.
      Proof.
        eapply closure_monotone.
        apply reflexive_transitive_closure_superrelation.
        apply reflexive_transitive_closure_minimal.
        apply reflexive_transitive_closure_valid.
      Qed.

    End Reflexive_Transitive_Closure_Monotone.

    (*************************************************)
    (** **** Reflexive transitive closure properties *)
    (*************************************************)
    Section Reflexive_Transitive_Closure_Reflexive.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
      .

      Instance reflexive_symmetric_closure_reflexive:
        Reflexive (reflexive_transitive_closure R).
      Proof.
        apply (reflexive_transitive_closure_valid);
          assumption.
      Qed.


    End Reflexive_Transitive_Closure_Reflexive.

    Section Reflexive_Transitive_Closure_Transitive.
      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
      .

      Definition reflexive_transitive_closure_transitive:
        Transitive (reflexive_transitive_closure R).
      Proof.
        unfold Transitive.
        intros x y z Cxy Cyz.
        unfold reflexive_transitive_closure in *.
        unfold reflexive_closure in *.
        destruct Cxy as [Exy | C'xy];
          destruct Cyz as [Eyz | C'yz].
        left; transitivity y; assumption.
        right; rewrite Exy; assumption.
        right; rewrite <- Eyz; assumption.
        right; apply (transitive_closure_transitivity C'xy C'yz).
      Qed.


    End Reflexive_Transitive_Closure_Transitive.

    Section Reflexive_Transitive_Closure_Symmetric.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
        `{R_symm: Symmetric _ R}
      .

      Definition reflexive_transitive_closure_symmetric:
        Symmetric (reflexive_transitive_closure R).
      Proof.
        unfold reflexive_transitive_closure.
        apply reflexive_closure_symmetric.
        apply transitive_closure_symmetric;
          assumption.
      Qed.


    End Reflexive_Transitive_Closure_Symmetric.

    Section Reflexive_Transitive_Closure_Dec.
      (* TODO *)
    End Reflexive_Transitive_Closure_Dec.


  End Reflexive_Transitive_Closure.

  (***********************************************)
  (** *** Reflexive transitive symmetric closure *)
  (***********************************************)
  Section Reflexive_Transitive_Symmetric_Closure.

    (********************)
    (** **** Definition *)
    (********************)
    Section Reflexive_Transitive_Symmetric_Closure_Definition.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
      .

      Definition reflexive_transitive_symmetric_closure :=
        (reflexive_transitive_closure (symmetric_closure R)).

      (* TODO: use general theorem for Proper of composition? *)
      Definition  reflexive_transitive_symmetric_closure_compat:
        Proper fequiv reflexive_transitive_symmetric_closure.
      Proof.
        unfold reflexive_transitive_symmetric_closure,
        reflexive_transitive_closure.
        unfold Proper.
        unfold respectful.
        intros x1 x2 eqx y1 y2 eqy.
        rewrite eqx.
        rewrite eqy.
        reflexivity.
      Qed.

    End Reflexive_Transitive_Symmetric_Closure_Definition.

    (****************************)
    (** **** Closure hypotheses *)
    (****************************)
    Section Reflexive_Transitive_Symmetric_Closure_Superrelation.

      Lemma reflexive_transitive_symmetric_closure_superrelation:
        closure_superrelation reflexive_transitive_symmetric_closure.
      Proof.
        unfold closure_superrelation.
        intros R R_compat.
        unfold reflexive_transitive_symmetric_closure.
        transitivity (symmetric_closure R).
        apply symmetric_closure_superrelation;
          assumption.
        apply reflexive_transitive_closure_superrelation.
        apply symmetric_closure_compat;
          assumption.
      Qed.

    End Reflexive_Transitive_Symmetric_Closure_Superrelation.

    Section Reflexive_Transitive_Symmetric_Closure_Minimal.

      Lemma reflexive_transitive_symmetric_closure_minimal:
        closure_minimal Equivalence reflexive_transitive_symmetric_closure.
      Proof.
        unfold closure_minimal.
        intros R1 R2 sub [R2_relf R2_trans R2_symm].
        unfold reflexive_transitive_symmetric_closure.
        apply reflexive_transitive_closure_minimal; blast.
        apply symmetric_closure_minimal; blast.
        apply Build_PreOrder;auto.
      Qed.

    End Reflexive_Transitive_Symmetric_Closure_Minimal.

    Section Reflexive_Transitive_Symmetric_Closure_Valid.

      Lemma reflexive_transitive_symmetric_closure_valid:
        closure_valid Equivalence reflexive_transitive_symmetric_closure.
      Proof.
        unfold closure_valid.
        intros R R_compat.
        unfold reflexive_transitive_symmetric_closure.
        split.
        apply reflexive_transitive_closure_valid.
        apply symmetric_closure_compat;
          assumption.
        apply reflexive_transitive_closure_symmetric.
        apply symmetric_closure_symmetric; assumption.
        apply reflexive_transitive_closure_transitive.
        apply symmetric_closure_compat; assumption.
      Qed.

    End Reflexive_Transitive_Symmetric_Closure_Valid.

    (****************************)
    (** **** Closure properties *)
    (****************************)
    Section Reflexive_Transitive_Symmetric_Closure_Monotone.

      Instance reflexive_transitive_symmetric_closure_monotone:
        Proper (subrelationA ==> subrelationA)
               reflexive_transitive_symmetric_closure.
      Proof.
        eapply closure_monotone.
        apply reflexive_transitive_symmetric_closure_superrelation.
        apply reflexive_transitive_symmetric_closure_minimal.
        apply reflexive_transitive_symmetric_closure_valid.
      Qed.

    End Reflexive_Transitive_Symmetric_Closure_Monotone.

    (***********************************************************)
    (** **** Reflexive transitive symmetric closure properties *)
    (***********************************************************)
    Section Reflexive_Transitive_Symmetric_Closure_Reflexive.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
      .
      
      Instance reflexive_transitive_symmetric_closure_reflexive:
        Reflexive (reflexive_transitive_symmetric_closure R).
      Proof.
        apply (reflexive_transitive_symmetric_closure_valid);
          assumption.
      Qed.

    End Reflexive_Transitive_Symmetric_Closure_Reflexive.

    Section Reflexive_Transitive_Symmetric_Closure_Transitive.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
      .

      Instance reflexive_transitive_symmetric_closure_transitive:
        Transitive (reflexive_transitive_symmetric_closure R).
      Proof.
        apply (reflexive_transitive_symmetric_closure_valid);
          assumption.
      Qed.

    End Reflexive_Transitive_Symmetric_Closure_Transitive.

    Section Reflexive_Transitive_Symmetric_Closure_Symmetric.
      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
      .

      Instance reflexive_transitive_symmetric_closure_symmetric:
        Symmetric (reflexive_transitive_symmetric_closure R).
      Proof.
        apply (reflexive_transitive_symmetric_closure_valid);
          assumption.
      Qed.

    End Reflexive_Transitive_Symmetric_Closure_Symmetric.

  End Reflexive_Transitive_Symmetric_Closure.

  (** * CHAINS of related elements *)
  Section Non_Empty_Chain.

    Section Non_Empty_Chain_Definition.

      Fixpoint non_empty_chain (R: relation A) x p y:=
        match p with
        | nil => R x y
        | h:: t => R x h /\ non_empty_chain R h t y
        end
      .

      Global Instance non_empty_chain_compat:
        Proper (pequiv ==> equiv ==> eqlistA equiv ==> fequiv) non_empty_chain.
      Proof.
        assert (half_lemma:
                  forall r1 r2,
                    r1 =~= r2 ->
                    forall x1 x2: A,
                      x1 == x2 ->
                      forall l1 l2: list A,
                        eqlistA equiv l1 l2 ->
                        forall y1 y2: A,
                          y1 == y2 -> (non_empty_chain r1 x1 l1 y1 ->
                                       non_empty_chain r2 x2 l2 y2)
               ).
        intros r1 r2 er x1 x2 eqx l1 l2 eql y1 y2 eqy p.
        revert x1 x2 eqx l1 eql p.
        induction l2 as [ | h2 t2 Hrec]; intros x1 x2 eqx l1 eql p.
        simpl.
        rewrite l_eqlistA_nil in eql; subst; blast.
        simpl in p.
        revert p; apply er; auto. 
        simpl.
        destruct (l_eqlistA_cons eql) as [h1 [t1 eq1]]; subst.
        rewrite cons_eqlistA_cons in eql.
        destruct eql as [eqh eqt].
        simpl in p.
        destruct p as [Rx1h1 p].
        split.
        revert Rx1h1; apply er; auto. 
        apply (Hrec h1 h2 eqh t1 eqt p).
        split; apply half_lemma; blast.
      Qed.
      
    End Non_Empty_Chain_Definition.

    Section Transitive_Closure_Left_Iff_Exists_Non_Empty_Chain.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
      .

      Lemma transitive_closure_left_iff_exists_non_empty_chain:
        forall x y,
          (transitive_closure_left R x y) <-> (exists c, non_empty_chain R x c y)
      .
        intros x z.
        split.
        intros Cxz.
        induction Cxz as [x z Rxz | x y z Rxy Cyz Hrec].
        exists nil.
        simpl.
        assumption.
        destruct Hrec as [c h].
        exists (y:: c).
        simpl.
        split; assumption.
        intros [c Hc].
        revert x Hc.
        induction c as [ | h t Hrec]; intros x Hc.
        simpl in Hc.
        left; assumption.
        simpl in Hc.
        destruct Hc as [Rxh Ht].
        apply (transitive_closure_left_transitivity x Rxh).
        apply (Hrec h).
        assumption.
      Qed.

    End Transitive_Closure_Left_Iff_Exists_Non_Empty_Chain.

    Section Transitive_Closure_Iff_Exists_Non_Empty_Chain.
      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
      .

      Lemma transitive_closure_iff_exists_non_empty_chain:
        forall x y,
          (transitive_closure R x y) <-> (exists c, non_empty_chain R x c y)
      .
      Proof.
        intros x y.
        transitivity (transitive_closure_left R x y).
        symmetry; apply transitive_closure_left_iff_transitive_closure.
        apply transitive_closure_left_iff_exists_non_empty_chain.
      Qed.

    End Transitive_Closure_Iff_Exists_Non_Empty_Chain.
    
  End Non_Empty_Chain.

  Section Chain.

    Section Chain_Definition.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
      .

      Fixpoint chain x p y:=
        match p with
        | nil => x == y
        | h :: t => R x h /\ chain h t y
        end
      .

      Global Instance chain_compat:
        Proper (equiv ==> eqlistA equiv ==> fequiv) chain.
      Proof.
        assert (half_lemma:
                  forall x1 x2: A,
                    x1 == x2 ->
                    forall l1 l2: list A,
                      eqlistA equiv l1 l2 ->
                      forall y1 y2: A, y1 == y2 ->
                                       (chain x1 l1 y1 -> chain x2 l2 y2)
               ).
        intros x1 x2 eqx l1 l2 eql y1 y2 eqy p.
        revert l1 eql x1 x2 eqx p.
        induction l2 as [ | h2 t2 Hrec]; intros l1 eql x1 x2 eqx p.
        simpl.
        rewrite l_eqlistA_nil in eql; blast.
        subst.
        simpl in p.
        transitivity x1.
        symmetry.
        assumption.
        transitivity y1; assumption.
        destruct (l_eqlistA_cons eql) as [h1 [t1 eq]].
        subst.
        rewrite cons_eqlistA_cons in eql.
        destruct eql as [eqh eqt].
        simpl.
        simpl in p.
        destruct p as [ph pt].
        split.
        rewrite <- eqx.
        rewrite <- eqh.
        assumption.
        specialize (Hrec t1 eqt h1 h2 eqh).
        apply Hrec.
        assumption.
        split; apply half_lemma; blast.
      Qed.

    End Chain_Definition.

    Lemma chain_equiv_equiv:
      forall (R1: relation A) (R2: relation A),
        R1 =~= R2 ->
        (equiv ==> eqlistA equiv ==> fequiv)%signature
                                            (chain R1)
                                            (chain R2).
    Proof.
      intros R1 R2 HR a1 a2 Ha l1 l2 Hl b1 b2 Hb.
      revert a1 a2 Ha b1 b2 Hb l2 Hl;
        induction l1 as [| x1 l1 IHl1];
        intros a1 a2 Ha b1 b2 Hb l2 Hl;
        destruct l2 as [| x2 l2].

      simpl.
      rewrite <- Ha.
      rewrite <- Hb.
      reflexivity.

      symmetry in Hl.
      rewrite l_eqlistA_nil in Hl.
      discriminate Hl.

      rewrite l_eqlistA_nil in Hl.
      discriminate Hl.

      simpl.
      apply cons_eqlistA_cons in Hl.
      destruct Hl as (Hx, Hl).

      split; intro HH; destruct HH as (HH, HH'); split.
      apply (HR a1 a2 Ha x1 x2 Hx); exact HH.
      specialize (IHl1 _ _ Hx _ _ Hb _ Hl).
      rewrite <- IHl1; exact HH'.
      apply (HR a1 a2 Ha x1 x2 Hx); exact HH.
      specialize (IHl1 _ _ Hx _ _ Hb _ Hl).
      rewrite IHl1; exact HH'.
    Qed.

    Section Chain_App.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
      .

      Lemma chain_app: forall x p1 p2 z,
          chain R x (p1 ++ p2) z <-> exists y, chain R x p1 y /\ chain R y p2 z
      .
      Proof.
        intros x p1 p2 z.
        split.
        intros c12.
        revert x c12.
        induction p1 as [ | h1 t1 Hrec]; intros x c12.
        simpl in c12.
        exists x.
        split.
        simpl.
        reflexivity.
        assumption.
        simpl in c12.
        destruct c12 as [c12_r c12_c].
        apply Hrec in c12_c.
        destruct c12_c as [y cyz].
        exists y.
        split.
        simpl.
        split.
        assumption.
        apply (proj1 cyz).
        apply (proj2 cyz).
        intros [y [cxy cyz]].
        revert x cxy.
        induction p1 as [ | h1 t1 Hrec]; intros x cxy.
        simpl in *.
        rewrite cxy.
        assumption.
        simpl in *.
        split.
        apply (proj1 cxy).
        apply Hrec.
        apply (proj2 cxy).
      Qed.

    End Chain_App.

    Section Chain_Cons_Iff_Non_Empty_Chain.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
      .

      Lemma chain_cons_iff_non_empty_chain:
        forall x z p,
          chain R x (p ++ (z :: nil)) z <-> non_empty_chain  R x p z
      .
      Proof.
        intros x z p.
        split.
        intros chain_x_pz_z.
        rewrite chain_app in chain_x_pz_z; blast.
        destruct chain_x_pz_z as [y [chain_x_p_y chain_y_z_z]].
        simpl in chain_y_z_z.
        destruct chain_y_z_z as [Ryz _].
        revert x chain_x_p_y.
        induction p as [ | h t Hrec]; intros x chain_x_p_y.
        simpl.
        simpl in chain_x_p_y.
        rewrite chain_x_p_y.
        assumption.
        simpl.
        simpl in chain_x_p_y.
        destruct chain_x_p_y as [Rxh chain_h_t_y].
        split; blast.
        intros chain'_x_p_z.
        revert x chain'_x_p_z.
        induction p as [ | h t Hrec]; intros x chain'_x_p_z.
        simpl.
        simpl in chain'_x_p_z.
        split; blast.
        simpl.
        simpl in chain'_x_p_z.
        destruct chain'_x_p_z as [Rxh chain'_h_t_z].
        apply Hrec in chain'_h_t_z.
        split; assumption.
      Qed.

    End Chain_Cons_Iff_Non_Empty_Chain.
    Section Chain_Exists_Iff_Reflexive_Transitive_Closure.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
      .

      Lemma list_snoc_inv: forall A (l: list A), l <> nil ->
                                                 exists h t, l = t ++ (h :: nil).
      Proof.
        intros b l neq_nil.
        induction l.
        elim neq_nil.
        reflexivity.
        clear neq_nil.
        destruct l as [ | u v].
        exists a.
        exists nil.
        simpl.
        reflexivity.
        assert (u :: v <> nil).
        discriminate.
        apply IHl in H.
        destruct H as [h [t p]].
        rewrite p.
        exists h.
        exists (a :: t).
        simpl.
        reflexivity.
      Qed.

      Lemma chain_last_eq:
        forall x t z y, chain R x (t ++ z :: nil) y -> z == y. 
      Proof.
        intros x t;revert x;induction t as [|a t ht];simpl.
        tauto.
        firstorder.
      Qed.

      Lemma chain_exists_iff_reflexive_transitive_closure:
        forall x y, ((exists p, chain R x p y) <->
                     (reflexive_transitive_closure R) x y)
      .
      Proof.
        intros x y.
        unfold reflexive_transitive_closure.
        unfold reflexive_closure.
        split.
        {
          intros [[|z p]];[left|right];auto.
          rewrite transitive_closure_iff_exists_non_empty_chain.
          destruct (list_snoc_inv (l:=z::p) ) as [h [t hh]].
          congruence.
          rewrite hh in H.
          assert (hhh:=chain_last_eq _ _ _ _ H).
          rewrite <- hhh in H.
          rewrite (chain_cons_iff_non_empty_chain) in H;auto.
          exists t;auto.
          rewrite <- hhh;auto.
        }
        {
          intros [h|h].
          exists nil.
          apply h.
          rewrite transitive_closure_iff_exists_non_empty_chain in h.
          destruct h as [l h].
          rewrite <- chain_cons_iff_non_empty_chain in h;auto.
          firstorder.
        }
      Qed.

    End Chain_Exists_Iff_Reflexive_Transitive_Closure.

    Section Chain_Dec.

      Context
        (R: relation A)
        `{R_compat: Proper _ fequiv R}
      .

      Lemma chain_dec:
        Decider equiv ->
        Decider R ->
        forall x p y, {chain R x p y}  + {~ (chain R x p y)}
      .
      Proof.
        intros eqA_dec R_dec x p y.
        revert x.
        induction p as [ | h t Hrec]; intros x.
        simpl.
        apply eqA_dec.
        simpl.
        destruct (R_dec x h) as [Rxh | nRxh].
        destruct (Hrec h) as [c | nc].
        left; split; assumption.
        right; intros [_ c]; apply nc; assumption.
        right; intros [Rxh _]; apply nRxh; assumption.
      Qed.
    End Chain_Dec.

  End Chain.

  Section Chain_Inclusion.

    Lemma chain_incl:
      forall (R: relation A) (R': relation A),
        inclusion R' R ->
        forall (a: A) (la: list A) (a': A),
          chain R' a' la a -> chain R a' la a.
    Proof.
      intros R R' Hincl a la.
      unfold inclusion in Hincl.
      induction la as [| a0 la Hla].
      simpl. trivial.
      simpl.
      intros a' (HR', Hchain).
      split.
      apply (Hincl a' a0).
      exact HR'.
      apply (Hla a0).
      exact Hchain.
    Qed.

    Lemma chain_compat_incl:
      forall (R: relation A) (R': relation A),
        Proper fequiv R ->
        Proper fequiv R' ->
        inclusion R' R ->
        forall (a: A) (b: A) (la: list A) (lb: list A)
               (a': A) (b': A),
          a == b -> a' == b' -> eqlistA equiv la lb
          -> chain R' a la a' -> chain R b lb b'.
    Proof.
      intros R R' HR HR' Hincl a b la lb a' b'
             Hab Hab' Hl.
      intro Hchain.
      eapply chain_compat.
      apply HR. symmetry. apply Hab.
      symmetry. apply Hl.
      symmetry. apply Hab'.
      apply chain_incl with (R:=R) (R':=R').
      apply Hincl.
      apply Hchain.
    Qed.

  End Chain_Inclusion.

  Section chain_cons_non_empty.
    Context
      (R: relation A)
      `{R_compat: Proper _ fequiv R}
    .

    Lemma chain_cons_non_empty: forall (a1 a2:A) (l:list A) ,
        chain R a1 (l ++ a2 :: nil) a2 <->
        non_empty_chain R a1 l a2.
    Proof.
      intros a1 a2 l;revert a1.
      induction l  as [ | a q hq ];intro a1.
      simpl;intuition.
      simpl in *.
      apply and_compat.
      reflexivity.
      apply hq.
    Qed.

  End chain_cons_non_empty.

End With_A.

Close Scope signature_scope.
Unset Implicit Arguments.
