(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Nov 2016
 *)
(* begin show *)


(** **** Global Imports *)
From Coq Require Import Setoid.
From Coq Require Import SetoidList.
From Coq Require Import Morphisms.
From Coq Require Import RelationClasses.
From Coq Require Import SetoidClass.
From Coq Require Import RelationPairs.
From Coq Require Import Arith.


(** **** Local Imports *)
From Padec Require Import OptionUtil.
From Padec Require Import SetoidUtil.
From Padec Require Import WellfoundedUtil.
From Padec Require Import RelationA.
From Padec Require Import Stream.


(* end show *)
Open Scope signature_scope.
Set Implicit Arguments.

(** * Simulation and Squeezing 
 
 Squeezing is an operator on streams that removes any consecutive duplicated 
 elements. It may remove finite numbers of duplicates. It also removes an 
 infinite number of duplicates provided that they are at the end of the stream.

 Squeezable establishes a condition under which Squeezing is actually computable.

 Simulation defines an order on streams: X <= Y when X contains less 
 consecutive duplicates thatn Y.
 
 For a stream X, (Squeeze X) is the minimum stream among the streams Y such
 that Y <= X.

 *)
Section Simulation_Squeezing.

  (** Simulation and Squeeze are defined assuming a partial setoid A 
   Equality relation on A is assumed to be weakly decidable
   *)
  Context {A: Type} `{SA: PartialSetoid A}.
  Context `{pequiv_dec: Weak_Dec_relation pequiv (A:=A)}.

  (** ** Squeezable: defines a condition under which a stream can be squeezed

   Nb: this condition should be in Type, to achieve constructive arguments that 
       can be used to actually build the operator Squeeze as a CoFixpoint.
   *)

  (** *** Rskip
   Relation on stream: (Rskip a s1 s2) means 
   - either the two streams are reduced to one element and equal; or
   - add element a to s1 to obtain s2 
   *)
  Definition Rskip (a: A): relation Stream :=
    (fun s1 s2 => Proper pequiv s2 ->
                  s_cons a s1 =~= s2 \/
                  s1 =~= s_one a /\ s_one a =~= s2).

  Global Instance Rskip_compat: Proper fequiv Rskip.
  Proof.
    intros a1 a2 ea s1 s2 es ss1 ss2 ess.
    apply Morphisms_Prop.iff_iff_iff_impl_morphism.
    split;unfold Proper;intro h.
    transitivity ss1;[symmetry|];apply ess.
    transitivity ss2;[|symmetry];apply ess.
    apply Morphisms_Prop.or_iff_morphism.
    rewrite ess.
    split; intros h; [symmetry|]; rewrite <- h; right; auto.
    apply Morphisms_Prop.and_iff_morphism.
    split; intros h; [rewrite <- es|rewrite es]; rewrite h;
      [|symmetry]; left; auto.
    split; intros h; [rewrite <- ess|rewrite ess]; rewrite <- h;
      [symmetry|]; left; auto.
  Qed.

  Global Instance Rskip_compat_right: forall a s, Proper fequiv (Rskip a s).
  Proof.
    intros a s s1 s2 e.
    apply Morphisms_Prop.iff_iff_iff_impl_morphism.
    split;unfold Proper;intro h; [rewrite <- e|rewrite e]; auto.
    apply Morphisms_Prop.or_iff_morphism.
    rewrite e; tauto.
    rewrite e; tauto.
  Qed.

  Lemma Rskip_hd :
    forall a s1 s2, Rskip a s1 s2 -> Proper pequiv s2 -> a =~= s_hd s2.
  Proof.
    intros a s1 s2 H ps2.
    destruct (H ps2) as [h|[h1 h]];
      destruct s2;simpl;try inversion h;assumption.
  Qed.

  Lemma Rskip_cons :
    forall a b x y,
      a =~= b -> x =~=y -> Rskip a x (s_cons b y).
  Proof.
    intros a b x y ab xy.
    left; right; auto.
  Qed.

  Lemma Rskip_left_unique:
    forall a s1 s2 s, Proper pequiv s -> Rskip a s1 s -> Rskip a s2 s ->
                      s1 =~= s2.
  Proof.
    intros until s;intros p h1 h2;specialize (h1 p);specialize (h2 p).
    revert h1 h2 ;intros [h1|[h11 h12]] [h2|[h21 h22] ].
    rewrite <- h2 in h1.
    inversion h1;auto.
    rewrite <- h22 in h1.
    inversion h1.
    rewrite <- h12 in h2.
    inversion h2.
    rewrite <- h21 in h11;auto.
  Qed.
  
  Lemma Rskip_cons_4_Skip (a: A) (pa: Proper pequiv a) (x: A)
        (ss: Stream)
        (EQ: Proper pequiv a -> Proper pequiv x -> pequiv a x):
    Rskip a ss (s_cons x ss).
  Proof.
    intros h; dup h h'; revert h'.
    ssplit h.
    apply Rskip_cons; auto.
    apply EQ; [auto|apply h].
    apply h.
  Qed.


  (** *** Skippable 
   Type on Stream: Skippable a s means that it can be decided is
   - whether s in constantly made of the element a, or
   - s begins with a finite number of elements, followed then by an element 
     which is different from a.
   *)
  Definition Skippable (a: A) (s: Stream): Type :=
    {is_constant a s}+{Acc (Rskip a) s}.

  Lemma Acc_Rskip_if_not_a:
    forall a s, Proper pequiv s -> ~ (a =~= s_hd s) -> Acc (Rskip a) s.
  Proof.
    intros a s p h.
    constructor;intros s0 hs.
    contradict h.
    apply Rskip_hd with s0.
    assumption.
    assumption.
  Qed.

  Lemma Skippable_if_not_a:
    forall a s, Proper pequiv s -> ~ (a =~= s_hd s) -> Skippable a s.
  Proof.
    intros a s p h; right.
    apply Acc_Rskip_if_not_a; assumption.
  Qed.

  Lemma is_constant_not_Acc:
    forall s a, is_constant a s -> ~Acc (Rskip a) s.
  Proof.
    intros s a i F.
    induction F.
    destruct x as [t | t x]; inversion i; subst.
    { eapply H0.
      2: apply i.
      right.
      simpl in H2.
      split; left; [|symmetry]; assumption.
    }{ eapply H0.
       2: apply H4.
       simpl in H3.
       left; right.
       symmetry; assumption.
       inversion H1; proper_.
     }
  Qed.

  Lemma Acc_not_is_constant:
    forall s a, Acc (Rskip a) s -> ~is_constant a s.
  Proof.
    intros s a acc F.
    contradict acc; apply is_constant_not_Acc; assumption.
  Qed.

  Lemma Acc_Rskip_cons:
    forall a s, Acc (Rskip a) (s_cons a s) -> Acc (Rskip a) s.
  Proof.
    intros a s acc.
    eapply Acc_inv.
    apply acc.
    intros h.
    left; right; ssplit h; apply h.
  Qed.
  
  Instance Acc_Rskip_compat: Proper fequiv (fun a => Acc (Rskip a)).
  Proof.
    assert (H:Proper (pequiv ==> pequiv ==> impl) (fun a => Acc (Rskip a))).
    {
      intros a b eab x y exy.
      apply Acc_per_impl;auto.
      apply Rskip_compat;symmetry;auto.
      intros xx yy hskip pxx.
      hnf in hskip.
      specialize (hskip pxx).
      intuition;try proper_.
      inversion H;proper_.
    }
    intros a b eab x y exy;split;apply H;auto;symmetry;auto.
  Qed.

  Lemma Skippable_if_move:
    forall a s s0, ~ (a =~= s_hd s0) -> s_cons a s0 =~= s ->
                   Skippable a s.
  Proof.
    intros a s s0 ne es.
    right.
    constructor.
    intros s0_ h0_.
    constructor.
    assert (hh: s0 =~= s0_).
    apply Rskip_left_unique with a s.
    unfold Proper;etransitivity;[symmetry|];eassumption.
    left;apply es.
    assumption.
    intros s1_ h1_.
    contradict ne.
    apply Rskip_hd with s1_.
    eapply Rskip_compat_right.
    apply hh.
    assumption.
    proper_.
  Qed.

  Lemma Skippable_if_move':
    forall a (pa: Proper pequiv a)
           s (ps: Proper pequiv s),
      ~ (a =~= s_hd s) -> Skippable a (s_cons a s).
  Proof.
    intros a pa s ps n.
    eapply Skippable_if_move.
    apply n.
    right; assumption.
  Qed.

  Lemma Skippable_compat:
    forall a b (eab: a =~= b)
           x y (exy: x =~= y),
      Skippable a x -> Skippable b y.
  Proof.
    intros a b eab x y exy [cst|acc].
    - left.
      eapply is_constant_compat.
      symmetry in eab; apply eab.
      symmetry in exy; apply exy.
      apply cst.
    - right.
      eapply Acc_per_impl.
      4: apply acc.
      3: apply exy.
      apply Rskip_compat; symmetry; auto.
      intros c v h p; specialize (h p);revert h;intros [sk| [sk1 sk2]].
      inversion sk; proper_.
      proper_.
  Qed.

  (** *** Squeezable
   Type on Stream: Squeezable s means that the stream is forever skippable,
   namely all along the stream, it is skippable using its head element.
   *)
  Definition Squeezable (s: Stream): Type :=
    Forever (fun s => Skippable (s_hd s) s) s.

  Lemma Squeezable_inv (s: Stream) (w: Squeezable s): Skippable (s_hd s) s.
  Proof.
    destruct w; auto.
  Qed.

  Lemma Squeezable_inv_cons (a: A) (s: Stream) (w: Squeezable (s_cons a s)):
    Squeezable s.
  Proof.
    inversion w; auto.
  Qed.

  Lemma Squeezable_one (a: A): Proper pequiv a -> Squeezable (s_one a).
  Proof.
    intro h; left; left; left; apply h.
  Qed.


  Lemma Squeezable_compat:
    forall x y (exy: x =~= y), Squeezable x -> Squeezable y.
  Proof.
    intros x y exy.
    unfold Squeezable.
    apply Forever_compat_impl.
    intros a b eab.
    apply Skippable_compat.
    apply hd_compat; apply eab.
    apply eab.
    apply exy.
  Qed.

  Lemma Squeezable_cons:
    forall (a : A) (s : Stream)
           (pa:Proper pequiv a) (ps:Proper pequiv s),
      Squeezable s -> Squeezable (s_cons a s).
  Proof.
    intros a s pa ps cs;right.
    2:apply cs.
    hnf;simpl.
    assert (hh:Skippable (s_hd s) s).
    {
      inversion cs;subst;auto.
    }
    destruct (pequiv_dec a (s_hd s)) as [EQ | NE].
    {
      destruct hh as [hconstant| hacc].
      {
        left.
        right;auto.
        revert hconstant;apply is_constant_compat;auto.
        apply EQ;auto.
        apply hd_compat;auto.
      }
      {
        assert (pas:Proper pequiv (s_cons a s)) by (right;auto).
        right.
        constructor.
        intros ss hskip.
        specialize (hskip pas).
        destruct hskip as [heq | [_ heq]];
          try inversion heq;subst.
        revert hacc;apply Acc_per_impl.
        {
          apply Rskip_compat,EQ;auto.
          apply hd_compat;auto.
        }
        {
          intros x y hskip px.
          hnf in hskip.
          specialize (hskip px).
          intuition;try proper_.
          inversion H;proper_.
        }
        {
          symmetry;auto.
        }
      }
    }
    {
      assert (pas:Proper pequiv (s_cons a s)) by (right;auto).
      right.
      constructor.
      intros ss hskip.
      specialize (hskip pas).
      destruct hskip as [heq | [_ heq]];try inversion heq;subst.
      apply (Acc_Rskip_if_not_a).
      proper_.
      intro h;apply NE.
      rewrite h;apply hd_compat,H4.
    }
  Qed.

  Lemma is_constant_Squeezable:
    forall a s, is_constant a s -> Squeezable s.
  Proof.
    intro a.
    cofix h.
    intros [b | b s] hh;[left|right;try eapply h].
    left;left;apply hd_compat;eapply is_constant_Proper_r;eauto.
    left;simpl.
    assert (p:=is_constant_Proper_r hh).
    generalize hh.
    apply is_constant_compat;auto.
    inversion_clear hh;subst;auto.
    inversion_clear hh;subst;auto.
  Qed.


  (** *** Skip 
     Operator that removes a finite prefix of a stream: Skip a s returns
     the suffix of the stream s from which every element a has been removed
     at the beginning.
     It requires Acc (Rskip a) s to be able to perform induction.
   *)
  Fixpoint Skip (a: A) (pa: Proper pequiv a)
           (s: Stream) (W: Acc (Rskip a) s): Stream :=
    match s as es return Acc (Rskip a) es -> Stream with
    | s_one x => fun _ => s_one x
    | s_cons x ss =>
      fun w =>
        match (pequiv_dec a x) with
        | left EQ => Skip pa (Acc_inv w _ (Rskip_cons_4_Skip a pa EQ ))
        | right _ => s_cons x ss
        end end W.

  Global Instance Skip_compat:
    forall a (pa: Proper pequiv a)
           (s: Stream) (ps: Proper pequiv s)
           (W: Acc (Rskip a) s),
      Proper pequiv (Skip pa W).
  Proof.
    intros a pa s ps W.
    induction W as [s fW hW] using Acc_ind_dep.
    simpl.
    destruct s as [x | x ss].
    - assumption.
    - destruct (pequiv_dec a x) as [ E | NE];simpl.
      + apply hW. ssplit ps; apply ps.
      + assumption.
  Qed.

  Lemma Skip_compat':
    forall (a: A) (pa: Proper pequiv a)
           (b: A) (pb: Proper pequiv b)
           (x y: Stream)
           (Wax: Acc (Rskip a) x) (Wby: Acc (Rskip b) y),
      a =~= b -> x =~= y -> Skip pa Wax =~= Skip pb Wby.
  Proof.
    intros a pa b pb x y Wax Wby eab exy.
    revert pa b pb y Wby eab exy;
      induction Wax using Acc_ind_dep;
      intros pa b pb y Wby eab exy;
      destruct Wby as [Wby].
    simpl.
    destruct x as [hx | hx x]; destruct y as [hy | hy y].
    - apply exy.
    - inversion exy.
    - inversion exy.
    - case (pequiv_dec a hx); case (pequiv_dec b hy); intros Ea Eb.
      + apply H; try assumption.
        ssplit exy; apply exy.
      + contradict Ea; inversion exy; subst; simpl in *.
        rewrite <- eab, <- H3; apply Eb;try assumption.
        inversion exy;unfold Proper; etransitivity;[|symmetry];eassumption.
      + contradict Eb; inversion exy; subst; simpl in *;
          rewrite eab, H3; apply Ea;try assumption.
        inversion exy;unfold Proper; etransitivity;[symmetry|];eassumption.
      + apply exy.
  Qed.

  Lemma Skip_hd:
    forall a (pa: Proper pequiv a) s (W: Acc (Rskip a) s),
      ~ (a =~= s_hd (Skip pa W)).
  Proof.
    intros a pa s W.
    induction W as [s H IH] using Acc_ind_dep.
    simpl.
    destruct s as [x | x s].
    { simpl. intros F.
      assert (tmp: Rskip a (s_one x) (s_one x)).
      right; split; left; [symmetry|]; apply F.
      assert (tmpp := Acc_not_is_constant (H _ tmp)).
      contradict tmpp.
      left; simpl; symmetry; apply F.
    }
    { case (pequiv_dec a x).
      - intros e.
        apply IH.
      - intros n; simpl; apply n.
    }
  Qed.

  Lemma Skip_suffix:
    forall (a: A) (pa: Proper pequiv a)
           (s: Stream) (ps: Proper pequiv s)
           (W: Acc (Rskip a) s),
      is_suffix s (Skip pa W).
  Proof.
    intros a pa s ps W.
    revert ps pa.
    induction W using Acc_ind_dep.
    intros px pa.
    simpl.
    destruct x as [hx | hx x].
    - left; apply px.
    - case (pequiv_dec a hx).
      + intros e.
        right.
        apply H.
        ssplit px; apply px.
      + left; apply px.
  Qed.

  Lemma Skip_cons:
    forall (a: A) (pa: Proper pequiv a) (s: Stream)
           (W: Acc (Rskip a) (s_cons a s)) (W': Acc (Rskip a) s),
      Proper pequiv s -> Skip pa W =~= Skip pa W'.
  Proof.
    intros a pa s W W' ps.
    destruct W as (W); simpl.
    case (pequiv_dec a a).
    - intros e.
      eapply Skip_compat'; assumption.
    - intros ne; contradict ne; apply pa.
  Qed.

  (** A squeezable stream from which has been removed its finite prefix of
     duplicates (Skip) is still squeezable *)
  Lemma Squeezable_Skip:
    forall a (pa:Proper pequiv a)
           (s: Stream)
           (SQ: Squeezable s)
           (W : Acc (Rskip a) s),
      Squeezable (Skip pa W).
  Proof.
    intros a pa s SQ W.
    induction W as [s fW hW] using Acc_rec_dep;simpl.
    destruct s as [x | x s].
    - assumption.
    - destruct (pequiv_dec a x) as [EQ| NE].
      + apply hW;eapply Squeezable_inv_cons;eassumption.
      + assumption.
  Qed.

  (** *** Squeeze
     Operator that removes every consecutive duplicates from the stream, included
     at its (potentially infinite) suffix.
     It requires the stream to be squeezable.
     
     Precisely,
     (Skip (s_hd s) s) if it exists implies that s can be written as
     s = a^+.s' with a = (s_hd s) and s' a non-empty stream.
     Then, (Skip (s_hd s) s) returns s'.

     - If (is_constant s), (Squeeze s) returns (s_one (s_hd s))
     - Otherwise:
       + s is not (s_one _) and s can be written as a^+.s' with s' 
         a non-empty stream and ~(pequiv a (s_hd s'))
       + (Squeeze a^+.s') returns a.(Squeeze s')

       Precisely, (Skip a a^+.s') returns s' and
       (Squeeze a^+.s') returns a.(Squeeze (Skip a a^+.s'))
 
       Note that (Skip a a^+.s') is itself Squeezable from Lemma Squeezable_Skip.
   *)

  CoFixpoint Squeeze (s: Stream) (p: Proper pequiv s) (CC: Squeezable s):
    Stream :=
    match s as op return Proper pequiv op -> Squeezable op -> Stream with
    | s_one x => fun _ _ => s_one x
    | s_cons x ss => fun ps cc =>
                       match Squeezable_inv cc with
                       | left _ (*is_constant x s*) =>
                         s_one x
                       | right hW (*Acc (Rskip x) s*) =>
                         s_cons x (Squeeze
                                     (Skip_compat (hd_compat ps) ps hW)
                                     (Squeezable_Skip (hd_compat ps) cc hW))
                       end end p CC.

  Lemma Squeeze_unfold:
    forall s p (CC: Squeezable s),
      Squeeze p CC =
      match s as op return Proper pequiv op -> Squeezable op -> Stream with
      | s_one x => fun _ _ => s_one x
      | s_cons x ss => fun ps cc =>
                         match Squeezable_inv cc with
                         | left _ (*is_constant x s*) =>
                           s_one x
                         | right hW (*Acc (Rskip x) s*) =>
                           s_cons x (Squeeze
                                       (Skip_compat (hd_compat ps) ps hW)
                                       (Squeezable_Skip (hd_compat ps) cc hW))
                         end end p CC.
  Proof.
    intros s p W.
    rewrite Stream_decomposition_lemma with (Squeeze p W).
    destruct s as [x | x ss]; simpl.
    - reflexivity.
    - case_eq (Squeezable_inv W); simpl;
        intros a c; reflexivity.
  Qed.

  Lemma Squeeze_compat:
    forall x (px: Proper pequiv x) (cx: Squeezable x)
           y (py: Proper pequiv y) (cy: Squeezable y),
      x =~= y -> Squeeze px cx =~= Squeeze py cy.
  Proof.
    cofix h.
    intros x px cx y py cy e.
    rewrite Squeeze_unfold.
    rewrite Squeeze_unfold.
    revert e; destruct x as [hx | hx x]; destruct y as [hy | hy y].
    - auto.
    - intros F; ssplit F.
    - intros F; ssplit F.
    - intros e.
      destruct (Squeezable_inv cx);
        destruct (Squeezable_inv cy); simpl.
      + left; ssplit e; apply e.
      + contradict a; apply is_constant_not_Acc; auto.
        eapply is_constant_compat.
        3: apply i.
        simpl; symmetry; ssplit e; apply e.
        symmetry; apply e.
      + contradict a; apply is_constant_not_Acc; auto.
        eapply is_constant_compat.
        3: apply i.
        simpl; ssplit e; apply e.
        apply e.
      + right.
        ssplit e; apply e.
        apply h.
        eapply Skip_compat'.
        ssplit e; apply e.
        apply e.
  Qed.

  Lemma Squeeze_hd: forall (s: Stream) (p: Proper pequiv s) (CC: Squeezable s),
      s_hd (Squeeze p CC) = s_hd s.
  Proof.
    intros [x | x s] p cc; rewrite Squeeze_unfold.
    reflexivity.
    destruct (Squeezable_inv cc); simpl; reflexivity.
  Qed.

  Lemma Squeeze_one: forall (a: A) (s: Stream)
                            (p: Proper pequiv s) (CC: Squeezable s),
      s_one a =~= Squeeze p CC -> is_constant a s.
  Proof.
    cofix h.
    intros a [x  | x s] p CC E; rewrite Squeeze_unfold in E.
    - left; simpl; ssplit E; symmetry; auto.
    - case_eq ( Squeezable_inv CC ).
      + intros i ic; rewrite ic in E.
        simpl in i.
        eapply is_constant_compat; try assumption.
        3: apply i.
        ssplit E; auto.
        apply p.
      + intros i ic; rewrite ic in E; ssplit E.
  Qed.

  (** *** cons_plus 
     Predicate, (cons_plus a s) means that s = a^+.s', namely
     s begins with a finite positive number of a.
   *)
  Inductive cons_plus (a: A): Stream -> Stream -> Prop :=
  | cons_plus_one: forall b s1 s2, a =~= b -> s1 =~= s2 ->
                                   cons_plus a s1 (s_cons b s2)
  | cons_plus_add: forall b s1 s2, a =~= b -> cons_plus a s1 s2 ->
                                   cons_plus a s1 (s_cons b s2).

  Lemma cons_plus_hd: forall a s1 s2, cons_plus a s1 s2 -> a =~= s_hd s2.
  Proof.
    intros a s1 [b | b s2] h; inversion_clear h; auto.
  Qed.

  Lemma cons_plus_Proper_1:
    forall a s1 s2, cons_plus a s1 s2 -> Proper pequiv a.
  Proof.
    intros a s1 [b | b s2] h; inversion_clear h; proper_.
  Qed.

  Lemma cons_plus_Proper_2:
    forall a s1 s2, cons_plus a s1 s2 -> Proper pequiv s1.
  Proof.
    intros a s1 s2 h.
    induction h; [proper_|auto].
  Qed.

  Lemma cons_plus_Proper_3:
    forall a s1 s2, cons_plus a s1 s2 -> Proper pequiv s2.
  Proof.
    intros a s1 s2 h.
    induction h; right; proper_.
  Qed.

  (** *** Simulation 
     Relation on streams, Simulation X Y means that X contains less consecutive
     duplicates than Y.
     The number of duplicates removed between X and Y may be infinite provided
     that there were at the end of Y (Y has an infinite suffix of duplicates).
     It is finite in any other case.

     Simulation has 2 constructors:
     - Simulation (s_one a) s is obtained when s is constantly made of a
     - Simulation (s_cons a s1) s3 is obtained when 
          + s1 is itself smaller than a stream s2 such that
          + s3 is equal to s2 at the beginning of which a finite number of a has
            been added.

     We also provide an alternative definition which is less difficult to read
     (one more constructor) but easier to use.
   *)
  CoInductive Simulation_alt: Stream -> Stream -> Prop :=
  | simalt_constant: forall a b s, a =~= b ->
                                   is_constant b s ->
                                   Simulation_alt (s_one a) s
  | simalt_cons: forall a b s1 s2 s3, a =~= b ->
                                      Simulation_alt s1 s2 ->
                                      cons_plus b s2 s3 ->
                                      Simulation_alt (s_cons a s1) s3.


  (** *** is_cons_star 
     Predicate, (is_cons_star a s) means that s = a^*.s', namely
     s begins with a finite number (may be 0) of a.
   *)
  Inductive is_cons_star (a: A): Stream -> Stream -> Prop :=
  | cons_star_no: forall s1 s2, s1 =~= s2 -> is_cons_star a s1 s2
  | cons_star_add: forall b s1 s2, a =~= b -> is_cons_star a s1 s2 ->
                                   is_cons_star a s1 (s_cons b s2).

  Global Instance is_cons_star_compat: Proper fequiv is_cons_star.
  Proof.
    cut (forall a a' x y z t, a =~= a' -> x =~= y -> z =~= t ->
                              is_cons_star a x z -> is_cons_star a' y t).
    { intros H a a' ea' x y exy z t ezt.
      split; [|symmetry in ea', exy, ezt]; apply H; assumption.
    }
    { intros a a' x y z t ea' exy ezt H.
      revert exy ea' ezt; revert a' y t; induction H;
        intros a' y t exy ea' ezt.
      - apply cons_star_no.
        rewrite <- exy, H; assumption.
      - inversion ezt; subst.
        apply cons_star_add.
        rewrite <- ea', <- H3; assumption.
        apply IHis_cons_star; assumption.
    }
  Qed.

  Lemma is_cons_star_Proper_r: forall a x y,
      is_cons_star a x y -> Proper pequiv y.
  Proof.
    intros a x y h;induction h.
    proper_.
    right;proper_.
  Qed.

  Lemma is_cons_star_Proper_l: forall a x y,
      is_cons_star a x y -> Proper pequiv x.
  Proof.
    intros a x y h;induction h.
    proper_.
    auto.
  Qed.

  Lemma is_cons_star_Skip:
    forall (a: A) (pa: Proper pequiv a) (s: Stream) (W: Acc (Rskip a) s),
      Proper pequiv s -> is_cons_star a (Skip pa W) s.
  Proof.
    intros a pa s W ps.
    induction W as [s H IH] using Acc_ind_dep.
    destruct s as [x | x s]; simpl.
    { apply cons_star_no.
      apply ps.
    }
    { dup ps pps; ssplit ps.
      case (pequiv_dec a x).
      - intros e.
        apply cons_star_add.
        apply e; [apply pa | apply ps].
        apply IH.
        apply ps.
      - intros ne.
        apply cons_star_no.
        apply pps.
    }
  Qed.

  Lemma Skip_is_cons_star:
    forall a pa s ss ws,
      is_cons_star a ss s -> ~ a =~= s_hd ss ->
      Skip (a:=a) pa (s:=s) ws =~= ss.
  Proof.
    Arguments Skip : clear implicits.
    intros a pa s ss ws ics neq.
    revert pa ws.
    induction ics.
    {
      intros pa [Fws].
      destruct s2 as [hd2| hd2 tl2];simpl in *;
        try (symmetry;auto).
      destruct (pequiv_dec a hd2) as [WEQ| NEQ2];auto.
      {
        contradict neq.
        apply hd_compat in H;rewrite H.
        apply WEQ;auto.
        proper_.
      }
    }
    {
      intros pa ws.
      specialize (IHics neq).
      assert (pb:Proper pequiv b) by proper_.
      assert (pbs2:Proper pequiv (s_cons b s2)) by
          (right;
           [assumption|
            apply (is_cons_star_Proper_r ics)]).
      assert (s_cons a s2 =~=s_cons b s2).
      {
        right;auto.
        apply (is_cons_star_Proper_r ics).
      }
      assert (ww:Acc (Rskip a) (s_cons a s2)).
      rewrite  (fun hh => Acc_per_morphism 
                            (R:= Rskip a)
                            (Rskip_compat _ _ pa) hh _ _ H0).
      auto.
      unfold Rskip;intuition.
      inversion H1;proper_.
      proper_.
      erewrite (Skip_compat' pa pa ws ww pa).
      2:symmetry;auto.
      assert (www:Acc (Rskip a) s2).
      {
        destruct ww as [Fw0];apply Fw0.
        left;auto.
      }
      rewrite (Skip_cons pa ww www).
      eapply IHics.
      inversion H0;auto.
    }
    Arguments Skip: default implicits.
  Qed.

  Lemma is_cons_star_suffix:
    forall a s1 s2, is_cons_star a s1 s2 -> is_suffix s2 s1.
  Proof.
    intros a s1 s2 C.
    induction C.
    - left; symmetry; assumption.
    - right; apply IHC.
  Qed.

  Lemma cons_star_suffix_bis:
    forall a s suf, Proper pequiv a -> Proper pequiv s ->
                    is_suffix s suf -> is_cons_star a suf s ->
                    is_suffix (s_cons a s) (s_cons a suf).
  Proof.
    intros a s suf pa ps hsuf hstar.
    induction hstar.
    - left; right; [| symmetry in H]; auto.
    - apply is_cons_star_suffix in hstar.
      right.
      { eapply is_suffix_compat.
        3: apply IHhstar; auto.
        symmetry; right; [auto | ssplit ps; apply ps].
        right; auto.
        apply is_suffix_Proper in hstar; auto.
        ssplit ps; apply ps.
      }
  Qed.

  Lemma is_cons_star_cons: forall a x sx,
      is_cons_star a (s_cons a sx) x -> is_cons_star a sx x.
  Proof.
    intros a x sx h.
    generalize (is_cons_star_Proper_l h).
    unfold Proper.
    change ((fun x1 x2 =>
               x1 =~=  s_cons a sx ->
               is_cons_star a
                            (match x1 with s_one aa => s_one aa
                                      | s_cons _ xx => xx end) x2)
              (s_cons a sx) x) .
    induction h;
      destruct s1 as [aa| aa xx];intro hh;ssplit hh;
        destruct hh as [ha hx].
    {
      assert (pxx: Proper pequiv xx) by proper_;
        assert (pa: Proper pequiv a) by proper_.
      rewrite <- H.
      right;[symmetry;assumption|left;assumption].
    }
    {
      right;auto.
      apply IHh.
      right;auto.
    }
  Qed.

  Lemma is_cons_star_trans:
    forall a, Proper pequiv a ->
              forall x y z, is_cons_star a x y ->
                            is_cons_star a y z ->
                            is_cons_star a x z.
  Proof.
    intros a pa ssx sx x h1.
    revert x.
    induction h1.
    {
      intros x h2.
      assert (px:=is_cons_star_Proper_r h2).
      rewrite H;auto.
    }
    {
      intros x h2.
      apply IHh1.
      clear h1 IHh1.
      apply is_cons_star_cons.
      assert (px:=is_cons_star_Proper_r h2).
      assert (pas2:=is_cons_star_Proper_l h2).
      revert h2.
      eapply proj2,is_cons_star_compat;try proper_.
      right.
      assumption.
      ssplit pas2;tauto.
    }
  Qed.

  Lemma cons_star_plus:
    forall a b s1 s2, a =~= b ->
                      (is_cons_star a s1 s2 <-> cons_plus a s1 (s_cons b s2)).
  Proof.
    intros a b s1 s2 ab; split; intros h.
    { revert ab; revert b; induction h; intros bb abb.
      left; auto.
      right; auto. }
    { remember (s_cons b s2) as bs2.
      assert (hss2: bs2 =~= s_cons b s2).
      { rewrite <- Heqbs2.
        apply cons_plus_Proper_3 in h; auto. }
      clear Heqbs2; revert hss2; revert s2; induction h; intros ss2 hss2.
      left; inversion hss2; auto.
      destruct s2 as [a2 | a2 s2]; [inversion h|].
      ssplit hss2; destruct hss2 as [h2 hss2].
      destruct ss2 as [aa2 | aa2 ss2]; inversion_clear hss2.
      right.
      apply cons_plus_hd in h.
      rewrite h; auto.
      apply IHh.
      right; auto.
      apply cons_plus_hd in h.
      rewrite <- h; auto.
    }
  Qed.
  
  Lemma is_cons_star_epsilon:
    forall x s s',
      is_cons_star x s s' -> ~(x =~= s_hd s') -> s =~= s'.
  Proof.
    intros x s s' IC N.
    destruct s' as [x' | x' s']; inversion IC; subst; clear IC.
    apply H.
    apply H.
    simpl in N; tauto.
  Qed.

  Lemma star_constant:
    forall b s, is_cons_star b (s_one b) s -> is_constant b s.
  Proof.
    intros b s C.
    remember (s_one b) as t.
    induction C.
    - rewrite Heqt in H.
      assert (pb: Proper pequiv b).
      { apply hd_compat in H; simpl in H.
        proper_. }
      eapply is_constant_compat.
      apply pb.
      symmetry; apply H.
      left; simpl; apply pb.
    - right.
      simpl.
      symmetry; assumption.
      apply IHC; assumption.
  Qed.

  Lemma is_cons_star_is_constant:
    forall a x y,
      is_constant a x -> is_cons_star a x y -> is_constant a y.
  Proof.
    intros a x y const ic.
    induction ic.
    { assert (pa := is_constant_Proper_l const).
      revert const; apply is_constant_compat; [| symmetry]; auto.
    }
    { right.
      simpl; symmetry; auto.
      apply IHic; auto.
    }
  Qed.

  Lemma is_cons_star_is_constant':
    forall a x y,
      is_constant a x -> is_cons_star a y x -> is_constant a y.
  Proof.
    intros a x y const ic.
    induction ic.
    { assert (pa := is_constant_Proper_l const).
      revert const; apply is_constant_compat; auto.
    }
    { apply IHic.
      inversion const; auto.
    }
  Qed.

  (** Simulation: alternative definition (the one used in the proof) 
   *)
  CoInductive Simulation: Stream -> Stream -> Prop :=
  | sim_one_one: forall a b, a =~= b ->
                             Simulation (s_one a) (s_one b)
  | sim_one_cons: forall a b s, a =~= b -> is_constant a s ->
                                Simulation (s_one a) (s_cons b s)
  | sim_cons: forall a b s1 s2 s3, a =~= b ->
                                   Simulation s1 s2 ->
                                   is_cons_star a s2 s3 ->
                                   Simulation (s_cons a s1) (s_cons b s3).

  Lemma Simulation_hd: forall x y, Simulation x y -> s_hd x =~= s_hd y.
  Proof.
    intros [x | x sx] [y | y sy] H; inversion H; subst; simpl; assumption.
  Qed.

  Global Instance Simulation_Proper_l:
    forall x y, Simulation x y -> Proper pequiv x.
  Proof.
    cofix h.
    intros [x | x sx] [y | y sy] H; inversion H; subst; hnf.
    left;proper_.
    left;proper_.
    right;try proper_.
    apply h with s2;auto.
  Qed.

  Global Instance Simulation_Proper_r:
    forall x y, Simulation x y -> Proper pequiv y.
  Proof.
    cofix h.
    intros [x | x sx] [y | y sy] H; inversion H; subst; hnf.
    left;proper_.
    right;try proper_.
    apply is_constant_Proper_r with x;auto.
    right;try proper_.
    apply is_cons_star_Proper_r with x s2; auto.
  Qed.

  Global Instance Simulation_compat: Proper fequiv Simulation.
  Proof.
    cut (forall x y a b, x =~= y -> a =~= b ->
                         Simulation x a -> Simulation y b).
    { intros H x y exy a b eab.
      split; [| symmetry in eab, exy]; apply H; assumption.
    }
    { cofix h.
      intros [x | x sx] [y | y sy] [a | a sa] [b | b sb] exy eab sim;
        inversion sim; ssplit exy; ssplit eab; subst.
      - apply sim_one_one; rewrite <- exy, H1; auto.
      - apply sim_one_cons.
        rewrite <- exy, H2;apply eab.
        eapply is_constant_compat.
        symmetry; apply exy.
        symmetry; apply eab.
        assumption.
      - destruct exy as [exy esxy]; destruct eab as [eab esab].
        apply sim_cons with s2; auto.
        rewrite <- exy, <- eab; assumption.
        apply h with sx s2; auto.
        apply Simulation_Proper_r in H4; auto.
        eapply is_cons_star_compat; auto.
        4: apply H5.
        symmetry; auto.
        apply Simulation_Proper_r in H4; auto.
        symmetry; auto.
    }
  Qed.

  (** Both definition of Simulation are indeed equivalent *)
  Lemma Simulation_alt_equiv:
    forall s1 s2, Simulation s1 s2 <-> Simulation_alt s1 s2.
  Proof.
    intros s1 s2; split; revert s1 s2.
    { cofix h.
      intros [a | a s1] [b | b s2] hh; inversion_clear hh.
      { left with b; auto.
        left; proper_.
      }
      { left with b; auto.
        right; [proper_|].
        assert (ps2 := is_constant_Proper_r H0).
        revert H0; apply is_constant_compat; symmetry; auto.
      }
      { eapply cons_star_plus in H1.
        2: apply H.
        right with a s3; auto.
      }
    }
    { cofix h.
      intros [a | a s1] [b | b s2] hh; inversion_clear hh.
      { constructor 1.
        inversion_clear H0; auto.
      }
      { constructor 2; inversion_clear H0; auto.
        assert (ps2 := is_constant_Proper_r H2).
        revert H2; apply is_constant_compat; auto.
      }
      { inversion H1. }
      { constructor 3 with s3.
        apply cons_plus_hd in H1; rewrite <- H1; auto.
        auto.
        apply cons_star_plus in H1; auto.
        assert (ps2 := is_cons_star_Proper_r H1).
        assert (ps3 := is_cons_star_Proper_l H1).
        revert H1; apply is_cons_star_compat; auto.
        apply cons_plus_hd in H1; auto.
      }
    }
  Qed.

  Lemma Simulation_inv_one:
    forall a x, Simulation (s_one a) x -> is_constant a x.
  Proof.
    intros a x sim;inversion sim;subst.
    left;symmetry;assumption.
    right.
    symmetry;assumption.
    assumption.
  Qed.

  Lemma Simulation_inv_cons_one:
    forall a x b, ~ Simulation (s_cons a x) (s_one b).
  Proof.
    intros a x b sim;inversion sim;subst.
  Qed.

  Lemma Simulation_inv_cons: forall a x b y,
      Simulation (s_cons a x) (s_cons b y) ->
      Simulation x y \/Simulation (s_cons a x) y.
  Proof.
    intros g1 e1 g2 e2 sim;inversion sim;subst.
    destruct e2 as [g2_ | g2_ e2_];(inversion H5;subst).
    { left.
      revert H4.
      intro h;assert (pe1:=Simulation_Proper_l h);revert h.
      eapply proj1.
      apply Simulation_compat;auto.
    }
    { left.
      revert H4.
      intro h;assert (pe1:=Simulation_Proper_l h);revert h.
      eapply proj1.
      apply Simulation_compat;auto.
    }
    { right;constructor 3 with s2;auto. }
  Qed.

  Lemma Simulation_inv_cons_star: forall a x b y,
      Simulation (s_cons a x) (s_cons b y) ->
      exists _y, Simulation x _y /\ is_cons_star a _y y.
  Proof.
    intros g1 e1 g2 e2 sim;inversion sim;subst.
    exists s2;subst;split;auto.
  Qed.

  Lemma is_constant_Simulation:
    forall a y, is_constant a y -> Simulation (s_one a) y.
  Proof.
    intros a [b|b y];
      constructor.
    apply is_constant_hd in H;symmetry;assumption.
    apply is_constant_hd in H;symmetry;assumption.
    inversion H;subst;auto.
  Qed.

  Lemma Simulation_remove_hd:
    forall x y a,
      a =~= s_hd x -> Simulation (s_cons a x) y -> Simulation x y.
  Proof.
    intros x y a hd h; inversion_clear h.
    induction H1 as [s1 s2 hss | c s1 s2 hss ic ih].
    { destruct x as [aa | aa x]; simpl in H.
      { constructor 2; auto.
        cut (is_constant aa s1).
        { apply is_constant_compat; symmetry; auto. }
        destruct s1 as [sx | sx s1]; inversion_clear H0.
        left; symmetry; auto.
        right; [symmetry|]; auto.
      }
      { destruct s1 as [sx | sx s1]; inversion_clear H0.
        destruct s2 as [sy |sy s2]; [inversion hss|].
        ssplit hss; destruct hss as [hss hhss].
        constructor 3 with s3; auto.
        rewrite <- H; symmetry; apply hd.
        right.
        rewrite H1; auto.
        revert H3; apply is_cons_star_compat; symmetry; auto.
        apply Simulation_Proper_r in H2; auto.
      }
    }
    { destruct x as [aa | aa x]; simpl in H, hss, ic.
      { constructor 2; auto.
        specialize (ih H0).
        inversion_clear ih.
        right; simpl; [symmetry|]; auto.
        rewrite <- hss, H; auto.
      }
      { specialize (ih H0).
        inversion_clear ih.
        constructor 3 with s3; auto.
        right; auto.
        rewrite <- hss, H; auto.
      }
    }
  Qed.

  Lemma Simulation_add_hd:
    forall x y a, s_hd x =~= a ->
                  Simulation x y -> Simulation x (s_cons a y).
  Proof.
    intros [b | b x] y a hd hs.
    { constructor 2; auto.
      apply Simulation_inv_one; auto. }
    { inversion_clear hs.
      constructor 3 with s2; auto.
      right; auto.
    }
  Qed.

  Lemma Simulation_is_constant:
    forall a x y, is_constant a x -> Simulation x y -> is_constant a y.
  Proof.
    cofix h.
    intros a [aa | aa x] [bb | bb y] ic hs;
      inversion_clear ic; inversion_clear hs.
    { left; simpl; rewrite <- H0; auto. }
    { right.
      simpl; rewrite <- H0; auto.
      assert (py := is_constant_Proper_r H1).
      revert H1; apply is_constant_compat; auto.
      symmetry; auto.
    }
    { right.
      simpl; rewrite <- H1; auto.
      apply h with x; auto.
      simpl in H; revert H2 H0; revert x.
      induction H3; intros.
      { assert (px := Simulation_Proper_l H2).
        revert H2; apply Simulation_compat; symmetry; auto. }
      {
        { apply Simulation_add_hd; auto.
          apply is_constant_hd in H4.
          rewrite H4, <- H0; symmetry; auto.
        }
      }
    }
  Qed.

  Lemma Simulation_is_constant':
    forall a x y, is_constant a y -> Simulation x y -> is_constant a x.
  Proof.
    intros a; cofix h.
    intros [b | b x] [c | c y] const sim; inversion_clear sim.
    { assert (pa := is_constant_Proper_l const).
      revert const; apply is_constant_compat; [|left]; auto.
    }
    { left; simpl.
      apply is_constant_hd in const; symmetry; auto.
    }
    { right.
      simpl; apply is_constant_hd in const; symmetry; auto.
      apply h with s2; auto.
      apply is_cons_star_is_constant' with y.
      inversion const; auto.
      assert (py :=  is_cons_star_Proper_r H1).
      revert H1; apply is_cons_star_compat; auto.
      apply is_constant_hd in const; symmetry; auto.
      apply Simulation_Proper_r in H0; auto.
    }
  Qed.

  (** **** Simulation is an order, i.e. a reflexive, transitive and 
     antisymmetric relation. Nb: Reflexivity is obtained only for compatible
     streams. *)
  Lemma Simulation_refl: forall x, Proper pequiv x -> Simulation x x.
  Proof.
    cofix h.
    intros [a | a x] px; ssplit px.
    constructor 1; auto.
    constructor 3 with x; try apply px.
    apply h, px.
    apply cons_star_no, px.
  Qed.

  Lemma Simulation_antisym:
    forall x y, Simulation x y -> Simulation y x -> x =~= y.
  Proof.
    cofix h.
    intros [a | a x] [b | b y] h1 h2;
      assert (hd := Simulation_hd h1); simpl in hd.
    { left; auto. }
    { inversion h2. }
    { inversion h1. }
    { right; auto.
      apply Simulation_inv_cons in h1.
      apply Simulation_inv_cons in h2.
      destruct h1 as [h1|h1]; destruct h2 as [h2|h2].
      { apply h; auto. }
      { apply h; auto.
        clear hd; assert (hd := Simulation_hd h1).
        assert (hdd := Simulation_hd h2); simpl in hdd.
        clear a; destruct x as [ a | a x]; simpl in hd, hdd.
        inversion h2.
        destruct y as [bb | bb y]; [inversion h1 |]; simpl in hd.
        cut (Simulation (s_cons a y) (s_cons a x)).
        { apply Simulation_compat; right; auto.
          symmetry; auto.
          apply Simulation_Proper_r in h1; ssplit h1; apply h1.
          apply Simulation_Proper_l in h1; ssplit h1; apply h1.
        }
        apply Simulation_remove_hd with a.
        simpl; proper_.
        revert h2.
        apply Simulation_compat; right; auto.
        symmetry; auto.
        right; auto.
        apply Simulation_Proper_r in h1; ssplit h1; apply h1.
        apply Simulation_Proper_l in h1; ssplit h1; apply h1.
      }
      { apply h; auto.
        clear hd; assert (hd := Simulation_hd h2).
        assert (hdd := Simulation_hd h1); simpl in hdd.
        clear b; destruct y as [ b | b y]; simpl in hd, hdd.
        inversion h1.
        destruct x as [aa | aa x]; [inversion h2 |]; simpl in hd.
        cut (Simulation (s_cons a x) (s_cons a y)).
        { apply Simulation_compat; right; auto.
          apply Simulation_Proper_r in h2; ssplit h2; apply h2.
          symmetry; auto.
          apply Simulation_Proper_l in h2; ssplit h2; apply h2.
        }
        apply Simulation_remove_hd with a.
        simpl; proper_.
        revert h1.
        apply Simulation_compat; right; auto.
        right; auto.
        apply Simulation_Proper_r in h2; ssplit h2; apply h2.
        apply Simulation_Proper_l in h2; ssplit h2; apply h2.
      }
      { assert (hd1 := Simulation_hd h1).
        assert (hd2 := Simulation_hd h2).
        destruct x as [aa | aa x]; [inversion h2|].
        destruct y as [bb | bb y]; [inversion h1|].
        simpl in hd1, hd2.
        apply Simulation_remove_hd in h1.
        2: simpl; rewrite hd; auto.
        apply Simulation_remove_hd in h2.
        2: simpl; rewrite <- hd; auto.
        apply h; auto.
      }
    }
  Qed.

  Lemma Simulation_trans:
    forall x y z,
      Simulation x y -> Simulation y z -> Simulation x z.
  Proof.
    cofix h.
    intros [a | a x] [b | b y] [c | c z] h1 h2.
    { inversion_clear h1; inversion_clear h2.
      constructor 1.
      transitivity b; auto. }
    { inversion_clear h1; inversion_clear h2.
      constructor 2.
      transitivity b; auto.
      assert (D := is_constant_Proper_r H1).
      rewrite H; auto. }
    { inversion h2. }
    { inversion_clear h1; inversion_clear h2.
      constructor 2.
      transitivity b; auto.
      cut (is_constant a s2).
      { clear H1 H2.
        induction H3.
        apply is_constant_compat; symmetry; auto.
        intros hh.
        right.
        simpl; auto.
        apply IHis_cons_star; auto.
      }
      apply Simulation_is_constant with y; auto.
    }
    { inversion h1. }
    { inversion h1. }
    { inversion h2. }
    {
      destruct (Simulation_inv_cons_star h1) as [sx [hsx hhxs]].
      assert (ha := Simulation_hd h1).
      simpl in ha.
      assert (pcz:=Simulation_Proper_r h2).
      assert (py:=is_cons_star_Proper_r hhxs).
      rewrite <- ha in h2.
      clear h1 b ha.
      assert (hhh:exists ssx,
                 Simulation sx ssx /\ is_cons_star a ssx z).
      2: {
        destruct hhh as [ssx [hss hss2]].
        constructor 3 with ssx;auto.
        assert (hb := Simulation_hd h2).
        simpl in hb;auto.
        eapply h;eauto.
      }
      clear h py pcz.
      destruct (Simulation_inv_cons_star h2) as [sy [hsy hhys]].
      clear h2 x hsx.
      revert sy hsy hhys.
      induction hhxs;intros sy hsy hhys.
      {
        split with sy;split;auto.
        assert (psy:=Simulation_Proper_r hsy).
        rewrite H;auto.
      }
      {
        assert (ps2:=is_cons_star_Proper_r hhxs).
        assert (psy:=Simulation_Proper_r hsy).
        rewrite <- H in hsy;clear H b psy ps2.
        destruct sy as [b | b sy].
        { apply Simulation_inv_cons_one in hsy;contradict hsy. }
        {
          destruct (Simulation_inv_cons_star hsy) as [s3 [hs3 hh3s]].
          apply (IHhhxs s3 hs3);clear IHhhxs.
          assert (pz:=is_cons_star_Proper_r hhys).
          assert (psy:=is_cons_star_Proper_r hh3s).
          assert (pa:=hd_compat (Simulation_Proper_l hsy):Proper pequiv a).
          apply Simulation_hd in hsy;simpl in hsy.
          rewrite <- hsy in hhys.
          clear hs3 pz psy hsy b c s1 s2 hhxs.
          apply is_cons_star_cons in hhys.
          apply is_cons_star_trans with sy;auto.
        }
      }
    }
  Qed.

  (** *** Monotony / Comonotony
     Monotony and comonotony is a tool to define preservation of some property
     which is preserved between a stream and a smaller (greater) one wrt 
     Simulation.
   *)
  Definition Simulation_monotonic (P: Stream -> Prop) :=
    forall X Y, Simulation X Y -> P Y -> P X.

  Definition Simulation_comonotonic (P: Stream -> Prop) :=
    forall X Y, Simulation X Y -> P X -> P Y.

  (** **** Instant: for a property that applies on the first element of the 
     stream only
   *)
  Definition Instant  (P: Stream -> Prop) :=
    forall x y, (s_hd x) =~= (s_hd y) -> P x <-> P y.

  Lemma Instant_Proper:
    forall P, Instant P -> Proper fequiv P.
  Proof.
    intros P Pinst e1 e2 ee.
    apply Pinst,hd_compat,ee.
  Qed.

  Theorem Instant_Simulation_monotonic:
    forall P, Instant P -> Simulation_monotonic P.
  Proof.
    intros P instP X Y SXY.
    assert (h:=Simulation_hd SXY).
    eapply proj2.
    apply instP.
    auto.
  Qed.

  Theorem Instant_Simulation_comonotonic:
    forall P, Instant P -> Simulation_comonotonic P.
  Proof.
    intros P instP X Y SXY.
    assert (h:=Simulation_hd SXY).
    eapply proj1.
    apply instP.
    auto.
  Qed.

  (** **** Stepwise: for a property that applies 
     - on the single element of the stream or   
     - on the two first elements, in which case, they should be different
   *)
  Definition Stepwise (P: Stream -> Prop): Prop :=
    (forall a b, a =~= b -> P (s_one a) <-> P (s_one b)) /\
    (forall a x, P (s_cons a x) -> ~ (a =~= s_hd x)) /\
    (forall a b x y, a=~=b -> s_hd x =~= s_hd y ->
                     P (s_cons a x) <-> P (s_cons b y)).
  Lemma Stepwise_Proper:
    forall P, Stepwise P -> Proper fequiv P.
  Proof.
    intros P [Pone [Pstep Pcons]] [x|x e1] [y|y e2] ee;ssplit ee;intuition.
    apply Pone,ee.
    apply Pcons;auto.
    apply hd_compat;auto.
  Qed.

  Theorem Stepwise_Simulation_monotonic:
    forall P, Stepwise P -> Simulation_monotonic P.
  Proof.
    intros P [must_step [real_step only_step]] [x | x sx] [y | y sy] H;
      inversion H; subst; hnf.
    firstorder.
    intro h; apply real_step in h;contradict h.
    rewrite <- H3.
    symmetry;auto using is_constant_hd.
    destruct H6.
    assert (_tmp:=Simulation_Proper_l H5).
    rewrite H0 in H5.
    eapply proj2.
    apply only_step;auto using Simulation_hd.
    intro h; apply real_step in h;contradict h.
    transitivity x;[symmetry|];auto.
  Qed.

  Lemma Stepwise_moves:  Stepwise moves.
  (* good candidate for new congruence automation!! *)
  Proof.
    unfold Stepwise,moves;intuition;apply H1.
    rewrite H,H0;auto.
    rewrite <- H,<- H0;auto.
  Qed.

  Lemma Stepwise_strict_moves:  Stepwise strict_moves.
  (* good candidate for new congruence automation!! *)
  Proof.
    unfold Stepwise,strict_moves;intuition;apply H1.
    rewrite H,H0;auto.
    rewrite <- H,<- H0;auto.
  Qed.

  Theorem Stepwise_Always_Simulation_monotonic:
    forall P, Stepwise P -> (forall a, Proper pequiv a -> P (s_one a)) ->
              Simulation_monotonic (Always (fun e => moves e -> P e)).
  Proof.
    intros P SwP;generalize SwP;
      intros [Hmust_step [Hstrict_moves Honly_step]] Pone.
    hnf.
    assert (PPP:Proper fequiv (fun e => moves e -> P e)).
    {
      intros e1 e2 ee;apply impl_compat.
      apply moves_compat,ee.
      apply Stepwise_Proper;auto.
    }
    cofix cfx.
    intros [x | x sx] [y | y sy] H h;
      inversion H;subst;inversion h;subst;
        constructor.
    intros _;apply Pone;proper_.
    intros _;apply Pone;proper_.
    {
      clear H;induction H6.
      {
        intros hh;hnf in hh.
        assert (hh2:moves (s_cons y s2)).
        {
          intros hh2;apply hh.
          rewrite H4.
          rewrite (Simulation_hd H5).
          rewrite (hd_compat H).
          auto.
        }
        apply H2 in hh2;revert hh2;eapply proj2;apply Honly_step;auto.
        rewrite (Simulation_hd H5).
        apply hd_compat,H.
      }
      {
        eapply IHis_cons_star;eauto.
        {
          revert H3.
          eapply proj2,Always_compat;auto.
          right.
          transitivity x;auto.
          symmetry;auto.
          eapply is_cons_star_Proper_r;eauto.
        }
        {
          inversion H3;subst.
          revert H7;eapply proj2,PPP.
          right.
          transitivity x;auto.
          symmetry;auto.
          eapply is_cons_star_Proper_r;eauto.
        }
        {
          inversion H3;subst;auto.
        }
      }
    }
    {
      eapply cfx;eauto;clear cfx.
      clear H;induction H6.
      {
        revert H3.
        eapply proj2,Always_compat;auto;apply PPP.
      }
      {
        apply IHis_cons_star;auto.
        {
          revert H3.
          eapply proj2,Always_compat;auto.
          right.
          transitivity x;auto.
          symmetry;auto.
          eapply is_cons_star_Proper_r;eauto.
        }
        {
          inversion H3;subst.
          revert H7;eapply proj2,PPP.
          right.
          transitivity x;auto.
          symmetry;auto.
          eapply is_cons_star_Proper_r;eauto.
        }
        {
          inversion H3;subst;auto.
        }
      }
    }
  Qed.

  (** **** Eventually *)
  Theorem Eventually_Simulation_monotonic:
    forall P, Simulation_monotonic P -> Simulation_monotonic (Eventually P).
  Proof.
    intros P smP X Y SXY hy.
    revert X SXY.
    induction hy as [y|b y].
    left;apply (smP _ _ SXY H).
    intros [a|a x] sxy.
    {
      apply IHhy.
      inversion sxy;subst.
      inversion H3;simpl in H;symmetry in H;constructor;auto.
    }
    {
      destruct (Simulation_inv_cons sxy).
      right;apply IHhy,H.
      apply IHhy,H.
    }
  Qed.

  Theorem Eventually_Simulation_comonotonic:
    forall P,Proper fequiv P ->
             Simulation_comonotonic P -> Simulation_comonotonic (Eventually P).
  Proof.
    intros P pP smP X Y SXY hx.
    revert Y SXY.
    induction hx as [x|a x].
    left;apply (smP _ _ SXY H).
    intros [b|b y] sxy.
    {
      inversion sxy.
    }
    {
      destruct (Simulation_inv_cons_star sxy) as [yy [h1 h2]].
      right.
      assert (hyy: Eventually P yy) by auto.
      clear - h2 hyy pP.
      induction h2.
      revert hyy;eapply proj1.
      apply Eventually_compat;auto.
      right;apply IHh2,hyy.
    }
  Qed.

  Theorem Simulation_Eventually:
    forall (P: Stream -> Prop)
           (P_compat: Proper fequiv P)
           (P_hd: Instant P)
           x y, Simulation x y ->
                Eventually P x <-> Eventually P y.
  Proof.
    intros P P_compat P_hd e1 e2 sim;hnf in P_hd;split.
    apply Eventually_Simulation_comonotonic;auto.
    apply Instant_Simulation_comonotonic;auto.
    apply Eventually_Simulation_monotonic;auto.
    apply Instant_Simulation_monotonic;auto.
  Qed.

  (** **** Always *)
  Theorem Always_Simulation_monotonic:
    forall P, Proper fequiv P ->
              Simulation_monotonic P -> Simulation_monotonic (Always P).
  Proof.
    intros P pP smP;hnf;cofix hh.
    intros [x | x sx] [y | y sy] H; inversion H; subst; hnf.
    intro h;inversion h;subst;left.
    eapply smP;eauto.
    intro h;inversion h;subst;left.
    eapply smP;eauto.
    intro h;inversion h;subst;right.
    eapply smP;eauto.
    assert (hs2: Always P s2).
    {
      clear - H3 H6 pP.
      induction H6.
      symmetry in H.
      revert H3;eapply proj1,Always_compat;eauto.
      inversion H3;subst;auto.
    }
    eapply hh;eauto.
  Qed.

  (** **** Implies *)
  Lemma Simulation_Impl:
    forall (P Q: Stream -> Prop),
      Proper fequiv P -> Proper fequiv Q ->
      Simulation_comonotonic P -> Simulation_monotonic Q ->
      Simulation_monotonic (fun s => P s -> Q s).
  Proof.
    intros P Q pP pQ sP sQ X Y sim.
    intros h hP.
    apply sQ with Y; auto.
    apply h.
    apply sP with X; auto.
  Qed.

  (** **** Acc Rskip *)
  Lemma Acc_Rskip_Simulation_monotonic:
    forall a, Proper pequiv a -> Simulation_monotonic (fun e => Acc (Rskip a) e).
  Proof.
    intros u pu x y simu accy.
    apply Transitive_Closure.Acc_clos_trans in accy.
    revert x simu;induction accy as [y Faccy Haccy];intros [a | a x] simu.
    {
      constructor;intros sx hsx.
      assert (pax:=Simulation_Proper_l simu).
      specialize (hsx pax).
      assert (eau:a=~=u).
      {
        hnf in hsx;intuition.
        symmetry;apply (hd_compat H).
        symmetry;apply (hd_compat H1).
      }
      apply Simulation_inv_one in simu.
      assert (py:=is_constant_Proper_r simu).
      apply is_constant_not_Acc in simu.
      contradict simu.
      apply (Acc_intro y) in  Faccy.
      eapply Inclusion.Acc_incl in Faccy.
      2: apply Transitive_Closure.incl_clos_trans.
      revert Faccy;apply Acc_Rskip_compat;auto.
    }
    {
      destruct y as [b|b y].
      inversion simu.
      constructor;intros sx hsx.
      assert (pax:=Simulation_Proper_l simu).
      specialize (hsx pax).
      assert (eau:a=~=u).
      {
        hnf in hsx;intuition.
        symmetry;apply (hd_compat H).
        symmetry;apply (hd_compat H1).
      }
      assert (eab:=Simulation_hd simu);simpl in eab.
      apply Simulation_inv_cons_star in simu.
      destruct simu as [sy [simu ics]].
      assert (lem:Relation_Operators.clos_trans (Rskip u) sy (s_cons b y)).
      {
        clear Faccy Haccy sx hsx x simu pax.
        assert (psy:=is_cons_star_Proper_l ics).
        assert (py:=is_cons_star_Proper_r ics).
        rewrite eau in ics,eab;clear a eau.
        clear py psy.
        revert b eab.
        induction ics.
        {
          intros b eb;left.
          hnf.
          intros.
          left;right;auto.
        }
        {
          intros bb ebb;right with (s_cons b s2).
          apply IHics.
          transitivity bb;auto.
          left.
          intros _.
          left; right; auto; right;auto.
          apply is_cons_star_Proper_r in ics; auto.
        }
      }
      eapply Haccy;eauto.
      assert (x =~= sx).
      {
        intuition.
        symmetry;inversion H;auto.
        ssplit H1.
      }
      assert (psy:=is_cons_star_Proper_l ics).
      rewrite H in simu;auto.
    }
  Qed.

  (** **** Simulation preserves moves *)
  Definition two_fst_eq (s s': Stream): Prop :=
    match s with
      s_one _ => False
    | s_cons a ss =>
      match s' with s_one _ => False
               | s_cons a' ss' =>
                 a =~= a' /\ s_hd ss =~= s_hd ss' end end.

  Lemma Simulation_preserves_move:
    forall s suf,
      is_suffix s suf -> strict_moves suf ->
      forall sim, Simulation sim s ->
                  exists suf_sim, is_suffix sim suf_sim /\
                                  two_fst_eq suf suf_sim.
  Proof.
    intros s suf S M.
    induction S.
    {
      intros sim SS.
      exists sim.
      split.
      left.
      apply Simulation_Proper_l with s; try assumption.
      destruct suf; simpl in M; [tauto|].
      destruct sim as [xsim | xsim sim].
      - inversion SS; subst; clear SS.
        inversion H; elim H.
        contradict M.
        apply is_constant_hd in H2.
        inversion H; subst.
        rewrite <- (hd_compat H7), <- H5, <- H1;
          symmetry; assumption.
      - simpl.
        assert (tmp: Simulation (s_cons xsim sim) (s_cons a suf)).
        eapply Simulation_compat; try assumption.
        3: apply SS.
        apply Simulation_Proper_l with s; try assumption.
        symmetry; apply H.
        clear SS; inversion tmp; subst; clear tmp.
        split.
        + symmetry; assumption.
        + apply Simulation_hd in H5; try assumption.
          rewrite H5.
          apply hd_compat.
          symmetry; apply is_cons_star_epsilon with xsim.
          assumption.
          intros F; contradict M; rewrite <- F; symmetry; assumption.
    }{
      intros sim SS.
      destruct sim as [xsim | xsim sim];
        inversion SS; subst; clear SS.
      - assert (T := constant_suffix S H3).
        destruct suf; simpl in M; [tauto|].
        contradict M.
        inversion T; subst; simpl in H1.
        rewrite H1.
        apply Always_P_P in H4; symmetry; assumption.
      - destruct s as [xs | xs s].
        + inversion S; subst; clear S.
          inversion H; subst; clear H; simpl in M; tauto.
        + inversion H5; subst; clear H5.
          * specialize (IHS sim).
            assert (tmp: Simulation sim (s_cons xs s)).
            eapply Simulation_compat; try assumption.
            apply Simulation_Proper_l with s2; try assumption.
            symmetry; apply H.
            assumption.
            specialize (IHS tmp).
            clear tmp.
            destruct IHS as [suf_sim [SS EQ]].
            exists suf_sim.
            split.
            right; apply SS.
            apply EQ.
          * assert (T: Simulation (s_cons xsim sim) (s_cons xs s)).
            apply sim_cons with s2; assumption.
            specialize (IHS _ T).
            destruct IHS as [suf_sim [SS EQ]].
            exists suf_sim.
            split.
            apply SS.
            apply EQ.
    }
  Qed.

  (** **** (Squeeze s) is smallest than s *)
  Lemma Simulation_Squeeze:
    forall (s: Stream) (p: Proper pequiv s) (w: Squeezable s),
      Simulation (Squeeze p w) s.
  Proof.
    cofix h.
    intros [x | x s] p w; rewrite Squeeze_unfold.
    {
      apply sim_one_one.
      inversion p; assumption.
    }
    {
      assert (px: Proper pequiv x).
      inversion p; assumption.
      assert (ps: Proper pequiv s).
      inversion p; assumption.

      case (Squeezable_inv w).
      - intros i; simpl in i.
        apply sim_one_cons.
        apply px.
        apply is_constant_inv with x; assumption.
      - intros a.
        eapply sim_cons with (Skip (hd_compat p) a).
        apply px.
        apply h.
        assert (a' := Acc_Rskip_cons a).
        assert (tmp := (Skip_cons (hd_compat p) a a' ps)).
        rewrite (is_cons_star_compat _ _ px tmp ps).
        apply is_cons_star_Skip.
        apply ps.
    }
  Qed.

  Lemma Simulation_Skip :
    forall a x b y
           (pa:Proper pequiv a) (pb:Proper pequiv b)
           (haccx: Acc (Rskip a) x) (haccy: Acc (Rskip b) y),
      Simulation (s_cons a x) (s_cons b y) ->
      Simulation (Skip pa haccx) (Skip pb haccy).
  Proof.
    Arguments Skip : clear implicits.
    intros a x b y pa pb haccx;revert b pb y.
    induction haccx as [x Faccx Haccx] using Acc_ind_dep;
      intros b pb y haccy simu.
    destruct x as [aa | aa x];simpl.
    {
      apply is_constant_Simulation.
      assert (eab:=Simulation_hd simu).
      apply Simulation_inv_cons_star in simu.
      destruct simu as [sy [simu ics]].
      simpl in eab.
      apply Simulation_inv_one in simu.
      destruct (pequiv_dec a aa) as [WEQ| NEQ].
      {
        assert (eaa: a=~=aa).
        {
          apply WEQ;try proper_.
          eapply is_constant_Proper_l;eauto.
        }
        assert (ht:is_constant a y).
        {
          apply is_cons_star_is_constant with sy;auto.
          assert (psy:=is_constant_Proper_r simu).
          rewrite eaa;assumption.
        }
        eelim (is_constant_not_Acc);eauto.
        revert haccy;apply Acc_Rskip_compat;auto.
        apply (is_constant_Proper_r ht).
      }
      {
        cut (forall p w, sy =~= Skip b p y w).
        {
          assert (paa:=is_constant_Proper_l simu).
          intro h;rewrite <- h;auto.
        }
        {
          rewrite eab in NEQ.
          assert (psy:=is_cons_star_Proper_l ics).
          assert (py:=is_cons_star_Proper_r ics).
          rewrite eab in ics.
          intros P W;symmetry;apply Skip_is_cons_star;auto.
          intro h;apply NEQ.
          rewrite h.
          apply is_constant_hd in simu.
          auto.
        }
      }
    }
    {
      destruct (pequiv_dec a aa) as [WEQ| NEQ].
      {
        assert (eaa: a=~=aa).
        {
          apply WEQ;try proper_.
          apply Simulation_Proper_l in simu.
          ssplit simu;intuition.
          ssplit H0;intuition.
        }
        assert (eab:=Simulation_hd simu);simpl in eab.
        eapply Haccx.
        assert (hh:=simu).
        apply Simulation_remove_hd in hh.
        revert hh;apply Simulation_compat.
        right;auto.
        apply Simulation_Proper_l in simu.
        inversion simu;subst.
        inversion H4;subst;auto.
        apply Simulation_Proper_r in simu;auto.
        auto.
      }
      {
        assert (eab:=Simulation_hd simu);simpl in eab.
        apply Simulation_inv_cons_star in simu.
        destruct simu as [sy [simu ics]].
        assert (eaa:=Simulation_hd simu);simpl in eaa.
        assert (paax:=Simulation_Proper_l simu).
        revert simu;apply Simulation_compat;auto.
        rewrite eaa in NEQ.
        clear aa eaa paax Faccx Haccx.
        rewrite eab in NEQ.
        assert (psy:=is_cons_star_Proper_l ics).
        assert (py:=is_cons_star_Proper_r ics).
        rewrite eab in ics.
        apply Skip_is_cons_star;auto.
      }
    }
    Arguments Skip: default implicits.
  Qed.

  (** (Squeeze s) is the minimum stream among the ones smaller than s *)
  Lemma Simulation_Squeeze_min:
    forall (s: Stream) (p:Proper pequiv s) (c: Squeezable s) x,
      Simulation x s -> Simulation (Squeeze p c) x.
  Proof.
    cofix cfx.
    intros [a | a s] p w x hh; rewrite Squeeze_unfold.
    { inversion_clear hh.
      constructor 1; symmetry; auto. }
    case (Squeezable_inv w).
    { intros const; simpl in const.
      destruct x as [b | b x].
      constructor 1.
      apply Simulation_hd in hh; symmetry; auto.
      constructor 2.
      apply Simulation_hd in hh; symmetry; auto.
      inversion_clear hh.
      apply Simulation_is_constant' with s2; auto.
      apply is_cons_star_is_constant' with s.
      inversion const; auto.
      assert (pa := is_constant_Proper_l const).
      revert H1; apply is_cons_star_compat; auto.
      eapply Simulation_Proper_r; eauto.
      ssplit p; apply p.
    }
    { destruct x as [b | b x].
      { inversion_clear hh.
        assert (hh0: is_constant a (s_cons a s)).
        { right; [symmetry|]; auto.
          revert H0; apply is_constant_compat;
            [symmetry; auto|ssplit p; apply p]. }
        apply is_constant_not_Acc in hh0; tauto.
      }
      { simpl.
        intros Aas.
        assert (pb := hd_compat (Simulation_Proper_l hh)); simpl in pb.
        assert (Abx: Acc (Rskip b) (s_cons b x)).
        {
          assert (eab:= Simulation_hd hh).
          assert (pas:= Simulation_Proper_r hh).
          eapply Acc_Rskip_Simulation_monotonic in hh.
          apply hh.
          revert Aas; apply Acc_per_impl.
          apply Rskip_compat
          ; auto.
          2:auto.
          intros xx yy hskip pxx.
          hnf in hskip.
          specialize (hskip pxx).
          intuition;try proper_.
          inversion H;proper_.
          auto.
        }
        assert (As := Acc_Rskip_cons Aas).
        assert (Ax := Acc_Rskip_cons Abx).
        assert (pas:=Simulation_Proper_r hh).
        assert (pbx:=Simulation_Proper_l hh).
        ssplit pas;ssplit pbx.
        destruct pas;destruct pbx.
        constructor 3 with (Skip pb Abx).
        2: {
          apply cfx.
          cut (Simulation (Skip pb Ax) (Skip (hd_compat p) As)).
          apply Simulation_compat; apply Skip_cons;auto.
          apply Simulation_Skip; auto.
        }
        {
          apply Simulation_hd in hh;symmetry;apply hh.
        }
        {
          cut (is_cons_star b (Skip pb Ax) x).
          apply is_cons_star_compat; auto.
          apply Simulation_hd in hh; auto.
          apply Skip_cons;auto.
          apply is_cons_star_Skip; auto.
        }
      }
    }
  Qed.

  (** (Squeeze s) contains no duplicate *)
  Lemma  Squeeze_no_repeat:
    forall (s: Stream) (p: Proper pequiv s) (w: Squeezable s),
      Always moves (Squeeze p w).
  Proof.
    cofix h.
    intros [x | x s] p w; rewrite Squeeze_unfold.
    {
      apply always_one.
      unfold moves.
      tauto.
    }
    {
      assert (px: Proper pequiv x).
      inversion p; assumption.
      assert (ps: Proper pequiv s).
      inversion p; assumption.

      case (Squeezable_inv w).
      - intros i.
        apply always_one.
        unfold moves.
        tauto.
      - intros a.
        apply always_cons.
        + unfold moves.
          rewrite Squeeze_hd.
          rewrite (Skip_cons (hd_compat p) a (Acc_Rskip_cons a) ps).
          apply Skip_hd.
        + apply h.
    }
  Qed.

End Simulation_Squeezing.

Unset Implicit Arguments.
Close Scope signature_scope.
