(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Nov 2016
 *)
(* begin show *)

(** **** Global imports **)
From Coq Require Import Setoid.
From Coq Require Import SetoidList.
From Coq Require Import Morphisms.
From Coq Require Import RelationClasses.
From Coq Require Import SetoidClass.
From Coq Require Import RelationPairs.
From Coq Require Import Arith.

(** **** Local imports **)
From Padec Require Import OptionUtil.
From Padec Require Import SetoidUtil.
From Padec Require Import WellfoundedUtil.
From Padec Require Import RelationA.

(* end show *)
Open Scope signature_scope.
Set Implicit Arguments.

Scheme Acc_ind_dep := Induction for Acc Sort Prop.
Scheme Acc_rec_dep := Induction for Acc Sort Type.

(** * Stream Definition
 Finite or infinite sequence of at least one element of a given type A
 *)
Section StreamA.

  Context {A: Type} `{SA: PartialSetoid A}.

  CoInductive Stream: Type :=
  | s_one: A -> Stream
  | s_cons: A -> Stream -> Stream.

  (** * Equality on Stream [eqStream]
   Pointwise equality on streams, based on equality on elements
   *)
  CoInductive eqStream: relation Stream :=
  | eqStream_one: forall a1 a2, pequiv a1 a2 -> eqStream (s_one a1) (s_one a2)
  | eqStream_cons: forall a1 a2 s1 s2,
      pequiv a1 a2 -> eqStream s1 s2 -> eqStream (s_cons a1 s1) (s_cons a2 s2).

  (** eqStream is a partial equivalence relation (PER) *)
  Global Instance eqStream_Symmetry: Symmetric eqStream.
  Proof.
    cofix h.
    intros [a1 | a1 s1] [a2 | a2 s2] e; inversion e; subst.
    - left; symmetry; auto.
    - right.
      symmetry; auto.
      apply h; auto.
  Qed.

  Global Instance eqStream_Transitive: Transitive eqStream.
  Proof.
    cofix h.
      intros [a1 | a1 s1] [a2 | a2 s2] [a3 | a3 s3] e12 e23;
        inversion e12; inversion e23; subst.
      - left; transitivity a2; auto.
      - right.
        transitivity a2; auto.
        apply h with s2; auto.
  Qed.

  (** Stream is a partial setoid: *)
  Global Instance Stream_PartialSetoid: PartialSetoid Stream :=
    @Build_PartialSetoid _ eqStream (Build_PER _ _ _).

  (** eqStream: inversion tools *)
  Lemma eqStream_inv: forall a1 a2 s1 s2,
      s_cons a1 s1 =~= s_cons a2 s2 -> s1 =~= s2.
  Proof.
    intros a1 a2 s1 s2 h; inversion h; auto.
  Qed.

  Lemma eqStream_inv_l: forall a1 a2 s1 s2,
      s_cons a1 s1 =~= s_cons a2 s2 -> a1 =~= a2.
  Proof.
    intros a1 a2 s1 s2 h;inversion h;auto.
  Qed.

  Lemma eqStream_inv_one: forall a1 a2,
      s_one a1 =~= s_one a2 -> a1 =~= a2.
  Proof.
    intros;inversion H;auto.
  Qed.

  Lemma eqStream_inv_cons: forall a1 a2 s1 s2,
      s_cons a1 s1 =~= s_cons a2 s2 -> a1 =~= a2 /\ s1 =~= s2.
  Proof.
    intros a1 a2 s1 s2 h;inversion h;auto.
  Qed.

  Lemma eqStream_inv_False_1: forall a1 a2 s1,
      s_cons a1 s1 =~= s_one a2 -> False.
  Proof.
    intros a1 a2 s1 h;inversion h;auto.
  Qed.
  Lemma eqStream_inv_False_2: forall a1 a2 s2,
      s_one a1 =~= s_cons a2 s2 -> False.
  Proof.
    intros a1 a2 s2 h;inversion h;auto.
  Qed.

  (** * Tools on the Stream definition *)
  Global Instance s_one_compat:  Proper fequiv s_one.
  Proof. intros a b eab; left; assumption. Qed.

  Global Instance s_cons_compat: Proper fequiv s_cons.
  Proof. intros a b eab x y exy; right; assumption. Qed.

  (** Access to the head of a stream: *)
  Definition s_hd (s: Stream): A :=
    match s with (s_one a | s_cons a _) => a end.

  (** Access to the tail of a stream: *)
  Definition s_tl (s: Stream): option Stream :=
    match s with
    | (s_cons _ tl) => Some tl
    | _ => None
    end.

  (** Access to the n-th suffix of a stream: *)
  Fixpoint s_nth (n: nat) (s: Stream): Stream :=
    match n with
      O => s
    | S n' =>
      match s with
        s_one a => s_one a
      | s_cons a s' => s_nth n' s'
      end
    end.

  Lemma s_nth_one: forall (a: A) n, s_nth n (s_one a) = (s_one a).
  Proof.
    intros a;induction n as [|n hn];simpl;congruence.
  Qed.

  Global Instance hd_compat: Proper fequiv s_hd.
  Proof.
    intros s1 s2 h; inversion h; auto.
  Qed.

  Global Instance tl_compat: Proper fequiv s_tl.
  Proof.
    intros s1 s2 h; inversion h; simpl; auto.
  Qed.

  Global Instance nth_compat: Proper ((@eq nat) ==> fequiv) s_nth.
  Proof.
    intros n_ n en;subst.
    induction n as [|n hrec].
    intros e1 e2 ee;simpl;auto.
    intros e1 e2 ee;destruct ee;simpl.
    constructor 1;auto.
    apply hrec; auto.
  Qed.

  (** ** Stream unfolding *)
  Definition Stream_decompose (s: Stream): Stream :=
    match s with
      s_one x => s_one x | s_cons x s => s_cons x s
    end.

  Lemma Stream_decomposition_lemma: forall (s: Stream), s = Stream_decompose s.
  Proof.
    intros s; destruct s; simpl; reflexivity.
  Qed.

  Global Instance Stream_decompose_compat: Proper fequiv Stream_decompose.
  Proof.
    intros s1 s2 hs.
    destruct hs as [ a1 a2 ha | a1 a2 s1 s2 ha hs];simpl;[left|right];auto.
  Qed.

  Lemma Stream_decompose_inv:
    forall s1 s2,
      (Stream_decompose s1) =~= (Stream_decompose s2) ->
      s1 =~= s2.
  Proof.
    intros s1 s2 hs;destruct s1;destruct s2;apply hs.
  Qed.

End StreamA.

(** Tactic for eqStream *)
Ltac ssplit pe :=
  (*    hnf in pe;*)
  try unfold Proper in pe;
  match (type of pe) with
  | s_cons ?a1 ?s1 =~= s_cons ?a2 ?s2 =>
    let X := fresh "X" in
    assert (X := eqStream_inv_cons pe);
    clear pe; rename X into pe
  | s_one ?a1 =~= s_one ?a2 =>
    let X := fresh "X" in
    assert (X := eqStream_inv_one pe);
    clear pe; rename X into pe
  | s_one ?a1 =~= s_cons ?a2 ?s2 =>
    let X := fresh "X" in
    assert (X := eqStream_inv_False_2 pe);
    tauto
  | s_cons ?a1 ?s1 =~= s_one ?a2 =>
    assert (X := eqStream_inv_False_1 pe);
    tauto
  end.

(** * Some LTL Operators on Stream *)
Section LTL_StreamA.

  Context {A: Type} `{SA: PartialSetoid A}.
  Notation Stream := (Stream (A := A)).
  
  (** ** Definitions *)

  (** *** Eventually, definitions and tools *)
  Inductive Eventually (P: Stream -> Prop): Stream -> Prop :=
  | eventually_now: forall s, P s -> Eventually P s
  | eventually_later: forall a s, Eventually P s -> Eventually P (s_cons a s).

  (** Finally is the counter part of Eventually in Type *)
  Inductive Finally (P: Stream -> Prop): Stream -> Type :=
  | finally_now: forall s, P s -> Finally P s
  | finally_not_yet: forall a s, ~P (s_cons a s) -> Finally P s ->
                                 Finally P (s_cons a s).

  Lemma Finally_Eventually (P: Stream -> Prop):
    forall s, Finally P s -> Eventually P s.
  Proof.
    intros s h; induction h; [left|right];assumption.
  Qed.

  Lemma Eventually_Dec_Finally (P: Stream -> Prop)
        (Pdec: forall s, {P s} + {~P s}):
    forall s, Eventually P s -> Finally P s.
  Proof.
    pose (R s1 s2 := exists a, s_cons a s1 = s2 /\ ~P s2).
    intros s Ev.
    assert (W: Acc R s).
    {
      induction Ev.
      constructor.
      intros y [_ [_ h]].
      elim (h H).
      constructor.
      intros y [a' [e h]].
      inversion e;subst.
      apply IHEv.
    }
    {
      induction W.
      destruct (Pdec x).
      left;auto.
      destruct x as [a|a xx].
      contradict n.
      inversion Ev;assumption.
      right.
      apply n.
      apply X.
      split with a;split;auto.
      inversion Ev.
      congruence.
      assumption.
    }
  Qed.

  Lemma Eventually_Dec_Finally'
        (P: Stream -> Prop) (Pdec: Weak_Dec pequiv P):
    forall s, Proper pequiv s -> Eventually P s -> Finally P s.
  Proof.
    pose (R s1 s2 := exists a, s_cons a s1 = s2 /\ ~P s2).
    intros s ps Ev.
    assert (W: Acc R s).
    {
      induction Ev.
      constructor.
      intros y [_ [_ h]].
      elim (h H).
      constructor.
      intros y [a' [e h]].
      inversion e;subst.
      apply IHEv.
      inversion ps; assumption.
    }
    {
      induction W.
      destruct (Pdec x).
      left;auto.
      destruct x as [a|a xx].
      contradict n.
      inversion Ev;assumption.
      right.
      apply n.
      apply X.
      split with a;split;auto.
      inversion ps; assumption.
      inversion Ev.
      congruence.
      assumption.
    }
  Qed.

  (** *** Always *)
  CoInductive Always (P: Stream -> Prop): Stream -> Prop  :=
  | always_one: forall a, P (s_one a) -> Always P (s_one a)
  | always_cons: forall a s, P (s_cons a s) -> Always P s ->
                             Always P (s_cons a s).

  (** Finally is the counter part of Eventually in Type *)
  CoInductive Forever (P: Stream -> Type): Stream -> Type :=
  | forever_one: forall a, P (s_one a) -> Forever P (s_one a)
  | forever_cons: forall a s, P (s_cons a s) -> Forever P s ->
                              Forever P (s_cons a s).

  (** *** (Weak) Until *)
  CoInductive WUntil (P Q: Stream -> Prop): Stream -> Prop :=
  | wuntil_now: forall s, Q s -> WUntil P Q s
  | wuntil_later: forall a s, P (s_cons a s) -> WUntil P Q s ->
                              WUntil P Q (s_cons a s).

  Inductive Until (P Q: Stream -> Prop): Stream -> Prop :=
  | until_now: forall s, Q s -> Until P Q s
  | until_later: forall a s, P (s_cons a s) -> Until P Q s ->
                             Until P Q (s_cons a s).

  (** ** Properties *)
  Lemma Always_inv_2: forall (P: Stream -> Prop) a s,
      Always P (s_cons a s) -> Always P s.
  Proof. intros P a s aps; inversion aps; auto. Qed.

  Lemma Always_P_P:
    forall P (s: Stream), Always P s -> P s.
  Proof.
    intros P e AA.
    destruct e; inversion AA; subst; auto.
  Qed.

  Lemma Forever_P_P:
    forall P (s: Stream), Forever P s -> P s.
  Proof.
    intros P [g | g e] h; inversion h; subst; auto.
  Qed.    

  Lemma Always_impl_P:
    forall (P Q: Stream -> Prop),
      (forall s, P s -> Q s) -> forall s, Always P s -> Always Q s.
  Proof.
    intros P Q impl.
    cofix Always_impl_P.
    intros e AA.
    destruct e.
    apply always_one; apply impl; inversion AA; subst; assumption.
    apply always_cons.
    apply impl; inversion AA; subst; assumption.
    apply Always_impl_P.
    inversion AA; subst; assumption.
  Qed.

  Lemma Always_MP:
    forall (P Q: Stream -> Prop) s,
      Always (fun s => P s -> Q s) s -> Always P s -> Always Q s.
  Proof.
    intros P Q;cofix Always_MP.
    intros s Aimpl AP.
    destruct s as [a| a s];inversion Aimpl;subst;
      inversion AP;subst;[left|right];auto.
  Qed.

  Lemma Always_conj:
    forall (P Q: Stream -> Prop) s,
      Always (fun s => P s /\ Q s) s <-> Always P s /\ Always Q s.
  Proof.
    intros P Q; split; intros h.
    { split; eapply Always_impl_P; [|apply h| |apply h];
        simpl; tauto. }
    { revert h; revert s; cofix hh.
      intros [a| a s] [hp hq].
      left; split; [inversion hp | inversion hq]; auto.
      right.
      split; [inversion hp | inversion hq]; auto.
      apply hh; split; [inversion hp | inversion hq]; auto.
    }
  Qed.

  Lemma Always_Always:
    forall (P: Stream -> Prop) s, Always P s <-> Always (Always P) s.
  Proof.
    intros P s; split; revert s; cofix h.
    { intros [a | a s] hh.
      left; auto.
      right; auto.
      apply h; inversion hh; auto.
    }
    { intros [a | a s] hh.
      inversion hh; auto.
      inversion hh; auto.
    }
  Qed.

  Lemma Always_forall' (P Q: Stream -> Prop):
    (forall s, Q s -> P s) -> forall s, Always Q s -> Always P s.
  Proof.
    intros f; cofix h.
    intros [a | a s]; intros q.
    - left; apply f.
      apply Always_P_P in q; auto.
    - right; [apply f | apply h].
      apply Always_P_P in q; auto.
      apply Always_inv_2 in q; auto.
  Qed.

  Lemma Forever_forall (P Q: Stream -> Prop):
    (forall s, Q s -> P s) -> forall s, Always Q s -> Forever P s.
  Proof.
    intros f; cofix h.
    intros [a | a s]; intros q.
    - left; apply f.
      apply Always_P_P in q; auto.
    - right; [apply f | apply h].
      apply Always_P_P in q; auto.
      apply Always_inv_2 in q; auto.
  Qed.

  Lemma Forever_forall_Type (P: Stream -> Type) (Q: Stream -> Prop):
    (forall s, Q s -> P s) -> forall s, Always Q s -> Forever P s.
  Proof.
    intros f; cofix h.
    intros [a | a s]; intros q.
    - left; apply f.
      apply Always_P_P in q; auto.
    - right; [apply f | apply h].
      apply Always_P_P in q; auto.
      apply Always_inv_2 in q; auto.
  Qed.
  
  Lemma Eventually_impl_P:
    forall (P Q: Stream -> Prop),
      (forall s, P s -> Q s) -> forall s, Eventually P s -> Eventually Q s.
  Proof.
    intros P Q impl e E.
    induction E.
    apply eventually_now; apply impl; assumption.
    apply eventually_later; assumption.
  Qed.

  Lemma Eventually_P_Always_not_P:
    forall P (s: Stream), Eventually P s -> Always (fun s' => ~ P s') s -> False.
  Proof.
    intros P e E AA.
    induction E.
    destruct AA; auto.
    inversion AA; subst; auto.
  Qed.

  Lemma Always_P_Eventually_not_P:
    forall P (s: Stream), Always P s -> Eventually (fun s' => ~ P s') s -> False.
  Proof.
    intros P e AA E.
    apply Eventually_P_Always_not_P in E.
    tauto.
    eapply Always_impl_P.
    2: apply AA.
    intros ee Pee.
    tauto.
  Qed.

  Lemma AE_not_EA:
    forall P (s: Stream), Always (Eventually P) s ->
                          ~ Eventually (Always (fun s' => ~P s')) s.
  Proof.
    intros P e AA E.
    induction E.
    2: inversion AA; apply IHE; assumption.
    apply Always_P_Eventually_not_P in H.
    tauto.
    apply Always_P_P in AA.
    destruct AA as [ e N | g e L].
    apply eventually_now; tauto.
    apply eventually_later; eapply Eventually_impl_P.
    2: apply L.
    tauto.
  Qed.

  Lemma AE_tool:
    forall (P Q: Stream -> Prop) s,
      Always (fun s': Stream => Q s' -> Eventually P s') s ->
      Eventually Q s -> Eventually P s.
  Proof.
    intros P Q e AA E.
    induction E.

    apply Always_P_P in AA.
    auto.

    apply eventually_later.
    apply IHE.
    inversion AA; subst; assumption.
  Qed.

  Lemma AE_tool2:
    forall P (s: Stream) a,
      Always (Eventually P) s -> Always (Eventually P) (s_cons a s).
  Proof.
    intros P e g AA.
    apply always_cons.
    2: apply AA.
    apply eventually_later.
    inversion AA; subst; assumption.
  Qed.

  Lemma AE_tool3:
    forall (P Q: Stream -> Prop) s,
      Always (fun s': Stream => Q s' -> Eventually P s') s ->
      Always (Eventually Q) s -> Always (Eventually P) s.
  Proof.
    intros P Q.
    cofix AE_tool3.
    destruct s as [g | g e].

    (* e_one *)
    intros A1 A2.
    inversion A1; subst.
    inversion A2; subst.
    inversion H1; subst.
    apply always_one; apply H0; apply H.

    (* e_cons *)
    intros A1 A2.
    inversion A1; subst; clear A1.
    inversion A2; subst; clear A2.
    apply always_cons.
    2: apply AE_tool3; assumption.
    clear AE_tool3.

    eapply AE_tool.
    2: apply H3.
    simpl.
    apply always_cons; assumption.
  Qed.

  Lemma AE_tool4:
    forall (P Q: Stream -> Prop)
           (Q_dec: forall s, { Q s } + { ~Q s }) s,
      Always (fun s': Stream => Q s' -> Eventually P s') s ->
      ~ Eventually (Always (fun s': Stream => Q s' /\ ~ P s')) s.
  Proof.
    intros P Q Q_dec e AA E.
    specialize (Always_P_Eventually_not_P AA); clear AA.
    intros EE; apply EE; clear EE.
    eapply Eventually_impl_P.
    2: apply E.
    clear E e.
    intros e AA.
    specialize (Always_P_Eventually_not_P AA); intros FF.
    intros F.
    case (Q_dec e); intros EN.
    specialize (F EN); clear EN.
    apply FF; clear AA FF.
    eapply Eventually_impl_P.
    2: apply F.
    clear F e.
    intros e X; simpl in X; tauto.
    apply Always_P_P in AA.
    tauto.
  Qed.

    Lemma Eventually_Dec_Finally'' (P: Stream -> Prop):
    forall s, Proper pequiv s -> Forever (fun s => {P s} + {~P s}) s ->
              Eventually P s -> Finally P s.
  Proof.
    pose (R s1 s2 := exists a, s_cons a s1 = s2 /\ ~P s2).
    intros s ps fdec Ev.
    assert (W: Acc R s).
    {
      induction Ev.
      constructor.
      intros y [_ [_ h]].
      elim (h H).
      constructor.
      intros y [a' [e h]].
      inversion e;subst.
      apply IHEv.
      inversion ps; assumption.
      inversion fdec; auto.
    }
    {
      induction W.
      assert (dec: {P x} + {~ P x}) by (inversion fdec; auto).
      destruct dec.
      left;auto.
      destruct x as [a|a xx].
      contradict n.
      inversion Ev;assumption.
      right.
      apply n.
      apply X.
      split with a;split;auto.
      inversion ps; assumption.
      inversion fdec; auto.
      inversion Ev; subst; [tauto|auto].
    }
  Qed.

  (** ** Compatibility *)
  Global Instance Eventually_compat: Proper fequiv Eventually.
  Proof.
    assert (ok: forall P1 P2 e1 e2,
               P1 =~= P2 -> e1 =~= e2 -> Eventually P1 e1 ->
               Eventually P2 e2).
    {
      intros P1 P2 e1 e2 eP ee E.
      revert ee. revert e2.
      induction E; intros e2 ee.
      - apply eventually_now.
        eapply eP.
        apply ee.
        assumption.
      - inversion ee; subst.
        apply eventually_later.
        apply IHE.
        assumption.
    }{
      intros P1 P2 eP e1 e2 ee.
      split.
      apply ok; assumption.
      apply ok.
      intros x b eb; symmetry; apply eP; symmetry; assumption.
      symmetry; assumption.
    }
  Qed.

  Lemma Finally_compat_impl:
    forall (P Q: Stream -> Prop) (pq: P =~= Q) x y (exy: x =~= y),
      Finally P x -> Finally Q y.
  Proof.
    intros P1 P2 eP e1 e2 ee E.
    revert ee. revert e2.
    induction E; intros e2 ee.
    - left.
      apply eP with s; auto.
    - destruct e2 as [a2 | a2 e2].
      + ssplit ee.
      + right.
        intro h; contradict n.
        eapply eP; [apply ee | apply h].
        apply IHE.
        ssplit ee; apply ee.
  Qed.

  Global Instance Always_compat: Proper fequiv Always.
  Proof.
    assert (ok: forall P1 P2 e1 e2, P1 =~= P2 -> e1 =~= e2 ->
                                    Always P1 e1 -> Always P2 e2).
    { intros P1 P2 e1 e2 eP ee AA.
      revert ee. revert e2. revert AA. revert e1.
      cofix h.
      intros [g1 | g1 e1] AA [g2 | g2 e2] ee;
        inversion ee; subst.
      - left.
        apply eP with (s_one g1); auto.
        apply Always_P_P; auto.
      - right.
        eapply eP; [apply ee |]; auto.
        apply Always_P_P; auto.
        apply h with e1; auto.
        inversion AA; assumption.
    }
    { intros P1 P2 eP e1 e2 ee.
      split.
      apply ok; assumption.
      apply ok; symmetry; auto.
    }
  Qed.

  Global Instance WUntil_compat: Proper fequiv WUntil.
  Proof.
    assert (ok: forall P1 P2 Q1 Q2 e1 e2,
               P1 =~= P2 -> Q1 =~= Q2 -> e1 =~= e2 ->
               WUntil P1 Q1 e1 -> WUntil P2 Q2 e2).
    { intros P1 P2 Q1 Q2; cofix h.
      intros e1 e2 eP eQ ee wu.
      inversion wu;subst.
      constructor 1.
      revert H;eapply proj1,eQ,ee.
      inversion ee;subst.
      constructor 2.
      revert H;eapply proj1,eP,ee.
      apply h with s;auto.
    }
    { intros P1 P2 eP Q1 Q2 eQ e1 e2 ee;split;apply ok;auto.
      intros a1 a2 ea;symmetry in ea;symmetry;auto.
      intros a1 a2 ea;symmetry in ea;symmetry;auto.
      symmetry;auto.
    }
  Qed.

  Global Instance Until_compat: Proper fequiv Until.
  Proof.
    cut (forall P1 P2 Q1 Q2 e1 e2,
            P1 =~= P2 -> Q1 =~= Q2 -> e1 =~= e2 ->
            Until P1 Q1 e1 -> Until P2 Q2 e2).
    { intros ok P1 P2 eP Q1 Q2 eQ e1 e2 ee;split;apply ok;auto.
      intros a1 a2 ea;symmetry in ea;symmetry;auto.
      intros a1 a2 ea;symmetry in ea;symmetry;auto.
      symmetry;auto.
    }
    { intros P1 P2 Q1 Q2 e1 e2 EP EQ ee U1.
      revert P2 EP Q2 EQ e2 ee; induction U1 as [ s1 IH| a1 s1 hp U1 IH]; intros.
      - left; rewrite <- (EQ _ _ ee); auto.
      - destruct e2 as [a2 | a2 s2].
        inversion ee.
        right.
        rewrite <- (EP _ _ ee); auto.
        apply IH; auto.
        inversion ee; auto.
    }
  Qed.

  Lemma Forever_compat_impl:
    forall P Q (pq: forall a b, a =~= b -> P a -> Q b)
           x y (exy: x =~= y), Forever P x -> Forever Q y.
  Proof.
    intros P Q epq.
    cofix h.
    intros [u | u y] [t | t x] exy Fx.
    - left.
      inversion Fx.
      eapply epq.
      apply exy.
      apply X.
    - ssplit exy.
    - ssplit exy.
    - right.
      eapply epq.
      apply exy.
      inversion Fx; assumption.
      eapply h.
      ssplit exy; apply exy.
      inversion Fx; assumption.
  Qed.
  
  Lemma Always_forall: forall X (P: X -> Stream -> Prop) s,
      (forall x ,(Always (P x)) s) <-> Always (fun s_ => forall x, P  x s_) s.
  Proof.
    intros;split;revert s.
    cofix Always_forall.
    intros [ a | a s] F.
    left;intro x; specialize (F x);inversion F;subst;assumption.
    right.
    intro x;specialize (F x);inversion F;subst;assumption.
    apply Always_forall.
    intro x;specialize (F x);inversion F;subst;assumption.
    intros s F x;revert s F;cofix Always_forall.
    intros [ a | a s] F.
    constructor 1.
    inversion F;subst;apply H0.
    constructor 2.
    inversion F;subst;apply H1.
    inversion F;subst;apply Always_forall,H2.
  Qed.

  Lemma Always_Proper:
    forall s, Proper pequiv s -> Always (fun s => Proper pequiv s) s.
  Proof.
    cofix h.
    intros [a | a s]; intros ps.
    - left; apply ps.
    - right.
      apply ps.
      apply h.
      ssplit ps; apply ps.
  Qed.

  Lemma forall_WUntil:
    forall (P Q: Stream -> Prop),
      (forall s, P s -> Q s) ->
      forall s, Eventually P s -> Always (fun s => P s \/~ P s) s  ->
                WUntil (fun s => ~P s) Q s.
  Proof.
    intros P Q incl.
    cofix forall_WUntil.
    intros [a|a s] h h';inversion h';subst.
    inversion h;subst.
    constructor 1.
    apply incl;assumption.
    destruct H1 as [yes|no].
    constructor 1.
    apply incl,yes.
    constructor 2.
    apply no.
    apply forall_WUntil.
    inversion h;tauto.
    assumption.
  Qed.

  Lemma Forall_Always:
    forall (Q: Stream -> Prop), (forall s, Q s) -> forall s, Always Q s.
  Proof.
    intros Q hQ; cofix Forall_Always.
    intros [a | a  s].
    constructor 1; apply hQ.
    constructor 2.
    apply hQ.
    apply Forall_Always.
  Qed.

  (** if C is stable on a stream and P implies Q under hypothesis C
      and C is true on Stream s then Always P s implies Always Q s.
      *)
  Lemma Always_impl_2':
    forall (C: Stream -> Prop) (hC: forall (s: Stream), C s -> Always C s),
    forall (P Q: Stream -> Prop)
           (P_implies_Q: forall (s: Stream), C s -> P s -> Q s),
    forall (s: Stream), C s -> Always P s -> Always Q s.
  Proof.
    intros C hC P Q hPQ.
    cofix h.
    intros [a | a s] hs ha; inversion ha as [_1 io|_1 _2 hp haa]; subst.
    - left; apply hPQ; auto.
    - right.
      apply hPQ; auto.
      apply h; auto.
      specialize (hC _ hs).
      apply Always_P_P; inversion hC; auto.
  Qed.


  (* if C is stable on a stream and P implies Q under hypothesis C and
      C is true on Stream s then Eventually P s implies Eventually Q
      s.  *)
  Lemma Eventually_impl_2':
    forall (C: Stream -> Prop) (hC: forall (s: Stream), C s -> Always C s),
    forall (P Q: Stream -> Prop)
           (P_implies_Q: forall (s: Stream), C s -> P s -> Q s),
    forall (s: Stream), C s -> Eventually P s -> Eventually Q s.
  Proof.
    intros C hC P Q hPQ.
    induction 2.
    apply eventually_now.
    apply hPQ;auto.
    apply hC in H.
    inversion H.
    apply eventually_later.
    inversion H4; subst s s0;
      apply IHEventually;auto.
  Qed.

End LTL_StreamA.


Section Stream_Properties.
  
  Context {A: Type} `{SA: PartialSetoid A}.
  Notation Stream := (Stream (A := A)).

  Definition moves (s: Stream): Prop :=
    match s with
      s_one _ => True
    | s_cons a ss => ~(a =~= s_hd ss)
    end.

  Global Instance moves_compat: Proper fequiv moves.
  Proof.
    intros [x | x sx] [y|y sy] ee;ssplit ee;simpl;intuition.
    apply H1.
    rewrite H,H2.
    symmetry;apply hd_compat,H0.
    apply H1.
    rewrite <- H,H2.
    apply hd_compat,H0.
  Qed.
  
  Definition strict_moves (s: Stream): Prop :=
    match s with
      s_one _ => False
    | s_cons a ss => ~(a =~= s_hd ss)
    end.

  Global Instance strict_moves_compat: Proper fequiv moves.
  Proof.
    intros [x | x sx] [y|y sy] ee;ssplit ee;simpl;intuition.
    apply H1.
    rewrite H,H2.
    symmetry;apply hd_compat,H0.
    apply H1.
    rewrite <- H,H2.
    apply hd_compat,H0.
  Qed.

  (** * Finite Streams VS Infinite Streams *)

  (** Finite or Infinite stream? *)
  Definition finite: Stream -> Prop :=
    Eventually (fun s => match s with
                         | s_one _ => True | s_cons _ _ => False end).

  Global Instance finite_compat: Proper fequiv finite.
  Proof.
    apply Eventually_compat; auto.
    intros [a | a s] [a' | a' s'] es;split; try tauto; inversion es.
  Qed.

  Lemma finite_inv: forall (a: A) s, finite (s_cons a s) -> finite s.
  Proof.
    intros a s fin; inversion fin; subst; tauto.
  Qed.

  Definition infinite: Stream -> Prop :=
    Always (fun s => match s with | s_one _ => False | s_cons _ _ => True end).

  Global Instance infinite_compat: Proper fequiv infinite.
  Proof.
    apply Always_compat; auto.
    intros [a | a s] [a' | a' s'] es;split; try tauto; inversion es.
  Qed.

  Lemma infinite_not_finite:
    forall (s: Stream), infinite s <-> ~ finite s.
  Proof.
    intros s; split.
    { (* -> *)
      intros inf fin.
      induction fin as [ [a | a s] hs | a s hs IH ].
      inversion inf; tauto.
      tauto.
      apply IH.
      inversion inf; assumption.
    }
    { (* <- *)
      revert s.
      cofix infinite_not_finite.
      destruct s as [ a | a s ].
      intros fin; contradict fin; left; tauto.
      intros nfin.
      right.
      tauto.
      apply infinite_not_finite.
      intros fin; contradict nfin.
      right; assumption.
    }
  Qed.

  (** Removing a finite prefix from a stream *)
  Section When.

    Variable (P: Stream -> Prop).
    Hypothesis P_compat: Proper fequiv P.

    Fixpoint When e (W: Finally P e): Stream :=
      match W with
        finally_not_yet ee _ WW => When WW
      | finally_now _ _ _ => e
      end.


    (* Eval cbv in respectful_dep_param pequiv (Stream::nil) (pequiv,tt) (inr (Finally P)::nil). *)

    Local Notation depR := (fun F G : forall st : Stream, Finally P st -> Stream =>
                              forall x y : Stream, eqStream x y ->
                              forall (hx : Finally P x) (hy : Finally P y),
                                (F x hx) =~= (G y hy)).

    Global Instance When_compat:
      Proper depR When.
    Proof.
      intros e1 e2 ee W1 W2;revert e2 ee W2;
        induction W1 as [e1 hp1|g1 e1 hnp1 W1 hrec];
        intros e2 ee W2;dependent inversion W2;subst;simpl;auto.
      rewrite ee in hp1;tauto.
      rewrite <- ee in p;tauto.
      apply hrec.
      inversion ee;subst;auto.
    Qed.

    Lemma When_P: forall e (W: Finally P e), P (When W).
    Proof.
      intros e W;induction W;simpl;auto.
    Qed.

    Lemma Always_When:
      forall Q e (W: Finally P e), Always Q e -> Always Q (When W).
    Proof.
      intros Q e W hQ;induction W;simpl;auto.
      apply IHW.
      inversion hQ;auto.
    Qed.

    Lemma When_eq:
      forall e (fe: Finally P e),
        Proper pequiv e -> P e -> When fe = e.
    Proof.
      intros e [ee fe | fe] pe hp; simpl; [auto | tauto].
    Qed.

  Lemma When_fst:
    forall g e (pge: Proper pequiv (s_cons g e))
           (P_compat: Proper fequiv P)
           (fe: Finally P e) (fge: Finally P (s_cons g e))
           (nP: ~ P (s_cons g e)),
      When fge =~= When fe.
  Proof.
    intros.
    cut (When (finally_not_yet _ nP fe) = When fe).
    { intros tmp; rewrite <- tmp; clear tmp.
      apply When_compat; auto.
    }
    { simpl; auto. }
  Qed.

  End When.

  (** * Suffix of Stream *)

  (** ** Definition *)
  Definition is_suffix (s s_suf: Stream): Prop :=
    Eventually (fun s => s =~= s_suf) s.

  (** ** Compatibility *)
  Global Instance is_suffix_compat: Proper fequiv is_suffix.
  Proof.
    assert (ok: forall a a' b b', a =~= a' -> b =~= b' ->
                                  is_suffix a b -> is_suffix a' b').
    { intros a a' b b' ea eb suf.
      revert ea eb. revert a' b'.
      induction suf.
      - left.
        rewrite <- ea, H; assumption.
      - intros a' b' ea eb.
        inversion ea; subst.
        apply eventually_later.
        apply IHsuf.
        assumption.
        apply eb.
    }
    { intros a a' ea b b' eb.
      split; intros H.
      - eapply ok.
        apply ea.
        apply eb.
        assumption.
      - eapply ok.
        symmetry; apply ea.
        symmetry; apply eb.
        assumption.
    }
  Qed.

  Lemma is_suffix_inv_cons1:
    forall s suf b, is_suffix s (s_cons b suf) -> is_suffix s suf.
  Proof.
    intros s suf b H.
    induction H.
    - inversion H; subst.
      right; left; assumption.
    - right; assumption.
  Qed.

  Lemma is_suffix_inv_cons2:
    forall s suf a b, is_suffix (s_cons a s) (s_cons b suf) -> is_suffix s suf.
  Proof.
    intros s suf a b H.
    inversion H;subst.
    - left; inversion H0; assumption.
    - apply is_suffix_inv_cons1  in H1; assumption.
  Qed.

  Lemma is_suffix_id: forall s, Proper pequiv s -> is_suffix s s.
  Proof. left; auto. Qed.

  Lemma is_suffix_Proper:
    forall s suf, is_suffix s suf -> Proper pequiv suf.
  Proof.
    intros s suf hsuf.
    induction hsuf; auto.
    proper_.
  Qed.      

  Lemma suffix_of_suffix:
    forall e c cc, is_suffix e c -> is_suffix c cc -> is_suffix e cc.
  Proof.
    intros e c cc h; induction h; intros hc.
    { dup hc hh.
      apply is_suffix_Proper in hc; rewrite H; auto. }
    { right; auto.
      apply IHh; auto. }
  Qed.

  Lemma Always_suffix:
    forall P (P_compat: Proper pequiv P) s suf,
          Always P s -> is_suffix s suf -> Always P suf.
  Proof.
    intros P P_compat s suf al isf.
    induction isf.
    - revert al; apply Always_compat; symmetry; auto.
    - apply IHisf.
      apply Always_inv_2 with a; auto.
  Qed.

  Lemma When_suffix:
    forall P (P_compat: Proper fequiv P)
           (s: Stream) (ps: Proper pequiv s)  (W: Finally P s),
      is_suffix s (When W).
  Proof.
    intros.
    revert ps;induction W;intro ps;simpl.
    left;assumption.
    inversion ps;subst.
    right.
    apply IHW.
    assumption.
  Qed.
  
  Lemma Eventually_suffix:
    forall P s, Proper pequiv s -> Proper pequiv P ->
                (Eventually P s <->
                 exists suf, is_suffix s suf /\ Eventually P suf).
  Proof.
    intros P s ps pp; split.
    { intros h; induction h as [s hs | a s hs ih].
      - exists s; split; auto; left; auto.
      - destruct ih as [suf [hsuf ev]].
        inversion ps; auto.
        exists suf; split; auto.
        right; auto.
    }
    { intros [suf [hsuf esuf]].
      induction hsuf as [s e| a s es ih].
      - revert esuf; apply Eventually_compat; auto.
      - right; apply ih.
        inversion ps; auto.
    }
  Qed.

  Lemma Eventually_conj:
    forall P Q s, Proper pequiv s -> Eventually P s ->
                  (forall s', is_suffix s s' -> P s' -> Always P s') ->
                  (forall s', is_suffix s s' -> P s' -> Eventually Q s') ->
                  Eventually (fun s => P s /\ Q s) s.
  Proof.
    intros P Q s ps evP; induction evP as [s hs| a s hs ih]; intros ap eq.
    { specialize (ap s (is_suffix_id ps) hs).
      specialize (eq s (is_suffix_id ps) hs).
      induction eq as [s hq | a s hq ih].
      left; split; auto.
      right; apply ih.
      inversion ps; auto.
      apply Always_P_P; inversion ap; auto.
      inversion ap; auto. }
    { right.
      apply ih.
      inversion ps; auto.
      intros s' hhs pps.
      apply ap; auto; right; auto.
      intros s' hhs pps.
      apply eq; auto; right; auto. }
  Qed.

  (** * Parcimonious
   *)
  Section Parcimonious.

    Context {B: Type} `{SB: PartialSetoid B}.

    Fixpoint non_empty_parcimonious (s: Stream) (b: B) (l: list B): Prop :=
      match s with
      | s_one a =>  match l with
                    | nil => True
                    | b'::q => b =~= b' /\ non_empty_parcimonious (s_one a) b' q
                    end
      | s_cons a e' =>  match l with
                        | nil => True
                        | b'::q =>
                          (a =~= s_hd e' -> b =~= b') /\
                          non_empty_parcimonious e' b' q
                        end
      end.

    Global Instance non_empty_parcimonious_compat:
      Proper (pequiv ==> pequiv ==> eqlistA pequiv ==> iff)
             non_empty_parcimonious.
    Proof.
      intros sa1 sa2 esa b1 b2 eb l1 l2 el;
        revert  sa1 sa2 esa b1 b2 eb l2 el;
        induction l1 as [|b1_ l1 hrec];
        intros sa1 sa2 esa b1 b2 eb [|b2_ l2] el;
        inversion el;subst;
          destruct sa1 as [a1| a1 sa1]; destruct sa2 as [a2| a2 sa2];
            inversion esa;subst;
              simpl;reflexivity || apply Morphisms_Prop.and_iff_morphism.
      apply PER_morphism;auto;apply SB.
      apply hrec;auto.
      apply Morphisms_Prop.iff_iff_iff_impl_morphism.
      apply PER_morphism;auto; apply SA.
      apply hd_compat; symmetry; apply H6.
      apply PER_morphism;auto; apply SB.
      apply hrec;auto.
    Qed.

    Definition parcimonious (s: Stream) (l: list B): Prop :=
      match l with
      | nil => True
      | a::l => non_empty_parcimonious s a l
      end.

    Lemma non_empty_parcimonious_parcimonious: forall a s b l,
        non_empty_parcimonious (s_cons a s) b l ->
        parcimonious s l.
    Proof.
      intros a s b [|b_ l] h;simpl.
      auto.
      simpl in h;tauto.
    Qed.

    Global Instance parcimonious_compat:
      Proper (pequiv ==> eqlistA pequiv ==> iff) parcimonious.
    Proof.
      intros sa1 sa2 esa [|b1_ l1] [|b2_ l2] el;
        inversion el;subst;simpl;try tauto.
      apply non_empty_parcimonious_compat;auto.
    Qed.

  End Parcimonious.

  (** ** Stream constantly made of a single element *)
  Section is_constant.
    
    Definition is_constant (a: A): Stream -> Prop :=
      Always (fun s => s_hd s =~= a).
    
    Global Instance is_constant_compat: Proper fequiv is_constant.
    Proof.
      intros a b eab.
      unfold is_constant.
      eapply Always_compat.
      intros x y exy.
      assert (X := hd_compat exy).
      split; intros H.
      rewrite <- eab, <- exy; assumption.
      rewrite eab, exy; assumption.
    Qed.

    Lemma is_constant_inv:
      forall a x s, is_constant a (s_cons x s) -> is_constant a s.
    Proof.
      intros a x s C; inversion C; assumption.
    Qed.

    Lemma is_constant_hd:
      forall a s, is_constant a s -> s_hd s =~= a.
    Proof.
      intros a [x | x s] i; inversion i; subst; simpl in *; assumption.
    Qed.

    Lemma is_constant_Proper_l:
      forall a s, is_constant a s -> Proper pequiv a.
    Proof.
      intros a [x | x s] i; inversion i; subst.
      - simpl in H0; unfold Proper; rewrite <- H0 at 1; assumption.
      - simpl in H1; unfold Proper; rewrite <- H1 at 1; assumption.
    Qed.

    Lemma is_constant_Proper_r:
      forall a s, is_constant a s -> Proper pequiv s.
    Proof.
      intros a.
      cofix is_constant_Proper_r.
      intros [x | x s]; intros i; inversion i; subst.
      - simpl in H0; unfold Proper; left; rewrite H0 at 2; assumption.
      - unfold Proper; right.
        + simpl in H1; unfold Proper; rewrite H1 at 2; assumption.
        + apply is_constant_Proper_r.
          apply H2.
    Qed.

    Lemma constant_suffix:
      forall s suf, is_suffix s suf ->
                    forall x, is_constant x s -> is_constant x suf.
    Proof.
      intros s suf S x C.
      apply Always_suffix with s; auto.
      intros a b ab; split;
        intros h; rewrite <- h;
          apply hd_compat; [symmetry|]; auto.      
    Qed.

  End is_constant.

End Stream_Properties.

(** ** Map on stream *)
CoFixpoint s_map {A B: Type} (f: A -> B) (s: Stream) :=
  match s with
    | s_one a => @s_one B (f a)
    | s_cons a s => s_cons (f a) (s_map f s)
  end.

Global Instance s_map_compat {A: Type} `{SA: PartialSetoid A}
       {B: Type} `{SB: PartialSetoid B}:
  Proper (fequiv ==> pequiv ==> pequiv) (@s_map A B).
Proof.
  intros f1 f2 hf;intros s1 s2 hs;apply  Stream_decompose_inv.
  revert  s1 s2 hs;cofix s_map_compat;intros s1 s2 hs.
  destruct hs  as [ a1 a2 ha | a1 a2 s1 s2 ha hs];simpl;[left|right];auto.
  specialize (s_map_compat _ _ hs).
  (* PC : must rewrite in hypothesis to preserve guardedness *)
  do 2 rewrite <- Stream_decomposition_lemma in s_map_compat.
  apply s_map_compat.
Qed.

Unset Implicit Arguments.
Close Scope signature_scope.
