(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- September 2017
*)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Lia.

(** **** Local Imports *)
From Padec Require Import Stream.
From Padec Require Import SetoidUtil.

(* end show *)
Open Scope nat_scope.
Open Scope signature_scope.
Set Implicit Arguments.

(** * Stream Length
 Tools to measure the length of 
   - a (finite) stream
   - a prefix of a stream before it reaches some predicate
*)
Section Stream_Length.

  (** Stream of elements of type A *)
  Context {A: Type} `{SA: PartialSetoid A}.

  (** Number of steps in a stream until reaching some predicate P.

      (Relational definition)

      Note that the definition is made deterministic, namely, we require 
      that 
      - counting stops at P
      - before stopping, ~P should hold
  *)
  Section with_P.

    Variable (P: Stream (A:=A) -> Prop).
    Inductive length_until_P: Stream -> nat -> Prop :=
    | in_P: forall s, P s -> length_until_P s 0
    | add_step: forall a s n, ~ P (s_cons a s) ->
                              length_until_P s n ->
                              length_until_P (s_cons a s) (S n) 
    .

    Lemma length_until_P_inv_0: forall s, length_until_P s 0 -> P s.
    Proof. intros s l; inversion l; auto. Qed.

    Lemma length_until_P_inv_1:
      forall a s n, length_until_P (s_cons a s) (S n) -> ~ P (s_cons a s).
    Proof. intros a s n l; inversion l; auto. Qed.

    Lemma length_until_P_inv_2: forall a s n, length_until_P (s_cons a s) (S n) -> length_until_P s n.
    Proof. intros a s n l; inversion l; auto. Qed.

  End with_P.

  (** Length of a finite stream 
      
      Relational definition
  *)
  Definition has_length (s: Stream) (n: nat): Prop :=
    length_until_P
      (fun s => match s with s_one _ => True
                        | s_cons _ _ => False
                end) s n.

  (** Compatibility *)
  Global Instance length_until_P_compat: Proper fequiv length_until_P.
  Proof.
    cut (forall P Q (hPQ: P =~= Q)
                s1 s2 (hs: s1 =~= s2)
                n1 n2 (hn: n1 = n2),
            length_until_P P s1 n1 -> length_until_P Q s2 n2).
    { intros ok P Q hPQ s1 s2 hs n1 n2 hn;
        split; apply ok; auto; [intros a b eab|];symmetry;
          [apply hPQ; symmetry|]; auto. }
    { intros P Q hPQ s1 s2 hs n1 n2 hn h.
      revert hs hn; revert s2 n2;
        induction h as [s1 h1|a s1 n1 nP lg ih]; intros.
      - rewrite <- hn; apply in_P.
        eapply hPQ.
        apply hs.
        auto.
      - inversion hs; subst.
        apply add_step.
        intros F; contradict nP; eapply hPQ.
        apply hs.
        apply F.
        apply ih; auto.
    }
  Qed.

  Global Instance has_length_compat: Proper fequiv has_length.
  Proof.
    intros sym; apply length_until_P_compat; auto.
    intros [a1 |a1 s1] [a2|a2 s2] hs; 
      inversion hs; simpl; tauto.
  Qed.

  (** A stream has a length iff it is finite. *)
  Lemma has_length_finite:
    forall (s: Stream), (exists n, has_length s n) <-> finite s.
  Proof.
    intros s; split; [intros [n lg] | intros fin].
    { induction lg as [s fin| a s n _ _ ih].
      - destruct s as [a | a s]; [left|]; tauto.
      - right; apply ih.
    }
    { induction fin as [s fin| a s fin ih]. 
      - destruct s as [a | a s]; [exists 0; left|]; tauto.
      - destruct ih as [n lg].
        exists (S n); right; [tauto| apply lg].
    }
  Qed.
  
  (** ** Functional definitions 

      Actually counting the number of steps until reaching P.

      Requires that the stream actually reaches P.

      Either reachability of P is required with
      - [Finally P] (in Type)
      - or [Eventually P] (in Prop), in which case, P is required to be 
        weakly decidable
   *)
  Fixpoint count_steps_aux (P: Stream -> Prop) (s: Stream)
           (F: @Finally A P s): nat :=
    match F in Finally _ _ return  nat with
    | finally_now _ _ _ => 0%nat
    | finally_not_yet a s' F' => S (count_steps_aux F')
    end.

  Definition count_steps
             (P: Stream -> Prop) (s: Stream)
             (P_dec: Weak_Dec pequiv P)
             (ps: Proper pequiv s) (hs: Eventually P s): nat :=
    count_steps_aux (Eventually_Dec_Finally' P_dec ps hs).

  (** Compatibility *)
  Global Instance count_steps_aux_compat:
    Proper (respectful_dep_param
              (equiv (A:=nat))
              ((Stream -> Prop)::Stream::nil)
              (fequiv,(pequiv,tt))
              (inr Finally::nil))
           (@count_steps_aux).
    Proof.
      intros P1 P2 hP s1 s2 hs F1 F2.
      revert s2 hs F2.
      induction F1 as [s1 hP1| a s1 nP F1 ih];
        destruct F2 as [| a2 s2 nP2 F2];
        simpl in *;specialize (hP  _ _ hs).
      reflexivity.
      tauto.
      tauto.
      apply f_equal.
      apply ih.
      inversion hs;auto.
    Qed.

    Global Instance count_steps_compat:
      Proper (respectful_dep_param
              (equiv (A:=nat))
              ((Stream -> Prop)::Stream::nil)
              (fequiv,(pequiv,tt))
              (inr (fun P s => Weak_Dec pequiv P)
                   ::inl (fun P s => Proper pequiv s)
                   ::inl Eventually :: nil))
             (@count_steps).
    Proof.
      intros P1 P2 hP s1 s2 hs P1_dec ps1 F1 P2_dec ps2 F2.
      apply count_steps_aux_compat; auto.
    Qed.

    (** Relational and functional counting provides the same result! 
     NB: requires P to be weakly decicable so that counting can decide
     when to stop.
    *)
    Lemma steps_until_P_def_equiv:
      forall (P: Stream -> Prop) (P_dec: Weak_Dec pequiv P)
             (s: Stream) (ps: Proper pequiv s) (hs: Eventually P s)
             (n: nat),
        count_steps P_dec ps hs = n <-> length_until_P P s n.
    Proof.
      intros P P_dec s ps hs n.
      split; intros h.
      { unfold count_steps in h.
        revert n h; induction (Eventually_Dec_Finally' P_dec ps hs).
        intros n h; simpl in h.
        rewrite <- h; apply in_P; assumption.
        intros [|N] h.
        simpl in h; inversion h.
        apply add_step.
        apply n.
        apply IHf.
        ssplit ps; apply ps. 
        apply Finally_Eventually; assumption.
        inversion h; auto.
      }
      { unfold count_steps.
        revert h. revert n.
        induction (Eventually_Dec_Finally' P_dec ps hs); simpl.
        intros n h; inversion h; subst; tauto.
        ssplit ps; destruct ps as [_ ps].
        specialize (IHf ps (Finally_Eventually f)).
        intros nn h; inversion h; subst; try tauto.
        apply f_equal.
        apply IHf; assumption.
      }
    Qed.
    
    (** When predicate P is reachable by some stream s, s has finite 
      length until P.

      NB: P is required to be weakly classical to be able to stop.
    *) 
    Lemma Reachable_Steps:
      forall (P: Stream -> Prop),
        (forall s, (Proper pequiv s -> P s) \/ ~ P s) ->
        forall s, Proper pequiv s -> Eventually P s ->
                  exists n, length_until_P P s n.
    Proof.
      intros P weakly_classical s ps ev.
      induction ev as [s now| a s ev later].
      - exists 0; left; auto.
      - destruct (weakly_classical (s_cons a s)) as [okP|nP].
        * exists 0; left; apply okP, ps.
        * ssplit ps; destruct ps as [_ ps].
          specialize (later ps).
          destruct later as [n lg].
          exists (S n); right; auto.
    Qed.

    (** Conversely, when a stream has some length until P, it 
      eventually reaches P *) 
    Lemma Reachable_steps_converse:
      forall (P: Stream -> Prop) (s: Stream) (n: nat),
        length_until_P P s n -> Eventually P s.
    Proof.
      intros P s n lg.
      induction lg.
      left; assumption.
      right; assumption.
    Qed.

    (** Furthermore, if P is weakly decidable, the stream finally 
      reaches P *)
    Lemma Reachable_steps_converse':
      forall (P: Stream -> Prop),
        Weak_Dec pequiv P ->
        forall (s: Stream), Proper pequiv s ->
                            forall n, length_until_P P s n -> Finally P s.
    Proof.
      intros P wd s ps n lg.
      eapply Eventually_Dec_Finally'.
      apply wd.
      apply ps.
      eapply Reachable_steps_converse.
      apply lg.
    Qed.

    (** Inclusion of predicates: the stream reaches the bigger one first *)
    Lemma length_until_P_le:
      forall (P: Stream -> Prop) (s: Stream) (n: nat),
        length_until_P P s n ->
        forall (Q: Stream -> Prop) (incl: forall s, P s -> Q s) (n': nat),
          length_until_P Q s n' -> n' <= n.
    Proof.
      intros P s n hSU Q incl.
      induction hSU as [a s | a s n nPas hSU ih]; intros n' hSU'.
      + destruct hSU' as [| a' s' n' nQas hSU']; [reflexivity|].
        contradict nQas; apply incl; auto.
      + destruct n' as [| n'].
        lia.
        inversion hSU' as [|_2 _3 _4 nQge hSU'' _6]; subst.
        apply le_n_S, ih,hSU''.
    Qed.
  
    Lemma length_until_P_le_ex':
      forall (P: Stream -> Prop) (s: Stream) (n: nat),
        length_until_P P s n ->
        forall (Q: Stream -> Prop)
               (Q_weak_clas: forall s, (Proper pequiv s -> Q s) \/ ~ Q s)
               (incl: forall s, P s -> Q s),
          Proper pequiv s -> exists n', length_until_P Q s n'.
    Proof.
      intros P s n hSU Q Q_weak_clas incl ps.
      induction hSU as [s Ps | a s n nPas hSU ih]. 
      * exists O; left; apply incl; auto.
      * destruct (Q_weak_clas (s_cons a s)) as [q | q].
      + exists O; left; apply q; auto.
      + destruct ih as [n' hSU']; auto.
        ssplit ps; apply ps.
        exists (S n').
        right; auto.
    Qed.
  
    Lemma length_until_P_le_ex:
      forall (P: Stream -> Prop) (s: Stream) (n: nat),
        length_until_P P s n ->
        forall (Q: Stream -> Prop)
               (Q_weak_clas: forall s, (Proper pequiv s -> Q s) \/ ~ Q s)
               (incl: forall s, P s -> Q s),
          Proper pequiv s -> exists n', n' <= n /\ length_until_P Q s n'.
    Proof.
      intros P s n hSU Q Q_weak_clas incl ps.
      dup hSU hSU'.
      apply length_until_P_le_ex' with (Q := Q) in hSU; auto.
      destruct hSU as [n' sn'].
      exists n'.
      split; auto.
      apply length_until_P_le with (P := P) (Q := Q) (s := s); auto.
    Qed.

    (** If a finite stream reaches P, it reaches it in less than its 
        length*)
    Lemma longer_length:
      forall (P: Stream -> Prop) (s: Stream) (n n': nat),
        length_until_P P s n -> has_length s n' -> n' >= n. 
    Proof.
      intros P s n n' lgP lg.
      revert s lgP lg. revert n'.
      induction n as [| n ih].
      { intros; lia. }
      { intros n' s lgP lg.
        destruct s as [a | a s];
          inversion_clear lgP as [|_1 _2 _3 nP llgP]; subst.
        destruct n' as [| n']; inversion lg as [|_1 _2 _3 _ lg'];
          subst; try tauto.
        specialize (ih _ _ llgP lg'); lia.
      }
    Qed.

    (** * Decreasing potential:
     This section provides a tool to prove that a relational stream 
     reaches a given predicate within at least n steps.
     *)
    Section Decreasing_Potential.

      (** Relation that links elements in a relational stream until dum.
 
        NB: there is no assumption on dum, this could be (fun _ => True)
       *)
      Variable (R: relation A) (dum: A -> Prop).
      
      (** Predicate to be reached by the relational stream *)
      Variable (P: A -> Prop)
               (P_weak_classical: forall a,
                   (Proper pequiv a -> P a) \/ ~ P a).

      (** Potential that measures an element of the relational 
          stream *)
      Variable (Pot: nat -> A -> Prop).

      (** Hypothesis that inductively makes the potential decrease
          step by step in the relational stream *)
      Hypothesis (base: forall a, Pot 0 a -> P a).
      Hypothesis (step: forall n a a', Pot (S n) a -> R a' a -> Pot n a').

      (** For a given relational stream, if its head has Potential n, 
          then either it is too short (length < n) or it reaches P 
          in less than n steps *)
      Lemma pot_invariant_rel_stream:
        forall (n: nat) (s: Stream),
          Proper pequiv s ->
          Always (fun s =>
                    match s with
                      s_one a => dum a
                    | s_cons a s => R (s_hd s) a end) s ->
          Pot n (s_hd s) ->
          (exists n', n' < n /\ has_length s n')
          \/
          exists n', n' <= n /\
                     length_until_P (fun s => P (s_hd s)) s n'.
      Proof.
        intros n.
        induction n as [| n ih]. 
        { intros s ps rs pot.
          right; exists 0.
          split; [lia|].
          left.
          apply base, pot.
        }
        { intros s ps rs pot.
          case (P_weak_classical (s_hd s)); intros Pv.
          * right; exists 0.
            split; [lia|].
            left; apply Pv, hd_compat, ps.
          * destruct s as [a | a s].
          - left; exists 0; split; [lia|left]; auto.
          - assert (hyp_n: Pot n (s_hd s)).
            { eapply step.
              apply pot.
              inversion rs; auto. 
            }
            inversion rs; subst.
            ssplit ps; destruct ps as [_ ps].
            specialize (ih _ ps H2 hyp_n).
            destruct ih as [ [n' [hn' lg]] | [n' [hn' st]] ].
            { left; exists (S n'); split; [lia|].
              right; [tauto|auto]. }
            { right; exists (S n'); split; [lia|].
              right; auto. }
        }
      Qed.
  
      Lemma pot_invariant_rel_stream':
        forall n s,
          Proper eqStream s ->
                    Always (fun s =>
                    match s with
                      s_one a => dum a
                    | s_cons a s => R (s_hd s) a end) s ->
          Eventually (fun s => P (s_hd s)) s ->
          Pot n (s_hd s) ->
          exists n', n' <= n /\
                     length_until_P (fun s => P (s_hd s)) s n'.
      Proof.
        intros n s ps rs ev pot.
        apply Reachable_Steps in ev; auto.
        2: intros ee;
          destruct (P_weak_classical (s_hd ee)) as [okP|nP];
          [left|right; auto].
        2: intros pee; apply okP; unfold Proper; apply hd_compat, pee.
        destruct ev as [n' st].
        destruct (pot_invariant_rel_stream ps rs pot) as [ok|ok];
          auto.
        destruct ok as [nn' [hnn' lg]].
        assert (x := longer_length st lg).
        exists n'; split; [lia|auto].
      Qed.
    
      Variable (N: nat).
      Hypothesis (bound: forall a, Pot N a).

      Lemma length_pot_bound:
        forall (s: Stream),
          Proper pequiv s ->
          Always (fun s =>
                    match s with
                      s_one a => dum a
                    | s_cons a s => R (s_hd s) a end) s ->
          Eventually (fun s => P (s_hd s)) s ->
          exists n, n <= N /\
                    length_until_P (fun s => P (s_hd s)) s n.
      Proof.
        intros s ps re ev.
        apply pot_invariant_rel_stream'; auto.
      Qed.

    End Decreasing_Potential.

End Stream_Length.

(* * Annotations:
  Annotation on a stream is a stream where each element has been 
  decorated with some additional information called Annotation.

  Every result on Stream is also obviously available for annotated 
  stream.

  This section provides definitions and also to pass from stream
  to annotated stream and reverse.
Section Annotations.

  (** streams on elements of type A: they will be annotated be elements 
      of type Annotation; 
      this produces streams of type AnnotatedStream *)
  Context {A: Type} `{SA: PartialSetoid A}.
  Context {Annotation: Type} `{SAnn: PartialSetoid Annotation}.

  Notation Annotated := (A * Annotation)%type.
  Notation getA := (fst (A:=A) (B:=Annotation)).
  Notation getAnn := (snd (A:=A) (B:=Annotation)).

  Notation eqA := (pequiv (A:=A)).
  Notation eqAnn := (pequiv (A:=Annotation)).
  Notation eqAAnn := (eqA * eqAnn).

  Notation StreamA := (Stream (A:=A)).
  Notation AnnotatedStream := (Stream (A := Annotated)).

  CoFixpoint getStream (ans: AnnotatedStream): StreamA :=
    match ans with
    | s_one aa => s_one (getA aa)
    | s_cons aa ans => s_cons (getA aa) (getStream ans)
    end.

  Lemma getStream_unfold:
    forall ans, getStream ans =
               match ans with
               | s_one aa => s_one (getA aa)
               | s_cons aa ans => s_cons (getA aa) (getStream ans)
               end.
  Proof.
    intros ans.
    rewrite Stream_decomposition_lemma with (s := getStream ans).
    unfold Stream_decompose; simpl.
    destruct ans as [aa | aa ans]; reflexivity.
  Qed.

  Lemma getStream_hd:
    forall ans, (s_hd (getStream ans)) = (getA (s_hd ans)).
  Proof.
    intros [aa | aa ans]; simpl; reflexivity.
  Qed.

  Global Instance getStream_compat: Proper fequiv getStream.
  Proof.
(*    intros ans ans' eans; simpl in eans; hnf; revert eans. revert ans ans'.*)
    intros s s'  es.
    rewrite Stream_decomposition_lemma.
    rewrite Stream_decomposition_lemma at 1.
    revert s s' es.
    cofix h.
    intros s s' [ aa aa' ea  | aa aa' ans ans' ea es ];[left|right];try eapply fst_compat,ea.
    specialize (h _ _ es).
    do 2 rewrite <- Stream_decomposition_lemma in h.
    apply h.
  Qed.

  Global Instance getA_compat: Proper fequiv getA.
  Proof. apply fst_compat. Qed.

  Global Instance getAnn_compat: Proper fequiv getAnn.
  Proof. apply snd_compat. Qed.

  (** relation between two elements of an annotated stream *)
  Variable (AnnotatedStep: relation Annotated).

  (** We consider that a stream is an annotated stream of a stream when any two consecutive 
      elements are linked using AnnotatedStep and the projection allows to recover the initial 
      stream by projection *)
  Definition is_Annotation (s: StreamA) (ans: AnnotatedStream): Prop :=
    s =~= getStream ans /\
    Relational_Stream AnnotatedStep (fun _ => True) ans.

  Lemma is_Annotation_inv:
    forall a s aa ans, is_Annotation (s_cons a s) (s_cons aa ans)
                      -> is_Annotation s ans.
  Proof.
    intros a s aa ans [prof rel].
    rewrite getStream_unfold in prof.
    ssplit prof.
    apply Relational_Stream_inv_2 in rel.
    split; [apply prof|apply rel].
  Qed.

  Lemma is_Annotation_inv':
    forall a s aa ans, is_Annotation (s_cons a s) (s_cons aa ans)
                      -> AnnotatedStep (s_hd ans) aa.
  Proof.
    intros a s aa ans [_ rel].
    eapply Relational_Stream_inv_1, rel.
  Qed.

  (** Annotated streams have the same length as their initial stream *)
  Lemma Ann_preserve_length:
    forall s ans, is_Annotation s ans -> forall P (P_compat: Proper fequiv P) n,
        length_until_P (fun ans => P (getStream ans)) ans n
        <->
        length_until_P P s n.
  Proof.
    intros s ans han P P_compat n.
    destruct han as [han _].
    revert s ans han.
    induction n as [| n ih]; intros s ans han; split; intros L.
    { left.
      apply length_until_P_inv_0 in L; auto.
      eapply P_compat.
      2: apply L.
      apply han.
    }{ left.
       apply length_until_P_inv_0 in L; auto.
       eapply P_compat.
       2: apply L.
       symmetry; apply han.
     }{ destruct s as [a | a s]; destruct ans as [aa | aa ans]; rewrite getStream_unfold in han;
          inversion han; inversion L; subst.
        right.
        + intros F; contradict H8; eapply P_compat.
          2: apply F.
          rewrite getStream_unfold; symmetry; auto.
        + rewrite <- (ih _ _ H4); auto.
      }{ destruct s as [a | a s]; destruct ans as [aa | aa ans]; rewrite getStream_unfold in han;
          inversion han; inversion L; subst.
         right.
         + intros F; contradict H8; eapply P_compat.
           2: apply F.
          rewrite getStream_unfold; auto.          
         + rewrite (ih _ _ H4); auto.
       }
  Qed.

  (** Similarly, annotation preserves the fact that a stream reaches some propery. *)
  Lemma Ann_preserve_eventually:
    forall s ans, is_Annotation s ans -> forall P (P_compat: Proper fequiv P),
        Eventually (fun ans => P (getStream ans)) ans
        <->
        Eventually P s.
  Proof.
    intros s ans han P P_compat.
    destruct han as [han _].
    split; intros ev.
    { revert s han; induction ev as [ans now | aa ans later ih]. 
      + intros e es.
        left.
        eapply P_compat, now; auto.
      + intros s es.
        rewrite getStream_unfold in es.
        destruct s as [ a | a s]; inversion es; subst.
        right.
        apply ih; auto.
    }
    { revert ans han; induction ev as [s now | a s later ih].
      + intros ans es.
        left.
        eapply P_compat, now; symmetry; auto.
      + intros ans es.
        destruct ans as [ aa | aa ans]; rewrite getStream_unfold in es;
          inversion es; subst.
        right.
        apply ih; auto.
    }
  Qed.

  (** We consider streams that eventually reaches P and measures their length until P. *)
  Variable (P: A -> Prop)
           (P_compat: Proper fequiv P)
           (P_weak_classical: forall a, (Proper eqA a -> P a) \/ ~ P a).

  (** Potential used to measure annotated elements *)
  Variable (Pot: nat -> Annotated -> Prop).

  (** We require that 
      - at potential 0, P is actually reached
      - crossing an annotated step increases the potential by one *)
  Hypothesis (base: forall aa, Pot 0 aa -> P (getA aa)).
  Hypothesis (step: forall n aa aa', Pot (S n) aa ->
                                     AnnotatedStep aa' aa ->
                                     Pot n aa').

  (** With base and step hypothesis, we can prove that 
   if a stream eventually reaches P and if it has a corresponding annotation stream, then
        if the head of the annotated stream has potential n, the length of the initial stream
        until P is less than P.
  *)
  Lemma annotation_pot_reach:
    forall (s: StreamA),
      Eventually (fun s => P (s_hd s)) s ->
      forall (ans: AnnotatedStream), Proper pequiv ans -> is_Annotation s ans ->
                                     forall n, Pot n (s_hd ans) ->
                                               exists n', n' <= n /\
                                                          length_until_P (fun s => P (s_hd s)) s n'
  .
  Proof.
    intros s ev ans pans hans n pot.
    assert (wk: forall aa, (Proper pequiv aa -> P (getA aa)) \/ ~ P (getA aa)).
    { intros aa; destruct (P_weak_classical (getA aa)) as [pa | npa].
      - left.
        intros paa; apply pa.
        eapply fst_compat, paa.
      - right; auto.       
    }
    assert (rans: Relational_Stream AnnotatedStep (fun _ : Annotated => True) ans).
    { apply hans. }
    assert (evans: Eventually (fun s : AnnotatedStream => (fun aa : Annotated => P (getA aa)) (s_hd s)) ans).
    { 
      rewrite <- Ann_preserve_eventually in ev; auto.
      3: intros as1 as2 eas; apply P_compat, hd_compat, eas.
      2: apply hans.
      
      clear -ev.
      induction ev.
      left; rewrite <- getStream_hd; auto.
      right; auto.
    }
    assert (ok := @pot_invariant_rel_stream' _ _ AnnotatedStep (fun _ => True) _ wk
                                             _ base step n _ pans rans evans pot).
    destruct ok as [n' [hn' lg]].
    exists n'; split; auto.
    eapply Ann_preserve_length with ans; auto.
    + intros s1 s2 es; apply P_compat, hd_compat,es.
    + eapply length_until_P_compat.
      4: apply lg.
      3: reflexivity.
      2: apply pans.
      intros as1 as2 eas; rewrite getStream_hd.
      apply P_compat, getA_compat, hd_compat, eas.
  Qed.    

  (** The goal now is to show that a given set of streams, Q, has a length until P bounded by a 
      given value, N. *)
  Variable (Q: StreamA -> Prop)
           (N: nat).

  (** We assume a way to build an annotated stream from a stream
      such that, for any stream in Q
        - it provides a compatible result and
        - it is actually an annotation 
   *)
  Hypothesis (compute_Annotation: StreamA -> AnnotatedStream)
             (compute_Annotation_ok: forall s, Q s -> is_Annotation s (compute_Annotation s))
             (compute_Annotation_compat: forall s, Q s -> Proper pequiv (compute_Annotation s)).

  (** We also assume that every annotated element has potential N *)
  Hypothesis (bound: forall aa, Pot N aa).

  (** If the hypothesis bound is met, then every stream in Q has length smaller than N until P *)
  Lemma annotation_pot_bound:
    forall (s: StreamA),
      Q s -> Eventually (fun s => P (s_hd s)) s ->
      exists n, n <= N /\
                length_until_P (fun s => P (s_hd s)) s n.
  Proof.
    intros s qs ev.
    eapply annotation_pot_reach; auto.
    apply compute_Annotation_compat, qs.
  Qed.
  
End Annotations.
*)

Close Scope nat_scope.
Close Scope signature_scope.
Unset Implicit Arguments.
