(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- September 2017
 *)
(* begin show *)

From Coq Require Import Wellfounded.
From Coq Require Import List.
From Coq Require Import SetoidList.

(* end show *)

Definition relation A := A -> A -> Prop.

Section Relations.
  Variable A:Type.
  Variable eqA: relation A.
  Variable Adec:forall x y:A,{eqA x y}+{~eqA x y}.

  Hypothesis eqA_equiv: Equivalence eqA.
  
  Definition transitive (R:relation A) :=
    forall x y z:A, R x y -> R y z -> R x z.

  Definition included (R1 R2:relation A) :=
    forall x y:A, R1 x y -> R2 x y.

  Inductive transitive_closure R :  A -> A -> Prop :=
    TC_incl : included R (transitive_closure R)
  | TC_trans : transitive (transitive_closure R).

  Inductive transitive_closure_alt R (a b:A) :Prop :=
  |TC_step: R a b->transitive_closure_alt R a b
  |TC_ind : forall y, transitive_closure_alt R a y ->
                      transitive_closure_alt R y b ->
                      transitive_closure_alt R a b.

  Lemma transitive_closure_transitive_closure_alt_if:
    forall R, included (transitive_closure R) (transitive_closure_alt R).
  Proof.
    intros R a b H;induction H.
    constructor 1;auto.
    constructor 2 with (y:=y);auto.
  Qed.


  Lemma transitive_closure_alt_transitive_closure_if:
    forall R:A->A->Prop, included (transitive_closure_alt R)
                                  (transitive_closure R) .
  Proof.
    intros R a b H;induction H.
    constructor 1;auto.
    constructor 2 with (y:=y);auto.
  Qed.

  Lemma transitive_closure_alt_transitive_closure_iff:
    forall (R:A->A->Prop) x y,
      (transitive_closure_alt R) x y <-> (transitive_closure R) x y.
  Proof.
    intros R x y;split;
      [apply transitive_closure_alt_transitive_closure_if
      |apply transitive_closure_transitive_closure_alt_if].
  Qed.
  
  Lemma transitive_closure_is_smallest :
    forall R R0:relation A, included R R0 -> transitive R0 ->
                            included (transitive_closure R) R0.
  Proof.
    intros R R0 R_incl_R0 R0_trans.
    intros x y hxy.
    induction hxy.

    apply R_incl_R0.
    assumption.

    apply R0_trans with y.
    assumption.
    assumption.
  Qed.

  Lemma transitive_closure_monotonic :
    forall R R0:relation A, included R R0 ->
                            included (transitive_closure R)
                                     (transitive_closure R0).
  Proof.
    intros R R0 R_incl_R0 x y hxy.
    induction hxy.
    left;apply R_incl_R0;apply H.
    right with y;auto.
  Qed.

  Lemma transitive_closure_split_right :
    forall R (x y:A), transitive_closure R x y ->
                      R x y \/
                      exists z, (transitive_closure R x z /\ R z y).
    intros R x y hxy;induction hxy as [x y hxy|x z y tcxz hrecxz tczy hreczy].
    { left;assumption. }
    { destruct hreczy as [ hzy | [zz [tczzz hzzy]]].
      { right;exists z;split;assumption. }
      { right;exists zz;split.
        { right with z;assumption. }
        { assumption. } } }
  Qed.

  Section R.
    Variable R:relation A.

    Hypothesis R_compat : Proper (eqA==>eqA==>iff) R.

    
    Scheme Acc_ind_dep := Induction for Acc Sort Prop.
    (** Any well-founded relation cannot be reflexive! *)
    Lemma acc_rel_is_non_reflexive:
      forall (x: A), Acc R x -> ~R x x.
      (********)
      intros x W.
      induction W using Acc_ind_dep.
      unfold not; intro false.
      specialize (H x false).
      auto.
    Qed.

    CoInductive path (x:A): Type :=
      path_cont : forall y, R y x -> path y -> path x
    | path_end : path x.

    Inductive finite_path : forall x, path x -> Prop   :=
      finite_cont: forall x y (h:R y x) (p:path y),
        finite_path y p -> finite_path x (path_cont x y h p)
    | finite_end : forall x, finite_path x (path_end x).

    Theorem wf_finite_path : well_founded R -> forall x p, finite_path x p.
      intros wf x p; specialize (wf x).
      induction wf.
      destruct p as [y hxy p_|].
      apply finite_cont.
      apply H0.
      apply hxy.
      apply finite_end.
    Qed.
  End R.

  Section finite.
    Variable lA:list A.
    Hypothesis A_finite: forall a:A, InA eqA a lA.

    Section R.
      Variable R:relation A.

      Hypothesis R_compat : Proper (eqA==>eqA==>iff) R.

      Inductive l_transitive_closure l : relation A :=
        LTC_incl: included R (l_transitive_closure l)
      | LTC_l_trans : forall x y z, InA eqA y l ->
                                    l_transitive_closure l x y ->
                                    l_transitive_closure l y z ->
                                    l_transitive_closure l x z.

      Lemma l_transitive_closure_compat:
        forall l, Proper (eqA==>eqA==>impl) (l_transitive_closure l).
      Proof.
        intros l a b eq0 c d eq1 H.
        revert b d eq0 eq1.
        induction H.
        intros b d eq0 eq1;  rewrite (R_compat _ _ eq0 _ _ eq1) in H.
        constructor;auto.
        intros b d eq0 eq1.
        constructor 2 with (y:=y).
        assumption.
        apply (IHl_transitive_closure1 b y);try reflexivity;try assumption.
        apply (IHl_transitive_closure2 y d);try reflexivity;try assumption.
      Qed.
      
      
      

      Lemma TC_LTC_lA_incl:
        included (transitive_closure R) (l_transitive_closure lA).
        intros x y hxy; induction hxy.
        left;auto.
        right with y;auto.
      Qed.

      Lemma LTC_TC_incl:
        forall l, included (l_transitive_closure l) (transitive_closure R).
        intros l x y hxy; induction hxy.
        left;auto.
        right with y;auto.
      Qed.

      Lemma LTC_nil_R_incl:included (l_transitive_closure nil) R.
        intros x y hxy; induction hxy.
        auto.
        simpl in H.
        inversion H.
      Qed.

      Lemma R_LTC_nil_incl:included R (l_transitive_closure nil).
        intros x y hxy.
        left.
        assumption.
      Qed.

      Lemma LTC_cons_incl:
        forall a l, included
                      (l_transitive_closure l)
                      (l_transitive_closure (a::l)).
        intros a l x y hxy; induction hxy;[left|right with y].
        apply H.
        simpl;auto.
        assumption.
        assumption.
      Qed.

      Variable Rdec:forall x y:A,{R x y}+{~R x y}.

      Lemma trans_dec: forall x y,
          {transitive_closure R x y}+{~transitive_closure R x y}.
        cut (forall x y,
                {l_transitive_closure lA x y}+{~l_transitive_closure lA x y}).
        { intro h;intros x y;destruct (h x y) as [YES|NO];[left|right].
          { apply (LTC_TC_incl lA).
            assumption. }
          { intro hh;apply NO.
            apply TC_LTC_lA_incl.
            assumption. } }
        { generalize lA;intro l.
          induction l.
          { intros x y.
            destruct (Rdec x y) as [YES|NO].
            { left;apply R_LTC_nil_incl.
              assumption. }
            { right;intro hh;apply NO.
              apply LTC_nil_R_incl.
              assumption. } }
          { intros x y;destruct (IHl x y) as [YES|NO].
            { left.
              apply LTC_cons_incl.
              assumption. }
            { destruct (IHl x a) as [YESa|NOa].
              { destruct (IHl a y) as [aYES|aNO].
                { left;right with a.
                  {rewrite InA_cons;left;reflexivity. }
                  { apply LTC_cons_incl. assumption. }
                  { apply LTC_cons_incl. assumption. } }
                { right;intro hh;apply NO.
                  revert aNO.
                  clear NO YESa.
                  induction hh.
                  { left;assumption. }
                  { rewrite InA_cons in H; destruct H.
                    { 
                      intro h.
                      
                      specialize (IHhh2 h).
                      rewrite (l_transitive_closure_compat
                                 l _ _ H _ _ (Equivalence_Reflexive z))
                        in IHhh2.
                      specialize (h IHhh2).
                      contradiction. }
                    { intro hh;right with y.
                      { assumption. }
                      { apply IHhh1.
                        intro h;apply hh.
                        { right with y;try assumption.
                          apply IHhh2.
                          assumption. } }
                      { apply IHhh2.
                        assumption. } } } } }
              { right;intro hh;apply NO.
                revert NOa.
                clear NO.
                induction hh.
                { left;assumption. }
                { rewrite InA_cons in H;destruct H.
                  { symmetry in H;subst.
                    intro h.
                    specialize (IHhh1 h).
                    symmetry in H.
                    rewrite (l_transitive_closure_compat
                               l _ _ (Equivalence_Reflexive x) _ _ H) in IHhh1.
                    
                    specialize (h IHhh1).
                    contradiction. }
                  { intro hh;right with y;try assumption.
                    { apply IHhh1.
                      assumption.
                    }
                    { apply IHhh2.
                      intro h;apply hh.
                      { right with y;try assumption.
                        apply IHhh1.
                        assumption. } } } } } } } }
      Qed.
    End R.

  End finite.
End Relations.