(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** **** Global Imports *)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Wellfounded.
From Coq Require Import RelationPairs.
From Coq Require Import Arith.
From Coq Require Import Morphisms_Prop.
From Coq Require Import Bool.
From Coq Require Import Lia.

(** **** Local Imports *)

From CoLoR Require RelUtil.

From Padec Require Import SetoidUtil.
From Padec Require Import ListUtils.
From Padec Require Import ZUtils.
From Padec Require Import RelationA.

(* end show *)
Open Scope signature_scope.
Set Implicit Arguments.

(** * Tools for proving well-foundedness on transformed relations *)

Global Instance Acc_per_morphism {A: Type} `{SA: PartialSetoid A}
         (R: A -> A -> Prop) `(pR: Proper _ fequiv R)
         `(closure: forall x y, R y x -> Proper pequiv x -> Proper pequiv y):
  Proper fequiv (Acc R).
Proof.
  cut (Proper (pequiv ==> impl) (Acc R)).
  - intros h x1 x2 ex;split.
    * apply h;assumption.
    * apply h;symmetry;assumption.
  - intros x1 x2 ex wf.
    revert x2 ex;induction wf as [x1 _ Hwf] using Acc_ind_dep.
    intros x2 ex.
    constructor.
    intros y2 h2.
    assert (h: Proper pequiv y2).
    * apply closure with x2;auto.
      hnf;transitivity x1;[symmetry|];assumption.
    * apply Hwf with y2;auto.
      apply (proj2 (pR _ _ h _ _ ex) h2).
Qed.

Lemma Acc_per_impl {A: Type} `{SA: PartialSetoid A}
      (R1 R2: A -> A -> Prop) `(pR: R1 =~= R2)
      `(closure: forall x y, R1 y x -> Proper pequiv x -> Proper pequiv y):
  (pequiv ==> impl) (Acc R2) (Acc R1).
Proof.
  intros x1 x2 ex wf.
  revert x2 ex;induction wf as [x1 _ Hwf] using Acc_ind_dep.
  intros x2 ex.
  constructor.
  intros y2 h2.
  assert (h: Proper pequiv y2).
  * eapply closure.
    apply h2.
    unfold Proper; rewrite <- ex at 1; assumption.
  * eapply Hwf .
    2: apply h.
    eapply pR.
    apply h.
    symmetry; apply ex.
    apply h2.
Qed.

Lemma Acc_per_morphism' {A: Type} `{SA: PartialSetoid A}
      (R1 R2: A -> A -> Prop) `(pR: R1 =~= R2)
      `(closure_R1: forall x y, R1 y x -> Proper pequiv x -> Proper pequiv y)
      `(closure_R2: forall x y, R2 y x -> Proper pequiv x -> Proper pequiv y):
  (pequiv ==> iff) (Acc R1) (Acc R2).
Proof.
  intros a1 a2 ea.
  split.
  - eapply Acc_per_impl; auto.
    symmetry; auto.
  - eapply Acc_per_impl; auto.
    symmetry; auto.
Qed.

Section Well_Founded.

  (** Two equivalent relations are equivalently well founded. *)
  Lemma wf_rel_equiv:
    forall (A: Type) (R1: relation A) (R2: relation A),
      (forall a b, R1 a b <-> R2 a b) -> (well_founded R1 <-> well_founded R2).
  Proof.
    intros A R1 R2 hr.
    split; intros W a; specialize (W a);
      induction W as [ a W IH] using Acc_ind_dep.
    apply Acc_intro; intros b hb.
    apply IH.
    rewrite <- hr in hb.
    exact hb.
    apply Acc_intro; intros b hb.
    apply IH.
    rewrite hr in hb.
    exact hb.
  Qed.

  (** Any well-founded relation cannot be reflexive! *)
  Lemma acc_rel_is_non_reflexive:
    forall (A: Type) (R: relation A) (x: A), Acc R x -> ~R x x.
  Proof.
    intros A R x W.
    induction W using Acc_ind_dep.
    unfold not; intro false.
    specialize (H x false).
    auto.
  Qed.

  Lemma wf_rel_is_non_reflexive:
    forall (A: Type) (R: relation A),
      well_founded R -> forall x, ~R x x.
  Proof.
    intros A R W x.
    apply acc_rel_is_non_reflexive.
    apply W.
  Qed.

  (** Equivalence between two different definitions of the transitive
      of a relation.
   *)
  Lemma equiv_clos_trans_transitive_closure:
    forall (A: Type) (R: relation A) (x y: A),
      Relation_Operators.clos_trans R x y <-> transitive_closure R x y.
  Proof.
    intros A R x y.
    split; intros hr; induction hr as [ x y hr | x y z hr IH IH' IH''].
    apply (transitive_closure_step R x y hr).
    apply (transitive_closure_transitivity IH IH'').
    apply (Relation_Operators.t_step A R x y hr).
    apply (Relation_Operators.t_trans A R x y z IH IH'').
  Qed.

  Lemma equiv_wf_clos_trans_wf_transitive_closure:
    forall (A: Type) (R: relation A),
      well_founded (Relation_Operators.clos_trans R) <->
      well_founded (transitive_closure R).
  Proof.
    intros A R.
    apply wf_rel_equiv.
    apply equiv_clos_trans_transitive_closure.
  Qed.

  (** A well-founded relation has a well-founded transitive closure. *)
  Lemma wf_rel_has_wf_transitive_closure:
    forall (A: Type) (R: relation A),
      well_founded R -> well_founded (transitive_closure R).
  Proof.
    intros A R wf.
    rewrite <- equiv_wf_clos_trans_wf_transitive_closure.
    apply wf_clos_trans.
    exact wf.
  Qed.

  (** A well-founded has a non reflexive transitive closure. *)
  Lemma wf_rel_has_non_reflexive_transitive_closure:
    forall (A: Type) (R: relation A),
      well_founded R -> forall x, ~(transitive_closure R) x x.
  Proof.
    intros A R wf.
    apply wf_rel_is_non_reflexive with (R := (transitive_closure R)).
    apply wf_rel_has_wf_transitive_closure.
    exact wf.
  Qed.

  (** ** With Setoid *)
  (** Two equivalent relations are equivalently well-founded. *)
  Lemma well_founded_equiv_equiv {A: Type} `{SA: Setoid A}:
    forall (R1 R2: A -> A -> Prop),
      R1 =~= R2 -> (well_founded R1 <-> well_founded R2).
  Proof.
    intros R1 R2 H.
    apply wf_rel_equiv.
    intros a b; apply H; reflexivity.
  Qed.

  (** The flips (transpositions) of two equivalent relations are equivalent. *)
  Lemma flip_equiv_equiv {A: Type} `{SA: Setoid A}:
    forall (R1 R2: A -> A -> Prop),
      R1 =~= R2 -> (equiv ==> equiv ==> iff) (flip R1) (flip R2).
  Proof.
    intros R1 R2 H a a' Ha b b' Hb.
    unfold flip; apply H; auto.
  Qed.

  (** The flip of two equivalent relations are equivalently well-founded. *)
  Lemma well_founded_flip_equiv_equiv {A: Type} `{SA: Setoid A}:
    forall (R1 R2: A -> A -> Prop),
      R1 =~= R2 -> (well_founded (flip R1) <-> well_founded (flip R2)).
  Proof.
    intros R1 R2 H.
    apply well_founded_equiv_equiv.
    apply flip_equiv_equiv.
    apply H.
  Qed.

  (** A well-founded relation cannot be reflexive (using equality on elements
      and assuming compatibility of the elements. *)
  Lemma wf_rel_for_non_equiv_elements {A: Type} `{SA: Setoid A}:
    forall (R: relation A) (R_compat: Proper fequiv R),
      well_founded R -> forall x y, R x y -> ~x == y.
  Proof.
    intros R R_compat W x y Rxy e.
    rewrite e in Rxy.
    generalize Rxy.
    apply (wf_rel_is_non_reflexive W).
  Qed.

  (** For any compatible relation, flip (transposition) is also compatible. *)
  Lemma flip_proper_rel {A: Type} `{SA: Setoid A}:
    forall (R: relation A) (R_compat: Proper fequiv R),
      Proper fequiv (flip R).
  Proof.
    intros R R_compat a1 a2 ea b1 b2 eb; unfold flip.
    apply R_compat; auto.
  Qed.

  (** For any well-founded relation,
           for any point x and z,
                x is not z and there exists a path from x to z
                   is equivalent to
                x has a neighbor y and there exists a path from y to z
   *)
  Lemma well_founded_strict {A: Type} `{SA: Setoid A}:
    forall (R: relation A) (R_compat: Proper fequiv R),
      well_founded R ->
      forall x z,
        (exists y, R x y /\ reflexive_transitive_closure R y z)
        <->
        (~x == z /\ reflexive_transitive_closure R x z).
  Proof.
    intros R R_compat W x z.
    assert (tcR_compat: Proper fequiv (transitive_closure R)).
    { apply transitive_closure_compat.
      apply R_compat.
    }
    split; intro X.
    destruct X as (y, (Rxy, yz)).
    split.

    2:
      {
      unfold reflexive_transitive_closure in *.
      unfold reflexive_closure in *.
      right.
      destruct yz as [eqyz | yz].
      rewrite eqyz in Rxy.
      apply transitive_closure_step.
      exact Rxy.
      apply transitive_closure_transitivity with (y := y).
      apply transitive_closure_step; apply Rxy.
      exact yz.
    }
    (* 1 *)
    destruct yz as [eqyz | yz].
    rewrite eqyz in Rxy.
    apply (wf_rel_for_non_equiv_elements R_compat W x z Rxy).
    apply (wf_rel_for_non_equiv_elements) with (R := (transitive_closure R)).
    apply tcR_compat.
    apply wf_rel_has_wf_transitive_closure.
    apply W.
    apply transitive_closure_transitivity with (y := y).
    apply (transitive_closure_step R x y Rxy).
    exact yz.

    unfold reflexive_transitive_closure in *.
    unfold reflexive_closure in *.
    destruct X as (n, t).
    destruct t as [t|t].
    elim n; exact t.
    induction t as [ x y st | x y z tr IH tr' IH'].
    exists y; intuition.

    assert (nxy: ~x == y).
    apply wf_rel_for_non_equiv_elements with (R := (transitive_closure R)).
    apply tcR_compat.
    apply wf_rel_has_wf_transitive_closure.
    apply W.
    exact tr.
    specialize (IH nxy).

    assert (nyz: ~y == z).
    apply wf_rel_for_non_equiv_elements with (R := (transitive_closure R)).
    apply tcR_compat.
    apply wf_rel_has_wf_transitive_closure.
    apply W.
    exact tr'.
    specialize (IH' nyz).

    destruct IH as (y0, (xy0, [eqy0y | y0y]));
      destruct IH' as (z0, (yz0, [eqz0y |z0y])).

    rewrite eqy0y in xy0.
    rewrite eqz0y in yz0.
    exists y.
    split.
    exact xy0.
    right.
    apply transitive_closure_step.
    exact yz0.

    rewrite eqy0y in xy0.
    exists y.
    split.
    exact xy0.
    right.
    apply transitive_closure_transitivity with (y := z0).
    apply transitive_closure_step.
    exact yz0.
    exact z0y.

    rewrite eqz0y in yz0.
    exists y0.
    split.
    exact xy0.
    right.
    apply transitive_closure_transitivity with (y := y).
    exact y0y.
    exact tr'.

    exists y0.
    split.
    exact xy0.
    right.
    apply transitive_closure_transitivity with (y := z0).
    apply transitive_closure_transitivity with (y := y).
    exact y0y.
    apply transitive_closure_step; exact yz0.
    exact z0y.
  Qed.

  (** * Well_founded / Finite elements
   The goal of the section is to show that
   a relation on a finite set of elements is a well-founded order

   iff

   the transitive closure of the relation is not reflexive,
   i.e., there is no cycle in the relation.

   We deduce that in case of finite elements, a relation is well-founded
   iff its inverse is well-founded

   Lemma wf_rel_has_non_reflexive_transitive_closure <<(well_founded R -> forall x, ~(transitive_closure R) x x)>>
   provides one implication, with no requirement on the set of elements.

   We prove here the inverse implication, i.e. (see above)

   << Lemma  finite_acyclic_wf:                                            >>
   <<   forall (lA: list A) (all: forall a, InA eqA a lA)                  >>
   <<          (R: relation A) (R_dec: forall a1 a2, {R a1 a2}+{~R a1 a2}) >>
   <<          (R_compat: Proper (eqA ==> eqA ==> iff) R),                 >>
   <<     (forall x: A, ~ transitive_closure R x x) -> well_founded R.     >>

   **)
  Section Wf_flip_finite.

    Context {A: Type} `{SA: Setoid A}
            {A_dec: Decider (equiv (A:=A))}.

    (** ** TCR and properties
     The proof of the result uses a subrelation of the transitive
     closure, i.e., the transitive closure between elements of a given list
     only
     *)
    Inductive TCR (l: list A) (R: relation A): relation A :=
    | TCR_step: forall x y, R x y -> TCR l R x y
    | TCR_transitivity: forall x y z, InA equiv y l ->
                                      TCR l R x y -> TCR l R y z -> TCR l R x z.

    Lemma TCR_gen:
      forall (l: list A) (R: relation A) (x y: A),
        TCR l R x y -> transitive_closure R x y.
    Proof.
      intros l R x y h;
        induction h as [x y hxy| x y z hy hxy IHhxy hyz IHhyz ];
        [left|right with y];auto.
    Qed.

    Lemma TCR_all:
      forall (lA: list A) (all: forall a, InA equiv a lA)
             (R: relation A) (x y: A),
        transitive_closure R x y -> TCR lA R x y.
    Proof.
      intros lA all R x y h;
        induction h as [x y hxy| x y z hxy IHhxy hyz IHhyz ];
        [left|right with y];auto.
    Qed.

    Lemma TCR_all_equiv:
      forall (lA: list A) (all: forall a, InA equiv a lA)
             (R: relation A) (x y: A),
        transitive_closure R x y <-> TCR lA R x y.
    Proof.
      intros lA all R x y;split;intro h;
        [apply TCR_all;auto| apply TCR_gen with lA;auto].
    Qed.

    Instance TCR_compat:
      Proper (equivlistA equiv ==> fequiv ==> equiv ==> equiv ==> iff) TCR.
    Proof.
      intros l1 l2 el R1 R2 Rc x1 x2 eqx y1 y2 eqy.
      unfold equivlistA in el.
      split; intros Sx1y1.
      revert x2 y2 eqx eqy.
      induction Sx1y1 as [x y Rxy| x y z Iy Sxy Sx2y2 Syz Sy2z2];
        intros x2 y2 eqx eqy.
      left; rewrite <- (Rc _ _ eqx _ _ eqy); assumption.
      specialize (Sx2y2 x2 y eqx (reflexivity y)).
      specialize (Sy2z2 y y2 (reflexivity y) eqy).
      right with y; [rewrite <- el| |]; assumption.

      revert x1 y1 eqx eqy.
      induction Sx1y1 as [x y Rxy| x y z Iy Sxy Sx2y2 Syz Sy2z2];
        intros x2 y2 eqx eqy.
      left; rewrite (Rc _ _ eqx _ _ eqy); assumption.
      specialize (Sx2y2 x2 y eqx (reflexivity y)).
      specialize (Sy2z2 y y2 (reflexivity y) eqy).
      right with y; [rewrite el| |]; assumption.
    Qed.

    Lemma TCR_incl:
      forall (R: relation A) (l l': list A),
        inclA equiv l l' -> forall a1 a2, TCR l R a1 a2 -> TCR l' R a1 a2.
    Proof.
      intros R l l' inc a1 a2 H.
      induction H.
      left; assumption.
      right with y; [ apply inc | | ]; assumption.
    Qed.

    Inductive TCR_left (l: list A) (R: relation A): relation A :=
    | TCR_left_step: forall x y, R x y -> TCR_left l R x y
    | TCR_left_transitivity:
        forall x y z, InA equiv y l -> R x y ->
                      TCR_left l R y z -> TCR_left l R x z.

    Lemma TCR_left_TCR_equiv:
      forall (R: relation A) (l: list A) (x y: A),
        TCR l R x y <-> TCR_left l R x y.
    Proof.
      intros R l x y.
      split; intros H.

      induction H.
      left; assumption.
      clear H0 H1.
      induction IHTCR1.
      right with y; assumption.
      right with y.
      assumption.
      assumption.
      apply IHIHTCR1; assumption.

      induction H.
      left; assumption.
      right with y.
      assumption.
      left; assumption.
      assumption.
    Qed.

    Inductive TCR_right (l: list A) (R: relation A): relation A :=
    | TCR_right_step: forall x y, R x y -> TCR_right l R x y
    | TCR_right_transitivity:
        forall x y z, InA equiv y l -> R y z ->
                      TCR_right l R x y -> TCR_right l R x z.

    Lemma TCR_right_TCR_equiv:
      forall (R: relation A) (l: list A) (x y: A),
        TCR l R x y <-> TCR_right l R x y.
    Proof.
      intros R l x y.
      split; intros H.

      induction H.
      left; assumption.
      clear H0 H1.
      induction IHTCR2.
      right with x0; assumption.
      right with y.
      assumption.
      assumption.
      apply IHIHTCR2; assumption.

      induction H.
      left; assumption.
      right with y.
      assumption.
      assumption.
      left; assumption.
    Qed.

    (** TCR does not require that the first and last elements of the chain be
     in the list.
     Hence the relation is preserved when removing one the them from the list.
     *)
    Lemma TCR_remove_1:
      forall (R: relation A) (R_compat: Proper fequiv R) (l: list A) (a1 a2: A),
        TCR l R a1 a2 -> TCR (removeA A_dec a1 l) R a1 a2.
    Proof.
      intros R Rc l a1 a2 H.
      rewrite TCR_right_TCR_equiv in H.
      rewrite TCR_right_TCR_equiv.
      induction H.
      left; assumption.
      case (A_dec x y); intros.
      left;  rewrite (Rc _ _ e _ _ (Equivalence_Reflexive _) ); assumption.
      right with y.
      rewrite removeA_InA; split; auto; apply SA.
      assumption.
      assumption.
    Qed.

    Lemma TCR_remove_2:
      forall (R: relation A) (R_compat: Proper fequiv R) (l: list A) (a1 a2: A),
        TCR l R a1 a2 -> TCR (removeA A_dec a2 l) R a1 a2.
    Proof.
      intros R Rc l a1 a2 H.
      rewrite TCR_left_TCR_equiv in H.
      rewrite TCR_left_TCR_equiv.
      induction H.
      left; assumption.
      case (A_dec z y); intros.
      left; rewrite (Rc _ _ (Equivalence_Reflexive _) _ _ e ); assumption.
      right with y.
      rewrite removeA_InA; split; auto; apply SA.
      assumption.
      assumption.
    Qed.

    Lemma remove_incl:
      forall (l: list A) (a: A), inclA equiv (removeA A_dec a l) l.
    Proof.
      unfold inclA.
      intros l a x H.
      rewrite removeA_InA in H; [ apply H| apply SA].
    Qed.

    Lemma TCR_remove_1_iff:
      forall (R: relation A) (R_compat: Proper fequiv R) (l: list A) (a1 a2: A),
        TCR l R a1 a2 <-> TCR (removeA A_dec a1 l) R a1 a2.
    Proof.
      intros R Rc l a1 a2.
      split; [apply TCR_remove_1 | apply TCR_incl; apply remove_incl].
      assumption.
    Qed.

    Lemma TCR_remove_2_iff:
      forall (R: relation A) (R_compat: Proper fequiv R) (l: list A) (a1 a2: A),
        TCR l R a1 a2 <-> TCR (removeA A_dec a2 l) R a1 a2.
    Proof.
      intros R Rc l a1 a2.
      split; [apply TCR_remove_2 | apply TCR_incl; apply remove_incl].
      assumption.
    Qed.

    (** Checking if there exists an element with some property in a list
        is decidable, provided that the property itself is decidable.
     *)
    Lemma finite_dec:
      forall (l: list A) (P: A -> Prop) (P_compat: Proper fequiv P)
             (P_dec: forall a, { P a } + { ~ P a }),
        { (exists a, InA equiv a l /\ P a) } +
        { ~(exists a, InA equiv a l /\ P a) }.
    Proof.
      intros l P P_compat P_dec.
      induction l as [| a l Hl].
      right; intros [ a [na _]].
      rewrite InA_nil in na; elim na.
      case (P_dec a); intros Pa.
      left; exists a; split;
        [rewrite InA_cons; left; reflexivity|assumption].
      destruct Hl as [Hl|Hl]; [left|right].
      destruct Hl as [x [lx Px]].
      exists x; split; [rewrite InA_cons; right|]; assumption.
      intros F; contradict Hl.
      destruct F as [x [lx Px]].
      rewrite InA_cons in lx.
      destruct lx as [lx|lx].
      contradict Pa; apply (P_compat _ _ lx); assumption.
      exists x; split; assumption.
    Qed.

    Lemma finite_dec':
      forall (lA: list A) (all: forall a, InA equiv a lA)
             (P: A -> Prop) (P_compat: Proper fequiv P)
             (P_dec: forall a, { P a } + { ~ P a }),
        { (exists a, P a) } + { ~(exists a, P a) }.
      intros lA all P P_compat P_dec.
      destruct (finite_dec lA P_compat P_dec) as [X | X].
      left; destruct X as [a [_ Pa]]; exists a; apply Pa.
      right; intros [a Pa].
      contradict X.
      exists a; split; [apply all | apply Pa].
    Qed.

    (** length of a list when an element is removed *)
    Lemma removeA_length_leq:
      forall (l: list A) (a: A), (length (removeA A_dec a l)) <= length l.
    Proof.
      intros l a.
      induction l; simpl.
      firstorder.
      case (A_dec a a0); simpl; intros.
      firstorder.
      apply Peano.le_n_S;auto.
    Qed.

    Lemma removeA_length_le:
      forall (l: list A) (a: A),
        InA equiv a l -> (length (removeA A_dec a l)) < length l.
    Proof.
      intros l a nina.
      induction l; simpl.
      rewrite InA_nil in nina; elim nina.
      case (A_dec a a0); simpl; intros.
      + assert (tmp := removeA_length_leq l a).
        lia.
      + rewrite InA_cons in nina.
        destruct nina.
        - contradict n; assumption.
        - specialize (IHl H); lia.
    Qed.

    (** ** Decidability of transitive closure over finite sets

     If the set is finite, checking that a pair of elements are
     related with transitive closure is decidable (provided that the
     relation itself is decidable).

     The proof shows the result for TCR and for any list of any length
     as intermediate goal: this proof is done using induction on the
     lenght of the list.
     *)
    Lemma  finite_transitive_closure_dec:
      forall (lA: list A) (all: forall a, InA equiv a lA)
             (R: relation A) (R_dec: forall a1 a2, {R a1 a2}+{~R a1 a2})
             (R_compat: Proper fequiv R),
      forall (a1 a2: A), {transitive_closure R a1 a2} +
                         {~transitive_closure R a1 a2}.
    Proof.
      intros lA all R R_dec R_compat.
      assert (TCR_dec : forall n, forall l,
                   length l <= n ->
                   forall a1 a2,
                     {TCR l R a1 a2}+{~TCR l R a1 a2}).
      induction n as [|h hrecn].
      intros l hl a1 a2.
      destruct (R_dec a1 a2) as [OK|KO];
        [left;left;auto|right;intros h;
                        destruct h as [a1 a2 h|a1 a2 a3 ha2 _ _ ]].
      auto.
      destruct l.
      rewrite InA_nil in ha2.
      auto.
      simpl in hl;lia.
      intros l hl a1 a2.
      destruct (R_dec a1 a2) as [OK|nh12];[left;left;auto|].
      pose (TCR2 := exists a3, InA equiv a3 l /\ TCR l R a1 a3 /\ TCR l R a3 a2).
      assert (hh:{TCR2} + {~ TCR2}).
      unfold TCR2;clear TCR2.

      apply finite_dec' with lA.
      apply all.
      intros x1 x2 ex.
      apply and_compat.
      apply InA_compat; [apply SA| auto| reflexivity].
      apply and_compat.
      apply TCR_compat.
      reflexivity.
      apply R_compat.
      reflexivity.
      assumption.
      apply TCR_compat.
      reflexivity.
      apply R_compat.
      assumption.
      reflexivity.

      intros a.
      case (InA_dec A_dec a l); [intros ina | intros nina].
      2: right; intros [F _]; contradict nina; apply F.

      assert (lr: length (removeA A_dec a l) <= h).
      assert (lst := (removeA_length_le ina)).
      lia.
      generalize hrecn; intros H1; specialize (H1 (removeA A_dec a l) lr a1 a).
      generalize hrecn; intros H2; specialize (H2 (removeA A_dec a l) lr a a2).
      clear hrecn.
      destruct H1 as [H1|H1]; destruct H2 as [H2|H2].
      left; split; [apply ina|].
      split; [ rewrite TCR_remove_2_iff | rewrite  (TCR_remove_1_iff)];
        assumption.
      right; intros [_ [_ F]]; contradict H2; rewrite  TCR_remove_1_iff in F;
        assumption.
      right; intros [_ [F _]]; contradict H1; rewrite  TCR_remove_2_iff in F;
        assumption.
      right; intros [_ [F _]]; contradict H1; rewrite  TCR_remove_2_iff in F;
        assumption.

      (***************************************************)
      destruct hh as [hh|nhh];[left|right;intro hh;apply nhh].
      destruct hh as[a3 [ha3 [h13 h32]]];right with a3;auto.
      destruct hh as [a1 a2 hh |a1 a2 a3 ha2 h12 h23 ].
      elim (nh12 hh).
      split with a2;repeat split;assumption.

      intros a1 a2;destruct (TCR_dec (length lA) lA (le_refl _) a1 a2);
        [left|right]; rewrite (TCR_all_equiv all);assumption.
    Qed.

    (** ** Pointwise strict order on list*)
    Inductive PW_lt (X: Type) (ltX eqX: relation X): relation (list X) :=
    | PWcons1: forall a a' l l',
        ltX a a' -> PW_lt ltX eqX l l' -> PW_lt ltX eqX (a::l) (a'::l')
    | PWcons2: forall a a' l l',
        ltX a a' -> eqlistA eqX l l' -> PW_lt ltX eqX (a::l) (a'::l')
    | PWcons3: forall a a' l l',
        eqX a a' -> PW_lt ltX eqX l l' -> PW_lt ltX eqX (a::l) (a'::l').

    Lemma PW_length:
      forall (X: Type) (ltX eqX: relation X),
      forall  (l1 l2: list X), PW_lt ltX eqX l1 l2 -> length l1 = length l2.
    Proof.
      intros X ltX eqX l1 l2 h; induction  h;simpl;auto.
      apply f_equal; apply eqlistA_length with eqX. auto.
    Qed.

    Definition PW_lt_n (X: Type) (ltX eqX: relation X) (n: nat)
               (l1 l2: list X): Prop :=
      (length l1 = n /\ length l2 = n) /\ PW_lt ltX eqX l1 l2.

    Lemma PW_lt_n_length_l:
      forall (X: Type) (ltX eqX: relation X) (l1 l2: list X),
        PW_lt eqX ltX l1 l2 -> PW_lt_n eqX ltX (length l1) l1 l2.
    Proof.
      intros X ltX eqX l1 l2 h;repeat split.
      symmetry.
      apply (PW_length h).
      assumption.
    Qed.

    Lemma PW_lt_n_length_r:
      forall (X: Type) (ltX eqX: relation X) (l1 l2: list X),
        PW_lt eqX ltX l1 l2 -> PW_lt_n eqX ltX (length l2) l1 l2.
    Proof.
      intros X ltX eqX l1 l2 h;repeat split.
      apply (PW_length h).
      assumption.
    Qed.

    Lemma Acc_PW_lt_n:
      forall (X: Type) (ltX eqX: relation X) (n: nat) (lx: list X),
        length lx = n -> Acc (PW_lt_n ltX eqX n) lx -> Acc (PW_lt ltX eqX) lx.
    Proof.
      intros X ltX eqX n lx hx  w; induction w as [lx_ w hrecw];split;
        intros ly hly.
      apply hrecw.
      subst.
      auto using PW_lt_n_length_r.
      subst.
      eauto using PW_length.
    Qed.

    Lemma PW_lt_n_wf:
      forall (X: Type) (ltX eqX: relation X) (eqX_equiv: Equivalence eqX)
             (px: Proper (eqX ==> eqX ==> iff) ltX) (wfX: well_founded ltX),
      forall (n: nat), well_founded (PW_lt_n ltX eqX n).
    Proof.
      intros X ltX eqX eqX_equiv px wfX n;induction n as [|n hrecn].
      intro l.
      split;intros l' [[h1 h2] h].
      inversion h;subst;simpl in h1;congruence.
      intro l.
      destruct l as [|x l].
      split;intros l' [[h1 h2] h].
      inversion h.
      revert l.
      induction (wfX x).
      intro l.
      clear H;revert x H0.
      induction (hrecn l) as [l _ hrecl].
      intros x hrecx.
      split;intros l' [[h1 h2] h];inversion h;subst.
      auto.
      auto.
      apply hrecl;auto.
      simpl in h1;injection h1;clear h1;intro h1;subst.
      apply PW_lt_n_length_l;assumption.
      intros y hy;apply hrecx.
      rewrite H2 in hy.
      assumption.
    Qed.

    (** The point strict order on list is well-founded when the strict
     order on elements is well-founded. *)
    Lemma PW_lt_wf:
      forall (X: Type) (ltX eqX: relation X) (eqX_equiv: Equivalence eqX)
             (px: Proper (eqX ==> eqX ==> iff) ltX) (wfX: well_founded ltX),
        well_founded (PW_lt ltX eqX).
    Proof.
      intros X ltX eqX eqX_equiv px wfX lX.
      apply Acc_PW_lt_n with (length lX).
      split.
      apply PW_lt_n_wf;auto.
    Qed.

    (** ** A relation with non-reflexive transitive closure over finite sets
    is well-founded.

    For the proof, we build for any element a list of nat as follows:

       <<R_trans_dec := (finite_transitive_closure_dec all R_dec R_compat)>>
       <<left_cone_charac := fun a0 a : A =>                             >>
       <<                    if R_trans_dec a a0 then 1%nat else 0%nat   >>
       <<measure := fun a0 : A => map (left_cone_charac a0) lA           >>

    And we show that for any a1, a2 such that R a1 a2 we have:

       <<PW_lt lt eq (measure a1) (measure a2)>>

    Inclusion of well-founded order and Lemma PW_lt_wf lead to
    the conclusion.
     *)
    Lemma  finite_acyclic_wf:
      forall (lA: list A) (all: forall a, InA equiv a lA)
             (R: relation A) (R_compat: Proper fequiv R)
             (R_dec: forall a1 a2, {R a1 a2} + {~R a1 a2}),
        (forall x: A, ~ transitive_closure R x x) -> well_founded R.
    Proof.
      intros lA all R R_compat R_dec R_acyclic.
      assert (R_trans_dec := (finite_transitive_closure_dec all R_dec R_compat)).

      pose (left_cone_charac (a0:A) := (fun a => if (R_trans_dec a a0)
                                                 then 1%nat else 0%nat)).
      pose (measure (a0:A) := List.map (left_cone_charac a0) lA).
      apply Inclusion.wf_incl with  ((PW_lt lt eq) @@ measure)%signature.

      2: apply (Inverse_Image.wf_inverse_image _ _ (PW_lt lt eq) measure).
      2: apply PW_lt_wf.
      2: apply eq_equivalence.
      Import Lia.
      2: intros x1 x2 hx y1 y2 hy;lia.
      2: apply lt_wf.

      intros a1 a2 h12.
      unfold "@@".
      assert (hle:forall a, left_cone_charac a1 a <= left_cone_charac a2 a).
      intro a; unfold left_cone_charac.
      destruct (R_trans_dec a a1) as [h1|h1];
        destruct (R_trans_dec a a2) as [h2|h2];auto.
      contradict h2;right with a1;[assumption|left;assumption].
      assert (hlt: left_cone_charac a1 a1 < left_cone_charac a2 a1).
      unfold left_cone_charac; destruct (R_trans_dec a1 a1) as [h1|h1].
      elim (R_acyclic _ h1).
      destruct (R_trans_dec a1 a2) as [h2|h2];auto.
      contradict h2;left;assumption.
      assert (Plcc: Proper (equiv ==> equiv ==> eq) left_cone_charac).
      intros aa1 aa2 ha ab1 ab2 hb.
      unfold left_cone_charac.
      assert (hh:=transitive_closure_compat R_compat _ _ hb _ _ ha).
      destruct (R_trans_dec ab1 aa1) as [test1|test1];
        rewrite (transitive_closure_compat R_compat _ _ hb _ _ ha) in test1;
        destruct (R_trans_dec ab2 aa2) as [test2|test2];auto||tauto.

      generalize (all a1);
        apply proj1 with (PW_lt lt eq (measure a1) (measure a2) \/
                          eqlistA eq (measure a1) (measure a2)).
      unfold measure; clear measure all;induction lA as [|a lA [hrec1 hrec2]].
      split.
      rewrite InA_nil;intros [].
      right;left.
      split.
      rewrite InA_cons;intros [ha|hlA].
      destruct hrec2 as [hrec2|hrec2].
      constructor 1.
      rewrite <- ha;assumption.
      apply hrec2.
      constructor 2.
      rewrite <- ha;assumption.
      apply hrec2.
      destruct (le_lt_eq_dec (hle a)) as [LT | EQ].
      destruct hrec2 as [hrec2|hrec2].
      constructor 1;auto.
      constructor 2;auto.
      constructor 3;auto.
      destruct (le_lt_eq_dec (hle a)) as [LT | EQ].
      left.
      destruct hrec2 as [hrec2|hrec2].
      constructor 1;auto.
      constructor 2;auto.
      destruct hrec2 as [hrec2|hrec2].
      left;constructor 3;auto.
      right;right;auto.
    Qed.

    (** ** Consequences on inverse relation *)
    Lemma transitive_closure_flip:
      forall (R: relation A) (x y: A),
        transitive_closure R x y <-> transitive_closure (flip R) y x.
    Proof.
      intros R x y.
      rewrite <- transitive_closure_left_iff_transitive_closure.
      rewrite <- transitive_closure_right_iff_transitive_closure.
      split; intros H.
      induction H.
      apply transitive_closure_right_step; unfold flip; assumption.
      apply transitive_closure_right_transitivity with y;
        [| unfold flip]; assumption.
      induction H.
      apply transitive_closure_left_step; unfold flip; assumption.
      apply transitive_closure_left_transitivity with y;
        [| unfold flip]; assumption.
    Qed.

    Lemma tc_equiv:
      forall (R1: relation A) (R2: relation A) (equiv: R1 =~= R2),
      forall (x y: A),
        transitive_closure R1 x y <-> transitive_closure R2 x y.
    Proof.
      intros.
      rewrite <- transitive_closure_left_iff_transitive_closure.
      rewrite <- transitive_closure_left_iff_transitive_closure.
      split; intros H.
      induction H.
      apply transitive_closure_left_step.
      rewrite <- (equiv _ _ (Equivalence_Reflexive x) _ _
                        (Equivalence_Reflexive y)); assumption.
      apply transitive_closure_left_transitivity with y.
      rewrite <- (equiv _ _ (Equivalence_Reflexive x) _ _
                        (Equivalence_Reflexive y)); assumption.
      assumption.
      induction H.
      apply transitive_closure_left_step.
      rewrite (equiv _ _ (Equivalence_Reflexive x) _ _
                     (Equivalence_Reflexive y)); assumption.
      apply transitive_closure_left_transitivity with y.
      rewrite (equiv _ _ (Equivalence_Reflexive x) _ _
                     (Equivalence_Reflexive y)); assumption.
      assumption.
    Qed.

    Lemma flip_flip:
      forall (R: relation A) (R_compat: Proper fequiv R),
        (flip (flip R)) =~= R.
    Proof.
      intros R Rc x1 x2 ex y1 y2 ey.
      unfold flip.
      apply Rc; assumption.
    Qed.

    Lemma  finite_acyclic_wf_flip:
      forall (lA: list A) (all: forall a, InA equiv a lA)
             (R: relation A) (R_compat: Proper fequiv R)
             (R_dec: forall a1 a2, {R a1 a2}+{~R a1 a2}),
        (forall x, ~ transitive_closure R x x) -> well_founded (flip R).
    Proof.
      intros lA all R R_compat R_dec R_acyclic.
      apply  finite_acyclic_wf with lA.
      apply all.
      apply flip_proper; assumption.
      intros a1 a2;apply R_dec.
      intros x; rewrite transitive_closure_flip.
      rewrite tc_equiv.
      2: apply flip_flip.
      apply R_acyclic.
      assumption.
    Qed.

    Lemma wf_flip_finite:
      forall (lA: list A) (all: forall a, InA equiv a lA)
             (R: relation A) (R_compat: Proper fequiv R)
             (R_dec: forall a1 a2, {R a1 a2}+{~R a1 a2}),
        well_founded R <-> well_founded (flip R).
    Proof.
      intros.
      split; intros H;
        apply  finite_acyclic_wf with lA;try assumption.
      apply flip_proper; assumption.
      intros a1 a2;unfold flip;auto.
      intros x; rewrite <- transitive_closure_flip.
      apply wf_rel_has_non_reflexive_transitive_closure; assumption.
      intros x; rewrite transitive_closure_flip.
      apply wf_rel_has_non_reflexive_transitive_closure.
      assumption.
    Qed.

  End Wf_flip_finite.

End Well_Founded.

Lemma not_not_dec:
  forall (A: Type) (R: relation A)
         (R_dec: forall a b: A, { R a b } + { ~R a b }),
  forall a b, R a b <-> ~ ~ R a b.
Proof.
  intros A R R_dec a b; split; intros H.
  intros F; contradict F; assumption.
  destruct (R_dec a b).
  assumption.
  unfold not in H, n.
  specialize (H n).
  elim H.
Qed.

Lemma not_exists_forall:
  forall (A: Type) (R: relation A),
  forall (y: A),  ~(exists x, R x y) <-> (forall x, ~R x y).
Proof.
  intros y; split; intros H.
  intros x F.
  contradict H.
  exists x; assumption.

  intros [x Rxy].
  specialize (H x).
  tauto.
Qed.

Lemma existsb_sig {A: Type} `{SA: Setoid A}:
  forall (f: A -> bool) (f_compat: Proper fequiv f) (l: list A),
    existsb f l = true -> {x: A | InA equiv x l /\ f x = true}.
Proof.
  intros f f_compat l; induction l as [|a l Hl];simpl.
  congruence.
  intro h.
  case_eq (f a); intro e.
  exists a.
  rewrite InA_cons.
  intuition.
  rewrite e in h;clear e.
  simpl in h.
  specialize (Hl h).
  destruct Hl as [x [hx e]].
  exists x.
  rewrite InA_cons;intuition.
Qed.


Section Wf_finite.

  Lemma exists_forall_finite_dec {A: Type} `{SA: Setoid A}:
    forall (lA: list A) (all: forall a, InA equiv a lA),
    forall (R: relation A) (R_dec: forall a b, { R a b } + { ~R a b })
           (R_compat: Proper fequiv R),
    forall y, { x | R x y } + { forall x, ~R x y }.
  Proof.
    intros lA all R R_dec R_compat y.
    destruct (Bool.bool_dec (forallb (fun x => if (R_dec x y)
                                               then false else true) lA) true).
    rewrite forallb_forall in e.
    right.
    intros x F.
    specialize (all x).
    rewrite InA_alt in all.
    destruct all as [xx [eqxx inxx]].
    specialize (e _ inxx).
    eapply R_compat in F.
    2: symmetry; apply eqxx.
    2: reflexivity.
    destruct (R_dec xx y); congruence.

    left.
    rewrite forallb_existsb with
        (notf := (fun x : A => if R_dec x y then true else false)) in n.
    2: intros a; destruct (R_dec a y); symmetry;
      [rewrite negb_false_iff | rewrite negb_true_iff]; reflexivity.
    specialize (Bool.eq_true_negb_classical _ n); intros H; clear n.
    pose (f x:= if R_dec x y then true else false).
    assert (pf:Proper fequiv f).
    {
      unfold f;intros x x' hx; destruct (R_dec x y), (R_dec x' y);
        reflexivity||(rewrite hx in *; tauto).
    }
    edestruct (existsb_sig pf lA H) as [a [ina ha]].
    exists a.
    unfold f in ha.
    destruct (R_dec a y) as [E | NE];auto.
    congruence.
  Qed.

  Lemma wf_finite_path {A: Type} `{SA: Setoid A}:
    forall (lA: list A) (all: forall a, InA equiv a lA),
    forall (R: relation A)
           (R_dec: forall a b, { R a b } + { ~R a b })
           (R_compat: Proper fequiv R),
      well_founded R ->
      forall x, exists (p: list A) y, chain R y p x /\ forall x, ~R x y.
  Proof.
    intros lA all R R_dec R_compat wf x.
    specialize (wf x).
    induction wf.
    destruct (exists_forall_finite_dec all R_dec R_compat x) as [ [y S] | S].
    2: exists nil, x; simpl; split; auto; reflexivity.
    destruct (H0 _ S) as [p [ yy [C N]]].
    exists (p ++ (x::nil)), yy.
    rewrite chain_app; simpl; auto.
    split; [|apply N].
    exists y.
    split; [assumption | split; [assumption |reflexivity]].
  Qed.

End Wf_finite.

(** Lexicographic ordering on pairs
    Well founded property *)
Section Lexico_pairs.

  Context {A B: Type}.
  Variable (eqA ltA: relation A) (ltB: relation B).

  Inductive lexico_pairs: relation (A * B) :=
  | lpA: forall a a' b b', ltA a a' -> lexico_pairs (a, b) (a', b')
  | lpB: forall a a' b b', eqA a a' -> ltB b b' -> lexico_pairs (a, b) (a', b').

  Lemma lexico_pairs_inv:
    forall a a' b b', lexico_pairs (a, b) (a', b') ->
                      ltA a a' \/ eqA a a' /\ ltB b b'.
  Proof. intros a a' b b' h; inversion h; intuition. Qed.

  Lemma lexico_pairs_intro:
    forall a a' b b', ltA a a' \/ eqA a a' /\ ltB b b' ->
                      lexico_pairs (a, b) (a', b').
  Proof. intros a a' b b' [h|[h h']]; [left|right]; auto. Qed.

  Lemma lexico_pairs_Acc:
    Transitive eqA ->
    inclusion (RelUtil.compose ltA eqA) ltA ->
    well_founded ltB -> forall a, Acc ltA a ->
                                  forall b, Acc lexico_pairs (a, b).
  Proof.
    intros trans incl wfB a accA.
    induction accA as [a ha iha].
    intros b; specialize (wfB b); induction wfB as [b hb ihb].
    apply Acc_intro.
    intros [a' b'] h'.
    apply lexico_pairs_inv in h'; destruct h' as [h' | [h' h'']].
    - apply iha; auto.
    - cut (Acc lexico_pairs (a, b')).
      { intros H.
        inversion H.
        apply Acc_intro.
        intros [a'' b''] hh.
        apply H0.
        apply lexico_pairs_inv in hh; destruct hh as [hh | hh].
        - left; apply incl; unfold RelUtil.compose.
          exists a'; split; auto.
        - right.
          + apply (trans a'' a'); intuition.
          + intuition.
      }
      apply ihb; auto.
  Qed.

  Lemma lexico_pairs_wf:
    Transitive eqA ->
    inclusion (RelUtil.compose ltA eqA) ltA ->
    well_founded ltA -> well_founded ltB -> well_founded lexico_pairs.
  Proof.
    intros trans incl wfA wfB [a b].
    apply lexico_pairs_Acc; auto.
  Qed.

End Lexico_pairs.

Close Scope signature_scope.
Unset Implicit Arguments.
