(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
*)
(* begin show *)

(** **** Global imports *)
From Coq Require Import Relations.
From Coq Require Import SetoidClass.
From Coq Require Import ZArith.
From Coq Require Import Lia.

(** **** Local imports **)
From Padec Require Import SetoidUtil.

(* end show *)
Open Scope Z_scope.
Open Scope signature_scope.
Set Implicit Arguments.

Lemma le_plus:
  forall x y, x <= y -> exists i, y = x + Z.of_nat i.
Proof.
  intros x y le'.
  assert (h: 0 <= y - x).
  lia.
  apply Z_of_nat_complete in h.
  destruct h as [i p].
  exists i.
  lia.
Qed.

Lemma gt_plus:
  forall x y, x > y -> exists i, y = x - 1 - Z.of_nat i.
Proof.
  intros x y le'.
  assert (h: 0 <= x - y - 1).
  lia.
  apply Z_of_nat_complete in h.
  destruct h as [i p].
  exists i.
  lia.
Qed.

Lemma Z_zero_place: forall a b: Z, 0 <= a * b < a -> b = 0.
Proof.
  cut (forall a b: nat, (a * b < a -> b = 0)%nat).
  { intros ok.
    intros a b h.
    assert (x: 0 < a).
    { lia. }
    assert (xx: 0 <= b).
    { case (Z_dec b 0); [|intros f; lia].
      intros [f | f]; [|lia].
      assert (xxx := Z.mul_pos_neg _ _ x f).
      lia.
    }
    apply Z2Nat.inj; try lia.
    apply ok with (Z.to_nat a).
    rewrite <- Z2Nat.inj_mul; try lia.
  }
  { intros a [| b]; auto.
    intros h.
    rewrite Nat.mul_succ_r in h.
    assert (t: forall x: nat, (x + a >= a)%nat).
    intros x; lia.
    specialize (t (a * b)%nat).
    lia.
  }
Qed.

Lemma Z_no_place: forall a b: Z, ~ 0 < a * b < a.
Proof.
  intros a b F.
  assert (x: 0 <= a * b < a).
  lia.
  rewrite (Z_zero_place _ x) in F.
  rewrite Z.mul_0_r in F; lia.
Qed.

(** * Some tools about Z/kZ *)
Section Tools_mod_k.

  Variable (k: Z) (k_pos: k > 0).

  Definition eqk: relation Z := fun x y => exists q, x = y + q*k.
  Definition eqk_alt: relation Z :=
    fun x y => Z.modulo x k = Z.modulo y k.
  Lemma eqk_def_equiv: (eq ==> eq ==> iff) eqk eqk_alt.
  Proof.
    intros x x' ex y y' ey.
    rewrite <- ex, <- ey; clear x' y' ey ex.
    split; intros h.
    { destruct h as [q hq]; rewrite hq; clear hq x.
      unfold eqk_alt.
      apply Z_mod_plus_full.
    }
    { unfold eqk_alt, Z.modulo in h.
      assert (Y := Z_div_mod y k k_pos).
      remember (Z.div_eucl y k) as ykd.
      destruct ykd as (yk,ykm); clear Heqykd.
      assert (X := Z_div_mod x k k_pos).
      remember (Z.div_eucl x k) as xkd.
      destruct xkd as (xk,xkm); clear Heqxkd.
      destruct Y as [Y hY].
      destruct X as [X hX].
      exists (xk - yk).
      rewrite Z.mul_comm, Z.mul_sub_distr_l; lia.
    }
  Qed.

  Lemma eqk_def_equiv': forall x y, eqk x y <-> eqk_alt x y.
  Proof. intros x y; apply eqk_def_equiv; reflexivity. Qed.

  Global Instance eqk_equiv: Equivalence eqk.
  Proof.
    split.
    - intros x; rewrite eqk_def_equiv'; unfold eqk_alt; auto.
    - intros x y h; rewrite eqk_def_equiv' in *;
        unfold eqk_alt in *; auto.
    - intros x y z h h'; rewrite eqk_def_equiv' in *;
        unfold eqk_alt in *; rewrite h; auto.
  Qed.

  Global Instance eqk_Setoid: Setoid Z :=
    @Build_Setoid _ _ eqk_equiv.

  Lemma eqk_dec: forall x y, {eqk x y} + {~eqk x y}.
  Proof.
    intros x y.
    destruct (Z.eq_dec (x mod k) (y mod k));
      [left | right];
      rewrite eqk_def_equiv'; unfold eqk_alt; auto.
  Qed.

  Definition gtk: relation Z := (fun x y => x mod k > y mod k).

  Lemma gtk_dec: forall x y, {gtk x y} + {~gtk x y}.
  Proof.
    intros x y; unfold gtk.
    destruct (Z_gt_dec (x mod k) (y mod k));
      [left | right]; auto.
  Qed.

  Global Instance gtk_compat: Proper (eqk ==> eqk ==> iff) gtk.
  Proof.
    intros x x' ex y y' ey; unfold gtk in *;
      rewrite eqk_def_equiv' in *;
      unfold eqk_alt in *; rewrite ex, ey; tauto.
  Qed.

  Lemma gt_diff: forall x y, k - x mod k + y mod k > 0.
  Proof.
    intros x y.
    assert (t := Z_mod_lt x _ k_pos).
    assert (tt := Z_mod_lt y _ k_pos).
    lia.
  Qed.

  Lemma mod_nil: forall a, 0 <= a < k -> a mod k = a.
  Proof.
    intros a h.
    symmetry; apply Zmod_unique with 0; auto.
    rewrite Z.mul_0_r; lia.
  Qed.

  Lemma mod_add_dist:
    forall a b, a mod k + b mod k < k ->
                (a + b) mod k = a mod k + b mod k.
  Proof.
    intros a b h.
    assert (hh: 0 <= a mod k + b mod k < k).
    assert (dum := Z_mod_lt a _ k_pos).
    assert (dum' := Z_mod_lt b _ k_pos).
    lia.
    rewrite Z.add_mod; [|lia].
    rewrite (mod_nil hh); auto.
  Qed.

  Lemma mod1: forall x, x mod 1 = 0.
  Proof.
    intros a.
    assert (dum: 1 > 0).
    lia.
    assert (t := Z_mod_lt a _ dum).
    lia.
  Qed.

  Lemma mod1_eq: k = 1 -> forall x y, eqk x y.
  Proof.
    intros h1 x y; rewrite eqk_def_equiv'; unfold eqk_alt.
    rewrite h1, (mod1 x), (mod1 y); auto.
  Qed.

  Lemma mod1_gt: k = 1 -> forall x y, ~gtk x y.
  Proof.
    intros h1 x y.
    assert (t := mod1_eq h1 x y).
    rewrite eqk_def_equiv' in t; unfold eqk_alt in t.
    unfold gtk; lia.
  Qed.

  Lemma mod_plus_1: forall y, y mod k <> k - 1 ->
                              (y + 1) mod k =  y mod k + 1.
  Proof.
    intros y h.
    assert (Y := Z_div_mod y k k_pos).
    unfold Z.modulo in *; remember (Z.div_eucl y k) as ykd.
    destruct ykd as (yk, ykm); clear Heqykd.
    symmetry; apply Zmod_unique with yk; lia.
  Qed.

  Lemma mod_plus_1': forall y, y mod k = k - 1 ->
                               (y + 1) mod k =  0.
  Proof.
    intros y h.
    assert (Y := Z_div_mod y k k_pos).
    unfold Z.modulo in *; remember (Z.div_eucl y k) as ykd.
    destruct ykd as (yk,ykm); clear Heqykd.
    rewrite h in Y.
    symmetry; eapply Zmod_unique with (yk + 1); [lia|].
    rewrite Z.mul_add_distr_l; lia.
  Qed.

  Lemma mod_minus_1: forall y, y mod k <> 0 ->
                               (y - 1) mod k =  y mod k - 1.
  Proof.
    intros y h.
    assert (Y := Z_div_mod y k k_pos).
    unfold Z.modulo in *; remember (Z.div_eucl y k) as ykd.
    destruct ykd as (yk,ykm); clear Heqykd.
    symmetry; apply Zmod_unique with yk; try lia.
  Qed.

  Lemma mod_minus_1': forall y, y mod k = 0 ->
                                (y - 1) mod k =  k - 1.
  Proof.
    intros y h.
    assert (Y := Z_div_mod y k k_pos).
    unfold Z.modulo in *; remember (Z.div_eucl y k) as ykd.
    destruct ykd as (yk,ykm); clear Heqykd.
    rewrite h in Y.
    symmetry; eapply Zmod_unique with (yk - 1); [lia|].
    rewrite Z.mul_sub_distr_l; lia.
  Qed.

End Tools_mod_k.

(** * Boolean Test *)
Section Boolean_Test.

  Context {A: Type} `{SA: Setoid A}.
  Variable (eqA_dec: Decider (equiv (A:=A))).

  Definition dec2b x y := if eqA_dec x y then true else false.

  Global Instance dec2b_compat: Proper fequiv dec2b.
  Proof.
    intros x1 x2 eqx y1 y2 eqy.
    unfold dec2b; destruct (eqA_dec x1 y1);
    destruct (eqA_dec x2 y2); tauto || auto;
    rewrite eqx in *; rewrite eqy in *; simpl; tauto.
  Qed.

  Lemma dec2b_spec:
    forall x1 x2 y1 y2, dec2b x1 y1 = dec2b x2 y2 <->
                        (x1 == y1 <-> x2 == y2).
  Proof.
    intros x1 x2 y1 y2.
    unfold dec2b; destruct (eqA_dec x1 y1); destruct (eqA_dec x2 y2);
      split; tauto || auto;
    intros f; discriminate f.
  Qed.

  Lemma dec2b_spec_true: forall x y,
      dec2b x y = true <-> x == y.
  Proof.
    intros x y; unfold dec2b.
    destruct (eqA_dec x y);
      split; auto.
    intros f; discriminate f.
  Qed.

  Lemma dec2b_spec_false: forall x y,
      dec2b x y = false <-> ~(x == y).
  Proof.
    intros x y; unfold dec2b.
    destruct (eqA_dec x y); split; tauto || auto.
    intros f; discriminate f.
  Qed.

End Boolean_Test.

Close Scope Z_scope.
Close Scope signature_scope.
Unset Implicit Arguments.
